/**************************************************************************
 ** INTEL CONFIDENTIAL
 ** Copyright 2011 - 2014 Intel Corporation All Rights Reserved.
 **
 ** The source code contained or described herein and all documents related to
 ** the source code ("Material") are owned by Intel Corporation or its suppliers or
 ** licensors. Title to the Material remains with Intel Corporation or its suppliers
 ** and licensors. The Material contains trade secrets and proprietary and confidential
 ** information of Intel or its suppliers and licensors. The Material is protected by
 ** worldwide copyright and trade secret laws and treaty provisions. No part of the
 ** Material may be used, copied, reproduced, modified, published, uploaded, posted,
 ** transmitted, distributed, or disclosed in any way without Intel's prior express written
 ** permission.
 **
 ** No license under any patent, copyright, trade secret or other intellectual property
 ** right is granted to or conferred upon you by disclosure or delivery of the Materials,
 ** either expressly, by implication, inducement, estoppel or otherwise. Any license under
 ** such intellectual property rights must be express and approved by Intel in writing.
 *************************************************************************/

/**
 *
 * Event Machine Chaining example.
 *  This sample demonstrates event chaining of events. It includes receiving events from
 *  remote SoC and sending events to remote SoC. It also includes reassembly of received
 *  fragments in receive path and fragmentation of events in transmit Path.
 *
 *  Creates One Execution Object (EO) (em_eo_create(), em_eo_start() )
 *  Create 1 EM queue with em_queue_create_static() - Use some fixed queue value.
 *  Add queues to EO and enable them. ( Use API em_queue_enable_all())
 *  Start the traffic from trafgen. This is pre-configured event chaining traffic
 *  as if coming from remote SoC. This traffic is received at receive function.
 *  In receive function, print the packet and send the packet to IP module.
 * 
 *  This sample also demonstrates basic EM <--> ADK EM communication.
 *  This creates 2 queues (static) attached to EO.
 *  ADK application sends events to these and on receiving this
 *  application sends event to remote SOC.
 *  EM application creates queues for EM->ADK in the EO upon calling em_init with nonem_cfg info
 *  filled in _pre_init() method.
 *  
 *  _pre_init() section does adk_dp_setup and fill up nonem_cfg struct. ADK specific command
 *  line arguments are passed through SNF_EM_ADK_config.xml which is used for adk_dp_setup. 
 *
 *  Command line arguments: ./eventMachine -c 0xff -d librte_bus_vmbus.so  -- -t -a test_appl_adk_ec_perf_multi_flow
 *			      -f ../programs/common/config/SNF_EM_Config.xml
 *			      -f $ADK_SRC/ncp_adk/test/adk_dp_app/SNF_EM_ADK_Config.xml
 */
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wsign-conversion"

#include <libgen.h>

#include "adk_dp_app_config.h"
#include "adk_sys.h"
#include "adk_ph_dev.h"
#include "adk_pal.h"
#include "adk_modules_enabled.h"
#include "adk_ethernet.h"
#include "adk_ip.h"
#include "adk_ipsec.h"

/*
 * Test configuration
 */

/** The number of events to allocate and send */
#define NUM_NONFRAG_EVENTS 100
#define NUM_FRAG_EVENTS 0	//10
#define APP_TIME_OUT 10

#define STATIC_QUEUE_ID_SOC(N) ( ( (N) << 16 | (EM_QUEUE_STATIC_MIN << 8)) + 1)	/* Tx Q */
#define FRAG_SIZE 3000
#define NONFRAG_SIZE 128

#define PRINT_PKT 0
#define PRINT_MSG 0

/*
 * Test configuration
 */
#define IP_HDR_LEN 20
#define UDP_HDR_LEN 8
#define DATA_LEN 18

#define VIRTUAL_CORE_ID_FOR_IPV4_ENTRY 25
#define VIRTUAL_CORE_ID_FOR_IPV6_ENTRY 26
#define NUM_QIDS_PER_WORKER 4
/* ADK Entry Point
 * 0 - Unused  TODO: Need to check
 * 1 - IPv4
 * 2 - IPv6
 */
#define NUM_ADK_ENTRY_POINT 4

#define EM_RECV_QUEUE_ID0_IP   	3276801 //0x320001
#define EM_RECV_QUEUE_ID1_IP   	3276802 //0x320001
#define EM_RECV_QUEUE_ID2_IP   	3276803 //0x320001
#define EM_RECV_QUEUE_ID3_IP   	3276804 //0x320001
/*#define EM_RECV_QUEUE_ID4_IP   	3276805 //0x320001
#define EM_RECV_QUEUE_ID5_IP   	3276805 //0x320001*/
#define EM_RECV_QUEUE_ID_IPV6 	3277057 //0x320101

#define NONEM_QUEUE_IPV6  	(long unsigned int)0xa00
#define BATCH_SIZE		32
em_queue_t g_nonem_queue_ip[20];
em_queue_type_t	emQueueType = EM_QUEUE_TYPE_PARALLEL_ORDERED;
em_queue_t g_ul_queue[20];
#if PRINT_MSG
static void printPacket(unsigned char *pkt, int pktLen);
#endif
typedef struct adk_app_msg_t
{
        unsigned char ipHdr[IP_HDR_LEN];
        unsigned char udpHdr[UDP_HDR_LEN];
}__attribute__((packed))adk_app_msg_t;


ENV_LOCAL em_event_t batch_buf[BATCH_SIZE];
ENV_LOCAL uint32_t batch_count;
/* Handcrafted UDP/IP packet. Size 60 bytes */

/*static adk_app_msg_t ipPacket = {
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x64, 0x64, 0x64, 0x01, 0xc8, 0xc8, 0xc8, 0x01},
    .udpHdr = {0x00, 0x3F, 0x00, 0x3F, 0x02, 0x40, 0xA2, 0xE8}};*/

static adk_app_msg_t ipPacket_flow[] = {{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x09, 0x01, 0x01, 0x01, 0x31, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0A, 0x01, 0x01, 0x01, 0x32, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0B, 0x01, 0x01, 0x01, 0x33, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0C, 0x01, 0x01, 0x01, 0x34, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0D, 0x01, 0x01, 0x01, 0x35, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0E, 0x01, 0x01, 0x01, 0x36, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0F, 0x01, 0x01, 0x01, 0x37, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x10, 0x01, 0x01, 0x01, 0x38, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x11, 0x01, 0x01, 0x01, 0x39, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x12, 0x01, 0x01, 0x01, 0x3A, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x13, 0x01, 0x01, 0x01, 0x3B, 0x01, 0x01, 0x64},
    .udpHdr = {0x00, 0x64, 0x08, 0x68, 0x01, 0xD6, 0xC1, 0x35}},
	/*Plain text*/{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0A, 0x01, 0x01, 0x01, 0x32, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0B, 0x01, 0x01, 0x01, 0x33, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0C, 0x01, 0x01, 0x01, 0x34, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0D, 0x01, 0x01, 0x01, 0x35, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0E, 0x01, 0x01, 0x01, 0x36, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x0F, 0x01, 0x01, 0x01, 0x37, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x10, 0x01, 0x01, 0x01, 0x38, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x11, 0x01, 0x01, 0x01, 0x39, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x12, 0x01, 0x01, 0x01, 0x3A, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}},{
    .ipHdr = {0x45, 0x00, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x40, 0x11,
              0x66, 0xBD, 0x13, 0x01, 0x01, 0x01, 0x3B, 0x01, 0x02, 0x64},
    .udpHdr = {0x00, 0x16, 0x00, 0x16, 0x01, 0xD6, 0xC1, 0x35}}};

static void
create_nonem_queues(void);
adk_xml_cfg_t adkCfg;

/** Spin value before restarting the test */
#define DELAY_SPIN_COUNT                500000

#define TIME_OUT                100000
#define NONEM_QUEUE0  (long unsigned int)0x400

extern void *shm_app_addr;
extern em_conf_t em_conf;
static shared_app_common_t *s;
static em_nonem_common_t *ns;
extern uint8_t is_unified_em_adk_set;
extern uint8_t number_em_cores;
extern uint8_t em_start_port;

/**
 * The event for this test case.
 */
static shared_app_common_t *s;

extern uint16_t rem_soc[4096];
extern char *config_xml;
/*
 * Local function prototypes
 */
void test_appl_adk_ec_perf_multi_flow_init_local(void);
void test_appl_adk_ec_perf_multi_flow_start(void);
void test_appl_adk_ec_perf_multi_flow_stop(void);
void test_appl_adk_ec_perf_multi_flow_pre_init(void * , void *);

/**
 * @private
 *
 * EO start function.
 *
 */
static em_status_t
ec_perf_start( __attribute__ ((unused)) eo_context_common_t * eo_ctx,
	       __attribute__ ((unused)) em_eo_t eo)
{
	em_event_t event;
	event_test_t *my_event;
	em_status_t ret;
	uint8_t data_pattern = 0x10;

	(void)memset(&s->core_stats, 0, sizeof(s->core_stats));

	/* Allocate and send first CPU generated event message to remote soc */
	event = em_alloc(NONFRAG_SIZE, EM_EVENT_TYPE_SW, EM_POOL_DEFAULT);
	CHECK_RET(event == EM_EVENT_UNDEF, "Failure in event allocation \n");
	s->core_stats[em_core_id()].stat.alloc++;
	my_event = em_event_pointer(event);
	memset(my_event->data, data_pattern, NONFRAG_SIZE);
	ret = em_send(event, STATIC_QUEUE_ID_SOC(rem_soc[0]));
	CHECK_RET(ret != EM_OK,
		  "Event send ret= (%u)! queue: %" PRI_QUEUE " \n", ret,
		  (em_queue_t) STATIC_QUEUE_ID_SOC(rem_soc[0]));
	printf("Sending first event to remote SOC on queue 0x%x \n",
	       STATIC_QUEUE_ID_SOC(rem_soc[0]));

	dump_mac_table();


	if (ret == EM_OK)
		s->core_stats[em_core_id()].stat.snd_ev_cnt++;

	printf("\nSEND TRAFFIC\n");

	return EM_OK;
}

/**
 * @private
 *
 * EO stop function.
 *
 */
static em_status_t
ec_perf_stop(eo_context_common_t * eo_ctx, __attribute__ ((unused)) em_eo_t eo)
{
	em_status_t ret;
	ret = em_eo_delete(eo_ctx->this_eo_id);
	CHECK_RET(ret != EM_OK, "em_eo_delete ret=(%u). EO: %" PRI_EO "\n", ret,
		  eo_ctx->this_eo_id);

	return EM_OK;
}

/**
 * @private
 *
 * EO receive function.
 *
 */
static void
ec_perf_receive(eo_context_common_t * eo_ctx, em_event_t event,
		__attribute__ ((unused)) em_event_type_t type,em_queue_t queue,
		__attribute__ ((unused)) queue_context_common_t * q_ctx)
{
	int i;
	em_status_t ret=EM_OK;
	int core = em_core_id();
	uint16_t size = 0, udplen=0;
    	adk_app_msg_t *msg = NULL;
	int pktLen = 0;
	em_queue_t dq, dest_queue=EM_QUEUE_UNDEF;
	uint64_t snd_cnt = 0, rcv_cnt = 0;
	struct adk_sys_event_metadata *ev = (struct adk_sys_event_metadata *)em_event_to_nonem_param (event);

	/* update the count of data events received on this core */
	s->core_stats[core].stat.rcv_ev_cnt += 1;

	#if PRINT_MSG
	printf("Received EVENT size %d num_seg %d: in EM QUEUE ID %" PRI_QUEUE "\n", em_ext_event_size(event),
	em_ext_segment_count(event), queue);
	#endif
	pktLen = em_ext_event_size(event);
	size = htons(pktLen);
	/*UDP length*/
	udplen= htons(pktLen - 20);
	msg = (adk_app_msg_t *)em_event_pointer(event);
	#if PRINT_MSG
	printPacket((unsigned char *)msg, pktLen);
	#endif
	switch(queue) {

	 	case  EM_RECV_QUEUE_ID0_IP:
		case  EM_RECV_QUEUE_ID_IPV6:
			// Received packet from ADK
			//Add bip header and send it to Remote SOC
			//18 is the ethernet header 16+2
			em_event_adj(event, 18);
			uint8_t *p = (uint8_t*)msg;
			dq = STATIC_QUEUE_ID_SOC(rem_soc[p[11]]);
			
			#if PRINT_MSG
			printf("Sending events to remote SOC on queue 0x%lx \n", dq);
			#endif
			ret = em_send(event, dq);
			CHECK_RET(ret != EM_OK, "Event send ret= (%u)! queue: 0x%lx" PRI_QUEUE " \n", ret, dq);
			s->core_stats[core].stat.snd_ev_cnt++;
			break;
        default:
#if PRINT_MSG
            printf("Default Queue ID: %" PRI_QUEUE "\n", queue);
#endif
            ev = (struct adk_sys_event_metadata *)em_event_to_nonem_param (event);
            ev->udp_cksum_gen = 1;
	    batch_buf[batch_count++] = event;
            {
		uint8_t flowId;
		/* first byte of payload indicates which flow to use */
		flowId = (msg->ipHdr[0] & 0xff);
		dest_queue = g_nonem_queue_ip[flowId];
		memcpy(&ipPacket_flow[flowId].ipHdr[2], &size, 2);
		memcpy(&ipPacket_flow[flowId].udpHdr[4], &udplen, 2);
		memcpy((void *)msg, (void *)&ipPacket_flow[flowId], sizeof(ipPacket_flow[flowId]));
            }
#if PRINT_MSG
            printf("Sending event %p to queue 0x%lx\n", event, dest_queue);
#endif
	    if(batch_count < BATCH_SIZE && queue == em_ext_get_next_queue())
		return;
	    /*em_free(event);
	    s->core_stats[core].stat.free++;*/
            //ret = em_send(event, dest_queue);
	    ret = em_send_multi(batch_buf, batch_count, dest_queue);
            if (ret != batch_count)
            {
		uint32_t j;
		for( j = ret; j < batch_count; j ++) {
	                s->core_stats[core].stat.free++;
			em_free(batch_buf[j]);
		}
                s->core_stats[core].stat.snd_ev_cnt += ret;
            }
            else {
#if PRINT_MSG
                CHECK_RET(ret != EM_OK, "Event send ret= (%u)! queue: 0x%lx" PRI_QUEUE " \n", ret, dest_queue);
#endif
                s->core_stats[core].stat.snd_ev_cnt += batch_count;
            }
	    batch_count = 0;
            break;
	}



#if PRINT_MSG
	printf("Core %u Rcv %" PRIu64 " Snd %" PRIu64 " free %" PRIu64 " alloc %" PRIu64 "\n",
	       em_core_id(), s->core_stats[core].stat.rcv_ev_cnt,
	       s->core_stats[core].stat.snd_ev_cnt, s->core_stats[core].stat.free,
	       s->core_stats[core].stat.alloc);
#endif
	for (i = 0; i < em_core_count(); ++i) {
		snd_cnt += s->core_stats[i].stat.snd_ev_cnt;
		rcv_cnt += s->core_stats[i].stat.rcv_ev_cnt;
	}
	if (rcv_cnt == NUM_NONFRAG_EVENTS/2) {
		em_eo_accept_scattered_events(eo_ctx->this_eo_id);
	}

	if (snd_cnt == (NUM_NONFRAG_EVENTS + 1)
	    && rcv_cnt == NUM_NONFRAG_EVENTS) {
		s->app_done = 1;
	}
}


void test_appl_adk_ec_perf_multi_flow_pre_init(void *em_config , void *adk_config_xml)
{
	em_conf_t *emcfg = (em_conf_t *)em_config;

	/* Parse common params for all test cases */
	if (adk_non_em_common_pre_init(em_config, adk_config_xml, &adkCfg) < 0)
	{
		ADK_APPL_EXIT_FAILURE("Failed to parse and configure XML!");
	}
	/******* ADK DP SETUP STARTS FROM HERE ***************/

	adk_sys_handle_t sys_handle;
	int argc = 1;
	char *argv [20];

	/* Setting the Unified DP setup mode*
	 * EAL and HLP initialization wont be happening in DP setup under this mode*
	 */
	is_unified_em_adk_set = 1;
	/* Number of EM cores used, will be passed to ADK *
	 * And ADK will create eventDev port for EM cores *
	 */
	number_em_cores = emcfg->core_count;
	emQueueType = adkCfg.nonEm_queue_type;

	char str[512];
	strncpy(str, adkCfg.cmd_line_args, 512);

	char *p2 = strtok(str, " ");
	while (p2)
	{
		argv[argc++] = p2;
		p2 = strtok(0, " ");
	}
	argv[argc] = 0;
        argv[0] = NULL;

	int arg_c = argc;

	printf("Following are the arguments passed for ADK-DP setup :\n");
	while(arg_c ){
		printf("argv[%d] %s ", argc-arg_c, argv[argc-arg_c]);
		arg_c--;
	}

	if (adk_dp_setup(argc, argv) != 0) {
		printf("adk_dp_setup() failed");
		return ;
	}

	sys_handle = init_adk_modules();
	if(!sys_handle){
		printf("Failed in System init \n");
		return ;
	}

	if (adk_dp_init(sys_handle)){
		printf("Failed to init DP \n");
		return ;
	}

	emcfg->nonem_cfg[INDEX].start_port_id = em_start_port ; // Number of evdev Ports used by ADK

	printf("Successfully completed DP initializations!\n");
}

/**
 * Global init and startup for the test application.
 *
 */
void
test_appl_adk_ec_perf_multi_flow_start(void)
{
	em_eo_t eo;
	em_status_t ret;
	em_core_mask_t ul_mask;
	em_core_mask_t dl_mask;
	em_queue_prio_t prio = EM_QUEUE_PRIO_HIGHEST;
	queue_context_common_t *q_ctx;
	em_queue_group_t group_dl, group_ul;
	static char q_grp_name_dl[] = "qgr_dl";
	static char q_grp_name_ul[] = "qgr_ul";
	int i=0;

	ret = setup_port_to_queue_mapping();
	IS_ERROR(ret != EM_OK, "setup_port_to_queue_mapping ret=(%u). \n", ret);
#ifdef EMCFG_ETH_TX_FLOW
	if (!config_xml)
#endif
	{
		ret = em_app_ec_conf();
		IS_ERROR(ret != EM_OK, "em_app_ec_conf ret=(%u). \n", ret);
	}

	/*
	 * Create the EO and 2 parallel queues, add them to the EO
	 */
	memset(&s->eo_context[0], 0, sizeof(eo_context_common_t));

	eo = em_app_eo_create("perf EC appl", (em_start_func_t) ec_perf_start, NULL,
			      (em_stop_func_t) ec_perf_stop, NULL,
			      (em_receive_func_t) ec_perf_receive, &s->eo_context[0]);
	IS_ERROR(eo == EM_EO_UNDEF, "Error in creating EO \n");

	/* Init EO context */
	s->eo_context[0].this_eo_id = eo;
	s->eo_context[0].core_id = 0;
	s->eo_context[0].qgrp_num = 0;

	q_ctx = &s->queue_context[s->eo_context[0].qgrp_num];
	memset(q_ctx, 0, sizeof(queue_context_common_t));
	/*create DL EM queues */
	em_core_mask_zero(&dl_mask);
	dl_mask._u64[0]=0xC;
	printf("Test uses dl_mask.u64[0]:%" PRIu64 "\n", dl_mask._u64[0]);

	group_dl = em_app_queue_group_create(q_grp_name_dl, &dl_mask, 0, NULL);
	q_ctx = &s->queue_context[s->eo_context[0].qgrp_num];
        for (i = 0; i < 2; i++) {
		q_ctx->queue_id = ((my_device_id << 16) | ( (EM_QUEUE_STATIC_MIN + i) << 8)) + 1;
		ret = em_queue_create_static("staticQ", EM_QUEUE_TYPE_PARALLEL, EM_QUEUE_PRIO_LOW, group_dl,
				q_ctx->queue_id);
		CHECK_RET(ret != -EM_OK, "DL Static queue creation EO: %" PRI_EO ", queue: 0x%lx" PRI_QUEUE "\n",
				eo, q_ctx->queue_id);
		printf("DL Static queue created for IPvX EO: %" PRI_EO ", queue: 0x%lx" PRI_QUEUE "\n", eo,
				q_ctx->queue_id);
                q_ctx->queue_group = group_dl;
                ret = em_queue_set_context(q_ctx->queue_id, q_ctx);
                CHECK_RET(ret != EM_OK, "em_queue_set_context ret=(%u):queue: %" PRI_QUEUE " \n",
                          ret, q_ctx->queue_id);

                ret = em_eo_add_queue(eo, q_ctx->queue_id);
                CHECK_RET(ret != EM_OK,
                          "add queue ret=(%u). EO: %" PRI_EO " queue: %" PRI_QUEUE "\n", ret, eo,
                          q_ctx->queue_id);

                q_ctx++;
        }
	/* EM queue create for traffic from Remote SOC */
	em_core_mask_zero(&ul_mask);
	em_core_mask_set_count(em_core_count(), &ul_mask);
	printf("Test uses ul_mask.u64[0]:%" PRIu64 "\n", ul_mask._u64[0]);
	group_ul = em_app_queue_group_create(q_grp_name_ul, &ul_mask, 0, NULL);
	for (i = 0; i< 20; ++i) {
		g_ul_queue[i] = em_queue_create("DynamicQ", EM_QUEUE_TYPE_PARALLEL, prio, group_ul);
		IS_ERROR(g_ul_queue[i] == EM_QUEUE_UNDEF, "em_queue_create() failed\n");
		printf("UL Created Q to receive from remote SOC 0x%x\n", (unsigned int)g_ul_queue[i]);

		q_ctx->queue_id = g_ul_queue[i];
		ret = em_queue_set_context(q_ctx->queue_id, q_ctx);
		IS_ERROR(ret != EM_OK, "queue context set  EO: %" PRI_EO ", queue: %" PRI_QUEUE "\n", eo,
				q_ctx->queue_id);
                ret = em_eo_add_queue(eo, q_ctx->queue_id);
                CHECK_RET(ret != EM_OK,
                          "add queue ret=(%u). EO: %" PRI_EO " queue: %" PRI_QUEUE "\n", ret, eo,
                          q_ctx->queue_id);

                q_ctx++;
	}
	//Create ADK entry point queues
	create_nonem_queues();

} /* test_appl_adk_ec_perf_multi_flow_start */

void
test_appl_adk_ec_perf_multi_flow_init_local(void)
{
	EM_APP_INIT_SHM(shared_app_common_t*);
	ns = (em_nonem_common_t *) (((uint8_t *) shm_app_addr) + EM_APP_DATA_SIZE -
                                        sizeof(em_nonem_common_t));
}

void
test_appl_adk_ec_perf_multi_flow_stop(void)
{
	test_appl_common_end(APP_TIME_OUT, (char *)__FILE__, NULL);
}

#if PRINT_MSG
static void
printPacket(unsigned char *pkt, int pktLen)
{
        int i;

        printf("\n================= Len = %d =====================\n", pktLen);
        printf("Packet : ");
        for (i = 0; i < pktLen; i++)
        {
                printf("%02x ", pkt[i]);
                if ( i%16 == 0 )
                        printf("\n");
        }
        printf("\n==============================================\n");
}
#endif

static void
create_nonem_queues(void)
{
	em_core_mask_t mask;
	em_queue_group_t nonem_group0_ip;
	int i;
	#define MAX_CHARS 100
	char str1 [MAX_CHARS];
	/* Create queues for EM to ADK (non EM) */
	em_core_mask_zero(&mask);
	em_core_mask_set(VIRTUAL_CORE_ID_FOR_IPV4_ENTRY, &mask);
	nonem_group0_ip = em_queue_group_create("ip0", &mask, 0, NULL);
	if (emQueueType == EM_QUEUE_TYPE_ATOMIC)
  strncpy(str1,"EM_QUEUE_TYPE_ATOMIC",MAX_CHARS);
	else if (emQueueType == EM_QUEUE_TYPE_PARALLEL_ORDERED)
  strncpy(str1,"EM_QUEUE_TYPE_PARALLEL_ORDERED",MAX_CHARS);
	else if (emQueueType == EM_QUEUE_TYPE_PARALLEL)
  strncpy(str1,"EM_QUEUE_TYPE_PARALLEL",MAX_CHARS);
	else if (emQueueType == EM_QUEUE_TYPE_UNSCHEDULED)
  strncpy(str1,"EM_QUEUE_TYPE_UNSCHEDULED",MAX_CHARS);
	else
  strncpy(str1,"NOT A KNOWN VALUE",MAX_CHARS);

	printf("NON-EM Entry point QUEUE TYPE is *** %s *** \n",str1);

	for(i = 0; i < 20; i ++) {
		/*Dynamic Queue Creation*/
		sprintf(str1, "dynamic%d", i);
	        g_nonem_queue_ip[i] = em_queue_create(str1, emQueueType, EM_QUEUE_PRIO_LOW, nonem_group0_ip);
		CHECK_RET(g_nonem_queue_ip[i] == EM_QUEUE_UNDEF, "Dynamic queue creation failed: %" PRI_QUEUE "\n", g_nonem_queue_ip[i]);
		printf("UL: Dynamic queue from non EM IPv4 created, queue: 0x%lx" PRI_QUEUE "\n",
				g_nonem_queue_ip[i]);
	}
} /* create_nonem_queues */
