/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_dpdk_ports.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _adkSamples ADK Sample Applications
 * @{
 *
 * @brief Code to poll the DPDK ports, returning the physical and veth ports
 *
 *
 ***********************************************************************/

#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wsign-conversion"

#include "rte_ethdev.h"
#if ADK_DPDK_VERSION >= 18050016
#include "rte_bus_pci.h"
#endif

#include "adk_sys.h"
#include "adk_ip.h"
#include "adk_dp_app_config.h"
#include "adk_test_shm.h"
#include "adk_sample_dpdk_ports.h"

extern uint16_t bh_eth_dev_id;

/*!
 * =============================================================
 * dpdk_to_adk_port_number_get
 * @param[in,out]  numPhysPorts - The number of physical ports to return
 * @param[in,out]  numVethPorts - The number of veth ports to return
 * @param[in,out]  physPort - array of to hold the physical ports
 * @param[in,out]  vethPort - array of to hold the veth ports
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval 
 *
 * @brief  Query DPDK and return the physical and veth port numbers
 *
 * The routine will look through all the DPDK ports, returning the first N
 * ports of each type. The assumption is that if more DPDK ports are bound or
 * veth exits, the first N of each will be returned. The ports are returned in
 * the order that DPDK has them, which is most likely the order they would
 * bound to to DPDK (physical) or exist in the ifconfig.
 *
 * =============================================================
 */
void
dpdk_to_adk_port_number_get (
    void)
{
    uint8_t                     numEthDev;
    uint8_t                     portId, ethDevId = 0;
    uint8_t                     physIndex = 0;
    uint8_t                     vethIndex = 0;
    struct rte_eth_dev_info     ethDev;
    char                        name[TEST_MAX_PCI_STR_LEN] = "";
    dpdk_eth_dev_pci_t         *ethDevInfo_p;

    /*
     * Get the number of DPDK ports
     */
    numEthDev = rte_eth_dev_count_avail();
    if (0 == numEthDev)
    {
        printf("No Ethernet device initialized!\n");
        goto _return;
    }

    printf("Found %d Eth devices initialized\n", numEthDev);

    /*
     * Poll the devices. If devices have been dynamically removed, the device
     * IDs may not be contiguous, so we need to keep quering until we get
     * the number of ethernet devices defined.
     */
    {
        rte_eth_dev_info_get(bh_eth_dev_id, &ethDev);

        rte_eth_dev_get_name_by_port(bh_eth_dev_id, name);

        ethDevInfo_p = &(testConfShm_p->ethDev.ethDevInfo_p[bh_eth_dev_id]);

        ethDevInfo_p->ethDevId = ethDevId;
        ethDevInfo_p->ifIndex = ethDev.if_index;
        memcpy(ethDevInfo_p->deviceName, name, TEST_MAX_PCI_STR_LEN-1);
        memcpy(ethDevInfo_p->driverName, ethDev.driver_name,
                TEST_MAX_PCI_STR_LEN-1);
        ethDevInfo_p->deviceName[TEST_MAX_PCI_STR_LEN-1] = '\000';
        ethDevInfo_p->driverName[TEST_MAX_PCI_STR_LEN-1] = '\000';

        if (!strcmp(ethDevInfo_p->driverName, "net_ice_dsi"))
        {
            /* On SNR, there is only one ethernet (CPK)
             * device, having all HLP ports behind it */
            /* ADK is aware of only one bridge port 
             * No Physical port representation in ADK
             */
            testConfShm_p->ethDev.tnPorts_p[physIndex++] = bh_eth_dev_id;
        }


        printf("ethDevId: %d\n    driver: '%s', ifId: %d, pciAddr: "
               "%x:%x:%x.%x, name: '%s'\n",
               ethDevId, ethDevInfo_p->driverName,
               ethDevInfo_p->ifIndex,
               ethDevInfo_p->addr.domain,
               ethDevInfo_p->addr.bus,
               ethDevInfo_p->addr.devid,
               ethDevInfo_p->addr.function,
               ethDevInfo_p->deviceName);
        printf("    class: 0x%x vendor: 0x%x, devId: 0x%x, "
               "sVndId: 0x%x, sDevId: 0x%x\n",
               ethDevInfo_p->id.class_id,
               ethDevInfo_p->id.vendor_id,
               ethDevInfo_p->id.device_id,
               ethDevInfo_p->id.subsystem_vendor_id,
               ethDevInfo_p->id.subsystem_device_id);
    }

    testConfShm_p->ethDev.numTnPorts = physIndex;
    testConfShm_p->ethDev.numHostPorts = vethIndex;
    testConfShm_p->ethDev.numEthDevs = ethDevId;

    for (portId = 0; portId < testConfShm_p->ethDev.numTnPorts; portId++)
        printf("Found TN Port: %d\n", testConfShm_p->ethDev.tnPorts_p[portId]);
    for (portId = 0; portId < testConfShm_p->ethDev.numHostPorts; portId++)
        printf("Found Veth ID: %d\n", testConfShm_p->ethDev.hostPorts_p[portId]);

_return:
    return;
}    /* -----  end of function dpdk_to_adk_port_number_get  ----- */

/*! end addtogroup _adkSamples
 * @}
 */
