/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_helper.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file has all the helper functions required in the
 *          Ethernet module for control plane APIs.
 *
 *  **********************************************************************/


#include <adk_eth.h>

/* Function to validate the ethernet module handle */
bool
adk_eth_valid_handle (
    adk_module_handle_t eth_mod_handle)
{
    adk_ethernet_internal_data_t *eth_internal_data_p = NULL;

    if (eth_mod_handle == 0)
        return FALSE;

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (eth_internal_data_p == NULL)
        return FALSE;

    if (eth_internal_data_p->cookie != ADK_ETHERNET_MODULE_COOKIE)
        return FALSE;

    return TRUE;
}

/* Function to validate an ethernet interface handle */
bool
adk_eth_valid_if_handle (
    adk_interface_handle_t if_handle)
{
    adk_eth_intf_internal_data_t    *if_data_p = NULL;

    if (if_handle == 0)
        return FALSE;

    if_data_p = (adk_eth_intf_internal_data_t *)
                ADK_SYS_INTERFACE_ATTR_PTR_GET(if_handle);

    if (if_data_p == NULL)
        return FALSE;

    if (if_data_p->cookie != ADK_ETHERNET_INTF_COOKIE)
        return FALSE;

    return TRUE;
}

/* Function to validate an ethernet flow handle */
bool
adk_eth_valid_flow_handle (
    adk_flow_handle_t flow_handle)
{
    adk_eth_flow_internal_data_t *flow_data_p = NULL;

    if (flow_handle == 0)
        return FALSE;

    flow_data_p = (adk_eth_flow_internal_data_t *)
                  ADK_SYS_FLOW_ATTR_PTR_GET(flow_handle);

    if (flow_data_p == NULL)
        return FALSE;

    if (flow_data_p->cookie != ADK_ETHERNET_FLOW_COOKIE)
        return FALSE;

    return TRUE;
}

adk_status_t
adk_eth_map_sys_log (
    adk_module_handle_t eth_mod_handle,
    adk_log_level_t sys_log_level,
    adk_ethernet_log_level_t *eth_log_level)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

#if 0   /* No conversion needed as the values are same */
    switch (sys_log_level)
    {
    case ADK_LOG_LEVEL_INFO:
        *eth_log_level = ADK_ETHERNET_LOG_LEVEL_INFO; break;

    case ADK_LOG_LEVEL_WARNING:
        *eth_log_level = ADK_ETHERNET_LOG_LEVEL_WARNING; break;

    case ADK_LOG_LEVEL_ERROR:
        *eth_log_level = ADK_ETHERNET_LOG_LEVEL_ERROR; break;

    case ADK_LOG_LEVEL_FATAL:
        *eth_log_level = ADK_ETHERNET_LOG_FATAL; break;

    default:
        ADK_ETH_LOG_WARNING(eth_mod_handle,
                            "Unsupported log-level in Ethernet!");
        ret = ADK_E_UNSUPPORTED;
        break;
    }
#endif

    *eth_log_level = sys_log_level;

    return (ret);
}

adk_log_level_t
adk_eth_map_eth_log (
    adk_module_handle_t eth_mod_handle)
{
    adk_ethernet_log_level_t eth_log_level;

    eth_log_level = ADK_MODULE_LOG_LEVEL_GET(eth_mod_handle);

    if (eth_log_level & ADK_ETHERNET_LOG_INFO)
        return (ADK_LOG_LEVEL_INFO);
    else if (eth_log_level & ADK_ETHERNET_LOG_WARNING)
        return (ADK_LOG_LEVEL_WARNING);
    else if (eth_log_level & ADK_ETHERNET_LOG_ERROR)
        return (ADK_LOG_LEVEL_ERROR);

    return (ADK_LOG_LEVEL_FATAL);
}

/* Function to create an interface ID for VLAN sub interface */
uint32_t
adk_eth_create_subif_id (
    uint32_t port,
    adk_ethernet_vlan_id_t vlan_id)
{
    uint32_t sub_if_num;

    port &= ETH_PHY_PORT_MASK;
    vlan_id &= VLAN_ID_MASK;

    sub_if_num = (vlan_id << ETH_PHY_PORT_WIDTH) + port;

    return (sub_if_num);
}

/* Function to create an bridge interface ID from NS ID */
uint32_t adk_eth_create_bridge_if_id(uint32_t ns_id)
{
    uint32_t br_if_num;

    ns_id &= 0xF;

    br_if_num = (ns_id << 4) + 0xF;

    return(br_if_num);
}

/* Function to create an bridge interface ID from NS ID */
uint32_t adk_eth_get_bridge_if_id(uint32_t if_id)
{
    uint32_t ns_id;

    ns_id = if_id >> 4;

    ns_id &= 0xF;

    return(ns_id);
}

/* Function to create an interface ID from NS ID */
uint32_t
adk_eth_create_virtual_if_id (
    uint32_t ns_id)
{
    uint32_t v_if_num;

    ns_id &= VLAN_ID_MASK;

    v_if_num = (ns_id << ETH_PHY_PORT_WIDTH) + 0xE;

    return (v_if_num);
}

/* Function to get NS ID from virtual interface ID */
uint32_t
adk_eth_get_virtual_if_ns_id (
    uint32_t if_id)
{
    uint32_t ns_id;

    ns_id = if_id >> ETH_PHY_PORT_WIDTH;

    ns_id &= VLAN_ID_MASK;

    return (ns_id);
}

adk_status_t
adk_eth_valid_ingress_flow_data (
    adk_module_handle_t eth_mod_handle,
    const adk_ethernet_ingress_flow_data_t *in_flow_data)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    if (!(in_flow_data->mask & ADK_ETHERNET_FLOW_MASK_ETYPE))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Ethertype must be specified\n");
        goto _adk_eth_valid_ingress_flow_data_exit;
    }

    if (ADK_ETH_IS_PROTO_INVALID(in_flow_data->proto))
    {
        ret = ADK_ETHERNET_E_INVALID_PROTO_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid ether type 0x%04x\n", in_flow_data->proto);
        goto _adk_eth_valid_ingress_flow_data_exit;
    }

    if ((in_flow_data->mask & ADK_ETHERNET_FLOW_MASK_DMAC) &&
        ADK_ETH_IS_MAC_SUBNET_LEN_INVALID((&(in_flow_data->dest_mac))))
    {
        ret = ADK_ETHERNET_E_INVALID_MAC_SUBNET;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid subnet MAC address length %d\n",
                          in_flow_data->dest_mac.num_valid_msb_bits);
        goto _adk_eth_valid_ingress_flow_data_exit;
    }

    /* If eth_type and LLC are wild-carded, the destination MAC address and vlan must
     * also be wild-carded */
    if (!(in_flow_data->mask & ADK_ETHERNET_FLOW_MASK_ETYPE) &&
        (in_flow_data->mask & ADK_ETHERNET_FLOW_MASK_DMAC)) 
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "DMAC cannot be specified if proto is wild-carded\n");
        goto _adk_eth_valid_ingress_flow_data_exit;
    }


    if (!(ADK_ETHERNET_INTERFACE_TYPE_BRIDGE == in_flow_data->intfType) &&
        !(ADK_ETHERNET_INTERFACE_TYPE_PHY == in_flow_data->intfType))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Interface type should be either bridge or physical\n");
        goto _adk_eth_valid_ingress_flow_data_exit;
    }

_adk_eth_valid_ingress_flow_data_exit:

    return (ret);
}

adk_status_t
adk_eth_valid_ingress_flow_action (
    adk_module_handle_t eth_mod_handle,
    const adk_ethernet_ingress_flow_data_t *in_flow_data,
    const adk_ethernet_ingress_flow_action_t *in_flow_action)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);


    if (ADK_ETH_IS_INGRESS_FLOW_ACTION_TYPE_INVALID(
            in_flow_action->action_type))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid ingress flow action type %s\n",
                          adk_eth_flow_action_name(
                              in_flow_action->action_type));
        goto _adk_eth_valid_ingress_flow_action_exit;
    }
    if (in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_HOST)
    {
        if(in_flow_action->enablePolicing && NULL == in_flow_action->policingHandle)
        {
            ret = ADK_E_INTERNAL_ERROR;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                    "Invalid policing handle");
            goto _adk_eth_valid_ingress_flow_action_exit;
        }
    }
    if(in_flow_data->intfType == ADK_ETHERNET_INTERFACE_TYPE_PHY &&
        (in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_IPV4 ||
           in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_IPV6))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid ingress flow action type %s for interface type PHY\n",
                adk_eth_flow_action_name(
                    in_flow_action->action_type));
        goto _adk_eth_valid_ingress_flow_action_exit;
    }

    if((in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_IPV4 ||
           in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_IPV6) &&
            in_flow_action->enablePolicing)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Policing is not supported");
        goto _adk_eth_valid_ingress_flow_action_exit;
    }

    if((in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_IPV4) &&
        (in_flow_data->proto != ETHER_TYPE_IPV4))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid ingress flow action type %s for etherType\n",
                adk_eth_flow_action_name(
                    in_flow_action->action_type));
        goto _adk_eth_valid_ingress_flow_action_exit;
    }

    if(in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_IPV6  &&
        in_flow_data->proto != ETHER_TYPE_IPV6)
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid ingress flow action type %s for etherType\n",
                adk_eth_flow_action_name(
                    in_flow_action->action_type));
        goto _adk_eth_valid_ingress_flow_action_exit;
    }
_adk_eth_valid_ingress_flow_action_exit:

    return (ret);
}


#ifdef NOT_NOW_L3CALL

adk_status_t
adk_eth_valid_association_flow_action (
    adk_module_handle_t eth_mod_handle,
    bool override_ingress_flow,
    const adk_ethernet_ingress_flow_action_t *in_flow_action)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_eth_intf_internal_data_t *if_data_p;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    if ((in_flow_action->action_type != ADK_ETHERNET_ASSOCIATION_RULE_TYPE_FORWARD) &&
        (override_ingress_flow == TRUE))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Incompatible override flag and action type\n");
        goto _adk_eth_valid_association_flow_action_exit;
    }

    if (ADK_ETH_IS_INGRESS_FLOW_ACTION_TYPE_INVALID(
            in_flow_action->action_type))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid ingress flow action type %s\n",
                          adk_eth_flow_action_name(
                              in_flow_action->action_type));
        goto _adk_eth_valid_association_flow_action_exit;
    }

    if (in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_CPU)
    {
        if (ADK_ETH_IS_PRIORITY_INVALID(in_flow_action->u.cpu_data.priority))
        {
            ret = ADK_ETHERNET_E_INVALID_PRIORITY;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Invalid priority %d\n",
                              in_flow_action->u.cpu_data.priority);
            goto _adk_eth_valid_association_flow_action_exit;
        }
    }

#ifndef ADK_QOS_GEN3_INTERFACE
    if (in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_NCA)
#else
    if (in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_SP_STACK)
#endif
    {
        if (ADK_ETH_IS_PRIORITY_INVALID(in_flow_action->u.nca_data.priority))
        {
            ret = ADK_ETHERNET_E_INVALID_PRIORITY;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Invalid priority %d\n",
                              in_flow_action->u.nca_data.priority);
            goto _adk_eth_valid_association_flow_action_exit;
        }
    }

    if (in_flow_action->action_type == ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_PORT)
    {
        if (!adk_eth_valid_if_handle(in_flow_action->u.output_if_handle))
        {
            ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Invalid Interface handle %"ADKPRIh "\n",
                              in_flow_action->u.output_if_handle);
            goto _adk_eth_valid_association_flow_action_exit;
        }

        if_data_p = (adk_eth_intf_internal_data_t *)
                    ADK_SYS_INTERFACE_ATTR_PTR_GET(in_flow_action->u.output_if_handle);

        if ((if_data_p->type != INTERFACE_TYPE_PHY) &&
            (if_data_p->type != INTERFACE_TYPE_AGGR))
        {
            ret = ADK_ETHERNET_E_INCOMPAT_IF_TYPE;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Incompatible Interface type %s\n",
                              adk_eth_intf_type_name(if_data_p->type));
            goto _adk_eth_valid_association_flow_action_exit;
        }
    }

_adk_eth_valid_association_flow_action_exit:

    return (ret);
}

#else

adk_status_t
adk_eth_valid_association_flow_action (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    bool __attribute__ ((unused)) override_ingress_flow,
    const adk_ethernet_ingress_flow_action_t __attribute__ ((unused)) *in_flow_action)
{
    return ADK_SUCCESS;
}

#endif

const char *
adk_eth_intf_type_name (
    adk_ethernet_intf_type_t type)
{
    switch (type)
    {
    case INTERFACE_TYPE_INVALID:        return ("INTERFACE_TYPE_INVALID");

    case INTERFACE_TYPE_PHY:            return ("INTERFACE_TYPE_PHY");

    case INTERFACE_TYPE_VLAN:           return ("INTERFACE_TYPE_VLAN");

    case INTERFACE_TYPE_BRIDGE:         return ("INTERFACE_TYPE_BRIDGE");

    case INTERFACE_TYPE_SWITCH_VIRTUAL: return ("INTERFACE_TYPE_SWITCH_VIRTUAL");

    case INTERFACE_TYPE_MAX:            return ("INTERFACE_TYPE_MAX");

    default:                            return ("INTERFACE_TYPE_UNKNOWN");
    }
}

const char *
adk_ethernet_encap_type_name (
    adk_ethernet_encap_t encap_type)
{
    switch (encap_type)
    {
    case ADK_ETHERNET_ENCAP_INVALID:     return ("ADK_ETHERNET_ENCAP_INVALID");

    case ADK_ETHERNET_ENCAP_DIX:         return ("ADK_ETHERNET_ENCAP_DIX");

    case ADK_ETHERNET_ENCAP_LLC_SNAP:    return ("ADK_ETHERNET_ENCAP_LLC_SNAP");

    case ADK_ETHERNET_ENCAP_MAX:         return ("ADK_ETHERNET_ENCAP_MAX");

    default:                    return ("ADK_ETHERNET_ENCAP_UNKNOWN");
    }
}

const char *
adk_eth_flow_type_name (
    adk_ethernet_flow_type_t flow_type)
{
    switch (flow_type)
    {
    case ETH_FLOW_TYPE_INVALID:
        return ("ETH_FLOW_TYPE_INVALID");

    case ETH_FLOW_TYPE_SIMPLE_INGRESS:
        return ("ETH_FLOW_TYPE_SIMPLE_INGRESS");

    case ETH_FLOW_TYPE_EXT_INGRESS_ASSOC:
        return ("ETH_FLOW_TYPE_EXT_INGRESS_ASSOC");

    case ETH_FLOW_TYPE_POLICING:
        return ("ETH_FLOW_TYPE_POLICING");

    case ETH_FLOW_TYPE_SIMPLE_EGRESS:
        return ("ETH_FLOW_TYPE_SIMPLE_EGRESS");

    case ETH_FLOW_TYPE_EGRESS_ENCAP:
        return ("ETH_FLOW_TYPE_EGRESS_ENCAP");

    case ETH_FLOW_TYPE_IPV4_UDP_ENCAP:
        return ("ETH_FLOW_TYPE_IPV4_UDP_ENCAP");

    case ETH_FLOW_TYPE_HASH:
        return ("ETH_FLOW_TYPE_HASH");

    case ETH_FLOW_TYPE_CPU_ORIGIN:
        return ("ETH_FLOW_TYPE_CPU_ORIGIN");

    case ETH_FLOW_TYPE_MAX:
        return ("ETH_FLOW_TYPE_MAX");

    default:
        return ("ETH_FLOW_TYPE_UNKNOWN");
    }
}

const char *
adk_eth_flow_action_name (
    adk_ethernet_ingress_flow_action_type_t flow_action)
{
    switch (flow_action)
    {
    case ADK_ETHERNET_ASSOCIATION_RULE_TYPE_INVALID:
        return ("ADK_ETHERNET_ASSOCIATION_RULE_TYPE_INVALID");

    case ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_HOST:
        return ("ADK_ETHERNET_ASSOCIATION_RULE_TYPE_SEND_TO_HOST");

    case ADK_ETHERNET_ASSOCIATION_RULE_TYPE_DROP:
        return ("ADK_ETHERNET_ASSOCIATION_RULE_TYPE_DROP");

    case ADK_ETHERNET_ASSOCIATION_RULE_TYPE_MAX:
        return ("ADK_ETHERNET_ASSOCIATION_RULE_TYPE_MAX");

    default:
        return ("ADK_ETHERNET_ASSOCIATION_RULE_TYPE_UNKNOWN");
    }
}

const char *
adk_eth_error_name (
    adk_status_t err)
{
    switch (err)
    {
    case ADK_ETHERNET_E_INVALID_IF_HANDLE:
        return ("ADK_ETHERNET_E_INVALID_IF_HANDLE");

    case ADK_ETHERNET_E_INVALID_FLOW_HANDLE:
        return ("ADK_ETHERNET_E_INVALID_FLOW_HANDLE");

    case ADK_ETHERNET_E_INVALID_IF_TYPE:
        return ("ADK_ETHERNET_E_INVALID_IF_TYPE");

    case ADK_ETHERNET_E_INVALID_PORT_NUMBER:
        return ("ADK_ETHERNET_E_INVALID_PORT_NUMBER");

    case ADK_ETHERNET_E_INVALID_PROTO_TYPE:
        return ("ADK_ETHERNET_E_INVALID_PROTO_TYPE");

    case ADK_ETHERNET_E_INVALID_LLC_TYPE:
        return ("ADK_ETHERNET_E_INVALID_LLC_TYPE");

    case ADK_ETHERNET_E_INVALID_VLAN_ID:
        return ("ADK_ETHERNET_E_INVALID_VLAN_ID");

    case ADK_ETHERNET_E_INVALID_MCAST_GROUP_HANDLE:
        return ("ADK_ETHERNET_E_INVALID_MCAST_GROUP_HANDLE");

    case ADK_ETHERNET_E_INVALID_MAC_ADDR:
        return ("ADK_ETHERNET_E_INVALID_MAC_ADDR");

    case ADK_ETHERNET_E_INVALID_IPV4_ADDR:
        return ("ADK_ETHERNET_E_INVALID_IPV4_ADDR");

    case ADK_ETHERNET_E_INVALID_MAC_SUBNET:
        return ("ADK_ETHERNET_E_INVALID_MAC_SUBNET");

    case ADK_ETHERNET_E_INVALID_PRIORITY:
        return ("ADK_ETHERNET_E_INVALID_PRIORITY");

    case ADK_ETHERNET_E_INVALID_POLICING_LEVEL:
        return ("ADK_ETHERNET_E_INVALID_POLICING_LEVEL");

    case ADK_ETHERNET_E_INVALID_OUTPUT_RATE:
        return ("ADK_ETHERNET_E_INVALID_OUTPUT_RATE");

    case ADK_ETHERNET_E_INVALID_VLAN_ASSIGN_POLICY_TYPE:
        return ("ADK_ETHERNET_E_INVALID_VLAN_ASSIGN_POLICY_TYPE");

    case ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE:
        return ("ADK_ETHERNET_E_INVALID_FLOW_ACTION_TYPE");

    case ADK_ETHERNET_E_INVALID_DUPLEX_MODE:
        return ("ADK_ETHERNET_E_INVALID_DUPLEX_MODE");

    case ADK_ETHERNET_E_INVALID_ENCAP:
        return ("ADK_ETHERNET_E_INVALID_ENCAP");

    case ADK_ETHERNET_E_INVALID_FWD_MODE:
        return ("ADK_ETHERNET_E_INVALID_FWD_MODE");

    case ADK_ETHERNET_E_INVALID_STATS_TYPE:
        return ("ADK_ETHERNET_E_INVALID_STATS_TYPE");

    case ADK_ETHERNET_E_INVALID_ASSOCIATION_MASK:
        return ("ADK_ETHERNET_E_INVALID_ASSOCIATION_MASK");

    case ADK_ETHERNET_E_INVALID_MTU:
        return ("ADK_ETHERNET_E_INVALID_MTU");

    case ADK_ETHERNET_E_INCOMPAT_PORT_TYPE:
        return ("ADK_ETHERNET_E_INCOMPAT_PORT_TYPE");

    case ADK_ETHERNET_E_INCOMPAT_IF_TYPE:
        return ("ADK_ETHERNET_E_INCOMPAT_IF_TYPE");

    case ADK_ETHERNET_E_INCOMPAT_STATS_TYPE:
        return ("ADK_ETHERNET_E_INCOMPAT_STATS_TYPE");

    case ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE:
        return ("ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE");

    case ADK_ETHERNET_E_INCOMPAT_IF_STATE:
        return ("ADK_ETHERNET_E_INCOMPAT_IF_STATE");

    case ADK_ETHERNET_E_INCOMPAT_FWD_MODE:
        return ("ADK_ETHERNET_E_INCOMPAT_FWD_MODE");

    case ADK_ETHERNET_E_INCOMPAT_VLAN_ASSIGN_POLICY_TYPE:
        return ("ADK_ETHERNET_E_INCOMPAT_VLAN_ASSIGN_POLICY_TYPE");

    case ADK_ETHERNET_E_INCOMPAT_EIOA_VLAN_ETH_TYPE:
        return ("ADK_ETHERNET_E_INCOMPAT_EIOA_VLAN_ETH_TYPE");

    case ADK_ETHERNET_E_INCONSIST_IF_STATE:
        return ("ADK_ETHERNET_E_INCONSIST_IF_STATE");

    case ADK_ETHERNET_E_UNSUPPORT_STATS_TYPE:
        return ("ADK_ETHERNET_E_UNSUPPORT_STATS_TYPE");

    case ADK_ETHERNET_E_INTF_ALREADY_EXISTS:
        return ("ADK_ETHERNET_E_INTF_ALREADY_EXISTS");

    case ADK_ETHERNET_E_NON_EXISTENT_INTF:
        return ("ADK_ETHERNET_E_NON_EXISTENT_INTF");

    case ADK_ETHERNET_E_INTF_IS_ALREADY_MEMBER_OF_MCAST_GROUP:
        return ("ADK_ETHERNET_E_INTF_IS_ALREADY_MEMBER_OF_MCAST_GROUP");

    case ADK_ETHERNET_E_INTF_IS_NOT_MEMBER_OF_MCAST_GROUP:
        return ("ADK_ETHERNET_E_INTF_IS_NOT_MEMBER_OF_MCAST_GROUP");

    case ADK_ETHERNET_E_EIOA_VLAN_ETH_TYPE_FULL:
        return ("ADK_ETHERNET_E_EIOA_VLAN_ETH_TYPE_FULL");

    case ADK_ETHERNET_E_EIOA_VLAN_ETH_TYPE_ABSENT:
        return ("ADK_ETHERNET_E_EIOA_VLAN_ETH_TYPE_ABSENT");

    default:
        break;
    }

    return ("ADK_ETHERNET_E_UNKNOWN");
}

/*! end addtogroup _ethernet
 * @}
 */
