/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_hw.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file has all the low level APIs which interact
 *          directly with hardware to configure, add/rem entries,
 *          read configuration/statistics, etc.
 *
 *  **********************************************************************/


#include <adk_eth.h>

/* This function must be called with valid Ethernet module handle and while
 * holding the module level lock. As return, it provides a boolean flag, which
 * indicates if synchronization is needed or not. If true, sync is needed else
 * not. That is, the caller can proceed and free up objects if this function
 * returns false */
bool adk_eth_check_rcu_sync(adk_module_handle_t eth_mod_handle)
{
    bool sync_needed = false;
    uint32_t i = 0;
    adk_ethernet_internal_data_t *eth_internal_data_p = NULL;
    adk_dp_eth_gl_stats_t *dp_gl_stats_p;

    /* Get the module info pointer from the handle */
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(eth_mod_handle);

    dp_gl_stats_p = eth_internal_data_p->dp_pvt_p->gl_stats_p;

    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        if (dp_gl_stats_p->proc_cntr & 0x1)
        {
            sync_needed = true;
            eth_internal_data_p->sync_cntr[i] = dp_gl_stats_p->proc_cntr;
            eth_internal_data_p->sync_pending[i] = true;
        }
        dp_gl_stats_p++;
    }

    return sync_needed;
}

/* This function must be called with valid Ethernet module handle and while
 * holding the module level lock. As return, it provides a boolean flag, which
 * indicates if pending synchronization is complete or not. If true, sync is 
 * complete else not. That is, the caller can proceed and free up objects if
 * this function returns true */
bool adk_eth_check_rcu_status(adk_module_handle_t eth_mod_handle)
{
    bool sync_complete = true;
    uint32_t i = 0;
    adk_ethernet_internal_data_t *eth_internal_data_p = NULL;
    adk_dp_eth_gl_stats_t *dp_gl_stats_p;

    /* Get the module info pointer from the handle */
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(eth_mod_handle);

    dp_gl_stats_p = eth_internal_data_p->dp_pvt_p->gl_stats_p;

    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        if (eth_internal_data_p->sync_pending[i] == true)
        {
            if ((dp_gl_stats_p->proc_cntr & 0x1) &&
                (eth_internal_data_p->sync_cntr[i] ==
                 dp_gl_stats_p->proc_cntr))
            {
                sync_complete = false;
            }
            else
            {
                eth_internal_data_p->sync_pending[i] = false;
            }
        }
        dp_gl_stats_p++;
    }

    return sync_complete;
}

/* This function must be called with valid Ethernet module handle and while
 * holding the module level lock. This function is a blocking call which waits
 * until the RCU sync is complete. */
void adk_eth_wait_rcu_sync(adk_module_handle_t eth_mod_handle)
{
    uint32_t i = 0;
    adk_ethernet_internal_data_t *eth_internal_data_p = NULL;
    adk_dp_eth_gl_stats_t *dp_gl_stats_p;

    /* Get the module info pointer from the handle */
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(eth_mod_handle);

    dp_gl_stats_p = eth_internal_data_p->dp_pvt_p->gl_stats_p;

    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        if (eth_internal_data_p->sync_pending[i] == true)
        {
            while ((dp_gl_stats_p->proc_cntr & 0x1) &&
                   (eth_internal_data_p->sync_cntr[i] ==
                    dp_gl_stats_p->proc_cntr))
            {
/*                usleep(ADK_ETH_RCU_POLL_PERIOD);*/
            }
            eth_internal_data_p->sync_pending[i] = false;
        }
        dp_gl_stats_p++;
    }

    return;
}

adk_status_t
adk_eth_port_stats_get (
    adk_module_handle_t eth_mod_handle,
    uint32_t port,
    adk_ethernet_port_statistics_t *stats)
{
    adk_status_t ret = ADK_SUCCESS;
    int xtats_array_sz, k;
    struct rte_eth_xstat *xstats = NULL;
    struct rte_eth_xstat_name *xstats_names = NULL;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    xstats = (struct rte_eth_xstat *)
        ADK_RAM_CALLOC(ADK_ETH_MAX_STATS, sizeof(struct rte_eth_xstat));
    xstats_names = (struct rte_eth_xstat_name *)
        ADK_RAM_CALLOC(ADK_ETH_MAX_STATS, sizeof(struct rte_eth_xstat_name));

    if ((xstats == NULL) || (xstats_names == NULL))
    {
        ret = ADK_E_MEM_UNAVAILABLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to allocate memory to gather port %d stats!",
                          port);
        goto _adk_eth_port_stats_get_exit;
    }

    xtats_array_sz = rte_eth_xstats_get_names(port, xstats_names,
                                              ADK_ETH_MAX_STATS);
    if ((xtats_array_sz < 0) || (xtats_array_sz > ADK_ETH_MAX_STATS))
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get extended stats names for port %d."
                          " Array size required %d\n", port, xtats_array_sz);
        goto _adk_eth_port_stats_get_exit;
    }

    for (k = 0; k < xtats_array_sz; k++)
    {
        xstats[k].id = k;
        xstats[k].value = 0;
    }

    xtats_array_sz = rte_eth_xstats_get(port, xstats, ADK_ETH_MAX_STATS);
    if ((xtats_array_sz < 0) || (xtats_array_sz > ADK_ETH_MAX_STATS))
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get extended stats for port %d."
                          " Array size required %d\n", port, xtats_array_sz);
        goto _adk_eth_port_stats_get_exit;
    }

    for (k = 0; k < xtats_array_sz; k++)
    {
        if (!xstats[k].value)
            continue;

        if (!ADK_STRNCMP(xstats_names[k].name, "rx_good_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_frm += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_good_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_frm += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_good_bytes",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_octets += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_good_bytes",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_octets += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_frm += xstats[k].value;
            stats->rx_bad_frm += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_frm += xstats[k].value;
            stats->tx_bad_frm += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_unicast_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_multicast_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_mcast += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_broadcast_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_bcast += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_dropped",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_drop += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_unicast_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_multicast_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_mcast += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_broadcast_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_bcast += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_dropped",
                         ADK_ETH_STATS_NAME_SIZE))
        {
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_crc_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_fcs += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_illegal_byte_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_error_bytes",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_octets += xstats[k].value;
            stats->rx_bad_octets += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_length_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_length += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_64_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_64 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_65_to_127_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_65 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_128_to_255_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_128 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_256_to_511_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_256 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_512_to_1023_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_512 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_1024_to_1522_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_1024 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_size_1523_to_max_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_1519 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_undersized_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_usize += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_oversize_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_osize += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_mac_short_dropped",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_urun += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_fragmented_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_frag += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "rx_jabber_errors",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_jab += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_64_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_64 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_65_to_127_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_65 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_128_to_255_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_128 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_256_to_511_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_256 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_512_to_1023_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_512 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_1024_to_1522_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->tx_1024 += xstats[k].value;
        }
        if (!ADK_STRNCMP(xstats_names[k].name, "tx_size_1523_to_max_packets",
                         ADK_ETH_STATS_NAME_SIZE))
        {
            stats->rx_1519 += xstats[k].value;
        }
    }

    /* Include FCS bytes in the octet counters */
    stats->rx_octets += (stats->rx_frm * ADK_ETH_CRC_SIZE);
    stats->tx_octets += (stats->tx_frm * ADK_ETH_CRC_SIZE);

_adk_eth_port_stats_get_exit:

    if (xstats)
    {
        ADK_RAM_FREE(xstats);
    }
    if (xstats_names)
    {
        ADK_RAM_FREE(xstats_names);
    }

    return ret;
}

adk_status_t
adk_eth_port_stats_reset (
    uint32_t port)
{
    rte_eth_xstats_reset(port);

    return ADK_SUCCESS;
}

adk_status_t
adk_eth_dp_interface_stats_get (
    adk_module_handle_t eth_mod_handle,
    uint32_t interfaceID,
    adk_eth_intf_internal_data_t *if_data_p,
    void *stats)
{
    adk_ethernet_internal_data_t *eth_internal_data_p;
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(eth_mod_handle);
    adk_dp_eth_stats_t *dp_intf_stats_p;
    adk_ethernet_vlan_interface_statistics_t *vlan_stats;
    adk_ethernet_bridge_interface_statistics_t *bridge_stats;
    uint32_t i = 0;

    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        if (if_data_p->type == INTERFACE_TYPE_SWITCH_VIRTUAL)
        {
            dp_intf_stats_p = &(if_data_p->dp_if_ptr->intf_stats[i]);
            vlan_stats = (adk_ethernet_vlan_interface_statistics_t*)stats;

            vlan_stats->inEthTypeMismatch += dp_intf_stats_p->inEthTypeMismatch;

            vlan_stats->inFrames += (dp_intf_stats_p->inFrames +
                    dp_intf_stats_p->inIpMcastFrames +
                    dp_intf_stats_p->inIpBcastFrames +
                    dp_intf_stats_p->inNonIpMcastFrames +
                    dp_intf_stats_p->inNonIpBcastFrames);

            vlan_stats->inOctets += (dp_intf_stats_p->inOctets +
                    dp_intf_stats_p->inIpMcastOctets +
                    dp_intf_stats_p->inIpBcastOctets +
                    dp_intf_stats_p->inNonIpMcastOctets +
                    dp_intf_stats_p->inNonIpBcastOctets);

            vlan_stats->OutFrames+= (dp_intf_stats_p->OutFrames +
                    dp_intf_stats_p->OutIpMcastFrames +
                    dp_intf_stats_p->OutIpBcastFrames +
                    dp_intf_stats_p->OutNonIpMcastFrames +
                    dp_intf_stats_p->OutNonIpBcastFrames);


            vlan_stats->OutOctets += (dp_intf_stats_p->OutOctets +
                    dp_intf_stats_p->OutIpMcastOctets +
                    dp_intf_stats_p->OutIpBcastOctets +
                    dp_intf_stats_p->OutNonIpMcastOctets +
                    dp_intf_stats_p->OutNonIpBcastOctets);


            vlan_stats->inMcastFrames += (dp_intf_stats_p->inIpMcastFrames + dp_intf_stats_p->inIpBcastFrames);
            vlan_stats->OutMcastFrames += (dp_intf_stats_p->OutIpMcastFrames + dp_intf_stats_p->OutIpBcastFrames);
        }
        else
        {
            dp_intf_stats_p = &(eth_internal_data_p->dp_pvt_p->bridge_intf[interfaceID]->intf_stats[i]);
            bridge_stats = (adk_ethernet_bridge_interface_statistics_t*)stats;

            bridge_stats->inUnknownVlan += dp_intf_stats_p->inUnknownVlan;
            bridge_stats->inEthTypeMismatch += dp_intf_stats_p->inEthTypeMismatch;

            bridge_stats->inFrames += (dp_intf_stats_p->inFrames +
                                       dp_intf_stats_p->inIpMcastFrames +
                                       dp_intf_stats_p->inIpBcastFrames +
                                       dp_intf_stats_p->inNonIpMcastFrames +
                                       dp_intf_stats_p->inNonIpBcastFrames);

            bridge_stats->inOctets += (dp_intf_stats_p->inOctets +
                                       dp_intf_stats_p->inIpMcastOctets +
                                       dp_intf_stats_p->inIpBcastOctets +
                                       dp_intf_stats_p->inNonIpMcastOctets +
                                       dp_intf_stats_p->inNonIpBcastOctets);

            bridge_stats->OutFrames+= (dp_intf_stats_p->OutFrames +
                                       dp_intf_stats_p->OutIpMcastFrames +
                                       dp_intf_stats_p->OutIpBcastFrames +
                                       dp_intf_stats_p->OutNonIpMcastFrames +
                                       dp_intf_stats_p->OutNonIpBcastFrames);

            bridge_stats->OutOctets += (dp_intf_stats_p->OutOctets +
                                        dp_intf_stats_p->OutIpMcastOctets +
                                        dp_intf_stats_p->OutIpBcastOctets +
                                        dp_intf_stats_p->OutNonIpMcastOctets +
                                        dp_intf_stats_p->OutNonIpBcastOctets);


            bridge_stats->inPrioTaggedFrames+= dp_intf_stats_p->inPrioTaggedFrames;
            bridge_stats->inMcastFrames += (dp_intf_stats_p->inIpMcastFrames + dp_intf_stats_p->inIpBcastFrames);
            bridge_stats->OutMcastFrames += (dp_intf_stats_p->OutIpMcastFrames + dp_intf_stats_p->OutIpBcastFrames);
        }
    }

    return (ADK_SUCCESS);
}

adk_status_t
adk_eth_dp_global_stats_get (
    adk_module_handle_t enetHandle,
    adk_ethernet_global_statistics_t *stats)
{
    uint32_t i = 0;
    adk_dp_eth_gl_stats_t *dp_gl_stats_p;
    adk_ethernet_internal_data_t *eth_internal_data_p;

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(enetHandle);

    ADK_MEMSET(stats, 0, sizeof(adk_ethernet_global_statistics_t));

    dp_gl_stats_p = eth_internal_data_p->dp_pvt_p->gl_stats_p;

    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        /* accumulate for all cores */
        stats->intf_not_defined += dp_gl_stats_p->intf_not_defined;
        stats->port_not_defined += dp_gl_stats_p->port_not_defined;
        stats->time_index_alloc_error += dp_gl_stats_p->time_index_alloc_error;
        stats->ts_tx_complete_notification_failed += dp_gl_stats_p->ts_tx_complete_notification_failed;    
        stats->pkt_clone_failure += dp_gl_stats_p->pkt_clone_failure;
        stats->ingressflow_lookup_failure += dp_gl_stats_p->ingressflow_lookup_failure;
        stats->ingressflowDrops += dp_gl_stats_p->ingressflowDrops;
        stats->ethPacketsRateLimitingDrops += dp_gl_stats_p->ethPacketsRateLimitingDrops;
        stats->arp_buffer_full_pkt_rollover_discard += dp_gl_stats_p->arp_buffer_full_pkt_rollover_discard;
        stats->arp_buffer_fragment_pkt_rollover_discard += dp_gl_stats_p->arp_buffer_fragment_pkt_rollover_discard;
        stats->nd_buffer_full_pkt_rollover_discard += dp_gl_stats_p->nd_buffer_full_pkt_rollover_discard;
        stats->nd_buffer_fragment_pkt_rollover_discard += dp_gl_stats_p->nd_buffer_fragment_pkt_rollover_discard;
        stats->exceptionDrops += dp_gl_stats_p->exceptionDrops;
        stats->arp_buffer_arp_entry_rem_drop += dp_gl_stats_p->arp_buffer_arp_entry_rem_drop;
        stats->nd_buffer_neigh_entry_rem_drop += dp_gl_stats_p->nd_buffer_neigh_entry_rem_drop;
        stats->neigh_state_unreachable_drop += dp_gl_stats_p->neigh_state_unreachable_drop;
        stats->non_l2term_frames_discard += dp_gl_stats_p->non_l2term_frames_discard;
        stats->ethtype_queue_map_not_found_drop += dp_gl_stats_p->ethtype_queue_map_not_found_drop;
        stats->netd_origin_non_bridge_frames_discard += dp_gl_stats_p->netd_origin_non_bridge_frames_discard;
        stats->unknown_output_interface_drop += dp_gl_stats_p->unknown_output_interface_drop;
        stats->buffering_memory_not_available_drop += dp_gl_stats_p->buffering_memory_not_available_drop;
        stats->nic_send_failure_drop += dp_gl_stats_p->nic_send_failure_drop;
        stats->nexthop_buffer_stage_not_found_drop += dp_gl_stats_p->nexthop_buffer_stage_not_found_drop;
        stats->nexthop_buffer_proc_drop += dp_gl_stats_p->nexthop_buffer_proc_drop;
        stats->nexthop_buffer_timeout_error += dp_gl_stats_p->nexthop_buffer_timeout_error;
        stats->failed_to_sent_next_stage_drop += dp_gl_stats_p->failed_to_sent_next_stage_drop;
        stats->vlan_header_stripping_failure_drop += dp_gl_stats_p->vlan_header_stripping_failure_drop;
        stats->egress_encap_pkt_failure_drop += dp_gl_stats_p->egress_encap_pkt_failure_drop;
        stats->unknown_multicast_addr_drop += dp_gl_stats_p->unknown_multicast_addr_drop;
        dp_gl_stats_p++;
    }

    return (ADK_SUCCESS);
}

/*!
 * =============================================================
 * @internal
 * adk_eth_if_info_se_set
 *
 * @param[in]   eth_module_handle   Ethernet module handle
 * @param[in]   ns_intf_flow_id     The flow ID assigned to physical or
 *                                  aggregate interface
 * @param[in]   src_mac             The source MAC to be set
 * @param[in]   encap               The encapsulation format to be set
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @brief  This API is used to store the data for a physical or aggregate
 *         interface in the SE memory.
 *
 *  This stores the source MAC address and encapsulation type, which are used in
 *  the data plane for appending Ethernet header for packets sent by the higher
 *  layer module to Ethernet module. These attributes are also used by all
 *  logical interfaces based upon this physical or aggregate interface.
 *
 * =============================================================
 */
adk_status_t
adk_eth_if_info_se_set (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    uint32_t __attribute__ ((unused)) ns_intf_flow_id,
    const adk_ethernet_mac_addr_t __attribute__ ((unused)) src_mac,
    adk_ethernet_encap_t __attribute__ ((unused)) encap)
{
    return (ADK_SUCCESS);
}


/*!
 * =============================================================
 * @internal
 * adk_eth_if_info_se_reset
 *
 * @param[in]   eth_module_handle   Ethernet module handle
 * @param[in]   ns_intf_flow_id     The flow ID assigned to physical or
 *                                  aggregate interface
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @brief  This API is used to clear the data for a physical or aggregate
 *         interface in the SE memory.
 *
 *  This clears the source MAC address and encapsulation type for the physical
 *  or aggregate interface.
 *
 * =============================================================
 */
adk_status_t
adk_eth_if_info_se_reset (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    uint32_t __attribute__ ((unused)) ns_intf_flow_id)
{
    return (ADK_SUCCESS);
}

/*!
 * =============================================================
 * @internal
 * adk_eth_clear_se_state
 *
 * @param[in]   eth_module_handle   Ethernet module handle
 * @param[in]   ns_id               The namespace ID for which the state
 *                                  needs to be cleared
 * @param[in]   size                The size of the SE memory for the
 *                                  Namespace ID to be cleared
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @brief  API to reset internal interface statistics maintained in SE memory
 *
 * =============================================================
 */
adk_status_t
adk_eth_clear_se_state (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    uint32_t __attribute__ ((unused)) ns_id,
    uint32_t __attribute__ ((unused)) size)
{
    return ADK_SUCCESS;
}

/*! end addtogroup _ethernet
 * @}
 */
