/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_list.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file has all the helper functions required in the
 *          Ethernet module for managing all pending lists. The lists
 *          are implemented as stacks using arrays.
 *
 *  **********************************************************************/

#include <adk_eth.h>

adk_status_t
adk_eth_list_create(
    uint32_t max_elem,
    adk_eth_list_t *list_p)
{
    if ((list_p == NULL) || (max_elem == 0))
    {
        return ADK_E_INVALID_PARAMETER;
    }

    list_p->elem_p = (uintptr_t *) ADK_NVM_CALLOC(ADK_ETHERNET_MODULE_ID, ADK_ETH_LIST_OBJ, max_elem, sizeof(uintptr_t));
    if (list_p->elem_p == NULL)
    {
        return ADK_E_MEM_UNAVAILABLE;
    }
    
    list_p->max_elements = max_elem;
    list_p->num_elements = 0;

    return ADK_SUCCESS;
}

adk_status_t
adk_eth_list_destroy(
    adk_eth_list_t *list_p)
{
    if ((list_p == NULL) || (list_p->elem_p == NULL))
    {
        return ADK_E_INVALID_PARAMETER;
    }

    /* List must be empty */
    if (list_p->num_elements)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ADK_NVM_FREE(list_p->elem_p);

    return ADK_SUCCESS;
}

void
adk_eth_list_add(
    adk_ethernet_internal_data_t *eth_internal_data_p,
    adk_eth_list_id_t list_id,
    uintptr_t data)
{
    adk_eth_list_t *list_p;

    list_p = &eth_internal_data_p->eth_list[list_id];
    if (list_p->num_elements == list_p->max_elements)
    {
        adk_rcu_distributed_wait();
        adk_eth_lists_clear(eth_internal_data_p);
    }
    list_p->elem_p[list_p->num_elements++] = data;

    return;
}

void
adk_eth_lists_clear(
    adk_ethernet_internal_data_t *eth_internal_data_p)
{
    adk_eth_list_t *list_p;

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_BRIDGE_ID_LIST_ID];
    while (list_p->num_elements)
    {
        adk_rm_resource_free(eth_internal_data_p->bridgeIf_id_ns_hdl,
                             (uint32_t) list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_SVIF_ID_LIST_ID];
    while (list_p->num_elements)
    {
        adk_rm_resource_free(eth_internal_data_p->svif_id_ns_hdl,
                             (uint32_t) list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_INGRESS_FLOW_LIST_ID];
    while (list_p->num_elements)
    {
        ADK_DP_SHM_FREE((adk_dp_eth_ingress_flow_t *)
                        list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_CPU_FLOW_LIST_ID];
    while (list_p->num_elements)
    {
        ADK_DP_SHM_FREE((adk_dp_eth_cpu_flow_result_t *)
                        list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_ING_FLOW_ID_LIST_ID];
    while (list_p->num_elements)
    {
        adk_rm_resource_free(eth_internal_data_p->eth_flows_ns_hdl,
                             (uint32_t) list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_BRIDGE_LIST_ID];
    while (list_p->num_elements)
    {
        ADK_DP_SHM_FREE((adk_dp_eth_bridge_intf_t *)
                        list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_SVIF_INTF_LIST_ID];
    while (list_p->num_elements)
    {
        ADK_DP_SHM_FREE((adk_dp_eth_svif_intf_t *)
                        list_p->elem_p[--list_p->num_elements]);
    }

    list_p = &eth_internal_data_p->eth_list[ADK_ETH_ARP_BUFFER_LIST_ID];
    while (list_p->num_elements)
    {
        ADK_DP_SHM_FREE((adk_dp_eth_next_hop_buffer_t *)
                        list_p->elem_p[--list_p->num_elements]);
    }
    list_p = &eth_internal_data_p->eth_list[ADK_ETH_TYPE_PCP_MAP_LIST_ID];
    while (list_p->num_elements)
    {
        ADK_DP_SHM_FREE((adk_ethernet_priority_t *)
                        list_p->elem_p[--list_p->num_elements]);
    }

    return;
}

adk_status_t
adk_eth_list_dump(
    adk_eth_list_t *list_p)
{
    uint32_t idx;

    if ((list_p == NULL) || (list_p->elem_p == NULL))
    {
        return ADK_E_INVALID_PARAMETER;
    }

    ADK_INFO_WRITE("List 0x%p of size %d has %d elements:\n",
                   list_p, list_p->max_elements, list_p->num_elements);

    for (idx = 0; idx < list_p->num_elements; idx++)
    {
        ADK_INFO_WRITE("\t0x%lx\n", list_p->elem_p[idx]);
    }

    return ADK_SUCCESS;
}

/*! end addtogroup _ethernet
 * @}
 */
