/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2017 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_eth_mac_table.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernet  Ethernet Module
 * @{
 *      @addtogroup _ethernetFunctions  Ethernet Functions
 *      @{
 *
 * @brief   This files has the source code for the MAC Table APIs
 *
 *  **********************************************************************/

#include <adk_eth.h>
#include <rte_hash_crc.h>
#include "adk_shm.h"

/*!
 * =============================================================
 * @internal
 * adk_eth_mac_fdb_dump
 *
 * @param[in]   hashTable        Linked list of MAC FDB
 * @param[in]   dump_options     MAC table dumption options
 *
 * @retval None
 *
 *  This function does a callback, if the entry is found in CP MAC FDB 
 *
 * =============================================================
 */
static void 
adk_eth_mac_fdb_dump(adk_eth_fdb_mac_table_entry_t *hashTable, adk_ethernet_mac_fdb_dump_cb_t *dump_options)
{
    uint32_t                    if_num=0;
    adk_eth_fdb_mac_table_entry_t *mac_entry = NULL;
    uint16_t if_id, bg_id;

    if((dump_options->dump_options.mask == ADK_ETH_MAC_TABLE_INTF_LEVEL) ||
       (dump_options->dump_options.mask == ADK_ETH_MAC_TABLE_INTF_BGID_LEVEL)) {
        /* Get the interface ID from the passed interface handle */
        if_num = ADK_ETH_INTERFACE_ID_GET(dump_options->dump_options.if_handle);
    }

    for (mac_entry = hashTable; mac_entry != NULL; mac_entry= mac_entry->next) {
        if_id = mac_entry->fwd_info.if_id;
        bg_id = mac_entry->fwd_info.bg_id;

        if(dump_options->dump_options.mask == ADK_ETH_MAC_TABLE_INTF_LEVEL) {
            if(if_num == if_id) {
                    dump_options->dumpCb(dump_options->rock, &mac_entry->fwd_info);
            }
        }
        else if(dump_options->dump_options.mask == ADK_ETH_MAC_TABLE_INTF_BGID_LEVEL) {
            if( (if_num == if_id) &&
                 (dump_options->dump_options.bg_id == bg_id) ) {
                if(mac_entry->is_alloted == TRUE)
                    dump_options->dumpCb(dump_options->rock, &mac_entry->fwd_info);

            }
        }
        else if(dump_options->dump_options.mask == 0) {
            if(mac_entry->is_alloted == TRUE)
                    dump_options->dumpCb(dump_options->rock, &mac_entry->fwd_info);
        }

    }
}

/*!
 * =============================================================
 *
 * adk_ethernet_mac_table_dump
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 * @param[in]   dump_options    Dump Options  
 * @param[in]   mac_dump_cb     Call Back Function 
 * @param[in]   handle          Handle to be passed along with call back fn
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 * @retval  ADK_E_MEM_UNAVAILABLE   Memory allocation failed
 * @retval  ADK_E_SE_STATE_GET_FAILED SE Memory read failed
 *
 * @brief   This function is used to dump MAC Table 
 *
 * This function is used to dump MAC Table from MPP and PCX
 *
 * =============================================================
 */
adk_status_t adk_ethernet_mac_table_dump(adk_module_handle_t         eth_mod_handle,
                                         adk_eth_mac_table_options_t dump_options,
                                         adk_eth_mac_dump_cb         mac_dump_cb,
                                         void                        *handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t  sys_handle;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_dp_pvt_data_t *eth_dp_pvt_p;
    adk_ethernet_mac_fdb_dump_cb_t   mac_fdb_dump_cb_data;
    bool moduleLockTaken = false;
    uint32_t i=0;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n", 
                          eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);
    moduleLockTaken = true;

    /* Get the module info pointer from the handle */
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);
    eth_dp_pvt_p = eth_internal_data_p->dp_pvt_p;

    mac_fdb_dump_cb_data.rock = handle;
    mac_fdb_dump_cb_data.dumpCb = mac_dump_cb;

    mac_fdb_dump_cb_data.dump_options.mask = dump_options.mask;
    mac_fdb_dump_cb_data.dump_options.if_handle = dump_options.if_handle;
    mac_fdb_dump_cb_data.dump_options.bg_id = dump_options.bg_id;

    if((dump_options.mask == ADK_ETH_MAC_TABLE_INTF_LEVEL) ||
       (dump_options.mask == ADK_ETH_MAC_TABLE_INTF_BGID_LEVEL)) {
        if (!adk_eth_valid_if_handle(dump_options.if_handle)) {
            ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
             ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Interface handle %"ADKPRIh "\n", dump_options.if_handle);
             goto _adk_ethernet_mac_table_dump_exit;
        }
    }

    if(dump_options.mask == ADK_ETH_MAC_TABLE_INTF_BGID_LEVEL) {
        if (ADK_ETH_IS_BG_ID_INVALID(dump_options.bg_id)) {
            ret = ADK_ETHERNET_E_INVALID_BG_ID;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid BG_ID 0x%04x\n", dump_options.bg_id);
        }
    }

    ADK_ACQUIRE_LOCK(eth_internal_data_p->macfdb_lock, ret);

    for(i=0; i<ADK_ETH_MAC_FDB_HASH_MAX; i++) {
        adk_eth_mac_fdb_dump(eth_dp_pvt_p->hashIndex[i], &mac_fdb_dump_cb_data);
    }

    ADK_RELEASE_LOCK(eth_internal_data_p->macfdb_lock, ret);

_adk_ethernet_mac_table_dump_exit:

    moduleLockTaken = false;
    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    if (moduleLockTaken == true)
    {
        ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
    }

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!     end addtogroup _ethernetFunction
 *      @}
 *  end addtogroup _ethernet
 * @}
 */
