/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ethernet.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernet  Ethernet Module
 * @{
 *      @addtogroup _ethernetFunctions  Ethernet Functions
 *      @{
 *
 * @brief   This files has the source code for the top module level 
 *          APIs that manage properties of Ethernet module as a whole.
 *          This includes all the initialization and shutdown routines 
 *          as well as service functions.
 *
 *  **********************************************************************/

#include <adk_eth.h>
#include "adk_shm.h"
#include "adk_ph_dev_dp.h"
#include "adk_ph_dev.h"

uint8_t adk_eth_quad_group_port_map[ADK_ETHERNET_QUAD_GROUP_MAX][ADK_ETH_MAX_NUM_PORT_PER_QUAD] =
               {{0,1,2,3},{4,5,6,7},{8,9,10,11},{12,13,14,15},{16,17,18,19}};

/*!
 * =============================================================
 *
 * adk_ethernet_init
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_E_MEM_UNAVAILABLE   Operation failed due to insufficient memory
 *
 * @brief       Function to initialise Ethernet module
 *
 * This function fills up the internal elements and data structures of the
 * Ethernet module. This includes allocating namespace resources, getting
 * tree IDs, allocating private data structures and registering service
 * functions.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called indirectly via adk_module_start() API. For that to
 * happen application must pass adk_ethernet_init() function as the parameter
 * to the adk_module_register() API at the time of Ethernet module registration.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_init (
    adk_module_handle_t eth_mod_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_module_service_fn_t service_funcs;
    adk_log_level_t sys_log_level;
    adk_ethernet_log_level_t eth_log_level;
    uint16_t i;
    adk_sys_handle_t sys_handle;
    adk_eth_dp_pvt_data_t *dp_pvt_p = NULL;

    /* Validate the passed handles */
    if (!eth_mod_handle)
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Ethernet module failed to initialize\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    /* Check if the handle has already been initialized */
    if (adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE(
            "Ethernet Module handle %"ADKPRIh " is already initialized\n",
            eth_mod_handle);
        return (ADK_SUCCESS);
    }

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    sys_log_level = ADK_MODULE_LOG_LEVEL_GET(eth_mod_handle);
    ret = adk_eth_map_sys_log(eth_mod_handle, sys_log_level, &eth_log_level);
    if (ret != ADK_SUCCESS)
        goto _adk_ethernet_init_exit;

    ADK_MODULE_LOG_LEVEL_SET(eth_mod_handle, eth_log_level);

    ADK_ETH_LOG_INFO(eth_mod_handle,
                     "Initializing at log-level 0x%04x\n",
                     eth_log_level);

    /* Create the internal data structure and store pointer to it */
    eth_internal_data_p =
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(eth_mod_handle), ADK_ETH_INTF_OBJ, 1, sizeof(adk_ethernet_internal_data_t));
    if (eth_internal_data_p == NULL)
    {
        ADK_ETH_LOG_FATAL(eth_mod_handle, ADK_E_MEM_UNAVAILABLE,
                          "Memory allocation failed\n");
        ret = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ethernet_init_exit;
    }

    ADK_MODULE_INFO_PTR_SET(eth_mod_handle, eth_internal_data_p);

    /* Create memzone to store the structure which will have pointers to all
     * the data plane tables */
    ret = adk_shm_create(ADK_ETHERNET_SHM_MEMZONE_NAME,
                        sizeof(adk_eth_dp_pvt_data_t),
                        ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                        (adk_shm_handle_t **)&dp_pvt_p);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_FATAL(eth_mod_handle, ADK_E_MEM_UNAVAILABLE,
                          "Failed to create Ethernet memzone %s\n",
                          ADK_ETHERNET_SHM_MEMZONE_NAME);
        ret = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ethernet_init_exit;
    }

    ret = adk_eth_dp_table_init(eth_mod_handle, dp_pvt_p);
    if (ret)
    {
        ADK_ETH_LOG_FATAL(eth_mod_handle, ret,
                          "Failed to initialize Ethernet data plane tables!\n");
        goto _adk_ethernet_init_exit;
    }

    /*
     * Create user defined namespaces ID for flows and interfaces
     */

    eth_internal_data_p->eth_flows_ns_hdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(eth_mod_handle), ADK_ETH_INGRESS_RES_OBJ,
            ADK_ETH_INGRESS_FLOWS_MAX);
    if (!eth_internal_data_p->eth_flows_ns_hdl)
    {
        ret = ADK_E_NS_RSRC_CREATE_FAILED;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to create %s namespace handle\n",
                "ethernet_flows");
        goto _adk_ethernet_init_exit;
    }

    eth_internal_data_p->svif_id_ns_hdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(eth_mod_handle), ADK_ETH_SVIF_IF_RES_OBJ,
            ADK_ETH_IF_SVIF_IF_MAX);
    if (!eth_internal_data_p->svif_id_ns_hdl)
    {
        ret = ADK_E_NS_RSRC_CREATE_FAILED;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to create %s namespace handle\n",
                "svif_id");
        goto _adk_ethernet_init_exit;
    }

    eth_internal_data_p->bridgeIf_id_ns_hdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(eth_mod_handle), ADK_ETH_BRIDGE_IF_RES_OBJ,
            ADK_ETH_NUM_BRIDGE);
    if (!eth_internal_data_p->bridgeIf_id_ns_hdl)
    {
        ret = ADK_E_NS_RSRC_CREATE_FAILED;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to create %s namespace handle\n",
                "bridge_id");
        goto _adk_ethernet_init_exit;
    }

    for (i = 0; i < ADK_ETH_MAX_LIST_ID; i++)
    {
        ret = adk_eth_list_create(ADK_ETH_LIST_MAX_SIZE,
                                  &eth_internal_data_p->eth_list[i]);
        if (ret != ADK_SUCCESS)
        {
            ADK_ETH_LOG_FATAL(eth_mod_handle, ret,
                              "Failed to create Ethernet list-%d!\n", i);
            goto _adk_ethernet_init_exit;
        }
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Ethernet list-%d %p created\n", i,
                         &eth_internal_data_p->eth_list[i]);
    }

    /* Intialize time slot index management data */
    eth_internal_data_p->adk_eth_time_slot_data.eioa0_slot1_namespace_id = 0x0;
    eth_internal_data_p->adk_eth_time_slot_data.eioa1_slot1_namespace_id = 0x0;
    eth_internal_data_p->adk_eth_time_slot_data.allocation_bit_map = 0x0;

    /* Initialize internal data */
    ADK_MEMSET(&service_funcs, 0x0, sizeof(adk_module_service_fn_t));

    service_funcs.moduleTerminate = adk_ethernet_shutdown;
    service_funcs.moduleBind = (module_bind_fn_t) adk_ethernet_custom_bind;
    service_funcs.moduleUnbind =
        (module_unbind_fn_t) adk_ethernet_custom_unbind;
    service_funcs.moduleGetImc = adk_ethernet_get_imc;
    service_funcs.moduleGetEntryFunc = adk_ethernet_get_entry_func;
    /* Ethernet is not interested in MAC events */
    service_funcs.moduleLogSet = adk_ethernet_sys_log;
    service_funcs.errorString = adk_eth_error_name;

    service_funcs.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_ETH_ULI_FN_ID;
    service_funcs.cpuEntryFn_ULI[_DEST_FUNC_RING] = ADK_ETH_ULI_RING_FN_ID;

    ret = adk_module_service_function_register(sys_handle, eth_mod_handle,
                                               &service_funcs);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to register Ethernet module's service"
                          " functions\n");
        goto _adk_ethernet_init_exit;
    }

    /* Hook up with Ethernet data plane tables to the module handle structure.
     * All the control plane access to the DP tables will be done via module
     * handle in the CP APIs. */
    eth_internal_data_p->dp_pvt_p = dp_pvt_p;

    eth_internal_data_p->macfdb_lock = (void *) ADK_ALLOC_LOCK();
    ADK_INIT_LOCK(eth_internal_data_p->macfdb_lock, TRUE, 1);

    eth_internal_data_p->cookie = ADK_ETHERNET_MODULE_COOKIE;


    ret = adk_sif_module_egress_callback_register(sys_handle,
            ADK_ETHERNET_MODULE_ID,
            (adk_sif_module_egress_cb_fn_t)adk_eth_dp_egress_stats);
    if(ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ADK_E_INTERNAL_ERROR,
                "Failed to register DP callback function for Ethernet Egress stats\n");
        ret = ADK_E_INTERNAL_ERROR;
        goto _adk_ethernet_init_exit;
    }


    /* Register callback function for policer drop in send to host classifier action*/
    ret = adk_sif_module_slowpath_callback_register(sys_handle,
            ADK_ETHERNET_MODULE_ID,
            (adk_sif_module_slowpath_cb_fn_t)adk_eth_dp_slowpath_policer_stats);
    if(ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ADK_E_INTERNAL_ERROR,
                "Failed to register DP callback function for slowpath policer stats \n");
        ret = ADK_E_INTERNAL_ERROR;
        goto _adk_ethernet_init_exit;
    }
_adk_ethernet_init_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Ethernet module %"ADKPRIh " initialized\n",
                         eth_mod_handle);
    }
    else
    {
        if (eth_internal_data_p)
        {
            if (eth_internal_data_p->macfdb_lock)
            {
                ADK_FREE_LOCK(eth_internal_data_p->macfdb_lock);
            }
            for (i = 0; i < ADK_ETH_MAX_LIST_ID; i++)
            {
                ret = adk_eth_list_destroy(&eth_internal_data_p->eth_list[i]);
                if (ret != ADK_SUCCESS)
                {
                    ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Failed to free"
                                      " Ethernet list-%d\n", i);
                }
            }
            if (dp_pvt_p)
            {
                ret = adk_eth_dp_table_shutdown(eth_mod_handle, dp_pvt_p);
                if (ret != ADK_SUCCESS)
                {
                    ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Failed to free"
                                      " memory allocated for DP tables\n");
                }
                ret = adk_shm_destroy(ADK_ETHERNET_SHM_MEMZONE_NAME);
                if (ret != ADK_SUCCESS)
                {
                    ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Failed to destroy"
                                      " Ethernet shared memory zone!\n");
                }
            }
            ADK_NVM_FREE(eth_internal_data_p);
        }
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Ethernet module %"ADKPRIh " failed to initialize\n",
                          eth_mod_handle);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_service_function_update
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 *
 * @brief       Function to register the CPU based service fns
 *
 * This function registers the CPU based service functions with
 * the SIF.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called from adk_module_handle_get() API.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_service_function_update (
    adk_module_handle_t eth_mod_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_module_service_fn_t service_funcs;
    adk_sys_handle_t sys_handle;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Ethernet module failed to register"
                          " service fns\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    /* Get the module info pointer from the handle */
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    /* Initialize internal data */
    ADK_MEMSET(&service_funcs, 0x0, sizeof(adk_module_service_fn_t));

    service_funcs.moduleTerminate = adk_ethernet_shutdown;
    service_funcs.moduleBind = (module_bind_fn_t) adk_ethernet_custom_bind;
    service_funcs.moduleUnbind =
        (module_unbind_fn_t) adk_ethernet_custom_unbind;
    service_funcs.moduleGetEntryFunc = adk_ethernet_get_entry_func;
    service_funcs.moduleLogSet = adk_ethernet_sys_log;
    service_funcs.errorString = adk_eth_error_name;

    /* return the function pointer stored in module info during ethernet_init */
    service_funcs.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_ETH_ULI_FN_ID;
    service_funcs.cpuEntryFn_ULI[_DEST_FUNC_RING] = ADK_ETH_ULI_RING_FN_ID;

    /*Register ethernet parser function in PH to Decode exception params*/
    ret = adk_ph_dev_rx_parse_register((void*)adk_ethernet_ph_rx_parse,
                                    ADK_ETHERNET_MODULE_ID);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_WARNING(eth_mod_handle,
                "Failed to register the ph_dev_rx_parse function for ethernet in CP\n");
    }

    ret = adk_module_service_function_update(
        sys_handle,
        eth_mod_handle,
        &service_funcs);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to register Ethernet module's service"
                          " functions\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_shutdown
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 *
 * @brief   Function to shutdown Ethernet module
 *
 * This function is used to shutdown Ethernet module by freeing up all its
 * internal data structures.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called indirectly and automatically via adk_module_stop() API.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_shutdown (
    adk_module_handle_t eth_mod_handle)
{
    adk_log_level_t sys_log_level;
    adk_ethernet_internal_data_t *eth_internal_data_p = NULL;
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    uint16_t i;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Ethernet module failed to shutdown\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    /* Get the module info pointer from the handle */
    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    adk_rcu_distributed_wait();
    adk_eth_lists_clear(eth_internal_data_p);
    ADK_ETH_LOG_INFO(eth_mod_handle,
                     "All ethernet lists cleared\n");


    for (i = 0; i < ADK_ETH_MAX_LIST_ID; i++)
    {
        ret = adk_eth_list_destroy(&eth_internal_data_p->eth_list[i]);
        if (ret != ADK_SUCCESS)
        {
            ADK_ETH_LOG_FATAL(eth_mod_handle, ret,
                              "Failed to destroy Ethernet list-%d!\n", i);
            goto _adk_ethernet_shutdown_exit;
        }
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Ethernet list-%d %p destroyed\n", i,
                         &eth_internal_data_p->eth_list[i]);
    }

    /* Destroy all the resource handles */
    adk_rm_resource_destroy(eth_internal_data_p->eth_flows_ns_hdl);
    adk_rm_resource_destroy(eth_internal_data_p->svif_id_ns_hdl);
    adk_rm_resource_destroy(eth_internal_data_p->bridgeIf_id_ns_hdl);
          
    /* Reset the data plane tables */
    eth_internal_data_p->dp_pvt_p->cookie = 0;

    adk_rcu_synchronize();

    ADK_FREE_LOCK(eth_internal_data_p->macfdb_lock);

    ret = adk_eth_dp_table_shutdown(eth_mod_handle,
                                    eth_internal_data_p->dp_pvt_p);
    if (ret)
    {
        ADK_ETH_LOG_FATAL(eth_mod_handle, ret,
                          "Failed to free Ethernet data plane tables!\n");
        goto _adk_ethernet_shutdown_exit;
    }
    
    eth_internal_data_p->dp_pvt_p = 0;

    /* Free up the Ethernet specific memzone */
    ret = adk_shm_destroy(ADK_ETHERNET_SHM_MEMZONE_NAME);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_WARNING(eth_mod_handle,
                            "Failed to destroy Ethernet shared memory zone %s\n",
                            ADK_ETHERNET_SHM_MEMZONE_NAME);
    }

    /*Unregister ethernet parser function in PH to Decode exception params*/
    ret = adk_ph_dev_rx_parse_unregister(ADK_ETHERNET_MODULE_ID);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_WARNING(eth_mod_handle,
                "Failed to unregister the ph_dev_rx_parse function for ethernet\n");
    }

    /*Unregister ethernet egress callback function*/
    ret = adk_sif_module_egress_callback_unregister(sys_handle,
            ADK_ETHERNET_MODULE_ID);
    if(ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_FATAL(eth_mod_handle, ret,
                "Failed to unregister ethernet egress callback function!\n");
        goto _adk_ethernet_shutdown_exit;
    }

    /* Finally, change the log level as per SIF before finishing */
    sys_log_level = adk_eth_map_eth_log(eth_mod_handle);
    ADK_MODULE_LOG_LEVEL_SET(eth_mod_handle, sys_log_level);

_adk_ethernet_shutdown_exit:

    if (ret == ADK_SUCCESS)
    {
        eth_internal_data_p->cookie = 0;
        ADK_NVM_FREE(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Ethernet module %"ADKPRIh " shutdown\n",
                         eth_mod_handle);
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Ethernet module %"ADKPRIh " failed to shutdown\n",
                          eth_mod_handle);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_sys_log
 *
 * @param[in]  eth_mod_handle   Ethernet Module Handle
 * @param[in]  log_level        SIF defined log level
 *
 * @retval ADK_SUCCESS          Successful completion
 * @retval ADK_E_INVALID_HANDLE Invalid Ethernet module handle parameter
 *
 * @brief  Function to set the log-level for Ethernet module through service
 *         funtion registered with SIF.
 *
 * This function is used to set the log-level for Ethernet module via SIF using
 * adk_module_attr_set() API. This function must not be called directly.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_sys_log (
    adk_module_handle_t eth_mod_handle,
    adk_log_level_t sys_log_level)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_log_level_t eth_log_level;
    adk_sys_handle_t sys_handle;

    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to set log-level 0x%x\n",
                          sys_log_level);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    ret = adk_eth_map_sys_log(eth_mod_handle, sys_log_level, &eth_log_level);
    if (ret != ADK_SUCCESS)
        goto _adk_ethernet_sys_log_exit;

    ADK_MODULE_LOG_LEVEL_SET(eth_mod_handle, eth_log_level);

_adk_ethernet_sys_log_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Ethernet module %"ADKPRIh " log-level set to 0x%x\n",
                         eth_mod_handle, eth_log_level);
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to set log-level %d\n", sys_log_level);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_log_level_set
 *
 * @param[in]  eth_mod_handle   Ethernet Module Handle
 * @param[in]  log_level        Logical OR of different log types
 *
 * @retval ADK_SUCCESS          Successful completion
 * @retval ADK_E_INVALID_HANDLE Invalid Ethernet module handle parameter
 *
 * @brief  Function to set the log-level for Ethernet module
 *
 * This function is used to set the log-level for Ethernet module. This
 * completely overwrites the existing log-level setting.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_log_level_set (
    adk_module_handle_t eth_mod_handle,
    adk_ethernet_log_level_t log_level)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to set Ethernet log-level 0x%x\n",
                          log_level);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    ADK_MODULE_LOG_LEVEL_SET(eth_mod_handle, log_level);

    ADK_ETH_LOG_INFO(eth_mod_handle,
                     "Ethernet module %"ADKPRIh " log-level set to 0x%x\n",
                     eth_mod_handle, log_level);

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_log_enable
 *
 * @param[in]  eth_mod_handle   Ethernet Module Handle
 * @param[in]  log_level        Logical OR of different log types to be enabled
 *
 * @retval ADK_SUCCESS          Successful completion
 * @retval ADK_E_INVALID_HANDLE Invalid Ethernet module handle parameter
 *
 * @brief  Function to enable one or more log types
 *
 * This function is used to enable one or more log type without affecting the
 * other log-types which are already enabled or disabled.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_log_enable (
    adk_module_handle_t eth_mod_handle,
    adk_ethernet_log_level_t log_level)
{
    adk_ethernet_log_level_t curr_log_level;
    adk_sys_handle_t sys_handle;

    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to enable log 0x%x\n", log_level);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    curr_log_level = ADK_MODULE_LOG_LEVEL_GET(eth_mod_handle);
    curr_log_level |= log_level;

    ADK_MODULE_LOG_LEVEL_SET(eth_mod_handle, curr_log_level);

    ADK_ETH_LOG_INFO(eth_mod_handle,
                     "Enabled log-level 0x%x for Ethernet module %"ADKPRIh "\n",
                     log_level, eth_mod_handle);

    return (ADK_SUCCESS);
}

/*!
 * =============================================================
 *
 * adk_ethernet_log_disable
 *
 * @param[in]  eth_mod_handle   Ethernet Module Handle
 * @param[in]  log_level        Logical OR of different log types to be disabled
 *
 * @retval ADK_SUCCESS          Successful completion
 * @retval ADK_E_INVALID_HANDLE Invalid Ethernet module handle parameter
 *
 * @brief  Function to disable one or more log types
 *
 * This function is used to disable one or more log type without affecting the
 * other log-types which are already enabled or disabled.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_log_disable (
    adk_module_handle_t eth_mod_handle,
    adk_ethernet_log_level_t log_level)
{
    adk_ethernet_log_level_t curr_log_level;
    adk_sys_handle_t sys_handle;

    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to disable log 0x%x\n", log_level);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    curr_log_level = ADK_MODULE_LOG_LEVEL_GET(eth_mod_handle);
    curr_log_level &= ~log_level;

    ADK_MODULE_LOG_LEVEL_SET(eth_mod_handle, curr_log_level);

    ADK_ETH_LOG_INFO(eth_mod_handle,
                     "Disabled log-level 0x%x for Ethernet module %"ADKPRIh "\n",
                     log_level, eth_mod_handle);

    return (ADK_SUCCESS);
}

adk_status_t adk_ethernet_rsrc_notify_fn_cbfn(adk_module_handle_t mod_handle, 
                                         adk_sys_rsrc_oper_t oper,
                                         adk_sys_resource_type_t type,
                                         adk_sys_resource_info_t rsrc)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(mod_handle);

    RTE_SET_USED(mod_handle);
    RTE_SET_USED(oper);
    RTE_SET_USED(type);
    RTE_SET_USED(rsrc);
    ADK_ETH_LOG_INFO(mod_handle, "Ethernet module %"ADKPRIh "\n",
                     mod_handle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 *
 * adk_ethernet_arp_entry_add
 *
 * @param[in]   eth_mod_handle      Handle to the Ethernet module
 * @param[in]   ip_addr             The next hop IP address
 * @param[in]   dest_mac            The MAC address of the next hop IP address
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_E_INTERNAL_ERROR    Internal error while adding arp entry
 *
 * @brief  API to add an entry in the ARP table. 
 *
 * The API takes the next hop IP address and maps it to the destination 
 * MAC address.If entry already exists then it would be overwritten. 
 *
 * =============================================================
 */
adk_status_t adk_ethernet_arp_entry_add(
                    adk_module_handle_t eth_mod_handle,
                    adk_ethernet_ipv4_addr_t ip_addr,
                    adk_ethernet_mac_addr_t dest_mac,
                    adk_ethernet_arp_status_t arp_status)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to add arp entry\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if(arp_status >= ADK_ETHERNET_NEIGHBOR_MAX)
    {
        dp_ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret, "Invalid arp status value\n");
        return dp_ret;
    }

    dp_ret = adk_eth_dp_arp(eth_mod_handle,
                                     ADK_ETH_TREE_OPER_ADD,
                                     ip_addr, dest_mac, arp_status);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to add entry to arp table.\n");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Add arp entry successful for ip_addr %d \n",
                         ip_addr);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (dp_ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_arp_entry_rem
 *
 * @param[in]   eth_mod_handle      Handle to the Ethernet module
 * @param[in]   ip_addr             The next hop IP address
 *
 * @retval  ADK_SUCCESS                        Successful completion
 * @retval  ADK_E_INVALID_HANDLE               Invalid Ethernet module handle specified
 * @retval  ADK_E_INVALID_PARAMETER            Arp entry not found
 *
 * @retval ADK_SUCCESS
 *
 * @brief  API to remove an entry from the ARP table. 
 *
 * The API takes the next hop IP address and removes the MAC 
 * address mapping for it.
 *
 * =============================================================
 */
adk_status_t adk_ethernet_arp_entry_rem(
                    adk_module_handle_t eth_mod_handle,
                    adk_ethernet_ipv4_addr_t ip_addr)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_ethernet_mac_addr_t dest_mac;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to delete arp entry\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    dp_ret = adk_eth_dp_arp(eth_mod_handle,
                                    ADK_ETH_TREE_OPER_DEL,
                                    ip_addr, dest_mac /* NOT USED */,
                                    ADK_ETHERNET_NEIGHBOR_REACHABLE /* NOT USED */);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                            "Failed to delete entry from arp table.\n");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                            "Delete arp entry successful for ip_addr %d \n",
                            ip_addr);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, dp_ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

   return (dp_ret);
}

/* =============================================================
*
* adk_ethernet_get_arp_entries_all
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
* @param[out]  numArpEntries       Number of ARP entries
* @param[out]  arp_entries_p       List of ARP entries
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INVALID_PARAMETER Invalid input pointers specfied
* @retval  ADK_E_INTERNAL_ERROR    Failed to list the ARP entries from Datapath
* @retval  ADK_E_MEM_UNAVAILABLE   Failed to allocate memory to store
*                                  ARP entries
*
* @brief  API to get all the entries in the ARP table.
*
* @details
* The API returns the list of ARP entries present in data path
* along with the number of entries.
* The application is expected to call rte_free for the memory allocated by
* the API in output parameter arp_entries_p, when the API returns
* non-zero number of entries.
*
* =============================================================
*/

adk_status_t adk_ethernet_get_arp_entries_all(
                    adk_module_handle_t eth_mod_handle,
                    uint32_t *numArpEntries,
                    adk_ethernet_arp_entry_t **arp_entries_p)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get arp entries\n");
        return (ADK_E_INVALID_HANDLE);
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if((NULL == numArpEntries) || (NULL == arp_entries_p))
    {
        dp_ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret, "Invalid Input Parameters\n");
        return dp_ret;
    }

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }


    dp_ret = adk_eth_dp_arp_entries_get(&eth_internal_data_p->dp_pvt_p->eth_arp_hash_table,
                                                    numArpEntries, arp_entries_p);
    if (dp_ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret,
                          "Failed to get arp enties from datapath");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Get arp entries successful from datapath\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (dp_ret);
}
/*!
 * =============================================================
 *
 * adk_ethernet_arp_entry_get
 *
 * @param[in]   eth_mod_handle      Handle to the Ethernet module
 * @param[in]   ip_addr             The next hop IP address
 * @param[out]  dest_mac            The MAC address of the next hop IP address
 * @param[out]  arp_status          Arp entry reachable status
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_E_INVALID_PARAMETER Invalid parameter
 *
 * @brief  API to get an entry from the ARP table. 
 *
 * The API takes the next hop IP address and gets the destination 
 * MAC address and arp reachability status. 
 *
 * =============================================================
 */
adk_status_t adk_ethernet_arp_entry_get(
                    adk_module_handle_t eth_mod_handle,
                    adk_ethernet_ipv4_addr_t ip_addr,
                    adk_ethernet_mac_addr_t dest_mac,
                    adk_ethernet_arp_status_t* arp_status)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to add arp entry\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if(NULL == arp_status)
    {
        dp_ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret, "Invalid arp status pointer\n");
        return dp_ret;
    }

    dp_ret = adk_eth_dp_arp_get(&eth_internal_data_p->dp_pvt_p->eth_arp_hash_table,
                                            ip_addr, dest_mac, arp_status);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get arp entry for ip_addr %d.\n", ip_addr);
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Get arp entry successful for ip_addr %d \n",
                         ip_addr);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (dp_ret);
}

#ifdef ADK_DEBUG

/*!
 * =============================================================
 *
 * adk_ethernet_arp_buffer_nexthop_debug 
 *
 * @param[in]   eth_mod_handle           Handle to the Ethernet module
 * @param[in]   ip_addr                  The next hop IP address
 * @param[in]   num_inflight_pkts        The number of in flight packets for this next hop
 * @param[in]   num_full_buffered_pkts   The number of full packets buffered for this next hop
 * @param[in]   num_frag_buffered_pkts   The number of fragmented packets buffered for this next hop
 *
 * @retval  ADK_SUCCESS                        Successful completion
 * @retval  ADK_E_INVALID_HANDLE               Invalid Ethernet module handle specified
 * @retval  ADK_E_INVALID_PARAMETER            Invalid parameter
 *
 * @retval ADK_SUCCESS
 *
 * @brief  Debug API to figure out buffer packet statistics per next hop
 *
 * The API can be used to figure out inflight count for a next hop. Inflight
 * count says how many packets are been sent to atomic buffer stage for this
 * next hop. Additionally it can be used to figure out full or fragmented
 * packets currently buffered for this next hop 
 *
 * =============================================================
 */
adk_status_t adk_ethernet_arp_buffer_nexthop_debug(
                    adk_module_handle_t eth_mod_handle,
                    adk_ethernet_ipv4_addr_t ip_addr,
                    uint32_t* num_inflight_pkts,
                    uint32_t* num_full_buffered_pkts,
                    uint32_t* num_frag_buffered_pkts)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get arp buffer debug counters for a next hop\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if(NULL == num_inflight_pkts || NULL == num_full_buffered_pkts
        || NULL == num_frag_buffered_pkts)
    {
        dp_ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret, "Invalid parameter\n");
        return dp_ret;
    }

    *num_inflight_pkts = *num_full_buffered_pkts = *num_frag_buffered_pkts = 0;

    dp_ret = adk_eth_dp_arp_buffer_statistics_get(&eth_internal_data_p->dp_pvt_p->eth_arp_hash_table,
                                     ip_addr, num_inflight_pkts, num_full_buffered_pkts,
                                        num_frag_buffered_pkts);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get arp buffer staticstics for ip_addr %x.\n", ip_addr);
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Arp buffer statistics get successful for ip_addr %x \n",
                         ip_addr);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (dp_ret);
}

#endif  /*ADK_DEBUG*/

/*!
 * =============================================================
 *
 * adk_ethernet_priority_tagged_encap_config
 *
 * @param[in]   eth_mod_handle          Handle to the Ethernet module
 * @param[in]   if_handle               Bridge interface handle
 * @param[in]   priority_tagging_enable Flag to enable/disable priority tagging
 *
 * @retval  ADK_SUCCESS                      Successful completion
 * @retval  ADK_E_INVALID_HANDLE             Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE Invalid Interface Handle specified
 * @retval  ADK_ETHERNET_E_INCOMPAT_IF_TYPE  Interface specified is not a bridge Interface Handle
 *
 * @brief   API to enable/disable priority tagging on the specified interface
 *
 * This API enables or disables priority tagging on the specified interface.
 * This configuration is allowed only on bridge interface. 
 * If priority tagging is enabled on bridge interface then priority
 * tagged packet goes out else untagged packet goes out.
 * =============================================================
 */
adk_status_t  adk_ethernet_priority_tagged_encap_config(
        adk_module_handle_t eth_mod_handle,
        adk_interface_handle_t if_handle,
        bool priority_tagging_enable)
{
    adk_status_t                    ret = ADK_SUCCESS;
    adk_sys_handle_t                sys_handle;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_intf_internal_data_t    *if_data_p = NULL;
    uint32_t                        statsId = 0;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to create sub interface\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);
    
    ADK_ETH_LOG_INFO( eth_mod_handle,"Input Parameters:  Interface Handle %p, Priority Tagging 0x%x",
                        (adk_interface_handle_t *) if_handle, priority_tagging_enable);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (!adk_eth_valid_if_handle(if_handle))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Interface Handle %p\n",(adk_interface_handle_t *) if_handle);
        goto _adk_ethernet_priority_tagged_encap_config;
    }

    if_data_p = (adk_eth_intf_internal_data_t *) 
        ADK_SYS_INTERFACE_ATTR_PTR_GET(if_handle);

    /* Check the interface type */
    if (if_data_p->type != INTERFACE_TYPE_BRIDGE)
    {
        ret = ADK_ETHERNET_E_INCOMPAT_IF_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Incompatible Interface type %s\n",
                adk_eth_intf_type_name(if_data_p->type));
        goto _adk_ethernet_priority_tagged_encap_config;
    }

    if_data_p->u.bridge_if.priority_tag = priority_tagging_enable;

    statsId = eth_internal_data_p->dp_pvt_p->bridge_map_id[if_data_p->u.bridge_if.l2_domain_id];
    eth_internal_data_p->dp_pvt_p->bridge_intf[statsId]->priority_tag = priority_tagging_enable;

    ret = adk_sys_flow_cache_update_version(sys_handle);
    if(ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to update global flow cache version");
        goto _adk_ethernet_priority_tagged_encap_config;
    }

_adk_ethernet_priority_tagged_encap_config:

    if (ret == ADK_SUCCESS) {
        ADK_ETH_LOG_INFO(
                eth_mod_handle, 
                "Interface Handle %p| Priority tagging Status %d\n",
               (adk_interface_handle_t *) if_handle, priority_tagging_enable);
    } else {
        ADK_ETH_LOG_ERROR(
                eth_mod_handle, ret,
                "Failed to set priority tag encapsulation config \n");
    }

    ADK_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!
 * =============================================================
 * adk_ethernet_register_exception
 * @param[in]  eth_mod_handle ETH module handle.
 *
 * @param[in]   exceptionHandler_p  Pointer to adk_ethernet_exception_handler_t
 *
 * @retval   ADK_SUCCESS - successfully registered exception
 * handler
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @retval  ADK_E_INVALID_HANDLE - Invalid ethernet handle specified 
 *
 * @retval ADK_IP_E_INVALID_EXCEPTION_TYPE  - Invalid exception type.
 *
 * @brief
 *
 * This function allows the user to register exception handlers
 * For various Ethernet exceptions. enum adk_ethernet_exception_type_t defines
 * The type of exceptions that are supported by ADK-ETH. Application
 * is allowed to provide an CPU callback functionas
 * Exception handler through adk_ph_dev_callback_fn_add.
 * Application can modify the parameter of already registered exception 
 * using the same register API 
 * * =============================================================
 */
adk_status_t
adk_ethernet_register_exception(  adk_module_handle_t eth_mod_handle,
        const adk_ethernet_exception_handler_t  *exceptionHandler_p)
{

    adk_status_t                    ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_ethernet_dp_exception_handler_t *dpExceptionHdl_p=NULL,*temp;
    adk_sys_handle_t                sys_handle;
    uint8_t dpExcepIndex;
    uint32_t                        policerFlowId = ADK_QOS_INVALID_POLICER_FLOW_ID;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to configure ethernet exception\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if (!exceptionHandler_p)
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "NULL pointer parameter\n");
        goto _adk_eth_exception_handler_register_exit;
    }

    switch(exceptionHandler_p->type)
    {
        case ADK_ETHERNET_EXCEPTION_TYPE_UNRESOLVED_ARP:
        case ADK_ETHERNET_EXCEPTION_TYPE_UNRESOLVED_IPV6_NEIGHBOR:
        case ADK_ETHERNET_EXCEPTION_TYPE_NEIGHBOR_STALE_ENTRY_USED:
            break;
        default:
            ret = ADK_ERROR_SEV_LEVEL_SET(ADK_ETHERNET_E_INVALID_EXCEPTION_TYPE);
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                    "Invalid Exception type \n");
            goto _adk_eth_exception_handler_register_exit;
    }
    dpExcepIndex = exceptionHandler_p->type - 1;
    dpExceptionHdl_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(eth_mod_handle),
            ADK_ETH_EXCEP_OBJ, 1, sizeof(adk_ethernet_dp_exception_handler_t), ADK_ETH_DP_OBJECT_ALIGN);
    if (!dpExceptionHdl_p)
    {
        ret = ADK_E_MEM_UNAVAILABLE;
        ADK_ETH_LOG_FATAL(eth_mod_handle, ret,
                "Could not build the input pattern\n");
        goto _adk_eth_exception_handler_register_exit;
    }

    /*change device and queue id, if host policer enabled*/
    if (exceptionHandler_p->policer_handle)
    {
        ret = adk_qos_policing_flow_id_get(sys_handle, exceptionHandler_p->policer_handle, &policerFlowId);
        if (ADK_SUCCESS != ret) {
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                    "Invalid policer handle\n");
            goto _adk_eth_exception_handler_register_exit;
        }
    }

    dpExceptionHdl_p->type = exceptionHandler_p->type;
    dpExceptionHdl_p->devId = exceptionHandler_p->devId;
    dpExceptionHdl_p->queueId = exceptionHandler_p->queueId;
    dpExceptionHdl_p->priority = exceptionHandler_p->priority;
    dpExceptionHdl_p->cbFnId = exceptionHandler_p->cbFnHandle;
    dpExceptionHdl_p->policerId = policerFlowId;

    if(eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex])
    {
        temp = eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex];

        eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex] = dpExceptionHdl_p;

        adk_rcu_synchronize();
        ADK_DP_SHM_FREE(temp);
    }
    else
    {
        eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex] = dpExceptionHdl_p;
    }

_adk_eth_exception_handler_register_exit:

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!
 * =============================================================
 * adk_ethernet_ph_rx_parse adk_ethernet.h "adk_ethernet.h"
 *
 * @param[in]   task Pointer to rte_mbuf.
 *
 * @param[out]  dp_message   Pointer to structure having the decoded parameters
 *                            sent by eth data path code to application.
 *
 * @retval   ADK_SUCCESS - successfully decoded the parameters
 * handler
 *
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 *                                    passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function helps in interpreting the parameters present in the task
 * Parameters that Ethernet module send along with the packet to the Application
 * via PH Dev.
 *
 * =============================================================
 */

adk_status_t
adk_ethernet_ph_rx_parse (
        void *task,
        void **dp_message,
        adk_module_handle_t eth_mod_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    struct rte_mbuf *mbuf_p;
    adk_metadata_t *adk_md_ptr;
    adk_ph_dev_dp_metadata_t *phMeta_p;
    eth_ph_metadata_t *ph_metadata;
    adk_sys_handle_t    sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        return ADK_E_INTERNAL_ERROR;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    if (!task || !dp_message)
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ethernet_ph_rx_parse_decode_exit;
    }

    mbuf_p = (struct rte_mbuf *)task;
    adk_md_ptr = ADK_METADATA_PTR(mbuf_p);
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(adk_md_ptr->imc_param_start);
    ph_metadata = (eth_ph_metadata_t *)&(phMeta_p->imc_param_start);

    *dp_message = (adk_ethernet_recv_exception_param_t*)ph_metadata;

_adk_ethernet_ph_rx_parse_decode_exit:

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);
    ADK_RETURN(ret);
}

/*!
 * =============================================================
 * adk_ethernet_unregister_exception
 * @param[in]  eth_mod_handle ETH module handle.
 *
 * @param[in]   exceptionType of  adk_ethernet_exception_type_t
 *
 * @retval   ADK_SUCCESS - successfully unregistered exception
 * handler
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @retval  ADK_E_INVALID_HANDLE - Invalid ethernet handle specified 
 *
 * @retval ADK_IP_E_INVALID_EXCEPTION_TYPE  - Invalid exception type.
 *
 * @brief
 *
 * This function allows the user to unregister exception handlers
 * For various Ethernet exceptions. enum adk_ethernet_exception_type_t defines
 * The type of exceptions that are supported by ADK-ETH.
 * * =============================================================
 */
adk_status_t
adk_ethernet_unregister_exception(  adk_module_handle_t eth_mod_handle,
        const adk_ethernet_exception_type_t  exceptionType)
{

    adk_status_t                    ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_ethernet_dp_exception_handler_t *dpExceptionHdl_p=NULL,*temp;
    adk_sys_handle_t                sys_handle;
    uint8_t dpExcepIndex;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {   
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to unregisterd ethernet exception\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }


    switch(exceptionType)
    {
        case ADK_ETHERNET_EXCEPTION_TYPE_UNRESOLVED_ARP:
        case ADK_ETHERNET_EXCEPTION_TYPE_UNRESOLVED_IPV6_NEIGHBOR:
        case ADK_ETHERNET_EXCEPTION_TYPE_NEIGHBOR_STALE_ENTRY_USED:
            break;
        default:
            ret = ADK_ERROR_SEV_LEVEL_SET(ADK_ETHERNET_E_INVALID_EXCEPTION_TYPE);
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                    "Invalid Exception type \n");
            goto _adk_eth_exception_handler_unregister_exit;
    }

    dpExcepIndex = exceptionType - 1;
    if(eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex])
    {
        temp = eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex];
        eth_internal_data_p->dp_pvt_p->exceptionHandlerTble[dpExcepIndex] = dpExceptionHdl_p;

        adk_rcu_synchronize();
        ADK_DP_SHM_FREE(temp);
    }
    else
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Not a valid registered exception \n");
        goto _adk_eth_exception_handler_unregister_exit;
    }

_adk_eth_exception_handler_unregister_exit:

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!
* =============================================================
*
* adk_ethernet_neighbor_entry_add
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
* @param[in]   ipv6_addr           The next hop IPv6 address
* @param[in]   dest_mac            The MAC address of the next hop IPv6 address
* @param[in]   neigh_status        status of Neighbor entry ex: INCOMPLETE, RECHABLE, etc.
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INVALID_PARAMETER Invalid parameter
* @retval  ADK_E_INTERNAL_ERROR    Internal error while adding neigh entry
*
* @brief  API to add an entry in the IPv6 Neighbor table.
*
* The API takes the next hop IPv6 address and maps it to the destination
* MAC address.If entry already exists then it would be overwritten.
*
* =============================================================
*/

adk_status_t
adk_ethernet_neighbor_entry_add(
                    adk_module_handle_t eth_mod_handle,
                    adk_ethernet_ipv6_addr_t ipv6_addr,
                    adk_ethernet_mac_addr_t dest_mac,
                    adk_ethernet_neigh_status_t neigh_status)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to add neigh entry\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

     if(neigh_status >= ADK_ETHERNET_NEIGHBOR_MAX)
    {
        dp_ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret, "Invalid neigh status value\n");
        return dp_ret;
    }

    dp_ret = adk_eth_dp_neigh(eth_mod_handle,
                                     ADK_ETH_TREE_OPER_ADD,
                                     &ipv6_addr, dest_mac, neigh_status);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to add entry to neigh table.\n");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Added neigh entry successfully for ipv6_addr %x:%x:%x:%x \n",
                         ipv6_addr.addr[0],ipv6_addr.addr[1],ipv6_addr.addr[2],ipv6_addr.addr[3]);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (dp_ret);
}


/*!
* =============================================================
*
* adk_ethernet_neighbor_entry_rem
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
* @param[in]   ipv6_addr           The next hop IPv6 address
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INVALID_PARAMETER neighbor entry not found
*
* @brief  API to remove an entry in the IPv6 Neighbor table.
*
* The API takes the next hop IPv6 address and removes the MAC
* address mapping for it.
* =============================================================
*/
adk_status_t
adk_ethernet_neighbor_entry_rem(
                    adk_module_handle_t eth_mod_handle,
                    adk_ethernet_ipv6_addr_t ipv6_addr)
{

    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_ethernet_mac_addr_t dest_mac;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to remove neigh entry\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    dp_ret = adk_eth_dp_neigh(eth_mod_handle,
                                     ADK_ETH_TREE_OPER_DEL,
                                     &ipv6_addr, dest_mac /*NOT USED*/,
                                    ADK_ETHERNET_NEIGHBOR_REACHABLE /* NOT USED */);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to remove entry from neighbor table.\n");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Removed neigh entry successfully for ipv6_addr %x:%x:%x:%x \n",
                         ipv6_addr.addr[0],ipv6_addr.addr[1],ipv6_addr.addr[2],ipv6_addr.addr[3]);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/* =============================================================
*
* adk_ethernet_get_neighbor_entries_all
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
* @param[out]  numArpEntries       Number of Neighbor entries
* @param[out]  neigh_entries_p     List of Neighbor entries
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INVALID_PARAMETER Invalid input pointers specfied
* @retval  ADK_E_INTERNAL_ERROR    Failed to list the Neighbor entries from Datapath
* @retval  ADK_E_MEM_UNAVAILABLE   Failed to allocate memory to store
*                                  Neighbor entries
*
* @brief  API to get all the entries in the Neighbor table.
*
* @details
* The API returns the list of Neighbor entries present in data path
* along with the number of entries.
* The application is expected to call rte_free for the memory allocated by
* the API in output parameter arp_entries_p, when the API returns
* non-zero number of entries.
*
* =============================================================
*/

adk_status_t adk_ethernet_get_neighbor_entries_all(
                    adk_module_handle_t eth_mod_handle,
                    uint32_t *numNeighborEntries,
                    adk_ethernet_neighbor_entry_t **neighbor_entries_p)
{
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get neighbor entries\n");
        return (ADK_E_INVALID_HANDLE);
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if((NULL == numNeighborEntries) || (NULL == neighbor_entries_p))
    {
        dp_ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret, "Invalid Input Parameters\n");
        return dp_ret;
    }

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }


    dp_ret = adk_eth_dp_neighbor_entries_get(&eth_internal_data_p->dp_pvt_p->eth_neighbor_hash_table,
                                                    numNeighborEntries, neighbor_entries_p);
    if (dp_ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, dp_ret,
                          "Failed to get neighbor enties from datapath");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Get neighbor entries successful from datapath\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (dp_ret);
}
/*!
* =============================================================
*
* adk_ethernet_neighbor_entry_get
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
* @param[in]   ipv6_addr           The next hop IPv6 address
*
* @param[out]  dest_mac            The MAC address of the next hop IPv6 address
* @param[out]  *neigh_status       Neighbor Entry Reachable Status 
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INVALID_PARAMETER Invalid Parameter 
*
* @brief  API to get an entry from the IPv6 Neigh table. 
*
* The API takes the next hop IPv6 address and gets the destination 
* MAC address and neighbor reachability status.  
* =============================================================
*/

adk_status_t
adk_ethernet_neighbor_entry_get(
        adk_module_handle_t eth_mod_handle,
        adk_ethernet_ipv6_addr_t ipv6_addr,
        adk_ethernet_mac_addr_t dest_mac,
        adk_ethernet_neigh_status_t *neigh_status)
{

    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get neigh entry\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    dp_ret = adk_eth_dp_neigh_get(eth_mod_handle,
            &ipv6_addr, dest_mac, neigh_status);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to Failed entry to neigh table.\n");
    }
    else
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                "Get neighbor entry successful for ipv6_addr %x:%x:%x:%x \n",
                ipv6_addr.addr[0],ipv6_addr.addr[1],ipv6_addr.addr[2],ipv6_addr.addr[3]);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}
/*!
* =============================================================
*
* adk_ethernet_autoconnection_enable
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INTERNAL_ERROR    Internal error while enabling auto connection
*
* @brief API to enable auto connection
 *
* The API will allow VLAN traffic in ingress/egress when SVIF is not created
*
* =============================================================
*/

adk_status_t adk_ethernet_autoconnection_enable (
                    adk_module_handle_t eth_mod_handle)

{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to enable the auto connection\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    *(eth_internal_data_p->dp_pvt_p->autoconn_status) = ADK_DP_ETH_AUTO_CONN_ENABLE;

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
* =============================================================
*
* adk_ethernet_autoconnection_disable
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INTERNAL_ERROR    Internal error while disabling auto connection
*
* @brief API to disable auto connection
 *
* The API will not allow VLAN traffic in ingress/egress when SVIF is not created
*
* =============================================================
*/
adk_status_t adk_ethernet_autoconnection_disable (
                    adk_module_handle_t eth_mod_handle)
{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to disable the auto connection\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    *(eth_internal_data_p->dp_pvt_p->autoconn_status) = ADK_DP_ETH_AUTO_CONN_DISABLE;

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
* =============================================================
*
* adk_ethernet_autoconnection_status_get
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INTERNAL_ERROR    Internal error while getting auto-connection status
*
* @brief API to get the auto connection status
*
* The API will return whether auto-connection is enabled or disabled
*
* =============================================================
*/
adk_status_t adk_ethernet_autoconnection_status_get(
                    adk_module_handle_t eth_mod_handle, bool *status)

{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get the auto connection status\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    *status = *(eth_internal_data_p->dp_pvt_p->autoconn_status);

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}


/*!
* =============================================================
*
* adk_ethernet_bridge_interface_mac_addr_modify
*
* @param[in] eth_mod_handle    Handle to the Ethernet module
* @param[in] src_mac           MAC address to be used in Ethernet header
* @param[in] bridge_if_handle  Bridge Interface Handle
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE    Invalid bridge interface handle
* @retval  ADK_ETHERNET_E_INVALID_MAC_ADDR  Invalid Mac address.
*
* @brief   API to modify source mac address of a bridge interface
*
*  This API modifies the MAC address of the bridge interface. The MAC
*  address must not be a multicast address or zero. The MAC address
*  is used for encapsulation of packets sent by higher layer module to
*  Ethernet module.This src mac is used to determine L2 termination on the ingress path.
*  Packets will not be terminated to higher layer on ingress until this API is called .
* =============================================================
*/
adk_status_t adk_ethernet_bridge_interface_mac_addr_modify(adk_module_handle_t eth_mod_handle,
        adk_ethernet_mac_addr_t   src_mac,
        adk_interface_handle_t bridge_if_handle)
{

    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_intf_internal_data_t    *b_if_data_p = NULL;
    adk_dp_eth_bridge_intf_t        *dp_bridge_intf_p = NULL;
    uint32_t                        vsi_id;
    uint32_t                        if_num;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to modify the bridge interface MAC\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (!adk_eth_valid_if_handle(bridge_if_handle))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Interface handle %"ADKPRIh "\n", bridge_if_handle);
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    if (src_mac == NULL )
    {
        ret = ADK_ETHERNET_E_INVALID_MAC_ADDR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Source mac\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    if (ADK_ETH_IS_MAC_MULTICAST(src_mac)) {
        ret = ADK_ETHERNET_E_INVALID_MAC_ADDR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Source MAC address"
                " %02x-%02x-%02x-%02x-%02x-%02x must not be"
                " multicast\n",
                ADK_ETH_MAC_ADDR_BYTES(src_mac));
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    if (ADK_ETH_IS_MAC_ZERO(src_mac)) {
        ret = ADK_ETHERNET_E_INVALID_MAC_ADDR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Invalid Source MAC address"
                " %02x-%02x-%02x-%02x-%02x-%02x\n",
                ADK_ETH_MAC_ADDR_BYTES(src_mac));
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    ADK_ETH_LOG_INFO(eth_mod_handle,"Input Parameters: Bridge handle %p, Source MAC: %02x:%02x:%02x:%02x:%02x:%02x \n",
            (adk_interface_handle_t *) bridge_if_handle,
            ADK_ETH_MAC_ADDR_BYTES(src_mac));

    b_if_data_p = (adk_eth_intf_internal_data_t *)
        ADK_SYS_INTERFACE_ATTR_PTR_GET(bridge_if_handle);


    ADK_MEMCPY(&b_if_data_p->u.bridge_if.src_mac, src_mac,
            sizeof(adk_ethernet_mac_addr_t));

    if_num = ADK_SYS_INTERFACE_ID_GET(bridge_if_handle);
    dp_bridge_intf_p = eth_internal_data_p->dp_pvt_p->bridge_intf[if_num];

    ADK_MEMCPY(&dp_bridge_intf_p->src_mac, src_mac,
                   sizeof(adk_ethernet_mac_addr_t));

    /*if Termination rule already added ,delete old one*/
    if(b_if_data_p->hos_l2_term_br_intf_rule_id)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,"Delete already existing termination rule on"
                            " bridge interface %d \n",(uint32_t)b_if_data_p->hos_l2_term_br_intf_rule_id);
        ret = adk_hos_rule_delete(ADK_HOS_L2_TERM_INTF_TBL,
                b_if_data_p->hos_l2_term_br_intf_rule_id);
        if (ret != ADK_SUCCESS)
        {
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                    "Failed to delete Bridge MAC-Term Interface HOS rule\n");
            goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
        }
    }

    /* TODO: Configure only bridge MAC without l2_domain.
     * Need reference count for bridge mac configured for deletion
     * */

    uint64_t bridge_mac =  src_mac[0];
    bridge_mac = (bridge_mac << 8) | src_mac[1];
    bridge_mac = (bridge_mac << 8) | src_mac[2];
    bridge_mac = (bridge_mac << 8) | src_mac[3];
    bridge_mac = (bridge_mac << 8) | src_mac[4];
    bridge_mac = (bridge_mac << 8) | src_mac[5];

    ret = adk_sys_eth_dev_vsi_get(dp_bridge_intf_p->eth_dev_id,
                                  &vsi_id);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get the vsi number from the eth device ID\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    ret = adk_hos_mac_mapper_add(bridge_mac, UNICAST_MAC_ADDR_MASK_BITS, ADK_ETH_BRIDGE_MAC_UNICAST_CMAC_1,
            &b_if_data_p->hos_mac_mapper_rule_id); 
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to add Bridge MAC-mapper HOS rule\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    ret = adk_hos_mac_mapper_add(MULTICAST_MAC_ADDR, MULTICAST_MAC_ADDR_MASK_BITS, ADK_ETH_BRIDGE_MAC_M_B_CMAC_1,
            &b_if_data_p->hos_multicast_mac_mapper_rule_id);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to add multicast MAC-mapper HOS rule\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    ret = adk_hos_mac_mapper_add(MULTICAST_MAC_ADDR_V6, MULTICAST_MAC_ADDR_V6_MASK_BITS, ADK_ETH_BRIDGE_MAC_M_B_CMAC_2,
            &b_if_data_p->hos_multicast_v6_mac_mapper_rule_id);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to add multicast_v6 MAC-mapper HOS rule\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    ret = adk_hos_mac_mapper_add(BROADCAST_MAC_ADDR, BROADCAST_MAC_ADDR_MASK_BITS, ADK_ETH_BRIDGE_MAC_M_B_CMAC_3,
            &b_if_data_p->hos_broadcast_mac_mapper_rule_id);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to add broadcast MAC-mapper HOS rule\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    adk_hos_l2_term_intf_tbl_rule_attr_t hos_l2_term_intf_rule =
    {
        .l2_domain = b_if_data_p->u.bridge_if.l2_domain_id,
        .dmac = bridge_mac,
        .vlan_id = 0,
        .l2_interface = ADK_HOS_DEFAULT_L2_INTERFACE,
        .vsi_id = vsi_id,
        .vsi_q = 0,
        .is_l2_term = true,
        .is_l2_m_b = false
    };

    ret = adk_hos_rule_add(ADK_HOS_L2_TERM_INTF_TBL,
                               (adk_hos_rule_attr_t *) &hos_l2_term_intf_rule,
                               &b_if_data_p->hos_l2_term_br_intf_rule_id);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to add Bridge MAC-Term interface HOS rule\n");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

    ret = adk_sys_flow_cache_update_version(sys_handle);
    if(ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to update global flow cache version");
        goto _adk_ethernet_bridge_interface_mac_addr_modify_exit;
    }

_adk_ethernet_bridge_interface_mac_addr_modify_exit:

    ADK_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:

    if (ret == ADK_SUCCESS) {
        ADK_ETH_LOG_INFO(eth_mod_handle, "Bridge handle modified %p| Source MAC: %02x:%02x:%02x:%02x:%02x:%02x \n",(adk_interface_handle_t *) bridge_if_handle,
                ADK_ETH_MAC_ADDR_BYTES(src_mac));
    } else {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Failed to modify src MAC on bridge handle %p \n",(adk_interface_handle_t *) bridge_if_handle);
    }

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_ethertype_p_bits_map_add
 *
 * @param[in]   eth_mod_handle         Handle to the Ethernet module
 * @param[in]   eth_type               Ethertype
 * @param[in]   pcp                    The PCP value used while
 *                                     sending packet out
 *
 * @retval  ADK_SUCCESS                        Successful completion
 * @retval  ADK_E_INVALID_HANDLE               Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_PROTO_TYPE  Invalid Ether_type value
 * @retval  ADK_ETHERNET_E_INVALID_PRIORITY    Invalid 3-bit PCP specified
 *
 * @brief   API to add Ethertype to PCP mapping
 *
 * This API maps ethertype to 3-bit PCP. This mapping is used
 * for non IP traffic going out via bridge interface/SVIF
  * =============================================================
 */
adk_status_t adk_ethernet_ethertype_p_bits_map_add(
        adk_module_handle_t eth_mod_handle,
        adk_ethernet_proto_t eth_type,
        adk_ethernet_priority_t pcp)
{
    adk_status_t        ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %p\n",
                (adk_module_handle_t *) eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle,ret);

    ADK_ETH_LOG_INFO(eth_mod_handle,"Input Parameters:  eth_type 0x%x, pcp value 0x%x", eth_type, pcp);

    if (ADK_ETH_IS_PROTO_INVALID(eth_type))
    {
        ret = ADK_ETHERNET_E_INVALID_PROTO_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Ether Type 0x%04x\n", eth_type);
        goto _adk_ethernet_ethertype_p_bits_map_add;
    }

    if (ADK_ETH_IS_PRIORITY_INVALID(pcp)) {
        ret = ADK_ETHERNET_E_INVALID_PRIORITY;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid priority value %d\n",
                pcp);
        goto _adk_ethernet_ethertype_p_bits_map_add;
    }
    ret = adk_eth_map_proto_to_pcp(eth_mod_handle, ADK_ETH_TREE_OPER_ADD,
            eth_type, pcp);
    if (ret != ADK_SUCCESS) {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to set the mapping between ethertype and PCP\n");
        goto _adk_ethernet_ethertype_p_bits_map_add;
    }

_adk_ethernet_ethertype_p_bits_map_add:

    ADK_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:

    if (ret == ADK_SUCCESS) {
        ADK_ETH_LOG_INFO(
                eth_mod_handle,
                "Ethertype 0x%x mapping added and set to %d\n",
                eth_type, pcp);
    } else
    {
        ADK_ETH_LOG_ERROR(
                eth_mod_handle, ret,
                "Failed to map proto to PCP value 0x%x\n",
                pcp);
    }

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);


    return(ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_ethertype_p_bits_map_del
 *
 * @param[in]   eth_mod_handle         Handle to the Ethernet module
 * @param[in]   eth_type               Ethertype
 *
 * @retval  ADK_SUCCESS                      Successful completion
 * @retval  ADK_E_INVALID_HANDLE             Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_PROTO_TYPE   Invalid Ether_type value
 *
 * @brief   API to delete Ethertype to PCP mapping
 *
 * This API deletes ethertype to 3-bit PCP mapping.
 * =============================================================
 */
adk_status_t adk_ethernet_ethertype_p_bits_map_del(
        adk_module_handle_t eth_mod_handle,
        adk_ethernet_proto_t eth_type)

{
    adk_status_t        ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %p\n", (adk_module_handle_t *) eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    ADK_ETH_LOG_INFO( eth_mod_handle,"Input Parameters:  eth_type 0x%x", eth_type);

    if (ADK_ETH_IS_PROTO_INVALID(eth_type))
    {
        ret = ADK_ETHERNET_E_INVALID_PROTO_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Ether Type 0x%04x\n", eth_type);
        goto _adk_ethernet_ethertype_p_bits_map_del;
    }

    ret = adk_eth_map_proto_to_pcp(eth_mod_handle, ADK_ETH_TREE_OPER_DEL,
            eth_type, 0);
    if (ret != ADK_SUCCESS) {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to remove the mapping between ethertype and PCP\n");
        goto _adk_ethernet_ethertype_p_bits_map_del;
    }
_adk_ethernet_ethertype_p_bits_map_del:

    ADK_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:

    if (ret == ADK_SUCCESS) {
        ADK_ETH_LOG_INFO(
                eth_mod_handle,
                "Ethertype 0x%x to PCP bits mapping removed\n", eth_type);
    } else
    {
        ADK_ETH_LOG_ERROR(
                eth_mod_handle, ret,
                "Failed to remove the Ether type 0x%x mapping \n",eth_type);
    }

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!
* =============================================================
*
* adk_ethernet_dscp_p_bits_map_add
*
* @param[in]   eth_mod_handle      Handle to the Ethernet module
* @param[in]   dscp                The input DSCP code point to be mapped
* @param[in]   pcp                 The output 3-bit PCP priority
*
* @retval  ADK_SUCCESS                          Successful completion
* @retval  ADK_E_INVALID_HANDLE                 Invalid Ethernet module handle specified
* @retval  ADK_ETHERNET_E_INVALID_PRIORITY      Invalid output 3-bit PCP specified
* @retval  ADK_ETHERNET_E_INVALID_DSCP          Invalid DSCP Value
* @retval  ADK_E_INTERNAL_ERROR                 Internal error while adding MAP 
 *
* @brief   Function to add DSCP code point to PCP mapping
*
* This function add a map from specific DSCP code point to 3-bit PCP.
*
* =============================================================
*/
adk_status_t
adk_ethernet_dscp_p_bits_map_add (
    adk_module_handle_t  eth_mod_handle,
    uint8_t dscp,
    adk_ethernet_priority_t  pcp)
{
    adk_status_t        ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %p\n",
                (adk_module_handle_t *) eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle,ret);

    ADK_ETH_LOG_INFO(eth_mod_handle,"Input Parameters:  DSCP 0x%x, PCP 0x%x", dscp, pcp);

    if (ADK_ETH_IS_DSCP_INVALID(dscp))
    {
        ret = ADK_ETHERNET_E_INVALID_DSCP;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid DSCP 0x%04x\n", dscp);
        goto _adk_ethernet_dscp_p_bits_map_add;
    }

    if (ADK_ETH_IS_PRIORITY_INVALID(pcp)) {
        ret = ADK_ETHERNET_E_INVALID_PRIORITY;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid priority value %d\n",
                pcp);
        goto _adk_ethernet_dscp_p_bits_map_add;
    }

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    eth_internal_data_p->dp_pvt_p->eth_dscp_to_pcp_map[dscp] = pcp;

    ret = adk_sys_flow_cache_update_version(sys_handle);
    if(ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to update global flow cache version");
        goto _adk_ethernet_dscp_p_bits_map_add;
    }

_adk_ethernet_dscp_p_bits_map_add:
    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
* =============================================================
*
* adk_ethernet_dscp_p_bits_map_del
*
* @param[in]   eth_mod_handle  Handle to the Ethernet module
* @param[in]   input           The input DSCP code point for which
*                              mapping needs to be removed
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_E_INVALID_PARAMETER Invalid parameter
* @retval  ADK_E_INTERNAL_ERROR    Internal error while removing Map
*
* @brief   Function to remove mapping for a DSCP code point
*
* This function removes mapping for specific DSCP code point.
* =============================================================
*/
adk_status_t
adk_ethernet_dscp_p_bits_map_del (
    adk_module_handle_t  eth_mod_handle,
    uint8_t dscp)
{
    adk_status_t        ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %p\n",
                (adk_module_handle_t *) eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle,ret);

    ADK_ETH_LOG_INFO(eth_mod_handle,"Input Parameters:  DSCP 0x%x", dscp);

    if (ADK_ETH_IS_DSCP_INVALID(dscp))
    {
        ret = ADK_ETHERNET_E_INVALID_DSCP;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid DSCP 0x%04x\n", dscp);
        goto _adk_ethernet_dscp_p_bits_map_del;
    }

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if(eth_internal_data_p->dp_pvt_p->eth_dscp_to_pcp_map[dscp] == 0)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "PCP Value not set for DSCP 0x%04x\n", dscp);
        goto _adk_ethernet_dscp_p_bits_map_del;
    }
    else
    {
        eth_internal_data_p->dp_pvt_p->eth_dscp_to_pcp_map[dscp] = 0;
    }

    ret = adk_sys_flow_cache_update_version(sys_handle);
    if(ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to update global flow cache version");
        goto _adk_ethernet_dscp_p_bits_map_del;
    }

_adk_ethernet_dscp_p_bits_map_del:
    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);

}

/*!
 * =============================================================
 *
 * adk_ethernet_vlan_eth_type_add
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   if_handle       The interface handle of the trunk interface
 *                              for which the ether type for VLAN needs to be
 *                              configured
 * @param[in]   eth_type        The eth type for the VLAN tag to be added
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE Invalid interface handle specified
 * @retval  ADK_ETHERNET_E_INCOMPAT_IF_TYPE Operation not permitted on passed
 * interface type
 * @retval  ADK_ETHERNET_E_INCOMPAT_FWD_MODE Operation not permitted for passed
 * interface due to incompatible forwarding mode
 * @retval  ADK_ETHERNET_E_INVALID_PROTO_TYPE Invalid Ether Type field specified
 *
 * @brief   Function to add allowed eth type for the VLAN tag on the trunk port.
 * While Ethernet allow any number of eth types in data plane software, only
 * two can be programmed in hardware. Programming in hardware gives better
 * performance in case IP traffic is being carried over and packet is getting
 * routed using the switch virtual interface concept. This API keeps the
 * latest two programmed in the hardware.
 *
 * This API is not supported
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_vlan_eth_type_add (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    adk_interface_handle_t __attribute__ ((unused)) if_handle,
    adk_ethernet_proto_t __attribute__ ((unused)) eth_type)
{
    return (ADK_SUCCESS);
}

/*!
* =============================================================
*
* adk_ethernet_ethertype_based_qos_policy_set
*
* @param[in]   eth_mod_handle    Handle to the Ethernet module
* @param[in]   ifHandle          Interface handle of bridge
* @param[in]   eth_type          Ethertype (e.g SyncE, PTP/Eth, ARP etc)
*
* @param[in]   queueID           Queue Mapped which implements the
*                                Buffer Management, Shaping and
*                                Scheduling policies
*
* @retval  ADK_SUCCESS                       Successful completion
* @retval  ADK_E_INVALID_HANDLE              Invalid Handle passed as NULL
* @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE  Invalid Interface handle
* @retval  ADK_ETHERNET_E_INVALID_PROTO_TYPE Invalid Ether type value
* @retval  ADK_E_INVALID_PARAMETER           Invaid parameter
* @retval  ADK_E_INTERNAL_ERROR              Internal error while setting Ether Type to Queue Mapping
*
* @brief   API to configure a QoS policy based on Ether Type.
*
* This API configures a non IP L3 QoS policy to be applied on the packet based on
* ethertype for CPU originated packets.
* =============================================================
*/
adk_status_t adk_ethernet_ethertype_based_qos_policy_set(
        adk_module_handle_t eth_mod_handle,
        adk_interface_handle_t ifHandle,
        adk_ethernet_proto_t ethType,
        adk_qos_queue_id_t  queueId)
{
    adk_status_t        ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %p\n",
                (adk_module_handle_t *) eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle,ret);

    ADK_ETH_LOG_INFO(eth_mod_handle,"Input Parameters:  eth_type 0x%x, qos_queue_id 0x%x\n", ethType, queueId);

    if (!adk_eth_valid_if_handle(ifHandle))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Interface handle %"ADKPRIh "\n", ifHandle);
        goto _adk_ethernet_ethertype_based_qos_policy_set;
    }

    if (ADK_ETH_IS_PROTO_INVALID(ethType))
    {
        ret = ADK_ETHERNET_E_INVALID_PROTO_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Ether Type 0x%04x\n", ethType);
        goto _adk_ethernet_ethertype_based_qos_policy_set;
    }

    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                      "Invalid queue id.\n");
        goto _adk_ethernet_ethertype_based_qos_policy_set;
    }

    ret = adk_eth_map_ethertype_to_queue(eth_mod_handle, ifHandle, ADK_ETH_TREE_OPER_ADD,
            ethType, queueId);
    if (ret != ADK_SUCCESS) {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to set qos policy for etherType \n");
        goto _adk_ethernet_ethertype_based_qos_policy_set;
    }

_adk_ethernet_ethertype_based_qos_policy_set:

    ADK_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:

    if (ret == ADK_SUCCESS) {
        ADK_ETH_LOG_INFO(
                eth_mod_handle,
                "Successfully to set qos policy[QueueID=%d] for etherType %x\n",
                queueId, ethType);
    } else
    {
        ADK_ETH_LOG_ERROR(
                eth_mod_handle, ret,
                "Failed to set qos policy[QueueID=%d] for etherType %x\n",
                queueId, ethType);
    }

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}


/*!
* =============================================================
*
* adk_ethernet_ethertype_based_qos_policy_rem
*
* @param[in]   eth_mod_handle   Handle to the Ethernet module
* @param[in]   ifHandle         Interface handle of bridge
* @param[in]   eth_type         Ethertype (e.g SyncE, PTP/Eth, ARP etc)
*
* @retval  ADK_SUCCESS                            Successful completion
* @retval  ADK_E_INVALID_HANDLE                   Invalid Handle passed as NULL
* @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE  Invalid Interface handle
* @retval  ADK_ETHERNET_E_INVALID_PROTO_TYPE      Invalid Ether type value
* @retval  ADK_E_INTERNAL_ERROR                   Internal error while removing Ether Type to Queue Mapping
*
* @brief   API to de-configure a QoS policy based on EtherType
*
* This API de-configures a non IP L3 QoS policy to be applied on the packet based on
* ethertype for CPU originated packets.
* =============================================================
*/
adk_status_t adk_ethernet_ethertype_based_qos_policy_rem(
        adk_module_handle_t eth_mod_handle,
        adk_interface_handle_t ifHandle,
        adk_ethernet_proto_t ethType)
{
    adk_status_t        ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %p\n",
                (adk_module_handle_t *) eth_mod_handle);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle,ret);

    ADK_ETH_LOG_INFO(eth_mod_handle,"Input Parameters:  eth_type 0x%x \n", ethType);

    if (!adk_eth_valid_if_handle(ifHandle))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Interface handle %"ADKPRIh "\n", ifHandle);
        goto _adk_ethernet_ethertype_based_qos_policy_rem;
    }

    if (ADK_ETH_IS_PROTO_INVALID(ethType))
    {
        ret = ADK_ETHERNET_E_INVALID_PROTO_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Ether Type 0x%04x\n", ethType);
        goto _adk_ethernet_ethertype_based_qos_policy_rem;
    }

    ret = adk_eth_map_ethertype_to_queue(eth_mod_handle, ifHandle, ADK_ETH_TREE_OPER_DEL,
            ethType, ADK_QOS_INVALID_QUEUE_ID);
    if (ret != ADK_SUCCESS) {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Failed to remove qos policy for etherType \n");
        goto _adk_ethernet_ethertype_based_qos_policy_rem;
    }

_adk_ethernet_ethertype_based_qos_policy_rem:

    ADK_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:

    if (ret == ADK_SUCCESS) {
        ADK_ETH_LOG_INFO(
                eth_mod_handle,
                "Successfully removed qos policy for etherType %x\n",
                ethType);
    } else
    {
        ADK_ETH_LOG_ERROR(
                eth_mod_handle, ret,
                "Failed to remove qos policy for etherType %x\n",
                 ethType);
    }

    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return(ret);
}

/*!
* =============================================================
*
* adk_ethernet_non_bridge_origin_queue_id_set
*
* @param[in]   eth_mod_handle    Handle to the Ethernet module
*
* @param[in]   queueID          Queue Mapped which implements the
*                               Buffer Management, Shaping and
*                               Scheduling policies
*
* @retval  ADK_SUCCESS                      Successful completion
* @retval  ADK_E_INVALID_HANDLE            Invalid Handle
* @retval  ADK_E_INVALID_PARAMETER          Invalid parameter
* @retval  ADK_E_INTERNAL_ERROR    Internal error while setting Queue ID
*
* @brief   API to configure QueueID for non-bridge originated netd traffic .
*
* This API configures QueueID to be applied on the non-bridge originated traffic 
* =============================================================
*/
adk_status_t adk_ethernet_non_bridge_origin_queue_id_set(
        adk_module_handle_t eth_mod_handle,
        adk_qos_queue_id_t  queueId)
{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to set non bridge origin traffic queue id\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid queue id.\n");
        goto _adk_ethernet_non_bridge_origin_queue_id_set;
    }

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    *(eth_internal_data_p->dp_pvt_p->eth_non_bridge_origin_queue_id) = queueId;

_adk_ethernet_non_bridge_origin_queue_id_set:
    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
* =============================================================
*
* adk_ethernet_non_bridge_origin_queue_id_reset
*
* @param[in]   eth_mod_handle   Handle to the Ethernet module
*
* @retval  ADK_SUCCESS                  Successful completion
* @retval  ADK_E_INVALID_HANDLE         Invalid Handle passed as NULL
* @retval  ADK_E_INTERNAL_ERROR         Internal error while resetting the Queue ID
*
* @brief   API to reset a queueID for non-bridge originate netd traffic
*
* This API reset a queueID to be applied on the packet non bridge originated traffic
* =============================================================
*/
adk_status_t adk_ethernet_non_bridge_origin_queue_id_reset(
        adk_module_handle_t eth_mod_handle)
{

    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to set non bridge origin traffic queue id\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    *(eth_internal_data_p->dp_pvt_p->eth_non_bridge_origin_queue_id) = ADK_QOS_INVALID_QUEUE_ID;

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
* =============================================================
*
* adk_ethernet_bridge_timestamp_quad_group_set
*
* @param[in]   eth_mod_handle    Handle to the Ethernet module
*
* @param[in]   quadGroup         QaudGroup of HLP ports on which timestamp
*                                Should be read
*
* @retval  ADK_SUCCESS                      Successful completion
* @retval  ADK_E_INVALID_HANDLE             Invalid Handle
* @retval  ADK_E_INVALID_PARAMETER          Invalid parameter
* @retval  ADK_E_INTERNAL_ERROR             Internal error while setting Quad Group
*
* @brief   API to configure  Quad Group of HLP ports for which timestamp has to be read for bridge interface
*
* This API configures Quad Group of HLP ports to getting timestamp for bridge interface
* QUAD-0  {port-0,1,2,3}
* QUAD-1  {port-4,5,6,7}
* QUAD-2  {port-8,9,10,11}
* QUAD-3  {port-12,13,14,15}
* QUAD-4  {port-16,17,18,19}
*
* =============================================================
*/
adk_status_t adk_ethernet_bridge_timestamp_quad_group_set (
        adk_module_handle_t eth_mod_handle,
        adk_ethernet_quad_group_t quadGroup)
{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to set bridge timestamp quad group\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    if (quadGroup < ADK_ETHERNET_QUAD_GROUP_0 ||
          quadGroup >= ADK_ETHERNET_QUAD_GROUP_MAX){
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                "Invalid Quad group.\n");
        goto _adk_ethernet_bridge_timestamp_quad_group_set;
    }

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    *(eth_internal_data_p->dp_pvt_p->eth_bridge_timestamp_quad_group) = quadGroup;
    *(eth_internal_data_p->dp_pvt_p->eth_bridge_timestamp_quad_group_port) =
                        adk_eth_quad_group_port_map[quadGroup][ADK_ETH_FIRST_QUAD_PORT_INDEX];

_adk_ethernet_bridge_timestamp_quad_group_set:
    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}
/*!     end addtogroup _ethernetFunction
 *      @}
 *  end addtogroup _ethernet
 * @}
 */
