/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_ethernet_bind.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernet  Ethernet Module
 * @{
 *      @addtogroup _ethernetFunctions  Ethernet Functions
 *      @{
 *
 * @brief   This files has the functions which are used for binding Ethernet to
 *          other modules. These functions should not be called directly by the
 *          application but they are invoked indirectly via ADK SIF.
 *
 *  **********************************************************************/

#include <adk_eth.h>

/*!
 * =============================================================
 * @internal
 * adk_ethernet_get_imc
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   bindInfo_p      Binding information
 * @param[out]  imcFnId         IMC function ID to return
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 * @retval  ADK_E_IMC_NOT_FOUND     There is no IMC defined.
 *
 * @brief   Function to get the IMC function ID.
 *
 * This function is used to get the IMC function when binding Ethernet module
 * as source module. This should not be called directly from the application but
 * is called indirectly from SIF.
 *
 * NOTE: that if there is no IMC function required, the IMC function ID should
 * be set to 0 and ADK_SUCCESS should be returned.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_get_imc (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    void __attribute__ ((unused))*bindInfo_p,
    uint32_t *imcFnId)
{
    *imcFnId = 0;
    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * @internal
 * adk_ethernet_get_entry_func
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   bindInfo_p      Binding information
 * @param[out]  entry_fn_id     Entry function ID to return
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 * @retval  ADK_E_IMC_NOT_FOUND     There is no IMC defined.
 *
 * @brief   Function to get the Entry function ID.
 *
 * This function is used to get the Entry function when binding to Ethernet
 * module as destination module. This is called from the ADK SIF. This should
 * not be called directly from the application.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_get_entry_func (
    adk_module_handle_t __attribute__ ((unused)) eth_mod_handle,
    void __attribute__ ((unused))*bindInfo_p,
    uint32_t *entry_fn_id)
{
    *entry_fn_id = 0;
    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * @internal
 * adk_ethernet_custom_bind
 *
 * @param[in]   sysHandle       ADK System Handle
 * @param[in]   bindInfo_p      Binding information
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Function to perform custom bind from Ethernet to any higher layer
 *          module
 *
 * This function is used to bind Ethernet to any higher layer module without
 * going through the ADK SIF fast path code. This custom binding provides better
 * throughput compared to regular SIF binding.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called indirectly and automatically via adk_bind() API, if
 * isCustomBind parameter to adk_bind() API is set to TRUE.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_custom_bind (
    adk_sys_handle_t sysHandle,
    adk_bind_info_t *bindInfo_p)
{
    adk_status_t ret = ADK_SUCCESS;
#ifndef ADK_DEVICE_SNR 
    adk_module_handle_t eth_mod_handle;
    adk_eth_flow_internal_data_t *flow_data_p = NULL;
    adk_dp_eth_ingress_flow_t *dp_flow_p;
    adk_dp_eth_cpu_flow_result_t *dp_cpu_flow_p;
    adk_sys_handle_t sys_handle;
    uint32_t destModId;

    eth_mod_handle = bindInfo_p->srcModuleHandle;

    if (!adk_eth_valid_handle(eth_mod_handle) || !sysHandle)
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to create Custom bind entry \n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    /* Validate the parameters in the bind_info */

    /* Custom bind must be a direct bind */
    if (bindInfo_p->path != ADK_BIND_PATH_DIRECT)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Only ADK_BIND_PATH_DIRECT is supported\n");
        goto _adk_ethernet_custom_bind_exit;
    }

    /* Only F2M or F2F binding is supported via custom bind */
    if ((bindInfo_p->type != ADK_BIND_FLOW_TO_MODULE) &&
        (bindInfo_p->type != ADK_BIND_FLOW_TO_FLOW))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid or Unsupported Binding Type %d\n",
                          bindInfo_p->type);
        goto _adk_ethernet_custom_bind_exit;
    }

    if (!adk_eth_valid_flow_handle(bindInfo_p->srcFlowHandle))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_HANDLE;
        ADK_ETH_LOG_ERROR(
            eth_mod_handle, ret,
            "Invalid Flow handle %"ADKPRIh "\n", bindInfo_p->srcFlowHandle);
        goto _adk_ethernet_custom_bind_exit;
    }

    /* get destination module id */
    if (bindInfo_p->destModuleHandle)
    {
        switch (ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle))
        {
        case ADK_IP_MODULE_ID:
            if (bindInfo_p->direction == ADK_BIND_INGRESS_PROCESSING)
                destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV4_LLI;
            else
                destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV4_ULI;
            break;

        case ADK_IPV6_MODULE_ID:
            if (bindInfo_p->direction == ADK_BIND_INGRESS_PROCESSING)
                destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV6_LLI;
            else
                destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV6_ULI;
            break;

        case ADK_ETH_OAM_MODULE_ID:
            /* TODO: Set the ID to correct value */
            destModId = ADK_ETH_OAM_MODULE_ID;
            break;

        default:
            ret = ADK_E_INVALID_PARAMETER;
            goto _adk_ethernet_custom_bind_exit;
        }
    }
    else
    {
        ret = ADK_E_INVALID_PARAMETER;
        goto _adk_ethernet_custom_bind_exit;
    }

    flow_data_p = (adk_eth_flow_internal_data_t *)
        ADK_SYS_FLOW_ATTR_PTR_GET(bindInfo_p->srcFlowHandle);

    if (bindInfo_p->direction == ADK_BIND_INGRESS_PROCESSING)
    {
        if (flow_data_p->type == ETH_FLOW_TYPE_SIMPLE_INGRESS)
        {
            dp_flow_p = flow_data_p->dp_u.dp_ing_flow_p;

            ret = adk_pal_dp_stage_info_get(ADK_MODULE_CPU_INGRESS_FN_GET(
                                                bindInfo_p->destModuleHandle,
                                                _DEST_FUNC_DEFAULT),
                                         &dp_flow_p->stage_info_p);
            if (ret != ADK_SUCCESS)
            {
                ret = ADK_E_INTERNAL_ERROR;
                ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                                  "Failed to get stage info for dest module\n");
                goto _adk_ethernet_custom_bind_exit;
            }

            flow_data_p->destFlow = bindInfo_p->destFlowHandle;

            dp_flow_p->flowId = ADK_SYS_FLOWID_GET(bindInfo_p->srcFlowHandle);
            dp_flow_p->destFunc[_DEST_FUNC_DEFAULT] = ADK_MODULE_CPU_INGRESS_FN_GET(
                bindInfo_p->destModuleHandle, _DEST_FUNC_DEFAULT);
            dp_flow_p->destFunc[_DEST_FUNC_RING] = ADK_MODULE_CPU_INGRESS_FN_GET(
                bindInfo_p->destModuleHandle, _DEST_FUNC_RING);
            if (bindInfo_p->destFlowHandle)
            {
                dp_flow_p->destFlow_valid = TRUE;
                dp_flow_p->destFlowId = ADK_SYS_FLOWID_GET(bindInfo_p->destFlowHandle);
            }
            else
            {
                dp_flow_p->destFlow_valid = FALSE;
                dp_flow_p->destFlowId = 0;
            }
            /* Lastly, set the dest module ID to flag the completion of bind */
            dp_flow_p->destModId = destModId;
        } else
        {
            ret = ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Incompatible flow type %s for ingres binding\n",
                              adk_eth_flow_type_name(flow_data_p->type));
            goto _adk_ethernet_custom_bind_exit;
        }
    }
    else
    {
        /* Check if Ethernet flow type is support for custom binding */
        if (flow_data_p->type != ETH_FLOW_TYPE_CPU_ORIGIN)
        {
            ret = ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Incompatible flow type %s for egress binding\n",
                              adk_eth_flow_type_name(flow_data_p->type));
            goto _adk_ethernet_custom_bind_exit;
        }

        flow_data_p->destFlow = bindInfo_p->destFlowHandle;

        dp_cpu_flow_p = flow_data_p->dp_u.dp_cpu_flow_p;

        dp_cpu_flow_p->destModId = destModId;
        dp_cpu_flow_p->destFunc[_DEST_FUNC_DEFAULT] = ADK_MODULE_CPU_EGRESS_FN_GET(
            bindInfo_p->destModuleHandle, _DEST_FUNC_DEFAULT);
        dp_cpu_flow_p->destFunc[_DEST_FUNC_RING] = ADK_MODULE_CPU_EGRESS_FN_GET(
            bindInfo_p->destModuleHandle, _DEST_FUNC_RING);

        if ((ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle) ==
             ADK_IP_MODULE_ID) ||
            (ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle) ==
             ADK_IPV6_MODULE_ID))
        {
            ret = adk_eth_cpu_flow(eth_mod_handle,
                                   ADK_ETH_TREE_OPER_ADD,
                                   flow_data_p);
            if (ret != ADK_SUCCESS)
            {
                ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Bind failed. \n");
                goto _adk_ethernet_custom_bind_exit;
            }
        }
    }

_adk_ethernet_custom_bind_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle, "Custom Bind Successfully done.\n");
    }
    else
    {
        ADK_ETH_LOG_FORCE("Failed to perform custom binding.\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);
#else
    RTE_SET_USED(sysHandle);
    RTE_SET_USED(bindInfo_p);
#endif
    return (ret);
}

/*!
 * =============================================================
 * @internal
 * adk_ethernet_custom_unbind
 *
 * @param[in]   sysHandle       ADK System Handle
 * @param[in]   bindInfo_p      Binding information
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Function to perform custom unbind between Ethernet and any higher
 *          layer module
 *
 * This function is used to unbind Ethernet from any higher layer module without
 * going through the ADK SIF fast path code.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called indirectly and automatically via adk_unbind() API, if
 * isCustomBind parameter to adk_unbind() API is set to TRUE.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_custom_unbind (
    adk_sys_handle_t sysHandle,
    adk_bind_info_t *bindInfo_p)
{
    adk_status_t ret = ADK_SUCCESS;
#ifndef ADK_DEVICE_SNR
    adk_module_handle_t eth_mod_handle;
    adk_dp_eth_ingress_flow_t *dp_flow_p;
    adk_dp_eth_cpu_flow_result_t *dp_cpu_flow_p;
    adk_eth_flow_internal_data_t *flow_data_p = NULL;
    adk_sys_handle_t sys_handle;

    eth_mod_handle = bindInfo_p->srcModuleHandle;
    if (!adk_eth_valid_handle(eth_mod_handle) || !sysHandle)
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to remove Custom bind entry \n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    /* Validate the parameters in the bind_info */

    /* Custom bind must be a direct bind */
    if (bindInfo_p->path != ADK_BIND_PATH_DIRECT)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Only ADK_BIND_PATH_DIRECT is supported\n");
        goto _adk_ethernet_custom_unbind_exit;
    }

    /* Only F2M or F2F binding is supported via custom bind */
    if ((bindInfo_p->type != ADK_BIND_FLOW_TO_MODULE) &&
        (bindInfo_p->type != ADK_BIND_FLOW_TO_FLOW))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid or Unsupported Binding Type %d\n",
                          bindInfo_p->type);
        goto _adk_ethernet_custom_unbind_exit;
    }

    if (!adk_eth_valid_flow_handle(bindInfo_p->srcFlowHandle))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Flow handle %"ADKPRIh "\n",
                          bindInfo_p->srcFlowHandle);
        goto _adk_ethernet_custom_unbind_exit;
    }

    flow_data_p = (adk_eth_flow_internal_data_t *)
                  ADK_SYS_FLOW_ATTR_PTR_GET(bindInfo_p->srcFlowHandle);

    if (bindInfo_p->direction == ADK_BIND_INGRESS_PROCESSING)
    {
        if (flow_data_p->type != ETH_FLOW_TYPE_SIMPLE_INGRESS)
        {
            ret = ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Incompatible flow type %s for ingress unbinding\n",
                              adk_eth_flow_type_name(flow_data_p->type));
            goto _adk_ethernet_custom_unbind_exit;
        }

        dp_flow_p = flow_data_p->dp_u.dp_ing_flow_p;

        flow_data_p->destFlow = 0;
        dp_flow_p->destModId = 0;

        adk_rcu_synchronize();

        dp_flow_p->destFunc[_DEST_FUNC_DEFAULT] = 0;
        dp_flow_p->destFunc[_DEST_FUNC_RING] = 0;
        dp_flow_p->destFlow_valid = FALSE;
        dp_flow_p->destFlowId = 0;
        dp_flow_p->stage_info_p = NULL;
    }
    else
    {
        /* Check if Ethernet flow type is support for custom binding */
        if (flow_data_p->type != ETH_FLOW_TYPE_CPU_ORIGIN)
        {
            ret = ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE;
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Incompatible flow type %s for egress unbinding\n",
                              adk_eth_flow_type_name(flow_data_p->type));
            goto _adk_ethernet_custom_unbind_exit;
        }

        if ((ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle) ==
             ADK_IP_MODULE_ID) ||
            (ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle) ==
             ADK_IPV6_MODULE_ID))
        {
            ret = adk_eth_cpu_flow(eth_mod_handle,
                                   ADK_ETH_TREE_OPER_DEL,
                                   flow_data_p);
            if (ret != ADK_SUCCESS)
            {
                ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Unbind failed. \n");
                goto _adk_ethernet_custom_unbind_exit;
            }
        }

        dp_cpu_flow_p = flow_data_p->dp_u.dp_cpu_flow_p;

        dp_cpu_flow_p->destFunc[_DEST_FUNC_DEFAULT] = 0;
        dp_cpu_flow_p->destFunc[_DEST_FUNC_RING] = 0;
        
        adk_rcu_synchronize();

        dp_cpu_flow_p->destModId = 0;
        
        flow_data_p->destFlow = 0;
    }

_adk_ethernet_custom_unbind_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle, "Custom Unbind Successfully done.\n");
    }
    else
    {
        ADK_ETH_LOG_FORCE("Failed to perform custom unbinding.\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);
#else
    RTE_SET_USED(sysHandle);
    RTE_SET_USED(bindInfo_p);
#endif
    return (ret);
}

/*!     end addtogroup _ethernetFunctions
 *      @}
 *  end addtogroup _ethernet
 * @}
 */
