/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

/*!
 * @file  adk_ethernet_cp_utility_counters.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernet Ethernet Module
 * @{
 *      @addtogroup _ethernetUtilityCounters Ethernet Utility Counters
 *      @{
 *
 * @brief
 *
 * The file contains APIs to configure Ethernet utility counters from the application
 * and get APIs to fetch utility counters from DP.
 ***********************************************************************/
#include <adk_eth.h>
#include "adk_ethernet_utility_counters.h"
/*!
 * =============================================================
 *
 * adk_eth_set_utility_counters
 *
 * @param[in]  ethHandle  - Ethernet module handle. 
 * @param[in]   utility_cnt_state - Boolean variable to enable/disable
                                    utility counter.
 *
 * @brief   Function to configure utility counters.
 *
 * This function configures utility counters.On disabling the utility
 * counters, all the values will get reset to zero.
 *
 * =============================================================
 */
adk_status_t
adk_eth_set_utility_counters(adk_module_handle_t ethHandle, bool utility_cnt_state)
{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(ethHandle))
    {   
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                ethHandle);
        ADK_ETH_LOG_FORCE("Failed to set ethernet utility counter status\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ethHandle);

    ADK_ETH_LOG_FUNC_ENTRY(ethHandle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(ethHandle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(ethHandle);

    if(utility_cnt_state)
    {
        eth_internal_data_p->dp_pvt_p->isUtilityCntEnabled = true;
    }
    else
    {
        eth_internal_data_p->dp_pvt_p->isUtilityCntEnabled = false;
        /* Reset the utility counters */
        memset(eth_internal_data_p->dp_pvt_p->ethUtilityCnt, 0,
                sizeof(adk_eth_utility_counters_t)*ADK_MAX_NUM_CORES);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(ethHandle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(ethHandle);

    return (ret);
}


/*!
 * =============================================================
 * @fn adk_eth_get_utility_counters
 *
 * @param[in]    ethHandle - Ethernet module handle.
 *
 * @param[out]   ethUtilityCounter_p - Ethernet related utility counter data structure pointer
 *               gets set by the Ethernet ADK
 *
 * @retval ADK_SUCCESS - Successfully printed utility counters
 * @retval ADK_E_INVALID_HANDLE - eth module handle is invalid.
 * @retval ADK_E_INVALID_PARAMETER - Invalid parameter
 *
 * @brief
 *
 * This API allows the application to retreive all Ethernet related utility counters
 * maintained by the ETH ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size adk_eth_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_eth_get_utility_counters(adk_module_handle_t           ethHandle,
                               adk_eth_utility_counters_t *ethUtilityCnt_p)
{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_utility_counters_t *utilityCounter;
    uint8_t i;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(ethHandle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                ethHandle);
        ADK_ETH_LOG_FORCE("Failed to set ethernet utility counter status\n");
        return (ADK_E_INVALID_HANDLE);
    }
    

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ethHandle);

    ADK_ETH_LOG_FUNC_ENTRY(ethHandle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(ethHandle, ret);

    if ( NULL == ethUtilityCnt_p) {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(ethHandle, ret,
                "ethUtilityCnt_p is NULL\n");
        goto _adk_eth_get_utility_counters;
    }


    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(ethHandle);


    ADK_MEMSET (ethUtilityCnt_p, 0x00, sizeof(adk_eth_utility_counters_t));

    for (i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        utilityCounter = &eth_internal_data_p->dp_pvt_p->ethUtilityCnt[i];
        
        /*Ingress Utility Counters*/
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkt_proc_lli_entry_count
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkt_proc_lli_entry_count; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkt_proc_lli_drop
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkt_proc_lli_drop; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_non_l2term_frames_received
            +=utilityCounter->ingressUtilityCounters.eth_ing_non_l2term_frames_received; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_phy_term_pkt_sent_to_netd
            +=utilityCounter->ingressUtilityCounters.eth_ing_phy_term_pkt_sent_to_netd; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkt_sent_to_netd
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkt_sent_to_netd; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkts_sent_to_l3
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkts_sent_to_l3; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkts_sent_to_ipv6
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkts_sent_to_ipv6; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkts_sent_to_ipv4
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkts_sent_to_ipv4; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkts_sent_to_policer_stage
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkts_sent_to_policer_stage; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkts_recv_in_policer_stage
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkts_recv_in_policer_stage; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkts_vlan_tag_stripped
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkts_vlan_tag_stripped; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_multicast_packets_received
            +=utilityCounter->ingressUtilityCounters.eth_ing_multicast_packets_received; 
        ethUtilityCnt_p->ingressUtilityCounters.eth_ing_pkt_sent_to_sw_ingress_flow_lookup
            +=utilityCounter->ingressUtilityCounters.eth_ing_pkt_sent_to_sw_ingress_flow_lookup; 
        /*Egress Utility Counters*/
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_proc_uli_entry_count
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_proc_uli_entry_count; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv4_pkts_received
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv4_pkts_received; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv6_pkts_received
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv6_pkts_received; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv6_multicast_pkts_received
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv6_multicast_pkts_received; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkts_received_from_netd
            +=utilityCounter->egressUtilityCounters.eth_eg_pkts_received_from_netd; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_received
            +=utilityCounter->egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_received; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_transmitted
            +=utilityCounter->egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_transmitted; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkts_sent_to_nic
            +=utilityCounter->egressUtilityCounters.eth_eg_pkts_sent_to_nic; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkts_sent_to_ph
            +=utilityCounter->egressUtilityCounters.eth_eg_pkts_sent_to_ph; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_proc_uli_drop
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_proc_uli_drop; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_arp_exception_sent
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_arp_exception_sent; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_nd_exception_sent
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_nd_exception_sent; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_stale_entry_used_exception_sent
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_stale_entry_used_exception_sent; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_sent_to_buffering_stage
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_sent_to_buffering_stage; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_recv_in_buffering_stage
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_recv_in_buffering_stage; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkt_dropped_in_buffering_stage
            +=utilityCounter->egressUtilityCounters.eth_eg_pkt_dropped_in_buffering_stage; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error
            +=utilityCounter->egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_nexthop_buffer_lock_alloc_error
            +=utilityCounter->egressUtilityCounters.eth_eg_nexthop_buffer_lock_alloc_error; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_overflow_error
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_overflow_error; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_create_failure
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_create_failure; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timeout_count
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timeout_count; 
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_restart_error
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_restart_error;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_overflow_error
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_overflow_error;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_create_failure
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_create_failure;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timeout_count
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timeout_count;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_restart_error
            +=utilityCounter->egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_restart_error;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkts_tstamp_alloc_requests
            +=utilityCounter->egressUtilityCounters.eth_eg_pkts_tstamp_alloc_requests;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkts_tstamp_free_requests
            +=utilityCounter->egressUtilityCounters.eth_eg_pkts_tstamp_free_requests;
        ethUtilityCnt_p->egressUtilityCounters.eth_eg_pkts_tstamp_alloc_success
            +=utilityCounter->egressUtilityCounters.eth_eg_pkts_tstamp_alloc_success;

    }

_adk_eth_get_utility_counters:
    ADK_ETH_RELEASE_MODULE_LOCK(ethHandle, ret);
_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(ethHandle);

    return (ret);
}

/*!
 * =============================================================
 * @fn adk_eth_print_utility_counters
 *
 * @param[in]    ethHandle - Ethernet module handle.
 *
 * @retval ADK_SUCCESS - Successfully printed utility counters
 * @retval ADK_E_INVALID_HANDLE - eth module handle is invalid.
 * @retval ADK_E_INTERNAL_ERROR -  Internal error
 *
 * @brief
 *
 * This API allows the application to print all ethernet related utility counters
 * maintained by the Ethernet ADK module.
 *
 * =============================================================
 */

adk_status_t
adk_eth_print_utility_counters(adk_module_handle_t     ethHandle)
{
    adk_sys_handle_t sys_handle;
    adk_status_t ret = ADK_SUCCESS;
    adk_eth_utility_counters_t utilityCounter;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(ethHandle))
    {   
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                ethHandle);
        ADK_ETH_LOG_FORCE("Failed to set ethernet utility counter status\n");
        return (ADK_E_INVALID_HANDLE);
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ethHandle);

    ADK_ETH_LOG_FUNC_ENTRY(ethHandle);
    
    ret = adk_eth_get_utility_counters(ethHandle, &utilityCounter);    
   
    if(ret != ADK_SUCCESS)
        return ADK_E_INTERNAL_ERROR;  

    printf("\n==========================Ethernet ingress utility counters=========================\n");
    printf("Total of ingress eth lli pkt received : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkt_proc_lli_entry_count);
    printf("Total of ingress eth lli pkt processing drop : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkt_proc_lli_drop);
    printf("Total of ingress eth lli non l2 term pkts received : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_non_l2term_frames_received);
    printf("Total of ingress physical terminating pdu's sent to netd  : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_phy_term_pkt_sent_to_netd);
    printf("Total of ingress pdu's sent to netd  : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkt_sent_to_netd);
    printf("Total of ingress pdu's sent to L3 module : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkts_sent_to_l3);
    printf("Total of ingress pdu's sent to IPv6 Module : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkts_sent_to_ipv6);
    printf("Total of ingress pdu's sent to IPv4 Module : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkts_sent_to_ipv4);
    printf("Total of ingress pdu's sent to policer stage : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkts_sent_to_policer_stage);
    printf("Total of ingress pdu's received in policer stage : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkts_recv_in_policer_stage);
    printf("Total of ingress pdu's vlan tag stripped : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkts_vlan_tag_stripped);
    printf("Total of ingress multicast pdu's received : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_multicast_packets_received);
    printf("Total of ingress pdu's sent for SW ingress flow lookup : %ld\n",
            utilityCounter.ingressUtilityCounters.eth_ing_pkt_sent_to_sw_ingress_flow_lookup);
    printf("==========================Ethernet egress utility counters=========================\n");
    printf("Total of egress pdu's received in eth ULI : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_proc_uli_entry_count);
    printf("Total of egress ipv4 pdu's received : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv4_pkts_received);
    printf("Total of egress IPv6 pdu's received  : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv6_pkts_received);
    printf("Total of egress IPv6 Multicast pdu's received : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv6_multicast_pkts_received);
    printf("Total of egress pdu's received from netd : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkts_received_from_netd);
    printf("Total of egress non bridge netd origin pdu's received : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_received);
    printf("Total of egress non bridge netd origin pdu's transmitted : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_transmitted);
    printf("Total of egress pdu's sent to nic: %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkts_sent_to_nic);
    printf("Total of egress pdu's sent to ph: %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkts_sent_to_ph);
    printf("Total of egress eth uli pkt processing drop : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_proc_uli_drop);
    printf("Total of egress pdu's ARP exception sent: %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_arp_exception_sent);
    printf("Total of egress pdu's ND exception sent: %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_nd_exception_sent);
    printf("Total of egress pdu's STALE Entry Used exception sent: %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_stale_entry_used_exception_sent);
    printf("Total of egress pdu's sent to buffering stage: %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_sent_to_buffering_stage);
    printf("Total of egress packets receive in buffering stage  : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_recv_in_buffering_stage);
    printf("Total of egress packets dropped in buffering stage : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkt_dropped_in_buffering_stage);
    printf("Total of egress nexthop buffer info free error : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error);
    printf("Total of egress nexthop buffer lock alloc error : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_nexthop_buffer_lock_alloc_error);
    printf("Total of egress ipv4 nexthop buffer overflow error : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_overflow_error);
    printf("Total of egress ipv4 nexthop buffer timer create failure : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_create_failure);
    printf("Total of egress ipv4 nexthop buffer timeout counter : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timeout_count);
    printf("Total of egress ipv4 nexthop buffer timer restart error : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_restart_error);
    printf("Total of egress ipv6 nexthop buffer overflow error : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_overflow_error);
    printf("Total of egress ipv6 nexthop buffer timer create failure : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_create_failure);
    printf("Total of egress ipv6 nexthop buffer timeout counter : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timeout_count);
    printf("Total of egress ipv6 nexthop buffer timer restart error : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_restart_error);
    printf("Total of egress pdu's timstamp alloc requests : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkts_tstamp_alloc_requests);
    printf("Total of egress pdu's timstamp free requests : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkts_tstamp_free_requests);
    printf("Total of egress pdu's timstamp alloc success : %ld\n",
            utilityCounter.egressUtilityCounters.eth_eg_pkts_tstamp_alloc_success);


    return ret;
}
