/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ethernet_flow.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernet  Ethernet Module
 * @{
 *      @addtogroup _ethernetFunctions  Ethernet Functions
 *      @{
 *
 * @brief   This files has the source code for the top module level 
 *          APIs that manage Ethernet flows. This includes Ethernet
 *          ingress flows, egress encap flows, etc.
 *
 *  **********************************************************************/

#include <adk_eth.h>

/*!
 * =============================================================
 *
 * adk_ethernet_ingress_flow_create
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   flow_data       The input fields for ingress flow classification
 * @param[in]   flow_action     The action to be taken for this ingress flow
 * @param[out]  flow_handle_p   The pointer to the ingress flow handle created
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE Invalid interface handle specified
 * @retval  ADK_E_INVALID_PARAMETER Pointer passed as NULL
 * @retval  ADK_E_RESOURCE_UNAVAILABLE  Out of resource to create this ingress
 *                                      flow
 * @retval  ADK_E_MEM_UNAVAILABLE   Out of memory to create this ingress flow
 *
 * @brief   Function to create ingress flows along with the flow action.
 *
 * This function is used to create an ingress Ethernet flow. The ingress flow is
 * used for filtering packets (drop or send to host). The default action is to
 * send L2 terminating packet to higher layer module based on ether type.
 *
 * =============================================================
 */

adk_status_t
adk_ethernet_ingress_flow_create (
    adk_module_handle_t eth_mod_handle,
    const adk_ethernet_ingress_flow_data_t *flow_data,
    const adk_ethernet_ingress_flow_action_t *flow_action,
    adk_flow_handle_t *flow_handle_p)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_flow_internal_data_t    *flow_data_p = NULL;
    adk_dp_eth_ingress_flow_t       *dp_ingress_flow_p = NULL;
    uint32_t flow_id;
    uint32_t    policerFlowId = ADK_QOS_INVALID_POLICER_FLOW_ID;
    uint64_t dmac_local, i; 

    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to create Ingress Flow Handle\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if ((flow_data == NULL) || (flow_action == NULL))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "NULL Flow Data or Action Pointer\n");
        goto _adk_ethernet_ingress_flow_create_exit;
    }

    if (flow_handle_p == NULL)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Null Flow Handle Pointer\n");
        goto _adk_ethernet_ingress_flow_create_exit;
    }

    *flow_handle_p = 0;

    ret = adk_eth_valid_ingress_flow_data(eth_mod_handle, flow_data);
    if (ret != ADK_SUCCESS)
        goto _adk_ethernet_ingress_flow_create_exit;

    ret = adk_eth_valid_ingress_flow_action(eth_mod_handle,
                                            flow_data,
                                            flow_action);
    if (ret != ADK_SUCCESS)
        goto _adk_ethernet_ingress_flow_create_exit;

    if(flow_action->enablePolicing)
    {
        ret = adk_qos_policing_flow_id_get(sys_handle, flow_action->policingHandle, &policerFlowId);
        if (ADK_SUCCESS != ret) {
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret, "Invalid policer handle\n");
            goto _adk_ethernet_ingress_flow_create_exit;
        }
    }
    flow_id = adk_rm_resource_reserve(eth_internal_data_p->eth_flows_ns_hdl);
    if (flow_id == ADK_RM_RSRC_INVALID)
    {
        ret = ADK_E_RESOURCE_UNAVAILABLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Could not allocate Flow ID For ingress flow\n");
        goto _adk_ethernet_ingress_flow_create_exit;
    }

    ADK_ETH_LOG_INFO(eth_mod_handle, "Allocated Ingress Flow NS ID 0x%08x\n",
                     flow_id);

    ADK_SYS_FLOW_HANDLE_ALLOC(eth_mod_handle, ADK_ETH_INGRESS_FLOW_OBJ,
                              sizeof(adk_eth_flow_internal_data_t),
                              flow_handle_p);
    if (*flow_handle_p == 0)
    {
        ADK_ETH_LOG_FATAL(eth_mod_handle, ADK_E_MEM_UNAVAILABLE,
                          "Memory allocation failed\n");
        ret = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ethernet_ingress_flow_create_exit;
    }

    flow_data_p = (adk_eth_flow_internal_data_t *)
                  ADK_SYS_FLOW_ATTR_PTR_GET(*flow_handle_p);

    ADK_SYS_FLOWID_SET(*flow_handle_p, flow_id);
    flow_data_p->cookie = ADK_ETHERNET_FLOW_COOKIE;
    flow_data_p->type = ETH_FLOW_TYPE_SIMPLE_INGRESS;

    ADK_MEMCPY((void *) &(flow_data_p->u.ingress_flow),
               (const void *) flow_data,
               sizeof(adk_ethernet_ingress_flow_data_t));
    ADK_MEMCPY((void *) &(flow_data_p->flow_action),
               (const void *) flow_action,
               sizeof(adk_ethernet_ingress_flow_action_t));

    /* Create DP table element for the sub-interface */
    dp_ingress_flow_p = (adk_dp_eth_ingress_flow_t *)
                        ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(eth_mod_handle), ADK_ETH_INGRESS_FLOW_OBJ,
					 1, sizeof(adk_dp_eth_ingress_flow_t),
                                          ADK_ETH_DP_OBJECT_ALIGN);
    if (dp_ingress_flow_p == NULL)
    {
        ADK_ETH_LOG_FATAL(eth_mod_handle, ADK_E_MEM_UNAVAILABLE,
                          "Memory allocation failed\n");
        ret = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ethernet_ingress_flow_create_exit;
    }

    dp_ingress_flow_p->ether_type = flow_data->proto;
    ADK_MEMCPY((void *) &dp_ingress_flow_p->dmac_subnet, (void *) &flow_data->dest_mac,
               sizeof(adk_ethernet_mac_subnet_t));
    dp_ingress_flow_p->flowId = flow_id;
    dp_ingress_flow_p->actionType = flow_data_p->flow_action.action_type;
    dp_ingress_flow_p->enablePolicing = flow_action->enablePolicing;
    dp_ingress_flow_p->policerId = policerFlowId;
    dp_ingress_flow_p->intfType = flow_data->intfType;

    /* Hook up the DP table element with CP Flow struct */
    flow_data_p->dp_u.dp_ing_flow_p = dp_ingress_flow_p;
    eth_internal_data_p->dp_pvt_p->eth_ingress_flow_list[flow_id] = dp_ingress_flow_p; 

    dp_ret = adk_eth_dp_ingress_flow(eth_mod_handle,
                                     ADK_ETH_TREE_OPER_ADD,
                                     flow_data_p);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to add entry to flow table.\n");
        goto _adk_ethernet_ingress_flow_create_exit;
    }


    adk_hos_l2_ingress_tbl_rule_attr_t hos_l2_ingress_rule = 
    {
        .l2_domain = *(eth_internal_data_p->dp_pvt_p->default_l2_domain),
        .ether_type = flow_data->proto,
        .l2_ingress = flow_id 
    };

    if (flow_data->mask & ADK_ETHERNET_FLOW_MASK_DMAC) 
    {
        dmac_local =  (flow_data->dest_mac.mac)[0];
        dmac_local = (dmac_local << 8) | (flow_data->dest_mac.mac)[1];
        dmac_local = (dmac_local << 8) | (flow_data->dest_mac.mac)[2];
        dmac_local = (dmac_local << 8) | (flow_data->dest_mac.mac)[3];
        dmac_local = (dmac_local << 8) | (flow_data->dest_mac.mac)[4];
        dmac_local = (dmac_local << 8) | (flow_data->dest_mac.mac)[5];
        hos_l2_ingress_rule.dmac = dmac_local;
        i = (uint64_t)1 << flow_data->dest_mac.num_valid_msb_bits;
        hos_l2_ingress_rule.dmac_mask = ((i - 1) << (MAC_ADDR_WIDTH - flow_data->dest_mac.num_valid_msb_bits));
    }
    else
    {
        hos_l2_ingress_rule.dmac = 0;
        hos_l2_ingress_rule.dmac_mask = 0;
    }

    ret = adk_hos_rule_add(ADK_HOS_L2_INGRESS_TBL,
                               (adk_hos_rule_attr_t *) &hos_l2_ingress_rule,
                               &flow_data_p->hos_l2_ingress_rule_id);

    /* HOS rule addition can fail when table is full*/
    if (ret != ADK_SUCCESS)
    {
        if(ret == ADK_E_HOS_TABLE_FULL)
        { 
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                              "Failed to add L2 ingress flow HOS rule."
                              " Logical space is not available\n");
            ret = ADK_E_RESOURCE_UNAVAILABLE;
            flow_data_p->hos_l2_ingress_rule_id = 0;
        }
        else
        {
            ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to add L2 ingress flow HOS rule\n");
            flow_data_p->hos_l2_ingress_rule_id = 0;
        }
    }

_adk_ethernet_ingress_flow_create_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Created Ingress Flow Handle %"ADKPRIh "\n",
                         *flow_handle_p);
    }
    else
    {
        if (flow_data_p && *flow_handle_p)
        {
            flow_data_p->cookie = 0;
            ADK_SYS_FLOW_HANDLE_FREE(*flow_handle_p);
            if (dp_ingress_flow_p)
                ADK_DP_SHM_FREE(dp_ingress_flow_p);
            *flow_handle_p = 0x0;
        }

        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to create Ingress Flow Handle\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}


/*!
 * =============================================================
 *
 * adk_ethernet_ingress_flow_delete
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   flow_handle     The ingress flow handle to be deleted
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_FLOW_HANDLE Invalid flow handle specified
 * @retval  ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE  Unexpected flow type
 *
 * @brief   Function to delete ingress flow
 *
 * This function is used to delete an ingress Ethernet flow. This function
 * deallocates the reserved resources and frees up the internal data
 * structure for the flow.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_ingress_flow_delete (
    adk_module_handle_t eth_mod_handle,
    adk_flow_handle_t flow_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_flow_internal_data_t    *flow_data_p;
    adk_dp_eth_ingress_flow_t       *dp_ingress_flow_p;
    uint32_t flow_id;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_ret = ADK_SUCCESS;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to delete Ingress Flow Handle %"ADKPRIh "\n",
                          flow_handle);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if (!adk_eth_valid_flow_handle(flow_handle))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Flow handle %"ADKPRIh "\n", flow_handle);
        goto _adk_ethernet_ingress_flow_delete_exit;
    }

    flow_data_p = (adk_eth_flow_internal_data_t *)
                  ADK_SYS_FLOW_ATTR_PTR_GET(flow_handle);

    if (flow_data_p->type != ETH_FLOW_TYPE_SIMPLE_INGRESS)
    {
        ret = ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Incompatible flow type %s\n",
                          adk_eth_flow_type_name(flow_data_p->type));
        goto _adk_ethernet_ingress_flow_delete_exit;
    }

    flow_id = ADK_SYS_FLOWID_GET(flow_handle);

    dp_ingress_flow_p = flow_data_p->dp_u.dp_ing_flow_p;

    dp_ret = adk_eth_dp_ingress_flow(eth_mod_handle,
                                     ADK_ETH_TREE_OPER_DEL,
                                     flow_data_p);
    if (dp_ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Deleting entry from flow table failed.\n");
        goto _adk_ethernet_ingress_flow_delete_exit;
    }

    if(flow_data_p->hos_l2_ingress_rule_id)
    {
        ret = adk_hos_rule_delete(ADK_HOS_L2_INGRESS_TBL,
                               flow_data_p->hos_l2_ingress_rule_id);
    }

    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to delete Ingress flow ID HOS rule\n");
        goto _adk_ethernet_ingress_flow_delete_exit;
    }

    adk_rcu_distributed_ver_update();

    adk_eth_list_add(eth_internal_data_p, ADK_ETH_ING_FLOW_ID_LIST_ID,
                     (uintptr_t) flow_id);

    adk_eth_list_add(eth_internal_data_p, ADK_ETH_INGRESS_FLOW_LIST_ID,
                     (uintptr_t) dp_ingress_flow_p);

    eth_internal_data_p->dp_pvt_p->eth_ingress_flow_list[flow_id] = NULL;


_adk_ethernet_ingress_flow_delete_exit:

    if (ret == ADK_SUCCESS)
    {
        flow_data_p->cookie = 0;
        ADK_SYS_FLOW_HANDLE_FREE(flow_handle);
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Deleted Ingress Flow Handle %"ADKPRIh "\n", flow_handle);
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to delete Ingress Flow Handle %"ADKPRIh "\n",
                          flow_handle);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

/*!
 * =============================================================
 *
 * adk_ethernet_ingress_flow_statistics_get
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   flow_handle     The ingress flow handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_E_INVALID_PARAMETER Invalid parameter
 * @retval  ADK_ETHERNET_E_INVALID_FLOW_HANDLE Invalid flow handle specified
 * @retval  ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE  Unexpected flow type
 *
 * @brief   Function to get ingress flow statistics
 *
 * This function is used to get the ingress Ethernet flow statistics.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_ingress_flow_statistics_get (
    adk_module_handle_t eth_mod_handle,
    adk_flow_handle_t flow_handle,
    adk_ethernet_ingress_flow_stats_t* flow_stats)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_internal_data_t    *eth_internal_data_p = NULL;
    adk_eth_flow_internal_data_t    *flow_data_p;
    adk_dp_eth_ingress_flow_t       *dp_ingress_flow_p;
    adk_sys_handle_t sys_handle;
    adk_dp_eth_ingress_flow_stats_t* dp_flow_stats;
    uint32_t i = 0;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to delete Ingress Flow Handle %"ADKPRIh "\n",
                          flow_handle);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    eth_internal_data_p = ADK_MODULE_INFO_PTR_GET(eth_mod_handle);

    if (adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
    {
        adk_eth_lists_clear(eth_internal_data_p);
        ADK_ETH_LOG_INFO(eth_mod_handle, "All ethernet lists cleared\n");
    }

    if (NULL == flow_stats)
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Flow handle %"ADKPRIh "\n", flow_handle);
        goto _adk_ethernet_ingress_flow_statistics_get_exit;
    }

    ADK_MEMSET(flow_stats, 0, sizeof(adk_ethernet_ingress_flow_stats_t));

    if (!adk_eth_valid_flow_handle(flow_handle))
    {
        ret = ADK_ETHERNET_E_INVALID_FLOW_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Flow handle %"ADKPRIh "\n", flow_handle);
        goto _adk_ethernet_ingress_flow_statistics_get_exit;
    }

    flow_data_p = (adk_eth_flow_internal_data_t *)
                  ADK_SYS_FLOW_ATTR_PTR_GET(flow_handle);

    if (flow_data_p->type != ETH_FLOW_TYPE_SIMPLE_INGRESS)
    {
        ret = ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Incompatible flow type %s\n",
                          adk_eth_flow_type_name(flow_data_p->type));
        goto _adk_ethernet_ingress_flow_statistics_get_exit;
    }

    dp_ingress_flow_p = flow_data_p->dp_u.dp_ing_flow_p;

    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        /* accumulate for all cores */
        dp_flow_stats = &(dp_ingress_flow_p->flow_stats[i]);
        flow_stats->rx_pkts += dp_flow_stats->numPkts;
        flow_stats->pkts_sent_to_host += dp_flow_stats->pkts_sent_to_host;
        flow_stats->pkts_sent_to_ipv4 += dp_flow_stats->pkts_sent_to_ipv4;
        flow_stats->pkts_sent_to_ipv6 += dp_flow_stats->pkts_sent_to_ipv6;
        flow_stats->pkts_drop += dp_flow_stats->pkts_drop;
    }
    
_adk_ethernet_ingress_flow_statistics_get_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Get Ingress Flow statistics successful\n");
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get Ingress Flow statistics for flow handle %"ADKPRIh "\n",
                          flow_handle);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return ret;
}
#ifdef ADK_DEBUG
/*!
 * =============================================================
 *
 * adk_ethernet_ingress_flow_statistics_print
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   flow_handle     The ingress flow handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_FLOW_HANDLE Invalid flow handle specified
 * @retval  ADK_ETHERNET_E_INCOMPAT_FLOW_TYPE  Unexpected flow type
 *
 * @brief   Function to print ingress flow statistics
 *
 * This function is used to get the ingress Ethernet flow statistics.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_ingress_flow_statistics_print (
    adk_module_handle_t eth_mod_handle,
    adk_flow_handle_t flow_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_ingress_flow_stats_t flow_stats;

    ADK_MEMSET(&flow_stats, 0, sizeof(adk_ethernet_ingress_flow_stats_t));

    /* Get the interface statistics using Ethernet API */
    ret = adk_ethernet_ingress_flow_statistics_get(eth_mod_handle, flow_handle,
                                                   &flow_stats);
    if (ret != ADK_SUCCESS)
        return (ret);

    ADK_INFO_WRITE("\nEthernet ingress flow Stats:"
                   " (Drop counters with zero count not shown):\n\n");

    /* Rx Counters */

    ADK_INFO_WRITE("\tPkts Received:                %" PRId64 "\n",
                   flow_stats.rx_pkts);

    ADK_ETH_NON_ZERO_CNTR_WRITE("\tPkts sent to host:      %" PRId64 "\n",
                   flow_stats.pkts_sent_to_host);

    ADK_ETH_NON_ZERO_CNTR_WRITE("\tPkts sent to ipv4:      %" PRId64 "\n",
            flow_stats.pkts_sent_to_ipv4);

    ADK_ETH_NON_ZERO_CNTR_WRITE("\tPkts sent to ipv6:      %" PRId64 "\n",
            flow_stats.pkts_sent_to_ipv6);

    ADK_ETH_NON_ZERO_CNTR_WRITE("\tPkts drop for ingress flow action drop:      %" PRId64 "\n",
                                flow_stats.pkts_drop);

    return ret;
}
#endif

/*!     end addtogroup _ethernetFunction
 *      @}
 *  end addtogroup _ethernet
 * @}
 */
