/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ethernet_stats.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernet  Ethernet Module
 * @{
 *      @addtogroup _ethernetFunctions  Ethernet Functions
 *      @{
 *
 * @brief   This files has the source code for the top module level
 *          APIs to get and print statistics. This include module
 *          level global statistics, interface level and flow level
 *          statistics.
 *
 *  **********************************************************************/

#include <adk_eth.h>
/*!
 * =============================================================
 *
 * adk_ethernet_global_statistics_get
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[out]  stats           The pointer to the structure in which the
 *                              retrieved global statistics are written
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_E_INVALID_PARAMETER Pointer passed as NULL
 *
 * @brief   Function to get the global statistics for Ethernet module
 *
 * This function gets all the global statistics maintained by the
 * Ethernet module.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_global_statistics_get (
    adk_module_handle_t eth_mod_handle,
    adk_ethernet_global_statistics_t *stats)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get global stats\n");
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    if (stats == NULL)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Null stats pointer parameter\n");
        goto _adk_ethernet_global_statistics_get_exit;
    }

    ret = adk_eth_dp_global_stats_get(eth_mod_handle, stats);
    if (ret != ADK_SUCCESS)
    {
        goto _adk_ethernet_global_statistics_get_exit;
    }

_adk_ethernet_global_statistics_get_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Retrieved the Ethernet Global Stats\n");
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get Ethernet Global Stats\n");
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

#ifdef ADK_DEBUG
/*!
 * =============================================================
 *
 * adk_ethernet_global_stats_print
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Helper function to print all the global statistics
 *
 * This function is used to dump all the global statistics maintained by
 * Ethernet module.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_global_stats_print (
    adk_module_handle_t eth_mod_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_global_statistics_t gl_stats;

    ADK_MEMSET(&gl_stats, 0, sizeof(adk_ethernet_global_statistics_t));

    /* No need to get semaphores or log entry exit function as they will be done
     * in below API call */

    /* Get the interface statistics using Ethernet API */
    ret = adk_ethernet_global_statistics_get(eth_mod_handle, &gl_stats);
    if (ret != ADK_SUCCESS)
        return (ret);

    ADK_INFO_WRITE("\nEthernet Global Stats"
                   " (Drop counters with zero count not shown):\n\n");

    /* Drop Counters */
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tUnknown Port:              \t%10u\n",
                                gl_stats.port_not_defined);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tUnknown Interface:    \t%10u\n",
                                gl_stats.intf_not_defined);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEgress time index allocation failure:  \t%10u\n",
                                gl_stats.time_index_alloc_error);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tTime Stamp tx complete notification failed:    \t%10u\n", 
                                gl_stats.ts_tx_complete_notification_failed);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tPacket clone failures:    \t%10u\n",
                                gl_stats.pkt_clone_failure);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthernet Ingress lookup failure:    \t%10u\n",
                                gl_stats.ingressflow_lookup_failure);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthernet Ingress flow Drops:    \t%10u\n",
                                gl_stats.ingressflowDrops);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthernet Ingress ratelimiter drops:    \t%10u\n",
                                gl_stats.ethPacketsRateLimitingDrops);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tFull packet roll over discard:    \t%10u\n",
                                gl_stats.arp_buffer_full_pkt_rollover_discard);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tFragment packet roll over discard:    \t%10u\n",
                                gl_stats.arp_buffer_fragment_pkt_rollover_discard);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tFragment packet roll over discard from ND buffer stage:    \t%10u\n",
                                gl_stats.nd_buffer_fragment_pkt_rollover_discard);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tFull packet roll over discard from ND buffer stage :    \t%10u\n",
                                gl_stats.nd_buffer_full_pkt_rollover_discard);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t exception drop:    \t%10u\n",
                                gl_stats.exceptionDrops);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tBuffered packet drop due to arp entry remove:    \t%10u\n",
                                gl_stats.arp_buffer_arp_entry_rem_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tND Buffered packet drop due to neigh entry remove:    \t%10u\n",
                                gl_stats.nd_buffer_neigh_entry_rem_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tNeighbor unreachable, hence packet drop:    \t%10u\n",
                                gl_stats.neigh_state_unreachable_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tNon L2 Termincated traffic drop:    \t%10u\n",
                                gl_stats.non_l2term_frames_discard);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tNonIP etherType to queue map not found drop :    \t%10u\n",
                                gl_stats.ethtype_queue_map_not_found_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tNon bridge orginated traffic received from netd got Discarded:    \t%10u\n",
                                gl_stats.netd_origin_non_bridge_frames_discard);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEgress packets unknown output interface drop:    \t%10u\n",
                                gl_stats.unknown_output_interface_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Packet buffering memory not available drop:    \t%10u\n",
                                gl_stats.buffering_memory_not_available_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Egress packets nic send failure drop:    \t%10u\n",
                                gl_stats.nic_send_failure_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Nexthop Buffer stage is not found drop:    \t%10u\n",
                                gl_stats.nexthop_buffer_stage_not_found_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Nexthop Buffer processing drop:    \t%10u\n",
                                gl_stats.nexthop_buffer_proc_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Nexthop Buffer timeout error:    \t%10u\n",
                                gl_stats.nexthop_buffer_timeout_error);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Ethernet failed to send packets to next stage drop:    \t%10u\n",
                                gl_stats.failed_to_sent_next_stage_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Ethernet ingress pkts failed to strip vlan tag drop:    \t%10u\n",
                                gl_stats.failed_to_sent_next_stage_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Ethernet egress pkts failed to encap drop:    \t%10u\n",
                                gl_stats.egress_encap_pkt_failure_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\t Ethernet ingress unknown multicast address pkts drop:    \t%10u\n",
                                gl_stats.unknown_multicast_addr_drop);

    return (ret);
}
#endif
/*!
 * =============================================================
 *
 * adk_ethernet_port_statistics_get
 *
 * @param[in]   eth_mod_handle  Handle to the Ethernet module
 * @param[in]   port            External port or CPK port
 * @param[out]  stats           The pointer to the structure in which the
 *                              retrieved statistics are written
 * @param[in]   clear_on_read   The boolean flag, if set, clears the statistics
 *                              in the hardware upon read
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
 * @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE Invalid interface handle specified
 * @retval  ADK_E_INVALID_PARAMETER Pointer passed as NULL
 * @retval  ADK_ETHERNET_E_INCOMPAT_IF_TYPE Operation not permitted on passed
 * interface type
 *
 * @brief   Function to get the statistics for a physical ethernet port
 *
 * This function gets the HLP port level Ethernet statistics. These are
 * retrieved from the HLP hardware.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_port_statistics_get (
    adk_module_handle_t eth_mod_handle,
    uint32_t port,
    adk_ethernet_port_statistics_t *stats,
    bool clear_on_read)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE("Failed to get port stats for port %d\n",
                          port);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    if (port > ADK_HOS_CPK_PORT)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid port %d\n", port);
        goto _adk_ethernet_port_statistics_get_exit;
    }

    if (stats == NULL)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Null stats pointer parameter\n");
        goto _adk_ethernet_port_statistics_get_exit;
    }

    if (clear_on_read)
    {
        ret = ADK_E_UNSUPPORTED;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Clear-On-Read Stats Not Supported for Interface"
                          " %d\n", port);
        goto _adk_ethernet_port_statistics_get_exit;
    }

    ADK_MEMSET(stats, 0x0, sizeof(adk_ethernet_port_statistics_t));

    ret = adk_hos_mac_stats_get(port,
                                (adk_hos_port_statistics_t *) stats);
    if (ret != ADK_SUCCESS)
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get port %d stats from hardware!\n",
                          port);
        goto _adk_ethernet_port_statistics_get_exit;
    }

_adk_ethernet_port_statistics_get_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Retrieved the port stats for port %d\n",
                         port);
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get port stats for port %d\n",
                          port);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}
#ifdef ADK_DEBUG
/*!
 * =============================================================
 *
 * adk_ethernet_port_stats_print
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 * @param[in]   port            Physical port or CPK port
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Helper function to print statistics for any physical HLP port
 *
 * This function is used to dump the statistics for any physical interface.
 * Calling this function does not clear the stats for that HLP port.
 *
 * =============================================================
 */
adk_status_t
adk_ethernet_port_stats_print (
    adk_module_handle_t eth_mod_handle,
    uint32_t port)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_port_statistics_t port_stats;

    /* No need to get semaphores or log entry exit function as they will be done
     * in below API call */

    ADK_MEMSET(&port_stats, 0, sizeof(adk_ethernet_port_statistics_t));
    /* Get the interface statistics using Ethernet API */
    ret = adk_ethernet_port_statistics_get(eth_mod_handle, port,
                                           &port_stats, FALSE);
    if (ret != ADK_SUCCESS)
    {
        return (ret);
    }

    ADK_INFO_WRITE("\nEthernet Port Stats (Counters with non-zero count"
                   " only except for tx_frm and rx_frm):\n\n");

    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_octets:            \t%20lu\n",
                                port_stats.tx_octets);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_bad_octets         \t%20lu\n",
                                port_stats.tx_bad_octets);
    ADK_INFO_WRITE("\ttx_frm                \t%20lu\n",
                   port_stats.tx_frm);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_bad_frm            \t%20lu\n",
                                port_stats.tx_bad_frm);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_mcast              \t%20lu\n",
                                port_stats.tx_mcast);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_bcast              \t%20lu\n",
                                port_stats.tx_bcast);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_pause              \t%20lu\n",
                                port_stats.tx_pause);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_fcs                \t%20lu\n",
                                port_stats.tx_fcs);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_defer              \t%20lu\n",
                                port_stats.tx_defer);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_xsdfr              \t%20lu\n",
                                port_stats.tx_xsdfr);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_1col               \t%20lu\n",
                                port_stats.tx_1col);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_mcol               \t%20lu\n",
                                port_stats.tx_mcol);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_lcol               \t%20lu\n",
                                port_stats.tx_lcol);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_xscol              \t%20lu\n",
                                port_stats.tx_xscol);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_frag               \t%20lu\n",
                                port_stats.tx_frag);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_usize              \t%20lu\n",
                                port_stats.tx_usize);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_64                 \t%20lu\n",
                                port_stats.tx_64);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_65                 \t%20lu\n",
                                port_stats.tx_65);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_128                \t%20lu\n",
                                port_stats.tx_128);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_256                \t%20lu\n",
                                port_stats.tx_256);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_512                \t%20lu\n",
                                port_stats.tx_512);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_1024               \t%20lu\n",
                                port_stats.tx_1024);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_1519               \t%20lu\n",
                                port_stats.tx_1519);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_osize              \t%20lu\n",
                                port_stats.tx_osize);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_jab                \t%20lu\n",
                                port_stats.tx_jab);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_urun               \t%20lu\n",
                                port_stats.tx_urun);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\ttx_stat_ovfl          \t%20lu\n",
                                port_stats.tx_stat_ovfl);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_octets             \t%20lu\n",
                                port_stats.rx_octets);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_bad_octets         \t%20lu\n",
                                port_stats.rx_bad_octets);
    ADK_INFO_WRITE("\trx_frm                \t%20lu\n",
                   port_stats.rx_frm);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_bad_frm            \t%20lu\n",
                                port_stats.rx_bad_frm);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_mcast              \t%20lu\n",
                                port_stats.rx_mcast);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_bcast              \t%20lu\n",
                                port_stats.rx_bcast);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_ctl                \t%20lu\n",
                                port_stats.rx_ctl);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_pause              \t%20lu\n",
                                port_stats.rx_pause);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_unk_op             \t%20lu\n",
                                port_stats.rx_unk_op);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_fcs                \t%20lu\n",
                                port_stats.rx_fcs);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_align              \t%20lu\n",
                                port_stats.rx_align);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_length             \t%20lu\n",
                                port_stats.rx_length);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_code               \t%20lu\n",
                                port_stats.rx_code);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_frag               \t%20lu\n",
                                port_stats.rx_frag);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_usize              \t%20lu\n",
                                port_stats.rx_usize);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_64                 \t%20lu\n",
                                port_stats.rx_64);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_65                 \t%20lu\n",
                                port_stats.rx_65);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_128                \t%20lu\n",
                                port_stats.rx_128);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_256                \t%20lu\n",
                                port_stats.rx_256);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_512                \t%20lu\n",
                                port_stats.rx_512);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_1024               \t%20lu\n",
                                port_stats.rx_1024);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_1519               \t%20lu\n",
                                port_stats.rx_1519);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_osize              \t%20lu\n",
                                port_stats.rx_osize);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_jab                \t%20lu\n",
                                port_stats.rx_jab);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_drop               \t%20lu\n",
                                port_stats.rx_drop);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_carrier            \t%20lu\n",
                                port_stats.rx_carrier);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_urun               \t%20lu\n",
                                port_stats.rx_urun);
    ADK_ETH_NON_ZERO_CNTR_WRITE("\trx_stat_ovfl          \t%20lu\n",
                                port_stats.rx_stat_ovfl);

    return (ret);
}
#endif

/*!
* =============================================================
*
* adk_ethernet_bridge_interface_statistics_get
*
* @param[in]   eth_mod_handle  Handle to the Ethernet module
* @param[in]   if_handle       The interface handle of bridge for which the statistics
*                              need to be retrieved.
* @param[out]  stats           The pointer to the structure in which the
*                              retrieved statistics are written
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE Invalid interface handle specified
* @retval  ADK_E_INVALID_PARAMETER Pointer passed as NULL
* @retval  ADK_ETHERNET_E_INVALID_IF_TYPE Interface type not supported for this
* operation
*
* @brief   Function to get the statistics for an Ethernet interface
*
* This function gets the statistics for a bridge interface maintained in the ADK. These include the number of
* packet received and transmitted and the number of packets dropped.
*
* =============================================================
*/
adk_status_t
adk_ethernet_bridge_interface_statistics_get (
    adk_module_handle_t eth_mod_handle,
    adk_interface_handle_t if_handle,
    adk_ethernet_bridge_interface_statistics_t *stats)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_eth_intf_internal_data_t    *if_data_p = NULL;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE(
            "Failed to get interface stats for interface %"ADKPRIh "\n",
            if_handle);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    if (!adk_eth_valid_if_handle(if_handle))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Interface handle %"ADKPRIh "\n", if_handle);
        goto _adk_ethernet_bridge_interface_statistics_get_exit;
    }

    if_data_p = (adk_eth_intf_internal_data_t *)
        ADK_SYS_INTERFACE_ATTR_PTR_GET(if_handle);

    if (stats == NULL)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Null stats pointer parameter\n");
        goto _adk_ethernet_bridge_interface_statistics_get_exit;
    }

    if ((if_data_p->type != INTERFACE_TYPE_BRIDGE))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Incompatible Interface type %s\n",
                          adk_eth_intf_type_name(if_data_p->type));
        goto _adk_ethernet_bridge_interface_statistics_get_exit;
    }

    ret = adk_eth_dp_interface_stats_get(eth_mod_handle, ADK_SYS_INTERFACE_ID_GET(if_handle),
                                            if_data_p, stats);
    if (ret != ADK_SUCCESS)
        goto _adk_ethernet_bridge_interface_statistics_get_exit;


_adk_ethernet_bridge_interface_statistics_get_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Retrieved the interface stats for interface %"ADKPRIh "\n",
                         if_handle);
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get interface stats for interface %"ADKPRIh "\n",
                          if_handle);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);

}
/*!
* =============================================================
*
* adk_ethernet_vlan_interface_statistics_get
*
* @param[in]   eth_mod_handle  Handle to the Ethernet module
* @param[in]   if_handle       The interface handle of switch virtual interface for which the statistics
*                              need to be retrieved. 
 * @param[out]  stats           The pointer to the structure in which the
*                              retrieved statistics are written
*
* @retval  ADK_SUCCESS             Successful completion
* @retval  ADK_E_INVALID_HANDLE    Invalid Ethernet module handle specified
* @retval  ADK_ETHERNET_E_INVALID_IF_HANDLE Invalid interface handle specified
* @retval  ADK_E_INVALID_PARAMETER Pointer passed as NULL
* @retval  ADK_ETHERNET_E_INVALID_IF_TYPE Interface type not supported for this
* operation
*
* @brief   Function to get the statistics for an Ethernet interface
*
* This function gets the statistics for a switch virtual interface maintained in the ADK. These include the number of
* packet received and transmitted and the number of packets dropped.
*
* =============================================================
*/
adk_status_t
adk_ethernet_vlan_interface_statistics_get (
    adk_module_handle_t eth_mod_handle,
    adk_interface_handle_t if_handle,
    adk_ethernet_vlan_interface_statistics_t *stats)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_eth_intf_internal_data_t    *if_data_p = NULL;
    adk_sys_handle_t sys_handle;

    /* Validate the passed handles */
    if (!adk_eth_valid_handle(eth_mod_handle))
    {
        ADK_ETH_LOG_FORCE("Invalid Ethernet module handle %"ADKPRIh "\n",
                          eth_mod_handle);
        ADK_ETH_LOG_FORCE(
            "Failed to get interface stats for interface %"ADKPRIh "\n",
            if_handle);
        return (ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(eth_mod_handle);

    ADK_ETH_LOG_FUNC_ENTRY(eth_mod_handle);

    ADK_ETH_ACQUIRE_MODULE_LOCK(eth_mod_handle, ret);

    if (!adk_eth_valid_if_handle(if_handle))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_HANDLE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Invalid Interface handle %"ADKPRIh "\n", if_handle);
        goto _adk_ethernet_vlan_interface_statistics_get_exit;
    }

    if_data_p = (adk_eth_intf_internal_data_t *)
        ADK_SYS_INTERFACE_ATTR_PTR_GET(if_handle);

    if (stats == NULL)
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Null stats pointer parameter\n");
        goto _adk_ethernet_vlan_interface_statistics_get_exit;
    }

    if ((if_data_p->type != INTERFACE_TYPE_SWITCH_VIRTUAL))
    {
        ret = ADK_ETHERNET_E_INVALID_IF_TYPE;
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Incompatible Interface type %s\n",
                          adk_eth_intf_type_name(if_data_p->type));
        goto _adk_ethernet_vlan_interface_statistics_get_exit;
    }

    ret = adk_eth_dp_interface_stats_get(eth_mod_handle, ADK_SYS_INTERFACE_ID_GET(if_handle), 
                                            if_data_p, stats);
    if (ret != ADK_SUCCESS)
        goto _adk_ethernet_vlan_interface_statistics_get_exit;


_adk_ethernet_vlan_interface_statistics_get_exit:

    if (ret == ADK_SUCCESS)
    {
        ADK_ETH_LOG_INFO(eth_mod_handle,
                         "Retrieved the interface stats for interface %"ADKPRIh "\n",
                         if_handle);
    }
    else
    {
        ADK_ETH_LOG_ERROR(eth_mod_handle, ret,
                          "Failed to get interface stats for interface %"ADKPRIh "\n",
                          if_handle);
    }

    ADK_ETH_RELEASE_MODULE_LOCK(eth_mod_handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_ETH_LOG_FUNC_EXIT(eth_mod_handle);

    return (ret);
}

#if ADK_DEBUG
/*!
 * =============================================================
 *
 * adk_ethernet_bridge_interface_statistics_print
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 * @param[in]   if_handle       Interface handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Helper function to print statistics for any interface
 *
 * This function is used to dump the bridge statistics for any interface.
 *
 * =============================================================
 */

adk_status_t
adk_ethernet_bridge_interface_statistics_print (
    adk_module_handle_t eth_mod_handle,
    adk_interface_handle_t if_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_bridge_interface_statistics_t bridge_intf_stats;

    ADK_MEMSET(&bridge_intf_stats, 0, sizeof(adk_ethernet_bridge_interface_statistics_t));

    /* Get the interface statistics using Ethernet API */
    ret = adk_ethernet_bridge_interface_statistics_get(eth_mod_handle, if_handle,
                                                   &bridge_intf_stats);
    if (ret != ADK_SUCCESS)
        return (ret);

    ADK_INFO_WRITE("\nEthernet bridge Interface Stats:"
                   " (Drop counters with zero count not shown):\n\n");



    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthIfInUnknownVlan:  %" PRId32 "\n",
                                bridge_intf_stats.inUnknownVlan);

    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthIfInEthTypeMismatch: %" PRId32 "\n",
                                bridge_intf_stats.inEthTypeMismatch);

    ADK_INFO_WRITE("\tEthIfInFrames: %" PRId64 "\n",
                   bridge_intf_stats.inFrames);

    ADK_INFO_WRITE("\tEthIfInOctets: %" PRId64 "\n",
                   bridge_intf_stats.inOctets);

    ADK_INFO_WRITE("\tEthIfOutFrames : %" PRId64 "\n",
                   bridge_intf_stats.OutFrames);

    ADK_INFO_WRITE("\tEthIfOutOctets:  %" PRId64 "\n",
                   bridge_intf_stats.OutOctets);

    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthIfInPrioTaggedFrames:  %" PRId32 "\n",
                                bridge_intf_stats.inPrioTaggedFrames);
    /* Rx Mcast/Bcast Counters */
    ADK_INFO_WRITE("\tEthIfInMcastFrames:  %" PRId64 "\n",
                   bridge_intf_stats.inMcastFrames);

    /* Tx Mcast/Bcast Counters */
    ADK_INFO_WRITE("\tEthOutMcastFrames:   %" PRId64 "\n",
                   bridge_intf_stats.OutMcastFrames);

    return (ret);
}
/*!
 * =============================================================
 *
 * adk_ethernet_vlan_interface_statistics_print
 *
 * @param[in]   eth_mod_handle  Ethernet module handle
 * @param[in]   if_handle       Interface handle
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Helper function to print statistics for any interface
 *
 * This function is used to dump the vlan statistics for any interface.
 *
 * =============================================================
 */

adk_status_t
adk_ethernet_vlan_interface_statistics_print (
    adk_module_handle_t eth_mod_handle,
    adk_interface_handle_t if_handle)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_vlan_interface_statistics_t vlan_intf_stats;

    ADK_MEMSET(&vlan_intf_stats, 0, sizeof(adk_ethernet_vlan_interface_statistics_t));

    /* Get the interface statistics using Ethernet API */
    ret = adk_ethernet_vlan_interface_statistics_get(eth_mod_handle, if_handle,
                                                   &vlan_intf_stats);
    if (ret != ADK_SUCCESS)
        return (ret);

    ADK_INFO_WRITE("\nEthernet vlan Interface Stats:"
                   " (Drop counters with zero count not shown):\n\n");


    ADK_ETH_NON_ZERO_CNTR_WRITE("\tEthIfInEthTypeMismatch: %" PRId32 "\n",
                                vlan_intf_stats.inEthTypeMismatch);

    ADK_INFO_WRITE("\tEthIfInFrames: %" PRId64 "\n",
                   vlan_intf_stats.inFrames);

    ADK_INFO_WRITE("\tEthIfInOctets: %" PRId64 "\n",
                   vlan_intf_stats.inOctets);

    ADK_INFO_WRITE("\tEthIfOutFrames : %" PRId64 "\n",
                   vlan_intf_stats.OutFrames);

    ADK_INFO_WRITE("\tEthIfOutOctets:  %" PRId64 "\n",
                   vlan_intf_stats.OutOctets);

    /* Rx Mcast/Bcast Counters */
    ADK_INFO_WRITE("\tEthIfInMcastFrames:  %" PRId64 "\n",
                   vlan_intf_stats.inMcastFrames);

    /* Tx Mcast/Bcast Counters */
    ADK_INFO_WRITE("\tEthOutMcastFrames:   %" PRId64 "\n",
                   vlan_intf_stats.OutMcastFrames);

    return (ret);
}
#endif

/*!     end addtogroup _ethernetFunction
 *      @}
 *  end addtogroup _ethernet
 * @}
 */
