/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp_bm.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file contains the Ethernet data plane software for batch
 *          processing.
 *
 *  **********************************************************************/

#include "adk_sif_dp_log_lib.h"
#include <adk_eth_dp_pvt.h>

#ifndef ADK_DP_NO_BATCH_PROC

/*!
 * =============================================================
 * adk_eth_dp_process_pkt
 *
 * Structure of task params:
 *
 * taskParams[0] : Incoming port number in upper 8-bits.
 * taskParams[7] : PDU buffer pointing to Ethernet header.
 *
 * Brief:
 *
 * This function is the entry point for processing of ethernet packets received
 * from NW or DSP. The function is called in a while(1) loop inside packet
 * processing (or worker) thread. When measuring cycle counts, this function is
 * the starting point to get first snapshot of cycle count.
 *
 * The function does following:
 *
 * a. When measurement of cycle counts is enabled, get first snapshot of cycle
 * count.
 *
 * b. Parses the ethernet header. Currently processing assumes singly tagged
 * packet always. Packets from NW are distinguished from DSP using ether_type.
 *
 * c. Lookup sub interface based on incoming port + incoming vlan tag.
 *
 * d. Check forwarding mode of incoming port. When forwarding mode as HL_FWD,
 * match the destination mac address from packet with the incoming port mac address.
 * If match is found jump to IPv4 LLI bind function.
 *
 * Branches to IPv4 LLI processing for packets received from NW.
 * Branches to GTPu ULI processing for packets received from DSP.
 * =============================================================
 */

adk_status_t
adk_eth_dp_process_pkt (
    void **task_p,
    uint8_t task_count,
    uint32_t task_mask,
    __attribute__ ((unused)) void *ring,
    __attribute__ ((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
    int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    uint8_t i, drop_count = 0;
    uint32_t drop_mask = 0, cur_bit;
    uint16_t destFunc = 0;
    void *ring_hdl = NULL;
    adk_sif_cfg_pipeline_info_t *next_stage_entry_hdl = NULL;

    uint32_t dst_mod_id = 0;
    uint32_t dst_mod_task_count[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {0};
    uint32_t dst_mod_task_mask[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {0};
    uint16_t dst_mod_dst_func[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {0};
    void *dst_mod_ring_hdl[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {NULL};
    adk_sif_cfg_pipeline_info_t
    *dst_mod_next_stage_entry_hdl[ADK_SIF_CFG_PIPELINE_MODULE_MAX] =
    {NULL};

    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_LLI_RX_TS();

    for (i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, cur_bit = cur_bit << 1)
    {
        if (!(task_mask & cur_bit))
            continue;

        ret = adk_eth_dp_process_pkt_single(task_p[i], coreIndex, &destFunc,
                                            &dst_mod_id, &ring_hdl,
                                            &next_stage_entry_hdl);

#ifdef OPTIMIZED_FOR_ATOM_CLASS
        adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task_p[i]);
        rte_prefetch0((void *) adk_md_p);
#endif
        if (likely((ADK_E_INTERNAL_ERROR != ret) &&
                   (ADK_E_TASK_CONSUMED != ret)))
        {
            dst_mod_task_mask[dst_mod_id] |= cur_bit;
            dst_mod_task_count[dst_mod_id]++;
            dst_mod_dst_func[dst_mod_id] = destFunc;
            dst_mod_ring_hdl[dst_mod_id] = ring_hdl;
            dst_mod_next_stage_entry_hdl[dst_mod_id] = next_stage_entry_hdl;
            /* TODO: Currently assuming all the packets to follow the same path.
             * Enhance to handle mutiple paths for each of the packet in
             * the list of tasks */
        }
        else
        {
            if (ADK_E_INTERNAL_ERROR == ret)
            {
                /* Free the task */
                adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
                drop_mask |= cur_bit;
                drop_count++;
            }
            else if (ADK_E_TASK_CONSUMED == ret)
            {
                /* Task consumed. Do not free the task */
                drop_mask |= cur_bit;
                drop_count++;
            }
        }
    }

    task_count -= drop_count;
    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_LLI_TX_TS();

    for (i = 0; i < ADK_SIF_CFG_PIPELINE_MODULE_MAX; ++i)
    {
        if (adkDestEntryFunction_g[dst_mod_dst_func[i]])
            (adkDestEntryFunction_g[dst_mod_dst_func[i]])(task_p, dst_mod_task_count[i],
                                                          dst_mod_task_mask[i],
                                                          dst_mod_ring_hdl[i], dst_mod_next_stage_entry_hdl[i], coreIndex);
    }

    return ret;

}

/*!
 * =============================================================
 * adk_eth_dp_pkt_proc_uli()
 *
 * Structure of task params :
 *
 * taskParams[1] : Lower 24-bits MTM queue id or ADK_QOS_GET_QUEUE_ID
 * taskParams[7] : PDU buffer pointing to higher layer header (IP header)
 * destFlow	 : Encap flow based on F->F bind between IP and Ethernet
 *
 * Brief:
 * This function completes the ethernet processing of packet before sending to
 * NW or WIFI port. This function is called from IP ULI processing and does
 * following:
 *
 * a. Encapsulate the IP packet with new ethernet header (with vlan tag).
 *
 * b. Looks up QoS queue id for output sub-interface + pcp, when Queue Id is not
 * provided by IP module.
 *
 * c. Enqueues the PDU to QoS queue for shaping/scheduling.
 *
 * =============================================================
 */

adk_status_t
adk_eth_dp_pkt_proc_uli (
    void **task_p,
    uint8_t task_count,
    uint32_t task_mask,
    __attribute__ ((unused)) void *ring,
    __attribute__ ((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
    int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    uint8_t i, drop_count = 0;
    uint32_t drop_mask = 0, cur_bit;
    void * next_task_p = NULL;

    adk_sif_cfg_pipeline_info_t *pipeline_info = NULL;
    adk_sif_cfg_pipeline_exit_module_t *pipeline_src_info = NULL;
    adk_sif_cfg_pipeline_dest_module_t *pipeline_dest_info = NULL;

    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_ULI_RX_TS();
    for (i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if (!(task_mask & cur_bit))
            continue;

        next_task_p = (void *)(ADK_METADATA_PARAM((adk_pktio_task_t *)(task_p[i]), next_frag));
        ret = adk_eth_dp_pkt_proc_uli_single(task_p[i], coreIndex);

        if (unlikely(ADK_E_INTERNAL_ERROR == ret) ||
            (ADK_E_TASK_CONSUMED == ret))
        {
            if (ADK_E_INTERNAL_ERROR == ret)
            {
                /*Free the task*/
                adk_pktio_task_buffer_free(task_p[i]);
                drop_mask |= cur_bit;
                drop_count++;
            }
            else if (ADK_E_TASK_CONSUMED == ret)
            {
                /*Task consumed. Do not free the task*/
                drop_mask |= cur_bit;
                drop_count++;
            }
        }
        else
        {
            while (unlikely(0 != next_task_p))
            {
                /*Chanined task*/
                ret = adk_eth_dp_pkt_proc_uli_single(next_task_p, coreIndex);
                if (unlikely(ADK_E_INTERNAL_ERROR == ret))
                {
                    /* Free the task */
                    adk_pktio_task_buffer_free(next_task_p);
                    (ADK_METADATA_PARAM((adk_pktio_task_t *)(next_task_p), next_frag)) = NULL;
                    break;
                }
                next_task_p = (void *)(ADK_METADATA_PARAM((adk_pktio_task_t *)(next_task_p), next_frag));
            }
        }
    }
    task_count -= drop_count;

    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_ULI_TX_TS();

    if (likely(task_count))
    {
        /* For all packets received in a burst, destination cane be either port tx on same core or
         * port tx running on a different core. Hence we assume only one destination per burst.
         */

        /* Get the pipeline info */
        pipeline_info = (adk_sif_cfg_pipeline_info_t *)
                        ADK_TASK_USER_PARAMS((adk_pktio_task_t *)(task_p[0]), 8);
        if (pipeline_info)
        {
            pipeline_src_info =
                &pipeline_info->exit_module[ADK_SIF_CFG_PIPELINE_MODULE_ETHERNET_ULI];
            pipeline_dest_info =
                &pipeline_src_info->dest_module[ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX];

            if (pipeline_dest_info->valid == _CONFIG)
            {
                ret = adk_sys_dp_pkt_send_to_port(task_p, task_count,
                                                  (task_mask & ~drop_mask), coreIndex);
            }
            else
            {
                pipeline_dest_info =
                    &pipeline_src_info->dest_module[ADK_SIF_CFG_PIPELINE_MODULE_RING_TX];
                if (pipeline_dest_info->valid == _CONFIG)
                {
                    ret = adk_pktio_mod_ring_tx(task_p, task_count, (task_mask & ~drop_mask),
                                                pipeline_dest_info->ring,
                                                pipeline_dest_info->next_entry, coreIndex);
                }
            }
        }
        else
        {
            /*No pipeline info. Use default function*/
            ret = adk_sys_dp_pkt_send_to_port(task_p, task_count,
                                              (task_mask & ~drop_mask), coreIndex);
        }
    }

    if (unlikely(ret != ADK_SUCCESS))
    {
        /* Actual error logging should be available from SIF */
        ADK_DP_LOG_INFO(ADK_ETHERNET_MODULE_ID, "Error (ULI): Failed to send"
                        " packet to the output port!");
    }

    return ret;
}

#endif

adk_status_t
adk_eth_dp_process_pkt_ring (
    void **task_p,
    uint8_t __attribute__ ((unused)) task_count,
    uint32_t task_mask,
    void *ring,
    adk_sif_cfg_pipeline_info_t *next_stage_entry,
    __attribute__ ((unused)) int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t **task = (adk_pktio_task_t **)task_p;
    uint16_t nb_tx;
    uint32_t i, num = 0;
    uint32_t cur_bit;
    struct rte_mbuf
    *tx_task[ADK_DP_BULK_RX_SIZE * ADK_DP_TASK_MULTIPLY_MAX] = {0};

    /* Tasks may not be packed since some tasks are dropped/consumed/fragmented.
     * Pack the tasks */
    for (i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if (task_mask & cur_bit)
        {
            tx_task[num] = (struct rte_mbuf *)(task[i]);
            ADK_TASK_USER_PARAMS(tx_task[num], 8) =
                ADK_CONVERT_TO_TPARAMS(next_stage_entry);
            ADK_TASK_USER_PARAMS(tx_task[num], 9) =
                ADK_CONVERT_TO_TPARAMS(ADK_ETH_LLI_FN_ID);
            num++;
        }
    }

    nb_tx = rte_ring_sp_enqueue_burst((struct rte_ring *)ring, (void *)tx_task, num, NULL);
    if (unlikely(nb_tx != num))
    {
        for (i = nb_tx; i < num; ++i)
            rte_pktmbuf_free(tx_task[i]);
    }

    return ret;
}

adk_status_t
adk_eth_dp_pkt_proc_ring_uli (
    void **task_p,
    __attribute__ ((unused)) uint8_t task_count,
    uint32_t task_mask,
    void *ring,
    adk_sif_cfg_pipeline_info_t *next_stage_entry,
    __attribute__ ((unused)) int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t **task = (adk_pktio_task_t **)task_p;
    uint16_t nb_tx;
    uint32_t i, num = 0;
    uint32_t cur_bit;
    struct rte_mbuf *tx_task[ADK_DP_BULK_RX_SIZE * ADK_DP_TASK_MULTIPLY_MAX] = {0};

    /* Tasks may not be packed since some tasks are dropped/consumed/fragmented.
     * Pack the tasks */
    for (i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if (task_mask & cur_bit)
        {
            tx_task[num] = (struct rte_mbuf *)(task[i]);
            ADK_TASK_USER_PARAMS(tx_task[num], 8) =
                ADK_CONVERT_TO_TPARAMS(next_stage_entry);
            ADK_TASK_USER_PARAMS(tx_task[num], 9) =
                ADK_CONVERT_TO_TPARAMS(ADK_ETH_ULI_FN_ID);
            num++;
        }
    }

    nb_tx = rte_ring_sp_enqueue_burst((struct rte_ring *)ring, (void *)tx_task, num, NULL);
    if (unlikely(nb_tx != num))
    {
        for (i = nb_tx; i < num; ++i)
            rte_pktmbuf_free(tx_task[i]);
    }

    return ret;
}

 /*!  end addtogroup _ethernetFunctions  Ethernet Module
 * @}
 */
