/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp_helper.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file has all the helper functions required by the Ethernet
 *          data plane software.
 *
 *  **********************************************************************/

#include <adk_eth_dp_pvt.h>

/* Module info pointer stored for DP process to access tables */
extern adk_eth_dp_pvt_data_t *adk_eth_dp_pvt_g_p;

/* Ethernet global stats pointers */
extern adk_dp_eth_gl_stats_t *adk_eth_gl_stats_g_p;

/* Pointer to the list of Ethernet sv-interface table */
extern adk_dp_eth_svif_if_hash_t *adk_eth_svif_if_hash_table_g_p;

/* List of all Ethernet Egress VLAN Info */
extern adk_dp_eth_bridge_intf_t **adk_dp_eth_bridge_intf_t_g;

/* Map of bridge l2 domain to interface ID */
extern uint32_t* adk_dp_eth_bridge_map_id_g;

/* Default L2 domain for L2 call */
extern uint32_t* adk_dp_eth_default_l2_domain_g;

/*Auto connection status*/
extern uint8_t* adk_dp_eth_auto_conn_status_g;


void
adk_eth_dp_print_mac (
    uint8_t *macAddr)
{
    /*
     * Use this for all builds/combinations as debug_print() is now under
     * a double set of compile flags; ADK_DEBUG and ADK_INTERNAL_PRINT.
     * Setting this outside the conditionals will ensure no future build
     * breaks.
     */
    RTE_SET_USED(macAddr);

    debug_print("%02x-%02x-%02x-%02x-%02x-%02x",
                macAddr[0], macAddr[1], macAddr[2],
                macAddr[3], macAddr[4], macAddr[5]);
    return;
}

void
adk_eth_dp_print_hdr (
    adk_dp_eth_hdr_t *eth_hdr)
{
    debug_print("smac - ");
    adk_eth_dp_print_mac(eth_hdr->smac);
    debug_print(" dmac - ");
    adk_eth_dp_print_mac(eth_hdr->dmac);
    debug_print(" vlan_tpid - 0x%x, vlan_pri - 0x%x,"
                " vlan_id - 0x%x, ether_type - 0x%x\n",
                eth_hdr->vlan.tpid, eth_hdr->vlan.pbits,
                eth_hdr->vlan.vlanid, eth_hdr->ether_type);
    fflush(stdout);
}

void
adk_eth_dp_print_pdu_hdr (
    uint8_t *pdu)
{
    uint16_t *tmp_16 = NULL;
    uint16_t vlan = 0;
    uint16_t ether_type = 0;

    /* mac addresses */
    debug_print("dmac - ");
    adk_eth_dp_print_mac(&pdu[0]);
    debug_print(" smac - ");
    adk_eth_dp_print_mac(&pdu[6]);

    /* ether type */
    tmp_16 = (uint16_t *)(void *)&pdu[12];
    ether_type = adk_ntohs(*tmp_16);
    if (ether_type == 0x8100)
    {
        /* vlan header */
        tmp_16 = (uint16_t *)(void *)&pdu[14];
        vlan = adk_ntohs(*tmp_16);

        /* ether type */
        tmp_16 = (uint16_t *)(void *)&pdu[16];
        ether_type = adk_ntohs(*tmp_16);
    }
    debug_print(" vlan_tpid - 0x%x, vlan_pri - 0x%x,"
                " vlan_id - 0x%x, ether_type - 0x%x\n",
                0x8100, (vlan >> 13) & 0x7, vlan & 0xFFF,
                ether_type);

    fflush(stdout);
}

adk_dp_status_t
adk_eth_vlan_header_strip(adk_pktio_task_t *task, uint8_t **pduData, int coreIndex)
{
    uint8_t eth_mac_hdr[ETH_HDR_VLAN_HDR_OFFSET] = {0};
    ADK_MEMCPY(eth_mac_hdr, *pduData, ETH_HDR_VLAN_HDR_OFFSET);
    ADK_PDU_HEAD_PULL(task, *pduData, (uint8_t *), ETH_VLAN_HDR_LEN);
    ADK_MEMCPY(*pduData, eth_mac_hdr, ETH_HDR_VLAN_HDR_OFFSET);

    task->l2_len -= ETH_VLAN_HDR_LEN;
    INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].ingressUtilityCounters.eth_ing_pkts_vlan_tag_stripped);
    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Vlan Header Successfully stripped");

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_eth_vlan_header_proc(
    adk_pktio_task_t *task,
    uint8_t **pduData,
    adk_dp_eth_bridge_intf_t *dp_bridge_if_p,
    int coreIndex,
    bool *strip_vlan)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    uint8_t *pdu = *pduData;

    /*priority tagged packets*/
    if((adk_ntohs(*(uint16_t*)&pdu[ETH_HDR_VLAN_ID_OFFSET]) & 0xfff) == 0)
    {
        if(!dp_bridge_if_p->priority_tag)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Priority tagging is not enabled"
                        " Traffic with Vlan ID 0 dropped");
            return ADK_E_INTERNAL_ERROR;
        }
        else
        {
            ret = adk_eth_vlan_header_strip(task, pduData, coreIndex);
            if(unlikely(ret != ADK_SUCCESS))
            {
                adk_eth_gl_stats_g_p[coreIndex].vlan_header_stripping_failure_drop++;
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Vlan Header stripping failed");
                return ADK_E_INTERNAL_ERROR;
            }
            dp_bridge_if_p->intf_stats[coreIndex].inPrioTaggedFrames++;
            return ADK_SUCCESS;
        }
    }
    else
    {
        /*Vlan untagged list packets*/
        if((ADK_DP_ETH_AUTO_CONN_DISABLE == *adk_dp_eth_auto_conn_status_g))
        {
            if((pdu[0] & 0x01))
            {
                *strip_vlan = true;
                return ADK_SUCCESS;
            }
            else
            {
                ret = adk_eth_vlan_header_strip(task, pduData, coreIndex);
                if(unlikely(ret != ADK_SUCCESS))
                {   
                    adk_eth_gl_stats_g_p[coreIndex].vlan_header_stripping_failure_drop++;
                    ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Vlan Header stripping failed");
                    return ADK_E_INTERNAL_ERROR;
                }
            }
        }
    }
    return ADK_SUCCESS;
}

adk_dp_status_t
adk_eth_dp_parse_hdr (
    void *pduData,
    adk_dp_eth_hdr_t *eth_hdr)
{
    uint8_t *pdu;
    uint16_t *tmp_16;
    uint16_t vlan;

    pdu = (uint8_t *) pduData;

    /* MAC Addresses */
    ADK_MEMCPY(eth_hdr->dmac, &pdu[0], sizeof(uint8_t) * 6);
    ADK_MEMCPY(eth_hdr->smac, &pdu[6], sizeof(uint8_t) * 6);

    /* Ether type */
    tmp_16 = (uint16_t *)(void *)&pdu[12];
    eth_hdr->ether_type = adk_ntohs(*tmp_16);

    /* Parse VLAN tag, if present */
    if (eth_hdr->ether_type == 0x8100)
    {
        /* vlan header parsing */
        tmp_16 = (uint16_t *)(void *)&pdu[14];
        vlan = adk_ntohs(*tmp_16);
        eth_hdr->vlan.vlanid = vlan & 0xFFF;
        eth_hdr->vlan.tpid = 0x8100;
        eth_hdr->vlan.pbits = (vlan >> 13) & 0x7;

        if(eth_hdr->vlan.vlanid == 0)
            eth_hdr->vlan.priorityTagged = true;
        else
            eth_hdr->vlan.priorityTagged = false;

        /* ether type */
        tmp_16 = (uint16_t *)(void *)&pdu[16];
        eth_hdr->ether_type = adk_ntohs(*tmp_16);
        eth_hdr->hdr_len = 18;
    }
    else
    {
        eth_hdr->vlan.vlanid = 0;
        eth_hdr->vlan.tpid = 0;
        eth_hdr->vlan.pbits = 0;
        eth_hdr->vlan.cfi = 0;
        eth_hdr->hdr_len = 14;
        eth_hdr->vlan.priorityTagged = false;
    }

    return ADK_SUCCESS;
}


adk_status_t
adk_eth_dp_interface_id_get(void *task_p,
                            uint32_t *interface_id)
{
    adk_pktio_task_t *task = (adk_pktio_task_t *)task_p;
    adk_dp_eth_hdr_t eth_hdr;
    uint32_t bridge_stats_id = 0;
    adk_dp_eth_svif_intf_t *dp_svif_if_p = NULL;
    adk_dp_eth_bridge_intf_t *dp_bridge_if_p = NULL;
    void *pduData_p;

    ADK_PDU_PTR_GET(task, pduData_p, void *);

    /* parse ethernet header and extract fields */
    adk_eth_dp_parse_hdr(pduData_p, &eth_hdr);

    if(eth_hdr.vlan.vlanid)
    {
        adk_eth_dp_svif_get(adk_eth_svif_if_hash_table_g_p, *adk_dp_eth_default_l2_domain_g, eth_hdr.vlan.vlanid,
                &dp_svif_if_p);
        if (!dp_svif_if_p)
        {
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "SVIF for vlan %d"
                    " not created. Packet drop!", eth_hdr.vlan.vlanid);
            return ADK_E_INTERNAL_ERROR;
        }
        *interface_id = (dp_svif_if_p->statsId + ADK_ETH_NUM_BRIDGE);
    }
    else
    {
        bridge_stats_id = adk_dp_eth_bridge_map_id_g[*adk_dp_eth_default_l2_domain_g];
        if(bridge_stats_id >= ADK_ETH_MAX_BG_ID)
        {
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                    " not defined for L2 domain %d", *adk_dp_eth_default_l2_domain_g);
            return ADK_E_INTERNAL_ERROR;
        }

        dp_bridge_if_p = adk_dp_eth_bridge_intf_t_g[bridge_stats_id];
        if(NULL == dp_bridge_if_p)
        {
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                    " not defined for L2 domain %d", *adk_dp_eth_default_l2_domain_g);
            return ADK_E_INTERNAL_ERROR;
        }
        *interface_id = dp_bridge_if_p->statsId;
    }

    return ADK_SUCCESS;
}

uint8_t adk_eth_task_chained_buffer_free(adk_pktio_task_t *task_p)
{
    adk_pktio_task_t *cur_task_p, *next_task_p = NULL;
    uint8_t tasks_freed = 0;

    /* Free the entire chained task */
    cur_task_p = (adk_pktio_task_t *)(task_p);
    while(cur_task_p)
    {
        next_task_p = ADK_METADATA_PARAM(cur_task_p, next_frag);
        debug_print("Calling adk_pal_dp_release_send from adk_eth_task_chained_buffer_free \n");
        adk_pal_dp_release_send(cur_task_p, ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
        adk_pktio_task_buffer_free(cur_task_p);
        cur_task_p = next_task_p;
        tasks_freed++;
    }
    return tasks_freed;
}
 /*!  end addtogroup _ethernetFunctions  Ethernet Module
 * @}
 */
