/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp_netd.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file contains the Ethernet data plane software to handle
 *          packets injected from Linux.
 *
 *          Note: The tx pkts counter, which counts all packets given by the
 *          higher layer, is not blindly incremented at the beginning because
 *          packet may get bound to IPvX module and thus, may get counted
 *          twice. So, this counter is only incremented if either the packet is
 *          dropped or fails to bind with any other module.
 *
 *          The tx octet counters, which counts only those octets which are
 *          sent out successfully, is not incremented if pkt binds to any other
 *          module. Only if the packet fails to bind and is not dropped, this
 *          counters is incremented.
 *
 *  **********************************************************************/

#include <adk_eth_dp_pvt.h>
#include "adk_ipsec_export.h"

/* Module info pointer stored for DP process to access tables */
extern adk_eth_dp_pvt_data_t *adk_eth_dp_pvt_g_p;

/* Ethernet global stats pointers */
extern adk_dp_eth_gl_stats_t *adk_eth_gl_stats_g_p;

/* Pointer to the list of Ethernet sv-interface table */
extern adk_dp_eth_svif_if_hash_t *adk_eth_svif_if_hash_table_g_p;

/* List of all Ethernet bridge interfaces */
extern adk_dp_eth_bridge_intf_t **adk_dp_eth_bridge_intf_t_g;

/* Map of bridge l2 domain to interface ID */
extern uint32_t* adk_dp_eth_bridge_map_id_g;

/* Default L2 domain for L2 call */
extern uint32_t* adk_dp_eth_default_l2_domain_g;

/*Auto Connectio status*/
extern uint8_t* adk_dp_eth_auto_conn_status_g;

/*The eth type to pcp map table*/
extern adk_lpm_handle_t adk_eth_ether_type_pcp_map_tbl_g;

/*The etherType to Queue Map hash table*/
extern adk_dp_eth_ethertype_queue_map_hash_t *adk_eth_ether_type_queue_map_tbl_g;

/*The ethernet non bridge origin queue ID*/
extern uint16_t *adk_eth_non_bridge_origin_queue_id_g;

/*The ethernet bridge timestamp quad group port*/
extern uint8_t *adk_eth_bridge_timestamp_quad_group_port_g;

/*The ethernet DSCP to PCP Map*/
extern uint8_t *adk_eth_dscp_to_pcp_map_g;

adk_status_t
adk_eth_dp_pkt_proc_netd_v2 (
    void **task_p,
    uint8_t task_count,
    uint32_t task_mask,
    void *ptr1,
    adk_sif_cfg_pipeline_info_t *ptr2,
    int coreIndex)
{

    RTE_SET_USED(ptr1);
    RTE_SET_USED(ptr2);
    adk_eth_dp_pkt_proc_netd(task_p, task_count, task_mask, coreIndex);
    return ADK_SUCCESS;
}

static inline adk_status_t adk_eth_dp_nic_send(adk_pktio_task_t* task, int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    uint8_t pktColor;
    uint16_t dev_tx_queue_id;
    pktColor = ADK_METADATA_PARAM((adk_pktio_task_t *)(task), pktColor);
    dev_tx_queue_id = ADK_METADATA_PARAM((adk_pktio_task_t *)(task), dev_tx_queue_id);

    ret = adk_pal_dp_nic_send(task->port, pktColor, dev_tx_queue_id, task, ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
    if(ret == ADK_E_INTERNAL_ERROR)
    {
        adk_pal_dp_release_send((adk_pktio_task_t *)(task), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
        adk_pktio_task_buffer_free((adk_pktio_task_t *)(task));
        adk_eth_gl_stats_g_p[coreIndex].nic_send_failure_drop++;
    }
    return ret; 
}
adk_status_t
adk_eth_dp_pkt_proc_netd (
    void **task_p,
    uint8_t task_count,
    uint32_t task_mask,
    int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0, out_port=0;
    bool eth_initialized = true;
    uint8_t *pdu = NULL;
    adk_dp_eth_qos_info_t qos_queue;
#ifdef ADK_DEBUG
    adk_dp_eth_hdr_t eth_hdr;
#endif
    adk_pktio_task_t *task;
    adk_metadata_t *adk_md_p;
    adk_dsi_tx_metadata_t *dsi_md_p;
    uint32_t bridge_stats_id = 0;
    adk_dp_eth_bridge_intf_t *dp_bridge_if_p = NULL;
    RTE_SET_USED(task_mask);

    if (unlikely((adk_eth_dp_pvt_g_p == NULL) ||
                 (adk_eth_dp_pvt_g_p->cookie != ADK_ETHERNET_MODULE_COOKIE) ||
                 (adk_eth_non_bridge_origin_queue_id_g == NULL) || 
                 (adk_eth_bridge_timestamp_quad_group_port_g == NULL)))
    {
        /* This error is expected when packets are received before Ethernet
         * module has been initialized. */
        ADK_DP_LOG_INFO(ADK_ETHERNET_MODULE_ID, "Error (NetD): Ethernet module"
                        " uninitialized! Dropping pkt(s)!");
        eth_initialized = false;
    }
    else
    {
        adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;
    }

    /* Note: Burst packets are not supported */
    for (i = 0; i < task_count; ++i)
    {
        if (unlikely(eth_initialized == false))
        {
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }

        INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_pkts_received_from_netd);
        INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_non_bridge_netd_origin_pkts_received);

        task = (adk_pktio_task_t *)task_p[i];
        adk_md_p = ADK_METADATA_PTR(task);

        out_port = ADK_MD_GET_LINUX_INTF_ID(task);

        if (unlikely(out_port >= ADK_ETH_PORTS_MAX))
        {
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Error (NetD): Received"
                                " pkt on invalid port %d! Dropping pkt!",
                                out_port);
            adk_eth_gl_stats_g_p[coreIndex].port_not_defined++;
            adk_eth_gl_stats_g_p[coreIndex].netd_origin_non_bridge_frames_discard++;
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }

        ADK_PDU_PTR_GET(task, pdu, uint8_t *);

#ifdef ADK_DEBUG
        debug_print("NetD Packet Injection: port %d, cpuID %d\n",
                    out_port, coreIndex);

        adk_eth_dp_parse_hdr(pdu, &eth_hdr);
        debug_print("NetD Packet Injection with following Ethernet header:\n");
        adk_eth_dp_print_hdr(&eth_hdr);
#endif
        if(ADK_MD_GET_TS_MSG_TYPE(task) == ADK_NETD_TS_IDX_RELEASE)
        {
            if(*adk_eth_non_bridge_origin_queue_id_g == 1)
            {
                ret = adk_netd_fp_ts_index_free(out_port,ADK_MD_GET_TS_INDEX(task),ADK_MD_GET_TS_CORR(task));
                if(ret != ADK_SUCCESS)
                {
                    ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Failed to free the Timestamp Index %d",ADK_MD_GET_TS_INDEX(task));
                }
            }
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }
        dsi_md_p = ADK_DSI_TX_METADATA_PTR(task);
        dsi_md_p->l2_domain = *adk_dp_eth_default_l2_domain_g;

        bridge_stats_id = adk_dp_eth_bridge_map_id_g[*adk_dp_eth_default_l2_domain_g];
        if(bridge_stats_id >= ADK_ETH_MAX_BG_ID)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            adk_eth_gl_stats_g_p[coreIndex].netd_origin_non_bridge_frames_discard++;
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                                  " not defined for L2 domain %d", *adk_dp_eth_default_l2_domain_g);
            ret = ADK_E_INTERNAL_ERROR;
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
         }

        dp_bridge_if_p = adk_dp_eth_bridge_intf_t_g[bridge_stats_id];
        if(NULL == dp_bridge_if_p)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            adk_eth_gl_stats_g_p[coreIndex].netd_origin_non_bridge_frames_discard++;
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                                  " not defined for L2 domain %d", *adk_dp_eth_default_l2_domain_g);
            ret = ADK_E_INTERNAL_ERROR;
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }
        if(*adk_eth_non_bridge_origin_queue_id_g != ADK_QOS_INVALID_QUEUE_ID)
        {
            ADK_METADATA_PARAM(task, dev_tx_queue_id) = *adk_eth_non_bridge_origin_queue_id_g;
        }
        else
        {
            adk_eth_gl_stats_g_p[coreIndex].netd_origin_non_bridge_frames_discard++;
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Queue not configured for non bridge originated traffic\n");
            ret = ADK_E_INTERNAL_ERROR;
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }

		bool isPHBQueue = false;
		uint32_t qosStatsId;
		/*Find the stats id for QoS queue*/
		qosStatsId = adk_qos_dp_get_queue_stats_id(adk_md_p->dev_tx_queue_id, &isPHBQueue);
		adk_md_p->qosStatsId = qosStatsId; 
		adk_md_p->isPHBQueue = isPHBQueue;
		adk_md_p->isIpPkt = false;

        adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_ETHERNET_INTERFACE_STATS].stats_id =
            ADK_SYS_EGRESS_STATS_ID_INVALID;

        adk_eth_dp_tx_md_update(task, dp_bridge_if_p->eth_dev_id, out_port, &qos_queue, coreIndex);

        adk_md_p->saID = ADK_IPSEC_INVALID_SA_ID; 

        ret = adk_eth_dp_nic_send(task, coreIndex);
        if(ret != ADK_E_INTERNAL_ERROR)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.
                    eth_eg_non_bridge_netd_origin_pkts_transmitted);
        }
    }

    if (eth_initialized == true)
    {
        adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;
    }

    return ADK_SUCCESS;
}


static inline void adk_eth_dp_priority_tag_add(adk_pktio_task_t *task,
                                 adk_dp_eth_hdr_t *eth_hdr)
{
    uint8_t *pdu = NULL;
    uint32_t vlan = 0;

    ADK_PDU_HEAD_PUSH(task, pdu, (uint8_t *), ETH_VLAN_HDR_LEN);

    ADK_MEMCPY(pdu, eth_hdr, ETH_HDR_VLAN_HDR_OFFSET);
    vlan = (0x8100 << 16);
    uint32_t *tmp_32 = (uint32_t *) (void *) &pdu[ETH_HDR_VLAN_HDR_OFFSET];
    *tmp_32 = adk_htonl(vlan);
    uint16_t *tmp_16 = (uint16_t *) (void *) &pdu[ETH_HDR_VLAN_HDR_OFFSET+ETH_VLAN_HDR_LEN];
    *tmp_16 = adk_htons(eth_hdr->ether_type);
    eth_hdr->vlan.priorityTagged = true;
}

static inline void adk_eth_dp_update_pcp(adk_pktio_task_t *task,
        adk_dp_eth_hdr_t *eth_hdr)
{
    uint8_t *pdu = NULL;
    adk_ethernet_priority_t *pcp_val = NULL;
    adk_handle_t lpmVal;

    if(!eth_hdr->vlan.vlanid &&
            !eth_hdr->vlan.priorityTagged)
    {
        ADK_ETH_DP_LOG_INFO_DEBUG("non vlan tag or priority tag packets");
        return;
    }
    if(eth_hdr->ether_type == ETHER_TYPE_IPV4)
    {
        ADK_ETH_DP_LOG_INFO_DEBUG("IPv4 packets");
        return;
    }

    /*IPv6 LinkLocal and Multicast traffic VLAN PCP value 
      should get updated based on DSCP value from IMC param*/
    if(eth_hdr->ether_type == ETHER_TYPE_IPV6) 
    {
        adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
        adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
        eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
            &adk_md_p->imc_param_start; 
        pcp_val = &adk_eth_dscp_to_pcp_map_g[eth_uli_imc_param_p->dscp];
    }
    else
    {
        lpmVal = adk_lpm_lookup(adk_eth_ether_type_pcp_map_tbl_g, (uint8_t *) &eth_hdr->ether_type);
        if((adk_handle_t) (uintptr_t)ADK_LPM_VALUE_INVALID == lpmVal)
        {
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "PCP map not found for ethertype=%x", eth_hdr->ether_type);
            return;
        }
        pcp_val = (adk_ethernet_priority_t *)(uintptr_t)lpmVal;
    }
    ADK_PDU_PTR_GET(task, pdu, uint8_t *);

    pdu[14] |= ((*pcp_val & ETH_PRI_MASK) << 5);

    return;
}

static inline bool adk_eth_dp_update_queue(
        adk_pktio_task_t *task,
        adk_dp_eth_bridge_intf_t *dp_bridge_if_p,
        adk_dp_eth_hdr_t *eth_hdr)
{
    adk_dp_eth_ethertype_queue_map_hash_t entry;
    uint32_t i, hash_val, hash_idx;

    if((eth_hdr->ether_type == ETHER_TYPE_IPV4) || 
        (eth_hdr->ether_type == ETHER_TYPE_IPV6))
    {
        return true;
    }

    hash_val = ADK_ETHERNET_HASH(dp_bridge_if_p->bridgeIntfId, eth_hdr->ether_type);
    hash_idx = hash_val & ADK_ETH_ETHTYPE_QUEUE_MAP_MASK;

    for (i = 0; i < 8; i++)
    {
        entry.hash_entry = adk_eth_ether_type_queue_map_tbl_g[hash_idx].hash_entry;
        if (entry.ethType == eth_hdr->ether_type)
        {
            ADK_METADATA_PARAM(task, dev_tx_queue_id) = entry.queueId;
            return true;
        }
        if ((hash_idx & 0x7) == 0x7)
            hash_idx = hash_idx & ~0x7;
        else
            hash_idx++;
    }

    hash_idx = (hash_idx & ~0x7) + 0x8;

    for (i = 0; i < ADK_ETH_ETHTYPE_QUEUE_MAP_MAX - 8; i++, hash_idx++)
    {
        if (hash_idx == ADK_ETH_ETHTYPE_QUEUE_MAP_MAX)
        {
            hash_idx = 0;
        }
        entry.hash_entry = adk_eth_ether_type_queue_map_tbl_g[hash_idx].hash_entry;
        if (entry.ethType == eth_hdr->ether_type)
        {
            ADK_METADATA_PARAM(task, dev_tx_queue_id) = entry.queueId;
            return true;
        }
    }

    return false;
}

#ifdef ADK_DEVICE_SNR
adk_status_t
adk_eth_dp_pkt_proc_netd_bridge (
    void **task_p,
    uint8_t task_count,
    uint32_t task_mask,
    void *ptr1,
    adk_sif_cfg_pipeline_info_t *ptr2,
    int coreIndex)
{

    RTE_SET_USED(ptr1);
    RTE_SET_USED(ptr2);

    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0, out_port=0;
    bool eth_initialized = true;
    uint8_t *pdu = NULL;
    adk_dp_eth_qos_info_t qos_queue;
    adk_dp_eth_hdr_t eth_hdr;
    adk_pktio_task_t *task;
    uint16_t bg_id;
    adk_dsi_tx_metadata_t *dsi_md_p;
    uint32_t bridge_stats_id = 0;
    adk_dp_eth_bridge_intf_t *dp_bridge_if_p = NULL;
    adk_dp_eth_svif_intf_t *dp_svif_if_p = NULL;
    adk_metadata_t *adk_md_p;
    RTE_SET_USED(task_mask);

    if (unlikely((adk_eth_dp_pvt_g_p == NULL) ||
                 (adk_eth_dp_pvt_g_p->cookie != ADK_ETHERNET_MODULE_COOKIE) ||
                 (adk_eth_dscp_to_pcp_map_g == NULL )))
    {
        /* This error is expected when packets are received before Ethernet
         * module has been initialized. */
        ADK_DP_LOG_INFO(ADK_ETHERNET_MODULE_ID, "Error (NetD): Ethernet module"
                        " uninitialized! Dropping pkt(s)!");
        eth_initialized = false;
    }
    else
    {
        adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;
    }

    /* Note: Burst packets are not supported */
    for (i = 0; i < task_count; ++i)
    {
        if (unlikely(eth_initialized == false))
        {
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }

        INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_pkts_received_from_netd);

        task = (adk_pktio_task_t *)task_p[i];

        /* Get Linux interface ID and mask out BGID from it */
        bg_id = (ADK_MD_GET_LINUX_INTF_ID(task)) - ADK_ETH_BG_ID_OFFSET;

        if (unlikely(bg_id >= ADK_ETH_MAX_BG_ID))
        {
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Error (NetD): Received"
                                " pkt on invalid bridge %d! Dropping pkt!",
                                bg_id);
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }

        dsi_md_p = ADK_DSI_TX_METADATA_PTR(task);
        dsi_md_p->l2_domain = *adk_dp_eth_default_l2_domain_g;

        bridge_stats_id = adk_dp_eth_bridge_map_id_g[*adk_dp_eth_default_l2_domain_g];
        if(bridge_stats_id >= ADK_ETH_MAX_BG_ID)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                                  " not defined for L2 domain %d", *adk_dp_eth_default_l2_domain_g);
            ret = ADK_E_INTERNAL_ERROR;
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
         }

        dp_bridge_if_p = adk_dp_eth_bridge_intf_t_g[bridge_stats_id];
        if(NULL == dp_bridge_if_p)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                                  " not defined for L2 domain %d", *adk_dp_eth_default_l2_domain_g);
            ret = ADK_E_INTERNAL_ERROR;
            adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
            continue;
        }

        ADK_DP_LOG_INFO_1D(ADK_ETHERNET_MODULE_ID, "Retrieved bridge interface for L2 domain %d", *adk_dp_eth_default_l2_domain_g);

        out_port = ADK_HOS_CPK_PORT; 
        ADK_PDU_PTR_GET(task, pdu, uint8_t *);

        adk_md_p = ADK_METADATA_PTR(task);

        adk_eth_dp_parse_hdr(pdu, &eth_hdr);

#ifdef ADK_DEBUG
        debug_print("NetD Packet Injection: port %d, cpuID %d\n",
                    out_port, coreIndex);

        debug_print("l2 domain %d\n", dsi_md_p->l2_domain);
        debug_print("NetD Packet Injection with following Ethernet header:\n");
        adk_eth_dp_print_hdr(&eth_hdr);
#endif
        
        /*
         * Lookup sub interface table to get sub interface from incoming port &
         * vlan id field.
         */
        if(ADK_DP_ETH_AUTO_CONN_DISABLE == *adk_dp_eth_auto_conn_status_g)
        {
            if(eth_hdr.vlan.vlanid)
            {
                adk_eth_dp_svif_get(adk_eth_svif_if_hash_table_g_p, *adk_dp_eth_default_l2_domain_g, eth_hdr.vlan.vlanid,
                        &dp_svif_if_p);

                if (!dp_svif_if_p)
                {
                    ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "SVIF for vlan %d"
                            " not created. Packet drop!", eth_hdr.vlan.vlanid);
                    adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
                    adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
                    adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
                    continue;
                }

                ADK_DP_LOG_INFO_1P(ADK_ETHERNET_MODULE_ID, "Retrieved svif - %p\n", dp_svif_if_p);
                ADK_DP_LOG_INFO_1P(ADK_ETHERNET_MODULE_ID, "Packet received on svif - %p\n", dp_svif_if_p);
            }
            else if(dp_bridge_if_p->priority_tag)
            {
                adk_eth_dp_priority_tag_add(task, &eth_hdr);
            }

            adk_eth_dp_update_pcp(task, &eth_hdr);

            if(!adk_eth_dp_update_queue(task, dp_bridge_if_p, &eth_hdr))
            {
                ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Failed to find the queue mapping for ethertype %d\n",eth_hdr.ether_type);
                adk_eth_gl_stats_g_p[coreIndex].ethtype_queue_map_not_found_drop++;
                adk_pal_dp_release_send((adk_pktio_task_t *)(task_p[i]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
                adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[i]));
                continue;
            }
        }
        else
        {
            ADK_METADATA_PARAM(task, dev_tx_queue_id) = adk_qos_dp_get_default_tx_queue();
            adk_md_p->saID = ADK_IPSEC_INVALID_SA_ID;
        }
        if(task->ol_flags & PKT_TX_IPV6)
        {
            if(task->ol_flags & PKT_TX_OUTER_IPV6)
            {
                task->outer_l2_len = (eth_hdr.vlan.vlanid || eth_hdr.vlan.priorityTagged) ? ETH_VLAN_HEADER_LEN : ETH_NO_VLAN_HEADER_LEN;
            }
            else
            {
                task->l2_len = (eth_hdr.vlan.vlanid || eth_hdr.vlan.priorityTagged) ? ETH_VLAN_HEADER_LEN : ETH_NO_VLAN_HEADER_LEN;
            }
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_ipv6_multicast_pkts_received);
        }


		bool isPHBQueue = false;
		uint32_t qosStatsId;
		/*Find the stats id for QoS queue*/
		qosStatsId = adk_qos_dp_get_queue_stats_id(adk_md_p->dev_tx_queue_id, &isPHBQueue);
		adk_md_p->qosStatsId = qosStatsId; 
		adk_md_p->isPHBQueue = isPHBQueue;
		adk_md_p->isIpPkt = false;

        if (IS_BROADCAST_MAC(eth_hdr.dmac)) {
            if(eth_hdr.ether_type == ETHER_TYPE_IPV6 || 
                    eth_hdr.ether_type == ETHER_TYPE_IPV4)
            {
                adk_md_p->isEthernetIpBcast = true;
                adk_md_p->isEthernetNonIpBcast = false;
            }
            else
            {
                adk_md_p->isEthernetNonIpBcast = true;
                adk_md_p->isEthernetIpBcast = false;
            }
            adk_md_p->isEthernetIpMcast = false;
            adk_md_p->isEthernetNonIpMcast = false;
        }else if ((eth_hdr.dmac[0] & 0x01)) {
            if(eth_hdr.ether_type == ETHER_TYPE_IPV6 || 
                    eth_hdr.ether_type == ETHER_TYPE_IPV4)
            {
                adk_md_p->isEthernetIpMcast = true;
                adk_md_p->isEthernetNonIpMcast = false;
            }
            else
            {
                adk_md_p->isEthernetNonIpMcast = true;
                adk_md_p->isEthernetIpMcast = false;
            }
            adk_md_p->isEthernetIpBcast = false;
            adk_md_p->isEthernetNonIpBcast = false;
        }else {
            adk_md_p->isEthernetNonIpBcast = false;
            adk_md_p->isEthernetNonIpMcast = false;
            adk_md_p->isEthernetIpMcast = false;
            adk_md_p->isEthernetIpBcast = false;
        }

        adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_ETHERNET_INTERFACE_STATS].stats_id =
            (dp_svif_if_p == NULL) ? dp_bridge_if_p->statsId : (dp_svif_if_p->statsId + ADK_ETH_NUM_BRIDGE);

        adk_eth_dp_tx_md_update(task, dp_bridge_if_p->eth_dev_id, out_port, &qos_queue, coreIndex);

        if((eth_hdr.ether_type != ETHER_TYPE_IPV6) &&
            (eth_hdr.ether_type != ETHER_TYPE_IPV4))
        {
            adk_md_p->saID = ADK_IPSEC_INVALID_SA_ID;
        }
        adk_eth_dp_nic_send(task, coreIndex);
    }

    if (eth_initialized == true)
    {
        adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;
    }

    return ret;
}

/* Function to be called from adk_sys_dp_egress_stats_cb for successfully transmitted packets with time stamp capture enabled */
adk_status_t
adk_eth_dp_ts_pkt_transmitted(
    void **task_p,
    __attribute__ ((unused)) uint8_t task_count,
    __attribute__ ((unused)) uint32_t task_mask,
    __attribute__ ((unused)) void *ring,
    __attribute__ ((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
    int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t *cur_task_p;

    cur_task_p = (adk_pktio_task_t *)(task_p[0]);

    ret = adk_netd_fp_ts_pkt_transmitted(ADK_MD_GET_LINUX_INTF_ID(cur_task_p),
     ADK_MD_GET_TS_INDEX(cur_task_p), ADK_MD_GET_TS_CORR(cur_task_p),
     cur_task_p);

    /* If tx complete notification send failed, increment error counter */
    if(ret != ADK_SUCCESS)
    {
        adk_eth_gl_stats_g_p[coreIndex].ts_tx_complete_notification_failed++;
    }
 
    return ret;
}   

#endif /*ADK_DEVICE_SNR*/


 /*!  end addtogroup _ethernetFunctions  Ethernet Module
 * @}
 */
