/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp_tables.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file has all the low level functions which operate upon
 *          data plane tables.
 *
 *  **********************************************************************/

#include <adk_eth_dp_pvt.h>
#include <adk_eth_dp.h>
/* NOTE: This header file will be removed shortly. 
    data plane code should not include this header file.
    This is added to get workaround of add/remove/update in LPM
    tree with multiple calls */

adk_dp_status_t
adk_eth_dp_svif_get (
    adk_dp_eth_svif_if_hash_t *svif_if_hash_table_p,
    uint16_t l2_domain_id,
    uint16_t vlan,
    adk_dp_eth_svif_intf_t **dp_svif_intf_p)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    uint32_t i, hash_val, hash_idx;
    adk_dp_eth_svif_if_hash_t entry;

    debug_print("Searching entry for l2_domain_id %x, vlan 0x%x in svif table ...\n",
                l2_domain_id, vlan);

    hash_val = ADK_ETHERNET_HASH(l2_domain_id, vlan);
    hash_idx = hash_val & ADK_ETH_SVIF_INTF_MASK;

    for (i = 0; i < 7; i++)
    {
        entry.hash_entry = svif_if_hash_table_p[hash_idx].hash_entry;
        if ((entry.l2_domain_id == l2_domain_id) && (entry.vlan == vlan) &&
            (entry.stats_id != ETH_INVALID_NS_ID))
        {
            debug_print("Found entry %"PRIu64"\n", entry.hash_entry);
            goto _adk_eth_dp_svif_get_fill;
        }

        if ((hash_idx & 0x7) == 0x7)
            hash_idx = hash_idx & ~0x7;
        else
            hash_idx++;
    }

    hash_idx = (hash_idx & ~0x7) + 0x8;

    for (i = 0; i < ADK_ETH_SVIF_INTF_MAX - 8; i++, hash_idx++)
    {
        if (hash_idx == ADK_ETH_SVIF_INTF_MAX)
            hash_idx = 0;

        entry.hash_entry = svif_if_hash_table_p[hash_idx].hash_entry;
        if ((entry.l2_domain_id ==  l2_domain_id) && (entry.vlan == vlan) &&
            (entry.stats_id != ETH_INVALID_NS_ID))
        {
            goto _adk_eth_dp_svif_get_fill;
        }
    }

    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Error: Entry not found in"
                          " svif Table!");
    *dp_svif_intf_p = NULL;
    ret = ADK_DP_E_ERROR;
    goto _adk_eth_dp_svif_get_exit;

_adk_eth_dp_svif_get_fill:

    *dp_svif_intf_p = adk_eth_svif_if_g[entry.stats_id];

_adk_eth_dp_svif_get_exit:

    return ret;
}

adk_status_t
adk_eth_dp_arp_get (
    adk_dp_eth_arp_table_t *eth_arp_table,
    adk_ethernet_ipv4_addr_t ip_addr,
    adk_ethernet_mac_addr_t dest_mac,
    adk_ethernet_arp_status_t* arp_status)
{

    adk_dp_eth_arp_entry_t *arp_entry;

    ADK_ETH_DP_RCU_DISTRIBUTED_CHECK(eth_arp_table);
    ADK_ETH_DP_RCU_DISTRIBUTED_CHECK_OVERFLOW_TBL(eth_arp_table);

    if(ADK_SUCCESS !=  adk_eth_dp_arp_entry_lookup(eth_arp_table, ip_addr, &arp_entry))
    {
        ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Error: Failed to find the ARP Entry");

        return ADK_E_INVALID_PARAMETER;
    }
    else
    {
        memcpy(dest_mac, arp_entry->dest_mac, sizeof(adk_ethernet_mac_addr_t));
        *arp_status = arp_entry->arp_status;
    }

    return (ADK_SUCCESS);
}

adk_status_t
adk_eth_dp_arp_entries_get (
        adk_dp_eth_arp_table_t *eth_arp_table,
        uint32_t *numArpEntries, 
        adk_ethernet_arp_entry_t **arp_entries_p)
{

    adk_ethernet_arp_entry_t *arpInfo_p = NULL;
    uint16_t i = 0,j = 0;

    ADK_ETH_DP_RCU_DISTRIBUTED_CHECK(eth_arp_table);
    ADK_ETH_DP_RCU_DISTRIBUTED_CHECK_OVERFLOW_TBL(eth_arp_table);

    *numArpEntries = 0;

    /*count the number of ARP entries*/
    for(i = 0;i < ADK_ETH_NO_OF_NEIGH_ENTRY; i++)
    {
        if(eth_arp_table->tbl[i].is_enabled)
        {
            (*numArpEntries)++;
        }
    }

    if(!(*numArpEntries))
    {
        *arp_entries_p = NULL;
        return ADK_SUCCESS;
    }
    arpInfo_p = (adk_ethernet_arp_entry_t *) 
                ADK_DP_SHM_CALLOC(ADK_ETHERNET_MODULE_ID, 
                                  ADK_ETH_ARP_ENTRY_OBJ, 
                                  (*numArpEntries),
                                  sizeof(adk_ethernet_arp_entry_t),
                                  ADK_ETH_DP_PTR_ALIGN);
    if(NULL == arpInfo_p)
    {
        return ADK_E_MEM_UNAVAILABLE;
    }

    for(i = 0,j = 0;i < ADK_ETH_NO_OF_NEIGH_ENTRY; i++)
    {
        if(eth_arp_table->tbl[i].is_enabled)
        {
            arpInfo_p[j].ip_addr = eth_arp_table->tbl[i].ipv4_addr;
            arpInfo_p[j].arp_status = eth_arp_table->tbl[i].arp_status;
            if((arpInfo_p[j].arp_status == ADK_ETHERNET_NEIGHBOR_REACHABLE) ||
                (arpInfo_p[j].arp_status == ADK_ETHERNET_NEIGHBOR_STALE))
            {
               memcpy(arpInfo_p[j].dest_mac, eth_arp_table->tbl[i].dest_mac, sizeof(adk_ethernet_mac_addr_t));     
            }
            j++;
        }
    }
    *arp_entries_p = arpInfo_p;
    return ADK_SUCCESS;
}

adk_status_t
adk_eth_dp_neighbor_entries_get (
        adk_dp_eth_neighbor_table_t *eth_neigh_table,
        uint32_t *numNeighEntries, 
        adk_ethernet_neighbor_entry_t **neigh_entries_p)
{

    adk_ethernet_neighbor_entry_t *neighInfo_p = NULL;
    uint16_t i = 0,j = 0;

    ADK_ETH_DP_RCU_DISTRIBUTED_CHECK(eth_neigh_table);
    ADK_ETH_DP_RCU_DISTRIBUTED_CHECK_OVERFLOW_TBL(eth_neigh_table);

    *numNeighEntries = 0;

    /*count the number of ARP entries*/
    for(i = 0;i < ADK_ETH_NO_OF_NEIGH_ENTRY; i++)
    {
        if(eth_neigh_table->tbl[i].is_enabled)
        {
            (*numNeighEntries)++;
        }
    }

    if(!(*numNeighEntries))
    {
        *neigh_entries_p = NULL;
        return ADK_SUCCESS;
    }

    neighInfo_p = (adk_ethernet_neighbor_entry_t *) 
                ADK_DP_SHM_CALLOC(ADK_ETHERNET_MODULE_ID, 
                                  ADK_ETH_NEIGH_ENTRY_OBJ, 
                                  (*numNeighEntries),
                                  sizeof(adk_ethernet_neighbor_entry_t),
                                  ADK_ETH_DP_PTR_ALIGN);
    if(NULL == neighInfo_p)
    {
        return ADK_E_MEM_UNAVAILABLE;
    }

    for(i = 0,j = 0;i < ADK_ETH_NO_OF_NEIGH_ENTRY; i++)
    {
        if(eth_neigh_table->tbl[i].is_enabled)
        {
            neighInfo_p[j].ipv6_addr.addr[0] = adk_ntohl(eth_neigh_table->tbl[i].ipv6_addr.addr[0]); 
            neighInfo_p[j].ipv6_addr.addr[1] = adk_ntohl(eth_neigh_table->tbl[i].ipv6_addr.addr[1]); 
            neighInfo_p[j].ipv6_addr.addr[2] = adk_ntohl(eth_neigh_table->tbl[i].ipv6_addr.addr[2]); 
            neighInfo_p[j].ipv6_addr.addr[3] = adk_ntohl(eth_neigh_table->tbl[i].ipv6_addr.addr[3]); 
            neighInfo_p[j].neigh_status = eth_neigh_table->tbl[i].neigh_status;
            if((neighInfo_p[j].neigh_status == ADK_ETHERNET_NEIGHBOR_REACHABLE) ||
                (neighInfo_p[j].neigh_status == ADK_ETHERNET_NEIGHBOR_STALE))
            {
               memcpy(neighInfo_p[j].dest_mac, eth_neigh_table->tbl[i].dest_mac, sizeof(adk_ethernet_mac_addr_t));     
            }
            j++;
        }
    }
    *neigh_entries_p = neighInfo_p;
    return ADK_SUCCESS;
}
static inline adk_status_t neighbor_entry_lookup
(
adk_dp_eth_neighbor_table_t *eth_neigh_table,
uint8_t *neighIpv6Addr,
adk_dp_eth_neighbor_entry_t **neigh_entry,
uint32_t key,
uint16_t hash_key
)
{
    uint16_t i; 
    adk_dp_eth_neighbor_hash_entry_t  *neigh_hash_entry = NULL;

    /* Lookup the Hash entry*/
    hash_key = hash_key*ADK_ETH_NO_OF_NEIGH_COLLISON_ENTRY;
    neigh_hash_entry = &eth_neigh_table->hash_tbl[hash_key];

    for(i=0; i<ADK_ETH_NO_OF_NEIGH_COLLISON_ENTRY; i++, neigh_hash_entry++)
    {
        if(neigh_hash_entry->hash_value != key)
            continue;

        
        *neigh_entry = &eth_neigh_table->tbl[neigh_hash_entry->index];

        if((*((uint64_t *)&neighIpv6Addr[0]) !=
                *((uint64_t *)&((*neigh_entry)->ipv6_addr.addr[0]))))
            continue;
    
        if((*((uint64_t *)&neighIpv6Addr[8]) !=
                *((uint64_t *)&((*neigh_entry)->ipv6_addr.addr[2]))))
            continue;

        return ADK_SUCCESS;
    }

    return ADK_DP_E_ERROR;
}

adk_dp_status_t
adk_eth_dp_arp_entry_lookup
(
 adk_dp_eth_arp_table_t *eth_arp_table,
 uint32_t ip_addr,
 adk_dp_eth_arp_entry_t **arp_entry
)
{
    adk_dp_eth_neighbor_hash_entry_t  *arp_hash_entry = NULL;
    uint32_t   key = 0;
    uint16_t hash_key, i;

    key = ADK_ETHERNET_HASH(key, ip_addr);

    hash_key = key & ADK_ETH_NEIGHBOR_HASH_MASK;

    /* Lookup the Hash entry*/
    hash_key = hash_key*ADK_ETH_NO_OF_NEIGH_COLLISON_ENTRY;
    arp_hash_entry = &eth_arp_table->hash_tbl[hash_key];

    for(i=0; i<ADK_ETH_NO_OF_NEIGH_COLLISON_ENTRY; i++, arp_hash_entry++)
    {   
        if(arp_hash_entry->hash_value != hash_key)
            continue;

        *arp_entry = &eth_arp_table->tbl[arp_hash_entry->index];

        if((ip_addr) != (*arp_entry)->ipv4_addr)
            continue;

        return ADK_SUCCESS;
    }

    /* check in the overflow list */
    adk_dp_eth_overflow_node_t *node;
    node=eth_arp_table->overflow_list.head;
    while(node)
    {
        /*compare destination address */
        *arp_entry = &eth_arp_table->tbl[node->idx];

        if((ip_addr) == (*arp_entry)->ipv4_addr)
        {
            return(ADK_SUCCESS);
        }
        node = node->next;
    }

    *arp_entry = NULL;

    return ADK_DP_E_ERROR;
}

adk_dp_status_t
adk_eth_dp_neighbor_entry_lookup
(
adk_dp_eth_neighbor_table_t *eth_neigh_table,
uint8_t *neighIpv6Addr,
adk_dp_eth_neighbor_entry_t **neigh_entry
)
{
    uint32_t key;
    uint16_t first_key, second_key;

    /* Generate key for neighbor dest table lookup */
    adk_eth_dp_neighbor_key_generate((uint32_t *)neighIpv6Addr,
            &key,
            &first_key,
            &second_key);

    /* Lookup the Hash entry using first key */
    if(ADK_SUCCESS == neighbor_entry_lookup(eth_neigh_table, neighIpv6Addr,
                        neigh_entry, key, first_key))
        return ADK_SUCCESS;


    /* Lookup the Hash entry using second key */
    if(ADK_SUCCESS == neighbor_entry_lookup(eth_neigh_table, neighIpv6Addr,
                        neigh_entry, key, second_key))
        return ADK_SUCCESS;

    /* check in the overflow list */
    adk_dp_eth_overflow_node_t *node;
    node=eth_neigh_table->overflow_list.head;
    while(node)
    {
        /*compare destination address */
        *neigh_entry = &eth_neigh_table->tbl[node->idx];

        if((*((uint64_t *)&neighIpv6Addr[0]) ==
                    *((uint64_t *)&((*neigh_entry)->ipv6_addr.addr[0]))) &&
                (*((uint64_t *)&neighIpv6Addr[8]) ==
                    *((uint64_t *)&((*neigh_entry)->ipv6_addr.addr[2]))))
        {
            return(ADK_SUCCESS);
        }
        node = node->next;
    }

    *neigh_entry = NULL;

    return ADK_DP_E_ERROR;
}

 /*!  end addtogroup _ethernetFunctions  Ethernet Module
 * @}
 */
