/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp_tx.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file contains the Ethernet data plane software for
 *          encapsulating L3 packet within L2 frame and preparing all
 *          the metadata for final transmission. It also updates the
 *          required port/VLAN level TX stats as, packet is not expected
 *          to get dropped after returning successfully from here.
 *
 *          Note: The tx pkts and octet counters, which counts all packets
 *          and octets given by the higher layer, is incremented at the
 *          end at the common exit point, by when output interface is known.
 *          While packet counter is incremented unconditionally, octet counter
 *          is incremented only when packet is not being dropped. Note that,
 *          there is no protocol specific drops possible in this Ethernet
 *          processing. All drops are only due to configuration errors. Also,
 *          note that the final octet stats reflects the Etheret header
 *          modifications done in this ULI processing.
 *
 *  **********************************************************************/

#include <adk_eth_dp_pvt.h>

/* Module info pointer stored for DP process to access tables */
extern adk_eth_dp_pvt_data_t *adk_eth_dp_pvt_g_p;

/* Ethernet global stats pointers */
extern adk_dp_eth_gl_stats_t *adk_eth_gl_stats_g_p;

/*The ethernet bridge timestamp quad group port*/
extern uint8_t *adk_eth_bridge_timestamp_quad_group_port_g;

void
adk_eth_dp_tx_md_update (
    adk_pktio_task_t *task_out,
    uint16_t eth_dev_id,
    uint32_t out_port,
    adk_dp_eth_qos_info_t *qos_queue,
    __attribute__ ((unused)) int coreIndex)
{
    adk_sys_egress_stats_mdata_t *sys_eg_stats_mdata_p = NULL;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task_out);

    adk_dsi_tx_metadata_t *dsi_md_p;
    adk_status_t ret = ADK_SUCCESS;
    uint8_t ts_index;
    
    dsi_md_p = ADK_DSI_TX_METADATA_PTR(task_out);

    ADK_MD_SET_ADK_DEV_ID(task_out, eth_dev_id);

    task_out->ol_flags |= PKT_TX_METADATA;
    dsi_md_p->overwrite = 1;
    ((adk_pktio_task_t *)task_out)->ol_flags &= ~PKT_TX_IEEE1588_TMST;
    dsi_md_p->dst_port = out_port;

	/*
	 * If time stamp capture and index allocation is enabled for the packet for which MAC lookup is successful,
	 * ADK need to allocate an index and notify Netdriver. 
	*/
	if(unlikely(ADK_MD_GET_TS_MSG_TYPE(task_out) == ADK_NETD_TS_CAPTURE_IDX_ALLOC))
	{
        INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_pkts_tstamp_alloc_requests);
        /*if the timestamp is required for bridge interface originated packets then the HLP quad group port 
        has to be set in netDriver metadata*/
        if(task_out->port >= ADK_ETH_BG_ID_OFFSET)
        {
           task_out->port = *adk_eth_bridge_timestamp_quad_group_port_g; 
        }
	    ret = adk_netd_fp_ts_index_alloc(task_out->port, ADK_MD_GET_TS_CORR(task_out), &ts_index);
	    /* 
	     * If index allocation failed, drop the packet 
	    */
	    if(ret != ADK_SUCCESS)
	    {
		    adk_eth_gl_stats_g_p[coreIndex].time_index_alloc_error++;
		    ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Error (ULI):"
		     " Time stamp index allocation failed. Dropping pkt!");
		    ret = ADK_E_INTERNAL_ERROR;
	    }   
        else
        {    /* Index allocation successful, set the DSI metadata fields */
            ((adk_pktio_task_t *)task_out)->ol_flags |= PKT_TX_IEEE1588_TMST;
		    dsi_md_p->tsyn_reg = ts_index;
            ADK_MD_SET_LINUX_INTF_ID(task_out, task_out->port);
            ADK_MD_SET_DEV_TX_NOTIFY_FN_ID(task_out, ADK_ETH_TS_PKT_TX_FN_ID);
            ADK_MD_SET_TS_INDEX(task_out, ts_index);
	    }
	}
	else if(unlikely(adk_md_p->ts_msg_type == ADK_NETD_TS_CAPTURE)) 
	{                                                                  
         ((adk_pktio_task_t *)task_out)->ol_flags |= PKT_TX_IEEE1588_TMST;
	     dsi_md_p->tsyn_reg = adk_md_p->timestamp_idx;
	}
    
    task_out->port = eth_dev_id;

#ifdef ADK_QOS_GEN3_INTERFACE
    ADK_METADATA_PARAM(task_out, dev_tx_queue_id) = qos_queue->queue_id;

    /* Update metadata for egress QoS queue statistics */
    sys_eg_stats_mdata_p =
        &adk_md_p->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_QUEUE_LOC];
    sys_eg_stats_mdata_p->stats_info = qos_queue->stats_id;
#else
    RTE_SET_USED(qos_queue);
#endif

    /* Update len correction metadata at IP egress stats location */
    sys_eg_stats_mdata_p =
        &adk_md_p->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_IP_LOC];
    if (sys_eg_stats_mdata_p->flags)
    {
        sys_eg_stats_mdata_p->len_correction =
            adk_eth_get_egress_pad_fcs_len_corr(task_out);
    }

    return;
}

 /*!  end addtogroup _ethernetFunctions  Ethernet Module
 * @}
 */
