/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp_uli.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ethernetFunctions  Ethernet Module
 *
 * @{
 *
 * @brief   This file contains the Ethernet data plane software to handle
 *          packets received from upper layer protocol modules.
 *
 *          Note: The tx pkts and octet counters, which counts all packets
 *          and octets given by the higher layer, is incremented at the
 *          end at the common exit point, by when output interface is known.
 *          While packet counter is incremented unconditionally, octet counter
 *          is incremented only when packet is not being dropped. Note that,
 *          there is no protocol specific drops possible in this Ethernet
 *          processing. All drops are only due to configuration errors. Also,
 *          note that the final octet stats reflects the Etheret header
 *          modifications done in this ULI processing.
 *
 *  **********************************************************************/

#include <adk_eth_dp_pvt.h>
#include <adk_error.h>
#if 0
#include <rte_pmd_ice_dsi.h>
#include <ies_sdk.h>
#include <ies_event_types.h>
#include <ies_api_port.h>
#include <ies_api_addr.h>
#endif

/* Next hop buffer timeout in milli second */
#define ETH_NEXT_HOP_BUFFER_TIMEOUT 10000

__thread adk_sif_dp_scratchpad_space_t *eth_scratch_pad_p = NULL;
__thread adk_sys_eth_dp_uli_flow_cache_entry_t *eth_action_params = NULL;

/* Module info pointer stored for DP process to access tables */
extern adk_eth_dp_pvt_data_t *adk_eth_dp_pvt_g_p;

/* Ethernet global stats pointers */
extern adk_dp_eth_gl_stats_t *adk_eth_gl_stats_g_p;

/* Macro definition for module level cycle count */
extern uint64_t adk_eth_uli_rx_ts_g, adk_eth_uli_tx_ts_g;
extern uint64_t adk_eth_lli_rx_ts_g, adk_eth_lli_tx_ts_g;

/* Pointer to the list of Ethernet bridge interface table */
extern adk_dp_eth_bridge_intf_t **adk_dp_eth_bridge_intf_t_g;

/* Map of bridge l2 domain to interface ID */
extern uint32_t* adk_dp_eth_bridge_map_id_g;

/* Arp table */
extern adk_dp_eth_arp_table_t *adk_eth_arp_hash_tbl_g;

/* Arp buffer table */
extern adk_eth_dp_arp_buffer_tbl_t *adk_eth_arp_buffer_tbl_g;

/* IPv6 Neighbor table*/
extern adk_dp_eth_neighbor_table_t *adk_eth_neighbor_hash_table_g_p;

/* IPv6 neighbor buffer table */
extern adk_eth_dp_neighbor_buffer_tbl_t *adk_eth_neighbor_buffer_tbl_g;

/* The ipv4 nexthop timer index table*/
extern adk_eth_dp_nexthop_v4_timer_index_tbl_t *adk_eth_nexthop_v4_timer_index_tbl_g;

/* The ipv6 nexthop timer index table*/
extern adk_eth_dp_nexthop_v6_timer_index_tbl_t *adk_eth_nexthop_v6_timer_index_tbl_g;

/*The eth buffer info mempool*/
extern struct rte_mempool *adk_eth_buffer_info_pool_g;

 /*The eth nexthop buffer mempool*/
extern struct rte_mempool *adk_eth_nexthop_buffer_pool_g;

/*The eth type to pcp map table*/
extern adk_lpm_handle_t adk_eth_ether_type_pcp_map_tbl_g;

/*The ethernet DSCP to PCP Map*/
extern uint8_t *adk_eth_dscp_to_pcp_map_g;

typedef struct task_ring_elm_s
{
    void* task;
    struct task_ring_elm_s* next;
    struct task_ring_elm_s* prev;
} task_ring_elm_t;

typedef struct task_ring_s
{
    uint32_t num_elms;
    uint32_t max_elms;
    task_ring_elm_t* head;
} task_ring_t;

static adk_status_t
adk_eth_dp_pkt_proc_uli_flow_cache_pkt_path_single (
        void *task,
        int coreIndex,
        uint16_t *destFunc);

adk_status_t adk_eth_dp_task_ring_create
(
task_ring_t** task_ring_p,
uint32_t max_elms
);

adk_status_t adk_eth_dp_task_ring_enqueue
(
task_ring_t* task_ring,
void* task,
void** last_task_removed
);

adk_status_t adk_eth_dp_task_ring_dequeue
(
task_ring_t* task_ring,
void** task
);

static void eth_dp_nexthop_buffer_timeout
(
void *timer_cb_info,
__attribute__((unused)) adk_timer_event_type_t event_type
);

static adk_dp_status_t eth_nexthop_v4_timer_create
(
adk_dp_eth_buffer_info_t* next_hop_buffer,
adk_ethernet_ipv4_addr_t *nexthop_addr
);

static adk_dp_status_t eth_nexthop_v6_timer_create
(
adk_dp_eth_buffer_info_t* next_hop_buffer,
adk_ethernet_ipv6_addr_t *nexthop_addr
);

uint32_t adk_eth_dp_task_ring_numElms(task_ring_t* task_ring);

adk_status_t adk_eth_dp_task_ring_create
(
task_ring_t** task_ring_p,
uint32_t max_elms
)
{
    adk_status_t ret = ADK_SUCCESS;
    if(NULL == task_ring_p)
    {
        ret = ADK_E_INVALID_PARAMETER; 
        return ret;
    }

    *task_ring_p = ADK_NVM_CALLOC(ADK_ETHERNET_MODULE_ID, 0, 1, sizeof(task_ring_t));
    if(NULL == *task_ring_p)
    {
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Memory allocation for task_ring fails!");
        ret = ADK_E_MEM_UNAVAILABLE;
        return ret;
    }
    (*task_ring_p)->max_elms = max_elms;
    return ret;
}

adk_status_t adk_eth_dp_task_ring_enqueue
(
task_ring_t* task_ring,
void* task,
void** last_task_removed
)
{
    adk_status_t ret = ADK_SUCCESS;
    task_ring_elm_t* task_ring_elm;
    task_ring_elm_t* task_ring_prev;

    if(NULL == last_task_removed)
    {
        ret = ADK_E_INVALID_PARAMETER;
        return ret;
    }

    *last_task_removed = NULL;

    if(task_ring->num_elms < task_ring->max_elms)
    {
        task_ring_elm = ADK_NVM_CALLOC(ADK_ETHERNET_MODULE_ID, 0, 1, sizeof(task_ring_elm_t));
        if(NULL == task_ring_elm)
        {
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Memory allocation for task_ring elm fails!");
            ret = ADK_E_MEM_UNAVAILABLE;
            return ret;
        }

        task_ring_elm->task = task;
        if(!task_ring->num_elms)
        {
            task_ring->head = task_ring_elm;
            task_ring_elm->next = task_ring_elm;
            task_ring_elm->prev = task_ring_elm;
            (task_ring->num_elms)++;
            return ret;
        }

        task_ring_prev = task_ring->head->prev;
        task_ring_prev->next = task_ring_elm;
        task_ring_elm->prev = task_ring_prev;
        task_ring_elm->next = task_ring->head;
        task_ring->head->prev = task_ring_elm;
        (task_ring->num_elms)++;
        return ret;
    }
    else
    {
        *last_task_removed = task_ring->head->task;
        task_ring->head->task = task; 
        task_ring->head = task_ring->head->next;
        return ret;
    }

    return ret;
}

adk_status_t adk_eth_dp_task_ring_dequeue
(
task_ring_t* task_ring,
void** task
)
{
    adk_status_t ret = ADK_SUCCESS;
    task_ring_elm_t* task_ring_elm_temp;

    if(NULL == task)
    {
        ret = ADK_E_INVALID_PARAMETER;
        return ret;
    }

    if(!task_ring->num_elms)
    {
        *task = NULL;
        ret = ADK_E_INCONSISTENT_STATE;
        return ret;
    }

    if(1 == task_ring->num_elms)
    {
        *task = task_ring->head->task;
        ADK_NVM_FREE(task_ring->head);
        task_ring->head = NULL;
        (task_ring->num_elms)--;
    }
    else
    {
        *task = task_ring->head->task;
        task_ring_elm_temp = task_ring->head;
        task_ring->head = task_ring->head->next;
        task_ring_elm_temp->prev->next = task_ring->head;
        task_ring->head->prev = task_ring_elm_temp->prev;
        ADK_NVM_FREE(task_ring_elm_temp);
        (task_ring->num_elms)--;
        return ADK_SUCCESS;
    }

    return ADK_SUCCESS;
}

uint32_t adk_eth_dp_task_ring_numElms(task_ring_t* task_ring)
{
    if(NULL != task_ring)
        return task_ring->num_elms;

    return 0;
} 

#ifdef ADK_DP_NO_BATCH_PROC

/*!
 * =============================================================
 * adk_eth_dp_pkt_proc_uli()
 *
 * Ethernet entry function for packets received from upper layer 
 * modules to be sent out.
 * =============================================================
 */

adk_status_t
adk_eth_dp_pkt_proc_uli (
        void **task_p,
        uint8_t task_count,
        uint32_t task_mask,
        __attribute__ ((unused)) void *ring,
        __attribute__ ((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
        int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    uint8_t pkt_drops;
    uint16_t destFunc = 0;
    bool netd_orig = false;
    eth_scratch_pad_p = ADK_SIF_DP_GET_SCRATCHPAD_PTR(coreIndex);

    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_ULI_RX_TS();

    if ((ADK_E_FLOW_CACHE_PKT_FAST_PATH == eth_scratch_pad_p->pktPath) &&
            (eth_scratch_pad_p->flow_cache_entry->eth_version == adk_eth_dp_pvt_g_p->eth_fc_ver)) {
            eth_action_params = &eth_scratch_pad_p->flow_cache_entry->ethActionParams;
        ret = adk_eth_dp_pkt_proc_uli_flow_cache_pkt_path_single(task_p[0], coreIndex, &destFunc);
    } else {
        if(NULL == eth_scratch_pad_p->flow_cache_entry)
        {
            eth_action_params = (adk_sys_eth_dp_uli_flow_cache_entry_t*)&eth_scratch_pad_p->data[0];
            netd_orig = true;
        }
        else
        {
            eth_action_params = &eth_scratch_pad_p->flow_cache_entry->ethActionParams;
        }

        ret = adk_eth_dp_pkt_proc_uli_single(task_p[0], coreIndex, &destFunc);
        if (likely(ADK_E_INTERNAL_ERROR != ret) && (!netd_orig))
        {
            eth_scratch_pad_p->flow_cache_entry->eth_version = adk_eth_dp_pvt_g_p->eth_fc_ver;
        }
    }

    /* If any fragment runs into error, drop entire original packet */
    if (unlikely(ADK_E_INTERNAL_ERROR == ret))
    {
        pkt_drops = adk_eth_task_chained_buffer_free(task_p[0]);
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                                                egressUtilityCounters.eth_eg_pkt_proc_uli_drop,pkt_drops);
        return ret;
    }

    /* If here, the ret value must be SUCCESS or CONSUMED. In either case, we
     * assume that this ret value would be same for all fragments including
     * the last fragment. */

    if (ADK_SUCCESS == ret)
    {
        /* Next stage info structure would be filled only if packet needs to go to some stage*/
        if (adkDestEntryFunction_g[destFunc])
        {
#if defined (ADK_PAL_DYNAMIC_PPL)
            ret = adk_pal_dp_stage_send((void **)task_p, task_count, task_mask, destFunc, 0x00,
                    0x00, ADK_PAL_DP_BATCH_CB_FN_ETH, ADK_PAL_DP_BATCH_CB_EVENT_STAGE_BUFFER, NULL);
            if(ret == ADK_E_INTERNAL_ERROR)
            {
                adk_eth_gl_stats_g_p[coreIndex].failed_to_sent_next_stage_drop++;
                adk_pal_dp_release_send((task_p[0]), ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
                adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[0]));
                ret = ADK_DP_E_ERROR;
                return ret;
            }
#else
            (adkDestEntryFunction_g[destFunc])(task_p, task_count, task_mask, ring_hdl,
                    next_stage_entry_hdl, coreIndex);
#endif
        }
    }
    else if (ADK_E_TASK_CONSUMED != ret)
    {
        /* Actual error logging should be available from SIF */
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Error (ULI): Failed to send"
                " packet!");
    }

    return ret;
}

#endif

adk_dp_status_t adk_eth_dp_encap_and_send_pkt
(
void* task,
uint8_t *dest_mac,
int coreIndex,
bool flush
)
{
    adk_dp_eth_qos_info_t qos_queue;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_pktio_task_t *next_task_p = NULL;
    adk_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t *task_out = (adk_pktio_task_t *)task;
    uint32_t vlan = 0;
    uint16_t *tmp_16 = NULL;
    uint32_t *tmp_32 = NULL;
    uint8_t *pdu = NULL;
    uint8_t pkt_drops = 0;
    uint8_t pktColor;
    uint16_t dev_tx_queue_id;
    uint16_t frag_count = 0;
    adk_dsi_tx_metadata_t *dsi_md_p = NULL;

    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    frag_count = eth_uli_imc_param_p->frag_count;
    pktColor = ADK_METADATA_PARAM(task_out, pktColor);
    dev_tx_queue_id = ADK_METADATA_PARAM(task_out, dev_tx_queue_id);

    do {
        next_task_p = ADK_METADATA_PARAM(task_out, next_frag);
        /* Use this as default as of now */
        qos_queue.info = 0;

        if ((eth_action_params->vlanId) ||
                (eth_action_params->priorityTag))
        {
            /* encapsulation with ethernet header */
            ADK_PDU_HEAD_PUSH(task_out, pdu, (uint8_t *), ETH_VLAN_HEADER_LEN);

            if(pdu == NULL)
            {
                pkt_drops = adk_eth_task_chained_buffer_free(task_out);
                adk_eth_gl_stats_g_p[coreIndex].egress_encap_pkt_failure_drop += pkt_drops;
                return ADK_E_TASK_CONSUMED;
            }

            if (IS_BROADCAST_MAC(dest_mac)) {
                adk_md_p->isEthernetIpBcast = true;
                adk_md_p->isEthernetIpMcast = false;
            }else if ((dest_mac[0] & 0x01)) {
                adk_md_p->isEthernetIpMcast = true;
                adk_md_p->isEthernetIpBcast = false;
            }else {
                adk_md_p->isEthernetIpBcast = false;
                adk_md_p->isEthernetIpMcast = false;
            }
            adk_md_p->isEthernetNonIpBcast = false;
            adk_md_p->isEthernetNonIpMcast = false;

            ADK_MEMCPY(&pdu[0], dest_mac, sizeof(uint8_t) * 6);
            ADK_MEMCPY(&pdu[6], eth_action_params->srcMac, sizeof(uint8_t) * 6);
            vlan = ((0x8100 << 16) |
                    (eth_action_params->pcpVal << 13) |
                    (eth_action_params->vlanId & 0xFFF));

            tmp_32 = (uint32_t *) (void *) &pdu[12];
            *tmp_32 = adk_htonl(vlan);
            tmp_16 = (uint16_t *) (void *) &pdu[16];

        }
        else
        {
            /* encapsulation with ethernet header */
            ADK_PDU_HEAD_PUSH(task_out, pdu, (uint8_t *), ETH_NO_VLAN_HEADER_LEN);

            if(pdu == NULL)
            {
                pkt_drops = adk_eth_task_chained_buffer_free(task_out);
                adk_eth_gl_stats_g_p[coreIndex].egress_encap_pkt_failure_drop += pkt_drops;
                return ADK_E_TASK_CONSUMED;
            }

            ADK_MEMCPY(&pdu[0], dest_mac, sizeof(uint8_t) * 6);
            ADK_MEMCPY(&pdu[6], eth_action_params->srcMac, sizeof(uint8_t) * 6);
            tmp_16 = (uint16_t *) (void *) &pdu[12];
        }

        if(eth_action_params->isV6)
        {
            *tmp_16 = adk_htons(ETHER_TYPE_IPV6);
        }
        else
        {
            *tmp_16 = adk_htons(ETHER_TYPE_IPV4);
        }

#ifdef ADK_DEBUG
        adk_eth_dp_print_pdu_hdr(pdu);
#endif

#ifdef ADK_TX_CKSUM_OL
        if((task_out->ol_flags & PKT_TX_OUTER_IPV4) || (task_out->ol_flags & PKT_TX_OUTER_IPV6))
        {
            task_out->outer_l2_len = (vlan) ? ETH_VLAN_HEADER_LEN : ETH_NO_VLAN_HEADER_LEN;
        }
        else if ((task_out->ol_flags & PKT_TX_IPV4) || (task_out->ol_flags & PKT_TX_IPV6))
        {
            task_out->l2_len = (vlan) ? ETH_VLAN_HEADER_LEN : ETH_NO_VLAN_HEADER_LEN;
        }
#endif
        adk_md_p = ADK_METADATA_PTR(task_out);

        /*Updating the IP offset after encapsulation*/
        adk_md_p->ip_offset = (vlan) ? ETH_VLAN_HEADER_LEN : ETH_NO_VLAN_HEADER_LEN; 
        adk_md_p->l4_offset += (vlan) ? ETH_VLAN_HEADER_LEN : ETH_NO_VLAN_HEADER_LEN;

        adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_ETHERNET_INTERFACE_STATS].stats_id =
            eth_action_params->ethIntfStatsId;

        adk_eth_dp_tx_md_update(task_out, eth_action_params->ethDevId,
                ADK_HOS_CPK_PORT, &qos_queue, coreIndex);
        dsi_md_p = ADK_DSI_TX_METADATA_PTR(task_out);
        dsi_md_p->l2_domain = eth_action_params->l2Domain;

        if(!flush && !next_task_p) 
        {
            ADK_METADATA_PARAM(task_out, task_type) = ADK_TASK_TYPE_FWD;
        }
        else
        {
            ADK_METADATA_PARAM(task_out, task_type) = ADK_TASK_TYPE_NEW;
        }

        ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_ULI_TX_TS();
        ret = adk_pal_dp_nic_send(task_out->port, pktColor, dev_tx_queue_id, task_out, ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
        if(ret == ADK_E_INTERNAL_ERROR)
        {
            pkt_drops = adk_eth_task_chained_buffer_free(task_out);
            adk_eth_gl_stats_g_p[coreIndex].nic_send_failure_drop += pkt_drops;
            return ADK_E_TASK_CONSUMED;
        }
        INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                egressUtilityCounters.eth_eg_pkts_sent_to_nic);
        task_out = next_task_p;
    }while(task_out);

    return ADK_E_TASK_CONSUMED;
}

adk_status_t adk_eth_dp_decode_interface
(
void* task,
adk_dp_eth_svif_intf_t **dp_svif_intf_pp,
adk_dp_eth_bridge_intf_t  **dp_bridge_intf_pp,
int coreIndex
)
{
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    adk_dsi_tx_metadata_t *dsi_md_p = ADK_DSI_TX_METADATA_PTR(task);;
    adk_dp_eth_bridge_intf_t *dp_bridge_intf_p;
    uint32_t bridge_stats_id;
    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    if (eth_uli_imc_param_p->output_intf_id >= ADK_ETH_NUM_BRIDGE)
    {
#ifdef ADK_DEBUG
       debug_print("ULI IMC output_intf_id %d\n", eth_uli_imc_param_p->output_intf_id);
#endif
        *dp_svif_intf_pp = adk_eth_svif_if_g[eth_uli_imc_param_p->output_intf_id - ADK_ETH_NUM_BRIDGE];
        if(NULL == *dp_svif_intf_pp)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "svif is not defined for output_intf_id"
                              " %d", eth_uli_imc_param_p->output_intf_id);
            return ADK_E_INTERNAL_ERROR;            
        }

        dsi_md_p->l2_domain = (*dp_svif_intf_pp)->bgid;
        bridge_stats_id = adk_dp_eth_bridge_map_id_g[(*dp_svif_intf_pp)->bgid];
        dp_bridge_intf_p = adk_dp_eth_bridge_intf_t_g[bridge_stats_id];
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->ifHandle,
                                                    (*dp_svif_intf_pp)->svifIntfHandle);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->vlanId,
                                                    (*dp_svif_intf_pp)->vlan);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->ethIntfStatsId,
                                                    ((*dp_svif_intf_pp)->statsId + ADK_ETH_NUM_BRIDGE));
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->ethDevId,
                                                    dp_bridge_intf_p->eth_dev_id);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->priorityTag,
                                                    false);
        ADK_MEMCPY(eth_action_params->srcMac,dp_bridge_intf_p->src_mac,
                                                    sizeof(eth_action_params->srcMac));
#ifdef ADK_DEBUG
        debug_print("dsi_md_p->l2_domain in uli %d\n", dsi_md_p->l2_domain);
        debug_print("dsi_md_p->dst_port in uli %d\n", dsi_md_p->dst_port);
#endif
    }
    else
    {
        *dp_bridge_intf_pp = adk_dp_eth_bridge_intf_t_g[eth_uli_imc_param_p->output_intf_id];   
        if(NULL == *dp_bridge_intf_pp)
        {
            adk_eth_gl_stats_g_p[coreIndex].intf_not_defined++;
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Bridge interface is"
                              " not defined for L2 domain %d", dsi_md_p->l2_domain);
            return ADK_E_INTERNAL_ERROR;            
        }
 
        dsi_md_p->l2_domain = (*dp_bridge_intf_pp)->l2_domain_id;
#ifdef ADK_DEBUG
        debug_print("dsi_md_p->l2_domain in uli %d\n", dsi_md_p->l2_domain);
        debug_print("dsi_md_p->dst_port in uli %d\n", dsi_md_p->dst_port);
#endif
        ADK_MEMCPY(eth_action_params->srcMac,(*dp_bridge_intf_pp)->src_mac,
                sizeof(eth_action_params->srcMac));
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->ifHandle,
                (*dp_bridge_intf_pp)->bridgeIntfHandle);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->vlanId,0);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->priorityTag,
                (*dp_bridge_intf_pp)->priority_tag);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->ethIntfStatsId,
                (*dp_bridge_intf_pp)->statsId);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->ethDevId,
                (*dp_bridge_intf_pp)->eth_dev_id);

    }
    ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->l2Domain,
            dsi_md_p->l2_domain);
    return ADK_SUCCESS;
}


static adk_status_t adk_eth_dp_send_exception(
void *task,
adk_ethernet_neigh_status_t nexthop_status,
int coreIndex
)
{
    adk_status_t ret = ADK_SUCCESS;
    struct rte_mbuf *new_task;
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_ethernet_dp_next_hop_exception_data_t nexthop_exception_data;
    uint64_t *nexthop_exception_sent_count = NULL;
    uint8_t exception_type = ADK_ETHERNET_EXCEPTION_TYPE_INVALID;
    adk_pktio_task_t *cur_task_p, *next_task_p = NULL;
    uint8_t frag_proc = 0;
 
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);

    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;
    cur_task_p = (adk_pktio_task_t *)(task);

    while(cur_task_p)
    {
        next_task_p = ADK_METADATA_PARAM(cur_task_p, next_frag);
        nexthop_exception_data.ifHandle = eth_action_params->ifHandle;
        nexthop_exception_data.vlanId = eth_action_params->vlanId;

        if(ADK_ETHERNET_NEIGHBOR_STALE == nexthop_status)
        {
            nexthop_exception_sent_count = &adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_pkt_stale_entry_used_exception_sent;
            exception_type = ADK_ETHERNET_EXCEPTION_TYPE_NEIGHBOR_STALE_ENTRY_USED;
        }

        if(!eth_uli_imc_param_p->next_hop.is_ipv6)
        {
            nexthop_exception_data.next_hop.v4_address = eth_uli_imc_param_p->next_hop.u.ipv4_addr;
            nexthop_exception_data.isIPv6 = false;

            if(ADK_ETHERNET_NEIGHBOR_INVALID == nexthop_status) 
            {
                nexthop_exception_sent_count = &adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_pkt_arp_exception_sent;
                exception_type = ADK_ETHERNET_EXCEPTION_TYPE_UNRESOLVED_ARP;
            }
        }
        else
        {
            nexthop_exception_data.next_hop.v6_address.addr[0] = adk_ntohl(eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[0]);
            nexthop_exception_data.next_hop.v6_address.addr[1] = adk_ntohl(eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[1]);
            nexthop_exception_data.next_hop.v6_address.addr[2] = adk_ntohl(eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[2]);
            nexthop_exception_data.next_hop.v6_address.addr[3] = adk_ntohl(eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[3]);
            nexthop_exception_data.isIPv6 = true;

            if(ADK_ETHERNET_NEIGHBOR_INVALID == nexthop_status) 
            {
                exception_type = ADK_ETHERNET_EXCEPTION_TYPE_UNRESOLVED_IPV6_NEIGHBOR;
                nexthop_exception_sent_count = &adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].egressUtilityCounters.eth_eg_pkt_nd_exception_sent;
            }
        }

        if(exception_type == ADK_ETHERNET_EXCEPTION_TYPE_INVALID)
        {
            return ADK_SUCCESS;
        }
        ret = adk_eth_packet_clone(task, &new_task, ADK_ETH_PACKET_CLONE_MAX_SIZE, ADK_ETH_PACKET_CLONE_SEGMENTS);

        if(ret != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Failed to clone the packet for raising the exception\n");
            adk_eth_gl_stats_g_p[coreIndex].pkt_clone_failure++;
            adk_eth_gl_stats_g_p[coreIndex].exceptionDrops += (eth_uli_imc_param_p->frag_count-frag_proc);
            return ADK_E_INTERNAL_ERROR;
        }

        ret = adk_eth_dp_handle_exception(new_task, exception_type, &nexthop_exception_data);
        if (ADK_E_DROP_TASK == ret)
        {
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(new_task));
            /*send the packets out if entry in stale state eventhough
              exception could not be sent*/
            if(ADK_ETHERNET_EXCEPTION_TYPE_NEIGHBOR_STALE_ENTRY_USED == exception_type)
            {
                ADK_ETH_DP_LOG_INFO_DEBUG("Failed to raise the exception for stale entry used ");
                return ADK_SUCCESS;
            }
            else
            {
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Failed to raise the exception,Dropping packets\n");
                adk_eth_gl_stats_g_p[coreIndex].exceptionDrops += (eth_uli_imc_param_p->frag_count-frag_proc);
                return ADK_E_INTERNAL_ERROR;
            }
        }

        ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Successfully raised the exception\n");
        INCREMENT_ETH_UTILITY_COUNTER_P(nexthop_exception_sent_count);
        INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                                                egressUtilityCounters.eth_eg_pkts_sent_to_ph);
        cur_task_p = next_task_p;
        frag_proc++;
    }
    /* For everything else, it's a SUCCESS */
    return ADK_SUCCESS;
}

static adk_status_t eth_nexthop_buffer_info_free
(
adk_dp_eth_buffer_info_t *next_hop_buffer_info,
bool release_lock
)
{
    adk_status_t ret = ADK_SUCCESS;
    void* nh_lock;
    if(NULL != next_hop_buffer_info)
    {
	    if(NULL != next_hop_buffer_info->nh_lock)
	    {
		    nh_lock = next_hop_buffer_info->nh_lock;
		    if(release_lock)
		    {
			    next_hop_buffer_info->nh_lock = NULL;
			    ADK_RELEASE_LOCK(nh_lock,ret);
		    }
		    ADK_DESTROY_LOCK(nh_lock);
		    ADK_FREE_LOCK(nh_lock);
	    }
	    rte_mempool_put(adk_eth_buffer_info_pool_g, next_hop_buffer_info);
    }
_lock_release_failed:
    return ret;
}

static adk_status_t eth_dp_ipv6_nexthop_buffer_timeout
(
uint64_t timer_index
)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_dp_eth_buffer_info_t* next_hop_buffer_p = NULL;
    adk_dp_eth_neighbor_entry_t *neigh_entry = NULL;
    bool nexthop_entry_found = false;

    INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                            egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timeout_count);
    ADK_DP_LOG_INFO_4D_DEBUG(ADK_ETHERNET_MODULE_ID,"nexthop buffer timeout called for IPV6 Addr : %x:%x:%x:%x",
            adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[0],
            adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[1],
            adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[2],
            adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[3]);

    ret = adk_eth_dp_neighbor_entry_lookup(adk_eth_neighbor_hash_table_g_p,
            (uint8_t*)adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr,
            &neigh_entry);
    if (ADK_SUCCESS == ret)
    {
        nexthop_entry_found = true;
    }

    ADK_ACQUIRE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);

    adk_eth_dp_neighbor_buffer_get(adk_eth_neighbor_buffer_tbl_g,
            (uint8_t*)adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr,
            &next_hop_buffer_p);

    if(NULL == next_hop_buffer_p)
    {
        ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);
        ADK_ETH_DP_LOG_INFO_DEBUG("nexhthop buffer is NULL in IPv6 Timeout callback");
        return ret;
    }

    ADK_ACQUIRE_LOCK(next_hop_buffer_p->nh_lock, ret);
    if(!nexthop_entry_found)
    {
        adk_eth_dp_neighbor_buffer_rem(adk_eth_neighbor_buffer_tbl_g,
                (uint8_t*)adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr);
        adk_eth_dp_nexthop_buffer_free(next_hop_buffer_p);
        ret = eth_nexthop_buffer_info_free(next_hop_buffer_p, true);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
        }
    }
    else if(!neigh_entry->have_inflight && !next_hop_buffer_p->inflight_count)
    {
        adk_eth_dp_neighbor_buffer_rem(adk_eth_neighbor_buffer_tbl_g,
                (uint8_t*)adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr);
        ret = eth_nexthop_buffer_info_free(next_hop_buffer_p, true);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
        }
    }
    else
    {
        /*restart the timer since packets are in inflight*/
        ADK_RELEASE_LOCK(next_hop_buffer_p->nh_lock, ret);
        ret = eth_nexthop_v6_timer_create(next_hop_buffer_p, &adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_restart_error);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"Could not restart the timer after timer expiry\n");
        }
    }
    ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);
_lock_release_failed:
_lock_acquire_failed:
    return ret;
}

static adk_status_t eth_dp_ipv4_nexthop_buffer_timeout
(
uint32_t timer_index
)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_dp_eth_buffer_info_t* next_hop_buffer_p = NULL;
    adk_dp_eth_arp_entry_t* arp_entry = NULL;
    bool nexthop_entry_found = false;

    INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                        egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timeout_count);
    ADK_DP_LOG_INFO_1D_DEBUG(ADK_ETHERNET_MODULE_ID,"nexthop buffer timeout called for IPv4 Addr %x",
            adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr);

    ret =  adk_eth_dp_arp_entry_lookup(adk_eth_arp_hash_tbl_g, 
                adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr, &arp_entry);
    if(ADK_DP_E_ERROR != ret)
    {
        nexthop_entry_found = true;
    }

    ADK_ACQUIRE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);

    adk_eth_dp_arp_buffer_get(adk_eth_arp_buffer_tbl_g,
            adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr,
            &next_hop_buffer_p);

    if(NULL == next_hop_buffer_p)
    {
        ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);
        ADK_ETH_DP_LOG_INFO_DEBUG("nexhthop buffer is NULL in IPv4 Timeout callback");
        return ret;
    }

    ADK_ACQUIRE_LOCK(next_hop_buffer_p->nh_lock, ret);
    if(!nexthop_entry_found)
    {
        adk_eth_dp_arp_buffer_rem(adk_eth_arp_buffer_tbl_g,
                adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr);
        adk_eth_dp_nexthop_buffer_free(next_hop_buffer_p);
        ret = eth_nexthop_buffer_info_free(next_hop_buffer_p, true);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
        }
    }
    else if (!arp_entry->have_inflight && !next_hop_buffer_p->inflight_count)
    {
        adk_eth_dp_arp_buffer_rem(adk_eth_arp_buffer_tbl_g,
                adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr);
        ret = eth_nexthop_buffer_info_free(next_hop_buffer_p, true);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
        }
    }
    else
    {
        /*restart the timer since packets are in inflight*/
        ADK_RELEASE_LOCK(next_hop_buffer_p->nh_lock, ret);
        ret = eth_nexthop_v4_timer_create(next_hop_buffer_p,&adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_restart_error);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"Could not restart the timer after timer expiry\n");
        }
    }
    ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);
_lock_release_failed:
_lock_acquire_failed:
    return ret;
}

static void eth_dp_nexthop_buffer_timeout
(
 void *timer_cb_info,
 __attribute__((unused)) adk_timer_event_type_t event_type
)
{
    uint64_t timer_index = *(uint32_t*)timer_cb_info;
    adk_dp_status_t ret = ADK_SUCCESS;
 
    ADK_DP_LOG_INFO_1D_DEBUG(ADK_ETHERNET_MODULE_ID, "Nexthop buffer timeout called for timer index %lu\n",timer_index);

    if(ADK_TIMER_EVENT_TYPE_TIMEOUT == event_type)
    {
        if(timer_index < ADK_ETH_ARP_BUFFER_TIMER_INDEX_MAX)
        {
            ret = eth_dp_ipv4_nexthop_buffer_timeout(timer_index);
        }
        else
        {
            timer_index -= ADK_ETH_ARP_BUFFER_TIMER_INDEX_MAX;
            ret = eth_dp_ipv6_nexthop_buffer_timeout(timer_index);
        }
    }
    if(ret != ADK_SUCCESS)
    {
        adk_eth_gl_stats_g_p[ADK_LCORE_INDEX].nexthop_buffer_timeout_error++;
    }
}

static adk_dp_status_t eth_nexthop_v4_timer_create
(
adk_dp_eth_buffer_info_t* next_hop_buffer,
adk_ethernet_ipv4_addr_t *nexthop_addr
)
{
    uint64_t timer_index;
    /* start the nexthop timer */
    adk_timer_info_t timer_info;

    timer_info.interval_milli_sec = ETH_NEXT_HOP_BUFFER_TIMEOUT;

    timer_index = next_hop_buffer->hash_index;
    adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].next_hop_addr = *nexthop_addr;
    adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].nexthop_timer_index = timer_index;

    /* TODO: Timer APIs should follow the ADK convention of returns */ 
    int ret = adk_timer_create(/*coreIndex*/ ADK_LCORE_INDEX, 
            timer_info, 
            eth_dp_nexthop_buffer_timeout, 
            &adk_eth_nexthop_v4_timer_index_tbl_g[timer_index].nexthop_timer_index,
            &(next_hop_buffer->timer_hdl)); 

    if(!ret)
    {
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                "Error whle registering ipv4 nexthop buffer timer!!!");
        return ADK_E_INTERNAL_ERROR;
    }

    RTE_SET_USED(next_hop_buffer);
    return ADK_SUCCESS;
}

static adk_dp_status_t eth_nexthop_v6_timer_create
(
adk_dp_eth_buffer_info_t* next_hop_buffer,
adk_ethernet_ipv6_addr_t *nexthop_addr
)
{
    uint64_t timer_index;
    /* start the nexthop timer */
    adk_timer_info_t timer_info;

    timer_info.interval_milli_sec = ETH_NEXT_HOP_BUFFER_TIMEOUT;

    timer_index = next_hop_buffer->hash_index;
    adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[0] = nexthop_addr->addr[0];
    adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[1] = nexthop_addr->addr[1];
    adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[2] = nexthop_addr->addr[2];
    adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].next_hop_addr.addr[3] = nexthop_addr->addr[3];
    adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].nexthop_timer_index = timer_index + ADK_ETH_ARP_BUFFER_TIMER_INDEX_MAX;

    /* TODO: Timer APIs should follow the ADK convention of returns */
    int ret = adk_timer_create(/*coreIndex*/ ADK_LCORE_INDEX,
            timer_info,
            eth_dp_nexthop_buffer_timeout,
            &adk_eth_nexthop_v6_timer_index_tbl_g[timer_index].nexthop_timer_index,
            &(next_hop_buffer->timer_hdl));

    if(!ret)
    {
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                            "Error whle registering nexthop buffer timer!!!");
        return ADK_E_INTERNAL_ERROR;
    }

    RTE_SET_USED(next_hop_buffer);
    return ADK_SUCCESS;
}

static adk_dp_status_t eth_nexthop_buffer_info_create
(
adk_dp_eth_buffer_info_t **next_hop_buffer_p,
adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p
)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_dp_eth_buffer_info_t* next_hop_buffer;

    /* TODO: RTE mempool should be used for DP object creation.
       That will make things much faster */
    if(rte_mempool_get(adk_eth_buffer_info_pool_g, (void **)next_hop_buffer_p))
    {
        ret = ADK_E_MEM_UNAVAILABLE;
        adk_eth_gl_stats_g_p[ADK_LCORE_INDEX].buffering_memory_not_available_drop += eth_uli_imc_param_p->frag_count;
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                "Failed to allocate DP memory for buffering nexthop");
    }
    next_hop_buffer = *next_hop_buffer_p;

    memset(next_hop_buffer, 0x00, sizeof(adk_dp_eth_buffer_info_t));

    /* Create nexthop lock */
    next_hop_buffer->nh_lock = (void *) ADK_ALLOC_LOCK();
    if(NULL == next_hop_buffer->nh_lock)
    {
        ret = ADK_E_MEM_UNAVAILABLE;
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                egressUtilityCounters.eth_eg_nexthop_buffer_lock_alloc_error,
                                                eth_uli_imc_param_p->frag_count);
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                "could not allocate memory for lock\n");
        return ret;
    }
    ADK_INIT_LOCK(next_hop_buffer->nh_lock,
            ADK_LOCK_SHARED,
            ADK_LOCK_VALUE);
    return ret;
}

static adk_dp_status_t eth_nexthop_buffer_update(
        void *task,
        adk_dp_eth_buffer_info_t *next_hop_buffer_p,
        int coreIndex
        )
{
    adk_dp_status_t ret = ADK_SUCCESS;
    bool send_exception = false;

    /* Once the hash table operation completes, subsequent operations
       can be done using next hop lock.*/

    ADK_ACQUIRE_LOCK(next_hop_buffer_p->nh_lock, ret);
    if(next_hop_buffer_p->pkt_count == 0)
    {
        send_exception = true;
    }
    ADK_RELEASE_LOCK(next_hop_buffer_p->nh_lock, ret);

    if((send_exception) && (ADK_ETHERNET_NEIGHBOR_INVALID == eth_action_params->neighStatus))
    {
        if(ADK_SUCCESS != (ret = adk_eth_dp_send_exception(task,
                        eth_action_params->neighStatus, coreIndex)))
        {
            return ret;
        }
    }

    ADK_ACQUIRE_LOCK(next_hop_buffer_p->nh_lock, ret);

    next_hop_buffer_p->pkt_count++;
    if(next_hop_buffer_p->pkt_count >= ADK_ETH_UNRESOLVED_ARP_EXCEPTION_PKT_COUNT)
    {
        next_hop_buffer_p->pkt_count = 0;
    }
    /* Atomic increment of inflight packet count
    TODO: It might be sufficient to take cache line
    lock for this */
    next_hop_buffer_p->inflight_count++;

    ADK_RELEASE_LOCK(next_hop_buffer_p->nh_lock, ret);
_lock_release_failed:
_lock_acquire_failed:
     return ret;
}

static adk_dp_status_t eth_nexthop_buffer_ipv4_proc
(
adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p,
adk_dp_eth_buffer_info_t **next_hop_buffer_p
)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    
    /* Aquire hash lock */
    ADK_ACQUIRE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);

    adk_eth_dp_arp_buffer_get(adk_eth_arp_buffer_tbl_g,  eth_uli_imc_param_p->next_hop.u.ipv4_addr,
            next_hop_buffer_p);

    if(NULL == *next_hop_buffer_p)
    {
        if(ADK_SUCCESS != eth_nexthop_buffer_info_create(next_hop_buffer_p, eth_uli_imc_param_p))
        {
            ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);
            ret = eth_nexthop_buffer_info_free(*next_hop_buffer_p, false);  
            if(ADK_SUCCESS != ret)
            {
                INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                            egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error,
                                                            eth_uli_imc_param_p->frag_count);
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
            }
            return ret;
        }

        (*next_hop_buffer_p)->is_ipv6 = false;
        ret = adk_eth_dp_arp_buffer_add(adk_eth_arp_buffer_tbl_g,  eth_uli_imc_param_p->next_hop.u.ipv4_addr,
                *next_hop_buffer_p);

        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                        egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_overflow_error,
                                                        eth_uli_imc_param_p->frag_count);
            ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                    "could not insert next hop buffer in hash table for IPv4\n");
            ret = eth_nexthop_buffer_info_free(*next_hop_buffer_p, false);
            if(ADK_SUCCESS != ret)
            {
                INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                            egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error,
                                                            eth_uli_imc_param_p->frag_count);
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
            }

            return ret;
        }
        /* Create timer object */
        ret = eth_nexthop_v4_timer_create(*next_hop_buffer_p,&eth_uli_imc_param_p->next_hop.u.ipv4_addr);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                    egressUtilityCounters.eth_eg_ipv4_nexthop_buffer_timer_create_failure,
                                                    eth_uli_imc_param_p->frag_count);
            return ret;
        }
    }

    ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);

_lock_release_failed:
_lock_acquire_failed:
     return ret;
}

static adk_dp_status_t eth_nexthop_buffer_ipv6_proc
(
adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p,
adk_dp_eth_buffer_info_t **next_hop_buffer_p
)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    
    /* Aquire hash lock */
    ADK_ACQUIRE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);

    adk_eth_dp_neighbor_buffer_get(adk_eth_neighbor_buffer_tbl_g,
        (uint8_t*)eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr, next_hop_buffer_p);

    if(NULL == *next_hop_buffer_p)
    {
        if(ADK_SUCCESS != eth_nexthop_buffer_info_create(next_hop_buffer_p, eth_uli_imc_param_p))
        {
            ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);
            ret = eth_nexthop_buffer_info_free(*next_hop_buffer_p, false);       
            if(ADK_SUCCESS != ret)
            {
                INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                        egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error,
                                                        eth_uli_imc_param_p->frag_count);
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
            }
            return ret;
        }

        (*next_hop_buffer_p)->is_ipv6 = true;
        ret = adk_eth_dp_neighbor_buffer_add(adk_eth_neighbor_buffer_tbl_g,
                (uint8_t*)eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr, *next_hop_buffer_p);

        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                    egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_overflow_error,
                                                    eth_uli_imc_param_p->frag_count);
            ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                    "could not insert next hop buffer in hash table for IPv6\n");
            ret = eth_nexthop_buffer_info_free(*next_hop_buffer_p, false);
            if(ADK_SUCCESS != ret)
            {
                INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                        egressUtilityCounters.eth_eg_nexthop_buffer_info_free_error,
                                                        eth_uli_imc_param_p->frag_count);
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"nexthop buffer info free failed\n");
            }
            return ret;
        }
        /* Create timer object */
        ret = eth_nexthop_v6_timer_create(*next_hop_buffer_p,&eth_uli_imc_param_p->next_hop.u.ipv6_addr);
        if(ADK_SUCCESS != ret)
        {
            INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                                                        egressUtilityCounters.eth_eg_ipv6_nexthop_buffer_timer_create_failure,
                                                        eth_uli_imc_param_p->frag_count);
            return ret;
        }
    }
    ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);

_lock_release_failed:
_lock_acquire_failed:
     return ret;
}

static adk_dp_status_t eth_nexthop_buffer_proc
(
adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p,
adk_dp_eth_buffer_info_t **next_hop_buffer_p
)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    if(!eth_uli_imc_param_p->next_hop.is_ipv6)
    {
        ret = eth_nexthop_buffer_ipv4_proc(eth_uli_imc_param_p,
            next_hop_buffer_p);
        return ret;
    }
    else
    {
        ret = eth_nexthop_buffer_ipv6_proc(eth_uli_imc_param_p,
            next_hop_buffer_p);
        return ret;
    }

    return ret;
}

static adk_status_t eth_handle_slowpath(
        void *task,
        int coreIndex,
        uint16_t *destFunc,
        int16_t *stale_exception_raised)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_dp_eth_buffer_info_t *next_hop_buffer = NULL;
    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    if((ADK_ETHERNET_NEIGHBOR_STALE == eth_action_params->neighStatus) &&
            (stale_exception_raised != NULL))
    {
        if(rte_atomic16_test_and_set((rte_atomic16_t*)stale_exception_raised))
        {
            if(ADK_SUCCESS != (ret = adk_eth_dp_send_exception(task,
                            eth_action_params->neighStatus, coreIndex)))
            {
                return ret;
            }
        }
    }

    if((ADK_ETHERNET_NEIGHBOR_REACHABLE == eth_action_params->neighStatus ||
                ADK_ETHERNET_NEIGHBOR_STALE == eth_action_params->neighStatus) &&
            (eth_action_params->haveInflight &&
             !(*eth_action_params->haveInflight)))
    {
        return ret;
    }


    if(ADK_ETHERNET_NEIGHBOR_INVALID == eth_action_params->neighStatus ||
            ADK_ETHERNET_NEIGHBOR_BUFFER_WITHOUT_NOTIFYING == eth_action_params->neighStatus ||
            (eth_action_params->haveInflight &&
             *eth_action_params->haveInflight))
    {
        *destFunc = ADK_ETH_ULI_PKT_BUFFER_FN_ID;
        /* Lookup hash table for arp/neighbor buffer*/
        ret = eth_nexthop_buffer_proc(eth_uli_imc_param_p, &next_hop_buffer);
        if(ADK_SUCCESS != ret)
        {
            adk_eth_gl_stats_g_p[coreIndex].nexthop_buffer_proc_drop += eth_uli_imc_param_p->frag_count;
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                    "Nexthop buffer create/get fails");
            return ret;
        }

        /*This function will update the packet count and 
          raise the exception 1 in n. This will will also take 
          the nexthop lock and Update infligh count for arp/neighbor buffer */
        ret = eth_nexthop_buffer_update(task, next_hop_buffer, coreIndex);
        if(ADK_SUCCESS != ret)
        {
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                    "Nexthop buffer Update fails");
            return ret;
        }


        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                egressUtilityCounters.eth_eg_pkt_sent_to_buffering_stage,
                eth_uli_imc_param_p->frag_count);
        /* Delete any pending timer objects.
           This is slow path. So it should not be a problem */

        adk_md_p->neigh_status = eth_action_params->neighStatus;
        adk_md_p->next_hop_buffer = (uintptr_t)next_hop_buffer;

        memcpy(adk_md_p->dest_mac, eth_action_params->destMac,
                sizeof(adk_ethernet_mac_addr_t));
    }
    return ret;
}

adk_status_t
adk_eth_dp_pkt_proc_uli_single (
        void *task,
        int coreIndex,
        uint16_t *destFunc)
{
    adk_dp_eth_svif_intf_t *dp_svif_intf_p = NULL;
    uint16_t l3_offset;
    uint32_t flow_id;
    uint8_t *pdu = NULL;
    adk_status_t ret = ADK_SUCCESS;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_dp_eth_bridge_intf_t  *dp_bridge_intf_p = NULL;
    adk_dp_eth_arp_entry_t* arp_entry = NULL;
    adk_dp_eth_neighbor_entry_t *neigh_entry = NULL;
    adk_dp_eth_buffer_info_t *next_hop_buffer = NULL;
    bool* have_inflight_p = NULL;
    int16_t* stale_exception_raised = NULL;

    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Ethernet ULI function for packet");

    if (unlikely((adk_eth_dp_pvt_g_p == NULL) ||
                (adk_eth_dp_pvt_g_p->cookie != ADK_ETHERNET_MODULE_COOKIE)))
    {
        /* This error is expected when packets are received before Ethernet
         * module has been initialized. */
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID, "Error (ULI): Ethernet module"
                " uninitialized! Dropping pkt!");
        return (ADK_E_INTERNAL_ERROR);
    }

    adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;

    l3_offset = adk_md_p->ip_offset;
#ifdef ADK_DEBUG
    debug_print("ULI IMC l3_offset %d\n", l3_offset);
#endif

    /* Strip off anything before L3 offset if sent from OS */
    if (l3_offset != 0)
    {
        ADK_PDU_HEAD_PULL(task, pdu, (uint8_t *), l3_offset);
    }

    flow_id = adk_md_p->pdu_flow_id;
    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                                            egressUtilityCounters.eth_eg_pkt_proc_uli_entry_count,
                                            eth_uli_imc_param_p->frag_count);

    ((adk_pktio_task_t *)task)->ol_flags &= ~PKT_TX_IEEE1588_TMST;
    /* Send the packet unmodified through the output interface specified as the IMC parameter. */
    ret = adk_eth_dp_decode_interface(task, &dp_svif_intf_p, &dp_bridge_intf_p, coreIndex);
    if(ADK_SUCCESS != ret)
    {
        adk_eth_gl_stats_g_p[coreIndex].unknown_output_interface_drop += eth_uli_imc_param_p->frag_count;
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                "Failed to decode output ethernet interface\n");
        return ret;
    }
#ifdef ADK_DEBUG
    debug_print("Parameters are - PDU:%p, PduSize:%d, l3Offset: %d, Task:%p,"
            " \n", pdu, ADK_TASK_PDU_SIZE((adk_pktio_task_t *)task),
            l3_offset, task);
#endif
    if(!eth_uli_imc_param_p->next_hop.is_ipv6)
    {
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                egressUtilityCounters.eth_eg_ipv4_pkts_received,
                eth_uli_imc_param_p->frag_count);
        /* If arp entry is not found, raise exception and buffer packet */
        ret =  adk_eth_dp_arp_entry_lookup(adk_eth_arp_hash_tbl_g,
                        eth_uli_imc_param_p->next_hop.u.ipv4_addr, &arp_entry);

        if(ADK_DP_E_ERROR == ret)
        {
            ADK_MEMSET(eth_action_params->destMac, 0,
                                sizeof(eth_action_params->destMac));
            ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->neighStatus,
                                ADK_ETHERNET_NEIGHBOR_INVALID);
            ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Get Arp entry for next_hop %x fails\n",
                    eth_uli_imc_param_p->next_hop.u.ipv4_addr);
        }
        else
        {
            if(arp_entry->have_inflight)
            {
                ADK_ACQUIRE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);
                adk_eth_dp_arp_buffer_get(adk_eth_arp_buffer_tbl_g,
                                            eth_uli_imc_param_p->next_hop.u.ipv4_addr,
                                            &next_hop_buffer);
                if(NULL == next_hop_buffer)
                    arp_entry->have_inflight = false;
                ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->arp_buffer_lock, ret);
            }
            have_inflight_p = &arp_entry->have_inflight;
            ADK_DP_LOG_INFO_1D_DEBUG(ADK_ETHERNET_MODULE_ID, "Get Arp entry for next_hop %x successful\n",
                    eth_uli_imc_param_p->next_hop.u.ipv4_addr);
            ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->neighStatus,
                                                                        arp_entry->arp_status);
            if(ADK_ETHERNET_NEIGHBOR_REACHABLE == arp_entry->arp_status ||
                    ADK_ETHERNET_NEIGHBOR_STALE == arp_entry->arp_status)
            {
                ADK_MEMCPY(eth_action_params->destMac, arp_entry->dest_mac,
                                            sizeof(eth_action_params->destMac));
                stale_exception_raised = &arp_entry->stale_exception_raised;
            }
            else
            {
                ADK_MEMSET(eth_action_params->destMac, 0,
                                    sizeof(eth_action_params->destMac));
            }
        }
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->isV6, false);
    }
    else 
    {
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                egressUtilityCounters.eth_eg_ipv6_pkts_received,
                eth_uli_imc_param_p->frag_count);
        ret = adk_eth_dp_neighbor_entry_lookup(adk_eth_neighbor_hash_table_g_p,
                (uint8_t*)eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr,
                &neigh_entry);
        if (ADK_DP_E_ERROR == ret)
        {  
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"Failed to find ipv6 neighbor entry in ethernet neigh table for ");
            ADK_DP_LOG_ERROR_2D(ADK_ETHERNET_MODULE_ID,"%x:%x",eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[0],
                    eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[1]);
            ADK_DP_LOG_ERROR_2D(ADK_ETHERNET_MODULE_ID,"%x:%x\n",eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[2],
                    eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[3]);
            ADK_MEMSET(eth_action_params->destMac, 0,
                            sizeof(eth_action_params->destMac));
            ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->neighStatus,
                            ADK_ETHERNET_NEIGHBOR_INVALID);
        }
        else
        {
            if(neigh_entry->have_inflight)
            {
                /* Aquire hash lock */
                ADK_ACQUIRE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);
                adk_eth_dp_neighbor_buffer_get(adk_eth_neighbor_buffer_tbl_g,
                        (uint8_t*)eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr, &next_hop_buffer);
                if(NULL == next_hop_buffer)
                    neigh_entry->have_inflight = false;
                ADK_RELEASE_LOCK(adk_eth_dp_pvt_g_p->neighbor_buffer_lock, ret);
            }
            have_inflight_p = &neigh_entry->have_inflight; 
            ADK_DP_LOG_INFO_4D_DEBUG(ADK_ETHERNET_MODULE_ID, "Get Neigh entry for next_hop %x:%x:%x:%x successful\n",
                                                        eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[0],
                                                        eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[1],
                                                        eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[2],
                                                        eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr[3]);
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->neighStatus,
                                                                    neigh_entry->neigh_status);
            if(ADK_ETHERNET_NEIGHBOR_REACHABLE == neigh_entry->neigh_status ||
                    ADK_ETHERNET_NEIGHBOR_STALE == neigh_entry->neigh_status)
            {
                ADK_MEMCPY(eth_action_params->destMac, neigh_entry->dest_mac,
                                                sizeof(eth_action_params->destMac));
                stale_exception_raised = &neigh_entry->stale_exception_raised;
            }
            else
            {
                ADK_MEMSET(eth_action_params->destMac, 0,
                                                sizeof(eth_action_params->destMac));
            }
        }
        ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->isV6, true);
    }

    ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->haveInflight, have_inflight_p);
    ret = eth_handle_slowpath(task, coreIndex, destFunc, stale_exception_raised);
    if(ret == ADK_SUCCESS)
    {
        if(*destFunc == ADK_ETH_ULI_PKT_BUFFER_FN_ID)
        {
            if(have_inflight_p != NULL)
            {
                *have_inflight_p = true;
            }
        }
        else
        {
            ADK_SYS_DP_FLOW_CACHE_SET_PARAM(eth_action_params->pcpVal,
                    adk_eth_dscp_to_pcp_map_g[eth_uli_imc_param_p->dscp]);
            ret = adk_eth_dp_encap_and_send_pkt(task, eth_action_params->destMac, coreIndex, false);

            if(ADK_E_TASK_CONSUMED != ret)
            {
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                        "Failed to encaptulate egress packets\n");
                return ret;
            }
        }
    }

    adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;
_lock_release_failed:
_lock_acquire_failed:
    return (ret);
}

static adk_status_t
adk_eth_dp_pkt_proc_uli_flow_cache_pkt_path_single (
        void *task,
        int coreIndex,
        uint16_t *destFunc)
{
    uint16_t l3_offset;
    uint32_t flow_id;
    uint8_t *pdu = NULL;
    adk_status_t ret = ADK_SUCCESS;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_dsi_tx_metadata_t *dsi_md_p = ADK_DSI_TX_METADATA_PTR(task);;

    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Ethernet ULI function for packet flow cache path ");

    adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;

    l3_offset = eth_scratch_pad_p->ip_offset;
#ifdef ADK_DEBUG
    debug_print("ULI IMC l3_offset %d\n", l3_offset);
#endif
    /* Strip off anything before L3 offset if sent from OS */
    if (l3_offset != 0)
    {
        ADK_PDU_HEAD_PULL(task, pdu, (uint8_t *), l3_offset);
    }

    flow_id = adk_md_p->pdu_flow_id;
    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;


    INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
            egressUtilityCounters.eth_eg_pkt_proc_uli_entry_count,
            eth_uli_imc_param_p->frag_count);

    ((adk_pktio_task_t *)task)->ol_flags &= ~PKT_TX_IEEE1588_TMST;

    dsi_md_p->l2_domain = eth_action_params->l2Domain;

#ifdef ADK_DEBUG
    debug_print("Parameters are - PDU:%p, PduSize:%d, l3Offset: %d, Task:%p,"
            " \n", pdu, ADK_TASK_PDU_SIZE((adk_pktio_task_t *)task),
            l3_offset, task);
    debug_print("dsi_md_p->l2_domain in uli %d\n", dsi_md_p->l2_domain);
#endif
    if(!eth_action_params->isV6)
    {
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[ADK_LCORE_INDEX].
                egressUtilityCounters.eth_eg_ipv4_pkts_received,
                eth_uli_imc_param_p->frag_count);
    }
    else
    {
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                egressUtilityCounters.eth_eg_ipv6_pkts_received,
                eth_uli_imc_param_p->frag_count);
    }
    ret = eth_handle_slowpath(task,coreIndex, destFunc, NULL);
    if(ret == ADK_SUCCESS)
    {
        if(*destFunc == 0)
        {
            ret = adk_eth_dp_encap_and_send_pkt(task,
                    eth_action_params->destMac, coreIndex, false);
            if(ADK_E_TASK_CONSUMED != ret)
            {
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                        "Failed to encaptulate egress packets\n");
                return ret;
            }
        }
    }
    adk_eth_gl_stats_g_p[coreIndex].proc_cntr++;

    return (ret);
}

#ifdef ADK_DP_NO_BATCH_PROC

adk_status_t
adk_eth_dp_buffer_packet(
    void **task_p,
    __attribute__ ((unused))uint8_t task_count,
    uint32_t task_mask,
    __attribute__ ((unused)) void *ring,
    __attribute__ ((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
    int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    uint8_t pkt_drops = 0;

    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_ULI_RX_TS();

    ret = adk_eth_dp_buffer_packet_single(task_p[0], task_mask, coreIndex);
    /* If any fragment runs into error, drop entire original packet */
    if (unlikely(ADK_E_INTERNAL_ERROR == ret))
    {
        pkt_drops = adk_eth_task_chained_buffer_free(task_p[0]);
        INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                                                egressUtilityCounters.eth_eg_pkt_dropped_in_buffering_stage,
                                                pkt_drops);
        return ret;
    }

    ADK_DP_MEASURE_BLOCK_CYCLE_ETHERNET_ULI_TX_TS();

    return ret;
}

#endif

adk_status_t adk_eth_dp_flush_pkt
(
 void* task,
 __attribute__((unused)) uint32_t task_mask,
 adk_ethernet_mac_addr_t dest_mac,
 int coreIndex,
 bool flush
 )
{
    adk_status_t ret = ADK_E_TASK_CONSUMED;
    adk_dp_eth_svif_intf_t *dp_svif_intf_p = NULL;
    adk_dp_eth_bridge_intf_t *dp_bridge_intf_p = NULL;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    eth_scratch_pad_p = ADK_SIF_DP_GET_SCRATCHPAD_PTR(coreIndex);
    eth_action_params = (adk_sys_eth_dp_uli_flow_cache_entry_t*)&eth_scratch_pad_p->data[0];
    eth_action_params->isV6 = eth_uli_imc_param_p->next_hop.is_ipv6;
    eth_action_params->pcpVal = adk_eth_dscp_to_pcp_map_g[eth_uli_imc_param_p->dscp];

    ret = adk_eth_dp_decode_interface(task, &dp_svif_intf_p, &dp_bridge_intf_p, coreIndex);
    if(ADK_SUCCESS != ret)
    {
        adk_eth_gl_stats_g_p[coreIndex].unknown_output_interface_drop += eth_uli_imc_param_p->frag_count;
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                "Failed to decode output ethernet interface while arp buffer flush\n");
        return ret;
    }

    ret = adk_eth_dp_encap_and_send_pkt(task, dest_mac, coreIndex, flush);

    if(ADK_E_TASK_CONSUMED != ret)
    {   
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                "Failed to encaptulate and send egress packets\n");
    }
    return ret;
}

adk_status_t adk_eth_dp_next_hop_buffer_store
(
 adk_dp_eth_buffer_info_t* nexthop_buffer_info,
 void* task,
 uint32_t task_mask,
 int coreIndex
 )
{
    RTE_SET_USED(task_mask);
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_dp_eth_next_hop_buffer_t* nexthop_buffer;
    void* task_to_discard = NULL;
    uint8_t pkt_drops;

    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);

    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    if(NULL == nexthop_buffer_info->nexthop_buffer)
    { 
        /* Init the DP element for the next hop buffer structure*/ 

        if(rte_mempool_get(adk_eth_nexthop_buffer_pool_g, (void **)&nexthop_buffer_info->nexthop_buffer))
        {
            ret = ADK_E_MEM_UNAVAILABLE;
            adk_eth_gl_stats_g_p[coreIndex].buffering_memory_not_available_drop += eth_uli_imc_param_p->frag_count;
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                    "Failed to allocate DP memory for buffering next hop");
        }
        memset(nexthop_buffer_info->nexthop_buffer, 0x00, sizeof(adk_dp_eth_next_hop_buffer_t));
    }

    nexthop_buffer = nexthop_buffer_info->nexthop_buffer;

    task_to_discard = (nexthop_buffer->buffer_entries)[nexthop_buffer->current_packet_flow_index].task;
    if(NULL != task_to_discard)
    {
        pkt_drops = adk_eth_task_chained_buffer_free(task_to_discard);
        if(pkt_drops > 1)
            (*nexthop_buffer_info->fragment_pkt_rollover_discard_count) += pkt_drops;
        else
            (*nexthop_buffer_info->full_pkt_rollover_discard_count)++;

        nexthop_buffer_info->inflight_count--;
    }

    adk_pal_dp_release_send(task, ADK_PAL_DP_BATCH_CB_FN_ETH, NULL);
   
    ADK_METADATA_PARAM(task, task_type) = ADK_TASK_TYPE_NEW;

    (nexthop_buffer->buffer_entries)[nexthop_buffer->current_packet_flow_index].task = task;
    nexthop_buffer->current_packet_flow_index++;
    if(nexthop_buffer->current_packet_flow_index >= NEXTHOP_BUFFER_MAX_FLOW)
        nexthop_buffer->current_packet_flow_index = 0;
    ADK_DP_LOG_INFO_1P_DEBUG(ADK_ETHERNET_MODULE_ID,
            "task %p stored in nexthop buffer", task);

    return ret;
}

adk_status_t adk_eth_dp_flush_pkts
(
adk_dp_eth_buffer_info_t* nexthop_buffer_info,
uint32_t task_mask,
adk_ethernet_mac_addr_t dest_mac,
int coreIndex
)
{
    uint32_t nexthop_entry_index;
    adk_dp_eth_next_hop_buffer_t* nexthop_buffer = nexthop_buffer_info->nexthop_buffer;
    adk_status_t ret = ADK_E_TASK_CONSUMED;
    uint32_t iCurrPacketFlush;

    if(NULL == nexthop_buffer)
        return ret;

    iCurrPacketFlush = nexthop_buffer->current_packet_flow_index;

    for(nexthop_entry_index = 0; nexthop_entry_index < NEXTHOP_BUFFER_MAX_FLOW; nexthop_entry_index++)
    {
        if(NULL != (nexthop_buffer->buffer_entries)[iCurrPacketFlush].task)
        {
            ret = adk_eth_dp_flush_pkt((nexthop_buffer->buffer_entries)[iCurrPacketFlush].task,
                    task_mask, dest_mac, coreIndex, true);
            if(ADK_E_TASK_CONSUMED != ret)
            {
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                        "Failed to flush buffered packet");
                return ret;
            }
            nexthop_buffer_info->inflight_count--;
        }
        iCurrPacketFlush++;
        if(iCurrPacketFlush >= NEXTHOP_BUFFER_MAX_FLOW)
           iCurrPacketFlush = 0; 
    }

    /* Clean memory right away */
    rte_mempool_put(adk_eth_nexthop_buffer_pool_g, nexthop_buffer_info->nexthop_buffer);
    nexthop_buffer_info->nexthop_buffer = NULL;

    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Buffered packets flushed successfully");
    return ret;
}

static adk_dp_status_t eth_inflight_status_change
(
adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p,
int coreIndex
)
{
    /* This is critical. Inflight count status and count should be accessible
       from both worker core and buffer core. And Hash table should be accessed only
       in slow path. The moment all packets are flushed and inflight count is zero in
       buffer stage, this information needs to be transmitted to worker core. So a look
       up of CP table is required. The pointer to arp entry can't be passed from worker
       core to atomic core. This lookup must be done to register this thread as reader
       thread for RCU protection. Inflight status change is done with next hop lock */ 

    adk_status_t ret = ADK_SUCCESS;
    adk_dp_eth_arp_entry_t* arp_entry = NULL;
    adk_dp_eth_neighbor_entry_t *neigh_entry = NULL;

    if(!eth_uli_imc_param_p->next_hop.is_ipv6)
    {
        ret =  adk_eth_dp_arp_entry_lookup(adk_eth_arp_hash_tbl_g,
                            eth_uli_imc_param_p->next_hop.u.ipv4_addr, &arp_entry);

            if(ADK_DP_E_ERROR == ret)
            {   
                adk_eth_gl_stats_g_p[coreIndex].arp_entry_not_found_drop++;
                ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                        "Failed to find  entry in arp table in arp buffer stage\n");
                return ADK_E_INTERNAL_ERROR;  
            }

        arp_entry->have_inflight = false;           
    }
    else
    {
        ret = adk_eth_dp_neighbor_entry_lookup(adk_eth_neighbor_hash_table_g_p,
                (uint8_t*)eth_uli_imc_param_p->next_hop.u.ipv6_addr.addr,
                &neigh_entry);
        if (ADK_DP_E_ERROR == ret)
        {
            adk_eth_gl_stats_g_p[coreIndex].neigh_entry_not_found_drop++;
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,"Failed to find ipv6 neighbor entry in ethernet neigh table\n");
            return ADK_E_INTERNAL_ERROR;
        }

        neigh_entry->have_inflight = false;
    }

    return ret;
}

static void eth_update_counter_address
(
adk_ethernet_dp_uli_imc_param_t* eth_uli_imc_param_p,
adk_dp_eth_buffer_info_t* next_hop_buffer,
int coreIndex
)
{
    if(next_hop_buffer->update_counter_address)
        return;

    if(!eth_uli_imc_param_p->next_hop.is_ipv6)
    {
        /*Update pointer to counters*/
        next_hop_buffer->fragment_pkt_rollover_discard_count = &adk_eth_gl_stats_g_p[coreIndex].arp_buffer_fragment_pkt_rollover_discard;
        next_hop_buffer->full_pkt_rollover_discard_count = &adk_eth_gl_stats_g_p[coreIndex].arp_buffer_full_pkt_rollover_discard;
        next_hop_buffer->nexthop_entry_rem_buffer_drop_count = &adk_eth_gl_stats_g_p[coreIndex].arp_buffer_arp_entry_rem_drop;
    }
    else
    {
        /*Update pointer to counters*/
        next_hop_buffer->fragment_pkt_rollover_discard_count = &adk_eth_gl_stats_g_p[coreIndex].nd_buffer_fragment_pkt_rollover_discard;
        next_hop_buffer->full_pkt_rollover_discard_count = &adk_eth_gl_stats_g_p[coreIndex].nd_buffer_full_pkt_rollover_discard;
        next_hop_buffer->nexthop_entry_rem_buffer_drop_count = &adk_eth_gl_stats_g_p[coreIndex].nd_buffer_neigh_entry_rem_drop;
    }

    next_hop_buffer->update_counter_address = true;
}
adk_status_t adk_eth_dp_next_hop_buffer
(
void* task,
uint32_t task_mask,
int coreIndex
)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_ethernet_dp_uli_imc_param_t *eth_uli_imc_param_p;
    adk_dp_eth_buffer_info_t *next_hop_buffer = NULL;
    adk_ethernet_neigh_status_t nexthop_status;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task);
    uint8_t *next_hop_mac;

    eth_uli_imc_param_p = (adk_ethernet_dp_uli_imc_param_t *)
        &adk_md_p->imc_param_start;

    INCREMENT_ETH_UTILITY_COUNTER_COUNT(adk_eth_dp_pvt_g_p->ethUtilityCnt[coreIndex].
                                            egressUtilityCounters.eth_eg_pkt_recv_in_buffering_stage,
                                            eth_uli_imc_param_p->frag_count);
    nexthop_status = (adk_ethernet_neigh_status_t)adk_md_p->neigh_status;
    next_hop_mac = adk_md_p->dest_mac;

    /* TODO: What is the guarantee that this pointer remains valid given the
       timer expiry mechanism. Time out value should be greater than the time
       required to enqueue the event and event scheduling. And there is no harm
       in keeping timeout value large enough since there is a limit on how many
       parallel next hop resolutio can happen in the system. So memory bloating
       will not happen */
    next_hop_buffer = (adk_dp_eth_buffer_info_t*)adk_md_p->next_hop_buffer;

    /* Only next hop lock is sufficient. This also enables parallel
       buffer or flush of packets on different cores */
    ADK_ACQUIRE_LOCK(next_hop_buffer->nh_lock, ret);

    eth_update_counter_address(eth_uli_imc_param_p, next_hop_buffer, coreIndex);

    /* If arp/Neighbor entry is not resolved, buffer the packet */
    if(nexthop_status == ADK_ETHERNET_NEIGHBOR_INVALID ||
       nexthop_status == ADK_ETHERNET_NEIGHBOR_BUFFER_WITHOUT_NOTIFYING)
    {
        ret = adk_eth_dp_next_hop_buffer_store(next_hop_buffer, task, task_mask, coreIndex);
        if (ADK_SUCCESS != ret)
        {
            ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);
            ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                              "Failed to store packet for nexthop buffer");
            return ret;
        }

        ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);
        ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID,
                         "Stored packet successfully for nexthop buffer");

        return ret;
    }

    /* Arp/Neighbor entry is resolved. Send buffered packets if any */
    ret = adk_eth_dp_flush_pkts(next_hop_buffer, task_mask, next_hop_mac, coreIndex);
    if(ADK_E_TASK_CONSUMED != ret)
    {
        ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                          "Flush error for resolved nexthop bufferred packets");
        return ret;
    }

    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID,
                    "Flushed packets successfully for resolved next hop");

    /* Send this packet as well */
    ret = adk_eth_dp_flush_pkt(task, task_mask, next_hop_mac, coreIndex, false);
    if(ADK_E_TASK_CONSUMED != ret)
    {
        ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);
        ADK_DP_LOG_ERROR(ADK_ETHERNET_MODULE_ID,
                  "Failed to flush buffered packet");
        return ret;
    }
    else
    {
        ret = ADK_SUCCESS;
    }

    next_hop_buffer->inflight_count--;
    if(!next_hop_buffer->inflight_count)
    {
        ret = eth_inflight_status_change(eth_uli_imc_param_p, coreIndex);
        if(ADK_SUCCESS != ret)
        {
            ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);
            return ret;
        }
        ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);
        return ret;
    }

    ADK_RELEASE_LOCK(next_hop_buffer->nh_lock, ret);

_lock_release_failed:
_lock_acquire_failed:
    return ret;
}

adk_status_t
adk_eth_dp_buffer_packet_single
(
void *task,
uint32_t task_mask,
int coreIndex
)
{
    adk_status_t ret = ADK_SUCCESS;

    ret = adk_eth_dp_next_hop_buffer(task, task_mask, coreIndex);    
    if(ret != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Packet buffer fails while nexthop is getting resolved %p\n",
                task);
        return ADK_E_INTERNAL_ERROR;
    }

    ADK_DP_LOG_INFO_1D_DEBUG(ADK_ETHERNET_MODULE_ID, "Packet buffer successful for packet %p\n",
            task);
    return ADK_E_TASK_CONSUMED;
}

void adk_eth_dp_nexthop_buffer_free(adk_dp_eth_buffer_info_t* nexthop_buffer_info)
{
    uint32_t nexthop_entry_index;
    uint8_t pkt_drop;

    if(NULL == nexthop_buffer_info->nexthop_buffer)
    {
        return;
    }
    for(nexthop_entry_index = 0; nexthop_entry_index < NEXTHOP_BUFFER_MAX_FLOW; nexthop_entry_index++)
    {
        if(nexthop_buffer_info->nexthop_entry_rem_buffer_drop_count)
        {
            pkt_drop = adk_eth_task_chained_buffer_free((adk_pktio_task_t *)(nexthop_buffer_info->nexthop_buffer->buffer_entries)[nexthop_entry_index].task);
            (*nexthop_buffer_info->nexthop_entry_rem_buffer_drop_count) += pkt_drop;
        }
    }

    /* Let's clean memory right away as of now */
    rte_mempool_put(adk_eth_nexthop_buffer_pool_g, nexthop_buffer_info->nexthop_buffer);

    ADK_DP_LOG_INFO_DEBUG(ADK_ETHERNET_MODULE_ID, "Buffered packets deleted successfully");
}

#ifdef ADK_DEBUG

/* task_ring_t is opaque for the rest of the ethernet module, hence need to
    get the stat from here. Need to review this later */ 
adk_status_t adk_eth_dp_arp_buffer_statistics_get
(
adk_dp_eth_arp_table_t *eth_arp_table,
adk_ethernet_ipv4_addr_t ip_addr,
uint32_t* num_inflight_pkts,
uint32_t* num_full_buffered_pkts,
uint32_t* num_frag_buffered_pkts
);

adk_status_t adk_eth_dp_arp_buffer_statistics_get
(
adk_dp_eth_arp_table_t *eth_arp_table,
adk_ethernet_ipv4_addr_t ip_addr,
uint32_t* num_inflight_pkts,
uint32_t* num_full_buffered_pkts,
uint32_t* num_frag_buffered_pkts
)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_dp_eth_arp_entry_t* arp_entry;
    adk_dp_eth_buffer_info_t* arp_buffer = NULL;
    task_ring_t* task_list;
    uint32_t iarp_entry;

    if (ADK_SUCCESS !=  adk_eth_dp_arp_entry_lookup(eth_arp_table,
                            ip_addr,&arp_entry))
    {
        ret = ADK_E_INVALID_PARAMETER;
        ADK_DP_LOG_ERROR_1D(ADK_ETHERNET_MODULE_ID, "Arp entry not present for IP addr %x\n",
                ip_addr);
        return ret;
    }

    *num_inflight_pkts = arp_buffer->inflight_count;

    for(iarp_entry = 0; iarp_entry < NEXTHOP_BUFFER_MAX_FLOW; iarp_entry++)
    {
        if(0 == (arp_buffer->nexthop_buffer->buffer_entries)[iarp_entry].ip_Id &&
            (NULL != (arp_buffer->nexthop_buffer->buffer_entries)[iarp_entry].task))
        {
            (*num_full_buffered_pkts)++;
        }
        else
        {
            task_list = (task_ring_t*)((arp_buffer->nexthop_buffer->buffer_entries)[iarp_entry].task_list);
            if(NULL != task_list)
            {
                (*num_frag_buffered_pkts)+= task_list->num_elms; 
            } 
        }
    }

    return ret;
}

#endif /*ADK_DEBUG*/
 
 /*!  end addtogroup _ethernetFunctions  Ethernet Module
 * @}
 */
