/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_eth_dp.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ethernet  Ethernet Module
 *
 * @{
 *
 * @brief   This file has all the macros and data types which are shared by
 *          control plane and data plane software. This file does not have
 *          any function prototype as no function is shared between control
 *          plane and data plane software components.
 *
 *  **********************************************************************/
#ifndef __ADK_ETH_DP_H__
#define __ADK_ETH_DP_H__

#include <adk_sys.h>
#include <adk_ethernet.h>
#include <adk_eth_constants.h>
#include <adk_ethernet_export.h>
#include <adk_ip_dp_export.h>
#include <adk_ipv6_dp_export.h>
#include <adk_timer.h>
#include <adk_hos.h>

#include <adk_pktio.h>
#include <adk_queue.h>
#include <adk_lpm.h>
#include <rte_hash_crc.h>
#ifdef ADK_MODULE_EXCEPTIONS_USE_PH_DEV
#include "adk_ph_dev_dp.h"
#endif
#include <adk_ethernet_utility_counters.h>
#include "adk_sys_dpdk.h"
/*
 * The Ethernet data plane tables are essentially arrays arranged in a specific
 * way. Two common ways are generally used depending upon how the array index
 * is determined.
 *
 * (i) The first way, which is also the easier way, is when the index is
 * provided as parameter by sending entity. For example, physical interface ID
 * is provided by the NIC driver for packet received form external interface.
 * Or, the Ethernet encap flow ID is provided by the sending IP layer module.
 * Since the array index is already known, the information can be retrieved
 * easily from the array using the ID as array index. So, the table entry just
 * consists of the pointers to the structures and not structures themselves.
 * This allow for RCU based atomic updates of the individual array elements
 * using simple single word writes.
 *
 * (ii) The other way, involves computing or looking up the index in the data
 * plane. This is similar to the classification table lookup, where multiple
 * input fields are used to determine the array index. For example, hashing can
 * be performed over the input fields to get the index. In case of hashing based
 * lookup, the array elements are not just pointers but also contain the hash
 * value. The hash value is used to detect collision upfront, with a very high
 * probability. So, the table entry consists of a pair of elements - hash value
 * and the pointer to the structure. This allow for RCU based atomic updates of
 * individual array elements (double words) using compare-and-exchange atomic
 * instruction.
 *
 */

#define ADK_ETHERNET_SHM_MEMZONE_NAME   "adk_ethernet_shm_zone"

/****************************************
************** CONSTANTS ****************
****************************************/
#define ETH_BCAST_MAC_ADDR_BYTE 0xFF
#define ETH_INVALID_NS_ID       0xFFFF

/****************************************
******* MACROS FOR COMPUTATION *********
****************************************/
#define ADK_ETHERNET_HASH(init, hashInput) \
    _mm_crc32_u32(init, hashInput)
    /* Not using DPDK rte_hash_crc_4byte() function as it would
     * fallback upon software implementation, without any warning,
     * in case below SSE instruction is not supported. */

#define IS_BROADCAST_MAC(mac) \
    ((mac[0] == ETH_BCAST_MAC_ADDR_BYTE) && \
     (mac[1] == ETH_BCAST_MAC_ADDR_BYTE) && \
     (mac[2] == ETH_BCAST_MAC_ADDR_BYTE) && \
     (mac[3] == ETH_BCAST_MAC_ADDR_BYTE) && \
     (mac[4] == ETH_BCAST_MAC_ADDR_BYTE) && \
     (mac[5] == ETH_BCAST_MAC_ADDR_BYTE))

/********************************************
******* MACROS SPECIFIC TO MAC FDB **********
*********************************************/
#define ADK_ETH_MAX_NUM_OF_MAC_FDB_ENTRIES    ADK_ETH_NUM_MACS_MAX

#define ADK_ETH_MAC_FDB_HASH_WIDTH            11
#define ADK_ETH_MAC_FDB_HASH_MAX              (1<<ADK_ETH_MAC_FDB_HASH_WIDTH)
#define ADK_ETH_MAC_FDB_HASH_MASK             (ADK_ETH_MAC_FDB_HASH_MAX - 1)

/* Arp related constants */
#define LPM_ARP_STORE_MAC 0
#define LPM_ARP_STORE_BUFFER_PKTS 1

#define ADK_ETHERNET_ETHERTYPE_SUPPORTED_MAX  10

#define EXCEPTION_NOT_RAISED  0
#define EXCEPTION_RAISED 1

typedef enum adk_eth_list_id_s
{
    ADK_ETH_INTF_LIST_ID = 0,
    ADK_ETH_SVIF_ID_LIST_ID,
    ADK_ETH_BRIDGE_ID_LIST_ID,
    ADK_ETH_INGRESS_FLOW_LIST_ID,
    ADK_ETH_CPU_FLOW_LIST_ID,
    ADK_ETH_ING_FLOW_ID_LIST_ID,
    ADK_ETH_BRIDGE_LIST_ID,
    ADK_ETH_SVIF_INTF_LIST_ID,
    ADK_ETH_ARP_BUFFER_LIST_ID,
    ADK_ETH_TYPE_PCP_MAP_LIST_ID,
    ADK_ETH_MAX_LIST_ID
} adk_eth_list_id_t;

/* Data type for managing pending list of objects waiting to be freed */
typedef struct adk_eth_list_s
{
    uint32_t max_elements;
    uint32_t num_elements;
    uintptr_t *elem_p;
} adk_eth_list_t;

/* obj types for dynamic memory */
typedef enum adk_eth_obj_s
{
    ADK_ETH_INTF_OBJ,
    ADK_ETH_VLAN_INFO_OBJ,
    ADK_ETH_SUB_INTF_OBJ,
    ADK_ETH_SVIF_INTF_OBJ,
    ADK_ETH_LIST_OBJ,
    ADK_ETH_STATS_OBJ,
    ADK_ETH_SVIF_HASH_INTF_OBJ,
    ADK_ETH_ARP_OBJ,
    ADK_ETH_ARP_ENTRY_OBJ,
    ADK_ETH_ARP_BUFFER_OBJ,
    ADK_ETH_NEIGH_OBJ,
    ADK_ETH_NEIGH_ENTRY_OBJ,
    ADK_ETH_NEIGH_BUFFER_OBJ,
    ADK_ETH_NEXTHOP_V4_TIMER_INDEX_OBJ,
    ADK_ETH_NEXTHOP_V6_TIMER_INDEX_OBJ,
    ADK_ETH_INGRESS_FLOW_OBJ,
    ADK_ETH_INGRESS_FLOW_LIST_OBJ,
    ADK_ETH_INGRESS_RES_OBJ,
    ADK_ETH_BRIDGE_IF_RES_OBJ,
    ADK_ETH_SVIF_IF_RES_OBJ,
    ADK_ETH_BRIDGE_INTF_OBJ,
    ADK_ETH_DEFAULT_BRIDGE,
    ADK_ETH_NEXTHOP_BUFFER_INFO_OBJ,
    ADK_ETH_NEXTHOP_BUFFER_OBJ,
    ADK_ETH_AUTO_CONN_OBJ,
    ADK_ETH_TYPE_PCP_MAP_OBJ,
    ADK_ETH_TYPE_QUEUE_MAP_OBJ,
    ADK_ETH_NON_BRIDGE_ORIGIN_QUEUE_OBJ,
    ADK_ETH_BRIDGE_TIMESTAMP_QUAD_GROUP_OBJ,
    ADK_ETH_BRIDGE_TIMESTAMP_QUAD_GROUP_PORT_OBJ,
    ADK_ETH_DSCP_TO_PCP_MAP_OBJ,
    ADK_ETH_EXCEP_OBJ
} adk_eth_obj_t;

/*************************************
******* MACROS FOR PARSING **********
*************************************/
typedef struct adk_dp_vlan_hdr_s
{
    uint16_t tpid;
    uint8_t pbits:3;
    uint8_t cfi:1;
    uint16_t vlanid:12;
    bool priorityTagged;
} adk_dp_vlan_hdr_t;

typedef struct adk_dp_eth_hdr_s
{
    uint8_t dmac[6];
    uint8_t smac[6];
    adk_dp_vlan_hdr_t vlan;
    uint16_t ether_type;
    uint16_t hdr_len;
} adk_dp_eth_hdr_t;

/***********************************************
******* DATA TYPES USED IN DP TABLES **********
***********************************************/
/* Global pata plane stats */
typedef struct adk_dp_eth_gl_stats_s
{
    uint32_t intf_not_defined;
    uint32_t port_not_defined;
    uint32_t time_index_alloc_error;
    uint32_t ts_tx_complete_notification_failed;
    uint32_t pkt_clone_failure;
    uint32_t ingressflow_lookup_failure;
    uint32_t ingressflowDrops;
    uint32_t ethPacketsRateLimitingDrops;
    uint32_t arp_entry_not_found_drop;
    uint32_t arp_buffer_full_pkt_rollover_discard;
    uint32_t arp_buffer_fragment_pkt_rollover_discard;
    uint32_t nd_buffer_full_pkt_rollover_discard;
    uint32_t nd_buffer_fragment_pkt_rollover_discard;
    uint32_t exceptionDrops;
    uint32_t arp_buffer_arp_entry_rem_drop;
    uint32_t nd_buffer_neigh_entry_rem_drop;
    uint32_t neigh_entry_not_found_drop;
    uint32_t neigh_state_unreachable_drop; 
    /* The below element in this structure is not for data plane statistics
     * but rather for RCU synchronization. This is here as this also needs to
     * be maintained on a per core basis */

    /* This counter increments any time Ethernet module starts packet
     * processing of any kind. And it again increments when Ethernet module
     * packet processing ends. So, while the LSB bit of these counters 
     * indicate if there is any packet processing going on anywhere, the
     * rest of the 31-bit value counts the number of times Ethernet module
     * has been invoked for packet processing. */
    uint32_t proc_cntr;
    uint32_t non_l2term_frames_discard;
    uint32_t netd_origin_non_bridge_frames_discard;
    uint32_t ethtype_queue_map_not_found_drop;
    uint32_t unknown_output_interface_drop;
    uint32_t buffering_memory_not_available_drop;
    uint32_t nic_send_failure_drop;
    uint32_t nexthop_buffer_stage_not_found_drop;
    uint32_t nexthop_buffer_proc_drop;
    uint32_t nexthop_buffer_timeout_error;
    uint32_t failed_to_sent_next_stage_drop;
    uint32_t vlan_header_stripping_failure_drop;
    uint32_t egress_encap_pkt_failure_drop;
    uint32_t unknown_multicast_addr_drop;
} __attribute__ ((aligned(64))) adk_dp_eth_gl_stats_t;

/* Interface level data plane stats */
typedef struct adk_dp_eth_stats_s
{
    uint64_t inFrames;  /*unicast*/
    uint64_t inOctets;  /*unicast*/
    uint64_t OutFrames; /*unicast*/
    uint64_t OutOctets; /*unicast*/

    uint64_t inIpMcastFrames;
    uint64_t inIpBcastFrames;
    uint64_t inIpMcastOctets;
    uint64_t inIpBcastOctets;

    uint64_t OutIpMcastFrames;
    uint64_t OutIpBcastFrames;
    uint64_t OutIpMcastOctets;
    uint64_t OutIpBcastOctets;

    uint64_t inNonIpMcastFrames;
    uint64_t inNonIpBcastFrames;
    uint64_t inNonIpMcastOctets; 
    uint64_t inNonIpBcastOctets;

    uint64_t OutNonIpMcastFrames; 
    uint64_t OutNonIpBcastFrames;
    uint64_t OutNonIpMcastOctets; 
    uint64_t OutNonIpBcastOctets; 

    uint32_t inEthTypeMismatch;
    uint32_t inUnknownVlan;
    uint32_t inPrioTaggedFrames;

} __attribute__ ((aligned(64))) adk_dp_eth_stats_t;

/* Ingress flow level data plane stats */
typedef struct adk_dp_eth_ingress_flow_stats_s
{
    uint64_t numPkts;
    uint64_t pkts_sent_to_host;
    uint64_t pkts_sent_to_ipv4; 
    uint64_t pkts_sent_to_ipv6;
    uint64_t pkts_drop;
} __attribute__ ((aligned(64))) adk_dp_eth_ingress_flow_stats_t;

typedef enum
{
    ADK_DP_ETH_AUTO_CONN_DISABLE = 0,
    ADK_DP_ETH_AUTO_CONN_ENABLE
}adk_dp_eth_auto_conn_status_t;

typedef struct adk_eth_dp_policer_meta_data_s
{
    uint32_t flowId;
    uint32_t policerFlowId;
}adk_eth_dp_policer_meta_data_t;
/* QoS Queue info */
typedef union adk_dp_eth_qos_info_s
{
    uint64_t info;
    struct {
        uint16_t dev_id;
        uint16_t queue_id;
        uint32_t stats_id;
    };
} __attribute__((aligned(8))) adk_dp_eth_qos_info_t;

/* Ethernet Bridge interface */
typedef struct adk_dp_eth_bridge_intf_s
{
    bool priority_tag;
    uint16_t eth_dev_id;
    uint16_t l2_domain_id; /* Equivalent to L2 Domain */
    uint32_t bridgeIntfId;
    uint32_t statsId;
    adk_ethernet_mac_addr_t src_mac;
    adk_dp_eth_stats_t intf_stats[ADK_MAX_NUM_CORES];
    adk_interface_handle_t bridgeIntfHandle;
} __attribute__ ((aligned(64))) adk_dp_eth_bridge_intf_t;

/* Ethernet SVIF interface */
typedef struct adk_dp_eth_svif_intf_s
{
    uint16_t vlan;
    uint16_t bgid; /* Equivalent to L2 Domain */
    uint32_t statsId;
    adk_dp_eth_stats_t intf_stats[ADK_MAX_NUM_CORES];
    adk_interface_handle_t svifIntfHandle;
} __attribute__ ((aligned(64))) adk_dp_eth_svif_intf_t;

/* Ethernet ingress flow */
typedef struct adk_dp_eth_ingress_flow_s
{
    bool     enablePolicing;
    uint16_t ether_type;
    uint32_t flowId;
    uint32_t policerId;
    adk_ethernet_ingress_flow_action_type_t actionType;
    adk_ethernet_interface_type_t intfType;
    adk_ethernet_mac_subnet_t dmac_subnet;
    adk_dp_eth_ingress_flow_stats_t flow_stats[ADK_MAX_NUM_CORES];
} __attribute__ ((aligned(64))) adk_dp_eth_ingress_flow_t;

/***********************************************
******* HASH BASED DP TABLE ELEMENTS **********
***********************************************/

typedef struct adk_dp_eth_svif_if_hash_s
{
    union {
        uint64_t hash_entry;    /* Combined for atomic access */
        struct {
            uint16_t l2_domain_id;
            uint16_t vlan;
            uint32_t stats_id;
        };
    };
} __attribute__ ((aligned(8))) adk_dp_eth_svif_if_hash_t;

typedef struct adk_dp_eth_ethertype_queue_map_hash_s
{
    union {
        uint32_t hash_entry;    /* Combined for atomic access */
        struct {
            uint16_t ethType;
            uint16_t queueId;
        };
    };
}__attribute__ ((aligned(8)))adk_dp_eth_ethertype_queue_map_hash_t;
/****************************************
******* LPM DP TABLE ELEMENTS **********
****************************************/
typedef struct adk_dp_eth_ingress_flow_key_s
{
    uint16_t ether_type;
    uint8_t dmac[6];
} __attribute__ ((packed)) adk_dp_eth_ingress_flow_key_t;

typedef struct adk_dp_eth_cpu_flow_result_s
{
    uint32_t destModId;
    uint16_t destFunc[_DEST_FUNC_MAX];
} adk_dp_eth_cpu_flow_result_t;

typedef struct adk_eth_fdb_mac_table_entry_s
{
    adk_ethernet_mac_table_entry_t          fwd_info;
    bool                                    is_alloted;
    struct adk_eth_fdb_mac_table_entry_s    *next;
} adk_eth_fdb_mac_table_entry_t;

/* Buffered packets for a flow */
typedef struct adk_eth_next_hop_buffer_entry_s
{
    uint16_t ip_Id;

    /* Unfragmented packets are simply stored here instead
        of another memory application on heap */
    void* task;
 
    /* Fragmented packets are stored here */
    void* task_list;
} adk_eth_next_hop_buffer_entry_t;

/* Arp/Neigh buffer storage structure.
    Packets stored for 3 flows */
typedef struct adk_dp_eth_next_hop_buffer_s
{
    adk_eth_next_hop_buffer_entry_t buffer_entries[NEXTHOP_BUFFER_MAX_FLOW];

    /* denotes where to store the current packet. Increment this 
    After storing packet at this index with wrap around 3 */
    uint32_t current_packet_flow_index;
} adk_dp_eth_next_hop_buffer_t;

typedef struct adk_dp_eth_buffer_info_s
{
    uint32_t inflight_count;
    void* nh_lock;
    adk_timer_hdl_t timer_hdl;
    /* Used for directing indexing into hash for this entry */
    uint32_t hash_index;    
    bool is_ipv6;

    bool update_counter_address;
    uint8_t pkt_count;
    uint32_t *full_pkt_rollover_discard_count;
    uint32_t *fragment_pkt_rollover_discard_count;
    uint32_t *nexthop_entry_rem_buffer_drop_count;

    /* Buffered packets are stored here */
    adk_dp_eth_next_hop_buffer_t* nexthop_buffer;
}  __attribute__ ((aligned(8))) adk_dp_eth_buffer_info_t;

/* This is only accessible in data plane */
typedef struct adk_eth_dp_arp_buffer_tbl_s
{
    uint32_t nh_ipAddr;
    adk_dp_eth_buffer_info_t* buffer_info;
}adk_eth_dp_arp_buffer_tbl_t;

/* This is only accessible in data plane */
typedef struct adk_eth_dp_nexthop_v4_timer_index_tbl_s
{
    uint64_t  nexthop_timer_index;
    adk_ethernet_ipv4_addr_t next_hop_addr;
}adk_eth_dp_nexthop_v4_timer_index_tbl_t;

/* This is only accessible in data plane */
typedef struct adk_eth_dp_nexthop_v6_timer_index_tbl_s
{   
    uint64_t  nexthop_timer_index;
    adk_ethernet_ipv6_addr_t next_hop_addr;
}adk_eth_dp_nexthop_v6_timer_index_tbl_t;


typedef struct adk_ethernet_dp_next_hop_exception_data_s
{
    adk_interface_handle_t   ifHandle;
    adk_ethernet_vlan_id_t   vlanId;
    union {
        adk_ethernet_ipv4_addr_t v4_address;
        adk_ethernet_ipv6_addr_t v6_address;
    }next_hop;
    bool isIPv6;
} adk_ethernet_dp_next_hop_exception_data_t;

typedef struct adk_ethernet_arp_dp_exception_data_s
{
    adk_interface_handle_t   ifHandle;
    adk_ethernet_vlan_id_t   vlanId;
    adk_ethernet_ipv4_addr_t next_hop_address;
} adk_ethernet_arp_dp_exception_data_t;

typedef struct adk_ethernet_neigh_dp_exception_data_s
{
    adk_interface_handle_t   ifHandle;
    adk_ethernet_vlan_id_t   vlanId;
    adk_ethernet_ipv6_addr_t v6_neigh_address;
} adk_ethernet_neigh_dp_exception_data_t;

typedef adk_ethernet_dp_next_hop_exception_data_t adk_ethernet_neigh_dp_stale_entry_exception_data_t;

typedef struct eth_ph_metadata_s
{
    adk_ethernet_exception_type_t exception_type;
    union {
        adk_ethernet_arp_dp_exception_data_t arp_exception_data;
        adk_ethernet_neigh_dp_exception_data_t neigh_exception_data;
        adk_ethernet_neigh_dp_stale_entry_exception_data_t neigh_stale_exception_data;
    }exception_dp_data;

} __attribute__((aligned(ADK_METADATA_ALIGN))) eth_ph_metadata_t;

typedef struct adk_ethernet_dp_exception_handler_s {
    /*!
     * type field decides the type of exception.
     */
    uint8_t devId;
    uint8_t  priority;
    uint8_t  cbFnId;
    uint16_t queueId;
    uint32_t policerId;
    adk_ethernet_exception_type_t type;
}__attribute__((aligned(ADK_METADATA_ALIGN))) adk_ethernet_dp_exception_handler_t;


/*
 * Etherent neighbor Table hash Entry
 */
typedef struct adk_dp_eth_neighbor_hash_entry_s
{
    uint8_t     is_enabled;
    uint8_t     reserved;
    uint16_t    index;
    uint32_t    hash_value;
}__attribute__ ((aligned(8))) adk_dp_eth_neighbor_hash_entry_t;


/*
 * Ethernet neighbor Table Entry
 */
typedef struct adk_dp_eth_neighbor_entry_s
{
    uint16_t                    idx;
    uint16_t                    next_idx;
    uint8_t                     is_enabled:1;
    uint8_t                     reserved1:7;
    adk_ethernet_ipv6_addr_t    ipv6_addr;
    adk_ethernet_mac_addr_t     dest_mac;
    adk_ethernet_neigh_status_t neigh_status;
    /* This is just a flag to send packet to 
       buffer stage if inflight_count is non-zero */
    bool                        have_inflight;
    int16_t                     stale_exception_raised;
}__attribute__ ((aligned(64))) adk_dp_eth_neighbor_entry_t;


/*
 * Ethernet neighbor Table Entry
 */
typedef struct adk_dp_eth_arp_entry_s
{
    uint16_t                    idx;
    uint16_t                    next_idx;
    uint8_t                     is_enabled:1;
    uint8_t                     reserved1:7;
    adk_ethernet_ipv4_addr_t    ipv4_addr;
    adk_ethernet_mac_addr_t     dest_mac;
    adk_ethernet_arp_status_t arp_status;
    /* This is just a flag to send packet to 
       buffer stage if inflight_count is non-zero */
    bool                        have_inflight;
    int16_t                     stale_exception_raised;
}__attribute__ ((aligned(64))) adk_dp_eth_arp_entry_t;


typedef struct adk_dp_eth_overflow_node_s
{
    uint16_t idx;
    struct adk_dp_eth_overflow_node_s *next;
}adk_dp_eth_overflow_node_t;

typedef struct adk_eth_dp_linked_list_s
{
    adk_dp_eth_overflow_node_t *head;
    uint16_t count;
}adk_dp_eth_linked_list_t;

typedef struct adk_ipv6_dp_array_list_s
{
    uint16_t    head_idx;
    uint16_t    tail_idx;
    uint32_t    count;
}adk_dp_eth_array_list_t;
/*
 * Ethernet Neighbor Table (Hash Table)
 */
typedef struct adk_dp_eth_arp_table_s
{
    adk_dp_eth_neighbor_hash_entry_t    *hash_tbl;
    adk_dp_eth_arp_entry_t              *tbl;
    adk_dp_eth_linked_list_t            overflow_list;
    adk_dp_eth_array_list_t             pending_list;
    adk_dp_eth_linked_list_t            overflow_pending_list;
    adk_dp_eth_array_list_t             free_list;
}adk_dp_eth_arp_table_t;


/*
 * Ethernet ARP Table (Hash Table)
 */
typedef struct adk_dp_eth_neighbor_table_s
{
    adk_dp_eth_neighbor_hash_entry_t    *hash_tbl;
    adk_dp_eth_neighbor_entry_t         *tbl;
    adk_dp_eth_linked_list_t            overflow_list;
    adk_dp_eth_array_list_t             pending_list;
    adk_dp_eth_linked_list_t            overflow_pending_list;
    adk_dp_eth_array_list_t             free_list;
}adk_dp_eth_neighbor_table_t;

/* This is only accessible in data plane */
typedef struct adk_eth_dp_neighbor_buffer_tbl_s
{
    
    uint8_t enable:1;
    uint8_t reserved1:7;
    uint32_t hash_value;
    adk_ethernet_ipv6_addr_t nh_ipv6Addr;
    adk_dp_eth_buffer_info_t* buffer_info;
}adk_eth_dp_neighbor_buffer_tbl_t;


/***********************************************************************
******* MODULE LEVEL STRUCT HAVING POINTERS TO ALL DP TABLES **********
***********************************************************************/
typedef struct adk_eth_dp_pvt_data_s
{
    /* Cookie for validation */
    uint32_t cookie;
    /* Per core global stats */
    adk_dp_eth_gl_stats_t *gl_stats_p;
    /* ethernet nexthop flow cache version*/
    uint8_t eth_fc_ver;
    /* List of all Ethernet switch-virtual-interfaces */
    adk_dp_eth_svif_intf_t **eth_svif_if;
    /* Ethernet svif-interface hash table */
    adk_dp_eth_svif_if_hash_t *eth_svif_if_hash_table;
    /* Ethernet Bridge interface*/
    adk_dp_eth_bridge_intf_t        **bridge_intf;
    /* Bridge L2 domain to stats ID map */
    uint32_t bridge_map_id[ADK_ETH_MAX_BG_ID];
    /* default L2 domain for L3 call */
    uint32_t* default_l2_domain;
    /*Ethernet Autoconnection status*/
    uint8_t *autoconn_status;
    /* Ethernet arp tables */
    adk_dp_eth_arp_table_t eth_arp_hash_table;
    /* Ethernet neighbor hash table */
    adk_dp_eth_neighbor_table_t eth_neighbor_hash_table;
    /*Ethernet DSCP to PCP Map*/
    uint8_t   *eth_dscp_to_pcp_map;
    /* Ethernet ingress flow offload list */
    adk_dp_eth_ingress_flow_t **eth_ingress_flow_list;
    /* Ethernet Type to PCP Mapping table */
    adk_lpm_handle_t ethTypePcpMapTbl;
    /*Ethernet Type to Queue Mapping Table*/
    adk_dp_eth_ethertype_queue_map_hash_t *eth_ethertype_queue_map_hash_table;
    /*Default QueueID for Non bridge originated netd traffic*/
    uint16_t  *eth_non_bridge_origin_queue_id;
    /*Bride timestamp Quad Group*/
    adk_ethernet_quad_group_t  *eth_bridge_timestamp_quad_group;
    /*Bridge timestamp Quad Group Port*/
    uint8_t  *eth_bridge_timestamp_quad_group_port;
    /*Ethernet exception table*/
    adk_ethernet_dp_exception_handler_t  *exceptionHandlerTble[ETH_EXC_CODE_MAX_NUM];
    /* Ethernet arp buffer table.
    Only accessible through DP, the critical design pillar */
    adk_eth_dp_arp_buffer_tbl_t *eth_arp_buffer_tbl;
    /* Global lock for accessing eth_arp_buffer_tbl */ 
    void* arp_buffer_lock;
    adk_eth_dp_nexthop_v4_timer_index_tbl_t *eth_nexthop_v4_timer_index_tbl;
    adk_eth_dp_nexthop_v6_timer_index_tbl_t *eth_nexthop_v6_timer_index_tbl;
    /* Ethernet neighbor buffer info table. 
    Only accessible through DP, the critical design pillar */
    adk_eth_dp_neighbor_buffer_tbl_t *eth_neighbor_buffer_tbl;
    /* Global lock for accessing eth_neighbor_buffer_tbl.
    Separate lock is required since IPv4 and IPv6 can be processed
    in parallel on different worker cores */ 
    void* neighbor_buffer_lock;
    /*Ethernet buffer info mempool*/
    struct rte_mempool *eth_buffer_info_pool;
    /*Ethernet nexthop buffer mempool*/
    struct rte_mempool *eth_nexthop_buffer_pool;
    /* Supported ethernet types */
    adk_hos_rule_id_t adk_eth_ether_type[ADK_ETHERNET_ETHERTYPE_SUPPORTED_MAX];
     /* Ethernet ingress flow LPM table */
    adk_lpm_handle_t ingressFlowTbl;
    /* Hash Table List */
    adk_eth_fdb_mac_table_entry_t   *hashIndex[ADK_ETH_MAC_FDB_HASH_MAX];
    /*Ethernet utility counter status*/
    bool                        isUtilityCntEnabled;
    /*Ethernet utility counters */
    adk_eth_utility_counters_t  ethUtilityCnt[ADK_MAX_NUM_CORES];
} adk_eth_dp_pvt_data_t;

/* Macros used by control plane to operate upon DP objects */

#define ADK_ETH_INIT_QOS_QUEUES(dp_if_p) \
{   \
    int qos_pcp;    \
    for (qos_pcp = 0; qos_pcp < ADK_ETH_MAX_PCP; qos_pcp++) \
    {   \
        dp_if_p->qos_queue[qos_pcp].queue_id = ETH_INVALID_QOS_QID; \
    }   \
    dp_if_p->default_qos_queue.queue_id = ETH_INVALID_QOS_QID;  \
}


#define ADK_ETH_DP_NEIGH_TABLE_IDX_INIT(dp_tbl, num_entries) \
    do { \
        uint16_t ne; \
        for(ne=0; ne<(num_entries-1); ne++) \
        { \
            dp_tbl.tbl[ne].idx = ne; \
            dp_tbl.tbl[ne].next_idx = ne+1; \
        } \
        dp_tbl.tbl[ne].idx = ne; \
        dp_tbl.tbl[ne].next_idx = UINT16_MAX; \
        dp_tbl.free_list.head_idx = 0; \
        dp_tbl.free_list.tail_idx = num_entries-1; \
        dp_tbl.free_list.count = num_entries; \
        dp_tbl.pending_list.head_idx = UINT16_MAX; \
        dp_tbl.pending_list.tail_idx = UINT16_MAX; \
        dp_tbl.pending_list.count = 0; \
    }while(0);

#define ADK_ETH_DP_CLEANUP_TBL(pending_list, free_list, tbl)    \
    do{\
        if(pending_list.count) \
        {\
            if(!(free_list).count) \
            { \
                (free_list).head_idx = pending_list.head_idx; \
                (free_list).tail_idx = pending_list.tail_idx; \
            } \
            else \
            { \
                tbl[(free_list).tail_idx].next_idx = tbl[pending_list.head_idx].idx; \
                (free_list).tail_idx = pending_list.tail_idx; \
            } \
            (free_list).count += pending_list.count; \
            pending_list.count = 0; \
        }\
    }while(0);

#define ADK_ETH_DP_RCU_DISTRIBUTED_CHECK(dp_tbl)\
    do{\
        if(((dp_tbl)->free_list.count==0) && ((dp_tbl)->pending_list.count!=0)) \
        { \
            adk_rcu_distributed_wait(); \
            ADK_ETH_DP_CLEANUP_TBL((dp_tbl)->pending_list, \
                    (dp_tbl)->free_list, (dp_tbl)->tbl); \
        } \
        else if(adk_rcu_distributed_ver_verify() == ADK_SUCCESS) \
        { \
            ADK_ETH_DP_CLEANUP_TBL((dp_tbl)->pending_list, \
                    (dp_tbl)->free_list, (dp_tbl)->tbl); \
        }\
    }while(0)

#define ADK_ETH_DP_CLEANUP_OVERFLOW_TBL(dp_tbl) \
    do{ \
        adk_dp_eth_overflow_node_t *node, *next_node; \
        node = dp_tbl->overflow_pending_list.head; \
        while(node) \
        { \
            next_node = node->next; \
            ADK_DP_SHM_FREE(node); \
            node = next_node; \
        }\
        dp_tbl->overflow_pending_list.count = 0; \
        dp_tbl->overflow_pending_list.head = NULL; \
    }while(0)

#define ADK_ETH_DP_RCU_DISTRIBUTED_CHECK_OVERFLOW_TBL(dp_tbl)\
    do{\
        if(((dp_tbl)->free_list.count==0) && ((dp_tbl)->overflow_pending_list.count!=0)) \
        { \
            adk_rcu_distributed_wait(); \
            ADK_ETH_DP_CLEANUP_OVERFLOW_TBL(dp_tbl); \
        } \
        else if(adk_rcu_distributed_ver_verify() == ADK_SUCCESS) \
        { \
            ADK_ETH_DP_CLEANUP_OVERFLOW_TBL(dp_tbl); \
        }\
    }while(0)

#define ADK_ETH_DP_RSRV_IDX(free_list, tbl, index) \
        do{ \
            if((free_list).count) \
            { \
                (index) = (free_list).head_idx; \
                (free_list).head_idx = tbl[(index)].next_idx; \
                (free_list).count--; \
            } \
            else \
            { \
                ADK_ETH_LOG_INFO(eth_mod_handle, "Error: no free resources \n"); \
                return ADK_DP_E_ERROR; \
            } \
        }while(0);

#define ADK_ETH_DP_FREE_IDX(list, tbl, index) \
    do{ \
        if(!(list).count) \
        { \
            (list).head_idx = index; \
            (list).tail_idx = index; \
        } \
        else \
        { \
            tbl[(list).tail_idx].next_idx = tbl[(index)].idx; \
            (list).tail_idx = index; \
        } \
        tbl[(index)].next_idx = UINT16_MAX; \
        (list).count++; \
    }while(0);


#define ADK_ETH_DP_ADD_OVERFLOW_ENTRY(overflow_list, \
        free_list, tbl, tbl_index)\
    do{ \
        adk_dp_eth_overflow_node_t *node = ADK_DP_SHM_CALLOC(ADK_ETHERNET_MODULE_ID, ADK_ETH_NEIGH_OBJ, 1, sizeof(adk_dp_eth_overflow_node_t), ADK_ETH_DP_PTR_ALIGN);\
        if(node == NULL) \
        { \
            printf("unable to allocate memory \n"); \
            /* release the allocated index*/ \
            /* need not to wait for RCU, becuase this entry is not visible to fast-path yet */ \
            ADK_ETH_DP_FREE_IDX(free_list, tbl, tbl_index); \
            return(ADK_E_INTERNAL_ERROR); \
        } \
        node->idx = tbl_index; \
        node->next = NULL; \
        if(overflow_list.head) \
        { \
            node->next = overflow_list.head; \
            overflow_list.head = node; \
        } \
        else \
        { \
            overflow_list.head = node; \
        } \
        overflow_list.count++; \
    }while (0)


#define ADK_ETH_DP_RELEASE_DISTRIBUTED_IDX(pending_list, free_list, tbl, index) \
    do{ \
        ADK_ETH_DP_FREE_IDX(pending_list, tbl, index) \
        adk_rcu_distributed_ver_update(); \
    }while(0);

#define ADK_ETH_DP_RELEASE_OVERFLOW_ENTRY(of_list, of_pending_list, \
        pending_list, free_list, tbl, tbl_index) \
    do{\
        /* check for ovrflow list */ \
        adk_dp_eth_overflow_node_t *node, *prev=NULL; \
        node = of_list.head; \
        while(node) \
        { \
            if( node->idx == tbl_index) \
            { \
                /* detatch ll-node from overflow_list */ \
                if(node == of_list.head) \
                { \
                    of_list.head = node->next; \
                } \
                else \
                { \
                    prev->next = node->next; \
                } \
              of_list.count--; \
                /* add ll-node to overflow pending list */ \
                node->next = of_pending_list.head; \
                of_pending_list.head = node; \
                of_pending_list.count++; \
                /* release the tbl_idx to pending list*/ \
                ADK_ETH_DP_RELEASE_DISTRIBUTED_IDX(pending_list, \
                        free_list, \
                        tbl, \
                        tbl_index); \
                return ADK_SUCCESS; \
            } \
            prev = node; \
            node = node->next; \
        } \
    }while(0)

#ifdef ADK_UTILITY_COUNTER_ENABLE

#define INCREMENT_ETH_UTILITY_COUNTER_P(counter) \
    do  \
    {       \
        if(adk_eth_dp_pvt_g_p->isUtilityCntEnabled)  \
        {  \
            (*counter)++;  \
        }  \
    }while(0);

#define INCREMENT_ETH_UTILITY_COUNTER(counter) \
    do  \
    {       \
        if(adk_eth_dp_pvt_g_p->isUtilityCntEnabled)  \
        {  \
            counter++;  \
        }  \
    }while(0);
#define INCREMENT_ETH_UTILITY_COUNTER_COUNT(counter,count) \
    do  \
    {       \
        if(adk_eth_dp_pvt_g_p->isUtilityCntEnabled)  \
        {  \
            counter = counter + count;  \
        }  \
    }while(0);

#else
#define INCREMENT_ETH_UTILITY_COUNTER(counter)
#define INCREMENT_ETH_UTILITY_COUNTER_COUNT(counter,count)
#define INCREMENT_ETH_UTILITY_COUNTER_P(counter)
#endif
/* Inline DP functions */

static inline void adk_eth_dp_update_rx_stats(
            adk_dp_eth_stats_t *dp_intf_stats_p,
            adk_pktio_task_t *task_in,
            uint8_t *pdu)
{
    /* Increment [UMB]cast packets received based upon DMAC */
    if (likely(!(pdu[0] & 0x01)))
    {
        dp_intf_stats_p->inFrames++;
        dp_intf_stats_p->inOctets += ADK_TASK_PDU_SIZE(task_in) + ADK_ETH_CRC_SIZE;
    }
    else if (IS_BROADCAST_MAC(pdu))
    {
        if(((*(uint16_t*)&pdu[(task_in->l2_len - ETHER_TYPE_SIZE)]) == adk_htons(ETHER_TYPE_IPV4)) ||
                ((*(uint16_t*)&pdu[(task_in->l2_len - ETHER_TYPE_SIZE)]) == adk_htons(ETHER_TYPE_IPV6)))
        {
            dp_intf_stats_p->inIpBcastFrames++;
            dp_intf_stats_p->inIpBcastOctets += ADK_TASK_PDU_SIZE(task_in) + ADK_ETH_CRC_SIZE;
        }
        else
        {
            dp_intf_stats_p->inNonIpBcastFrames++; 
            dp_intf_stats_p->inNonIpBcastOctets += ADK_TASK_PDU_SIZE(task_in) + ADK_ETH_CRC_SIZE;
        }
    }
    else
    {
        if(((*(uint16_t*)&pdu[(task_in->l2_len - ETHER_TYPE_SIZE)]) == adk_htons(ETHER_TYPE_IPV4)) ||
                ((*(uint16_t*)&pdu[(task_in->l2_len - ETHER_TYPE_SIZE)]) == adk_htons(ETHER_TYPE_IPV6)))
        { 
            dp_intf_stats_p->inIpMcastFrames++;
            dp_intf_stats_p->inIpMcastOctets += ADK_TASK_PDU_SIZE(task_in) + ADK_ETH_CRC_SIZE;
        }
        else
        {
            dp_intf_stats_p->inNonIpMcastFrames++;
            dp_intf_stats_p->inNonIpMcastOctets += ADK_TASK_PDU_SIZE(task_in) + ADK_ETH_CRC_SIZE;
        }
    }
}

static inline void adk_eth_dp_update_tx_stats(
            adk_dp_eth_stats_t *dp_intf_stats_p,
            adk_pktio_task_t *task_out,
            adk_metadata_t *adk_md_p)
{
    uint64_t pktOctets = 0;
    /* Increment bytes transmitted statistics including padding and CRC */
    if (ADK_TASK_PDU_SIZE(task_out) < (ADK_ETH_MIN_PKT_SIZE -
                                       ADK_ETH_CRC_SIZE))
    {
        pktOctets += ADK_ETH_MIN_PKT_SIZE;
    }
    else
    {
        pktOctets += (ADK_TASK_PDU_SIZE(task_out) +
                                     ADK_ETH_CRC_SIZE);
    }

    if (adk_md_p->isEthernetIpMcast)
    {
        dp_intf_stats_p->OutIpMcastFrames++;
        dp_intf_stats_p->OutIpMcastOctets += pktOctets;
    }
    else if (adk_md_p->isEthernetIpBcast)
    {
        dp_intf_stats_p->OutIpBcastFrames++;
        dp_intf_stats_p->OutIpBcastOctets += pktOctets;
    }
    if (adk_md_p->isEthernetNonIpMcast)
    {
        dp_intf_stats_p->OutNonIpMcastFrames++;
        dp_intf_stats_p->OutNonIpMcastOctets += pktOctets;
    }
    else if (adk_md_p->isEthernetNonIpBcast)
    {
        printf("OutNonIpBcastFrames\n");
        dp_intf_stats_p->OutNonIpBcastFrames++;
        dp_intf_stats_p->OutNonIpBcastOctets += pktOctets;
    }
    else
    {
        dp_intf_stats_p->OutFrames++;
        dp_intf_stats_p->OutOctets += pktOctets;
    }
}

static inline uint16_t adk_eth_get_egress_pad_fcs_len_corr(
            adk_pktio_task_t *task_out)
{
    if (ADK_TASK_PDU_SIZE(task_out) <
        (ADK_ETH_MIN_PKT_SIZE - ADK_ETH_CRC_SIZE))
    {
        return (ADK_ETH_MIN_PKT_SIZE - ADK_TASK_PDU_SIZE(task_out));
    }
    else
    {
        return ADK_ETH_CRC_SIZE;
    }
}


void
adk_eth_dp_tx_md_update (
    adk_pktio_task_t *task_out,
    uint16_t eth_dev_id,
    uint32_t out_port,
    adk_dp_eth_qos_info_t *qos_queue,
    int coreIndex);

adk_status_t
adk_eth_dp_handle_exception(
    adk_pktio_task_t *ph_task,
    uint8_t type,
    void *exception_data);

void adk_eth_dp_nexthop_buffer_free
(
adk_dp_eth_buffer_info_t* nexthop_buffer_info
);

void adk_eth_dp_task_ring_free
(
void* task_list,
uint32_t *drop_count_p
);

adk_dp_status_t
adk_eth_dp_arp_buffer_add
(
adk_eth_dp_arp_buffer_tbl_t* arp_buf_tbl,
uint32_t nh_ipAddr,
adk_dp_eth_buffer_info_t* arp_buffer
);

adk_dp_status_t
adk_eth_dp_arp_buffer_rem
(
adk_eth_dp_arp_buffer_tbl_t* arp_buf_tbl,
uint32_t nh_ipAddr
);

adk_dp_status_t
adk_eth_dp_arp_buffer_get
(
adk_eth_dp_arp_buffer_tbl_t* arp_buf_tbl,
uint32_t nh_ipAddr,
adk_dp_eth_buffer_info_t** arp_buffer_p
);

#ifdef ADK_DEBUG

adk_status_t adk_eth_dp_arp_buffer_statistics_get
(
adk_dp_eth_arp_table_t *eth_arp_table,
adk_ethernet_ipv4_addr_t ip_addr,
uint32_t* num_inflight_pkts,
uint32_t* num_full_buffered_pkts,
uint32_t* num_frag_buffered_pkts
);

#endif

adk_dp_status_t
adk_eth_dp_arp_entry_lookup
(
 adk_dp_eth_arp_table_t *eth_arp_table,
 uint32_t ip_addr,
 adk_dp_eth_arp_entry_t **arp_entry
 );

adk_status_t
adk_eth_dp_arp_entry_add(
        adk_dp_eth_arp_table_t *eth_arp_table,
        uint32_t ip_addr,
        uint16_t index);

adk_status_t
adk_eth_dp_arp_entry_del(
        adk_dp_eth_arp_table_t *eth_arp_table,
        uint32_t ip_addr);


adk_status_t
adk_eth_dp_neighbor_key_generate(uint32_t *ipv6_addr,
        uint32_t *key,
        uint16_t *key1,
        uint16_t *key2);

adk_status_t
adk_eth_dp_neighbor_hash_entry_add(adk_dp_eth_neighbor_table_t *eth_neighbor_table, uint8_t *neighIpv6Addr,
                                        uint16_t index);

adk_status_t
adk_eth_dp_neighbor_hash_entry_lookup(adk_dp_eth_neighbor_table_t *eth_neighbor_table, uint8_t *neighIpv6Addr,
                            adk_dp_eth_neighbor_hash_entry_t **neigh_looked_up_entry);

adk_status_t
adk_eth_dp_neighbor_hash_overflow_lookup(adk_dp_eth_neighbor_table_t *eth_neighbor_table, uint8_t *neighIpv6Addr,
                            adk_dp_eth_overflow_node_t **overflow_looked_up_node);

adk_dp_status_t
adk_eth_dp_neighbor_buffer_add
(
adk_eth_dp_neighbor_buffer_tbl_t* neigh_buf_tbl,
uint8_t* nh_ipv6Addr,
adk_dp_eth_buffer_info_t* buffer
);

adk_dp_status_t
adk_eth_dp_neighbor_buffer_rem
(
adk_eth_dp_neighbor_buffer_tbl_t* neigh_buf_tbl,
uint8_t* nh_ipv6Addr
);

adk_dp_status_t
adk_eth_dp_neighbor_buffer_get
(
adk_eth_dp_neighbor_buffer_tbl_t* neigh_buf_tbl,
uint8_t* nh_ipv6Addr,
adk_dp_eth_buffer_info_t** buffer_p
);

void adk_eth_dp_slowpath_policer_stats(void *user_params);

adk_status_t adk_eth_dp_pkt_proc_police (void **task_p,
          __attribute__((unused)) uint8_t task_count,
          __attribute__((unused)) uint32_t task_mask,
          __attribute__((unused)) void *ring,
          __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
          int coreIndex);

void
adk_eth_dp_egress_stats(
struct rte_mbuf *pkt,
void *data);

void adk_eth_dp_pal_batch_cb_fn(
        adk_pktio_task_t *dpdkMbuf,
        uint8_t       cbEventType,
        void          *cbParam,
        bool          sendSuccess);

#endif  /* #ifndef __ADK_ETH_DP_H__ */

/*!  end addtogroup _ethernet  Ethernet Module
 * @}
 */
