/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *
 * @brief
 *
 * This file contains all data types and API prototypes for IP module.
 *
 ***********************************************************************/
#ifndef _ADK_IP_H_
#define _ADK_IP_H_

#include <rte_mbuf.h>
#include "adk_ip_config.h"
#include "adk_ph_dev.h"
#include "adk_domain.h"
#include "adk_qos_include.h"

/*!
 * =============================================================
 *
 * @typedef adk_ip_log_level_t
 *
 * @brief   The 32-bit IPv4 module's log-level
 *
 *  This data type is used for setting the log-level of IPv4 module. It is
 *  represented using unsigned long integer and is a logical OR or following
 *  log-types defined by IPv4 module:
 *
 *  ADK_IP_LOG_FATAL       For logging any unrecoverable fatal error
 *
 *  ADK_IP_LOG_ERROR       For logging all other errors
 *
 *  ADK_IP_LOG_WARNING     For logging any any warnings
 *
 *  ADK_IP_LOG_INFO        For logging all resource allocations, etc
 *
 *  ADK_IP_LOG_FUNCS       For logging entry to all internal/external functions
 *
 *  ADK_IP_LOG_HW          For logging all access to EIOA hardware
 *
 *  ADK_IP_LOG_PTN         For logging all learning/unlearning of Tree entries
 *
 *  ADK_IP_LOG_SE_MEM      For logging all access to SE memory
 *
 *  ADK_IP_LOG_EXT_API     For logging all external API calls
 *
 *  ADK_IP_LOG_ALL         For logging all above
 *
 *  The following are some of the pre-defined commonly used log-levels.
 *
 *  ADK_IP_LOG_LEVEL_ERROR     enable ADK_IP_LOG_ERROR & ADK_IP_LOG_FATAL only
 *
 *  ADK_IP_LOG_LEVEL_WARNING   enable ADK_IP_LOG_WARNING, ADK_IP_LOG_ERROR &
 *                           ADK_IP_LOG_FATAL only
 *
 *  ADK_IP_LOG_LEVEL_INFO      enable ADK_IP_LOG_INFO, ADK_IP_LOG_WARNING,
 *                           ADK_IP_LOG_ERROR & ADK_IP_LOG_FATAL only
 *
 *  ADK_IP_LOG_LEVEL_VERBOSE   enable all types of logs. Same as ADK_IP_LOG_ALL
 *
 *  ADK_IP_LOG_LEVEL_DEFAULT   Compile time default log level setting
 *
 * =============================================================
 */
typedef uint32_t adk_ip_log_level_t;

#define ADK_IP_LOG_DISABLE      (0 << 0)
#define ADK_IP_LOG_FATAL        (1 << 0)
#define ADK_IP_LOG_ERROR        (1 << 1)
#define ADK_IP_LOG_WARNING      (1 << 2)
#define ADK_IP_LOG_INFO         (1 << 3)
#define ADK_IP_LOG_FUNCS        (1 << 4)
#define ADK_IP_LOG_HW           (1 << 5)
#define ADK_IP_LOG_PTN          (1 << 6)
#define ADK_IP_LOG_SE_MEM       (0 << 0) /*Inherited from previous releases, but not used anymore. so disabling it.*/
#define ADK_IP_LOG_SUBNET       (1 << 7)
#define ADK_IP_LOG_EXT_API      (1 << 8)

#define ADK_IP_LOG_ALL          0xFFFFFFFF

#define ADK_IP_LOG_LEVEL_ERROR         (ADK_IP_LOG_ERROR | ADK_IP_LOG_FATAL)
#define ADK_IP_LOG_LEVEL_WARNING       (ADK_IP_LOG_LEVEL_ERROR | ADK_IP_LOG_WARNING)
#define ADK_IP_LOG_LEVEL_INFO          (ADK_IP_LOG_LEVEL_WARNING | ADK_IP_LOG_INFO)
#define ADK_IP_LOG_LEVEL_VERBOSE       ADK_IP_LOG_ALL

#define ADK_IP_LOG_LEVEL_DEFAULT       ADK_IP_LOG_LEVEL_ERROR

/*!
 * =============================================================
 * @enum adk_ip_status_t
 *
 * @brief
 *
 * This enumeration defines all error codes return by IP module
 * external APIs.
 *
 * =============================================================
 */
typedef enum adk_ip_status_e {
    /*! Successfully completed requested operation */
    ADK_IP_SUCCESS = ADK_SUCCESS,
    /*! Failed to allocate NAT pipeline flow. */
    ADK_IP_E_NAPT_FLOW_UNAVAILABLE = ((ADK_IP_MODULE_ID << 24) + 1),
    /*! Failed to allocate ingress state engine memory block */
    ADK_IP_E_INGRESS_STATE_MEMORY_UNAVAILABLE,
    /*! Failed to find or get Namespace ID */
    ADK_IP_E_INGRESS_STATE_ID_UNAVAILABLE,
    /*! Failed to allocate egress state engine memory block */
    ADK_IP_E_EGRESS_STATE_MEMORY_UNAVAILABLE,
    /*! Invalid IP association flow passed as input */
    ADK_IP_E_INVALID_IP_ASSOCIATION,
    /*! Invalid IP interface passed as input */
    ADK_IP_E_INVALID_IP_INTERFACE,
    /*! Invalid interface state */
    ADK_IP_E_INVALID_INTERFACE_STATE,
    /*! Invalid IP version type in the IP address field */
    ADK_IP_E_INVALID_IP_VERSION,
    /*! Invalid Source address paramter. IP flow classification does
     * not support source IP address subnet currently*/
    ADK_IP_E_INVALID_SRC_IP_ADDRESS,
    /*! Invalid IP Qos profile type information.
     */
    ADK_IP_E_INVALID_QOS_PROFILE_TYPE,
    /*! Maximum number of classes that can be configured in single
     * API call is exceeded.
     */
    ADK_IP_E_MAX_CLASS_CONFIG_EXCEEDED,
    /*! Invalid DSCP profile handle or DSCL class type spefiied.
     */
    ADK_IP_E_INVALID_DSCP_PROFILE,
    /*!
     * Input/Ouptut (or both) address fields missing for NAPT rule
     * configuration API.
     */
    ADK_IP_E_NAPT_NO_ADDRESS,
    /*!
     * Output L4 port not specified along with Input L4 port.
     */
    ADK_IP_E_NAPT_NO_OUT_L4PORT,
    /*!
     * Invalid IPv4 address prefix.
     */
    ADK_IP_E_INVALID_ADDR_PREFIX,
    /*!
      * IP Forwarding table programming Failures. This can happen when the
      * hardware update fails.
     */
    ADK_IP_E_FIB_UPDATE_FAILURE,
    /*!
      * Invalid IP host or interface address
     */
    ADK_IP_E_INVALID_IP_ADDRESS,
    /*!
      * Invalid IP exception type.
     */
    ADK_IP_E_INVALID_EXCEPTION_TYPE,
    /*!
      * Invalid MTU configuration.
     */
    ADK_IP_E_INVALID_MTU,
    /*!
     * Committed (rate or burst) value is bigger than peak value
     * (rate or burst) for policing flow configuration.
     */
    ADK_IP_E_INCONSISTENT_RATES,
    /*!
     * Out of internal state memory
     */
    ADK_IP_E_INTERNAL_MEMORY_UNAVAILABLE,
    /*!
     * Invalid PBR policy set
     */
    ADK_IP_E_INVALID_PBR_POLICY_SET,
    /*!
     * Invalid route table id
     */
    ADK_IP_E_INVALID_ROUTE_TABLE_ID,

    /*!
     * Max number of routes already configured in the system.
     */
    ADK_IP_E_MAX_ROUTES_PROGRAMMED,

    /*!
     * Max ip address limit exceeded
     */
    ADK_IP_E_MAX_IP_ADDR_EXCEEDED,
    /*!
     * Max interface limit exceeded
     */
    ADK_IP_E_MAX_IP_INTERFACES_EXCEEDED,
} adk_ip_status_t;

/*!
 * =============================================================
 * @typedef adk_ip_addr_t
 *
 * @brief
 *
 * The 32-bit IPv4 address
 *
 * =============================================================
 */
typedef uint32_t adk_ip_addr_t;

/*!
 * =============================================================
 * @struct adk_ip_addr_prefix_t
 *
 * @brief
 *
 * The IPv4 subnet address which consists of 32-bit IP address and prefix
 * length.
 *
 * =============================================================
 */
typedef struct adk_ip_addr_prefix_s {
    /*! 32-bit IPv4 address. */
    adk_ip_addr_t addr;
    /*! 32-bit prefix value that indicates how many bits in the
     * addr field are used for subnet definition.
     */
    uint32_t prefixWidth;
} adk_ip_addr_prefix_t;

/*!
 * =============================================================
 * @struct adk_ip_range_t
 *
 * @brief
 *
 * The IPv4 range data type is used to specify a contiguous range
 * of values e.g. range of L4 port numbers or range of IP protocol
 * types, etc. This type of data type is used in PBR configuration.
 *
 * =============================================================
 */
typedef struct adk_ip_range_s {
    uint32_t begin;
    uint32_t end;
} adk_ip_range_t;


/*!
 * =============================================================
 * @enum adk_ip_frag_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Defines the packet type w.r.t. to fragmentation. All combinations are
 * provided so that application does not have to add multiple rules for the same
 * flow.
 *
 * =============================================================
 */
typedef enum adk_ip_frag_e {
    /*! Unfragment packet */
    ADK_IP_UNFRAGMENTED_PKT = 0,
    /*! Either Unfragment packet or first fragment. This has L4 fields available
     * in the packet */
    ADK_IP_L4_AVAILABLE_PKT,
    /*! A fragmented packet which can be first (offset = 0) or a non-first
     * fragment (offset != 0) */
    ADK_IP_FRAGMENTED_PKT,
    /*! First fragment of a packet (offset = 0) */
    ADK_IP_FIRST_FRAGMENT,
    /*! Non first fragment of a packet (offset != 0) */
    ADK_IP_NON_FIRST_FRAGMENT
} adk_ip_frag_t;


/*!
 *=============================================================
 * @enum adk_ip_interface_attr_mask_t "adk_ip.h"
 *
 * @brief
 *
 * This enumeration defines the attributes supported 
 * for an IPv4 interface.
 *=============================================================
 */
typedef enum adk_ip_interface_attr_mask_e {

    /*! Disable QoS & FWD on an IP interface*/
    ADK_IP_INTERFACE_MASK_DEFAULT= 0x00,
    /*! Enable forwarding for the interface */
    ADK_IP_INTERFACE_MASK_FORWARDING_ENABLE = 0x01,
    /*! Enable QoS on an IP interface*/
    ADK_IP_INTERFACE_MASK_QOS_ENABLE = 0x02,
    ADK_IP_INTERFACE_MASK_MAX = 0x04

}adk_ip_interface_attr_mask_t;

/*!
 * =============================================================
 * @struct adk_ip_interface_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * This data structure contains parameters that are required
 * to create IP interface.
 *
 * =============================================================
 */
typedef struct adk_ip_interface_s {
    /*! Layer-2 interface handle. E.g. This could be ethernet
     *  interface handle or VLAN based sub-interface handle.
     */
    adk_interface_handle_t layer2InterfaceHandle;

    /*! Address prefix for this IP interface. This will add two entries
     * to the routing table - one for directly connected hosts and another
     * for locally terminate traffic */
    adk_ip_addr_prefix_t ipAddrPrefix;

    /*!
     * Interface MTU value. This value is used to overwrite the
     * default interface MTU. The range of values supported for MTU
     * configuration is between 64 and 9000 bytes. If specified as 0,
     * then default MTU value is used.
     */
    uint16_t mtu;

    /*! Attributes for an IP interface. Enable these attributes by 
     * setting the correct mask values while creating IP interface. 
     */
    adk_ip_interface_attr_mask_t ipIntfAttr;

} adk_ip_interface_t;

/*!
 * =============================================================
 * @struct adk_ip_virtual_interface_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * This data structure contains parameters that are required
 * to create virtual IP interface.
 *
 * =============================================================
 */
typedef struct adk_ip_virtual_interface_s {
    /*! The 32-bit IPv4 Address for this IP interface. This will add an entry
     * to the routing table for L3 termination */
    adk_ip_addr_t ipAddress;
} adk_ip_virtual_interface_t;


/*!
 * =============================================================
 * @struct adk_ip_interface_ingress_policing_attr_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * This data structure is used for configuring IP interface level
 * ingress policing policy. This consists of a flag to enable or
 * disable policing and the policing flow handle (when policing is
 * enabled).
 *
 * =============================================================
 */
typedef struct adk_ip_interface_ingress_policing_attr_s {
    /*!
     * enablePolicing. This enable/disable policing of traffic received on
     * this interface. Possible values are TRUE or FALSE.
     */
    bool enablePolicing;
    /*!
     * policeFlowHandle. This configures the policing flow handle
     * to be used for policing traffic received on this interface. This is
     * valid only if enablePolicing is set to TRUE. Application must create
     * the policing flow using adk_qos_policing_flow_create API (with type
     * set to ADK_QOS_POLICE_TYPE_TRTCM_COLOR_BLIND).
     */
    adk_flow_handle_t policeFlowHandle;
} adk_ip_interface_ingress_policing_attr_t;

/*!
 *
 * Exception handler specific data structures.
 */


/*!
 * =============================================================
 * @enum adk_ip_direction_t
 *
 * @brief
 *
 * Defines the direction in which configuration is applied. This
 * can only be ingress or egress.
 *
 * =============================================================
 */
typedef enum adk_ip_direction_s {
        ADK_IP_DIR_INGRESS = 0,
        ADK_IP_DIR_EGRESS,
        ADK_IP_DIR_MAX
} adk_ip_direction_t;


/*!
 * =============================================================
 * @enum adk_ip_exception_type_t
 *
 * @brief
 *
 * Defines the supported IP exception types.
 *
 * =============================================================
 */
typedef enum adk_ip_exception_type_e {
    /*! Packets with no Exception */
    ADK_IP_NO_EXCEPTION = 0,
    /*! Packets with IP options */
    ADK_IP_OPTIONS_EXCEPTION = 1,
    /*! Failed to fragment large IP datagram received when Do not
     * fragment configuration is set*/
    ADK_IP_FRAGMENTATION_EXCEPTION = 2,
    /*! No forwarding rule or valid next hop to forward the packet */
    ADK_IP_NO_ROUTE_EXCEPTION = 3,
    /*! TTL expired exception */
    ADK_IP_TTL_EXPIRED_EXCEPTION = 4,
    /*! Unsupported L4 protocol received */
    ADK_IP_UNSUPPORTED_PROTO_EXCEPTION = 5,
    /*! Forwarding Disabled exception */
    ADK_IP_FORWARDING_DISABLED_EXCEPTION = 6,
    /*! Packets with invalid IP version */
    ADK_IP_INVALID_VERSION_EXCEPTION = 7,
    /*! Invalid Exception type */
    ADK_IP_EXCEPTION_MAX,
} adk_ip_exception_type_t;

/*!
 * =============================================================
 *
 * @struct adk_ip_exception_handler_t
 *
 * @brief   This structure defines the parameters used by exception handler
 * function.
 *
 * type       Type field decides the type of exception
 *
 * devId      PH Device ID returned by adk_ph_dev_device_configure
 *
 * queueId    PH QueueID used in adk_ph_dev_rx_queue_setup
 *
 * cbFnHandle PH Callback Fn handle return by adk_ph_dev_callback_fn_add()
 *
 * policer_handle Handle for host policer
 *
 * =============================================================
 */
typedef struct adk_ip_exception_handler_s {
    /*!
     * type field decides the type of exception.
     */
    adk_ip_exception_type_t type;
    uint8_t devId;
    uint16_t queueId;
    uint8_t priority;
    adk_ph_dev_cbfn_handle_t cbFnHandle;
    adk_handle_t policer_handle;
} adk_ip_exception_handler_t;

typedef struct adk_ip_options_exception_param_s
{
    adk_interface_handle_t ifaceHandle;
}adk_ip_options_exception_param_t;

typedef struct adk_ip_frag_exception_param_s
{
    adk_interface_handle_t ifaceHandle;
}adk_ip_frag_exception_param_t;


/* For ADK_IP_NO_ROUTE_EXCEPTION*/
typedef struct adk_ip_noroute_exception_param_s
{
    uint8_t routeTableId;
}adk_ip_noroute_exception_param_t;

typedef struct adk_ip_ttl_expired_exception_param_s
{
    adk_interface_handle_t ifaceHandle;
}adk_ip_ttl_expired_exception_param_t;

typedef struct adk_ip_unsup_proto_exception_param_s
{
    adk_interface_handle_t ifaceHandle;
}adk_ip_unsup_proto_exception_param_t;

/* For ADK_IP_FORWARDING_DISABLED_EXCEPTION exception */
typedef struct adk_ip_fwding_disabled_exception_param_s
{
    adk_interface_handle_t ifaceHandle;
    adk_ip_direction_t dir;
}adk_ip_fwding_disabled_exception_param_t;

typedef struct adk_ip_invalid_version_param_s
{
    adk_interface_handle_t ifaceHandle;
}adk_ip_invalid_version_param_t;

typedef struct adk_ip_recv_exception_params_s
{
    /*!
     * exception type when packet is sent to CPU due to some exception.
     */
    adk_ip_exception_type_t exception_type;

    /* rte mbuf structure pointer */
    struct rte_mbuf *mbuf;
    uint16_t ipOffset;

    union
    {
        adk_ip_options_exception_param_t options;
        adk_ip_frag_exception_param_t frag;
        adk_ip_noroute_exception_param_t noRoute;
        adk_ip_ttl_expired_exception_param_t ttl;
        adk_ip_unsup_proto_exception_param_t proto;
        adk_ip_fwding_disabled_exception_param_t fwdingDisabled;
        adk_ip_invalid_version_param_t version;
    }exception_data;

}adk_ip_recv_exception_params_t;

/*!
 * =============================================================
 * @struct adk_ip_route_info_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * This data structure contains information that are required to
 * update IP forwarding table. It is used as input parameter to
 * adk_ip_route_add and adk_ip_route_delete APIs.
 *
 * =============================================================
 */
typedef struct adk_ip_route_info_s {
    /* Routing table ID. Route will be configured on this table.
     */
    uint8_t routeTableId;

    /* Flag to indicate if this is a connected route.
     * When true - l2 interface id must be valid and packet will be sent over that interface.
     * When false - nextHop must be valid.
     */
    bool connectedRoute;
    /*!
     * Destination IP address prefix. This can be a fully specified IP address
     * or a partically specified destination IP network address.
     */
    adk_ip_addr_prefix_t destAddr;

    /*!
     * Layer2  interface handle where packet has to be sent. Used when
     * 'connectedRoute' flag is true.
     */
    adk_interface_handle_t l2IfHd;

    /*!
     * IP address of the next router (hop/gateway) for the destination
     * address. Will be used when 'connected_route' is false.
     */
    adk_ip_addr_t nextHop;

    /*!
     * MTU value that will be used for packets matching
     * this particular route. If set to 0, MTU value
     * from outgoing interface will get used.
     */
    uint16_t routeMTU;
} adk_ip_route_info_t;

/*!
 * =============================================================
 * @enum adk_ip_pbr_field_mask_t
 *
 * @brief
 *
 * This enum contains the bit mask for various fields used as input
 * for Policy Based Routing
 *
 * =============================================================
 */
typedef enum adk_ip_pbr_field_mask_s {
    /*!  IP destination address mask */
    ADK_IP_PBR_SRC_ADDR_MASK = 0x01,
} adk_ip_pbr_field_mask_t;

/*!
 * =============================================================
 * @struct adk_ip_pbr_rule_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Definition of adk_ip_pbr_rule_t
 *
 * This data structure is used as input for Policy based Routing. Since Policy
 * based routing can take arbitrary fields into account for lookup, all fields
 * in here except mask field and policy set are optional. The mask field
 * indicates which fields are specified and which ones are not. If a field is
 * indicated to be valid by setting corresponding bit in the bitmask, then a
 * valid value for that field must be specified.
 *
 * Note that, except for the policy set handle, all other fields are list of
 * values. Further, while the IP addresses are expressed in prefix-length
 * format, all other fields are expressed as contiguous range of values.
 *
 * =============================================================
 */
typedef struct adk_ip_pbr_rule_s {
    /*!
     * IP PBR rule selection mask. Logical OR of bit masks as defined in
     * adk_ip_pbr_field_mask_t data structure
     */
    adk_ip_pbr_field_mask_t pbrRuleMask;
    /*!
     * srcAddrInfo gives the source IP address for this PBR rule. The source
     * IP address can also be partially or fully specified using the prefix
     * length format.
     */
    adk_ip_addr_prefix_t srcAddrInfo;
} adk_ip_pbr_rule_t;

/*!
 * =============================================================
 * @struct adk_ip_pbr_action_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * This data structure contains information about the action for
 * any PBR rule. It is used as input parameter for IP PBR rule
 * creation and for modification of a PBR rule action.
 *
 * =============================================================
 */
typedef struct adk_ip_pbr_action_s {
    uint8_t routeTableId;
} adk_ip_pbr_action_t;

/*!
 * =============================================================
 * @struct adk_ip_pbr_info_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * This data structure contains information that is required to create
 * any IP PBR rule with action. It is used as input parameter to
 * adk_ip_pbr_entry_create API.
 *
 * =============================================================
 */
typedef struct adk_ip_pbr_info_s {
    /*!
     * Input PBR Rule Criteria having all the input fields for policy based
     * routing.
     */
    adk_ip_pbr_rule_t pbrRule;
    /*!
     * Output PBR Rule Action. This part of the rule can be modified later
     * without performing commit operation.
     */
    adk_ip_pbr_action_t pbrAction;
} adk_ip_pbr_info_t;


/*!
 * =============================================================
 * @enum adk_ip_protocol_t
 *
 * @brief
 *
 * defines well known IP protocol fileds.
 *
 * =============================================================
 */
typedef enum adk_ip_protocol_e {
    /*! ICMP protocol type */
    ADK_IP_PROTOCOL_ICMP = 1,
    /*! IGMP protocol type */
    ADK_IP_PROTOCOL_IGMP = 2,
    /*! TCP protocol type */
    ADK_IP_PROTOCOL_TCP = 6,
    /*! UDP protocol type */
    ADK_IP_PROTOCOL_UDP = 17,
    /*! ESP protocol type */
    ADK_IP_PROTOCOL_ESP = 50,
    /*! AH protocol type */
    ADK_IP_PROTOCOL_AH = 51,
    /*! SCTP protocol type */
    ADK_IP_PROTOCOL_SCTP = 132,
    /*! UDP lite protocol type */
    ADK_IP_PROTOCOL_UDP_LITE = 136
} adk_ip_protocol_t;



/*!
 * =============================================================
 * @struct adk_ip_address_stats_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Structure for maintaining IPv4 address stats
 * =============================================================
 */
typedef struct adk_ip_address_stats_s {

    /*! total number of received octets */
    uint64_t ipInOctets;

    /*! total number of received packets */
    uint64_t ipInPkts;

    /*! total number of transmitted octets */
    uint64_t ipOutOctets;

    /*! total number of transmitted packets */
    uint64_t ipOutPkts;

    /*! total number of reassembled packets */
    uint64_t ipInReasmPkts;

    /*! total number of received packets with errors */
    uint32_t ipInErrors;

} adk_ip_address_stats_t;

typedef adk_qos_ingress_stats_t adk_ip_ingress_qos_stats_t;

/*!
* =============================================================
* @struct adk_ip_interface_stats_t adk_ip.h "adk_ip.h"
*
* @brief
*
* Structure to maintain IP interface statistics.
* =============================================================
*/
typedef struct adk_ip_interface_stats_s {

    /*! Number of octets received in IPv4 packets on input interface,
     * including those received in error (e.g. chksum error).*/
    uint64_t ifInOctets;

    /*! Number of IP packets received on input IP interface,
     * including those received in error (e.g. chksum error). */
    uint64_t ifInPkts;

    /*! Number of octets transmitted in IPv4 packets on output interface. */
    uint64_t ifOutOctets;

    /*! Number of transmitted IPv4 packets on output interface */
    uint64_t ifOutPkts;

    /*! Number of IPv4 packets drop due to IPv4 checksum failure on input interface*/
    uint32_t ifInIpCksumErrDrops;

    /*! Number of input IPv4 packets dropped as Destination MAC was either multicast or broadcast but
     * destination ip address was unicast address on input interface
     */
    uint32_t ifInMBcastDMacUcastDipErrDrops;

    /*! Number of input IPv4 packets dropped due to incorrect ip length on input interface.
     * packet length less than minimum ip header length or truncated packet received.
     */
    uint32_t ifInBadIpLengthErrDrops;

    /*! Number of non-terminating IPv4 packets received on input interface but dropped
     * because of TTL expiry (i.e. TTL was 0 or 1) */
    uint32_t ifOutTtlExceededErrDrops;

    /*! Number of IPv4 packets received on input interface dropped due to IP version mismatch */
    uint32_t ifInVersionMismatchErrDrops;

    /*! Number of IPv4 packets received on input interface dropped due to invalid version in IP header */
    uint32_t ifInInvalidVersionErrDrops;

    /*! Number of IPv4 packets received on input interface dropped due to IP options present in IP header */
    uint32_t ifInIpOptionsPresentDrops;

    /*! Number of IPv4 packets received on input interface dropped due to protocol not allowed. */
    uint32_t ifInUnknownProtoErrDrops;

    /*! Number of IPv4 packets received on input interface dropped due to UDP checksum errors */
    uint32_t ifInUDPCksumErrDrops;

    /*! Number of IPv4 packets received on input interface dropped due to ICMP checksum errors */
    uint32_t ifInICMPCksumErrDrops;

    /*! Number of IPv4 packets received on input interface dropped due to destination address 0.0.0.0
     */
    uint32_t ifInZeroDipDrops;

    /*! Number of IPv4 packets that have been successfully fragmented.
     * This includes packets that get dropped at later stage in processing
     */
    uint64_t ifOutFragPkts;

    /*! Number of output fragments generated because of IPv4 fragmentation.
     * This includes packets that get dropped at later stage in processing
     */
    uint64_t ifOutFragments;

    /*! Number of multicast IP packets received on input IP interface,
     * including those received in error (e.g. chksum error).
     */
    uint64_t ifInMcastPkts;

    /*! Number of octets received in IPv4 multicast packets on input interface,
     * including those received in error (e.g. chksum error).
     */
    uint64_t ifInMcastOctets;

    /*! Number of multicast IP packets sent on output IP interface */
    uint64_t ifOutMcastPkts;

    /*! Number of octets sent in IPv4 multicast packets on output interface  */
    uint64_t ifOutMcastOctets;

    /*! Number of IPv4 packets dropped as NE is not part of that multicast
        group */
    uint32_t ifInDiscardedMCastPkts;

    /*! Number of IPv4 octets dropped as NE is not part of that multicast
        group */
    uint32_t ifInDiscardedMCastOctets;

    /*! Number of discarded IPv4 packets at an output interface. */
    uint32_t ifOutDroppedPkts;

    /*! Number of octets of discarded IPv4 packets at an output interface. */
    uint32_t ifOutDroppedOctets;

    /*! Number of IPv4 packets that got dropped as routing was disabled on the interface */
    uint32_t ifRoutingDisabledDrops;

    /*! Number of IPv4 packets dropped due to IP QoS policy lookup failure
        or no default queue configured when QoS is disabled */
    uint32_t ifOutQoSLookupFailDrops;


    /*! Rx QoS Statistics for PHB */
    adk_ip_ingress_qos_stats_t ifRxQoSstats[ADK_QOS_MAX_PHB_CLASS_IDX];
} adk_ip_interface_stats_t;

/*!
 * =============================================================
 * @struct adk_ip_global_stats_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * IP global statistics data structure.
 *
 * =============================================================
 */
typedef struct adk_ip_global_stats_s {

    /*! Number of IPv4 packets dropped as they were too big to be fragmented.
     * Would be num-framgs exceeds max-frag-limit.
     * */
    uint32_t ipOutTooBigDrops;

    /*! Number of IPv4 packets (fragments) dropped after Reassembly timeout (reassembly could
     * not be completed)
     */
    uint32_t ipInReasmToutFragDrops;

    /*! Number of IPv4 packets (fragments) dropped due to lack of resources.
     * This includes:
     *  - Session allocation failures (no-memory),
     *  - Hash allocation failures (collision list exhausted)
     *  - Timer allocation failures
     */
    uint32_t ipInReasmNoResourceDrops;

    /*! Number of IPv4 packets (fragments) dropped as fragment buffer is full
     * (max number of supported fragments already parked in reassembly context)
     */
    uint32_t ipInReasmFragBufferFullDrops;

    /*! Number of IPv4 packets (fragments) dropped as maximum supported reassembly holes
     * reached.
     */
    uint32_t ipInReasmHolesExceededDrops;

    /*! Number of IPv4 packets dropped because matching ingress classifier rule was not found.
     * This is same as ipRmDroppedPktsFiltering.
     */
    uint32_t ipInClassifierDefaultDrops;

    /*! Number of IPv4 packets dropped where ADK tried to send exception but failed as
     * no exception handler registered by application.
     */
    uint32_t ipNoExceptionRegisteredDrops;

    /*! Number of IPv4 packets dropped where ADK tried to send exception but failed
     * dueto exception send failure.
     */
    uint32_t ipExceptionSendFailureDrops;

    /*! Number of IPv4 packets dropped due to policing violation.
     * Same as ipRmDroppedPktsRateLimiting. */
    uint32_t ipInPolicerDrops;

    /*! Number of locally terminating IPv4 packets dropped as per the DROP
     * action configured in the matching ingress classifier rule */
    uint32_t ipInClassifierDrops;

    /*! Number of ingress IPv4 packets dropped due to failure
     * to send the packet to EO.
     */
    uint32_t ipInSendToEoFailDrops;

    /*! Number of ingress IPv4 packets dropped due to failure
     * to send the packet to netdriver.
     */
    uint32_t ipInSendToHostFailDrops;

    /*! Number of ingress IPv4 packets dropped due to unknown classifier action */
    uint32_t ipInClassifierUnknownActionDrops;

    /*! Number of ingress IPv4 packets dropped during reassembly due to
     * overlapping packet offsets (Duplicate packet)
     */
    uint32_t ipInReasmDuplicateDrops;

    /*! Number of egress IPv4 packets dropped at IPv4 layer due to invalid IP
     * header length.
     */
    uint32_t ipOutInvalidIpHdrLenDrops;

    /*! Number of egress IPv4 packets dropped due to non local source address for packets
     * originating from EO.
     */
    uint32_t ipOutNonLocalSipFromEoDrops;

    /*! Number of mbuf allocation failures for reassembly timeout processing */
    uint32_t ipInReasmToutMbufAllocFails;

    /*! Number of egress IPv4 packets dropped due to mbuf allocation failures during fragmentation*/
    uint32_t ipOutFragMbufAllocFailDrops;

    /*! Number of egress IPv4 packets dropped due to actual PDU len received
     * less than len field in IP header.
     */
    uint32_t ipOutTruncatedPktDrops;

    /*! Number of ingress IPv4 packets dropped due to invalid IP interface. */
    uint32_t ipInInvalidIfDrops;

    /*! Number of egress IPv4 packets dropped due to invalid IP interface. */
    uint32_t ipOutInvalidIfDrops;

    /*! Number of egress IPv4 packets dropped due to route lookup failure */
    uint32_t ipOutNoRouteFoundDrops;

    /* Number of egress IPv4 packets dropped as fragmentation is not allowed via DF bit
     * in IPv4 header.
     */
    uint32_t ipOutFragDFSetDrops;

    /*! Number of IPv4 reassemblies failed due to timer expiry. */
    uint32_t ipInReasmTouts;

} adk_ip_global_stats_t;


/*!
 * =============================================================
 * @struct adk_ip_ingress_classifier_stats_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * IP ingress association specific data structure.
 *
 * =============================================================
 */
typedef struct adk_ip_ingress_classifier_stats_s {
    /*! total number of packets received from lower layer */
    uint32_t rxPkts;
    /*! total number of packets transmitted to upper layer or IP ULI */
    uint32_t txPkts;
    /*! total number of packets dropped by policer. 
     *  Same as ipRmDroppedPktsRateLimitingPerRule. */
    uint32_t ipPolicerDroppedPkts;
} adk_ip_ingress_classifier_stats_t;


typedef struct adk_ip_l4_stats_s
{
    /*! Total number of TCP protocol ip packets received.
     *  Fragments are not included as individual packets.
     */
    uint64_t ipTcpInPackets;
    /*! Total number of TCP protocol ip octets received.
     *  Fragments are not included as individual packets.
     */
    uint64_t ipTcpInOctets;
    /*! Total number of TCP protocol ip packets transmitted.
     *  Fragmented packets are not included as individual packets.
     */
    uint64_t ipTcpOutPackets;
    /*! Total number of TCP protocol ip octets transmitted.
     *  Fragmented packets are not included as individual packets.
     */
    uint64_t ipTcpOutOctets;
    /*! Total number of SCTP protocol ip packets received.
     *  Fragments are not included as individual packets.
     */
    uint64_t ipSctpInPackets;
    /*! Total number of SCTP protocol ip octets received.
     *  Fragments are not included as individual packets.
     */
    uint64_t ipSctpInOctets;
    /*! Total number of SCTP protocol ip packets transmitted.
     *  Fragmented packets are not included as individual packets.
     */
    uint64_t ipSctpOutPackets;
    /*! Total number of SCTP protocol ip octets transmitted.
     *  Fragmented packets are not included as individual packets.
     */
    uint64_t ipSctpOutOctets;
}adk_ip_l4_stats_t;

/*!
 * =============================================================
 * @enum adk_ip_qos_profile_type_t
 *
 * @brief
 *
 * IP Qos profile type. The profile type determines whether QoS
 * profile is DIFFSERV based or user defined.
 *
 * =============================================================
 */
typedef enum adk_ip_qos_profile_type_e {
    /*! DiffServ QoS profile */
    ADK_IP_QOS_PROFILE_DIFFSERV = 1,
    /*! User defined QoS profile */
    ADK_IP_QOS_PROFILE_USER_DEFINED
} adk_ip_qos_profile_type_t;


/*!
 * =============================================================
 * @struct adk_ip_qos_diffserv_class_selector_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Defines the DSCP class to Detailed description starts here
 *
 * =============================================================
 */
typedef struct adk_ip_qos_diffserv_class_selector_s {
    /*!
     * dscp
     * DSCP bits from IP packet header or based on mapping
     * policy configured by the user for this type of packet.
     */
    uint8_t dscp;
    /*!
     * classType
     * DSCP class type for the above dscp value. Possible values
     * are ADK_QOS_DS_PHB_EF, ADK_QOS_DS_PHB_AF and ADK_QOS_DS_PHB_BE.
     */
    adk_qos_ds_phb_type_t classType;
    /*!
     * classId
     * DSCP class Id. The value should be 1 for EF/BE classes and 1-4
     * for AF class.
     */
    uint8_t classId;
} adk_ip_qos_diffserv_class_selector_t;


/*!
 * =============================================================
 * @struct adk_ip_qos_diffserv_info_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Diffserv specific information for setting up IP Qos. User has
 * to make sure that a DSCP profile is already created using
 * adk_qos_ip_dscp_policy_create API. The profile handle returned
 * from this API is an input to this data structure.
 *
 * =============================================================
 */
typedef struct adk_ip_qos_diffserv_info_s {
    /*!
     * DSCP profile handle returned from adk_qos_ip_dscp_policy_create
     * API. The user should create DSCP profile using apporiate
     * class and rates.
     */
    adk_handle_t dscpProfileHandle;
    /*!
     * numClassSelectors
     * This defines the number of class selectors that the
     * user would like to configure in this request.
     */
    uint8_t numClassSelectors;
    /*!
     * Array of data structure of type adk_ip_diffserv_class_selector.
     * Size of the array here depends on how many different class
     * mappings that user wanted to establish. Multiple Diffserv
     * code point values can be mapped to the same PHB or DSCP class.
     * E.g. DSCP 46 -> EF, DSCP 12 -> BE, DSCP 0 -> BE etc. The user
     * is expected to allocate memory for this data structure plus
     * (numClassSelector -1) * sizeof (adk_ip_diffserv_class_selector_t)
     * to hold information about all class selectors.
     */
    adk_ip_qos_diffserv_class_selector_t classSelector;
} adk_ip_qos_diffserv_info_t;


/*!
 * =============================================================
 * @struct adk_ip_qos_queue_selector_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Mapping from the DSCP value to MTM queue.
 *
 * =============================================================
 */
typedef struct adk_ip_qos_queue_selector_s {
    /*!
     * 6-bit DSCP bit value for the packet. This value
     * can come from the packet header or from the mapping
     * policies that the user set up for this type of packet.
     */
    uint8_t dscp;
    /*!
     * MTM Queue handle for the packets with above 6-bit DSCP value.
     * User is expected to program the queue and associated
     * scheduler (if applicable) before calling this API.
     */
    adk_handle_t queueHandle;
} adk_ip_qos_queue_selector_t;


/*!
 * =============================================================
 * @struct adk_ip_qos_user_profile_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * User defined profile for setting  up the appropriate traffic
 * shaping queue based on the nominal value of 6-bit DSCP value
 * that is assigned to the packet (The 6-bit DSCP value can come
 * from the packet or from user specified mapping policies).
 *
 * =============================================================
 */
typedef struct adk_ip_qos_user_profile_s {
    /*!
     * numQueueSelectors
     * number of Traffic shaping Queue selectors that user would
     * like to configure in this call.
     */
    uint8_t numQueueSelectors;

    /*!
     * queueSelector
     * Array of data structure of type adk_ip_qos_queue_selctor_t.
     * Size of the array is equal to numQueueSelectors. The caller
     * of the API should allocate the memory for structure
     * adk_ip_qos_user_profile_t plus ((numQueueSelectors - 1) *
     * sizeof(adk_ip_qos_queue_selector_t)).
     */
    adk_ip_qos_queue_selector_t queueSelector;
} adk_ip_qos_user_profile_t;

/*!
 * =============================================================
 * @enum adk_ip_ing_classifier_rulemask_e
 *
 * @brief
 *
 * IP ingress classifier rule mask definitions. This definition should be used to
 * set rulesMask field in adk_ip_ing_classifier_rulemask_e data structure.
 *
 * =============================================================
 */
typedef enum {

    /*! IP protocol type mask */
    ADK_IP_ING_CLASSIFIER_PROTO_MASK     = 0x01,
    /*! Destination address mask */
    ADK_IP_ING_CLASSIFIER_DEST_ADDR_MASK = 0x02,
    /*! Source address mask */
    ADK_IP_ING_CLASSIFIER_SRC_ADDR_MASK  = 0x04,
    /*! Interface Handle mask */
    ADK_IP_ING_CLASSIFIER_INTF_HANDLE_MASK = 0x08,
    /*! Source Port mask */
    ADK_IP_ING_CLASSIFIER_SRC_PORT_MASK = 0x10,
    /*! Destination Port */
    ADK_IP_ING_CLASSIFIER_DEST_PORT_MASK  = 0x20,
    /*! ICMP Type MASK */
    ADK_IP_ING_CLASSIFIER_ICMP_TYPE_MASK  = 0x40,
    /*! ICMP Code MASK */
    ADK_IP_ING_CLASSIFIER_ICMP_CODE_MASK  = 0x80,

} adk_ip_ing_classifier_rulemask_e;
/*!
 * =============================================================
 * @struct adk_ip_l4_port_info_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Defines L4 port info. Applicable for UDP/TCP/SCTP specific fields
 *
 * =============================================================
 */
typedef struct adk_ip_l4_port_info_s {
    /*! source port number */
    adk_ip_range_t srcPort;
    /*! destination port number */
    adk_ip_range_t destPort;
} adk_ip_l4_port_info_t;

/*!
 * =============================================================
 * @struct adk_ip_icmp_filter_info_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * Defines ICMP specific fields in the filter rule.
 *
 * =============================================================
 */
typedef struct adk_icmp_filter_info_s {
    /*! ICMP packet message type */
    adk_ip_range_t type;
    /*! ICMP packet message code */
    adk_ip_range_t code;
} adk_ip_icmp_filter_info_t;

typedef adk_ip_l4_port_info_t adk_ip_udp_port_info_t;
typedef adk_ip_l4_port_info_t adk_ip_tcp_port_info_t;
typedef adk_ip_l4_port_info_t adk_ip_sctp_port_info_t;

/*!
* =============================================================
* @struct adk_ip_ingress_classifier_rule_t adk_ip.h "adk_ip.h"
*
* @brief
*
* This structure contains parameters that defines a ip ingress
* classifier rule. Ingress Classifier is an ordered tree.
*
* =============================================================
*/
typedef struct {
    /*!
     * Filter rule mask identifies the valid parameters in this
     * data structure.
     */
    adk_ip_ing_classifier_rulemask_e ruleMask;
    /*!
     * Identifies IP protocol type.
     * Additional L4 fields can be specified only for following protocols (using protoInfo_u member):
     *  1. TCP
     *  2. UDP
     *  3. ICMP
     *  4. SCTP
     */
    uint8_t protocol;
    /*!
     * destination IP address prefix. This field can be a fully specified
     * address too. This field is matched against the destination IP address
     * in IP header.
     */
    adk_ip_addr_prefix_t destAddr;
    /*!
     * source IP address prefix. This field can be a fully specified
     * address too. This field is matched against the source IP address
     * in IP header.
     */
    adk_ip_addr_prefix_t srcAddr;
    /*!
     * IP interface handle
     */
    adk_interface_handle_t ipIntfHandle;
    /*!
     * protocol specific filter info. Only protocols defined in adk_ip_protocol_t
     * can have one of the associated proto info below.
     */
    union {
        /*! icmp specific filter fields */
        adk_ip_icmp_filter_info_t icmp;
        /*! tcp specific filter fields */
        adk_ip_tcp_port_info_t tcp;
        /*! udp specific filter fields */
        adk_ip_udp_port_info_t udp;
        /*! sctp specific filter fields */
        adk_ip_sctp_port_info_t sctp;
    } protoInfo_u;

} adk_ip_ingress_classifier_rule_t;



/*!
 * =============================================================
 * @enum adk_ip_ingress_classifier_action_type_e
 *
 * @brief
 *
 * IP Ingress Classifier action types applicable
 * when specifing adk_ip_ingress_classifier_action_t
 *
 * =============================================================
 */
typedef enum {
    /*! Drop this packet immediately. A global stats counter will
     * be incremented by default.
     */
    ADK_IP_ING_CLASSIFIER_ACTION_DROP = 1,
    /*! Forward to a program running on host processor */
    ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_HOST = 2,
    /*! Forward to a EVENT OBJECT running in EVENT MACHINE */
    ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_EO = 3,
} adk_ip_ingress_classifier_action_type_e;
/*!
 * =============================================================
 * @struct adk_ip_ingress_classifier_action_t adk_ip.h "adk_ip.h"
 *
 * @brief
 *
 * IP Ingress classifier action. This action (set of actions) is applied to
 * the packet when the packet fields match the configured ingress classifier
 * tree rule.
 *
 * =============================================================
 */
typedef struct {
    /*! action->  adk_ip_ingress_classifier_action_type_e */
    adk_ip_ingress_classifier_action_type_e action;
    /*! 1 indicates policing enabled and 0 indicates policing disabled */
    bool enablePolicing;
    /*!
     * policingHandle, This field determines the policing profile
     * handle. Relevant only if enablePolicing flag is set to
     * set to 1. Application must create this policing flow handle using
     * adk_qos_policing_flow_create API
     */
    adk_flow_handle_t policingHandle;
    union
    {
        /*!
         * Applicable for ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_EO */
        uint32_t  eventMachineQueueId;
      
    } actionAttr_u;
} adk_ip_ingress_classifier_action_t;

typedef adk_qos_ingress_info_t adk_ip_ingress_qos_info_t;

/*!
* =============================================================
* @struct adk_ip_qos_classifier_rule_t
*
* @brief
*
* This data structure contains IP Egress QoS lookup parameters
*
* =============================================================
*/
typedef struct adk_ip_qos_classifier_rule_s {
    /*!
     * QoS classifier rule mask. It identifies the valid
     * parameters in this data structure.
     */
    adk_qos_classifier_field_mask_t ruleMask;

   /*!
    * source IP address. This field can be IPv4 or IPv6 address.
    * This field can be a fully specified address or prefix.
    * This field is matched against the source IP address
    * in IPv4 or IPv6 header.
    */
    adk_ip_address_t srcAddr;

   /*!
    * destination IP address. This field can be IPv4 or IPv6 address.
    * This field can be a fully specified address or prefix.
    * This field is matched against the destination IP address
    * in IPv4 or IPv6 header.
    */
    adk_ip_address_t destAddr;
    /*!
     * 8-bit IP protocol type field.
     */
    uint8_t protocol;

    /*!
     * DSCP value for the packet. This value
     * can come from the packet header or from the mapping
     * policies that the user set up for this type of packet.
     * Note: It is expected in lower 6 bits.
     */
    uint8_t dscp;

} adk_ip_qos_classifier_rule_t;   /* ------  end of struct adk_ip_qos_classifier_rule_t  ------ */

/*!
* =============================================================
* @struct adk_ip_qos_classifier_action_t
*
* @brief
*
* This data structure contains IP Egress QoS action parameters
*
* =============================================================
*/
typedef struct adk_ip_qos_classifier_action_s {
    /*!
     * queueId for the packets.
     * User is expected to program the queue before calling this API.
     */
    adk_qos_queue_id_t queueId;
    /*!
     * Packet color for the packets matching the rule.
     * If queue CMAN type is not WRED then this value is ignored. 
     */
     enum rte_tm_color pktColor;
} adk_ip_qos_classifier_action_t;   /* ------  end of struct adk_ip_qos_classifier_action_t  ------ */

/*!
* =============================================================
* @struct adk_ip_qos_info_t
*
* @brief
*
* This data structure contains IP Egress QoS related parameters
*
* =============================================================
*/
typedef struct adk_ip_qos_info_s {
    /*QoS Classifier rule parameters*/
    adk_ip_qos_classifier_rule_t qosRule;
    /*QoS Classifier action parameters*/
    adk_ip_qos_classifier_action_t qosAction;
} adk_ip_qos_info_t;   /* ------  end of struct adk_ip_qos_info_t  ------ */

/*!
* =============================================================
* @struct adk_ip_ingress_classifier_info_t 
*
* @brief
*
* This data structure is polulated by adk_ip_get_ingress_classifier_all
* API. 
* =============================================================
*/

typedef struct adk_ip_ingress_classifier_info_s {
    /*! IP ingress classifier rule */
    adk_ip_ingress_classifier_rule_t ClassifierRule;
    /*! IP ingress classifier action */
    adk_ip_ingress_classifier_action_t ClassifierAction;
} adk_ip_ingress_classifier_info_t;

/*!
 * @addtogroup _ipFunctions  IPv4 Functions
 * @{
 */

adk_status_t adk_ip_init(
    adk_module_handle_t);

adk_status_t adk_ip_service_function_update(
    adk_module_handle_t);

adk_status_t adk_ip_log_level_set(
    adk_module_handle_t ipHandle,
    adk_log_level_t log_level);

adk_status_t adk_ip_log_enable(
    adk_module_handle_t ipHandle,
    adk_log_level_t log_level);

adk_status_t adk_ip_log_disable(
    adk_module_handle_t ipHandle,
    adk_log_level_t log_level);

adk_status_t adk_ip_interface_create(
    adk_module_handle_t ipHandle,
    const adk_ip_interface_t  *ipIf_p,
    adk_interface_handle_t *ipIfHandle);

adk_status_t adk_ip_interface_delete(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle);

adk_status_t adk_ip_interface_addr_add(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    const adk_ip_addr_prefix_t *ipAddrPrefix);

adk_status_t adk_ip_interface_addr_rem(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    const adk_ip_addr_prefix_t *ipAddrPrefix);

adk_status_t adk_ip_virtual_interface_create(
    adk_module_handle_t ipHandle,
    const adk_ip_virtual_interface_t  *ipVIf_p,
    adk_interface_handle_t *ipVIfHandle);

adk_status_t adk_ip_virtual_interface_delete(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipVIfHandle);

adk_status_t adk_ip_interface_priority_map_profile_config(
    adk_module_handle_t	ipHandle,
    adk_interface_handle_t ipIfHandle,
    adk_ip_direction_t direction,
    uint32_t profile_id);

adk_status_t adk_ip_exception_handler_register (
    adk_module_handle_t ipHandle,
    const adk_ip_exception_handler_t *exceptionHandler_p);

adk_status_t adk_ip_exception_handler_unregister (
    adk_module_handle_t ipHandle,
    adk_ip_exception_type_t type);

adk_status_t adk_ip_route_add (
    adk_module_handle_t ipHandle,
    const adk_ip_route_info_t *ipRoute_p);

adk_status_t adk_ip_route_delete (
    adk_module_handle_t ipHandle,
    const adk_ip_route_info_t *ipRoute_p);

adk_status_t
adk_ip_pbr_configure_default_route_table (
        adk_module_handle_t ipHandle,
        uint8_t  defRtTbl);

adk_status_t
adk_ip_pbr_add_policy (
        adk_module_handle_t ipHandle,
        uint32_t ruleIndex,
        adk_ip_pbr_info_t *pbrInfo_p);

adk_status_t
adk_ip_pbr_del_policy (
        adk_module_handle_t ipHandle,
    uint32_t  ruleIndex);

adk_status_t adk_ip_pbr_commit (
    adk_module_handle_t ipHandle);

adk_status_t adk_ip_qos_policy_add (
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint32_t ruleIndex,
    const adk_ip_qos_info_t *qosInfo_p);

adk_status_t adk_ip_qos_policy_delete (
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint32_t ruleIndex);

adk_status_t adk_ip_qos_policy_build (
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle);

adk_status_t adk_ip_interface_mtu_set(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint16_t mtu);

adk_status_t adk_ip_global_stats_get (
    adk_module_handle_t ipHandle,
    adk_ip_global_stats_t *stats_p);

adk_status_t adk_ip_interface_stats_get (
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    adk_ip_interface_stats_t *stats_p);


adk_status_t adk_ip_address_stats_get(
        adk_module_handle_t ipHandle,
        adk_ip_addr_t ipAddress,
        adk_ip_address_stats_t *stats_p);

adk_status_t adk_ip_ingress_classifier_stats_get (
    adk_module_handle_t ipHandle,
	const uint16_t ruleIndex,
    adk_ip_ingress_classifier_stats_t *ipStats_p);

adk_status_t adk_ip_dscp_pcp_map_add(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint8_t dscp,
    uint8_t pcp);

adk_status_t adk_ip_dscp_pcp_map_delete(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint8_t dscp);

adk_status_t adk_ip_dp_init(void);

adk_status_t adk_ip_ingress_classifier_rule_add(adk_module_handle_t ipHandle,
        const adk_ip_ingress_classifier_rule_t *ipIngClassifierRule_p,
        const adk_ip_ingress_classifier_action_t *ipIngClassifierAction_p,
        const uint16_t ruleIndex);

adk_status_t adk_ip_ingress_classifier_rule_delete(adk_module_handle_t ipHandle,
        const uint16_t ruleIndex);

adk_status_t adk_ip_ingress_classifier_build(adk_module_handle_t ipHandle);

adk_status_t adk_ip_reassembly_timeout_config (
        adk_module_handle_t ipHandle,
        uint32_t timeval,
        uint8_t percentChangeInTimeout,
        uint32_t minTimeout);

adk_status_t adk_ip_configure_supported_protocol(
    adk_module_handle_t ipHandle,
       uint8_t protocol);

adk_status_t
adk_ip_remove_supported_protocol (
        adk_module_handle_t ipHandle,
        uint8_t protocol);

adk_status_t adk_ip_ingress_qos_policy_add (
        adk_module_handle_t ipHandle,
        adk_interface_handle_t ipIfHandle,
        adk_ip_ingress_qos_info_t *qosInfo_p);

adk_status_t adk_ip_ingress_qos_policy_delete (
        adk_module_handle_t ipHandle,
        adk_interface_handle_t ipIfHandle,
        adk_ip_ingress_qos_info_t *qosInfo_p);


adk_status_t adk_ip_sse_ip_addr_add(adk_module_handle_t ipHandle, adk_ip_addr_prefix_t ipAddrPrefix);

adk_status_t adk_ip_sse_ip_addr_del(adk_module_handle_t ipHandle, adk_ip_addr_prefix_t ipAddrPrefix);

adk_status_t adk_ip_get_pbr_policies_all( 
        adk_module_handle_t      ipHandle,
        uint32_t                 *numPbrPolicies,
        adk_ip_pbr_info_t        **ipPbrPol_p);

adk_status_t adk_ip_get_ingress_classifiers_all(
        adk_module_handle_t               ipHandle,
        uint32_t                          *numClassifiers,
        adk_ip_ingress_classifier_info_t  **ingressClassifier_p);

adk_status_t adk_ip_get_routes_all(
       adk_module_handle_t      ipHandle,
       uint8_t                  ipRouteTblId,
       uint32_t                 *numIpRoutes,
       adk_ip_route_info_t      **ipRoute_p);

adk_status_t adk_ip_get_qos_policies_all( 
        adk_module_handle_t          ipHandle,
        adk_interface_handle_t       ipIfHandle,
        uint32_t                     *numQosPolicies,
        adk_ip_qos_info_t            **ipQosPol_p);


adk_status_t adk_ip_l4_stats_get (
    adk_module_handle_t ipHandle,
    adk_ip_l4_stats_t *ipL4Stats_p);

#endif

/*! end addtogroup _ipFunctions
 * @}
 *
 * end addtogroup _ip
 * @}
 */
