/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

/*!
 * @file  adk_ip_utility.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipUtilityCounters
 * @{
 *
 * @brief
 *
 * This file contains ADK APIs to enable/disable IPv4 utility counters and get API to fetch the counter values.
 *
 ***********************************************************************/

#ifndef _ADK_IP_UTILITY_H_
#define _ADK_IP_UTILITY_H_

/*!
 * =============================================================
 * @internal
 * @struct adk_ip_utility_stats_t "adk_ip_utility.h"
 *
 * @brief
 *
 * Data structure for IP utility counters.
 * =============================================================
 */

typedef struct adk_ip_utility_stats_s
{
    /* Number of packets (fragments) received in LLI that needed to be sent for reassembly */
    uint64_t ipReasmReqPkts;

    /* Number of packets (fragments) queued/parked for reassembly */
    uint64_t ipReasmEnqueuedPkts;

    /* Number of packets received by IP module that originated from Host/NetD */
    uint64_t ipOutNetDOrigPkts;

    /* Number of packets received by IP module that originated from EO */
    uint64_t ipOutEOOrigPkts;

    /* Number of packets sent to Host after ingress processing */
    uint64_t ipInSentToHostPkts;

    /* Number of packets sent to EO after ingress processing */
    uint64_t ipInSentToEoPkts;

    /* Number of packets sent to IPSec during ingress processing.
     * Note: ALL packets will be sent to IPSec if ipsec module is NOT bypassed */
    uint64_t ipInSentToIPSecPkts;

    /* Number of packets sent to IPSec during egress processing.
     * Note: ALL packets will be sent to IPSec if ipsec module is NOT bypassed */
    uint64_t ipOutSentToIPSecPkts;

    /* Number of packets received from IPSec during egress processing. */
    uint64_t ipOutRcvdFromIPSecPkts;

    /* Number of packets sent to Ethernet module during egress processing. */
    uint64_t ipOutSentToEnetPkts;

    /* Number of timeout events for which reassembly context has already been cleared.
     * Possible Race condition. This event will be ignored. */
    uint32_t ipReasmTimerCbAfterCleanup;

    /* Reassembly processing indicated invalid action. This would indicate memory corruption */
    uint32_t ipReasmInvalidReasmAction;

    /* Number of packets received by IP module that originated in Host/NetD and has non-local
     * source address.
     */
    uint64_t ipOutNonLocalSipFromHostPkts;

    /* Number of UDP packets received in ingress */
    uint64_t ipInUDPPkts;

    /* Number of UDP octets received in ingress */
    uint64_t ipInUDPOctets;

    /* Number of UDP packets sent out in egress direction */
    uint64_t ipOutUDPPkts;

    /* Number of UDP octets sent out in egress direction */
    uint64_t ipOutUDPOctets;

    /* Number of fragments created (in plntxt as well as pre fragmentation) */
    uint64_t ipIpsecFragsCreated;

    /* Number of times fragmentation is requested (plntxt as well as pre fragmentation (before encryption) */
    uint64_t ipIpsecFragsRequested;

    /* IP LLI received packet with corrupted sendr imc field */
    uint64_t ipInUnknownSender;

    /* IP ULI received packet with corrupted sendr imc field */
    uint64_t ipOutUnknownSender;
}adk_ip_utility_stats_t;


/*!
 * =============================================================
 * @fn adk_ip_utility_stats_enable
 *
 * @param[in]    ipHandle - IP module handle.
 * @param[in]    state    - TRUE to enable utility counters
 *                        - FALSE to disable utility counters
 *
 * @retval ADK_SUCCESS - Successfully configuration
 * @retval ADK_E_INVALID_HANDLE - IP module handle is invalid.
 * @retval  ADK_E_INTERNAL_ERROR -  Internal error
 *
 * @brief  This API allows the application to enable/disable utility counters
 *
 * =============================================================
 */

adk_status_t adk_ip_utility_stats_enable(
    adk_module_handle_t ipHandle,
    bool state
);

/*!
 * =============================================================
 * @fn adk_ip_utility_stats_get
 *
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[out]   stats - pointer to adk_ip_utility_stats_t
 * data structure.
 *
 * @retval ADK_SUCCESS - Successfully populated utility counters
 * @retval ADK_E_INVALID_HANDLE - IP module handle is invalid.
 * @retval  ADK_E_INVALID_PARAMETER - Invalid parameter
 *
 * @brief
 *
 * This function reads the statistics from data path
 * and fill the adk_ip_utility_stats_t pointer with appropriate values.
 * It is expected that 'stats' points to valid memory location and has
 * enough memory (sizeof(adk_ip_utility_stats_t))
 *
 * =============================================================
 */
adk_status_t adk_ip_utility_stats_get(
    adk_module_handle_t ipHandle,
    adk_ip_utility_stats_t *stats_p
);

typedef enum
{
    ADK_IP_GLOBAL_STATS         = 1,
    ADK_IP_IFACE_STATS          = 2,
    ADK_IP_ADDR_STATS           = 4,
    ADK_IP_L4_STATS             = 8,
    ADK_IP_UTILITY_STATS        = 16,
    ADK_IP_ING_CLASSIFIER_STATS = 32,
    ADK_IP_STATS_ALL            = 63
}adk_ip_stats_type_t;

adk_status_t adk_ip_stats_clear(
        adk_module_handle_t ipHandle,
        adk_ip_stats_type_t stats_type
);

#endif /* _ADK_IP_UTILITY_H_ */
