/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains initialization function and all module level support
 * APIs required from the IP module e.g. for IPv4 logs, global statistics,
 * receiving or injecting PDU from CPU, exception configurations, etc.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include "adk_ph_dev_dp.h"
#include <adk_cksum.h>
#include <adk_timer.h>
#include "adk_ip_dp_internal.h"

adk_status_t adk_ip_ph_rx_parse(
        void *mbuf,
        void **dp_message);


/*!
 * =============================================================
 * adk_ip_init
 * @param[in]  ipHandle  - IP module handle. Return at the
 * time of IP module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully initialized IP module.
 *
 * @retval ADK_E_NS_RSRC_CREATE_FAILED - Failed to create user
 * define resource (interface and egress flow) with NCP
 * resource manager.
 *
 * @retval ADK_IP_E_INGRESS_STATE_ID_UNAVAILABLE - Failed to allocate
 * memory to store global statistics. The memory here comes from ingress
 * statistics pool.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid IP mdoule handle.
 *
 *
 * @brief
 *
 * This function initializeds the ADK IP module. The primary
 * responsibility of the function are
 *  - create the user defined resources consumed by IP module.
 *  - get the IDs of all objects and cache them for futur
 *    use.
 *  - memory for internal data structure for the IP module is
 *  allocated in NVM (optionally) and set in the main module
 *  handle.
 *
 * =============================================================
 */
adk_status_t
adk_ip_init (adk_module_handle_t ipHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_module_service_fn_t svcFn;
    ip_module_info_t *ipModuleInfo_p;
    adk_log_level_t sys_log_level;
    adk_ip_log_level_t ipv4_log_level;
    adk_sys_handle_t sys_handle;

    uint32_t i=0;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        return retValue;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    sys_log_level = ADK_MODULE_LOG_LEVEL_GET(ipHandle);
    retValue = adk_ipv4_map_sys_log(ipHandle, sys_log_level, &ipv4_log_level);
    if (retValue != ADK_SUCCESS)
        return retValue;

    ADK_MODULE_LOG_LEVEL_SET(ipHandle, ipv4_log_level);

    ADK_IPV4_LOG_INFO(ipHandle, "Initializing at log-level 0x%04x\n",
            ipv4_log_level);

    /*
     * We assume that module handle is validated at this point
     * and start using it.
     */
    ipModuleInfo_p = (ip_module_info_t *)
                             ADK_NVM_CALLOC (ADK_MODULE_MODULEID_GET(ipHandle),
                                     ADK_IP_MOD_INFO_OBJ, 1,
                                     sizeof(ip_module_info_t));
    if (!ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_MEM_UNAVAILABLE,
                "Failed to allocate memory\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto adk_ip_init_return;
    }
    else
    {
        ADK_MODULE_INFO_PTR_SET(ipHandle, ipModuleInfo_p);

        ipModuleInfo_p->cookie = ADK_IPV4_MODULE_COOKIE;

        ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
        svcFn.moduleTerminate = adk_ip_shutdown;
        svcFn.srcModuleBind = (module_bind_fn_t) NULL;
        svcFn.srcModuleUnbind = (module_bind_fn_t) NULL;
        svcFn.moduleGetImc = adk_ip_get_imc;
        svcFn.moduleLogSet = adk_ipv4_sys_log;
        svcFn.errorString = adk_ipv4_error_name;
        svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_IPV4_LLI_FN_ID;
        svcFn.cpuEntryFn_LLI[_DEST_FUNC_RING] = ADK_IPV4_LLI_RING_FN_ID;
        svcFn.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_IPV4_ULI_FN_ID;
        svcFn.cpuEntryFn_ULI[_DEST_FUNC_RING] = ADK_IPV4_ULI_RING_FN_ID;
        ipModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_IPV4_LLI_FN_ID;
        ipModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_RING] = ADK_IPV4_LLI_RING_FN_ID;
        ipModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_IPV4_ULI_FN_ID;
        ipModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_RING] = ADK_IPV4_ULI_RING_FN_ID;

        retValue = adk_module_service_function_register(sys_handle,
                ipHandle, &svcFn);
        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to register service function\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }
        /* Create rule id resource for ingress classifier otree.
         * Id 0 is reserved to indicate invalid rule
         * This is to be done before adk_ip_dp_table_init as it
         * internally allocates one id from this while creating 
         * DEFAULT DENY rule
         */
        ipModuleInfo_p->ipIngClassifierRuleIDHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_ING_CLASSIFIER_RULE_INFO_OBJ,
                        ADK_IP_MAX_INGRESS_CLASSIFIER_RULES+1);
        if(NULL == ipModuleInfo_p->ipIngClassifierRuleIDHandle)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate rm IP ipIngClassifierRuleIDHandle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }
        adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipIngClassifierRuleIDHandle,
                0);

        for (i = 0; i < ADK_DP_MAX_IP_INTERFACE; i++) {
            ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[i] =
                    adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                            ADK_IP_EGRESS_QOS_CLASSIFIER_INFO_OBJ,
                            ADK_IP_MAX_EGRESS_QOS_CLASSIFIER_RULES+1);
            if(NULL == ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle)
            {
                ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                        "Failed to allocate rm IP ipEgressQoSClassifierRuleIDHandle\n");
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
                goto adk_ip_init_return;
            }
        }

        ipModuleInfo_p->ipPbrPolicyRuleIdHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_PBR_RULE_OBJ,
                        ADK_IP_MAX_PBR_RULES+2); /* id:0 is reserved. and id:1 is used for default rule */
        if(NULL == ipModuleInfo_p->ipPbrPolicyRuleIdHandle)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate rm IP PbrPolicyID Handle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }
        adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipPbrPolicyRuleIdHandle,0);

        ipModuleInfo_p->ipVrdIDHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_VRD_OBJ,
                        /* ADK_IP_DEFAULT_VRD_ID */ 1);
        if(NULL == ipModuleInfo_p->ipVrdIDHandle)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate rm IP VrdID Handle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }
        adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipVrdIDHandle,
                /*ADK_IP_DEFAULT_VRD_ID*/1);

        ipModuleInfo_p->ipStatsHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_MAX_OBJ,
                        ADK_DP_MAX_IP_STATS+1);

        if(ipModuleInfo_p->ipStatsHandle == NULL )
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate IP global stats Handle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }


        adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipStatsHandle, 0);

        ipModuleInfo_p->ipAddrStatsHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_ADDR_OBJ,
                        ADK_IP_MAX_ADDRESSES+1);

        if(ipModuleInfo_p->ipAddrStatsHandle == NULL )
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate IP address stats Handle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }

        adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipAddrStatsHandle, 0);

        ipModuleInfo_p->ipIngFlowStatsHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_INGRESS_ASSOC_OBJ,
                        ADK_DP_MAX_IP_ING_ASSOC_FLOWS);

        /*Id 0 is reserved for invalid case.*/
        ipModuleInfo_p->ipRmEgFlowHandle =
                adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipHandle),
                        ADK_IP_DP_MAX_IP_EG_OBJ,
                        ADK_DP_MAX_IP_EG_FLOWS+1);
        if(NULL == ipModuleInfo_p->ipRmEgFlowHandle)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate rm IP EgFlow Handle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }
        adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipRmEgFlowHandle,
                0);

        retValue = adk_sif_pcap_id_alloc(sys_handle,
                &(ipModuleInfo_p->sif_pcap_id));
        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to allocate SIF pcap id\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }

        retValue = adk_sif_pcap_id_config(sys_handle,
                ipModuleInfo_p->sif_pcap_id,
                ADK_IPV4_PCAP_FN_ID);
        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to config SIF pcap id\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }
        retValue = adk_sif_module_egress_callback_register(sys_handle,
                ADK_IP_MODULE_ID,
                (adk_sif_module_egress_cb_fn_t)adk_ip_dp_egress_stats);
        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to register DP callback function \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }

        /* Register callback function for policer drop in send to host classifier action*/
        retValue = adk_sif_module_slowpath_callback_register(sys_handle,
                ADK_IP_MODULE_ID,
                (adk_sif_module_slowpath_cb_fn_t)adk_ip_dp_slowpath_policer_stats);
        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to register DP callback function \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ip_init_return;
        }


        adk_acl_info_t cpRouteInfo={0};
        cpRouteInfo.user = ADK_ACL_IPV4_ROUTE_TABLE;
        if(ADK_SUCCESS != adk_acl_create(&(ipModuleInfo_p->routeInfo),
                &cpRouteInfo))
        {
            retValue = ADK_E_MEM_UNAVAILABLE;
            ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                    "Could not allocate CP list for IP Routing table!\n");
            goto adk_ip_init_return;
        }

#ifdef ADK_HOS_OLD_IMPL
        adk_hos_pcap_attr_t pcap_attr;
        uint32_t vsi_id;
        retValue = adk_sif_pcap_vsi_id_get (sys_handle, &vsi_id);
        if(retValue == ADK_SUCCESS)
        {
            pcap_attr.vsi_id = vsi_id;
            pcap_attr.vsi_q = 0;
            pcap_attr.pcap_data = ipModuleInfo_p->sif_pcap_id;
            retValue = adk_hos_pcap_create(&pcap_attr,
                    &(ipModuleInfo_p->hos_pcap_id));
            if(retValue != ADK_SUCCESS)
            {
                ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                        "Failed to create HOS pcap mirror id\n");
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
                goto adk_ip_init_return;
            }
            ipModuleInfo_p->hw_pcap_enabled = true;
        }
        else
        {
            ADK_IPV4_LOG_WARNING(ipHandle, "Failed to get VSI Id for pcap\n");
            ipModuleInfo_p->hw_pcap_enabled = false;
        }
#else
        ADK_IPV4_LOG_WARNING(ipHandle, "HW assistance for pcap is disabled by HOS\n");
        ipModuleInfo_p->hw_pcap_enabled = false;
#endif

        ipModuleInfo_p->ipv4IfaceCount = 0;
        ipModuleInfo_p->numRoutesProgrammed = 0;
        ipModuleInfo_p->aggrIpv4AddrCount = 0;
        ipModuleInfo_p->aggrIpv4VirtAddrCount = 0;
        ipModuleInfo_p->numPbrPolicies = 0;

        /* initialize ip dataplane tables */
        retValue = adk_ip_dp_table_init(ipHandle);
        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                    "Failed to init IP DP tables\n");
            goto adk_ip_init_return;
        }


    }

    retValue = ADK_IP_SUCCESS;

    adk_ip_init_return:
    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_init  ----- */




/*!
 * =============================================================
 * adk_ip_ph_rx_parse adk_ip.h "adk_ip.h"
 *
 * @param[in]   task Pointer to rte_mbuf.
 *
 * @param[out]  dp_message   Pointer to structure having the decoded parameters
 *                            sent by eth data path code to application.
 *
 * @retval   ADK_SUCCESS - successfully decoded the parameters
 * handler
 *
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 *                                    passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function helps in interpreting the parameters present in the task
 * Parameters that ip module send along with the packet to the Application
 * via PH Dev.
 *
 * =============================================================
 */

adk_status_t
adk_ip_ph_rx_parse (
        void *task,
        void **dp_message)
{
    adk_status_t ret = ADK_SUCCESS;
    struct rte_mbuf *mbuf_p;
    adk_metadata_t *adk_md_ptr;
    adk_ph_dev_dp_metadata_t *phMeta_p;
    ip_ph_metadata_t dpException;
    adk_ip_recv_exception_params_t *cpException;
    if (!task || !dp_message)
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ph_ip_rx_parse_decode_exit;
    }

    mbuf_p = (struct rte_mbuf *)task;
    adk_md_ptr = ADK_METADATA_PTR(mbuf_p);
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(ADK_MD_GET_IMC_PARAM_START(mbuf_p));

    /* Copy the params locally as we are going to reuse the IMC space in rte mbuf and provide the data to user in adk_ip_recv_exception_params_t format */
    dpException = *((ip_ph_metadata_t *)&(phMeta_p->imc_param_start));

    cpException = (adk_ip_recv_exception_params_t *)&(phMeta_p->imc_param_start);

    cpException->exception_type = dpException.exception_type;
    cpException->mbuf = mbuf_p;
    cpException->ipOffset = ADK_MD_GET_IP_OFF(mbuf_p);

    switch(cpException->exception_type)
    {
    case ADK_IP_NO_EXCEPTION:
        break;
    case ADK_IP_OPTIONS_EXCEPTION:
        cpException->exception_data.options.ifaceHandle = ip_dp_tbl_info_p->ifaceHandles[dpException.interface_id];
        break;
    case ADK_IP_FRAGMENTATION_EXCEPTION:
        cpException->exception_data.frag.ifaceHandle = ip_dp_tbl_info_p->ifaceHandles[dpException.interface_id];
        break;
    case ADK_IP_NO_ROUTE_EXCEPTION:
        cpException->exception_data.noRoute.routeTableId = dpException.routeTableId;
        break;
    case ADK_IP_TTL_EXPIRED_EXCEPTION:
        cpException->exception_data.ttl.ifaceHandle = ip_dp_tbl_info_p->ifaceHandles[dpException.interface_id];
        break;
    case ADK_IP_UNSUPPORTED_PROTO_EXCEPTION:
        cpException->exception_data.proto.ifaceHandle = ip_dp_tbl_info_p->ifaceHandles[dpException.interface_id];
        break;
    case ADK_IP_FORWARDING_DISABLED_EXCEPTION:
        cpException->exception_data.fwdingDisabled.ifaceHandle = ip_dp_tbl_info_p->ifaceHandles[dpException.interface_id];
        cpException->exception_data.fwdingDisabled.dir = dpException.dir;
        break;
    case ADK_IP_INVALID_VERSION_EXCEPTION:
        cpException->exception_data.version.ifaceHandle = ip_dp_tbl_info_p->ifaceHandles[dpException.interface_id];
        break;
    case ADK_IP_EXCEPTION_MAX:
        break;
    }


    *dp_message = (adk_ip_recv_exception_params_t*)cpException;

_adk_ph_ip_rx_parse_decode_exit:

    ADK_RETURN(ret);
}




/*!
 * =============================================================
 * adk_ip_service_function_update
 *
 * @param[in]  ipHandle  - IP module handle. Return at the
 * time of IP module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully initialized IP module.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid IP mdoule handle.
 *
 * @brief
 *
 * This function registers the CPU based service functions with
 * SIF. This function is called by adk_module_handle_get
 * function.
 *
 * =============================================================
 */
adk_status_t
adk_ip_service_function_update (adk_module_handle_t ipHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_module_service_fn_t svcFn;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ipModuleInfo_p = (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipHandle);
    if (!ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid IP handle \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);

        goto adk_ip_service_function_update_return;
    }

    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleTerminate = adk_ip_shutdown;
    svcFn.srcModuleBind = (module_bind_fn_t) NULL;
    svcFn.srcModuleUnbind = (module_bind_fn_t) NULL;
    svcFn.moduleGetImc = adk_ip_get_imc;
    svcFn.moduleLogSet = adk_ipv4_sys_log;
    svcFn.errorString = adk_ipv4_error_name;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ipModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_DEFAULT];
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_RING] = ipModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_RING];
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ipModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_DEFAULT];
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_RING] = ipModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_RING];

    /*Register ip parser function in PH to Decode exception params*/
    retValue = adk_ph_dev_rx_parse_register((void*)adk_ip_ph_rx_parse,
            ADK_IP_MODULE_ID);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle,
                                   retValue, "Failed to register the ph_dev_rx_parse function for IP in CP\n");
    }
    retValue = adk_module_service_function_update (
                   sys_handle,
                   ipHandle,
                   &svcFn);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle,
                           retValue, "Failed to register the functions\n");
    }

    ADK_RETURN(retValue);

adk_ip_service_function_update_return:

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_service_function_update  ----- */


/*!
 * =============================================================
 * adk_ip_shutdown
 * @param[in]  ipHandle  - IP module handle. Return at the
 * time of IP module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully initialized IP module.
 *
 * @brief
 *
 * This function de-initializeds the ADK IP module. The primary
 * responsibility of the function are
 *  - release module level resource. The resources that are
 *  allocated by module's objects won't be cleared.
 *  - Free up memory allocated for module data structure.
 *
 * =============================================================
 */
adk_status_t
adk_ip_shutdown (
    adk_module_handle_t ipHandle
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ip_shutdown_return;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ipModuleInfo_p = (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipHandle);
    if (!ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid IP handle \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ip_shutdown_return;
    }

    retValue = adk_sif_module_egress_callback_unregister(sys_handle,
                ADK_IP_MODULE_ID);
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                "Failed to unregister DP callback function \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        goto adk_ip_shutdown_return;
    }

    retValue = adk_sif_module_slowpath_callback_unregister(sys_handle,
                ADK_IP_MODULE_ID);
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INTERNAL_ERROR,
                "Failed to unregister DP callback function \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        goto adk_ip_shutdown_return;
    }

    adk_rm_resource_destroy(ipModuleInfo_p->ipVrdIDHandle);
    adk_rm_resource_destroy(ipModuleInfo_p->ipStatsHandle);
    adk_rm_resource_destroy(ipModuleInfo_p->ipAddrStatsHandle);
    adk_rm_resource_destroy(ipModuleInfo_p->ipIngFlowStatsHandle);
    adk_rm_resource_destroy(ipModuleInfo_p->ipIngClassifierRuleIDHandle);
    adk_rm_resource_destroy(ipModuleInfo_p->ipPbrPolicyRuleIdHandle);

    adk_rm_resource_destroy(ipModuleInfo_p->ipRmEgFlowHandle);

    /* Free CP routing table */
    adk_acl_rule_t route;
    adk_acl_get_reset(ipModuleInfo_p->routeInfo);
    while(ADK_SUCCESS == adk_acl_next_rule_get(
    		ipModuleInfo_p->routeInfo, &route))
    {
        if(NULL != route.user_data)
        {
            ADK_NVM_FREE(route.user_data);
            route.user_data = NULL;
        }
    }

    adk_acl_free(ipModuleInfo_p->routeInfo);


    /* Free the DP tables */
    retValue = adk_ip_dp_table_shutdown(ipHandle);
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to free IP DP tables \n");
        goto adk_ip_shutdown_return;
    }

    /* Clear the internal cookie */
    ipModuleInfo_p->cookie = 0;

    ADK_NVM_FREE(ipModuleInfo_p);

adk_ip_shutdown_return:
    ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * adk_ip_destination_mtu_add
 *
 * @param[in] ipHandle          IP module handle
 *
 * @param[in] dipAddr           Destination IP address
 *
 * @param[in] mtu               MTU for the Destination IP address. The range of
 * values supported for MTU configuration is between 64 and 9000 bytes.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IP module handle
 * specified
 *
 * @retval ADK_E_INVALID_PARAMETER          Invalid input parameter
 *
 * @retval ADK_E_LEARN_FAILED               Failed to add the entry in the
 * DP table
 *
 * @brief
 *
 * This API adds an entry to the Dest IP address based MTU table. This table is
 * looked up only for originating traffic. Invoking this API with same
 * input(VrdId, DIP) will result in overwritting the old entry.
 *
 * Destination MTU gets precedence over interface mtu. Whenever destination mtu is configured and is non-zero,
 * it is used for fragmentation decision.
 *
 * Note: Add parameter 'uint32_t ipVrdId' to function signature and remove local variable ipVrdId to enable vrd feature
 * =============================================================
 */
adk_status_t adk_ip_destination_mtu_add(
    adk_module_handle_t ipHandle,
    adk_ip_addr_t dipAddr,
    uint32_t mtu)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dp_status_t dp_status = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    ip_module_info_t        *ipModuleInfo_p = NULL;
    uint32_t ipVrdId = ADK_IP_DEFAULT_VRD_ID;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_destination_mtu_add_exit;
    }

    /* Validate Routing Domain */
    if (ipVrdId >= ADK_IP_VRD_ID_MAX)
    {
    	retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
    	ADK_IPV4_LOG_FORCE("Invalid IPv4 routing domain Id %d\n",
    			ipVrdId);
    	goto _adk_ip_destination_mtu_add_exit;
    }

    /* Validate MTU */
    if ((mtu < IP_IF_MTU_MIN) || (mtu > IP_IF_MTU_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_HANDLE,
                           "Invalid MTU Value %d\n",
                           mtu);
        goto _adk_ip_destination_mtu_add_exit;
    }

    /* Add the entry */
    dp_status = adk_ip_dp_dest_mtu_add(ipHandle, ipVrdId, adk_htonl(dipAddr), mtu);
    if (dp_status != ADK_SUCCESS)
    {
        retValue = ADK_E_LEARN_FAILED;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Failed to add the rule in dp table\n");
        goto _adk_ip_destination_mtu_add_exit;
    }

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[ipVrdId])++;

_adk_ip_destination_mtu_add_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ip_destination_mtu_rem
 *
 * @param[in] ipHandle          IP module handle
 *
 * @param[in] dipAddr           Destination IP address
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IP module handle
 * specified
 *
 * @retval ADK_E_INVALID_PARAMETER          Invalid input parameter
 *
 * @retval ADK_E_UNLEARN_FAILED             Failed to remove the entry in the
 * DP table
 *
 * @brief
 *
 * This API removes an entry from the Dest IP address based MTU table.
 * This API is deprecated.
 *
 * Note: Add parameter 'uint32_t ipVrdId' to function signature and remove local variable ipVrdId to enable vrd feature
 * =============================================================
 */
adk_status_t adk_ip_destination_mtu_rem(
    adk_module_handle_t ipHandle,
    adk_ip_addr_t dipAddr)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dp_status_t dp_status = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    ip_module_info_t        *ipModuleInfo_p = NULL;
    uint32_t ipVrdId = ADK_IP_DEFAULT_VRD_ID;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_destination_mtu_rem_exit;
    }

    /* Validate Routing Domain */
    if (ipVrdId >= ADK_IP_VRD_ID_MAX)
    {
    	retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
    	ADK_IPV4_LOG_FORCE("Invalid IPv4 routing domain Id %d\n",
    			ipVrdId);
    	goto _adk_ip_destination_mtu_rem_exit;
    }

    /* Remove the entry */
    dp_status = adk_ip_dp_dest_mtu_rem(ipHandle, ipVrdId, adk_htonl(dipAddr));
    if (dp_status != ADK_SUCCESS)
    {
        retValue = ADK_E_LEARN_FAILED;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Failed to remove the rule from dp table\n");
        goto _adk_ip_destination_mtu_rem_exit;
    }

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[ipVrdId])++;

_adk_ip_destination_mtu_rem_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

static inline const char * exceptionEnumToStr(adk_ip_exception_type_t type)
{
    switch(type)
    {
    case ADK_IP_NO_EXCEPTION:
        return "ADK_IP_NO_EXCEPTION";
    case ADK_IP_OPTIONS_EXCEPTION:
        return "ADK_IP_OPTIONS_EXCEPTION";
    case ADK_IP_FRAGMENTATION_EXCEPTION:
        return "ADK_IP_FRAGMENTATION_EXCEPTION";
    case ADK_IP_NO_ROUTE_EXCEPTION:
        return "ADK_IP_NO_ROUTE_EXCEPTION";
    case ADK_IP_TTL_EXPIRED_EXCEPTION:
        return "ADK_IP_TTL_EXPIRED_EXCEPTION";
    case ADK_IP_UNSUPPORTED_PROTO_EXCEPTION:
        return "ADK_IP_UNSUPPORTED_PROTO_EXCEPTION";
    case ADK_IP_FORWARDING_DISABLED_EXCEPTION:
        return "ADK_IP_FORWARDING_DISABLED_EXCEPTION";
    case ADK_IP_INVALID_VERSION_EXCEPTION:
        return "ADK_IP_INVALID_VERSION_EXCEPTION";
    default:
        return "INVALID EXCEPTION";
    }
}

/*!
 * =============================================================
 * adk_ip_exception_handler_register
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[in]   exceptionHandler_p  Pointer to adk_ip_exception_handler_t.
 *
 * @retval   ADK_SUCCESS - successfully registerd exception
 * handler
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @retval ADK_IP_E_INVALID_EXCEPTION_TYPE  - Invalid exception type.
 *
 * @brief
 *
 * This function allows the user to register exception handlers
 * for various exceptions. enum adk_ip_exception_type_t defines
 * the type of exceptions that are supported by ADK.
 *
 * =============================================================
 */
adk_status_t
adk_ip_exception_handler_register (
    adk_module_handle_t ipHandle,
    const adk_ip_exception_handler_t  *exceptionHandler_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_dp_exception_handler_t *dpExceptionHdl_p=NULL,*temp;
    uint8_t type;
#ifdef ADK_QOS_GEN3_INTERFACE
    adk_qos_queue_id_info_t qid_info = {.policerId = UINT_MAX};
#endif
    uint32_t policerFlowId = ADK_QOS_INVALID_POLICER_FLOW_ID;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_exception_handler_register_exit;
    }

    if (!exceptionHandler_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "NULL pointer parameter\n");
        goto _adk_ip_exception_handler_register_exit;
    }


    switch(exceptionHandler_p->type)
    {
    case ADK_IP_OPTIONS_EXCEPTION:
    case ADK_IP_FRAGMENTATION_EXCEPTION:
    case ADK_IP_NO_ROUTE_EXCEPTION:
    case ADK_IP_TTL_EXPIRED_EXCEPTION:
    case ADK_IP_UNSUPPORTED_PROTO_EXCEPTION:
    case ADK_IP_FORWARDING_DISABLED_EXCEPTION:
    case ADK_IP_INVALID_VERSION_EXCEPTION:
        break;
    default:
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_EXCEPTION_TYPE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Exception type \n");
        goto _adk_ip_exception_handler_register_exit;
    }

    type = exceptionHandler_p->type;

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    dpExceptionHdl_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                   ADK_IP_EXCEP_OBJ, 1, sizeof(adk_ip_dp_exception_handler_t), 0);

    if (!dpExceptionHdl_p)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_exception_handler_register_exit;
    }

    /*change device and queue id, if host policer enabled*/
    if (exceptionHandler_p->policer_handle)
    {
        retValue = adk_qos_policing_flow_id_get(sys_handle, exceptionHandler_p->policer_handle, &policerFlowId);
        if (ADK_SUCCESS != retValue) {
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid policer handle\n");
            return retValue;
        }
        }
        //TODO update for PH path
    dpExceptionHdl_p->type = exceptionHandler_p->type;
    dpExceptionHdl_p->devId = exceptionHandler_p->devId;
    dpExceptionHdl_p->queueId = exceptionHandler_p->queueId;
    dpExceptionHdl_p->priority = exceptionHandler_p->priority;
    dpExceptionHdl_p->cbFnId = exceptionHandler_p->cbFnHandle;
    dpExceptionHdl_p->policerId = policerFlowId;

    if(ipModuleInfo_p->ip_dp_tbl_info_p->exceptionHandlerTble[type])
    {
        temp = ipModuleInfo_p->ip_dp_tbl_info_p->
               exceptionHandlerTble[type];
        ipModuleInfo_p->ip_dp_tbl_info_p->exceptionHandlerTble[type] =
                                                               dpExceptionHdl_p;

        adk_rcu_synchronize();
        ADK_NVM_FREE(temp);
    }
    else
    {
        ipModuleInfo_p->ip_dp_tbl_info_p->exceptionHandlerTble[type] =
                                                               dpExceptionHdl_p;
    }

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_exception_handler_register: Registered ip exception handler with config: exception=%s, devId=%d, queueId=%d, priority=%d, policer_handle=0x%lx, callback=0x%lx  ", exceptionEnumToStr(exceptionHandler_p->type),
            exceptionHandler_p->devId, exceptionHandler_p->queueId, exceptionHandler_p->priority, (uint64_t)exceptionHandler_p->policer_handle, (uint64_t)exceptionHandler_p->cbFnHandle);

_adk_ip_exception_handler_register_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_exception_handler_register  ----- */


/*!
 * =============================================================
 * adk_ip_exception_handler_unregister
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[in]   exceptionHandler_p  Pointer to adk_ip_exception_handler_t.
 *
 * @retval   ADK_SUCCESS - successfully unregisterd exception
 * handler
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @retval  ADK_E_LEARN_FAILED - Failed to add entry into DP
 *
 * @retval ADK_IP_E_INVALID_EXCEPTION_TYPE  - Invalid exception type.
 *
 * @brief
 *
 * This function allows the user to unregister exception handlers
 * for various exceptions. enum adk_ip_exception_type_t defines
 * the type of exceptions that are supported by ADK. Application
 * is allowed to provide a CPU function as
 * exception handler to unregister. For CPU based exception
 * functions, application
 * is expected to remove the PDU handler transport path (creation
 * of receiveGroupHandle->receiveQueue->vpBind->callbackRegister)
 * after calling this API.
 *
 * =============================================================
 */
adk_status_t
adk_ip_exception_handler_unregister (
    adk_module_handle_t ipHandle,
    adk_ip_exception_type_t type
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_dp_exception_handler_t *temp;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_exception_handler_unregister_exit;
    }

    switch(type)
    {
    case ADK_IP_OPTIONS_EXCEPTION:
    case ADK_IP_FRAGMENTATION_EXCEPTION:
    case ADK_IP_NO_ROUTE_EXCEPTION:
    case ADK_IP_TTL_EXPIRED_EXCEPTION:
    case ADK_IP_UNSUPPORTED_PROTO_EXCEPTION:
    case ADK_IP_FORWARDING_DISABLED_EXCEPTION:
    case ADK_IP_INVALID_VERSION_EXCEPTION:
        break;
    default:
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_EXCEPTION_TYPE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Exception type \n");
        goto _adk_ip_exception_handler_unregister_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if(ipModuleInfo_p->ip_dp_tbl_info_p->exceptionHandlerTble[type])
    {
        temp = ipModuleInfo_p->ip_dp_tbl_info_p->
               exceptionHandlerTble[type];
        ipModuleInfo_p->ip_dp_tbl_info_p->
                        exceptionHandlerTble[type] =NULL;

        adk_rcu_synchronize();
        ADK_NVM_FREE(temp);
    }
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Not a valid registered exception \n");
        goto _adk_ip_exception_handler_unregister_exit;
    }

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_exception_handler_unregister: Unregistered exception handler for exception : %s", exceptionEnumToStr(type));

_adk_ip_exception_handler_unregister_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_exception_handler_unregister  ----- */

/*!
 * =============================================================
 * adk_ip_tables_dump
 *
 * @param[in] ipHandle      IP module handle.
 *
 * @param[in] table_group   Group of tables to dump
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv4 module handle specified
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid table group type specified
 *
 * @brief
 *
 * This function dumps the specified group of tables.
 * =============================================================
 */
adk_status_t adk_ip_tables_dump(
    adk_module_handle_t  ipHandle,
    adk_ip_table_group_t table_group)
{

    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle 0x%"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle 0x%"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_tables_dump_exit;
    }

    /* Table group validation */
    if ((table_group <= ADK_IP_TABLES_NONE)
        || (table_group > ADK_IP_TABLES_ALL))
    {
        ADK_IPV4_LOG_FORCE("Invalid table group %d\n", table_group);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_tables_dump_exit;
    }

    retValue = adk_ipv4_dump_tables(ipHandle, table_group);
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(retValue);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to dump IP trees\n");
        goto _adk_ip_tables_dump_exit;
    }

_adk_ip_tables_dump_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:

    return(ADK_SUCCESS);
}

/*!
* @param[in] ipHandle        IP module handle
* @param[in] timeval         Reassembly timeout in milliseconds
* @param[in]   percentageChangeInTimeout - Percentage change in the timeout
*              value from previous stage as number of occupied contexts increase.
*              Acceptable values are integers between 1 & 100.
* @param[in] minTimeout      Minimum timeout to be used. This will be used for for stage-4.
*               If timeval is going below minTimeout in any of the stages, minTimeout will
*               be used for that stage.
*
* This function configures the global IPv4 reassembly timeout value.
* Timeout value greater than 2000 seconds is not supported.
* The default reassembly timeout value is 100 milliseconds. Minimum reassembly
* timeout value that can be configured is 100 milliseconds.
*
* @retval ADK_E_INVALID_PARAMETER  Invalid IP module handle or invalid defrag timer interval or invalid percentage change.
*
* @retval ADK_E_UNSUPPORTED Timeout value greater than 2000 seconds is not supported.
*
* The total number of reassembly contexts, is divided into four stages.
* Stage 1: supports timeout value timeval : for contexts occupancy of fist 1/3rd of total contexts
* Stage 2: supports timeout value as percentageChangeInTimeout of timeval: for contexts occupancy of next 1/6th of total contexts
* stage 3: supports timeout value as percentageChangeInTimeout of stage 2 timeout value: for contexts occupancy of next 1/6th of total contexts
* stage 4: supports timeout value minTimeVal : for contexts occupancy of remaining contexts
*
*/
adk_status_t adk_ip_reassembly_timeout_config (
        adk_module_handle_t ipHandle,
        uint32_t timeval,
        uint8_t percentChangeInTimeout,
        uint32_t minTimeout)
{
    ip_module_info_t *ipModuleInfo_p;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_reassembly_timeout_config_exit;
    }

    ipModuleInfo_p = (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipHandle);

    if(timeval > 2000000)
    {
        ADK_IPV4_LOG_FORCE("Invalid timeout value %d", timeval);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_reassembly_timeout_config_exit;
    }
    if(percentChangeInTimeout > 100)
    {
        ADK_IPV4_LOG_FORCE("Invalid percentChangeInTimeout value %d", percentChangeInTimeout);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_reassembly_timeout_config_exit;
    }
    if(minTimeout < 100)
    {
        ADK_IPV4_LOG_FORCE("Invalid minTimeout value %d", minTimeout);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_reassembly_timeout_config_exit;
    }

    ipModuleInfo_p->ip_dp_tbl_info_p->reasmTout = timeval;
    ipModuleInfo_p->ip_dp_tbl_info_p->reasmPrcntToutChange = percentChangeInTimeout;
    ipModuleInfo_p->ip_dp_tbl_info_p->reasmMinTout = minTimeout;

    ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage1 = ipModuleInfo_p->ip_dp_tbl_info_p->reasmTout;

    if ( ((ipModuleInfo_p->ip_dp_tbl_info_p->reasmTout * (100-ipModuleInfo_p->ip_dp_tbl_info_p->reasmPrcntToutChange))/100) > (ipModuleInfo_p->ip_dp_tbl_info_p->reasmMinTout) )
        ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage2 = ((ipModuleInfo_p->ip_dp_tbl_info_p->reasmTout * (100-ipModuleInfo_p->ip_dp_tbl_info_p->reasmPrcntToutChange))/100);
    else
        ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage2 = (ipModuleInfo_p->ip_dp_tbl_info_p->reasmMinTout);

    if ( ((ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage2 * (100-ipModuleInfo_p->ip_dp_tbl_info_p->reasmPrcntToutChange))/100) > (ipModuleInfo_p->ip_dp_tbl_info_p->reasmMinTout) )
        ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage3 = ((ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage2 * (100-ipModuleInfo_p->ip_dp_tbl_info_p->reasmPrcntToutChange))/100);
    else
        ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage3 = (ipModuleInfo_p->ip_dp_tbl_info_p->reasmMinTout);

    ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage4 = (ipModuleInfo_p->ip_dp_tbl_info_p->reasmMinTout);

    ADK_IPV4_LOG_INFO(ipHandle, "IPv4 reassembly config: Received: Tout=%d, %%Toutchange=%d, minTout=%d, \
            \n Configured: stage1Tout=%d, stage2Tout=%d, stage3Tout=%d,stage4Tout=%d\n", timeval, percentChangeInTimeout,\
            minTimeout, ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage1, ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage2,\
            ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage3, ipModuleInfo_p->ip_dp_tbl_info_p->reasmStage4);

_adk_ip_reassembly_timeout_config_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_reassembly_timeout_config----- */


/*!
* =========================================================================================================
* adk_ip_configure_supported_protocol
*
* @param[in]   ipHandle    IP module handle.
*
* @param[in] protocol  Packets with this protocol value will be allowed to be
*            processed if they are locally terminating packets.
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_IP_E_INVALID_PARAMETER  invalid IPv4 module handle
*
* @retval ADK_E_INTERNAL_ERROR - Failed to configure supported protocols.
*
* @brief
*
* This function configures the allowed protocols for locally
* terminating IPv4 packets.
*
* =========================================================================================================
*/
adk_status_t adk_ip_configure_supported_protocol(
    adk_module_handle_t ipHandle,
       uint8_t protocol)
{
    ip_module_info_t *ipModuleInfo_p;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ip_configure_supported_protocol_exit;
    }

    ipModuleInfo_p = (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipHandle);

    ipModuleInfo_p->ip_dp_tbl_info_p->allowedL4protos[protocol/(sizeof(ipModuleInfo_p->ip_dp_tbl_info_p->allowedL4protos[0])*8)] |= (1 << (protocol%(sizeof(ipModuleInfo_p->ip_dp_tbl_info_p->allowedL4protos[0])*8)));

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_configure_supported_protocol: Added protocol %d to supported proto list", protocol);

adk_ip_configure_supported_protocol_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_configure_supported_protocol----- */



/*!
* =========================================================================================================
* adk_ip_remove_supported_protocol
*
* @param[in] ipHandle IP module handle.
*
* @param[in] protocol  Packets with this protocol value will not allowed to be
*              processed if they are locally terminating packets.
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_IP_E_INVALID_PARAMETER  invalid IPv4 module handle
*
* @brief
*
* This function removes the protocol from  the allowed protocols list for locally
* terminating IPv4 packets.
*
* =========================================================================================================
*/
adk_status_t
adk_ip_remove_supported_protocol (
        adk_module_handle_t ipHandle,
        uint8_t protocol)
{
    ip_module_info_t *ipModuleInfo_p;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ip_remove_supported_protocol_exit;
    }

    ipModuleInfo_p = (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipHandle);

    ipModuleInfo_p->ip_dp_tbl_info_p->allowedL4protos[protocol/(sizeof(ipModuleInfo_p->ip_dp_tbl_info_p->allowedL4protos[0])*8)] &= ~(1 << (protocol%(sizeof(ipModuleInfo_p->ip_dp_tbl_info_p->allowedL4protos[0])*8)));

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_remove_supported_protocol: Removed protocol %d to supported proto list", protocol);
adk_ip_remove_supported_protocol_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_remove_supported_protocol----- */


/**
*=============================================================
* adk_ip_sse_ip_addr_add
* @param[in]  ipHandle - IP module handle.
*
* @param[in] ipAddr - pointer to the IP address and its
*                       prefix Length.
*
* @retval ADK_SUCCESS Successfully configured SSE address
*
* @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
* handle pointer might be NULL.
*
* @brief
*
* This function configures SSE address to be terminated locally. User should add
*  a classifier entry for this address
* =============================================================
*/

adk_status_t adk_ip_sse_ip_addr_add(adk_module_handle_t ipHandle, adk_ip_addr_prefix_t ipAddrPrefix)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    uint32_t tblId=0, idx=0;
    adk_ip_dp_route_info_t * dipResult_p = NULL;
    adk_dp_status_t dp_st = 0;
    bool hos_programmed=false;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    if( !ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                "Invalid input param: moduleInfo\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_sse_ip_addr_add_exit;
    }

    if (ipAddrPrefix.prefixWidth > 32)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_ADDR_PREFIX,
                "Invalid IP address prefix\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
        goto _adk_ip_sse_ip_addr_add_exit;
    }


    adk_hos_ip_term_tbl_rule_attr_t hos_ip_term_rule =
    {
        .precedence = -1,
        .pkt_type = ADK_HOS_L2_TERMINATING,
        .is_ipv6 = false,
        .ip_addr_mask[0] = adk_htonl(ip_bit_mask[ipAddrPrefix.prefixWidth]),
        .ip_addr[0] = adk_htonl(ipAddrPrefix.addr),
        .is_ip_term = true,
        .is_bcast = false,
        .is_mcast = false,
        .ip_addr_stats_id = ADK_HOS_DEFAULT_IP_STATS_ID,
    };

    for(idx=0;idx<sizeof(ipModuleInfo_p->sseAddrRuleIds)/sizeof(ipModuleInfo_p->sseAddrRuleIds[0]);idx++)
    {

        if(!ipModuleInfo_p->sseAddrRuleIds[idx].inUse)
            break;
        else if(ipModuleInfo_p->sseAddrRuleIds[idx].ip.addr>> (32-ipModuleInfo_p->sseAddrRuleIds[idx].ip.prefixWidth) ==
                ipAddrPrefix.addr>>(32-ipAddrPrefix.prefixWidth) )
        {
            ADK_IPV4_LOG_INFO(ipHandle, "ip-addr: 0x%x/%d aleady present\n", ipAddrPrefix.addr, ipAddrPrefix.prefixWidth);
            retValue = ADK_SUCCESS;
            goto _adk_ip_sse_ip_addr_add_exit;
        }
    }

    if(idx == sizeof(ipModuleInfo_p->sseAddrRuleIds)/sizeof(ipModuleInfo_p->sseAddrRuleIds[0]))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                         "Max limit reached for SSE addresses. Can't add more\n");
        goto _adk_ip_sse_ip_addr_add_exit;
    }

    // We have found a free slot at 'idx'. Now insert this address at 'idx' if successful

    retValue = adk_hos_rule_add(ADK_HOS_IP_TERM_TBL,
                                (adk_hos_rule_attr_t *) &hos_ip_term_rule,
                                &(ipModuleInfo_p->sseAddrRuleIds[idx].ruleId));
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                 "Failed to add IP-Term HOS rule for ip-intf address\n");
        goto _adk_ip_sse_ip_addr_add_exit;
    }
    hos_programmed = true;

    for(tblId=0; tblId < ADK_IP_MAX_ROUTING_TABLES; tblId++)
    {
        dipResult_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                ADK_IP_DIP_OBJ, 1, sizeof(adk_ip_dp_route_info_t));
        if (dipResult_p == NULL)
        {
            ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE, "Memory allocation failed\n");
            retValue = ADK_E_MEM_UNAVAILABLE;
            goto _adk_ip_sse_ip_addr_add_exit;
        }

        dipResult_p->isTerminating = TRUE;
        dipResult_p->u.termination.ip_addr_stats_id = ADK_HOS_DEFAULT_IP_STATS_ID;

        dp_st = adk_ip_dp_dip_add(ipHandle, tblId, adk_htonl(ipAddrPrefix.addr), ipAddrPrefix.prefixWidth,
                ADK_CONVERT_TO_LPM_DATA(dipResult_p));
        if(dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in adding dest ip dataplane\n");
            goto _adk_ip_sse_ip_addr_add_exit;
        }
    }

    ADK_IPV4_LOG_INFO(ipHandle, "Added ipv4 SSE address: 0x%x/%d with ruleId=%x\n", ipAddrPrefix.addr,
            ipAddrPrefix.prefixWidth, ipModuleInfo_p->sseAddrRuleIds[idx].ruleId);
    ipModuleInfo_p->sseAddrRuleIds[idx].inUse = true;
    ipModuleInfo_p->sseAddrRuleIds[idx].ip.addr = ipAddrPrefix.addr;
    ipModuleInfo_p->sseAddrRuleIds[idx].ip.prefixWidth = ipAddrPrefix.prefixWidth;

_adk_ip_sse_ip_addr_add_exit:
    if(ADK_SUCCESS != retValue)
    {
        if(true == hos_programmed)
        {
            adk_hos_rule_delete(ADK_HOS_IP_TERM_TBL,
                    ipModuleInfo_p->sseAddrRuleIds[idx].ruleId);
            ipModuleInfo_p->sseAddrRuleIds[idx].inUse = false;
        }
        for(idx=0;idx<tblId;idx++)
        {
            dp_st = adk_ip_dp_dip_remove(ipHandle, idx, adk_htonl(ipAddrPrefix.addr), ipAddrPrefix.prefixWidth);
        }
    }
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

    _lock_release_failed:
    _lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

} /* end of adk_ip_sse_ip_addr_add */


/**
*=============================================================
* adk_ip_sse_ip_addr_del
* @param[in]  ipHandle - IP module handle.
*
* @param[in] ipAddr - pointer to the IP address and its
*                       prefix Length.
*
* @retval ADK_SUCCESS Successfully removed route entry.
*
* @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
* handle pointer might be NULL.
*
* @brief
*
* This function removes the SSE addresses configured previously using adk_ip_sse_ip_addr_add() API
* =============================================================
*/

adk_status_t adk_ip_sse_ip_addr_del(adk_module_handle_t ipHandle, adk_ip_addr_prefix_t ipAddrPrefix)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    uint32_t tblId, idx;
    adk_dp_status_t dp_st;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    if( !ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                "Invalid input parameter\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_sse_ip_addr_del_exit;
    }

    if (ipAddrPrefix.prefixWidth > 32)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_ADDR_PREFIX,
                "Invalid IP address prefix\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
        goto _adk_ip_sse_ip_addr_del_exit;
    }

    for(idx=0;idx<sizeof(ipModuleInfo_p->sseAddrRuleIds)/sizeof(ipModuleInfo_p->sseAddrRuleIds[0]);idx++)
    {
        if(ipModuleInfo_p->sseAddrRuleIds[idx].ip.addr>> (32-ipModuleInfo_p->sseAddrRuleIds[idx].ip.prefixWidth) ==
                ipAddrPrefix.addr>>(32-ipAddrPrefix.prefixWidth) )
            break;
    }

    if(idx == sizeof(ipModuleInfo_p->sseAddrRuleIds)/sizeof(ipModuleInfo_p->sseAddrRuleIds[0]))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                "Couldn't find ip address\n");
        goto _adk_ip_sse_ip_addr_del_exit;
    }

    retValue = adk_hos_rule_delete(ADK_HOS_IP_TERM_TBL,
            ipModuleInfo_p->sseAddrRuleIds[idx].ruleId);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                "Failed to remove IP-Term HOS rule for ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_sse_ip_addr_del_exit;
    }
    ipModuleInfo_p->sseAddrRuleIds[idx].inUse = false;

    /* Delete from all tables */
    for(tblId=0; tblId < ADK_IP_MAX_ROUTING_TABLES; tblId++)
    {

        dp_st = adk_ip_dp_dip_remove(ipHandle, tblId, adk_htonl(ipAddrPrefix.addr), ipAddrPrefix.prefixWidth);
        if(dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in removing dest ip to dataplane\n");
            goto _adk_ip_sse_ip_addr_del_exit;
        }
    }
    ADK_IPV4_LOG_INFO(ipHandle, "Removed ipv4 SSE address: 0x%x/%d\n", ipAddrPrefix.addr, ipAddrPrefix.prefixWidth);

_adk_ip_sse_ip_addr_del_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

    _lock_release_failed:
    _lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
} /* end of adk_ip_sse_ip_addr_del */

/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
