/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <adk_include.h>
#include <adk_ip_acl.h>
#include <adk_ip_common.h>

#include "adk_dpdk.h"

/*
 * Rule and trace formats definitions.
 */

struct ipv4_7tuple {
        uint8_t  proto;
        uint32_t ip_src;
        uint32_t ip_dst;
        uint16_t port_src;
        uint16_t port_dst;
        uint16_t  dscp;
        uint16_t  frag_type;
}__attribute__((__packed__));

enum {
        PROTO_FIELD_IPV4,
        SRC_FIELD_IPV4,
        DST_FIELD_IPV4,
        SRCP_FIELD_IPV4,
        DSTP_FIELD_IPV4,
        DSCP_FIELD_IPV4,
        FRAG_FIELD_IPV4,
        NUM_FIELDS_IPV4
};

enum {
        PROTO_INPUT_IPV4,
        SRC_INPUT_IPV4,
        DST_INPUT_IPV4,
        L4PORTS_INPUT_IPV4,
        DSCP_FRAG_INPUT_IPV4
};

static struct rte_acl_field_def ipv4_defs[NUM_FIELDS_IPV4] = {
        {
                .type = RTE_ACL_FIELD_TYPE_RANGE,
                .size = sizeof(uint8_t),
                .field_index = PROTO_FIELD_IPV4,
                .input_index = PROTO_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, proto),
        },
        {
                .type = RTE_ACL_FIELD_TYPE_MASK,
                .size = sizeof(uint32_t),
                .field_index = SRC_FIELD_IPV4,
                .input_index = SRC_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, ip_src),
        },
        {
                .type = RTE_ACL_FIELD_TYPE_MASK,
                .size = sizeof(uint32_t),
                .field_index = DST_FIELD_IPV4,
                .input_index = DST_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, ip_dst),
        },
        {
                .type = RTE_ACL_FIELD_TYPE_RANGE,
                .size = sizeof(uint16_t),
                .field_index = SRCP_FIELD_IPV4,
                .input_index = L4PORTS_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, port_src),
        },
        {
                .type = RTE_ACL_FIELD_TYPE_RANGE,
                .size = sizeof(uint16_t),
                .field_index = DSTP_FIELD_IPV4,
                .input_index = L4PORTS_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, port_dst),
        },
        {
                .type = RTE_ACL_FIELD_TYPE_RANGE,
                .size = sizeof(uint16_t),
                .field_index = DSCP_FIELD_IPV4,
                .input_index = DSCP_FRAG_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, dscp),
        },
        {
                .type = RTE_ACL_FIELD_TYPE_RANGE,
                .size = sizeof(uint16_t),
                .field_index = FRAG_FIELD_IPV4,
                .input_index = DSCP_FRAG_INPUT_IPV4,
                .offset = offsetof(struct ipv4_7tuple, frag_type),
        },
};

RTE_ACL_RULE_DEF(acl_rule, NUM_FIELDS_IPV4);

adk_ip_acl_tree_t *adk_ip_acl_create(adk_module_handle_t ipHandle, uint16_t ruleSetId, uint8_t index)
{
    char name[20];
    struct rte_acl_ctx * ctx=NULL;
    struct rte_acl_param prm;

    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    snprintf(name, 20, "ADK_IP_ACL%3d_%1d", ruleSetId, index);
    prm.name = name;
    prm.socket_id = SOCKET_ID_ANY;
    prm.rule_size = RTE_ACL_RULE_SZ(NUM_FIELDS_IPV4);
    prm.max_rule_num = IP_ACL_MAX_RULES_PER_SET + 1;

    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_create", "creating new ACL context: %s\n", prm.name);
    ctx = rte_acl_create(&prm);
#if 0
    rte_acl_set_ctx_classify(ctx, RTE_ACL_CLASSIFY_SSE);
#endif

    return((adk_ip_acl_tree_t *)ctx);

}

int adk_ip_acl_ruleset_free(adk_module_handle_t ipHandle, adk_ip_acl_tree_t *acl_tree)
{
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_free", "de-allocating all memory used by ACL context\n");
    rte_acl_free((struct rte_acl_ctx *)acl_tree);
    return 1;
}

int adk_ip_acl_rule_add(adk_module_handle_t ipHandle,
		 	adk_ip_acl_tree_t *acl_tree,
		 	adk_ip_acl_rule_t *acl_rule,
		 	adk_ip_acl_val_t  acl_val,
		 	uint32_t precedence)
{
    int rc;
    struct acl_rule v;

    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_MEMSET(&v, 0, sizeof(v));
    v.field[SRC_FIELD_IPV4].value.u32 = acl_rule->sip_addr;
    v.field[SRC_FIELD_IPV4].mask_range.u32 = acl_rule->sip_mask;
    v.field[DST_FIELD_IPV4].value.u32 = acl_rule->dip_addr;
    v.field[DST_FIELD_IPV4].mask_range.u32 = acl_rule->dip_mask;
    v.field[SRCP_FIELD_IPV4].value.u16 = acl_rule->sport_begin;
    v.field[SRCP_FIELD_IPV4].mask_range.u16 = acl_rule->sport_end;
    v.field[DSTP_FIELD_IPV4].value.u16 = acl_rule->dport_begin;
    v.field[DSTP_FIELD_IPV4].mask_range.u16 = acl_rule->dport_end;
    v.field[PROTO_FIELD_IPV4].value.u8 = acl_rule->proto_begin;
    v.field[PROTO_FIELD_IPV4].mask_range.u8  = acl_rule->proto_end;
    v.field[DSCP_FIELD_IPV4].value.u16 = acl_rule->dscp_begin;
    v.field[DSCP_FIELD_IPV4].mask_range.u16  = acl_rule->dscp_end;
    v.field[FRAG_FIELD_IPV4].value.u16 = acl_rule->frag_type_begin;
    v.field[FRAG_FIELD_IPV4].mask_range.u16  = acl_rule->frag_type_end;

    v.data.category_mask = 1 /*LEN2MASK(RTE_ACL_MAX_CATEGORIES)*/;
    v.data.priority = precedence;
    v.data.userdata = (uint32_t) (size_t)acl_val;

    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_add_rules", "adding acl rules to ACL context\n");
    rc = rte_acl_add_rules((struct rte_acl_ctx *)acl_tree, (struct rte_acl_rule *)&v, 1);
    if (rc != 0) {
	return 0;
    }

    return 1;
}

int adk_ip_acl_ruleset_commit(adk_module_handle_t ipHandle, adk_ip_acl_tree_t *acl_tree)
{
    struct rte_acl_config cfg;
    int ret;

    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_MEMSET(&cfg, 0, sizeof(cfg));
    cfg.num_fields = NUM_FIELDS_IPV4;
    ADK_MEMCPY(&cfg.defs, ipv4_defs, sizeof(ipv4_defs));
    cfg.num_categories = 1;
    cfg.max_size = 0;

    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_build", "building ACL context\n");
    ret = rte_acl_build((struct rte_acl_ctx *)acl_tree, &cfg);
    if (ret != 0)
    {
        return 0;
    }
    return 1;

}

void adk_ip_acl_ruleset_clear(adk_module_handle_t ipHandle, adk_ip_acl_tree_t *acl_tree)
{
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    /* Remove all the rules. Does not clear the internal run time memory */
    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_reset_rules", "deleting all rules from ACL context\n");
    rte_acl_reset_rules((struct rte_acl_ctx *)acl_tree);
}
