/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_flow.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains the APIs to manage all kinds of IPv4 classification flows
 * e.g. IP ingress association flows, IP egress associations flows, IP egress
 * flows, etc. It also has APIs to get and print the statistics regarding these
 * flows.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include "adk_ethernet.h"
#include "adk_ethernet_export.h"
#include "adk_dpdk.h"

/*!
 * adk_ip_ingress_classifier_rule_handle_t
 * IP Ingress classifier flow handle. This handle is returned at the time of IP ingress
 * classifier rule creation
 */
typedef adk_flow_handle_t adk_ip_ingress_classifier_rule_handle_t;



void
adk_ip_lpm_rcu_distributed_check(adk_rm_hdl_t ipRmHandle,
         adk_ip_lpm_pending_list_t *ipDpPendingList_p, uint32_t flow_id)
{
    if((flow_id == ADK_RM_RSRC_INVALID)&& (ipDpPendingList_p->list_count != 0))
    {
        /* Wait for all workers to synchronize*/
        adk_rcu_distributed_wait();
        /*move all entries from pending list to free list*/
        adk_ip_lpm_clear_pending_list(ipRmHandle, ipDpPendingList_p);
    }
    else if(ipDpPendingList_p->list_count != 0)
    {
        if(adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
        {
            /*move all entries from pending list to free list*/
            adk_ip_lpm_clear_pending_list(ipRmHandle, ipDpPendingList_p);
        }
    }
}


void
adk_ip_rcu_distributed_check(adk_rm_hdl_t ipRmHandle,
         adk_ip_hash_pending_list_t *ipDpPendingList_p, uint32_t flow_id)
{
    if((flow_id == ADK_RM_RSRC_INVALID)&& (ipDpPendingList_p->count != 0))
    {
        /* Wait for all workers to synchronize*/
        adk_rcu_distributed_wait();
        /*move all entries from pending list to free list*/
        adk_ip_clear_pending_list(ipRmHandle, ipDpPendingList_p);
    }
    else if(ipDpPendingList_p->count != 0)
    {
        if(adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
        {
            /*move all entries from pending list to free list*/
            adk_ip_clear_pending_list(ipRmHandle, ipDpPendingList_p);
        }
    }
}

/*!
 * =============================================================
 * adk_ip_ingress_classifier_stats_get
 * @param[in]   ipHandle  - IP module handle.
 *
 * @param[in]   ruleIndex - Rule index to get stats from.
 *
 * @param[out]  ipStats_p    - pointer to adk_ip_ingress_classifier_stats_t
 * data structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER  invalid input parameter.
 *
 *
 * @brief
 *
 * This API read the IP classifier specific statistics
 * information from the state engine memory and copy them to
 * ipStats_p data structure.
 *
 * =============================================================
 */
adk_status_t
adk_ip_ingress_classifier_stats_get(
    adk_module_handle_t  ipHandle,
	const uint16_t ruleIndex,
    adk_ip_ingress_classifier_stats_t  *ipStats_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    ip_flow_info_t          *flow_info_p;
    adk_sys_handle_t        sys_handle;
    ip_module_info_t *ipModuleInfo_p;
	uint16_t index;
    adk_ip_ing_classifer_info_t* ingClassifierInfo_p = NULL;
    adk_ip_ingress_classifier_rule_handle_t classFlowHd=0;
    adk_acl_rule_t cpRule;
    size_t ingClassifRuleId = ADK_RM_RSRC_INVALID;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        return (retValue);
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ingClassifierInfo_p =  ((adk_ip_ing_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierInfo_p);

    adk_acl_get_reset(ingClassifierInfo_p->cpList);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
    		ingClassifierInfo_p->cpList, &cpRule))
    {
    	if(index==ruleIndex)
    	{
    		ingClassifRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);
    		classFlowHd = (adk_ip_ingress_classifier_rule_handle_t)cpRule.user_data;
    		break;
    	}
    	index++;
    }


    if (!ADK_SYS_VALID_FLOW_HANDLE(classFlowHd))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid Index.. %d\n", ruleIndex);
        goto _adk_ip_ingress_classifier_stats_get_exit;
    }

    flow_info_p = (ip_flow_info_t *)
        ADK_SYS_FLOW_ATTR_PTR_GET(classFlowHd);

    if ((flow_info_p->cookie != ADK_IPV4_FLOW_COOKIE) ||
        (flow_info_p->type != IP_FLOW_TYPE_INGRESS_CLASSIFIER_FLOW))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid rule index: %d\n", ruleIndex);
        goto _adk_ip_ingress_classifier_stats_get_exit;
    }

    retValue = adk_ip_dp_ingress_classifier_stats_get(ipHandle, ADK_SYS_FLOWID_GET(classFlowHd), ipStats_p);
    
_adk_ip_ingress_classifier_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * adk_ip_ingress_classifier_stats_print
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[in]   ipFlowHandle Ingress association flow handle.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER  invalid IP module handle or
 * flow handle specified.
 *
 * @retval ADK_E_STATS_READ_FAILURE  failed to read statistics
 * from the hardware.
 *
 * @brief
 *
 * This function dumps the IPv4 ingress flow statistics.
 *
 * =============================================================
 */
#if 0
adk_status_t adk_ip_ingress_classifier_stats_print (
    adk_module_handle_t ipHandle,
    adk_flow_handle_t ipFlowHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ip_ingress_classifier_stats_t stats_p;

    retValue = adk_ip_ingress_classifier_stats_get(ipHandle, ipFlowHandle,
               &stats_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_INFO_WRITE("Failed to get IP ingress flow statistics\n");
        return retValue;
    }

    ADK_INFO_WRITE("\n IP Ingress Flow %"ADKPRIh" Statistics\n", ipFlowHandle);
    ADK_INFO_WRITE("Total Rx Pkts       : %d\n", stats_p.rxPkts);
    ADK_INFO_WRITE("Total Tx Pkts       : %d\n", stats_p.txPkts);

    return ADK_SUCCESS;
}
#endif
/*************************************************** INGRESS ASSOCIATED ORDERED TREE START ***********************************************/

/***************** RTE ACL DEFINITION *******************
 */
enum {
    IP_ING_CLASSIFIER_PROTO_FIELD,
    IP_ING_CLASSIFIER_SRC_FIELD,
    IP_ING_CLASSIFIER_DST_FIELD,
    IP_ING_CLASSIFIER_SP_FIELD,
    IP_ING_CLASSIFIER_DP_FIELD,
    IP_ING_CLASSIFIER_INTF_FIELD_HDL,
    ADK_IP_ING_CLASSIFIER_NUM_FIELDS
};

enum {
    IP_ING_CLASSIFIER_PROTO_INPUT,
    IP_ING_CLASSIFIER_SRC_INPUT,
    IP_ING_CLASSIFIER_DST_INPUT,
    IP_ING_CLASSIFIER_L4PORTS_INPUT,
    IP_ING_CLASSIFIER_INTF_INPUT,
	IP_ING_CLASSIFIER_POST_IPSEC_INPUT
};


RTE_ACL_RULE_DEF(adk_ip_ingress_classifier_rule_rte, ADK_IP_ING_CLASSIFIER_NUM_FIELDS);

static  struct  rte_acl_field_def adk_ip_ing_classifier_def[ADK_IP_ING_CLASSIFIER_NUM_FIELDS]  = {
    {
        .type = RTE_ACL_FIELD_TYPE_RANGE,
        .size = sizeof(uint8_t),
        .field_index  = IP_ING_CLASSIFIER_PROTO_FIELD,
        .input_index  = IP_ING_CLASSIFIER_PROTO_INPUT,
        .offset = offsetof(adk_ip_ing_classifier_ptn_t, proto),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_ING_CLASSIFIER_SRC_FIELD,
        .input_index = IP_ING_CLASSIFIER_SRC_INPUT,
        .offset = offsetof(adk_ip_ing_classifier_ptn_t, sip),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_ING_CLASSIFIER_DST_FIELD,
        .input_index = IP_ING_CLASSIFIER_DST_INPUT,
        .offset = offsetof(adk_ip_ing_classifier_ptn_t, dip),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_RANGE,
        .size = sizeof(uint16_t),
        .field_index = IP_ING_CLASSIFIER_SP_FIELD,
        .input_index = IP_ING_CLASSIFIER_L4PORTS_INPUT,
        .offset = offsetof(adk_ip_ing_classifier_ptn_t, sPort),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_RANGE,
        .size = sizeof(uint16_t),
        .field_index = IP_ING_CLASSIFIER_DP_FIELD,
        .input_index = IP_ING_CLASSIFIER_L4PORTS_INPUT,
        .offset = offsetof(adk_ip_ing_classifier_ptn_t, dPort),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_RANGE,
        .size = sizeof(uint32_t),
        .field_index = IP_ING_CLASSIFIER_INTF_FIELD_HDL,
        .input_index = IP_ING_CLASSIFIER_INTF_INPUT,
        .offset = offsetof(adk_ip_ing_classifier_ptn_t, ipIntfHandle),
    },

};

/**
 *=============================================================
 * @internal
 * adk_ip_ingress_classifier_create
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[out] ingClassifierInfo_pp - Address of pointer to context which is 
               populated and returned by this API
 *
 * @retval ADK_SUCCESS Successfully created ingress classifer context.
 *
 * @retval ADK_E_MEM_UNAVAILABLE Failed to create as memory not available
 * @brief
 *
 * This function creates ingrees classifier ordered tree
 * This is called at initialization time where validation of params
 * is already done. Not input arg validation, No lock required.
 * =============================================================
 */
adk_status_t adk_ip_ingress_classifier_create(
        adk_module_handle_t ipHandle,
        adk_ip_ing_classifer_info_t  **ingClassifierInfo_pp)
{
    struct rte_acl_param prm;
    uint8_t   i;
    uint16_t  actionIndexDp;
    char name[50];
    adk_status_t    retValue;
    adk_sys_handle_t sys_handle;
    adk_acl_info_t cpListInfo={0};
    adk_ip_ingress_classifier_rule_t ipIngClassifierRule;
    adk_ip_ingress_classifier_action_t ipIngClassifierAction;
#define ACTIVE_TREE_INDEX_INIT 0
#define SHADOW_TREE_INDEX_INIT 1

    adk_ip_ing_classifer_info_t *ingClassifierInfo_p = NULL;

    ip_module_info_t * ipModuleInfo_p =
             (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ingClassifierInfo_p = (adk_ip_ing_classifer_info_t *)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_ING_CLASSIFIER_INFO_OBJ, 1,
			 sizeof(adk_ip_ing_classifer_info_t));

    if (ingClassifierInfo_p == NULL)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                           "Memory allocation failed\n");
        return ADK_E_MEM_UNAVAILABLE;
    }

    *ingClassifierInfo_pp = ingClassifierInfo_p;

    prm.socket_id = SOCKET_ID_ANY;
    prm.rule_size = RTE_ACL_RULE_SZ(ADK_IP_ING_CLASSIFIER_NUM_FIELDS);
    prm.max_rule_num = ADK_IP_MAX_INGRESS_CLASSIFIER_RULES + 1;


    actionIndexDp = adk_rm_resource_reserve(ipModuleInfo_p->ipIngClassifierRuleIDHandle);
    if (1 != actionIndexDp)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_RESOURCE_UNAVAILABLE,
                "Could not get resource id ING ORDERED CLASSIFIER using rte_acl_create!\n");
        return ADK_E_RESOURCE_UNAVAILABLE;
    }

    for (i=0; i<2; i ++)
    {
        /* Initialize */
        ingClassifierInfo_p->ipIngClassifierCtx[i] = NULL;
      
        snprintf(name, sizeof(name), "ADK_IP_INGRESS_ORDERED_CLASSIFIER_%1d",i);
        prm.name = name;
        ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_create",
                    "creating new Ing Ordered classifier context: %s\n", prm.name);

        ingClassifierInfo_p->ipIngClassifierCtx[i] =
            (adk_ip_ing_classifier_ctx_t*) rte_acl_create(&prm);
        if (NULL == ingClassifierInfo_p->ipIngClassifierCtx[i])
        {
            if (1 == i)
            {
                ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                        "Could not create ING ORDERED CLASSIFIER using rte_acl_create!\n");
                /* free previous one */
                rte_acl_free(ingClassifierInfo_p->ipIngClassifierCtx[0]);
            }
            return ADK_E_MEM_UNAVAILABLE;
        }
    }

    ingClassifierInfo_p->aclTreeIndex = ACTIVE_TREE_INDEX_INIT;
    cpListInfo.user = ADK_ACL_IPV4_ING_CLASSIFIER;
    if(ADK_SUCCESS != adk_acl_create(&(ingClassifierInfo_p->cpList),
                                        &cpListInfo))
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not allocate CP list for IP ING CLASSIFIER!\n");
        goto free_rte_acl_ctx;
    }

    /* 
     * Set first default Rule as ANY and action as DENY 
     * it goes with highest priority
     */
    /* Set Rule mask as zero, indicating no mask */
    ipIngClassifierRule.ruleMask = 0;
    memset(&ipIngClassifierAction, 0x0, sizeof(ipIngClassifierAction));
    ipIngClassifierAction.action = ADK_IP_ING_CLASSIFIER_ACTION_DROP;

    if(ADK_SUCCESS != classifier_add_rule_to_hw (
            ipHandle,
            ADK_IP_CLASSIFIER_RULE_MAX_PRIORITY,
            actionIndexDp,
            &ipIngClassifierRule,
            &ipIngClassifierAction,
			ACTIVE_TREE_INDEX_INIT
            ))
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not add default DENY rule to IP ING CLASSIFIER active tree!\n");
        goto free_rte_acl_ctx;
    }
    /* Build */
    if( ADK_SUCCESS != classifier_build_ordered_tree(ipHandle,
                ingClassifierInfo_p->ipIngClassifierCtx[ACTIVE_TREE_INDEX_INIT]))
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not BUILD default DENY rule to IP ING CLASSIFIER active tree!\n");
        goto free_rte_acl_ctx;
    }

    ipIngClassifierRule.ruleMask = 0;
    memset(&ipIngClassifierAction, 0x0, sizeof(ipIngClassifierAction));
    ipIngClassifierAction.action = ADK_IP_ING_CLASSIFIER_ACTION_DROP;

    if(ADK_SUCCESS != classifier_add_rule_to_hw (
    		ipHandle,
			ADK_IP_CLASSIFIER_RULE_MAX_PRIORITY,
			actionIndexDp,
			&ipIngClassifierRule,
			&ipIngClassifierAction,
			SHADOW_TREE_INDEX_INIT
    ))
    {
    	retValue = ADK_E_HW_PROGRAMMING_ERROR;
    	ADK_IPV4_LOG_FATAL(ipHandle, retValue,
    			"Could not add default DENY rule to IP ING CLASSIFIER shadow tree!\n");
    	goto free_rte_acl_ctx;
    }
    /* Build */
    if( ADK_SUCCESS != classifier_build_ordered_tree(ipHandle,
    		ingClassifierInfo_p->ipIngClassifierCtx[SHADOW_TREE_INDEX_INIT]))
    {
    	retValue = ADK_E_HW_PROGRAMMING_ERROR;
    	ADK_IPV4_LOG_FATAL(ipHandle, retValue,
    			"Could not BUILD default DENY rule to IP ING CLASSIFIER shadow tree!\n");
    	goto free_rte_acl_ctx;
    }
    return ADK_SUCCESS;

free_rte_acl_ctx:
    adk_rm_resource_free(ipModuleInfo_p->ipIngClassifierRuleIDHandle,
            actionIndexDp);

    if(ingClassifierInfo_p->ipIngClassifierCtx[0])
        rte_acl_free(ingClassifierInfo_p->ipIngClassifierCtx[0]);

    if(ingClassifierInfo_p->ipIngClassifierCtx[1])
        rte_acl_free(ingClassifierInfo_p->ipIngClassifierCtx[1]);

    if(ingClassifierInfo_p->cpList)
        adk_acl_free(ingClassifierInfo_p->cpList);

    return retValue;
} /* adk_ip_ingress_classifier_create */


void adk_ip_ingress_classifier_delete(
        adk_module_handle_t ipHandle,
        adk_ip_ing_classifer_info_t  **ingClassifierInfo_pp)
{
    /* TO DO: Delete CP list as well */
    ip_module_info_t * ipModuleInfo_p =
                (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    adk_ip_ing_classifer_info_t *ingClassifierInfo_p = *ingClassifierInfo_pp;

    if(ingClassifierInfo_p->aclTreeIndex)
    {
        adk_ip_ing_classifier_delete(ipHandle, ingClassifierInfo_p->ipIngClassifierCtx[0]);
        adk_rcu_synchronize();
        adk_ip_ing_classifier_delete(ipHandle, ingClassifierInfo_p->ipIngClassifierCtx[1]);
    }
    else
    {
        adk_ip_ing_classifier_delete(ipHandle, ingClassifierInfo_p->ipIngClassifierCtx[1]);
        adk_rcu_synchronize();
        adk_ip_ing_classifier_delete(ipHandle, ingClassifierInfo_p->ipIngClassifierCtx[0]);
    }

    adk_acl_free(ingClassifierInfo_p->cpList);
    ingClassifierInfo_p->cookie = 0;
    ADK_NVM_FREE(ingClassifierInfo_p);
    *ingClassifierInfo_pp = NULL;
    adk_rm_resource_free(ipModuleInfo_p->ipIngClassifierRuleIDHandle, 1);

}

/**
 *=============================================================
 * adk_ip_ing_classifier_delete
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[out] ctx - pointer to context to be deleted
 *
 * @retval ADK_SUCCESS Successfully deleted ingress classifer context.
 *
 * @brief
 *
 * This function deletes ingrees classifer ordered tree
 * =============================================================
 */
adk_status_t adk_ip_ing_classifier_delete(
        adk_module_handle_t ipHandle,
        adk_ip_ing_classifier_ctx_t *ctx)
{
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_free", "de-allocating all memory used by Ingress Association Ordered Tree context\n");
    rte_acl_free(ctx);
    return ADK_SUCCESS;
} /* adk_ip_ing_classifier_delete */
/*!
 * =============================================================
 * adk_ip_classifer_add_validate_arg
 * validates the input parameter for ingress Classifier rule add 
 * =============================================================
 */
adk_status_t adk_ip_classifer_add_validate_arg(
        adk_module_handle_t ipHandle,
        const adk_ip_ingress_classifier_rule_t *ipIngClassifierRule_p,
        const adk_ip_ingress_classifier_action_t *ipIngClassifierAction_p)
{
    adk_status_t    retValue;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    uint32_t    policerFlowId = 0;

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipIngClassifierRule_p || !ipIngClassifierAction_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid NULL input parameter\n");
        goto return_with_error;
    }

    if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_ADDR_MASK)
    {
        if((ipIngClassifierRule_p->destAddr.prefixWidth == 0) ||
                (ipIngClassifierRule_p->destAddr.prefixWidth > 32))
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid Destination IP address prefix width\n");
            goto return_with_error;
        }
    }

    if ((ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_SRC_ADDR_MASK) &&
            ((ipIngClassifierRule_p->srcAddr.prefixWidth == 0) ||
             (ipIngClassifierRule_p->srcAddr.prefixWidth > 32)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid Source IP address prefix width: %d\n", ipIngClassifierRule_p->srcAddr.prefixWidth);
        goto return_with_error;
    }

    if (ipIngClassifierAction_p->enablePolicing) {
        retValue = adk_qos_policing_flow_id_get(sys_handle, ipIngClassifierAction_p->policingHandle, &policerFlowId);
        if (ADK_SUCCESS != retValue) {
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid policer handle\n");
            return retValue;
        }
        (void)policerFlowId;
    }
    switch (ipIngClassifierAction_p->action)
    {
    case ADK_IP_ING_CLASSIFIER_ACTION_DROP:

    /*! Forward to a program running on host processor */
    case ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_HOST:

    /*! Forward to a EVENT OBJECT running in EVENT MACHINE */
    case ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_EO:
        break;
    default:
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid ACTION input parameter\n");
        goto return_with_error;
    }
    return ADK_SUCCESS;
return_with_error:
    return retValue;
} /* adk_ip_classifer_add_validate_arg */

/*!
 * =============================================================
 * @internal
 * classifier_add_rule_to_hw
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief
 *
 * Internal support function for ACL rule addition.
 *
 * =============================================================
 */
adk_status_t
classifier_add_rule_to_hw (
    adk_module_handle_t ipHandle,
    uint32_t precedence,
    uint32_t ingClassifierFlowId,
    const adk_ip_ingress_classifier_rule_t *ipIngClassifierRule_p,
    const adk_ip_ingress_classifier_action_t *ipIngClassifierAction_p,
    uint8_t treeIndex
    )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ip_module_info_t *ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    struct adk_ip_ingress_classifier_rule_rte v = {0};
    adk_ip_dp_ing_classifier_action_info_t *classifierResult_p;

    /*****   Build the Individual Patterns ******/

    /* Set the result before programming ID in dp
     * set all the fields as is -
     * control the access to field from DP.
     * for e.g. only netd path supports policing as of now,
     * no check required here - In DP it will access policingHandle only if
     * action is send to netd
     * Same principle applies for other fields
     */
    classifierResult_p = &(ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierActionDp_p[treeIndex][ingClassifierFlowId]);
    classifierResult_p->action = ipIngClassifierAction_p->action;
    classifierResult_p->policingEnabled = ipIngClassifierAction_p->enablePolicing;
    if (classifierResult_p->policingEnabled) {
        retValue = adk_qos_policing_flow_id_get(sys_handle, ipIngClassifierAction_p->policingHandle, &classifierResult_p->policerFlowId);
        if (ADK_SUCCESS != retValue) {
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid policer handle\n");
            return retValue;
        }
    }else {
        classifierResult_p->policerFlowId = ADK_QOS_INVALID_POLICER_FLOW_ID;
    }
    classifierResult_p->eventMachineQueueId =
        ipIngClassifierAction_p->actionAttr_u.eventMachineQueueId;
    ADK_IPV4_LOG_INFO(ipHandle,"\tIPv4 ingress classifier - building rule: FlowId=%d, treeIndex=%d\n", ingClassifierFlowId, treeIndex);
    /* Set protocol and corresponding port information */
    if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_PROTO_MASK)
    {
    	ADK_IPV4_LOG_INFO(ipHandle,"\t\tproto=%d\n", ipIngClassifierRule_p->protocol);
        /* Set RTE_ACL structure */
        v.field[IP_ING_CLASSIFIER_PROTO_FIELD].value.u8 =
            ipIngClassifierRule_p->protocol;
        v.field[IP_ING_CLASSIFIER_PROTO_FIELD].mask_range.u8  =
            ipIngClassifierRule_p->protocol;
        /* Set protocol specific port values */
        switch(ipIngClassifierRule_p->protocol)
        {
        case ADK_IP_PROTOCOL_TCP:
        case ADK_IP_PROTOCOL_UDP:
        case ADK_IP_PROTOCOL_SCTP:
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_SRC_PORT_MASK)
            {
            	ADK_IPV4_LOG_INFO(ipHandle,"\t\tsPort=%d-%d\n", ipIngClassifierRule_p->protoInfo_u.udp.srcPort.begin, ipIngClassifierRule_p->protoInfo_u.udp.srcPort.end);
                v.field[IP_ING_CLASSIFIER_SP_FIELD].value.u16 =
                    ipIngClassifierRule_p->protoInfo_u.udp.srcPort.begin;
                v.field[IP_ING_CLASSIFIER_SP_FIELD].mask_range.u16 =
                    ipIngClassifierRule_p->protoInfo_u.udp.srcPort.end;
            }
            else
            {
                v.field[IP_ING_CLASSIFIER_SP_FIELD].value.u16 = 0;
                v.field[IP_ING_CLASSIFIER_SP_FIELD].mask_range.u16 = 0xFFFF;
            }
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_PORT_MASK)
            {
            	ADK_IPV4_LOG_INFO(ipHandle,"\t\tdPort=%d-%d\n", ipIngClassifierRule_p->protoInfo_u.udp.destPort.begin, ipIngClassifierRule_p->protoInfo_u.udp.destPort.end);
                v.field[IP_ING_CLASSIFIER_DP_FIELD].value.u16 =
                    ipIngClassifierRule_p->protoInfo_u.udp.destPort.begin;
                v.field[IP_ING_CLASSIFIER_DP_FIELD].mask_range.u16 =
                    ipIngClassifierRule_p->protoInfo_u.udp.destPort.end;
            }
            else
            {
            	v.field[IP_ING_CLASSIFIER_DP_FIELD].value.u16 = 0;
            	v.field[IP_ING_CLASSIFIER_DP_FIELD].mask_range.u16 = 0xFFFF;
            }
            break;
        case ADK_IP_PROTOCOL_ICMP:
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_TYPE_MASK)
            {
            	ADK_IPV4_LOG_INFO(ipHandle,"\t\ticmpType=%d-%d\n", ipIngClassifierRule_p->protoInfo_u.icmp.type.begin, ipIngClassifierRule_p->protoInfo_u.icmp.type.end);
                v.field[IP_ING_CLASSIFIER_SP_FIELD].value.u16 =
                    ipIngClassifierRule_p->protoInfo_u.icmp.type.begin;
                v.field[IP_ING_CLASSIFIER_SP_FIELD].mask_range.u16 =
                    ipIngClassifierRule_p->protoInfo_u.icmp.type.end;
            }
            else
            {
            	v.field[IP_ING_CLASSIFIER_SP_FIELD].value.u16 = 0;
            	v.field[IP_ING_CLASSIFIER_SP_FIELD].mask_range.u16 = 0xFFFF;
            }

            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_CODE_MASK)
            {
            	ADK_IPV4_LOG_INFO(ipHandle,"\t\ticmpCode=%d-%d\n", ipIngClassifierRule_p->protoInfo_u.icmp.code.begin, ipIngClassifierRule_p->protoInfo_u.icmp.code.end);
                v.field[IP_ING_CLASSIFIER_DP_FIELD].value.u16 =
                    ipIngClassifierRule_p->protoInfo_u.icmp.code.begin;
                v.field[IP_ING_CLASSIFIER_DP_FIELD].mask_range.u16 =
                    ipIngClassifierRule_p->protoInfo_u.icmp.code.end;
            }
            else
            {
            	v.field[IP_ING_CLASSIFIER_DP_FIELD].value.u16 = 0;
            	v.field[IP_ING_CLASSIFIER_DP_FIELD].mask_range.u16 = 0xFFFF;
            }
            break;
        default:
            /* Set ANY for SP and DP */
            v.field[IP_ING_CLASSIFIER_SP_FIELD].value.u16 = 0;
            v.field[IP_ING_CLASSIFIER_SP_FIELD].mask_range.u16 = 0xFFFF;
            v.field[IP_ING_CLASSIFIER_DP_FIELD].value.u16 = 0;
            v.field[IP_ING_CLASSIFIER_DP_FIELD].mask_range.u16 = 0xFFFF;
            break;
        }
    }
    else
    {
        /* Set ANY for protocol */
        v.field[IP_ING_CLASSIFIER_PROTO_FIELD].value.u8 = 0;
        v.field[IP_ING_CLASSIFIER_PROTO_FIELD].mask_range.u8  = 0xFF;
        /* Set ANY for SP and DP */
        v.field[IP_ING_CLASSIFIER_SP_FIELD].value.u16 = 0;
        v.field[IP_ING_CLASSIFIER_SP_FIELD].mask_range.u16 = 0xFFFF;
        v.field[IP_ING_CLASSIFIER_DP_FIELD].value.u16 = 0;
        v.field[IP_ING_CLASSIFIER_DP_FIELD].mask_range.u16 = 0xFFFF;
    }
    /* Set Source address */
    if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_SRC_ADDR_MASK)
    {
    	ADK_IPV4_LOG_INFO(ipHandle,"\t\tsip=0x%x/%d\n", ipIngClassifierRule_p->srcAddr.addr, ipIngClassifierRule_p->srcAddr.prefixWidth);
        v.field[IP_ING_CLASSIFIER_SRC_FIELD].value.u32 =
            ipIngClassifierRule_p->srcAddr.addr;
        v.field[IP_ING_CLASSIFIER_SRC_FIELD].mask_range.u32  =
            ipIngClassifierRule_p->srcAddr.prefixWidth;
    }
    else
    {
        /* Set ANY */
        v.field[IP_ING_CLASSIFIER_SRC_FIELD].value.u32 = 0;
        v.field[IP_ING_CLASSIFIER_SRC_FIELD].mask_range.u32  = 0xFFFFFFFF;
    }
    /* Set Destination address */
    if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_ADDR_MASK)
    {
    	ADK_IPV4_LOG_INFO(ipHandle,"\t\tdip=0x%x/%d\n", ipIngClassifierRule_p->destAddr.addr, ipIngClassifierRule_p->destAddr.prefixWidth);
        v.field[IP_ING_CLASSIFIER_DST_FIELD].value.u32 =
            ipIngClassifierRule_p->destAddr.addr;
        v.field[IP_ING_CLASSIFIER_DST_FIELD].mask_range.u32  =
            ipIngClassifierRule_p->destAddr.prefixWidth;
    }
    else
    { /* SET ANY */
        v.field[IP_ING_CLASSIFIER_DST_FIELD].value.u32 = 0;
        v.field[IP_ING_CLASSIFIER_DST_FIELD].mask_range.u32  = 0xFFFFFFFF;
    }

    /* Set Interface handle */
    if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_INTF_HANDLE_MASK)
    {
    	ADK_IPV4_LOG_INFO(ipHandle,"\t\tifaceHd=%x\n", (uint32_t)(size_t) ADK_SYS_INTERFACE_ID_GET(ipIngClassifierRule_p->ipIntfHandle));
        /* Store ID instead of handle */
        v.field[IP_ING_CLASSIFIER_INTF_FIELD_HDL].value.u32 =
            (uint32_t)(size_t) ADK_SYS_INTERFACE_ID_GET(ipIngClassifierRule_p->ipIntfHandle);
        v.field[IP_ING_CLASSIFIER_INTF_FIELD_HDL].mask_range.u32  =
            (uint32_t)(size_t) ADK_SYS_INTERFACE_ID_GET(ipIngClassifierRule_p->ipIntfHandle);
    }
    else
    {
        /* Set ANY */
        v.field[IP_ING_CLASSIFIER_INTF_FIELD_HDL].value.u32 = 0;
        v.field[IP_ING_CLASSIFIER_INTF_FIELD_HDL].mask_range.u32  = 0xFFFFFFFF;
    }

    switch(ipIngClassifierAction_p->action)
    {
    case ADK_IP_ING_CLASSIFIER_ACTION_DROP:
        ADK_IPV4_LOG_INFO(ipHandle,"\t\taction=DROP\n");
        break;
    case ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_EO:
        ADK_IPV4_LOG_INFO(ipHandle,"\t\taction=SEND_TO_EO\n");
        break;
    case ADK_IP_ING_CLASSIFIER_ACTION_SEND_TO_HOST:
        ADK_IPV4_LOG_INFO(ipHandle,"\t\taction=SEND_TO_HOST\n");
        break;
    }
    v.data.category_mask = 1;
    v.data.priority = precedence;
    v.data.userdata = ingClassifierFlowId;

    if (rte_acl_add_rules(
          ((adk_ip_ing_classifer_info_t*) ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierInfo_p)->ipIngClassifierCtx[treeIndex],
            (struct rte_acl_rule *)&v, 1) < 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error inserting ACL rule in IP\n");
        return ADK_E_OTREE_INSERT_FAILED ;
    }

    /* */
    ADK_RETURN(retValue);

}/* classifier_add_rule_to_hw */
/**
 *=============================================================
 * classifier_build_ordered_tree
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ctx - Ingress classifer Context
 *
 * @retval ADK_SUCCESS Successfully built ingress classifer context.
 *
 * @retval ADK_E_OTREE_BUILD_FAILED - Build Failed
 *
 * @brief
 *
 * This function builds ingress classifier ordered tree
 * =============================================================
 */
adk_status_t classifier_build_ordered_tree(adk_module_handle_t ipHandle, adk_ip_ing_classifier_ctx_t *ctx)
{
    struct rte_acl_config cfg;
    int ret=0;

    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_MEMSET(&cfg, 0, sizeof(cfg));
    cfg.num_fields = ADK_IP_ING_CLASSIFIER_NUM_FIELDS;
    ADK_MEMCPY(&cfg.defs, adk_ip_ing_classifier_def, sizeof(adk_ip_ing_classifier_def));
    cfg.num_categories = 1;
    cfg.max_size = 0;

    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_build", "building IP Ingress Classifier context\n");

    ret = rte_acl_build(ctx, &cfg);
    if (ret != 0)
    {
    	ADK_IPV4_LOG_ERROR(ipHandle, ret, "rte_acl_build failed\n");
        return ADK_E_OTREE_BUILD_FAILED;
    }

    return ADK_SUCCESS;
} /* classifier_build_ordered_tree */

/*!
 * =============================================================
 * @internal
 * adk_ip_ingress_convert_to_hos_rule
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief
 *
 * Internal support function for ACL rule conversion.
 *
 * =============================================================
*/
static
adk_status_t adk_ip_ingress_convert_to_hos_rule(adk_module_handle_t ipHandle,
        uint32_t ingClassifierFlowId,
        const adk_ip_ingress_classifier_rule_t *ipIngClassifierRule_p,
        const adk_ip_ingress_classifier_action_t *ipIngClassifierAction_p,
        uint16_t *num_hos_rules_p,
        adk_hos_ip_classifier_tbl_rule_attr_t **rule_attr)
{
    uint32_t sport_begin = 0, sport_end = 0xFFFF;
    uint32_t dport_begin = 0, dport_end = 0xFFFF;
    adk_acl_range_bitmask_t *sport_bitmask_p = NULL, *dport_bitmask_p = NULL;
    adk_acl_range_bitmask_t *tmp_sport_p, *tmp_dport_p;
    int sport_rule_count=1, dport_rule_count=1;
    int total_rule_count, loop1, loop2;
    int rule_count;
    adk_status_t retValue= ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_module_handle_t eth_mod_handle;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    RTE_SET_USED(ipIngClassifierAction_p);

    if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_PROTO_MASK)
    {
        /* All Known protocol specific information is processed here */
        switch (ipIngClassifierRule_p->protocol)
        {
        case ADK_IP_PROTOCOL_ICMP:
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_TYPE_MASK) {
                sport_begin = ipIngClassifierRule_p->protoInfo_u.icmp.type.begin;
                sport_end = ipIngClassifierRule_p->protoInfo_u.icmp.type.end;
            }
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_CODE_MASK) {
                dport_begin = ipIngClassifierRule_p->protoInfo_u.icmp.code.begin;
                dport_end = ipIngClassifierRule_p->protoInfo_u.icmp.code.end;
            }
            break;

        case ADK_IP_PROTOCOL_TCP:
        case ADK_IP_PROTOCOL_SCTP:
        case ADK_IP_PROTOCOL_UDP:
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_SRC_PORT_MASK) {
                sport_begin = ipIngClassifierRule_p->protoInfo_u.udp.srcPort.begin;
                sport_end = ipIngClassifierRule_p->protoInfo_u.udp.srcPort.end;
            }
            if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_PORT_MASK) {
                dport_begin = ipIngClassifierRule_p->protoInfo_u.udp.destPort.begin;
                dport_end = ipIngClassifierRule_p->protoInfo_u.udp.destPort.end;
            }
            break;

        default:
            break;
        }               /* -----  end switch  ----- */
    }

    if ((ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_SRC_PORT_MASK) ||
        (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_TYPE_MASK))
    {
        /*Allocate memory for storing bitmask and Range info*/
        sport_bitmask_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
			 ADK_IP_HOS_RULE_OBJ,
			 1,sizeof(adk_acl_range_bitmask_t )*
                            ADK_ACL_MAX_HOS_RULES_PER_RULE);
        if(!sport_bitmask_p)
        {
            retValue = ADK_E_MEM_UNAVAILABLE;
            goto _adk_ip_ingress_convert_to_hos_rule_exit;
        }

        sport_rule_count = adk_acl_convert_range_bitmask(sport_begin,sport_end, 16, sport_bitmask_p);

        if(sport_rule_count <= 0)
        {
            ADK_IPV4_LOG_INFO(ipHandle,"Rule splitting not possible for sport\n");
            retValue = ADK_DP_E_ERROR;
            goto _adk_ip_ingress_convert_to_hos_rule_exit;
        }
    }

    if ((ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_PORT_MASK) ||
        (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_CODE_MASK))
    {
        /*Allocate memory for storing bitmask and Range info*/
        dport_bitmask_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
			 ADK_IP_HOS_RULE_OBJ,
			 1,sizeof(adk_acl_range_bitmask_t )*
                            ADK_ACL_MAX_HOS_RULES_PER_RULE);
        if(!dport_bitmask_p)
        {
            retValue = ADK_E_MEM_UNAVAILABLE;
            goto _adk_ip_ingress_convert_to_hos_rule_exit;
        }

        dport_rule_count = adk_acl_convert_range_bitmask(dport_begin, dport_end, 16, dport_bitmask_p);
        if(dport_rule_count <= 0)
        {
            ADK_IPV4_LOG_INFO(ipHandle, "Rule splitting not possible for dport\n");
            retValue = ADK_DP_E_ERROR;
            goto _adk_ip_ingress_convert_to_hos_rule_exit;
        }
    }

    total_rule_count = sport_rule_count * dport_rule_count;
 
    ADK_IPV4_LOG_INFO(ipHandle, "sport_rule_count: %d, dport_rule_count: %d \n",
            sport_rule_count, dport_rule_count);

    if(total_rule_count > ADK_ACL_MAX_HOS_RULES_PER_RULE)
    {
        ADK_IPV4_LOG_INFO(ipHandle, "Rule splitting crossing limits for hos rule count \n");
        retValue = ADK_DP_E_ERROR;
        goto _adk_ip_ingress_convert_to_hos_rule_exit;
    }

    for(loop1 =0; loop1< total_rule_count; loop1++)
    {
        rule_attr[loop1] = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_HOS_RULE_OBJ,
				 1, sizeof(adk_hos_ip_classifier_tbl_rule_attr_t));
        if(!rule_attr[loop1])
        {
            retValue = ADK_E_MEM_UNAVAILABLE;
            goto _adk_ip_ingress_convert_to_hos_rule_exit;
        }
    }

    *num_hos_rules_p = total_rule_count;  //Update hos Rule count

    tmp_dport_p = dport_bitmask_p;
    tmp_sport_p = sport_bitmask_p;

    rule_count = 0;
    adk_hos_ip_classifier_tbl_rule_attr_t *hos_rule = rule_attr[rule_count];

    for(loop1 = 0; loop1 < sport_rule_count; loop1++)
    {
        for(loop2=0; loop2 < dport_rule_count; loop2++)
        {
                hos_rule->is_ipv6 = false;
#if 0
                hos_rule->precedence = precedence;
#endif
                hos_rule->is_disabled = false;
                hos_rule->rule_id = ingClassifierFlowId;

                if(ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_INTF_HANDLE_MASK)
                {
                    uint32_t interface_id = ADK_SYS_INTERFACE_ID_GET(ipIngClassifierRule_p->ipIntfHandle);
                    retValue = adk_module_handle_get(sys_handle, ADK_ETHERNET_MODULE_ID,
                                    adk_ethernet_service_function_update, &eth_mod_handle);
                    if(ADK_SUCCESS != retValue)
                    {
                        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Ethernet module get handle failed\n");
                        retValue = ADK_E_MODULE_NOT_REGISTERED;
                        goto _adk_ip_ingress_convert_to_hos_rule_exit;
                    }
                    
                    hos_rule->vlan_id = ADK_HOS_ANY_VLAN; 
                    adk_eth_intf_get_vlan(eth_mod_handle, interface_id, &hos_rule->vlan_id);
                    hos_rule->useVlan = true;
                } 
                if ((ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_SRC_PORT_MASK) ||
                    (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_TYPE_MASK))
                {
                    hos_rule->l4_info1 = tmp_sport_p->val;
                    hos_rule->l4_info1_mask = tmp_sport_p->mask;
                }
                if ((ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_PORT_MASK) ||
                    (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_ICMP_CODE_MASK))
                {
                    hos_rule->l4_info2 = tmp_dport_p->val;
                    hos_rule->l4_info2_mask = tmp_dport_p->mask;
                }
                if (ipIngClassifierRule_p->ruleMask &
                    ADK_IP_ING_CLASSIFIER_SRC_ADDR_MASK)
                {
                    hos_rule->sip_addr[0] =
                        adk_htonl(ipIngClassifierRule_p->srcAddr.addr);
                    hos_rule->sip_addr_mask[0] =
                        adk_htonl(ip_bit_mask[ipIngClassifierRule_p->
                                              srcAddr.prefixWidth]);
                }

                if (ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_DEST_ADDR_MASK)
                {
                    hos_rule->dip_addr[0] =
                        adk_htonl(ipIngClassifierRule_p->destAddr.addr);
                    hos_rule->dip_addr_mask[0] =
                        adk_htonl(ip_bit_mask[ipIngClassifierRule_p->
                                              destAddr.prefixWidth]);
                }

                if(ipIngClassifierRule_p->ruleMask & ADK_IP_ING_CLASSIFIER_PROTO_MASK)
                {
                    hos_rule->proto = ipIngClassifierRule_p->protocol;
                    hos_rule->proto_mask = HOS_IP_PROTO_MASK;
                }

                rule_count++;
                hos_rule = rule_attr[rule_count]; //Move to next hos rule
            tmp_dport_p +=1; // Move to next dport mask
        } /* Dport loop ends*/
        tmp_sport_p +=1;    //move to next sport mask
        tmp_dport_p = dport_bitmask_p; //dport start
    }/*Sport loop ends*/

_adk_ip_ingress_convert_to_hos_rule_exit:
    if(dport_bitmask_p)
        ADK_NVM_FREE(dport_bitmask_p);
    if(sport_bitmask_p)
        ADK_NVM_FREE(sport_bitmask_p);
    return retValue;
}

/*!
 * =============================================================
 * adk_ip_ingress_classifier_rule_add
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[in]    ruleIndex - index at which the rule will be inserted
 *
 * @param[in]    ipIngClassifierRule_p  - pointer to adk_ip_ingress_classifier_rule_t 
 * data structure.
 *
 * @param[in]    ipIngClassifierAction_p  - pointer to adk_ip_ingress_classifier_action_t
 * data structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE - IP inteface handle
 * not present or invalid.
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid.
 *
 * @brief
 * This API validates the input parameter and adds a new rule to the ip ingress
 * association ordered tree. Note that the rule does not get reflected in
 * data plane until the adk_ip_ingress_classifier_build() API is called for this.
 *
 * =============================================================
 */
adk_status_t adk_ip_ingress_classifier_rule_add(adk_module_handle_t ipHandle,
        const adk_ip_ingress_classifier_rule_t *ipIngClassifierRule_p,
        const adk_ip_ingress_classifier_action_t *ipIngClassifierAction_p,
        const uint16_t ruleIndex
        )
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    ip_flow_info_t *myFlowInfo_p;
    adk_sys_handle_t sys_handle;
    adk_acl_rule_t *cpRule_p;
    size_t ingClassifRuleId = ADK_RM_RSRC_INVALID;
    uint32_t i=0;
    adk_ip_ingress_classifier_rule_handle_t classFlowHd = 0;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    /* Validate input args */
    retValue = adk_ip_classifer_add_validate_arg( ipHandle, ipIngClassifierRule_p,
            ipIngClassifierAction_p);

    if (ADK_SUCCESS != retValue)
        goto release_lock_and_return;

    ingClassifRuleId = adk_rm_resource_reserve(
            ipModuleInfo_p->ipIngClassifierRuleIDHandle);
    if (ADK_RM_RSRC_INVALID == ingClassifRuleId)
    {
        retValue = ADK_E_RESOURCE_UNAVAILABLE;
        goto release_lock_and_return;
    }
    /* Clearing to indicate no handle is allocated yet */
    classFlowHd = 0x0;
    /*Set the flow stats to 0*/
    for(i=0; i <ADK_MAX_NUM_CORES; ++i)
    {
        (ipModuleInfo_p->ip_dp_tbl_info_p->ipIngFlowStats)[i*ADK_DP_MAX_IP_ING_ASSOC_FLOWS+ingClassifRuleId].rxPkts = 0;
        (ipModuleInfo_p->ip_dp_tbl_info_p->ipIngFlowStats)[i*ADK_DP_MAX_IP_ING_ASSOC_FLOWS+ingClassifRuleId].txPkts = 0;
    }

    /* Allocate flow handle */
    ADK_SYS_FLOW_HANDLE_ALLOC(ipHandle, ADK_IP_ING_CLASSIFIER_RULE_INFO_OBJ,
            sizeof(ip_flow_info_t),
            &classFlowHd);

    /************/
    if(!(myFlowInfo_p = (ip_flow_info_t *)
            ADK_SYS_FLOW_ATTR_PTR_GET(classFlowHd)))
    {
        retValue=ADK_E_INTERNAL_ERROR;
        goto release_lock_and_return;
    }

    myFlowInfo_p->cookie = ADK_IPV4_FLOW_COOKIE;
    myFlowInfo_p->type = IP_FLOW_TYPE_INGRESS_CLASSIFIER_FLOW;
    ADK_SYS_FLOWID_SET(classFlowHd, ingClassifRuleId);

    ADK_MEMCPY((void *) &(myFlowInfo_p->u.ingClassifierFlow.rule),
            (const void *) ipIngClassifierRule_p,
            sizeof(adk_ip_ingress_classifier_rule_t));
    ADK_MEMCPY((void *) &(myFlowInfo_p->u.ingClassifierFlow.action),
            (const void *) ipIngClassifierAction_p,
            sizeof(adk_ip_ingress_classifier_action_t));
    /*
     * Add rule to ADK CP rule list.
     * The rules will be added during the next commit.
     */
    cpRule_p = &(myFlowInfo_p->u.ingClassifierFlow.cp_rule);
    cpRule_p->user_data = (void *)(classFlowHd);   /* IMP: We are storing the value of classFlowHd in cpList.
     	 	 	 	 	 	 	 	 	 	 	 	  Hence, we don't need to maintain it, ClassFlowHd, in turn, stored the
     	 	 	 	 	 	 	 	 	 	 	 	  start address of calloced block. So, effectively user_data is pointer to flow info structure */

    /* Convert this rule to a expanded set of hos rules */
    adk_hos_ip_classifier_tbl_rule_attr_t *hos_rule[ADK_ACL_MAX_HOS_RULES_PER_RULE];

    retValue = adk_ip_ingress_convert_to_hos_rule(ipHandle, ingClassifRuleId,
            ipIngClassifierRule_p, ipIngClassifierAction_p,
            &(cpRule_p->num_hos_rules), hos_rule);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to convert ingress classifier"
                           " rule to HOS rule\n");
        goto release_lock_and_return;
    }

    for(i = 0; i< cpRule_p->num_hos_rules; i++)
    {
        cpRule_p->hos_rule_attr[i] = (adk_hos_rule_attr_t *)hos_rule[i];
    }

    /************/

    retValue = adk_acl_rule_add(
            ((adk_ip_ing_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierInfo_p)->cpList,
            ruleIndex,
            cpRule_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to add the rule in CP list\n");
        goto release_lock_and_return;
    }

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);
    return ADK_SUCCESS;
release_lock_and_return:
    /* Upon error, release the reserved resources */
    if (ADK_RM_RSRC_INVALID != ingClassifRuleId)
    {
        adk_rm_resource_free(ipModuleInfo_p->ipIngClassifierRuleIDHandle, ingClassifRuleId);
    }

    if (retValue != ADK_SUCCESS) {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Failed to add classifier rule\n");
        if (classFlowHd)
            ADK_SYS_FLOW_HANDLE_FREE(classFlowHd);
    }

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:

    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
} /* adk_ip_ingress_classifier_rule_add */
/*!
 * =============================================================
 * adk_ip_ingress_classifier_rule_delete
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[in]    ruleIndex - index at which the rule is to be deleted.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid. Possibly rule index specified is either
 * out of sequence or there is no classifier entry at index 0.
 *
 * @retval ADK_E_MEM_UNAVAILABLE Exceeds the number of operation available
 * to be carried out before 'build'
 *
 * @brief
 * This API validates the input parameters and deletes the rule at the given
 * index from the ip ingress association ordered tree.
 * Note that the rule update does not get reflected in
 * data plane until the adk_ip_ingress_classifier_build() API is called for this.
 *
 * Note: Once a rule is deleted, all further rules will have their index decremented
 * by 1 (e.g. after deleting rule[3], the rule which was at index 4 will
 * become the rule at index 3 and same applies for all further rules).
 * =============================================================
 */
adk_status_t adk_ip_ingress_classifier_rule_delete(adk_module_handle_t ipHandle,
        const uint16_t ruleIndex)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_ing_classifer_info_t* ingClassifierInfo_p = NULL;
    uint16_t index=0;
    adk_acl_rule_t cpRule;
    size_t ingClassifRuleId = ADK_RM_RSRC_INVALID;
    adk_ip_ingress_classifier_rule_handle_t ipIngClassifierFlowHandle=0;
    adk_ip_dp_cppi_stats_t *CppiStats = NULL;


    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    ingClassifierInfo_p =  ((adk_ip_ing_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierInfo_p);

    adk_acl_get_reset(ingClassifierInfo_p->cpList);
    index=0;


    while(ADK_SUCCESS == adk_acl_next_rule_get(
    		ingClassifierInfo_p->cpList, &cpRule))
    {
    	if(index==ruleIndex)
    	{
    		ingClassifRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);
    		ipIngClassifierFlowHandle = (adk_ip_ingress_classifier_rule_handle_t)cpRule.user_data;
    		break;
    	}
    	index++;
    }

    if (ADK_RM_RSRC_INVALID != ingClassifRuleId)
    {
    	adk_rm_resource_free(ipModuleInfo_p->ipIngClassifierRuleIDHandle, ingClassifRuleId);
    }

    /* Take a backup of CPPI policer drops for global stats & reset the same. */
    CppiStats = &ipModuleInfo_p->ip_dp_tbl_info_p->ipCppiStats[ingClassifRuleId];

    if(CppiStats == NULL)
    {
      ADK_IPV4_LOG_ERROR(ipHandle,ADK_E_INTERNAL_ERROR,"Failed to reset Classifier Cppi Stats");
      goto _lock_release_failed;
    }
    
    ipModuleInfo_p->ipGlobalStatsTable_bkp.ipPolicerDrop += CppiStats->cppiPolicerDropPkts;
    ADK_MEMSET(CppiStats,0,sizeof(adk_ip_dp_cppi_stats_t));

    if (ipIngClassifierFlowHandle)
    	ADK_SYS_FLOW_HANDLE_FREE(ipIngClassifierFlowHandle);

    /*** ***/

    retValue = adk_acl_rule_delete(ingClassifierInfo_p->cpList,
            ruleIndex);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to delete the rule from CP list\n");
        goto release_lock_rule_delete;
    }
release_lock_rule_delete:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
} /* adk_ip_ingress_classifier_rule_delete */
/*!
 * =============================================================
 * adk_ip_ingress_classifier_build
 * @param[in]    ipHandle - IP module handle.
 *
 * @brief
 * This API validates the input parameter and adds a new rule to the ip ingress
 * association ordered tree. Note that the rule does not get reflected in
 * data plane until the adk_ip_ingress_classifier_build() API is called for this.
 */
adk_status_t adk_ip_ingress_classifier_build(adk_module_handle_t ipHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_ip_ing_classifer_info_t  *ingClassifierInfo_p = NULL;
    adk_sys_handle_t sys_handle;
    uint8_t i, activeTreeId, shadowTreeId;
    uint16_t ruleIndex;
    ip_flow_info_t *myFlowInfo_p;
    size_t ingClassifRuleId;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    adk_acl_rule_t cpRule;
    adk_ip_ingress_classifier_rule_t *ipIngClassifierRule_p;
    adk_ip_ingress_classifier_action_t *ipIngClassifierAction_p;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    ingClassifierInfo_p = (adk_ip_ing_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierInfo_p;
    /*
     * Clear the Ruleset & add the rules from the ADK CP rule list.
     * Then commit the ruleset.
     */
    if(ingClassifierInfo_p->aclTreeIndex)
    {
        activeTreeId = 1;
        shadowTreeId = 0;
    }
    else
    {
        activeTreeId = 0;
        shadowTreeId = 1;
    }
    /* Only update Shadow tree. Active tree update is not really required
     * Next build will atomatically update active now which will be shadow then
     */
    rte_acl_reset_rules(ingClassifierInfo_p->ipIngClassifierCtx[shadowTreeId]);
    /*Update CP version id for VRDs*/
    for(i=0;i<=ADK_IP_DEFAULT_VRD_ID;++i)
        (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[i])++;
    /* Update Active copy. Get the rules from CP list & add them to S/W ACL*/
    adk_acl_get_reset(ingClassifierInfo_p->cpList);
    ruleIndex = 0;


    while(ADK_SUCCESS == adk_acl_next_rule_get(
                ingClassifierInfo_p->cpList, &cpRule))
    {
        myFlowInfo_p = (ip_flow_info_t *)
            ADK_SYS_FLOW_ATTR_PTR_GET(cpRule.user_data);
        ipIngClassifierRule_p = &(myFlowInfo_p->u.ingClassifierFlow.rule);
        ipIngClassifierAction_p = &(myFlowInfo_p->u.ingClassifierFlow.action);
        ingClassifRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);
        /*TODO: enhance this call to perform bulk add*/
        if( ADK_SUCCESS != classifier_add_rule_to_hw(
                ipHandle,
                (ADK_IP_CLASSIFIER_RULE_MAX_PRIORITY - ruleIndex),
                ingClassifRuleId,
                ipIngClassifierRule_p,
                ipIngClassifierAction_p,
                shadowTreeId))
        {
            retValue = ADK_E_HW_PROGRAMMING_ERROR;
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Failed to add the rule in hardware\n");
            goto release_lock_and_return;
        }
        ruleIndex++;
    }

    /* Add ANY - DENY rule at the end */
    /* Set Rule mask as zero, indicating no mask */
    adk_ip_ingress_classifier_rule_t ipIngClassifierRule = {0};
    adk_ip_ingress_classifier_action_t ipIngClassifierAction = {0};
    ipIngClassifierRule.ruleMask = 0;
    ipIngClassifierAction.action = ADK_IP_ING_CLASSIFIER_ACTION_DROP;
    if(ADK_SUCCESS != classifier_add_rule_to_hw (
            ipHandle,
			(ADK_IP_CLASSIFIER_RULE_MAX_PRIORITY - ruleIndex),
            1, /* zero id reserved for default rule */
            &ipIngClassifierRule,
            &ipIngClassifierAction,
            shadowTreeId
            ))
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not add default DENY rule to IP ING CLASSIFIER!\n");
        goto release_lock_and_return;
    }

    /*Update shadow copy*/
    retValue = classifier_build_ordered_tree(ipHandle,
            ingClassifierInfo_p->ipIngClassifierCtx[shadowTreeId]);
    if(ADK_SUCCESS !=  retValue)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not build IP Ingress CLASSIFEER...\n");
        goto release_lock_and_return;
    }
    /* Toggle the index */
    ingClassifierInfo_p->aclTreeIndex = shadowTreeId;
    /* RCU sync */
    adk_rcu_synchronize();

    /* CP ACL List HOS Commit */
    retValue = adk_acl_hos_commit(ingClassifierInfo_p->cpList);
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error: Could not commit in HOS...\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto release_lock_and_return;
    }

release_lock_and_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);
    return retValue;

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retValue);
} /* adk_ip_ingress_classifier_build */

/*!
 * =============================================================
 *  adk_ip_get_ingress_classifiers_all
 *
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[out]   *numClassifiers - Number of ingress classifier rules
 *
 * @param[out]   **ingressClassifier_p - List of ingress classifier rules
 *
 * @retval ADK_SUCCESS - Successful completion
 *
 * @retval ADK_IP_E_INVALID_PARAMETER - Invalid input parameters
 * 
 * @retval ADK_E_INVALID_HANDLE - Invalid module handle
 * 
 * @brief
 *
 * This API will fetch ingress classifier rules in IP module.
 *
 * @details
 *
 * This API gets all the ingress classifier rules along with their count.
 * The array index of a rule represents its index in the tree.
 * The application is expected to free the memory allocated by the API in 
 *  variable *ingressClassifier_p.
 * =============================================================
 */
adk_status_t adk_ip_get_ingress_classifiers_all( 
	adk_module_handle_t               ipHandle, 
	uint32_t                          *numClassifiers, 
	adk_ip_ingress_classifier_info_t  **ingressClassifier_p)
{
    adk_status_t retVal = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p = NULL;
    uint32_t num_acl_nodes,i;
    adk_ip_ing_classifer_info_t *ingClassifierInfo_p;
    adk_ip_ingress_classifier_info_t *classifier_temp_info;
    adk_acl_rule_t classifier_rule;
    ip_flow_info_t *flow_info_p;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
	ADK_FORCE_LOG_ERROR("Invalid IP module handle \n");
	return retVal;
    }

    if(!(sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle)))
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Invalid sys handle \n");
	return retVal;
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    /*Validate input parameters*/
    if ((numClassifiers == NULL) || (ingressClassifier_p == NULL))
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
	ADK_FORCE_LOG_ERROR("NULL pointer passed, invalid input parameter \n");
	return retVal;
    }

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retVal);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    if(ipModuleInfo_p == NULL)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Invalid pointer to IP module info \n");
	goto _adk_ip_get_ingress_classifiers_all_return; 
    }

    ingClassifierInfo_p = (adk_ip_ing_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->ingClassifierInfo_p;

    if (ingClassifierInfo_p == NULL)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Invalid pointer to ingress classifier info\n");
	goto _adk_ip_get_ingress_classifiers_all_return;
    }

    /* Get the number of entries in the ingress classifier tree ==> Active tree in CP is always in sync with the committed rules,
     * hence fetch the ingress classifier info from active tree. */
    if((retVal = adk_acl_get_num_nodes(ingClassifierInfo_p->cpList,
		    &num_acl_nodes)) != ADK_SUCCESS)
    {
	ADK_FORCE_LOG_ERROR("Failed to get number of entries in ingress classifier tree. \n");
	goto _adk_ip_get_ingress_classifiers_all_return;
    }

    /* Allocate the memory to store the classifier info.
     * NOTE: Application should free the memory allocated in this API. */
    classifier_temp_info = (adk_ip_ingress_classifier_info_t *) malloc(num_acl_nodes * sizeof(adk_ip_ingress_classifier_info_t));
    adk_acl_get_reset(ingClassifierInfo_p->cpList);

    /* Now we have number of entries in the classifier tree, hence fetch the classifier info.*/
    for(i=0;i<num_acl_nodes;i++)
    {
        retVal = adk_acl_next_rule_get(ingClassifierInfo_p->cpList, &classifier_rule);
        if (retVal != ADK_SUCCESS)
        {
            retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LIST_NOT_FOUND);
            ADK_FORCE_LOG_ERROR("Classifier rule not found in ACL list\n");
            goto _adk_ip_get_ingress_classifiers_all_return;
        }

        flow_info_p = (ip_flow_info_t *)
            ADK_SYS_FLOW_ATTR_PTR_GET(classifier_rule.user_data);

        ADK_MEMCPY(&((classifier_temp_info+i)->ClassifierRule), &flow_info_p->u.ingClassifierFlow.rule,sizeof(adk_ip_ingress_classifier_rule_t));
        ADK_MEMCPY(&((classifier_temp_info+i)->ClassifierAction), &flow_info_p->u.ingClassifierFlow.action,sizeof(adk_ip_ingress_classifier_action_t));
    }

    /* Return the fetched values in the pointer */
    *numClassifiers = num_acl_nodes;
    *ingressClassifier_p = classifier_temp_info;

_adk_ip_get_ingress_classifiers_all_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle,retVal);	

_lock_acquire_failed:
_lock_release_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retVal);
}
/*************************************************** INGRESS  ORDERED CLASSIFIER END ***********************************************/

/*!  @} end addtogroup _ipFunctions
 *
 * @} end addtogroup _ip  IPv4 Module
 */
