/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_interface.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs to manage IPv4 interfaces. This includes creation,
 * deletion of interfaces, setting their attributes, getting or printing
 * statistics for the interfaces, etc.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include <adk_ethernet_export.h>
#include <adk_cksum.h>
#include <adk_timer.h>
#include "adk_ip_dp_internal.h"


#if 0
static adk_status_t adk_ip_invalid_sip(
    adk_module_handle_t ipHandle,
    const adk_ip_addr_prefix_t *sipAddrPrefix_p,
    adk_ipv4_oper_t oper);
#endif

static adk_status_t adk_ip_create_egress_qos_resources(
       adk_module_handle_t ipHandle,
       uint32_t ifId);
 
static adk_status_t adk_ip_delete_egress_qos_resources(
       adk_module_handle_t ipHandle,
       uint32_t ifId);


static inline adk_status_t add_route_to_all_tables(
    adk_sys_handle_t sys_handle,
    adk_module_handle_t ipHandle,
    uint32_t ip_addr_stats_id,
    adk_ip_addr_prefix_t ipAddrPrefix
)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t tblId = 0;
    adk_ip_dp_route_info_t *dipResult_p = NULL;
    adk_dp_status_t dp_st;

    for (tblId = 0; tblId < ADK_IP_MAX_ROUTING_TABLES; tblId++)
    {
        /* Add route for the interface address itself. Here the advanced lookup is
     * always enabled. Re-assembly can also be optionally enabled */

        dipResult_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                                     ADK_IP_DIP_OBJ,
                                     1, sizeof(adk_ip_dp_route_info_t));
        if (dipResult_p == NULL)
        {
            ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                               "Memory allocation failed\n");
            retValue = ADK_E_MEM_UNAVAILABLE;
            goto add_route_to_all_tables_exit;
        }
        dipResult_p->isTerminating = TRUE;
        dipResult_p->u.nh_if.interfaceId = ADK_DP_MAX_IP_INTERFACE;
        /* ip addr stats id will be set only for terminating addr */
        dipResult_p->u.termination.ip_addr_stats_id = ip_addr_stats_id;
        dipResult_p->prefixWidth = 32;

        dp_st = adk_ip_dp_dip_add(ipHandle, tblId, adk_htonl(ipAddrPrefix.addr),
                                  32, ADK_CONVERT_TO_LPM_DATA(dipResult_p));
        if (dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                               "Error in adding dest ip dataplane\n");
            goto add_route_to_all_tables_exit;
        }

    }

add_route_to_all_tables_exit:
    return retValue;

}

static inline adk_status_t delete_route_from_all_tables(
    adk_module_handle_t ipHandle,
    adk_ip_addr_prefix_t ipAddrPrefix
)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t tblId = 0;
    adk_dp_status_t dp_st;

    for (tblId = 0; tblId < ADK_IP_MAX_ROUTING_TABLES; tblId++)
    {

        dp_st = adk_ip_dp_dip_remove(ipHandle, tblId, adk_htonl(ipAddrPrefix.addr),
                                     32);
        if (dp_st != ADK_SUCCESS)
        {
            retValue = ADK_E_INTERNAL_ERROR;
            ADK_IPV4_LOG_FORCE("%s:%d,Error in removing dest ip from dataplane.\n",
                               __FILE__, __LINE__);
            goto delete_route_from_all_tables_exit;
        }
    }

delete_route_from_all_tables_exit:
    return retValue;
}




/**
 *=============================================================
 * adk_ip_interface_create
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[in]  ipIf_p  - pointer to adk_ip_interface_t data
 * structure.
 *
 * @param[out] ipIfHandle - pointer to the IP interface handle
 * which gets created by this API.
 *
 * @retval ADK_SUCCESS Successfully created ADK IP interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @retval ADK_E_MEM_UNAVAILABLE - failed to allocate memory
 * for interface handle.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @retval  ADK_IP_E_INTERFACE_DUPLICATE_ADDRESS - Specified IP address is
 * already configured for another interface.
 *
 * @retval ADK_IP_E_INVALID_IP_ADDRESS - Multicast/Broadcast address specified
 * for the the interface.
 *
 * @retval ADK_IP_E_INVALID_MTU - Invalid MTU configruation. Supported
 * range is from 64B to 9000B.
 *
 * @brief
 *
 * This function create an IP interface handle to be used across all the IPv4
 * APIs (route updates, associations, etc). The basis for an IPv4
 * interface is some lower layer interface which can be an L2 physical
 * interface or any logical interface.
 * =============================================================
 */
adk_status_t
adk_ip_interface_create(
    adk_module_handle_t ipHandle,
    const adk_ip_interface_t *ipIf_p,
    adk_interface_handle_t *ipIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t l2IfId, host_addr, host_mask;
    uint32_t vrdId = ADK_IP_DEFAULT_VRD_ID;
    ip_module_info_t *ipModuleInfo_p;
    ip_interface_info_t *if_pvt_info = NULL;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_st;
    uint32_t                stats_id, ip_addr_stats_id;
    bool enableQoS, enableRouting ;
    uint32_t ipIfId;


    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipModuleInfo_p || !ipIf_p || !ipIfHandle)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "NULL pointer parameter specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ip_interface_create_return;
    }

    /*Validate the attribute mask */
    if(!(ipIf_p->ipIntfAttr < ADK_IP_INTERFACE_MASK_MAX))
    {
	ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
		"Invalid attribute mask specified\n");
	retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
	goto adk_ip_interface_create_return;

    }

    *ipIfHandle = 0;

    /* Validate the address */
    if (ADK_IPV4_IS_IP_INTF_ADDR_INVALID(ipIf_p->ipAddrPrefix))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_IP_ADDRESS,
                           "Invalid interface address specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        goto adk_ip_interface_create_return;
    }

    /* The host part should not be a reserved one unless it is a P2P
     * Ethernet link */
    host_mask = ~ip_bit_mask[ipIf_p->ipAddrPrefix.prefixWidth];
    host_addr = (ipIf_p->ipAddrPrefix.addr & host_mask);
    if ((ipIf_p->ipAddrPrefix.prefixWidth != IP_P2P_IF_ADDR_WIDTH) &&
        ((host_addr == 0) || (host_addr == host_mask)))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_IP_ADDRESS,
                           "Invalid interface address specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        goto adk_ip_interface_create_return;
    }

    /* If MTU is specified, then check its validity */
    if (ipIf_p->mtu) {
        if ((ipIf_p->mtu < IP_IF_MTU_MIN) || (ipIf_p->mtu > IP_IF_MTU_MAX))
        {
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_MTU,
                               "Invalid MTU Value %d\n",
                               ipIf_p->mtu);
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_MTU);
            goto adk_ip_interface_create_return;
        }
    }

    /* Validate the interface handle */
    if (!ADK_SYS_VALID_INTERFACE_HANDLE(ipIf_p->layer2InterfaceHandle))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_IP_INTERFACE,
                           "Invalid Layer-2 interface handle \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_INTERFACE);
        goto adk_ip_interface_create_return;
    }

    if(ipModuleInfo_p->aggrIpv4AddrCount >= ADK_IP_MAX_ADDRESSES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_MAX_IP_ADDR_EXCEEDED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Max allowed ip addresses (%d) already configured in the system\n", ADK_IP_MAX_ADDRESSES);
        goto adk_ip_interface_create_return;
    }

    if(ipModuleInfo_p->ipv4IfaceCount >= ADK_IP_MAX_INTERFACES)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_IP_INTERFACE, "Max allowed ipv4 interfaces (%d) already created in the system\n",
                ADK_IP_MAX_INTERFACES);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_INTERFACE);
        goto adk_ip_interface_create_return;
    }

#ifdef ENABLE_VRD
    /* Validate VRD handle */
    if (!adk_ipv4_vrd_valid_handle(ipIf_p->ipVrdHandle, &vrdId))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_INTERFACE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid VIrtual Routing Domain handle \n");
        goto adk_ip_interface_create_return;
    }
#else
	vrdId = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE_VRD */

    ADK_SYS_INTERFACE_HANDLE_ALLOC(ipHandle, ADK_IP_INTF_OBJ,
                                   sizeof(ip_interface_info_t),
                                   ipIfHandle);
    if (*ipIfHandle == 0) {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                           "Memory allocation failed\n");
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto adk_ip_interface_create_return;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(*ipIfHandle);

    ADK_MEMCPY(&if_pvt_info->if_info, ipIf_p, sizeof(adk_ip_interface_t));

    if_pvt_info->cookie = ADK_IPV4_INTF_COOKIE;

    stats_id = adk_rm_resource_reserve(ipModuleInfo_p->ipStatsHandle);
    if(stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to allocate IP IF Stats ID");
        goto adk_ip_interface_create_return;
    }

    ip_addr_stats_id = adk_rm_resource_reserve(ipModuleInfo_p->ipAddrStatsHandle);
    if(ip_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Failed to allocate IP Address Stats ID");
        goto adk_ip_interface_create_return;
    }

    if_pvt_info->ip_stats_id = stats_id;
    if_pvt_info->ip_addr_stats_id = ip_addr_stats_id;


    retValue = adk_sys_egress_stats_id_alloc(sys_handle,
                                     &(if_pvt_info->sys_egress_stats_id));
    if(retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to allocate sys egress stats id fot IP IF");
        goto adk_ip_interface_create_return;
    }

    /* If MTU is not specified, assign the default value */
    if (if_pvt_info->if_info.mtu == 0) {
        if_pvt_info->if_info.mtu = ADK_IP_INTERFACE_DEFAULT_MTU;
    }

    /*Ingress ACL may or may not be enabled, but ruleset Id is not mapped*/
    if_pvt_info->ingress_ruleset_id = 0;

    l2IfId = ADK_SYS_INTERFACE_ID_GET(ipIf_p->layer2InterfaceHandle);

    if_pvt_info->vrd_id = vrdId;
    if(likely(l2IfId >= ADK_ETH_NUM_BRIDGE))
        ipIfId = l2IfId - ADK_ETH_NUM_BRIDGE + ADK_ETH_BR_MAX;
    else
        ipIfId = l2IfId;

    ADK_SYS_INTERFACE_ID_SET(*ipIfHandle, ipIfId);
    ADK_SYS_VRD_ID_SET(*ipIfHandle, vrdId);

    dp_st = adk_ip_dp_interface_add(ipHandle, *ipIfHandle, stats_id,
                                    if_pvt_info->sys_egress_stats_id);
    if(dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error in adding IPv4 interface to dataplane\n");
        goto adk_ip_interface_create_return;
    }

    ADK_IPV4_LOG_INFO(ipHandle, "Allocated IP interface ID %d for l2IfId %d\n", ipIfId, l2IfId);

    dp_st = adk_ip_create_egress_qos_resources(ipHandle, ipIfId);

    if(dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error in Creating IP QOS resources in dataplane\n");
        goto adk_ip_interface_create_return;
    }
    
    add_route_to_all_tables(sys_handle, ipHandle, ip_addr_stats_id, ipIf_p->ipAddrPrefix);

    /*Update CP version id*/
    ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId]++;

    adk_module_handle_t eth_mod_handle=NULL; 
    retValue = adk_module_handle_get(sys_handle, ADK_ETHERNET_MODULE_ID, NULL,
                                            &eth_mod_handle);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Ethernet module handle not found!\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_create_return;
    }

    adk_hos_ip_term_tbl_rule_attr_t hos_ip_term_rule =
    {
        .precedence = -1,
        .pkt_type = ADK_HOS_L2_TERMINATING,
        .is_ipv6 = false,
        .ip_addr_mask[0] = adk_htonl(0xFFFFFFFF),
        .ip_addr[0] = adk_htonl(ipIf_p->ipAddrPrefix.addr),
        .is_ip_term = true,
        .is_bcast = false,
        .is_mcast = false,
        .ip_addr_stats_id = ip_addr_stats_id,
    };

    retValue = adk_hos_rule_add(ADK_HOS_IP_TERM_TBL,
                                (adk_hos_rule_attr_t *) &hos_ip_term_rule,
                                &(if_pvt_info->prim_addr_term_rule_id[0]));
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                 "Failed to add IP-Term HOS rule for ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_create_return;
    }


#ifdef ADK_HOS_OLD_IMPL
    retValue = adk_hos_rule_add(ADK_HOS_INNER_IP4_TERM_TBL,
                                (adk_hos_rule_attr_t *) &hos_ip_term_rule,
                                &(if_pvt_info->prim_addr_term_rule_id[1]));
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                 "Failed to add Inner IP-Term HOS rule for ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_create_return;
    }
#endif

#if 0 /* TO DO: Remove if not needed */
    retValue = adk_eth_intf_vrd_config(eth_mod_handle,
            ipIf_p->layer2InterfaceHandle, vrdId);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                          "Failed to update Outer IP Id HOS rule with VRD Id\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_create_return;
    }
#endif

    /*TODO: add other entries*/
    /* Configure the invalid SIP addresses corresponding to interface address */
    /* Then program layer-2 interface tree with its source IP address and MTU
     * for the attached subnet mask... This is for ULI */

    /* ip_dp_tbl_info_p global variable gets updated only in dp_init() that happens only in dp_process(). It is uninitialized in Application process. Hence setting it here */
    ip_dp_tbl_info_p = ipModuleInfo_p->ip_dp_tbl_info_p;

    /* We are accessing shared memory table through interface handles mapping as ip module handle is not available here. */
    ipModuleInfo_p->ip_dp_tbl_info_p->ifaceHandles[ipIfId] = *ipIfHandle;
    ++ipModuleInfo_p->ipv4IfaceCount;
    ++ipModuleInfo_p->aggrIpv4AddrCount;


    enableQoS = (ipIf_p->ipIntfAttr & ADK_IP_INTERFACE_MASK_QOS_ENABLE );
    enableRouting = (ipIf_p->ipIntfAttr & ADK_IP_INTERFACE_MASK_FORWARDING_ENABLE);

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_interface_create: Created interface: 0x%lx, ip=0x%x/%d, l2IfHandle=%p, "
            "ipIfId=%d, iface-mtu=%d, enableRouting=%d, enableQos=%d\n",
                    (uint64_t)*ipIfHandle, ipIf_p->ipAddrPrefix.addr, ipIf_p->ipAddrPrefix.prefixWidth,
                    ipIf_p->layer2InterfaceHandle, ipIfId, ipIf_p->mtu, enableRouting , enableQoS);

adk_ip_interface_create_return:

    if (retValue != ADK_SUCCESS)
        if (ipIfHandle != NULL)
            if (*ipIfHandle != 0)
                ADK_SYS_INTERFACE_HANDLE_FREE(*ipIfHandle);

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    
    ADK_RETURN(retValue);

}    /* -----  end of function adk_ip_interface_create----- */

/**
 *=============================================================
 * adk_ip_interface_delete
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[out] ipIfHandle - IP interface handle to be deleted
 *
 * @retval ADK_SUCCESS Successfully created ADK IP interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This function destroys an IP interface handle.
 * TO DO: Routes for secondary ip addresses are not removed
 * =============================================================
 */
adk_status_t
adk_ip_interface_delete(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t vrdId = ADK_IP_DEFAULT_VRD_ID;
    ip_module_info_t *ipModuleInfo_p;
    ip_interface_info_t *if_pvt_info = NULL;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_st;
    adk_ip_interface_t      *ipIf_p = NULL;
    adk_ip_interface_stats_t ipIfStats;
    uint16_t ip_addr_stats_id;
    adk_ip_dp_ip_addr_stats_t *dp_stats;
    uint32_t i=0, j=0;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "NULL pointer parameter specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_interface_delete_return;
    }

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_interface_delete_return;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);

    ipIf_p = &if_pvt_info->if_info;
#ifdef ENABLE_VRD
	vrdId = if_pvt_info->vrd_id;
#else
	vrdId = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE_VRD */

    dp_st = adk_ip_dp_interface_stats_get(ipHandle, ipIfHandle, &ipIfStats);
    if(dp_st != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "IP Interface Stats get failed\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_interface_delete_return;
    }

    //Enable for debug stats
    /* ipModuleInfo_p->ipGlobalStatsTable_bkp.ipInPkts += ipIfStats.ifInPackets;
       ipModuleInfo_p->ipGlobalStatsTable_bkp.ipOutPkts += ipIfStats.ifOutPackets;*/

    ADK_IPV4_LOG_INFO(ipHandle, "Deleting IP interface with ID 0x%x\n",
        ADK_SYS_INTERFACE_ID_GET(ipIfHandle));

    dp_st = adk_ip_delete_egress_qos_resources(ipHandle, ADK_SYS_INTERFACE_ID_GET(ipIfHandle));

    if(dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error in Deleting IP QOS resources in dataplane\n");
        goto _adk_ip_interface_delete_return;
    }

    dp_st = adk_ip_dp_interface_remove(ipHandle, ipIfHandle);
    if(dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error in  removing IPv4 interface from dataplane\n");
        goto _adk_ip_interface_delete_return;
    }

    delete_route_from_all_tables(ipHandle, ipIf_p->ipAddrPrefix);

    /*Update CP version id*/
    ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId] ++;

    retValue = adk_hos_rule_delete(ADK_HOS_IP_TERM_TBL,
                                   if_pvt_info->prim_addr_term_rule_id[0]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                 "Failed to remove IP-Term HOS rule for ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_interface_delete_return;
    }


#ifdef ADK_HOS_OLD_IMPL
    retValue = adk_hos_rule_delete(ADK_HOS_INNER_IP4_TERM_TBL,
                                   if_pvt_info->prim_addr_term_rule_id[1]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                 "Failed to remove Inner IP-Term HOS rule for ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_interface_delete_return;
    }
#endif
    adk_rm_resource_free(ipModuleInfo_p->ipStatsHandle, if_pvt_info->ip_stats_id);
    if_pvt_info->ip_stats_id = 0;

    /*Now clean up*/
    ip_addr_stats_id = if_pvt_info->ip_addr_stats_id;
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipAddrStatsTble[i][ip_addr_stats_id]);
        ADK_MEMSET(dp_stats, 0, sizeof(adk_ip_dp_ip_addr_stats_t));
    }
    adk_rm_resource_free(ipModuleInfo_p->ipAddrStatsHandle, if_pvt_info->ip_addr_stats_id);
    if_pvt_info->ip_addr_stats_id = 0;

    for(j=0;j<IP_SEC_ADDR_MAX;j++)
    {
        if (if_pvt_info->ip_prefix[j].addr)
        {
            delete_route_from_all_tables(ipHandle, if_pvt_info->ip_prefix[j]);

            retValue = adk_hos_rule_delete(ADK_HOS_IP_TERM_TBL, if_pvt_info->sec_addr_term_rule_id[0][j]);
            if (retValue != ADK_SUCCESS)
            {
                ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                                   "Failed to remove IP-Term HOS rule for ip-intf sec address\n");
                retValue = ADK_E_INTERNAL_ERROR;
                goto _adk_ip_interface_delete_return;
            }

            ip_addr_stats_id = if_pvt_info->ip_prefix_stats_id[j];
            for (i = 0; i < ADK_MAX_NUM_CORES; i++)
            {
                dp_stats = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipAddrStatsTble[i][ip_addr_stats_id]);
                ADK_MEMSET(dp_stats, 0, sizeof(adk_ip_dp_ip_addr_stats_t));
            }
            adk_rm_resource_free(ipModuleInfo_p->ipAddrStatsHandle, if_pvt_info->ip_prefix_stats_id[j]);

            if_pvt_info->ip_prefix[j].addr = 0;
            if_pvt_info->ip_prefix[j].prefixWidth = 0;
            if_pvt_info->ip_prefix_stats_id[j] = 0;

            /* For Secondary interface removal */
            --ipModuleInfo_p->aggrIpv4AddrCount;
            --if_pvt_info->numSecAddr;
        }
    }

    retValue = adk_sys_egress_stats_id_free(sys_handle, if_pvt_info->sys_egress_stats_id);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                          "Failed to free egress stats id for ip-intf\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_interface_delete_return;
    }

    /*TODO: remove other entries*/
    /* Configure the invalid SIP addresses corresponding to interface address */
    /* Then program layer-2 interface tree with its source IP address and MTU
     * for the attached subnet mask... This is for ULI */

    /* Clear the internal cookie */
    if_pvt_info->cookie = 0;

    ipModuleInfo_p->ip_dp_tbl_info_p->ifaceHandles[ADK_SYS_INTERFACE_ID_GET(ipIfHandle)] = 0;
    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_interface_delete: Deleted interface 0x%lx\n", (uint64_t)ipIfHandle);

    /* For Primary interface removal */
    --ipModuleInfo_p->aggrIpv4AddrCount;
    --ipModuleInfo_p->ipv4IfaceCount;
    /* Free the interface data structure */
    ADK_SYS_INTERFACE_HANDLE_FREE(ipIfHandle);

_adk_ip_interface_delete_return:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    
    ADK_RETURN(retValue);

}    /* -----  end of function adk_ip_interface_delete ----- */

/*!
 * =============================================================
 * adk_ip_interface_addr_add
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ipIfHandle  - IP interface handle.
 * @param[in]  ipAddrPrefix_p - Pointer to the additional IP address/subnet
 * to be set for the interface.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. Module
 * handle may be invalid.
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE- Invalid IP interface handle passed.
 *
 * @brief
 *
 * This API sets additional IP address for an existing IP interface. Using this
 * API an IP interface can be assigned multiple IP addresses. However, the IP
 * addresses must not have overlapping subnets. The API does not perform check
 * for overlapping subnets and hence it is application's responsibility not to
 * configure overlapping IP subnets on an interface. All the IP addresses that
 * are added using this API are secondary IP addresses while the one specified
 * at the time of interface creation is the primary IP address. All these IP
 * addresses work the same way - they only differ in how they are configured.
 * =============================================================
 */
adk_status_t adk_ip_interface_addr_add(
    adk_module_handle_t   ipHandle,
    adk_interface_handle_t ipIfHandle,
    const adk_ip_addr_prefix_t   *ipAddrPrefix_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t idx, host_addr, host_mask;
    uint32_t ipIfId;
    ip_interface_info_t *if_pvt_info = NULL;
    adk_ip_interface_t *ipIf_p = NULL;
    adk_sys_handle_t sys_handle;
    uint32_t vrdId = ADK_IP_DEFAULT_VRD_ID;
    ip_module_info_t *ipModuleInfo_p = NULL;
    uint32_t ip_addr_stats_id;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %" ADKPRIh "\n", ipHandle);
        return (ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %" ADKPRIh "\n", ipHandle);
        goto adk_ip_interface_addr_add_return;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /* Validate the address */
    if (ADK_IPV4_IS_IP_INTF_ADDR_INVALID((*ipAddrPrefix_p)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid interface address specified\n");
        goto adk_ip_interface_addr_add_return;
    }

    if(ipModuleInfo_p->aggrIpv4AddrCount >= ADK_IP_MAX_ADDRESSES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_MAX_IP_ADDR_EXCEEDED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Max allowed ip addresses (%d) already configured in the system\n", ADK_IP_MAX_ADDRESSES);
        goto adk_ip_interface_addr_add_return;
    }

    /* The host part should not be a reserved one unless it is a P2P
     * Ethernet link */
    host_mask = ~ip_bit_mask[ipAddrPrefix_p->prefixWidth];
    host_addr = (ipAddrPrefix_p->addr & host_mask);
    if ((ipAddrPrefix_p->prefixWidth != IP_P2P_IF_ADDR_WIDTH) &&
        ((host_addr == 0) || (host_addr == host_mask)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid interface address specified\n");
        goto adk_ip_interface_addr_add_return;
    }

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto adk_ip_interface_addr_add_return;
    }

    if_pvt_info = (ip_interface_info_t *)
        ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);

    ipIf_p = &if_pvt_info->if_info;

    ipIfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    if((if_pvt_info->numSecAddr+1) >= ADK_IP_MAX_ADDR_PER_INTERFACE)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Max allowed ip addresses per interface (%d) are already configured\n", ADK_IP_MAX_ADDR_PER_INTERFACE);
        goto adk_ip_interface_addr_add_return;
    }

#ifdef ENABLE_VRD
    /* Validate VR Domain Handle */
    if (!adk_ipv4_vrd_valid_handle(ipIf_p->ipVrdHandle, &vrdId))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid VIrtual Routing Domain handle \n");
        goto adk_ip_interface_addr_add_return;
    }
#else
    vrdId = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE_VRD */

    /* Find free space for the address in our internal data structure */
    for (idx = 0; idx < IP_SEC_ADDR_MAX; idx++)
    {
        if ((if_pvt_info->ip_prefix[idx].addr == 0) &&
            (if_pvt_info->ip_prefix[idx].prefixWidth == 0))
        {
            break;
        }
    }
    /* If not found, return error */
    if (idx == IP_SEC_ADDR_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "No space left for any more addr!\n");
        goto adk_ip_interface_addr_add_return;
    }

    ip_addr_stats_id = adk_rm_resource_reserve(ipModuleInfo_p->ipAddrStatsHandle);
    if (ip_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to allocate IP Address Stats ID");
        goto adk_ip_interface_addr_add_return;
    }

    ipIfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    add_route_to_all_tables(sys_handle, ipHandle, ip_addr_stats_id, *ipAddrPrefix_p);

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId])++;

    /* Term Address offload */
    adk_hos_ip_term_tbl_rule_attr_t hos_ip_term_rule =
        {
            .precedence = -1,
            .pkt_type = ADK_HOS_L2_TERMINATING,
            .is_ipv6 = false,
            .ip_addr_mask[0] = adk_htonl(0xFFFFFFFF),
            .ip_addr[0] = adk_htonl(ipAddrPrefix_p->addr),
            .is_ip_term = true,
            .is_bcast = false,
            .is_mcast = false,
            .ip_addr_stats_id = ip_addr_stats_id,
        };

    retValue = adk_hos_rule_add(ADK_HOS_IP_TERM_TBL,
                                (adk_hos_rule_attr_t *)&hos_ip_term_rule,
                                &(if_pvt_info->sec_addr_term_rule_id[0][idx]));
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Failed to add IP-Term HOS rule for ip-intf sec address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_addr_add_return;
    }

#ifdef ADK_HOS_OLD_IMPL
    retValue = adk_hos_rule_add(ADK_HOS_INNER_IP4_TERM_TBL,
                                (adk_hos_rule_attr_t *)&hos_ip_term_rule,
                                &(if_pvt_info->sec_addr_term_rule_id[1][idx]));
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Failed to add IP-Term HOS rule for ip-intf sec address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_addr_add_return;
    }
#endif

    /* Store the IP address internally */
    if_pvt_info->ip_prefix[idx].addr = ipAddrPrefix_p->addr;
    if_pvt_info->ip_prefix[idx].prefixWidth = ipAddrPrefix_p->prefixWidth;
    if_pvt_info->ip_prefix_stats_id[idx] = ip_addr_stats_id;
    if_pvt_info->numSecAddr++;
    ++ipModuleInfo_p->aggrIpv4AddrCount;

    /*TODO: add other entries*/
    /* Configure the invalid SIP addresses corresponding to interface address */
    /* Then program layer-2 interface tree with its source IP address and MTU
     * for the attached subnet mask... This is for ULI */

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_interface_addr_add: Added address 0x%x/%d to interface 0x%lx\n", ipAddrPrefix_p->addr, ipAddrPrefix_p->prefixWidth,
                      (uint64_t)ipIfHandle);
adk_ip_interface_addr_add_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    return retValue;
} /* -----  end of function adk_ip_interface_addr_add ----- */

/*!
 * =============================================================
 * adk_ip_interface_addr_rem
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ipIfHandle  - IP interface handle.
 * @param[in]  ipAddrPrefix_p - Pointer to the IP address/subnet to
 * be removed from the interface.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. Module
 * handle may be invalid.
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE- Invalid IP interface handle passed.
 *
 * @brief
 *
 * This API remove an existing IP address for an IP interface. The IP address
 * must be a secondary IP address. This API cannot be used to remove the primary
 * IP address. Even if all the secondary IP addresses are removed using this
 * API, the primary IP address always remains in existence. Primary IP address
 * can only be removed by deleting the interface.
 * =============================================================
 */
adk_status_t adk_ip_interface_addr_rem(
    adk_module_handle_t   ipHandle,
    adk_interface_handle_t   ipIfHandle,
    const adk_ip_addr_prefix_t   *ipAddrPrefix_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t host_addr, host_mask, idx = 0;
    uint32_t ipIfId;
    ip_interface_info_t     *if_pvt_info = NULL;
    adk_ip_interface_t      *ipIf_p = NULL;
    adk_sys_handle_t sys_handle;
    uint32_t vrdId = ADK_IP_DEFAULT_VRD_ID;
    ip_module_info_t *ipModuleInfo_p = NULL;
    bool is_primary = FALSE;
    adk_ip_dp_interface_entry_t *dp_entry;
    uint32_t ifres_index, i;
    adk_hos_rule_id_t hos_rule_id = {0}, hos_rule_id_old = {0};
    uint16_t ip_addr_stats_id;
    adk_ip_dp_ip_addr_stats_t *dp_stats;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto adk_ip_interface_addr_rem_return;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /* Validate the address */
    if (ADK_IPV4_IS_IP_INTF_ADDR_INVALID((*ipAddrPrefix_p)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid interface address specified\n");
        goto adk_ip_interface_addr_rem_return;
    }

    /* The host part should not be a reserved one unless it is a P2P
     * Ethernet link */
    host_mask = ~ip_bit_mask[ipAddrPrefix_p->prefixWidth];
    host_addr = (ipAddrPrefix_p->addr & host_mask);
    if ((ipAddrPrefix_p->prefixWidth != IP_P2P_IF_ADDR_WIDTH) &&
        ((host_addr == 0) || (host_addr == host_mask)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid interface address specified\n");
        goto adk_ip_interface_addr_rem_return;
    }

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto adk_ip_interface_addr_rem_return;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);

    ipIf_p = &if_pvt_info->if_info;

#ifdef ENABLE_VRD
    /* Validate VR Domain Handle */
    if (!adk_ipv4_vrd_valid_handle(ipIf_p->ipVrdHandle, &vrdId))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid VIrtual Routing Domain handle \n");
        goto adk_ip_interface_addr_rem_return;
    }
#else
	vrdId = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE_VRD */

    ipIfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /* Find the address in our internal data structure */
    if ((ipIf_p->ipAddrPrefix.addr == ipAddrPrefix_p->addr) &&
        (ipIf_p->ipAddrPrefix.prefixWidth == ipAddrPrefix_p->prefixWidth))
    {
        is_primary = TRUE;
        /* Make sure that there is atleast one secondary IP address that can be
         * made the primary IP address */
        if (if_pvt_info->numSecAddr == 0)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                               "Cannot remove the one and only IP address\n");
            goto adk_ip_interface_addr_rem_return;
        }
    }
    else
    {
        for (idx = 0; idx < if_pvt_info->numSecAddr; idx++)
        {
            if ((if_pvt_info->ip_prefix[idx].addr == ipAddrPrefix_p->addr) &&
                (if_pvt_info->ip_prefix[idx].prefixWidth ==
                 ipAddrPrefix_p->prefixWidth))
            {
                break;
            }
        }
        /* If not found, return error */
        if (idx == if_pvt_info->numSecAddr)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                               "Cannot find the specified addr to remove!\n");
            goto adk_ip_interface_addr_rem_return;
        }
    }

    delete_route_from_all_tables(ipHandle, *ipAddrPrefix_p);

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId])++;

    /*TODO: add other entries*/
    /* Configure the invalid SIP addresses corresponding to interface address */
    /* Then program layer-2 interface tree with its source IP address and MTU
     * for the attached subnet mask... This is for ULI */

    /* Remove the IP address from internal data structure */
    uint8_t lastAddrIndex = if_pvt_info->numSecAddr-1;
    if (is_primary == TRUE)
    {
        /* Reset and free the address stats */
        ip_addr_stats_id = if_pvt_info->ip_addr_stats_id;
        for (i = 0; i < ADK_MAX_NUM_CORES; i++)
        {
            dp_stats = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipAddrStatsTble[i][ip_addr_stats_id]);
            ADK_MEMSET(dp_stats, 0, sizeof(adk_ip_dp_ip_addr_stats_t));
        }
        adk_rm_resource_free(ipModuleInfo_p->ipAddrStatsHandle, if_pvt_info->ip_addr_stats_id);
        if_pvt_info->ip_addr_stats_id = 0;
        /* */

        if_pvt_info->if_info.ipAddrPrefix.addr =
            if_pvt_info->ip_prefix[lastAddrIndex].addr;
        if_pvt_info->if_info.ipAddrPrefix.prefixWidth =
            if_pvt_info->ip_prefix[lastAddrIndex].prefixWidth;
        if_pvt_info->ip_addr_stats_id = if_pvt_info->ip_prefix_stats_id[lastAddrIndex];

        /* update primary address in DP structure */
        ifres_index = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);
        if (ifres_index >= ADK_DP_MAX_IP_INTERFACE)
        {   /* TO DO: Returning error from here leaves if_pvt_info->if_info.ipAddrPrefix wrong. Fix it */
            ADK_IP_DP_LOG_ERROR("IP interface id larger than the max supported\n");
            retValue = ADK_E_INTERNAL_ERROR;
            ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                               "Could not program LPM rule in hardware\n");
            goto adk_ip_interface_addr_rem_return;
        }

        dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceTble[ifres_index]);
        dp_entry->if_addr = adk_htonl(if_pvt_info->if_info.ipAddrPrefix.addr);
        hos_rule_id = if_pvt_info->prim_addr_term_rule_id[0];
        hos_rule_id_old = if_pvt_info->prim_addr_term_rule_id[1];
        if_pvt_info->prim_addr_term_rule_id[0] =
            if_pvt_info->sec_addr_term_rule_id[0][lastAddrIndex];
    }
    else
    {
        /* Reset and free the address stats */
        ip_addr_stats_id = if_pvt_info->ip_prefix_stats_id[idx];
        for(i=0;i<ADK_MAX_NUM_CORES;i++)
        {
            dp_stats = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipAddrStatsTble[i][ip_addr_stats_id]);
            ADK_MEMSET(dp_stats, 0, sizeof(adk_ip_dp_ip_addr_stats_t));
        }
        adk_rm_resource_free(ipModuleInfo_p->ipAddrStatsHandle, if_pvt_info->ip_prefix_stats_id[idx]);
        if_pvt_info->ip_prefix_stats_id[idx] = 0;
        /* */


        if_pvt_info->ip_prefix[idx].addr =
            if_pvt_info->ip_prefix[lastAddrIndex].addr;
        if_pvt_info->ip_prefix[idx].prefixWidth =
            if_pvt_info->ip_prefix[lastAddrIndex].prefixWidth;
        if_pvt_info->ip_prefix_stats_id[idx] = if_pvt_info->ip_prefix_stats_id[lastAddrIndex];

        hos_rule_id = if_pvt_info->sec_addr_term_rule_id[0][idx];
        /* hos old impl */
        hos_rule_id_old = if_pvt_info->sec_addr_term_rule_id[1][idx];
        /* */

        if_pvt_info->sec_addr_term_rule_id[0][idx] =
            if_pvt_info->sec_addr_term_rule_id[0][lastAddrIndex];
        /* hos old impl */
        if_pvt_info->sec_addr_term_rule_id[1][idx] =
            if_pvt_info->sec_addr_term_rule_id[1][lastAddrIndex];
        /* */
    }

    if_pvt_info->ip_prefix[lastAddrIndex].addr = 0;
    if_pvt_info->ip_prefix[lastAddrIndex].prefixWidth = 0;
    if_pvt_info->ip_prefix_stats_id[lastAddrIndex] = 0;

    retValue = adk_hos_rule_delete(ADK_HOS_IP_TERM_TBL, hos_rule_id);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
             "Failed to remove IP-Term HOS rule for ip-intf sec address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_addr_rem_return;
    }
    
#ifdef ADK_HOS_OLD_IMPL
    retValue = adk_hos_rule_delete(ADK_HOS_INNER_IP4_TERM_TBL,
                                   hos_rule_id[1]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
           "Failed to remove Inner IP-Term HOS rule for ip-intf sec address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto adk_ip_interface_addr_rem_return;
    }
#endif

    --if_pvt_info->numSecAddr;
    --ipModuleInfo_p->aggrIpv4AddrCount;

    ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_interface_addr_rem: Removed address 0x%x/%d from interface 0x%lx\n", ipAddrPrefix_p->addr, ipAddrPrefix_p->prefixWidth,
            (uint64_t)ipIfHandle);

adk_ip_interface_addr_rem_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    
    return retValue;
}    /* -----  end of function adk_ip_interface_addr_rem ----- */

static adk_status_t adk_ip_create_egress_qos_resources(
       adk_module_handle_t ipHandle,
       uint32_t ifId)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;

    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

       if((ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle)) == NULL)
    {
	    ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
			    "NULL pointer parameter specified\n");
	    return ADK_E_INVALID_PARAMETER;
    }

    /* Allocate and initialize QoS egress classifier resources for this interface*/
    ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT] =
        (adk_ip_dp_egress_qos_classifier_action_info_t*)
        ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                ADK_IP_EGRESS_QOS_CLASSIFIER_INFO_OBJ, 1,
                ((ADK_IP_MAX_EGRESS_QOS_CLASSIFIER_RULES+ 1) *
                 sizeof(adk_ip_dp_egress_qos_classifier_action_info_t)),
                0);
    if(!ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT])
    {
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_create_egress_qos_resources_exit;
    }
    ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_SHADOW_TREE_INDEX_INIT] =
        (adk_ip_dp_egress_qos_classifier_action_info_t*)
        ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                ADK_IP_EGRESS_QOS_CLASSIFIER_INFO_OBJ, 1,
                ((ADK_IP_MAX_EGRESS_QOS_CLASSIFIER_RULES+ 1) *
                 sizeof(adk_ip_dp_egress_qos_classifier_action_info_t)),
                0);
    if(!ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_SHADOW_TREE_INDEX_INIT])
    {
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_create_egress_qos_resources_exit;
    }
    retValue = adk_ip_egress_qos_classifier_create(ipHandle, ifId, &(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId]));
    if (ADK_SUCCESS != retValue)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_create_egress_qos_resources_exit;
    }
    adk_sys_flow_cache_update_version(sys_handle);

    return retValue;

_adk_ip_create_egress_qos_resources_exit:
    if (ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT]) {
        ADK_DP_SHM_FREE(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT]);
    }
    if (ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_SHADOW_TREE_INDEX_INIT]) {
        ADK_DP_SHM_FREE(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_SHADOW_TREE_INDEX_INIT]);
    }
    return retValue;
}
        
static adk_status_t adk_ip_delete_egress_qos_resources(
       adk_module_handle_t ipHandle,
       uint32_t ifId)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    if((ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle))==NULL)
    {
    	return ADK_E_INTERNAL_ERROR;
    }

    if(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT])
    {
        ADK_DP_SHM_FREE(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT]);
    }
    if(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_SHADOW_TREE_INDEX_INIT])
    {
        ADK_DP_SHM_FREE(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][ADK_IP_QOS_SHADOW_TREE_INDEX_INIT]);
    }
    retValue = adk_ip_egress_qos_classifier_delete(ipHandle, ifId);
    if (ADK_SUCCESS != retValue)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Failed to delete egress QoS classifier context ");
        return ADK_E_INTERNAL_ERROR;
    }
    adk_sys_flow_cache_update_version(sys_handle);

    return retValue;
}

/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
