/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_interface_common.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs to manage IPv4 interfaces. This includes creation,
 * deletion of interfaces, setting their attributes, getting or printing
 * statistics for the interfaces, etc.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include <adk_ethernet_export.h>


/*!
 * =============================================================
 * adk_ip_interface_priority_map_profile_config
 *
 * @param[in] ipHandle          IP module handle.
 *
 * @param[in] ipIfHandle        IP interface handle with which the
 * profile ID needs to be associated with.
 *
 * @param[in] direction         The direction in which profile ID
 * needs to be associated with the interface.
 *
 * @param[in] profile_id        The profile ID with which the interface
 * needs to be associated with. Specifying this as 0 clears up existing
 * association.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          Invalid input parameter
 *
 * @retval ADK_IP_E_INVALID_IP_INTERFACE    Invalid IP interface handle passed
 *
 * @brief
 *
 * This API associates an IP interface with DSCP mapping profile ID in a given
 * direction. By specifying profile ID as 0, existing association for the
 * interface can be removed for the specified direction.
 *
 * Before deleting the IP interface, this API must be called to dissociate IP 
 * interface from the profile ID.
 *
 * The interface can be a regular IPoE interface.
 * =============================================================
 */
adk_status_t adk_ip_interface_priority_map_profile_config(
    adk_module_handle_t	ipHandle,
    adk_interface_handle_t ipIfHandle,
    adk_ip_direction_t direction,
    uint32_t profile_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_dp_interface_entry_t *dp_entry;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_interface_priority_map_profile_config_exit;
    }

    if (!(adk_ipv4_interface_valid_handle(ipIfHandle)) )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto _adk_ip_interface_priority_map_profile_config_exit;
    }

    if (direction != ADK_IP_DIR_INGRESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid or unsupported direction %d specified!\n",
                           direction);
        goto _adk_ip_interface_priority_map_profile_config_exit;
    }

    if (profile_id > ADK_QOS_CP_MAPPING_PROFILE_ID_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Profile ID %d out of valid range!\n", profile_id);
        goto _adk_ip_interface_priority_map_profile_config_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /*Update dp*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                 ipInterfaceTble[ADK_SYS_INTERFACE_ID_GET(ipIfHandle)]);
    dp_entry->ingress_qos_prof_id = profile_id;

_adk_ip_interface_priority_map_profile_config_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    return(retValue);
}

/*!
 * =============================================================
 * adk_ip_interface_mtu_set
 *
 * @param[in] ipHandle      IP module handle.
 *
 * @param[in] ipIfHandle    IP interface handle whose MTU needs to be set.
 *
 * @param[in] mtu           The new MTU to be set.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          Invalid input parameter
 *
 * @retval ADK_IP_E_INVALID_IP_INTERFACE    Invalid Layer-2 interface passed
 *
 * @brief
 *
 * This API sets MTU for an IPoE. The MTU specified must
 * be valid.
 * Note: The use of this API is limited only to IPoE interface.
 * =============================================================
 */
adk_status_t adk_ip_interface_mtu_set(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint16_t mtu)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    ip_interface_info_t     *if_pvt_info = NULL;
    adk_sys_handle_t sys_handle;
    adk_ip_dp_interface_entry_t *dp_entry;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_interface_mtu_set_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    
    /* Check validity of MTU */
    if ((mtu < IP_IF_MTU_MIN) || (mtu > IP_IF_MTU_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_MTU);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Invalid MTU Value %d\n", mtu);
        goto _adk_ip_interface_mtu_set_exit;
    }

    if (adk_ipv4_interface_valid_handle(ipIfHandle)) {
        if_pvt_info = (ip_interface_info_t *)
            ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);
        if_pvt_info->if_info.mtu = mtu;
    } else {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle or type\n");
        goto _adk_ip_interface_mtu_set_exit;
    }

    /*Update dp*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                 ipInterfaceTble[ADK_SYS_INTERFACE_ID_GET(ipIfHandle)]);
    dp_entry->mtu = mtu;
    adk_sys_flow_cache_update_version(sys_handle);


    /* Update the CP version for flow cache aging */
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[dp_entry->vrd_id])++;

_adk_ip_interface_mtu_set_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    

    return(retValue);
}



/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
