/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_ip_misc.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @brief
 *
 * This file contains internal IPv4 functions registered with SIF for overall
 * integration with system and other modules. This includes functions needed for
 * binding, IMC selection and other service functions needed by SIF.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"

/*!
 * =============================================================
 * @internal
 * adk_ip_get_imc
 *
 * @param[in]   ipHandle  Handle to the IP module
 * @param[in]   bindInfo_p Binding information
 * @param[out]  imcFnId IMC function ID to return
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 * @retval  ADK_E_IMC_NOT_FOUND     There is no IMC defined.
 *
 * @brief   Function to get the IMC function ID.
 *
 * This function is used to get the IMC function. This is called
 * from the ADK SIF. This should not be called directly from
 * the application.
 *
 * NOTE: that if there is no IMC function required, ADK_SUCCESS
 * should be returned.
 *
 * =============================================================
 */
adk_status_t adk_ip_get_imc (adk_module_handle_t  __attribute__((unused))  ipHandle,
                             void  __attribute__((unused))  *bindInfo_p,
                             uint32_t  __attribute__((unused))  *imcFnId)
{
    return ADK_SUCCESS;
}


const char *adk_ipv4_error_name(adk_status_t err)
{
    switch (err)
    {
    case ADK_IP_E_NAPT_FLOW_UNAVAILABLE:
        return("ADK_IP_E_NAPT_FLOW_UNAVAILABLE");
    case ADK_IP_E_INGRESS_STATE_MEMORY_UNAVAILABLE:
        return("ADK_IP_E_INGRESS_STATE_MEMORY_UNAVAILABLE");
    case ADK_IP_E_INGRESS_STATE_ID_UNAVAILABLE:
        return("ADK_IP_E_INGRESS_STATE_ID_UNAVAILABLE");
    case ADK_IP_E_EGRESS_STATE_MEMORY_UNAVAILABLE:
        return("ADK_IP_E_EGRESS_STATE_MEMORY_UNAVAILABLE");
    case ADK_IP_E_INVALID_INTERFACE_STATE:
        return("ADK_IP_E_INVALID_INTERFACE_STATE");
    case ADK_IP_E_INVALID_IP_VERSION:
        return("ADK_IP_E_INVALID_IP_VERSION");
    case ADK_IP_E_INVALID_SRC_IP_ADDRESS:
        return("ADK_IP_E_INVALID_SRC_IP_ADDRESS");
    case ADK_IP_E_INVALID_QOS_PROFILE_TYPE:
        return("ADK_IP_E_INVALID_QOS_PROFILE_TYPE");
    case ADK_IP_E_MAX_CLASS_CONFIG_EXCEEDED:
        return("ADK_IP_E_MAX_CLASS_CONFIG_EXCEEDED");
    case ADK_IP_E_INVALID_DSCP_PROFILE:
        return("ADK_IP_E_INVALID_DSCP_PROFILE");
    case ADK_IP_E_NAPT_NO_ADDRESS:
        return("ADK_IP_E_NAPT_NO_ADDRESS");
    case ADK_IP_E_NAPT_NO_OUT_L4PORT:
        return("ADK_IP_E_NAPT_NO_OUT_L4PORT");
    case ADK_IP_E_INVALID_ADDR_PREFIX:
        return("ADK_IP_E_INVALID_ADDR_PREFIX");
    case ADK_IP_E_FIB_UPDATE_FAILURE:
        return("ADK_IP_E_FIB_UPDATE_FAILURE");
    case ADK_IP_E_INVALID_IP_ADDRESS:
        return("ADK_IP_E_INVALID_IP_ADDRESS");
    case ADK_IP_E_INVALID_EXCEPTION_TYPE:
        return("ADK_IP_E_INVALID_EXCEPTION_TYPE");
    case ADK_IP_E_INVALID_MTU:
        return("ADK_IP_E_INVALID_MTU");
    case ADK_IP_E_INCONSISTENT_RATES:
        return("ADK_IP_E_INCONSISTENT_RATES");
    default:
        break;
    }

    return("ADK_E_IPV4_UNKNOWN");

}

adk_status_t adk_ipv4_map_sys_log(adk_module_handle_t ipHandle,
                                  adk_log_level_t sys_log_level,
                                  adk_log_level_t *ipv4_log_level)
{
    adk_status_t                    ret = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

#if 0   /* No conversion needed as the values are same */
    switch (sys_log_level) {
    case ADK_LOG_LEVEL_INFO:
        *ipv4_log_level = (adk_log_level_t)ADK_IP_LOG_LEVEL_INFO;
        break;
    case ADK_LOG_LEVEL_WARNING:
        *ipv4_log_level = (adk_log_level_t)ADK_IP_LOG_LEVEL_WARNING;
        break;
    case ADK_LOG_LEVEL_ERROR:
        *ipv4_log_level = (adk_log_level_t)ADK_IP_LOG_LEVEL_ERROR;
        break;
    case ADK_LOG_LEVEL_FATAL:
        *ipv4_log_level = (adk_log_level_t)ADK_IP_LOG_FATAL;
        break;
    default:
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_UNSUPPORTED,
                           "Unsupported log-level in IPv4!");
        ret = ADK_E_UNSUPPORTED;
        break;
    }
#endif

    *ipv4_log_level = sys_log_level;

    return(ret);
}

#if 0

/*!
 * =============================================================
 * @internal
 * adk_ip_interface_egress_dscp_pcp_map_enable
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ipIfHandle  - IP interface handle for which the
 * DSCP-to-PCP mapping needs to be enabled.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. Module
 * handle may be invalid.
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE- Invalid IP interface handle passed.
 *
 * @brief
 *
 * This API enables the DSCP-to-PCP mapping for the specified IP interface
 * at egress.
 *
 * Note that, the individual mappings may continue to exist but they will not
 * be used unless the mapping is enabled using this API.
 * =============================================================
 */
adk_status_t adk_ip_interface_egress_dscp_pcp_map_enable(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t if_egress_flags;
    ip_interface_info_t *if_pvt_info = NULL;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle 0x%08x\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_interface_egress_dscp_pcp_map_enable_exit;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);

    if_egress_flags = (if_pvt_info->if_egress_flags |
                       (1 << IP_IF_DSCP_ASSIGN_FLAG_OFFSET));

    /* Store the QoS attributes in the private data struture */
    if_pvt_info->if_egress_flags = if_egress_flags;

_adk_ip_interface_egress_dscp_pcp_map_enable_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return(retValue);
}

/*!
 * =============================================================
 * @internal
 * adk_ip_interface_egress_dscp_pcp_map_disable
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ipIfHandle  - IP interface handle for which the
 * DSCP-to-PCP mapping needs to be disabled.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. Module
 * handle may be invalid.
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE- Invalid IP interface handle passed.
 *
 * @brief
 *
 * This API disables the DSCP-to-PCP mapping for the specified IP interface
 * at egress.
 *
 * Note that, the individual mappings may continue to exist but they will not
 * be used after calling this API.
 * =============================================================
 */
adk_status_t adk_ip_interface_egress_dscp_pcp_map_disable(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t if_egress_flags;
    ip_interface_info_t *if_pvt_info = NULL;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle 0x%08x\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_interface_egress_dscp_pcp_map_disable_exit;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);

    if_egress_flags = (if_pvt_info->if_egress_flags &
                       ~(1 << IP_IF_DSCP_ASSIGN_FLAG_OFFSET));

    /* Store the QoS attributes in the private data struture */
    if_pvt_info->if_egress_flags = if_egress_flags;

_adk_ip_interface_egress_dscp_pcp_map_disable_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return(retValue);
}

/*!
 * =============================================================
 * @internal
 * adk_ip_interface_egress_dscp_pcp_map_add
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ipIfHandle  - IP interface handle.
 * @param[in]  ipIfDscpPcpMap_p - Pointer to the DSCP-to-PCP mappings
 * that need to be added for the IP interface.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. Module
 * handle may be invalid.
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE- Invalid IP interface handle passed.
 *
 * @brief
 *
 * This API adds the mapping from the DSCP to PCP for the egress IP interface.
 *
 * Before removing an IP interface, all DSCP-to-PCP mappings for the interface
 * must be removed.
 *
 * NOTE: THIS API IS UNDER WORK AND IS NOT OPERATIONAL AND SUBJECT
 * TO CHANGE. PLEASE DO NOT CALL THIS API WHICH CAN CORRUPT INTERNAL
 * STATE OF IPV4 MODULE.
 *
 * =============================================================
 */
adk_status_t adk_ip_interface_egress_dscp_pcp_map_add(
    adk_module_handle_t  __attribute__((unused))  ipHandle,
    adk_interface_handle_t  __attribute__((unused))  ipIfHandle,
    const adk_ip_interface_egress_dscp_pcp_map_t  __attribute__((unused))  *ipIfDscpPcpMap_p)
{
    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * @internal
 * adk_ip_interface_egress_dscp_pcp_map_delete
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ipIfHandle  - IP interface handle.
 * @param[in]  ipIfDscpPcpMap_p - Pointer to the DSCP-to-PCP mappings
 * that need to be removed for the IP interface.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. Module
 * handle may be invalid.
 *
 * @retval ADK_IP_E_INVALID_L2_INTERFACE- Invalid IP interface handle passed.
 *
 * @brief
 *
 * This API removes the mapping from the DSCP to PCP for the egress IP
 * interface.
 *
 * NOTE: THIS API IS UNDER WORK AND IS NOT OPERATIONAL AND SUBJECT
 * TO CHANGE. PLEASE DO NOT CALL THIS API WHICH CAN CORRUPT INTERNAL
 * STATE OF IPV4 MODULE.
 *
 * =============================================================
 */
adk_status_t adk_ip_interface_egress_dscp_pcp_map_delete(
    adk_module_handle_t  __attribute__((unused))  ipHandle,
    adk_interface_handle_t  __attribute__((unused))  ipIfHandle,
    const adk_ip_interface_egress_dscp_pcp_map_t  __attribute__((unused))  *ipIfDscpPcpMap_p)
{
    return ADK_SUCCESS;
}

#endif



