/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_packet_capture.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains packet capture APIs for
 * receiving or injecting PDU from CPU and vice versa.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include "adk_ph_dev_dp.h"
#include "adk_ethernet_export.h"

/*!
 * =============================================================
 * adk_ip_packet_capture_start
 *
 * @param[in] ipHandle          IP module handle.
 *
 * @param[in] pktCapInfo_p      Pointer to adk_ip_pktcap_info_t data
 * structure which defines the capture point and its attributes.
 *
 * @param[out] pktCapHandle     Pointer to output capture point handle.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IP module handle
 * specified
 *
 * @retval ADK_E_INVALID_PARAMETER          Invalid input parameter
 * passed e.g. NULL pointer parameter or invalid direction.
 *
 * @retval ADK_IP_E_INVALID_IP_INTERFACE    Invalid IP interface specified
 *
 * @retval ADK_E_MEM_UNAVAILABLE            Failed to allocate memory for
 * the packet capture point
 *
 * @brief
 *
 * This API create and starts a capture point. Once the capture point is
 * created, packets seen at the capture point are copied and sent to 
 * application via PH. The original packet continues to be processed 
 * normally while the copy is delivered to application. Note that,
 * the IP interface must be kept enabled for packets, received/sent 
 * from that IP interface, to get captured.
 *
 * Creating packet capture points affects the throughput of packet
 * crossing the capture point. That is, forwarding performance of the
 * original packets, that continue to be processes normally, is also hit.
 * This feature must be enabled for troubleshooting purposes only, keeping
 * this performance impact in consideration.
 *
 * Also note that, an application must not re-create capture point with
 * existing capture rule (with same or different attributes) without
 * first deleting the existing capture point with same capture rule.
 * =============================================================
 */
adk_status_t adk_ip_packet_capture_start(
    adk_module_handle_t ipHandle,
    const adk_ip_pktcap_info_t *pktCapInfo_p,
    adk_ip_pktcap_handle_t *pktCapHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    uint32_t l2IfId;
    adk_ip_direction_t dir = ADK_IP_DIR_INGRESS;
    ip_interface_info_t *if_pvt_info = NULL;
    adk_ip_dp_interface_entry_t *dp_entry;
    ip_pktcap_info_t *pktcap_info_p = NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    
    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_packet_capture_start_exit; 
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!pktCapInfo_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid NULL pointer parameter\n");
        goto _adk_ip_packet_capture_start_exit;
    }

    /* The interface must be regular IPoE only */
    if (adk_ipv4_interface_valid_handle(pktCapInfo_p->pktCapRule.ipIfHandle))
    {
        if_pvt_info = (ip_interface_info_t *)
            ADK_SYS_INTERFACE_ATTR_PTR_GET(pktCapInfo_p->pktCapRule.ipIfHandle);
    }
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_INTERFACE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid Interface Handle type\n");
        goto _adk_ip_packet_capture_start_exit;
    }

    l2IfId = ADK_SYS_INTERFACE_ID_GET(pktCapInfo_p->pktCapRule.ipIfHandle);

    if ((pktCapInfo_p->pktCapRule.direction != ADK_IP_DIR_INGRESS) &&
        (pktCapInfo_p->pktCapRule.direction != ADK_IP_DIR_EGRESS))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid direction specified");
        goto _adk_ip_packet_capture_start_exit;
    }

    dir = pktCapInfo_p->pktCapRule.direction;

    pktcap_info_p = (ip_pktcap_info_t *)
            ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                           ADK_IP_PKT_CAP_OBJ, 1, sizeof(ip_pktcap_info_t));
    if (pktcap_info_p == NULL)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                "Memory allocation failed\n");
        goto _adk_ip_packet_capture_start_exit;
    }

    /*update dp_entry*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                 ipInterfaceTble[l2IfId]);
#if 0
    dp_entry->pkt_cap_info[dir].devId = pktCapInfo_p->pktCapAttr.phAttr.devId;
    dp_entry->pkt_cap_info[dir].cbFnId = pktCapInfo_p->pktCapAttr.phAttr.cbFnId;
    dp_entry->pkt_cap_info[dir].queueId = pktCapInfo_p->pktCapAttr.phAttr.queueId;
#endif
    dp_entry->pkt_cap_info[dir].context = pktCapInfo_p->pktCapAttr.context;
    dp_entry->pkt_cap_info[dir].enableCapture = true;
    dp_entry->pkt_cap_info[dir].enableHwCapture = false;

#ifdef ADK_HOS_OLD_IMPL
    if(if_pvt_info &&
       ipModuleInfo_p->hw_pcap_enabled)
    {
        adk_module_handle_t eth_mod_handle=NULL;
        bool at_ingress = (dir == ADK_IP_DIR_INGRESS)? true: false;
        retValue = adk_module_handle_get(sys_handle, ADK_ETHERNET_MODULE_ID,
                                            NULL, &eth_mod_handle);
        if (retValue == ADK_SUCCESS)
        {
            retValue = adk_eth_intf_pcap_config(eth_mod_handle,
                        if_pvt_info->if_info.layer2InterfaceHandle,
                        false, at_ingress, ipModuleInfo_p->hos_pcap_id, true);
            if(retValue == ADK_SUCCESS)
            {
                dp_entry->pkt_cap_info[dir].enableHwCapture = true;
            }
            else
            {
                ADK_IPV4_LOG_WARNING(ipHandle,
                                     "Failed enable H/W pkt capture");
            }
        }
        else
        {
            ADK_IPV4_LOG_WARNING(ipHandle,
                                 "Ethernet module handle not found!\n");
        }
    }
#endif

    /* Store the packet capture info in the internal data structure */
    ADK_MEMCPY((void *) &(pktcap_info_p->pktcap_info),
                   (const void *) pktCapInfo_p,
                    sizeof(adk_ip_pktcap_info_t));
    pktcap_info_p->cookie = ADK_IPV4_PKT_CAP_COOKIE;

    *pktCapHandle = (adk_ip_pktcap_handle_t) pktcap_info_p;
         
_adk_ip_packet_capture_start_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * adk_ip_packet_capture_stop
 *
 * @param[in] ipHandle          IP module handle.
 *
 * @param[out] pktCapHandle     Capture point handle to be stopped and
 * destroyed
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IP module handle
 * or capture point handle
 *
 * @brief
 *
 * This API stops and deletes specified capture point. Once the capture
 * point has been deleted using this API, the packets will not be captured.
 * This API must be called for each capture point when capturing is not
 * needed anymore.
 * =============================================================
 */
adk_status_t adk_ip_packet_capture_stop(
    adk_module_handle_t ipHandle,
    adk_ip_pktcap_handle_t pktCapHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    ip_pktcap_info_t *pktcap_info_p = NULL;
    ip_module_info_t *ipModuleInfo_p;
    uint32_t l2IfId;
    adk_ip_direction_t dir = ADK_IP_DIR_INGRESS;
    adk_ip_dp_interface_entry_t *dp_entry;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_packet_capture_stop_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    pktcap_info_p = (ip_pktcap_info_t *) pktCapHandle;

    if ((NULL == pktcap_info_p) ||
        (pktcap_info_p->cookie != ADK_IPV4_PKT_CAP_COOKIE))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid packet capture handle\n");
        goto _adk_ip_packet_capture_stop_exit;
    }

    l2IfId = ADK_SYS_INTERFACE_ID_GET(
                pktcap_info_p->pktcap_info.pktCapRule.ipIfHandle);
    dir = pktcap_info_p->pktcap_info.pktCapRule.direction;
    
    /*update dp_entry*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                 ipInterfaceTble[l2IfId]);
    dp_entry->pkt_cap_info[dir].enableCapture = FALSE;
    pktcap_info_p->cookie = 0;

    adk_rcu_synchronize();

#ifdef ADK_HOS_OLD_IMPL
    if(dp_entry->pkt_cap_info[dir].enableHwCapture)
    {
        adk_module_handle_t eth_mod_handle=NULL; 
        bool at_ingress = (dir == ADK_IP_DIR_INGRESS)? true: false;
        retValue = adk_module_handle_get(sys_handle, ADK_ETHERNET_MODULE_ID,
                                            NULL, &eth_mod_handle);
        if (retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                            "Ethernet module handle not found!\n");
            retValue = ADK_E_INTERNAL_ERROR;
            goto _adk_ip_packet_capture_stop_exit;
        }
        retValue = adk_eth_intf_pcap_config(eth_mod_handle,
                    if_pvt_info->if_info.layer2InterfaceHandle,
                    false, at_ingress, ipModuleInfo_p->hos_pcap_id, false);
        if(retValue != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Failed disable H/W pkt capture");
            goto _adk_ip_packet_capture_stop_exit;
        }
    }
#endif

    ADK_NVM_FREE(pktcap_info_p);
_adk_ip_packet_capture_stop_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue); 
}

/*!
 * =============================================================
 * adk_ip_packet_capture_status_get
 *
 * @param[in] ipHandle          IP module handle.
 *
 * @param[out] pktCapHandle     Capture point handle whose status
 * needs to be retrieved
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IP module handle
 * or capture point handle
 *
 * @retval ADK_E_INVALID_PARAMETER          Invalid input parameter
 * passed e.g. NULL pointer parameter or invalid direction.
 *
 * @brief
 *
 * This API gets the status of a capture point. The status tells us
 * if the capture has been autonomously disabled by ADK or not. This
 * automatic disabling by ADK is done when packets cannot be buffered
 * for delivery to application due to lack of space in the buffer area.
 * Once a capture point enters this state, it remains in this state
 * until application re-enables the packet capture. 
 *
 * Note that, a disabled capture point is not same as stopped capture
 * point. In disabled state, the capture point still exist. For capture
 * points in disabled state, application can either stop the capture
 * (thereby, remove/delete the capture point) or re-enable the capturing
 * using the adk_ip_packet_capture_enable() API.
 *
 * From the time the packet capture was disabled by ADK autonomously
 * to the time when it is enabled again by application, there may have
 * been packets which were captured but were not delivered to application.
 * These packets are not buffered by ADK but instead dropped. And thus,
 * there may exist discontinuity in the captured packet flow during
 * this transition.
 * =============================================================
 */
adk_status_t adk_ip_packet_capture_status_get(
    adk_module_handle_t ipHandle,
    adk_ip_pktcap_handle_t pktCapHandle,
    adk_ip_pktcap_status_t *pktCapStatus_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_pktcap_info_t *pktcap_info_p = NULL;
    adk_sys_handle_t sys_handle;
    adk_ip_dp_interface_entry_t *dp_entry;
    ip_module_info_t *ipModuleInfo_p;
    adk_ip_direction_t dir;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_packet_capture_status_get_exit;
    }

    if (!pktCapStatus_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid NULL pointer parameter\n");
        goto _adk_ip_packet_capture_status_get_exit;
    }

    pktcap_info_p = (ip_pktcap_info_t *) pktCapHandle;
    if ((NULL == pktcap_info_p) ||
        (pktcap_info_p->cookie != ADK_IPV4_PKT_CAP_COOKIE))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                   "Invalid packet capture handle\n");
        goto _adk_ip_packet_capture_status_get_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    dir = pktcap_info_p->pktcap_info.pktCapRule.direction;
    
    /*Get dp_entry*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                 ipInterfaceTble[ADK_SYS_INTERFACE_ID_GET(
                 pktcap_info_p->pktcap_info.pktCapRule.ipIfHandle)]);
    
    /*Update Packet capture status*/
    pktCapStatus_p->is_enabled =  dp_entry->pkt_cap_info[dir].enableCapture;

_adk_ip_packet_capture_status_get_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ip_packet_capture_enable
 *
 * @param[in] ipHandle          IP module handle.
 *
 * @param[out] pktCapHandle     Capture point handle to be enabled
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IP module handle
 * or capture point handle
 *
 * @brief
 * This API re-enables a capture point which has been disabled
 * autonomously by ADK. Please refer to the description of
 * adk_ip_packet_capture_status_get() API for more details on this.
 * =============================================================
 */
adk_status_t adk_ip_packet_capture_enable(
    adk_module_handle_t  ipHandle,
    adk_ip_pktcap_handle_t pktCapHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_pktcap_info_t *pktcap_info_p = NULL;
    adk_sys_handle_t sys_handle;
    adk_ip_dp_interface_entry_t *dp_entry;
    ip_module_info_t *ipModuleInfo_p;
    adk_ip_direction_t dir;
    
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_packet_capture_enable_exit;
    }

    pktcap_info_p = (ip_pktcap_info_t *) pktCapHandle;

    if ((NULL == pktcap_info_p) ||
        (pktcap_info_p->cookie != ADK_IPV4_PKT_CAP_COOKIE))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid packet capture handle\n");
        goto _adk_ip_packet_capture_enable_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    dir = pktcap_info_p->pktcap_info.pktCapRule.direction;
    
    /*Get dp_entry*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                 ipInterfaceTble[ADK_SYS_INTERFACE_ID_GET(
                 pktcap_info_p->pktcap_info.pktCapRule.ipIfHandle)]);
    
    /*Enable Packet capture */
    dp_entry->pkt_cap_info[dir].enableCapture = TRUE;

_adk_ip_packet_capture_enable_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

#if 0
/*!
 * =============================================================
 * adk_ip_cpu_recv_task_param
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[in]   task_params_p Pointer to 32-byte task parameter array.
 *
 * @param[out]  ip_cpu_params_p Pointer to structure having the decoded
 *                              parameters sent by IPv4 fast path code to CPU.
 *
 * @retval   ADK_SUCCESS - successfully unregisterd exception
 * handler
 *
 * @retval  ADK_E_INVALID_HANDLE -  Invalid IPv4 module handle specified or
 *                                  non IPv4 task parameters received
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 *                                       passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function helps in interpreting the parameters present in the task
 * parameters that IPv4 module send along with the packet to the CPU via PH.
 * This API is used for PHv1 and will be deprecated later.
 * =============================================================
 */
adk_status_t adk_ip_cpu_recv_task_param(
    adk_module_handle_t ipHandle,
    const uintptr_t *task_params_p,
    adk_ip_cpu_recv_param_t *ip_cpu_params_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    uint32_t pduParam;
    uint8_t type;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    if (!task_params_p || !ip_cpu_params_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid NULL parameter\n");
        goto _adk_ip_cpu_recv_task_param_exit;
    }

    /* Zero out the returned param structure before starting to fill */
    ADK_MEMSET(ip_cpu_params_p, 0x0, sizeof(adk_ip_cpu_recv_param_t));
    
    pduParam = ADK_CONVERT_TPARAMS_TO_UINT32(ADK_PH_PARAMS(task_params_p, 2));
    ip_cpu_params_p->port = (adk_port_number_t)(pduParam >> 16);
    ip_cpu_params_p->interface_id = pduParam & 0xFFFF;
    ip_cpu_params_p->linux_intf_id = ip_cpu_params_p->port;

    pduParam = ADK_CONVERT_TPARAMS_TO_UINT32(ADK_PH_PARAMS(task_params_p, 4));
    ip_cpu_params_p->context = pduParam;

    pduParam = ADK_CONVERT_TPARAMS_TO_UINT32(ADK_PH_PARAMS(task_params_p, 3));
    ip_cpu_params_p->ip_header_offset = (pduParam >> 16) & 0xFF;
    type = (pduParam >> 8) & 0xFF;
    switch(type)
    {
    case ADK_IP_OPTIONS_EXCEPTION:
        ip_cpu_params_p->exception_type = ADK_IP_OPTIONS_EXCEPTION;
        break;
    case ADK_IP_FRAGMENTATION_EXCEPTION:
        ip_cpu_params_p->exception_type = ADK_IP_FRAGMENTATION_EXCEPTION;
        break;
    case ADK_IP_NO_ROUTE_EXCEPTION:
        ip_cpu_params_p->exception_type = ADK_IP_NO_ROUTE_EXCEPTION;
        break;
    case ADK_IP_TTL_EXPIRED_EXCEPTION:
        ip_cpu_params_p->exception_type = ADK_IP_TTL_EXPIRED_EXCEPTION;
        break;
    case IP_EXC_CODE_BAD_HDR:
        ip_cpu_params_p->exception_type = ADK_IP_BAD_HDR_EXCEPTION;
        break;
    default:
        ip_cpu_params_p->exception_type = ADK_IP_NO_EXCEPTION;
        break;
    }

_adk_ip_cpu_recv_task_param_exit:

    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ip_ph_dev_rx_param_decode 
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[in]   pduMetaData_p Pointer to rte_mbuf.
 *
 * @param[out]  ip_params_p   Pointer to structure having the decoded parameters
 *                            sent by IPv4 data path code to application.
 *
 * @retval   ADK_SUCCESS - successfully unregisterd exception
 * handler
 *
 * @retval  ADK_E_INVALID_HANDLE -  Invalid IPv4 module handle specified or
 *                                  non IPv4 task parameters received
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 *                                    passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function helps in interpreting the parameters present in the task
 * parameters that IPv4 module send along with the packet to the Application 
 * via PH Dev.
 *
 * =============================================================
 */
adk_status_t
adk_ip_ph_dev_rx_param_decode (
    adk_module_handle_t ipHandle,
    const void *pduMetaData_p,
    adk_ip_ph_dev_rx_param_t *ip_params_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    struct rte_mbuf *mb_p;
    adk_ph_dev_dp_metadata_t *phMeta_p;
    ip_ph_metadata_t *ph_metadata;
    adk_metadata_t *adk_md_ptr;
    uint8_t type;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    if (!pduMetaData_p || !ip_params_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid NULL parameter\n");
        goto _adk_ip_ph_dev_rx_param_decode_exit;
    }

    mb_p = (struct rte_mbuf *)pduMetaData_p;
    adk_md_ptr = ADK_METADATA_PTR(mb_p);
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(adk_md_ptr->imc_param_start);
    ph_metadata = (ip_ph_metadata_t *)&(phMeta_p->imc_param_start);
    /* Zero out the returned param structure before starting to fill */
    ADK_MEMSET(ip_params_p, 0x0, sizeof(adk_ip_ph_dev_rx_param_t));
    
    if(adk_md_ptr->exception_domain == ADK_TN_PORT_DOMAIN)
    {
        ip_params_p->port = (adk_port_number_t)ADK_PACKET_INPUT_PORT(mb_p);
        ip_params_p->linux_intf_id = ADK_MD_GET_LINUX_INTF_ID(mb_p);
        ip_params_p->interface_id = adk_md_ptr->input_intf_id;
    }
    else
    {
        ip_params_p->port = 0;
        ip_params_p->linux_intf_id = 0;
        ip_params_p->interface_id = 0;
    }
    ip_params_p->context = adk_md_ptr->pdu_flow_id;
    ip_params_p->ip_header_offset = ADK_MD_GET_IP_OFF(mb_p);

    ip_params_p->mtu =  ph_metadata->mtu;
    type = ph_metadata->exception_type;
    switch(type)
    {
    case ADK_IP_OPTIONS_EXCEPTION:
        ip_params_p->exception_type = ADK_IP_OPTIONS_EXCEPTION;
        break;
    case ADK_IP_FRAGMENTATION_EXCEPTION:
        ip_params_p->exception_type = ADK_IP_FRAGMENTATION_EXCEPTION;
        break;
    case ADK_IP_NO_ROUTE_EXCEPTION:
        ip_params_p->exception_type = ADK_IP_NO_ROUTE_EXCEPTION;
        break;
    case ADK_IP_TTL_EXPIRED_EXCEPTION:
        ip_params_p->exception_type = ADK_IP_TTL_EXPIRED_EXCEPTION;
        break;
    case IP_EXC_CODE_BAD_HDR:
        ip_params_p->exception_type = ADK_IP_BAD_HDR_EXCEPTION;
        break;
    default:
        ip_params_p->exception_type = ADK_IP_NO_EXCEPTION;
        break;
    }

_adk_ip_ph_dev_rx_param_decode_exit:

    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

#endif
/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
