/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_pbr.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs to manage Policy Based Routing (PBR) policies. This
 * includes enabling or disabling PBR for locally originated traffic or for an
 * interface, adding or deleting PBR rules, committing PBR rules into hardware,
 * etc.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"

#define IP_MAX_PBR_RULE_PRIORITY 0xFFFF

/*!
 * =============================================================
 * @internal
 * adk_ip_pbr_dpdk_sw_rule
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief
 *
 * Internal support function for PBR ACL rule addition.
 *
 * =============================================================
 */
static adk_status_t
adk_ip_pbr_dpdk_sw_rule (
    adk_module_handle_t ipHandle,
    adk_ip_pbr_tree_t *tree,
    uint32_t precedence,
    uint32_t pbrRuleId,
    const adk_ip_pbr_rule_t *pbrRules_p,
    adk_ip_pbr_action_t *pbrAction_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ip_pbr_acl_rule_t pbr_rule;
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ip_module_info_t *ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    int ret;
    adk_ip_dp_pbr_action_t *action = NULL;
    /*****   Build the Individual Patterns ******/
    ADK_IPV4_LOG_INFO(ipHandle,"\tIPv4 PBR - building rule: pbrRuleId=%d\n", pbrRuleId);
    memset(&pbr_rule, 0, sizeof(pbr_rule));

    if (pbrRules_p->pbrRuleMask & ADK_IP_PBR_SRC_ADDR_MASK)
    {
        pbr_rule.sip_addr = pbrRules_p->srcAddrInfo.addr;
        pbr_rule.sip_mask = pbrRules_p->srcAddrInfo.prefixWidth;
    }
    else
    {
        /* Set ANY */
        pbr_rule.sip_addr = 0;
        /* Add an invalid value so that adk_ip_pbr_rule_add will set the default */
        pbr_rule.sip_mask = 0xFF;
    }

    ADK_IPV4_LOG_INFO(ipHandle,"\t\tsip=0x%x/%d\n", pbr_rule.sip_addr, pbr_rule.sip_mask);
    /* PBR rule add will not touch the lookup DB.
     * Hence add rule to both active & shadow copy.
     * TO DO: Clean up the return value .. its confusing
     */
    ret = adk_ip_pbr_rule_add(ipHandle, tree,
       &pbr_rule, (adk_ip_pbr_val_t)(uintptr_t)pbrRuleId, precedence);
    if(!ret)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error inserting PBR rule in IP\n");
        goto _adk_ip_pbr_dpdk_sw_rule_exit;
    }

    action = &(ipModuleInfo_p->ip_dp_tbl_info_p->pbrActionTable_p[pbrRuleId]);
    memset(action, 0, sizeof(adk_ip_dp_pbr_action_t));
    *action =  *pbrAction_p;

_adk_ip_pbr_dpdk_sw_rule_exit:

    ADK_RETURN(retValue);

}    /* -----  end of function adk_ip_pbr_dpdk_sw_rule  ----- */

/*!
* =============================================================
* adk_ip_pbr_configure_default_route_table
*
* @param[in] ipHandle IP module handle. Returned at the
* time of IP module's registration with the module manager.
*
* @param[in] defRtTbl - Default Route table Id
*
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_IP_E_INVALID_PARAMETER - Invalid input parameter.
* The pointer might be NULL.
*

 * @retval ADK_IP_E_INVALID_ROUTE_TABLE_ID - Invalid route table Id.
*
* @brief IP API to program default route table Id.
*
* This API programs the default route table Id
 * All packets which do not match to any specific policy
 * configured by the application always match to default policy
*. Such packets are processed based on action configured
* for this default policy given by input parameter defRtTbl
* This defRtTbl  is used in below 2 cases:
* 1. when a route is not present in route table Id given by specific policy
* 2. when packet doesn't match the configured PBR policy.
*
* Note: In the absense of explicit default value configured
* through this API, implicit default table id of 0 will be used.
* =============================================================
*/
adk_status_t
adk_ip_pbr_configure_default_route_table (
        adk_module_handle_t ipHandle,
        uint8_t  defRtTbl)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IP module handle \n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipHandle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /*Validate Ip private data and cookie */
    if(!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipHandle,
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ip_pbr_configure_default_route_table_return;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (defRtTbl > ADK_IP_MAX_ROUTING_TABLES)
    {
        ADK_LOG_ERROR(sys_handle, ipHandle,
            " Invalid default route table ID %u\n", defRtTbl);
        goto adk_ip_pbr_configure_default_route_table_return;
    }

    ipModuleInfo_p->ip_dp_tbl_info_p->defRouteTblId = defRtTbl;
    adk_sys_flow_cache_update_version(sys_handle);

    ADK_LOG_INFO(sys_handle, ipHandle,"Successfully confgirued def route table : %u \n",
            ipModuleInfo_p->ip_dp_tbl_info_p->defRouteTblId);

adk_ip_pbr_configure_default_route_table_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipHandle);

    ADK_RETURN(retValue);
} /* end of adk_ip_pbr_configure_default_route_table */


adk_status_t adk_ip_pbr_create(
        adk_module_handle_t  ipHandle,
        adk_ip_pbr_ctx_t  **ipPbrInfo_pp)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ip_pbr_ctx_t *pbrInfo_p = NULL;
    uint16_t  actionIndexDp;
    adk_acl_info_t cpListInfo={0};
    adk_ip_pbr_rule_t pbrRule;
    adk_ip_pbr_action_t pbrAction;

    ip_module_info_t * ipModuleInfo_p =
            (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    pbrInfo_p = (adk_ip_pbr_ctx_t *)
                           ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_PBR_POLICY_OBJ, 1,
                                   sizeof(adk_ip_pbr_ctx_t));

    if (pbrInfo_p == NULL)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                "Memory allocation failed\n");
        return ADK_E_MEM_UNAVAILABLE;
    }

    *ipPbrInfo_pp = pbrInfo_p;

    actionIndexDp = adk_rm_resource_reserve(ipModuleInfo_p->ipPbrPolicyRuleIdHandle);
    if (1 != actionIndexDp)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_RESOURCE_UNAVAILABLE,
                "Could not get resource id PBR using rte_acl_create!\n");
        return ADK_E_RESOURCE_UNAVAILABLE;
    }

    pbrInfo_p->cookie = ADK_IPV4_POLICY_COOKIE;
    /* Create a shadow copy as well for RCU */
    pbrInfo_p->pbr_tree_index = 0;
    pbrInfo_p->pbr_tree[0] = adk_ip_pbr_acl_create(ipHandle, 0);
    pbrInfo_p->pbr_tree[1] = adk_ip_pbr_acl_create(ipHandle, 1);
    if((pbrInfo_p->pbr_tree[0] == NULL) ||
            (pbrInfo_p->pbr_tree[1] == NULL))
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_INTERNAL_ERROR,
                "Could not allocate IP PBR policy!\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_pbr_create_exit;
    }

    /*Create CP list*/
    cpListInfo.user = ADK_ACL_IPv4;
    if(adk_acl_create(&(pbrInfo_p->cp_list), &cpListInfo) != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                "Could not allocate CP list for IP PBR policy!\n");
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ip_pbr_create_exit;
    }

    /*
     * Add default PBR rule, to prevent crash in SW ACL
     * when try to do ACL lookup before commit
     */
    memset(&pbrRule, 0, sizeof(pbrRule));
    pbrAction.routeTableId = ADK_IP_MAX_ROUTING_TABLES;
    retValue = adk_ip_pbr_dpdk_sw_rule(ipHandle, pbrInfo_p->pbr_tree[0],
            RTE_ACL_MAX_PRIORITY,
            actionIndexDp,
            &pbrRule,
            &pbrAction);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Failed to add the rule in dpdk s/w\n");
        goto _adk_ip_pbr_create_exit;
    }

    /*Update both copies*/
    retValue = adk_ip_pbr_ruleset_commit(ipHandle, pbrInfo_p->pbr_tree[0]);
    if(!retValue)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not commit PBR ACL rules...\n");
        goto _adk_ip_pbr_create_exit;
    }

    memset(&pbrRule, 0, sizeof(pbrRule));
    pbrAction.routeTableId = ADK_IP_MAX_ROUTING_TABLES;
    retValue = adk_ip_pbr_dpdk_sw_rule(ipHandle, pbrInfo_p->pbr_tree[1],
            RTE_ACL_MAX_PRIORITY,
            actionIndexDp,
            &pbrRule,
            &pbrAction);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Failed to add the rule in dpdk s/w\n");
        goto _adk_ip_pbr_create_exit;
    }
    retValue = adk_ip_pbr_ruleset_commit(ipHandle, pbrInfo_p->pbr_tree[1]);
    if(!retValue)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not commit PBR ACL rules...\n");
        goto _adk_ip_pbr_create_exit;
    }
    retValue = ADK_SUCCESS;
_adk_ip_pbr_create_exit:
if (retValue != ADK_SUCCESS)
    {
        if (pbrInfo_p != NULL)
        {
            if(pbrInfo_p->pbr_tree[0])
                adk_ip_pbr_ruleset_free(ipHandle, pbrInfo_p->pbr_tree[0]);

            if(pbrInfo_p->pbr_tree[1])
                adk_ip_pbr_ruleset_free(ipHandle, pbrInfo_p->pbr_tree[1]);

            if(pbrInfo_p->cp_list)
                adk_acl_free(pbrInfo_p->cp_list);

            adk_rm_resource_free(ipModuleInfo_p->ipPbrPolicyRuleIdHandle, 1);

            ADK_NVM_FREE(pbrInfo_p);
        }
    }

    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

adk_status_t adk_ip_pbr_delete(
        adk_module_handle_t  ipHandle,
        adk_ip_pbr_ctx_t  **ipPbrInfo_pp)
{
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_pbr_ctx_t  *ipPbrInfo_p = *ipPbrInfo_pp;
    
    if (!adk_ipv4_valid_handle(ipHandle))
    {
	    ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
	    return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
    }

    if(ipPbrInfo_p->pbr_tree_index)
    {
        /* Free the shadow copy */
        adk_ip_pbr_ruleset_free(ipHandle, ipPbrInfo_p->pbr_tree[0]);
        /* Wait for readers to complete */
        adk_rcu_synchronize();
        /* Free the active copy */
        adk_ip_pbr_ruleset_free(ipHandle, ipPbrInfo_p->pbr_tree[1]);
    }
    else
    {
        /* Free the shadow copy */
        adk_ip_pbr_ruleset_free(ipHandle, ipPbrInfo_p->pbr_tree[1]);
        /* Wait for readers to complete */
        adk_rcu_synchronize();
        /* Free the active copy */
        adk_ip_pbr_ruleset_free(ipHandle, ipPbrInfo_p->pbr_tree[0]);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (ipModuleInfo_p == NULL)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "NULL pointer parameter specified\n");
        return ADK_E_INVALID_PARAMETER;
    }



    adk_acl_free(ipPbrInfo_p->cp_list);

    /* Clear the rule set information and free the internal data structure */
    ipPbrInfo_p->cookie = 0;
    ADK_NVM_FREE(ipPbrInfo_p);
    adk_rm_resource_free(ipModuleInfo_p->ipPbrPolicyRuleIdHandle, 1);
    *ipPbrInfo_pp = NULL;

    return ADK_SUCCESS;
} /* end of adk_ip_pbr_delete */


adk_status_t
adk_ip_pbr_add_policy (
        adk_module_handle_t ipHandle,
        uint32_t ruleIndex,
        adk_ip_pbr_info_t *pbrInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    ip_module_info_t *ipModuleInfo_p = NULL;
    adk_ip_pbr_ctx_t * pbrCtx_p = NULL;
    adk_flow_handle_t   ruleHandle_p = 0;
    uint32_t pbrRuleId = 0, i;
    ip_flow_info_t * flow_info_p = NULL;
    adk_acl_rule_t *cpRule_p = NULL;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /* Validate input args */
    if (!pbrInfo_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ipHandle,
                "Invalid NULL input parameter\n");
        goto _adk_ip_pbr_add_policy_exit;
    }

    if ((pbrInfo_p->pbrRule.pbrRuleMask & ADK_IP_PBR_SRC_ADDR_MASK) &&
               ((pbrInfo_p->pbrRule.srcAddrInfo.prefixWidth == 0) ||
                (pbrInfo_p->pbrRule.srcAddrInfo.prefixWidth > 32)))
       {
           retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
           ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                   "Invalid Source IP address prefix width: %d\n", pbrInfo_p->pbrRule.srcAddrInfo.prefixWidth);
           goto _adk_ip_pbr_add_policy_exit;
       }


    if (pbrInfo_p->pbrAction.routeTableId >= ADK_IP_MAX_ROUTING_TABLES)
    {
        ADK_LOG_ERROR(sys_handle, ipHandle,
                " Invalid route table ID %u\n", pbrInfo_p->pbrAction.routeTableId);
        goto _adk_ip_pbr_add_policy_exit;
    }
    /* -- */

    pbrCtx_p = (adk_ip_pbr_ctx_t *) ipModuleInfo_p->ip_dp_tbl_info_p->ipPbrInfo_p;

    /* Clearing to indicate no handle is allocated yet */
    ruleHandle_p = 0x0;

    ADK_SYS_FLOW_HANDLE_ALLOC(ipHandle, ADK_IP_PBR_RULE_OBJ, sizeof(ip_flow_info_t),
            &ruleHandle_p);
    if (ruleHandle_p == 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sys_handle, ipHandle,
                "Failed to allocate memory for PBR rule\n");
        goto _adk_ip_pbr_add_policy_exit;
    }

    /* Allocating rule id for each rule*/
    pbrRuleId = adk_rm_resource_reserve(ipModuleInfo_p->ipPbrPolicyRuleIdHandle);
    if (pbrRuleId == ADK_RM_RSRC_INVALID)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_MEM_UNAVAILABLE,
                "Failed to allocate id for PBR rule\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _adk_ip_pbr_add_policy_exit;
    }

    flow_info_p = (ip_flow_info_t *) ADK_SYS_FLOW_ATTR_PTR_GET(ruleHandle_p);

    flow_info_p->cookie = ADK_IPV4_FLOW_COOKIE;
    flow_info_p->type = IP_FLOW_TYPE_PBR_RULE;
    ADK_SYS_FLOWID_SET(ruleHandle_p, pbrRuleId);

    flow_info_p->u.pbr_flow.rule_index = ruleIndex;

    ADK_MEMCPY((void *) &(flow_info_p->u.pbr_flow.rule),
            (const void *) &(pbrInfo_p->pbrRule),
            sizeof(adk_ip_pbr_rule_t));
    ADK_MEMCPY((void *) &(flow_info_p->u.pbr_flow.action),
            (const void *) &(pbrInfo_p->pbrAction),
            sizeof(adk_ip_pbr_action_t));

    /*
     * Add rule to ADK CP rule list.
     * The rules will be added during the next commit.
     */
    cpRule_p = &(flow_info_p->u.pbr_flow.cp_rule);
    cpRule_p->user_data = (void *)(ruleHandle_p);
    retValue = adk_acl_rule_add(pbrCtx_p->cp_list,
            ruleIndex,
            cpRule_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipHandle,
                "Failed to add the rule in CP list\n");
        goto _adk_ip_pbr_add_policy_exit;
    }

    printf("add_policy: ruleHandle_p=%p, ruleId=%d\n", ruleHandle_p, ADK_SYS_FLOWID_GET(ruleHandle_p));
    /* Set the statistics for this rule to 0 */
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        ipModuleInfo_p->ip_dp_tbl_info_p->pbrRuleStats[i*ADK_IP_MAX_PBR_RULES+pbrRuleId-1] = 0;
    }

_adk_ip_pbr_add_policy_exit:

/* Upon error, release the reserved resources */
    if (retValue != ADK_SUCCESS) {
        if (ruleHandle_p)
            ADK_SYS_FLOW_HANDLE_FREE(ruleHandle_p);

    }

    if (ADK_RM_RSRC_INVALID == pbrRuleId)
    {
        adk_rm_resource_free(ipModuleInfo_p->ipPbrPolicyRuleIdHandle, pbrRuleId);
    }
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

} /* end of adk_ip_pbr_add_policy */



adk_status_t
adk_ip_pbr_del_policy (
        adk_module_handle_t ipHandle,
        uint32_t  ruleIndex)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p = NULL;
    adk_sys_handle_t sys_handle;
    adk_ip_pbr_ctx_t * pbrCtx_p = NULL;
    uint32_t pbrRuleId = ADK_RM_RSRC_INVALID, idx, index=0;
    adk_acl_rule_t cpRule;
    adk_flow_handle_t   ruleHandle_p = 0;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    pbrCtx_p = (adk_ip_pbr_ctx_t *) ipModuleInfo_p->ip_dp_tbl_info_p->ipPbrInfo_p;

    /* Make sure we are not deleting more than we can hold for commit */
    if (pbrCtx_p->num_pending_rule_ids >= ADK_IP_MAX_PBR_RULES)
    {
        ADK_LOG_ERROR(sys_handle, ipHandle,
                "Cannot delete any more rules until previously"
                " deleted or modified rules are committed\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_UNSUPPORTED);
        goto _adk_ip_pbr_del_policy_exit;
    }

    adk_acl_get_reset(pbrCtx_p->cp_list);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(pbrCtx_p->cp_list, &cpRule))
    {
        if(index==ruleIndex)
        {
            ruleHandle_p = (adk_flow_handle_t)cpRule.user_data;
            pbrRuleId = ADK_SYS_FLOWID_GET(ruleHandle_p);
            break;
        }
        index++;
    }

    /* Make sure, the rule is not in commit pending list */
    for (idx = 0; idx < pbrCtx_p->num_pending_rule_ids; idx++)
    {
        if (pbrCtx_p->pendingRuleIds[idx] == pbrRuleId)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_UNSUPPORTED);
            ADK_LOG_ERROR(sys_handle, ipHandle,
                    "Cannot delete a rule with pending commit\n");
            goto _adk_ip_pbr_del_policy_exit;
        }
    }

    if (ADK_RM_RSRC_INVALID != pbrRuleId)
    {
        retValue = adk_acl_rule_delete(pbrCtx_p->cp_list,
                ruleIndex);
        if (retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipHandle,
                    "Failed to delete the rule from CP list\n");
            goto _adk_ip_pbr_del_policy_exit;
        }

        adk_rm_resource_free(ipModuleInfo_p->ipPbrPolicyRuleIdHandle, pbrRuleId);
        /*
         * There are no PBR rule delete APIs. Hence remove from ADK CP rule list,
         * and during the commit, flush all the entries & add the rules from the ADK
         * CP rule list.
         */

        pbrCtx_p->pendingRuleIds[idx] = pbrRuleId;
        pbrCtx_p->num_pending_rule_ids++;
    }

    if (ruleHandle_p)
        ADK_SYS_FLOW_HANDLE_FREE(ruleHandle_p);

_adk_ip_pbr_del_policy_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

    _lock_release_failed:
    _lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipHandle);
    ADK_RETURN(retValue);

} /* end of adk_ip_pbr_del_policy */


adk_status_t adk_ip_pbr_commit (
        adk_module_handle_t ipHandle)
{
    adk_sys_handle_t sys_handle;
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    ip_flow_info_t *flow_info_p;
    adk_ip_pbr_ctx_t * pbrCtx_p = NULL;
    uint8_t activeTreeId, shadowTreeId;
    adk_acl_rule_t cpRule;
    adk_ip_pbr_rule_t *pbrRules_p = NULL, pbrRule;
    adk_ip_pbr_action_t *pbrAction_p = NULL, pbrAction;
    uint32_t pbrRuleId = 0, idx;
    uint32_t  ruleIndex=0;

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    pbrCtx_p = (adk_ip_pbr_ctx_t *)ipModuleInfo_p->ip_dp_tbl_info_p->ipPbrInfo_p;

    /*
     * Clear the Ruleset & add the rules from the ADK CP rule list.
     * Then commit the ruleset.
     */
    if(pbrCtx_p->pbr_tree_index)
    {
        activeTreeId = 1;
        shadowTreeId = 0;
    }
    else
    {
        activeTreeId = 0;
        shadowTreeId = 1;
    }

    /* Only update Shadow tree. Active tree update is not really required
     * Next build will atomatically update active now which will be shadow then
     */
    adk_ip_pbr_ruleset_clear(ipHandle, pbrCtx_p->pbr_tree[shadowTreeId]);

    /*Get the rules from CP list & add them to S/W ACL*/
    adk_acl_get_reset(pbrCtx_p->cp_list);
    ruleIndex = 0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
            pbrCtx_p->cp_list, &cpRule))
    {
        flow_info_p = (ip_flow_info_t *)
                                      ADK_SYS_FLOW_ATTR_PTR_GET(cpRule.user_data);

        pbrRules_p = &(flow_info_p->u.pbr_flow.rule);
        pbrAction_p = &(flow_info_p->u.pbr_flow.action);
        pbrRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);

        printf("pbr_commit: ruleHandle_p=%p, ruleId=%d\n", cpRule.user_data, ADK_SYS_FLOWID_GET(cpRule.user_data));
        /*TODO: enhance this call to perform bulk add*/
        retValue = adk_ip_pbr_dpdk_sw_rule(ipHandle, pbrCtx_p->pbr_tree[shadowTreeId],
                (IP_MAX_PBR_RULE_PRIORITY - ruleIndex),
                pbrRuleId,
                pbrRules_p,
                pbrAction_p);
        if (retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Failed to add the rule in DPDK S/W\n");
            goto _adk_ip_pbr_commit_exit;
        }
        ruleIndex++;
    }

    /*
     * Add a default to S/W PBR rule to mark as not found
     */
    memset(&pbrRule, 0, sizeof(pbrRule));
    pbrAction.routeTableId = ADK_IP_MAX_ROUTING_TABLES;
    retValue = adk_ip_pbr_dpdk_sw_rule(ipHandle, pbrCtx_p->pbr_tree[shadowTreeId],
            (IP_MAX_PBR_RULE_PRIORITY - ruleIndex),
            1,
            &pbrRule,
            &pbrAction);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Failed to add the PBR rule in DPDK S/W\n");
        goto _adk_ip_pbr_commit_exit;
    }

    /*Update shadow copy*/
    retValue = adk_ip_pbr_ruleset_commit(ipHandle, pbrCtx_p->pbr_tree[shadowTreeId]);
    if(!retValue)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not commit PBR ACL rules...\n");
        goto _adk_ip_pbr_commit_exit;
    }
    adk_sys_flow_cache_update_version(sys_handle);

    /* Toggle the index */
    pbrCtx_p->pbr_tree_index = shadowTreeId;
    adk_rcu_synchronize();

    /* CP ACL List update to re-arrange the List */
    retValue = adk_acl_sw_only_commit(pbrCtx_p->cp_list);
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not re-arrange CP PBR ACL List...\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_pbr_commit_exit;
    }

    for (idx = 0; idx < pbrCtx_p->num_pending_rule_ids;
            idx++)
    {
        pbrRuleId = pbrCtx_p->pendingRuleIds[idx];
        /*TODO: Need to relook at pendingRuleIds 
         * Please refer JIRA-362*/
        if(pbrRuleId)
           adk_rm_resource_free(ipModuleInfo_p->ipPbrPolicyRuleIdHandle, pbrRuleId);
    }
    pbrCtx_p->num_pending_rule_ids = 0;

_adk_ip_pbr_commit_exit:
ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
   ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

   ADK_RETURN(retValue);
} /* end of adk_ip_pbr_commit */

/*!
 * =============================================================
 * adk_ip_get_pbr_policies_all
 *
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[out]   *numPbrPolicies - Number of PBR policies retrieved.
 *
 * @param[out]   **ipPbrPol_p - Array of numPbrPolicies PBR policies.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE  Invalid module handle passed
 *
 * @retval  ADK_IP_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @brief
 * This API gets all the entires in PBR policy table.
 *
 * @details
 * This API will fetch all IPv4 PBR policies along with the number of policies.
 * The array index of a policy represents its index in the PBR table.
 * The application is expected to free the memory allocated by the 
 *  API in variable *ipPbrPol_p.
 * =============================================================
 */
adk_status_t adk_ip_get_pbr_policies_all(
        adk_module_handle_t      ipHandle,
        uint32_t                 *numPbrPolicies,
        adk_ip_pbr_info_t        **ipPbrPol_p)
{
    adk_status_t retVal = ADK_SUCCESS;
    uint32_t num_acl_nodes = 0,i;
    adk_ip_pbr_ctx_t *pbrCtx_p = NULL;
    adk_ip_pbr_info_t *ipPbrPol_temp = NULL;
    adk_acl_rule_t pbr_rule; 
    ip_flow_info_t *flow_info_p = NULL;
    ip_module_info_t *ipModuleInfo_p = NULL;
    adk_sys_handle_t sys_handle;

    /* Validate module handle*/
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
	ADK_FORCE_LOG_ERROR("Invalid IP module handle \n");
	return retVal;
    }

    if(!(sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle)))
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Invalid sys handle \n");
	return retVal;
    }


    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    /*Validate input parameters*/
    if ((numPbrPolicies == NULL) || (ipPbrPol_p == NULL))
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
	ADK_FORCE_LOG_ERROR("NULL pointer passed, invalid input parameter \n");
	return retVal;
    }

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retVal);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    if(ipModuleInfo_p == NULL)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Invalid pointer to IP module info \n");
    }

    pbrCtx_p = (adk_ip_pbr_ctx_t *)ipModuleInfo_p->ip_dp_tbl_info_p->ipPbrInfo_p;

    if(pbrCtx_p == NULL)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Invalid pointer to PBR info \n");
    }

    /* Get the number of entries in the PBR tree. */
    retVal = adk_acl_get_num_nodes(pbrCtx_p->cp_list, &num_acl_nodes);

    if(retVal != ADK_SUCCESS)
    {
	ADK_FORCE_LOG_ERROR("Failed to get number of entries in PBR tree. \n");
	goto _adk_ip_get_pbr_policies_all_return;
    }

    /* Allocate the memory to store the PBR info.
     * NOTE: Application should free the memory allocated in this API. */
    ipPbrPol_temp = (adk_ip_pbr_info_t *) malloc(num_acl_nodes *  sizeof(adk_ip_pbr_info_t));

     adk_acl_get_reset(pbrCtx_p->cp_list);

    /* Now we have number of entries in PBR tree, hence fetch the PBR info.*/
    for(i=0;i<num_acl_nodes;i++)
    {
	retVal = adk_acl_next_rule_get(pbrCtx_p->cp_list, &pbr_rule);
	if (retVal != ADK_SUCCESS)
	{
	    retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LIST_NOT_FOUND);
	    ADK_FORCE_LOG_ERROR("PBR rule not found in ACL list\n");
	    goto _adk_ip_get_pbr_policies_all_return;
	}  

	flow_info_p = (ip_flow_info_t *)
	    ADK_SYS_FLOW_ATTR_PTR_GET(pbr_rule.user_data);

	ADK_MEMCPY(&((ipPbrPol_temp+i)->pbrRule), &(flow_info_p->u.pbr_flow.rule), sizeof(adk_ip_pbr_rule_t));
	ADK_MEMCPY(&((ipPbrPol_temp+i)->pbrAction), &(flow_info_p->u.pbr_flow.action), sizeof(adk_ip_pbr_action_t));
    }
    
    /* Return the fetched values in the pointer */
    *ipPbrPol_p = ipPbrPol_temp;
    *numPbrPolicies = num_acl_nodes;

_adk_ip_get_pbr_policies_all_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle,retVal);	

_lock_acquire_failed:
_lock_release_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retVal);
} 

/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
