/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_qos.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains all APIs related to QoS settings. This includes Diffserv
 * policies for outgoing traffic, Policing flows, rate limiting policies towards
 * CPU, DSCP re-mappings, etc.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"

static
adk_status_t adk_ip_qos_classifer_validate_arg(
        adk_module_handle_t ipHandle,
        const adk_ip_qos_classifier_rule_t *ipEgressQoSClassifierRule_p,
        const adk_ip_qos_classifier_action_t *ipEgressQoSClassifierAction_p);

static 
adk_status_t
adk_ip_egress_qos_classifier_hw_rule (
    adk_module_handle_t ipHandle,
    uint32_t ifId,
    uint32_t precedence,
    uint32_t egressQoSClassifierFlowId,
    const adk_ip_qos_info_t *ipEgressQoSClassifierInfo_p,
    uint8_t treeIndex);

static 
adk_status_t adk_ip_qos_policy_build_internal(adk_module_handle_t ipHandle, adk_ip_egress_qos_classifier_ctx_t *ctx);

static adk_status_t
adk_ip_qos_policy_delete_all(
        adk_module_handle_t ipHandle,
        uint32_t            ifId);

/***************** RTE ACL DEFINITION *******************
 */
enum {
    IP_EGRESS_QOS_CLASSIFIER_PROTO_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD,
    IP_EGRESS_QOS_CLASSIFIER_DSCP_FIELD, /*use 32 bits space for DSCP lookup*/
    ADK_IP_EGRESS_QOS_CLASSIFIER_NUM_FIELDS
};

enum {
    IP_EGRESS_QOS_CLASSIFIER_PROTO_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_SRC0_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_SRC1_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_SRC2_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_SRC3_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_DEST0_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_DEST1_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_DEST2_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_DEST3_INPUT,
    IP_EGRESS_QOS_CLASSIFIER_DSCP_INPUT, /*use 32 bits space for DSCP lookup*/
};


RTE_ACL_RULE_DEF(adk_ip_egress_qos_classifier_rule_rte, ADK_IP_EGRESS_QOS_CLASSIFIER_NUM_FIELDS);

static  struct  rte_acl_field_def adk_ip_egress_qos_classifier_def[ADK_IP_EGRESS_QOS_CLASSIFIER_NUM_FIELDS]  = {
    {
        .type = RTE_ACL_FIELD_TYPE_RANGE,
        .size = sizeof(uint8_t),
        .field_index  = IP_EGRESS_QOS_CLASSIFIER_PROTO_FIELD,
        .input_index  = IP_EGRESS_QOS_CLASSIFIER_PROTO_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, proto),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_SRC0_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, sip0),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_SRC1_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, sip1),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_SRC2_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, sip2),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_SRC3_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, sip3),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_DEST0_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, dip0),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_DEST1_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, dip1),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_DEST2_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, dip2),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_DEST3_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, dip3),
    },
    {
        .type = RTE_ACL_FIELD_TYPE_MASK,
        .size = sizeof(uint32_t),
        .field_index = IP_EGRESS_QOS_CLASSIFIER_DSCP_FIELD,
        .input_index = IP_EGRESS_QOS_CLASSIFIER_DSCP_INPUT,
        .offset = offsetof(adk_ip_egress_qos_classifier_ptn_t, dscp),
    },
};

/*!
 * =============================================================
 * adk_ip_qos_policy_add
 * @param[in] ipHandle   - ADK IP module handle.
 *
 * @param[in] ipIfHandle - ADK IP interface handle.
 *
 * @param[in] ruleIndex  - rule index
 *
 * @param[in] qosInfo_p  - pointer to data structure
 *                         adk_ip_qos_info_t.
 *
 *
 * @retval ADK_IP_SUCCESS Successfully added QoS policy.
 *
 * @retval ADK_E_INVALID_HANDLE invalid module handle or
 * interface handle.
 *
 * @retval ADK_E_INVALID_PARAMETER invalid parameter. 
 * 
 *
 * @retval ADK_IP_E_INVALID_ADDR_PREFIX invalid address prefix.
 *
 * @brief
 * This API adds the egress QoS classifier rule on a given 
 * IP interface.
 * =============================================================
 */
adk_status_t
adk_ip_qos_policy_add(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint32_t ruleIndex,
    const adk_ip_qos_info_t *qosInfo_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    ip_flow_info_t *ipFlowInfo_p;
    adk_sys_handle_t sys_handle;
    adk_acl_rule_t *cpRule_p;
    size_t classifierRuleId = ADK_RM_RSRC_INVALID;
    adk_flow_handle_t flowHandle  = 0;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    uint32_t ifId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    /* Validate input args */
    retValue = adk_ip_qos_classifer_validate_arg( ipHandle, &qosInfo_p->qosRule, &qosInfo_p->qosAction);

    if (ADK_SUCCESS != retValue)
        goto _adk_ip_qos_policy_add_exit;

    classifierRuleId = adk_rm_resource_reserve(
                               ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId]);
    if (ADK_RM_RSRC_INVALID == classifierRuleId)
    {
        goto _adk_ip_qos_policy_add_exit;
    }
    /* Clearing to indicate no handle is allocated yet */
    flowHandle = 0x0;

    /* Allocate flow handle */
    ADK_SYS_FLOW_HANDLE_ALLOC(ipHandle, ADK_IP_EGRESS_QOS_CLASSIFIER_INFO_OBJ,
                              sizeof(ip_flow_info_t), &flowHandle);

    /************/
    if(!(ipFlowInfo_p = (ip_flow_info_t *)
                ADK_SYS_FLOW_ATTR_PTR_GET(flowHandle)))
    {
        retValue=ADK_E_INTERNAL_ERROR;
        goto _adk_ip_qos_policy_add_exit;
    }

    ipFlowInfo_p->cookie = ADK_IPV4_FLOW_COOKIE;
    ipFlowInfo_p->type = IP_FLOW_TYPE_EGRESS_QOS_CLASSIFIER_FLOW;
    ADK_SYS_FLOWID_SET(flowHandle, classifierRuleId);

    ADK_MEMCPY((void *) &(ipFlowInfo_p->u.egress_qos_flow.flow_info),
               (const void *) qosInfo_p,
               sizeof(adk_ip_qos_info_t));
    /*
     * Add rule to ADK CP rule list.
     * The rules will be added during the next commit.
     */
    cpRule_p = &(ipFlowInfo_p->u.egress_qos_flow.cp_rule);
    cpRule_p->user_data = (void *)(flowHandle);   

    retValue = adk_acl_rule_add(
            ((adk_ip_egress_qos_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId])->cpList,
            ruleIndex,
            cpRule_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to add the rule in CP list\n");
        goto _adk_ip_qos_policy_add_exit;
    }

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);
    return ADK_SUCCESS;
_adk_ip_qos_policy_add_exit:
	/* Upon error, release the reserved resources */
	if (ADK_RM_RSRC_INVALID != classifierRuleId)
	{
		adk_rm_resource_free(ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId], classifierRuleId);
	}

    if (retValue != ADK_SUCCESS) {
        if (flowHandle)
            ADK_SYS_FLOW_HANDLE_FREE(flowHandle);
    }

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
    return ADK_SUCCESS;
}    /* -----  end of function adk_ip_qos_policy_add----- */

/*!
 * =============================================================
 * adk_ip_qos_policy_delete
 * @param[in] ipHandle    - ADK IP module handle.
 * 
 * @param[in] ipIfHandle  - ADK IP interface handle.
 *
 * @param[in] ruleIndex   - rule index.
 *
 * @retval ADK_IP_SUCCESS Successfully deleted QoS policy.
 *
 * @retval ADK_E_INVALID_HANDLE invalid module handle or
 * interface handle.
 *
 * @retval ADK_E_INVALID_PARAMETER invalid parameter. 
 *
 * @brief
 * This API deletes the egress QoS classifier rule on a given 
 * IP interface.
 * =============================================================
 */
adk_status_t
adk_ip_qos_policy_delete (
    adk_module_handle_t  ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint32_t ruleIndex 
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_egress_qos_classifer_info_t* egressQoSClassifierInfo_p = NULL;
    uint16_t index=0;
    adk_acl_rule_t cpRule;
    size_t classifierRuleId = ADK_RM_RSRC_INVALID;
    adk_flow_handle_t flowHandle = 0;
    uint32_t ifId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);


    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    egressQoSClassifierInfo_p =  ((adk_ip_egress_qos_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId]);

    adk_acl_get_reset(egressQoSClassifierInfo_p->cpList);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
    		egressQoSClassifierInfo_p->cpList, &cpRule))
    {
    	if(index==ruleIndex)
    	{
    		classifierRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);
    		flowHandle = (adk_flow_handle_t)cpRule.user_data;
    		break;
    	}
    	index++;
    }

    if (ADK_RM_RSRC_INVALID != classifierRuleId)
    {
    	adk_rm_resource_free(ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId], classifierRuleId);
    }

    if (flowHandle)
    	ADK_SYS_FLOW_HANDLE_FREE(flowHandle);

    /*** ***/


    retValue = adk_acl_rule_delete(egressQoSClassifierInfo_p->cpList,
                                           ruleIndex);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to delete the rule from CP list\n");
        goto _adk_ip_qos_policy_delete_exit;
    }
_adk_ip_qos_policy_delete_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_qos_policy_delete ----- */

/*!
 *=============================================================
 * adk_ip_dscp_pcp_map_add
 *
 * @param[in]   ipHandle        IP module handle
 *
 * @param[in]   ipIfHandle      IP interface for which the mapping needs to be
 * added
 *
 * @param[in]   dscp            The DSCP value to be mapped
 *
 * @param[in]   pcp             The PCP value to which DSCP needs to be mapped
 *
 * @retval ADK_SUCCESS              Successfully added dscp to pcp map
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid module or interface handle
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid DSCP or PCP value (exceeding max)
 *
 * @retval ADK_E_LEARN_FAILED       Failed to learn the pattern in the
 * hardware
 *
 * @brief
 *
 * This function adds maps a DSCP value to a PCP value for a given IP
 * interface. The IP interface must an IPoE type interface i.e. it cannot be
 * virtual interface.
 * =============================================================
 */
adk_status_t adk_ip_dscp_pcp_map_add(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint8_t dscp,
    uint8_t pcp)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    uint32_t l2IfId;
    ip_interface_info_t     *if_pvt_info = NULL;
    adk_ip_dp_interface_entry_t *dp_entry;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_dscp_pcp_map_add_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto _adk_ip_dscp_pcp_map_add_exit;
    }

    /* Validate the input DSCP and PCP parameters */
    if ((dscp >= IP_NUM_DSCP_VALUES) || (pcp >= IP_NUM_PCP_VALUES))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid DSCP (%d) or PCP (%d) value\n", dscp, pcp);
        goto _adk_ip_dscp_pcp_map_add_exit;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);

    l2IfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /*Update dp*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceTble[l2IfId]);
    dp_entry->dscp_lkup[dscp] = pcp;
    adk_sys_flow_cache_update_version(sys_handle);

    /* Update CP version id */
    ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[dp_entry->vrd_id]++;

_adk_ip_dscp_pcp_map_add_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}

/*!
 *=============================================================
 * adk_ip_dscp_pcp_map_delete
 *
 * @param[in]   ipHandle        IP module handle
 *
 * @param[in]   ipIfHandle      IP interface for which the mapping needs to be
 * removed
 *
 * @param[in]   dscp            The DSCP value for which mapping needs to be
 * removed
 *
 * @retval ADK_SUCCESS              Successfully removed DSCP to PCP mapping
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid module or interface handle
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid DSCP value (exceeding max)
 *
 * @retval ADK_E_LEARN_FAILED       Failed to learn the pattern in the
 * hardware
 *
 * @retval ADK_E_UNLEARN_FAILED     Failed to unlearn the pattern from the
 * hardware
 *
 * @brief
 *
 * This function removes mapping from DSCP to PCP value for a given IP
 * interface. The IP interface must an IPoE type interface i.e. it cannot be
 * virtual interface.
 * =============================================================
 */
adk_status_t adk_ip_dscp_pcp_map_delete(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ipIfHandle,
    uint8_t dscp)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    uint32_t l2IfId;
    ip_interface_info_t     *if_pvt_info = NULL;
    adk_ip_dp_interface_entry_t *dp_entry;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_dscp_pcp_map_delete_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto _adk_ip_dscp_pcp_map_delete_exit;
    }

    /* Validate the input DSCP parameter */
    if (dscp >= IP_NUM_DSCP_VALUES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid DSCP (%d) value\n", dscp);
        goto _adk_ip_dscp_pcp_map_delete_exit;
    }

    if_pvt_info = (ip_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHandle);
    l2IfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /*Update dp*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceTble[l2IfId]);
    dp_entry->dscp_lkup[dscp] &= 0xF0;
    adk_sys_flow_cache_update_version(sys_handle);

    /* Update CP version id */
    ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[dp_entry->vrd_id]++;

_adk_ip_dscp_pcp_map_delete_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}

/*!
 *=============================================================
 * adk_ip_ingress_qos_policy_add 
 *
 * @param[in] ipHandle   - ADK IP module handle.
 *
 * @param[in] ipIfHandle - ADK IP interface handle.
 *
 * @param[in] qosInfo_p  - pointer to data structure
 *                         adk_ip_ingress_qos_info_t.
 *
 * @retval ADK_IP_SUCCESS Successfully added QoS policy.
 *
 * @retval ADK_IP_E_INVALID_PARAMETER - invalid module handle or
 * qosInfo_p.
 *
 * @retval ADK_E_INVALID_HANDLE - IP interface handle
 * not present or invalid.
 *
 * @brief
 *
 * This API creates the IP ingress Qos classifier policy.
 * =============================================================
 */
adk_status_t
adk_ip_ingress_qos_policy_add(
        adk_module_handle_t ipHandle,
        adk_interface_handle_t ipIfHandle, 
        adk_ip_ingress_qos_info_t *qosInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    uint32_t l2IfId;
    adk_ip_dp_interface_entry_t *dp_entry;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_ingress_qos_policy_add_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto _adk_ip_ingress_qos_policy_add_exit;
    }

    l2IfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /*Update dp table*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceTble[l2IfId]);
    dp_entry->dscpToPhbClass[qosInfo_p->dscp] = qosInfo_p->phbClass;

_adk_ip_ingress_qos_policy_add_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}

/*!
 *=============================================================
 * adk_ip_ingress_qos_policy_delete 
 *
 * @param[in] ipHandle   - ADK IP module handle.
 *
 * @param[in] ipIfHandle - ADK IP interface handle.
 *
 * @param[in] qosInfo_p  - pointer to data structure
 *                         adk_ip_ingress_qos_info_t.
 *
 * @retval ADK_IP_SUCCESS Successfully delete QoS policy.
 *
 * @retval ADK_IP_E_INVALID_PARAMETER - invalid module handle or
 * qosInfo_p.
 *
 * @retval ADK_E_INVALID_HANDLE - IP interface handle
 * not present or invalid.
 *
 * @brief
 *
 * This API creates the IP ingress Qos classifier policy.
 * =============================================================
 */
adk_status_t
adk_ip_ingress_qos_policy_delete(
        adk_module_handle_t ipHandle,
        adk_interface_handle_t ipIfHandle, 
        adk_ip_ingress_qos_info_t *qosInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    uint32_t l2IfId;
    adk_ip_dp_interface_entry_t *dp_entry;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_ingress_qos_policy_delete_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!adk_ipv4_interface_valid_handle(ipIfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid Interface Handle type\n");
        goto _adk_ip_ingress_qos_policy_delete_exit;
    }

    l2IfId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /*Update dp table*/
    dp_entry = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceTble[l2IfId]);
    /*Set the default PHB for this dscp*/
    dp_entry->dscpToPhbClass[qosInfo_p->dscp] = ADK_QOS_DS_PHB_CLASS_NONE;
    
_adk_ip_ingress_qos_policy_delete_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}


/**
 *=============================================================
 * @internal
 * adk_ip_egress_qos_classifier_create
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[out] egressQoSClassifierInfo_p - Address of pointer to context which is 
               populated and returned by this API
 *
 * @retval ADK_SUCCESS Successfully created ingress classifer context.
 *
 * @retval ADK_E_MEM_UNAVAILABLE Failed to create as memory not available
 * @brief
 *
 * This function creates egress QoS classifier ordered tree
 * This is called at initialization time where validation of params
 * is already done.
 * =============================================================
 */
adk_status_t adk_ip_egress_qos_classifier_create(
        adk_module_handle_t ipHandle,
        uint32_t ifId,
        adk_ip_egress_qos_classifer_info_t  **ipEgressQoSClassifierInfo_p)
{
    struct rte_acl_param prm;
    uint8_t   i;
    uint16_t  actionIndexDp;
    char name[50];
    adk_status_t    retValue;
    adk_sys_handle_t sys_handle;
    adk_acl_info_t cpListInfo={0};
    adk_ip_qos_info_t qosInfo;
    adk_ip_egress_qos_classifer_info_t *egressQoSClassifierInfo_p = NULL;

    ip_module_info_t * ipModuleInfo_p =
             (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    egressQoSClassifierInfo_p = (adk_ip_egress_qos_classifer_info_t*)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_EGRESS_QOS_CLASSIFIER_INFO_OBJ, 1,
			 sizeof(adk_ip_egress_qos_classifer_info_t));

    if (egressQoSClassifierInfo_p == NULL)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                           "Memory allocation failed\n");
        return ADK_E_MEM_UNAVAILABLE;
    }

    *ipEgressQoSClassifierInfo_p = egressQoSClassifierInfo_p;

    prm.socket_id = SOCKET_ID_ANY;
    prm.rule_size = RTE_ACL_RULE_SZ(ADK_IP_EGRESS_QOS_CLASSIFIER_NUM_FIELDS);
    prm.max_rule_num = ADK_IP_MAX_EGRESS_QOS_CLASSIFIER_RULES + 1;


    if (!adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId], 0)) {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_RESOURCE_UNAVAILABLE,
                "Could not get resource id ING ORDERED CLASSIFIER using rte_acl_create!\n");
        return ADK_E_RESOURCE_UNAVAILABLE;
    }

    /*Default rule use index 0*/
    actionIndexDp = 0;

    for (i=0; i<2; i ++)
    {
        /* Initialize */
        egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[i] = NULL;
      
        snprintf(name, sizeof(name), "ADK_IP_EGRESS_QOS_CLASSIFIER_%u_%1d",ifId,i);
        prm.name = name;
        ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_create",
                    "creating new Egress QOS Ordered classifier context: %s\n", prm.name);

        egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[i] =
            (adk_ip_egress_qos_classifier_ctx_t*) rte_acl_create(&prm);
        if (NULL == egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[i])
        {
            if (1 == i)
            {
                ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                        "Could not create EGRESS QOS ORDERED CLASSIFIER using rte_acl_create!\n");
                /* free previous one */
                rte_acl_free(egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[0]);
            }
            return ADK_E_MEM_UNAVAILABLE;
        }
    }

    egressQoSClassifierInfo_p->aclTreeIndex = ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT;
    cpListInfo.user = ADK_ACL_IPV4_EGRESS_QOS_CLASSIFIER;
    if(ADK_SUCCESS != adk_acl_create(&(egressQoSClassifierInfo_p->cpList),
                                &cpListInfo))
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not allocate CP list for IP EGRESS QOS CLASSIFIER!\n");
        goto free_rte_acl_ctx;
    }

    /* 
     * Set first default Rule as ANY 
     */

    memset(&qosInfo, 0x0, sizeof(adk_ip_qos_info_t));
    qosInfo.qosRule.ruleMask = 0;
    qosInfo.qosAction.queueId = ADK_QOS_INVALID_QUEUE_ID;
    qosInfo.qosAction.pktColor = RTE_TM_GREEN;

    if(ADK_SUCCESS != adk_ip_egress_qos_classifier_hw_rule(
            ipHandle,
            ifId,
            ADK_IP_EGRESS_QOS_CLASSIFIER_RULE_MAX_PRIORITY,
            actionIndexDp,
            &qosInfo,
			ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT
            ))
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not add default DENY rule to IP ING CLASSIFIER active tree!\n");
        goto free_rte_acl_ctx;
    }
    /* Build */
    if( ADK_SUCCESS != adk_ip_qos_policy_build_internal(ipHandle,
                egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[ADK_IP_QOS_ACTIVE_TREE_INDEX_INIT]))
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not BUILD default DENY rule to IP ING CLASSIFIER active tree!\n");
        goto free_rte_acl_ctx;
    }

    if(ADK_SUCCESS != adk_ip_egress_qos_classifier_hw_rule(
    		ipHandle,
            ifId,
			ADK_IP_EGRESS_QOS_CLASSIFIER_RULE_MAX_PRIORITY,
			actionIndexDp,
			&qosInfo,
		    ADK_IP_QOS_SHADOW_TREE_INDEX_INIT	
    ))
    {
    	retValue = ADK_E_HW_PROGRAMMING_ERROR;
    	ADK_IPV4_LOG_FATAL(ipHandle, retValue,
    			"Could not add default DENY rule to IP ING CLASSIFIER shadow tree!\n");
    	goto free_rte_acl_ctx;
    }
    /* Build */
    if( ADK_SUCCESS != adk_ip_qos_policy_build_internal(ipHandle,
    		egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[ADK_IP_QOS_SHADOW_TREE_INDEX_INIT]))
    {
    	retValue = ADK_E_HW_PROGRAMMING_ERROR;
    	ADK_IPV4_LOG_FATAL(ipHandle, retValue,
    			"Could not BUILD default DENY rule to IP ING CLASSIFIER shadow tree!\n");
    	goto free_rte_acl_ctx;
    }
    return ADK_SUCCESS;

free_rte_acl_ctx:
    adk_rm_resource_free(ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId],
            actionIndexDp);
    rte_acl_free(egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[0]);
    rte_acl_free(egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[1]);
    return retValue;
} /* adk_ip_egress_qos_classifier_create*/

/**
 *=============================================================
 * adk_ip_egress_qos_classifier_delete 
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[out] ctx - pointer to context to be deleted
 *
 * @retval ADK_SUCCESS Successfully deleted ingress classifer context.
 *
 * @brief
 *
 * This function deletes ingrees classifer ordered tree
 * =============================================================
 */
adk_status_t adk_ip_egress_qos_classifier_delete(
        adk_module_handle_t ipHandle,
        uint32_t ifId)
{
    uint8_t   i;
    adk_status_t    retValue;
    adk_sys_handle_t sys_handle;
    adk_ip_egress_qos_classifer_info_t *egressQoSClassifierInfo_p = NULL;

    ip_module_info_t * ipModuleInfo_p =
             (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    egressQoSClassifierInfo_p =  ((adk_ip_egress_qos_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId]);

    if (egressQoSClassifierInfo_p) {

        /*Delete all the rules from the table and delete both ACL context*/

        retValue = adk_ip_qos_policy_delete_all(ipHandle, ifId);

        if (retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Failed to delete all rules from CP list \n");
            goto _adk_ip_egress_qos_classifier_delete_exit;
        }
        for (i=0; i<2; i ++)
        {
            /*Delete ACL context*/
            if (egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[i]) {
                rte_acl_reset(egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[i]);
                rte_acl_free(egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[i]);
            }
        }

        /*Free egress QoS classifier structure*/
        ADK_NVM_FREE(egressQoSClassifierInfo_p);
    }

_adk_ip_egress_qos_classifier_delete_exit:
    return ADK_SUCCESS;
} /* adk_ip_egress_qos_classifier_delete*/

/*!
 * =============================================================
 * adk_ip_qos_classifer_validate_arg
 * validates the input parameter for IP QOS Classifier rule add 
 * =============================================================
 */
static
adk_status_t adk_ip_qos_classifer_validate_arg(
        adk_module_handle_t ipHandle,
        const adk_ip_qos_classifier_rule_t *ipEgressQoSClassifierRule_p,
        const adk_ip_qos_classifier_action_t *ipEgressQoSClassifierAction_p)
{
    adk_status_t    retValue;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipEgressQoSClassifierRule_p || !ipEgressQoSClassifierAction_p) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid NULL input parameter\n");
        goto return_with_error;
    }

    if (ipEgressQoSClassifierRule_p->ruleMask & ADK_QOS_CLASSIFIER_DEST_ADDR_MASK) {
        if ((ADK_IPV4_ADDRESS == ipEgressQoSClassifierRule_p->destAddr.type) &&
            (ipEgressQoSClassifierRule_p->destAddr.addr_u.v4Addr.prefixWidth > ADK_IPV4_MAX_PREFIX_WIDTH))
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid Destination IP address prefix width\n");
            goto return_with_error;
        } else if ((ADK_IPV6_ADDRESS == ipEgressQoSClassifierRule_p->destAddr.type) &&
            (ipEgressQoSClassifierRule_p->destAddr.addr_u.v6Addr.prefixWidth > ADK_IPV6_MAX_PREFIX_WIDTH)) {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
                ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                        "Invalid Destination IP address prefix width\n");
                goto return_with_error;
        }
    }

    if (ipEgressQoSClassifierRule_p->ruleMask & ADK_QOS_CLASSIFIER_SRC_ADDR_MASK) {
        if ((ADK_IPV4_ADDRESS == ipEgressQoSClassifierRule_p->srcAddr.type) &&
            (ipEgressQoSClassifierRule_p->srcAddr.addr_u.v4Addr.prefixWidth > ADK_IPV4_MAX_PREFIX_WIDTH))
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid Source IP address prefix width: %d\n", ipEgressQoSClassifierRule_p->srcAddr.addr_u.v4Addr.prefixWidth);
            goto return_with_error;
        } else if ((ADK_IPV6_ADDRESS == ipEgressQoSClassifierRule_p->srcAddr.type) &&
                   (ipEgressQoSClassifierRule_p->srcAddr.addr_u.v6Addr.prefixWidth > ADK_IPV6_MAX_PREFIX_WIDTH)) {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
                ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                        "Invalid Destination IP address prefix width\n");
                goto return_with_error;
        }
    }

    if ((ipEgressQoSClassifierRule_p->ruleMask & ADK_QOS_CLASSIFIER_SRC_ADDR_MASK) && 
        (ipEgressQoSClassifierRule_p->ruleMask & ADK_QOS_CLASSIFIER_DEST_ADDR_MASK) && 
        (ipEgressQoSClassifierRule_p->destAddr.type != ipEgressQoSClassifierRule_p->srcAddr.type)) {
            retValue = ADK_IP_E_INVALID_IP_ADDRESS;
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "IP address type mismatch : %u\n", ipEgressQoSClassifierRule_p->destAddr.type);
            goto return_with_error;
        }

    if (ipEgressQoSClassifierRule_p->dscp >= IP_NUM_DSCP_VALUES) {
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid dscp value: %u\n", ipEgressQoSClassifierRule_p->dscp);
            goto return_with_error;
    }

    if (ADK_QOS_INVALID_QUEUE_ID == ipEgressQoSClassifierAction_p->queueId) {
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Invalid queue Id : %u\n", ipEgressQoSClassifierAction_p->queueId);
            goto return_with_error;
    }

    switch (ipEgressQoSClassifierAction_p->pktColor)
    {
    case RTE_TM_GREEN:
    case RTE_TM_YELLOW:
    case RTE_TM_RED:
         break;
    default:
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid packet color \n");
        goto return_with_error;
    }
    return ADK_SUCCESS;
return_with_error:
    return retValue;
} /* adk_ip_qos_classifer_validate_arg*/

/*!
 * =============================================================
 * @internal
 * adk_ip_egress_qos_classifier_hw_rule
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief
 *
 * Internal support function for ACL rule addition.
 *
 * =============================================================
 */
static 
adk_status_t
adk_ip_egress_qos_classifier_hw_rule (
    adk_module_handle_t ipHandle,
    uint32_t ifId, 
    uint32_t precedence,
    uint32_t egressQoSClassifierFlowId,
    const adk_ip_qos_info_t *ipEgressQoSClassifierInfo_p,
    uint8_t treeIndex
    )
{
    adk_status_t retValue = ADK_SUCCESS;
    int ret = 0;
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ip_module_info_t *ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    struct adk_ip_egress_qos_classifier_rule_rte rule = {0};
    adk_ip_dp_egress_qos_classifier_action_info_t *classifierResult_p;
    uint8_t prefixWidth = 0;

    /*****   Build the Individual Patterns ******/

    classifierResult_p = &(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[ifId][treeIndex][egressQoSClassifierFlowId]);
    classifierResult_p->queueId = ipEgressQoSClassifierInfo_p->qosAction.queueId;
    classifierResult_p->pktColor = ipEgressQoSClassifierInfo_p->qosAction.pktColor;
    ADK_IPV4_LOG_INFO(ipHandle,"\tIPv4 QoS classifier - building rule: treeIndex=%d\n", treeIndex);

    /* Set protocol information */
    if (ipEgressQoSClassifierInfo_p->qosRule.ruleMask & ADK_QOS_CLASSIFIER_PROTOCOL_TYPE_MASK)
    {
    	ADK_IPV4_LOG_INFO(ipHandle,"\t\tproto=%d\n", ipEgressQoSClassifierInfo_p->qosRule.protocol);
        /* Set RTE_ACL structure */
        rule.field[IP_EGRESS_QOS_CLASSIFIER_PROTO_FIELD].value.u8 =
            ipEgressQoSClassifierInfo_p->qosRule.protocol;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_PROTO_FIELD].mask_range.u8  =
            ipEgressQoSClassifierInfo_p->qosRule.protocol;
    }
    else
    {
        /* Set ANY for protocol */
        rule.field[IP_EGRESS_QOS_CLASSIFIER_PROTO_FIELD].value.u8 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_PROTO_FIELD].mask_range.u8  = 0xFF;
    }
    /* Set Source address */
    if (ipEgressQoSClassifierInfo_p->qosRule.ruleMask & ADK_QOS_CLASSIFIER_SRC_ADDR_MASK)
    {
        if (ADK_IPV4_ADDRESS == ipEgressQoSClassifierInfo_p->qosRule.srcAddr.type) {
            ADK_IPV4_LOG_INFO(ipHandle,"\t\tsip=0x%x/%d\n", ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v4Addr.addr,
                                                            ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v4Addr.prefixWidth);
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD].value.u32 =
                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v4Addr.addr;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD].mask_range.u32  =
                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v4Addr.prefixWidth;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD].value.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD].mask_range.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD].value.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD].mask_range.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD].value.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD].mask_range.u32 = 0;
        }else {
            ADK_IPV4_LOG_INFO(ipHandle,"\t\tsip=0x%x:0x%x:0x%x:0x%x /%d\n",
                                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[0],
                                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[1],
                                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[2],
                                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[3],
                                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.prefixWidth);
            prefixWidth = ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.prefixWidth;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD].value.u32 =
                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[0];
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD].value.u32 = 
                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[1];
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD].value.u32 = 
                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[2];
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD].value.u32 = 
                ipEgressQoSClassifierInfo_p->qosRule.srcAddr.addr_u.v6Addr.addr[3];
        
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD].mask_range.u32 = (prefixWidth> 32) ? 32 : prefixWidth;
            prefixWidth = (prefixWidth > 32) ? (prefixWidth - 32) : 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
            prefixWidth = (prefixWidth > 32) ? (prefixWidth - 32) : 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
            prefixWidth = (prefixWidth > 32) ? (prefixWidth - 32) : 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
        }
    }
    else
    {
        /* Set ANY */
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC0_FIELD].mask_range.u32  = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC1_FIELD].mask_range.u32  = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC2_FIELD].mask_range.u32  = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_SRC3_FIELD].mask_range.u32  = 0;
    }

    /* Set Destination address */
    if (ipEgressQoSClassifierInfo_p->qosRule.ruleMask & ADK_QOS_CLASSIFIER_DEST_ADDR_MASK)
    {
        if (ADK_IPV4_ADDRESS == ipEgressQoSClassifierInfo_p->qosRule.destAddr.type) {
            ADK_IPV4_LOG_INFO(ipHandle,"\t\tdip=0x%x/%d\n",
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v4Addr.addr,
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v4Addr.prefixWidth);
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD].value.u32 =
                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v4Addr.addr;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD].mask_range.u32  =
                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v4Addr.prefixWidth;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD].value.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD].mask_range.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD].value.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD].mask_range.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD].value.u32 = 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD].mask_range.u32 = 0;
        }else {
            ADK_IPV4_LOG_INFO(ipHandle,"\t\tdip=0x%x:0x%x:0x%x:0x%x /%d\n",
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[0],
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[1],
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[2],
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[3],
                                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.prefixWidth);
            prefixWidth = ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.prefixWidth;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD].value.u32 =
                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[0];
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD].value.u32 = 
                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[1];
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD].value.u32 = 
                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[2];
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD].value.u32 = 
                ipEgressQoSClassifierInfo_p->qosRule.destAddr.addr_u.v6Addr.addr[3];
        
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
            prefixWidth = (prefixWidth > 32) ? (prefixWidth - 32) : 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
            prefixWidth = (prefixWidth > 32) ? (prefixWidth - 32) : 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
            prefixWidth = (prefixWidth > 32) ? (prefixWidth - 32) : 0;
            rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD].mask_range.u32 = (prefixWidth > 32) ? 32 : prefixWidth;
        }
    }
    else
    {
        /* Set ANY */
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST0_FIELD].mask_range.u32  = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST1_FIELD].mask_range.u32  = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST2_FIELD].mask_range.u32  = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DEST3_FIELD].mask_range.u32  = 0;
    }

    /* Set dscp*/
    if (ipEgressQoSClassifierInfo_p->qosRule.ruleMask & ADK_QOS_CLASSIFIER_DSCP_MASK)
    {
    	ADK_IPV4_LOG_INFO(ipHandle,"\t\tdscp=0x%x\n", ipEgressQoSClassifierInfo_p->qosRule.dscp);
        /* Store ID instead of handle */
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DSCP_FIELD].value.u32 = adk_htonl(ipEgressQoSClassifierInfo_p->qosRule.dscp);
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DSCP_FIELD].mask_range.u32  = 8;
    }
    else
    {
        /* Set ANY */
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DSCP_FIELD].value.u32 = 0;
        rule.field[IP_EGRESS_QOS_CLASSIFIER_DSCP_FIELD].mask_range.u32  = 0;
    }

    ADK_IPV4_LOG_INFO(ipHandle,"\t\tqueueId=%d\n", ipEgressQoSClassifierInfo_p->qosAction.queueId);
    ADK_IPV4_LOG_INFO(ipHandle,"\t\tpktColor=%d\n", ipEgressQoSClassifierInfo_p->qosAction.pktColor);
    rule.data.category_mask = 1;
    rule.data.priority = precedence;
    /*Pack the action in userdata field itself*/
    rule.data.userdata = egressQoSClassifierFlowId;

    ret = rte_acl_add_rules(
            ((adk_ip_egress_qos_classifer_info_t*) ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId])->ipEgressQoSClassifierCtx[treeIndex],
            (struct rte_acl_rule *)&rule, 1);
    if (ret< 0)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error inserting ACL rule in IP\n");
        return ADK_E_OTREE_INSERT_FAILED ;
    }

    /* */
    ADK_RETURN(retValue);

}/* adk_ip_egress_qos_classifier_hw_rule */

/**
 *=============================================================
 * @internal 
 * adk_ip_qos_policy_build_internal 
 * @param[in]  ipHandle - IP module handle.
 * @param[in]  ctx - IP egress QoS classifer Context
 *
 * @retval ADK_SUCCESS Successfully built ingress classifer context.
 *
 * @retval ADK_E_OTREE_BUILD_FAILED - Build Failed
 *
 * @brief
 *
 * This function builds IP QoS classifier ordered tree
 * =============================================================
 */
static 
adk_status_t adk_ip_qos_policy_build_internal(adk_module_handle_t ipHandle, adk_ip_egress_qos_classifier_ctx_t *ctx)
{
    struct rte_acl_config cfg;
    int ret=0;

    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_MEMSET(&cfg, 0, sizeof(cfg));
    cfg.num_fields = ADK_IP_EGRESS_QOS_CLASSIFIER_NUM_FIELDS;
    ADK_MEMCPY(&cfg.defs, adk_ip_egress_qos_classifier_def, sizeof(adk_ip_egress_qos_classifier_def));
    cfg.num_categories = 1;
    cfg.max_size = 0;

    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_acl_build", "building IP Egress QoS Classifier context\n");

    ret = rte_acl_build(ctx, &cfg);
    if (ret != 0)
    {
    	ADK_IPV4_LOG_ERROR(ipHandle, ret, "rte_acl_build failed\n");
        return ADK_E_OTREE_BUILD_FAILED;
    }

    return ADK_SUCCESS;
} /* adk_ip_qos_policy_build_internal */

/**
 *=============================================================
 * adk_ip_qos_policy_build 
 *
 * @param[in]  ipHandle   - ADK IP module handle.
 *
 * @param[in]  ipIfHandle - ADK IP interface handle. 
 *
 * @retval ADK_SUCCESS Successfully built ingress classifer context.
 *
 * @retval ADK_E_INVALID_HANDLE invalid module or interface handle.
 *
 * @retval ADK_E_HW_PROGRAMMING_ERROR - Failed to add rule to hardware. 
 *
 * @retval ADK_E_OTREE_BUILD_FAILED - Build Failed
 *
 * @brief
 *
 * This function builds IP QoS classifier ordered tree.
 * =============================================================
 */
adk_status_t adk_ip_qos_policy_build(adk_module_handle_t ipHandle, adk_interface_handle_t ipIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_ip_egress_qos_classifer_info_t  *egressQoSClassifierInfo_p = NULL;
    adk_sys_handle_t sys_handle;
    uint8_t i, activeTreeId, shadowTreeId;
    uint16_t ruleIndex;
    ip_flow_info_t *ipFlowInfo_p;
    size_t classifierRuleId;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    adk_acl_rule_t cpRule;
    adk_ip_qos_info_t *qosInfo_p;
    uint32_t ifId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    egressQoSClassifierInfo_p = (adk_ip_egress_qos_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId];
    /*
     * Clear the Ruleset & add the rules from the ADK CP rule list.
     * Then commit the ruleset.
     */
    if(egressQoSClassifierInfo_p->aclTreeIndex)
    {
        activeTreeId = 1;
        shadowTreeId = 0;
    }
    else
    {
        activeTreeId = 0;
        shadowTreeId = 1;
    }
    /* Only update Shadow tree. Active tree update is not really required
     * Next build will atomatically update active now which will be shadow then
     */
    rte_acl_reset_rules(egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[shadowTreeId]);
    /*Update CP version id for VRDs*/
    for(i=0;i<=ADK_IP_DEFAULT_VRD_ID;++i)
        (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[i])++;
    /* Update Active copy. Get the rules from CP list & add them to S/W ACL*/
    adk_acl_get_reset(egressQoSClassifierInfo_p->cpList);
    ruleIndex = 0;

    while(ADK_SUCCESS == adk_acl_next_rule_get(
                egressQoSClassifierInfo_p->cpList, &cpRule))
    {
        ipFlowInfo_p = (ip_flow_info_t *)
            ADK_SYS_FLOW_ATTR_PTR_GET(cpRule.user_data);
        qosInfo_p = &(ipFlowInfo_p->u.egress_qos_flow.flow_info);
        classifierRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);
        if( ADK_SUCCESS != adk_ip_egress_qos_classifier_hw_rule(
                ipHandle,
                ifId,
                (ADK_IP_EGRESS_QOS_CLASSIFIER_RULE_MAX_PRIORITY - ruleIndex),
                classifierRuleId,
                qosInfo_p,
                shadowTreeId))
        {
            retValue = ADK_E_HW_PROGRAMMING_ERROR;
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Failed to add the rule in hardware\n");
            goto release_lock_and_return;
        }
        ruleIndex++;
    }

    /* Add Default rule at the end */
    /* Set Rule mask as zero, indicating no mask */
    adk_ip_qos_info_t ipQoSDefaultRule = {0};
    ipQoSDefaultRule.qosAction.queueId = ADK_QOS_INVALID_QUEUE_ID;
    ipQoSDefaultRule.qosAction.pktColor = RTE_TM_GREEN;
    if(ADK_SUCCESS != adk_ip_egress_qos_classifier_hw_rule(
            ipHandle,
            ifId,
			(ADK_IP_EGRESS_QOS_CLASSIFIER_RULE_MAX_PRIORITY - ruleIndex),
            0, /* zero id reserved for default rule */
            &ipQoSDefaultRule,
            shadowTreeId
            ))
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not add default rule to IP QOS CLASSIFIER!\n");
        goto release_lock_and_return;
    }

    /*Update shadow copy*/
    retValue = adk_ip_qos_policy_build_internal(ipHandle,
            egressQoSClassifierInfo_p->ipEgressQoSClassifierCtx[shadowTreeId]);
    if(ADK_SUCCESS !=  retValue)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not build IP Ingress CLASSIFEER...\n");
        goto release_lock_and_return;
    }
    /* Toggle the index */
    egressQoSClassifierInfo_p->aclTreeIndex = shadowTreeId;
    adk_sys_flow_cache_update_version(sys_handle);
    /* RCU sync */
    adk_rcu_synchronize();
    /* CP ACL List update to re-arrange the List
     * To be called only if adk_cp_acl_list_hos_commit is not invoked*/
    retValue = adk_acl_sw_only_commit(egressQoSClassifierInfo_p->cpList);
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error: Could not re-arrange CP ACL List...\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto release_lock_and_return;
    }

release_lock_and_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);
    return retValue;

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retValue);
} /* adk_ip_qos_policy_build*/

static adk_status_t
adk_ip_qos_policy_delete_all(
        adk_module_handle_t ipHandle,
        uint32_t            ifId)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_ip_egress_qos_classifer_info_t* egressQoSClassifierInfo_p = NULL;
    uint16_t index=0;
    adk_acl_rule_t cpRule;
    size_t classifierRuleId = ADK_RM_RSRC_INVALID;
    adk_flow_handle_t flowHandle = 0;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    egressQoSClassifierInfo_p =  ((adk_ip_egress_qos_classifer_info_t*)ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId]);

    adk_acl_get_reset(egressQoSClassifierInfo_p->cpList);

    index=0;

    while(ADK_SUCCESS == adk_acl_next_rule_get(
    		egressQoSClassifierInfo_p->cpList, &cpRule))
    {
        classifierRuleId = ADK_SYS_FLOWID_GET(cpRule.user_data);
        flowHandle = (adk_flow_handle_t)cpRule.user_data;
        if (ADK_RM_RSRC_INVALID != classifierRuleId)
        {
            adk_rm_resource_free(ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId], classifierRuleId);
        }
        if (flowHandle)
            ADK_SYS_FLOW_HANDLE_FREE(flowHandle);

        retValue = adk_acl_rule_delete(egressQoSClassifierInfo_p->cpList, index);
        if (retValue != ADK_SUCCESS)
        {
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Failed to delete the rule from CP list\n");
            goto _adk_ip_qos_policy_delete_all_exit;
        }
    	index++;
    }

    /*Delete the control plane list*/
    if(ADK_SUCCESS != adk_acl_free(egressQoSClassifierInfo_p->cpList))
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not allocate CP list for IP EGRESS QOS CLASSIFIER!\n");
        goto _adk_ip_qos_policy_delete_all_exit;
    }

    /*Free 0th id for the default rule*/ 
    adk_rm_resource_free(ipModuleInfo_p->ipEgressQoSClassifierRuleIDHandle[ifId], 0);

_adk_ip_qos_policy_delete_all_exit:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * adk_ip_get_qos_policies_all
 *
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[in]    ipIfHandle - Interface handle for which the QoS policies
 *                            need to be retrieved.
 *
 * @param[out]   *numQosPolicies - Number of QoS policies retrieved.
 *
 * @param[out]   **ipQosPol_p - Array of numQosPolicies QoS policies.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE  Invalid module handle passed
 *
 * @retval  ADK_IP_E_INVALID_PARAMETER - Invalid input parameter
 *
 * @brief
 * This API will fetch all the IP QoS rules added for an IP interface.
 *
 * @details
 * This API gets all the IP QoS policies along with the number of policies.
 * The array index of a policy represents its index in the egress QoS 
 * classifier tree.
 * The application is expected to free the memory allocated by the API in 
 * variable *ipQosPol_p.
 * =============================================================
 */
adk_status_t adk_ip_get_qos_policies_all(
        adk_module_handle_t          ipHandle,
	adk_interface_handle_t       ipIfHandle,
        uint32_t                     *numQosPolicies,
        adk_ip_qos_info_t            **ipQosPol_p)
{
    adk_status_t retVal = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p = NULL;
    uint32_t num_acl_nodes,i,ifId;
    ip_flow_info_t *flow_info_p = NULL;
    adk_sys_handle_t sys_handle;
    adk_ip_egress_qos_classifer_info_t *my_egressQoSClassifierInfo_p = NULL;
    adk_ip_qos_info_t *temp_Qos_info_p = NULL;
    adk_acl_rule_t qos_classifier_acl_rule;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid IP module handle \n");
        return retVal;
    }

    if(!(sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle)))
    {
        retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_FORCE_LOG_ERROR("Invalid sys handle \n");
        return retVal;
    }

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    /*Validate input parameters*/
    if ((numQosPolicies == NULL)  || (ipQosPol_p == NULL) || (ipIfHandle == NULL))
    {
        retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("NULL pointer passed, invalid input parameter \n");
        return retVal;
    }

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retVal);

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    if(ipModuleInfo_p == NULL)
    {
        retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_FORCE_LOG_ERROR("NULL pointer to IP module info \n");
        goto _adk_ip_get_qos_policies_all_return;
    }

    ifId = ADK_SYS_INTERFACE_ID_GET(ipIfHandle);

    my_egressQoSClassifierInfo_p = ((adk_ip_egress_qos_classifer_info_t*)(ipModuleInfo_p->ip_dp_tbl_info_p->egressQoSClassifierInfo_p[ifId]));

    if(my_egressQoSClassifierInfo_p == NULL)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("NULL pointer to Egress Classifier info\n");
	goto _adk_ip_get_qos_policies_all_return;
    }

    /* Get the number of entries in the PBR tree ==> Active tree in CP is always in sync with the committed rules,
     * hence fetch the pbr info from active tree. */
    adk_acl_get_num_nodes(my_egressQoSClassifierInfo_p->cpList, &num_acl_nodes);

    /* Allocate the memory to store the QoS classifier info.
     * NOTE: Application should free the memory allocated in this API. */
    temp_Qos_info_p = (adk_ip_qos_info_t *) malloc(num_acl_nodes * sizeof(adk_ip_qos_info_t));

    if(temp_Qos_info_p == NULL)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
	ADK_FORCE_LOG_ERROR("Unable to allocate memory for egress QoS classifier info \n");
	goto _adk_ip_get_qos_policies_all_return;
    }

    retVal = adk_acl_get_reset(my_egressQoSClassifierInfo_p->cpList);
    if(retVal != ADK_SUCCESS)
    {
	retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	ADK_FORCE_LOG_ERROR("Unable to reset ACL list\n");
	goto _adk_ip_get_qos_policies_all_return;
    }

    /* Now we have number of entries in Egress QoS classifier tree, hence fetch the classifier info.*/
    for(i=0;i<num_acl_nodes;i++)
    {
	retVal = adk_acl_next_rule_get(my_egressQoSClassifierInfo_p->cpList, &qos_classifier_acl_rule);
	if (retVal != ADK_SUCCESS)
	{
	    retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LIST_NOT_FOUND);
	    ADK_FORCE_LOG_ERROR("QoS classifier rule not found in ACL list\n");
	    goto _adk_ip_get_qos_policies_all_return;
	}

	flow_info_p = (ip_flow_info_t *)
	    ADK_SYS_FLOW_ATTR_PTR_GET(qos_classifier_acl_rule.user_data);

	if(flow_info_p == NULL)
	{
	    retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	    ADK_FORCE_LOG_ERROR("NULL pointer to QoS Egress Classifier flow info\n");
	    goto _adk_ip_get_qos_policies_all_return;
	}

	ADK_MEMCPY(&((temp_Qos_info_p + i)->qosRule), &(flow_info_p->u.egress_qos_flow.flow_info.qosRule), sizeof(adk_ip_qos_classifier_rule_t));
	ADK_MEMCPY(&((temp_Qos_info_p + i)->qosAction), &(flow_info_p->u.egress_qos_flow.flow_info.qosAction), sizeof(adk_ip_qos_classifier_action_t));

    }

    /* Return the fetched values in the pointer */
    *numQosPolicies = num_acl_nodes;
    *ipQosPol_p = temp_Qos_info_p;


_adk_ip_get_qos_policies_all_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle,retVal);	

_lock_acquire_failed:
_lock_release_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retVal);

}
/*! end addtogroup _ipFunctions
 * @}
 *  end addtogroup _ip
 * @}
 */
