/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_route.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains API to add/delete routing entries to/from IP forwarding
 * table.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include <adk_cksum.h>
#include <adk_timer.h>
#include "adk_ip_dp_internal.h"
#include <adk_ethernet_conf.h>

static inline uint16_t adk_ip_route_find_cp_node(
        ip_module_info_t *ipModuleInfo_p,
        adk_acl_rule_t *route_p,
        const adk_ip_route_info_t *ipRoute_p
        )
{
    adk_acl_get_reset(ipModuleInfo_p->routeInfo);
    uint16_t index = 0;
    bool found = false;
    adk_acl_rule_t route;

    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipModuleInfo_p->routeInfo, &route))
    {
        if( (ipRoute_p->routeTableId == ((adk_ip_route_info_t *)route.user_data)->routeTableId) &&
           ( (ipRoute_p->destAddr.addr && ipRoute_p->destAddr.prefixWidth) ==
                        (((adk_ip_route_info_t *)route.user_data)->destAddr.addr &&
                        ((adk_ip_route_info_t *)route.user_data)->destAddr.prefixWidth))
        )
        {
            found = true;
            *route_p = route;
            break;
        }
        index++;
    }

    if(!found)
        index = 0xFFFF;

    return index;
}

static inline adk_status_t adk_ip_route_update_cp_list(
        bool addOperation,
    adk_module_handle_t ipHandle,
        ip_module_info_t *ipModuleInfo_p,
        adk_status_t dp_st,
        const adk_ip_route_info_t *ipRoute_p
)
{
    adk_sys_handle_t sys_handle;
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    adk_status_t ret = ADK_SUCCESS;

    if(addOperation)
    { /* add or update existing route */
        if(dp_st == ADK_SUCCESS)
        { /* New route was added to DP successfully. So, add new node here */
            adk_acl_rule_t route;
            route.user_data = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_ROUTE_ADD_OBJ, 1,
                    sizeof(adk_ip_route_info_t));
            if(NULL == route.user_data)
            {
                ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                        "Memory allocation failed\n");
                ret = ADK_E_MEM_UNAVAILABLE;
                goto __adk_ip_route_update_cp_list_exit;
            }

            (*(adk_ip_route_info_t *)route.user_data) = *ipRoute_p;

            if(ADK_SUCCESS !=adk_acl_rule_add(ipModuleInfo_p->routeInfo, 0, &route))
            {
                ADK_IPV4_LOG_ERROR(ipHandle, ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR),
                        "Error inserting ACL rule in IP route update\n");
                ret = ADK_E_OTREE_INSERT_FAILED ;
                goto __adk_ip_route_update_cp_list_exit;
            }

            ipModuleInfo_p->numRoutesProgrammed++;
            goto __adk_ip_route_update_cp_list_exit;
        }
        else if( ADK_E_ENTRY_OVERWRITTEN == dp_st)
        { /* Existing route was updated, now update CP list */
            adk_acl_rule_t route;
            uint16_t index = adk_ip_route_find_cp_node(ipModuleInfo_p, &route, ipRoute_p);
            if(0xFFFF == index)
            {
                /* somethings' not right. Entry updated in DP but not found in CP list */
                ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_INTERNAL_ERROR,
                          "Route Entry not found in CP list\n");
                ret = ADK_E_INTERNAL_ERROR;
                goto __adk_ip_route_update_cp_list_exit;
            }
            (*(adk_ip_route_info_t *)route.user_data) = *ipRoute_p;
            adk_acl_rule_modify(ipModuleInfo_p->routeInfo, index, &route);
            goto __adk_ip_route_update_cp_list_exit;
        }
    }
    else
    { /* delete route */
        adk_acl_rule_t route;
        uint16_t index = adk_ip_route_find_cp_node(ipModuleInfo_p, &route, ipRoute_p);
        if(0xFFFF == index)
        {
            /* somethings' not right. Entry updated in DP but not found in CP list */
            ret = ADK_E_INTERNAL_ERROR;
            goto __adk_ip_route_update_cp_list_exit;
        }

        if(ADK_SUCCESS == adk_acl_rule_delete(ipModuleInfo_p->routeInfo, index))
            ipModuleInfo_p->numRoutesProgrammed--;
        else
        {
            ret = ADK_E_INTERNAL_ERROR;
            goto __adk_ip_route_update_cp_list_exit;
        }
    }

__adk_ip_route_update_cp_list_exit:
    if(ADK_SUCCESS == ret)
        adk_acl_sw_only_commit(ipModuleInfo_p->routeInfo);

    return ret;

} /* adk_ip_route_update_cp_list */

/*!
 * =============================================================
 * adk_ip_route_add
 *
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[in]    ipRoute_p - pointer to adk_ip_route_info_t data
 * structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @retval  ADK_IP_E_INVALID_IP_VERSION - Unsupported IP version
 * value configured in one of the Address prefixes.
 *
 * @retval  ADK_IP_E_INVALID_ADDR_PREFIX - Invalid IPv4 address prefix.
 * It should be between 0 and 32.
 *
 * @retval ADK_IP_E_INVALID_IP_ADDRESS - Multicast/Broadcast address specified
 * for the the interface.
 *
 * @brief
 *
 * This API adds a new entry to IP forwarding table
 * (adk_ip_dest_addr_lookup) using the parameters provided
 * in the adk_ip_route_info_t data structure. This API can be used to modify route-mtu
 * by setting the same route again with new mtu value.
 *
 * When connectedRoute flag is set to true, we will assume that valid layer-2 interface hd
 * is provided and packets will be send on this l2 interface. If 'connectedRoute' flag is set to false,
 * we will assume the nextHop field is valid and it will be used in datapath for taking forwarding
 * decision.
 * =============================================================
 */
adk_status_t
adk_ip_route_add (
    adk_module_handle_t ipHandle,
    const adk_ip_route_info_t *ipRoute_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t vrdId;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_st;
    adk_ip_dp_route_info_t * dipResult_p = NULL;
    adk_ip_dp_route_info_t * ipRouteInfo_p = NULL;
    adk_handle_t lkup_result = NULL;

#ifdef ADK_HOS_OLD_IMPL
    size_t ip_term_hos_rule_id[2] = {0};
#endif
    /* validate input handles.
     */
    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    if (!ipRoute_p || !ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid input parameter\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ip_route_add_return;
    }

#ifdef ENABLE_VRD
     /* Validate Routing Domain */
    if (!adk_ipv4_vrd_valid_handle(ipRoute_p->ipVrdHandle, &vrdId))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 routing domain handle %"ADKPRIh"\n",
                           ipRoute_p->ipVrdHandle);
        goto adk_ip_route_add_return;
    }
#else
	vrdId = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE_VRD */

    /* Validation of the IP address and subnet masks are beyond
     * the scope of this API. It is assumed that caller validate
     * all these parameters.
     * Update destination Address lookup tree. The tree input is
     * destination IP address prefix. The return values are
     * structured like below.
     * real handle     - bit-3 is set to indicate IP fast forwarding.
     *                 - bits 4-23 ->20 LSBs of IP address.
     * virtual handle  - bits 0-11 -> 12 LSBs of next hop address.
     */
    if (ipRoute_p->destAddr.prefixWidth > 32)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_ADDR_PREFIX, "Invalid IP address prefix\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
        goto adk_ip_route_add_return;
    }

    if (ipRoute_p->routeTableId >= ADK_IP_MAX_ROUTING_TABLES)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_ROUTE_TABLE_ID,
                                   "Invalid route table id (%d)\n", ipRoute_p->routeTableId);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ROUTE_TABLE_ID);
        goto adk_ip_route_add_return;
    }

    /* Make sure that next hop address is a valid interface address if
     * specified */
    if ((ipRoute_p->nextHop != 0) && (ADK_IPV4_IS_IP_IF_ADDR_INVALID(ipRoute_p->nextHop)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Invalid address specified as next hop\n");
        goto adk_ip_route_add_return;
    }

    if ( ipRoute_p->routeMTU != 0 && ((ipRoute_p->routeMTU < IP_IF_MTU_MIN) || (ipRoute_p->routeMTU > IP_IF_MTU_MAX)) )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_MTU, "Invalid MTU Value %d\n", ipRoute_p->routeMTU);
        goto adk_ip_route_add_return;
    }

    if(ADK_IP_MAX_ROUTES == ipModuleInfo_p->numRoutesProgrammed)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_MAX_ROUTES_PROGRAMMED);
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_MAX_ROUTES_PROGRAMMED, "Maximum allowed routes already programmed (%d)\n", ADK_IP_MAX_ROUTES);
        goto adk_ip_route_add_return;
    }

    /* ip_dp_tbl_info_p global variable gets updated only in dp_init() that happens only in dp_process(). It is uninitialized in Application process. Hence setting it here */
    ip_dp_tbl_info_p = ipModuleInfo_p->ip_dp_tbl_info_p;

    /* Check for implicit route. If route is found AND prefix-width is same -> user is trying to update implicit route. this is not allowed */
    dp_st = adk_ip_dp_dip_get(ipRoute_p->routeTableId,
            adk_htonl(ipRoute_p->destAddr.addr & ip_bit_mask[ipRoute_p->destAddr.prefixWidth]), &lkup_result);
    if(likely(ADK_SUCCESS == dp_st) && likely(lkup_result))
    {
        ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
        if(ipRouteInfo_p->isTerminating)
        {
            /* This is a directly connected address. We just need to update the MTU */
            ADK_IPV4_LOG_INFO(ipHandle, "Modification of implicit route (termination) not allowed: dip=0x%x\n", ipRoute_p->destAddr.addr);
            goto adk_ip_route_add_return;
        }
    }

    dipResult_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_ROUTE_ADD_OBJ, 1, sizeof(adk_ip_dp_route_info_t));
    if (dipResult_p == NULL)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                          "Memory allocation failed\n");
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto adk_ip_route_add_return;
    }

    dipResult_p->isTerminating = FALSE;
    dipResult_p->routeMTU = ipRoute_p->routeMTU;
    if(ipRoute_p->connectedRoute)
    {
        uint16_t l2IfId, ipIfId;
        l2IfId = ADK_SYS_INTERFACE_ID_GET(ipRoute_p->l2IfHd);

        if(likely(l2IfId >= ADK_ETH_NUM_BRIDGE))
            ipIfId = l2IfId - ADK_ETH_NUM_BRIDGE + ADK_ETH_BR_MAX;
        else
            ipIfId = l2IfId;

        dipResult_p->u.nh_if.interfaceId = ipIfId;
        dipResult_p->u.nh_if.nextHop = 0;
    }
    else
    {
        dipResult_p->u.nh_if.nextHop = ipRoute_p->nextHop;
        dipResult_p->u.nh_if.interfaceId = ADK_DP_MAX_IP_INTERFACE;
    }

    dipResult_p->prefixWidth = ipRoute_p->destAddr.prefixWidth;

    dp_st = adk_ip_dp_dip_add(ipHandle, ipRoute_p->routeTableId, adk_htonl(ipRoute_p->destAddr.addr), ipRoute_p->destAddr.prefixWidth,
                        ADK_CONVERT_TO_LPM_DATA(dipResult_p));
    if(dp_st != ADK_SUCCESS && dp_st != ADK_E_ENTRY_OVERWRITTEN)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INTERNAL_MEMORY_UNAVAILABLE,
	                   "Error in adding dest ip to dataplane");
        goto adk_ip_route_add_return;
    }
    else
    {
        if(ADK_SUCCESS != adk_ip_route_update_cp_list(true, ipHandle, ipModuleInfo_p, dp_st, ipRoute_p))
        {
            adk_ip_dp_dip_remove(ipHandle, ipRoute_p->routeTableId, adk_htonl(ipRoute_p->destAddr.addr), ipRoute_p->destAddr.prefixWidth);
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INTERNAL_MEMORY_UNAVAILABLE,
                    "Error in adding dest ip to routing table");
            goto adk_ip_route_add_return;
        }
        ADK_IPV4_LOG_INFO(ipHandle, "Added/updated ipv4 route: route-table-id=%d dip=0x%x/%d, NH=0x%x, MTU=%d, total-num-routes=%d\n", ipRoute_p->routeTableId,
                ipRoute_p->destAddr.addr, ipRoute_p->destAddr.prefixWidth, ipRoute_p->nextHop, ipRoute_p->routeMTU, ipModuleInfo_p->numRoutesProgrammed);
    }

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId])++;

adk_ip_route_add_return:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_route_add  ----- */


/*!
 * =============================================================
 * adk_ip_route_delete
 *
 * @param[in]    ipHandle - IP module handle.
 *
 * @param[in]    ipRoute_p - pointer to adk_ip_route_info_t data
 * structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @retval  ADK_IP_E_INVALID_IP_VERSION - Unsupported IP version
 * value configured in one of the Address prefixes.
 *
 * @retval  ADK_IP_E_INVALID_ADDR_PREFIX - Invalid IPv4 address prefix.
 * It should be between 0 and 32.
 *
 * @brief
 *
 * This API deletes an entry from IP forwarding table
 * using the parameters provided
 * in the adk_ip_route_info_t data structure.
 *
 * =============================================================
 */
adk_status_t
adk_ip_route_delete (
    adk_module_handle_t ipHandle,
    const adk_ip_route_info_t *ipRoute_p
)
{
    adk_status_t        retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    adk_sys_handle_t        sys_handle;
    uint32_t             vrd_id;
    adk_dp_status_t         dp_st;
    adk_ip_dp_route_info_t * ipRouteInfo_p = NULL;
    adk_handle_t lkup_result = NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    /* Validate input handle */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_route_delete_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipRoute_p || !ipModuleInfo_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid input parameter\n");
        goto _adk_ip_route_delete_exit;
    }

#ifdef ENABLE_VRD
    /* Validate Routing Domain */
    if (!adk_ipv4_vrd_valid_handle(ipRoute_p->ipVrdHandle, &vrd_id))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 routing domain handle %"ADKPRIh"\n",
                           ipRoute_p->ipVrdHandle);
        goto _adk_ip_route_delete_exit;
    }
#else
    vrd_id = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE_VRD */

    /* Make sure that the destination IP address prefix length is valid */
    if (ipRoute_p->destAddr.prefixWidth > 32)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ADDR_PREFIX);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid IP address prefix\n");
        goto _adk_ip_route_delete_exit;
    }

    if (ipRoute_p->routeTableId >= ADK_IP_MAX_ROUTING_TABLES)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_IP_E_INVALID_ROUTE_TABLE_ID,
                "Invalid route table id (%d)\n", ipRoute_p->routeTableId);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_ROUTE_TABLE_ID);
        goto _adk_ip_route_delete_exit;
    }

    /* ip_dp_tbl_info_p global variable gets updated only in dp_init() that happens only in dp_process(). It is uninitialized in Application process. Hence setting it here */
    ip_dp_tbl_info_p = ipModuleInfo_p->ip_dp_tbl_info_p;

    /* Check for implicit route. If route is found AND prefix-width is same -> user is trying to update implicit route. this is not allowed */
    dp_st = adk_ip_dp_dip_get(ipRoute_p->routeTableId,
            adk_htonl(ipRoute_p->destAddr.addr & ip_bit_mask[ipRoute_p->destAddr.prefixWidth]), &lkup_result);
    if(likely(ADK_SUCCESS == dp_st) && likely(lkup_result))
    {
        ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
        if(ipRouteInfo_p->isTerminating)
        { /* This is a directly connected address. We just need to update the MTU */
            ADK_IPV4_LOG_INFO(ipHandle, "Modification of implicit terminating route not allowed: dip=0x%x\n", ipRoute_p->destAddr.addr);
            goto _adk_ip_route_delete_exit;
        }
    }

    /*TO DO: Do we need this? Delete the HW termination rule if present*/
    dp_st = adk_ip_dp_dip_hw_rule_remove(ipHandle,ipRoute_p->routeTableId, adk_htonl(ipRoute_p->destAddr.addr), ipRoute_p->destAddr.prefixWidth);
    if(dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
	                   "Error in removing dest ip to dataplane\n");
        goto _adk_ip_route_delete_exit;
    }
    /* Delete the route */
    dp_st = adk_ip_dp_dip_remove(ipHandle,ipRoute_p->routeTableId, adk_htonl(ipRoute_p->destAddr.addr), ipRoute_p->destAddr.prefixWidth);
    if(dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
	                   "Error in removing dest ip to dataplane\n");
        goto _adk_ip_route_delete_exit;
    }
    else
    {
        adk_ip_route_update_cp_list(false,  ipHandle,  ipModuleInfo_p, dp_st, ipRoute_p);
	    ADK_IPV4_LOG_INFO(ipHandle, "Removed ipv4 route: route-table-id=%d dip=0x%x/%d, NH=0x%x, MTU=%d, total-num-routes=%d\n", ipRoute_p->routeTableId,
			    ipRoute_p->destAddr.addr, ipRoute_p->destAddr.prefixWidth, ipRoute_p->nextHop, ipRoute_p->routeMTU, ipModuleInfo_p->numRoutesProgrammed);
    }

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrd_id])++;


_adk_ip_route_delete_exit:
        ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    
    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_route_delete ----- */


/*!
* =============================================================
* adk_ip_get_routes_all
*
* @param[in]    ipHandle - IP module handle.
*
* @param[in]    ipRouteTblId - Identifies one of the route table to be used.
 *
* @param[out]   *numIpRoutes - Number of routes in routing table
*
* @param[out]   **ipRoute_p - Array of numIpRoutes IP routes
*
* @retval ADK_SUCCESS - Successful completion
*
* @retval ADK_E_INVALID_HANDLE - Invalid module handle passed
*
* @retval ADK_IP_E_INVALID_PARAMETER - Invalid input parameter
* passed. The pointer might be NULL.
*
* @brief
*
* This API will fetch all routes in an IP forwarding table.
*
* @details
*
* This API retrieves all the route entries along with the number
* of entries. The application is
* expected to free the memory allocated by the API in
* variable *ipRoute_p.
*
* =============================================================
*/
adk_status_t adk_ip_get_routes_all(
       adk_module_handle_t      ipHandle,
       uint8_t                  ipRouteTblId,
       uint32_t                 *numIpRoutes,
       adk_ip_route_info_t      **ipRoute_p)
{
    adk_status_t        retValue = ADK_SUCCESS;
    ip_module_info_t        *ipModuleInfo_p;
    adk_sys_handle_t        sys_handle;
    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Validate input handle */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_get_routes_all_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipRoute_p || !ipModuleInfo_p || !numIpRoutes)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid input parameter\n");
        goto _adk_ip_get_routes_all_exit;
    }

    adk_acl_get_reset(ipModuleInfo_p->routeInfo);
    adk_acl_rule_t route;

    /* Count the routes in this routing table 'ipRouteTblId' */
    *numIpRoutes = 0;

    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipModuleInfo_p->routeInfo, &route))
    {
        if( (ipRouteTblId == ((adk_ip_route_info_t *)route.user_data)->routeTableId))
            (*numIpRoutes)++;
    }

    /* Allocate memory */
    if(0 ==  *numIpRoutes)
        goto _adk_ip_get_routes_all_exit;

    *ipRoute_p = malloc((*numIpRoutes) * sizeof(adk_ip_route_info_t));
    if(NULL == *ipRoute_p)
    {
        retValue = ADK_IP_E_INTERNAL_MEMORY_UNAVAILABLE;
        goto _adk_ip_get_routes_all_exit;
    }

    uint16_t i=0;
    adk_acl_get_reset(ipModuleInfo_p->routeInfo);
    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipModuleInfo_p->routeInfo, &route))
    {
        if( (ipRouteTblId == ((adk_ip_route_info_t *)route.user_data)->routeTableId))
            (*ipRoute_p)[i++] = *((adk_ip_route_info_t *)route.user_data);
    }


_adk_ip_get_routes_all_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:

    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    ADK_RETURN(retValue);

}


/*! 
 *      end addtogroup _ipFunctions  IPv4 Functions
 *      @}
 * end addtogroup _ipFunctions
 * @}
 */
