/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_stats.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs for getting or printing
 * statistics for the interfaces, module level global stats.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include <adk_ethernet_export.h>
#include <adk_cksum.h>
#include <adk_timer.h>
#include "adk_ip_dp_internal.h"

/*!
 * =============================================================
 * adk_ip_global_stats_get
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[out]  stats_p  pointer to adk_ip_global_stats_t
 * data structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER  invalid IP module handle or
 * stats_p pointer specified.
 *
 * @retval ADK_E_STATS_READ_FAILURE  failed to read statistics
 * from the hardware.
 *
 * @brief
 *
 * This function reads the statistics from data path
 * and fill the adk_ip_global_stats_t pointer with appropriate values.
 *
 * =============================================================
 */
adk_status_t
adk_ip_global_stats_get (
    adk_module_handle_t  ipHandle,
    adk_ip_global_stats_t   *stats_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_st;
    uint32_t i, j;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_global_stats_get_exit;
    }

    if (!stats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid input pointer parameter\n");
        goto _adk_ip_global_stats_get_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    dp_st = adk_ip_dp_global_stats_get(ipHandle, stats_p);
    if (dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Error in reading IPv4 global stats from dataplane\n");
        goto _adk_ip_global_stats_get_exit;
    }

    for(i=0; i<ADK_MAX_NUM_CORES; i++) {
        for(j=0; j<ADK_DP_MAX_IP_STATS; j++) {
            
            // To be enabled for debug stats
            /* stats_p->ipInPkts += ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceStatsTble[i][j].ifInPackets;
               stats_p->ipOutPkts += ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceStatsTble[i][j].ifOutBcastPkts;

               stats_p->ipOutPkts += ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceStatsTble[i][j].ifOutMcastPkts;
               stats_p->ipOutPkts += ipModuleInfo_p->ip_dp_tbl_info_p->ipInterfaceStatsTble[i][j].ifOutUcastPkts; */
        }
    }
    
    // To be enabled for debug stats
    /*stats_p->ipInPkts += ipModuleInfo_p->ipGlobalStatsTable_bkp.ipInPkts; 
      stats_p->ipOutPkts += ipModuleInfo_p->ipGlobalStatsTable_bkp.ipOutPkts; */

    /* Collect the backed up stats for interface/classifier rule deletion in global stats. */
    stats_p->ipInPolicerDrops += ipModuleInfo_p->ipGlobalStatsTable_bkp.ipPolicerDrop;

    
    // To be enabled for debug stats  
    /* Invalid Interface count needs to be accounted under total received packets */
    /* stats_p->ipInPkts += stats_p->ipInvalidInIfCnt; */

_adk_ip_global_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}    /* -----  end of function adk_ip_global_stats_get  ----- */


/*!
 * =============================================================
 * adk_ip_interface_stats_get
 *
 * @param[in] ipHandle      IP module handle.
 *
 * @param[in] ifHandle      IP interface handle.
 *
 * @param[out] stats_p      Pointer to adk_ip_interface_stats_t data
 * structure.
 *
 * @retval ADK_SUCCESS                  Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE         Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER      Invalid parameter e.g. NULL
 * stats_p pointer specified.
 *
 * @brief
 *
 * This function reads IP interface stats from the data path
 * and populates the adk_ip_interface_stats_t structure with
 * appropriate values.
 * =============================================================
 */
adk_status_t adk_ip_interface_stats_get(
    adk_module_handle_t ipHandle,
    adk_interface_handle_t ifHandle,
    adk_ip_interface_stats_t *stats_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    bool is_virtual_if = FALSE,
         l3_qos_enabled = FALSE, read_eg_stats = FALSE;
    uint32_t stats_id, eg_stats_id;
    adk_sys_handle_t sys_handle;
    ip_interface_info_t         *if_pvt_info = NULL;
    ip_virtual_interface_info_t *vif_pvt_info = NULL;
    adk_dp_status_t dp_st;
    adk_sys_egress_stats_t eg_stats;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_interface_stats_get_exit;
    }

    if (!stats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid input NULL pointer parameter\n");
        goto _adk_ip_interface_stats_get_exit;
    }

    if (adk_ipv4_interface_valid_handle(ifHandle)) {
        if_pvt_info = (ip_interface_info_t *) ADK_SYS_INTERFACE_ATTR_PTR_GET(ifHandle);
        dp_st = adk_ip_dp_interface_stats_get(ipHandle, ifHandle, stats_p);
        if (dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in reading IP interface stats from dataplane\n");
            goto _adk_ip_interface_stats_get_exit;
        }
        stats_id = if_pvt_info->ip_stats_id;
        eg_stats_id = if_pvt_info->sys_egress_stats_id;
        read_eg_stats = TRUE;
        l3_qos_enabled = ((if_pvt_info->if_info.ipIntfAttr) & ADK_IP_INTERFACE_MASK_QOS_ENABLE);
    } else if (adk_ipv4_vif_valid_handle(ifHandle)) {
        vif_pvt_info = (ip_virtual_interface_info_t *) ADK_SYS_INTERFACE_ATTR_PTR_GET(ifHandle);
        dp_st = adk_ip_dp_virtual_if_stats_get(ipHandle, vif_pvt_info->stats_id, stats_p);
        if (dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in reading IP interface stats from dataplane\n");
            goto _adk_ip_interface_stats_get_exit;
        }
        stats_id = vif_pvt_info->stats_id;
        is_virtual_if = TRUE;
    } else {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_INTERFACE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid IPv4 interface handle\n");
        goto _adk_ip_interface_stats_get_exit;
    }

    if(read_eg_stats)
    {
        retValue = adk_sys_egress_stats_get(sys_handle, eg_stats_id, &eg_stats);
        if(retValue != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in reading sys egress stats for IP interface\n");
            goto _adk_ip_interface_stats_get_exit;
        }

    }

_adk_ip_interface_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

}


/*!
* =============================================================
* adk_ip_address_stats_get
* @param[in]   ipHandle  IP module handle.
*
* @param[in]   ipAddress  IP Address to retrieve stats for.
*
* @param[out]  stats_p  pointer to adk_ip_address_stats_t
 * data structure.
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_E_INVALID_PARAMETER  invalid IP module handle or
* stats_p pointer specified.
*
* @brief
*
* This function reads IP address stats and fill the adk_ip_address_stats_t
* point with appropriate values.
*
* =============================================================
*/

adk_status_t adk_ip_address_stats_get(
        adk_module_handle_t ipHandle,
        adk_ip_addr_t ipAddress,
        adk_ip_address_stats_t *stats_p)
{
    adk_dp_status_t dp_st;
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t stats_id;
    adk_sys_handle_t sys_handle;
    adk_handle_t lkup_result = NULL;
    adk_ip_dp_route_info_t * ipRouteInfo_p=NULL;
    ip_module_info_t *ipModuleInfo_p=NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_addr_stats_get_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    if (!ipModuleInfo_p)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER, "Invalid input parameter\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_addr_stats_get_exit;
    }

    if (!stats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Invalid input NULL pointer parameter\n");
        goto _adk_ip_addr_stats_get_exit;
    }

    /* TO DO: We are doing route lookup from control path. This will increase API latency. This should be replaced by ip address table */

    /* Search for route entry with addr&prefix as ip-address and update the route-info with new routeMTU */

    dp_st = adk_ip_dp_dip_get(ADK_IP_DEFAULT_VRD_ID, adk_htonl(ipAddress), &lkup_result);
    if(likely(ADK_SUCCESS == dp_st) && likely(lkup_result))
    {
        if(unlikely((ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result) == NULL))
	goto _adk_ip_addr_stats_get_exit;
    }
    if(unlikely(!lkup_result))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "No existing route found for given dip: %x\n", ipAddress);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_CFG_ENTRY_NOT_FOUND);

        goto _adk_ip_addr_stats_get_exit;
    }

    if(FALSE == ipRouteInfo_p->isTerminating)
    {
        goto _adk_ip_addr_stats_get_exit;
    }

    stats_id = ipRouteInfo_p->u.termination.ip_addr_stats_id;

    dp_st = adk_ip_dp_ip_addr_stats_get(ipHandle, stats_id, stats_p);
    if (dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Error in reading IP address stats from dataplane\n");
        goto _adk_ip_addr_stats_get_exit;
    }


    _adk_ip_addr_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

    _lock_release_failed:
    _lock_acquire_failed:

    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ip_l4_stats_get
 * @param[in]   ipHandle  IP module handle.
 *
 * @param[out]  stats_p  pointer to adk_ip_l4_stats_t
 * data structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER  invalid IP module handle or
 * stats_p pointer specified.
 *
 * @brief
 *
 * This function reads the statistics from data path
 * and fill the adk_ip_l4_stats_t pointer with appropriate values.
 *
 * =============================================================
 */
adk_status_t
adk_ip_l4_stats_get (
    adk_module_handle_t  ipHandle,
    adk_ip_l4_stats_t   *stats_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_st;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_l4_stats_get_exit;
    }

    if (!stats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid input pointer parameter\n");
        goto _adk_ip_l4_stats_get_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    dp_st = adk_ip_dp_l4_stats_get(ipHandle, stats_p);
    if (dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error in reading IPv4 L4 stats from dataplane\n");
        goto _adk_ip_l4_stats_get_exit;
    }


_adk_ip_l4_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

} /* adk_ip_l4_stats_get */


/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
