/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include "adk_sif.h"
#include "adk_shm.h"
#include <adk_ip_common.h>
#include "adk_ip_utility.h"

/*!
 * =============================================================
 * adk_ip_dp_table_init()
 *
 * IP DP table Init function -- To be called in CP init process context
 * =============================================================
 */
#ifdef DPDK_IP_FRAG
extern struct rte_mempool *pktmbuf_pool_indirect_g;
#endif
adk_status_t
adk_ip_dp_table_init(adk_module_handle_t ipHandle)
{
    adk_status_t ret=ADK_SUCCESS;
    uint32_t ip_addr;
    uint32_t  i = 0, j = 0;
    adk_ip_dp_tbl_info_t *ip_dp_tbl_info_p = NULL;
    adk_sys_config_worker_lcore_info_t  *workerInfo;
    adk_sys_handle_t sys_handle;
    void *ind_pool;
    adk_ip_dp_lli_hash5_tbl_entry_t *lli_hash5_entry_p;
    adk_ip_dp_lli_hash3_tbl_entry_t *lli_hash3_entry_p;
    adk_ip_dp_uli_hash5_tbl_entry_t *uli_hash5_entry_p;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ip_module_info_t * ipModuleInfo_p =
             (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);


    /*
     * Lookup the worker info so we can get the NUMA socket to allocate the
     * memzone from
     */
    ret = adk_sys_config_worker_lcore_info_get(&workerInfo);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }

    if (workerInfo->adkMode  == ADK_SYS_CONFIG_MODE_STANDALONE)
    {
#ifdef DPDK_IP_FRAG
        ind_pool = (void *)pktmbuf_pool_indirect_g;
#endif
    }
    else
    {
        /* indirect pool */
        ret = adk_sys_config_ip_frag_mempool_get(&ind_pool);
        if(ret != ADK_SUCCESS){
            return ret;
        }
    }

    ADK_IPV4_LOG_INFO(ipHandle,
                     "First Worker Core ID = %d, socket ID = %d\n",
                     workerInfo->lcore[0].id, workerInfo->lcore[0].socket);

    /* Create memzone to store the structure which will have pointers to all
     * the data plane tables */
    ret = adk_shm_create(ADK_IP_SHM_MEMZONE_NAME,
                         sizeof(adk_ip_dp_tbl_info_t),
                         ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                         (adk_shm_handle_t **)&ip_dp_tbl_info_p);
    if (ret != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                          "Failed to create IP memzone %s\n",
                          ADK_IP_SHM_MEMZONE_NAME);
        return ADK_E_MEM_UNAVAILABLE;
    }

    /* Initialize all IP tables here */
    if(!ip_dp_tbl_info_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ipModuleInfo_p->ip_dp_tbl_info_p = ip_dp_tbl_info_p;

    /* 1. Initialize Destination IP Lookup table */
    ip_dp_tbl_info_p->dipTbl = adk_lpm_create(ADK_MODULE_MODULEID_GET(ipHandle),
					ADK_IP_DIP_TBL_OBJ, IP_DIP_TBL_WIDTH,
                                                ADK_IP_DIP_TBL_PENDING_SIZE);

    if(ip_dp_tbl_info_p->dipTbl == (adk_lpm_handle_t)NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 2. Initialize Source IP Validation table */
    ip_dp_tbl_info_p->sipTbl = adk_lpm_create(ADK_MODULE_MODULEID_GET(ipHandle),
					ADK_IP_SIP_TBL_OBJ, IP_SIP_TBL_WIDTH,
                                               ADK_IP_SIP_TBL_PENDING_SIZE);
    if(ip_dp_tbl_info_p->sipTbl == (adk_lpm_handle_t)NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* Default SIP filter entries
     * SIP 127.*.*.* -> DENY (loopback)
     * SIP 255.255.255.255 -> DENY (link-local broadcast)
     * SIP 224-239.*.*.* -> DENY (multicast)
     */
    ip_addr = adk_htonl(0x7F000000);
    if(ADK_SUCCESS != adk_ip_dp_sip_add(ipHandle, ADK_IP_DEFAULT_VRD_ID,
                                           ip_addr, 8, ADK_CONVERT_TO_LPM_DATA(ADK_IP_SIP_ADDR_ACTION_BLOCK_ALL)))
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_addr = adk_htonl(0xFFFFFFFF);
    if(ADK_SUCCESS != adk_ip_dp_sip_add(ipHandle, ADK_IP_DEFAULT_VRD_ID,
                                           ip_addr, 32, ADK_CONVERT_TO_LPM_DATA(ADK_IP_SIP_ADDR_ACTION_BLOCK_ALL)))
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_addr = adk_htonl(0xE0000000);
    if(ADK_SUCCESS != adk_ip_dp_sip_add(ipHandle, ADK_IP_DEFAULT_VRD_ID,
                                           ip_addr, 4, ADK_CONVERT_TO_LPM_DATA(ADK_IP_SIP_ADDR_ACTION_BLOCK_ALL)))
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 4. Ininilize interface table */
    ip_dp_tbl_info_p->ipInterfaceTble = (adk_ip_dp_interface_entry_t *)
       ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_VRD_HMODE_TBL_OBJ, 
			1, ADK_DP_MAX_IP_INTERFACE * sizeof(adk_ip_dp_interface_entry_t), 0);
    if(ip_dp_tbl_info_p->ipInterfaceTble == NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 5. Initilize VRD host mode table */
    ip_dp_tbl_info_p->vrdHostModeTble = (bool *)
       ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
	 ADK_IP_VRD_HMODE_TBL_OBJ, 1,
	 /* ADK_IP_VRD_ID_MAX * */ sizeof(bool), 0);
    if(ip_dp_tbl_info_p->vrdHostModeTble == NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 6. Initilize IP Reassembly lookup table and session table*/
    ip_dp_tbl_info_p->ipReassemblyTbl = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_REASS_TBL_OBJ, 1,
               (IP_REASSEMBLY_MAX * sizeof(adk_ip_dp_reassembly_tbl_t)), 0);
    if(!ip_dp_tbl_info_p->ipReassemblyTbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    
    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_mempool_create", "creating memory pool: %s\n", ADK_IP_DP_REASM_SESSION_TBL_NAME);
    ip_dp_tbl_info_p->ipReassemblySessionTbl =
            rte_mempool_create(ADK_IP_DP_REASM_SESSION_TBL_NAME,
            ADK_IP_MAX_REASSEMBLY_SESSIONS,
            sizeof(adk_ip_dp_reassembly_session_tbl_t),
            ADK_IP_DP_REASM_SESSION_CACHE_SIZE,
            0, NULL, NULL, NULL, NULL, rte_socket_id(), 0);

    if(ip_dp_tbl_info_p->ipReassemblySessionTbl == NULL)
        return ADK_E_INTERNAL_ERROR;

    ip_dp_tbl_info_p->reasmMinTout = ADK_REASSEMBLY_TIMEOUT_MILLISECONDS;
    ip_dp_tbl_info_p->reasmTout = ADK_REASSEMBLY_TIMEOUT_MILLISECONDS;
    ip_dp_tbl_info_p->reasmStage1 = ADK_REASSEMBLY_TIMEOUT_MILLISECONDS;
    ip_dp_tbl_info_p->reasmStage2 = ADK_REASSEMBLY_TIMEOUT_MILLISECONDS;
    ip_dp_tbl_info_p->reasmStage3 = ADK_REASSEMBLY_TIMEOUT_MILLISECONDS;
    ip_dp_tbl_info_p->reasmStage4 = ADK_REASSEMBLY_TIMEOUT_MILLISECONDS;

    /* Set all elements to 0. This indicates that all protocols are unsupported at start */
    ADK_MEMSET(&ip_dp_tbl_info_p->allowedL4protos, 0, sizeof(ip_dp_tbl_info_p->allowedL4protos));

    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_spinlock_init", "initialising reassembly table spin lock\n");
    for(i=0; i < IP_REASSEMBLY_MAX; i++)
    {
        rte_spinlock_init(&(ip_dp_tbl_info_p->ipReassemblyTbl[i].lock));
    }

    /* 7. Initialize IP Egress Flow table */
    ip_dp_tbl_info_p->ip_egress_hashtbl = (adk_ip_hash_tbl_t *)
            ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
			ADK_IP_EGRESS_TBL_OBJ,
			1, ADK_IPV4_EGRESS_TBL_HASH_MAX *
            ADK_IPV4_EGRESS_TBL_MAX_COLLISION * sizeof(adk_ip_hash_tbl_t), 0);

    if(!ip_dp_tbl_info_p->ip_egress_hashtbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->ip_egress_ovflow_list_head = NULL;

    ip_dp_tbl_info_p->ip_egflow_tbl= ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_EGRESS_TBL_OBJ, 1,
           ((ADK_DP_MAX_IP_EG_FLOWS+1) * sizeof(adk_ip_dp_eg_flow_info_t)), 0);
    if(!ip_dp_tbl_info_p->ip_egflow_tbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->ip_egress_flow_pending.resource = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_EGRESS_TBL_OBJ, 1,
            ((ADK_DP_MAX_IP_EG_FLOWS) * sizeof(uint16_t)), 0);
    if(!ip_dp_tbl_info_p->ip_egress_flow_pending.resource)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->ip_egress_flow_pending.count  = 0;
    ip_dp_tbl_info_p->ip_egress_flow_pending.ovflow_pending_p = NULL;

    /* 8. Initialize IP Qos table*/
    ip_dp_tbl_info_p->ipQosTbl = (adk_ip_qos_tbl_hdr_t *)
            ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
			ADK_IP_QOS_TBL,
			1,ADK_IPV4_QOS_TBL_HASH_MAX *
            ADK_IPV4_QOS_TBL_MAX_COLLISION * sizeof(adk_ip_qos_tbl_hdr_t), 0);
    if(!ip_dp_tbl_info_p->ipQosTbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->ipQos_ovflow_list_head = NULL;


    /* 10. Initialize interface stats table */
    ip_dp_tbl_info_p->ipInterfaceStatsTble = (adk_ip_dp_intf_stats_t**)
       ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_STATS_OBJ,
		1,ADK_MAX_NUM_CORES * sizeof(adk_ip_dp_intf_stats_t *), 0);
    if(ip_dp_tbl_info_p->ipInterfaceStatsTble == NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    for(i=0; i<(ADK_MAX_NUM_CORES);i++)
    {
        ip_dp_tbl_info_p->ipInterfaceStatsTble[i] = (adk_ip_dp_intf_stats_t*)
                    ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                    ADK_IP_STATS_OBJ, 1, ADK_DP_MAX_IP_STATS *sizeof(adk_ip_dp_intf_stats_t),
                    0);
        if(ip_dp_tbl_info_p->ipInterfaceStatsTble[i] == NULL)
        {
            return ADK_E_INTERNAL_ERROR;
        }
    }

    /* 11. Initialize ip-addr stats table */
    ip_dp_tbl_info_p->ipAddrStatsTble = (adk_ip_dp_ip_addr_stats_t **)
                   ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_ADDR_STATS_OBJ,
                           1, ADK_MAX_NUM_CORES * sizeof(adk_ip_dp_ip_addr_stats_t *), 0);
    if(ip_dp_tbl_info_p->ipAddrStatsTble == NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    for(i=0; i<(ADK_MAX_NUM_CORES);i++)
    {
        ip_dp_tbl_info_p->ipAddrStatsTble[i] =
                (adk_ip_dp_ip_addr_stats_t*)ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                ADK_IP_ADDR_STATS_OBJ, 1,
                ADK_IP_MAX_ADDRESSES * sizeof(adk_ip_dp_ip_addr_stats_t), 0);
        if(ip_dp_tbl_info_p->ipAddrStatsTble[i] == NULL)
        {
            return ADK_E_INTERNAL_ERROR;
        }
    }


    /* 12. Initialize exception handle table */
    for(j=0;j<ADK_IP_EXCEPTION_MAX;++j)
        ip_dp_tbl_info_p->exceptionHandlerTble[j]=NULL;

    /* 13. Initialize CP version Id */
    ip_dp_tbl_info_p->ip_cp_version_id = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_CP_VERID_OBJ, 1,
                                     (/* ADK_IP_VRD_ID_MAX * */ sizeof(uint16_t)), 0);
    if(!ip_dp_tbl_info_p->ip_cp_version_id)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 15. Initialize IP Dest MTU table */
    ip_dp_tbl_info_p->ipDestMtuTbl = (adk_ip_dest_mtu_tbl_t *)
            ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
			ADK_IP_DEST_MTU_OBJ,
			1, ADK_IPV4_DST_MTU_TBL_HASH_MAX *
            ADK_IPV4_DST_MTU_TBL_MAX_COLLISION * sizeof(adk_ip_dest_mtu_tbl_t),
            0);
    if(!ip_dp_tbl_info_p->ipDestMtuTbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 16. Reset global stats*/
    for(i=0; i<ADK_MAX_NUM_CORES;i++) {
        ADK_MEMSET(&ip_dp_tbl_info_p->ipGlobalStatsTable[i], 0, sizeof(adk_ip_dp_global_stats_t));
    }

    /* 17. Initialize IP Ing Assoc Flow Tbl & Stats */
    ip_dp_tbl_info_p->ipIngFlowTbl = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_IN_FLOW_TBL_OBJ, 1,
                   (ADK_DP_MAX_IP_ING_ASSOC_FLOWS * sizeof(adk_ip_dp_ing_flow_info_t)), 0);
    if(!ip_dp_tbl_info_p->ipIngFlowTbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->ipIngFlowStats = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_IN_FLOW_TBL_OBJ, 1,
                   (ADK_MAX_NUM_CORES * ADK_DP_MAX_IP_ING_ASSOC_FLOWS *
                    sizeof(adk_ip_dp_ing_flow_stats_t)), 0);
    if(!ip_dp_tbl_info_p->ipIngFlowStats)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 18. Initialize IP flow cache tables*/
    ip_dp_tbl_info_p->fc_global_maintenance_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
                    (ADK_MAX_NUM_CORES * sizeof(adk_ip_dp_fc_count_t)),0);
    if(!ip_dp_tbl_info_p->fc_global_maintenance_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /*  Initialize IP CPPI (slowpath) stats*/
    ip_dp_tbl_info_p->ipCppiStats = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_CPPI_STATS, 1, 
					              ADK_DP_MAX_IP_ING_ASSOC_FLOWS*sizeof(adk_ip_dp_cppi_stats_t), 0);
    if(!ip_dp_tbl_info_p->ipCppiStats)
    {
	return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->lliCacheTbl_5_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * IP_FLOW_HASH_MAX * IP_FLOW_HASH_COLLISION_SIZE *
                sizeof(adk_ip_dp_flow_cache_entry_t)), 0);
    if(!ip_dp_tbl_info_p->lliCacheTbl_5_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->lliNodeTbl_5_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * IP_MAX_FLOWS_PER_CPU *
                sizeof(adk_ip_dp_lli_hash5_tbl_entry_t)), 0);
    if(!ip_dp_tbl_info_p->lliNodeTbl_5_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->lliNodeTbl_5_count_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * sizeof(adk_ip_dp_fc_maintenance_t)), 0);
    if(!ip_dp_tbl_info_p->lliNodeTbl_5_count_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }
   /*Initialize next_free_index as next entry for each node
    * and store it's own index */
    lli_hash5_entry_p = ip_dp_tbl_info_p->lliNodeTbl_5_p;
    for(i=0; i< ADK_MAX_NUM_CORES; i++)
    {
        for(j=0; j < IP_MAX_FLOWS_PER_CPU; j++)
        {
            lli_hash5_entry_p[i*IP_MAX_FLOWS_PER_CPU +j].next_free_index = j+1;
            lli_hash5_entry_p[i*IP_MAX_FLOWS_PER_CPU +j].own_index = j;
        }
    }

    ip_dp_tbl_info_p->lliCacheTbl_3_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),  ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * IP_FLOW_HASH_MAX * IP_FLOW_HASH_COLLISION_SIZE *
                sizeof(adk_ip_dp_flow_cache_entry_t)), 0);
    if(!ip_dp_tbl_info_p->lliCacheTbl_3_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->lliNodeTbl_3_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * IP_MAX_FLOWS_PER_CPU *
                sizeof(adk_ip_dp_lli_hash3_tbl_entry_t)), 0);
    if(!ip_dp_tbl_info_p->lliNodeTbl_3_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->lliNodeTbl_3_count_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * sizeof(adk_ip_dp_fc_maintenance_t)), 0);
    if(!ip_dp_tbl_info_p->lliNodeTbl_3_count_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

   /*Initialize next_free_index as next entry for each node
    * and store it's own index */
    lli_hash3_entry_p = ip_dp_tbl_info_p->lliNodeTbl_3_p;
    for(i=0; i< ADK_MAX_NUM_CORES; i++)
    {
        for(j=0; j < IP_MAX_FLOWS_PER_CPU; j++)
        {
            lli_hash3_entry_p[i*IP_MAX_FLOWS_PER_CPU +j].next_free_index = j+1;
            lli_hash3_entry_p[i*IP_MAX_FLOWS_PER_CPU +j].own_index = j;
        }
    }

    ip_dp_tbl_info_p->uliCacheTbl_5_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * IP_FLOW_HASH_MAX * IP_FLOW_HASH_COLLISION_SIZE *
                sizeof(adk_ip_dp_flow_cache_entry_t)), 0);
    if(!ip_dp_tbl_info_p->uliCacheTbl_5_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->uliNodeTbl_5_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * IP_MAX_FLOWS_PER_CPU *
                sizeof(adk_ip_dp_uli_hash5_tbl_entry_t)), 0);
    if(!ip_dp_tbl_info_p->uliNodeTbl_5_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->uliNodeTbl_5_count_p = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_FLOW_CACHE_TBL_OBJ, 1,
               (ADK_MAX_NUM_CORES * sizeof(adk_ip_dp_fc_maintenance_t)), 0);
    if(!ip_dp_tbl_info_p->uliNodeTbl_5_count_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }
   /*Initialize next_free_index as next entry for each node
    * and store it's own index */
    uli_hash5_entry_p = ip_dp_tbl_info_p->uliNodeTbl_5_p;
    for(i=0; i< ADK_MAX_NUM_CORES; i++)
    {
        for(j=0; j < IP_MAX_FLOWS_PER_CPU; j++)
        {
            uli_hash5_entry_p[i*IP_MAX_FLOWS_PER_CPU +j].next_free_index = j+1;
            uli_hash5_entry_p[i*IP_MAX_FLOWS_PER_CPU +j].own_index = j;
        }
    }

    /* 19. Initialize IP Egress Assoc Table */
    ip_dp_tbl_info_p->ipEgressAssocTbl = (adk_ip_dp_udp_encap_info_t *)
            ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_EGRESS_ASSOC_TBL_OBJ, 1, ADK_DP_MAX_IP_EG_ASSOC_FLOWS *
                                 sizeof(adk_ip_dp_udp_encap_info_t), 0);
    if(!ip_dp_tbl_info_p->ipEgressAssocTbl)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* 20. Initialize the worker list*/
    ip_dp_tbl_info_p->num_worker_lcores = workerInfo->numLcores;
    for (i = 0; i < workerInfo->numLcores; i++)
    {
        ip_dp_tbl_info_p->worker_lcores_list[i] = workerInfo->lcore[i].index;
    }

    /* 21. Initialize indirect memory pool */
#ifdef DPDK_IP_FRAG
    ip_dp_tbl_info_p->packet_indirect_pool = (struct rte_mempool *)ind_pool;
#endif

    /* 22. Initialize IP PBR rule counter */
    ip_dp_tbl_info_p->pbrRuleStats = ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_PBR_RULE_STATS_OBJ, 1,
                   (ADK_MAX_NUM_CORES * ADK_IP_MAX_PBR_RULES*sizeof(uint32_t)), 0);
    if(!ip_dp_tbl_info_p->pbrRuleStats)
    {
        return ADK_E_INTERNAL_ERROR;
    }

#if 1
    /* 23. Allocate & initailise the ACL actions */
    ip_dp_tbl_info_p->pbrActionTable_p = (adk_ip_dp_pbr_action_t *)
                    ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_PBR_ACT_OBJ, 1,
                            ((ADK_IP_MAX_PBR_RULES + 1) * sizeof(adk_ip_dp_pbr_action_t)),
                                       0);
    if(!ip_dp_tbl_info_p->pbrActionTable_p)
    {
        return ADK_E_INTERNAL_ERROR;
    }
#endif

    /* 24. Initialize IP Dest MTU table */
    /* TO DO: Is it needed */
    ip_dp_tbl_info_p->ip_vrd_pbr_rule = (adk_ip_dp_vrd_pbr_rulesetid_t *)
            ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_VRD_PRB_RULE_OBJ, 1, /* ADK_IP_VRD_ID_MAX * */ sizeof(adk_ip_dp_vrd_pbr_rulesetid_t),
            0);
    if(!ip_dp_tbl_info_p->ip_vrd_pbr_rule)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    /* 25. Allocate & initailise the Ing Assoc Otree actions table shadow and active copies */
    ip_dp_tbl_info_p->ingClassifierActionDp_p[0] =
        (adk_ip_dp_ing_classifier_action_info_t *)
                    ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                            ADK_IP_ING_CLASSIFIER_RULE_ACT_OBJ, 1,
                            ((ADK_IP_MAX_INGRESS_CLASSIFIER_RULES + 1) *
                             sizeof(adk_ip_dp_ing_classifier_action_info_t)),
                                       0);
    if(!ip_dp_tbl_info_p->ingClassifierActionDp_p[0])
    {
        return ADK_E_INTERNAL_ERROR;
    }
    ip_dp_tbl_info_p->ingClassifierActionDp_p[1] =
        (adk_ip_dp_ing_classifier_action_info_t *)
                    ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle),
                            ADK_IP_ING_CLASSIFIER_RULE_ACT_OBJ, 1,
                            ((ADK_IP_MAX_INGRESS_CLASSIFIER_RULES + 1) *
                             sizeof(adk_ip_dp_ing_classifier_action_info_t)),
                                       0);
    if(!ip_dp_tbl_info_p->ingClassifierActionDp_p[1])
    {
        return ADK_E_INTERNAL_ERROR;
    }
    /* 26. Allocate and initialize ingress classifier.
     * Note that ingClassifierActionDp_p must be initialized before calling this.
     */
    ret = adk_ip_ingress_classifier_create( ipHandle, &ip_dp_tbl_info_p->ingClassifierInfo_p);
    if (ADK_SUCCESS != ret)
    {
        return ADK_E_INTERNAL_ERROR;
    }
    /* 27. Set the cookie to mark done from CP*/
    ip_dp_tbl_info_p->cookie = ADK_IPV4_MODULE_COOKIE;

    /* 28. Interface ID to CP interface handle mapping */
    ip_dp_tbl_info_p->ifaceHandles = (adk_interface_handle_t *)
                   ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_IFACE_HANDLE_OBJ,
                           1, ADK_DP_MAX_IP_INTERFACE * sizeof(adk_interface_handle_t), 0);
    if(ip_dp_tbl_info_p->ifaceHandles == NULL)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    /* Reset l4 stats table */
    for(i=0; i<ADK_MAX_NUM_CORES;i++) {
        ADK_MEMSET(&(ip_dp_tbl_info_p->ipL4Stats[i]), 0, sizeof(adk_ip_l4_stats_t));
     }

    ip_dp_tbl_info_p->defRouteTblId = ADK_IP_MAX_ROUTING_TABLES-1;

    ret = adk_ip_pbr_create(ipHandle, &ip_dp_tbl_info_p->ipPbrInfo_p);
    if (ADK_SUCCESS != ret)
    {
        return ADK_E_INTERNAL_ERROR;
    }

    ip_dp_tbl_info_p->utilCntrEnabled = TRUE;
    /* TO DO: Free up earlier allocations in case of any failure */
    return ADK_SUCCESS;
}
/*!
 * =============================================================
 * adk_ip_dp_table_shutdown()
 *
 * IP DP table shutdown function -- To be called in CP init process context
 * =============================================================
 */
adk_status_t
adk_ip_dp_table_shutdown(adk_module_handle_t ipHandle)
{
    uint32_t ip_addr;
    uint32_t  i = 0;


    ip_module_info_t * ipModuleInfo_p =
             (ip_module_info_t *) ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_ip_dp_tbl_info_t *ip_dp_tbl_info_p = ipModuleInfo_p->ip_dp_tbl_info_p;

    /* Free all IP tables here */

    /* 1. Free Destination IP Lookup table */
    adk_lpm_free(ip_dp_tbl_info_p->dipTbl);

    /* 2. Free Source IP Validation table */

    /* Remove Default SIP filter entries
     * SIP 127.*.*.* -> DENY (loopback)
     * SIP 255.255.255.255 -> DENY (link-local broadcast)
     * SIP 224-239.*.*.* -> DENY (multicast)
     */
    ip_addr = adk_htonl(0x7F000000);
    adk_ip_dp_sip_remove(ipHandle, ADK_IP_DEFAULT_VRD_ID, ip_addr, 8);

    ip_addr = adk_htonl(0xFFFFFFFF);
    adk_ip_dp_sip_remove(ipHandle, ADK_IP_DEFAULT_VRD_ID, ip_addr, 32);

    ip_addr = adk_htonl(0xE0000000);
    adk_ip_dp_sip_remove(ipHandle, ADK_IP_DEFAULT_VRD_ID, ip_addr, 4);

    adk_lpm_free(ip_dp_tbl_info_p->sipTbl);

    /* 4. Free interface table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipInterfaceTble);

    /* 5. Free VRD host mode table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->vrdHostModeTble);

    /* 6. Free IP Reassembly lookup table and session table*/
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipReassemblyTbl);

    /* TBD: This API is crashing when running with EM. Disabling to avoid the crash. This needs to be fixed properly
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    ADK_IPV4_LOG_EXT_API(ipHandle, "rte_mempool_free", "freeing reassembly session table memory pool\n");
    rte_mempool_free(ip_dp_tbl_info_p->ipReassemblySessionTbl);
    */


    /* 7. Free IP Egress Flow table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ip_egress_hashtbl);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ip_egflow_tbl);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ip_egress_flow_pending.resource);

    /* 8. Free IP Qos table*/
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipQosTbl);

    /* 10. Free interface table */

    for(i=0; i<(ADK_MAX_NUM_CORES);i++)
    {
        ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipInterfaceStatsTble[i]);
    }
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipInterfaceStatsTble);

    /* 11. Free ip-addr table */
    for(i=0; i<(ADK_MAX_NUM_CORES);i++)
    {
        ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipAddrStatsTble[i]);
    }
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipAddrStatsTble);

    /* 12. Free exception handle table. Nothing to be done */

    /* 13. Free CP version Id */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ip_cp_version_id);

    /* 15. Free IP Dest MTU table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipDestMtuTbl);

    /* 16. Global stats, nothing to be done*/

    /* 17. Free IP Ing Assoc Flow Tbl & stats */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipIngFlowTbl);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipIngFlowStats);

    /* 18. Free the flow cache tables */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->fc_global_maintenance_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->lliCacheTbl_5_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->lliNodeTbl_5_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->lliNodeTbl_5_count_p);

    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->lliCacheTbl_3_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->lliNodeTbl_3_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->lliNodeTbl_3_count_p);

    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->uliCacheTbl_5_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->uliNodeTbl_5_p);
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->uliNodeTbl_5_count_p);

    /* 19. Free IP Egress Assoc Table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipEgressAssocTbl);

    /* 20. Free the worker list*/
    ip_dp_tbl_info_p->num_worker_lcores = 0;
    for (i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        ip_dp_tbl_info_p->worker_lcores_list[i] = 0;
    }

    /* 21. Free PBR rule stats */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->pbrRuleStats);

    /* 22. Free the ACL actions table*/
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->pbrActionTable_p);
    
    /* 23. Free the VRD-ID to pbr policy set id mapping table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ip_vrd_pbr_rule);
    
    /* 24. Reset the cookie to mark not-ready from CP*/
    ip_dp_tbl_info_p->cookie = 0;

    /* Reset the allowed proto array */
    ADK_MEMSET(&ip_dp_tbl_info_p->allowedL4protos, 0, sizeof(ip_dp_tbl_info_p->allowedL4protos));

    /* Free the iface handle mapping table */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ifaceHandles);

    adk_ip_pbr_delete(ipHandle, &ip_dp_tbl_info_p->ipPbrInfo_p);

    adk_ip_ingress_classifier_delete(ipHandle, &ip_dp_tbl_info_p->ingClassifierInfo_p);

    /* Free the CPPI stats */
    ADK_DP_SHM_FREE(ip_dp_tbl_info_p->ipCppiStats); 

    /* 25. Free the memzone */
    if(ADK_SUCCESS != adk_shm_destroy(ADK_IP_SHM_MEMZONE_NAME))
    {
        return ADK_E_INTERNAL_ERROR;
    }

    return ADK_SUCCESS;
}



adk_dp_status_t
adk_ip_dp_interface_add(adk_module_handle_t ipHandle,
                        adk_interface_handle_t ipIfHdl,
                        uint16_t ip_stats,
                        uint32_t sys_egress_stats_id
                        )
{
    uint32_t ifres_index, i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    adk_ip_dp_interface_entry_t *dp_entry;
    bool routingEnabled;

    ifres_index = ADK_SYS_INTERFACE_ID_GET(ipIfHdl);
    if(ifres_index >= ADK_DP_MAX_IP_INTERFACE)
    {
	ADK_IPV4_LOG_ERROR(ipHandle, ADK_DP_E_ERROR,
		"IP interface id larger than the max supported\n");
	return ADK_DP_E_ERROR;
    }

    dp_entry = &(ip_module_info->ip_dp_tbl_info_p->ipInterfaceTble[ifres_index]);

    ADK_MEMSET(dp_entry, 0, sizeof(adk_ip_dp_interface_entry_t));

    ip_interface_info_t *if_pvt_info = (ip_interface_info_t *)
	ADK_SYS_INTERFACE_ATTR_PTR_GET(ipIfHdl);
    dp_entry->mtu = if_pvt_info->if_info.mtu;
    dp_entry->vrd_id = if_pvt_info->vrd_id;
    dp_entry->ingress_ruleset_id = if_pvt_info->ingress_ruleset_id;

    dp_entry->enableOutAcl = false;

    dp_entry->enableQoS = ((if_pvt_info->if_info.ipIntfAttr) & ADK_IP_INTERFACE_MASK_QOS_ENABLE);
    dp_entry->if_stats_id = ip_stats;
    dp_entry->sys_egress_stats_id = sys_egress_stats_id;
    
    routingEnabled = ((if_pvt_info->if_info.ipIntfAttr) & ADK_IP_INTERFACE_MASK_FORWARDING_ENABLE);
    dp_entry->routingDisabled = !(routingEnabled); 

    /* update interface address in dp structure */
    dp_entry->if_addr = adk_htonl(if_pvt_info->if_info.ipAddrPrefix.addr);
    for(i=0; i<IP_NUM_DSCP_VALUES; i++)
    {
	dp_entry->dscp_lkup[i] = IP_DEFAULT_PCP;
    }

    /*Set the default PHB - ADK_QOS_DS_PHB_BE for all dscp */
    for(i=0; i<IP_NUM_DSCP_VALUES; i++)
    {
	dp_entry->dscpToPhbClass [i] = ADK_QOS_DS_PHB_CLASS_NONE;
    }
    dp_entry->ifType = ADK_IP_DP_INTF_TYPE_ETH;

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_interface_remove(adk_module_handle_t ipHandle,
                           adk_interface_handle_t ipIfHdl)
{
    uint32_t ifres_index, i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    uint16_t if_stats_id;
    adk_ip_dp_interface_entry_t *dp_entry;
    adk_ip_dp_intf_stats_t *dp_stats;

    ifres_index = ADK_SYS_INTERFACE_ID_GET(ipIfHdl);
    if(ifres_index >= ADK_DP_MAX_IP_INTERFACE)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_DP_E_ERROR,
                           "IP interface id larger than the max supported\n");
        return ADK_DP_E_ERROR;
    }

    dp_entry = &(ip_module_info->ip_dp_tbl_info_p->
                 ipInterfaceTble[ifres_index]);
    /*First set the flag to Invalid*/
    dp_entry->ifType = ADK_IP_DP_INTF_TYPE_INVALID;

    /*Wait for the current readers to be done*/
    adk_rcu_synchronize();

    /*Now clean up*/
    if_stats_id = dp_entry->if_stats_id;
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->
                     ipInterfaceStatsTble[i][if_stats_id]);
        ADK_MEMSET(dp_stats, 0, sizeof(adk_ip_dp_intf_stats_t));
    }

    ADK_MEMSET(dp_entry, 0, sizeof(adk_ip_dp_interface_entry_t));

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_interface_stats_get(adk_module_handle_t ipHandle,
                              adk_interface_handle_t ipIfHdl,
                              adk_ip_interface_stats_t *stats_p)
{
    uint32_t ifres_index, i,j;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    uint16_t if_stats_id;
    adk_ip_dp_intf_stats_t *dp_stats;
    ifres_index = ADK_SYS_INTERFACE_ID_GET(ipIfHdl);
    if(ifres_index >= ADK_DP_MAX_IP_INTERFACE)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_DP_E_ERROR,
                "IP interface id larger than the max supported\n");
        return ADK_DP_E_ERROR;
    }

    if_stats_id = ip_module_info->ip_dp_tbl_info_p->
        ipInterfaceTble[ifres_index].if_stats_id;
    ADK_MEMSET(stats_p, 0, sizeof(adk_ip_interface_stats_t));
    for(i=0;i<ADK_MAX_NUM_CORES;i++) 
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->
                ipInterfaceStatsTble[i][if_stats_id]);

        stats_p->ifInOctets                         += dp_stats->ifInOctets;
        stats_p->ifInPkts                           += dp_stats->ifInPkts;

        stats_p->ifInMcastOctets                    += dp_stats->ifInMcastOctets;
        stats_p->ifInMcastPkts                      += dp_stats->ifInMcastPkts;

        stats_p->ifOutMcastOctets                   += dp_stats->ifOutMcastOctets;
        stats_p->ifOutMcastPkts                     += dp_stats->ifOutMcastPkts;

        stats_p->ifOutOctets                        += dp_stats->ifOutUcastOctets
                                                    +  dp_stats->ifOutMcastOctets
                                                    +  dp_stats->ifOutBcastOctets;

        stats_p->ifOutPkts                          += dp_stats->ifOutUcastPkts
                                                    +  dp_stats->ifOutMcastPkts
                                                    +  dp_stats->ifOutBcastPkts;

        stats_p->ifInIpCksumErrDrops                += dp_stats->ifInIpCksumErrDrops;
        stats_p->ifInMBcastDMacUcastDipErrDrops     += dp_stats->ifInMBcastDMacUcastDipErrDrops;
        stats_p->ifInBadIpLengthErrDrops            += dp_stats->ifInBadIpLengthErrDrops;
        stats_p->ifOutTtlExceededErrDrops           += dp_stats->ifOutTtlExceededErrDrops;
        stats_p->ifInVersionMismatchErrDrops        += dp_stats->ifInVersionMismatchErrDrops;
        stats_p->ifInInvalidVersionErrDrops         += dp_stats->ifInInvalidVersionErrDrops;
        stats_p->ifInIpOptionsPresentDrops          += dp_stats->ifInIpOptionsPresentDrops;
        stats_p->ifInUnknownProtoErrDrops           += dp_stats->ifInUnknownProtoErrDrops;
        stats_p->ifInUDPCksumErrDrops               += dp_stats->ifInUDPCksumErrDrops;
        stats_p->ifInICMPCksumErrDrops              += dp_stats->ifInIcmpCksumErrors;
        stats_p->ifInZeroDipDrops                   += dp_stats->ifInZeroDipDrops;

        stats_p->ifOutFragPkts                      += dp_stats->ifOutFragPkts;
        stats_p->ifOutFragments                     += dp_stats->ifOutFragments;
        stats_p->ifOutDroppedPkts                   += dp_stats->ifOutDroppedPkts;
        stats_p->ifOutDroppedOctets                 += dp_stats->ifOutDroppedOctets;
        stats_p->ifRoutingDisabledDrops             += dp_stats->ifRoutingDisabledDrops;
        stats_p->ifOutQoSLookupFailDrops            += dp_stats->ifOutQoSLookupFailDrops;

        for(j = 0; j < ADK_QOS_MAX_PHB_CLASS_IDX ; j++) {
            stats_p->ifRxQoSstats[j].rxPkts         += dp_stats->ifRxQoSstats[j].rxPkts;
            stats_p->ifRxQoSstats[j].rxBytes        += dp_stats->ifRxQoSstats[j].rxBytes;
        } 
        
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_interface_stats_get */

adk_status_t adk_ip_dp_interface_stats_clear(
        adk_module_handle_t ipHandle
)
{
    uint32_t i, j, k;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_ip_dp_intf_stats_t *dp_stats;

    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        for(j=0;j<ADK_DP_MAX_IP_STATS;j++)
        {
            dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipInterfaceStatsTble[i][j]);

            dp_stats->ifInOctets = 0;
            dp_stats->ifInPkts = 0;
            dp_stats->ifInMcastOctets = 0;
            dp_stats->ifInMcastPkts = 0;
            dp_stats->ifOutMcastOctets = 0;
            dp_stats->ifOutMcastPkts = 0;
            dp_stats->ifOutUcastOctets = 0;
            dp_stats->ifOutMcastOctets = 0;
            dp_stats->ifOutBcastOctets = 0;
            dp_stats->ifOutUcastPkts = 0;
            dp_stats->ifOutMcastPkts = 0;
            dp_stats->ifOutBcastPkts = 0;
            dp_stats->ifInIpCksumErrDrops = 0;
            dp_stats->ifInMBcastDMacUcastDipErrDrops = 0;
            dp_stats->ifInBadIpLengthErrDrops = 0;
            dp_stats->ifOutTtlExceededErrDrops = 0;
            dp_stats->ifInVersionMismatchErrDrops = 0;
            dp_stats->ifInInvalidVersionErrDrops = 0;
            dp_stats->ifInIpOptionsPresentDrops = 0;
            dp_stats->ifInUnknownProtoErrDrops = 0;
            dp_stats->ifInUDPCksumErrDrops = 0;
            dp_stats->ifInIcmpCksumErrors = 0;
            dp_stats->ifInZeroDipDrops = 0;
            dp_stats->ifOutFragPkts = 0;
            dp_stats->ifOutFragments = 0;
            dp_stats->ifOutDroppedPkts = 0;
            dp_stats->ifOutDroppedOctets = 0;
            dp_stats->ifRoutingDisabledDrops = 0;
            dp_stats->ifOutQoSLookupFailDrops = 0;

            for(k = 0; k < ADK_QOS_MAX_PHB_CLASS_IDX ; k++) {
                dp_stats->ifRxQoSstats[k].rxPkts = 0;
                dp_stats->ifRxQoSstats[k].rxBytes = 0;
            }
        }
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_interface_stats_clear */

adk_dp_status_t
adk_ip_dp_ip_addr_stats_get(adk_module_handle_t ipHandle,
                              uint16_t ip_addr_stats_id,
                              adk_ip_address_stats_t *stats_p)
{
    uint32_t i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_ip_dp_ip_addr_stats_t *dp_stats;

    ADK_MEMSET(stats_p, 0, sizeof(adk_ip_address_stats_t));


    for(i=0;i<ADK_MAX_NUM_CORES;i++) {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->
                ipAddrStatsTble[i][ip_addr_stats_id]);

        stats_p->ipInOctets     += dp_stats->ipInOctets;
        stats_p->ipInPkts       += dp_stats->ipInPkts;
        stats_p->ipOutOctets    += dp_stats->ipOutOctets;
        stats_p->ipOutPkts      += dp_stats->ipOutPkts;
        stats_p->ipInReasmPkts  += dp_stats->ipInReasmPkts;
        stats_p->ipInErrors     += dp_stats->ipInErrors;
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_ip_addr_stats_get */

adk_dp_status_t
adk_ip_dp_ip_addr_stats_clear(adk_module_handle_t ipHandle)
{
    uint32_t i, j;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_ip_dp_ip_addr_stats_t *dp_stats;

    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        for(j=0; j< (ADK_IP_MAX_ADDRESSES); j++)
        {
            dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipAddrStatsTble[i][j]);

            dp_stats->ipInOctets = 0;
            dp_stats->ipInPkts = 0;
            dp_stats->ipOutOctets = 0;
            dp_stats->ipOutPkts = 0;
            dp_stats->ipInReasmPkts = 0;
            dp_stats->ipInErrors = 0;
        }
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_ip_addr_stats_clear */



adk_dp_status_t
adk_ip_dp_virtual_if_stats_get(adk_module_handle_t ipHandle,
                               uint16_t if_stats_id,
                               adk_ip_interface_stats_t *stats_p)
{
    uint32_t i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_ip_dp_intf_stats_t *dp_stats;

    ADK_MEMSET(stats_p, 0, sizeof(adk_ip_interface_stats_t));
    for(i=0;i<ADK_MAX_NUM_CORES;i++) {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->
                     ipInterfaceStatsTble[i][if_stats_id]);
        stats_p->ifInPkts += dp_stats->ifInPkts;
    }

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_global_stats_get(adk_module_handle_t ipHandle,
                              adk_ip_global_stats_t *stats_p)
{
    uint32_t i;
    adk_ip_dp_global_stats_t *dp_stats;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    ADK_MEMSET(stats_p, 0, sizeof(adk_ip_global_stats_t));

    for(i=0;i<ADK_MAX_NUM_CORES;i++) 
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipGlobalStatsTable[i]);

        stats_p->ipOutTooBigDrops                   += dp_stats->ipOutTooBigDrops;
        stats_p->ipInReasmToutFragDrops             += dp_stats->ipInReasmToutFragDrops;

        stats_p->ipInReasmNoResourceDrops           += dp_stats->ipInReasmHashCollisionDrops
                                                    +  dp_stats->ipInReasmTimerCreateFailDrops
                                                    +  dp_stats->ipInReasmSessionAllocFailDrops;

        stats_p->ipInReasmFragBufferFullDrops       +=  dp_stats->ipInReasmFragBufferFullDrops;

        stats_p->ipInReasmHolesExceededDrops        +=  dp_stats->ipInReasmHolesExceededDrops;

        stats_p->ipInClassifierDefaultDrops         += dp_stats->ipInClassifierDefaultDrops;
        stats_p->ipNoExceptionRegisteredDrops       += dp_stats->ipNoExceptionRegisteredDrops;
        stats_p->ipExceptionSendFailureDrops        += dp_stats->ipExceptionSendFailureDrops;
        stats_p->ipInPolicerDrops                   += dp_stats->ipInPolicerDrops;
        stats_p->ipInClassifierDrops                += dp_stats->ipInClassifierDrops;
        stats_p->ipInSendToEoFailDrops              += dp_stats->ipInSendToEoFailDrops;
        stats_p->ipInSendToHostFailDrops            += dp_stats->ipInSendToHostFailDrops;
        stats_p->ipInClassifierUnknownActionDrops   += dp_stats->ipInClassifierUnknownActionDrops;
        stats_p->ipInReasmDuplicateDrops            += dp_stats->ipInReasmDuplicateDrops;
        stats_p->ipOutInvalidIpHdrLenDrops          += dp_stats->ipOutInvalidIpHdrLenDrops;
        stats_p->ipOutNonLocalSipFromEoDrops        += dp_stats->ipOutNonLocalSipFromEoDrops;
        stats_p->ipInReasmToutMbufAllocFails        += dp_stats->ipInReasmToutMbufAllocFails;
        stats_p->ipOutFragMbufAllocFailDrops        += dp_stats->ipOutFragMbufAllocFailDrops;
        stats_p->ipOutTruncatedPktDrops             += dp_stats->ipOutTruncatedPktDrops;
        stats_p->ipInInvalidIfDrops                 += dp_stats->ipInInvalidIfDrops;
        stats_p->ipOutInvalidIfDrops                += dp_stats->ipOutInvalidIfDrops;
        stats_p->ipOutNoRouteFoundDrops             += dp_stats->ipOutNoRouteFoundDrops;
        stats_p->ipOutFragDFSetDrops                += dp_stats->ipOutFragDFSetDrops;
        stats_p->ipInReasmTouts                     += dp_stats->ipInReasmTouts;
    }

    /* Aggregate CPPI stats into global stats. TO DO: this is wrong. we are not taking
     * counters from global backup table. Also, we don't need per flow counters. Remove it and
     * directly increment global counter */
    for(i=0;i<ADK_DP_MAX_IP_ING_ASSOC_FLOWS;i++)
    {
        stats_p->ipInPolicerDrops += ip_module_info->ip_dp_tbl_info_p->ipCppiStats[i].cppiPolicerDropPkts;
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_global_stats_get */


adk_dp_status_t
adk_ip_dp_global_stats_clear(adk_module_handle_t ipHandle)
{
    uint32_t i;
    adk_ip_dp_global_stats_t *dp_stats;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);


    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipGlobalStatsTable[i]);

        dp_stats->ipOutTooBigDrops = 0;
        dp_stats->ipInReasmToutFragDrops = 0;
        dp_stats->ipInReasmHashCollisionDrops = 0;
        dp_stats->ipInReasmTimerCreateFailDrops = 0;
        dp_stats->ipInReasmSessionAllocFailDrops = 0;
        dp_stats->ipInReasmFragBufferFullDrops = 0;
        dp_stats->ipInReasmHolesExceededDrops = 0;
        dp_stats->ipInClassifierDefaultDrops = 0;
        dp_stats->ipNoExceptionRegisteredDrops = 0;
        dp_stats->ipExceptionSendFailureDrops = 0;
        dp_stats->ipInPolicerDrops = 0;
        dp_stats->ipInClassifierDrops = 0;
        dp_stats->ipInSendToEoFailDrops = 0;
        dp_stats->ipInSendToHostFailDrops = 0;
        dp_stats->ipInClassifierUnknownActionDrops = 0;
        dp_stats->ipInReasmDuplicateDrops = 0;
        dp_stats->ipOutInvalidIpHdrLenDrops = 0;
        dp_stats->ipOutNonLocalSipFromEoDrops = 0;
        dp_stats->ipInReasmToutMbufAllocFails = 0;
        dp_stats->ipOutFragMbufAllocFailDrops = 0;
        dp_stats->ipOutTruncatedPktDrops = 0;
        dp_stats->ipInInvalidIfDrops = 0;
        dp_stats->ipOutInvalidIfDrops = 0;
        dp_stats->ipOutNoRouteFoundDrops = 0;
        dp_stats->ipOutFragDFSetDrops = 0;
        dp_stats->ipInReasmTouts = 0;
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_global_stats_clear */

adk_dp_status_t
adk_ip_dp_l4_stats_get(adk_module_handle_t ipHandle,
                              adk_ip_l4_stats_t *stats_p)
{
    uint32_t i;
    adk_ip_dp_l4_stats_t *dp_stats;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    ADK_MEMSET(stats_p, 0, sizeof(adk_ip_l4_stats_t));

    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipL4Stats[i]);

        stats_p->ipSctpInOctets         += dp_stats->ipSctpInOctets;
        stats_p->ipSctpInPackets        += dp_stats->ipSctpInPackets;
        stats_p->ipSctpOutOctets        += dp_stats->ipSctpOutOctets;
        stats_p->ipSctpOutPackets       += dp_stats->ipSctpOutPackets;

        stats_p->ipTcpInOctets          += dp_stats->ipTcpInOctets;
        stats_p->ipTcpInPackets         += dp_stats->ipTcpInPackets;
        stats_p->ipTcpOutOctets         += dp_stats->ipTcpOutOctets;
        stats_p->ipTcpOutPackets        += dp_stats->ipTcpOutPackets;
    }

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_l4_stats_clear(adk_module_handle_t ipHandle)
{
    uint32_t i;
    adk_ip_dp_l4_stats_t *dp_stats;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipL4Stats[i]);

        dp_stats->ipSctpInOctets = 0;
        dp_stats->ipSctpInPackets = 0;
        dp_stats->ipSctpOutOctets = 0;
        dp_stats->ipSctpOutPackets = 0;
        dp_stats->ipTcpInOctets = 0;
        dp_stats->ipTcpInPackets = 0;
        dp_stats->ipTcpOutOctets = 0;
        dp_stats->ipTcpOutPackets = 0;
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_l4_stats_clear */

adk_dp_status_t
adk_ip_dp_vrd_host_mode(adk_module_handle_t ipHandle,
                        uint8_t vrd_id,
                        bool host_mode)
{
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    ip_module_info->ip_dp_tbl_info_p->vrdHostModeTble[vrd_id] = host_mode;
    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_qos_policy_add(adk_module_handle_t ipHandle, uint32_t ifres_index,
                         uint8_t dscp, uint32_t queueId)
{
    uint32_t hash_val,hash_index,i;
    adk_ip_qos_tbl_t *qos_node;
    adk_ip_qostbl_ovflow_t *qos_ovflow_node,*temp_qos;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    if(ifres_index >= ADK_DP_MAX_IP_INTERFACE)
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_DP_E_ERROR,
                           "IP interface id larger than the max supported\n");
        return ADK_DP_E_ERROR;
    }

    hash_val = SISD_CRC_HASH(0,ifres_index);
    hash_val = SISD_CRC_HASH(hash_val, dscp);

    hash_index = hash_val & ADK_IPV4_QOS_TBL_HASH_MASK;
    hash_index *= ADK_IPV4_QOS_TBL_MAX_COLLISION;

    for(i=0;i<ADK_IPV4_QOS_TBL_MAX_COLLISION;i++)
    {
        if(ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].qos_entry_p == NULL)
        {
            qos_node = (adk_ip_qos_tbl_t *)ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_QOS_POL_ADD_OBJ,
				 1,sizeof(adk_ip_qos_tbl_t), 0);
            if(qos_node == NULL)
                return ADK_DP_E_ERROR;

            ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].hash_val = hash_val;
            qos_node->ifid = ifres_index;
            qos_node->dscp = dscp;
            qos_node->queueid = queueId;
            ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].qos_entry_p= qos_node;
            goto done;
        }
    }
    if(i == ADK_IPV4_QOS_TBL_MAX_COLLISION)
    {
        qos_ovflow_node = (adk_ip_qostbl_ovflow_t *)ADK_DP_SHM_CALLOC
                            (ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_QOS_POL_ADD_OBJ, 1, sizeof(adk_ip_qostbl_ovflow_t), 0);
        if(qos_ovflow_node == NULL)
            return ADK_DP_E_ERROR;
        qos_ovflow_node->ifid = ifres_index;
        qos_ovflow_node->dscp = dscp;
        qos_ovflow_node->queueid = queueId;
        qos_ovflow_node->next = NULL;

        if(ip_module_info->ip_dp_tbl_info_p->ipQos_ovflow_list_head == NULL)
        {
            ip_module_info->ip_dp_tbl_info_p->ipQos_ovflow_list_head = qos_ovflow_node;
        }
        else
        {
            temp_qos = ip_module_info->ip_dp_tbl_info_p->ipQos_ovflow_list_head;
            while(temp_qos->next != NULL)
                temp_qos = temp_qos->next;
            temp_qos->next = qos_ovflow_node;
        }
    }
done:
    return ADK_SUCCESS;
}

adk_dp_status_t adk_ip_dp_qos_policy_delete(adk_module_handle_t ipHandle,
                                         uint32_t ipIfId, uint8_t dscp)
{
    uint32_t hash_val,hash_index,i;
    adk_ip_qos_tbl_t *qos_node;
    adk_ip_qostbl_ovflow_t *prev,*current;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    hash_val = SISD_CRC_HASH(0,ipIfId);
    hash_val = SISD_CRC_HASH(hash_val, dscp);

    hash_index = hash_val & ADK_IPV4_QOS_TBL_HASH_MASK;
    hash_index *= ADK_IPV4_QOS_TBL_MAX_COLLISION;

    for(i=0; i<ADK_IPV4_QOS_TBL_MAX_COLLISION; i++)
    {
        if((ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].qos_entry_p != NULL)&&
            (ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].hash_val == hash_val))
        {
            if((ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].qos_entry_p->ifid == ipIfId)&&
                (ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].qos_entry_p->dscp == dscp))
            {
                qos_node = ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index +i].qos_entry_p;
                ip_module_info->ip_dp_tbl_info_p->ipQosTbl[hash_index + i].qos_entry_p = NULL;

                adk_rcu_synchronize();
                ADK_DP_SHM_FREE(qos_node);
                goto done;
            }
        }
    }
    if(i == ADK_IPV4_QOS_TBL_MAX_COLLISION)
    {
        current = ip_module_info->ip_dp_tbl_info_p->ipQos_ovflow_list_head;
        prev = NULL;
        if(current == NULL)
           return ADK_DP_E_ERROR;

        if((current->ifid == ipIfId)&&(current->dscp == dscp))
        {
            ip_module_info->ip_dp_tbl_info_p->ipQos_ovflow_list_head = current->next;
            adk_rcu_synchronize();
            ADK_DP_SHM_FREE(current);
            return ADK_SUCCESS;
        }
        while((current != NULL)&&
                !((current->ifid == ipIfId)&&(current->dscp == dscp)))
        {
            prev = current;
            current = current->next;
        }
        if(current == NULL)
            return ADK_DP_E_ERROR;

        prev->next = current->next;

        adk_rcu_synchronize();
        ADK_DP_SHM_FREE(current);
        return ADK_SUCCESS;
    }
done:
    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_dip_add(adk_module_handle_t ipHandle, uint8_t vrd_id,
                  uint32_t dip, uint8_t mask, adk_handle_t dip_result)
{
    unsigned char key[(IP_DIP_TBL_WIDTH/8)+1] = {};
    int size = 0;
    adk_handle_t dip_old_result=ADK_CONVERT_TO_LPM_DATA(NULL);
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    /* Clear pending List if distributed RCU is success */
    ADK_RCU_DISTRIBUTED_PENDING_LIST_CHECK(ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.u.pending_list,
                                           ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.list_count,
                                           adk_rcu_shm_free_cb_fn, NULL);


    if( ADK_DP_E_ERROR == adk_ip_dp_diptbl_key_create(vrd_id, dip, mask, key, &size))
    {
        return ADK_DP_E_ERROR;
    }

    if(adk_lpm_is_rule_present(ip_module_info->ip_dp_tbl_info_p->dipTbl, key, size, &dip_old_result))
    {
#if 0
        /*Free the old result*/
        if(dip_old_result)
        {
            ADK_DP_SHM_FREE(dip_old_result);
        }
#endif
    }

    if(adk_lpm_add(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_DIP_ADD_OBJ,
	ip_module_info->ip_dp_tbl_info_p->dipTbl, key, size, dip_result) != 0)
    {
        return ADK_DP_E_ERROR;
    }

    /* release the old index if overwritten */
    if(dip_old_result != ADK_CONVERT_TO_LPM_DATA(NULL))
    {
        ADK_IPV4_LOG_INFO(ipHandle,
                          "entry was overwritten, releasing the"
                          "old resources %"ADKPRIh"\n", dip_old_result);

        ADK_RCU_RELEASE_DISTRIBUTED_PENDING_LIST(ADK_MODULE_MODULEID_GET(ipHandle),
                                                 ADK_IP_DIP_TBL_OBJ,
                                                 ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.u.pending_list,
                                                 dip_old_result,
                                                 ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.list_count,
                                                 ADK_IP_DIP_TBL_PENDING_SIZE,
                                                 adk_rcu_shm_free_cb_fn, NULL);
        return ADK_E_ENTRY_OVERWRITTEN;
    }
    else
    {
        adk_sys_flow_cache_update_version(sys_handle);
        return ADK_SUCCESS;
    }
}

adk_dp_status_t
adk_ip_dp_dip_hw_rule_remove(adk_module_handle_t ipHandle, uint8_t vrd_id,
                     uint32_t dip, uint8_t mask)
{
    unsigned char key[(IP_DIP_TBL_WIDTH/8)+1] = {};
    int size = 0;
    adk_handle_t dip_result=ADK_CONVERT_TO_LPM_DATA(NULL);
    ip_module_info_t *ip_module_info =
        ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);
    adk_status_t retValue = ADK_SUCCESS;

    if( ADK_DP_E_ERROR ==
        adk_ip_dp_diptbl_key_create(vrd_id, dip, mask, key, &size))
    {
        return ADK_DP_E_ERROR;
    }

    ADK_IPV4_LOG_INFO(ipHandle,
                      "Removing HW rule for Dest IP 0x%x...\n",
                      dip);

    if(adk_lpm_is_rule_present(ip_module_info->ip_dp_tbl_info_p->
                               dipTbl, key, size, &dip_result))
    {
        if(dip_result)
        {
            /* Check if HW termination rule is present*/
            if(((adk_ip_dp_route_info_t *)dip_result)->isTerminating)
            {
                /*Free the HW rule from IP TERMINATION table*/
                retValue = adk_hos_rule_delete(
                               ADK_HOS_IP_TERM_TBL,
                               ((adk_ip_dp_route_info_t *)dip_result)
                                ->ip_term_hos_rule_id[0]);
                if (retValue != ADK_SUCCESS)
                {
                    ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                            "Failed to remove IP-Term HOS rule for DIP\n");
                    return ADK_DP_E_ERROR;
                }
#ifdef ADK_HOS_OLD_IMPL
                retValue = adk_hos_rule_delete(
                               ADK_HOS_INNER_IP4_TERM_TBL,
                               ((adk_ip_dp_route_info_t *)dip_result)
                                ->ip_term_hos_rule_id[1]);
                if (retValue != ADK_SUCCESS)
                {
                    ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                            "Failed to remove IP-Term HOS rule for DIP\n");
                    return ADK_DP_E_ERROR;
                }
#endif
            }
        }
    }
    return retValue;
}

adk_dp_status_t
adk_ip_dp_dip_remove(adk_module_handle_t ipHandle, uint8_t vrd_id,
                     uint32_t dip, uint8_t mask)
{
    unsigned char key[(IP_DIP_TBL_WIDTH/8)+1] = {};
    int size = 0;
    adk_handle_t dip_result=ADK_CONVERT_TO_LPM_DATA(NULL);
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    /* Clear pending List if distributed RCU is success */
    ADK_RCU_DISTRIBUTED_PENDING_LIST_CHECK(ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.u.pending_list,
                                           ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.list_count,
                                           adk_rcu_shm_free_cb_fn, NULL);

    if( ADK_DP_E_ERROR == adk_ip_dp_diptbl_key_create(vrd_id, dip, mask, key, &size))
    {
        return ADK_DP_E_ERROR;
    }

    ADK_IPV4_LOG_INFO(ipHandle,
                      "Removing Dest IP 0x%x...\n",
                      dip);

    if(adk_lpm_is_rule_present(ip_module_info->ip_dp_tbl_info_p->dipTbl, key, size, &dip_result))
    {
#if 0
        /*Free the result*/
        if(dip_result)
        {
            ADK_DP_SHM_FREE(dip_result);
        }
#endif
        if(adk_lpm_delete(ip_module_info->ip_dp_tbl_info_p->dipTbl, key, size) == 0)
        {

            if(dip_result)
            {
                ADK_RCU_RELEASE_DISTRIBUTED_PENDING_LIST(ADK_MODULE_MODULEID_GET(ipHandle),
                                                         ADK_IP_DIP_TBL_OBJ,
                                                         ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.u.pending_list,
                                                         dip_result,
                                                         ip_module_info->ip_dp_tbl_info_p->dipTbl_pending.list_count,
                                                         ADK_IP_DIP_TBL_PENDING_SIZE,
                                                         adk_rcu_shm_free_cb_fn, NULL);
            }

            adk_sys_flow_cache_update_version(sys_handle);
            return ADK_SUCCESS;
        }
    }
    return ADK_DP_E_ERROR;
}

adk_dp_status_t
adk_ip_dp_sip_add(adk_module_handle_t ipHandle, uint8_t vrd_id,
                  uint32_t sip, uint8_t mask, adk_handle_t sip_result)
{
    unsigned char key[(IP_SIP_TBL_WIDTH/8)+1] = {};
    int size = 0;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    if( ADK_DP_E_ERROR == adk_ip_dp_siptbl_key_create(vrd_id, sip, mask, key, &size))
    {
        return ADK_DP_E_ERROR;
    }

    ADK_IPV4_LOG_INFO(ipHandle,
                      "Inserting Vrd Id:0x%x, Src IP 0x%x with %"ADKPRIh"...\n",
                      vrd_id, sip, sip_result);

    if(adk_lpm_add(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_SIP_ADD_OBJ,
	ip_module_info->ip_dp_tbl_info_p->sipTbl, key, size, sip_result) == 0)
    {
        return ADK_SUCCESS;
    }
    return ADK_DP_E_ERROR;
}

adk_dp_status_t
adk_ip_dp_sip_remove(adk_module_handle_t ipHandle, uint8_t vrd_id,
                     uint32_t sip, uint8_t mask)
{
    unsigned char key[(IP_SIP_TBL_WIDTH/8)+1] = {};
    int size = 0;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    if( ADK_DP_E_ERROR == adk_ip_dp_siptbl_key_create(vrd_id, sip, mask, key, &size))
    {
        return ADK_DP_E_ERROR;
    }

    ADK_IPV4_LOG_INFO(ipHandle,
                      "Removing Src IP 0x%x...\n", sip);

    if(adk_lpm_delete(ip_module_info->ip_dp_tbl_info_p->sipTbl, key, size) == 0)
    {
        return ADK_SUCCESS;
    }
    return ADK_DP_E_ERROR;
}



adk_dp_status_t
adk_ip_dp_egress_flow_add(adk_module_handle_t ipHandle,
                          uint32_t ipIfId,
                          uint32_t nh,
                          uint32_t flow_id)
{
    uint32_t hash_val,hash_index,i;
    adk_ip_hash_ovflow_t *node,*temp;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_INFO(ipHandle,
                      "Inserting NH 0x%x, ipIfId 0x%x with 0x%x...\n",
                      nh, ipIfId, flow_id);

    hash_val = SISD_CRC_HASH(0,ipIfId);
    hash_val = SISD_CRC_HASH(hash_val,nh);

    hash_index = hash_val & ADK_IPV4_EGRESS_TBL_HASH_MASK;
    /*hash table is allocated as 1D array. Hence multiply with collision width*/
    hash_index *= ADK_IPV4_EGRESS_TBL_MAX_COLLISION;

    for(i=0; i< ADK_IPV4_EGRESS_TBL_MAX_COLLISION; i++)
    {
        if(!(ip_module_info->ip_dp_tbl_info_p->
            ip_egress_hashtbl[hash_index + i].flow_id))
        {
            ip_module_info->ip_dp_tbl_info_p->
                ip_egress_hashtbl[hash_index + i].hash_val = hash_val;
            ip_module_info->ip_dp_tbl_info_p->ip_egflow_tbl[flow_id].nh = nh;
            ip_module_info->ip_dp_tbl_info_p->ip_egflow_tbl[flow_id].ifid
                                                = ipIfId;
            ip_module_info->ip_dp_tbl_info_p->
                ip_egress_hashtbl[hash_index + i].flow_id = flow_id;
            return ADK_SUCCESS;
        }
    }

    /*i == ADK_IPV4_EGRESS_TBL_MAX_COLLISION*/
    node = (adk_ip_hash_ovflow_t *)ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_TUN_ADD_OBJ, 1,
                        sizeof(adk_ip_hash_ovflow_t), 0);
    if(node == NULL)
        return ADK_DP_E_ERROR;
    node->flow_id = flow_id;
    node->next = NULL;

    ip_module_info->ip_dp_tbl_info_p->ip_egflow_tbl[flow_id].nh = nh;
    ip_module_info->ip_dp_tbl_info_p->ip_egflow_tbl[flow_id].ifid = ipIfId;

    if(ip_module_info->ip_dp_tbl_info_p->ip_egress_ovflow_list_head == NULL)
        ip_module_info->ip_dp_tbl_info_p->ip_egress_ovflow_list_head = node;
    else
    {
        temp = ip_module_info->ip_dp_tbl_info_p->ip_egress_ovflow_list_head;
        while(temp->next != NULL)
            temp = temp->next;
        temp->next = node;
    }

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_egress_flow_remove(adk_module_handle_t ipHandle,
                             uint32_t ipIfId,
                             uint32_t nh,
                             uint32_t flow_id)
{
    uint32_t hash_val, hash_index,i;
    adk_ip_hash_ovflow_t *current,*prev;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_INFO(ipHandle,
                      "Deleting NH 0x%x, ipIfId 0x%x from egress flow lookup\n",
                      nh, ipIfId);

    hash_val = SISD_CRC_HASH(0,ipIfId);
    hash_val = SISD_CRC_HASH(hash_val,nh);

    hash_index = hash_val & ADK_IPV4_EGRESS_TBL_HASH_MASK;
    /*hash table is allocated as 1D array. Hence multiply with collision width*/
    hash_index *= ADK_IPV4_EGRESS_TBL_MAX_COLLISION;

    for(i=0; i< ADK_IPV4_EGRESS_TBL_MAX_COLLISION; i++)
    {
        if(ip_module_info->ip_dp_tbl_info_p->ip_egress_hashtbl[hash_index + i].
                                                    flow_id == flow_id)
        {
            ip_module_info->ip_dp_tbl_info_p->
                        ip_egress_hashtbl[hash_index + i].flow_id = 0;
            /*Move deleted entry to pending list*/
            ip_module_info->ip_dp_tbl_info_p->ip_egress_flow_pending.resource[
              ip_module_info->ip_dp_tbl_info_p->ip_egress_flow_pending.count] =
                                                                 flow_id;
            ip_module_info->ip_dp_tbl_info_p->ip_egress_flow_pending.count++;
            /*Update RCU version*/
            adk_rcu_distributed_ver_update();
            return ADK_SUCCESS;
        }
    }

    /*i == ADK_IPV4_EGRESS_TBL_MAX_COLLISION*/
    if(ip_module_info->ip_dp_tbl_info_p->ip_egress_ovflow_list_head == NULL)
        return ADK_DP_E_ERROR;
    current = ip_module_info->ip_dp_tbl_info_p->ip_egress_ovflow_list_head;
    prev = NULL;

    if(current->flow_id == flow_id)
    {
        ip_module_info->ip_dp_tbl_info_p->ip_egress_ovflow_list_head =
                                                        current->next;
        adk_ip_pushto_pending_list(current,
                &(ip_module_info->ip_dp_tbl_info_p->ip_egress_flow_pending));
        return ADK_SUCCESS;
    }
    while((current != NULL)&&
            !(current->flow_id == flow_id))
    {
        prev = current;
        current = current->next;
    }
    if(current == NULL)
        return ADK_DP_E_ERROR;

    prev->next = current->next;

    adk_ip_pushto_pending_list(current,
            &(ip_module_info->ip_dp_tbl_info_p->ip_egress_flow_pending));

    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_dest_mtu_add(adk_module_handle_t ipHandle,
                          uint16_t ipVrdId,
                          uint32_t dipAddr_nw,
                          uint16_t mtu)
{
    uint32_t hash_val,hash_index,i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_INFO(ipHandle,
           "Dest MTU Tbl: Inserting VrdId 0x%x, DIP 0x%x with MTU%d...\n",
           ipVrdId, adk_ntohl(dipAddr_nw), mtu);

    hash_val = SISD_CRC_HASH(0,ipVrdId);
    hash_val = SISD_CRC_HASH(hash_val,dipAddr_nw);

    hash_index = hash_val & ADK_IPV4_DST_MTU_TBL_HASH_MASK;
    /*hash table is allocated as 1D array. Hence multiply with collision width*/
    hash_index *= ADK_IPV4_DST_MTU_TBL_MAX_COLLISION;

    /*Add with same input should overwrite. Hence first find if entry is present*/
    for(i=0; i< ADK_IPV4_DST_MTU_TBL_MAX_COLLISION; i++)
    {
        /* DIP is overloaded to denote present flag
         * (DIP == 0) => not present*/
        if(ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw)
        {
            if((ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw == dipAddr_nw)
               &&(ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].vrdid == ipVrdId))
            {
                /*Exact match found. Atomically overwrite MTU*/
                ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].mtu = mtu;
                goto _adk_ip_dp_dest_mtu_add_done;
            }
        }
    }

    /*If here means entry not present & add to first free position*/
    for(i=0; i< ADK_IPV4_DST_MTU_TBL_MAX_COLLISION; i++)
    {
        if(!(ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw))
        {
            /*This entry is free, update the data*/
            ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].mtu = mtu;
            ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].vrdid = ipVrdId;
            /*DIP should be the last field to be updated*/
            ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw = dipAddr_nw;
            goto _adk_ip_dp_dest_mtu_add_done;
        }
    }

    /*If here means no free slot in collision list*/
    return ADK_DP_E_ERROR;

_adk_ip_dp_dest_mtu_add_done:
    return ADK_SUCCESS;
}

adk_dp_status_t
adk_ip_dp_dest_mtu_rem(adk_module_handle_t ipHandle,
                          uint16_t ipVrdId,
                          uint32_t dipAddr_nw)
{
    uint32_t hash_val,hash_index,i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_sys_handle_t sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_INFO(ipHandle,
           "Dest MTU Tbl: Removing VrdId 0x%x, DIP 0x%x ...\n",
           ipVrdId, adk_ntohl(dipAddr_nw));

    hash_val = SISD_CRC_HASH(0,ipVrdId);
    hash_val = SISD_CRC_HASH(hash_val,dipAddr_nw);

    hash_index = hash_val & ADK_IPV4_DST_MTU_TBL_HASH_MASK;
    /*hash table is allocated as 1D array. Hence multiply with collision width*/
    hash_index *= ADK_IPV4_DST_MTU_TBL_MAX_COLLISION;

    /*Find if entry is present*/
    for(i=0; i< ADK_IPV4_DST_MTU_TBL_MAX_COLLISION; i++)
    {
        /* DIP is overloaded to denote present flag
         * (DIP == 0) => not present*/
        if(ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw)
        {
            if((ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw == dipAddr_nw)
               &&(ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].vrdid == ipVrdId))
            {
                /*Exact match found. Reset DIP to denote entry is free*/
                ip_module_info->ip_dp_tbl_info_p->ipDestMtuTbl[hash_index + i].dipAddr_nw = 0;
                /*Wait for all current readers*/
                adk_rcu_synchronize();
                return ADK_SUCCESS;
            }
        }
    }

    /*If here means entry not present, hence returen error*/
    return ADK_DP_E_ERROR;
}

adk_dp_status_t
adk_ip_dp_utility_stats_get(adk_module_handle_t ipHandle,
                              adk_ip_utility_stats_t *stats_p)
{
    uint32_t i;
    adk_ip_dp_global_stats_t *dp_stats;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    ADK_MEMSET(stats_p, 0, sizeof(adk_ip_utility_stats_t));

    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipGlobalStatsTable[i]);

        stats_p->ipReasmReqPkts                    += dp_stats->UipReasmReqPkts;
        stats_p->ipReasmEnqueuedPkts               += dp_stats->UipReasmEnqueuedPkts;
        stats_p->ipOutNetDOrigPkts                 += dp_stats->UipOutNetDOrigPkts;
        stats_p->ipOutEOOrigPkts                   += dp_stats->UipOutEOOrigPkts;
        stats_p->ipInSentToHostPkts                += dp_stats->UipInSentToHostPkts;
        stats_p->ipInSentToEoPkts                  += dp_stats->UipInSentToEoPkts;
        stats_p->ipInSentToIPSecPkts               += dp_stats->UipInSentToIPSecPkts;
        stats_p->ipOutSentToIPSecPkts              += dp_stats->UipOutSentToIPSecPkts;
        stats_p->ipOutRcvdFromIPSecPkts            += dp_stats->UipOutRcvdFromIPSecPkts;
        stats_p->ipOutSentToEnetPkts               += dp_stats->UipOutSentToEnetPkts;
        stats_p->ipReasmTimerCbAfterCleanup        += dp_stats->UipReasmTimerCbAfterCleanup;
        stats_p->ipReasmInvalidReasmAction         += dp_stats->UipReasmInvalidReasmAction;
        stats_p->ipOutNonLocalSipFromHostPkts      += dp_stats->UipOutNonLocalSipFromHostPkts;
        stats_p->ipInUDPOctets                     += dp_stats->UipInUDPOctets;
        stats_p->ipInUDPPkts                       += dp_stats->UipInUDPPkts;
        stats_p->ipOutUDPOctets                    += dp_stats->UipOutUDPOctets;
        stats_p->ipOutUDPPkts                      += dp_stats->UipOutUDPPkts;
        stats_p->ipIpsecFragsCreated               += dp_stats->UipIpsecFragsCreated;
        stats_p->ipIpsecFragsRequested             += dp_stats->UipIpsecFragsRequested;
        stats_p->ipInUnknownSender                 += dp_stats->UipInUnknownSender;
        stats_p->ipOutUnknownSender                += dp_stats->UipOutUnknownSender;

    }

    return ADK_SUCCESS;
} /* adk_ip_dp_utility_stats_get */


adk_dp_status_t
adk_ip_dp_utility_stats_clear(adk_module_handle_t ipHandle)
{
    uint32_t i;
    adk_ip_dp_global_stats_t *dp_stats;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ip_module_info->ip_dp_tbl_info_p->ipGlobalStatsTable[i]);

        dp_stats->UipReasmReqPkts					= 0;
        dp_stats->UipReasmEnqueuedPkts					= 0;
        dp_stats->UipOutNetDOrigPkts					= 0;
        dp_stats->UipOutEOOrigPkts					= 0;
        dp_stats->UipInSentToHostPkts					= 0;
        dp_stats->UipInSentToEoPkts					= 0;
        dp_stats->UipInSentToIPSecPkts					= 0;
        dp_stats->UipOutSentToIPSecPkts					= 0;
        dp_stats->UipOutRcvdFromIPSecPkts					= 0;
        dp_stats->UipOutSentToEnetPkts					= 0;
        dp_stats->UipReasmTimerCbAfterCleanup					= 0;
        dp_stats->UipReasmInvalidReasmAction					= 0;
        dp_stats->UipOutNonLocalSipFromHostPkts					= 0;
        dp_stats->UipInUDPOctets					= 0;
        dp_stats->UipInUDPPkts					= 0;
        dp_stats->UipOutUDPOctets					= 0;
        dp_stats->UipOutUDPPkts					= 0;
        dp_stats->UipIpsecFragsCreated					= 0;
        dp_stats->UipIpsecFragsRequested					= 0;
        dp_stats->UipInUnknownSender					= 0;
        dp_stats->UipOutUnknownSender					= 0;

    }

    return ADK_SUCCESS;
} /* adk_ip_dp_utility_stats_clear */



adk_status_t
adk_ip_dp_ingress_classifier_stats_get(adk_module_handle_t ipHandle,
    uint32_t classFlowId,
        adk_ip_ingress_classifier_stats_t *stats_p)
{
    uint16_t i;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);

    stats_p->rxPkts = 0;
    stats_p->txPkts = 0;
    stats_p->ipPolicerDroppedPkts=0;

    for(i=0; i<ADK_MAX_NUM_CORES; ++i)
    {
        stats_p->rxPkts +=
                ip_module_info->ip_dp_tbl_info_p->ipIngFlowStats[i*ADK_DP_MAX_IP_ING_ASSOC_FLOWS+classFlowId].rxPkts;

        stats_p->txPkts +=
                ip_module_info->ip_dp_tbl_info_p->ipIngFlowStats[i*ADK_DP_MAX_IP_ING_ASSOC_FLOWS+classFlowId].txPkts;
    }

    /* Now collect the IP stats from CPPI BD (slowpath) */
    stats_p->txPkts += ip_module_info->ip_dp_tbl_info_p->ipCppiStats[classFlowId].cppiTxPkts;
    stats_p->ipPolicerDroppedPkts += ip_module_info->ip_dp_tbl_info_p->ipCppiStats[classFlowId].cppiPolicerDropPkts;

    return ADK_SUCCESS;
} /* adk_ip_dp_ingress_classifier_stats_get */


adk_status_t
adk_ip_dp_ingress_classifier_stats_clear(adk_module_handle_t ipHandle)
{
    uint16_t i,j;
    ip_module_info_t *ip_module_info = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipHandle);
    adk_ip_dp_ing_flow_stats_t *tmp;

    for(j=0; j<ADK_DP_MAX_IP_ING_ASSOC_FLOWS; ++j)
    {
        ip_module_info->ip_dp_tbl_info_p->ipCppiStats[j].cppiPolicerDropPkts = 0;
        ip_module_info->ip_dp_tbl_info_p->ipCppiStats[j].cppiTxPkts = 0;
        tmp = &(ip_module_info->ip_dp_tbl_info_p->ipIngFlowStats[j*ADK_MAX_NUM_CORES]);
        for(i=0; i<ADK_MAX_NUM_CORES; ++i)
        {
            tmp[i].rxPkts = 0;
            tmp[i].txPkts = 0;
        }
    }

    return ADK_SUCCESS;
} /* adk_ip_dp_ingress_classifier_stats_clear */
