/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_stats.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs for getting or printing
 * statistics for the interfaces, module level global stats.
 *
 ***********************************************************************/

#include "adk_ip_common.h"
#include "adk_ip_utility.h"

adk_status_t adk_ip_utility_stats_enable(
    adk_module_handle_t ipHandle,
    bool state
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_utility_stats_enable_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    ipModuleInfo_p->ip_dp_tbl_info_p->utilCntrEnabled = state;

    if(true == state)
        ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_utility_stats_enable: utility counters enabled");
    else
        ADK_IPV4_LOG_INFO(ipHandle, "adk_ip_utility_stats_enable: utility counters disabled");

_adk_ip_utility_stats_enable_exit:

        ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
        ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

        ADK_RETURN(retValue);

}

adk_status_t adk_ip_utility_stats_get(
    adk_module_handle_t ipHandle,
    adk_ip_utility_stats_t *stats_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;
    adk_dp_status_t dp_st;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_utility_stats_get_exit;
    }

    if (!stats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Invalid input pointer parameter\n");
        goto _adk_ip_utility_stats_get_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    dp_st = adk_ip_dp_utility_stats_get(ipHandle, stats_p);
    if (dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Error in reading IPv4 utility stats from dataplane\n");
        goto _adk_ip_utility_stats_get_exit;
    }


_adk_ip_utility_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

} /* adk_ip_utility_stats_get */


adk_status_t adk_ip_stats_clear(
        adk_module_handle_t ipHandle,
        adk_ip_stats_type_t stats_type
)
{
    adk_status_t retValue = ADK_SUCCESS, errRetVal = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ip_stats_clear_exit;
    }


    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if(stats_type & ADK_IP_GLOBAL_STATS)
        retValue = adk_ip_dp_global_stats_clear(ipHandle);

    if(ADK_SUCCESS != retValue)
        errRetVal = retValue;

    if(stats_type & ADK_IP_IFACE_STATS)
    {
        if(ADK_SUCCESS != (retValue = adk_ip_dp_interface_stats_clear(ipHandle)))
                ADK_IPV4_LOG_INFO(ipHandle, "Failed to clear all ipv4 interface stats\n");

    }
    if(ADK_SUCCESS != retValue)
        errRetVal = retValue;

    if(stats_type & ADK_IP_ADDR_STATS)
    {
        if(ADK_SUCCESS != (retValue = adk_ip_dp_ip_addr_stats_clear(ipHandle)))
            ADK_IPV4_LOG_INFO(ipHandle, "Failed to clear all ipv4 address stats\n");
    }

    if(ADK_SUCCESS != retValue)
        errRetVal = retValue;

    if(stats_type & ADK_IP_L4_STATS)
    {
        if(ADK_SUCCESS != (retValue = adk_ip_dp_l4_stats_clear(ipHandle)))
            ADK_IPV4_LOG_INFO(ipHandle, "Failed to clear ipv4 L4 stats\n");
    }

    if(ADK_SUCCESS != retValue)
        errRetVal = retValue;

    if(stats_type & ADK_IP_UTILITY_STATS)
    {
        if(ADK_SUCCESS != (retValue = adk_ip_dp_utility_stats_clear(ipHandle)))
            ADK_IPV4_LOG_INFO(ipHandle, "Failed to clear all ipv4 utility stats\n");
    }

    if(ADK_SUCCESS != retValue)
        errRetVal = retValue;

    if(stats_type & ADK_IP_ING_CLASSIFIER_STATS)
    {
        if(ADK_SUCCESS != (retValue = adk_ip_dp_ingress_classifier_stats_clear(ipHandle)))
            ADK_IPV4_LOG_INFO(ipHandle, "Failed to clear all ipv4 ingress classifier stats\n");
    }

    /* If any of the internal APIs gave error, we will return the last error. */
    retValue = errRetVal;

    _adk_ip_stats_clear_exit:
    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

    _lock_release_failed:
    _lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

} /* adk_ip_clear_stats */
