/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_virtual_interface.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs to manage IPv4 interfaces. This includes creation,
 * deletion of virtual interfaces, setting their attributes etc.
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include <adk_ethernet_export.h>

/**
 *=============================================================
 * adk_ip_virtual_interface_create
 *
 * @param[in] ipHandle      IP module handle.
 *
 * @param[in] ipVIf_p       Pointer to adk_ip_virtual_interface_t data
 * structure.
 *
 * @param[out] ipVIfHandle  Pointer to the virtual IP interface handle
 * which gets created by this API.
 *
 * @retval ADK_SUCCESS                  Successfully created ADK IP interface
 *
 * @retval ADK_E_INVALID_HANDLE         Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER      Invalid input parameter e.g. interface
 * attribute pointer might be NULL.
 *
 * @retval ADK_E_MEM_UNAVAILABLE        Failed to allocate memory
 * for interface handle.
 *
 * @retval ADK_IP_E_INVALID_IP_ADDRESS  Multicast/Broadcast address specified
 * for the the interface.
 *
 * @retval ADK_IP_E_INVALID_MTU         Invalid MTU configruation. Supported
 * range is between 64B and 8192B.
 *
 * @brief
 *
 * This function create a virtual IP interface handle to be used for L3
 * termination. The virtual interface has no underlying L2 physical or any
 * other type of interface. It is purely logical interface and has no subnet
 * associated with it.
 * =============================================================
 */
adk_status_t adk_ip_virtual_interface_create(
    adk_module_handle_t ipHandle,
    const adk_ip_virtual_interface_t  *ipVIf_p,
    adk_interface_handle_t *ipVIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p = NULL;
    ip_virtual_interface_info_t *vif_pvt_info = NULL;
    adk_sys_handle_t sys_handle;
    uint32_t vrdId = ADK_IP_DEFAULT_VRD_ID, tblId=0;
    adk_ip_dp_route_info_t  *dipResult_p;
    adk_dp_status_t dp_st;
    uint32_t                stats_id, ip_addr_stats_id;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_virtual_interface_create_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipVIf_p || !ipVIfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "NULL pointer parameter specified\n");
        goto _adk_ip_virtual_interface_create_exit;
    }

    *ipVIfHandle = 0;

#ifdef ENABLE_VRD
    /* Validate VRD handle */
    if (!adk_ipv4_vrd_valid_handle(ipVIf_p->ipVrdHandle, &vrdId))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_INTERFACE);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid VIrtual Routing Domain handle \n");
        goto _adk_ip_virtual_interface_create_exit;
    }
#else
	vrdId = ADK_IP_DEFAULT_VRD_ID;
#endif /* ENABLE VRD */

    /* Validate the address */
    if (ADK_IPV4_IS_IP_IF_ADDR_INVALID(ipVIf_p->ipAddress))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_INVALID_IP_ADDRESS);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid interface address specified\n");
        goto _adk_ip_virtual_interface_create_exit;
    }

    if(ipModuleInfo_p->aggrIpv4AddrCount >= ADK_IP_MAX_ADDRESSES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_MAX_IP_ADDR_EXCEEDED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Max allowed ip addresses (%d) already configured in the system\n", ADK_IP_MAX_ADDRESSES);
        goto _adk_ip_virtual_interface_create_exit;
    }

    /* Note: We use 'virtual interface' concept as a means to configure virtual addresses.
     * We decided to re_purpose std-adk virtual interface implementation like this.
     */
    if(ipModuleInfo_p->aggrIpv4VirtAddrCount >= ADK_IP_MAX_VIRTUAL_IP_ADDRESSES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IP_E_MAX_IP_ADDR_EXCEEDED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue, "Max allowed virtual ip addresses (%d) already configured in the system\n", ADK_IP_MAX_VIRTUAL_IP_ADDRESSES);
        goto _adk_ip_virtual_interface_create_exit;
    }

    ADK_SYS_INTERFACE_HANDLE_ALLOC(ipHandle, ADK_IP_VIRT_INTF_OBJ,
                                   sizeof(ip_virtual_interface_info_t),
                                   ipVIfHandle);
    if (*ipVIfHandle == 0) {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                           "Memory allocation failed\n");
        goto _adk_ip_virtual_interface_create_exit;
    }

    vif_pvt_info = (ip_virtual_interface_info_t *)
                   ADK_SYS_INTERFACE_ATTR_PTR_GET(*ipVIfHandle);

    ADK_MEMCPY(&vif_pvt_info->vif_info, ipVIf_p,
               sizeof(adk_ip_virtual_interface_t));

    vif_pvt_info->cookie = ADK_IPV4_V_INTF_COOKIE;

#if 0
    /* Reserve a statistics flow ID for the interface */
    retValue = ADK_SYS_INTERFACE_ID_ALLOC(sys_handle, &vif_pvt_info->stats_id);
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(retValue);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to allocate stats NS ID for VIP intf\n");
        goto _adk_ip_virtual_interface_create_exit;
    }

    ADK_IPV4_LOG_INFO(ipHandle, "Allocated interface NS ID 0x%x\n",
                      vif_pvt_info->stats_id);
#endif
    
    vif_pvt_info->vrd_id = vrdId;

    stats_id = adk_rm_resource_reserve(ipModuleInfo_p->ipStatsHandle);
    if(stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to allocate IP IF Stats ID");
        goto _adk_ip_virtual_interface_create_exit;
    }

    ip_addr_stats_id = adk_rm_resource_reserve(ipModuleInfo_p->ipAddrStatsHandle);
    if(ip_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                "Failed to allocate IP Address Stats ID");
        goto _adk_ip_virtual_interface_create_exit;
    }

    for(tblId=0; tblId < ADK_IP_MAX_ROUTING_TABLES; tblId++)
    {
        /* Add route for virtual IP address of virtual IP interface. The advanced
         * lookup is always enabled. Re-assembly can also be optionally enabled */
        dipResult_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_VIRT_INTF_OBJ, 1, sizeof(adk_ip_dp_route_info_t));
        if (dipResult_p == NULL)
        {
            ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                    "Memory allocation failed\n");
            retValue = ADK_E_MEM_UNAVAILABLE;
            goto _adk_ip_virtual_interface_create_exit;
        }
        dipResult_p->isTerminating = TRUE;
        dipResult_p->virtualIf = TRUE;
        dipResult_p->virtual_if_stats_id = stats_id;
        /* ip addr stats id will be set only for terminating addr */
        dipResult_p->u.termination.ip_addr_stats_id = ip_addr_stats_id;

        dp_st = adk_ip_dp_dip_add(ipHandle, tblId, adk_htonl(ipVIf_p->ipAddress),
                32, ADK_CONVERT_TO_LPM_DATA(dipResult_p));
        if(dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in adding dest ip to dataplane\n");
            goto _adk_ip_virtual_interface_create_exit;
        }
    }
    /*Update CP version id*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId])++;

    adk_hos_ip_term_tbl_rule_attr_t hos_ip_term_rule =
    {
        .precedence = -1,
        .pkt_type = ADK_HOS_L2_TERMINATING,
        .is_ipv6 = false,
        .ip_addr_mask[0] = adk_htonl(0xFFFFFFFF),
        .ip_addr[0] = adk_htonl(ipVIf_p->ipAddress),
        .is_ip_term = true,
        .is_bcast = false,
        .is_mcast = false,
        .ip_addr_stats_id = ADK_HOS_DEFAULT_IP_STATS_ID,
    };

    retValue = adk_hos_rule_add(ADK_HOS_IP_TERM_TBL,
                          (adk_hos_rule_attr_t *) &hos_ip_term_rule,
                          &(vif_pvt_info->ip_term_rule_id[0]));
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
          "Failed to add IP-Term HOS rule for virtual ip-intf address\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        goto _adk_ip_virtual_interface_create_exit;
    }
#ifdef ADK_HOS_OLD_IMPL
    retValue = adk_hos_rule_add(ADK_HOS_INNER_IP4_TERM_TBL,
                          (adk_hos_rule_attr_t *) &hos_ip_term_rule,
                          &(vif_pvt_info->ip_term_rule_id[1]));
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
          "Failed to add Inner IP-Term HOS rule for virtual ip-intf address\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        goto _adk_ip_virtual_interface_create_exit;
    }
#endif

    vif_pvt_info->stats_id = stats_id;
    vif_pvt_info->ip_addr_stats_id = ip_addr_stats_id;

    ++ipModuleInfo_p->aggrIpv4VirtAddrCount;
    ++ipModuleInfo_p->aggrIpv4AddrCount;

_adk_ip_virtual_interface_create_exit:

    if (retValue != ADK_SUCCESS) {
        if (ipVIfHandle != NULL)
            if (*ipVIfHandle != 0)
                ADK_SYS_INTERFACE_HANDLE_FREE(*ipVIfHandle);
    }

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    
    ADK_RETURN(retValue);

}

/**
 *=============================================================
 * adk_ip_virtual_interface_delete
 * @param[in]  ipHandle - IP module handle.
 *
 * @param[out] ipVIfHandle - Virtual IP interface handle to be deleted
 *
 * @retval ADK_SUCCESS Successfully created ADK IP interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This function destroys a virtual IP interface handle.
 * =============================================================
 */
adk_status_t adk_ip_virtual_interface_delete(
    adk_module_handle_t   ipHandle,
    adk_interface_handle_t   ipVIfHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p = NULL;
    ip_virtual_interface_info_t *vif_pvt_info = NULL;
    adk_sys_handle_t sys_handle;
    uint8_t vrdId = ADK_IP_DEFAULT_VRD_ID, tblId;
    adk_dp_status_t dp_st;
    adk_ip_dp_intf_stats_t *dp_stats;
    adk_ip_dp_ip_addr_stats_t *dp_addr_stats;
    uint16_t ip_addr_stats_id;
    uint32_t i=0;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);
    
    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_virtual_interface_delete_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipVIfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "NULL pointer parameter specified\n");
        goto _adk_ip_virtual_interface_delete_exit;
    }

    if (!adk_ipv4_vif_valid_handle(ipVIfHandle))
    {
        ADK_IPV4_LOG_ERROR(ipHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ip_virtual_interface_delete_exit;
    }

    vif_pvt_info = (ip_virtual_interface_info_t*)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipVIfHandle);

    vrdId = vif_pvt_info->vrd_id;

    for(tblId=0; tblId < ADK_IP_MAX_ROUTING_TABLES; tblId++)
    {
        dp_st = adk_ip_dp_dip_remove(ipHandle, tblId, adk_htonl(vif_pvt_info->vif_info.ipAddress), 32);
        if(dp_st != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                    "Error in removing dest ip to dataplane\n");
            goto _adk_ip_virtual_interface_delete_exit;
        }
    }

    /*Update CP version id*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrdId])++;
    retValue = adk_hos_rule_delete(ADK_HOS_IP_TERM_TBL,
                                   vif_pvt_info->ip_term_rule_id[0]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
             "Failed to remove IP-Term HOS rule for virtual ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_virtual_interface_delete_exit;
    }
#ifdef ADK_HOS_OLD_IMPL 
    retValue = adk_hos_rule_delete(ADK_HOS_INNER_IP_TERM_TBL,
                                   vif_pvt_info->ip_term_rule_id[1]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
        "Failed to remove Inner IP-Term HOS rule for virtual ip-intf address\n");
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ip_virtual_interface_delete_exit;
    }
#endif

    /*Now clean up stats*/
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ipModuleInfo_p->ip_dp_tbl_info_p->
                     ipInterfaceStatsTble[i][vif_pvt_info->stats_id]);
        ADK_MEMSET(dp_stats, 0, sizeof(adk_ip_dp_intf_stats_t));
    }
    adk_rm_resource_free(ipModuleInfo_p->ipStatsHandle, vif_pvt_info->stats_id);
    vif_pvt_info->stats_id = 0;

    ip_addr_stats_id = vif_pvt_info->ip_addr_stats_id;
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_addr_stats = &(ipModuleInfo_p->ip_dp_tbl_info_p->ipAddrStatsTble[i][ip_addr_stats_id]);
        ADK_MEMSET(dp_addr_stats, 0, sizeof(adk_ip_dp_ip_addr_stats_t));
    }
    adk_rm_resource_free(ipModuleInfo_p->ipAddrStatsHandle, vif_pvt_info->ip_addr_stats_id);
    vif_pvt_info->ip_addr_stats_id = 0;
                
    ADK_SYS_INTERFACE_HANDLE_FREE(ipVIfHandle);

    
#if 0
    /* Add route for the interface address itself. Here the advanced lookup is
     * always enabled. Re-assembly can also be optionally enabled */

    dipResult_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_VIRT_INTF_OBJ, 1, sizeof(adk_ip_dp_route_info_t));
    if (dipResult_p == NULL)
    {
        ADK_IPV4_LOG_FATAL(ipHandle, ADK_E_MEM_UNAVAILABLE,
                          "Memory allocation failed\n");
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ip_virtual_interface_delete_exit;
    }
#endif

    --ipModuleInfo_p->aggrIpv4VirtAddrCount;
    --ipModuleInfo_p->aggrIpv4AddrCount;


_adk_ip_virtual_interface_delete_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);
    
    ADK_RETURN(retValue);

}
/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
