/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ip_vrd.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ip  IPv4 Module
 * @{
 *      @addtogroup _ipFunctions  IPv4 Functions
 *      @{
 *
 * @brief
 *
 * This file contains APIs for managing VRD.
 * This includes creation, deletion of VRD, etc
 *
 ***********************************************************************/

/* IP internal header files */
#include "adk_ip_common.h"
#include "adk_ph_dev_dp.h"

#ifdef ENABLE_VRD

/*!
 * =============================================================
 * adk_ip_vrd_create
 *
 * @param[in] ipHandle          IP module handle.
 *
 * @param[in] vrdAttr           Pointer to data structure having
 * attributes of the virtual routing domain.
 *
 * @param[out] ipVrdHandle_p    Pointer to VRF handle created
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          One of the direct arguments
 * to the API call is invalid e.g. NULL pointer paramter
 *
 * @retval ADK_E_MEM_UNAVAILABLE            Failed to allocate memory for
 * virtual routing domain object
 *
 * @retval ADK_E_NS_RSRC_ALLOCATE_FAILED    Failed to allocate NS ID for
 * virtual routing domain object
 *
 * @brief
 *
 * This API creates a virtual routing domain based on the application
 * supplied domain attributes. The created routing domain handle is
 * returned to the application. Application has to use this domain handle while
 * associating a IPv4 object (like IPoE Interface, Route Entry, Ingress
 * association etc..) with this routing domain.
 * =============================================================
 */
adk_status_t adk_ip_vrd_create(
    adk_module_handle_t ipHandle,
    const adk_ip_vrd_t  *vrdAttr_p,
    adk_ip_vrd_handle_t *ipVrdHandle_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_vrd_info_t           *vrd_info_p = NULL;
    uint32_t vrd_id = 0;
    ip_module_info_t        *ipModuleInfo_p = NULL;
    ncp_st_t st = NCP_ST_SUCCESS;
    uint32_t ip_addr;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_vrd_create_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!ipVrdHandle_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid NULL pointer parameter\n");
        goto _adk_ip_vrd_create_exit;
    }

    if (!vrdAttr_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid NULL pointer parameter\n");
        goto _adk_ip_vrd_create_exit;
    }

    if (!adk_rm_resource_reserve_fixed(ipModuleInfo_p->ipVrdIDHandle,
                                      vrdAttr_p->vrd_id) )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to allocate vrd namespace ID,"
                           " RTE Error: %s\n", ncp_status_get(st));
        goto _adk_ip_vrd_create_exit;
    }

    vrd_id = vrdAttr_p->vrd_id;

    ADK_IPV4_LOG_INFO(ipHandle, "Allocated VRD ID 0x%x\n", vrd_id);

    vrd_info_p = (ip_vrd_info_t *)
                 ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipHandle), ADK_IP_VRD_OBJ, 1, sizeof(ip_vrd_info_t));
    if (vrd_info_p == NULL)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Memory allocation failed\n");
        goto _adk_ip_vrd_create_exit;
    }

    /* Add source IP address validation entries for this VRD */
    
    /* Default SIP filter entries
     * SIP 127.*.*.* -> DENY (loopback)
     * SIP 255.255.255.255 -> DENY (link-local broadcast)
     * SIP 224-239.*.*.* -> DENY (multicast)
     */

    ip_addr = adk_htonl(0x7F000000);
    if(ADK_SUCCESS != adk_ip_dp_sip_add(ipHandle, vrd_id, ip_addr, 8,
                ADK_CONVERT_TO_LPM_DATA(ADK_IP_SIP_ADDR_ACTION_BLOCK_ALL)))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_vrd_create_exit;
    }
    ip_addr = adk_htonl(0xFFFFFFFF);
    if(ADK_SUCCESS != adk_ip_dp_sip_add(ipHandle, vrd_id, ip_addr, 32,
                         ADK_CONVERT_TO_LPM_DATA(ADK_IP_SIP_ADDR_ACTION_BLOCK_ALL)))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_vrd_create_exit;
    }
    ip_addr = adk_htonl(0xE0000000);
    if(ADK_SUCCESS != adk_ip_dp_sip_add(ipHandle, vrd_id, ip_addr, 4,
                         ADK_CONVERT_TO_LPM_DATA(ADK_IP_SIP_ADDR_ACTION_BLOCK_ALL)))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_vrd_create_exit;
    }

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrd_id])++;
    /* Store the VRD information in the internal data structure */
    vrd_info_p->cookie = ADK_IPV4_VRD_COOKIE;
    vrd_info_p->vrd_id = (uint32_t) vrd_id;

    *ipVrdHandle_p = (adk_ip_vrd_handle_t) vrd_info_p;
    ipModuleInfo_p->vrdHandles[vrd_id] = *ipVrdHandle_p;

_adk_ip_vrd_create_exit:

    if (retValue != ADK_SUCCESS) {
        if (vrd_id && ipModuleInfo_p)
            adk_rm_resource_free(ipModuleInfo_p->ipVrdIDHandle, vrd_id);
        if (vrd_info_p != NULL)
            ADK_NVM_FREE(vrd_info_p);
    }

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

}    /* -----  end of function adk_ip_vrd_create ----- */

/*!
 * =============================================================
 * adk_ip_vrd_delete
 *
 * @param[in] ipHandle      IP module handle.
 *
 * @param[in] ipVrdHandle   VRD handle to be deleted.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          One of the direct arguments
 * to the API call is invalid e.g. invalid VRF handle
 *
 * @retval ADK_E_NS_RSRC_FREE_FAILED        Failed to free NS resource
 * ID for virtual routing domain object
 *
 * @brief
 *
 * This API deletes a routing domain instance and frees up the resources. The
 * VRF domain must be configured in routing mode before deletion.
 * =============================================================
 */
adk_status_t adk_ip_vrd_delete(
    adk_module_handle_t ipHandle,
    adk_ip_vrd_handle_t ipVrdHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_vrd_info_t           *vrd_info_p;
    uint32_t vrd_id;
    ip_module_info_t        *ipModuleInfo_p;
    uint32_t ip_addr;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_vrd_delete_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    if (!adk_ipv4_vrd_valid_handle(ipVrdHandle, &vrd_id) || !ipVrdHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid VRD Handle\n");
        goto _adk_ip_vrd_delete_exit;
    }

    vrd_info_p = (ip_vrd_info_t *) ipVrdHandle;

    adk_rm_resource_free(ipModuleInfo_p->ipVrdIDHandle, vrd_id);

    ADK_IPV4_LOG_INFO(ipHandle, "Freed VRD ID 0x%x\n", vrd_id);

    /* Remove source IP address validation entries for this VRD */
    /* Default SIP filter entries
     * SIP 127.*.*.* -> DENY (loopback)
     * SIP 255.255.255.255 -> DENY (link-local broadcast)
     * SIP 224-239.*.*.* -> DENY (multicast)
     */

    ip_addr = adk_htonl(0x7F000000);
    if(ADK_SUCCESS != adk_ip_dp_sip_remove(ipHandle, vrd_id, ip_addr, 8))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_vrd_delete_exit;
    }
    ip_addr = adk_htonl(0xFFFFFFFF);
    if(ADK_SUCCESS != adk_ip_dp_sip_remove(ipHandle, vrd_id, ip_addr, 32))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_vrd_delete_exit;
    }
    ip_addr = adk_htonl(0xE0000000);
    if(ADK_SUCCESS != adk_ip_dp_sip_remove(ipHandle, vrd_id, ip_addr, 4))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_IPV4_LOG_FATAL(ipHandle, retValue,
                           "Could not build the input pattern\n");
        goto _adk_ip_vrd_delete_exit;
    }

    /*Update CP version id the specific VRD*/
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrd_id])++;
    ipModuleInfo_p->vrdHandles[vrd_id] = 0;
    vrd_info_p->cookie = 0;
    ADK_NVM_FREE(vrd_info_p);

_adk_ip_vrd_delete_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);

}    /* -----  end of function adk_ip_vrd_delete ----- */


/*!
 * =============================================================
 * adk_ip_vrd_mode
 *
 * @param[in] ipHandle      IP module handle
 *
 * @param[in] ipVrdHandle   VRD handle to configure in Host or Router mode. If
 * passed as 0, the configuration applies to the default VRD.
 *
 * @param[in] host_most     Flag to enable host mode for the VRF. If set to
 * FALSE, the VRF is configured in Router mode.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv4 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          One of the direct arguments
 * to the API call is invalid e.g. invalid VRF handle
 *
 * @brief
 *
 * This API configures a VRF in router or host mode. The VRF must be configured
 * in Router mode before VRF is deleted.
 * =============================================================
 */
adk_status_t adk_ip_vrd_mode(
    adk_module_handle_t ipHandle,
    adk_ip_vrd_handle_t ipVrdHandle,
    bool host_mode)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t vrd_id = ADK_IP_DEFAULT_VRD_ID;
    adk_sys_handle_t sys_handle;
    ip_module_info_t *ipModuleInfo_p;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_vrd_mode_exit;
    }

    if (!adk_ipv4_vrd_valid_handle(ipVrdHandle, &vrd_id))
    {
    	retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    	ADK_IPV4_LOG_ERROR(ipHandle, retValue,
    			"Invalid VRD Handle type\n");
    	goto _adk_ip_vrd_mode_exit;
    }

    if(ADK_SUCCESS != adk_ip_dp_vrd_host_mode(ipHandle, vrd_id, host_mode))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Failed to update DP table\n");
        goto _adk_ip_vrd_mode_exit;
    }
    
    /*Update CP version id the specific VRD*/
    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);
    (ipModuleInfo_p->ip_dp_tbl_info_p->ip_cp_version_id[vrd_id])++;

_adk_ip_vrd_mode_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ip_vrd_handle_get
 *
 * @param[in] ipHandle      IP module handle.
 *
 * @param[in] vrd_id        VRD ID for which the VRD handle is to be retrieved.
 *
 * @param[in] ipVrdHandle_p Pointer to the VRD handle to be returned.
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv4 module handle
 *
 * @retval ADK_E_INTERNAL_ERROR             Non-existent VRD handle
 *
 * @retval ADK_E_INVALID_PARAMETER          One of the direct arguments
 * to the API call is invalid e.g. null pointer or invalid VRD ID.
 *
 * @brief
 *
 * This query API retrieves the VRD handle for a given VRD ID. Note
 * that, for default VR domain, this API returns success with VRD 
 * handle as 0x0.
 * =============================================================
 */
adk_status_t adk_ip_vrd_handle_get(
    adk_module_handle_t ipHandle,
    uint32_t vrd_id,
    adk_ip_vrd_handle_t *ipVrdHandle_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    ip_module_info_t *ipModuleInfo_p;
    adk_sys_handle_t sys_handle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipHandle))
    {
        ADK_IPV4_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipHandle);

    ADK_IPV4_LOG_FUNC_ENTRY(ipHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipHandle, retValue);

    /* Handle validations */
    if (!adk_ipv4_valid_handle(ipHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV4_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipHandle);
        goto _adk_ip_vrd_handle_get_exit;
    }

    if (!ipVrdHandle_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "Invalid NULL pointer parameter\n");
        goto _adk_ip_vrd_handle_get_exit;
    }

    if (vrd_id != ADK_IP_DEFAULT_VRD_ID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "VRD ID %d out of range\n", vrd_id);
        goto _adk_ip_vrd_handle_get_exit;
    }

    ipModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipHandle);

    /**ipVrdHandle_p = ipModuleInfo_p->vrdHandles[vrd_id];*/
    *ipVrdHandle_p = ipModuleInfo_p->vrdHandles[vrd_id];

    /* Return error if VRD handle does not exist */
    if ((*ipVrdHandle_p == 0) && (vrd_id != ADK_IP_DEFAULT_VRD_ID))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV4_LOG_ERROR(ipHandle, retValue,
                           "VRD handle for VRD ID %d does not exist\n", vrd_id);
        goto _adk_ip_vrd_handle_get_exit;
    }

_adk_ip_vrd_handle_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV4_LOG_FUNC_EXIT(ipHandle);

    ADK_RETURN(retValue);
}

#endif /* ENABLE_VRD */

/*!     end addtogroup _ipFunctions
 *      @}
 *  end addtogroup _ip
 * @}
 */
