/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

#include <math.h>
#include "adk_sys_lcore.h"
#include <adk_ip_common.h>
#ifdef IP_FLOW_HASH_CACHE
#ifdef ADK_MODULE_IPSEC_ENABLED
#include <adk_ipsec_export.h>
#endif
#endif
#include <adk_ethernet_export.h>
#ifdef ADK_MODULE_TWAMP_ENABLED
#include <adk_twamp_dp_export.h>
#endif
#ifdef ADK_MODULE_BFD_ENABLED
#include <adk_bfd_dp_export.h>
#endif
#include <adk_hos_dp.h>
#include <adk_qos_dp_export.h>
#include <adk_cksum.h>
#include <adk_timer.h>
#include <adk_pmu.h>
#include <adk_ip_dp_internal.h>
#include <adk_ip_dp_internal_fc.h>

#include "adk_ph_dev_dp.h"

/*!
 * =============================================================
 * adk_ip_dp_pkt_proc_ring_lli()
 *
 * Structure of task params :
 *
 * taskParams[0] : Upper 8-bits indicate EIOA port number.
 * taskParams[1] : Interface (or Sub interface) handle.
 * taskParams[2] : IPsec data if received on tunnel interface
 *                 (from IPsec after decryption).
 * taskParams[3] : PDU buffer pointing to L2 header.
 * taskParams[7] : PDU buffer pointing to IP header.
 *
 * Brief:
 * This function is an interface to ethernet module and performs
 * following:
 *
 * a. IP header checksum validation.
 * b. Route Lookup.
 * c. ACL lookup.
 * d. Interface-level policing.
 * e. Ingress association/Advance Lookup.
 * f. Reassembly of fragmented packets.
 *
 * This function binds to following modules depending on packet processing:
 * 1. Binds to IPsec LLI - adk_ipsec_dp_pkt_proc_lli for decryption of ESP
 * packet received from TN port.
 * 2. Binds to GTPu LLI - adk_gtpu_dp_pkt_proc_lli for termination of inner
 * packet received on IPsec tunnel.
 * 3. Binds to Ethernet ULI - adk_eth_dp_pkt_proc_uli for normal IPv4
 * forwarding (TN->WIFI, WIFI->TN).
 *
 * =============================================================
 */
adk_status_t adk_ip_dp_pkt_proc_ring_lli (void **task_p,
                        uint8_t __attribute__((unused)) task_count,
                        uint32_t task_mask,
                        void *ring,
                        adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                        __attribute__((unused)) int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t **task = (adk_pktio_task_t **)task_p;
    uint16_t nb_tx;
    uint32_t i, num = 0;
    uint32_t cur_bit;
    struct rte_mbuf *tx_task[ADK_DP_BULK_RX_SIZE * ADK_DP_TASK_MULTIPLY_MAX] = {0};

    /*Tasks may not be packed since some tasks are dropped/consumed/fragmented. Pack the tasks*/
    for(i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if(task_mask & cur_bit)
        {
            tx_task[num] = (struct rte_mbuf *)(task[i]);
            ADK_TASK_USER_PARAMS(tx_task[num], 8) = 
                ADK_CONVERT_TO_TPARAMS(next_stage_entry);
            ADK_TASK_USER_PARAMS(tx_task[num], 9) =
                ADK_CONVERT_TO_TPARAMS(ADK_IPV4_LLI_FN_ID);
            num++;
        }
    }

    nb_tx = rte_ring_sp_enqueue_burst((struct rte_ring *)ring, (void *)tx_task, num, NULL);
    if (unlikely(nb_tx != num))
    {
        for(i = nb_tx; i < num; ++i)
            rte_pktmbuf_free(tx_task[i]);
    }
   
    return ret;
}

#ifdef UDP_CHECKSUM_STAGE_AFTER_IP
adk_status_t adk_udp_dp_pkt_proc_lli (void **task_p,
                        uint8_t task_count,
                        uint32_t task_mask,
                        __attribute__((unused)) void *ring,
                        __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                        int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    uint32_t i = 0, cur_bit = 0, destModId = 0;
    adk_pktio_task_t *task = NULL;
    uint8_t *pdu = NULL;
    uint16_t udp_length;
    uint16_t *dst_mod_dst_func = NULL; 
    adk_sif_cfg_pipeline_info_t *pipeline_info = NULL;
    adk_sif_cfg_pipeline_exit_module_t *pipeline_src_info = NULL;
    adk_sif_cfg_pipeline_dest_module_t *pipeline_dest_info = NULL;
    
    /* Verify UDP checksum */
    for(i=0, cur_bit=1; i<ADK_DP_BULK_RX_SIZE; ++i,(cur_bit=cur_bit<<1))
    {
        if(!(task_mask & cur_bit))
            continue;

        task = (adk_pktio_task_t *)(task_p[i]);
        pdu = ADK_CONVERT_TPARAMS_TO_UINT8PTR(ADK_TASK_USER_PARAMS(task, 7));
        dst_mod_dst_func =
            ADK_CONVERT_TPARAMS_TO_UINT16PTR(ADK_TASK_USER_PARAMS(task, 12));
        destModId = ADK_CONVERT_TPARAMS_TO_UINT32(ADK_TASK_USER_PARAMS(task, 13));
        udp_length = ADK_CONVERT_TPARAMS_TO_UINT32(ADK_TASK_PDU_SIZE(task)) - 20;
        /* Verify UDP checksum */
        if(ADK_TASK_PDU_SIZE(task) < 
           ((ADK_MD_GET_L4_OFF(task) + udp_length)))
        {
            ADK_IP_DP_LOG_INFO("Truncated UDP packet! Dropping the packet.");
            return ADK_E_DROP_TASK;       
        }
        if(ipv4_udp_checksum(task, udp_length,
                                  *((uint32_t *)(pdu + 12)),
                                  *((uint32_t *)(pdu + 16))))
        {    
            ADK_IP_DP_LOG_INFO("Incorrect UDP Checksum! Dropping the packet.");
            return ADK_E_DROP_TASK;       
        }
    }

    task = (adk_pktio_task_t *)(task_p[0]);
    /* Get the pipeline info */                                                                         
    pipeline_info = ADK_CONVERT_TPARAMS_TO_PIPELINE_INFO(ADK_TASK_USER_PARAMS(task, 8));
    pipeline_src_info = &pipeline_info->exit_module[ADK_SIF_CFG_PIPELINE_MODULE_L4_LLI];
    pipeline_dest_info = &pipeline_src_info->dest_module[destModId];

    if(dst_mod_dst_func) {
        (adkDestEntryFunction_g[dst_mod_dst_func[pipeline_dest_info->func_idx]])(task_p, task_count,
                                                                            task_mask, pipeline_dest_info->ring, pipeline_dest_info->next_entry, coreIndex);
    }
    return ret;
}

adk_status_t adk_udp_dp_pkt_proc_ring_lli (void **task_p,
                        uint8_t __attribute__((unused)) task_count,
                        uint32_t task_mask,
                        void *ring,
                        adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                        __attribute__((unused)) int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t **task = (adk_pktio_task_t **)task_p;
    adk_pktio_task_t *task_tmp = NULL;
    uint16_t nb_tx;
    uint32_t i, num = 0;
    uint32_t cur_bit;
    struct rte_mbuf *tx_task[ADK_DP_BULK_RX_SIZE * ADK_DP_TASK_MULTIPLY_MAX] = {0};

    /*Tasks may not be packed since some tasks are dropped/consumed/fragmented. Pack the tasks*/
    for(i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if(task_mask & cur_bit)
        {
            task_tmp = (adk_pktio_task_t *)(task[i]);
            tx_task[num] = (struct rte_mbuf *)(task[i]);
            ADK_TASK_USER_PARAMS(tx_task[num], 8) =
                ADK_CONVERT_TO_TPARAMS(next_stage_entry);
            ADK_TASK_USER_PARAMS(tx_task[num], 9) =
                ADK_CONVERT_TO_TPARAMS(ADK_UDP_LLI_FN_ID);
            num++;
        }
    }

    nb_tx = rte_ring_sp_enqueue_burst((struct rte_ring *)ring, (void *)tx_task, num, NULL);
    if (unlikely(nb_tx != num))
    {
        for(i = nb_tx; i < num; ++i)
            rte_pktmbuf_free(tx_task[i]);
    }
   
    return ret;
}
#endif

/*!
 * =============================================================
 * adk_ip_dp_pkt_proc_ring_uli()
 * =============================================================
 */
adk_status_t adk_ip_dp_pkt_proc_ring_uli (void **task_p,
                        uint8_t __attribute__((unused)) task_count,
                        uint32_t task_mask,
                        void *ring,
                        adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                        __attribute__((unused)) int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t **task = (adk_pktio_task_t **)task_p;
    uint16_t nb_tx;
    uint32_t i, num = 0;
    uint32_t cur_bit;
    struct rte_mbuf *tx_task[ADK_DP_BULK_RX_SIZE * ADK_DP_TASK_MULTIPLY_MAX] = {0};
    struct rte_mbuf *chainTask =NULL; 

    /*Tasks may not be packed since some tasks are dropped/consumed/fragmented. Pack the tasks*/
    for(i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if(task_mask & cur_bit)
        {
            tx_task[num] = (struct rte_mbuf *)(task[i]);
            ADK_TASK_USER_PARAMS(tx_task[num], 8) =
                ADK_CONVERT_TO_TPARAMS(next_stage_entry);
            ADK_TASK_USER_PARAMS(tx_task[num], 9) = 
                ADK_CONVERT_TO_TPARAMS(ADK_IPV4_ULI_FN_ID);
            num++;
        }
    }

    nb_tx = rte_ring_sp_enqueue_burst((struct rte_ring *)ring, (void *)tx_task, num, NULL);
    if (unlikely(nb_tx != num))
    {
        for(i = nb_tx; i < num; ++i)
        {
            chainTask = (struct rte_mbuf *)ADK_TASK_USER_PARAMS(tx_task[i], 6);
            if(chainTask)
            {
                rte_pktmbuf_free(chainTask);
            }
            rte_pktmbuf_free(tx_task[i]);
        }
    }
   
    return ret;
}

#ifdef UDP_CHECKSUM_STAGE_AFTER_BB_IC_MODULE        
adk_status_t adk_udp_dp_pkt_proc_uli (void **task_p,
                        uint8_t task_count,
                        uint32_t task_mask,
                        __attribute__((unused)) void *ring,
                        __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                        int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    uint32_t i = 0, cur_bit = 0, destModId = 0;
    adk_pktio_task_t *task = NULL;
    uint8_t *pdu = NULL;
    uint16_t cksum;
    uint16_t *dst_mod_dst_func = NULL; 
    adk_sif_cfg_pipeline_info_t *pipeline_info = NULL;
    adk_sif_cfg_pipeline_exit_module_t *pipeline_src_info = NULL;
    adk_sif_cfg_pipeline_dest_module_t *pipeline_dest_info = NULL;
    
    /* Verify UDP checksum */
    for(i=0, cur_bit=1; i<ADK_DP_BULK_RX_SIZE; ++i,(cur_bit=cur_bit<<1))
    {
        if(!(task_mask & cur_bit))
            continue;

        task = (adk_pktio_task_t *)(task_p[i]);
        pdu = (uint8_t *)(ADK_TASK_USER_PARAMS(task, 7));
        dst_mod_dst_func = (uint16_t*)(ADK_TASK_USER_PARAMS(task, 12));
        destModId = (uint32_t)(ADK_TASK_USER_PARAMS(task, 13));
        cksum = adk_raw_cksum((void *)(pdu), (uint32_t)ADK_TASK_PDU_SIZE(task));
        ADK_TASK_USER_PARAMS(task, 0) = (uint32_t)cksum;            
    }

    task = (adk_pktio_task_t *)(task_p[0]);
    /* Get the pipeline info */                                                                         
    pipeline_info = (adk_sif_cfg_pipeline_info_t *)ADK_TASK_USER_PARAMS(task, 8);
    pipeline_src_info = &pipeline_info->exit_module[ADK_SIF_CFG_PIPELINE_MODULE_L4_ULI];
    pipeline_dest_info = &pipeline_src_info->dest_module[destModId];
    if(dst_mod_dst_func) {
        ret = (adkDestEntryFunction_g[dst_mod_dst_func[pipeline_dest_info->func_idx]])(task_p, task_count,
                                   task_mask, pipeline_dest_info->ring, pipeline_dest_info->next_entry, coreIndex);
    }

    return ret;
}

adk_status_t adk_udp_dp_pkt_proc_ring_uli (void **task_p,
                        uint8_t __attribute__((unused)) task_count,
                        uint32_t task_mask,
                        void *ring,
                        adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                        __attribute__((unused)) int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t **task = (adk_pktio_task_t **)task_p;
    adk_pktio_task_t *task_tmp = NULL;
    uint16_t nb_tx;
    uint32_t i, num = 0;
    uint32_t cur_bit;
    struct rte_mbuf *tx_task[ADK_DP_BULK_RX_SIZE * ADK_DP_TASK_MULTIPLY_MAX] = {0};

    /*Tasks may not be packed since some tasks are dropped/consumed/fragmented. Pack the tasks*/
    for(i = 0, cur_bit = 1; i < ADK_DP_BULK_RX_SIZE; ++i, (cur_bit = cur_bit << 1))
    {
        if(task_mask & cur_bit)
        {
            task_tmp = (adk_pktio_task_t *)(task[i]);
            tx_task[num] = (struct rte_mbuf *)(task[i]);
            ADK_TASK_USER_PARAMS(tx_task[num], 8) = (uint32_t)next_stage_entry;
            ADK_TASK_USER_PARAMS(tx_task[num], 9) = ADK_UDP_ULI_FN_ID;
            num++;
        }
    }

    nb_tx = rte_ring_sp_enqueue_burst((struct rte_ring *)ring, (void *)tx_task, num, NULL);
    if (unlikely(nb_tx != num))
    {
        for(i = nb_tx; i < num; ++i)
            rte_pktmbuf_free(tx_task[i]);
    }
   
    return ret;
}
#endif
