/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

#include <math.h>
#include "adk_sys_lcore.h"
#include <adk_ip_common.h>
#ifdef ADK_MODULE_IPSEC_ENABLED
#include <adk_ipsec_export.h>
#endif

#include <adk_ethernet_export.h>
#ifdef ADK_MODULE_TWAMP_ENABLED
#include <adk_twamp_dp_export.h>
#endif
#ifdef ADK_MODULE_BFD_ENABLED
#include <adk_bfd_dp_export.h>
#endif
#include <adk_hos_dp.h>
#include <adk_qos_dp_export.h>
#include <adk_cksum.h>
#include <adk_timer.h>
#include <adk_pmu.h>
#include <adk_ip_dp_internal.h>
#include <adk_ip_dp_internal_fc.h>

#include "adk_ph_dev_dp.h"
#include <adk_ethernet_conf.h>
#include <adk_sys.h>
#include "adk_em.h"
#include "adk_nd_common.h"
#include "adk_sif_dp.h"
#include "adk_sys_dp_flow_cache_uli_func.h"

#define IS_FRAGMENT(ipHdr) ( (*((uint8_t *)ipHdr+6) & 0x3f) || *((uint8_t *)ipHdr+7) )? true : false;

/* To take care of flow cache allocation failure */
#define FC_ALLOCED(m) do{\
    if(likely(scratch_pad_p->flow_cache_entry))\
    {\
        m;\
    }\
}while(0);


__thread adk_sif_dp_scratchpad_space_t *scratch_pad_p = NULL;

typedef struct common_params_per_run_s
{
    adk_pktio_task_t                * ph_task;              /* 8B */
    struct ipv4_hdr                 * ipHdr;                /* 8B */
    adk_ip_dp_interface_entry_t     * egIfEntry;            /* 8B */
    uint16_t                        egressIfId;             /* 2B */
    adk_ip_dp_intf_stats_t          * if_stats;             /* 8B */
    uint16_t                        mtu;                    /* 2B */
    adk_ip_addr_t                   nextHop;                /* 8B */
    uint8_t                         coreIndex;              /* 1B */
    uint8_t                         isNextEnet;             /* 1B */
}common_params_per_run_t;

common_params_per_run_t perCore[ADK_MAX_NUM_CORES];

static inline uint16_t getL2IfId(uint16_t ipIfId)
{
    if(ipIfId>=ADK_ETH_BR_MAX)
        return ipIfId + ADK_ETH_NUM_BRIDGE - ADK_ETH_BR_MAX;
    else
        return ipIfId;
}


static inline void ip_inc_l4_out_stats(int coreIndex, uint8_t proto, uint16_t len)
{

    if(ADK_IP_PROTOCOL_TCP == proto)
    {
        ++ip_dp_tbl_info_p->ipL4Stats[coreIndex].ipTcpOutPackets;
        ip_dp_tbl_info_p->ipL4Stats[coreIndex].ipTcpOutOctets += len;
    }
    else if(ADK_IP_PROTOCOL_SCTP == proto)
    {
        ++ip_dp_tbl_info_p->ipL4Stats[coreIndex].ipSctpOutPackets;
        ip_dp_tbl_info_p->ipL4Stats[coreIndex].ipSctpOutOctets += len;
    }
    else if(ADK_IP_PROTOCOL_UDP == proto)
    {
        INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutUDPPkts, 1);
        INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutUDPOctets, len);
    }
}

static adk_status_t uli_pass1(
        common_params_per_run_t *c
);

static adk_status_t uli_pass2(
    common_params_per_run_t *c
);

static adk_status_t adk_ip_dp_pkt_proc_uli_single(
    void **task_p,
    int coreIndex,
    uint8_t *isNextEnet
);

/*Generate UDP checksum*/
inline
uint16_t ipv4_udp_checksum_gen(adk_pktio_task_t *task,
                           uint16_t udp_len,
			   uint32_t src_addr_nw,
			   uint32_t dest_addr_nw)
{
    uint32_t cksum=0;
    struct ipv4_psd_header {
        uint32_t src_addr; /* IP address of source host. */
        uint32_t dst_addr; /* IP address of destination host. */
        uint8_t  zero;     /* zero. */
        uint8_t  proto;    /* L4 protocol type. */
        uint16_t len;      /* L4 length. */
    } psd_hdr;

    uint8_t *buff;
    uint16_t  total_len=0, seg_len = (task->data_len - ADK_MD_GET_L4_OFF(task));
    ADK_PDU_OFFSET_PTR_GET(task, buff, uint8_t *, ADK_MD_GET_L4_OFF(task));

    if(likely(udp_len <= seg_len))
    {
        cksum += adk_raw_cksum((void *)buff, udp_len);
    }
    else
    {
        bool odd_len = false;
        /*Multi segmented mbuf*/
        while((total_len+seg_len) < udp_len)
        {
            if(odd_len)
            {
                cksum += rte_bswap16(adk_raw_cksum((void *)buff, seg_len));
                odd_len = ((seg_len-1) % 2);
            }
            else
            {
                cksum += adk_raw_cksum((void *)buff, seg_len);
                odd_len = (seg_len % 2);
            }
            total_len+=seg_len;
            /*Access next segment*/
            task = (adk_pktio_task_t *)(task->next);
            buff = rte_pktmbuf_mtod(task, uint8_t *);
            seg_len = task->data_len;
        }

        if(odd_len)
        {
            cksum += rte_bswap16(adk_raw_cksum((void *)buff,
                    (udp_len - total_len)));
        }
        else
        {
            cksum += adk_raw_cksum((void *)buff, (udp_len - total_len));
        }
    }

    buff = (uint8_t *)&psd_hdr;
    psd_hdr.src_addr = src_addr_nw;
    psd_hdr.dst_addr = dest_addr_nw;
    psd_hdr.zero = 0;
    psd_hdr.proto = 17;
    psd_hdr.len = rte_cpu_to_be_16(udp_len);
    cksum += adk_raw_cksum((void *)buff, sizeof(psd_hdr));
    cksum = ((cksum & 0xffff0000) >> 16) + (cksum & 0xffff);
    cksum = (~cksum) & 0xffff;

    if(cksum)
        return cksum;
    else
        return 0xFFFF;
}



/*!
 * =============================================================
 * adk_ip_dp_pkt_proc_uli()
 * =============================================================
 */
adk_status_t adk_ip_dp_pkt_proc_uli (void **task_p,
          uint8_t task_count,
          uint32_t task_mask,
          __attribute__((unused)) void *ring,
          __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
          int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    /* We point this pointer to an allocated memory location which gives us next stage info.
     * adk_pal_dp_stage_info_get() assigns that memory location to this pointer
     */
    static adk_pal_dp_stage_info_t stage_info[2];
    adk_pal_dp_stage_info_t *stage_info_p = NULL;
    adk_pktio_task_t *main_task, *main_next, *main_prev = NULL,
                     *head_task = NULL;
    uint8_t isNextEnet;

    main_task = (adk_pktio_task_t *)(task_p[0]);
    scratch_pad_p = ADK_SIF_DP_GET_SCRATCHPAD_PTR(coreIndex);

    while(main_task)
    {
        main_next = (adk_pktio_task_t *)ADK_MD_GET_NEXT_FRAG(main_task);

        ret = adk_ip_dp_pkt_proc_uli_single((void **)(&main_task), coreIndex,
                            &isNextEnet);
        if(ADK_E_TASK_CONSUMED == ret) /* Whole chain would have been consumed */
            break;

        if(likely(ADK_SUCCESS == ret))
        {
            if(isNextEnet)
            {   /* next module is ethernet */
                INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutSentToEnetPkts, 1);
                if(likely(0 != stage_info[0].fn_id)) /* fn_id of 0 is for SYS. We will never send to it */
                    stage_info_p = &stage_info[0];
                else
                {
                    ret = adk_pal_dp_stage_info_get(
                            adkModuleMgrCpuServiceFn_g[ADK_ETHERNET_MODULE_ID].cpuEntryFn_ULI[_DEST_FUNC_DEFAULT], &stage_info_p);
                    stage_info[0] = *stage_info_p;
                }
            }
            else
            {   /* next module is ipsec */
                INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutSentToIPSecPkts, 1);
                if(likely(0 != stage_info[1].fn_id))
                    stage_info_p = &stage_info[1];
                else
                {
                    ret = adk_pal_dp_stage_info_get(
                            adkModuleMgrCpuServiceFn_g[ADK_IPSEC_MODULE_ID].cpuEntryFn_ULI[_DEST_FUNC_DEFAULT], &stage_info_p);
                    stage_info[1] = *stage_info_p;
                }
            }
        }

        if(unlikely(ADK_SUCCESS != ret))
        {
            /* processed pkt to be dropped or consumed. Remove from the main
             * chain */
            if(main_prev)
            {
                ADK_MD_GET_NEXT_FRAG(main_prev) = main_next;
            }
            if((ADK_E_INTERNAL_ERROR == ret)
               ||(ADK_E_DROP_TASK == ret))
            {
                /* Free the task */
                adk_pktio_task_t *cur_task, *tmp_task;
                cur_task = (adk_pktio_task_t *)(main_task);
                while(cur_task != main_next)
                {
                    tmp_task = (adk_pktio_task_t *)
                                ADK_MD_GET_NEXT_FRAG(cur_task);
                    adk_pal_dp_release_send(cur_task, ADK_PAL_DP_BATCH_CB_FN_IP, NULL);
                    ADK_MD_GET_NEXT_FRAG(cur_task) = NULL;
                    /*Free the task*/
                    adk_pktio_task_buffer_free(cur_task);
                    cur_task = tmp_task;
                }
            }
        }
        else
        {
            if(!head_task)
                head_task = main_task;

            main_prev = main_task;
        }
        main_task = main_next;
    }


    if(head_task)
    {
        task_p = (void **)(&head_task);
        if(stage_info_p && adkDestEntryFunction_g[stage_info_p->fn_id])
        {
            ret = adk_pal_dp_stage_send(task_p, task_count, task_mask, stage_info_p->fn_id, 0x00, 0x00, ADK_PAL_DP_BATCH_CB_FN_IP, ADK_PAL_DP_BATCH_CB_EVENT_MAX, NULL);
            if(unlikely(ADK_E_INTERNAL_ERROR == ret))
            {
                adk_pal_dp_release_send(task_p[0], ADK_PAL_DP_BATCH_CB_FN_IP, NULL);
                /* Free the task */
                adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[0]));
                ADK_IP_DP_LOG_INFO("Failed to enQ to next stage in IP ULI");
                ret = ADK_E_INTERNAL_ERROR;
            }
        }
    }

    return ret;
}


/* No error checking done. Seperate func to avoid code duplication */
static inline void inc_if_out_stats(adk_metadata_t *adk_metadata_p, adk_ip_dp_intf_stats_t *if_stats, uint16_t len)
{
    if(unlikely(adk_metadata_p->isIpMcastPkt))
    {   /* Mcast */
        ++if_stats->ifOutMcastPkts;
        if_stats->ifOutMcastOctets += len;
    }
    else if(unlikely(adk_metadata_p->isIpBcastPkt))
    {   /* Bcast */
        ++if_stats->ifOutBcastPkts;
        if_stats->ifOutBcastOctets += len;
    }
    else
    {   /* Ucast */
        ++if_stats->ifOutUcastPkts;
        if_stats->ifOutUcastOctets += len;
    }
} /* inc_if_out_stats */


static inline adk_status_t uli_egress_qos(
        common_params_per_run_t *c
)
{
    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(c->ph_task));
    /*Flags set by Netdriver for PTP traffic requiring egress timestamp*/
    if (unlikely((IP_ULI_SENDER_NETD == ipuliImc->pktOrigin) &&
            ((ADK_MD_GET_TS_MSG_TYPE(c->ph_task) == ADK_NETD_TS_CAPTURE_IDX_ALLOC) ||
                    (ADK_MD_GET_TS_MSG_TYPE(c->ph_task) == ADK_NETD_TS_CAPTURE))))
    {
        /*PTP queue lookup*/
        ADK_METADATA_PARAM(c->ph_task, dev_tx_queue_id) = adk_qos_dp_get_ptp_queue();
    }
    else
    {
        /* IP Qos Lookup */
        if(likely(c->egIfEntry->enableQoS))
        {
            /*Check metadata parameters received from IPsec
                  If innerIPVersion is ZERO then it is a plain text packet
                  in presence of non-ZERO value it could be either v4/v6
                  process accordingly*/
            uint8_t *pdu;
            adk_ip_egress_qos_classifier_ptn_t qosClassifierPtn;
            ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
            if (adk_md_p->innerIPVersion && adk_md_p->innerIPHdrOffset)
            {
                if (ADK_IPV4_VER == adk_md_p->innerIPVersion)
                {
                    struct ipv4_hdr *innerIPv4Hdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_INNER_IP_OFF(c->ph_task));
                    qosClassifierPtn.proto = ipuliImc->inner_l4_protocol;
                    qosClassifierPtn.dscp = (innerIPv4Hdr->type_of_service>>2);
                    qosClassifierPtn.sip0 = innerIPv4Hdr->src_addr;
                    qosClassifierPtn.sip1 = 0;
                    qosClassifierPtn.sip2 = 0;
                    qosClassifierPtn.sip3 = 0;
                    qosClassifierPtn.dip0 = innerIPv4Hdr->dst_addr;
                    qosClassifierPtn.dip1 = 0;
                    qosClassifierPtn.dip2 = 0;
                    qosClassifierPtn.dip3 = 0;
                }
                else if (ADK_IPV6_VER == adk_md_p->innerIPVersion)
                {
                    struct ipv6_hdr *innerIPv6Hdr  = (struct ipv6_hdr*)(pdu + ADK_MD_GET_INNER_IP_OFF(c->ph_task));
                    qosClassifierPtn.proto = ipuliImc->inner_l4_protocol;
                    qosClassifierPtn.dscp = (innerIPv6Hdr->vtc_flow >> 22) & 0xff;
                    qosClassifierPtn.sip0 = innerIPv6Hdr->src_addr[0];
                    qosClassifierPtn.sip1 = innerIPv6Hdr->src_addr[1];
                    qosClassifierPtn.sip2 = innerIPv6Hdr->src_addr[2];
                    qosClassifierPtn.sip3 = innerIPv6Hdr->src_addr[3];
                    qosClassifierPtn.dip0 = innerIPv6Hdr->dst_addr[0];
                    qosClassifierPtn.dip1 = innerIPv6Hdr->dst_addr[1];
                    qosClassifierPtn.dip2 = innerIPv6Hdr->dst_addr[2];
                    qosClassifierPtn.dip3 = innerIPv6Hdr->dst_addr[3];
                }
            }
            else
            {
                qosClassifierPtn.proto = c->ipHdr->next_proto_id;
                qosClassifierPtn.dscp = (c->ipHdr->type_of_service >> 2) & 0x3F;
                qosClassifierPtn.sip0 = c->ipHdr->src_addr;
                qosClassifierPtn.sip1 = 0;
                qosClassifierPtn.sip2 = 0;
                qosClassifierPtn.sip3 = 0;
                qosClassifierPtn.dip0 = c->ipHdr->dst_addr;
                qosClassifierPtn.dip1 = 0;
                qosClassifierPtn.dip2 = 0;
                qosClassifierPtn.dip3 = 0;
            }

            uint8_t aclTreeIndex = ip_dp_tbl_info_p->egressQoSClassifierInfo_p[c->egressIfId]->aclTreeIndex;
            uint32_t qosClassifierFlowId;
            adk_ip_dp_egress_qos_classifier_lookup(
                    ip_dp_tbl_info_p->egressQoSClassifierInfo_p[c->egressIfId]->ipEgressQoSClassifierCtx[aclTreeIndex],
                    &qosClassifierPtn, &qosClassifierFlowId);

            /*Fill up the transmit queue in ADK metadata*/
            adk_ip_dp_egress_qos_classifier_action_info_t *qosClassifierAction_p;
            qosClassifierAction_p = &(ip_dp_tbl_info_p->egressQoSClassifierActionDp_p[c->egressIfId][aclTreeIndex][qosClassifierFlowId]);

            ADK_METADATA_PARAM(c->ph_task, dev_tx_queue_id) = qosClassifierAction_p->queueId;
            ADK_METADATA_PARAM(c->ph_task, pktColor) = qosClassifierAction_p->pktColor;
        }
        else
        {
            ADK_METADATA_PARAM(c->ph_task, dev_tx_queue_id) = adk_qos_dp_get_default_tx_queue();
        }
    }

    if (ADK_QOS_INVALID_QUEUE_ID == adk_md_p->dev_tx_queue_id)
    {
        ADK_IP_DP_LOG_INFO_DEBUG("IP ULI: Invalid QoS queue id. Dropping the packet.");
        c->if_stats->ifOutQoSLookupFailDrops++;
        return ADK_E_DROP_TASK;
    }

    bool isPHBQueue = false;
    uint32_t qosStatsId;
    /*Find the stats id for QoS queue*/
    qosStatsId = adk_qos_dp_get_queue_stats_id(adk_md_p->dev_tx_queue_id, &isPHBQueue);
    
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->dev_tx_queue_id, adk_md_p->dev_tx_queue_id));
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->pktColor , adk_md_p->pktColor));
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->qosStatsId, qosStatsId));
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->isPHBQueue, isPHBQueue));

    adk_md_p->qosStatsId = qosStatsId;
    adk_md_p->isPHBQueue = isPHBQueue;
    adk_md_p->isIpPkt = true;

    return ADK_SUCCESS;
} /* uli_egress_qos */

/* TO DO: convert this to MACRO and take care of fastpath/slowpath difference */
static inline void set_eth_imc_and_md(
    common_params_per_run_t *c,
    uint8_t num_frags
)
{
    /* --  S E T   M E T A D A T A -- */
    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
    adk_ethernet_dp_uli_imc_param_t *eth_imc = (adk_ethernet_dp_uli_imc_param_t *)&(ADK_MD_GET_IMC_PARAM_START(c->ph_task));

    eth_imc->dscp                       = (c->ipHdr->type_of_service >> 2) & 0x3F;
    eth_imc->pcp                        = c->egIfEntry->dscp_lkup[eth_imc->dscp] & 0x0F;
    eth_imc->l4_proto                   = c->ipHdr->next_proto_id;
    eth_imc->output_intf_id             = getL2IfId(c->egressIfId);
    eth_imc->next_hop.u.ipv4_addr       = c->nextHop;
    eth_imc->next_hop.is_ipv6           = 0;
    eth_imc->frag_count                 = num_frags;

    c->ph_task->l2_len                  = 0;
    adk_md_p->pdu_flow_id            = 0; /* DestFlow */

    /* update the egress stats Id */
    adk_md_p->outerIPVersion            = ADK_IPV4_VER;
    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id =
            c->egIfEntry->if_stats_id;


    ADK_MD_SET_EXCP_DOMAIN(c->ph_task, ADK_OS_ORIGIN_DOMAIN);

    /* Set the next module as ETHERNET */
    c->isNextEnet = 1;

} /* set_eth_imc_and_md */

/*
 * CKSum handling
 *
 * Origination case
 * ----------------
 *
 *              ** BYPASS=TRUE **
 *             UDP                NON-UDP
 * plntxt   |  Set-flag     |     reset-flag
 * encr     |  NA           |     reset-flag
 *
 *              ** BYPASS=FALSE **
 *             UDP                NON-UDP
 * plntxt   |  Set-flag     |     reset-flag
 * encr     |  NA           |     dont-touch
 *
 *
 *
 * Forwarding case (basically, if forwarded packet, reset the flag)
 * ---------------
 *              ** BYPASS=TRUE **
 *             UDP                NON-UDP
 * plntxt   |  reset-flag   |     reset-flag
 * encr     |  NA           |     NA
 *
 *              ** BYPASS=FALSE **
 *             UDP                NON-UDP
 * plntxt   |  reset-flag   |     reset-flag
 * encr     |  NA           |     reset-flag
 *
 */

static inline void uli_handle_udp_cksum_gen(
    common_params_per_run_t *c,
    struct udp_hdr *udpHdr
        )
{
    if(likely(IP_ULI_SENDER_IPLLI != scratch_pad_p->pktOrigin))
    {
        if(likely(udpHdr))
        {/* If udpHdr is going to point to non-udp headers also,
            then this logic needs to add proto==UDP check */

            /* Set the UDP cksum*/
            udpHdr->dgram_cksum = 0;
            /* Generate UDP checksum flag is set, either generate it in SW OR generate pseudo header checksum and keep
             * the flag for HW
             */
            if(likely(c->ph_task->ol_flags & PKT_TX_UDP_CKSUM))
            {
                udpHdr->dgram_cksum = 0;
#ifdef UDP_CHECKSUM_ENABLED
                /* offload udp checksum calculation. */
                c->ph_task->l4_len = UDP_HDR_LEN;
                uint16_t tmp_16;
                tmp_16 = rte_ipv4_phdr_cksum(c->ipHdr, c->ph_task->ol_flags);
                udpHdr->dgram_cksum = tmp_16;
                ADK_IP_DP_LOG_INFO_DEBUG("pseudo UDP Checksum =%x, l4_offset=%d\n", udpHdr->dgram_cksum, ADK_MD_GET_L4_OFF(c->ph_task));
#else
                udpHdr->dgram_cksum = ipv4_udp_checksum_gen(c->ph_task,
                        rte_be_to_cpu_16(c->ipHdr->total_length)-IP_HDR_LEN,
                        c->ipHdr->src_addr, c->ipHdr->dst_addr);
                c->ph_task->ol_flags &= ~PKT_TX_UDP_CKSUM;
#endif
            }
            ADK_IP_DP_LOG_INFO_DEBUG("UDP Checksum=%x\n", udpHdr->dgram_cksum);
        } /* Generate checksum */
        else
        {
            if(unlikely(ADK_IP_PROTOCOL_ESP != c->ipHdr->next_proto_id && ADK_IP_PROTOCOL_AH != c->ipHdr->next_proto_id))
                c->ph_task->ol_flags &= ~PKT_TX_UDP_CKSUM; /* Non-udp non-encrypted packet */
        } /* Don't generated UDP checksum */
    } /* originated packet */
    else
            c->ph_task->ol_flags &= ~PKT_TX_UDP_CKSUM; /* Forwarded packet */

} /* uli_handle_udp_cksum_gen */


static inline adk_status_t uli_handle_fragmentation(
    common_params_per_run_t *c
)
{
    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
    ADK_IP_DP_LOG_INFO_DEBUG("mtu=%d, pkt-size=%d, fragmenting!\n", c->mtu, (ADK_TASK_PDU_SIZE(c->ph_task) - ADK_MD_GET_IP_OFF(c->ph_task)));

    /* Note: Assumption is that adk_ip_dp_fragment will *not* re-order and place new mbufs before parent mbuf while chaining. */
    uint8_t num_frags, loop_count;
    uint16_t total_payload;
    uint8_t genPartialL4Cksum=0;
    if( ((ADK_IP_PROTOCOL_UDP==c->ipHdr->next_proto_id) || (IP_FILTER_PROTOCOL_UDP_LITE==c->ipHdr->next_proto_id)) &&
            (c->ph_task->ol_flags & PKT_TX_UDP_CKSUM)
        )
    {
        genPartialL4Cksum = 1;
    }


    adk_ip_dp_frag_error_t frag_result = adk_ip_dp_fragment(&c->ph_task, c->mtu, &num_frags, &total_payload, genPartialL4Cksum);
    ADK_PDU_OFFSET_PTR_GET(c->ph_task, c->ipHdr, struct ipv4_hdr *, ADK_MD_GET_IP_OFF(c->ph_task));

    if(likely(frag_result == ADK_IP_DP_FRAG_OK))
    {
        ++c->if_stats->ifOutFragPkts;
        c->if_stats->ifOutFragments += num_frags;
        ADK_IP_DP_LOG_INFO_DEBUG("Total packet fragments: %d\n", num_frags);

        /* --  S E T   M E T A D A T A -- */
        /* Important: We don't support a case when a pre-fragmented packet is fragmented again. */
        set_eth_imc_and_md(c, num_frags);

        /*Copy the task params for the remaining fragments*/
        adk_pktio_task_t *temp_task = (adk_pktio_task_t *)ADK_MD_GET_NEXT_FRAG(c->ph_task);
        adk_metadata_t *adk_md_temp_p = ADK_METADATA_PTR(temp_task);

        for(loop_count = 1; loop_count < num_frags; loop_count++)
        {
            adk_md_temp_p->pdu_flow_id = adk_md_p->pdu_flow_id;
            ADK_MD_GET_EXCP_DOMAIN(temp_task) = ADK_MD_GET_EXCP_DOMAIN(c->ph_task);
            ADK_MD_GET_IP_OFF(temp_task) = ADK_MD_GET_IP_OFF(c->ph_task);
            adk_metadata_t *adk_md_frag_p = ADK_METADATA_PTR(temp_task);
            adk_md_frag_p->saID = ADK_IPSEC_INVALID_SA_ID;

            ADK_MEMCPY(&(ADK_MD_GET_IMC_PARAM_START(temp_task)), &(ADK_MD_GET_IMC_PARAM_START(c->ph_task)),
                    sizeof(adk_ethernet_dp_uli_imc_param_t));

            adk_md_temp_p-> egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id =
                    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id;

            adk_md_temp_p->egress_mdata_stats[ADK_STATS_MDATA_ID_OUTER_IP_STATS].stats_id =
                    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_OUTER_IP_STATS].stats_id;

            adk_md_temp_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id =
                    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id;

            adk_md_temp_p->outerIPVersion = adk_md_p->outerIPVersion;
            adk_md_temp_p->innerIPVersion = adk_md_p->innerIPVersion;
            adk_md_temp_p->isIpMcastPkt = adk_md_p->isIpMcastPkt;
            adk_md_temp_p->isIpBcastPkt = adk_md_p->isIpBcastPkt;
            /* */
            temp_task = (adk_pktio_task_t *)ADK_MD_GET_NEXT_FRAG(temp_task);
            adk_md_temp_p = ADK_METADATA_PTR(temp_task);
        }
    } /* fragmentation success */
    else
    {
        ++c->if_stats->ifOutDroppedPkts;
        c->if_stats->ifOutDroppedOctets += (ADK_TASK_PDU_SIZE(c->ph_task) - ADK_MD_GET_IP_OFF(c->ph_task));
        if((frag_result == ADK_IP_DP_FRAG_DFSET))
        {
            ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutFragDFSetDrops++;
            /* Not sure if this param is used by ethernet or HLP in forwarding path. Send_exception uses this field to get the interface id */
            adk_md_p->input_intf_id = c->egressIfId;
            int coreIndex = c->coreIndex;
            return adk_ip_dp_send_exception(c->ph_task, ADK_IP_FRAGMENTATION_EXCEPTION, coreIndex, 0, ADK_IP_DIR_EGRESS);
        }
        else if(unlikely((frag_result == ADK_IP_DP_FRAG_LIMIT_EXCEED)))
        {
            ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutTooBigDrops++;
            ADK_IP_DP_LOG_INFO_DEBUG("IP ULI: Packet can't be fragmented as it will exceed max-frag-limit. dropping the packet");
            return ADK_E_DROP_TASK;
        }

        ADK_IP_DP_LOG_INFO_DEBUG("IP ULI: Fragmentation failed. dropping the packet");
        return ADK_E_DROP_TASK;
    } /* fragmentation failure */

    return ADK_SUCCESS;
} /* uli_handle_fragmentation */


static inline adk_status_t proc_no_route_found(
    common_params_per_run_t *c,
    adk_ip_dp_pbr_action_t *pbrAct,
    uint8_t routeTblId
)
{
    ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutNoRouteFoundDrops++;
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->outerSipStatsId,
            (NULL==pbrAct)?routeTblId:pbrAct->routeTableId));

    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->egressIntfId,
            ADK_DP_MAX_IP_INTERFACE));

    return adk_ip_dp_send_exception(c->ph_task, ADK_IP_NO_ROUTE_EXCEPTION, c->coreIndex,
            (NULL==pbrAct)?routeTblId:pbrAct->routeTableId, ADK_IP_DIR_EGRESS);
} /* proc_no_route_found */

static inline adk_status_t pbr_and_route_lookup(
    common_params_per_run_t *c
)
{
    /*          P O L I C Y   B A S E D   R O U T I N G        */

    adk_ip_dp_pbr_action_t *pbrAct=NULL;
    bool lookup_defTbl=false;
    uint8_t routeTblId;
    adk_ip_pbr_val_t pbrActIndex;
    adk_ip_dp_route_info_t *ipRouteInfo_p;
    adk_dp_status_t dp_st;
    adk_handle_t lkup_result;

    /* if pbr lookup fails or route lookup fails on the resultant tableId, lookup default table */
    if(likely(TRUE == adk_ip_dp_pbr(c->ipHdr->src_addr, &pbrActIndex)))
    {
        pbrAct = (adk_ip_dp_pbr_action_t *)&ip_dp_tbl_info_p->pbrActionTable_p[(uintptr_t)pbrActIndex];

        if(likely(ADK_IP_MAX_ROUTING_TABLES != pbrAct->routeTableId))
        {
            ADK_IP_DP_LOG_INFO_DEBUG("PBR returned tableId=%d for IP:0x%x\n", pbrAct->routeTableId, adk_ntohl(c->ipHdr->dst_addr));

            (pbrRuleStats[c->coreIndex*ADK_IP_MAX_PBR_RULES + pbrActIndex - 1])++;
            routeTblId = pbrAct->routeTableId;

            dp_st = adk_ip_dp_dip_get(/*vrdId*/ routeTblId, c->ipHdr->dst_addr, &lkup_result);
            if(likely(ADK_SUCCESS == dp_st && (NULL != lkup_result)))
            {
                ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
            } /* dip lookup success */
            else
            {
                ADK_IP_DP_LOG_INFO_DEBUG("DIP Lookup Failed in PBR given table (%d) for IP:0x%x\n", pbrAct->routeTableId, adk_ntohl(c->ipHdr->dst_addr));
                lookup_defTbl = TRUE;
            } /* dip lookup failure */
        } /* Valid table id by PBR */
        else
        {
            ADK_IP_DP_LOG_INFO_DEBUG("No PBR policy found for dip=0x%x\n", adk_ntohl(c->ipHdr->dst_addr));
            lookup_defTbl = TRUE;
            pbrAct = NULL;
        } /* Invalid table id by pbr */
    } /* pbr lookup success */
    else
    {/* PBR lookup failed. Lookup default routing table for route */
        ADK_IP_DP_LOG_INFO_DEBUG("No PBR policy found for dip=0x%x\n", adk_ntohl(c->ipHdr->dst_addr));
        lookup_defTbl = TRUE;
    } /* pbr lookup failure */


    if(TRUE == lookup_defTbl)
    {   /* dip lookup in default table */

        ADK_IP_DP_LOG_INFO_DEBUG("DIP Lookup for IP:0x%x in default routing table (%d)\n", adk_ntohl(c->ipHdr->dst_addr), ip_dp_tbl_info_p->defRouteTblId);
        routeTblId = ip_dp_tbl_info_p->defRouteTblId;

        dp_st = adk_ip_dp_dip_get(/*vrdId*/routeTblId, c->ipHdr->dst_addr, &lkup_result);
        if(likely(ADK_SUCCESS == dp_st && (NULL != lkup_result)))
        {
            ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
        } /* dip lookup success */
        else
        {
            ADK_IP_DP_LOG_INFO_DEBUG(" Result> No match found!\n");
            return proc_no_route_found(c, pbrAct, routeTblId);
        } /* dip lookup failure */
    } /* Default table lookup */


    /*                R O U T E   L O O K U P   F O R    I F A C E    I N F O          */

    /* We want to use the routeMTU of the first route lookup */
    c->mtu = ipRouteInfo_p->routeMTU;
    c->nextHop = ipRouteInfo_p->u.nh_if.nextHop;

    /* We will have egressIfId Only when there is direct route and address is not subnet broadcast or self ip address */
    if(ADK_DP_MAX_IP_INTERFACE != ipRouteInfo_p->u.nh_if.interfaceId && FALSE == ipRouteInfo_p->isTerminating)
    {
        c->egressIfId = ipRouteInfo_p->u.nh_if.interfaceId;
        /* For direct route, next hop is nothing but the dip. ipRouteInfo_p->u.nh_if.nextHop will be invalid in that case */
        c->nextHop = rte_be_to_cpu_32(c->ipHdr->dst_addr);
    } /* connected route in 1st lookup */
    else
    {
        if(false == ((adk_ip_dp_route_info_t *)lkup_result)->isTerminating)
        {   /* Packet DIP is not local ip */

            dp_st = adk_ip_dp_dip_get(/*vrdId*/ routeTblId, rte_cpu_to_be_32(ipRouteInfo_p->u.nh_if.nextHop), &lkup_result);
            if(likely(ADK_SUCCESS == dp_st                                          &&
                    (NULL != lkup_result)                                           &&
                    FALSE == ((adk_ip_dp_route_info_t *)lkup_result)->isTerminating &&
                    ADK_DP_MAX_IP_INTERFACE != ((adk_ip_dp_route_info_t *)lkup_result)->u.nh_if.interfaceId
                    )
            )
            {
                ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
                c->egressIfId = ((adk_ip_dp_route_info_t *)lkup_result)->u.nh_if.interfaceId;
            } /* dip lookup success */
            else
            {
                ADK_IP_DP_LOG_INFO_DEBUG(" DIP lookup for next hop failed!\n");
                return proc_no_route_found(c, pbrAct, routeTblId);
            } /* dip lookup failure */
        }
        else
        {
            /* isTerminating check will go away with separate local address table. Getting own address as DIP is not
             * a practical use case for us.
             */
            return ADK_E_DROP_TASK;
        }
    } /* user configured route in 1st lookup */

    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->routeMtu, c->mtu));
    scratch_pad_p->mtu = c->mtu; /* TO DO: Why this when we have a fc entry? */
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->nextHop , c->nextHop));
    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->egressIntfId, c->egressIfId));

    return ADK_SUCCESS;
} /* pbr_and_route_lookup */

static inline adk_status_t set_outer_addr_stats_id(
    common_params_per_run_t *c
        )
{
    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(c->ph_task));
    /* Set outer ip stats id */
    if( likely(IP_ULI_SENDER_IPLLI != ipuliImc->pktOrigin))
    {
        /* sip will not be local in fwded case. Hence the above check (optimization) */
        /* ip-address counters - find if its a locally originated packet. */
        adk_handle_t lkup_result;
        adk_dp_status_t dp_st = adk_ip_dp_dip_get(ADK_IP_DEFAULT_VRD_ID, c->ipHdr->src_addr, &lkup_result);
        if(likely(ADK_SUCCESS == dp_st) && likely(lkup_result))
        {
            adk_ip_dp_route_info_t *ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
            if(ipRouteInfo_p->isTerminating)
            {
                adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
                FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->outerSipStatsId,
                        ipRouteInfo_p->u.termination.ip_addr_stats_id));

                adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_OUTER_IP_STATS].stats_id =
                        ipRouteInfo_p->u.termination.ip_addr_stats_id;
                adk_md_p->outerIPVersion = ADK_IPV4_VER;
                /* We don't want to increment counters twice if outer and inner are same */
                if (adk_md_p->innerIPVersion==ADK_IPV4_VER &&
                        (adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id == ipRouteInfo_p->u.termination.ip_addr_stats_id)
                )
                {
                    /* reset inner */
                    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->innerSipStatsId, (ADK_SYS_INVALID_MDATA_ID & 0xFFFF)));
                    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id = ADK_SYS_INVALID_MDATA_ID;
                }
            } /* sip local */
            else
            {
                if(IP_ULI_SENDER_EO == scratch_pad_p->pktOrigin)
                {
                    ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutNonLocalSipFromEoDrops++;
                    return ADK_E_DROP_TASK;
                }
            } /* sip not local */
        } /* sip found in local addr table */
        else
        {
            if(IP_ULI_SENDER_EO == scratch_pad_p->pktOrigin)
            {
                ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutNonLocalSipFromEoDrops++;
                return ADK_E_DROP_TASK;
            }
        } /* sip not found in local addr table */
    } /* set address stats id */

    return ADK_SUCCESS;
} /* set_outer_addr_stats_id */


static inline adk_status_t proc_forwarding_disabled(
    common_params_per_run_t *c
        )
{
    ADK_IP_DP_LOG_ERROR("Routing disabled on egress interface, dropping the forwarded packet\n");
    ADK_IP_DP_LOG_INFO_DEBUG("DIP=0x%x, interface-id=%d\n", adk_ntohl(c->ipHdr->dst_addr), c->egressIfId);
    /*Drop the packet*/
    c->if_stats->ifRoutingDisabledDrops++;
    c->if_stats->ifOutDroppedPkts++;
    c->if_stats->ifOutDroppedOctets += rte_be_to_cpu_16(c->ipHdr->total_length);
    ADK_METADATA_PTR(c->ph_task)->input_intf_id = c->egressIfId;
    return adk_ip_dp_send_exception(c->ph_task,
            ADK_IP_FORWARDING_DISABLED_EXCEPTION, c->coreIndex, 0, ADK_IP_DIR_EGRESS);
}

static adk_status_t uli_pass1(
    common_params_per_run_t *c
)
{
    adk_status_t ret;
    uint8_t *pdu;
    ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(c->ph_task));

    scratch_pad_p->ip_version = ADK_IPV4_VER;
    scratch_pad_p->ip_offset = 0;
    scratch_pad_p->pktOrigin = ipuliImc->pktOrigin;;

    c->ipHdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_IP_OFF(c->ph_task));

    if(likely( IP_HDR_LEN == (((uint8_t *)c->ipHdr)[0] & 0x0f) * 4))  // l3_len = IHL*4
    {
        if(likely(adk_ntohs(c->ipHdr->total_length) <= ADK_TASK_PDU_SIZE(c->ph_task)))
        {
            if( likely(IP_ULI_SENDER_IPLLI != scratch_pad_p->pktOrigin))
            {   /* originated packet. set addr stats id */
                struct udp_hdr *udpHdr;

                if(likely( (ADK_IP_PROTOCOL_UDP==c->ipHdr->next_proto_id) || (IP_FILTER_PROTOCOL_UDP_LITE==c->ipHdr->next_proto_id) ))
                    udpHdr = (struct udp_hdr *)(void *)(pdu + IP_HDR_LEN);
                else
                    udpHdr=NULL;


                if (likely(/* TO DO: remove udp header */udpHdr && IP_ULI_SENDER_EO == scratch_pad_p->pktOrigin) )
                { /* We are using flow-cache only for perf case right now. i.e. UDP and EO originated traffic */
                    uint32_t  hash_key;
                    ret = adk_sys_dp_uli_flow_cache_get((adk_ip_hdr_t *)c->ipHdr, ADK_IPV4, c->ipHdr->next_proto_id,
                            udpHdr->src_port, udpHdr->dst_port, &scratch_pad_p->flow_cache_entry, &hash_key);

                    if (unlikely(ADK_E_FLOW_CACHE_ALLOC_FAILED == ret))
                        scratch_pad_p->flow_cache_entry = NULL;

                    FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->nextHdr, c->ipHdr->next_proto_id));

                    if (likely(ADK_E_FLOW_CACHE_PKT_FAST_PATH == ret))
                    {
                        /*Active Flow cache .. go for cached action processing*/
                        scratch_pad_p->pktPath = ADK_E_FLOW_CACHE_PKT_FAST_PATH;
                        adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id = scratch_pad_p->flow_cache_entry->innerSipStatsId;
                        adk_md_p->innerIPVersion = ADK_IPV4_VER;
                        adk_md_p->innerIpPktLen = rte_be_to_cpu_16(c->ipHdr->total_length);
                        ADK_IP_DP_LOG_INFO_DEBUG("v4 innerIpStatsId(%d)=%d\n", ADK_STATS_MDATA_ID_INNER_IP_STATS, adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id);

                    } /* fc fast path */

                } /* l4 proto=udp.. fc get */

                if( unlikely(ADK_E_FLOW_CACHE_PKT_FAST_PATH != scratch_pad_p->pktPath))
                {
                    scratch_pad_p->pktPath = ADK_E_FLOW_CACHE_PKT_SLOW_PATH;
                    adk_dp_status_t dp_st;
                    adk_handle_t lkup_result;
                    dp_st = adk_ip_dp_dip_get(ADK_IP_DEFAULT_VRD_ID, c->ipHdr->src_addr, &lkup_result);
                    if(likely(ADK_SUCCESS == dp_st && lkup_result))
                    {
                        adk_ip_dp_route_info_t *ipRouteInfo_p = (adk_ip_dp_route_info_t *)lkup_result;
                        if(likely(ipRouteInfo_p->isTerminating))
                        {
                            FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->innerSipStatsId, ipRouteInfo_p->u.termination.ip_addr_stats_id));
                            ADK_METADATA_PTR(c->ph_task)->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id = ipRouteInfo_p->u.termination.ip_addr_stats_id;
                            adk_md_p->innerIPVersion = ADK_IPV4_VER;
                            adk_md_p->innerIpPktLen = rte_be_to_cpu_16(c->ipHdr->total_length);
                        } /* local sip */
                        else
                        {
                            if(IP_ULI_SENDER_EO == scratch_pad_p->pktOrigin)
                            {
                                ADK_IP_DP_LOG_INFO_DEBUG("EO path: v4 inner sip (%d) is not local. Dropping\n", c->ipHdr->src_addr);
                                ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutNonLocalSipFromEoDrops++;
                                return ADK_E_DROP_TASK;
                            }
                            else
                            {
                                ADK_IP_DP_LOG_INFO_DEBUG("NETD path: v4 inner sip (%d) is not local. Accepting it\n", c->ipHdr->src_addr);
                                INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].UipOutNonLocalSipFromHostPkts, 1);
                            }
                        } /* non local sip */
                    }
                    else
                    {
                        if(IP_ULI_SENDER_EO == scratch_pad_p->pktOrigin)
                        {
                            ADK_IP_DP_LOG_INFO_DEBUG("EO path: v4 inner sip (%d) is not local. Dropping\n", c->ipHdr->src_addr);
                            ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutNonLocalSipFromEoDrops++;
                            return ADK_E_DROP_TASK;
                        }
                        else
                        {
                            ADK_IP_DP_LOG_INFO_DEBUG("NETD path: v4 inner sip (%d) is not local. Accepting it\n", c->ipHdr->src_addr);
                            INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].UipOutNonLocalSipFromHostPkts, 1);
                        }
                    }
                } /* non udp or non eo or first udp packet of the flow */
                ip_inc_l4_out_stats(c->coreIndex, c->ipHdr->next_proto_id, rte_be_to_cpu_16(c->ipHdr->total_length) - IP_HDR_LEN);
            } /* originated packet */

            adk_ipsec_dp_uli_imc_param_t *ipsecUliImc = (adk_ipsec_dp_uli_imc_param_t *)&(ADK_MD_GET_IMC_PARAM_START(c->ph_task));
            ipsecUliImc->pktOrigin = scratch_pad_p->pktOrigin;
            c->isNextEnet = 0;

            return ADK_SUCCESS;

        } /* ip packet len < total pdu len */
        else
        {
            ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutTruncatedPktDrops++;
            return ADK_E_DROP_TASK;
        }
    } /* valid ip header length */
    else
    {
        ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutInvalidIpHdrLenDrops++;
        return ADK_E_DROP_TASK;
    } /* invalid ip header length */

} /* uli_pass1 */


static adk_status_t uli_pass2_fastpath(
    common_params_per_run_t *c
)
{
    /* TO DO: MAKE SURE THAT scratchpad->pktOrigin is updated for packet at adk entry */
    uint8_t *pdu;
    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
    ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
    c->ipHdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_IP_OFF(c->ph_task));
    struct udp_hdr *udpHdr;
    ADK_IP_DP_LOG_INFO_DEBUG("IP ULI_single entry flow cache path ..\n");
    /* Offloading outer cksum as well */
    c->ipHdr->hdr_checksum = 0;

    if(likely( (ADK_IP_PROTOCOL_UDP==c->ipHdr->next_proto_id) || (IP_FILTER_PROTOCOL_UDP_LITE==c->ipHdr->next_proto_id) ))
        udpHdr = (struct udp_hdr *)(void *)(pdu + IP_HDR_LEN);
    else
        udpHdr=NULL;

    if (unlikely(ADK_DP_MAX_IP_INTERFACE == scratch_pad_p->flow_cache_entry->egressIntfId))
    {
        ADK_IP_DP_LOG_INFO_DEBUG("Invalid interface id received from route lookup. Dropping the packet\n");
        ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutNoRouteFoundDrops++;
        return adk_ip_dp_send_exception(c->ph_task, ADK_IP_NO_ROUTE_EXCEPTION, c->coreIndex, scratch_pad_p->flow_cache_entry->outerSipStatsId, ADK_IP_DIR_EGRESS);
    }

    c->egressIfId   = scratch_pad_p->flow_cache_entry->egressIntfId;
    c->mtu          = scratch_pad_p->flow_cache_entry->routeMtu;
    c->nextHop      = scratch_pad_p->flow_cache_entry->nextHop;
    c->egIfEntry    = &(ip_dp_tbl_info_p->ipInterfaceTble[scratch_pad_p->flow_cache_entry->egressIntfId]);
    c->if_stats     = &(ip_dp_tbl_info_p->ipInterfaceStatsTble[c->coreIndex][c->egIfEntry->if_stats_id]);

    ADK_IP_DP_LOG_INFO_DEBUG(" Result> (NH:0x%x; Egress IP If Id:0x%x)\n", c->nextHop,
            scratch_pad_p->flow_cache_entry->egressIntfId);


    if(unlikely( (IP_ULI_SENDER_IPLLI==scratch_pad_p->pktOrigin) && c->egIfEntry->routingDisabled))
        return proc_forwarding_disabled(c);

    if(unlikely(ADK_IP_DP_INTF_TYPE_ETH != c->egIfEntry->ifType) )
    {/* TO DO: remove if counter not needed */
        ADK_IP_DP_LOG_INFO("IP interface not created. Dropping packet!");
        ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutInvalidIfDrops++;
        return ADK_E_DROP_TASK;
    }

    adk_md_p->dev_tx_queue_id = scratch_pad_p->flow_cache_entry->dev_tx_queue_id;
    adk_md_p->pktColor = scratch_pad_p->flow_cache_entry->pktColor;
    adk_md_p->qosStatsId = scratch_pad_p->flow_cache_entry->qosStatsId;
    adk_md_p->isPHBQueue = scratch_pad_p->flow_cache_entry->isPHBQueue;
    adk_md_p->isIpPkt = true;
    
    if (unlikely((c->ipHdr->dst_addr & 0x000000f0) == 0xE0)) {
        adk_md_p->isIpMcastPkt = true;
    }else if (unlikely(c->ipHdr->dst_addr == 0xFFFFFFFF)) {
        adk_md_p->isIpBcastPkt = true;
    }else{
        adk_md_p->isIpMcastPkt = false;
        adk_md_p->isIpBcastPkt = false;
    }

    if (unlikely(ADK_QOS_INVALID_QUEUE_ID == adk_md_p->dev_tx_queue_id))
    {
        ADK_IP_DP_LOG_INFO_DEBUG("IP ULI: Invalid QoS queue id. Dropping the packet.");
        c->if_stats->ifOutQoSLookupFailDrops++;
        return ADK_E_DROP_TASK;
    }

    /* TO DO: this needs to go in set_eth_imc_and_md()*/
    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id = c->egIfEntry->if_stats_id;
    adk_md_p->outerIPVersion = ADK_IPV4_VER;
    /* Set outer ip stats id */
    if( likely(IP_ULI_SENDER_IPLLI != scratch_pad_p->pktOrigin))
    {
        adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_OUTER_IP_STATS].stats_id = scratch_pad_p->flow_cache_entry->outerSipStatsId;
    }

    /*     F R A G M E N T A T I O N   A N D   U D P   C H E C K S U M     */

    /* Route MTU gets precedence. Hence, use it if not zero (else use interface mtu).
     * mtu is set to route-mtu from first route lookup.
     */
    if(0 == c->mtu)
        c->mtu = c->egIfEntry->mtu;

    if(likely(rte_be_to_cpu_16(c->ipHdr->total_length) <= c->mtu))
    {
        ADK_IP_DP_LOG_INFO_DEBUG("Not fragmenting. mtu=%d, pkt-size=%d\n", c->mtu, (ADK_TASK_PDU_SIZE(c->ph_task) - ADK_MD_GET_IP_OFF(c->ph_task)) );
        uli_handle_udp_cksum_gen(c, udpHdr);
        /* --  S E T   M E T A D A T A -- */
	    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(c->ph_task));
	    if(1 == ADK_SYS_IPSEC_BYPASS_GET)
	        set_eth_imc_and_md(c, 1);
	    else
	        set_eth_imc_and_md(c, ipuliImc->frag_count);
        adk_md_p->outerIpPktLen = rte_be_to_cpu_16(c->ipHdr->total_length); 

    } /* fragmentation not needed */
    else
    {
        return uli_handle_fragmentation(c);
    } /* fragmentation needed */

    return ADK_SUCCESS;
} /* uli_pass2_fastpath */


static adk_status_t uli_pass2(
    common_params_per_run_t *c
)
{
	uint8_t *pdu;
	ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
	c->ipHdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_IP_OFF(c->ph_task));
	struct udp_hdr *udpHdr;
	adk_status_t ret;

	/* Offloading outer cksum as well */
	c->ipHdr->hdr_checksum = 0;
    if(likely( (ADK_IP_PROTOCOL_UDP==c->ipHdr->next_proto_id) || (IP_FILTER_PROTOCOL_UDP_LITE==c->ipHdr->next_proto_id) ))
        udpHdr = (struct udp_hdr *)(void *)(pdu + IP_HDR_LEN);
    else
    {    udpHdr=NULL;     }

	if(likely(ADK_SUCCESS == (ret=pbr_and_route_lookup(c)) ))
	{
	    c->egIfEntry = &(ip_dp_tbl_info_p->ipInterfaceTble[c->egressIfId]);
	    c->if_stats = &(ip_dp_tbl_info_p->ipInterfaceStatsTble[c->coreIndex][c->egIfEntry->if_stats_id]);

	    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(c->ph_task));

	    if(unlikely(IP_ULI_SENDER_IPLLI == ipuliImc->pktOrigin && c->egIfEntry->routingDisabled))
	        return proc_forwarding_disabled(c);

	    if(unlikely(ADK_IP_DP_INTF_TYPE_ETH != c->egIfEntry->ifType) )
	    {/* TO DO: remove if counter not needed */
	        ADK_IP_DP_LOG_INFO("IP interface not created. Dropping packet!");
	        ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutInvalidIfDrops++;
	        return ADK_E_DROP_TASK;
	    }

	    ret = uli_egress_qos(c);
	    if(unlikely(ADK_SUCCESS != ret))
	        return ret;

	    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
        if (unlikely((c->ipHdr->dst_addr & 0x000000f0) == 0xE0)) {
            adk_md_p->isIpMcastPkt = true;
        }else if (unlikely(c->ipHdr->dst_addr == 0xFFFFFFFF)) {
            adk_md_p->isIpBcastPkt = true;
        }else{
            adk_md_p->isIpMcastPkt = false;
            adk_md_p->isIpBcastPkt = false;
        }

        adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id =
                c->egIfEntry->if_stats_id;
        ret=set_outer_addr_stats_id(c);
	    if(unlikely(ADK_SUCCESS != ret))
	        return ret;

	    /* Route MTU gets precedence. Hence, use it if not zero (else use interface mtu).
	     * mtu is set to route-mtu from first route lookup.
	     */
	    if(0 == c->mtu)
	        c->mtu = c->egIfEntry->mtu;


	    /*     F R A G M E N T A T I O N   A N D   U D P   C H E C K S U M     */

	    if(rte_be_to_cpu_16(c->ipHdr->total_length) > c->mtu)
	    {
	        ret = uli_handle_fragmentation(c);
	        if(unlikely(ADK_SUCCESS != ret))
	            return ret;
	    } /* fragmentation needed */
	    else
	    {
	        ADK_IP_DP_LOG_INFO_DEBUG("Not fragmenting. mtu=%d, pkt-size=%d\n", c->mtu, (ADK_TASK_PDU_SIZE(c->ph_task) - ADK_MD_GET_IP_OFF(c->ph_task)) );
	        uli_handle_udp_cksum_gen(c, udpHdr);
	        /* --  S E T   M E T A D A T A -- */
	        set_eth_imc_and_md(c, ipuliImc->frag_count);
            adk_md_p->outerIpPktLen = rte_be_to_cpu_16(c->ipHdr->total_length); 
	    } /* fragmentation not needed */
	} /* pbr_and_route_lookup success */
	else
	    return ret; /* pbr_and_route_lookup failed */


	return ADK_SUCCESS;
} /* uli_pass2 */


static inline adk_status_t uli_single_pass(
    common_params_per_run_t *c
)
{
    adk_status_t ret;
    uint8_t *pdu;
    adk_metadata_t * adk_md_p= ADK_METADATA_PTR(c->ph_task);
    adk_md_p->saID = ADK_IPSEC_INVALID_SA_ID;
    ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *)&(ADK_MD_GET_IMC_PARAM_START(c->ph_task));
    struct udp_hdr *udpHdr;

    scratch_pad_p->ip_version = ADK_IPV4_VER;
    scratch_pad_p->ip_offset = 0;
    scratch_pad_p->pktOrigin = ipuliImc->pktOrigin;

    c->ipHdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_IP_OFF(c->ph_task));

    if(likely( IP_HDR_LEN == (((uint8_t *)c->ipHdr)[0] & 0x0f) * 4))  // l3_len = IHL*4
    {
        if(likely(adk_ntohs(c->ipHdr->total_length) <= ADK_TASK_PDU_SIZE(c->ph_task)))
        {
            if(likely( (ADK_IP_PROTOCOL_UDP==c->ipHdr->next_proto_id) || (IP_FILTER_PROTOCOL_UDP_LITE==c->ipHdr->next_proto_id) ))
                udpHdr = (struct udp_hdr *)(void *)(pdu + IP_HDR_LEN);
            else
            {
                udpHdr=NULL;
                /* When ipsec module is bypassed and protocol is not udp, don't calculate UDP checksum in HW */
                c->ph_task->ol_flags &= ~PKT_TX_UDP_CKSUM;
            }

            if(IP_ULI_SENDER_IPLLI != scratch_pad_p->pktOrigin)
                ip_inc_l4_out_stats(c->coreIndex, c->ipHdr->next_proto_id, rte_be_to_cpu_16(c->ipHdr->total_length) - IP_HDR_LEN);

            if (likely(/* TO DO: remove udp header */udpHdr && IP_ULI_SENDER_EO == scratch_pad_p->pktOrigin) )
            { /* We are using flow-cache only for perf case right now. i.e. UDP and EO originated traffic */
                uint32_t  hash_key;
                ret = adk_sys_dp_uli_flow_cache_get((adk_ip_hdr_t *)c->ipHdr, ADK_IPV4, c->ipHdr->next_proto_id,
                        udpHdr->src_port, udpHdr->dst_port, &scratch_pad_p->flow_cache_entry, &hash_key);

                if (unlikely(ADK_E_FLOW_CACHE_ALLOC_FAILED == ret))
                {
                    scratch_pad_p->flow_cache_entry = NULL;
                }

                FC_ALLOCED(ADK_SYS_DP_FLOW_CACHE_SET_PARAM(scratch_pad_p->flow_cache_entry->nextHdr, c->ipHdr->next_proto_id));

                if (likely(ADK_E_FLOW_CACHE_PKT_FAST_PATH == ret))
                {
                    return uli_pass2_fastpath(c);
                }
            } /* l4 proto=udp.. fc get */


            if(likely(ADK_SUCCESS == (ret=pbr_and_route_lookup(c)) ))
            {
                c->egIfEntry = &(ip_dp_tbl_info_p->ipInterfaceTble[c->egressIfId]);
                c->if_stats = &(ip_dp_tbl_info_p->ipInterfaceStatsTble[c->coreIndex][c->egIfEntry->if_stats_id]);

                if(unlikely(IP_ULI_SENDER_IPLLI == ipuliImc->pktOrigin && c->egIfEntry->routingDisabled))
                    return proc_forwarding_disabled(c);

                if(unlikely(ADK_IP_DP_INTF_TYPE_ETH != c->egIfEntry->ifType) )
                {/* TO DO: remove if counter not needed */
                    ADK_IP_DP_LOG_INFO("IP interface not created. Dropping packet!");
                    ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutInvalidIfDrops++;
                    return ADK_E_DROP_TASK;
                }

                ret = uli_egress_qos(c);
                if(unlikely(ADK_SUCCESS != ret))
                    return ret;

                ret=set_outer_addr_stats_id(c);
                if(unlikely(ADK_SUCCESS != ret))
                    return ret;

                /* Route MTU gets precedence. Hence, use it if not zero (else use interface mtu).
                 * mtu is set to route-mtu from first route lookup.
                 */
                if(0 == c->mtu)
                    c->mtu = c->egIfEntry->mtu;

                /*     F R A G M E N T A T I O N   A N D   U D P   C H E C K S U M     */

                if(rte_be_to_cpu_16(c->ipHdr->total_length) > c->mtu)
                {
                    ret = uli_handle_fragmentation(c);
                    if(unlikely(ADK_SUCCESS != ret))
                        return ret;
                } /* fragmentation needed */
                else
                {
                    ADK_IP_DP_LOG_INFO_DEBUG("Not fragmenting. mtu=%d, pkt-size=%d\n", c->mtu, (ADK_TASK_PDU_SIZE(c->ph_task) - ADK_MD_GET_IP_OFF(c->ph_task)) );
                    /* --  S E T   M E T A D A T A -- */
                    uli_handle_udp_cksum_gen(c, udpHdr);
                    set_eth_imc_and_md(c,  1);
                } /* fragmentation not needed */

            } /* pbr_and_route_lookup success */
            else
                return ret; /* pbr_and_route_lookup failed */
        } /* ip packet len < total pdu len */
        else
        {
            ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutTruncatedPktDrops++;
            return ADK_E_DROP_TASK;
        }
    } /* valid ip header length */
    else
    {
        ip_dp_tbl_info_p->ipGlobalStatsTable[c->coreIndex].ipOutInvalidIpHdrLenDrops++;
        return ADK_E_DROP_TASK;
    } /* invalid ip header length */

    return ADK_SUCCESS;
} /* uli_single_pass */


static inline void uli_single_entry_logs(
    common_params_per_run_t *c,
    adk_ip_dp_uli_imc_data_t *ipuliImc
)
{
#ifdef ADK_DEBUG
    uint8_t *pdu;
    ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
    ADK_IP_DP_LOG_INFO_DEBUG("IP ULI_single entry::\n");
    ADK_IP_DP_LOG_INFO_DEBUG("Pdu:%p,PduSize:%d, Task:%p,\nsender:%d, ip-offset=%d, l4-offset=%d, udp-cksum-flag=%llx\n",
            pdu, ADK_TASK_PDU_SIZE(c->ph_task), c->ph_task, ipuliImc->sender, ADK_MD_GET_IP_OFF(c->ph_task), ADK_MD_GET_L4_OFF(c->ph_task), c->ph_task->ol_flags & PKT_TX_UDP_CKSUM);
	c->ipHdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_IP_OFF(c->ph_task));
    ADK_IP_DP_LOG_IP_HDR(c->ipHdr);
#else
    (void)c;
    (void)ipuliImc;
#endif

} /* uli_single_entry_logs */

static inline void uli_single_exit_logs(
    common_params_per_run_t *c,
    adk_ip_dp_uli_imc_data_t *ipuliImc
)
{
#ifdef ADK_DEBUG
    uint8_t *pdu;
    ADK_PDU_PTR_GET(c->ph_task, pdu, uint8_t *);
    ADK_IP_DP_LOG_INFO_DEBUG("IP ULI_single exit::\n");
    ADK_IP_DP_LOG_INFO_DEBUG("Pdu:%p,PduSize:%d, Task:%p,\nsender:%d, ip-offset=%d, l4-offset=%d, udp-cksum-flag=%llx\n",
            pdu, ADK_TASK_PDU_SIZE(c->ph_task), c->ph_task, ipuliImc->sender, ADK_MD_GET_IP_OFF(c->ph_task), ADK_MD_GET_L4_OFF(c->ph_task), c->ph_task->ol_flags & PKT_TX_UDP_CKSUM);
    c->ipHdr = (struct ipv4_hdr *)(pdu + ADK_MD_GET_IP_OFF(c->ph_task));
    ADK_IP_DP_LOG_IP_HDR(c->ipHdr);
#else
    (void)c;
    (void)ipuliImc;
#endif

} /* uli_single_entry_logs */

static adk_status_t adk_ip_dp_pkt_proc_uli_single(
    void **task_p,
    int coreIndex,
    uint8_t *isNextEnet
)
{
    adk_status_t ret;
    common_params_per_run_t *c = &perCore[coreIndex];
    c->ph_task = *task_p;
    c->coreIndex = coreIndex;

    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(c->ph_task));;

    uli_single_entry_logs(c, ipuliImc);

    if(likely(false == ADK_SYS_IPSEC_BYPASS_GET))
    {
        if(IP_ULI_SENDER_EO == ipuliImc->sender || IP_ULI_SENDER_NETD == ipuliImc->sender)
        {
            ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_RX_TS_1();
            ret = uli_pass1(c);
            ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_TX_TS_1();
        }
        else if(IP_ULI_SENDER_IPSEC == ipuliImc->sender)
        {
            ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_RX_TS_2();
            INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutRcvdFromIPSecPkts, 1);
            if (ADK_E_FLOW_CACHE_PKT_FAST_PATH == (ADK_SIF_DP_GET_SCRATCHPAD_PTR(coreIndex))->pktPath)
                ret = uli_pass2_fastpath(c);
            else
                ret = uli_pass2(c);
            ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_TX_TS_2();
        }
        else
        {
            INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutUnknownSender, 1);
            ret = ADK_E_DROP_TASK;
        }
    }
    else
    {
        ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_RX_TS_2();
        ret = uli_single_pass(c);
        ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_TX_TS_2();
    }


    uli_single_exit_logs(c, ipuliImc);
    *isNextEnet = c->isNextEnet;
    return ret;
} /* adk_ip_dp_pkt_proc_uli_single */



/* Set default values. Reset fields that are only set by code in certain packet path and avoid incorrect processing due to garbage from previous usage of this mbuf */
static inline void reset_mdata_for_orig(
        adk_pktio_task_t *ph_task,
         __attribute__((unused)) int coreIndex)
{
    struct ipv4_hdr *ipHdr;
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(ph_task);

    /* 1. Reset next_frag pointer */
    ADK_MD_GET_NEXT_FRAG(ph_task) = NULL;

    /* 2. Reset mdata stats_ids */
    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id = ADK_SYS_INVALID_MDATA_ID;
    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_OUTER_IP_STATS].stats_id = ADK_SYS_INVALID_MDATA_ID;
    adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id = ADK_SYS_INVALID_MDATA_ID;
    /* 3. Reset ip-version field */
    adk_md_p->innerIPVersion = 0;
    adk_md_p->outerIPVersion = 0;

    ADK_PDU_OFFSET_PTR_GET(ph_task, ipHdr, struct ipv4_hdr *, adk_md_p->ip_offset);

    /* 4. Set OL flags */
    ph_task->ol_flags |= (PKT_TX_IPV4 | PKT_TX_IP_CKSUM);

    /* 5. Set l3_len which is basically l3 header length */
    ph_task->l3_len = IP_HDR_LEN;

    /* 6. Reset checksum field in packet. Otherwise, HLP calculates wrong checksum */
    ipHdr->hdr_checksum = 0;

    /* 7. Set l4 offset field in adk metadata */
    ADK_MD_SET_L4_OFF(ph_task, (adk_md_p->ip_offset + (ipHdr->version_ihl & 0x0f) * 4));

    /* 8. Reset imc metadata field */
    memset(&(ADK_MD_GET_IMC_PARAM_START(ph_task)), 0, sizeof(adk_ip_dp_uli_imc_data_t));

} /* reset_mdata_for_orig */


/*!
 * =============================================================
 * adk_ip_dp_pkt_proc_netd()
 *
 * ADK entry point for packets coming from NETD
 * =============================================================
 */
adk_status_t adk_ip_dp_pkt_proc_netd (void **task_p,
          __attribute__((unused)) uint8_t task_count,
          __attribute__((unused)) uint32_t task_mask,
          __attribute__((unused)) void *ring,
          __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
          int coreIndex)
{
    adk_pktio_task_t *ph_task = (adk_pktio_task_t *)(*task_p);
    adk_ip_dp_uli_imc_data_t *ipuliImc;

    ADK_IP_DP_LOG_INFO_DEBUG("adk_ip_dp_pkt_proc_netd entered. PduSize:%d, Task:%p \n", ADK_TASK_PDU_SIZE(ph_task), ph_task);
    if (0 != ADK_MD_GET_IP_OFF(ph_task))
    {
        uint8_t * pdu;
        ADK_PDU_PTR_GET(ph_task, pdu, uint8_t *);
        ADK_PDU_HEAD_PULL(ph_task, pdu, (uint8_t *), ADK_MD_GET_IP_OFF(ph_task));
        ADK_MD_SET_IP_OFF(ph_task, 0);
    }

    reset_mdata_for_orig(ph_task, coreIndex);

    /* If origin=netd and global-flag=true -> generate udp checksum */
    if(ADK_SYS_UDP_GEN_CONFIG_GET)
    {
        ph_task->ol_flags |= PKT_TX_UDP_CKSUM;
    }
    else
    {
        ph_task->ol_flags &= ~PKT_TX_UDP_CKSUM; /*Explicitly disabling CKSUM HW offload */
    }

    /* Must be done after reset_mdata_for_orig() as it memsets this memory to 0 */
    ipuliImc = (adk_ip_dp_uli_imc_data_t *) &(ADK_MD_GET_IMC_PARAM_START(ph_task));
    ipuliImc->sender = IP_ULI_SENDER_NETD;
    ipuliImc->pktOrigin = IP_ULI_SENDER_NETD;

    INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[coreIndex].UipOutNetDOrigPkts, 1);
    return adk_ip_dp_pkt_proc_uli(task_p, 1, 1, NULL, NULL, coreIndex);
} /* adk_ip_dp_pkt_proc_netd */


/*!
 * =============================================================
 * adk_ip_dp_pkt_proc_em_uli()
 *
 * ADK entry point for packets coming from EM
 * =============================================================
 */
adk_status_t adk_ip_dp_pkt_proc_em_uli(void ** task)
{
    ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_RX_EM_INIT_TS();
    adk_pktio_task_t *ph_task = (adk_pktio_task_t *)(*task);

    ADK_MD_SET_IP_OFF(ph_task, 0);
    reset_mdata_for_orig(ph_task, ADK_LCORE_INDEX);

    adk_ip_dp_uli_imc_data_t *ipuliImc = (adk_ip_dp_uli_imc_data_t *)&(ADK_MD_GET_IMC_PARAM_START(ph_task));
    /* must be done after reset_mdata_for_orig() as it memsets this memory to 0 */
    ipuliImc->sender = IP_ULI_SENDER_EO;
    ipuliImc->pktOrigin = IP_ULI_SENDER_EO;

    ADK_MD_GET_TS_MSG_TYPE(ph_task) = ADK_NETD_TS_NONE;
    //ADK_METADATA_PARAM(ph_task, linux_intf_id) = 0;
    ADK_METADATA_PARAM(ph_task, ts_msg_type) = 0;
    ADK_METADATA_PARAM(ph_task, ts_correlator) = 0;


    /* If origin=eo and per-packet-flag=true -> generate udp checksum */
    /* TO DO: Set adk md ip offset from adk_sys_event_metadata->ip_offset */
    if(likely(ADK_EM_UDP_GEN_CONFIG_GET(ph_task)))
        ph_task->ol_flags |= PKT_TX_UDP_CKSUM;
    else  /*Explicitly disabling CKSUM HW offload */
        ph_task->ol_flags &= ~PKT_TX_UDP_CKSUM;

    ADK_DP_MEASURE_BLOCK_CYCLE_IP_ULI_TX_EM_INIT_TS();
    INC_IP_UTIL_CNTR(ip_dp_tbl_info_p->ipGlobalStatsTable[ADK_LCORE_INDEX].UipOutEOOrigPkts, 1);
    return adk_ip_dp_pkt_proc_uli(task, 1, 1, NULL, NULL, ADK_LCORE_INDEX);
} /* adk_ip_dp_pkt_proc_em_uli */



void adk_ip_dp_egress_stats(
    struct rte_mbuf *pkt, 
    void *data) 
{
    uint16_t egress_stats_id=0, outer_addr_stats_id=0, inner_addr_stats_id=0;

    adk_egress_user_param_t *params = NULL;
    adk_ip_dp_intf_stats_t *if_stats = NULL;
    adk_ip_dp_ip_addr_stats_t *inner_ip_addr_stats = NULL;
    adk_ip_dp_ip_addr_stats_t *outer_ip_addr_stats = NULL;

    params = (adk_egress_user_param_t *)data;

    ADK_IP_DP_LOG_INFO_DEBUG("### adk_ip_dp_egress_stats: is_xmit_cb:%u, stats_mask:0x%x\n", params->xmit_cb, params->stats_mask);

    if (params->stats_mask & ADK_STATS_IP_INTERFACE_MASK)
    {
        egress_stats_id = ADK_METADATA_PTR(pkt)->egress_mdata_stats[ADK_STATS_MDATA_ID_INTERFACE_STATS].stats_id;
        if_stats = &(ip_dp_tbl_info_p->ipInterfaceStatsTble[ADK_LCORE_INDEX][egress_stats_id]);
    }
    if (params->stats_mask & ADK_STATS_INNER_IP_ADDRESS_MASK)
    {
        inner_addr_stats_id = ADK_METADATA_PTR(pkt)->egress_mdata_stats[ADK_STATS_MDATA_ID_INNER_IP_STATS].stats_id;
        inner_ip_addr_stats = &(ip_dp_tbl_info_p->ipAddrStatsTble[ADK_LCORE_INDEX][inner_addr_stats_id]);
    }
    if (params->stats_mask & ADK_STATS_OUTER_IP_ADDRESS_MASK)
    {
        outer_addr_stats_id = ADK_METADATA_PTR(pkt)->egress_mdata_stats[ADK_STATS_MDATA_ID_OUTER_IP_STATS].stats_id;
        outer_ip_addr_stats = &(ip_dp_tbl_info_p->ipAddrStatsTble[ADK_LCORE_INDEX][outer_addr_stats_id]);
    }

    if(NULL == if_stats && NULL == inner_ip_addr_stats && NULL == outer_ip_addr_stats)
    {
        ADK_IP_DP_LOG_ERROR("Invalid counter type identifier\n");
        ADK_IP_DP_LOG_INFO_DEBUG("id=%d", params->stats_mask);
        return ;
    }

    /* E N Q U E U E   S T A T S*/
    if (TRUE != params->xmit_cb)
    {
        if(rte_pktmbuf_tx_enqueue_status(pkt) == RTE_MBUF_TX_ENQUEUE_SUCCESS)
        {
            ADK_IP_DP_LOG_INFO_DEBUG("### Enqueue Success\n");
            if(if_stats)
            {
                inc_if_out_stats(ADK_METADATA_PTR(pkt), if_stats, (ADK_METADATA_PTR(pkt)->outerIpPktLen));
            }
            if (inner_ip_addr_stats) {
                inner_ip_addr_stats->ipOutPkts++;
                inner_ip_addr_stats->ipOutOctets += ADK_METADATA_PTR(pkt)->innerIpPktLen;
            }
            if (outer_ip_addr_stats) {
                outer_ip_addr_stats->ipOutPkts++;
                outer_ip_addr_stats->ipOutOctets += ADK_METADATA_PTR(pkt)->outerIpPktLen;
            }
        }
        else
        {
            ADK_IP_DP_LOG_INFO("### Enqueue Drop\n");
            if(if_stats)
            {
                if_stats->ifOutDroppedPkts++;
                if_stats->ifOutDroppedOctets += ADK_METADATA_PTR(pkt)->outerIpPktLen;
            }
        }
    }
    /*T R A N S M I T   S T A T S*/
    else
    {
        if(rte_pktmbuf_tx_cmpltn_status(pkt) == RTE_MBUF_TX_CMPLTN_SUCCESS)
        {
            /*Transmit queue might be removed later. Hence incrementing the
             *TX packets/octets only in the Enqueue successfull case
             */
            ADK_IP_DP_LOG_INFO("### Xmit Success\n");

        }
        else
        {
            ADK_IP_DP_LOG_INFO("### Xmit Drop\n");
             if(if_stats)
            {
                if_stats->ifOutDroppedPkts++;
                if_stats->ifOutDroppedOctets += ADK_METADATA_PTR(pkt)->outerIpPktLen;
            }
       }
    }

    return;
} /* adk_ip_dp_egress_stats */

