/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_ipsec_struct.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipsec  IPsec Module
 * @{
 *
 * @brief
 *
 * This file contains all return error codes, structure definitions
 * required by the IPsec ADK
 *
 ***********************************************************************/

#ifndef _ADK_IPSEC_STRUCT_H_
#define _ADK_IPSEC_STRUCT_H_

#include <rte_memory.h>
#include <rte_crypto.h>
#include <rte_cryptodev.h>

#include "adk_ip.h"

/*! Value to indicate DSCP value wildcard in SPD table */
#define ADK_IPSEC_DSCP_ANY	0xFF

/*!
 * @brief Specifies the list of anti-replay wnidow
 * sizes supported by the SPP. They are
 * 64, 128, 256, 512, 1024.
 */
typedef enum adk_ipsec_arw_size_e
{
    ADK_IPSEC_ARW_MIN  = 64,
    ADK_IPSEC_ARW_64   = 64,
    ADK_IPSEC_ARW_128  = 128,
    ADK_IPSEC_ARW_256  = 256,
    ADK_IPSEC_ARW_512  = 512,
    ADK_IPSEC_ARW_1024 = 1024,
    ADK_IPSEC_ARW_4096 = 4096,
    ADK_IPSEC_ARW_MAX  = 4096
} adk_ipsec_arw_size_t;

/*!
 * @brief IPSec Encryption algorithms; the enum values are
 *        assigned to match the IKEv2.
 *
 * See RFC 4306 and also: <br>
 * http://www.iana.org/assignments/ikev2-parameters
 */

typedef enum adk_ipsec_encr_e {
    ADK_IPSEC_ENCR_DES_CBC       = 2,   /* 64-bit key */
    ADK_IPSEC_ENCR_3DES_CBC      = 3,
    ADK_IPSEC_ENCR_NULL          = 11,
    ADK_IPSEC_ENCR_AES_CBC       = 12,
    ADK_IPSEC_ENCR_AES_CTR       = 13,
    ADK_IPSEC_ENCR_AES_CCM_ICV8  = 14,
    ADK_IPSEC_ENCR_AES_CCM_ICV12 = 15,
    ADK_IPSEC_ENCR_AES_CCM_ICV16 = 16,
    ADK_IPSEC_ENCR_AES_GCM_ICV8  = 18,
    ADK_IPSEC_ENCR_AES_GCM_ICV12 = 19,
    ADK_IPSEC_ENCR_AES_GCM_ICV16 = 20,
    ADK_IPSEC_ENCR_NULL_GMAC     = 21,
    ADK_IPSEC_ENCR_MAX
} adk_ipsec_encr_t;

/*!
 * @brief IPSec Authentication/Integrity algorithms; the enum values are
 *        assigned to match the IKEv2.
 *
 * See RFC 4306 and also: <br>
 * http://www.iana.org/assignments/ikev2-parameters
 */
 /* from RFC4868
  *
   +------------------+--------+--------+--------+--------+------------+
   |    Algorithm     | Block  | Output | Trunc. |  Key   | Algorithm  |
   |       ID         |  Size  | Length | Length | Length |   Type     |
   +==================+========+========+========+========+============+
   | HMAC-SHA-256-128 |   512  |   256  |  128   |  256   | auth/integ |
   +------------------+--------+--------+--------+--------+------------+
   | HMAC-SHA-384-192 |  1024  |   384  |  192   |  384   | auth/integ |
   +------------------+--------+--------+--------+--------+------------+
   | HMAC-SHA-512-256 |  1024  |   512  |  256   |  512   | auth/integ |
   +------------------+--------+--------+--------+--------+------------+
  */
typedef enum adk_ipsec_auth_e {
    ADK_IPSEC_AUTH_NULL          = 0,
    ADK_IPSEC_AUTH_HMAC_MD5      = 1,
    ADK_IPSEC_AUTH_HMAC_SHA1     = 2,
    ADK_IPSEC_AUTH_AES_CMAC      = 8,
    ADK_IPSEC_AUTH_AES_XCBC_MAC  = 9, /* As per RFC 3566 */
    ADK_IPSEC_AUTH_HMAC_SHA_256  = 12,
    ADK_IPSEC_AUTH_HMAC_SHA_384  = 13,
    ADK_IPSEC_AUTH_HMAC_SHA_512  = 14,
    /* As per RFC 4543, AES-GMAC in ESP is identified by an encryption
     * transform identifier - that is used to reference
     * ADK_IPSEC_ENCR_NULL_GMAC. Hence assigning temporary values to
     * ADK_IPSEC_AUTH_AES_128_GMAC, ADK_IPSEC_AUTH_AES_192_GMAC and
     * ADK_IPSEC_AUTH_AES_256_GMAC.
     */
    ADK_IPSEC_AUTH_AES_128_GMAC  = 15,
    ADK_IPSEC_AUTH_AES_192_GMAC  = 16,
    ADK_IPSEC_AUTH_AES_256_GMAC  = 17,
    ADK_IPSEC_AUTH_MAX
} adk_ipsec_auth_t;

typedef enum adk_ipsec_mode_e
{
    ADK_IPSEC_TUNNEL,
    ADK_IPSEC_TRANSPORT,
} adk_ipsec_mode_t;


typedef enum adk_ipsec_protocol_e {
    ADK_IPSEC_AH,
    ADK_IPSEC_ESP
} adk_ipsec_protocol_t;

/*!
 * @enum adk_ipsec_ip_version_t "adk_ipsec_struct.h"
 *
 * @brief The enumeration defines ip version's supported
 */
typedef enum adk_ipsec_ip_version_e {
        /*! signifies value to be used for IPv4 support */
        ADK_IPSEC_IPV4 = 4,
        /*! signifies value to be ued for IPv6 support */
        ADK_IPSEC_IPV6 = 6
} adk_ipsec_ip_version_t;

/*!
 * @enum adk_ipsec_flag_t "adk_ipsec_struct.h"
 *
 * @brief The enumeration defines flag to set TRUE/FALSE value
 */
typedef enum adk_ipsec_flag_e {
        /*! signifies value set as false w.r.t IPsec ADK */
        ADK_IPSEC_FALSE = 0,
        /*! signifies value set as true w.r.t IPsec ADK */
        ADK_IPSEC_TRUE
} adk_ipsec_flag_t;

/*!
 * @enum adk_ipsec_dont_frag_config_t "adk_ipsec_struct.h"
 *
 * @brief The enumeration defines dont frag configurations that are supported
 */
typedef enum adk_ipsec_dont_frag_config_e {
    /*! security policy dont fragment config as copy inner IP header */
    ADK_IPSEC_SP_DF_COPY_INNER_IP= 0,
    /*! security policy dont fragment bit set */
    ADK_IPSEC_SP_DF_SET,
    /*! security policy dont fragment bit clear */
    ADK_IPSEC_SP_DF_CLEAR,
    /*! security policy dont fragment as invalid */
    ADK_IPSEC_SP_DF_INVALID
} adk_ipsec_dont_frag_config_t;

/*!
 * @enum adk_ipsec_sa_update_msg_type_t "adk_ipsec_struct.h"
 *
 * @brief The enumeration defines the SA message type
 */
typedef enum adk_ipsec_sa_update_msg_type_e {
        /*! security policy SA message type as acquire */
        ADK_IPSEC_SA_ACQUIRE = 1,
        /*! security policy SA message type as expire */
        ADK_IPSEC_SA_EXPIRE,
        /*! security policy SA message max */
        ADK_IPSEC_SA_MAX,
} adk_ipsec_sa_update_msg_type_t;

/*!
 * @brief Specifies the SA direction by IPsec ADK defined in rte_crypto.h of DPDK
 */
typedef enum adk_ipsec_sa_direction_e
{
    /*! SA type w.r.t direction as egress/outbound */
    ADK_IPSEC_OUTBOUND = RTE_CRYPTO_CIPHER_OP_ENCRYPT,
    /*! SA type w.r.t direction as ingress/inbound */
    ADK_IPSEC_INBOUND = RTE_CRYPTO_CIPHER_OP_DECRYPT
} adk_ipsec_sa_direction_t;


typedef enum adk_sa_expiry_type_s
{
    ADK_HARD_LIFE_EXPIRY = 1,
    ADK_SOFT_LIFE_EXPIRY = 2,
}adk_sa_expiry_type_t;

/*!
 * @enum adk_ipsec_flow_label_config_t "adk_ipsec_struct.h"
 *
 * @brief The enumeration defines flow label configuration for tunnel IPv6 header.
 * Applicable only if SA version is IPv6.
 */
typedef enum adk_ipsec_flow_label_config_e {
/*! IPv6 header flow label should be copied from inner IPv6 header.
 *! Note that if inner packet is IPv4(i.e. policy is IPv4),
 *! flow label is set to the 20 bit value from flow_label_value.
 */
    ADK_IPSEC_SA_FLOW_LABEL_COPY_INNER_IP = 0,
 /*! IPv6 header flow label is taken from configuration i.e. application provided value*/
    ADK_IPSEC_SA_FLOW_LABEL_FROM_CONFIGURATION
} adk_ipsec_flow_label_config_t;

/*!
 * The data structure is used to configure flow label in tunnel IPv6 header configuration.
 */
typedef struct adk_ipsec_flow_label_info_s {
    /*! Flow label configuration options */
    adk_ipsec_flow_label_config_t flow_label_config;

    /*! Flow label value if flow_label_config indicates ADK_IPSEC_SA_FLOW_LABEL_FROM_CONFIGURATION
     *! Only lower 20 bits are used from this value */
    uint32_t flow_label_value;
} adk_ipsec_flow_label_info_t; /* end of struct adk_ipsec_flow_label_info_t */

/*!
 * @enum adk_ipsec_sa_mode_t "adk_ipsec_struct.h"
 *
 * @brief 
 * The enumeration defines Inline/Lookaside configuration for particular SA.
 */
typedef enum adk_ipsec_sa_mode_s {
    /*! The packet will be processed with Inline operation.
     * By default, the SA Mode is Inline.
     * In outbound processing, packet will be processed with Inline operation.
     * In inbound processing, packet will be processed with either Inline or Lookaside operation.
     */
    ADK_IPSEC_SA_MODE_INLINE = 0,
    /*! The packet will be processed with Lookaside operation in both inbound and
     * outbound processing.
     */
    ADK_IPSEC_SA_MODE_LOOKASIDE
} adk_ipsec_sa_mode_t;

/*!
 * @enum adk_ipsec_pol_action_t "adk_ipsec_struct.h"
 *
 * @brief The enumeration defines policy action
 */
typedef enum adk_ipsec_pol_action_e {
        /*! security policy action as ipsec
         * It should be ignored while configuring SPD IN policy.
         */
        ADK_IPSEC_SP_ACTION_IPSEC = 1,
        /*! security policy action as drop  */
        ADK_IPSEC_SP_ACTION_DROP,
        /*! security policy action as bypass  */
        ADK_IPSEC_SP_ACTION_BYPASS,
        ADK_IPSEC_SP_ACTION_MAX
} adk_ipsec_pol_action_t;

/*!
 * @enum adk_ipsec_cookie_values_t "adk_ipsec_struct.h"
 *
 * @brief The anumeration defines cookie values used in IPsec ADK
 */
typedef enum adk_ipsec_cookie_values_e {

          /*! IPsec Module cookie */
          ADK_IPSEC_MODULE_COOKIE = ((ADK_IPSEC_MODULE_ID << 24) + 1000),
          /*! IPsec ingress flow cookie */
          ADK_IPSEC_INGRESS_FLOW_COOKIE,
          /*! IPsec egress flow cookie */
          ADK_IPSEC_EGRESS_FLOW_COOKIE,
} adk_ipsec_cookie_values_t;


/*!
 * @brief The enumeration defines all error codes returned by IPsec ADK
 */
typedef enum adk_ipsec_return_code_e {

    /*! Successfully completed requested operation */
	ADK_IPSEC_SUCCESS = ADK_SUCCESS,
    /*! IPsec module init failed */
    ADK_E_IPSEC_INIT_FAIL,
    /*! Incorrect IPsec module handle */
    ADK_E_IPSEC_INVALID_MODULE_HANDLE,
    /*! Incorrect IPsec policy configuration*/
    ADK_E_IPSEC_INVALID_POLICY,
    /*! Incorrect IPsec policy selector configuration*/
    ADK_E_IPSEC_INVALID_SEC_POL_SEL,
    /*! Failed to get a policy in given index, no such index exists */
    ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
    /*! Failed to get policy, policy does not exist */
    ADK_E_IPSEC_POLICY_NOT_FOUND,
    /*! Failed to get security association, it does not exist */
    ADK_E_IPSEC_SA_NOT_FOUND,
    /*! NULL SA i/p entry found */
    ADK_E_IPSEC_INVALID_SA,
    /*! Failed to get number of rules from SPD structure */
	ADK_E_IPSEC_GET_NUM_SPD_RULE_FAIL,
    /*! Failed to add flow entry into association flow structure */
	ADK_E_IPSEC_FLOW_ENTRY_ADD_FAIL,
    /*! SA entry already present */
    ADK_E_IPSEC_VALID_SA_EXISTS,
    /*! Failed to delete SPD entry */
    ADK_E_IPSEC_SPD_ENTRY_DEL_FAIL,
    /*! Failed to get SPI entry from SPI table */
    ADK_E_IPSEC_SPI_TABLE_NO_ENTRY_FOUND,
    /*! Failed to delete flow entry from association flow structure */
    ADK_E_IPSEC_FLOW_ENTRY_DELETE_FAIL,
    /*! Failed to add spi entry into spi table */
    ADK_E_IPSEC_SPI_ADD_ENTRY_FAIL,
    /*! Failed to add tunnel entry into tunnel structure */
    ADK_E_IPSEC_TUNNEL_ENTRY_ADD_FAIL,
    /*! Failed to get tunnel id */
    ADK_E_IPSEC_TUNNEL_GET_FAIL,
    /*! Flow already exists */
    ADK_E_IPSEC_FLOW_ALREADY_EXISTS,
    /*! Protocol value is a must as i/p paramter for flow */
    ADK_E_IPSEC_PROTOCOL_IS_MUST,
    /*! Failed to get statistics */
    ADK_E_IPSEC_GET_STAT_CALL_FAIL,
    /*! Failed to reset statistics */
    ADK_E_IPSEC_SET_STAT_CALL_FAIL,
    /*! Audit event already enabled */
    ADK_E_IPSEC_AUDIT_ALREADY_ENABLED,
    /*! Audit event not enabled */
    ADK_E_IPSEC_AUDIT_NOT_ENABLED,
    /*! SA acquire/expire update already enabled */
    ADK_E_IPSEC_SAUPDATE_ALREADY_ENABLED,
    /*! SA acquire/expire update not enabled */
    ADK_E_IPSEC_SAUPDATE_NOT_ENABLED,
    /*! Failed to modify SPD entry */
    ADK_E_IPSEC_SPD_ENTRY_MOD_FAIL,
    /*! Invalid security policy selector parameter */
    ADK_E_IPSEC_INVALID_SP_SEL_PARAM,
    /*! Invalid input parameter */
    ADK_E_IPSEC_INVALID_INPUT_PARAM,
    /*! Invalid input configuration */
    ADK_E_IPSEC_INVALID_CONFIG,
    /*! Invalid input anti-replay window size */
    ADK_E_IPSEC_INVALID_ANTIREPLAY_WINDOW_SIZE,
    /*! Failed to allocate flow handle */
    ADK_E_IPSEC_FLOW_HANDLE_ALLOC_FAIL,
    /*! Failed to add SPD entry */
    ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL,
    /*! Failed to get audit entry */
    ADK_E_IPSEC_GET_AUDIT_ENTRY_FAIL,
    /*! Failed to add audit entry */
    ADK_E_IPSEC_AUDIT_ENTRY_ADD_FAIL,
    /*! Failed to delete audit tree entry */
    ADK_E_IPSEC_AUDIT_TREE_DELETE_FAIL,
    /*! Failed to add SA update entry */
    ADK_E_IPSEC_SA_UPDATE_ENTRY_ADD_FAIL,
    /*! Failed to delete SA update entry */
    ADK_E_IPSEC_SA_UPDATE_DELETE_FAIL,
    /*! Failed to get SA update entry */
    ADK_E_IPSEC_GET_SA_UPDATE_ENTRY_FAIL,
    /*! Failed to delete SPI entry */
    ADK_E_IPSEC_SPI_DELETE_FAIL,
    /*! SPI value is in use */
    ADK_E_IPSEC_SPI_IN_USE,
    /*! Received an invalid Event Type */
    ADK_E_IPSEC_INVALID_EVENT_TYPE_RCVD,
    /*! Invalid IP version */
    ADK_E_IPSEC_INVALID_IP_VERSION,
    /*! SPD list rules failed */
    ADK_E_IPSEC_POLICY_LIST_RULES_FAILED,
    /*! SPD ruleset resource intialization failed */
    ADK_E_IPSEC_RULESETID_UNAVAILABLE,
    /*! Failed to allocate SPD rulserid */
    ADK_E_IPSEC_RULESETID_MEMORY_UNAVAILABLE,
    /*! Failed to allocate SP in memory structure */
    ADK_E_IPSEC_SP_MEMORY_UNAVAILABLE,

    /*! Cryptodev returned errors - included from rte_crypto.h in DPDK */
    /*! Operation not yet been processed by cryptodev */
    ADK_E_IPSEC_CRYPTO_OP_STATUS_NOT_PROCESSED =
        RTE_CRYPTO_OP_STATUS_NOT_PROCESSED,
    /*! Authentication verification failed in crypto device */
    ADK_E_IPSEC_CRYPTO_OP_STATUS_AUTH_FAILED =
        RTE_CRYPTO_OP_STATUS_AUTH_FAILED,
    /*! Invalid session arguments causing symmetric operation to fail in
     * cryptodev */
    ADK_E_IPSEC_CRYPTO_OP_STATUS_INVALID_SESSION =
        RTE_CRYPTO_OP_STATUS_INVALID_SESSION,
    /*! Invalid crypto request arguments leading to crypto operation failure */
    ADK_E_IPSEC_CRYPTO_OP_STATUS_INVALID_ARGS =
        RTE_CRYPTO_OP_STATUS_INVALID_ARGS,
    /*! Cryptodev error handling operation */
    ADK_E_IPSEC_CRYPTO_OP_STATUS_ERROR = RTE_CRYPTO_OP_STATUS_ERROR,

    /*! arw invalid params */
    ADK_E_IPSEC_ARW_INVALID_PARAMS,
    /*! arw reply */
    ADK_E_IPSEC_ARW_OUTDATED_VALUE,
    /*! arw reply */
    ADK_E_IPSEC_ARW_REPLAY,
    /*! arw reply */
    ADK_E_IPSEC_ARW_NO_REPLAY,
    /* Failed to build SPD tree */
    ADK_E_IPSEC_SPD_TREE_BUILD_FAIL,
    /*! General error code */
    ADK_E_IPSEC_ERROR

} adk_ipsec_return_code_t;

/*!
 * The data strcuture captures information w.r.t IPv6 address.
 * The information contains actual IP adress and subnet mask info
 */
typedef struct adk_ipsec_ipv6_addr_info_s {
    /*! 128-bit IPv6 adress. */
    uint32_t addr[4];
    /*! 32-bit prefix value that indicates how many bits in the
     * addr field are used for lookup. Set this value to 128 for
     * IPv6 host address.
     * These values are populated from the spMppContext namespace.
     */
    uint32_t prefixWidth;
} adk_ipsec_ipv6_addr_info_t; /* end of struct adk_ipsec_ipv6_addr_info_t */


/*!
 * The data strcuture captures information w.r.t IPv6 address selector of ipsec policy.
 * The information contains IP address range or actual IP adress and subnet mask info
 */
typedef struct adk_ipsec_sel_ipv6_addr_info_s {
    /*! add_range_sel need to set 1 if selector
      * added is range based else need to set as 0 */
    bool rangeSelFlag;
    union {
      struct addr_ipv6_range_s{
          /*! 128-bit Beginning IPv6 adress. */
          uint32_t  addrBegin[4];
          /*! 128-bit End IPv6 adress. */
          uint32_t  addrEnd[4];
      }addr_range_t;

      struct addr_ipv6_subnet_s{

          /*! 128-bit IPv6 adress. */
          uint32_t addr[4];
          /*! 128-bit prefix value that indicates how many bits in the
            * addr field are used for lookup. Set this value to 128 for
            * IPv6 host address.
            */
          uint32_t prefixWidth;
      } addr_subnet_t;

  } addr;
} adk_ipsec_sel_ipv6_addr_info_t; /* end of struct adk_ipsec_sel_ipv6_addr_info_t */


/*!
 * The data strcuture captures information w.r.t IPv4 address.
 * The information contains actual IP adress and subnet mask info
 */
typedef struct adk_ipsec_ipv4_addr_info_s {
    /*! 32-bit IPv4 adress. */
    uint32_t addr;
    /*! 32-bit prefix value that indicates how many bits in the
     * addr field are used for lookup. Set this value to 32 for
     * IPv4 host address.
     * These values are populated from the spMppContext namespace.
     */
    uint32_t prefixWidth;
} adk_ipsec_ipv4_addr_info_t; /* end of struct adk_ipsec_ipv4_addr_info_t */

/*!
 * The data strcuture captures information w.r.t IPv4 address selector of ipsec policy.
 * The information contains IP address range or actual IP adress and subnet mask info
 */
typedef struct adk_ipsec_sel_ipv4_addr_info_s {

  /*! add_range_sel need to set 1 if selector
    * added is range based else need to set as 0 */
  bool rangeSelFlag;
  union {
      struct addr_ipv4_range_s{
          /*! 32-bit Beginning IPv4 adress. */
          uint32_t  addrBegin;
          /*! 32-bit End IPv4 adress. */
          uint32_t  addrEnd;
      }addr_range_t;

      struct addr_ipv4_subnet_s{

          /*! 32-bit IPv4 adress. */
          uint32_t addr;
          /*! 32-bit prefix value that indicates how many bits in the
            * addr field are used for lookup. Set this value to 32 for
            * IPv4 host address.
            */
          uint32_t prefixWidth;
      } addr_subnet_t;

  } addr;
} adk_ipsec_sel_ipv4_addr_info_t; /* end of struct adk_ipsec_sel_ipv4_addr_info_t */

/*!
 * The data strcuture captures information w.r.t IPv4 tunnel address.
 * The information contains actual IPv4 tunnel source/destination address
 */
typedef struct adk_ipsec_ipv4_tunnel_addr_s {
    /*! 32-bit IPv4 source adress */
    uint32_t srcAddr;
    /*! 32-bit IPv4 destination adress */
    uint32_t dstAddr;
} adk_ipsec_ipv4_tunnel_addr_t; /* end of struct adk_ipsec_ipv4_tunnel_addr_t */

/*!
 * The data strcuture captures information w.r.t IPv6 tunnel address.
 * The information contains actual IPv6 tunnel source/destination address
 */
typedef struct adk_ipsec_ipv6_tunnel_addr_s {
    /*! 128-bit IPv6 source adress */
    uint32_t srcAddr[4];
    /*! 128-bit IPv6 destination adress */
    uint32_t dstAddr[4];
} adk_ipsec_ipv6_tunnel_addr_t; /* end of struct adk_ipsec_ipv6_tunnel_addr_t */

/*!
 * The data strcuture captures information w.r.t Layer 4 port.
 * The information contains actual port value and port mask/range info
 */
typedef struct adk_ipsec_port_info_s {

      /*! 16-bit layer 4 port. */
      uint16_t    port;
      /*! 16-bit prefix value that indicates how many bits in the
        * port field are used for lookup. Set this value to 16 for
        * actual port value to be used.
        * For TCP and UDP protocols these values are populated from
        * the spMppContext Namespace. For all other protocols these values
        * are not populated and is set to zero.
        */
      uint16_t    prefixWidth;


}adk_ipsec_port_info_t; /* end of struct adk_ipsec_port_info_t */

/*!
 * The data strcuture captures information w.r.t Layer 4 port.
 * This data structure is used to configure port information at the time of adding the policy.
 * The information contains actual port value and port mask/range info
 */
typedef struct adk_ipsec_sel_port_info_s {

    /*!16-bit layer 4 port begin value. */
    uint16_t        portBegin;
    /*!16-bit layer 4 port end value. */
    uint16_t        portEnd;

}adk_ipsec_sel_port_info_t; /* end of struct adk_ipsec_sel_port_info_t */

/*!
* The data structure captures information w.r.t icmp type.
* This data structure is used to configure icmp type information at the time of adding the policy.
*/
typedef struct adk_ipsec_sel_icmp_type_info_s {

    /*!8-bit icmp type begin value. */
    uint8_t        typeBegin;
    /*!8-bit icmp type end value. */
    uint8_t        typeEnd;

}adk_ipsec_sel_icmp_type_info_t; /* end of struct adk_ipsec_sel_icmp_type_info_t */

/*!
* The data structure captures information w.r.t icmp code.
* This data structure is used to configure icmp code information at the time of adding the policy.
*/
typedef struct adk_ipsec_sel_icmp_code_info_s {

    /*!8-bit icmp code begin value. */
    uint8_t        codeBegin;
    /*!8-bit icmp code end value. */
    uint8_t        codeEnd;

}adk_ipsec_sel_icmp_code_info_t; /* end of struct adk_ipsec_sel_code_info_t */


/*!
 * ================================================================
 * @struct adk_ipsec_sec_pol_sel_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines security policy selector values for SPD OUT Table
 *
 * ================================================================
 */
typedef struct adk_ipsec_sec_pol_sel_s {

    /*! 8-bit field carrying IP version information */
    adk_ipsec_ip_version_t      ipVersion;

    union ip_add {

        struct ipv6_addrs {
            /*! source IPv6 address info */
            adk_ipsec_sel_ipv6_addr_info_t    srcAddr;
            /*! destination IPv6 address info */
            adk_ipsec_sel_ipv6_addr_info_t    dstAddr;
        } ipv6Addr;

        struct ipv4_addrs {
            /*! source IPv4 address info */
            adk_ipsec_sel_ipv4_addr_info_t    srcAddr;
            /*! destination IPv4 address info */
            adk_ipsec_sel_ipv4_addr_info_t    dstAddr;
        } ipv4Addr;

    } ipAddr;

    /*! 8-bit IP header protocol field */
    uint8_t                   protocol;
    union pol_l4_info {

        struct port_info {
            /*!
             * Layer 4 source port info. This field can be specified
             * only if protocol is UDP or TCP or SCTP
             */
            adk_ipsec_sel_port_info_t         srcPort;
            /*!
             * Layer 4 destination port info. This field can be specified
             * only if protocol is UDP or TCP or SCTP
             */
            adk_ipsec_sel_port_info_t         dstPort;
        }portInfo;

        struct icmp_info {
            /*! icmp type info */
            adk_ipsec_sel_icmp_type_info_t       type;
            /*! icmp code info */
            adk_ipsec_sel_icmp_code_info_t       code;
        } icmpInfo;
    }l4Info;

    /*! 8-bit dscp value. Specify ADK_IPSEC_DSCP_ANY to wildcard dscp field */
    uint8_t                   dscpVal;

} adk_ipsec_sec_pol_sel_t; /* end of struct adk_ipsec_sec_pol_sel_t */

/*!
 * ================================================================
 * @struct adk_ipsec_sec_pol_in_sel_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines security policy selector values for SPD IN Table
 *
 * ================================================================
 */
typedef struct adk_ipsec_sec_pol_in_sel_s {

    /*! 8-bit field carrying IP version information */
    adk_ipsec_ip_version_t      ipVersion;

    union ip_add_spd_in {

        struct ipv6_addrs_spd_in {
            /*! source IPv6 address info */
            adk_ipsec_sel_ipv6_addr_info_t    srcAddr;
            /*! destination IPv6 address info */
            adk_ipsec_sel_ipv6_addr_info_t    dstAddr;
        } ipv6Addr;

        struct ipv4_addrs_spd_in {
            /*! source IPv4 address info */
            adk_ipsec_sel_ipv4_addr_info_t    srcAddr;
            /*! destination IPv4 address info */
            adk_ipsec_sel_ipv4_addr_info_t    dstAddr;
        } ipv4Addr;

    } ipAddr;

    /*! 8-bit IP header protocol field */
    uint8_t                   protocol;
    union pol_l4_info_spd_in {

        struct port_info_spd_in {
            /*!
             * Layer 4 source port info. This field can be specified
             * only if protocol is UDP or TCP or SCTP
             */
            adk_ipsec_sel_port_info_t         srcPort;
            /*!
             * Layer 4 destination port info. This field can be specified
             * only if protocol is UDP or TCP or SCTP
             */
            adk_ipsec_sel_port_info_t         dstPort;
        }portInfo;

        struct icmp_info_spd_in {
            /*! icmp type info */
            adk_ipsec_sel_icmp_type_info_t       type;
            /*! icmp code info */
            adk_ipsec_sel_icmp_code_info_t       code;
        } icmpInfo;
    }l4Info;
} adk_ipsec_sec_pol_in_sel_t; /* end of struct adk_ipsec_sec_pol_in_sel_t */

/*!
 * ================================================================
 * @struct adk_ipsec_sec_pol_attr_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines secutiy policy selector attributes and other
 * relevant configuration
 *
 * ================================================================
 */
typedef struct adk_ipsec_sec_pol_attr_s {
    /*! 2-bit policy action field */
    adk_ipsec_pol_action_t         action;
    /*! 2-bit dont fragment config */
    adk_ipsec_dont_frag_config_t	configDF;
    /*! 1-bit pre fragment enable config */
    bool                    preFragEnable;
    /*! 1-bit tunnel mode config */
    bool                    tunnelMode;
    /*1 bit security protocol flag */
    bool                    espProto;
    /*! 1-bit which tells single SA will be used for pre-fragmented traffic.
     *
     * When set means, only one SA will be used for initial fragment and non-initial
     * fragments. SA associated with initial fragment will be used for non-intial
     * fragments as well.
     *
     * When not set means, different SA's will be used for initial and non-initial
     * fragments. This implies that separate policies should be configured for
     * initial and non-initial fragments.
     *    -- In case of IPv4/IPv6, for initial fragments complete 6 tuple (SIP, DIP,SP,DP,protocol and DSCP)
     *    is used as lookup parameters in policy.
     *    -- In case of IPv4, for non-initial fragments SIP, DIP, Protocol and DSCP value
     *    is used for policy lookup.
     *    -- In case of IPv6, for non-initial fragments SIP, DIP and DSCP value is used
     *    for policy lookup.
     *
     */
    uint16_t    useSingleSAInPrefrag;

    /*! 16-bit path MTU config value which must includes extra bytes that will be added
     * in the packet after pre-fragmentation and IPsec outbound processing in case of tunnel mode.
     * If path MTU is configured as 0, it indicates no pre-fragmentation even if prefrag is enabled.
     * Range for path MTU is 1 to 16383.
     */
    uint16_t    pathMTU;

    /* No SA discard in outbond direction */
    uint32_t    saNotFoundErrDrop;
    /*! 16-bit unique identifier for the set of multi-selector policy rules that the application sees 
     * as single policy but are expanded in the ADK fast path policy table.
     * Every user level policy should have unique ID.
     * If the user level policy has multiple selectors then application will expand them into a list of
     * ADK policy rules. In multi-selector case, each such rule will have same spIDEx.
     */
    /* TODO: Currently support is not yet added. */ 
    uint16_t    spIDEx;

} adk_ipsec_sec_pol_attr_t; /* end of struct adk_ipsec_sec_pol_attr_t */

/*!
 * ================================================================
 * @struct adk_ipsec_sec_pol_in_attr_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines security policy selector attributes for SPD IN.
 *
 * ================================================================
 */
typedef struct adk_ipsec_sec_pol_in_attr_s {
    /*! 2-bit policy action field */
    adk_ipsec_pol_action_t         action;
} adk_ipsec_sec_pol_in_attr_t; /* end of struct adk_ipsec_sec_pol_in_attr_t */

/*!
 * ================================================================
 * @struct adk_ipsec_sec_pol_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines security policy selector and it's attributes for SDP OUT
 *
 * ================================================================
 */
typedef struct adk_ipsec_sec_pol_s {
    /*! policy selector info */
	adk_ipsec_sec_pol_sel_t		secPolSel;
    /*! policy attributes info */
	adk_ipsec_sec_pol_attr_t	secPolAttr;
} adk_ipsec_sec_pol_t; /* end of struct adk_ipsec_sec_pol_t */

/*!
 * ================================================================
 * @struct adk_ipsec_sec_pol_in_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines security policy selector and it's attributes for SDP IN
 *
 * ================================================================
 */

typedef struct adk_ipsec_sec_pol_in_s {
    /*! policy selector info */
    adk_ipsec_sec_pol_in_sel_t     secPolSel;
    /*! policy attributes info */
    adk_ipsec_sec_pol_in_attr_t    secPolAttr;
} adk_ipsec_sec_pol_in_t; /* end of struct adk_ipsec_sec_pol_in_t */

/*!
 * ================================================================
 * @struct adk_natt_ipsec_context_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines NATT related security context for secutiry
 * association
 *
 * ================================================================
 */
typedef struct adk_natt_ipsec_context_s {
    /*! 16-bit l4 NATT source port */
    uint16_t        NATTSrcPort;
    /*! 16-bit l4 NATT destination port */
    uint16_t        NATTDstPort;
    /*! 32-bit orig src IP address */
    uint32_t        origSrcAddr;
    /*! 32-bit orig dst IP address */
    uint32_t        origDstAddr;
} adk_natt_ipsec_context_t; /* end of struct adk_natt_ipsec_context_t */


/*!
 * ================================================================
 * @struct adk_crypto_ipsec_context_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief used to store security context used by the QAT crypto engine and IPsec ADK.
 *
 * ================================================================
 */

typedef struct adk_crypto_ipsec_context_s
{
    /*
     * If "lifetimesEnabled" is true, this value gives a soft
     * SA lifetime (in bytes) as defined by the RFC4301.
     * Once the Soft Lifetime expires, the SPP will set a flag
     * in the Output parameters telling the downstream engine to
     * trigger the auto-negotiation. If "lifetimesEnabled" is false,
     * this is a simple byte down-counter that can be used for statistics.
     */
    uint64_t             softByteLifetime;

    /*
     * If "lifetimesEnabled" is true, this value gives a hard SA
     * lifetime (in bytes) as defined by RFC4301. Once the Hard Lifetime
     * expires, no packets for that SA will be processed by the SPP.
     * If "lifetimesEnabled" is false, this is a simple byte
     * down-counter that can be used for statistics.
     */
    uint64_t             hardByteLifetime;
    adk_ipsec_sa_direction_t direction;
    /*
     * @brief Selects Tunnel or Transport Mode.
     */
    adk_ipsec_mode_t         mode;
    /*
     * @brief Selects ESP or AH protocol.
     */
    adk_ipsec_protocol_t     protocol;

    adk_ipsec_encr_t         encr;
    adk_byte_string_t        encrKey;
    adk_byte_string_t        salt;
    adk_ipsec_auth_t         auth;
    adk_byte_string_t        authKey;

    /*
     * Security Parameters Index (SPI) (as defined by RFC4301).
     * It is to be inserted into the packet for OUTBOUND processing.
     * The value is ignored for the INBOUND.
     */
    uint32_t             spi;

    /*
     * @brief Specifies whether the context uses 64-bit extended
     * sequence number or 32-bit sequence numbers (see RFC4302/4303).
     */
    bool               extSeqNum;
    union {
        /*
         * @brief Applies only to INBOUND Context.
         *        Enables the anti-replay checking by the SPP.
         *
         * It must be set to true if "extSeqNum" is true
         */
        bool           antiReplayEnable;
        /*
         * @brief Applies only to OUTBOUND. Specifies whether a
         *        Sequence Number counter is allowed to rollover.
         *
         * It must be set to false if "extSeqNum" is true.
         */
        bool           seqRollover;
    } sn;

    /*
     * @brief Specifies whether the SPP needs to check for
     * Soft/Hard byte lifetime expiration.
     *
     * If false, these lifetimes work as simple decrementing counters.
     */
    bool               lifetimesEnabled;
    /*
     * This value is used ONLY for the Inbound and ONLY
     * when anti-replay is enabled. The SPP only supports
     * ARW sizes specified by the enum values.
     */
    adk_ipsec_arw_size_t     antiReplayWindowSize;
} adk_crypto_ipsec_context_t;

/*!
 * ================================================================
 * @struct adk_ipsec_sa_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure is used to define security association used by the IPsec
 * ADK
 *
 * ================================================================
 */
typedef struct adk_ipsec_sa_s {

    /*! 8-bit field carrying IP version information */
    adk_ipsec_ip_version_t ipVersion;
    /*! Applicable only to ipsec egress SA configuration. If not configured IPSEC_SA_DEFAULT_TTL_HOPLIMIT_VALUE will be taken */
#define IPSEC_SA_DEFAULT_TTL_HOPLIMIT_VALUE 255
    union ttl_hopLimit {
        /*! 8-bit IPv4 ttl used in outer ipv4 header if per packet ttl configuration is not enabled*/
        uint8_t ttl;
        /*! 8-bit IPv6 hop limit used in outer ipv6 header if per packet hop limit configuration is not enabled*/
        uint8_t hopLimit;
    } eSaTttlHopLimit;
    /*! Applicable only if SA version is IPv6.
     *! Configuration for flow label value in tunnel IPv6 header. */
    adk_ipsec_flow_label_info_t     flowLabelInfo;
    union tun_addr {

        /*! IPv6 tunnel header IP address info */
        struct ipv6_tun_addr {
            /*! tunnel header source IPv6 address */
            uint32_t                tunnelSrcIP[4];
            /*! tunnel header destination IPv6 address */
            uint32_t                tunnelDstIP[4];
        } ipv6TunAddr;

        /*! IPv4 tunnel header IP address info */
        struct ipv4_tun_addr {
            /*! tunnel header source IPv4 address */
            uint32_t                tunnelSrcIP;
            /*! tunnel header destination IPv4 address */
            uint32_t                tunnelDstIP;
        } ipv4TunAddr;
    } tunAddr;
    /*! SASPPContext mostly contains the instantiated or run time
     * information of the SA as against the user configured information.
     */
    adk_crypto_ipsec_context_t SASPPContext;
    /*! NATT processing enable */
    bool                  isNATTEnable;
    /*! NATT processing context */
    adk_natt_ipsec_context_t    NATTContext;
    /*! 16-bit unique identifier for the set of multi-selector policy rules that the application sees 
     * as single policy but are expanded in the ADK fast path policy table.
     * Every user level policy should have unique ID.
     * If the user level policy has multiple selectors then application will expand them into a list of
     * ADK policy rules. In multi-selector case, each such rule will have same spIDEx.
     * This is unused for egress SA and used in ingress SA configuration.
     * It should be the same as the ID configured for the policy rules associated with ingress SA.
     */
	uint16_t    spIDEx;
    /* Inline/Lookaside configuration for particular SA */
    adk_ipsec_sa_mode_t saMode;

} adk_ipsec_sa_t; /*end of struct adk_ipsec_sa_t */

/*!
 * ================================================================
 * @struct adk_ipsec_upd_sa_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure is used to define part of security association required for SA update
 * used by the IPsec ADK
 *
 * ================================================================
 */
typedef struct adk_ipsec_upd_sa_s {

	/*! security association SPP context */
    adk_crypto_ipsec_context_t     SASPPContext;

} adk_ipsec_upd_sa_t; /*end of struct adk_ipsec_upd_sa_t */


/*!
 * ================================================================
 * @struct adk_ipsec_ing_stat_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure is used to define ingress side global IPsec statistics
 *
 * ================================================================
 */
typedef struct adk_ipsec_ing_stat_s {

    /*! number of ingress PDUs bypassed due to SPD IN policy look up action is bypass */
    uint64_t    ipsec_ing_global_spd_in_pol_bypass_count;
    /*! number of ingress security policy PDUs dropped */
    uint64_t    ipsec_ing_global_spd_in_pol_discard_drop;
    /*! number of ingress PDUs dropped due to SPD IN policy look up default discard */
    uint64_t    ipsec_ing_global_spd_in_default_discard_drop;
    /*! number of ingress keepalive PDUs dropped */
    uint32_t    ipsec_ing_global_keepalive_pdu_drop;
    /*! number of ingress unknown natt PDUs dropped */
    uint32_t    ipsec_ing_global_unknown_natt_pdu_drop;
    /*! number of ingress IP fragment PDUs dropped */
    uint32_t    ipsec_ing_global_ip_fragment_drop;
    /*! number of ingress security policy PDUs dropped */
    uint32_t    ipsec_ing_global_spd_pol_drop;
    /*! number of ingress PDUs dropped as no SA found for these PDUs */
    uint32_t    ipsec_ing_global_sa_not_present_drop;
    /*! number of ingress PDUs dropped due to authentication failure */
    uint32_t    ipsec_ing_global_auth_fail_drop;
    /*! number of ingress PDUs dropped due to pad verification failure */
    uint32_t    ipsec_ing_global_pad_verification_fail_drop;
    /*! number of ingress PDUs dropped due to sequence number check failure */
    uint32_t    ipsec_ing_global_seqno_check_fail_drop;
    /*! number of ingress PDUs dropped due to SA hard life expire */
    uint32_t    ipsec_ing_global_sa_hard_life_expire_drop;
    /*! number of ingress PDUs dropped due to SA validation failure */
    uint32_t    ipsec_ing_global_sa_validation_fail_drop;
    /*! number of ingress PDUs dropped due to SPD OUT policy look up default discard at the time of SA validation */
    uint64_t    ipsec_ing_global_spd_out_default_discard_drop;
    /*! number of ingress PDUs dropped due to SPD OUT lookup action is not IPsec at the time of SA validation */
    uint64_t    ipsec_ing_global_spd_out_not_ipsec_drop;
    /*! number of ingress side dummy pkt drop */
    uint32_t    ipsec_ing_global_dummy_pkt_drop;
    /*! number of TFC padding packets dropped */
    uint32_t    ipsec_ing_global_tfc_pad_drop;
    /*! number of ingress side exception packets dropped due to unregistered exception */
    uint32_t    ipsec_ing_global_exception_not_registered;
    /*! number of inline decryption failed packet count */
    uint32_t    ipsec_ing_global_inline_decryption_failed;
    /*! number of packets dropped while sending to crypto for decryption in lookaside */
    uint32_t    ipsec_ing_global_la_send_to_crypto_failed;
    /*! number of ingress packets dropped due to pad removal failure */
    uint32_t    ipsec_ing_global_pad_removal_fail_drop;
    /*! number of ingress packets dropped due to ARW update failure */
    uint32_t    ipsec_ing_global_arw_update_fail_drop;
    /*! number of packets dropped due to failure while sending the packet to next stage */
    uint32_t    ipsec_ing_global_send_to_next_stage_failed;
    /*! number of ingress packets processing failure drop */
    uint32_t    ipsec_ing_global_pkt_process_fail_drop;
    /*! number of ingress packets dropped while performing single sa buffered packet processing */
    uint32_t    ipsec_ing_global_buffered_single_sa_pkt_process_fail_drop;
    /*! number of packets dropped due to failure while creating clone packets. */
    uint32_t    ipsec_ing_global_mbuf_clone_failed;
    /*! number of too small packets received */
    uint32_t    ipsec_ing_global_too_small_pkt_drop;
    /*! number of packets dropped due to no crypto operation buffer allocated for the SA in lookaside */
    uint32_t    ipsec_ing_global_la_no_buffer_drop;
    /*! number of packets dropped due to pad buffer allocation failure */
    uint32_t    ipsec_ing_global_pad_alloc_fail_drop;
    /*! number of packets dropped due to reassembly failure*/
    uint32_t    ipsec_ing_global_reassembly_fail_drop;
    /*! number of packets dropped due to invalid destination function ID*/
    uint32_t    ipsec_ing_global_invalid_dst_fn_id_drop;
    /*! number of packets dropped due to spd in look up failure*/
    uint32_t    ipsec_ing_global_spd_in_lookup_fail_drop;
    /*! number of packets dropped due to unsupported l4 protocol*/
    uint32_t    ipsec_ing_global_unsupported_l4_proto_drop;
    /*! number of ingress side exception send failure drop count */
    uint32_t    ipsec_ing_global_exception_send_failure_drop;
} adk_ipsec_ing_stat_t __rte_cache_aligned; /* end of struct adk_ipsec_ing_stat_t */


/*!
 * ================================================================
 * @struct adk_ipsec_eg_stat_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure is used to define egress side global IPsec statistics
 *
 * ================================================================
 */
typedef struct adk_ipsec_eg_stat_s {

    /*! number of egress PDUs hitting bypass policy */
    uint64_t    ipsec_eg_global_spd_pol_bypassCnt;
    /*! number of egress PDUs policy discard drop count */
    uint64_t    ipsec_eg_global_spd_pol_discard_drop;
    /*! number of egress PDUs policy discard due to action is not ipsec */
    uint64_t    ipsec_eg_global_spd_out_not_ipsec_drop;
    /*! number of egress side exception packets dropped due to unregistered exception */
    uint32_t    ipsec_eg_global_exception_not_registered;
    /*! number of egress PDUs dropped due to default discard policy */
    uint64_t    ipsec_eg_global_spd_pol_default_discard_drop;
    /*! number of egress PDUs dropped due to sequence number overflow */
    uint32_t    ipsec_eg_global_seqno_overflow_drop;
    /*! number of egress PDUs dropped due to SA hardlife expire */
    uint32_t    ipsec_eg_global_sa_hardlife_expire_drop;
    /*! number of packets dropped since no matching SA found*/
    uint32_t    ipsec_eg_global_sa_not_found_drop;
    /*! number of packets dropped due to error in pre-fragmentation */
    uint32_t    ipsec_eg_global_prefrag_reqd_df_bit_set_drop_count;
    /*! number of packets dropped due to error returned in pre fragmentation */
    uint32_t    ipsec_eg_global_pre_frag_error_drop;
    /*! number of packets dropped due to pad buffer allocation failure */
    uint32_t    ipsec_eg_global_pad_alloc_fail_drop;
    /*! number of packets dropped due to no security session created for the SA in Inline */
    uint32_t    ipsec_eg_global_no_security_session_drop;
    /*! number of packets dropped while sending to crypto for encryption in lookaside */
    uint32_t    ipsec_eg_global_la_send_to_crypto_failed;
    /*! number of packets dropped due to no crypto operation buffer allocated for the SA in lookaside */
    uint32_t    ipsec_eg_global_la_no_buffer_drop;
    /*! number of too small packet dropped */
    uint32_t    ipsec_eg_global_too_small_pkt_drop;
    /*! number of egress packets dropped due to crypto encryption operation failure in lookaside mode */
    uint32_t    ipsec_eg_global_la_crypto_op_failed;
    /*! number of packets dropped due to failure while creating clone packets. */
    uint32_t    ipsec_eg_global_mbuf_clone_failed;
    /*! number packets dropped due to outer header buffer allocation failure */
    uint32_t    ipsec_eg_global_outer_hdr_buf_alloc_fail_drop;
    /*! number of egress side exception send failure drop count */
    uint32_t    ipsec_eg_global_exception_send_failure_drop;
} adk_ipsec_eg_stat_t __rte_cache_aligned; /* end of struct adk_ipsec_eg_stat_t */


/*!
 * ================================================================
 * @struct adk_ipsec_stat_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure is used to define global IPsec statistics maintained by
 * the IPsec ADK
 *
 * ================================================================
 */
typedef struct adk_ipsec_stat_s {

    /*! ingress side global statistics */
    adk_ipsec_ing_stat_t    ingressStats;
    /*! egress side global statistics */
    adk_ipsec_eg_stat_t     egressStats;

} adk_ipsec_stat_t __rte_cache_aligned; /* end of struct adk_ipsec_stat_t */



/*!
 * ================================================================
 * @struct adk_ipsec_ing_sa_stat_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief IPsec ADK ingress side SA statistics
 *
 * ================================================================
 */
typedef struct adk_ipsec_ing_sa_stat_s {

    /*! total ingress side PDUs received count */
	uint32_t	ipsec_ing_sa_tot_in_pkts;
    /*! ingress side PDUs cryptographic error count */
	uint32_t	ipsec_ing_sa_esp_cryptographic_error;
    /*! ingress side PDUs sequence number mismatch count */
	uint32_t	ipsec_ing_sa_seq_num_mismatch_drop;
    /*! total ingress packets dropped due to policy mismatch count*/
    uint32_t    ipsec_ing_sa_traffic_sel_mismatch_drop;
}__attribute__ ((aligned)) adk_ipsec_ing_sa_stat_t; /* end of struct adk_ipsec_ing_sa_stat_t */

/*!
 * ================================================================
 * @struct  adk_ipsec_eg_sa_stat_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief IPsec ADK egress side SA statistics
 *
 * ================================================================
 */
typedef struct adk_ipsec_eg_sa_stat_s {
    /*! total egress side PDUs transmit count*/
    uint32_t	ipsec_eg_sa_tot_out_pkts;
} __attribute__ ((aligned)) adk_ipsec_eg_sa_stat_t; /* end of struct adk_ipsec_eg_sa_stat_t */


/*!
 * ================================================================
 * @struct adk_ipsec_sa_stat_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * The data structure defines security association specific IPsec statistics
 * maintained by the IPsec ADK
 *
 * ================================================================
 */
typedef struct adk_ipsec_sa_stat_s {

    union s {
        /*! ingress SA side statistics */
        adk_ipsec_ing_sa_stat_t ing_sa_stat;
        /*! egress SA side statistics */
        adk_ipsec_eg_sa_stat_t  eg_sa_stat;
    }sa_stat;

} adk_ipsec_sa_stat_t __rte_cache_aligned; /* end of struct adk_ipsec_sa_stat_t */




/*!
 * ================================================================
 * @struct adk_ipsec_tcp_udp_audit_event_t adk_ipsec_struct.h
 * "adk_ipsec_struct.h"
 *
 * @brief event information in case of clear text packets
 *
 * ================================================================
 */
typedef struct adk_ipsec_tcp_udp_audit_event_s {

    /*! 16-bit layer 4 source port */
    uint16_t        srcPort;
    /*! 16-bit layer 4 destiantion port */
    uint16_t        dstPort;

} adk_ipsec_tcp_udp_audit_event_t; /* end of struct adk_ipsec_tcp_udp_audit_event_t */


/*!
 * ================================================================
 * @struct adk_ipsec_encr_audit_event_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief event information in case of cypher text packets
 *
 * ================================================================
 */
typedef struct adk_ipsec_encr_audit_event_s {

    /*! 32-bit spi value */
    uint32_t        spiVal;
    /*! 32-bit sequence number */
    uint32_t        seqNum;

} adk_ipsec_encr_audit_event_t; /* end of struct adk_ipsec_encr_audit_event_t */

/*!
 * ================================================================
 * @struct adk_ipsec_icmp_audit_event_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief event information in case of icmp packets
 *
 * ================================================================
 */
typedef struct adk_ipsec_icmp_audit_event_s {
    /*! 8-bit ICMP type field */
    uint8_t        type;
    /*! 8-bit ICMP code field */
    uint8_t        code;

} adk_ipsec_icmp_audit_event_t; /* end of struct adk_ipsec_icmp_audit_event_t */

/*!
 * ================================================================
 * @struct adk_ipsec_sa_msg_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief defines information to be captured in case of event generation
 *
 * ================================================================
 */

typedef struct adk_ipsec_sa_msg_s {

    /*! ipVersion */
    adk_ipsec_ip_version_t      ipVersion;

    union ip_addr {
        struct ipv4_addr {
            /*! IP packet source address information */
            adk_ipsec_ipv4_addr_info_t  srcAddr;
            /*! IP packet destination address information */
            adk_ipsec_ipv4_addr_info_t  dstAddr;
        }ipv4Addr;
        struct ipv6_addr {
            /*! IP packet source address information */
            adk_ipsec_ipv6_addr_info_t  srcAddr;
            /*! IP packet destination address information */
            adk_ipsec_ipv6_addr_info_t  dstAddr;
        }ipv6Addr;
    }ipAddr;
    union sa_msg_l4_info_s {
        struct sa_msg_port_info_s {
            /*! Layer 4 packet source port information */
            adk_ipsec_port_info_t       srcPort;
            /*! Layer 4 packet destination port information */
            adk_ipsec_port_info_t       dstPort;
        }portInfo;

        struct sa_msg_icmp_info_s {
            /*! icmp type information */
            uint8_t       type;
            /*! icmp code information */
            uint8_t       code;
        }icmpInfo;
    }l4Info;

    /*! IP packet protocol field information */
    uint8_t                 protocol;   /* from PDU */

}adk_ipsec_sa_msg_t;

/*!
 * ================================================================
 * @ adk_ipsec_audit_event_before_encr_pkt_data_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief Event information in case of,
 * 1. clear text packet in egress side.
 * 2. clear text inner packet (after decryption in the ingress side) for the below audit events,
 *    a. ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_MISMATCH
 *    b. ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_DISCARD
 *    c. ADK_IPSEC_AUDIT_ING_SPP_SPD_OUT_LKUP_NO_IPSEC_DROP
 * ================================================================
 */
typedef struct adk_ipsec_audit_event_before_encr_pkt_data_s {

    /*! IP version */
    adk_ipsec_ip_version_t   ipVersion;
    /*! protocol field of IP packet */
    uint8_t         protocol;
    /*! dscp field of IP packet */
    uint8_t         dscp;
    /*! flow label if IPv6 packet */
    uint32_t        flowLabel;
    /*! Source IP addresses as per IP version*/
    union ip_src_add_u {
        /*! ipv6 src address */
        uint32_t ipv6Addr[4];
        /*! ipv4 src address */
        uint32_t ipv4Addr;
    } ipSrcAddr_u;

    /*! Destination IP addresses as per IP version*/
    union ip_dst_add_u {
        /*! ipv6 dst address */
        uint32_t ipv6Addr[4];
        /*! ipv4 dst address */
        uint32_t ipv4Addr;
    } ipDstAddr_u;

    /*! l4Port info as per IP packet*/
    union _l4Port {
        /*! event information in case of TCP/UDP/SCTP packets */
        adk_ipsec_tcp_udp_audit_event_t     ipsecTCPUDPEvent;
        /*! event information in case of ICMP packets */
        adk_ipsec_icmp_audit_event_t        ipsecICMPEvent;
        /*! event information in case of ESP/AH packets */
        adk_ipsec_encr_audit_event_t      ipsecEspAhEvent;
    }l4Ports;
} adk_ipsec_audit_event_before_encr_pkt_data_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_encr_pkt_data_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief Event data for events that are generated because of encrypted packet
 *
 * RFC-Recommendation
 * The audit log entry for this event SHOULD include the SPI value, date/time received,
 * Source Address, Destination Address,Sequence Number,Protocol (ESP/AH) and (for IPv6) the flow ID.
 * ================================================================
 */
typedef struct adk_ipsec_audit_event_encr_pkt_data_s{
    /*! IP version */
    adk_ipsec_ip_version_t  ipVersion;
    /*! ipsec protocol value */
    uint8_t         secProtocol;
    /*! spi value */
    uint32_t        spiVal;
    /*! sequence number */
    uint32_t        seqNum;
    /*! flow label if IPv6 packet */
    uint32_t        flowLabel;

    /*! Source IP addresses as per IP version*/
    union ip_srcAddr_u {
        /*! ipv6 src address */
        uint32_t ipv6Addr[4];
        /*! ipv4 src address */
        uint32_t ipv4Addr;
    } ipSrcAddr_u;

    /*! Destination IP addresses as per IP version*/
    union ip_dstAddr_u {
        /*! ipv6 dst address */
        uint32_t ipv6Addr[4];
        /*! ipv4 dst address */
        uint32_t ipv4Addr;
    } ipDstAddr_u;

} adk_ipsec_audit_event_encr_pkt_data_t;

/*!
 * ================================================================
 * @ adk_ipsec_audit_ing_sa_validation_data_from_rx_pkt_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief Event information from incoming packet used in case of ingress SA validation
 * ================================================================
 */
typedef adk_ipsec_audit_event_before_encr_pkt_data_t adk_ipsec_audit_ing_sa_validation_data_from_rx_pkt_t;

/*!
 * ================================================================
 * @ adk_ipsec_audit_ing_sa_validation_data_from_expected_isa_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief Event information from expected SA used in case of ingress SA validation. This
 * is essentially an ingress SA handle from which application can get policy selectors associated with it.
 * ================================================================
 */
typedef adk_handle_t  adk_ipsec_audit_ing_sa_validation_data_from_expected_isa_t;

/*!
 * ================================================================
 * @ adk_ipsec_audit_event_ing_sa_validation_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief Event information in case of ingress SA validation
 * ================================================================
 */
typedef struct adk_ipsec_audit_event_ing_sa_validation_s {
    /*! SA Mismatch data from packet */
    adk_ipsec_audit_ing_sa_validation_data_from_rx_pkt_t            data_from_rx_pkt;
    /*! SA Mismatch data from SA */
    adk_ipsec_audit_ing_sa_validation_data_from_expected_isa_t      data_from_expected_ingress_sa;
    /*! spi value */
    uint32_t        spiVal;
    /*! ipsec protocol value */
    uint8_t         secProtocol;
    /*! IP version of SA */
    adk_ipsec_ip_version_t  ipVersionSA;

} adk_ipsec_audit_event_ing_sa_validation_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_sa_not_present_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_NOT_PRESENT
 *
 * RFC-Recommendation
 * SHOULD include the SPI value, date/time received, Source Address, Destination Address,
 * Sequence Number, and (in IPv6)the Flow ID.
 * ================================================================
 */

typedef adk_ipsec_audit_event_encr_pkt_data_t adk_ipsec_audit_event_ing_sa_not_present_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_spd_out_lookup_discard_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_DISCARD
 *
 * RFC-Recommendation
 *
 * If an IPsec system receives an inbound packet on an SA and the packet’s header fields
 * are not consistent with the selectors for the SA, it MUST discard the packet.
 * This is an auditable event. The audit log entry for this event SHOULD include the current date/time,
 * SPI, IPsec protocol(s), source and destination of the packet, any other selector values of
 * the packet that are available, and the selector values from the relevant SAD entry
 * ================================================================
 */
typedef adk_ipsec_audit_event_ing_sa_validation_t adk_ipsec_audit_event_ing_spd_out_lookup_discard_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_spd_out_lookup_no_ipsec_policy_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_NO_IPSEC_DROP
 *
 * RFC-Recommendation
 *
 * If an IPsec system receives an inbound packet on an SA and the packet’s header fields
 * are not consistent with the selectors for the SA, it MUST discard the packet.
 * This is an auditable event. The audit log entry for this event SHOULD include the current date/time,
 * SPI, IPsec protocol(s), source and destination of the packet, any other selector values of
 * the packet that are available, and the selector values from the relevant SAD entry
 * ================================================================
 */
typedef adk_ipsec_audit_event_ing_sa_validation_t adk_ipsec_audit_event_ing_spd_out_lookup_no_ipsec_policy_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_spd_in_lookup_discard_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_SPD_IN_LKUP_DISCARD
 *
 * RFC-Recommendation
 * If the packet is not addressed to the device or is addressed to
 * this device and is not AH or ESP, look up the packet header in
 * the (appropriate) SPD-I cache.  If there is a match and the
 * packet is to be discarded or bypassed, do so.  If there is no
 * cache match, look up the packet in the corresponding SPD-I and
 * create a cache entry as appropriate. (No SAs are created in
 * response to receipt of a packet that requires IPsec protection;
 * only BYPASS or DISCARD cache entries can be created this way.) If
 * there is no match, discard the traffic.  This is an auditable
 * event.  The audit log entry for this event SHOULD include the
 * current date/time, SPI if available, IPsec protocol if available,
 * source and destination of the packet, and any other selector
 * values of the packet that are available.
 * ================================================================
 */
typedef adk_ipsec_audit_event_before_encr_pkt_data_t adk_ipsec_audit_event_ing_spd_in_lookup_discard_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_sa_mismatch_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_MISMATCH
 *
 * RFC-Recommendation
 *
 * If an IPsec system receives an inbound packet on an SA and the packet’s header fields
 * are not consistent with the selectors for the SA, it MUST discard the packet.
 * This is an auditable event. The audit log entry for this event SHOULD include the current date/time,
 * SPI, IPsec protocol(s), source and destination of the packet, any other selector values of
 * the packet that are available, and the selector values from the relevant SAD entry
 * ================================================================
 */
typedef adk_ipsec_audit_event_ing_sa_validation_t adk_ipsec_audit_event_ing_sa_mismatch_t ;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_fragmented_ipsec_pkt_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_IP_FRAGMENT_DROP
 *
 * RFC-Recommendation
 * A packet offered to ESP for processing appears to be an IP fragment, i.e., the
 * OFFSET field is non-zero or the MORE FRAGMENTS flag is set. The audit log entry for this event
 * SHOULD include the SPI value, date/time received, Source Address, Destination Address,
 * Sequence Number, and (in IPv6)the Flow ID.
 * ================================================================
 */

typedef adk_ipsec_audit_event_encr_pkt_data_t  adk_ipsec_audit_event_ing_fragmented_ipsec_pkt_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_seq_check_failure_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_SEQNO_CHECK_FAILURE
 *
 * RFC-Recommendation
 *
 * The received packet fails the anti-replay checks. The audit log entry for this event
 * SHOULD include the SPI value, date/time received, Source Address, Destination Address,
 * the Sequence Number, and (in IPv6) the Flow ID.
 *
 * ================================================================
 */
typedef adk_ipsec_audit_event_encr_pkt_data_t adk_ipsec_audit_event_ing_seq_check_failure_t;
/*!
 * ================================================================
 * @adk_ipsec_audit_event_ing_auth_check_failure_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_ING_AUTH_CHECK_FAILURE
 *
 * RFC-Recommendation
 *
 * The integrity check fails. The audit log entry for this event SHOULD include the
 * SPI value, date/time received, Source Address, Destination Address, the Sequence Number,
 * and (for IPv6) the Flow ID.
 * ================================================================
 */
typedef adk_ipsec_audit_event_encr_pkt_data_t adk_ipsec_audit_event_ing_auth_check_failure_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_eg_spd_out_lookup_discard_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_EG_SPD_OUT_LKUP_DISCARD and
 * ADK_IPSEC_EXCEPTION_POLICY_2ND_FRAGMENT_2SA_PREFRAG
 *
 * RFC-Recommendation
 * If an IPsec system receives an outbound packet that it finds it must discard,
 * it SHOULD be capable of generating and sending an ICMP message to indicate to
 * the sender of the outbound packet that the packet was discarded. The reason SHOULD
 * be recorded in the audit log.
 * The audit log entry for this event SHOULD include the reason, current date/time,
 * and the selector values from the packet
 * ================================================================
 */
typedef  adk_ipsec_audit_event_before_encr_pkt_data_t adk_ipsec_audit_event_eg_spd_out_lookup_discard_t;

/*!
 * ================================================================
 * @adk_ipsec_audit_event_eg_seq_no_overflow_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_AUDIT_EG_SEQNO_OVERFLOW
 *
 * RFC-Recommendation
 *
 * Attempt to transmit a packet that would result in Sequence Number overflow.
 * The audit log entry for this event SHOULD include the SPI value, current date/time,
 * Source Address, Destination Address, Sequence Number, and (for IPv6) the cleartext Flow ID.
 *
 * ================================================================
 */
typedef adk_ipsec_audit_event_encr_pkt_data_t adk_ipsec_audit_event_eg_seq_no_overflow_t;

/*!
 * ================================================================
 * @ adk_ipsec_audit_event_eg_no_policy_2nd_fragment_2sa_prefrag_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_NO_POLICY_2ND_FRAGMENT_2SA_PREFRAG
 * ================================================================
 */
typedef  adk_ipsec_audit_event_before_encr_pkt_data_t adk_ipsec_eg_event_no_policy_2nd_fragment_2sa_prefrag_t;

/*!
 * ================================================================
 * @ adk_ipsec_audit_event_eg_prefrag_reqd_but_df_bit_set_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief audit event data for ADK_IPSEC_EXCEPTION_EG_PREFRAG_REQD_BUT_DF_BIT_SET
 * ================================================================
 */
typedef  adk_ipsec_audit_event_before_encr_pkt_data_t adk_ipsec_eg_event_prefrag_reqd_but_df_bit_set_t;

/*!
 * =============================================================
 * @struct adk_ipsec_exception_type_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * This structure defines the exception type used by exception handler
 * structure.
 *
 * =============================================================
 */
typedef enum adk_ipsec_exception_type_s {
    ADK_IPSEC_EXCEPTION_INVALID = 0,
    ADK_IPSEC_EXCEPTION_SA_ACQUIRE,
    ADK_IPSEC_EXCEPTION_SA_EXPIRE,
    ADK_IPSEC_EXCEPTION_NO_POLICY_2ND_FRAGMENT_2SA_PREFRAG,
    ADK_IPSEC_EXCEPTION_EG_PREFRAG_REQD_BUT_DF_BIT_SET,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_NOT_PRESENT,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_DISCARD,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_NO_IPSEC_DROP,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_SPD_IN_LKUP_DISCARD,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_MISMATCH,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_IP_FRAGMENT_DROP,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_SEQNO_CHECK_FAILURE,
    ADK_IPSEC_EXCEPTION_AUDIT_ING_AUTH_CHECK_FAILURE,
    ADK_IPSEC_EXCEPTION_AUDIT_EG_SPD_OUT_LKUP_DISCARD,
    ADK_IPSEC_EXCEPTION_AUDIT_EG_SEQNO_OVERFLOW,
    ADK_IPSEC_EXCEPTION_MAX,
}adk_ipsec_exception_type_t;

/*!
 * ================================================================
 * @struct adk_ipsec_audit_event_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief defines information to be captured in case of event generation
 * ================================================================
 */
typedef struct adk_ipsec_audit_event_s {
    /*! 32-bit MSW of time field */
    uint32_t        timeInSec;
    /*! 32-bit LSW of time field */
    uint32_t        timeInUsec;
    /*! Event Type specific fields for each event */
    union _event {
        /*! Ingress Events */
        adk_ipsec_audit_event_ing_sa_not_present_t                    ipsecIngSaNotPresentEvent;
        adk_ipsec_audit_event_ing_spd_out_lookup_discard_t            ipsecIngSpdOutLkupDiscardEvent;
        adk_ipsec_audit_event_ing_spd_out_lookup_no_ipsec_policy_t    ipsecIngSpdOutLkupNoIpsecPolicyEvent;
        adk_ipsec_audit_event_ing_spd_in_lookup_discard_t             ipsecIngSpdInLkupDiscardEvent;
        adk_ipsec_audit_event_ing_sa_mismatch_t                       ipsecIngSaMismatchEvent;
        adk_ipsec_audit_event_ing_fragmented_ipsec_pkt_t              ipsecIngFragmentedIpsecPktEvent;
        adk_ipsec_audit_event_ing_seq_check_failure_t                 ipsecIngSeqCheckFailureEvent;
        adk_ipsec_audit_event_ing_auth_check_failure_t                ipsecIngAuthCheckFailureEvent;

        /*! Egress Events */
        adk_ipsec_audit_event_eg_spd_out_lookup_discard_t             ipsecEgSpdOutLkupDiscardEvent;
        adk_ipsec_audit_event_eg_seq_no_overflow_t                    ipsecEgSeqNumOverFlowEvent;
        adk_ipsec_eg_event_no_policy_2nd_fragment_2sa_prefrag_t       ipsecEgNoPolicy2ndFrag2SaPrefragEvent;
        adk_ipsec_eg_event_prefrag_reqd_but_df_bit_set_t              ipsecEgPrefragReqdButDfBitSetEvent;
    } event;
} adk_ipsec_audit_event_t; /* end of struct adk_ipsec_audit_event_t */

/*!
 * =============================================================
 * @struct adk_ipsec_exception_handler_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 *
 * @brief
 *
 * type        Type field decides the type of exception
 *
 * devId      PH Device ID returned by adk_ph_dev_device_configure
 *
 * queueId    PH QueueID used in adk_ph_dev_rx_queue_setup
 *
 * Priority   This will be updated in eventDev metadata
 *
 * cbFnHandle PH Callback Fn handle return by adk_ph_dev_callback_fn_add()
 *
 * policerHandle Handle for host policer
 *
 * This structure defines the parameters used by ipsec register exception
 *
 * =============================================================
 */
typedef struct adk_ipsec_exception_handler_s {
    /*!
     * type field decides the type of exception.
     */
    adk_ipsec_exception_type_t exceptionType;
    uint8_t devId;
    uint16_t queueId;
    uint8_t priority;
    adk_ph_dev_cbfn_handle_t cbFnHandle;
    adk_handle_t policerHandle;
} adk_ipsec_exception_handler_t;



typedef struct adk_ipsec_sa_aquire_excep_data_s {
    /* Policy Index on which exception occurred */
    uint16_t                 polIndex;
    /* SA Message from packet info*/
    adk_ipsec_sa_msg_t       saMsg;
}adk_ipsec_sa_aquire_excep_data_t;


typedef struct adk_ipsec_sa_expiry_excep_data_s {
    /*! 8-bit direction field specifying SA direction */
    uint8_t                 direction;
    /*! ipVersion */
    adk_ipsec_ip_version_t      ipVersion;
    /*! IP packet destination address information */
    union ipAddInfo_s {
        /*! ipv4 destination address info */
        adk_ipsec_ipv4_addr_info_t ipv4dstAddr;
        /*! ipv6 destination address info */
        adk_ipsec_ipv6_addr_info_t ipv6dstAddr;
    } ipAddrInfo;

    /*! IP packet protocol field information */
    uint8_t                 protocol;   /* from PDU */
    /*! 32-bit spi value */
    uint32_t                spi;        /* SPP context */
    /*! 8-bit field specifying which type of expiry of SA.*/
    adk_sa_expiry_type_t     whichExpiry;
}adk_ipsec_sa_expiry_excep_data_t;

typedef struct adk_ipsec_recv_exception_param_s
{
    adk_ipsec_exception_type_t exceptionType;
    union {
        adk_ipsec_sa_aquire_excep_data_t saAquireData;
        adk_ipsec_sa_expiry_excep_data_t saExpiryData;
        adk_ipsec_audit_event_t          auditEventData;
    }exceptionData;
}adk_ipsec_recv_exception_param_t;

/*!
 * * =============================================================
 * * @enum adk_ipsec_table_t
 * *
 * * @brief
 * *
 * * This enum defines the IPsec tables
 * *
 * * =============================================================
 * */
typedef enum  adk_ipsec_table_e {
    /*! SPD In table */
    ADK_IPSEC_SPD_IN_TBL = 0x0,
    /*! SPD Out table */
    ADK_IPSEC_SPD_OUT_TBL,
    /*! MAX TABLE*/
    ADK_IPSEC_MAX_TBL,
}  adk_ipsec_table_t;   /* ------  end of enum  adk_ipsec_table_t ------- */

/*!
 * * ================================================================
 * * @struct adk_ipsec_sec_pol_sel_lkup_t adk_ipsec_struct.h "adk_ipsec_struct.h"
 * *
 * * @brief
 * *
 * * The data structure defines security policy selector values used for lookup
 * *
 * * ================================================================
 * */
typedef struct adk_ipsec_sec_pol_sel_lkup_s {
    /*! 8-bit field carrying IP version information */
    adk_ipsec_ip_version_t      ipVersion;
    union ip_addr_lkup {
        struct ipv6_addr_lkup {
            /*! source IPv6 address */
            uint32_t    srcAddr[4];
            /*! source IPv6 address */
            uint32_t    dstAddr[4];
        } ipv6Addr;
        struct ipv4_addr_lkup {
            /*! source IPv4 address */
            uint32_t    srcAddr;
            /*! source IPv4 address */
            uint32_t    dstAddr;
        } ipv4Addr;
    } ipAddr;
    /*! 8-bit IP header protocol field */
    uint8_t          protocol;
    union l4_info_s {
        struct l4_port_info_s {
            /*! Layer 4 packet source port information */
            uint16_t       srcPort;
            /*! Layer 4 packet destination port information */
            uint16_t       dstPort;
        }portInfo;

        struct icmp_info_s {
            /*! icmp type information */
            uint8_t       type;
            /*! icmp code information */
            uint8_t       code;
        }icmpInfo;
    }l4Info;
    /*! 8-bit dscp value. Specify ADK_IPSEC_DSCP_ANY to wildcard dscp field.
     * This should be ignored while doing lookup in SPD IN policy.
     */
    uint8_t          dscpVal;
} adk_ipsec_sec_pol_sel_lkup_t; /* end of struct adk_ipsec_sec_pol_sel_lkup_t */

typedef struct adk_ipsec_ph_dp_cb_params_s
{
    uint8_t version;
    uint8_t direction;
    uint16_t phDevId;
    uint16_t phQueueId;
}adk_ipsec_ph_dp_cb_params_t;

#endif

/*! end addtogroup _ipsec
 * @}
 */
