/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipsec_cp_utility_counters.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipsecUtilityCounters  IPsec Utility Counters
 * @{
 *
 * @brief
 *
 * This file contains ADK APIs to enable/disable utility counters and get API to fetch the counter values. 
 *
 ***********************************************************************/

#ifndef _ADK_IPSEC_UTILITY_COUNTER_H_
#define _ADK_IPSEC_UTILITY_COUNTER_H_

/*!
 * =============================================================
 * @internal
 * @struct ipsec_ing_utility_counters_t "adk_ipsec_utility_counters.h"
 *
 * @brief
 *
 * Data structure for ingress utility counters.
 * =============================================================
 */

typedef struct ipsec_ing_utility_counters_s
{
   /*! Sum of ingress pdu's received count */
   uint64_t    ipsec_ing_total_pdu_recv_cnt;
   /*! Sum of ingress pdu's sent for inbound processing in Look_Aside IPsec operation */
   uint64_t    ipsec_ing_la_num_pdus_sent_for_inbound_processing;
   /*! Sum of pdu's received after inbound processing in Look_Aside or Inline operation */
   uint64_t    ipsec_ing_num_pdus_rcvd_after_inbound_processing;
   /*! Sum of ingress side exceptions sent count */
   uint64_t    ipsec_ing_exception_sent;
   /*! Sum of all the packets that got decrypted successfully in HW */
   uint64_t    ipsec_ing_inline_decrypted;
   /*! number of ingress nonlocal pdu's count */
   uint64_t    ipsec_ing_nonlocal_pkts;
   /*! Total number of initial and non-initial fragments buffered in IPsec for Single SA use case */
   uint64_t    ipsec_ing_total_buffered_frags;
} ipsec_ing_utility_counters_t;

/*!
 * =============================================================
 * @internal
 * @struct ipsec_eg_utility_counters_t "adk_ipsec_utility_counters.h"
 *
 * @brief
 *
 * Data structure for egress utility counters.
 * =============================================================
 */

typedef struct ipsec_eg_utility_counters_s
{
    /*! Sum of egress pdu's received count */
    uint64_t   ipsec_eg_total_pdu_recv_cnt;
    /*! Sum of egress exceptions sent count */
    uint64_t   ipsec_eg_exception_sent;
    /*! Sum of egress pdu's sent for out_bound_processing in Look_Aside or Inline operation */
    uint64_t   ipsec_eg_num_pdus_sent_for_outbound_processing;
    /*! number of egress trnasmitted pdu's count */
    uint64_t    ipsec_eg_tx_pkts;
    /*! Sum of pdu's received after outbound processing in Look_Aside operation */
    uint64_t   ipsec_eg_la_num_pdus_rcvd_after_outbound_processing;
    /*! Sum of egress forwarded pdu's count */
    uint64_t    ipsec_eg_fwd_pkts;
} ipsec_eg_utility_counters_t;

/*!
 * ================================================================
 * @struct adk_ipsec_utility_counters_t adk_ipsec_utility_counters.h "adk_ipsec_utility_counters.h"
 *
 * @brief
 *
 * The data structure is used to define utility counters maintained by
 * the IPsec ADK
 *
 * ================================================================
 */
typedef struct adk_ipsec_utility_counters_s {

    /*! ingress side utility counters */
    ipsec_ing_utility_counters_t    ingressUtilityCounters;
    /*! egress side utility counters */
    ipsec_eg_utility_counters_t     egressUtilityCounters;

} adk_ipsec_utility_counters_t; /* end of struct adk_ipsec_utility_counters_t */


/*!
 * =============================================================
 * @internal
 * @struct ipsec_inbound_sa_utility_counters_t "adk_ipsec_utility_counters.h"
 *
 * @brief
 *
 * Data structure for inbound SA utility counters.
 * =============================================================
 */

typedef struct ipsec_inbound_sa_utility_counters_s
{
    /*! total ingress side PDU's received byte count */
    uint64_t   ipsec_ing_sa_tot_in_bytes;
    /*! total ingress side inflight PDU count */
    uint64_t   ipsec_ing_sa_in_flight_pkt_count;
} ipsec_inbound_sa_utility_counters_t;

/*!
 * =============================================================
 * @internal
 * @struct ipsec_outbound_sa_utility_counters_t "adk_ipsec_utility_counters.h"
 *
 * @brief
 *
 * Data structure for outbound SA utility counters.
 * =============================================================
 */

typedef struct ipsec_outbound_sa_utility_counters_s
{
    /*! total egress side PDU's transmit count */
    uint64_t   ipsec_eg_sa_tot_out_bytes;
    /*! total egress side in flight PDU count */
    uint64_t   ipsec_eg_sa_inflight_pkt;
} ipsec_outbound_sa_utility_counters_t;

/*!
 * ================================================================
 * @struct adk_ipsec_sa_utility_counters_t adk_ipsec_utility_counters.h "adk_ipsec_utility_counters.h"
 *
 * @brief
 *
 * The data structure is used to define SA utility counters maintained by
 * the IPsec ADK
 *
 * ================================================================
 */
typedef struct adk_ipsec_sa_utility_counters_s {

    /*! inbound SA utility counters */
    ipsec_inbound_sa_utility_counters_t    ingSaUtilityCounters;
    /*! outbound SA utility counters */
    ipsec_outbound_sa_utility_counters_t    egSaUtilityCounters;

} adk_ipsec_sa_utility_counters_t; /* end of adk_ipsec_sa_utility_counters_t */


/*!
 * =============================================================
 *
 * adk_ipsec_set_utility_counters
 *
 * @param[in]  ipsecHandle  - IPsec module handle. Returned at the
 *                            time of IPsec module's registration
 *                            with the module manager.
 * @param[in]   utility_cnt_state - Boolean variable to enable/disable
                                    utility counter.
 *
 * @brief   Function to configure utility counters.
 *
 * This function configures utility counters.On disabling the utility
 * counters, all the values will get reset to zero.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_set_utility_counters(adk_module_handle_t ipsecHandle, bool utility_cnt_state);

/*!
 * =============================================================
 * @fn adk_ipsec_get_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 * @param[in]    ipVersion   - IP version of the utility counter.
 *
 * @param[out]   ipsecUtilityCounter_p - IPsec related utility counter data structure pointer
 *               gets set by the IPsec ADK
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully fetched utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to retreive all IPsec related utility counters
 * maintained by the IPsec ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size adk_ipsec_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_utility_counters(adk_module_handle_t           ipsecHandle,
                               adk_ipsec_ip_version_t        ipVersion,
                               adk_ipsec_utility_counters_t *ipsecUtilityCnt_p);

/*!
 * =============================================================
 * @fn adk_ipsec_get_sa_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 *
 * @param[out]   ipseSaUtilityCounter_p - IPsec related SA utility counter data structure pointer
 *               gets set by the IPsec ADK
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully fetched SA utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to retreive all IPsec related SA utility counters
 * maintained by the IPsec ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size adk_ipsec_sa_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_sa_utility_counters(adk_module_handle_t  ipsecHandle,
                                 adk_ipsec_sa_utility_counters_t  *saUtilityCnt_p);

/*!
 * =============================================================
 * @fn adk_ipsec_print_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 * @param[in]    ipVersion   - IP version of the utility counter.
 *
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully printed utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to print all IPsec related utility counters
 * maintained by the IPsec ADK module.
 *
 * =============================================================
 */

adk_status_t
adk_ipsec_print_utility_counters(adk_module_handle_t     ipsecHandle,
                                 adk_ipsec_ip_version_t  ipVersion);

/*!
 * =============================================================
 * @fn adk_ipsec_print_sa_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 *
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully printed utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to print all IPsec related sa utility counters
 * maintained by the IPsec ADK module.
 *
 * =============================================================
 */

adk_status_t
adk_ipsec_print_sa_utility_counters(adk_module_handle_t  ipsecHandle);


#endif
