/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

/*!
 * @file  adk_ipsec_cp_utility_counters.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipsec IPsec Module
 * @{
 *      @addtogroup _ipsecUtilityCounters IPsec Utility Counters
 *      @{
 *
 * @brief
 *
 * The file contains APIs to configure IPsec utility counters from the application
 * and get APIs to fetch utility counters from DP.
 ***********************************************************************/
#include "adk_include.h"

/* ipsec header file */
#include "adk_ipsec.h"
#include "adk_ipsec_common.h"
#include "adk_ipsec_cp.h"

/*!
 * =============================================================
 *
 * adk_ipsec_set_utility_counters
 *
 * @param[in]  ipsecHandle  - IPsec module handle. Returned at the
 *                            time of IPsec module's registration
 *                            with the module manager.
 * @param[in]   utility_cnt_state - Boolean variable to enable/disable
                                    utility counter.
 *
 * @brief   Function to configure utility counters.
 *
 * This function configures utility counters.On disabling the utility
 * counters, all the values will get reset to zero.
 *
 * =============================================================
 */

adk_status_t
adk_ipsec_set_utility_counters(adk_module_handle_t ipsecHandle, bool utility_cnt_state)
{
    ipsec_module_info_t             *ipsecModuleInfo_p = NULL;
    adk_status_t                    retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t                sysHandle;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);

    if(utility_cnt_state)
    {
        ipsecModuleInfo_p->dpTbl->isUtilityCntEnabled = true;
    }
    else
    {
        ipsecModuleInfo_p->dpTbl->isUtilityCntEnabled = false;
        /* Reset the utility counters */
        memset(ipsecModuleInfo_p->dpTbl->ipsecUtilityCntV4, 0,
               sizeof(adk_ipsec_utility_counters_t)*ADK_MAX_NUM_CORES);
        memset(ipsecModuleInfo_p->dpTbl->ipsecUtilityCntV6, 0,
               sizeof(adk_ipsec_utility_counters_t)*ADK_MAX_NUM_CORES);
        /* reset SA utility counters */
        memset(ipsecModuleInfo_p->dpTbl->SAUtilityCnt, 0,
               sizeof(adk_ipsec_sa_utility_counters_t)*ADK_MAX_NUM_CORES*ADK_IPSEC_NUM_OF_SA);
    }
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * @fn adk_ipsec_get_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 * @param[in]    ipVersion   - IP version of the utility counter.
 *
 * @param[out]   ipsecUtilityCounter_p - IPsec related utility counter data structure pointer
 *               gets set by the IPsec ADK
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully fetched utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to retreive all IPsec related utility counters
 * maintained by the IPsec ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size adk_ipsec_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_utility_counters(adk_module_handle_t  ipsecHandle,
                               adk_ipsec_ip_version_t  ipVersion,
                               adk_ipsec_utility_counters_t    *ipsecUtilityCounter_p)
{
    adk_sys_handle_t      sysHandle;
    adk_status_t retValue = ADK_IPSEC_SUCCESS;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    adk_ipsec_utility_counters_t      *utilityCounter;

    uint16_t  i;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    if (ipsecUtilityCounter_p == NULL)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    ADK_MEMSET (ipsecUtilityCounter_p, 0x00, sizeof(adk_ipsec_utility_counters_t));

    if (ipVersion == ADK_IPSEC_IPV4)
    {
        for (i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            utilityCounter = &ipsecModuleInfo_p->dpTbl->ipsecUtilityCntV4[i];
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_total_pdu_recv_cnt
                += utilityCounter->ingressUtilityCounters.ipsec_ing_total_pdu_recv_cnt;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_la_num_pdus_sent_for_inbound_processing
                += utilityCounter->ingressUtilityCounters.ipsec_ing_la_num_pdus_sent_for_inbound_processing;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_num_pdus_rcvd_after_inbound_processing
                += utilityCounter->ingressUtilityCounters.ipsec_ing_num_pdus_rcvd_after_inbound_processing;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_exception_sent
                += utilityCounter->ingressUtilityCounters.ipsec_ing_exception_sent;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_inline_decrypted
                += utilityCounter->ingressUtilityCounters.ipsec_ing_inline_decrypted;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_total_buffered_frags
                += utilityCounter->ingressUtilityCounters.ipsec_ing_total_buffered_frags;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_total_pdu_recv_cnt
                += utilityCounter->egressUtilityCounters.ipsec_eg_total_pdu_recv_cnt;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_exception_sent
                += utilityCounter->egressUtilityCounters.ipsec_eg_exception_sent;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_num_pdus_sent_for_outbound_processing
                += utilityCounter->egressUtilityCounters.ipsec_eg_num_pdus_sent_for_outbound_processing;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_la_num_pdus_rcvd_after_outbound_processing
                += utilityCounter->egressUtilityCounters.ipsec_eg_la_num_pdus_rcvd_after_outbound_processing;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_tx_pkts
                += utilityCounter->egressUtilityCounters.ipsec_eg_tx_pkts;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_fwd_pkts
                += utilityCounter->egressUtilityCounters.ipsec_eg_fwd_pkts;
        }
    }
    else
    {
        for (i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            utilityCounter = &ipsecModuleInfo_p->dpTbl->ipsecUtilityCntV6[i];
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_total_pdu_recv_cnt
                += utilityCounter->ingressUtilityCounters.ipsec_ing_total_pdu_recv_cnt;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_la_num_pdus_sent_for_inbound_processing
                += utilityCounter->ingressUtilityCounters.ipsec_ing_la_num_pdus_sent_for_inbound_processing;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_num_pdus_rcvd_after_inbound_processing
                += utilityCounter->ingressUtilityCounters.ipsec_ing_num_pdus_rcvd_after_inbound_processing;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_exception_sent
                += utilityCounter->ingressUtilityCounters.ipsec_ing_exception_sent;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_inline_decrypted
                += utilityCounter->ingressUtilityCounters.ipsec_ing_inline_decrypted;
            ipsecUtilityCounter_p->ingressUtilityCounters.ipsec_ing_total_buffered_frags
                += utilityCounter->ingressUtilityCounters.ipsec_ing_total_buffered_frags;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_total_pdu_recv_cnt
                += utilityCounter->egressUtilityCounters.ipsec_eg_total_pdu_recv_cnt;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_exception_sent
                += utilityCounter->egressUtilityCounters.ipsec_eg_exception_sent;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_num_pdus_sent_for_outbound_processing
                += utilityCounter->egressUtilityCounters.ipsec_eg_num_pdus_sent_for_outbound_processing;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_la_num_pdus_rcvd_after_outbound_processing
                += utilityCounter->egressUtilityCounters.ipsec_eg_la_num_pdus_rcvd_after_outbound_processing;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_tx_pkts
                += utilityCounter->egressUtilityCounters.ipsec_eg_tx_pkts;
            ipsecUtilityCounter_p->egressUtilityCounters.ipsec_eg_fwd_pkts
                += utilityCounter->egressUtilityCounters.ipsec_eg_fwd_pkts;
        }
    }
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}


/*!
 * =============================================================
 * @fn adk_ipsec_get_sa_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 *
 * @param[out]   ipseSaUtilityCounter_p - IPsec related SA utility counter data structure pointer
 *               gets set by the IPsec ADK
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully fetched SA utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to retreive all IPsec related SA utility counters
 * maintained by the IPsec ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size adk_ipsec_sa_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_sa_utility_counters(adk_module_handle_t  ipsecHandle,
                                  adk_ipsec_sa_utility_counters_t  *ipsecSaUtilityCounter_p)
{
    adk_sys_handle_t      sysHandle;
    adk_status_t retValue = ADK_IPSEC_SUCCESS;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    adk_ipsec_sa_utility_counters_t      *saUtilityCounter;

    uint16_t  i, j;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    if (ipsecSaUtilityCounter_p == NULL)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    ADK_MEMSET (ipsecSaUtilityCounter_p, 0x00, sizeof(adk_ipsec_sa_utility_counters_t));

        for (i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            for(j = 0; j < ADK_IPSEC_NUM_OF_SA; j++)
            {
                saUtilityCounter = &ipsecModuleInfo_p->dpTbl->SAUtilityCnt[i][j];
                ipsecSaUtilityCounter_p->ingSaUtilityCounters.ipsec_ing_sa_tot_in_bytes
                    += saUtilityCounter->ingSaUtilityCounters.ipsec_ing_sa_tot_in_bytes;
                ipsecSaUtilityCounter_p->ingSaUtilityCounters.ipsec_ing_sa_in_flight_pkt_count
                    += saUtilityCounter->ingSaUtilityCounters.ipsec_ing_sa_in_flight_pkt_count;
                ipsecSaUtilityCounter_p->egSaUtilityCounters.ipsec_eg_sa_tot_out_bytes
                    += saUtilityCounter->egSaUtilityCounters.ipsec_eg_sa_tot_out_bytes;
                ipsecSaUtilityCounter_p->egSaUtilityCounters.ipsec_eg_sa_inflight_pkt
                    += saUtilityCounter->egSaUtilityCounters.ipsec_eg_sa_inflight_pkt;
            }
        }
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}


/*!
 * =============================================================
 * @fn adk_ipsec_print_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 * @param[in]    ipVersion   - IP version of the utility counter.
 *
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully printed utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to print all IPsec related utility counters
 * maintained by the IPsec ADK module.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_print_utility_counters(adk_module_handle_t  ipsecHandle,
                                 adk_ipsec_ip_version_t  ipVersion)
{
    adk_sys_handle_t      sysHandle;
    adk_status_t retValue = ADK_IPSEC_SUCCESS;
    adk_ipsec_utility_counters_t      utilityCounter;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    ADK_MEMSET (&utilityCounter, 0x00, sizeof(adk_ipsec_utility_counters_t));

    retValue = adk_ipsec_get_utility_counters(ipsecHandle, ipVersion, &utilityCounter);

    printf("\n==========================IPsec ingress utility counters=========================\n");
    printf("Total of ingress pdu's received : %ld\n",
               utilityCounter.ingressUtilityCounters.ipsec_ing_total_pdu_recv_cnt);
    printf("Total of ingress pdu's sent for inbound processing in Look_Aside IPsec operation: %ld\n",
               utilityCounter.ingressUtilityCounters.ipsec_ing_la_num_pdus_sent_for_inbound_processing);
    printf("Total of pdu's received after inbound processing in Look_Aside or Inline operation  : %ld\n",
               utilityCounter.ingressUtilityCounters.ipsec_ing_num_pdus_rcvd_after_inbound_processing);
    printf("Total of ingress exception sent count : %ld\n",
               utilityCounter.ingressUtilityCounters.ipsec_ing_exception_sent);
    printf("Total of all the packets that got decrypted successfully in HW : %ld\n",
               utilityCounter.ingressUtilityCounters.ipsec_ing_inline_decrypted);
    printf("Total number of initial/non-initial fragments buffered in IPsec for single SA : %ld\n",
               utilityCounter.ingressUtilityCounters.ipsec_ing_total_buffered_frags);
    printf("==========================IPsec egress utility counters=========================\n");
    printf("Total of egress pdu's received count : %ld\n",
               utilityCounter.egressUtilityCounters.ipsec_eg_total_pdu_recv_cnt);
    printf("Total of egress exceptions sent count : %ld\n",
               utilityCounter.egressUtilityCounters.ipsec_eg_exception_sent);
    printf("Total of egress pdu's sent for out_bound_processing in Look_Aside or Inline operation : %ld\n",
               utilityCounter.egressUtilityCounters.ipsec_eg_num_pdus_sent_for_outbound_processing);
    printf("Total of pdu's received after outbound processing in Look_Aside operation : %ld\n",
               utilityCounter.egressUtilityCounters.ipsec_eg_la_num_pdus_rcvd_after_outbound_processing);
    printf("Total of egress trnasmitted pdu's count : %ld\n",
               utilityCounter.egressUtilityCounters.ipsec_eg_tx_pkts);
    printf("Total of egress forward pdu's count : %ld\n",
               utilityCounter.egressUtilityCounters.ipsec_eg_fwd_pkts);
    printf("================================================================================\n");

    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * =============================================================
 * @fn adk_ipsec_print_sa_utility_counters
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 *
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully printed utility counters
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to print all IPsec related sa utility counters
 * maintained by the IPsec ADK module.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_print_sa_utility_counters(adk_module_handle_t  ipsecHandle)
{
    adk_sys_handle_t      sysHandle;
    adk_status_t retValue = ADK_IPSEC_SUCCESS;
    adk_ipsec_sa_utility_counters_t      saUtilityCounter;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    ADK_MEMSET (&saUtilityCounter, 0x00, sizeof(adk_ipsec_sa_utility_counters_t));

    retValue = adk_ipsec_get_sa_utility_counters(ipsecHandle, &saUtilityCounter);

    printf("\n==========================IPsec ingress SA utility counters=========================\n");
    printf("Total of ingress side PDU's received byte count : %ld\n",
               saUtilityCounter.ingSaUtilityCounters.ipsec_ing_sa_tot_in_bytes);
    printf("Total of ingress side inflight PDU count : %ld\n",
               saUtilityCounter.ingSaUtilityCounters.ipsec_ing_sa_in_flight_pkt_count);
    printf("==========================IPsec egress SA utility counters=========================\n");
    printf("Total of egress side PDU's transmit count : %ld\n",
               saUtilityCounter.egSaUtilityCounters.ipsec_eg_sa_tot_out_bytes);
    printf("Total of egress side in flight PDU count : %ld\n",
               saUtilityCounter.egSaUtilityCounters.ipsec_eg_sa_inflight_pkt);
    printf("================================================================================\n");


    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}
