/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

#include <adk_include.h>

/* DPDK/RDK include files */
#include <rte_common.h>
#include <rte_security.h>
#include <rte_eventdev.h>
#include <ice_dsi_ipsec.h>
#include <rte_bridge_crypto.h>
#include <adk_cp_acl_list.h>
#include <adk_hos.h>

/* ADK common headers */
#include <adk_rm.h>
#include <adk_hash.h>

/* IPsec module specific headers */
#include "adk_ipsec.h"
#include "adk_ipsec_common.h"
#include "adk_ipsec_cp.h"

#define INIT_SEQ_NUMBER 1

static void reset_stats (ipsec_module_info_t *ipsecModuleInfo_p,
                         uint32_t saIndex, adk_ipsec_sa_direction_t saDir)
{
    uint16_t i;

    if (saDir == ADK_IPSEC_OUTBOUND)
    {
        for (i = 0; i<ADK_MAX_NUM_CORES; i++)
        {
            ADK_MEMSET(&ipsecModuleInfo_p->dpTbl->egSAStats[i][saIndex],
                       0x00, sizeof(adk_ipsec_eg_sa_stat_t));
        }
    }
    else
    {
        for (i = 0; i<ADK_MAX_NUM_CORES; i++)
        {
            ADK_MEMSET(&ipsecModuleInfo_p->dpTbl->ingSAStats[i][saIndex],
                       0x00, sizeof(adk_ipsec_ing_sa_stat_t));
        }
    }
}

static void convert_32bit_addr_to_8bit_addr (struct in6_addr *byteIpv6,
                                             uint32_t *ipv6TunAddr)
{
    uint8_t i, j, k;

    j = 0;
    k = 0;
    for (i = 0; i < 16; i++)
    {
        byteIpv6->s6_addr[i] = ipv6TunAddr[j] >> (k*8);
        k++;
        if (i % 4 == 3)
        {
            j++;
            k = 0;
        }
    }
}

/* Function to map HW SA ID to SW SA ID used in ADK.
 * The mapping will be done by using hash table hwSAIdMap.
 */
static void map_hw_sa_adk_sa(ipsec_module_info_t *ipsecModuleInfo_p,
                             bool oper,
                             uint16_t adkSAId)
{
    uint32_t  hashVal, hwSAId;
    uint8_t i;
    hwSAId = ipsecModuleInfo_p->dpTbl->iSATbl[adkSAId].hwSAId;
    hashVal = hwSAId & ADK_IPSEC_HWSA_MAP_HASH_MASK;

    for(i=0; i< ADK_IPSEC_HWSA_MAP_MAX_COLLISION; i++)
    {
        if (TRUE == oper)
        {
            /* Mapping HW SA ID to SW SA ID */
            if(!ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].isValid)
            {
                ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].isValid = TRUE;
                ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].hwSAId = hwSAId;
                ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].adkSAId = adkSAId;
                return;
            }
        }
        else 
        {
            /* Unmap */
            if((ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].isValid) && 
               (adkSAId == ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].adkSAId))
            {
                ipsecModuleInfo_p->dpTbl->hwSAIdMap[hashVal][i].isValid = FALSE;
                return;
            }
        }
    }
}

/* Function to allocate crypto device and Q pair for SA.
 * Current scheme is simple - allocate the first crypto device
 * that is "available". Availability of device is purely based
 * on number of SAs that are offloaded to given device and
 * the user configured limit for that device.
 * Once the device is selected, Q pair for that device is selected
 * based on the direction of the SA and its "priority"
 */

static uint8_t allocate_qp(ipsec_module_info_t *ipsecModuleInfo_p,
                           uint8_t devId, uint8_t *qpId,
                           uint8_t  priority,
                           adk_ipsec_sa_dir direction)
{
    uint8_t j, k;
    static uint8_t  lastUsedBoth = 0xFF, lastUsedDL = 0xFF, lastUsedUL = 0xFF;
    /* now decide the Q pair id for lookaside device */
    for (j = 0; j < ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs; j++)
    {
        if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].usage ==
                ADK_SYS_RSRC_CRYPTO_DEV_QP_USE_IPSEC_BOTH)
        {
            if (ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs == 1)
            {
                *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                return j;
            }
            else
            {
                if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri == 0)
                {
                    /* need to assign Q pair in round robin manner only when
                     * no specific priorities are associated with Q pairs
                     */
                    if (lastUsedBoth == j)
                        continue;
                    else if (lastUsedBoth == ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs - 1)
                        lastUsedBoth = 0;
                    *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                    lastUsedBoth = j;
                    return j;
                }
                else
                {
                    for (k = 0; k < ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri; k++)
                    {
                        if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].pri_list[k] == priority)
                        {
                            *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                            return j;
                        }
                    }
                    if (k == ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri)
                    {
                        return 0xFF;
                    }
                }
            }
        }
        else if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].usage ==
                    ADK_SYS_RSRC_CRYPTO_DEV_QP_USE_IPSEC_DL)
        {
            if (direction == ADK_IPSEC_SA_UL)
            {
                /* not suitable crypto device */
                continue;
            }
            if (ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs == 1)
            {
                *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                return j;
            }
            else
            {
                if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri == 0)
                {
                    /* need to assign Q pair in round robin manner only when
                     * no specific priorities are associated with Q pairs
                     */
                    if (lastUsedDL == j)
                        continue;
                    else if (lastUsedDL == ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs - 1)
                        lastUsedDL = 0;
                    *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                    lastUsedDL = j;
                    return j;
                }
                else
                {
                    for (k = 0; k < ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri; k++)
                    {
                        if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].pri_list[k] == priority)
                        {
                            *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                            return j;
                        }
                    }
                    if (k == ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri)
                    {
                        return 0xFF;
                    }
                }
            }
        }
        else if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].usage ==
                    ADK_SYS_RSRC_CRYPTO_DEV_QP_USE_IPSEC_UL)
        {
            if (direction == ADK_IPSEC_SA_DL)
            {
                /* not suitable crypto device */
                continue;
            }
            if (ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs == 1)
            {
                *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                return j;
            }
            else
            {
                if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri == 0)
                {
                    /* need to assign Q pair in round robin manner only when
                     * no specific priorities are associated with Q pairs
                     */
                    if (lastUsedUL == j)
                        continue;
                    else if (lastUsedUL == ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs - 1)
                        lastUsedUL = 0;
                    *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                    lastUsedUL = j;
                    return j;
                }
                else
                {
                    for (k = 0; k < ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri; k++)
                    {
                        if (ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].pri_list[k] == priority)
                        {
                            *qpId = ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].queue_pair_id;
                            return j;
                        }
                    }
                    if (k == ipsecModuleInfo_p->cryptoDev[devId]->qp_info_p[j].num_pri)
                    {
                        return 0xFF;
                    }
                }
            }
        } /* end of usage check for QP */
    } /* end of for loop */
    if (j >= ipsecModuleInfo_p->cryptoDev[devId]->num_q_pairs)
        return 0XFF;
    else
        return j;
}
static struct rte_mempool *allocate_crypto_qp(ipsec_module_info_t *ipsecModuleInfo_p,
                                              uint8_t *devId, uint8_t *qpId,
                                              uint8_t  priority,
                                              adk_ipsec_sa_dir direction)
{
    uint8_t i, j = 0xFF;
    /* based on the number of SAs allowed for a crypto device, first decide
     * which looakside crypto device to use, there is no need to allocate
     * inline type device. As of now, there would be only one inline device
     * so all the SAs would be using it
     */
    for (i = 0; i < ipsecModuleInfo_p->numCryptoDevices; i++)
    {
        if ((ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].isValid) &&
                (ipsecModuleInfo_p->cryptoDev[i]->dev_type == ADK_SYS_RSRC_CRYPTO_DEV_TYPE_LA))
        {
            if (ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].numSAsUsing !=
                  ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].numSAsAllowed)
            {
                j = allocate_qp(ipsecModuleInfo_p, i, qpId, priority, direction);
                if (j != 0xFF)
                {
                    /* Found the correct device with correct Q */
                    break;;
                }
            }
        }
    }
    if (i == ipsecModuleInfo_p->numCryptoDevices)
    {
        /* no suitable crypto device found! */
        return NULL;
    }
    *devId = ipsecModuleInfo_p->cryptoDev[i]->dev_id;

    ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].numSAsUsing++;
    return ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].sessPool;
}

static adk_status_t allocate_sess_op_data(adk_sys_handle_t sysHandle,
                                      adk_module_handle_t ipsecHandle,
                                      ipsec_module_info_t *ipsecModuleInfo_p,
                                      uint32_t *saIndex,
                                      bool dir,
                                      adk_ipsec_sa_t *saConfig,
                                      uint8_t priority)
{
    adk_status_t  retVal = 0;
    char PoolName[18];
    struct rte_crypto_sym_xform cipher;
    struct rte_crypto_sym_xform auth;
    struct rte_crypto_sym_xform init_xform;
    uint8_t     cipherAlgo = 0, authAlgo = 0;
    uint16_t    blockSize = 0;
    uint32_t digestLen =0, ivLen = 0;
    struct rte_security_session_conf  secConf;
    struct rte_security_session *secSession = NULL;
    struct ice_dsi_crypto_session *devSecSess = NULL;
    struct rte_crypto_sym_xform aead;
    struct rte_mempool  *devSessPool = NULL;
    bool  modeInline;
    adk_ipsec_dp_lkaside_info_t *lkasideInfo = NULL;
    adk_ipsec_dp_ing_sa_t *ingSaPtr = NULL;
    adk_ipsec_dp_eg_sa_t *egSaPtr = NULL;

    ADK_MEMSET(&aead, 0, sizeof(struct rte_crypto_sym_xform));
    ADK_MEMSET(&cipher, 0, sizeof(struct rte_crypto_sym_xform));
    ADK_MEMSET(&auth, 0, sizeof(struct rte_crypto_sym_xform));
    switch (saConfig->SASPPContext.encr)
    {
    case ADK_IPSEC_ENCR_NULL:
        cipherAlgo = RTE_CRYPTO_CIPHER_NULL;;
        blockSize = 16;
        ivLen = 0;
        break;
    case ADK_IPSEC_ENCR_AES_CBC:
        cipherAlgo = RTE_CRYPTO_CIPHER_AES_CBC;;
        blockSize = 16;
        ivLen = 16;
        break;
    case ADK_IPSEC_ENCR_DES_CBC:
        cipherAlgo = RTE_CRYPTO_CIPHER_DES_CBC;
        blockSize = 8;
        ivLen = 8;
        break;
    case ADK_IPSEC_ENCR_3DES_CBC:
        cipherAlgo = RTE_CRYPTO_CIPHER_3DES_CBC;
        blockSize = 8;
        ivLen = 8;
        break;
    case ADK_IPSEC_ENCR_AES_CTR:
        cipherAlgo = RTE_CRYPTO_CIPHER_AES_CTR;
        blockSize = 16;
        ivLen = 8;
        break;
    case ADK_IPSEC_ENCR_AES_GCM_ICV8:
        cipherAlgo = RTE_CRYPTO_AEAD_AES_GCM;
        blockSize = 16;
        ivLen = 8;
        digestLen = 8;
        break;
    case ADK_IPSEC_ENCR_AES_GCM_ICV12:
        cipherAlgo = RTE_CRYPTO_AEAD_AES_GCM;
        blockSize = 16;
        ivLen = 8;
        digestLen = 12;
        break;
    case ADK_IPSEC_ENCR_AES_GCM_ICV16:
        cipherAlgo = RTE_CRYPTO_AEAD_AES_GCM;
        blockSize = 16;
        ivLen = 8;
        digestLen = 16;
        break;
    case ADK_IPSEC_ENCR_AES_CCM_ICV8:
        cipherAlgo = RTE_CRYPTO_AEAD_AES_CCM;
        blockSize = 16;
        ivLen = 8;
        digestLen = 8;
        break;
    case ADK_IPSEC_ENCR_AES_CCM_ICV12:
        cipherAlgo = RTE_CRYPTO_AEAD_AES_CCM;
        blockSize = 16;
        ivLen = 8;
        digestLen = 12;
        break;
    case ADK_IPSEC_ENCR_AES_CCM_ICV16:
        cipherAlgo = RTE_CRYPTO_AEAD_AES_CCM;
        blockSize = 16;
        ivLen = 8;
        digestLen = 16;
        break;
    case ADK_IPSEC_ENCR_NULL_GMAC:
        cipherAlgo = RTE_CRYPTO_CIPHER_NULL;
        ivLen = 8;
        blockSize = 16;
        break;
    default:
        break;
    }

    switch (saConfig->SASPPContext.auth)
    {
    case ADK_IPSEC_AUTH_NULL:
        authAlgo = RTE_CRYPTO_AUTH_NULL;
        break;
    case ADK_IPSEC_AUTH_HMAC_SHA1:
        authAlgo = RTE_CRYPTO_AUTH_SHA1_HMAC;
        digestLen = 12;
        break;
    case ADK_IPSEC_AUTH_HMAC_SHA_256:
        authAlgo = RTE_CRYPTO_AUTH_SHA256_HMAC;
        digestLen = 16;
        break;
    case ADK_IPSEC_AUTH_HMAC_SHA_384:
        authAlgo = RTE_CRYPTO_AUTH_SHA384_HMAC;
        digestLen = 24;
        break;
    case ADK_IPSEC_AUTH_HMAC_SHA_512:
        authAlgo = RTE_CRYPTO_AUTH_SHA512_HMAC;
        digestLen = 32;
        break;
    case ADK_IPSEC_AUTH_HMAC_MD5:
        authAlgo = RTE_CRYPTO_AUTH_MD5_HMAC;
        digestLen = 12;
        break;
    case ADK_IPSEC_AUTH_AES_XCBC_MAC:
        authAlgo = RTE_CRYPTO_AUTH_AES_XCBC_MAC;
        digestLen = 16;
        break;
    case ADK_IPSEC_AUTH_AES_CMAC :
        authAlgo = RTE_CRYPTO_AUTH_AES_CMAC;
        digestLen = 16;
        break;
    case ADK_IPSEC_AUTH_AES_128_GMAC:
    case ADK_IPSEC_AUTH_AES_192_GMAC:
    case ADK_IPSEC_AUTH_AES_256_GMAC:
        authAlgo = RTE_CRYPTO_AUTH_AES_GMAC;
        digestLen = 16;
        break;
    default:
        break;
    }

    if ( (saConfig->SASPPContext.encr == ADK_IPSEC_ENCR_AES_GCM_ICV8)  ||
         (saConfig->SASPPContext.encr == ADK_IPSEC_ENCR_AES_GCM_ICV12) ||
         (saConfig->SASPPContext.encr == ADK_IPSEC_ENCR_AES_GCM_ICV16) ||
         (saConfig->SASPPContext.encr == ADK_IPSEC_ENCR_AES_CCM_ICV8)  ||
         (saConfig->SASPPContext.encr == ADK_IPSEC_ENCR_AES_CCM_ICV12) ||
         (saConfig->SASPPContext.encr == ADK_IPSEC_ENCR_AES_CCM_ICV16) )
    {
        aead.type = RTE_CRYPTO_SYM_XFORM_AEAD;
        aead.aead.algo = cipherAlgo;
        aead.aead.key.data = (uint8_t *) saConfig->SASPPContext.encrKey.buf;
        aead.aead.key.length = saConfig->SASPPContext.encrKey.len;
        aead.aead.iv.offset = IV_OFFSET;
        aead.aead.iv.length = ivLen;
        aead.aead.digest_length = digestLen;
        aead.aead.aad_length = sizeof(struct esp_header);
        init_xform = aead;
        init_xform.next = NULL;
        if (dir == ADK_IPSEC_OUTBOUND)
        {
            aead.aead.op = RTE_CRYPTO_AEAD_OP_ENCRYPT;
        }
        else
        {
            aead.aead.op = RTE_CRYPTO_AEAD_OP_DECRYPT;
        }
    }
    else
    {
        cipher.type = RTE_CRYPTO_SYM_XFORM_CIPHER;
        cipher.cipher.algo = cipherAlgo;
        cipher.cipher.key.data = (uint8_t *) saConfig->SASPPContext.encrKey.buf;
        cipher.cipher.key.length = saConfig->SASPPContext.encrKey.len;
        cipher.cipher.iv.offset = IV_OFFSET;
        cipher.cipher.iv.length = ivLen;
        if (dir == ADK_IPSEC_OUTBOUND)
        {
            cipher.cipher.op = RTE_CRYPTO_CIPHER_OP_ENCRYPT;
            cipher.next = &auth;
            init_xform = cipher;
        }
        else
        {
            cipher.cipher.op = RTE_CRYPTO_CIPHER_OP_DECRYPT;
            cipher.next = NULL;
        }

        auth.type = RTE_CRYPTO_SYM_XFORM_AUTH;
        auth.auth.algo = authAlgo;
        auth.auth.key.data = (uint8_t *) saConfig->SASPPContext.authKey.buf;
        auth.auth.key.length = saConfig->SASPPContext.authKey.len;
        auth.auth.digest_length = digestLen;
        if (dir == ADK_IPSEC_OUTBOUND)
        {
            auth.auth.op = RTE_CRYPTO_AUTH_OP_GENERATE;
            auth.next = NULL;
        }
        else
        {
            auth.auth.op = RTE_CRYPTO_AUTH_OP_VERIFY;
            auth.next = &cipher;
            init_xform = auth;
        }
    }
    if (dir == ADK_IPSEC_OUTBOUND)
    {
        egSaPtr = &ipsecModuleInfo_p->dpTbl->eSATbl[*saIndex];
        egSaPtr->blockSize = blockSize;
        egSaPtr->digestLen = digestLen;
        egSaPtr->ivLen = ivLen;
    }
    else
    {
        ingSaPtr = &ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex];
        ingSaPtr->ivLen = ivLen;
        ingSaPtr->digestLen = digestLen;
    }
    
    if (ADK_IPSEC_SA_MODE_INLINE == saConfig->saMode)
        modeInline = true;
    else
        modeInline = false;

    /* In case of inbound dir, if SA mode is set as inline, packets may be processed with Inline or Lookaside
     * In case of outbound dir, if SA mode is set as inline, packets should be processed only with Inline.
     */
    if (modeInline)
    {
        if (ipsecModuleInfo_p->secCtxCP)
        {
            ADK_MEMSET(&secConf, 0, sizeof(struct rte_security_session_conf));
            secConf.action_type = RTE_SECURITY_ACTION_TYPE_INLINE_CRYPTO;
            secConf.protocol = RTE_SECURITY_PROTOCOL_IPSEC;
            /* SA information */
            secConf.ipsec.spi = saConfig->SASPPContext.spi;
            if (saConfig->SASPPContext.salt.len)
            {
                ADK_MEMCPY(&secConf.ipsec.salt,
                        saConfig->SASPPContext.salt.buf,
                        saConfig->SASPPContext.salt.len);
            }
            secConf.ipsec.options.esn = saConfig->SASPPContext.extSeqNum;
            secConf.ipsec.options.udp_encap = saConfig->isNATTEnable;
            if (dir == ADK_IPSEC_OUTBOUND)
                secConf.ipsec.direction = RTE_SECURITY_IPSEC_SA_DIR_EGRESS;
            else
                secConf.ipsec.direction = RTE_SECURITY_IPSEC_SA_DIR_INGRESS;
            secConf.ipsec.proto = RTE_SECURITY_IPSEC_SA_PROTO_ESP;
            secConf.ipsec.mode = RTE_SECURITY_IPSEC_SA_MODE_TUNNEL;

            if (saConfig->ipVersion == ADK_IPSEC_IPV4)
            {
                secConf.ipsec.tunnel.type = RTE_SECURITY_IPSEC_TUNNEL_IPV4;
                secConf.ipsec.tunnel.ipv4.src_ip.s_addr = saConfig->tunAddr.ipv4TunAddr.tunnelSrcIP;
                secConf.ipsec.tunnel.ipv4.dst_ip.s_addr = saConfig->tunAddr.ipv4TunAddr.tunnelDstIP;
            }
            else
            {
                secConf.ipsec.tunnel.type = RTE_SECURITY_IPSEC_TUNNEL_IPV6;
                convert_32bit_addr_to_8bit_addr (&secConf.ipsec.tunnel.ipv6.src_addr,
                        saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP);
                convert_32bit_addr_to_8bit_addr (&secConf.ipsec.tunnel.ipv6.dst_addr,
                        saConfig->tunAddr.ipv6TunAddr.tunnelDstIP);
            }

            secConf.crypto_xform = &init_xform;
            ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_security_session_create", "Create "
                    "security session for given security context and session mempool\n");
            secSession = rte_security_session_create(ipsecModuleInfo_p->secCtxCP, &secConf,
                    ipsecModuleInfo_p->dpTbl->inlineSessPool);
            if (secSession == NULL)
            {
                if (ADK_IPSEC_INBOUND == dir)
                {
                    modeInline = false;
                    /* Since modeInline is false, Lookaside should be attempted */
                    ADK_IPSEC_LOG_INFO(ipsecHandle, "Failed to create session\n");
                    ADK_IPSEC_LOG_INFO(ipsecHandle, "\n All packets on SA = %d would be processed in lookaside mode \n",
                            *saIndex);
                }
                else
                {
                    ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                            "\n Failed to create security session. Security session is needed since SA Mode is set as Inline in Outbound direction\n");
                    return ADK_E_RESOURCE_UNAVAILABLE; 
                }
            }
            else
            {
                devSecSess = (struct ice_dsi_crypto_session *)get_sec_session_private_data(secSession);
                if (devSecSess == NULL)
                {
                    if (ADK_IPSEC_INBOUND == dir)
                    {
                        /* Since modeInline is false, Lookaside should be attempted */
                        modeInline = false;
                        ADK_IPSEC_LOG_INFO(ipsecHandle, "Failed to fetch private session data\n");
                        ADK_IPSEC_LOG_INFO(ipsecHandle, "\n All packets on SA = %d would be processed in lookaside mode \n",
                                *saIndex);
                    }
                    else
                    {
                        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                                "Session private data is needed since SA Mode is set as Inline in Outbound direction\n");
                        return ADK_E_RESOURCE_UNAVAILABLE; 
                    }
                }
                else
                {
                    if (ADK_IPSEC_OUTBOUND == dir)
                    {
                        egSaPtr->secSession = secSession;
                    }
                    else
                    {    
                        ingSaPtr->secSession = secSession;
                        ingSaPtr->hwSAId = devSecSess->sa_idx;
                    }
                }
            }

            /*  Allocate index for Tx notification counters for SAs with
             *  possibility of inline operations. This is valid only for
             *  outbound
             */

            if ((dir == ADK_IPSEC_OUTBOUND) && (modeInline == true))
            {
                retVal = adk_sys_egress_stats_id_alloc (sysHandle,
                        &ipsecModuleInfo_p->dpTbl->eSATbl[*saIndex].egressStatsId);
                if (retVal != ADK_SUCCESS)
                {
                    retVal = ADK_ERROR_SEV_LEVEL_SET (ADK_E_NS_RSRC_ALLOCATE_FAILED);
                    ADK_IPSEC_LOG_ERROR (ipsecHandle, retVal,
                            "Failed to allocate egress stats ID for SA");
                    return retVal;
                }
            }
        }
        else
        {
            if (ADK_IPSEC_INBOUND == dir)
            {
                modeInline = false;
                ADK_IPSEC_LOG_INFO(ipsecHandle, "\n Inline device not registered as yet \n");
                ADK_IPSEC_LOG_INFO(ipsecHandle, "\n All packets on SA = %d would be processed in lookaside mode \n",
                        *saIndex);
            }
            else
            {
                ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                        "Inline device should be registered since SA mode is Inline in Outbound direction\n");
                return ADK_E_RESOURCE_UNAVAILABLE;
            }
        }
    } /* modeInline */ 
    
    /* In case of inbound and outbound, if SA mode is set as Lookaside, then packets 
     * should be processed only with Lookaside.
     */
    if (!modeInline || 
        (ADK_IPSEC_INBOUND == dir))
    {
        if (authAlgo == RTE_CRYPTO_AUTH_AES_GMAC)
        {
            auth.next = NULL;
            init_xform = auth;
        }

        lkasideInfo = ADK_DP_SHM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_DP_TBL_OBJ,
                                        1, sizeof(adk_ipsec_dp_lkaside_info_t), 0);
        if (NULL == lkasideInfo)
        {
            return ADK_E_RESOURCE_UNAVAILABLE;
        }

        devSessPool = allocate_crypto_qp(ipsecModuleInfo_p,
                &lkasideInfo->cryptoDevId, &lkasideInfo->cryptoQP,
                priority,
                (dir == ADK_IPSEC_OUTBOUND)? ADK_IPSEC_SA_UL:ADK_IPSEC_SA_DL);
        if (devSessPool == NULL)
        {
            /* TODO in case of inline, need to delete added SA?
            */
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                    "allocate_sess_op_data : failed to allocate crypto DEV/QP\n");
            ADK_DP_SHM_FREE(lkasideInfo);
            return ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }
        ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_cryptodev_sym_session_create", "Create "
                "cryptodev lookaside session\n");
        lkasideInfo->cryptoSession =
            rte_cryptodev_sym_session_create(devSessPool);
        if (lkasideInfo->cryptoSession == NULL)
        {
            /* TODO in case of inline, need to delete added SA?
            */
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                    "allocate_sess_op_data :Unable to create session\n");
            ADK_DP_SHM_FREE(lkasideInfo);
            return ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }

        ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_cryptodev_sym_session_init", "Initialize "
                "lookaside session with initial crypto transforms\n");
        retVal = rte_cryptodev_sym_session_init(lkasideInfo->cryptoDevId,
                lkasideInfo->cryptoSession,
                &init_xform,
                devSessPool);
        if (retVal != 0)
        {
            /* TODO in case of inline, need to delete added SA?
            */
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                    "allocate_sess_op_data :session init failed\n");
            ADK_DP_SHM_FREE(lkasideInfo);
            return ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }
        /* Create crypto mempool and allocate object from mempool for first crypto op */
        if (dir == ADK_IPSEC_OUTBOUND)
            snprintf(PoolName, 16, "egCryptoPool_%d", *saIndex);
        else
            snprintf(PoolName, 16, "ingCryptoPool_%d", *saIndex);

        ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_crypto_op_pool_create", "Create "
                "opdata pool for lookaside sessions\n");
        lkasideInfo->cryptoPool = rte_crypto_op_pool_create(PoolName,
                RTE_CRYPTO_OP_TYPE_SYMMETRIC, 128,
                0, ADK_IPSEC_MAXIMUM_IV_LENGTH, rte_socket_id());
        if (lkasideInfo->cryptoPool == NULL)
        {
            /* TODO in case of inline, need to delete added SA?
            */
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                    "allocate_sess_op_data :Unable to create OP pool :%s\n",
                    PoolName);
            ADK_DP_SHM_FREE(lkasideInfo);
            return ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }

        ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_crypto_op_alloc", "Allocate "
                "one entity of opdata from opdata pool\n");
        lkasideInfo->cryptoOp = rte_crypto_op_alloc(lkasideInfo->cryptoPool,
                RTE_CRYPTO_OP_TYPE_SYMMETRIC);
        if (lkasideInfo->cryptoOp == NULL)
        {
            /* TODO in case of inline, need to delete added SA?
            */
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_RESOURCE_UNAVAILABLE,
                    "allocate_sess_op_data :Unable to allocate OP pool :%s\n",
                    PoolName);
            ADK_DP_SHM_FREE(lkasideInfo);
            return ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }
    }
    if (ADK_IPSEC_OUTBOUND == dir)
    {
        egSaPtr->saMode = (modeInline) ? ADK_IPSEC_SA_INLINE : ADK_IPSEC_SA_LOOKASIDE;
        egSaPtr->lkasideInfo = lkasideInfo;
        if (ADK_IPSEC_SA_LOOKASIDE == egSaPtr->saMode)
        {
            rte_atomic32_init(&egSaPtr->lkasideInfo->seqNumber);
            rte_atomic32_set(&egSaPtr->lkasideInfo->seqNumber, 0);
        }
    }
    else
    {
       ingSaPtr->saMode = (modeInline) ? ADK_IPSEC_SA_BOTH : ADK_IPSEC_SA_LOOKASIDE;
       ingSaPtr->lkasideInfo = lkasideInfo;
    }
    return ADK_SUCCESS;
}

/* Function to fill the egress SA info required in DP 
 * based on application given SA config.
 */
static void 
adk_ipsec_fill_dp_eg_sa_info(adk_ipsec_dp_eg_sa_t *egSaPtr, adk_ipsec_sa_t *saConfig)
{
    if (ADK_IPSEC_IPV6 == saConfig->ipVersion)
        egSaPtr->ipVersion = 1;
    else
        egSaPtr->ipVersion = 0;
    
    egSaPtr->isNATTEnable = saConfig->isNATTEnable;
    egSaPtr->seqRollover = saConfig->SASPPContext.sn.seqRollover;
    egSaPtr->extSeqNum  =  saConfig->SASPPContext.extSeqNum;
    egSaPtr->lifetimesEnabled = saConfig->SASPPContext.lifetimesEnabled;
    egSaPtr->encr = saConfig->SASPPContext.encr;
    egSaPtr->spi = adk_htonl(saConfig->SASPPContext.spi);
    /* explicitly set the seqNumber to 1 */
    egSaPtr->seqNumber = INIT_SEQ_NUMBER;
    egSaPtr->hardByteLifetime = saConfig->SASPPContext.hardByteLifetime;
    egSaPtr->softByteLifetime = saConfig->SASPPContext.softByteLifetime;
    egSaPtr->flow_label_value = saConfig->flowLabelInfo.flow_label_value;
    egSaPtr->bytesUsed = 0;
    if (ADK_IPSEC_IPV4 == saConfig->ipVersion)
    {
        egSaPtr->saSIP[0] = adk_htonl(saConfig->tunAddr.ipv4TunAddr.tunnelSrcIP);
        egSaPtr->saDIP[0] = adk_htonl(saConfig->tunAddr.ipv4TunAddr.tunnelDstIP);
    }
    else
    {
        ADK_MEMCPY(egSaPtr->saSIP, saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP, 16);
        ADK_MEMCPY(egSaPtr->saDIP, saConfig->tunAddr.ipv6TunAddr.tunnelDstIP, 16);
    }
    if (saConfig->isNATTEnable)
    {
        egSaPtr->lkasideInfo->NATTSrcPort = adk_htons(saConfig->NATTContext.NATTSrcPort);
        egSaPtr->lkasideInfo->NATTDstPort = adk_htons(saConfig->NATTContext.NATTDstPort);
        egSaPtr->lkasideInfo->NATTSrcAddr = adk_htonl(saConfig->NATTContext.origSrcAddr);
        egSaPtr->lkasideInfo->NATTDstAddr = adk_htonl(saConfig->NATTContext.origDstAddr);
    }
    if (ADK_IPSEC_IPV4 == saConfig->ipVersion)
    {
        if (saConfig->eSaTttlHopLimit.ttl == 0)
            egSaPtr->ttl_hopLimit = IPSEC_SA_DEFAULT_TTL_HOPLIMIT_VALUE;
        else
            egSaPtr->ttl_hopLimit = saConfig->eSaTttlHopLimit.ttl;
    }
    else
    {
        if (saConfig->eSaTttlHopLimit.hopLimit == 0)
            egSaPtr->ttl_hopLimit = IPSEC_SA_DEFAULT_TTL_HOPLIMIT_VALUE;
        else
            egSaPtr->ttl_hopLimit = saConfig->eSaTttlHopLimit.hopLimit;
    }
}

adk_status_t adk_ipsec_add_eSA(adk_sys_handle_t sysHandle,
                               adk_module_handle_t ipsecHandle,
                               ipsec_module_info_t *ipsecModuleInfo_p,
                               uint32_t *saIndex,
                               adk_ipsec_sa_t *saConfig,
                               uint8_t  dscpVal)
{
    adk_status_t retVal = ADK_IPSEC_SUCCESS;

    retVal = allocate_sess_op_data(sysHandle, ipsecHandle, ipsecModuleInfo_p,
                                   saIndex, ADK_IPSEC_OUTBOUND,
                                   saConfig, dscpVal);
    if (retVal != ADK_SUCCESS)
    {
        return ADK_E_RESOURCE_UNAVAILABLE;
    }
    ADK_MEMCPY(&ipsecModuleInfo_p->eSAConfig[*saIndex], saConfig,
            sizeof(adk_ipsec_sa_t));
    if (ADK_IPSEC_ENCR_NULL != saConfig->SASPPContext.encr)
    {
        ipsecModuleInfo_p->eSAConfig[*saIndex].SASPPContext.encrKey.buf =
                ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_KEY_OBJ,
                               1, saConfig->SASPPContext.encrKey.len);
        ADK_MEMCPY(ipsecModuleInfo_p->eSAConfig[*saIndex].SASPPContext.encrKey.buf,
                   saConfig->SASPPContext.encrKey.buf,
                   saConfig->SASPPContext.encrKey.len);
    }
    if (ADK_IPSEC_AUTH_NULL != saConfig->SASPPContext.auth)
    {
        ipsecModuleInfo_p->eSAConfig[*saIndex].SASPPContext.authKey.buf =
                ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_KEY_OBJ,
                               1, saConfig->SASPPContext.authKey.len);
        ADK_MEMCPY(ipsecModuleInfo_p->eSAConfig[*saIndex].SASPPContext.authKey.buf,
                   saConfig->SASPPContext.authKey.buf,
                   saConfig->SASPPContext.authKey.len);
    }
    if (saConfig->SASPPContext.salt.len)
    {
        ipsecModuleInfo_p->eSAConfig[*saIndex].SASPPContext.salt.buf =
                ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_KEY_OBJ,
                               1, saConfig->SASPPContext.salt.len);
        ADK_MEMCPY(ipsecModuleInfo_p->eSAConfig[*saIndex].SASPPContext.salt.buf,
                   saConfig->SASPPContext.salt.buf,
                   saConfig->SASPPContext.salt.len);
    }
    adk_ipsec_fill_dp_eg_sa_info(&ipsecModuleInfo_p->dpTbl->eSATbl[*saIndex],
                                 saConfig);
    reset_stats (ipsecModuleInfo_p, *saIndex, ADK_IPSEC_OUTBOUND);

    return retVal;
}

adk_status_t adk_ipsec_stop_eSA(adk_module_handle_t ipsecHandle,
                        ipsec_module_info_t *ipsecModuleInfo_p,
                        uint32_t            saIdx,
                        uint8_t             numPol,
                        uint16_t            startPolIndex)
{
    uint16_t polIdx;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;
    adk_status_t retValue = ADK_IPSEC_SUCCESS;
    ADK_IPSEC_LOG_INFO(ipsecHandle, "\n STOP_SA :: %d\n", saIdx);

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
    if(numPol)
    {
        uint16_t index=0;
        while(ADK_SUCCESS == adk_acl_next_rule_get(
                    ipsecModuleInfo_p->cpSpdAclList, &cpRule))
        {
            if(index==startPolIndex)
            {
                ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
                break;
            }
            index++;
        }
        if (NULL == ipsecSecPolLocal)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                    "The policy is not found in CP ACL List\n");
            return ADK_E_IPSEC_POLICY_NOT_FOUND;
        }
        for (index=0; index<numPol; index++)
        {
            polIdx = ipsecSecPolLocal->policyResourceID;
            /* here the first step should be to stop accpeting packets on this
             * SA. The way to do that is to set state
             * to invalid in polSaInfo
             */
            if (saIdx ==
                    ipsecModuleInfo_p->dpTbl->polSAInfo[polIdx].eSAIdx)
            {
                ADK_IPSEC_LOG_INFO(ipsecHandle,
                        "\n STOP_SA :: stop traffic on this SA\n");
                ipsecModuleInfo_p->dpTbl->polSAInfo[polIdx].eSAstate =
                    INVALID_SA;
            }
            retValue = adk_acl_next_rule_get(ipsecModuleInfo_p->cpSpdAclList, &cpRule);
            /* Here, we are not resetting to HEAD in the CP ACL list. So we should return
             * error only when we didn't iterate the loop till numPol but returned value is
             * ADK_E_LIST_NOT_FOUND.
             */
            if ((ADK_SUCCESS != retValue) && (index+1 < numPol))
            {
                ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                        "The policy is not found in CP ACL List\n");
                return ADK_E_IPSEC_POLICY_NOT_FOUND;
            }
            ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
        }
    }
    return ADK_IPSEC_SUCCESS;
}

adk_status_t
adk_ipsec_del_eSA(adk_sys_handle_t sysHandle,
                  adk_module_handle_t ipsecHandle,
                  ipsec_module_info_t *ipsecModuleInfo_p,
                  uint32_t            saIdx)
{
    uint8_t i;
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipsec_dp_lkaside_info_t *lkasideInfo = ipsecModuleInfo_p->dpTbl->eSATbl[saIdx].lkasideInfo;
    ADK_IPSEC_LOG_INFO(ipsecHandle, "\n DEL_SA :: %d\n", saIdx);

    if (lkasideInfo != NULL)
    {
        if(lkasideInfo->cryptoOp)
        {
            ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_crypto_op_free", "Free "
                    "cryptodev operation buffer\n");
            rte_crypto_op_free(lkasideInfo->cryptoOp);
        }
        if(lkasideInfo->cryptoSession)
        {
            ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_cryptodev_sym_session_free", "Free "
                    "cryptodev lookaside session\n");
            rte_cryptodev_sym_session_clear(lkasideInfo->cryptoDevId,
                    lkasideInfo->cryptoSession);
            rte_cryptodev_sym_session_free(
                    lkasideInfo->cryptoSession);
        }
        ADK_DP_SHM_FREE(lkasideInfo);
        lkasideInfo = NULL;
    }
    if (ADK_IPSEC_ENCR_NULL != ipsecModuleInfo_p->eSAConfig[saIdx].SASPPContext.encr)
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->eSAConfig[saIdx].SASPPContext.encrKey.buf);
    }
    if (ADK_IPSEC_AUTH_NULL != ipsecModuleInfo_p->eSAConfig[saIdx].SASPPContext.auth)
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->eSAConfig[saIdx].SASPPContext.authKey.buf);
    }
    if (ipsecModuleInfo_p->eSAConfig[saIdx].SASPPContext.salt.len)
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->eSAConfig[saIdx].SASPPContext.salt.buf);
    }
    if (ipsecModuleInfo_p->secCtxCP)
    {
        if (ipsecModuleInfo_p->dpTbl->eSATbl[saIdx].secSession)
        {
            ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_security_session_destroy", "Destroy "
                            "inline security session\n");

            rte_security_session_destroy(ipsecModuleInfo_p->secCtxCP,
                                ipsecModuleInfo_p->dpTbl->eSATbl[saIdx].secSession);
            ipsecModuleInfo_p->dpTbl->eSATbl[saIdx].secSession = NULL;
        }
    }
    /* change crypto dev id count as well */
    for (i = 0; i < ipsecModuleInfo_p->numCryptoDevices; i++)
    {
        if ((lkasideInfo != NULL) &&
            (ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].devId ==
                               lkasideInfo->cryptoDevId))
            ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].numSAsUsing--;
    }

    if (ipsecModuleInfo_p->dpTbl->eSATbl[saIdx].saMode == ADK_IPSEC_SA_INLINE)
    {
        retValue = adk_sys_egress_stats_id_free (sysHandle,
                                ipsecModuleInfo_p->dpTbl->eSATbl[saIdx].egressStatsId);

        if (retValue != ADK_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR (ipsecHandle, retValue,
                                "Failed to free egress stats ID for SA\n");
            return ADK_E_INTERNAL_ERROR;
        }
    }

    if (adk_rm_resource_isset (ipsecModuleInfo_p->eSASetHdl, saIdx))
    {
        adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIdx);
    }

    return ADK_SUCCESS;
}

static void adk_ipsec_init_arw_params(ipsec_module_info_t *ipsecModuleInfo_p,
                                      uint32_t saIndex,
                                      adk_ipsec_arw_size_t winSize)
{
    uint32_t  winSizeBytes, winSizeDwords;

    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].window_size = winSize;
    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].window_position = 0;
    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].buffer_head = 0;
    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].buffer_size = 0;

    winSizeBytes = (winSize / 8) + ((winSize % 8) ? 1 : 0);
    winSizeDwords = (winSizeBytes / 4) + ((winSizeBytes % 4) ? 1 : 0);

    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].window_buffer =
                              ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_ARW_OBJ, 1, winSizeDwords * 4 );
}

adk_status_t
adk_ipsec_add_iSA(adk_sys_handle_t sysHandle,
                  adk_module_handle_t ipsecHandle,
                  ipsec_module_info_t *ipsecModuleInfo_p,
                  uint32_t *saIndex,
                  adk_ipsec_sa_t *saConfig,
                  uint8_t dscpVal)
{
    uint32_t hashVal, i;
    adk_status_t retVal = ADK_SUCCESS;

    retVal = allocate_sess_op_data(sysHandle, ipsecHandle, ipsecModuleInfo_p,
                                   saIndex,
                                   ADK_IPSEC_INBOUND,
                                   saConfig, dscpVal
                                  );
    if((ADK_SUCCESS == retVal) || ((saConfig->saMode == ADK_IPSEC_SA_MODE_INLINE)
       && (ADK_E_LA_CRYPTO_CONFIG_FAILED == retVal)))
    {
        hashVal = SISD_CRC_HASH(0, adk_htonl(saConfig->SASPPContext.spi));
        hashVal &=  ADK_IPSEC_SPI_TBL_HASH_MASK;

        if (ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].secSession)
        {
            map_hw_sa_adk_sa(ipsecModuleInfo_p, TRUE, *saIndex);
        }
        for(i=0; i< ADK_IPSEC_SPI_TBL_MAX_COLLISION; i++)
        {
            if(ipsecModuleInfo_p->dpTbl->spiHashTbl[hashVal][i].isValid != TRUE)
            {
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].iSAIndex = *saIndex;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].dscpVal = dscpVal;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].bytesUsed = 0;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].antiReplayEnable = saConfig->SASPPContext.sn.antiReplayEnable;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].lifetimesEnabled = saConfig->SASPPContext.lifetimesEnabled;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].hardByteLifetime = saConfig->SASPPContext.hardByteLifetime;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].softByteLifetime = saConfig->SASPPContext.softByteLifetime;
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].spi = adk_htonl(saConfig->SASPPContext.spi);
                ipsecModuleInfo_p->dpTbl->iSATbl[*saIndex].spIDEx = saConfig->spIDEx;
                ipsecModuleInfo_p->dpTbl->spiHashTbl[hashVal][i].isValid = TRUE;
                ipsecModuleInfo_p->dpTbl->spiHashTbl[hashVal][i].idx = *saIndex;
                adk_ipsec_init_arw_params(ipsecModuleInfo_p,
                        *saIndex, saConfig->SASPPContext.antiReplayWindowSize);
                reset_stats (ipsecModuleInfo_p, *saIndex, ADK_IPSEC_INBOUND);
            }
        }
        ADK_MEMCPY(&ipsecModuleInfo_p->iSAConfig[*saIndex], saConfig,
                sizeof(adk_ipsec_sa_t));
        if (ADK_IPSEC_ENCR_NULL != saConfig->SASPPContext.encr)
        {
            ipsecModuleInfo_p->iSAConfig[*saIndex].SASPPContext.encrKey.buf =
                ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_KEY_OBJ,
                        1, saConfig->SASPPContext.encrKey.len);
            ADK_MEMCPY(ipsecModuleInfo_p->iSAConfig[*saIndex].SASPPContext.encrKey.buf,
                    saConfig->SASPPContext.encrKey.buf,
                    saConfig->SASPPContext.encrKey.len);
        }
        if (ADK_IPSEC_AUTH_NULL != saConfig->SASPPContext.auth)
        {
            ipsecModuleInfo_p->iSAConfig[*saIndex].SASPPContext.authKey.buf =
                ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_KEY_OBJ,
                        1, saConfig->SASPPContext.authKey.len);
            ADK_MEMCPY(ipsecModuleInfo_p->iSAConfig[*saIndex].SASPPContext.authKey.buf,
                    saConfig->SASPPContext.authKey.buf,
                    saConfig->SASPPContext.authKey.len);
        }
        if (saConfig->SASPPContext.salt.len)
        {
            ipsecModuleInfo_p->iSAConfig[*saIndex].SASPPContext.salt.buf =
                ADK_NVM_CALLOC(ADK_IPSEC_MODULE_ID, ADK_IPSEC_KEY_OBJ,
                        1, saConfig->SASPPContext.salt.len);
            ADK_MEMCPY(ipsecModuleInfo_p->iSAConfig[*saIndex].SASPPContext.salt.buf,
                    saConfig->SASPPContext.salt.buf,
                    saConfig->SASPPContext.salt.len);
        }
    }
    return retVal;
}

adk_status_t adk_ipsec_del_iSA(adk_sys_handle_t sysHandle,
                        adk_module_handle_t ipsecHandle,
                        ipsec_module_info_t *ipsecModuleInfo_p,
                        uint32_t            saIdx)
{
    uint16_t i;
    ADK_IPSEC_LOG_INFO(ipsecHandle, "\n stop_isa :: %d\n", saIdx);

    adk_ipsec_dp_lkaside_info_t *lkasideInfo = ipsecModuleInfo_p->dpTbl->iSATbl[saIdx].lkasideInfo;
    if (lkasideInfo != NULL)
    {
        if(lkasideInfo->cryptoOp)
        {
            ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_crypto_op_free", "Free "
                    "cryptodev operation buffer\n");
            rte_crypto_op_free(lkasideInfo->cryptoOp);
        }
        if(lkasideInfo->cryptoSession)
        {
            ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_cryptodev_sym_session_free", "Free "
                    "cryptodev lookaside session\n");
            rte_cryptodev_sym_session_clear(lkasideInfo->cryptoDevId,
                                            lkasideInfo->cryptoSession);
            rte_cryptodev_sym_session_free(
                    lkasideInfo->cryptoSession);
        }
        ADK_DP_SHM_FREE(lkasideInfo);
        lkasideInfo = NULL;
    }

    if (ADK_IPSEC_ENCR_NULL != ipsecModuleInfo_p->iSAConfig[saIdx].SASPPContext.encr)
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->iSAConfig[saIdx].SASPPContext.encrKey.buf); 
    }
    
    if (ADK_IPSEC_AUTH_NULL != ipsecModuleInfo_p->iSAConfig[saIdx].SASPPContext.auth)
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->iSAConfig[saIdx].SASPPContext.authKey.buf);
    }
    
    if (ipsecModuleInfo_p->iSAConfig[saIdx].SASPPContext.salt.len)
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->iSAConfig[saIdx].SASPPContext.salt.buf);
    }

    /* change crypto dev id count as well */
    for (i = 0; i < ipsecModuleInfo_p->numCryptoDevices; i++)
    {
        if ((lkasideInfo != NULL) && 
            (ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].devId ==
                               lkasideInfo->cryptoDevId))
            ipsecModuleInfo_p->dpTbl->cryptoDevMap[i].numSAsUsing--;
    }

    if (adk_rm_resource_isset (ipsecModuleInfo_p->iSASetHdl, saIdx))
    {
        ADK_NVM_FREE(ipsecModuleInfo_p->dpTbl->iSATbl[saIdx].window_buffer);
        adk_rm_resource_free(ipsecModuleInfo_p->iSASetHdl, saIdx);
    }
    return ADK_IPSEC_SUCCESS;
}

void
adk_ipsec_invalidate_iSA(adk_sys_handle_t    sysHandle,
                  adk_module_handle_t ipsecHandle,
                  ipsec_module_info_t *ipsecModuleInfo_p,
                  uint32_t            saIdx)
{
    uint32_t hash_val, firstHash, i;
    adk_ipsec_dp_ing_sa_t *spiInfo;

    ADK_IPSEC_LOG_INFO(ipsecHandle, "\n invalidate_isa :: %d\n", saIdx);

    /* remove it from HASH tbl */
    spiInfo = &ipsecModuleInfo_p->dpTbl->iSATbl[saIdx];
    hash_val = SISD_CRC_HASH(0, spiInfo->spi);
    firstHash = hash_val & ADK_IPSEC_SPI_TBL_HASH_MASK;

    for(i=0; i< ADK_IPSEC_SPI_TBL_MAX_COLLISION; i++)
    {
        if(ipsecModuleInfo_p->dpTbl->spiHashTbl[firstHash][i].isValid == TRUE)
        {
            ipsecModuleInfo_p->dpTbl->spiHashTbl[firstHash][i].isValid = FALSE;
            ipsecModuleInfo_p->dpTbl->spiHashTbl[firstHash][i].idx = INVALID_SA;
        }
    }
    /*  If the SA mode isn't lookaside, the entry can be deleted
     *  from the CPM to gate future packets. This is the inline equivalent
     *  of the first stage of SA lookup validation in inbound processing
     */
    if (ipsecModuleInfo_p->dpTbl->iSATbl[saIdx].saMode ==
                ADK_IPSEC_SA_BOTH)
    {
        if (ipsecModuleInfo_p->secCtxCP)
        {
            if (ipsecModuleInfo_p->dpTbl->iSATbl[saIdx].secSession)
            {
                ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_security_session_destroy", "Destroy inline security session\n");
                rte_security_session_destroy(ipsecModuleInfo_p->secCtxCP,
                                    ipsecModuleInfo_p->dpTbl->iSATbl[saIdx].secSession);
                ipsecModuleInfo_p->dpTbl->iSATbl[saIdx].secSession = NULL;
                map_hw_sa_adk_sa(ipsecModuleInfo_p, FALSE, saIdx);
            }
        }
    }
}


/*
 * Rule and trace formats definitions.
 */

adk_ipsec_spd_t *adk_ipsec_spd_create (adk_sys_handle_t sysHandle,
                                       adk_module_handle_t ipsecHandle,
                                       uint8_t tblIdx)
{
    char aclName[RTE_ACL_NAMESIZE];
    struct rte_acl_ctx *aclCtx = NULL;
    struct rte_acl_param aclParams;

    ADK_MEMSET(&aclParams, 0, sizeof(struct rte_acl_param));
    snprintf (aclName, RTE_ACL_NAMESIZE, "ADK_IPSEC_ACL_%d", tblIdx);
    aclParams.name = aclName;
    aclParams.socket_id = SOCKET_ID_ANY;
    aclParams.rule_size = RTE_ACL_RULE_SZ (NUM_FIELDS_IPSEC);
    aclParams.max_rule_num = ADK_IPSEC_MAX_NUM_SPD_RULES;

    ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_acl_create", "Create "
                        "ACL for given rule set ID\n");
    aclCtx = rte_acl_create (&aclParams);

    if (aclCtx == NULL)
    {
        ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_IPSEC_RULESETID_MEMORY_UNAVAILABLE,
                             "\nFailed to create ACL context!\n");
    }

    return((adk_ipsec_spd_t *) aclCtx);
}

/*
 * Rule and trace formats definitions.
 */

adk_ipsec_spd_t *adk_ipsec_spd_in_create (adk_sys_handle_t sysHandle,
                                          adk_module_handle_t ipsecHandle,
                                          uint8_t tblIdx)
{
    char aclName[RTE_ACL_NAMESIZE];
    struct rte_acl_ctx *aclCtx = NULL;
    struct rte_acl_param aclParams;

    ADK_MEMSET(&aclParams, 0, sizeof(struct rte_acl_param));
    snprintf (aclName, RTE_ACL_NAMESIZE, "ADK_IPSEC_IN_ACL_%d", tblIdx);
    aclParams.name = aclName;
    aclParams.socket_id = SOCKET_ID_ANY;
    aclParams.rule_size = RTE_ACL_RULE_SZ (NUM_FIELDS_IPSEC_IN);
    aclParams.max_rule_num = ADK_IPSEC_MAX_NUM_SPD_IN_RULES;

    ADK_LOG_EXT_API (sysHandle, ipsecHandle, "rte_acl_create", "Create "
                     "SPD IN ACL\n");
    aclCtx = rte_acl_create (&aclParams);

    if (aclCtx == NULL)
    {
        ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_IPSEC_RULESETID_MEMORY_UNAVAILABLE,
                             "\nFailed to create SPD IN ACL context!\n");
    }

    return((adk_ipsec_spd_t *) aclCtx);
}

