/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <rte_eventdev.h>
#include <ice_dsi_ipsec.h>
#include <adk_include.h>
#include <adk_sys.h>
#include <adk_pmu.h>
#include <adk_metadata.h>
#include <adk_ipsec_struct.h>
#include "adk_ipsec_common.h"
#include "adk_ipsec_dp.h"
#include "adk_ipsec_export.h"

extern adk_ipsec_dp_table_info_t *adk_ipsec_dp_tbl_gbl_ptr;
extern ipsecIdSeqDB_t adk_ipsec_IdSeqDB[ADK_MAX_NUM_CORES];
extern ipsecIdSeqRange_t adk_ipsec_IdSeqRange[ADK_MAX_NUM_CORES];

static inline adk_status_t check_icmp_type(uint8_t type)
{
    switch (type)
    {
    case IPSEC_ICMP_MSGTYPE_DEST_UNREACHBLE:
    case IPSEC_ICMP_MSGTYPE_SRC_QUENCH:
    case IPSEC_ICMP_MSGTYPE_REDIRECT:
    case IPSEC_ICMP_MSGTYPE_TIME_EXCEEDED:
    case IPSEC_ICMP_MSGTYPE_PARAM_PROB:
        return ADK_SUCCESS;
    default:
        return ADK_IPSEC_ICMP_TYPE_NOT_ALLOWED;
    }
}

adk_status_t
adk_ipsec_outbound_icmp_transit(int coreIndex, adk_pktio_task_t  *taskInfo,
                                uint8_t version,
                                adk_ipsec_cache_frag_info_t *cacheInfo,
                                uint16_t *destFunc,
                                uint8_t *operType,
                                bool fragFlag)
{
    adk_status_t retVal = ADK_SUCCESS;
    struct rte_mbuf   *pktMbuf = (struct rte_mbuf *)taskInfo;
    adk_metadata_t    *adkMetaData = ADK_METADATA_PTR(taskInfo);
    adk_ipsec_dp_uli_imc_param_t  *ipImc =
        (adk_ipsec_dp_uli_imc_param_t *)&(adkMetaData->imc_param_start);
    ipsec_dp_outbound_meta_data_t *ipsecMetaData =
            (ipsec_dp_outbound_meta_data_t *)&(adkMetaData->imc_param_start);

    uint32_t saId, finalPktLen = 0, innerPktLen = 0;
    uint16_t  ipId, polIndex, innerIPHdrOffset = 0;
    uint8_t ipHdrLen, *pdu;
    struct ipv4_hdr *ipv4Hdr = NULL, *outerIpv4 = NULL;
    struct ipv6_hdr *ipv6Hdr = NULL, *outerIpv6 = NULL;
    struct udp_hdr  *udpHdr = NULL;
    struct udp_hdr  *outerUdp = NULL;
    struct icmp_hdr *icmpHdr = NULL;
    struct rte_crypto_op        *opData = NULL;

    adk_ipsec_spd_sel_t spdPtn;
    adk_ipsec_dp_eg_sa_t   *saInfo;

    adk_ipsec_stat_t *globalStats;
    uint16_t  fragInfo = 0;
    ADK_MEMSET(&spdPtn, 0, sizeof(adk_ipsec_spd_sel_t));
    ADK_PDU_PTR_GET(taskInfo, pdu, uint8_t *);

    if (version == ADK_IPSEC_IPV4)
    {
        globalStats = &adk_ipsec_dp_tbl_gbl_ptr->ipsecGblStatsV4[coreIndex];
        /* first of all, check if this is ICMP packet,
         * if so, check if the ICMP type is allowed
         */
        ipv4Hdr = (struct ipv4_hdr *)(pdu + adkMetaData->ip_offset);
        innerPktLen = ntohs(ipv4Hdr->total_length) + sizeof (struct ipv4_hdr);
        icmpHdr = (struct icmp_hdr *)(pdu + adkMetaData->l4_offset);
        retVal = check_icmp_type(icmpHdr->icmp_type);
        if (retVal != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit: Error: NON-ALLOWED ICMP TYPE!");
            return ADK_E_DROP_TASK;
        }
        /* move PDU to the real IP packet inside ICMP
         * let us assume right now that there will be IPv4 packet inside
         */
#if ADK_DEBUG
    rte_pktmbuf_dump(stdout, pktMbuf, pktMbuf->data_len);
#endif
        ipv4Hdr = (struct ipv4_hdr *)(pdu + adkMetaData->l4_offset + sizeof(struct icmp_hdr));
        udpHdr = (struct udp_hdr *)(ipv4Hdr+1);
        spdPtn.version = ADK_IPSEC_IPV4;
        spdPtn.sip_addr[0] = ipv4Hdr->dst_addr;
        spdPtn.dip_addr[0] = ipv4Hdr->src_addr;
        spdPtn.dscpVal = (ipv4Hdr->type_of_service >> 2) & 0x3F;
        /* this is required to be done for DSCP and protocol fields,
         * because in reality these are 8 bit fields!!
         */
        spdPtn.dscpVal = rte_cpu_to_be_16(spdPtn.dscpVal);
        spdPtn.protocol = rte_cpu_to_be_16(ipv4Hdr->next_proto_id);
        /* add a check for fragmentation here */
        fragInfo = rte_cpu_to_be_16(ipv4Hdr->fragment_offset);
        if ((unlikely(fragInfo & 0x3FFF)) && (unlikely(fragInfo != 0)))
        {
            spdPtn.sport = 0;
            spdPtn.dport = 0;
        }
        else
        {
            spdPtn.sport = udpHdr->dst_port;
            spdPtn.dport = udpHdr->src_port;
        }
    }
    else if(version == ADK_IPSEC_IPV6)
    {
        globalStats = &adk_ipsec_dp_tbl_gbl_ptr->ipsecGblStatsV6[coreIndex];
        ipv6Hdr = (struct ipv6_hdr *)(pdu + adkMetaData->ip_offset);
        /* let us ingnore the extension header for the time being
         */
        icmpHdr = (struct icmp_hdr *)(pdu + adkMetaData->l4_offset);
        retVal = check_icmp_type(icmpHdr->icmp_type);
        if (retVal != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit: Error: NON-ALLOWED ICMP TYPE!");
            return ADK_E_DROP_TASK;
        }
        /* move PDU to the real IP packet inside ICMP
         * let us assume right now that there will be IPv6 packet inside
         */
        ipv6Hdr = (struct ipv6_hdr *)(pdu + adkMetaData->l4_offset + sizeof(struct icmp_hdr));
        spdPtn.version = ADK_IPSEC_IPV6;
        rte_memcpy(spdPtn.sip_addr, ipv6Hdr->src_addr, 16);
        rte_memcpy(spdPtn.dip_addr, ipv6Hdr->dst_addr, 16);
        spdPtn.dscpVal = ipv6Hdr->vtc_flow >> 20;
        /* this is required to be done for DSCP and protocol fields,
         * because in reality these are 8 bit fields!!
         */
        spdPtn.dscpVal = rte_cpu_to_be_16(spdPtn.dscpVal);
        spdPtn.protocol = rte_cpu_to_be_16(ipv6Hdr->proto);
        /* add a check for fragmentation here */
        udpHdr = (struct udp_hdr *)(pdu + sizeof(struct ipv6_hdr));
        spdPtn.sport = udpHdr->src_port;
        spdPtn.dport = udpHdr->dst_port;
    }
    else
    {
        ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit:: Invalid IP version!!");
        return ADK_E_DROP_TASK;
    }


    retVal = adk_ipsec_spd_lkup(&spdPtn, &polIndex);
    if (retVal != ADK_SUCCESS)
    {
        return ADK_E_DROP_TASK;
    }
    if (polIndex == 0)
    {
        return ADK_E_DROP_TASK;
    }
    else
    {
        polIndex -= 1;

        /* first get all the required indices
         */
        saId = adk_ipsec_dp_tbl_gbl_ptr->polSAInfo[polIndex].eSAIdx;

        switch(adk_ipsec_dp_tbl_gbl_ptr->polSAInfo[polIndex].action)
        {
            case ADK_IPSEC_SP_ACTION_IPSEC:
                saInfo = &adk_ipsec_dp_tbl_gbl_ptr->eSATbl[saId];
                if (adk_ipsec_dp_tbl_gbl_ptr->polSAInfo[polIndex].eSAstate ==
                        INVALID_SA)
                {
                    /* as per RFC, this is only an auditable event, no need to
                     * create a new SA for this, and we don't have any audit
                     * for this type, so lets just return an error
                     */
                    ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit: Error: Egress SA not found!");
                    return ADK_E_DROP_TASK;
                }
                retVal = adk_ipsec_prepare_proto_op(coreIndex, taskInfo, adkMetaData, saInfo,
                                                 version, innerPktLen,
                                                 saId,
                                                 adk_ipsec_dp_tbl_gbl_ptr->polSAInfo[polIndex].configDF,
                                                 globalStats, &innerIPHdrOffset, false);
                if (retVal == ADK_SUCCESS)
                {
                    if (!saInfo->ipVersion)
                    {
                        ipHdrLen = 20;
                    }
                    else
                    {
                        ipHdrLen = sizeof(struct ipv6_hdr);
                    }

                    if (saInfo->saMode == ADK_IPSEC_SA_LOOKASIDE)
                        *operType = ADK_IPSEC_LOOKASIDE;
                    else if (ipImc->la_flag == 1)
                    {
                        /* this means, post-fragmentation would happen, so
                         * process this packet in LA way
                         */
                        *operType = ADK_IPSEC_LOOKASIDE;
                    }
                    
                    if(likely(ADK_IPSEC_INLINE == *operType))
                    {
                        finalPktLen = adk_ipsec_inline_prepare_crypto_op(taskInfo,
                                                        saInfo,
                                                        version,
                                                        innerPktLen);
                    }
                    else
                    {
                        /* proceed only if we have opData buffer
                         */
                        opData = saInfo->lkasideInfo->cryptoOp;
                        if (NULL == opData)
                        {
                            return ADK_E_DROP_TASK;
                        }

                        finalPktLen = adk_ipsec_outbound_la_prepare_crypto_op(taskInfo, opData,
                                                        saInfo,
                                                        ipHdrLen, version,
                                                        innerPktLen);
                    }
                    if (0 == finalPktLen)
                    {
                        return ADK_E_DROP_TASK;
                    }
                    saInfo->bytesUsed += finalPktLen;
                    /* modify outer IP header before sending packet for encryption */
                    /* now fix the total length field of outer ip */
                    if (!saInfo->ipVersion)
                    {
                        outerIpv4 = rte_pktmbuf_mtod(pktMbuf, struct ipv4_hdr *);
                        /*Set the IP Id*/
                        ipId = ( (adk_ipsec_IdSeqDB[coreIndex].ipIdSeq >= adk_ipsec_IdSeqRange[coreIndex].max) ?
                                (adk_ipsec_IdSeqDB[coreIndex].ipIdSeq = adk_ipsec_IdSeqRange[coreIndex].min) :
                                (adk_ipsec_IdSeqDB[coreIndex].ipIdSeq++) );
                        outerIpv4->packet_id = rte_cpu_to_be_16(ipId);
                        outerIpv4->total_length = adk_htons(finalPktLen+ipHdrLen+
                                                            saInfo->digestLen);
                        /* TODO do we really need this?
                         */
                        outerIpv4->hdr_checksum = rte_ipv4_cksum(outerIpv4);
                        if (saInfo->isNATTEnable)
                        {
                            outerUdp = (struct udp_hdr*)(outerIpv4 + 1);
                            outerUdp->dgram_len = adk_htons(finalPktLen + UDPV4_HDR_LEN
                                                            + saInfo->digestLen);
                            /* as per RFC 3948, section 2.1
                             * the IPv4 UDP Checksum SHOULD be transmitted as a zero value
                             */
                            outerUdp->dgram_cksum = 0;
                        }
                    }
                    else
                    {
                        outerIpv6 = rte_pktmbuf_mtod(pktMbuf, struct ipv6_hdr *);
                        outerIpv6->payload_len = adk_htons(finalPktLen +
                                                           saInfo->digestLen);
                    }
                    saInfo->inFlightCnt++;
                    adk_ipsec_dp_tbl_gbl_ptr->SAUtilityCnt[coreIndex][saId].egSaUtilityCounters.ipsec_eg_sa_inflight_pkt++;
                    INCREMENT_EGRESS_UTILITY_COUNTER(coreIndex, version, ipsec_eg_num_pdus_sent_for_outbound_processing);
                    INCREMENT_EGRESS_UTILITY_COUNTER(coreIndex, version, ipsec_eg_tx_pkts);
                    if (ADK_IPSEC_LOOKASIDE == *operType)
                    {
                        ipsecMetaData = (ipsec_dp_outbound_meta_data_t *)&(adkMetaData->imc_param_start);
                        ipsecMetaData->path = 0;
                        ipsecMetaData->saId = saId;
                        ipsecMetaData->dscpVal_version = saInfo->ipVersion ? ADK_IPSEC_IPV6 : ADK_IPSEC_IPV4;
                        ADK_METADATA_PARAM((adk_pktio_task_t *)(taskInfo), next_frag) = NULL;

                         retVal = adk_pal_dp_crypto_send(0,
                                                        ADK_IPSEC_ULI_FROM_CRYPTO_FN_ID,
                                                        0, /* priority */
                                                        taskInfo,
                                                        opData,
                                                        ADK_PAL_DP_BATCH_CB_FN_IPSEC,
                                                        NULL);

                        if (retVal != ADK_E_TASK_CONSUMED)
                        {
                            return ADK_E_INTERNAL_ERROR;
                        }
                        /* now that this packet got processed, correctly,
                         * lets set the cache data for next frags, if any
                         */
                        if (MORE_FRAGS == fragFlag)
                        {
                            cacheInfo->operType = *operType;
                            cacheInfo->innerVersion = version;
                            cacheInfo->unrollPolIndex = polIndex;
                            cacheInfo->saId = saId;
                            if (!saInfo->ipVersion) /* this is outer version */
                            {
                                rte_memcpy(&cacheInfo->outerHeader.ipv4Hdr,
                                           outerIpv4,
                                           IPV4_HDR_LEN);
                            }
                            else
                            {
                                rte_memcpy(&cacheInfo->outerHeader.ipv6Hdr,
                                           outerIpv6,
                                           sizeof(struct ipv6_hdr));
                            }
                            cacheInfo->saInfo = saInfo;
                        }
                        RTE_SET_USED(destFunc);
                    }
                    else
                    {
                        if (saInfo->secSession)
                        {
                            rte_security_set_pkt_metadata(adk_ipsec_dp_tbl_gbl_ptr->secCtxDP,
                                    saInfo->secSession,
                                    pktMbuf, NULL);
                        }
                        else
                        {
                            return ADK_E_DROP_TASK;
                        }

                        adkMetaData->exception_domain = ADK_FP_ORIGIN_DOMAIN;

                        /*  Set the egress stats fields in the metadata */
                        adkMetaData->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_IPSEC_LOC]
                            .flags = ADK_SYS_EGRESS_STATS_MDATA_FLAG_PKT_CNTR;
                        adkMetaData->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_IPSEC_LOC]
                            .stats_id = adk_ipsec_dp_tbl_gbl_ptr->eSATbl[saId].egressStatsId;

                        if (!saInfo->ipVersion)
                        {
                            *destFunc = adkModuleMgrCpuServiceFn_g[ADK_IP_MODULE_ID].cpuEntryFn_ULI[_DEST_FUNC_DEFAULT];
                            INCREMENT_UTILITY_COUNTER(adk_ipsec_dp_tbl_gbl_ptr->ipsecUtilityCntV4[coreIndex].egressUtilityCounters.ipsec_eg_tx_pkts);
                        }
                        else
                        {
                            *destFunc = adkModuleMgrCpuServiceFn_g[ADK_IPV6_MODULE_ID].cpuEntryFn_LLI[_DEST_FUNC_DEFAULT];
                            INCREMENT_UTILITY_COUNTER(adk_ipsec_dp_tbl_gbl_ptr->ipsecUtilityCntV6[coreIndex].egressUtilityCounters.ipsec_eg_tx_pkts);
                        }
                        ++adk_ipsec_dp_tbl_gbl_ptr->egSAStats[coreIndex][saId].ipsec_eg_sa_tot_out_pkts;
                        if (MORE_FRAGS == fragFlag)
                        {
                            cacheInfo->operType = *operType;
                            cacheInfo->innerVersion = version;
                            cacheInfo->unrollPolIndex = polIndex;
                            cacheInfo->saId = saId;
                            if (!saInfo->ipVersion)
                            {
                                rte_memcpy(&cacheInfo->outerHeader.ipv4Hdr,
                                           outerIpv4, IPV4_HDR_LEN);
                            }
                            else
                            {
                                rte_memcpy(&cacheInfo->outerHeader.ipv6Hdr,
                                           outerIpv6,
                                           sizeof(struct ipv6_hdr));
                            }
                            cacheInfo->saInfo = saInfo;
                        }
                    }
                }
                else if (retVal == ADK_E_DROP_TASK)
                {
                    return ADK_E_DROP_TASK;
                }
                else
                {
                    ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit:: Internal error");
                    return ADK_E_INTERNAL_ERROR;
                }
                break;
            case ADK_IPSEC_SP_ACTION_DROP:
                ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit:: SPD ACTION DROP!");
                return ADK_E_DROP_TASK;
            case ADK_IPSEC_SP_ACTION_BYPASS:
                ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_outbound_icmp_transit:: SPD ACTION BYPASS!");
                return ADK_E_DROP_TASK;
            default:
                if (ADK_IPSEC_IPV4 == version)
                {
                    adk_ipsec_dp_tbl_gbl_ptr->ipsecGblStatsV4[coreIndex].egressStats.ipsec_eg_global_spd_pol_default_discard_drop++;
                }
                else
                {
                    adk_ipsec_dp_tbl_gbl_ptr->ipsecGblStatsV6[coreIndex].egressStats.ipsec_eg_global_spd_pol_default_discard_drop++;
                }
                return ADK_E_DROP_TASK;
        }
    }
    return ADK_SUCCESS;
}


adk_status_t
adk_ipsec_inbound_icmp_transit(void *taskInfo, uint8_t *pdu,
                               uint8_t version,
                               uint16_t *polIndex)
{
    adk_metadata_t    *adkMetaData = ADK_METADATA_PTR(taskInfo);

    ipsec_dp_inbound_meta_data_t *ipsecMetaData =
            (ipsec_dp_inbound_meta_data_t *)&(adkMetaData->imc_param_start);

    adk_status_t retVal = ADK_SUCCESS;

    adk_ipsec_spd_sel_t spdPtn;
    struct ipv4_hdr *ipv4Hdr = NULL;
    struct ipv6_hdr *ipv6Hdr = NULL;
    struct icmp_hdr *icmpHdr;
    uint32_t saId = 0;
    uint16_t  fragInfo = 0;
    struct udp_hdr  *udpHdr;
    if (version == ADK_IPSEC_IPV4)
    {
        ipv4Hdr = (struct ipv4_hdr *)pdu;
        if (ipv4Hdr->next_proto_id != IPPROTO_ICMP)
        {
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_inbound_icmp_transit: Error: NOT ICMP PACKET!");
            /*  This scenario occurs when there's a non-ICMP packet
             *  that fails the SPD lookup prior to entering this function.
             *  Thus, the error returned should be policy not found
             */
            return ADK_E_IPSEC_POLICY_NOT_FOUND;
        }
        icmpHdr = (struct icmp_hdr *)(pdu + IPV4_HDR_LEN);
        retVal = check_icmp_type(icmpHdr->icmp_type);
        if (retVal != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_inbound_icmp_transit: Error: NON-ALLOWED ICMP TYPE!");
            return ADK_E_DROP_TASK;
        }
        /* now first of all do a reverse SPD lookup for the inner packet
         */
        /* move PDU to the real IP packet inside ICMP
         */
        pdu += IPV4_HDR_LEN + sizeof(struct icmp_hdr);
        /* let us assume right now that there will be IPv4 packet inside
         */
        ipv4Hdr = (struct ipv4_hdr *)pdu;
        spdPtn.version = 4;
        spdPtn.sip_addr[0] = ipv4Hdr->dst_addr;
        spdPtn.dip_addr[0] = ipv4Hdr->src_addr;
        /* this is required to be done for DSCP and protocol fields,
         * because in reality these are 8 bit fields!!
         */
        spdPtn.dscpVal = rte_cpu_to_be_16(ipsecMetaData->dscpVal_version);
        spdPtn.protocol = rte_cpu_to_be_16(ipv4Hdr->next_proto_id);
        /* add a check for fragmentation here */
        fragInfo = rte_cpu_to_be_16(ipv4Hdr->fragment_offset);
        if ((unlikely(fragInfo & 0x3FFF)) && (unlikely(fragInfo != 0)))
        {
            spdPtn.sport = 0;
            spdPtn.dport = 0;
        }
        else
        {
            udpHdr = (struct udp_hdr *)(pdu + IPV4_HDR_LEN);
            spdPtn.sport = udpHdr->dst_port;
            spdPtn.dport = udpHdr->src_port;
        }
    }
    else
    {
        /* now first of all do a reverse SPD lookup for the inner packet
         */
        ipv6Hdr = (struct ipv6_hdr *)pdu;
        if (ipv6Hdr->proto != IPPROTO_ICMPV6)
        {
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_inbound_icmp_transit: Error: NOT ICMP PACKET!");
            /*  This scenario occurs when there's a non-ICMP packet
             *  that fails the SPD lookup prior to entering this function.
             *  Thus, the error returned should be policy not found
             */
            return ADK_E_IPSEC_POLICY_NOT_FOUND;
        }
        icmpHdr = (struct icmp_hdr *)(pdu + sizeof(struct ipv6_hdr));
        retVal = check_icmp_type(icmpHdr->icmp_type);
        if (retVal != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_inbound_icmp_transit: Error: NON-ALLOWED ICMP TYPE!");
            return ADK_E_DROP_TASK;
        }
        /* now first of all do a reverse SPD lookup for the inner packet
         */
        /* move PDU to the real IP packet inside ICMP
         */
        pdu += sizeof(struct ipv6_hdr) + sizeof(struct icmp_hdr);
        /* let us assume right now that there will be IPv4 packet inside
         */
        ipv6Hdr = (struct ipv6_hdr *)pdu;
        spdPtn.version = 6;
        rte_memcpy(spdPtn.sip_addr, ipv6Hdr->dst_addr, 16);
        rte_memcpy(spdPtn.dip_addr, ipv6Hdr->src_addr, 16);
        /* this is required to be done for DSCP and protocol fields,
         * because in reality these are 8 bit fields!!
         */
        spdPtn.dscpVal = rte_cpu_to_be_16(ipsecMetaData->dscpVal_version);
        spdPtn.protocol = rte_cpu_to_be_16(ipv6Hdr->proto);

        udpHdr = (struct udp_hdr *)(ipv6Hdr + 1);
        /* add a check for fragmentation here */
        spdPtn.sport = udpHdr->dst_port;
        spdPtn.dport = udpHdr->src_port;
    }

    retVal = adk_ipsec_spd_lkup(&spdPtn, polIndex);
    if (retVal != ADK_SUCCESS)
    {
        return ADK_E_DROP_TASK;
    }
    if (*polIndex == 0)
    {
        return ADK_E_IPSEC_POLICY_NOT_FOUND;
    }
    else
    {
        *polIndex -= 1;
        /* first get all the required indices
         */
        if(adk_ipsec_dp_tbl_gbl_ptr->polSAInfo[*polIndex].action ==
                ADK_IPSEC_SP_ACTION_IPSEC)
        {
            if (adk_ipsec_dp_tbl_gbl_ptr->polSAInfo[*polIndex].spIDEx
                != adk_ipsec_dp_tbl_gbl_ptr->iSATbl[saId].spIDEx)
            {
                /*  TODO : Need to raise audit for SA mismatch for ICMP here  */
                ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_inbound_icmp_transit:: SA validation post decryption failed!");
                return ADK_E_DROP_TASK;
            }
        }
        else
        {
            /*  TODO : Need to raise audit for action DROP here for ICMP  */
            ADK_DP_LOG_ERROR (ADK_IPSEC_MODULE_ID, "adk_ipsec_inbound_icmp_transit:: SA validation post decryption failed!");
            return ADK_E_DROP_TASK;
        }
    }

    return ADK_SUCCESS;
}

