/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <adk_include.h>
#include <adk_pmu.h>
#include <adk_sys.h>

#include <adk_ipsec_struct.h>
#include <adk_ipsec_common.h>
#include <adk_ipsec_dp.h>

extern adk_ipsec_dp_table_info_t *adk_ipsec_dp_tbl_gbl_ptr;

static void convert_32bit_addr_to_8bit_addr (struct ipv6_hdr *outerIpv6,
                                             uint32_t *saSIP, uint32_t *saDIP)
{
    uint8_t i, j, k;

    j = 0;
    k = 0;
    for (i = 0; i < 16; i++)
    {
        outerIpv6->src_addr[i] = saSIP[j] >> (32 - ((k + 1)*8));
        outerIpv6->dst_addr[i] = saDIP[j] >> (32 - ((k + 1)*8));
        k++;
        if (i % 4 == 3)
        {
            j++;
            k = 0;
        }
    }
}

adk_status_t adk_ipsec_prepare_proto_op(int coreIndex, void *pkt, adk_metadata_t *adkMetaData,
                                        adk_ipsec_dp_eg_sa_t *saInfo,
                                        uint8_t innerVersion,
                                        uint32_t packetLen,
                                        uint32_t saId,
                                        adk_ipsec_dont_frag_config_t dfConfig,
                                        adk_ipsec_stat_t *globalStats,
                                        uint16_t *innerIPHdrOffset,
                                        uint8_t isFrag)
{
    int retValue = ADK_SUCCESS;
    struct rte_mbuf   *pktMbuf = (struct rte_mbuf *)pkt;
    adk_pktio_task_t  *taskInfo = (adk_pktio_task_t *)pkt;
    struct esp_header *espHdr = NULL;
    struct ipv4_hdr   *innerIpv4 = NULL, *outerIpv4 = NULL;
    struct ipv6_hdr   *innerIpv6 = NULL, *outerIpv6 = NULL;
    struct udp_hdr    *outerUdp = NULL;
    uint16_t offset = 0;
    uint64_t ivValue[2], saSeqNum;
    struct rte_mbuf *pktClone = NULL;
    uint8_t i = 0;
    
    if (likely(ADK_IPSEC_SA_INLINE == saInfo->saMode))
    {
        saSeqNum = saInfo->seqNumber;
    }
    else
    {
        saSeqNum = rte_atomic32_read(&saInfo->lkasideInfo->seqNumber);
    }
    if (FALSE == saInfo->seqRollover)
    {
        /* Raising exception when seqNumber is reached 75% of Max sequence number */
        if (unlikely((!saInfo->extSeqNum) &&
            (saSeqNum == MAX_SEQ_NUMBER_EXCEPTION)))
        {
            /* Filling the required info directly from SA Info since outer tunnel header is
             * not yet created and attached to the packet at this point.
             */
            adk_ipsec_audit_event_t auditEventData = {0};
            auditEventData.event.ipsecEgSeqNumOverFlowEvent.ipVersion = saInfo->ipVersion ? ADK_IPSEC_IPV6: ADK_IPSEC_IPV4;
            if (!saInfo->ipVersion)
            {
                auditEventData.event.ipsecEgSeqNumOverFlowEvent.ipSrcAddr_u.ipv4Addr = 
                               adk_ntohl(saInfo->saSIP[0]);
                auditEventData.event.ipsecEgSeqNumOverFlowEvent.ipDstAddr_u.ipv4Addr =
                               adk_ntohl(saInfo->saDIP[0]);
            } 
            else
            {
                for (i = 0; i < 4; i++)
                {
                    auditEventData.event.ipsecEgSeqNumOverFlowEvent.ipSrcAddr_u.ipv6Addr[i] = 
                           adk_ntohl(saInfo->saSIP[i]);
                    auditEventData.event.ipsecEgSeqNumOverFlowEvent.ipDstAddr_u.ipv6Addr[i] = 
                            adk_ntohl(saInfo->saDIP[i]);
                }
               auditEventData.event.ipsecEgSeqNumOverFlowEvent.flowLabel = saInfo->flow_label_value;
            }
            auditEventData.event.ipsecEgSeqNumOverFlowEvent.secProtocol = PROTO_ESP;
            auditEventData.event.ipsecEgSeqNumOverFlowEvent.spiVal = saInfo->spi;
            auditEventData.event.ipsecEgSeqNumOverFlowEvent.seqNum = saSeqNum;
            pktClone = adk_ipsec_pktmbuf_clone((struct rte_mbuf *)(taskInfo),
                                               ((struct rte_mbuf *)(taskInfo))->pool);
            if (unlikely(NULL == pktClone))
            {
                if (ADK_IPSEC_IPV4 == innerVersion)
                {
                    globalStats->egressStats.ipsec_eg_global_mbuf_clone_failed++;
                }
                else
                {
                    globalStats->egressStats.ipsec_eg_global_mbuf_clone_failed++;
                }
            }
            else
            {
                uint8_t ipVersion = saInfo->ipVersion ? ADK_IPSEC_IPV6 : ADK_IPSEC_IPV4;
                retValue = adk_ipsec_dp_handle_exception(pktClone, ADK_IPSEC_EXCEPTION_AUDIT_EG_SEQNO_OVERFLOW, 
                        &auditEventData, coreIndex, ipVersion, ADK_IPSEC_OUTBOUND, false);
                if (unlikely(ADK_E_TASK_CONSUMED != retValue))
                {
                    if(unlikely(ADK_E_INVALID_HANDLE == retValue))
                    {
                        globalStats->egressStats.ipsec_eg_global_exception_not_registered++;
                    }
                    adk_pal_dp_release_send(pktClone, ADK_PAL_DP_BATCH_CB_FN_IPSEC, NULL);
                    adk_pktio_task_buffer_free((adk_pktio_task_t *)(pktClone));
                    retValue = ADK_SUCCESS;
                }
                return retValue;
            }
        }
        else if (unlikely((!saInfo->extSeqNum) &&
                 (saSeqNum == MAX_SEQ_NUMBER)))
        {
            globalStats->egressStats.ipsec_eg_global_seqno_overflow_drop++;
            return ADK_E_DROP_TASK;
        }
    }
    
    if (ADK_IPSEC_TRUE == saInfo->lifetimesEnabled)
    {
        if(unlikely(saInfo->hardByteLifetime <= saInfo->bytesUsed))
        {
            adk_ipsec_sa_expiry_excep_data_t sa_expiry_excep_data = {0};
            globalStats->egressStats.ipsec_eg_global_sa_hardlife_expire_drop++;

            /*Revisit when IPSec Audit event is implemented*/  
            sa_expiry_excep_data.direction = ADK_IPSEC_OUTBOUND;
            sa_expiry_excep_data.whichExpiry = ADK_HARD_LIFE_EXPIRY;
            sa_expiry_excep_data.spi = saInfo->spi;
            uint8_t ipVersion = saInfo->ipVersion ? ADK_IPSEC_IPV6 : ADK_IPSEC_IPV4;  
            retValue = adk_ipsec_dp_handle_exception(taskInfo, ADK_IPSEC_EXCEPTION_SA_EXPIRE,
                                                     &sa_expiry_excep_data, coreIndex,
                                                     ipVersion, ADK_IPSEC_OUTBOUND,
                                                     isFrag);
            if(unlikely(ADK_E_INVALID_HANDLE == retValue))
            {
                globalStats->egressStats.ipsec_eg_global_exception_not_registered++;
            }
            return retValue;
        }

        if (unlikely(saInfo->softByteLifetime <=
                    (saInfo->bytesUsed + ESPV4_HDR_LEN + saInfo->ivLen + RTE_ALIGN_CEIL(packetLen  + 2, saInfo->blockSize))))
        {
            adk_ipsec_sa_expiry_excep_data_t sa_expiry_excep_data = {0};
            /*Revisit when IPSec Audit event is implemented*/
            sa_expiry_excep_data.direction = ADK_IPSEC_OUTBOUND;
            sa_expiry_excep_data.whichExpiry = ADK_SOFT_LIFE_EXPIRY;
            sa_expiry_excep_data.spi = saInfo->spi;

            /* Soft lifetime expiry or sampling : Need to clone the mbuf
             */
            pktClone = adk_ipsec_pktmbuf_clone((struct rte_mbuf *)(taskInfo),
                    ((struct rte_mbuf *)(taskInfo))->pool);
            if (unlikely(NULL == pktClone))
            {
                if (ADK_IPSEC_IPV4 == innerVersion)
                {
                    globalStats->egressStats.ipsec_eg_global_mbuf_clone_failed++;
                }
                else
                {
                    globalStats->egressStats.ipsec_eg_global_mbuf_clone_failed++;
                }
            }
            else
            {
                uint8_t ipVersion = saInfo->ipVersion ? ADK_IPSEC_IPV6 : ADK_IPSEC_IPV4;
                retValue = adk_ipsec_dp_handle_exception(pktClone, ADK_IPSEC_EXCEPTION_SA_EXPIRE,
                                                         &sa_expiry_excep_data, coreIndex,
                                                         ipVersion,ADK_IPSEC_OUTBOUND,
                                                         isFrag);
                if (unlikely(ADK_E_TASK_CONSUMED != retValue))
                {
                    if(unlikely(ADK_E_INVALID_HANDLE == retValue))
                    {
                        globalStats->egressStats.ipsec_eg_global_exception_not_registered++;
                    }
                    adk_pal_dp_release_send(pktClone, ADK_PAL_DP_BATCH_CB_FN_IPSEC, NULL);
                    adk_pktio_task_buffer_free((adk_pktio_task_t *)(pktClone));
                }
            }
        }
    }

    /* Removing the FCS bits */
    if (pktMbuf->pkt_len > packetLen)
    {
        retValue = rte_pktmbuf_trim(pktMbuf, (pktMbuf->pkt_len - packetLen));
    }
 
    adkMetaData->l4_offset -= adkMetaData->ip_offset;
    adkMetaData->ip_offset = 0;
    if (likely(ADK_IPSEC_SA_INLINE == saInfo->saMode))
    {
        adkMetaData->saID = saId;
    }
    else
    {
        adkMetaData->saID = ADK_IPSEC_INVALID_SA_ID;
    }
    if (ADK_IPSEC_IPV4 == innerVersion)
    {
        innerIpv4 = rte_pktmbuf_mtod(pktMbuf, struct ipv4_hdr * );
    }
    else
    {
        innerIpv6 = rte_pktmbuf_mtod(pktMbuf, struct ipv6_hdr *);
    }

    if (!saInfo->ipVersion) /* this is outer version */
    {
        if (saInfo->isNATTEnable)
            offset = IPV4_HDR_LEN + UDPV4_HDR_LEN + ESPV4_HDR_LEN
                     + saInfo->ivLen;
        else
            offset = IPV4_HDR_LEN + ESPV4_HDR_LEN + saInfo->ivLen;

        outerIpv4 = (struct ipv4_hdr *)rte_pktmbuf_prepend(pktMbuf, offset);
        if (unlikely(NULL == outerIpv4))
        {
            globalStats->egressStats.ipsec_eg_global_outer_hdr_buf_alloc_fail_drop++;
            return ADK_E_DROP_TASK;
        }
        outerIpv4->version_ihl = 0x45;
        if (ADK_IPSEC_IPV4 == innerVersion)
        {
            /* Copying 6 most significant bits out of total 8 bits for
             * DSCP
             */
            outerIpv4->type_of_service = (innerIpv4->type_of_service & 0xFC);
        }
        else
        {
            /* Right shifting vtc flow field to obtain DSCP/ECN field which
             * is 8 bits.
             * Copying 6 most significant bits out of total 8 bits for
             * DSCP
             * dscpVal = (ntohl(innerIpv6->vtc_flow) >> ADK_IPSEC_TC_OFFSET) & 0x3F
             */
             outerIpv4->type_of_service =
                  ((ntohl(innerIpv6->vtc_flow)>> ADK_IPSEC_TC_OFFSET) & 0x3F) << 2;
        }
        outerIpv4->fragment_offset = 0;

        switch(dfConfig)
        {
        case ADK_IPSEC_SP_DF_COPY_INNER_IP:
            if (innerVersion == ADK_IPSEC_IPV4)
                outerIpv4->fragment_offset |=
                    (innerIpv4->fragment_offset & rte_cpu_to_be_16(IPV4_HDR_DF_FLAG));
            break;
        case ADK_IPSEC_SP_DF_SET:
            outerIpv4->fragment_offset |= rte_cpu_to_be_16(IPV4_HDR_DF_FLAG);
            break;
        case ADK_IPSEC_SP_DF_CLEAR:
            outerIpv4->fragment_offset &= ~rte_cpu_to_be_16(IPV4_HDR_DF_FLAG);
            break;
        default:
            return ADK_E_INTERNAL_ERROR;
        }
        outerIpv4->time_to_live = adk_ipsec_dp_tbl_gbl_ptr->eSATbl[saId].ttl_hopLimit;
        outerIpv4->next_proto_id = IPPROTO_ESP;
        outerIpv4->hdr_checksum = 0;
        if (unlikely(saInfo->isNATTEnable))
        {
            outerIpv4->next_proto_id = IPPROTO_UDP;
            outerIpv4->src_addr = saInfo->lkasideInfo->NATTSrcAddr;
            outerIpv4->dst_addr = saInfo->lkasideInfo->NATTDstAddr;
            outerUdp = (struct udp_hdr *)(outerIpv4 +1);
            outerUdp->src_port = saInfo->lkasideInfo->NATTSrcPort;
            outerUdp->dst_port = saInfo->lkasideInfo->NATTDstPort;
            espHdr = (struct esp_header*)(outerUdp + 1);
        }
        else
        {
            outerIpv4->src_addr = saInfo->saSIP[0];
            outerIpv4->dst_addr = saInfo->saDIP[0];
            espHdr = (struct esp_header*)(outerIpv4 + 1);
        }
    }
    else /* ipv6 */
    {
        offset = sizeof(struct ipv6_hdr)+ sizeof(struct esp_header)
                            + saInfo->ivLen;

        outerIpv6 = (struct ipv6_hdr *)rte_pktmbuf_prepend(pktMbuf, offset);
        if (unlikely(NULL == outerIpv6))
        {
            globalStats->egressStats.ipsec_eg_global_outer_hdr_buf_alloc_fail_drop++;
            return ADK_E_DROP_TASK;
        }
        outerIpv6->payload_len = htons((rte_pktmbuf_data_len(pktMbuf) - sizeof(struct ipv6_hdr)));
        outerIpv6->hop_limits = adk_ipsec_dp_tbl_gbl_ptr->eSATbl[saId].ttl_hopLimit;
        outerIpv6->proto = IPPROTO_ESP;

        if (ADK_IPSEC_IPV4 == innerVersion)
        {
            /* Copying an 6-bit DSCP value from inner IPv4 header field
             * to a 32-bit outer IPv6 field. Hence, converting IPv4 field to 32-bits, and
             * left shifting it by 22 bits in order to align with DSCP/ECN bits
             * in IPv6 header. Using 0x3F mask to extract only DSCP bits.
             * dscpVal = (innerIpv4->type_of_service >> 2) & 3F
             */
             outerIpv6->vtc_flow = rte_cpu_to_be_32(ADK_IPSEC_DEF_IPV6_VTC_FLOW |
                                  (((innerIpv4->type_of_service >> 2) & 0x3F) << ADK_IPSEC_TC_OFFSET));
        }
        else
        {
             /* Easier assignment from v6 to v6. Thus only taking the bits
              * concerned (4-9 from the left, starting with 0) from inner header
              * dscpVal = (ntohl(innerIpv6->vtc_flow) >> ADK_IPSEC_TC_OFFSET) & 0x3F
              */
             outerIpv6->vtc_flow = rte_cpu_to_be_32(ADK_IPSEC_DEF_IPV6_VTC_FLOW |
                                  ((ntohl(innerIpv6->vtc_flow) >> ADK_IPSEC_TC_OFFSET) & 0x3F) << ADK_IPSEC_TC_OFFSET);
        }
        /* Need to convert Tunnel DIP and Tunnel SIP to 8 bit addresses for
         * DPDK compatibility
         */
        convert_32bit_addr_to_8bit_addr (outerIpv6,
                                         saInfo->saSIP,
                                         saInfo->saDIP);
        espHdr = (struct esp_header*)(outerIpv6 + 1);
    }
    *innerIPHdrOffset = offset;
    /* prepare ESP header and append it */
    espHdr->spi = saInfo->spi;
    if (unlikely(ADK_IPSEC_SA_MODE_LOOKASIDE == saInfo->saMode))
    {
        rte_atomic32_inc(&saInfo->lkasideInfo->seqNumber);
        saSeqNum = rte_atomic32_read(&saInfo->lkasideInfo->seqNumber);
        espHdr->seqNum = (uint32_t)(saSeqNum);
        espHdr->seqNum = rte_cpu_to_be_32(espHdr->seqNum);
    }

    saInfo->ivBuf = (uint8_t *)(espHdr + 1);
    ADK_MEMSET(saInfo->ivBuf, 0, saInfo->ivLen);

    switch (saInfo->encr)
    {
        case ADK_IPSEC_ENCR_NULL:
            break;
        case ADK_IPSEC_ENCR_AES_CBC:
            ivValue[0] = rte_rand();
            ivValue[1] = rte_rand();
            rte_memcpy(saInfo->ivBuf, &ivValue[0], sizeof(uint64_t)*2);
            break;
        case ADK_IPSEC_ENCR_DES_CBC:
        case ADK_IPSEC_ENCR_3DES_CBC:
            ivValue[0] = rte_rand();
            rte_memcpy(saInfo->ivBuf, &ivValue[0], sizeof(uint64_t));
            break;
        case ADK_IPSEC_ENCR_AES_GCM_ICV8:
        case ADK_IPSEC_ENCR_AES_GCM_ICV12:
        case ADK_IPSEC_ENCR_AES_GCM_ICV16:
        case ADK_IPSEC_ENCR_AES_CCM_ICV8:
        case ADK_IPSEC_ENCR_AES_CCM_ICV12:
        case ADK_IPSEC_ENCR_AES_CCM_ICV16:
        case ADK_IPSEC_ENCR_AES_CTR:
        case ADK_IPSEC_ENCR_NULL_GMAC:
            if (unlikely(ADK_IPSEC_SA_MODE_LOOKASIDE == saInfo->saMode))
            {
                saSeqNum = rte_cpu_to_be_64(saSeqNum);
                rte_memcpy(saInfo->ivBuf, &saSeqNum, sizeof(saSeqNum));
            }
            break;
        default:
            /* Error logging not required - unsupported algorithm would be caught
             * at CP configuration level itself.
             */
            return ADK_E_INTERNAL_ERROR;
    }
    return ADK_SUCCESS;
}

