/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_utility_counters.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipv6  IPv6 Module
 * @{
 *
 * @brief
 *
 * This file contains ADK APIs to enable and disable utility counters
 * also get APIs to fetch utility counter values.
 *
 ***********************************************************************/

#ifndef _ADK_IPV6_UTILITY_COUNTER_H_
#define _ADK_IPV6_UTILITY_COUNTER_H_

#include "adk_sys.h"

/*!
 * =============================================================
 * @internal
 * @struct ipv6_ing_utility_counters_t "adk_ipv6_utility_counters.h"
 *
 * @brief
 *
 * Data structure for ipv6 ingress utility counters.
 * =============================================================
 */
typedef struct ipv6_ing_utility_counters_s
{
    /* Total Rx packets send to ipsec lli */
    uint64_t ipv6RxPktSendToIpsecLli;
    /* Total Rx Packets send to ipv6 uli */
    uint64_t ipv6RxPktFwdToIpv6Uli;
    /* Total Rx packets send to EO */
    uint64_t ipv6RxClassifierSendToEo;
    /* Total Rx packets sent to linux */
    uint64_t ipv6RxClassifierSendToHost;
    /* Total Rx packets dropped due to IP version is invalid */
    uint32_t ipv6RxInvalidVersion;
    /* Total Rx packets dropped due to invalid Multicast mac address */
    uint32_t ipv6RxInvalidMcastMac;
    /* Total Rx packets dropped because of invalid Multicast group */
    uint32_t ipv6RxInvalidMcastGroup;
    /* Total Rx packets dropped due to invalid Multicast scope */
    uint32_t ipv6RxInvalidMcastScope;
    /* Total Rx packets dropped because of packet having dst ip as
     * unicast IP and dst Mac address is multicast Mac.
     */
    uint32_t ipv6RxUcastDstIpMcastDstMacDrop;
    /* Total Rx packets dropped due to hop limit value as zero */
    uint32_t ipv6RxHopLimitZero;
    /* Total Rx packets dropped due to invalid DIP lookup, means
     * DIP lookup action is neither terminate nor forward
     */
    uint32_t ipv6RxInvalidRouteAction;
    /* Total Rx packets dropped because of send to EO fail after
     * ingress classifier lookup successful.
     */
    uint32_t ipv6RxClassifierSendToEoFail;
    /* Total Rx packets dropped because of send to Linux fail after
     * ingress classifier lookup successful.
     */
    uint32_t ipv6RxCLassifierSendToHostFail;
    /* Total Rx packets dropped due to invalid ingress classifier lookup action */
    uint32_t ipv6RxInvalidClassifierAction;
}ipv6_ing_utility_counters_t;

/*!
 * =============================================================
 * @internal
 * @struct ipv6_egress_utility_counters_t "adk_ipv6_utility_counters.h"
 *
 * @brief
 *
 * Data structure for ipv6 Egress utility counters.
 * =============================================================
 */
typedef struct ipv6_egress_utility_counters_s
{
    /* Total Tx packets sent from IPv6 to IPSEC ULI */
    uint64_t ipv6TxSendToIpsecUli;
    /* Total Tx packets sent from IPv6 to ETH ULI*/
    uint64_t ipv6TxSendToEthUli;
    /* Total Tx packets dropped due to pbr is disabled on particular interface */
    uint32_t ipv6PbrDisabled;
    /* Total Tx packets dropped due to route action is drop */
    uint32_t ipv6TxRouteActionDrop;
    /* Total Tx packets dropped due to invalid route action */
    uint32_t ipv6TxInvalidRouteAction;
    /* Total Tx packets dropped because of ADK not able to add space for
     * fragment header while fragmenting the packet
     */
    uint32_t ipv6TxhdrRoomNotAvail;
}ipv6_egress_utility_counters_t;

/*!
 * =============================================================
 * @internal
 * @struct ipv6_utility_counters_t "adk_ipv6_utility_counters.h"
 *
 * @brief
 *
 * Data structure for Ingress/Egress utility counters.
 * =============================================================
 */
typedef struct ipv6_utility_counters_s
{
    /* Ingress utility counters */
    ipv6_ing_utility_counters_t ipv6IngressUtilityCounters;

    /* Egress utility counters */
    ipv6_egress_utility_counters_t ipv6EgressUtilityCounters;
}ipv6_utility_counters_t;

/*!
 * =============================================================
 *
 * adk_ipv6_set_utility_counters
 *
 * @param[in]  ipv6Handle  -  IPv6 module handle. Returned at the
 *                            time of IPv6 module's registration
 *                            with the module manager.
 * @param[in]   utility_cnt_state - Boolean variable to enable/disable
                                    utility counter.
 *
 * @brief   Function to configure utility counters.
 *
 * This function configures utility counters.On disabling the utility
 * counters, all the counter values will get reset to zero.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_set_utility_counters(adk_module_handle_t ipv6Handle, bool utility_cnt_state);

/*!
 * =============================================================
 * @fn adk_ipv6_utility_counters_get
 *
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[out]   ipv6_utility_counters - IPv6 related utility counter data structure pointer
 *               gets set by the IPv6 ADK
 *
 * @retval ADK_SUCCESS - Successfully fetched utility counters
 *
 * @brief
 *
 * This API allows the application to retreive all IPv6 related utility counters
 * maintained by the IPv6 ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size ipv6_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_utility_counters_get(adk_module_handle_t ipv6Handle, ipv6_utility_counters_t *ipv6_utility_counters);

/*!
 * =============================================================
 * @fn adk_ipv6_utility_counters_print
 *
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully printed utility counters
 *
 * @brief
 *
 * This API allows the application to print all IPv6 related utility counters
 * maintained by the IPv6 ADK module.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_utility_counters_print(adk_module_handle_t ipv6Handle);

#endif
