/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief  
 *
 * This file contains all the APIs provided by the IPV6 module.
 *
 ***********************************************************************/
/* Infrastructure header files */
#include "adk_include.h"
#include "adk_ph_dev.h"
#include "acp_gen_interface.h"
#include "adk_utils_tree_update.h"
#include <netinet/in.h>
#include <arpa/inet.h>

#include "adk_shm.h"

/* IPV6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

/* IPV6 internal header files */
#include "adk_ipv6_internal.h"
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_dp_export.h"
#include "adk_ipv6_utility_counters.h"
#include "adk_ph_dev_dp.h"
#include "adk_qos_include_internal.h"
#define ADK_CONVERT_TO_V6_VRD__HANDLE(__DATA) ((adk_ipv6_vrd_handle_t)__DATA)
#define ADK_CONVERT_TO_V6_FLOW__HANDLE(__DATA) ((adk_flow_handle_t)__DATA)

extern adk_ipv6_dp_tables_t    *adk_ipv6_dp_tables;
#if 0
adk_hos_rule_id_t adk_ipv6_supp_proto_ruleids[ADK_IPV6_MAX_SUPPORTED_PROTOCOLS];
#endif

/*!
 * =============================================================
 * adk_ipv6_init
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully initialized IPV6 module.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL
 *
 * @brief  Initialises the IPV6 module.
 *
 * This function initializes the ADK IPV6 module. It allocates
 * the memory for the module specific data structure.  
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_init (adk_module_handle_t ipv6Handle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dp_status_t dpRetVal = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_module_service_fn_t svcFn;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_addr_prefix_t ipv6_addr_info;
    adk_ipv6_dp_route_entry_t route_entry;
    adk_ipv6_addr_prefix_t ipv6AddrPrefix; 
    adk_ipv6_dp_ingress_flow_sip_entry_t ingress_flow_sip_entry;
    adk_ipv6_dp_tables_t            *dp_pvt_p;
    adk_pal_dp_stage_info_t  *stage_info_p;
    uint16_t  i = 0;

#ifdef ADK_HOS_OLD_IMPL
    uint32_t pcap_id, mirror_id;
    adk_hos_pcap_attr_t pcap_attr;
    uint32_t vsi_id;
#endif

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return(retValue);
    }

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    
    /* Check if the IPv6 handle has already been initialized */
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    
    if ((ipv6ModuleInfo_p!= NULL) && (ipv6ModuleInfo_p->cookie == ADK_IPV6_MODULE_COOKIE))
    {
        ADK_ERROR_WRITE("IPv6 Module handle 0x%"ADKPRIh" is already initialized\n",
        ipv6Handle);
        ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
        return(ADK_SUCCESS);
    }

    /*! It is assumed that the module handle is valid. */
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /* Create memzone to store the structure which will have pointers to all
     * the data plane tables */
    retValue = adk_shm_create(ADK_IPV6_SHM_MEMZONE_NAME,
                              sizeof(adk_ipv6_dp_tables_t),
                              ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                              (adk_shm_handle_t **)&dp_pvt_p);

    if (retValue != ADK_SUCCESS)
    {
        ADK_ERROR_WRITE("Failed to create IPv6 memzone %s\n",
                          ADK_IPV6_SHM_MEMZONE_NAME);
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
        return(retValue);
    }

    ADK_MEMSET(dp_pvt_p, 0, sizeof(adk_ipv6_dp_tables_t));

    ipv6ModuleInfo_p =
        (adk_ipv6_info_t *) ADK_NVM_CALLOC (ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_MOD_INFO_OBJ, 1, sizeof(adk_ipv6_info_t));
    if (ipv6ModuleInfo_p == NULL) 
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Failed to allocate memory\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);

        goto adk_ipv6_init_return;
    }

 
    ADK_MODULE_INFO_PTR_SET(ipv6Handle, ipv6ModuleInfo_p);
 
    /*! Get handles for module's MPP LLI and ULI functions and register
     *  them with module manager.
     */
    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleTerminate   = adk_ipv6_shutdown;
    svcFn.moduleGetImc      = adk_ipv6_get_imc;

    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT]    = ADK_IPV6_LLI_FN_ID;
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT]    = ADK_IPV6_ULI_FN_ID;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_RING]       = 0;
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_RING]       = 0;

    ipv6ModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_IPV6_LLI_FN_ID;
    ipv6ModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_IPV6_ULI_FN_ID;
    ipv6ModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_RING]    = 0;
    ipv6ModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_RING]    = 0;

    retValue = adk_module_service_function_register (
                    sys_handle,
                    ipv6Handle,
                    &svcFn);
    if (retValue != ADK_SUCCESS)
    {
        /*! No need to map as this is the result of the ADK
         *  function.
         */
        ADK_LOG_ERROR(sys_handle, ipv6Handle,
            "Failed to register the functions\n");

        goto adk_ipv6_init_return;
    }
    ipv6ModuleInfo_p->ipv6IngClassifierRuleIDHandle =
	    adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle),
			    ADK_IPV6_ING_CLASSIFIER_RULE_INFO_OBJ,
			    ADK_IPV6_INGRESS_CLASSIFIER_MAX_RULES+1);
    if(NULL == ipv6ModuleInfo_p->ipv6IngClassifierRuleIDHandle)
    {
	    ADK_LOG_ERROR(sys_handle, ipv6Handle,
			    "Failed to allocate rm IPv6 ipv6IngClassifierRuleIDHandle\n");
	    retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
	    goto adk_ipv6_init_return;
    }
    adk_rm_resource_reserve_fixed(ipv6ModuleInfo_p->ipv6IngClassifierRuleIDHandle,
		    0);

    for (i = 0; i < ADK_DP_MAX_IPV6_INTERFACE; i++) {
        ipv6ModuleInfo_p->ipv6EgressQoSClassifierRuleIDHandle[i] =
            adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle),
                    ADK_IPV6_EGRESS_QOS_CLASSIFIER_INFO_OBJ,
                    ADK_IPV6_EGRESS_QOS_CLASSIFIER_MAX_RULES+1);
        if(NULL == ipv6ModuleInfo_p->ipv6EgressQoSClassifierRuleIDHandle)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle, 
                    "Failed to allocate ipv6EgressQoSClassifierRuleIDHandle\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            goto adk_ipv6_init_return;
        }
    }


    /* Create User defined resources. The PBR policy-set ID 0 is reserved as
      * invalid policy-set ID */
    ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle =
                      adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_PBR_RULESETS_OBJ,
				ADK_IPV6_PBR_RULE_SETS_MAX+1);
    if(NULL == ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sys_handle, ipv6Handle, 
                           "Failed to Allocate rm IP PbrPolicySetID Handle\n");
        goto adk_ipv6_init_return;
     }
     adk_rm_resource_reserve_fixed(ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle, 0);

     ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle =
                   adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_PBR_MAX_RULES_OBJ,
				IPV6_PBR_MAX_NUM_RULES+1);
     if(NULL == ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle)
     {
         retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
         ADK_LOG_ERROR(sys_handle, ipv6Handle,
                            "Failed to allocate rm IP PbrPolicyID Handle\n");
         goto adk_ipv6_init_return;
     }
     adk_rm_resource_reserve_fixed(ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle, 0);

    /* Updating default module log level to info */
    dp_pvt_p->logLevel = ADK_LOG_LEVEL_INFO;

    ipv6ModuleInfo_p->adk_ipv6_dp_tables  = dp_pvt_p;

    /* Enabling by default utility counters to true */
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->isIpv6UtilityCntEnabled = true;

    dpRetVal = adk_ipv6_dp_table_init(ipv6Handle, dp_pvt_p);
    if(dpRetVal != ADK_SUCCESS)
    {
           ADK_LOG_ERROR(sys_handle, ipv6Handle,
            "Failed to allocate resourses for Dataplane DBs \n");

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE); 
        goto adk_ipv6_init_return;
    }

    retValue = adk_pal_dp_stage_info_get( ADK_IPV6_LLI_REASSEMBLY_FN_ID, &stage_info_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, 
                    "Invalid destination handle specified as part of bind config\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_init_return;
    }
    rte_memcpy(&dp_pvt_p->ipv6_dp_reassmbly_stage_info, stage_info_p, sizeof(adk_pal_dp_stage_info_t));

    ipv6ModuleInfo_p->ipv6_dp_virtual_if_rsrc_hdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_VIRT_IF_RES_OBJ,
							ADK_IPV6_NO_OF_IPV6_VIRTUAL_INTERFACES);


    /* initial rm resources for filters */
    ipv6ModuleInfo_p->ipv6_dp_acl_rsrc_id = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_ACL_RES_OBJ,
							ADK_IPV6_FILTER_RULE_SETS_MAX);
    ipv6ModuleInfo_p->ipv6_dp_acl_stats_rsrc_id = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_ACL_STATS_OBJ,
							ADK_IPV6_NO_OF_ACL_ENTRIES);



     ipv6ModuleInfo_p->ipv6AddrStatsHandle =
	           adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_ADDR_STATS_OBJ, 
				   ADK_DP_MAX_IPV6_STATS*ADK_IPV6_INTF_ADDR_MAX);
     if (ipv6ModuleInfo_p->ipv6AddrStatsHandle == NULL)
     {
         retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
         ADK_LOG_ERROR(sys_handle, ipv6Handle,
                            "Failed to allocate rm IP ipv6AddrStatsHandle Handle\n");
         goto adk_ipv6_init_return;
     }
     adk_rm_resource_reserve_fixed(ipv6ModuleInfo_p->ipv6AddrStatsHandle, 0);

    /* TODO: add tree default entries here */

    /* ADD STATIC ENTRY FOR Route lookup & SIP LOOKUP
     *   - unspecified IP address
     *   - Loopbak adress
     *   - Link local address
     *   - multicast address 
     */   

    ADK_MEMSET(&ipv6_addr_info, 0, sizeof(adk_ipv6_addr_prefix_t));
    ADK_MEMSET(&route_entry, 0, sizeof(adk_ipv6_dp_route_entry_t));

    /* unspecified IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0;
    ipv6_addr_info.Ipv6Addr.addr[1] = 0;
    ipv6_addr_info.Ipv6Addr.addr[2] = 0;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LENGTH);
    
    route_entry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_DROP_WITH_COUNT;

    dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle,
                                ADK_IPV6_LOCAL_ROUTING_TABLE,
                                ipv6_addr_info,
                                &route_entry, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }

    /* loopback IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[3] = 1;
    dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                                ADK_IPV6_LOCAL_ROUTING_TABLE,
                                ipv6_addr_info,
                                &route_entry, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }

    /* Link local address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFE800000;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LINK_LOCAL_PREFIX_LENGTH);
    dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                                ADK_IPV6_LOCAL_ROUTING_TABLE,
                                ipv6_addr_info,
                                &route_entry, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }


#if  0
    /* TODO - Multicast feature support */
    /* multicast IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFF000000;
    dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                                ADK_IPV6_DEFAULT_VRD_ID,
                                ipv6_addr_info,
                                &route_entry,
                                &index);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }
#endif

    /*ADD STATIC ENTRY FOR SIP LOOKUP */
    ADK_MEMSET(&ipv6AddrPrefix, 0, sizeof(adk_ipv6_addr_prefix_t));
    ADK_MEMSET(&ingress_flow_sip_entry, 0, sizeof(adk_ipv6_dp_ingress_flow_sip_entry_t));

    ipv6AddrPrefix.Ipv6Addr.addr[0] = 0;
    ipv6AddrPrefix.Ipv6Addr.addr[1] = 0;
    ipv6AddrPrefix.Ipv6Addr.addr[2] = 0;
    ipv6AddrPrefix.Ipv6Addr.addr[3] = 0;

    ipv6AddrPrefix.prefixLength = (ADK_IPV6_PDU_ADDR_LENGTH);
   
    ingress_flow_sip_entry.apply_to_fwd_traffic = TRUE;
    ingress_flow_sip_entry.action = ADK_IPV6_ACTION_COUNT_AND_DROP;


    /* unspecified IPv6 address */
    dpRetVal = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                            ADK_IPV6_DEFAULT_VRD_ID,
                            ipv6AddrPrefix,
                            &ingress_flow_sip_entry);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }


    /* Loopback address */
    ipv6AddrPrefix.Ipv6Addr.addr[3] = 1;
    ingress_flow_sip_entry.apply_to_fwd_traffic = FALSE;
    ingress_flow_sip_entry.action = ADK_IPV6_ACTION_COUNT_AND_DROP;
    dpRetVal = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                            ADK_IPV6_DEFAULT_VRD_ID,
                            ipv6AddrPrefix,
                            &ingress_flow_sip_entry);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }

    /* Multicast address */
    ipv6AddrPrefix.Ipv6Addr.addr[0] = 0xFF000000;
    ipv6AddrPrefix.Ipv6Addr.addr[3] = 0;
    ipv6AddrPrefix.prefixLength = (ADK_IPV6_PDU_ADDR_MULTICAST_PREFIX_LENGTH);
    dpRetVal = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                            ADK_IPV6_DEFAULT_VRD_ID,
                            ipv6AddrPrefix,
                            &ingress_flow_sip_entry);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }



#if 0
    retValue = adk_pal_stage_register("adk_ipv6_dp_pkt_proc_lli",
                              ADK_PAL_STAGE_LB_TYPE_ATOMIC,
                              ADK_IPV6_LLI_FN_ID,
                              ADK_IPV6_LLI_FN_ID);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_init_return;
    }

    retValue = adk_pal_stage_register("adk_ipv6_dp_reassembly",
                              ADK_PAL_STAGE_LB_TYPE_ATOMIC,
                              ADK_IPV6_LLI_REASSEMBLY_FN_ID,
                              ADK_IPV6_LLI_REASSEMBLY_FN_ID);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_init_return;
    }

    retValue = adk_pal_stage_register("adk_ipv6_dp_reassembly_timeout",
                              ADK_PAL_STAGE_LB_TYPE_ATOMIC,
                              ADK_IPV6_LLI_REASS_TIMER_FN_ID,
                              ADK_IPV6_LLI_REASS_TIMER_FN_ID);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_init_return;
    }



    retValue = adk_pal_stage_register("adk_ipv6_dp_pkt_proc_uli",
                              ADK_PAL_STAGE_LB_TYPE_ATOMIC,
                              ADK_IPV6_ULI_FN_ID,
                              ADK_IPV6_ULI_FN_ID);
    if (retValue  != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_init_return;
    }
#endif

#ifdef ADK_HOS_OLD_IMPL
    /*allocate resources for packet capture */
    /* get the vsi queue id */
    if((adk_sif_pcap_vsi_id_get(sys_handle, &vsi_id)) == ADK_SUCCESS)
    {
        /* allocate pcap id */
        retValue = adk_sif_pcap_id_alloc(sys_handle, &pcap_id);
        if(dpRetVal != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,"unable to alloc resource for pkt capture \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
            goto adk_ipv6_init_return;
        }

        /* configure the entry function */
        retValue = adk_sif_pcap_id_config(sys_handle, pcap_id,
                ADK_IPV6_PACKET_CAPTURE_FN_ID);
        if(dpRetVal != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,"unable to configure pkt capture \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
            goto adk_ipv6_init_pcap_id_release_return;
        }

        pcap_attr.vsi_id    = vsi_id;
        pcap_attr.vsi_q     = 0;
        pcap_attr.pcap_data = pcap_id;

        retValue = adk_hos_pcap_create(&pcap_attr, &mirror_id);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,
                    "Failed to set the fast-path table %d\n",retValue);
            goto adk_ipv6_init_pcap_id_release_return;
        }

        ipv6ModuleInfo_p->mirror_id = mirror_id;
        ipv6ModuleInfo_p->pcap_id = pcap_id;
        ipv6ModuleInfo_p->enableHwCap = TRUE;
        dp_pvt_p->enable_hw_cap = TRUE;
    }
#endif

    retValue = adk_sif_module_egress_callback_register(sys_handle,
                ADK_IPV6_MODULE_ID,
                (adk_sif_module_egress_cb_fn_t)adk_ipv6_dp_egress_stats);
    if (retValue != ADK_SUCCESS) {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to register DP callback function\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }

    retValue = adk_sif_module_slowpath_callback_register(sys_handle,
                ADK_IPV6_MODULE_ID,
                (adk_sif_module_slowpath_cb_fn_t)adk_ipv6_dp_slowpath_policer_stats);
    if (retValue != ADK_SUCCESS) {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to register DP callback function\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto adk_ipv6_init_return;
    }
    adk_acl_info_t cpRouteInfo={0};
    cpRouteInfo.user = ADK_ACL_IPV6_ROUTE_TABLE;
    if(ADK_SUCCESS != adk_acl_create(&(ipv6ModuleInfo_p->cp_list),
            &cpRouteInfo))
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Could not allocate CP list for IPv6 Route rules !\n");
        goto adk_ipv6_init_return;
    }
    /* Set IPv6 moule cookie */
    ipv6ModuleInfo_p->cookie = ADK_IPV6_MODULE_COOKIE;

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_RETURN( retValue );

#ifdef ADK_HOS_OLD_IMPL
adk_ipv6_init_pcap_id_release_return:
    adk_sif_pcap_id_free(sys_handle, pcap_id);
#endif

adk_ipv6_init_return:
    if (ipv6ModuleInfo_p != NULL)
    {
        adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle);
        adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6_dp_acl_stats_rsrc_id);
        adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6_dp_acl_rsrc_id);
        adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6_dp_virtual_if_rsrc_hdl);
        adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle);
        for (i = 0; i < ADK_DP_MAX_IPV6_INTERFACE; i++) {
            adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6EgressQoSClassifierRuleIDHandle[i]);
        }
        ADK_NVM_FREE (ipv6ModuleInfo_p); 
    }
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}    /* -----  end of function adk_ipv6_init  ----- */

/*!
 * =============================================================
 * adk_ipv6_exception_handler_register
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] exceptionHandler_p The exception handler function
 * information.
 *
 * @retval ADK_SUCCESS Successfully registered the exception
 * handler.
 * @retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @retval ADK_E_MEM_UNAVAILABLE There is not enough memory
 * available.
 *
 * @brief  Registers an exception handler function.
 *
 * This API allows the application to register exception
 * handlers for various exceptions. NOTE that it is the
 * responsibility of the caller to register the exception
 * function with the PH as call back function when the
 * exception handler resides on the CPU.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_exception_handler_register (adk_module_handle_t ipv6Handle,
    const adk_ipv6_exception_handler_t *exceptionHandler_p)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_dp_excep_entry_t excep_entry;
    uint32_t policerFlowId = ADK_QOS_INVALID_POLICER_FLOW_ID;

    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_exception_handler_register_null_handle_return;
    }

    /*! Validate some input parameters */
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
        goto adk_ipv6_exception_handler_register_null_handle_return;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);

    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_exception_handler_register_return;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (exceptionHandler_p == NULL)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Invalid argument\n");
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);

        goto adk_ipv6_exception_handler_register_return;
    }

    switch(exceptionHandler_p->exceptionType)
    {
        case ADK_IPV6_EXCEPTION_ROUTE_NOT_FOUND:
        case ADK_IPV6_UDP_INGRESS_CHECKSUM_INVALID:
        case ADK_IPV6_EXCEPTION_UNSUPPORTED_L4_PROTO:
        case ADK_IPV6_EXCEPTION_HOP_LIMIT_ZERO:
        case ADK_IPV6_INTERFACE_FORWARDING_DISABLE:
        case ADK_IPV6_EXCEPTION_INCORRECT_IP_VERSION:
        case ADK_IPV6_EXCEPTION_EXT_HEADER_OCCURED_TWICE:
        case ADK_IPV6_EXCEPTION_FRAG_OVERLAP:
        case ADK_IPV6_EXCEPTION_REASSEMBLY_TIMER_EXPIRED:
        case ADK_IPV6_EXCEPTION_MTU_VIOLATION:
        case ADK_IPV6_EXCEPTION_PARAM_PROBLEM:
            break;
        default:
            retValue1=ADK_ERROR_SEV_LEVEL_SET(ADK_IPV6_E_INVALID_EXCEPTION_TYPE);
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue1,
                               "Invalid Exception Type \n");
            goto adk_ipv6_exception_handler_register_return;
    }

    if (exceptionHandler_p->policer_handle)
    {
        retValue1 = adk_qos_policing_flow_id_get(sys_handle, exceptionHandler_p->policer_handle, &policerFlowId);
        if (ADK_SUCCESS != retValue1) {
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue1,
                    "Invalid policer handle\n");
            goto adk_ipv6_exception_handler_register_return;
        }
    }
    excep_entry.type = exceptionHandler_p->exceptionType;
    excep_entry.devId = exceptionHandler_p->devId;
    excep_entry.queueId = exceptionHandler_p->queueId;
    excep_entry.priority = exceptionHandler_p->priority;
    excep_entry.callback_fun_id = exceptionHandler_p->cbFnHandle;
    excep_entry.policerId = policerFlowId;

    if( adk_ipv6_dp_excep_entry_add(ipv6Handle,
                exceptionHandler_p->exceptionType,
                &excep_entry) != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,
                "Failed to enable IPv6 exception \n");

        goto adk_ipv6_exception_handler_register_return;
    }

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
    ADK_RETURN( retValue1 );
adk_ipv6_exception_handler_register_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

adk_ipv6_exception_handler_register_null_handle_return:
    ADK_RETURN( retValue1 );
}    /* -----  end of function adk_ipv6_exception_handler_register  ----- */

/*!
 * =============================================================
 * adk_ipv6_exception_handler_unregister
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] exceptionHandler_p The exception handler function
 * information to unregister.
 *
 * @retval ADK_SUCCESS Successfully un-registered the exception
 * handler.
 * @retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @retval ADK_E_MEM_UNAVAILABLE There is not enough memory
 * available.
 *
 * @brief  Un-registers an exception handler function.
 *
 * This API allows the application to unregister exception
 * handlers for various exceptions. NOTE that it is the
 * responsibility of the caller to unregister the exception
 * function with the PH as call back function when the
 * exception handler resides on the CPU.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_exception_handler_unregister (
    adk_module_handle_t ipv6Handle,
    const adk_ipv6_exception_handler_t *exceptionHandler_p)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;

    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_exception_handler_unregister_null_handle_return;
    }
  
    /*! Validate some input parameters */
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
       goto adk_ipv6_exception_handler_unregister_null_handle_return;
    }
 
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);


    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle,
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_exception_handler_unregister_return;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (exceptionHandler_p == NULL)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Invalid argument\n");
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);

        goto adk_ipv6_exception_handler_unregister_return;
    }

    switch(exceptionHandler_p->exceptionType)
    {
        case ADK_IPV6_EXCEPTION_ROUTE_NOT_FOUND:
        case ADK_IPV6_UDP_INGRESS_CHECKSUM_INVALID:
        case ADK_IPV6_EXCEPTION_UNSUPPORTED_L4_PROTO:
        case ADK_IPV6_EXCEPTION_HOP_LIMIT_ZERO:
        case ADK_IPV6_INTERFACE_FORWARDING_DISABLE:
        case ADK_IPV6_EXCEPTION_INCORRECT_IP_VERSION:
        case ADK_IPV6_EXCEPTION_FRAG_OVERLAP:
        case ADK_IPV6_EXCEPTION_REASSEMBLY_TIMER_EXPIRED:
        case ADK_IPV6_EXCEPTION_MTU_VIOLATION:
        case ADK_IPV6_EXCEPTION_PARAM_PROBLEM:
        case ADK_IPV6_EXCEPTION_EXT_HEADER_OCCURED_TWICE:
            break;
        default:
            retValue1=ADK_ERROR_SEV_LEVEL_SET(ADK_IPV6_E_INVALID_EXCEPTION_TYPE);
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue1,
                               "Invalid Exception Type \n");
            goto adk_ipv6_exception_handler_unregister_return;

    }

    if( adk_ipv6_dp_excep_entry_rem(ipv6Handle,
                exceptionHandler_p->exceptionType) != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,
                "Failed to enable IPv6 exception \n");
        goto adk_ipv6_exception_handler_unregister_return;
    }

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);

    ADK_RETURN( retValue1 );

adk_ipv6_exception_handler_unregister_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
adk_ipv6_exception_handler_unregister_null_handle_return:

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue1 );
}    /* -----  end of function adk_ipv6_exception_handler_unregister  ----- */

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_extract_bits
 *
 * @brief
 *
 * This function is used to parse the output values based on their size
 * =============================================================
 */
uint32_t adk_ipv6_extract_bits( uint32_t __attribute__((unused)) value, uint32_t __attribute__((unused)) Offset, uint32_t __attribute__((unused)) size)
{

    return (0);
}

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_tables_dump
 *
 * @brief
 *
 * This function is called to dump the fastpath tables in readable format
 * =============================================================
 */
adk_status_t
adk_ipv6_tables_dump(adk_module_handle_t ipv6Handle, 
                     adk_ipv6_tree_conf_t *ipv6TreeConf_p)
{

    adk_status_t      retValue = ADK_SUCCESS;
    adk_sys_handle_t  sys_handle;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        return retValue;
    }
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
       return retValue;
    }
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
 
    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_tree_dump_return;
    }

    if(ipv6TreeConf_p->group < ADK_IPV6_TABLES_NONE ||
            ipv6TreeConf_p->group > ADK_IPV6_TABLES_ALL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "Invalid IPv6 tree group \n");
        goto adk_ipv6_tree_dump_return;
    }

    if( ipv6TreeConf_p->treeType != ADK_IPV6_LPM_TREE )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "TREE type is not supported \n");
        goto adk_ipv6_tree_dump_return;
    }

    retValue = adk_ipv6_dump_tables(ipv6Handle, ipv6TreeConf_p); 
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(retValue);
        ADK_LOG_ERROR(sys_handle,ipv6Handle,
                           "Failed to dump IPv6 trees\n");
        goto adk_ipv6_tree_dump_return;
    }

adk_ipv6_tree_dump_return:    
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN (retValue);/* -----  end of function adk_ipv6_dscp_group_stats_get ----- */
}

/*!
 * =============================================================
 * adk_ipv6_shutdown
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully cleaned up IPV6 module.
 * @retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 *
 * @brief  Cleans up the IPV6 module.
 *
 * This function cleans up the ADK IPV6 module. It frees
 * the memory for the module specific data structure and other
 * spaces reserved for fast-path tables. NOTE that it is assumed
 * that the application will delete the flows if any it created
 * and unregister the exception handler functions.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_shutdown (adk_module_handle_t ipv6Handle)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_dp_status_t dpRetVal = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_module_service_fn_t svcFn;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_addr_prefix_t ipv6_addr_info;
    adk_ipv6_dp_route_entry_t route_entry;
    adk_ipv6_addr_prefix_t ipv6AddrPrefix; 
    adk_ipv6_dp_ingress_flow_sip_entry_t ingress_flow_sip_entry;
    const struct rte_memzone *mz = NULL;
    uint16_t i = 0;

    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_shutdown_return;
    }
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
       goto adk_ipv6_shutdown_return;
    }
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
 
    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_shutdown_lock_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    /*! Register NULL MPP LLI and ULI functions */
    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    retValue1 = adk_module_service_function_register (
                    sys_handle,
                    ipv6Handle,
                    &svcFn);
    if (retValue1 != ADK_SUCCESS)
    {
        /*! No need to map as this is the resule of the ADK
         *  function.
         */
        ADK_LOG_ERROR(sys_handle, ipv6Handle,
            "Failed to register NULL functions\n");
        /*! No need to quit, continue further */
    }

    retValue1 = adk_sif_module_egress_callback_unregister(sys_handle,
                ADK_IPV6_MODULE_ID);
    if (retValue1 != ADK_SUCCESS) {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,
                    "Failed to unregister DP callback function error - %d\n",retValue1);
            /*! No need to quit, continue further */ 
    }

    retValue1 = adk_sif_module_slowpath_callback_unregister(sys_handle,
                ADK_IPV6_MODULE_ID);
    if (retValue1 != ADK_SUCCESS) {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,
                    "Failed to unregister DP callback function error - %d\n",retValue1);
            /*! No need to quit, continue further */ 
    }

    /* release rm resources */
    adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6_dp_virtual_if_rsrc_hdl);
    adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6_dp_acl_rsrc_id);
    adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6_dp_acl_stats_rsrc_id);
    for (i = 0; i < ADK_DP_MAX_IPV6_INTERFACE; i++) {
        adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6EgressQoSClassifierRuleIDHandle[i]);
    }
    adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle);
    adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle);
    adk_rm_resource_destroy(ipv6ModuleInfo_p->ipv6AddrStatsHandle);

    /* DEL STATIC ENTRY FROM Route lookup & SIP LOOKUP
     *   - unspecified IP address
     *   - Loopbak adress
     *   - Link local address
     *   - multicast address 
     */   
    ADK_MEMSET(&ipv6_addr_info, 0, sizeof(adk_ipv6_addr_prefix_t));
    ADK_MEMSET(&route_entry, 0, sizeof(adk_ipv6_dp_route_entry_t));

    /* unspecified IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0;
    ipv6_addr_info.Ipv6Addr.addr[1] = 0;
    ipv6_addr_info.Ipv6Addr.addr[2] = 0;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LENGTH);
    
    route_entry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_DROP_WITH_COUNT;

    dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle,
                                ADK_IPV6_DEFAULT_VRD_ID,
                                ipv6_addr_info, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to rem the default route entry \n");
    }

    /* loopback IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[3] = 1;
    dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                                ADK_IPV6_DEFAULT_VRD_ID,
                                ipv6_addr_info, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to rem the default route entry \n");
    }

    /* Link Local IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFE800000;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LINK_LOCAL_PREFIX_LENGTH);
    dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                                ADK_IPV6_DEFAULT_VRD_ID,
                                ipv6_addr_info, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to rem the default route entry \n");
    }

    /*ADD STATIC ENTRY FOR SIP LOOKUP */
    ADK_MEMSET(&ipv6AddrPrefix, 0, sizeof(adk_ipv6_addr_prefix_t));
    ADK_MEMSET(&ingress_flow_sip_entry, 0, sizeof(adk_ipv6_dp_ingress_flow_sip_entry_t));

    ipv6AddrPrefix.Ipv6Addr.addr[0] = 0;
    ipv6AddrPrefix.Ipv6Addr.addr[1] = 0;
    ipv6AddrPrefix.Ipv6Addr.addr[2] = 0;
    ipv6AddrPrefix.Ipv6Addr.addr[3] = 0;
    ipv6AddrPrefix.prefixLength = (ADK_IPV6_PDU_ADDR_LENGTH);
  
    /* unspecified IPv6 address */
    dpRetVal = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                            ADK_IPV6_DEFAULT_VRD_ID,
                            ipv6AddrPrefix);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default SIP entry \n");
    }

    /* Loopback address */
    ipv6AddrPrefix.Ipv6Addr.addr[3] = 1;
    dpRetVal = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                            ADK_IPV6_DEFAULT_VRD_ID,
                            ipv6AddrPrefix);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default SIP entry \n");
    }

    /* Multicast address */
    ipv6AddrPrefix.Ipv6Addr.addr[0] = 0xFF000000;
    ipv6AddrPrefix.Ipv6Addr.addr[3] = 0;
    ipv6AddrPrefix.prefixLength = (ADK_IPV6_PDU_ADDR_MULTICAST_PREFIX_LENGTH);
    dpRetVal = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                            ADK_IPV6_DEFAULT_VRD_ID,
                            ipv6AddrPrefix);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,"Failed to add the default SIP entry \n");
    }
    /* Free CP routing table */
    adk_acl_rule_t route;
    adk_acl_get_reset(ipv6ModuleInfo_p->cp_list);
    while(ADK_SUCCESS == adk_acl_next_rule_get(
             ipv6ModuleInfo_p->cp_list, &route))
    {
        if(NULL != route.user_data)
        {
            ADK_NVM_FREE(route.user_data);
            route.user_data = NULL;
        }
    }
    adk_acl_free(ipv6ModuleInfo_p->cp_list);
#ifdef ADK_HOS_OLD_IMPL
    if(ipv6ModuleInfo_p->enableHwCap)
    {
        /*release the resources allocated for packet capture */
        retValue1 = adk_hos_pcap_delete(ipv6ModuleInfo_p->mirror_id);
        if(retValue1 != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,
                    "Failed to reset the fast-path table %d\n",retValue1);
        }

        /* allocate pcap id */
        retValue1 = adk_sif_pcap_id_free(sys_handle, ipv6ModuleInfo_p->pcap_id);
        if(dpRetVal != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,"unable to release resource of pkt capture \n");
        }
    }
#endif

    /* Lookup the shared memory that gets created by control plane IPv6
     * module initialization */
    ADK_IPV6_LOG_EXT_API(ipv6Handle, "rte_memzone_lookup", "Lookup the shared memory %s \n", ADK_IPV6_SHM_MEMZONE_NAME);
    mz = rte_memzone_lookup(ADK_IPV6_SHM_MEMZONE_NAME);
    if (mz != NULL)
    {

        retValue1 = adk_ipv6_dp_table_release(ipv6Handle, (adk_ipv6_dp_tables_t *)mz->addr);
        if (retValue1 != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sys_handle, ipv6Handle,
                    "Failed to register ipv6 dp tables \n");
            /*! No need to quit, continue further */
        }

        if (adk_shm_destroy(ADK_IPV6_SHM_MEMZONE_NAME))
        {
            ADK_ERROR_WRITE("Failed to release IPv6 memzone %s\n",
                    ADK_IPV6_SHM_MEMZONE_NAME);
        }
    }
    else
    {
        ADK_ERROR_WRITE("Failed to lookup IPv6 shared memory %s\n",
                ADK_IPV6_SHM_MEMZONE_NAME);
    }

    /*Unregister IPv6 parser function in PH to Decode exception params*/
    retValue1 = adk_ph_dev_rx_parse_unregister(ADK_IPV6_MODULE_ID);
    if (retValue1 != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,
                "Failed to unregister the ph_dev_rx_parse function for IPv6\n");
    }

    ADK_MODULE_INFO_PTR_SET(ipv6Handle, NULL);

    /* Set IPv6 moule cookie to zero */
    ipv6ModuleInfo_p->cookie = 0x0;

    /*! Free the IPV6 module info structure */
    ADK_NVM_FREE (ipv6ModuleInfo_p);
    ipv6ModuleInfo_p = NULL;

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
    ADK_RETURN( retValue1 );

adk_ipv6_shutdown_lock_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);

adk_ipv6_shutdown_return:

_lock_release_failed:
_lock_acquire_failed:
    if (ipv6ModuleInfo_p != NULL)
    {
        ADK_NVM_FREE (ipv6ModuleInfo_p);
    }

    ADK_RETURN( retValue1 );
}    /* -----  end of function adk_ipv6_shutdown  ----- */

/*!
 * =============================================================
 *
 * adk_ipv6_get_imc
 *
 * @param[in]   ipv6Handle  Handle to the IPv6 module
 * @param[in]   bindInfo_p Binding information
 * @param[out]  imcFnId IMC function ID to return
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @retval  ADK_E_IMC_NOT_FOUND     There is no IMC defined.
 *
 * @brief   
 *
 * This API is not supported
 *
 *Function to get the IMC function ID.
 *
 * This function is used to get the IMC function. This is called
 * from the ADK SIF. This should not be called directly from
 * the application.
 *
 * NOTE: that if there is no IMC function required, ADK_SUCCESS
 * should be returned.
 *
 * =============================================================
 */
adk_status_t adk_ipv6_get_imc (adk_module_handle_t __attribute__((unused)) ipv6Handle,
                             void __attribute__((unused)) *bindInfo_p,
                             uint32_t __attribute__((unused)) *imcFnId)
{
    return (ADK_SUCCESS);
}

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_get_entry_func
 *
 * @param[in] ipv6Handle      Handle to the IPv6 module
 *
 * @param[in] bindInfo_p    Binding information
 *
 * @param[out] entry_fn_id  Entry function ID to return
 *
 * @retval ADK_SUCCESS              Successful completion. If IMC is not
 * required than output IMC function will be set to 0.
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv6 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid parameter pass e.g. null pointer
 * parameter or invalid flow handle.
 *
 * @retval ADK_E_UNSUPPORTED        An unsupported binding requested. It could
 * be unsupported type (F2F, F2M, etc), path (DIRECT or INDIRECT), direction 
 * (INGRESS or EGRESS or location (CPU or ENGINE).
 *
 * @brief
 *
 * This API is not supported
 *
 * This function is used to get the Entry function when binding to IPv6 module
 * as the destination module. This is called from the ADK SIF. This function 
 * must not be called directly from the application.
 *
 * =============================================================
 */
adk_status_t adk_ipv6_get_entry_func(adk_module_handle_t __attribute__((unused)) ipv6Handle,
                                     void __attribute__((unused)) *bindInfo_p,
                                     uint32_t __attribute__((unused)) *entry_fn_id)
{
    return (ADK_SUCCESS);
}

/*!
 * =============================================================
 *  adk_ipv6_service_function_update
 *
 * @param[in]  ipv6Handle  - IPv6 module handle. Return at the
 * time of IPv6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully initialized IPv6 module.
 * 
 * @retval ADK_E_INVALID_PARAMETER - Invalid IPv6 mdoule handle.
 *
 * @brief  
 *
 * This function registers the CPU based service functions with
 * SIF. This function is called by adk_module_handle_get
 * function.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_service_function_update (adk_module_handle_t ipv6Handle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_module_service_fn_t svcFn;
    adk_ipv6_info_t *ipv6ModuleInfo_p;

    if (!ipv6Handle)
    {
      ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
      retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
      goto adk_ipv6_service_function_update_return;
    }
    /*! It is assumed that the module handle is valid. */
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    ipv6ModuleInfo_p =  ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (!ipv6ModuleInfo_p) 
    {
      ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
      retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);

      goto adk_ipv6_service_function_update_return;
    }

    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleTerminate = adk_ipv6_shutdown;
    svcFn.moduleGetImc = adk_ipv6_get_imc;
    svcFn.moduleGetEntryFunc = adk_ipv6_get_entry_func;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ipv6ModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_DEFAULT];
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ipv6ModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_DEFAULT];
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_RING]    = ipv6ModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_RING];
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_RING]    = ipv6ModuleInfo_p->cpuEntryFn_ULI[_DEST_FUNC_RING];

    /*Register IPv6 parser function in PH to Decode exception params*/
    retValue = adk_ph_dev_rx_parse_register((void*)adk_ph_ipv6_rx_parse,
                                ADK_IPV6_MODULE_ID);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle,
                "Failed to register the ph_dev_rx_parse function for ipv6 in CP\n");
    }
    retValue = adk_module_service_function_update (
                    sys_handle,
                    ipv6Handle,
                    &svcFn);
    if (retValue != ADK_SUCCESS)
    {
       ADK_LOG_ERROR(sys_handle, ipv6Handle, "Failed to register the functions\n");
       retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);  
    }
    ADK_RETURN( retValue );
adk_ipv6_service_function_update_return:

    ADK_RETURN( retValue );
}    /* -----  end of function adk_ipv6_service_function_update  ----- */

/*!
 * =============================================================
 * adk_ph_ipv6_rx_parse  adk_ipv6.c
 *
 * @param[in]   task Pointer to rte_mbuf.
 *
 * @param[out]  dp_message   Pointer to structure having the decoded parameters
 *                            sent by eth data path code to application.
 *
 * @retval   ADK_SUCCESS - successfully decoded the parameters
 * handler
 *
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 *                                    passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function helps in interpreting the parameters present in the task
 * Parameters that IPv6 module send along with the packet to the Application
 * via PH Dev.
 *
 * =============================================================
 */

adk_status_t
adk_ph_ipv6_rx_parse (
        void *task,
        void **dp_message)
{
    adk_status_t ret = ADK_SUCCESS;
    struct rte_mbuf *mbuf_p;
    adk_metadata_t *adk_md_ptr;
    adk_ph_dev_dp_metadata_t *phMeta_p;
    adk_ipv6_ph_dev_rx_param_t *ph_metadata;
    if (!task || !dp_message)
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ph_ipv6_rx_parse_decode_exit;
    }

    mbuf_p = (struct rte_mbuf *)task;
    adk_md_ptr = ADK_METADATA_PTR(mbuf_p);
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(adk_md_ptr->imc_param_start);
    ph_metadata = (adk_ipv6_ph_dev_rx_param_t*)&(phMeta_p->imc_param_start);

    *dp_message = (adk_ipv6_exception_task_param_t *)ph_metadata;

_adk_ph_ipv6_rx_parse_decode_exit:

    ADK_RETURN(ret);
}

/*!
 * =============================================================
 * @internal
 * adk_ipv6_prefix_mask
 * @param[in]    prefix  - IPV6 Address Prefix length.
 * @param[in]    mask    - Index for mask
 * @param[in,out]  
 *
 * @retval : None 
 *
 * @brief  
 * This API is not supported 
 *
 * Internal support function for prefix calculation.
 *
 * =============================================================
 */
void adk_ipv6_prefix_mask( uint32_t __attribute__((unused)) prefix,
                           uint32_t __attribute__((unused)) *mask)
{
} /* -----  end of function adk_ipv6_prefix_mask  ----- */


/*!
 * =============================================================
 * adk_ipv6_checksum_validation_enable
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully checksum validation is 
 * enabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 * It enables checksum validation for all packets
 *
 * This function enables checksum validation for all packets
 *
 * NOTE: This API is not yet supported
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_checksum_validation_enable (
        adk_module_handle_t __attribute__((unused)) ipv6Handle)
{

    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * adk_ipv6_checksum_validation_disable
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully checksum validation is 
 * disabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 *It disables checksum validation for all packets
 *
 * This function disables checksum validation for all packets
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_checksum_validation_disable (
        adk_module_handle_t __attribute__((unused)) ipv6Handle)
{
    adk_status_t retValue = ADK_SUCCESS;
    return retValue;
}


/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_vrd_validate_handle
 *
 * @param[in] ipv6VrdHandle    IPv6 VRD instance handle
 *
 * @retval TRUE         The IPv6 VRF handle is valid
 *
 * @retval FALSE        The IPv6 VRF handle is invalid
 *
 * @brief  
 *
 * API to validate an IPv6 VRF handle.
 * =============================================================
 */
bool adk_ipv6_vrd_valid_handle(adk_ipv6_vrd_handle_t ipv6VrdHandle,
                                     uint32_t        *vrdId)
{
    adk_ipv6_vrd_info_t       *vrd_info_p;

    vrd_info_p = (adk_ipv6_vrd_info_t *) (ipv6VrdHandle);
                  
    if (vrd_info_p == NULL)
    {
        *vrdId   = ADK_IPV6_DEFAULT_VRD_ID;
        return TRUE;
    }

    if (vrd_info_p->cookie != ADK_IPV6_VRD_ID_COOKIE)
        return FALSE;

    *vrdId = vrd_info_p->vrdAttr.vrd_id;

    return TRUE;
}


/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_valid_rule_set_handle
 *
 * @param[in] filterRuleSetHandle   IPv6 filter rule set handle
 *
 * @retval TRUE         The IPv6 filter rule set handle is valid
 *
 * @retval FALSE        The IPv6 filter rule set handle is invalid
 *
 * @brief  
 *
 * API to validate an IPv6 filter rule set handle. This API is not for
 * validating the filter rule handle but only the set.
 * =============================================================
 */
bool adk_ipv6_valid_rule_set_handle(
                adk_ipv6_filter_rule_set_handle_t filterRuleSetHandle)
{
    ipv6_filter_rule_set_info_t *filter_info_p = NULL;

    if (!filterRuleSetHandle)
        return FALSE;

    filter_info_p = (ipv6_filter_rule_set_info_t *) filterRuleSetHandle;

    if (filter_info_p->cookie != ADK_IPV6_RULESET_COOKIE)
        return FALSE;

    return TRUE;
}

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_valid_policy_set_handle
 *
 * @param[in] pbrPolicySetHandle   IPv6 pbr rule set handle
 *
 * @retval TRUE         The IPv6 pbr rule set handle is valid
 *
 * @retval FALSE        The IPv6 pbr rule set handle is invalid
 *
 * @brief  
 *
 * API to validate an IPv6 pbr rule set handle. This API is not for
 * validating the pbr rule handle but only the set.
 * =============================================================
 */
bool adk_ipv6_valid_policy_set_handle(
    adk_ipv6_pbr_policy_set_handle_t pbrPolicySetHandle)
{
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;

    if (!pbrPolicySetHandle)
        return FALSE;

    policy_info_p = (ipv6_pbr_rule_set_info_t *) pbrPolicySetHandle;

    if (policy_info_p->cookie != ADK_IPV6_PBR_RULESET_COOKIE)
        return FALSE;

    return TRUE;
}

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_valid_acl_rule_handle
 *
 * @param[in] filterRuleHandle   IPv6 filter rule handle
 *
 * @retval TRUE         The IPv6 filter rule  handle is valid
 *
 * @retval FALSE        The IPv6 filter rule  handle is invalid
 *
 * @brief  
 *
 * API to validate an IPv6 filter rule  handle. 
 * =============================================================
 */
bool adk_ipv6_valid_acl_rule_handle(
                adk_handle_t filterRuleHandle)
{
    adk_ipv6_filter_rule_handle_info_t *filter_info_p = NULL;

    if (!filterRuleHandle)
        return FALSE;

    filter_info_p = (adk_ipv6_filter_rule_handle_info_t *) filterRuleHandle;

    if (filter_info_p->cookie != ADK_IPV6_FILTER_RULE_HANDLE_COOKIE)
        return FALSE;

    return TRUE;
}

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_valid_pbr_rule_handle
 *
 * @param[in] pbrRuleHandle   IPv6 PBR rule handle
 *
 * @retval TRUE         The IPv6 PBR rule  handle is valid
 *
 * @retval FALSE        The IPv6 PBR rule  handle is invalid
 *
 * @brief  
 *
 * API to validate an IPv6 PBR rule  handle. 
 * =============================================================
*/
bool adk_ipv6_valid_pbr_rule_handle(
                adk_handle_t pbrRuleHandle)
{
    adk_ipv6_pbr_rule_handle_info_t *pbr_info_p = NULL;

    if (!pbrRuleHandle)
        return FALSE;

    pbr_info_p = (adk_ipv6_pbr_rule_handle_info_t *) ADK_SYS_FLOW_ATTR_PTR_GET(pbrRuleHandle);

    if (pbr_info_p->cookie != ADK_IPV6_PBR_RULE_HANDLE_COOKIE)
        return FALSE;

    return TRUE;
}

/*!
 * =============================================================
 * adk_ipv6_destination_mtu_add
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] vrd_id           IPv6 VRD instance Id
 * @param[in] dipAddr          IPv6 Destination address
 * @param[in] mtu              Path MTU for the dipAddr
 *
 * @retval ADK_SUCCESS Successfully checksum validation is 
 * enabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  API configures the IPv6 destination address based MTU 
 *  This API adds an entry to destination MTU table.
 *  PMTU lookup from this table applies only for locally orginated packets. 
 *
 * =============================================================
 */
adk_status_t   
adk_ipv6_destination_mtu_add(
                adk_module_handle_t ipv6Handle,
                uint32_t  vrd_id,
                adk_ipv6_addr_t dipAddr,
                uint16_t mtu)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ipv6_dp_dest_mtu_entry_t dest_mtu_entry;
    uint16_t flowId;

    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_destination_mtu_add_null_handle_return;
    }

    /*! Validate some input parameters */
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
       goto adk_ipv6_destination_mtu_add_null_handle_return;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);
 
    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_destination_mtu_add_return;
    }

    /* Validate MTU input */
    if ((mtu) && (mtu < ADK_IPV6_INTERFACE_MIN_MTU)) 
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                      "MTU must be between 1280 to 65535 \n");
        goto adk_ipv6_destination_mtu_add_return;
    }

    /*Validate Routing Domain */
    if (vrd_id > ADK_IPV6_DEFAULT_VRD_ID)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sys_handle,ipv6Handle,"Invalid IPv6 routing domain %d \n", vrd_id);
        goto adk_ipv6_destination_mtu_add_return;
    }

    ADK_MEMSET(&dest_mtu_entry, 0, sizeof(adk_ipv6_dp_dest_mtu_entry_t));

    dest_mtu_entry.mtu = mtu;
    dest_mtu_entry.vrd_id = vrd_id;
    ADK_MEMCPY(dest_mtu_entry.ipv6_addr.addr, 
                    dipAddr.addr, 
                    sizeof(adk_ipv6_addr_t));

    /* update the IPv6 destination MTU Table */
    retValue1 = adk_ipv6_dp_dest_mtu_entry_add(ipv6Handle,
                    vrd_id,
                    &dest_mtu_entry,
                    &flowId);
    if(retValue1 != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "failed to add dest mtu entry \n");
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_destination_mtu_add_return;
    }
    

adk_ipv6_destination_mtu_add_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

adk_ipv6_destination_mtu_add_null_handle_return:
    ADK_RETURN( retValue1 );
}


/*!
 * =============================================================
 * adk_ipv6_destination_mtu_modify
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] vrd_id           IPv6 VRD instance Id
 * @param[in] dipAddr          IPv6 Destination address
 * @param[in] mtu              Path MTU for the dipAddr
 *
 * @retval ADK_SUCCESS Successfully checksum validation is 
 * enabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  API Modifies the existing IPv6 destination address based MTU 
 *
 * =============================================================
 */
adk_status_t   
adk_ipv6_destination_mtu_modify(
                adk_module_handle_t ipv6Handle,
                uint32_t  vrd_id,
                adk_ipv6_addr_t dipAddr,
                uint16_t mtu)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ipv6_dp_dest_mtu_entry_t dest_mtu_entry;

    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_destination_mtu_add_null_handle_return;
    }

    /*! Validate some input parameters */
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
       goto adk_ipv6_destination_mtu_add_null_handle_return;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);
 
    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_destination_mtu_add_return;
    }

    /* Validate MTU input */
    if ((mtu) && (mtu < ADK_IPV6_INTERFACE_MIN_MTU)) 
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                      "MTU must be between 1280 to 65535 \n");
        goto adk_ipv6_destination_mtu_add_return;
    }

    /*Validate Routing Domain */
    if (vrd_id > ADK_IPV6_DEFAULT_VRD_ID)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sys_handle,ipv6Handle,"Invalid IPv6 routing domain %d \n", vrd_id);
        goto adk_ipv6_destination_mtu_add_return;
    }

    ADK_MEMSET(&dest_mtu_entry, 0, sizeof(adk_ipv6_dp_dest_mtu_entry_t));

    dest_mtu_entry.mtu = mtu;
    dest_mtu_entry.vrd_id = vrd_id;
    ADK_MEMCPY(dest_mtu_entry.ipv6_addr.addr, 
                    dipAddr.addr, 
                    sizeof(adk_ipv6_addr_t));

    /* update the IPv6 destination MTU Table */
    retValue1 = adk_ipv6_dp_dest_mtu_entry_modify(ipv6Handle,
                    vrd_id,
                    &dest_mtu_entry);
    if(retValue1 != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "failed to add dest mtu entry \n");
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_destination_mtu_add_return;
    }
    
    ADK_LOG_INFO(sys_handle, ipv6Handle, "added dest mtu entry \n");

adk_ipv6_destination_mtu_add_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

adk_ipv6_destination_mtu_add_null_handle_return:
    ADK_RETURN( retValue1 );
}



/*!
 * =============================================================
 * adk_ipv6_destination_mtu_rem
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] vrd_id           IPv6 VRD instance Id
 * @param[in] dipAddr          IPv6 Destination address
 *
 * @retval ADK_SUCCESS Successfully checksum validation is 
 * enabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  API removes an entry from IPv6 destination address based MTU table
 *
 * =============================================================
 */
adk_status_t   
adk_ipv6_destination_mtu_rem(
                adk_module_handle_t ipv6Handle,
                uint32_t vrd_id,
                adk_ipv6_addr_t dipAddr)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_destination_mtu_rem_null_handle_return;
    }

    /*! Validate some input parameters */
    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
       goto adk_ipv6_destination_mtu_rem_null_handle_return;
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);
 
    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_destination_mtu_rem_return;
    }

    /*Validate Routing Domain */
    if ( vrd_id >ADK_IPV6_DEFAULT_VRD_ID)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sys_handle,ipv6Handle,"Invalid IPv6 routing domain %d\n", vrd_id);
        goto adk_ipv6_destination_mtu_rem_return;
    }

    /* update the IPv6 destination MTU Table */
    retValue1 = adk_ipv6_dp_dest_mtu_entry_rem(ipv6Handle,
                    vrd_id,
                    (uint8_t *) &dipAddr.addr[0]);
    if(retValue1 != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "failed to rem dest mtu entry \n");
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_destination_mtu_rem_return;
    }
    
    ADK_LOG_INFO(sys_handle, ipv6Handle, "removed dest mtu entry \n");

adk_ipv6_destination_mtu_rem_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

adk_ipv6_destination_mtu_rem_null_handle_return:
    ADK_RETURN( retValue1 );
}

/*!
* @param[in] ipv6Handle        IPv6 module handle
* @param[in] timeval         Reassembly timeout in milliseconds
* @param[in]   percentageChangeInTimeout - Percentage change in the timeout
*              value from previous stage as number of occupied contexts increase.
*              Acceptable values are integers between 1 & 100.
* @param[in] minTimeout      Minimum timeout to be used. This will be used for for stage-4.
*               If timeval is going below minTimeout in any of the stages, minTimeout will
*               be used for that stage.
*
* This function configures the global IPv6 reassembly timeout value.
* Timeout value greater than 2000 seconds is not supported.
* The default reassembly timeout value is 100 milliseconds. Minimum reassembly
* timeout value that can be configured is 100 milliseconds.
*
* @retval ADK_E_INVALID_PARAMETER  Invalid IPv6 module handle or invalid defrag timer interval or invalid percentage change.
*
* @retval ADK_E_UNSUPPORTED Timeout value greater than 2000 seconds is not supported.
*
* The total number of reassembly contexts, is divided into four stages.
* Stage 1: supports timeout value timeval : for contexts occupancy of fist 1/3rd of total contexts
* Stage 2: supports timeout value as percentageChangeInTimeout of timeval: for contexts occupancy of next 1/6th of total contexts
* stage 3: supports timeout value as percentageChangeInTimeout of stage 2 timeout value: for contexts occupancy of next 1/6th of total contexts
* stage 4: supports timeout value minTimeVal : for contexts occupancy of remaining contexts
*
*/
adk_status_t adk_ipv6_reassembly_timeout_config (
        adk_module_handle_t ipv6Handle,
        uint32_t timeval,
        uint8_t percentChangeInTimeout,
        uint32_t minTimeout)
{
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv6 module handle %"ADKPRIh"\n", ipv6Handle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_ipv6_reassembly_timeout_config_exit;
    }

    ipv6ModuleInfo_p = (adk_ipv6_info_t *) ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if(timeval > ADK_IPV6_MAX_TIMEOUT)
    {
        ADK_IPV6_LOG_FORCE("Invalid timeout value %d", timeval);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipv6_reassembly_timeout_config_exit;
    }
    if(percentChangeInTimeout > ADK_IPV6_MAX_PERC_CHANGE_TIMEOUT)
    {
        ADK_IPV6_LOG_FORCE("Invalid percentChangeInTimeout value %d", percentChangeInTimeout);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipv6_reassembly_timeout_config_exit;
    }
    if(minTimeout < ADK_IPV6_MIN_TIMEOUT_VALUE)
    {
        ADK_IPV6_LOG_FORCE("Invalid minTimeout value %d", minTimeout);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipv6_reassembly_timeout_config_exit;
    }

    ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmTout = timeval;
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmPrcntToutChange = percentChangeInTimeout;
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout = minTimeout;

    ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage1 = ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmTout;

    if ( ((ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmTout * (100-ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmPrcntToutChange))/100) 
							> (ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout) )
    {
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage2 = ((ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmTout * 
							(100-ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmPrcntToutChange))/100);

	if ( ((ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage2 * (100-ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmPrcntToutChange))/100) 
							> (ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout) )
        	ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage3 = ((ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage2 * 
							(100-ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmPrcntToutChange))/100);
    	else
        	ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage3 = (ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout);


    }
    else
    {
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage3 = (ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout);
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage2 = (ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout);
    }

    
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage4 = (ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmMinTout);

    ADK_IPV6_LOG_INFO(ipv6Handle, "IPv6 reassembly config: Received: Tout=%d, %%Toutchange=%d, minTout=%d, \
            \n Configured: stage1Tout=%d, stage2Tout=%d, stage3Tout=%d,stage4Tout=%d\n", timeval, percentChangeInTimeout,\
            minTimeout, ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage1, ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage2,\
            ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage3, ipv6ModuleInfo_p->adk_ipv6_dp_tables->reasmStage4);

_adk_ipv6_reassembly_timeout_config_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);

    ADK_RETURN(retValue);
}


/*!
* =========================================================================================================
* adk_ipv6_configure_supported_protocol
*
* @param[in]   ipv6Handle    IPv6 module handle.
*
* @param[in] protocol  Packets with this protocol value will be allowed to be
*            processed if they are locally terminating packets.
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_IPV6_E_INVALID_PARAMETER  invalid IPv6 module handle
*
* @retval ADK_E_INTERNAL_ERROR - Failed to configure supported protocols.
*
* @brief
*
* This function configures the allowed protocols for locally
* terminating IPv6 packets.
*
* =========================================================================================================
*/
adk_status_t adk_ipv6_configure_supported_protocol(
    adk_module_handle_t ipv6Handle,
       uint8_t protocol)
{
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    //adk_hos_ipproto_tbl_rule_attr_t  hos_ip_proto_rule;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv6 module handle %"ADKPRIh"\n", ipv6Handle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ipv6_configure_supported_protocol_exit;
    }

    ipv6ModuleInfo_p = (adk_ipv6_info_t *) ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->allowedL4protos[protocol/64] |= ((uint64_t)1 << (protocol % 64));

/* TODO: Removing ip proto HLP offload for now, as it is conflicting with l4-cheksum offload
 * Will enable back, once we have right design in place
 */
#if 0
    hos_ip_proto_rule.proto = protocol;
    retValue = adk_hos_rule_add(ADK_HOS_SUPPORTED_IPPROTO_TBL, 
                &hos_ip_proto_rule,
                &adk_ipv6_supp_proto_ruleids[protocol]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_WARNING( sys_handle, ipv6Handle,
                "Failed to add HOS IP PROTO Table entry \n");
        goto adk_ipv6_configure_supported_protocol_exit;
    }
#endif

adk_ipv6_configure_supported_protocol_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);

    ADK_RETURN(retValue);
}



/*!
* =========================================================================================================
* adk_ipv6_remove_supported_protocol
*
* @param[in] ipv6Handle IPv6 module handle.
*
* @param[in] protocol  Packets with this protocol value will not allowed to be
*              processed if they are locally terminating packets.
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_IPV6_E_INVALID_PARAMETER  invalid IPv6 module handle
*
* @brief
*
* This function removes the protocol from  the allowed protocols list for locally
* terminating IPv6 packets.
*
* =========================================================================================================
*/
adk_status_t
adk_ipv6_remove_supported_protocol (
        adk_module_handle_t ipv6Handle,
        uint8_t protocol)
{
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipv6Handle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ipv6_remove_supported_protocol_exit;
    }

    ipv6ModuleInfo_p = (adk_ipv6_info_t *) ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->allowedL4protos[protocol/64] &= ~((uint64_t)1 << (protocol % 64));
#if 0
    retValue = adk_hos_rule_delete(ADK_HOS_SUPPORTED_IPPROTO_TBL,
            adk_ipv6_supp_proto_ruleids[protocol]);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_WARNING( sys_handle, ipv6Handle,
                "Failed to del HOS IP PROTO Table entry \n");
        goto adk_ipv6_remove_supported_protocol_exit;
    }
#endif

adk_ipv6_remove_supported_protocol_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);

    ADK_RETURN(retValue);
}

/*!
* =============================================================
* adk_ipv6_udp_port_exempt_zero_checksum_add
* @param[in] ipv6Handle IPV6 module handle. Returned at the
* time of IPV6 module's registration with the module manager.
* @param[in] udpPort UDP port that need to be added to
* exempted list to exempt from zero checksum check.
*
* @retval ADK_SUCCESS Successfully added given UDP port to
* exempted list.
* @retval ADK_E_INVALID_PARAMETER The input parameters are NULL.
*
* @brief  Adds an UDP dest port to the exempted list.
*
* This API adds an UDP port to the exempted list.
* For all incoming IPv6 traffic, checksum is mandatory except for
* configured destination port using this API
*
* =============================================================
*/
adk_status_t
adk_ipv6_udp_port_exempt_zero_checksum_add(
                                    adk_module_handle_t ipv6Handle,
                                    uint16_t udpPort)
{
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv6 module handle %"ADKPRIh"\n", ipv6Handle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ipv6_udp_port_exempt_zero_checksum_add_exit;
    }

    ipv6ModuleInfo_p = (adk_ipv6_info_t *) ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->allowedUdpPorts[udpPort/64] |= ((uint64_t)1 << (udpPort % 64));

adk_ipv6_udp_port_exempt_zero_checksum_add_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);

    ADK_RETURN(retValue);  
}


/*!
* =============================================================
* adk_ipv6_udp_port_exempt_zero_checksum_delete
* @param[in] ipv6Handle IPV6 module handle. Returned at the
* time of IPV6 module's registration with the module manager.
* @param[in] udpPort UDP port that need to be removed from
* exempted list.
*
* @retval ADK_SUCCESS Successfully removed the given UDP port from
* exempted list.
* @retval ADK_E_INVALID_PARAMETER The input parameters are NULL.
*                                 or udpPort not found in the configured list
*
* @brief  Removes an UDP port from the exempted list.
*
* This API removes an UDP port from the exempted list.
*
* =============================================================
*/
adk_status_t
adk_ipv6_udp_port_exempt_zero_checksum_delete(
                                adk_module_handle_t ipv6Handle,
                                uint16_t        udpPort)
{
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }


    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv4 module handle %"ADKPRIh"\n", ipv6Handle);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_ipv6_udp_port_exempt_zero_checksum_delete_exit;
    }

    ipv6ModuleInfo_p = (adk_ipv6_info_t *) ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->allowedUdpPorts[udpPort/64] &= ~((uint64_t)1 << (udpPort % 64));

adk_ipv6_udp_port_exempt_zero_checksum_delete_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);
    ADK_RETURN(retValue);
}

/*!
 * end addtogroup _ipv6Functions  IPv6 Module
 * @}
 */
