/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_address_validation.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief  
 *
 * This file contains all the APIs provided by the IPV6 module.
 *
 ***********************************************************************/
/* infrastructure header files */
#include "adk_include.h"
#include "acp_gen_interface.h"


/* IPv6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

/* IPv6 internal header files */
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_internal.h"
#include "adk_qos_include_internal.h"


/*!
 * =============================================================
 * adk_ipv6_address_validate
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] ipv6AddressValInfo_p pointer to 
 * adk_ipv6_address_val_info_t
 * @retval ADK_SUCCESS Successfully created the association.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 *@retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 * @retval ADK_E_MEM_UNAVAILABLE There is not enough memory
 * available.
 *
 * @brief  Adds entries to source ip and destination ip validation 
 *         trees depending on the flag
 *
 * =============================================================
 */
adk_status_t adk_ipv6_address_validate_add(
    adk_module_handle_t ipv6Handle,
    const adk_ipv6_address_val_info_t * ipv6AddressValInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_sys_handle_t sysHandle;
    uint32_t  vrd_id=0;
    /*adk_ipv6_dp_route_entry_t route_entry;*/
    adk_ipv6_dp_ingress_flow_sip_entry_t ingress_flow_sip_entry;
    adk_ipv6_dp_route_entry_t       dpRouteEntry;
       adk_handle_t                    queue_handle;
#ifdef ADK_QOS_GEN3_INTERFACE
    adk_qos_queue_id_info_t qid_info = {.queueId = IPV6_INVALID_QOS_QID};
#endif 
    
	/* Validate all the input arguments */
    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }
    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

       /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_address_validate_failure_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    /* !Validate input structure pointer */
    if (ipv6AddressValInfo_p == NULL) {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_address_validate_failure_return;
    }


     /* Validate  IPv6 VRD handle */
    if (!adk_ipv6_vrd_valid_handle(ipv6AddressValInfo_p->ipv6VrdHandle, &vrd_id))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Invalid  IPv6 Virtual Routing Domain handle \n");
        goto adk_ipv6_address_validate_failure_return ;
    }

    if ( (ipv6AddressValInfo_p->action != ADK_IPV6_ACTION_ALLOW)  &&
         (ipv6AddressValInfo_p->action != ADK_IPV6_ACTION_DROP) &&
         (ipv6AddressValInfo_p->action != ADK_IPV6_ACTION_COUNT_AND_DROP) &&
         (ipv6AddressValInfo_p->action != ADK_IPV6_ACTION_FORWARD_TO_CPU))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle,
                "This Action is not supported for Destination IP address\n"
                "Add route to forward/terminate for this destination address\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_address_validate_failure_return; 
    }

    ADK_MEMSET(&dpRouteEntry , 0, sizeof(adk_ipv6_dp_route_entry_t));
    ADK_MEMSET(&ingress_flow_sip_entry , 0, sizeof(adk_ipv6_dp_ingress_flow_sip_entry_t));

    if (ipv6AddressValInfo_p->ipv6_address_flag == ADK_IPV6_ADDRESS_DEST) 
    {
        /* ROUTE lookup tree used, need to change? */
        if(ipv6AddressValInfo_p->action == ADK_IPV6_ACTION_ALLOW)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Invalid Action Type for IPV6 destination address \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto adk_ipv6_address_validate_failure_return; 
        }

        if(ipv6AddressValInfo_p->action == ADK_IPV6_ACTION_DROP)
        {
            dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_DROP;
        }
        else if(ipv6AddressValInfo_p->action == ADK_IPV6_ACTION_COUNT_AND_DROP)
        {
            dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_DROP_WITH_COUNT;
        }
        else if(ipv6AddressValInfo_p->action == ADK_IPV6_ACTION_FORWARD_TO_CPU)
        {
            if(ipv6AddressValInfo_p->cpuPath == ADK_IPV6_PROCESSING_PATH_CPU_PH)
                dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_STH_VIA_PH;
            else if(ipv6AddressValInfo_p->cpuPath == 
#ifdef ADK_QOS_GEN3_INTERFACE
                                          ADK_IPV6_PROCESSING_PATH_CPU_SP_STACK)
#else
                                          ADK_IPV6_PROCESSING_PATH_CPU_NOPH)
#endif
                dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_STH_VIA_NCA;
            else
            {
                ADK_LOG_ERROR(sysHandle,
                        ipv6Handle,
                        "Invalid cpu path  \n");
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                goto adk_ipv6_address_validate_failure_return;
            }
        
#ifndef ADK_QOS_GEN3_INTERFACE
            queue_handle = ipv6AddressValInfo_p->mtmQueueHandle;
#else
            queue_handle = ipv6AddressValInfo_p->rl_queue_handle;
#endif
#ifdef ADK_QOS_GEN3_INTERFACE
            if (queue_handle) {
                retValue = adk_qos_queue_info_get(sysHandle,
                                           queue_handle, &qid_info);
            }
            if (retValue != ADK_SUCCESS)
            {
                ADK_LOG_ERROR(sysHandle, ipv6Handle, 
                    "Failed to get Queue ID from the Queue handle %"ADKPRIh "\n",
                    queue_handle);
                goto adk_ipv6_address_validate_failure_return;
            }
            dpRouteEntry.devId = qid_info.devId;
            dpRouteEntry.queueId = qid_info.queueId;
            dpRouteEntry.statsId = qid_info.queueStatsId;
            dpRouteEntry.policerId = qid_info.policerId;
#else
        dpRouteEntry.devId           = ipv6AddressValInfo_p->cpuPath_u.phAttr.devId;
        dpRouteEntry.queueId         = ipv6AddressValInfo_p->cpuPath_u.phAttr.queueId;
#endif
        }

        dpRouteEntry.ph_bind_handle  = ipv6AddressValInfo_p->cpuPath_u.phAttr.phBindHandle;
        dpRouteEntry.callback_fun_id = ipv6AddressValInfo_p->cpuPath_u.phAttr.cbFnId;

        /*! Add this route to the route_lookup_tree. */
        retValue = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                vrd_id, 
                ipv6AddressValInfo_p->ipv6AddrPrefix, 
                &dpRouteEntry, FALSE);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle,
                    ipv6Handle,
                    "Failed to add route entry \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR); 
            goto adk_ipv6_address_validate_failure_return;
        }
    } 
    /* Source address validation */
    else 
    {
        ingress_flow_sip_entry.apply_to_fwd_traffic = ipv6AddressValInfo_p->applyOnlyToForwarded; 
        ingress_flow_sip_entry.action               = ipv6AddressValInfo_p->action;


        if(ipv6AddressValInfo_p->action == ADK_IPV6_ACTION_FORWARD_TO_CPU)
        {
            if (ipv6AddressValInfo_p->cpuPath == ADK_IPV6_PROCESSING_PATH_CPU_PH) 
            {
                ingress_flow_sip_entry.cpuPath = ADK_IPV6_PROCESSING_PATH_CPU_PH;
                ingress_flow_sip_entry.ph_bind_handle = 
                    (ipv6AddressValInfo_p->cpuPath_u.phAttr.phBindHandle);
                ingress_flow_sip_entry.cbFnId         = ipv6AddressValInfo_p->cpuPath_u.phAttr.cbFnId;
                ingress_flow_sip_entry.devId          = ipv6AddressValInfo_p->cpuPath_u.phAttr.devId;
                ingress_flow_sip_entry.queueId        = ipv6AddressValInfo_p->cpuPath_u.phAttr.queueId;
            }
            else 
            {
#ifdef ADK_QOS_GEN3_INTERFACE
                ingress_flow_sip_entry.cpuPath = ADK_IPV6_PROCESSING_PATH_CPU_SP_STACK;
#else
                ingress_flow_sip_entry.cpuPath = ADK_IPV6_PROCESSING_PATH_CPU_NOPH;
#endif
            }
#ifndef ADK_QOS_GEN3_INTERFACE
            queue_handle = ipv6AddressValInfo_p->mtmQueueHandle;
#else
            queue_handle = ipv6AddressValInfo_p->rl_queue_handle;
#endif
#ifdef ADK_QOS_GEN3_INTERFACE
            if (queue_handle) {
                retValue = adk_qos_queue_info_get(sysHandle,
                                           queue_handle, &qid_info);
            }
            if (retValue != ADK_SUCCESS)
            {
                ADK_LOG_ERROR(sysHandle, ipv6Handle, 
                    "Failed to get Queue ID from the Queue handle %"ADKPRIh "\n",
                    queue_handle);
                goto adk_ipv6_address_validate_failure_return;
            }
            ingress_flow_sip_entry.devId = qid_info.devId;
            ingress_flow_sip_entry.queueId = qid_info.queueId;
            ingress_flow_sip_entry.statsId = qid_info.queueStatsId;
            ingress_flow_sip_entry.policerId = qid_info.policerId;
#endif
        }


        retValue = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                vrd_id,
                ipv6AddressValInfo_p->ipv6AddrPrefix,
                &ingress_flow_sip_entry);

        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to add the SIP entry %d\n",
                    (retValue));
            retValue =
                ADK_ERROR_SEV_LEVEL_SET
                (ADK_IPV6_E_SIP_STATE_MEMORY_UNAVAILABLE);
            goto adk_ipv6_address_validate_failure_return;
        }
    }

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
    ADK_RETURN(retValue);

adk_ipv6_address_validate_failure_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ipv6_address_validate_remove
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] ipv6AddressValInfo_p pointer to 
 * adk_ipv6_address_val_info_t
 * @retval ADK_SUCCESS Successfully created the association.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 *@retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 * @retval ADK_E_MEM_UNAVAILABLE There is not enough memory
 * available.
 *
 * @brief  Removes entries to source ip and destination ip validation 
 *         trees depending on the flag
 *
 * =============================================================
 */

adk_status_t adk_ipv6_address_validate_remove(
    adk_module_handle_t ipv6Handle,
    const adk_ipv6_address_val_info_t *ipv6AddressValInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_sys_handle_t sysHandle;
    uint32_t vrd_id=0;
    
	/* Validate all the input arguments */
    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }
    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
     }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

       /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_address_validate_remove_failure;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    /* !Validate input structure pointer */
    if (ipv6AddressValInfo_p == NULL) 
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);

        goto adk_ipv6_address_validate_remove_failure;
    }


    /* Validate  IPv6 VRD handle */
    if (!adk_ipv6_vrd_valid_handle(ipv6AddressValInfo_p->ipv6VrdHandle, &vrd_id))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Invalid  IPv6 Virtual Routing Domain handle \n");
        goto adk_ipv6_address_validate_remove_failure ;
    }

    if(ipv6AddressValInfo_p->ipv6_address_flag == ADK_IPV6_ADDRESS_DEST)
    {
        /* ROUTE lookup tree used, need to change? */
        if(ipv6AddressValInfo_p->action == ADK_IPV6_ACTION_ALLOW)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Invalid Action Type for IPV6 destination address \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto adk_ipv6_address_validate_remove_failure; 
        }

        /*! Add this route to the route_lookup_tree. */
        retValue = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                vrd_id, 
                ipv6AddressValInfo_p->ipv6AddrPrefix, FALSE);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle,
                    ipv6Handle,
                    "Failed to rem route entry \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR); 
            goto adk_ipv6_address_validate_remove_failure;
        }
    }
    else
    {
        retValue = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                    vrd_id,
                    ipv6AddressValInfo_p->ipv6AddrPrefix);

            if(retValue != ADK_SUCCESS)
            {
                ADK_LOG_ERROR(sysHandle, ipv6Handle,
                        "Failed to rem the SIP entry %d\n",
                        (retValue));
                retValue =
                    ADK_ERROR_SEV_LEVEL_SET
                    (ADK_IPV6_E_SIP_STATE_MEMORY_UNAVAILABLE);
                goto adk_ipv6_address_validate_remove_failure;
            }
    }

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_RETURN(retValue);
adk_ipv6_address_validate_remove_failure:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}
 /*!
 * end addtogroup _ipv6Functions  IPv6 Functions
 * @}
 */
