/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_interface.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief
 *
 * This file contains APIs to manage IPv6 interfaces. This includes creation,
 * deletion of interfaces, setting their attributes, getting or printing
 * statistics for the interfaces, etc.
 *
 ***********************************************************************/

/* Infrastructure header files */
#include "adk_include.h"
#include "adk_hos.h"
#include "adk_ph_dev.h"
#include "adk_utils_tree_update.h"
#include "adk_ethernet_export.h"

/* IPV6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

#define ADK_CONVERT_TO_V6_IF_HANDLE(__DATA) ((adk_interface_handle_t)__DATA)
#define ADK_CONVERT_TO_V6_IF_HANDLE(__DATA) ((adk_interface_handle_t)__DATA)
/* IPV6 internal header files */
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_internal.h"

/*
bool
adk_ipv6_compare_address(adk_ipv6_addr_prefix_t address1,
                         adk_ipv6_addr_prefix_t address2);

bool
adk_ipv6_address_validate(adk_module_handle_t ipv6Handle, adk_ipv6_addr_prefix_t address);
*/

adk_status_t
adk_ipv6_link_local_addr_db_update( adk_module_handle_t ipv6Handle, 
                                    uint32_t vrd_id, 
                                    adk_ipv6_addr_t Ipv6Addr, 
                                    bool add,
                                    bool *action);

static adk_status_t adk_ipv6_create_egress_qos_resources(
       adk_module_handle_t ipv6Handle,
       uint32_t ifId);

static adk_status_t adk_ipv6_delete_egress_qos_resources(
       adk_module_handle_t ipv6Handle,
       uint32_t ifId);
/**
 *=============================================================
 * adk_ipv6_interface_create
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in]  ipv6If_p  - pointer to adk_ipv6_interface_t data
 * structure.
 *
 * @param[out] ipv6IfHandle - pointer to the IPv6 interface handle
 * which gets created by this API.
 *
 * @retval ADK_SUCCESS Successfully created ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @retval ADK_E_MEM_UNAVAILABLE - failed to allocate memory
 * for interface handle.
 *
 * @retval ADK_E_LEARN_FAILED - failed to update MPP tree
 * entry.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This function create an IP interface handle to be used across all the IPv6
 * APIs (route updates, associations, filters, etc). The basis for an IPv6
 * interface is some lower layer interface which can be an L2 physical
 * interface or any logical interface.
 * =============================================================
 */
adk_status_t
adk_ipv6_interface_create(
    adk_module_handle_t ipv6Handle,
    const adk_ipv6_interface_t  *ipv6If_p,
    adk_interface_handle_t *ipv6IfHandle)
{
    adk_status_t 	 	 retValue = ADK_SUCCESS;
    adk_dp_status_t              dpRetVal = ADK_SUCCESS;          
    uint32_t 		         adk_if_id;
    ipv6_interface_info_t 	 *if_pvt_info = NULL;
    adk_sys_handle_t 		 sys_handle;
    adk_ipv6_info_t              *ipv6ModuleInfo_p = NULL;
    uint8_t                      dscp, numAddrIndex;
    bool                         linkLocalAction = TRUE;
    adk_ipv6_dp_route_entry_t    dpRouteEntry;
    adk_ipv6_addr_prefix_t       ipv6_addr_info;
    uint32_t ipv6_addr_stats_id;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);
    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,"Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_interface_create_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module info is NULL\n");
        goto adk_ipv6_interface_create_return;
    }
    if(ipv6ModuleInfo_p->totalIpv6Address >= ADK_IPV6_INTF_ADDR_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                       "Maximum (%d) IPv6 addresses are already added new interface "
                       "cannot be created\n", ADK_IPV6_INTF_ADDR_MAX);
        goto adk_ipv6_interface_create_return;
    }
    if (ipv6ModuleInfo_p->maxIpv6Interfaces >= ADK_IPV6_INTERFACES_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                       "Reached Max IP Interface Limit %d\n", ADK_IPV6_INTERFACES_MAX);
        goto adk_ipv6_interface_create_return;
    }

    if (!ipv6If_p || !ipv6IfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "NULL pointer parameter specified\n");
        goto adk_ipv6_interface_create_return;
    }

    *ipv6IfHandle = 0;


    /* Validate the L2 interface handle */
    if (!ADK_SYS_VALID_INTERFACE_HANDLE(
                    ipv6If_p->layer2InterfaceHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IPV6_E_INVALID_L2_INTERFACE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "Invalid Layer-2 interface handle \n");
        goto adk_ipv6_interface_create_return;
    }

    /* Validate MTU input */
    if ((ipv6If_p->mtu) && ((ipv6If_p->mtu < ADK_IPV6_INTERFACE_MIN_MTU) 
        || (ipv6If_p->mtu >= ADK_IPV6_INTERFACE_MAX_MTU)))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                      "MTU must be between 1280 to 32767 \n");
        goto adk_ipv6_interface_create_return;
    }

    ipv6_addr_stats_id = adk_rm_resource_reserve(ipv6ModuleInfo_p->ipv6AddrStatsHandle);
    if(ipv6_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to allocate IP Address Stats ID");
        goto adk_ipv6_interface_create_return;
    }

    if( (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[0] == 0) &&
        (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[1] == 0) &&
        (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[2] == 0) &&
        (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[3] == 1))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                      " Invalid IPv6 interface address (Loopback address)\n");
       /* Since IPv6 address is an optional Parameter, we are just
	* logging an Error , but not returning. Hence user/appln
	* has to ensure to pass valid IPv6 address*/	
    }
    
    ADK_SYS_INTERFACE_HANDLE_ALLOC(ipv6Handle, ADK_IPV6_INTF_OBJ,
                                   sizeof(ipv6_interface_info_t),
                                   ipv6IfHandle);
    if (*ipv6IfHandle == 0) {
        retValue = ADK_E_MEM_UNAVAILABLE;

        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "Memory allocation failed\n");
        goto adk_ipv6_interface_create_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(*ipv6IfHandle);

    ADK_MEMSET( if_pvt_info, 0, sizeof(ipv6_interface_info_t)); 
    ADK_MEMCPY(&(if_pvt_info->if_info_u.regularIf), ipv6If_p, sizeof(adk_ipv6_interface_t));

    if_pvt_info->cookie = ADK_IPV6_INTF_COOKIE;
    if_pvt_info->ifType = IPV6_INTERFACE_TYPE_ETH;
    if_pvt_info->ipv6_addr_stats_id = ipv6_addr_stats_id; 

    if ( ipv6If_p->mtu == 0)
    {
        if_pvt_info->if_info_u.regularIf.mtu = ADK_IPV6_INTERFACE_DEFAULT_MTU; 
    }

    /*IPv6 interface over Ethernet*/
    adk_if_id = ADK_SYS_INTERFACE_ID_GET(
                    ipv6If_p->layer2InterfaceHandle);
    ADK_SYS_INTERFACE_ID_SET(*ipv6IfHandle, adk_if_id);

    dpRetVal = adk_ipv6_create_egress_qos_resources(ipv6Handle, adk_if_id);

    if(dpRetVal != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                           "Error in Creating IPv6 QOS resources in dataplane\n");
        goto adk_ipv6_interface_create_return;
    }
    /* allocate statistics id check if it is common resource acrosss L2 & IPvX Module */
    if_pvt_info->stats_id = adk_if_id;

    /* adk_ipv6_dp_tables will be updated in ipv6_dp_init(), that is in DP process, below copying DP pointer to CP, so that Application can use it */
    adk_ipv6_dp_tables = ipv6ModuleInfo_p->adk_ipv6_dp_tables;
    /* Copying interface Handle to dp structure, so that ifHandle can be used while raising exceptions */
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6IfaceHandles[adk_if_id] = *ipv6IfHandle;

    /*using same stats id for ingress & egress 
     * (ingress & egress stats are seperate DP data tables */
    if_pvt_info->EgrStatsId = if_pvt_info->stats_id;
    /* Interface attributes */
    if_pvt_info->interfaceAttr = ipv6If_p->mask ;
    /* update the egress sys stats id */
    retValue = adk_sys_egress_stats_id_alloc(sys_handle, &(if_pvt_info->SysEgrStatsId)); 
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to reserve SYS egress stats id \n");
        goto adk_ipv6_interface_stats_alloc_failed;
    }
    /* Store the default stats id for all dscp values  */
    for (dscp = 0; dscp < IPV6_NUM_DSCP_VALUES; dscp++)
    {
        if_pvt_info->dscp_pcp[dscp] = IPV6_DEFAULT_PCP;
        if_pvt_info->dscp_stats_id[dscp] = if_pvt_info->stats_id;
    }

    if(!( (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[0] == 0) &&
          (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[1] == 0) &&
          (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[2] == 0) &&
          (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[3] == 0) &&
          (ipv6If_p->ipv6AddrPrefix.prefixLength == 0)))
    {
        numAddrIndex = if_pvt_info->numAddr;
        ADK_MEMCPY( &(if_pvt_info->ipv6_prefix[numAddrIndex]), &(ipv6If_p->ipv6AddrPrefix), 
                sizeof(adk_ipv6_addr_prefix_t));
        if_pvt_info->numAddr = 1;
        ipv6ModuleInfo_p->totalIpv6Address++;
        /* Configure Route Lookup Tree for Ingress Assocation */
        if(linkLocalAction)
        {
            ADK_MEMSET(&dpRouteEntry, 0, sizeof(adk_ipv6_dp_route_entry_t));

            ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[0]); 
            ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[1]); 
            ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[2]); 
            ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[3]); 
            ipv6_addr_info.prefixLength = ADK_IPV6_PDU_ADDR_LENGTH; 
            dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_TERMINATE;
            dpRouteEntry.ipv6_addr_stats_id = ipv6_addr_stats_id;
            dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info,
                    &dpRouteEntry, FALSE);
            if(dpRetVal != ADK_SUCCESS)
            {
                ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                        "Failed to add route entry for IP termination\n");
                goto adk_ipv6_interface_route_add_failed_return;
            }
            /*Implicit Onlink Route addition in ADK is not supported. 
             * Application has to configure route explicitly*/
        }
    }

    /* Configure the IPv6 interface table */
    dpRetVal = adk_ipv6_dp_interface_entry_add(ipv6Handle, if_pvt_info, 0 /* TBD default DSCP */ );
    if(dpRetVal != ADK_SUCCESS)
    {
            ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                           "Failed to Update Input Interface Table \n");
            goto adk_ipv6_in_interface_create_return; 
    }
    ipv6ModuleInfo_p->maxIpv6Interfaces++;

    dpRetVal = adk_ipv6_pbr_config(ipv6Handle, ipv6IfHandle, ipv6ModuleInfo_p->adk_ipv6_dp_tables->adk_ipv6_pbr_policy_set_handle_t);
    if(dpRetVal != ADK_SUCCESS)
    {
	    ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
			    "Failed to Update PBR config\n");
	    goto adk_ipv6_in_interface_create_return;
    }

#ifdef ADK_DEVICE_SNR
    adk_module_handle_t eth_mod_handle;
    retValue = adk_module_handle_get(sys_handle,
            ADK_ETHERNET_MODULE_ID,
            NULL,
            &eth_mod_handle);
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to get ethernet module handle \n");
        goto adk_ipv6_in_interface_create_return;
    }
#endif
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
    ADK_RETURN(retValue);

adk_ipv6_in_interface_create_return:
    if ( ipv6If_p->ipv6AddrPrefix.Ipv6Addr.addr[0] == 0xFE800000 )
    {
        if (adk_ipv6_link_local_addr_db_update( ipv6Handle, ADK_IPV6_LOCAL_ROUTING_TABLE, 
                                            ipv6If_p->ipv6AddrPrefix.Ipv6Addr,
                                            FALSE,
                                            &linkLocalAction) !=  ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                           "Failed to remove Link local IP address\n");
        }

        if(linkLocalAction)
        {
            ipv6_addr_info.Ipv6Addr.addr[0] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[0]); 
            ipv6_addr_info.Ipv6Addr.addr[1] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[1]); 
            ipv6_addr_info.Ipv6Addr.addr[2] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[2]); 
            ipv6_addr_info.Ipv6Addr.addr[3] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[3]); 
            ipv6_addr_info.prefixLength = ADK_IPV6_PDU_ADDR_LENGTH; 

            dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                    ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info, FALSE);
            if(dpRetVal != ADK_SUCCESS)
            {
                ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                        "Failed to remove route entry during interface delete \n");
            }
        }

    }
adk_ipv6_interface_route_add_failed_return:
    retValue = adk_sys_egress_stats_id_free(sys_handle, (if_pvt_info->SysEgrStatsId));
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to free SYS egress stats id \n");
    }

adk_ipv6_interface_stats_alloc_failed:
    if (ipv6IfHandle != NULL)
        if (*ipv6IfHandle != 0)
             ADK_SYS_INTERFACE_HANDLE_FREE(*ipv6IfHandle);

adk_ipv6_interface_create_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
    ADK_RETURN(retValue);

}    /* -----  end of function adk_ipv6_interface_create----- */

/**
 *=============================================================
 * adk_ipv6_interface_delete
 * @param[in]  ipv6Handle - IP module handle.
 *
 * @param[out] ipv6IfHandle - IP interface handle to be deleted
 *
 * @retval ADK_SUCCESS Successfully deleted ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle.
 *
 * @retval ADK_E_UNLEARN_FAILED - failed to update MPP tree
 * entry.
 *
 * @brief
 *
 * This function destroys an IPv6 interface handle.
 * =============================================================
 */
adk_status_t
adk_ipv6_interface_delete(
    adk_module_handle_t ipv6Handle,
    adk_interface_handle_t ipv6IfHandle)
{
    adk_status_t 		retValue = ADK_SUCCESS;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_sys_handle_t 		sys_handle;
    uint32_t                i=0, j=0;
    adk_dp_status_t         dpRetVal;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;
    uint16_t                if_id; 
    bool                    linkLocalAction = TRUE;
    adk_ipv6_addr_prefix_t  ipv6_addr_info;
    uint16_t ipv6_addr_stats_id;
    adk_ipv6_dp_ip_addr_stats_t *dp_stats;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_interface_delete_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module info is NULL\n");
        goto adk_ipv6_interface_delete_return;
    }

    if ( !ipv6IfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "NULL pointer parameter specified\n");
        goto adk_ipv6_interface_delete_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "Invalid Interface Handle type\n");
        goto adk_ipv6_interface_delete_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( (!if_pvt_info) || (if_pvt_info->ifType != IPV6_INTERFACE_TYPE_ETH))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "Invalid Interface Handle type\n");
        goto adk_ipv6_interface_delete_return;
    }

    /* Remove all DSCP Groups attached to the interface */
    for(i = 0; i < ADK_IPV6_MAX_DSCP_GROUP_PER_IF; i++)
    {
       if ( if_pvt_info->ifDscpGroupList[i].isValid == TRUE)
       {
           retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
           ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                              "All DSCP Groups attached to the interface are not removed\n");
           goto adk_ipv6_interface_delete_return;
       }
    }
    
    if (if_pvt_info->numAddr > 1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "More than one (%d) IPv6 addresses are associated "
                           "to the interface\n", if_pvt_info->numAddr);
        goto adk_ipv6_interface_delete_return;
    }
 
    if_id = ADK_SYS_INTERFACE_ID_GET(ipv6IfHandle);

    dpRetVal = adk_ipv6_delete_egress_qos_resources(ipv6Handle, if_id);

    if(dpRetVal != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                           "Error in Deleting IPv6 QOS resources in dataplane\n");
        goto adk_ipv6_interface_delete_return;
    }

    /* Configure the IPv6 interface table */
    dpRetVal = adk_ipv6_dp_interface_entry_rem(ipv6Handle, 0 /* TBD default DSCP */, if_id );
    if(dpRetVal != ADK_SUCCESS)
    {
            ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                           "Failed to Update Input Interface Table \n");
            goto adk_ipv6_interface_delete_return; 
    }

    if ( if_pvt_info->numAddr == 1)
    {
        if(linkLocalAction)
        {
            ipv6_addr_info.Ipv6Addr.addr[0] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[0]); 
            ipv6_addr_info.Ipv6Addr.addr[1] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[1]); 
            ipv6_addr_info.Ipv6Addr.addr[2] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[2]); 
            ipv6_addr_info.Ipv6Addr.addr[3] = (if_pvt_info->ipv6_prefix[0].Ipv6Addr.addr[3]); 
            ipv6_addr_info.prefixLength = ADK_IPV6_PDU_ADDR_LENGTH; 

            dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                    ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info, FALSE);
            if(dpRetVal != ADK_SUCCESS)
            {
                ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                        "Failed to remove route entry during interface delete \n");
                goto adk_ipv6_interface_delete_return;
            }
            ipv6ModuleInfo_p->totalIpv6Address--;
            adk_sys_flow_cache_update_version(sys_handle);
        }
    }

    /*! Release ipv6 address stats */
    ipv6_addr_stats_id = if_pvt_info->ipv6_addr_stats_id;
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        dp_stats = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6AddrStatsTble[i][ipv6_addr_stats_id]);
        ADK_MEMSET(dp_stats, 0, sizeof(adk_ipv6_dp_ip_addr_stats_t));
    }
    adk_rm_resource_free(ipv6ModuleInfo_p->ipv6AddrStatsHandle, if_pvt_info->ipv6_addr_stats_id);
    if_pvt_info->ipv6_addr_stats_id = 0;

    for(j=0;j<IPV6_SEC_ADDR_MAX;j++)
    {
        if(if_pvt_info->ipv6_prefix_stats_id[j])
        {
            ipv6_addr_stats_id = if_pvt_info->ipv6_prefix_stats_id[j];
            for(i=0; i<ADK_MAX_NUM_CORES; i++)
            {
                dp_stats = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6AddrStatsTble[i][ipv6_addr_stats_id]);
                ADK_MEMSET(dp_stats, 0, sizeof(adk_ipv6_dp_ip_addr_stats_t));
            }
            adk_rm_resource_free(ipv6ModuleInfo_p->ipv6AddrStatsHandle, if_pvt_info->ipv6_prefix_stats_id[j]);
            if_pvt_info->ipv6_prefix_stats_id[j] = 0;
        }
    }
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6IfaceHandles[ADK_SYS_INTERFACE_ID_GET(ipv6IfHandle)] = 0;
    ipv6ModuleInfo_p->maxIpv6Interfaces--;

    /* release the sys egress stats id */
    retValue = adk_sys_egress_stats_id_free(sys_handle, (if_pvt_info->SysEgrStatsId));
    if(retValue != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to free SYS egress stats id \n");
    }
    ADK_SYS_INTERFACE_HANDLE_FREE(ipv6IfHandle);

adk_ipv6_interface_delete_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/********** VIRTUAL ********************************/
adk_status_t
adk_ipv6_virtual_ip_addr_add (
    adk_module_handle_t ipv6Handle,
    adk_ipv6_addr_t *ipv6Addr)
{
    adk_status_t                    retValue = ADK_SUCCESS;
    adk_dp_status_t         dpRetVal = ADK_SUCCESS;
    adk_sys_handle_t        sys_handle;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;
    adk_ipv6_dp_route_entry_t   dpRouteEntry;
    adk_ipv6_addr_prefix_t  ipv6_addr_info;
    uint32_t ipv6_addr_stats_id;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_interface_addr_add_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module info is NULL\n");
        goto adk_ipv6_interface_addr_add_return;
    }
    if (ipv6ModuleInfo_p->totalIpv6Address >= ADK_IPV6_INTF_ADDR_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                       "Reached Max IP address Limit %d\n", ADK_IPV6_INTF_ADDR_MAX);
        goto adk_ipv6_interface_addr_add_return;
    }

    if (ipv6ModuleInfo_p->numAddrIndex > (ADK_IPV6_VIRTUAL_ADDR_MAX-1))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                       "Reached Virtual IP address Max Limit \n");
        goto adk_ipv6_interface_addr_add_return;
    }

    ipv6_addr_stats_id = adk_rm_resource_reserve(ipv6ModuleInfo_p->ipv6AddrStatsHandle);
    if(ipv6_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to allocate IP Address Stats ID");
        goto adk_ipv6_interface_addr_add_return;
    }
    ipv6ModuleInfo_p->ipv6_virtual_addr_stats_id[ipv6ModuleInfo_p->numAddrIndex] = ipv6_addr_stats_id;
    ADK_MEMCPY(&(ipv6ModuleInfo_p->ipv6_virtual_addr_prefix[ipv6ModuleInfo_p->numAddrIndex]), ipv6Addr, sizeof(adk_ipv6_addr_prefix_t));

    ADK_MEMSET(&dpRouteEntry, 0, sizeof(adk_ipv6_dp_route_entry_t));
    dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_TERMINATE;
    dpRouteEntry.ipv6_addr_stats_id = ipv6_addr_stats_id;

    ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6Addr->addr[0]);
    ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6Addr->addr[1]);
    ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6Addr->addr[2]);
    ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6Addr->addr[3]);
    ipv6_addr_info.prefixLength     = ADK_IPV6_PDU_ADDR_LENGTH;

    dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info,
                        &dpRouteEntry, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {    
        ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                           "Failed to add route entry for IP termination\n");
        goto adk_ipv6_interface_addr_add_learn_failed;
    } 
    adk_sys_flow_cache_update_version(sys_handle);
    ipv6ModuleInfo_p->numAddrIndex++;
    ipv6ModuleInfo_p->totalIpv6Address++;

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
    ADK_RETURN(retValue);

adk_ipv6_interface_addr_add_learn_failed:
adk_ipv6_interface_addr_add_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);


_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
    ADK_RETURN(retValue);
}


adk_status_t 
adk_ipv6_virtual_ip_addr_del (adk_module_handle_t ipv6Handle,
					adk_ipv6_addr_t *ipv6Addr_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_sys_handle_t            sys_handle;
    adk_dp_status_t         dpRetVal;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;
    bool                    linkLocalAction = TRUE;
    adk_ipv6_addr_prefix_t  ipv6_addr_info;
    adk_ipv6_dp_ip_addr_stats_t *dp_stats;
    uint32_t ipv6_addr_stats_id, i;

    if (!ipv6Handle)
    {   
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {   
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_virtual_ip_addr_del_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module info is NULL\n");
        goto adk_ipv6_virtual_ip_addr_del_return;
    }

    ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6Addr_p->addr[0]);
    ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6Addr_p->addr[1]);
    ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6Addr_p->addr[2]);
    ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6Addr_p->addr[3]);

    for (i =0; i < ipv6ModuleInfo_p->numAddrIndex; i++)
    {
        if (adk_ipv6_compare_address(ipv6ModuleInfo_p->ipv6_virtual_addr_prefix[i], ipv6_addr_info))
        {
            break;
        }
    }
    if ( i >= ipv6ModuleInfo_p->numAddrIndex)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                       "This IPv6 address is not already added to the interface\n");
        goto adk_ipv6_virtual_ip_addr_del_return;
    }

    if(linkLocalAction)
    {
        ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6Addr_p->addr[0]);
        ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6Addr_p->addr[1]);
        ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6Addr_p->addr[2]);
        ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6Addr_p->addr[3]);
        ipv6_addr_info.prefixLength = ADK_IPV6_PDU_ADDR_LENGTH;

        dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle,
                ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info, FALSE);
        if(dpRetVal != ADK_SUCCESS)
        {
            ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                    "Failed to remove route entry during interface delete \n");
            goto adk_ipv6_virtual_ip_addr_del_return;
        }  
        adk_sys_flow_cache_update_version(sys_handle);
    }

    ipv6_addr_stats_id = ipv6ModuleInfo_p->ipv6_virtual_addr_stats_id[i];
    for(uint32_t j=0;j<ADK_MAX_NUM_CORES;j++)
    {
        dp_stats = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6AddrStatsTble[j][ipv6_addr_stats_id]);
        ADK_MEMSET(dp_stats, 0, sizeof(adk_ipv6_dp_ip_addr_stats_t));
    }
    adk_rm_resource_free(ipv6ModuleInfo_p->ipv6AddrStatsHandle, ipv6ModuleInfo_p->ipv6_virtual_addr_stats_id[i]);
    ipv6ModuleInfo_p->ipv6_virtual_addr_stats_id[i] = 0;

    ADK_MEMSET(&(ipv6ModuleInfo_p->ipv6_virtual_addr_prefix[i]), 0, sizeof(adk_ipv6_addr_prefix_t));
    for (i=i; i< ipv6ModuleInfo_p->numAddrIndex; i++)
    {
        ADK_MEMCPY(&(ipv6ModuleInfo_p->ipv6_virtual_addr_prefix[i]), &(ipv6ModuleInfo_p->ipv6_virtual_addr_prefix[i+1]), sizeof(adk_ipv6_addr_prefix_t));
    }
    ipv6ModuleInfo_p->numAddrIndex--;
    ipv6ModuleInfo_p->totalIpv6Address--;
adk_ipv6_virtual_ip_addr_del_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/********************************************VIRTUAL *************************/

/**
 *=============================================================
 * adk_ipv6_interface_addr_add
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6IfHandle - pointer to the IPv6 interface handle
 *
 * @param[in] ipv6Addr - pointer to the IPv6 interface address
 *
 * @retval ADK_SUCCESS Successfully updated Ingress association
 *         rule mask
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @brief
 *
 * This function adds IPv6 address to an interface. It configues
 * Route for this IPv6 address for termination.This API is not applicable 
 * for Tunnel and virtual interface
 * =============================================================
 */
adk_status_t adk_ipv6_interface_addr_add(
    adk_module_handle_t  ipv6Handle,
    adk_interface_handle_t ipv6IfHandle,
    const adk_ipv6_addr_prefix_t *ipv6AddrPrefix)
{
    adk_status_t 		    retValue = ADK_SUCCESS;
    adk_dp_status_t         dpRetVal = ADK_SUCCESS;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_sys_handle_t 		sys_handle;
    uint32_t                i;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;
    bool                    linkLocalAction = TRUE;
    adk_ipv6_dp_route_entry_t   dpRouteEntry;
    adk_ipv6_addr_prefix_t  ipv6_addr_info;
    uint32_t ipv6_addr_stats_id;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_interface_addr_add_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module info is NULL\n");
        goto adk_ipv6_interface_addr_add_return;
    }

    if ( !ipv6IfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "NULL pointer parameter specified\n");
        goto adk_ipv6_interface_addr_add_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "Invalid Interface Handle type\n");
        goto adk_ipv6_interface_addr_add_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_TUNNEL) ||
         (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "This API is not supported for Tunnel and Vitual Interface\n");
        goto adk_ipv6_interface_addr_add_return;
    }

    if(ipv6ModuleInfo_p->totalIpv6Address >= ADK_IPV6_INTF_ADDR_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                       "Maximum (%d) IPv6 addresses are already added\n",
                       ADK_IPV6_INTF_ADDR_MAX);
        goto adk_ipv6_interface_addr_add_return;
    }
    else if ( if_pvt_info->numAddr >= IPV6_SEC_ADDR_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER); 
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                       "Maximum (%d) IPv6 addresses are already added to interface\n", 
                       ADK_IPV6_INTF_ADDR_MAX);
        goto adk_ipv6_interface_addr_add_return;
    }

    ipv6_addr_stats_id = adk_rm_resource_reserve(ipv6ModuleInfo_p->ipv6AddrStatsHandle);
    if(ipv6_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to allocate IP Address Stats ID");
        goto adk_ipv6_interface_addr_add_return;
    }

    if (!adk_ipv6_address_validate(ipv6Handle, *ipv6AddrPrefix))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER); 
        goto adk_ipv6_interface_addr_add_return;
    }
  
    for ( i = 0; i < if_pvt_info->numAddr; i++)
    {
        if (adk_ipv6_compare_address(if_pvt_info->ipv6_prefix[i], 
                                     *ipv6AddrPrefix))
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER); 
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "This IPv6 address is already added to interface\n");
            goto adk_ipv6_interface_addr_add_return;
        }
    }

    ADK_MEMCPY( &(if_pvt_info->ipv6_prefix[if_pvt_info->numAddr]), ipv6AddrPrefix, 
            sizeof(adk_ipv6_addr_prefix_t));
    if_pvt_info->ipv6_prefix_stats_id[if_pvt_info->numAddr] = ipv6_addr_stats_id;
    if_pvt_info->numAddr++;
    ipv6ModuleInfo_p->totalIpv6Address++;

    /* Configure Route Lookup Tree for Ingress Assocation */
    if(linkLocalAction)
    {
        ADK_MEMSET(&dpRouteEntry, 0, sizeof(adk_ipv6_dp_route_entry_t));

        ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6AddrPrefix->Ipv6Addr.addr[0]);
        ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6AddrPrefix->Ipv6Addr.addr[1]);
        ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6AddrPrefix->Ipv6Addr.addr[2]);
        ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6AddrPrefix->Ipv6Addr.addr[3]);
        ipv6_addr_info.prefixLength     = ADK_IPV6_PDU_ADDR_LENGTH;
        dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_TERMINATE;
        dpRouteEntry.ipv6_addr_stats_id = ipv6_addr_stats_id;

        dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info,
                &dpRouteEntry, FALSE);
        if(dpRetVal != ADK_SUCCESS)
        {
            ADK_IPV6_LOG_ERROR(ipv6Handle, dpRetVal,
                    "Failed to add route entry for IP termination\n");
            goto adk_ipv6_interface_addr_add_learn_failed;
        }
    }
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_RETURN(retValue);

adk_ipv6_interface_addr_add_learn_failed:
adk_ipv6_interface_addr_add_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}


/**
 *=============================================================
 * adk_ipv6_interface_addr_rem
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6IfHandle - pointer to the IPv6 interface handle
 *
 * @param[in] ipv6Addr - pointer to the IPv6 interface address
 *
 * @retval ADK_SUCCESS Successfully updated Ingress association
 *         rule mask
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @brief
 *
 * This function removes IPv6 address from an interface.
 * =============================================================
 */
adk_status_t adk_ipv6_interface_addr_rem(
    adk_module_handle_t  ipv6Handle,
    adk_interface_handle_t  ipv6IfHandle,
    const adk_ipv6_addr_prefix_t *ipv6AddrPrefix)
{
    adk_status_t 		    retValue = ADK_SUCCESS;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_sys_handle_t 		sys_handle;
    uint32_t                i, vrd_id;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;
    bool                    linkLocalAction = TRUE;
    adk_ipv6_addr_prefix_t  ipv6_addr_info;
    adk_ipv6_interface_t    *ipv6If_p = NULL;
    bool is_primary = FALSE;
    uint16_t ipv6_addr_stats_id;
    adk_ipv6_dp_ip_addr_stats_t *dp_stats;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module info is NULL\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    if ( !ipv6IfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "NULL pointer parameter specified\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "Invalid Interface Handle type\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    ipv6If_p = &if_pvt_info->if_info_u.regularIf;

    if ( (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_TUNNEL) ||
         (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                           "This API is not supported for Tunnel and Vitual Interface\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    /* Add IPv6 address to database */
    if ( if_pvt_info->numAddr < 1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER); 
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                       "No IPv6 address is associated with an interface\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    if (adk_ipv6_compare_address(ipv6If_p->ipv6AddrPrefix, *ipv6AddrPrefix))
    {
	    is_primary = TRUE;

	    if (if_pvt_info->numAddr == 0)
	    {
		    retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
		    ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Cannot remove the one and only IP address\n");
                    goto adk_ipv6_interface_addr_rem_return;

	    }
    }
  
    for ( i = 0; i < if_pvt_info->numAddr; i++)
    {
        if (adk_ipv6_compare_address(if_pvt_info->ipv6_prefix[i], 
                                     *ipv6AddrPrefix))
        {
            break;
        }
    }

    if ( i >= if_pvt_info->numAddr)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER); 
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, 
                       "This IPv6 address is not already added to the interface\n");
        goto adk_ipv6_interface_addr_rem_return;
    }

    vrd_id = ADK_IPV6_LOCAL_ROUTING_TABLE;

    ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6AddrPrefix->Ipv6Addr.addr[0]); 
    ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6AddrPrefix->Ipv6Addr.addr[1]); 
    ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6AddrPrefix->Ipv6Addr.addr[2]); 
    ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6AddrPrefix->Ipv6Addr.addr[3]); 
    ipv6_addr_info.prefixLength     = ADK_IPV6_PDU_ADDR_LENGTH; 

    if ( linkLocalAction == TRUE)
    {
        /* remove the route entry */
        retValue = adk_ipv6_dp_route_entry_rem(ipv6Handle,
                vrd_id, ipv6_addr_info, FALSE);

        if(retValue != ADK_SUCCESS)
        {
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                    "Error while removing route entry \n");
            goto adk_ipv6_interface_addr_rem_failed_return;
        }
    }
    /* Update the ipv6 interface internal structure */
    if_pvt_info->numAddr--;
    ipv6ModuleInfo_p->totalIpv6Address--;
    if (is_primary == TRUE)
    {
	ipv6_addr_stats_id = if_pvt_info->ipv6_addr_stats_id;
        for(i=0;i<ADK_MAX_NUM_CORES;i++)
        {
            dp_stats = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6AddrStatsTble[i][ipv6_addr_stats_id]);
            ADK_MEMSET(dp_stats, 0, sizeof(adk_ipv6_dp_ip_addr_stats_t));
        }
        adk_rm_resource_free(ipv6ModuleInfo_p->ipv6AddrStatsHandle, if_pvt_info->ipv6_addr_stats_id);
        if_pvt_info->ipv6_addr_stats_id = 0;

        if_pvt_info->if_info_u.regularIf.ipv6AddrPrefix.Ipv6Addr.addr[0] = if_pvt_info->ipv6_prefix[if_pvt_info->numAddr].Ipv6Addr.addr[0];
        if_pvt_info->if_info_u.regularIf.ipv6AddrPrefix.Ipv6Addr.addr[1] = if_pvt_info->ipv6_prefix[if_pvt_info->numAddr].Ipv6Addr.addr[1];
        if_pvt_info->if_info_u.regularIf.ipv6AddrPrefix.Ipv6Addr.addr[2] = if_pvt_info->ipv6_prefix[if_pvt_info->numAddr].Ipv6Addr.addr[2];
        if_pvt_info->if_info_u.regularIf.ipv6AddrPrefix.Ipv6Addr.addr[3] = if_pvt_info->ipv6_prefix[if_pvt_info->numAddr].Ipv6Addr.addr[3];
        if_pvt_info->if_info_u.regularIf.ipv6AddrPrefix.prefixLength = if_pvt_info->ipv6_prefix[if_pvt_info->numAddr].prefixLength;
        if_pvt_info->ipv6_addr_stats_id = if_pvt_info->ipv6_prefix_stats_id[if_pvt_info->numAddr];

    }
    else
    {
        ADK_MEMCPY( &(if_pvt_info->ipv6_prefix[i]), &(if_pvt_info->ipv6_prefix[if_pvt_info->numAddr]), 
                                                                     sizeof(adk_ipv6_addr_prefix_t));
        ADK_MEMSET( &(if_pvt_info->ipv6_prefix[if_pvt_info->numAddr]), 0, sizeof(adk_ipv6_addr_prefix_t));
    }
  
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_RETURN(retValue);

adk_ipv6_interface_addr_rem_failed_return:
adk_ipv6_interface_addr_rem_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}

/**
 *=============================================================
 * adk_ipv6_interface_mtu_set
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6IfHandle - pointer to the IPv6 interface handle
 *
 * @param[in] mtu -  configure MTU value for the interface
 *
 * @retval ADK_SUCCESS Successfully configured ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle might be NULL.
 *
 * @brief
 *
 * This function configures the IPv6 interface MTU
 * =============================================================
 */
adk_status_t
adk_ipv6_interface_mtu_set(
    adk_module_handle_t ipv6Handle,
    adk_interface_handle_t ipv6IfHandle,
    uint32_t mtu)
{
    adk_status_t        retValue = ADK_SUCCESS;
    ipv6_interface_info_t   *if_pvt_info = NULL;
    adk_sys_handle_t        sys_handle;
    uint32_t                preVal = ADK_IPV6_INTERFACE_DEFAULT_MTU;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_interface_mtu_set_return;
    }

    if ( !ipv6IfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                           "NULL pointer parameter specified\n");
        goto adk_ipv6_interface_mtu_set_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                           "Invalid Interface Handle type\n");
        goto adk_ipv6_interface_mtu_set_return;
    }
    /* Validate MTU input */
    if (mtu == 0)
    {
        mtu = ADK_IPV6_INTERFACE_DEFAULT_MTU;
    }
    else if ((mtu < ADK_IPV6_INTERFACE_MIN_MTU)
             || (mtu >= ADK_IPV6_INTERFACE_MAX_MTU))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                      "MTU must be between 1280 to 32767 \n");
        goto adk_ipv6_interface_mtu_set_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                      "This attribute is not applicable for Virtual Interface\n");
        goto adk_ipv6_interface_mtu_set_return;
    }

    ADK_IPV6_IF_INFO_SET(if_pvt_info,mtu,mtu);

    retValue = adk_ipv6_dp_interface_entry_mtu_set(ipv6Handle,
            if_pvt_info,
            mtu);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV6_IF_INFO_SET(if_pvt_info,mtu,preVal);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                           "Failed to configure the interface rule\n");
        goto adk_ipv6_interface_mtu_set_return;
    }
    adk_sys_flow_cache_update_version(sys_handle);

adk_ipv6_interface_mtu_set_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
    return retValue;
}    /* -----  end of function adk_ipv6_interface_mtu_set ----- */

/**
 *=============================================================
 * adk_ipv6_tunnel_interface_outer_pmtu_set
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6IfHandle - pointer to the IPv6 interface handle
 *
 * @param[in] outerPathMtu -  configure Outer Path MTU value for the Tunnel interface
 *
 * @retval ADK_SUCCESS Successfully configured ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle might be NULL.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 * This API will be deprecated.
 * To Set Tunnel Outer Path MTU use adk_ipv6_tunnel_interface_mtu_set
 * This function configures the IPv6 tunnel interface - Outer Path MTU
 * NOTE: This API is not supported
 * =============================================================
 */
adk_status_t
adk_ipv6_tunnel_interface_outer_pmtu_set(
    adk_module_handle_t  __attribute__((unused))ipv6Handle,
    adk_interface_handle_t  __attribute__((unused))ipv6IfHandle,
    uint32_t  __attribute__((unused))outerPathMtu)
{
    return ADK_SUCCESS;
}

/**
 *=============================================================
 * adk_ipv6_dscp_pcp_map_add 
 * @param[in] ipv6Handle    IPv6 module handle.
 *
 * @param[in] ipv6IfHandle  IPv6 interface for which the mapping needs to be
 * added
 *
 * @param[in] dscp          The DSCP value to be mapped
 *
 * @param[in] pcp           The PCP value to which DSCP needs to be mapped
 *
 * @retval ADK_SUCCESS Successfully configured ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle might be NULL or invalid.
 *
 * @brief
 *
 * This function maps a DSCP value to a PCP value for a given IPv6
 * interface. The IPv6 interface must an IPoE type interface i.e. it cannot be
 * tunnel IPv6 interface.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_dscp_pcp_map_add(
    adk_module_handle_t  ipv6Handle,
    adk_interface_handle_t  ipv6IfHandle,
    uint8_t  dscp,
    uint8_t  pcp)
{
    adk_status_t 		retValue = ADK_SUCCESS;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_sys_handle_t 		sys_handle;
    uint8_t                      preval = 0;

    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Module Info is NULL or cookie mismatches\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    if ( !ipv6IfHandle)
    {
        ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                           "NULL pointer parameter specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }
    /* Validate DSCP and PCP input */
    if ( dscp > 63) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                      "DSCP must be between 0 to 63 \n");
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    if ( pcp > 7) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                      "PCP must be between 0 to 7 \n");
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }
    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL) ||
         (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_TUNNEL))
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, 
                      "This attribute is not applicable for Virtual/Tunnel Interface\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    preval = if_pvt_info->dscp_pcp[dscp];
    if_pvt_info->dscp_pcp[dscp] = pcp;

    retValue = adk_ipv6_dp_interface_entry_dscp_pcp_set(ipv6Handle,
            if_pvt_info,
            dscp, pcp);
    if (retValue != ADK_SUCCESS)
    {
        
        if_pvt_info->dscp_pcp[dscp] = preval;
        ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                           "Failed to configure the interface rule\n");
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

adk_ipv6_interface_dscp_pcp_map_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

    return retValue;
}    /* -----  end of function adk_ipv6_dscp_pcp_map_add ----- */

/*!
 *=============================================================
 * adk_ipv6_dscp_pcp_map_delete
 *
 * @param[in]   ipv6Handle        IPv6 module handle
 *
 * @param[in]   ipv6IfHandle      IPv6 interface for which the mapping needs to be
 * removed
 *
 * @param[in]   dscp              The DSCP value for which mapping needs to be
 * removed
 *
 * @retval ADK_SUCCESS              Successfully removed DSCP to PCP mapping
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid module or interface handle
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid DSCP value (exceeding max)
 *
 * @brief
 *
 * This function removes mapping from DSCP to PCP value for a given IPv6
 * interface. The IPv6 interface must an IPoE type interface i.e. it cannot be
 * tunnel IPv6 interface.
 *
 * =============================================================
 */
adk_status_t adk_ipv6_dscp_pcp_map_delete(
    adk_module_handle_t  ipv6Handle,
    adk_interface_handle_t  ipv6IfHandle,
    uint8_t  dscp)
{
    adk_status_t 		retValue = ADK_SUCCESS;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_sys_handle_t 		sys_handle;
    uint8_t                    preval = 0;

    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Module Info is NULL or cookie mismatches\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    if ( !ipv6IfHandle)
    {
        ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                           "NULL pointer parameter specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }
    /* Validate DSCP and PCP input */
    if ( dscp > 63) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle, 
                      "DSCP must be between 0 to 63 \n");
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL) ||
         (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_TUNNEL))
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, 
                      "This attribute is not applicable for Virtual/Tunnel Interface\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

    preval = if_pvt_info->dscp_pcp[dscp];
    if_pvt_info->dscp_pcp[dscp] = 0;

    retValue = adk_ipv6_dp_interface_entry_dscp_pcp_set(ipv6Handle,
            if_pvt_info,
            dscp, 0);
    if (retValue != ADK_SUCCESS)
    {
        
        if_pvt_info->dscp_pcp[dscp] = preval;
        ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                           "Failed to configure the interface rule\n");
        goto adk_ipv6_interface_dscp_pcp_map_return;
    }

adk_ipv6_interface_dscp_pcp_map_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

    return retValue;
}

/**
 *=============================================================
 * adk_ipv6_interface_ingress_policing_config 
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6IfHandle - pointer to the IPv6 interface handle
 *
 * @param[in] enable -  Enable/disable policing based on the 
 *                      on this configured interface and 
 *                      configure the policing flow id
 *
 * @retval ADK_SUCCESS Successfully configured ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface handle might be NULL.
 *
 * @brief
 *
 * This API is not supported
 *
 * This function configures whether policing is enabled on the 
 * interface or not
 *
 * NOTE: This API is not yet supported.
 * =============================================================
 */
adk_status_t
adk_ipv6_interface_ingress_policing_config(
    adk_module_handle_t __attribute__((unused)) ipv6Handle,
    adk_interface_handle_t __attribute__((unused)) ipv6IfHandle,
    const adk_ipv6_interface_ingress_policing_attr_t __attribute__((unused)) *policingAttr_p)
{
    adk_status_t 		retValue = ADK_SUCCESS;
    return retValue;
}    /* -----  end of function adk_ipv6_interface_ingress_policing_config ----- */

/*!
 * =============================================================
 * adk_ipv6_interface_priority_map_profile_config
 *
 * @param[in] ipv6Handle              IPv6 module handle.
 *
 * @param[in] ipv6IfHandle            IPv6 interface handle for which the IPv6
 * filter rule set needs to be set.
 *
 * @param[in] direction             Direction (ingress or egress) in which
 * the filter rule set needs to be set for the interface.
 *
 * @param[in] profile_id            Qos Profile Id to be set
 * for the IP interface. 
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv6 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER  One of the direct arguments
 * to the API call is invalid.
 *
 * @brief
 *
 * This API is not supported
 *
 * This API configures IPv6 Qos Profile Id for an IPv6 interface. The 
 * packets received on the interface are then subject to the idscp remarking
 * based on the added profile id. Note that this API must be 
 * called for every IPv6 interface on which ingress or egress dscp remarking 
 * needs to be enabled. Also note that an IP interface can only be 
 * associated with one ingress and one egress profile id. So, calling this 
 * API will replace the existing association if any present.
 *
 * NOTE: This API is not yet supported.
 * =============================================================
 */
adk_status_t 
adk_ipv6_interface_priority_map_profile_config(
    adk_module_handle_t	    ipv6Handle,
    adk_interface_handle_t 	ipv6IfHandle,
    adk_ipv6_direction_t  	direction,
    uint32_t 		        profile_id)
{
    adk_status_t 		    retValue = ADK_SUCCESS;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_sys_handle_t 		sysHandle;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;

    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    if ( direction != ADK_IPV6_DIR_INGRESS)
    {
        ADK_FORCE_LOG_ERROR( "This direction (%d) is not supported\n", direction);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

    if ( profile_id > ADK_IPV6_DSCP_MAP_PROFILE_ID_MAX)
    {
        ADK_FORCE_LOG_ERROR( "Profile id (%d) is not present in valid range\n", profile_id);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);
    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Module Info is NULL or cookie mismatches\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_profile_config_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Module info is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_profile_config_return;
    }

    if ( !ipv6IfHandle)
    {
        ADK_LOG_ERROR( sysHandle, ipv6Handle, 
                           "NULL pointer parameter specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_interface_dscp_profile_config_return;
    }

    if (!adk_ipv6_interface_valid_handle(ipv6IfHandle))
    {
        ADK_LOG_ERROR( sysHandle, ipv6Handle, 
                           "Invalid Interface Handle type\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_interface_dscp_profile_config_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL)
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, 
                      "This attribute is not applicable for Virtual Interface\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_interface_dscp_profile_config_return;
    }

    if ( profile_id != 0 )
    {
        if_pvt_info->enableL3Remark = TRUE;
    }
    else
    {
        if_pvt_info->enableL3Remark = FALSE;
    }

    /* update profile id  */
    if_pvt_info->profile_Id  = profile_id;
    retValue = adk_ipv6_dp_interface_entry_profile_set(ipv6Handle, if_pvt_info, profile_id);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR( sysHandle, ipv6Handle, 
                           "Failed to update the profile id \n");
        goto adk_ipv6_interface_dscp_profile_config_return;
    }
    
adk_ipv6_interface_dscp_profile_config_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
    return retValue;
}

/*!
 * =============================================================
 * adk_ipv6_interface_filter_rule_set_map
 *
 * @param[in] ipv6Handle              IPv6 module handle.
 *
 * @param[in] ipv6IfHandle            IPv6 interface handle for which the IPv6
 * filter rule set needs to be set.
 *
 * @param[in] direction             Direction (ingress or egress) in which
 * the filter rule set needs to be set for the interface.
 *
 * @param[in] filterRuleSetHandle   Filter rule-set to be set
 * for the IP interface. 
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv6 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER  One of the direct arguments
 * to the API call is invalid.
 *
 * @brief
 *
 * This API is not supported
 *
 * This API configures IPv6 filter rule-set for an IPv6 interface. The 
 * packets received on the interface are then subject to the filtering
 * rules added to the specified rule set. Note that this API must be 
 * called for every IPv6 interface on which ingress or egress filtering 
 * needs to be enabled. Also note that an IP interface can only be 
 * associated with one ingress and one egress rule-set. So, calling this 
 * API will replace the existing association to any filter rule-set, 
 * if any present.
 *
 * =============================================================
 */
adk_status_t adk_ipv6_interface_filter_rule_set_map(
    adk_module_handle_t ipv6Handle,
    adk_interface_handle_t ipv6IfHandle,
    adk_ipv6_filter_direction_t direction,
    adk_ipv6_filter_rule_set_handle_t filterRuleSetHandle)
{
    adk_status_t                retValue = ADK_SUCCESS;
    // TODO: API will be removed/cleaned later, as its not used. Right now fixing compilation issue
    (void)ipv6Handle;
    (void)ipv6IfHandle;
    (void)direction;
    (void)filterRuleSetHandle;
#if ADK_IPV6_ACL_ENABLE
    ipv6_interface_info_t       *if_pvt_info = NULL;
    adk_sys_handle_t 		    sys_handle;
    uint32_t                    filterRuleSetId;
    ipv6_filter_rule_set_info_t *filter_info_p = NULL;

    /* Handle validations */
    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Module Info is NULL or cookie mismatches\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto _adk_ipv6_interface_filter_set_exit;
    }

    /*Check if valid input interface input is given*/
    if ( !adk_ipv6_interface_valid_handle(ipv6IfHandle) )
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Invalid IPv6 interface specified\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipv6_interface_filter_set_exit;
    }
    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if (!adk_ipv6_valid_rule_set_handle(filterRuleSetHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle,ipv6Handle,
                "Invalid IP Filter Rule Set Handle\n");
        goto _adk_ipv6_interface_filter_set_exit;
    }

    filter_info_p = (ipv6_filter_rule_set_info_t *) filterRuleSetHandle; 
    filterRuleSetId = filter_info_p->ipv6_filer_rule_set_id;

    if ( direction == ADK_IPV6_FILTER_IN)
    {
        if_pvt_info->ingressRuleSetHandle = filterRuleSetHandle;
    }
    else
    {
        if_pvt_info->egressRuleSetHandle = filterRuleSetHandle;
    }

    retValue = adk_ipv6_dp_interface_entry_filter_set(ipv6Handle,
                                                      if_pvt_info,
                                                      direction, 
                                                      filterRuleSetId);
   
#ifdef ADK_HOS_OLD_IMPL
    adk_status_t                ret = ADK_SUCCESS;
    if ((direction == ADK_IPV6_FILTER_IN) && if_pvt_info)
    {

        if(if_pvt_info->ifType == IPV6_INTERFACE_TYPE_ETH)
        {
            adk_module_handle_t eth_mod_handle=NULL;
            ret = adk_module_handle_get(sys_handle, ADK_ETHERNET_MODULE_ID, NULL,
                    &eth_mod_handle);
            if (ret != ADK_SUCCESS)
            {
                ADK_LOG_FATAL_ERROR(sys_handle, ipv6Handle, 
                        "Ethernet module handle not found!\n");

                /* do not return error if it fails to configure the hardware */
                goto _adk_ipv6_interface_filter_set_exit;
            }

            ret = adk_eth_intf_acl_config(eth_mod_handle,
                    if_pvt_info->if_info_u.regularIf.layer2InterfaceHandle,
                    if_pvt_info->if_info_u.regularIf.enableIngAcl,
                    false, true, filterRuleSetId);
            if (ret != ADK_SUCCESS)
            {
                ADK_LOG_WARNING(sys_handle, ipv6Handle, 
                        "Failed to set ACL ruleset in HOS table!\n");
                /* do not return error if it fails to configure the hardware */
                goto _adk_ipv6_interface_filter_set_exit;
            }
        }
        else if(if_pvt_info->ifType == IPV6_INTERFACE_TYPE_TUNNEL)
        {
            ret = adk_ipv6_tnl_intf_acl_config(ipv6Handle, ipv6IfHandle, TRUE);
            if (ret != ADK_SUCCESS)
            {
                ADK_LOG_WARNING(sys_handle, ipv6Handle, 
                        "Failed to set ACL ruleset in HOS table!\n");
                /* do not return error if it fails to configure the hardware */
                goto _adk_ipv6_interface_filter_set_exit;
            }
        }
    }
#endif

_adk_ipv6_interface_filter_set_exit:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);
#endif

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @internal
 *
 * adk_ipv6_config_tunnel_cpu_path
 *
 * @brief
 *
 * This API is not supported
 *
 * Internal support function for configuration of tunnel interface
 * for CPU path connectivity.
 *
 * NOTE: This API is not yet supported.
 * =============================================================
 */
adk_status_t adk_ipv6_config_tunnel_cpu_path(adk_module_handle_t __attribute__((unused)) ipv6Handle,
                                             uint32_t __attribute__((unused)) adk_interface_id,
                                             uint32_t __attribute__((unused)) cpu_interface_id,
                                             bool __attribute__((unused)) addEntry)
{
    adk_status_t                retValue = ADK_SUCCESS;
    return(retValue);
}

 
/*!
 * =============================================================
 * adk_ipv6_interface_statistics_get
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    interfaceHandle - Interface for which want to 
 * get statistics
 *
 * @param[out]   ipv6InterfaceStats_p - Interface statistics
 * structure
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 * @retval ADK_E_INVALID_PARAMETER - IPv6 inteface handle
 * not present or invalid.  
 * 
 * @retval ADK_IPV6_E_INVALID_PARAMETER -  one of the direct 
 * arguments to the API call is invalid.
 * 
 * @brief  
 * This API validates the input parameter and get the statistics
 * for the specified interface
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_interface_statistics_get (
        adk_module_handle_t  ipv6Handle,
        adk_interface_handle_t ipv6IfHandle,
        adk_ipv6_interface_statistics_t  *ipv6InterfaceStats_p)
{
    adk_status_t                     retValue = ADK_IPV6_SUCCESS;
    uint32_t                         i = 0, j = 0, if_id;
    adk_sys_handle_t                 sys_handle;
    adk_ipv6_info_t                  *ipv6ModuleInfo_p = NULL;
    adk_ipv6_dp_if_stats_entry_t     **if_stats = NULL;
    ipv6_interface_info_t            *if_pvt_info = NULL;
    adk_sys_egress_stats_t           eg_stats;
    adk_ipv6_dp_interface_entry_t    *ipv6_if_entry = NULL;

    /* Handle validations */
    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid module handle\n");
        return (retValue);
    }

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module Info is NULL or cookie mismatches\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    /*Check if IPv6 module info is not NULL*/
    if ( !ipv6InterfaceStats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IPV6_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid input parameter\n");
        goto adk_ipv6_get_interface_stats_return;
    }
    
    /*Check if valid input interface input is given*/
    if ( !adk_ipv6_interface_valid_handle(ipv6IfHandle) )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid IPv6 interface specified\n");
        goto adk_ipv6_get_interface_stats_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        goto adk_ipv6_get_interface_stats_return;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6IfHandle);

    if ( if_pvt_info == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Interface info is NULL\n");
        goto adk_ipv6_get_interface_stats_return;
    }

    ADK_MEMSET(ipv6InterfaceStats_p, 0, sizeof(adk_ipv6_interface_statistics_t));
    if ( (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_ETH) ||
         (if_pvt_info->ifType == IPV6_INTERFACE_TYPE_TUNNEL))
    {
        if_stats = ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_if_table.stats_tbl;
        if_id = ADK_SYS_INTERFACE_ID_GET(ipv6IfHandle);
    }
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Interface Type is not supported\n");
        goto adk_ipv6_get_interface_stats_return;
    }

    for(i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        ipv6InterfaceStats_p->ipv6IfInMCastPackets
            += if_stats[i][if_id].ipv6IfStatsInMcastPkts      ;

        ipv6InterfaceStats_p->ipv6IfInMCastOctets
            += if_stats[i][if_id].ipv6IfStatsHCInMcastOctets  ;

        ipv6InterfaceStats_p->ipv6IfOutMCastPackets
            += if_stats[i][if_id].ipv6IfStatsOutMcastPkts     ;
 
        ipv6InterfaceStats_p->ipv6IfOutMCastOctets
            += if_stats[i][if_id].ipv6IfStatsOutMcastOctets   ;

        ipv6InterfaceStats_p->ipv6IfInDiscardedMCastPackets
            += if_stats[i][if_id].ipv6IfStatsInMcastErrors    ;

        ipv6InterfaceStats_p->ipv6IfInDiscardedMCastOctets
            += if_stats[i][if_id].ipv6IfInDiscardedMCastOctets;

        ipv6InterfaceStats_p->ipv6IfInPackets
            += if_stats[i][if_id].ipv6IfStatsInReceives  +
               if_stats[i][if_id].ipv6IfStatsInMcastPkts      ;

        ipv6InterfaceStats_p->ipv6IfOutPackets
            += if_stats[i][if_id].ipv6IfStatsOutMcastPkts +
               if_stats[i][if_id].ipv6IfStatsOutUcastPkts     ;

        ipv6InterfaceStats_p->ipv6IfInOctets
            += if_stats[i][if_id].ipv6IfStatsHCInOctets +
               if_stats[i][if_id].ipv6IfStatsHCInMcastOctets  ;

        ipv6InterfaceStats_p->ipv6IfOutOctets
            += if_stats[i][if_id].ipv6IfStatsHCOutOctets +
               if_stats[i][if_id].ipv6IfStatsOutMcastOctets   ;

        ipv6InterfaceStats_p->ipv6IfOutFragPackets
            += if_stats[i][if_id].ipv6IfStatsOutFragReqds     ;

        ipv6InterfaceStats_p->ipv6IfOutFragments
            += if_stats[i][if_id].ipv6IfStatsOutFragOKs       ;

        ipv6InterfaceStats_p->ipv6IfInErrors
            += if_stats[i][if_id].ipv6IfStatsInHdrErrors  +
               if_stats[i][if_id].ipv6IfStatsInAddrErrors +
               if_stats[i][if_id].ipv6IfStatsInSrcAddrErrors +
               if_stats[i][if_id].ipv6IfStatsInTruncatedPkts +
               if_stats[i][if_id].ipv6IfStatsInMcastErrors   +
               if_stats[i][if_id].ipv6IfNotMulOfEightFragLen  +
               if_stats[i][if_id].ipv6IfInvalidReasmSessionId+
               if_stats[i][if_id].ipv6IfReasmOverlapFrag     +
               if_stats[i][if_id].ipv6IfExpectedReasmLengthExceed +
               if_stats[i][if_id].ipv6IfL4Hdrtruncated +
               if_stats[i][if_id].ipv6IfUDPCksumErrors +
               if_stats[i][if_id].ipv6IfICMPCksumErrors       ; 

        ipv6InterfaceStats_p->ipv6IfOutDroppedPackets
            += if_stats[i][if_id].ipv6IfStatsOutFragFails +
               if_stats[i][if_id].ipv6IfStatsOutDiscards      ;

        ipv6InterfaceStats_p->ipv6IfOutDroppedOctets
            += if_stats[i][if_id].ipv6IfStatsOutDroppedOctets ;

        ipv6InterfaceStats_p->ipv6IfNotMulOfEightFragLen
            += if_stats[i][if_id].ipv6IfNotMulOfEightFragLen   ;

        ipv6InterfaceStats_p->ipv6IfInvalidReasmSessionId
            += if_stats[i][if_id].ipv6IfInvalidReasmSessionId ;

        ipv6InterfaceStats_p->ipv6IfReasmOverlapFrag
            += if_stats[i][if_id].ipv6IfReasmOverlapFrag      ;

        ipv6InterfaceStats_p->ipv6IfTooManyFragsDropCount
            += if_stats[i][if_id].ipv6IfTooManyFragsDropCount ;

        ipv6InterfaceStats_p->ipv6IfExpectedReasmLengthExceed
            += if_stats[i][if_id].ipv6IfExpectedReasmLengthExceed  ;

        ipv6InterfaceStats_p->ipv6IfForwardingDisableDrop
            += if_stats[i][if_id].ipv6IfForwardingDisableDrop ;

        ipv6InterfaceStats_p->ipv6IfL4Hdrtruncated
            += if_stats[i][if_id].ipv6IfL4Hdrtruncated        ;

        ipv6InterfaceStats_p->ipv6IfUDPCksumErrors
            += if_stats[i][if_id].ipv6IfUDPCksumErrors        ;

        ipv6InterfaceStats_p->ipv6IfICMPCksumErrors
            += if_stats[i][if_id].ipv6IfICMPCksumErrors       ;

        ipv6InterfaceStats_p->ipv6IfUnsupportedProtoDrops
            += if_stats[i][if_id].ipv6IfUnsupportedProtoDrops ;

        ipv6InterfaceStats_p->ipv6IfStatsInTruncatedPkts
            += if_stats[i][if_id].ipv6IfStatsInTruncatedPkts  ;

        ipv6InterfaceStats_p->ipv6IfQoSLookupFailed
            += if_stats[i][if_id].ipv6IfQoSLookupFailed;

        for(j = 0; j < ADK_QOS_MAX_PHB_CLASS_IDX ; j++) {
            ipv6InterfaceStats_p->ipv6ifRxQoSstats[j].rxPkts += if_stats[i][if_id].ipv6ifRxQoSstats[j].rxPkts; 
            ipv6InterfaceStats_p->ipv6ifRxQoSstats[j].rxBytes += if_stats[i][if_id].ipv6ifRxQoSstats[j].rxBytes; 
        }
    }

    if(if_pvt_info->ifType == IPV6_INTERFACE_TYPE_ETH)
    {
        ipv6_if_entry = &ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_if_table.tbl[if_id];

        retValue = adk_sys_egress_stats_get(sys_handle, ipv6_if_entry->sys_egress_stats_id, &eg_stats);
        if(retValue != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                    "Error in reading sys egress stats for IPv6 interface err %d \n", retValue);
            goto adk_ipv6_get_interface_stats_return;
        }
    }

adk_ipv6_get_interface_stats_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

    ADK_RETURN ( retValue ); /* -----  end of function adk_ipv6_interface_statistics_get ----- */
}

#ifdef ADK_DEBUG
/*!
 * =============================================================
 * adk_ipv6_interface_statistics_reset
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    interfaceHandle - Interface for which want to 
 * get statistics
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 * @retval ADK_E_INVALID_PARAMETER - IPv6 inteface handle
 * not present or invalid.  
 * 
 * @retval ADK_IPV6_E_INVALID_PARAMETER -  one of the direct 
 * arguments to the API call is invalid.
 * 
 * @brief 
 *
 * This API is not supported
 * 
 * This API validates the input parameter and reset the statistics
 * for the specified interface
 *
 *
 * NOTE: This API is not yet supported.
 * =============================================================
 */
adk_status_t 
adk_ipv6_interface_statistics_reset ( 
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
        adk_interface_handle_t __attribute__((unused)) ipv6IfHandle)
{
    adk_status_t retValue = ADK_IPV6_SUCCESS;
    ADK_RETURN ( retValue ); /* -----  end of function adk_ipv6_interface_statistics_get ----- */
}
#endif
/*!
 * =============================================================
 * adk_ipv6_interface_statistics_print
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    ipv6IfHandle - Interface for which want to 
 * get statistics
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 *
 * @brief  
 * This API prints the interface statistics 
 *
 * =============================================================
 */
adk_status_t 
adk_ipv6_interface_statistics_print(
                             adk_module_handle_t ipv6Handle,
                             adk_interface_handle_t ifHandle)
{
    adk_status_t                       retValue = ADK_SUCCESS;
    adk_ipv6_interface_statistics_t    ipv6InterfaceStats;
    uint8_t                            i = 0;

    ADK_MEMSET(&ipv6InterfaceStats, 0, sizeof(adk_ipv6_interface_statistics_t));

    retValue = adk_ipv6_interface_statistics_get(ipv6Handle, ifHandle, &ipv6InterfaceStats);
    if (retValue != ADK_SUCCESS)
        return retValue;

    ADK_INFO_WRITE("\nIPv6 Interface 0x%08x Statistics\n", ADK_SYS_INTERFACE_ID_GET(ifHandle));
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Multicast packets Received                                : %"PRId64"\n", ipv6InterfaceStats.ipv6IfInMCastPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Multicast Octets Received                                 : %"PRId64"\n", ipv6InterfaceStats.ipv6IfInMCastOctets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Multicast packets transmitted                             : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutMCastPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Multicast octets transmitted                              : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutMCastOctets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Ingress Multicast packets dropped                         : %d\n", ipv6InterfaceStats.ipv6IfInDiscardedMCastPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Ingress Multicast octets dropped                          : %d\n", ipv6InterfaceStats.ipv6IfInDiscardedMCastOctets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Rx Pkts                                                   : %"PRId64"\n", ipv6InterfaceStats.ipv6IfInPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Rx Octets                                                 : %"PRId64"\n", ipv6InterfaceStats.ipv6IfInOctets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Tx Pkts                                                   : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Tx Octets                                                 : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutOctets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Fragmentation Reqd Pkts                                         : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutFragPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Fragmentation Ok Pkts                                           : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutFragments);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Rx Error pkts                                             : %d\n", ipv6InterfaceStats.ipv6IfInErrors);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Out Discarded Pkts                                        : %d\n", ipv6InterfaceStats.ipv6IfOutDroppedPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total Out Discarded Pkts                                        : %"PRId64"\n", ipv6InterfaceStats.ipv6IfOutDroppedOctets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Fragments with frag length is not multiple of 8 discarded       : %d\n", ipv6InterfaceStats.ipv6IfNotMulOfEightFragLen);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Fragments belong to invalid session discarded                   : %d\n", ipv6InterfaceStats.ipv6IfInvalidReasmSessionId);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Total overlap fragments discarded                               : %d\n", ipv6InterfaceStats.ipv6IfReasmOverlapFrag);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Too many frags received discarded                               : %d\n", ipv6InterfaceStats.ipv6IfTooManyFragsDropCount);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Fragments with max reassmbly pdu length discarded               : %d\n", ipv6InterfaceStats.ipv6IfExpectedReasmLengthExceed);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("QoS Lookup failed discarded Pkts                                : %d\n", ipv6InterfaceStats.ipv6IfQoSLookupFailed);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("ICMP6 checksum validation failed packets                        : %d\n", ipv6InterfaceStats.ipv6IfICMPCksumErrors);

    for(i = 0; i < ADK_QOS_MAX_PHB_CLASS_IDX ; i++) {
        if (ipv6InterfaceStats.ipv6ifRxQoSstats[i].rxPkts) {
            ADK_INFO_WRITE("Rx QoS Pkts  [PHB: %u]                                      : %lu\n", i, ipv6InterfaceStats.ipv6ifRxQoSstats[i].rxPkts);
        }
        if (ipv6InterfaceStats.ipv6ifRxQoSstats[i].rxBytes) {
            ADK_INFO_WRITE("Rx QoS Bytes [PHB: %u]                                      : %lu\n", i, ipv6InterfaceStats.ipv6ifRxQoSstats[i].rxBytes);
        }
    }

    return ADK_SUCCESS;
}

/* 
 * Internal function to compare ipv6 addresses
 */
bool
adk_ipv6_compare_address(adk_ipv6_addr_prefix_t address1,
                         adk_ipv6_addr_prefix_t address2)
{

   if( (address1.Ipv6Addr.addr[0] == address2.Ipv6Addr.addr[0]) &&
       (address1.Ipv6Addr.addr[1] == address2.Ipv6Addr.addr[1]) &&
       (address1.Ipv6Addr.addr[2] == address2.Ipv6Addr.addr[2]) &&
       (address1.Ipv6Addr.addr[3] == address2.Ipv6Addr.addr[3]))
   {
       return TRUE;
   }
   else
   {
       return FALSE;
   }
}

/* 
 * Internal function to validate ipv6 addresses
 */
bool
adk_ipv6_address_validate(adk_module_handle_t ipv6Handle,adk_ipv6_addr_prefix_t address)
{
   adk_sys_handle_t 		     sys_handle;
   sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

   if( (address.Ipv6Addr.addr[0] == 0) &&
       (address.Ipv6Addr.addr[1] == 0) &&
       (address.Ipv6Addr.addr[2] == 0) &&
       (address.Ipv6Addr.addr[3] == 0))
   {
       ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                      " Invalid IPv6 interface address (Unspecified address)\n"); 
       return FALSE;
   }

   if( (address.Ipv6Addr.addr[0] == 0) &&
       (address.Ipv6Addr.addr[1] == 0) &&
       (address.Ipv6Addr.addr[2] == 0) &&
       (address.Ipv6Addr.addr[3] == 1))
   {
       ADK_LOG_ERROR( sys_handle, ipv6Handle, 
                      " Invalid IPv6 interface address (Loopback address)\n"); 
       return FALSE;
   }
  
   return TRUE;
}
/*
 * Internal Function to update the Link Local address database
 */
adk_status_t
adk_ipv6_link_local_addr_db_update( adk_module_handle_t ipv6Handle, 
                                    uint32_t vrd_id, 
                                    adk_ipv6_addr_t Ipv6Addr, 
                                    bool add,
                                    bool *action)
{
    adk_status_t 		     retValue = ADK_SUCCESS;
    adk_sys_handle_t                 sys_handle;
    adk_ipv6_info_t                  *ipv6ModuleInfo_p = NULL;
    uint32_t                     arrayId = 0;
    bool                       entryFound = FALSE;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        return (retValue);
    }


    for ( arrayId = 0; arrayId < ADK_IPV6_MAX_LINK_LOCAL_ADDR; arrayId++)
    {
        if (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].count != 0)
        {
            if ( (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].vrd_id == vrd_id) &&
                 (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].Ipv6Addr.addr[3] == Ipv6Addr.addr[3]) &&
                 (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].Ipv6Addr.addr[2] == Ipv6Addr.addr[2]) &&
                 (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].Ipv6Addr.addr[1] == Ipv6Addr.addr[1]) &&
                 (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].Ipv6Addr.addr[0] == Ipv6Addr.addr[0]))
            {
                entryFound = TRUE;
                break;
            }
        }
        else
        {
            entryFound = FALSE;
            break;
        } 
    }

    if ( add == TRUE)
    {
        if ( entryFound == TRUE)
        {
            ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].count++;
            *action = FALSE;
        }
        else
        {
            if ( ipv6ModuleInfo_p->linkLocalAddrCount == ADK_IPV6_MAX_LINK_LOCAL_ADDR)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
                ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                        "Max supported linkLocal IPv6 address for system is %d\n",
                        ADK_IPV6_MAX_LINK_LOCAL_ADDR);
                return (retValue);

            }
            *action = TRUE;
            ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].count = 1;
            ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].vrd_id = vrd_id;
            ADK_MEMCPY(&(ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].Ipv6Addr), 
                       &Ipv6Addr,
                       sizeof(adk_ipv6_addr_t));
            ipv6ModuleInfo_p->linkLocalAddrCount++;
            *action = TRUE;
        }
    }
    else
    {
        if ( entryFound == TRUE)
        {
            ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].count--;
            *action = FALSE;
            if (ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].count == 0)
            {
                ipv6ModuleInfo_p->linkLocalAddrCount--;
                ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].count = 
                    ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].count;
                ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].vrd_id = 
                    ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].vrd_id;
                ADK_MEMCPY(&(ipv6ModuleInfo_p->linkLocalAddrDb[arrayId].Ipv6Addr),
                           &(ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].Ipv6Addr),
                           sizeof(adk_ipv6_addr_t));
                ipv6ModuleInfo_p->linkLocalAddrDb[ipv6ModuleInfo_p->linkLocalAddrCount].count = 0;
                *action = TRUE;
            }
        }
        else
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
            ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Link Local IPv6 address not found in db\n");
            return (retValue);
        }
    }
    return (retValue);
}
/**
*=============================================================
* adk_ipv6_add_sse_ip
* @param[in]  ipv6Handle - IPv6 module handle.
*
* @param[in] ipv6Addr - pointer to the IPv6 address and its
*                       prefix Length.
*
* @retval ADK_SUCCESS Successfully configured route entry.
*
* @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
* handle pointer might be NULL.
*
* @brief
*
* packets with destination IP address matching any of the addresses
* configured using this API are subjected to classifier lookup and action
* =============================================================
*/
adk_status_t
adk_ipv6_sse_ip_addr_add(adk_module_handle_t ipv6Handle, adk_ipv6_addr_prefix_t ipv6AddrPrefix)
{
    adk_status_t         retValue = ADK_SUCCESS;
    adk_dp_status_t         dpRetVal = ADK_SUCCESS;
    adk_sys_handle_t            sys_handle;
    adk_ipv6_dp_route_entry_t   dpRouteEntry;
    adk_ipv6_info_t         *ipv6ModuleInfo_p = NULL;
    adk_ipv6_addr_prefix_t  ipv6_addr_info;
    uint32_t   ipv6_addr_stats_id;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_sse_addr_add_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        goto adk_ipv6_sse_addr_add_return;
    }

    ipv6_addr_stats_id = adk_rm_resource_reserve(ipv6ModuleInfo_p->ipv6AddrStatsHandle);
    if(ipv6_addr_stats_id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to allocate IP Address Stats ID");
        goto adk_ipv6_sse_addr_add_return;
    }

    ipv6ModuleInfo_p->ipv6_sse_addr_stats_id[ipv6ModuleInfo_p->sseAddrIndex] = ipv6_addr_stats_id;
    ADK_MEMCPY(&(ipv6ModuleInfo_p->ipv6_sse_addr_prefix[ipv6ModuleInfo_p->sseAddrIndex]), &ipv6AddrPrefix, sizeof(adk_ipv6_addr_prefix_t));

    if (!adk_ipv6_address_validate(ipv6Handle, ipv6AddrPrefix))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_sse_addr_add_return;
    }

    ADK_MEMSET(&dpRouteEntry, 0, sizeof(adk_ipv6_dp_route_entry_t));

    ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6AddrPrefix.Ipv6Addr.addr[0]);
    ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6AddrPrefix.Ipv6Addr.addr[1]);
    ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6AddrPrefix.Ipv6Addr.addr[2]);
    ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6AddrPrefix.Ipv6Addr.addr[3]);
    ipv6_addr_info.prefixLength     = ipv6AddrPrefix.prefixLength;
    dpRouteEntry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_TERMINATE;
    dpRouteEntry.ipv6_addr_stats_id = ipv6_addr_stats_id;

    dpRetVal = adk_ipv6_dp_route_entry_add(ipv6Handle, ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info,
                        &dpRouteEntry, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,dpRetVal,
                                "Failed to add route entry for IP termination\n");
        goto adk_ipv6_sse_addr_add_learn_failed;
    }
    ipv6ModuleInfo_p->sseAddrIndex++;

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    ADK_RETURN(retValue);

adk_ipv6_sse_addr_add_learn_failed:
adk_ipv6_sse_addr_add_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}

/**
*=============================================================
* adk_ipv6_del_sse_ip
* @param[in]  ipv6Handle - IPv6 module handle.
*
* @param[in] ipv6Addr - pointer to the IPv6 address and it's [[GG]] its
 *                       prefix Length.
*
* @retval ADK_SUCCESS Successfully removed route entry.
*
* @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
* handle pointer might be NULL.
*
* @brief
*
* This function removes the SSE addresses configured previously
* using adk_ipv6_add_sse_ip() API
* =============================================================
*/

adk_status_t
adk_ipv6_sse_ip_addr_del(adk_module_handle_t ipv6Handle, adk_ipv6_addr_prefix_t ipv6AddrPrefix)
{
    adk_status_t                 retValue = ADK_SUCCESS;
    adk_sys_handle_t             sys_handle;
    adk_dp_status_t              dpRetVal;
    adk_ipv6_info_t              *ipv6ModuleInfo_p = NULL;
    adk_ipv6_addr_prefix_t       ipv6_addr_info;
    adk_ipv6_dp_ip_addr_stats_t  *dp_stats;
    uint32_t ipv6_addr_stats_id, i;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return (retValue);
    }

    /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid module handle\n");
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module Info is NULL or cookie mismatches\n");
        goto adk_ipv6_sse_ip_addr_del_return;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        goto adk_ipv6_sse_ip_addr_del_return;
    }

    ipv6_addr_info.Ipv6Addr.addr[0] = (ipv6AddrPrefix.Ipv6Addr.addr[0]);
    ipv6_addr_info.Ipv6Addr.addr[1] = (ipv6AddrPrefix.Ipv6Addr.addr[1]);
    ipv6_addr_info.Ipv6Addr.addr[2] = (ipv6AddrPrefix.Ipv6Addr.addr[2]);
    ipv6_addr_info.Ipv6Addr.addr[3] = (ipv6AddrPrefix.Ipv6Addr.addr[3]);
    ipv6_addr_info.prefixLength = ipv6AddrPrefix.prefixLength;

    for (i =0; i < ipv6ModuleInfo_p->sseAddrIndex; i++)
    {
        if (adk_ipv6_compare_address(ipv6ModuleInfo_p->ipv6_sse_addr_prefix[i], ipv6_addr_info))
        {
            break;
        }
    }
    if ( i >= ipv6ModuleInfo_p->sseAddrIndex)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                       "This IPv6 address is not already added to the interface\n");
        goto adk_ipv6_sse_ip_addr_del_return;
    }

    /* Deleting the route entry */
    dpRetVal = adk_ipv6_dp_route_entry_rem(ipv6Handle,
            ADK_IPV6_LOCAL_ROUTING_TABLE, ipv6_addr_info, FALSE);
    if(dpRetVal != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,dpRetVal,
                "Failed to remove route entry during interface delete \n");
        goto adk_ipv6_sse_ip_addr_del_return;
    }
    ipv6_addr_stats_id = ipv6ModuleInfo_p->ipv6_sse_addr_stats_id[i];
    for(uint32_t j=0;j<ADK_MAX_NUM_CORES;j++)
    {
        dp_stats = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6AddrStatsTble[j][ipv6_addr_stats_id]);
        ADK_MEMSET(dp_stats, 0, sizeof(adk_ipv6_dp_ip_addr_stats_t));
    }
    adk_rm_resource_free(ipv6ModuleInfo_p->ipv6AddrStatsHandle, ipv6ModuleInfo_p->ipv6_sse_addr_stats_id[i]);
    ipv6ModuleInfo_p->ipv6_sse_addr_stats_id[i] = 0;

    ADK_MEMSET(&(ipv6ModuleInfo_p->ipv6_sse_addr_prefix[i]), 0, sizeof(adk_ipv6_addr_prefix_t));
    for (i=i; i< ipv6ModuleInfo_p->sseAddrIndex; i++)
    {
        ADK_MEMCPY(&(ipv6ModuleInfo_p->ipv6_sse_addr_prefix[i]), &(ipv6ModuleInfo_p->ipv6_sse_addr_prefix[i+1]), sizeof(adk_ipv6_addr_prefix_t));
    }
    ipv6ModuleInfo_p->sseAddrIndex--;

adk_ipv6_sse_ip_addr_del_return:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

static adk_status_t adk_ipv6_create_egress_qos_resources(
       adk_module_handle_t ipv6Handle,
       uint32_t ifId)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipv6_info_t *ipv6ModuleInfo_p;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        goto _adk_ipv6_create_egress_qos_resources_exit;
    }

    /* Allocate and initialize QoS egress classifier resources for this interface*/
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_ACTIVE_TREE_INDEX_INIT] =
        (adk_ipv6_dp_egress_qos_classifier_action_info_t*)
        ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipv6Handle),
                ADK_IPV6_EGRESS_QOS_CLASSIFIER_INFO_OBJ, 1,
                ((ADK_IPV6_EGRESS_QOS_CLASSIFIER_MAX_RULES+ 1) *
                 sizeof(adk_ipv6_dp_egress_qos_classifier_action_info_t)),
                0);
    if(!ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_ACTIVE_TREE_INDEX_INIT])
    {
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ipv6_create_egress_qos_resources_exit;
    }
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_SHADOW_TREE_INDEX_INIT] =
        (adk_ipv6_dp_egress_qos_classifier_action_info_t*)
        ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(ipv6Handle),
                ADK_IPV6_EGRESS_QOS_CLASSIFIER_INFO_OBJ, 1,
                ((ADK_IPV6_EGRESS_QOS_CLASSIFIER_MAX_RULES+ 1) *
                 sizeof(adk_ipv6_dp_egress_qos_classifier_action_info_t)),
                0);
    if(!ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_SHADOW_TREE_INDEX_INIT])
    {
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ipv6_create_egress_qos_resources_exit;
    }
    retValue = adk_ipv6_egress_qos_classifier_create(ipv6Handle, ifId, &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierInfo_p[ifId]));
    if (ADK_SUCCESS != retValue)
    {
        retValue = ADK_E_INTERNAL_ERROR;
        goto _adk_ipv6_create_egress_qos_resources_exit;
    }
    adk_sys_flow_cache_update_version(sys_handle);

    return retValue;

_adk_ipv6_create_egress_qos_resources_exit:
    if (ipv6ModuleInfo_p != NULL)
    {
        if(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_ACTIVE_TREE_INDEX_INIT])
        {
            ADK_DP_SHM_FREE(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_ACTIVE_TREE_INDEX_INIT]);
        }
        if(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_SHADOW_TREE_INDEX_INIT])
        {
            ADK_DP_SHM_FREE(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_SHADOW_TREE_INDEX_INIT]);
        }
    }
    return retValue;
}

static adk_status_t adk_ipv6_delete_egress_qos_resources(
       adk_module_handle_t ipv6Handle,
       uint32_t ifId)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipv6_info_t *ipv6ModuleInfo_p;
    adk_sys_handle_t sys_handle;

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        return retValue;
    }

    if(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_ACTIVE_TREE_INDEX_INIT])
    {
        ADK_DP_SHM_FREE(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_ACTIVE_TREE_INDEX_INIT]);
    }
    if(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_SHADOW_TREE_INDEX_INIT])
    {
        ADK_DP_SHM_FREE(ipv6ModuleInfo_p->adk_ipv6_dp_tables->egressQoSClassifierActionDp_p[ifId][ADK_IPV6_QOS_SHADOW_TREE_INDEX_INIT]);
    }
    retValue = adk_ipv6_egress_qos_classifier_delete(ipv6Handle, ifId);
    if (ADK_SUCCESS != retValue)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                "Failed to delete egress QoS classifier context ");
        return ADK_E_INTERNAL_ERROR;
    }
    adk_sys_flow_cache_update_version(sys_handle);
    return retValue;
}
/*! end addtogroup _ipv6Functions
 * @}
 
*/
