/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_multicast.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief  
 *
 * This file contains all the APIs provided by the IPV6 module.
 *
 ***********************************************************************/
/* Infrastructure header files */
#include "adk_include.h"
#include "adk_ph_dev.h"
#include "acp_gen_interface.h"
#include "adk_utils_tree_update.h"
#include <netinet/in.h>
#include <arpa/inet.h>

#include "adk_shm.h"

/* IPV6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

/* IPV6 internal header files */
#include "adk_ipv6_internal.h"
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_dp_export.h"
#include "adk_ph_dev_dp.h"
#define ADK_CONVERT_TO_V6_VRD__HANDLE(__DATA) ((adk_ipv6_vrd_handle_t)__DATA)
#define ADK_CONVERT_TO_V6_FLOW__HANDLE(__DATA) ((adk_flow_handle_t)__DATA)

/*!
 * =============================================================
 * adk_ipv6_multicast_group_add
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] mcGrp_p pointer to multicast group information.
 * @param[out] mcGrpHdl_p Pointer to multicast group handle.
 * This is set by the API on successfull return.
 *
 * @retval ADK_SUCCESS Successfully multicast group is added.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @retval ADK_E_MEM_UNAVAILABLE There is not enough memory
 * available.
 * @brief  
 *
 *
 * Creates multicast group.
 * This function creates the multicast group on basis of input 
 * parameters provided and returns a group handle.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_group_add (
        adk_module_handle_t ipv6Handle,
        adk_ipv6_multicast_group_t *mcGrp_p,
        adk_flow_handle_t *mcGrpHdl_p)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_flow_info_t *ipv6FlowInfo_p = NULL;
    adk_ipv6_addr_prefix_t srcIpv6Addr;
    adk_dp_status_t dpRetVal = ADK_SUCCESS;
    adk_flow_handle_t ipv6MgrpHandle;   
    adk_ipv6_ingress_mgroup_info_t *ipv6IngressMgroup_p  = NULL;
    adk_ipv6_dp_ingress_mgroup_entry_t ingress_mgroup_entry;
    uint16_t index = 0;
    adk_ipv6_dp_ingress_mgroup_table_t *ingress_mgroup_tbl = NULL ; 
    adk_handle_t mcast_index;
    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        goto adk_ipv6_multicast_group_add_return;
    }

    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid module handle \n");
       goto adk_ipv6_multicast_group_add_return;
    }
    
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);
   /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
    
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if ((mcGrpHdl_p == NULL) || (mcGrp_p == NULL))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid argument\n");

        goto adk_ipv6_multicast_group_return;
    }
    
    /*! Check if valid interface handle given as input */
    /* TODO:Have to validate l2InterfaceHandle better using adk_eth_valid_if_handle(),
     * but since its defined in adk_eth.h, and this internally has dependency on other 
     * Eth DP .h files, couldn't use it right now.
     */
    if ((!mcGrp_p->l2InterfaceHandle))
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid L2 interface specified\n");
        goto adk_ipv6_multicast_group_return;
    }
      /*! Allocate the memory for the IPV6 flow handle */
    ADK_SYS_FLOW_HANDLE_ALLOC (ipv6Handle, ADK_IPV6_ING_MGROUP_OBJ,
        sizeof(adk_ipv6_flow_info_t),
        &ipv6MgrpHandle);
    if (ipv6MgrpHandle == (adk_flow_handle_t)NULL) 
    {    
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Failed to allocate memory "
            "for the IPV6 Mgroup handle\n");

        goto adk_ipv6_multicast_group_return;
    }    
    ADK_IPV6_LOG_INFO(ipv6Handle, "Successfully allocated  memory "
            "for IPV6 Mgroup handle: 0x%"ADKPRIh"\n", ipv6MgrpHandle);

    ipv6FlowInfo_p = (adk_ipv6_flow_info_t *)
                        ADK_SYS_FLOW_ATTR_PTR_GET (ipv6MgrpHandle);
    
    ipv6FlowInfo_p->type = ADK_IPV6_INGRESS_MGROUP_FLOW;
    ipv6IngressMgroup_p = &ipv6FlowInfo_p->ingMgroupFlow;

    rte_memcpy(srcIpv6Addr.Ipv6Addr.addr, mcGrp_p->srcAddr.addr, sizeof(adk_ipv6_addr_t));
    srcIpv6Addr.prefixLength = ADK_IPV6_PDU_ADDR_LENGTH; 
    if (TRUE == mcGrp_p->useSrcAddr)
    {
	    if (!adk_ipv6_address_validate(ipv6Handle, srcIpv6Addr))
	    {
		    retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
		    ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid Src IPv6 address specified\n");
		    goto adk_ipv6_multicast_group_return;
	    }
    }
 
   /*Check for valid multicast group action */
   if ((mcGrp_p->mcGrpAction < ADK_IPV6_MCGROUP_ACTION_ISLOCAL) ||
	               (mcGrp_p->mcGrpAction >= ADK_IPV6_MCGROUP_ACTION_MAX))

   {
	   retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
	   ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid Action specified\n");
	   goto adk_ipv6_multicast_group_return;
   }

  ingress_mgroup_tbl = &ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_ingress_mgroup_table;

  memset(&ingress_mgroup_entry, 0x0, sizeof(adk_ipv6_dp_ingress_mgroup_entry_t));
  dpRetVal = adk_ipv6_cp_ingress_mgroup_entry_lookup(ipv6Handle, ingress_mgroup_tbl, mcGrp_p, &mcast_index);
  if (ADK_DP_E_ERROR == dpRetVal)
  {
  	ADK_IPV6_LOG_INFO(ipv6Handle,"Root node not found, as its a first entry\n");
  }
  else
  {
         if(ingress_mgroup_tbl->tbl[(size_t)mcast_index].is_enabled)
         {   
		 ADK_SYS_FLOWID_SET(ipv6MgrpHandle, (size_t)mcast_index);
		 rte_memcpy((uint8_t *)&ipv6IngressMgroup_p->mGrpInfo,
		        (uint8_t *)mcGrp_p,
		        sizeof (adk_ipv6_multicast_group_t));
                 *mcGrpHdl_p = ipv6MgrpHandle;
                 ingress_mgroup_tbl->tbl[(size_t)mcast_index].ref_count ++;
                 ADK_IPV6_LOG_INFO(ipv6Handle,"Multicast Entry exists, just incrementing ref_count\n");
                 goto adk_ipv6_multicast_group_return;
         }
         else
         {
                 ADK_IPV6_LOG_INFO(ipv6Handle,"Multicast Entry not found, Adding it\n");
         }
  }

   memset(&ingress_mgroup_entry, 0x0, sizeof(adk_ipv6_dp_ingress_mgroup_entry_t));
   ingress_mgroup_entry.action = mcGrp_p->mcGrpAction; 
   dpRetVal = adk_ipv6_dp_ingress_mgroup_entry_add(ipv6Handle, mcGrp_p, &ingress_mgroup_entry, &index);
   if(dpRetVal != ADK_SUCCESS)
   {
        retValue1 = ADK_E_INTERNAL_ERROR;
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1,
            "Failed to create mgroup entry LPM tree entry "
            "error code:0x%x\n",dpRetVal);

        goto adk_ipv6_multicast_group_return;
   }
   ingress_mgroup_tbl->tbl[(size_t)index].ref_count ++;
   ADK_SYS_FLOWID_SET(ipv6MgrpHandle, index);
   /*! Copy the required parameters from the rules to flow */
   rte_memcpy((uint8_t *)&ipv6IngressMgroup_p->mGrpInfo,
		   (uint8_t *)mcGrp_p,
		   sizeof (adk_ipv6_multicast_group_t));

   *mcGrpHdl_p = ipv6MgrpHandle;
   /*! Release module lock */
   ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
   ADK_RETURN(retValue1);

   adk_ipv6_multicast_group_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

   adk_ipv6_multicast_group_add_return:

   _lock_release_failed:
   _lock_acquire_failed:
    ADK_RETURN(retValue1);
    
}


/*!
 * =============================================================
 * adk_ipv6_multicast_group_del
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] mcGrpHdl  multicast group handle.
 *
 * @retval ADK_SUCCESS Successfully multicast group is deleted.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * Deletes multicast group.
 *
 * This function deletes the created multicast group on basis 
 * of input parameter provided. 
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_group_del (
        adk_module_handle_t ipv6Handle,
        adk_flow_handle_t mcGrpHdl)
{
    adk_status_t retValue1 = ADK_SUCCESS;
    adk_sys_handle_t sys_handle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_dp_status_t dpRetVal = ADK_SUCCESS;
    adk_ipv6_ingress_mgroup_info_t *ipv6MgrpInfo = NULL;
    adk_ipv6_flow_info_t *ipv6FlowInfo_p = NULL;
    uint16_t index = 0;
    adk_ipv6_dp_ingress_mgroup_table_t *ingress_mgroup_tbl = NULL ;
    adk_handle_t mcast_index;
    if (!ipv6Handle)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        goto adk_ipv6_multicast_group_add_return;
    }

    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
       retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
       ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid module handle \n");
       goto adk_ipv6_multicast_group_add_return;
    }
    
    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);
   /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue1);
    
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (mcGrpHdl == NULL)
    {
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Invalid argument\n");

	goto adk_ipv6_multicast_group_return;
    }

    ipv6FlowInfo_p = (adk_ipv6_flow_info_t *)
	    ADK_SYS_FLOW_ATTR_PTR_GET (mcGrpHdl);
    if (ipv6FlowInfo_p == NULL)
    {
	    retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
	    ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Ingress Multicast info is NULL\n");
	    goto adk_ipv6_multicast_group_return;
    }

    ipv6MgrpInfo = &ipv6FlowInfo_p->ingMgroupFlow;

    if (ipv6MgrpInfo == NULL)
    {   
        retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1, "Incorrect Multicast Group handle\n");

        goto adk_ipv6_multicast_group_return;
    }
    index = ADK_SYS_FLOWID_GET (mcGrpHdl);
    
    ingress_mgroup_tbl = &ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_ingress_mgroup_table;
    dpRetVal = adk_ipv6_cp_ingress_mgroup_entry_lookup(ipv6Handle, ingress_mgroup_tbl, &ipv6MgrpInfo->mGrpInfo, &mcast_index);
    if (ADK_DP_E_ERROR == dpRetVal)
    { 
        /*should not come here*/
        ADK_IPV6_LOG_ERROR(ipv6Handle,dpRetVal,"Root node not found, as its a first entry\n");
        goto adk_ipv6_multicast_group_return;
    }
    else
    {
	  if(ingress_mgroup_tbl->tbl[(size_t)mcast_index].is_enabled)
	  {
		  if(ingress_mgroup_tbl->tbl[(size_t)mcast_index].ref_count > 1)
		  {   
			ingress_mgroup_tbl->tbl[(size_t)mcast_index].ref_count --;
                        ADK_SYS_FLOW_HANDLE_FREE (mcGrpHdl);
			ADK_IPV6_LOG_INFO(ipv6Handle,"Multiple multicast Entry exists, deleting handle and just decrementing ref_count\n");
			goto adk_ipv6_multicast_group_return;
		  }
		  else
		  {
			ADK_IPV6_LOG_INFO(ipv6Handle,"Deleting the multicast Entry found\n");
			goto adk_ipv6_multicast_group_rem;
		  }
	  }
         else
         {
             retValue1 = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
             ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1,"Multicast Entry not found\n");
             goto adk_ipv6_multicast_group_return;
         }
    }
   adk_ipv6_multicast_group_rem:
    dpRetVal = adk_ipv6_dp_ingress_mgroup_entry_rem(ipv6Handle, &ipv6MgrpInfo->mGrpInfo, (uint16_t)index);
    if(dpRetVal != ADK_SUCCESS)
    {
        retValue1 = ADK_E_INTERNAL_ERROR;
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue1,
            "Failed to remove mgroup entry from LPM tree "
            "error code:0x%x\n",dpRetVal);

        goto adk_ipv6_multicast_group_return;
    }
    ingress_mgroup_tbl->tbl[(size_t)index].ref_count --;
    ADK_SYS_FLOW_HANDLE_FREE (mcGrpHdl);
    /*! Release module lock */
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
    ADK_RETURN(retValue1);

   adk_ipv6_multicast_group_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue1);
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

   adk_ipv6_multicast_group_add_return:

   _lock_release_failed:
   _lock_acquire_failed:
    ADK_RETURN(retValue1);
}

/*!
 * =============================================================
 * adk_ipv6_multicast_flow_add
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] mcFlow_p poniter to multicast flow information. 
 * @param[out] mcFlowHdl_p pointer to multicast flow handle. 
 * This is set by the API.
 *
 * @retval ADK_SUCCESS Successfully multicast flow is added.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 * Adds multicast flow to a multicast group.
 *
 * This function adds a multicast flow to already created 
 * multicast group on basis of input parameters provided.
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_flow_add (
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
        const adk_ipv6_multicast_flow_t __attribute__((unused)) *mcFlow_p,
        adk_flow_handle_t __attribute__((unused)) *mcFlowHdl_p)
{
    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * adk_ipv6_multicast_flow_del
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] mcFlowHdl multicast flow handle. 
 *
 * @retval ADK_SUCCESS Successfully multicast flow is deleted.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 * This API is not supported
 * Deletes a multicast flow from a multicast group.
 *
 * This function deletes a multicast flow from a 
 * multicast group on basis of input parameters provided.
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_flow_del (
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
        adk_flow_handle_t __attribute__((unused)) mcFlowHdl)
{
    return ADK_SUCCESS;
}


/*!
 * =============================================================
 * adk_ipv6_multicast_group_stat_get
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] mcGrpHdl multicast group handle.
 * @param[out]  mcGrpStat_p pointer to multicast group statistics.
 *
 * @retval ADK_SUCCESS Successfully read the multicast group
 * statistics.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 * Gets statistics for a multicast group
 *
 * This function gets statistics for a multicast group identified
 * by mcGrphdl parameter.
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_group_stat_get (
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
        adk_flow_handle_t __attribute__((unused)) mcGrpHdl,
        adk_ipv6_multicast_group_stat_t __attribute__((unused)) *mcGrpStat_p)
{
    return ADK_SUCCESS;
}    /* -----  end of function adk_ipv6_multicast_group_stat_get  ----- */

/*!
 * =============================================================
 * adk_ipv6_multicast_group_statistics_print
 * @param[in]    ipv6Handle - IPv6 module handle.
 * @param[in]    mcGrpHdl - multicast group handle.
 *  
 *
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 *
 * @brief  
 * This API prints the multicast group statistics 
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t 
adk_ipv6_multicast_group_statistics_print (adk_module_handle_t ipv6Handle, adk_flow_handle_t mcGrpHdl)
{
    adk_status_t                       retValue = ADK_SUCCESS;
    adk_ipv6_multicast_group_stat_t    ipv6MulticastGroupStats;

    ADK_MEMSET(&ipv6MulticastGroupStats, 0, sizeof(adk_ipv6_multicast_group_stat_t));

    retValue = adk_ipv6_multicast_group_stat_get(ipv6Handle, mcGrpHdl, &ipv6MulticastGroupStats);
    if (retValue != ADK_SUCCESS)
        return retValue;

    ADK_INFO_WRITE("\nIPv6 Module Multicast Group Statistics\n" );
    ADK_INFO_WRITE("Total Ingress Bytes Received               : %d\n", ipv6MulticastGroupStats.byteLLIRcvCount);
    ADK_INFO_WRITE("Total Ingress Packets Received             : %d\n", ipv6MulticastGroupStats.pktLLIRcvCount);
    ADK_INFO_WRITE("Total Egress Bytes Received                : %d\n", ipv6MulticastGroupStats.byteULIRcvCount);
    ADK_INFO_WRITE("Total Egress Packets Received              : %d\n", ipv6MulticastGroupStats.pktULIRcvCount);
    ADK_INFO_WRITE("Total Forwarded Bytes                      : %d\n", ipv6MulticastGroupStats.byteFwdCount);
    ADK_INFO_WRITE("Total Forwarded Packets                    : %d\n", ipv6MulticastGroupStats.pktFwdCount);
    ADK_INFO_WRITE("Total Bytes Sent to CPU using PH           : %d\n", ipv6MulticastGroupStats.byteSendToPHCount);
    ADK_INFO_WRITE("Total Packets Sent to CPU using PH         : %d\n", ipv6MulticastGroupStats.pktSendToPHCount);
    return ADK_SUCCESS;
}

#ifdef ADK_DEBUG
/*!
 * =============================================================
 * adk_ipv6_multicast_group_stat_reset
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[in] mcGrpHdl multicast group handle.
 *
 * @retval ADK_SUCCESS Successfully reset the multicast group
 * statistics.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 * Reset statistics for a multicast group
 *
 * This function resets statistics for a multicast group identified
 * by mcGrphdl parameter.
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_group_stat_reset (
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
        adk_flow_handle_t __attribute__((unused)) mcGrpHdl)
{
    return ADK_SUCCESS;
}
#endif

/*!
 * =============================================================
 * adk_ipv6_multicast_rpf_check_enable
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully multicast rpf check is 
 * enabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 * It enables RPF check for all multicast packets
 *
 * This function enables RPF check for all multicast packets
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_rpf_check_enable (
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
    	adk_ipv6_vrd_handle_t __attribute__((unused)) ipv6VrdHandle)
{
    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * adk_ipv6_multicast_rpf_check_disable
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully multicast rpf check is 
 * disabled.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 * @brief  
 *
 * This API is not supported
 * It disables RPF check for all multicast packets
 *
 * This function disables RPF check for all multicast packets
 *
 * NOTE: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_multicast_rpf_check_disable (
        adk_module_handle_t __attribute__((unused)) ipv6Handle,
        adk_ipv6_vrd_handle_t  __attribute__((unused)) ipv6VrdHandle
       )
{
    return ADK_SUCCESS;
}


 /*!
 * end addtogroup _ipv6Functions  IPv6 Functions
 * @}
 */
