/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_packet_capture.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief
 *
 * This file contains APIs to manage IPv6 interfaces. This includes creation,
 * deletion of interfaces, setting their attributes, getting or printing
 * statistics for the interfaces, etc.
 *
 ***********************************************************************/

/* Infrastructure header files */
#include "adk_include.h"
#include "adk_hos.h"
#include "adk_ph_dev.h"
#include "adk_utils_tree_update.h"
#include "adk_ethernet_export.h"

/* IPV6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

#define ADK_CONVERT_TO_V6_IF_HANDLE(__DATA) ((adk_interface_handle_t)__DATA)
#define ADK_CONVERT_TO_V6_IF_HANDLE(__DATA) ((adk_interface_handle_t)__DATA)
/* IPV6 internal header files */
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_internal.h"


/**
 *=============================================================
 * adk_ipv6_packet_capture_start
 *
 *
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in]  ipv6IfPktCapInfo  - pointer to adk_ipv6_pktcap_info_t data
 * structure.
 *
 * @param[out] ipv6PktCapHandle - pointer to the IPv6 capture point handle
 * which gets created by this API.
 *
 * @retval ADK_SUCCESS Successfully created ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @retval ADK_E_MEM_UNAVAILABLE - failed to allocate memory
 * for interface handle.
 *
 * @retval ADK_E_LEARN_FAILED - failed to update MPP tree
 * entry.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This API is not supported
 *
 * This function start an IPv6 Packet capturing at th interface level.
 * This function returns packet capture handle which can be used to
 * get status, enable and stop the packet capture.  
 *
 * Note: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_packet_capture_start(
    adk_module_handle_t           ipv6Handle,
    const adk_ipv6_pktcap_info_t  *ipv6PktCapInfo_p,
    adk_flow_handle_t             *ipv6PktCapHandle_p)
{
	adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    adk_flow_handle_t ipv6PktCapHandle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_pktcap_internal_info_t *ipv6FlowInfo_p = NULL;
    ipv6_interface_info_t 	*if_pvt_info = NULL;
    adk_ipv6_dp_pkt_capture_info_t pkt_capture_info;
    bool at_ingress;
#ifdef ADK_HOS_OLD_IMPL
    adk_module_handle_t eth_mod_handle;
#endif
    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }

    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }
 
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_packet_capture_create_return;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    
    if ((ipv6PktCapHandle_p == NULL) || (ipv6PktCapInfo_p == NULL))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);

        goto adk_ipv6_packet_capture_create_return;
    }

    if(!adk_ipv6_interface_valid_handle(ipv6PktCapInfo_p->ipv6IfHandle))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument, Interface handle is NULL \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_packet_capture_create_return;
    }

    /*! Allocate the memory for the IPV6 flow handle */
    ADK_SYS_FLOW_HANDLE_ALLOC (ipv6Handle, ADK_IPV6_PKT_CAP_OBJ,
        sizeof(adk_ipv6_pktcap_internal_info_t),
        &ipv6PktCapHandle);
    if (ipv6PktCapHandle == (adk_flow_handle_t)NULL) 
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Failed to allocate memory "
            "for the IPV6 flow handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);

        goto adk_ipv6_packet_capture_create_return;
    }
    ADK_LOG_INFO(sysHandle, ipv6Handle, "Successfully allocated  memory "
            "for IPV6 flow handle: 0x%"ADKPRIh" \n",ipv6PktCapHandle);

    ipv6FlowInfo_p = (adk_ipv6_pktcap_internal_info_t *)
                        ADK_SYS_FLOW_ATTR_PTR_GET (ipv6PktCapHandle);
    ipv6FlowInfo_p->cookie = ADK_IPV6_PACKET_CAPTURE_COOKIE;

    ADK_MEMCPY( &(ipv6FlowInfo_p->flowInfo) , ipv6PktCapInfo_p, sizeof(adk_ipv6_pktcap_info_t));

    /* update the fast-path table */
    pkt_capture_info.is_enabled = TRUE;
    pkt_capture_info.context = ipv6PktCapInfo_p->context;
    pkt_capture_info.devId = ipv6PktCapInfo_p->phAttr.devId;
    pkt_capture_info.queueId = ipv6PktCapInfo_p->phAttr.queueId;
    pkt_capture_info.cbFnId = ipv6PktCapInfo_p->phAttr.cbFnId;

    /* TBD: update the DP/CP interface details */
    if_pvt_info = (ipv6_interface_info_t *)
                          ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6PktCapInfo_p->ipv6IfHandle);

    if ( ipv6PktCapInfo_p->direction == ADK_IPV6_DIR_INGRESS)
    {
        if_pvt_info->enableIngressPacketCapture = TRUE;
    
        retValue = adk_ipv6_dp_ing_pkt_capture_set(ipv6Handle,
                ADK_SYS_INTERFACE_ID_GET( ipv6PktCapInfo_p->ipv6IfHandle),
                &pkt_capture_info);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to set the fast-path table %d\n",retValue);
            goto adk_ipv6_packet_capture_if_rule_failure_return;
        }
        at_ingress = TRUE;
    }
    else
    {
        if_pvt_info->enableEgressPacketCapture = TRUE;

        retValue = adk_ipv6_dp_eg_pkt_capture_set(ipv6Handle,
                ADK_SYS_INTERFACE_ID_GET( ipv6PktCapInfo_p->ipv6IfHandle),
                &pkt_capture_info);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to set the fast-path table %d\n",retValue);
            goto adk_ipv6_packet_capture_if_rule_failure_return;
        }
        at_ingress = FALSE;
    }

#ifdef ADK_HOS_OLD_IMPL
    if(ipv6ModuleInfo_p->enableHwCap && if_pvt_info->ifType == IPV6_INTERFACE_TYPE_ETH)
    {
        retValue = adk_module_handle_get(sysHandle, ADK_ETHERNET_MODULE_ID, 
                NULL, &eth_mod_handle);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to get the Ethernet module handle %d\n",retValue);
            goto adk_ipv6_packet_capture_if_rule_failure_return;
        }

        retValue = adk_eth_intf_pcap_config(eth_mod_handle, 
                if_pvt_info->if_info_u.regularIf.layer2InterfaceHandle, 
                TRUE, at_ingress, ipv6ModuleInfo_p->mirror_id, TRUE);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to configure packet capture %d\n",retValue);
            goto adk_ipv6_packet_capture_if_rule_failure_return;
        }
    }
#endif

    *ipv6PktCapHandle_p = ipv6PktCapHandle;
    
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);
    return retValue;

adk_ipv6_packet_capture_if_rule_failure_return:
    ADK_SYS_FLOW_HANDLE_FREE (ipv6PktCapHandle);

adk_ipv6_packet_capture_create_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

adk_ipv6_packet_capture_null_handle_return:

_lock_release_failed:
_lock_acquire_failed:
    return retValue;

}

/**
 *=============================================================
 * adk_ipv6_packet_capture_stop
 *
 *
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6PktCapHandle - pointer to the IPv6 capture point handle
 * which gets created by this API.
 *
 * @retval ADK_SUCCESS Successfully created ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @retval ADK_E_MEM_UNAVAILABLE - failed to allocate memory
 * for interface handle.
 *
 * @retval ADK_E_LEARN_FAILED - failed to update MPP tree
 * entry.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This API is not supported
 *
 * This function stops an IPv6 Packet capturing at th interface level.
 *   
 * Note: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_packet_capture_stop(
    adk_module_handle_t  ipv6Handle,
    adk_flow_handle_t    ipv6PktCapHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    adk_ipv6_pktcap_internal_info_t *ipv6FlowInfo_p = NULL;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    ipv6_interface_info_t 	     *if_pvt_info = NULL;
    adk_ipv6_dp_pkt_capture_info_t pkt_capture_info;
    bool at_ingress;
#ifdef ADK_HOS_OLD_IMPL
    adk_module_handle_t eth_mod_handle;
#endif
    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }

    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }
 
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_packet_capture_stop_return;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    
    if(!adk_ipv6_if_pktcap_valid_handle(ipv6PktCapHandle))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument, packet capture handle is NULL \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_packet_capture_stop_return;
    }

    ipv6FlowInfo_p = (adk_ipv6_pktcap_internal_info_t *)
                        ADK_SYS_FLOW_ATTR_PTR_GET (ipv6PktCapHandle);

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(ipv6FlowInfo_p->flowInfo.ipv6IfHandle);

    /* update the fast-path table */
    ADK_MEMSET(&pkt_capture_info, 0, sizeof(pkt_capture_info));

    if ( ipv6FlowInfo_p->flowInfo.direction == ADK_IPV6_DIR_INGRESS)
    {
        if_pvt_info->enableIngressPacketCapture = FALSE;

        retValue = adk_ipv6_dp_ing_pkt_capture_set(ipv6Handle,
                ADK_SYS_INTERFACE_ID_GET(ipv6FlowInfo_p->flowInfo.ipv6IfHandle),
                &pkt_capture_info);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to set the fast-path table %d\n",retValue);
            goto adk_ipv6_packet_capture_stop_return;
        }
        at_ingress = TRUE;
    }
    else
    {
        if_pvt_info->enableEgressPacketCapture = FALSE;

        retValue = adk_ipv6_dp_eg_pkt_capture_set(ipv6Handle,
                ADK_SYS_INTERFACE_ID_GET(ipv6FlowInfo_p->flowInfo.ipv6IfHandle),
                &pkt_capture_info);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to set the fast-path table %d\n",retValue);
            goto adk_ipv6_packet_capture_stop_return;
        }
        at_ingress = FALSE;
    }

#ifdef ADK_HOS_OLD_IMPL
    if(ipv6ModuleInfo_p->enableHwCap && if_pvt_info->ifType == IPV6_INTERFACE_TYPE_ETH)
    {
        retValue = adk_module_handle_get(sysHandle, ADK_ETHERNET_MODULE_ID, 
                NULL, &eth_mod_handle);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to get the Ethernet module handle %d\n",retValue);
            goto adk_ipv6_packet_capture_stop_return;
        }

        retValue = adk_eth_intf_pcap_config(eth_mod_handle, 
                if_pvt_info->if_info_u.regularIf.layer2InterfaceHandle,
                TRUE, at_ingress, ipv6ModuleInfo_p->mirror_id, FALSE);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle,
                    "Failed to configure packet capture %d\n",retValue);
            goto adk_ipv6_packet_capture_stop_return;
        }
    }
#endif

    ADK_SYS_FLOW_HANDLE_FREE (ipv6PktCapHandle);

adk_ipv6_packet_capture_stop_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

adk_ipv6_packet_capture_null_handle_return:

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}


/**
 *=============================================================
 * adk_ipv6_packet_capture_status
 *
 *
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in] ipv6PktCapHandle - pointer to the IPv6 capture point handle
 * which gets created by this API.
 *
 * @param[out] status - pointer to the packet capture status
 *
 * @retval ADK_SUCCESS Successfully created ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This API is not supported
 *
 * This API gets the status of a capture point. The status tells us
 * if the capture has been autonomously disabled by ADK or not. This
 * automatic disabling by ADK is done when packets cannot be buffered
 * for delivery to application due to lack of space in the buffer area.
 * Once a capture point enters this state, it remains in this state
 * until application re-enables the packet capture. In this state, i.e.
 * when packet capture has been disabled, then even though the traffic
 * is being captured, it is not being delivered to CPU.
 *
 * Note that, a disabled capture point is not same as stopped capture
 * point. In disabled state, the capture point still exist. For capture
 * points in disabled state, application can either stop the capture
 * (thereby, remove/delete the capture point) or re-enable the capturing
 * using the adk_ip_packet_capture_enable() API.
 *
 * From the time the packet capture was disabled by ADK autonomously
 * to the time when it is enabled again by application, there may have
 * been packets which were captured but were not delivered to application.
 * These packets are not buffered by ADK but instead dropped. And thus,
 * there may exist discontinuity in the captured packet flow during
 * this transition.
 *   
 * Note: This API is not yet supported
 * =============================================================
 */
adk_status_t
adk_ipv6_packet_capture_status(
    adk_module_handle_t ipv6Handle,
    adk_flow_handle_t   ipv6PktCapHandle,
    adk_ipv6_pktcap_status_t  *status)
{

    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    adk_ipv6_pktcap_internal_info_t *ipv6FlowInfo_p = NULL;
    uint16_t if_id;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }

    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }
 
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_packet_capture_status_return;
    }
    
    if(!adk_ipv6_if_pktcap_valid_handle(ipv6PktCapHandle))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument, packet capture handle is NULL \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_packet_capture_status_return;
    }

    ipv6FlowInfo_p = (adk_ipv6_pktcap_internal_info_t *)
                        ADK_SYS_FLOW_ATTR_PTR_GET (ipv6PktCapHandle);

    if_id = ADK_SYS_INTERFACE_ID_GET(ipv6FlowInfo_p->flowInfo.ipv6IfHandle);

    if(ipv6FlowInfo_p->flowInfo.direction == ADK_IPV6_DIR_INGRESS)
    {
        retValue = adk_ipv6_dp_ing_pkt_capture_status_get(ipv6Handle,
                if_id, &status);
    }
    else
    {
        retValue = adk_ipv6_dp_eg_pkt_capture_status_get(ipv6Handle,
                if_id, &status);
    }
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR( sysHandle, ipv6Handle,
                           "Failed to get the packet capture status \n");
        goto adk_ipv6_packet_capture_status_return;
    }

adk_ipv6_packet_capture_status_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

adk_ipv6_packet_capture_null_handle_return:

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/**
 *=============================================================
 * adk_ipv6_packet_capture_enable
 *
 *
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[out] status - pointer to adk_ipv6_pktcap_status_t structure
 * @retval ADK_SUCCESS Successfully created ADK IPv6 interface.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or interface attribute pointer might be NULL.
 *
 * @retval  ADK_IP_E_INTERFACE_TABLE_ACCESS_FAILURE - Failed to IP interface
 * hardware table.
 *
 * @brief
 *
 * This API is not supported
 *
 * This API re-enables a capture points which has been disabled
 * autonomously by ADK. Please refer to the description of
 * adk_ip_packet_capture_status_get() API for more details on this.
 *   
 * Note: This API is not yet supported
 * =============================================================
 */

adk_status_t
adk_ipv6_packet_capture_enable(
    adk_module_handle_t ipv6Handle,
    adk_flow_handle_t ipv6PktCapHandle)
{

    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    adk_ipv6_pktcap_internal_info_t *ipv6FlowInfo_p = NULL;
    uint16_t if_id;

    if (!ipv6Handle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Module handle is NULL \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }

    /* validate IPv6 Module Handle*/
    if(!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Invalid module handle \n");
        goto adk_ipv6_packet_capture_null_handle_return;
    }
 
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_packet_capture_enable_return;
    }
    
    if(!adk_ipv6_if_pktcap_valid_handle(ipv6PktCapHandle))
    {
        ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid argument, packet capture handle is NULL \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_ipv6_packet_capture_enable_return;
    }

    ipv6FlowInfo_p = (adk_ipv6_pktcap_internal_info_t *)
                        ADK_SYS_FLOW_ATTR_PTR_GET (ipv6PktCapHandle);

    if_id = ADK_SYS_INTERFACE_ID_GET(ipv6FlowInfo_p->flowInfo.ipv6IfHandle);

    if(ipv6FlowInfo_p->flowInfo.direction == ADK_IPV6_DIR_INGRESS)
    {
        retValue = adk_ipv6_dp_ing_pkt_capture_status_set(ipv6Handle,
                if_id);
    }
    else
    {
        retValue = adk_ipv6_dp_eg_pkt_capture_status_set(ipv6Handle,
                if_id);
    }
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR( sysHandle, ipv6Handle,
                           "Failed to enable packet capture \n");
        goto adk_ipv6_packet_capture_enable_return;
    }

adk_ipv6_packet_capture_enable_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

adk_ipv6_packet_capture_null_handle_return:

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

 /*!
 * end addtogroup _ipv6Functions  IPv6 Module
 * @}
 */
