/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_pbr.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief  
 *
 * This file contains APIs to manage Policy Based Routing (PBR) policies.
 *
 ***********************************************************************/


/* Infrastructure header files */
#include "adk_include.h"
#include "adk_ph_dev.h"
#include "adk_utils_tree_update.h"

/* IPV6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

/* IPV6 internal header files */
#include "adk_ipv6_internal.h"
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"

#if ADK_IPV6_PBR_ENABLE
#define IPV6_MAX_PBR_RULE_PRIORITY 0xFFFF

static adk_status_t adk_ipv6_pbr_dpdk_sw_rule (
                           adk_module_handle_t ipv6Handle,
                           ipv6_pbr_rule_set_info_t *policy_info_p,
                           uint32_t precedence,
                           uint32_t pbrRuleStatsId,
                           const adk_ipv6_pbr_rule_t *pbrRules_p);
#endif

/*!
 * =============================================================
 * adk_ip_pbr_policy_set_create
 *
 * @param[in] pv6Handle                  IPv6 module handle
 *
 * @param[out] pbrRuleSetHandle_p     Pointer to the PBR policy-set
 * handle.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE       Invalid IPv6 module handle specified
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid.
 *
 * @retval ADK_E_MEM_UNAVAILABLE - Failed to allocate
 * memory to store policy state. This memory comes from pbr policy
 * namespace.
 *
 * @brief
 *
 * This API creates a PBR policy set to be used across multiple interfaces
*/

adk_status_t adk_ipv6_pbr_policy_set_create (
    adk_module_handle_t ipv6Handle,
    adk_ipv6_pbr_policy_set_handle_t *pbrRuleSetHandle_p,
    uint32_t *pbr_policy_set_id)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_info_t	*ipv6ModuleInfo_p=NULL;
    adk_sys_handle_t sysHandle ;
    uint32_t pbrPolicySetId = 0, ret;
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;
    adk_acl_info_t cp_list_info={0};
    adk_ipv6_pbr_rule_t pbrRule;

    ipv6ModuleInfo_p = (adk_ipv6_info_t *) ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module info is NULL\n");
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    if (!pbrRuleSetHandle_p)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid NULL input parameter \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }


    pbrPolicySetId = adk_rm_resource_reserve(
                               ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle);

    if (pbrPolicySetId == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_FATAL_ERROR(sysHandle,ipv6Handle, 
                           "PBR policy set allocation failed\n");
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }

    policy_info_p = (ipv6_pbr_rule_set_info_t *)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_PBR_RULESETS_OBJ, 1, sizeof(ipv6_pbr_rule_set_info_t));

    if (policy_info_p == NULL)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_FATAL_ERROR(sysHandle,ipv6Handle, 
                           "Memory allocation failed\n");
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }

    /* Store the policy set information in the internal data structure */
    policy_info_p->cookie =  ADK_IPV6_PBR_RULESET_COOKIE;
    policy_info_p->ipv6_pbr_rule_set_id = pbrPolicySetId;
    /* Create a shadow copy as well for RCU */
    policy_info_p->pbr_tree_index = 0;
    policy_info_p->pbr_tree[0] = adk_ipv6_pbr_create(sysHandle, ipv6Handle, pbrPolicySetId, 0);
    policy_info_p->pbr_tree[1] = adk_ipv6_pbr_create(sysHandle, ipv6Handle, pbrPolicySetId, 1);
    if((policy_info_p->pbr_tree[0] == NULL) ||
       (policy_info_p->pbr_tree[1] == NULL))
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_FATAL_ERROR(sysHandle,ipv6Handle, 
                           "PBR Rule create failed\n");
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }
    /*Create CP list*/
    cp_list_info.user = ADK_ACL_IPv6;
    if(adk_acl_create(&(policy_info_p->cp_list),
                                &cp_list_info) != ADK_SUCCESS)
    {
        ADK_LOG_FATAL_ERROR(sysHandle,ipv6Handle, 
                           "Could not allocate CP list for IPv6 ACL Ruleset!\n");
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }

    *pbrRuleSetHandle_p = (adk_ipv6_pbr_policy_set_handle_t) policy_info_p;
    ipv6ModuleInfo_p->policySetHandles[pbrPolicySetId-1] =
        *pbrRuleSetHandle_p;

    ipv6ModuleInfo_p->adk_ipv6_dp_tables->policySetHandles[pbrPolicySetId-1] = 
        *pbrRuleSetHandle_p;
    /*
     * Add default PBR rule, to prevent crash in SW ACL
     * when try to do ACL lookup before commit
     */
    memset(&pbrRule, 0, sizeof(pbrRule));
    retValue = adk_ipv6_pbr_dpdk_sw_rule(ipv6Handle, policy_info_p,
            IPV6_MAX_PBR_RULE_PRIORITY, 0/*pbrPolicySetId*/,
            &pbrRule); 
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                "Failed to add the rule in dpdk s/w\n");
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }
    /*Update both copies*/
    ret = adk_ipv6_pbr_ruleset_commit(sysHandle, ipv6Handle, policy_info_p->pbr_tree[0]);
    if(!ret)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                "Error: Could not commit PBR ACL rules...\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }

    ret = adk_ipv6_pbr_ruleset_commit(sysHandle, ipv6Handle, policy_info_p->pbr_tree[1]);
    if(!ret)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                "Error: Could not commit PBR ACL rules...\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        goto _adk_ipv6_pbr_policy_set_create_exit;
    }
    *pbr_policy_set_id = pbrPolicySetId;

_adk_ipv6_pbr_policy_set_create_exit:

    if (retValue != ADK_SUCCESS)
    {
        if (pbrPolicySetId != 0)
        {
            adk_rm_resource_free(ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle,
                              pbrPolicySetId);
        }

        if (policy_info_p != NULL)
        {
            if(policy_info_p->pbr_tree[0])
                adk_ipv6_pbr_ruleset_free(sysHandle, ipv6Handle, policy_info_p->pbr_tree[0]);

            if(policy_info_p->pbr_tree[1])
                adk_ipv6_pbr_ruleset_free(sysHandle, ipv6Handle, policy_info_p->pbr_tree[1]);

            if(policy_info_p->cp_list)
                adk_acl_free(policy_info_p->cp_list);
            ADK_NVM_FREE(policy_info_p);
        }
    }

#endif

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ip_pbr_policy_set_delete
 *
 * @param[in] ipv6Handle                  IPv6 module handle
 *
 * @param[in] pbrRuleSetHandle            PBR Policy-Set handle to be deleted.
 *
 * @retval ADK_SUCCESS                    Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE           Invalid IPv6 module handle specified
 *
 * @retval ADK_E_INVALID_PARAMETER        Invalid parameter e.g. invalid
 * PBR policy-set handle.
 *
 * @brief
 *
 * This API deletes an existing PBR policy set. Note that, before
 * deleting the PBR policy set, all interfaces associated with it
 * must be detached.
 * ===================================================
*/
adk_status_t adk_ipv6_pbr_policy_set_delete(
    adk_module_handle_t ipv6Handle,
    adk_ipv6_pbr_policy_set_handle_t pbrRuleSetHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_info_t    *ipv6ModuleInfo_p=NULL;
    adk_sys_handle_t sysHandle;
    uint32_t pbrPolicySetId = 0;
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Module Info NULL or cookie mismatch \n");
        goto _adk_ipv6_pbr_policy_set_delete_exit;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle); 

    if (!adk_ipv6_valid_policy_set_handle(pbrRuleSetHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid NULL input parameter \n");
        goto _adk_ipv6_pbr_policy_set_delete_exit;
    }

    policy_info_p = (ipv6_pbr_rule_set_info_t *) pbrRuleSetHandle;
    pbrPolicySetId = policy_info_p->ipv6_pbr_rule_set_id;

    ipv6ModuleInfo_p->policySetHandles[pbrPolicySetId-1] = 0;
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->policySetHandles[pbrPolicySetId-1] = 0;

    if(policy_info_p->pbr_tree_index)
    {
        /* Free the shadow copy */
        adk_ipv6_pbr_ruleset_free( sysHandle, ipv6Handle, policy_info_p->pbr_tree[0]);
        /* Wait for readers to complete */
        adk_rcu_synchronize();
        /* Free the active copy */
        adk_ipv6_pbr_ruleset_free( sysHandle, ipv6Handle, policy_info_p->pbr_tree[1]);
    }
    else
    {
        /* Free the shadow copy */
        adk_ipv6_pbr_ruleset_free(sysHandle, ipv6Handle, policy_info_p->pbr_tree[1]);
        /* Wait for readers to complete */
        adk_rcu_synchronize();
        /* Free the active copy */
        adk_ipv6_pbr_ruleset_free(sysHandle, ipv6Handle, policy_info_p->pbr_tree[0]);
    }

    adk_rm_resource_free(ipv6ModuleInfo_p->ipv6PbrPolicySetIDHandle,
                           pbrPolicySetId);

    adk_acl_free(policy_info_p->cp_list);

    /* Clear the rule set information and free the internal data structure */
    policy_info_p->cookie = 0;
    ADK_NVM_FREE(policy_info_p);

_adk_ipv6_pbr_policy_set_delete_exit:
#endif
    ADK_RETURN(retValue);
}

/*!
* =============================================================
* adk_ipv6_pbr_configure_default_route_table
*
* @param[in] ipv6Handle IPV6 module handle. Returned at the
* time of IPV6 module's registration with the module manager.
*
* @param[in] defRtTbl - Default Route table Id
*
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_IPV6_E_INVALID_PARAMETER - Invalid input parameter.
* The pointer might be NULL.
*

 * @retval ADK_IPV6_E_INVALID_ROUTE_TABLE_ID - Invalid route table Id.
*
* @brief IPv6 API to program default route table Id into policy based
* roting tree.
*
* This API programs the default route table Id 
 * All packets which do not match to any specific policy 
 * configured by the application always match to default policy
*. Such packets are processed based on action configured
* for this default policy given by input parameter defRtTbl
* This defRtTbl  is used in below 2 cases:
* 1. when a route is not present in route table Id given by specific policy 
* 2. when packet doesn’t match the configured PBR policy.
*
* =============================================================
*/
adk_status_t
adk_ipv6_pbr_configure_default_route_table (
        adk_module_handle_t ipv6Handle,
        uint8_t  defRtTbl)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_sys_handle_t sysHandle;
    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_pbr_configure_default_route_table_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (defRtTbl > ADK_IPV6_MAX_ROUTING_TABLES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                " Invalid default route table ID %u\n", defRtTbl);
        goto adk_ipv6_pbr_configure_default_route_table_return;
    }
 
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->defaultRouteTable = defRtTbl;
    adk_sys_flow_cache_update_version(sysHandle);
    ADK_IPV6_LOG_INFO(ipv6Handle,"Successfully confgirued def route table : %u \n", 
            ipv6ModuleInfo_p->adk_ipv6_dp_tables->defaultRouteTable);

adk_ipv6_pbr_configure_default_route_table_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ipv6_pbr_add_policy
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    ruleIndex - location in the PBR table where
 * the rule needs to be inserted.
 *
 * @param[in]    pbrInfo_p - Pointer to adk_ipv6_pbr_info_t
 * data structure.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid.
 *
 * @retval ADK_E_MEM_UNAVAILABLE - Failed to allocate memory
 *
 * @brief
 * This API validates the input parameter and adds a new rule to the PBR
 * policy set associated with the specified IP interface at the specified policy
 * rule index. Once the rule is programmed successfully, it applies to
 * all IP interfaces that are associated with same PBR policy set. That is,
 * adding a rule for one IP interface, adds it for all IP interfaces associated
 * with same PBR policy set. This API does not perform check for redundant
 * policy rules (caller is expected to perform this check).  
 * =============================================================
 */
adk_status_t
adk_ipv6_pbr_add_policy (
        adk_module_handle_t  	ipv6Handle,
        uint32_t   	        ruleIndex,
        adk_ipv6_pbr_info_t  	*pbrInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_pbr_rule_handle_info_t *flow_info_p = NULL;
    adk_ipv6_dp_pbr_action_entry_t *pbr_action_p=NULL;
    uint32_t pbrRuleStatsId = 0, i;
    adk_sys_handle_t sysHandle;
    ipv6_pbr_rule_set_info_t   *policy_info_p = NULL;
    adk_acl_rule_t *cp_rule_p;
    adk_flow_handle_t  	ruleHandle_p = 0;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_pbr_entry_create_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (!pbrInfo_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Invalid NULL input parameter\n");
        goto adk_ipv6_pbr_entry_create_return;
    }
    if (ipv6ModuleInfo_p->maxIpv6PbrRules > ADK_IPV6_PBR_RULES_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue," Max Pbr Rules limit reached:%d\n",
                ADK_IPV6_PBR_RULES_MAX);
        goto adk_ipv6_pbr_entry_create_return;
    }

    if (pbrInfo_p->pbrAction.routeTableId > ADK_IPV6_MAX_ROUTING_TABLES)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                " Invalid route table ID %u\n", pbrInfo_p->pbrAction.routeTableId);
        goto adk_ipv6_pbr_entry_create_return;
    }
    /* Clearing to indicate no handle is allocated yet */
    ruleHandle_p = 0x0;

    policy_info_p = (ipv6_pbr_rule_set_info_t *) ipv6ModuleInfo_p->adk_ipv6_dp_tables->adk_ipv6_pbr_policy_set_handle_t;

    ADK_SYS_FLOW_HANDLE_ALLOC(ipv6Handle, ADK_IPV6_PBR_RULESETS_OBJ, sizeof(adk_ipv6_pbr_rule_handle_info_t),
                              &ruleHandle_p);
    if (ruleHandle_p == 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Failed to allocate memory for PBR rule\n");
        goto adk_ipv6_pbr_entry_create_return;
    }

    /* Allocating stats id for each rule*/
    pbrRuleStatsId = adk_rm_resource_reserve(ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle);
    if (pbrRuleStatsId == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Failed to allocate id for PBR rule\n");
        goto adk_ipv6_pbr_entry_create_return;
    }

    flow_info_p = (adk_ipv6_pbr_rule_handle_info_t *)
                  ADK_SYS_FLOW_ATTR_PTR_GET(ruleHandle_p);

    flow_info_p->cookie = ADK_IPV6_PBR_RULE_HANDLE_COOKIE;
    ADK_SYS_FLOWID_SET(ruleHandle_p, pbrRuleStatsId);

    flow_info_p->rule_index = ruleIndex;
    ADK_MEMCPY((void *) &(flow_info_p->rule),
               (const void *) &(pbrInfo_p->pbrRule),
               sizeof(adk_ipv6_pbr_rule_t));
    ADK_MEMCPY((void *) &(flow_info_p->action),
               (const void *) &(pbrInfo_p->pbrAction),
               sizeof(adk_ipv6_pbr_action_t));

    /*
     * Add rule to ADK CP rule list.
     * The rules will be added during the next commit.
     */
    cp_rule_p = &(flow_info_p->cp_rule);
    cp_rule_p->user_data = (void *)(ruleHandle_p);
    retValue = adk_acl_rule_add(policy_info_p->cp_list,
                             ruleIndex,
                             cp_rule_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Failed to add the rule in CP list\n");
        goto adk_ipv6_pbr_entry_create_return;
    }

    pbr_action_p = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_pbr_table.tbl[pbrRuleStatsId]);
    memset(pbr_action_p, 0, sizeof(adk_ipv6_dp_pbr_action_entry_t));
    pbr_action_p->stats_id = pbrRuleStatsId;
    pbr_action_p->routeTableId= pbrInfo_p->pbrAction.routeTableId;

    /* Set the statistics for this rule to 0 */
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_pbr_table.pbrRuleStats[i*IPV6_PBR_MAX_RULES_PER_SET+pbrRuleStatsId-1] = 0;
    }
    ipv6ModuleInfo_p->maxIpv6PbrRules++;

adk_ipv6_pbr_entry_create_return:

    /* Upon error, release the reserved resources */
    if (retValue != ADK_SUCCESS) {
        if (ruleHandle_p)
            ADK_SYS_FLOW_HANDLE_FREE(ruleHandle_p);

        if (ruleHandle_p)
            ruleHandle_p = 0;
    }

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
#endif

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ipv6_pbr_del_policy 
 *
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    ruleIndex - location in the PBR table from where
 * the rule needs to be deleted.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid.
 *
 *
 * @brief
 * This API validates the input parameter and deletes the PBR rule
 * that is already programmed to the PBR policy set. Note that, both PBR
 * policy handle and PBR rule index are mandatory parameters as the PBR
 * index is not internally tracked by ADK. This PBR rule index changes
 * when other PBR rules are added or deleted in the same policy set.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_pbr_del_policy (
        adk_module_handle_t ipv6Handle,
        uint32_t  ruleIndex)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_pbr_rule_handle_info_t *flow_info_p = NULL;
    adk_ipv6_pbr_rule_t *pbrRules_p;
    adk_ipv6_pbr_action_t *pbrAction_p;
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;
    uint32_t pbrRuleStatsId = ADK_RM_RSRC_INVALID, idx, index=0;
    adk_sys_handle_t sysHandle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_flow_handle_t   ruleHandle_p = 0;
    adk_acl_rule_t cpRule;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid IPv6 module handle \n");
        return (retValue);
    }
    
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Module Info NULL or cookie mismatch \n");
        goto _adk_ipv6_pbr_entry_delete_exit;
    }

    policy_info_p = (ipv6_pbr_rule_set_info_t *) ipv6ModuleInfo_p->adk_ipv6_dp_tables->adk_ipv6_pbr_policy_set_handle_t;
    adk_acl_get_reset(policy_info_p->cp_list);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(policy_info_p->cp_list, &cpRule))
    {
         if(index==ruleIndex)
         {
                pbrRuleStatsId = ADK_SYS_FLOWID_GET(cpRule.user_data);
                ruleHandle_p = (adk_flow_handle_t)cpRule.user_data;
                break;  
         }
         index++;
    }       
    if (ADK_RM_RSRC_INVALID != pbrRuleStatsId)
    {                     
         adk_rm_resource_free(ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle, pbrRuleStatsId);
    }   

    flow_info_p = (adk_ipv6_pbr_rule_handle_info_t *) ADK_SYS_FLOW_ATTR_PTR_GET(ruleHandle_p);
    if (flow_info_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,"pbr flow handle is null");
        goto _adk_ipv6_pbr_entry_delete_exit;
    }

    pbrRules_p = &(flow_info_p->rule);
    pbrAction_p = &(flow_info_p->action);

    /* Make sure we are not deleting more than we can hold for commit */
    if (policy_info_p->num_pending_rule_ids >=
        IPV6_PBR_MAX_RULES_PER_SET)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_UNSUPPORTED);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Cannot delete any more rules until previously"
                           " deleted or modified rules are committed\n");
        goto _adk_ipv6_pbr_entry_delete_exit;
    }

    /* Make sure, the rule is not in commit pending list */
    for (idx = 0; idx < policy_info_p->num_pending_rule_ids; idx++)
    {
        if (policy_info_p->pendingRuleIds[idx] == pbrRuleStatsId)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_UNSUPPORTED);
            ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                               "Cannot delete a rule with pending commit\n");
            goto _adk_ipv6_pbr_entry_delete_exit;
        }
    }

    /*
     * There are no PBR rule delete APIs. Hence remove from ADK CP rule list,
     * and during the commit, flush all the entries & add the rules from the ADK
     * CP rule list.
     */
    retValue = adk_acl_rule_delete(policy_info_p->cp_list,
                             ruleIndex);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Failed to delete the rule from CP list\n");
        goto _adk_ipv6_pbr_entry_delete_exit;
    }

    policy_info_p->pendingRuleIds[idx] = pbrRuleStatsId;
    policy_info_p->num_pending_rule_ids++;
    ipv6ModuleInfo_p->maxIpv6PbrRules--;

    /* Clear the internal cookie */
    flow_info_p->cookie = 0;
    if (ruleHandle_p)
         ADK_SYS_FLOW_HANDLE_FREE(ruleHandle_p);

_adk_ipv6_pbr_entry_delete_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
#endif
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ipv6_pbr_commit
 *
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid.
 *
 * @retval ADK_E_INTERNAL_ERROR - Failed to build the
 * PBR lookup tree.
 *
 * @brief
 * This API commits all additions and deletions of the rules for a
 * set or all sets since last commit. Since commit is an expensive
 * operation, it is recommended that rules be committed only after
 * all desired rules have been added or deleted.
 *
 * =============================================================
 */
adk_status_t 
adk_ipv6_pbr_commit (adk_module_handle_t ipv6Handle)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    uint32_t pbrRuleStatsId=0, idx, cnt;
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;
    adk_sys_handle_t sysHandle;
    adk_ipv6_pbr_rule_handle_info_t *flow_info_p = NULL;
    adk_ipv6_pbr_rule_t *pbrRules_p, pbrRule;
    uint16_t ret, rule_index;
    uint8_t old_index, new_index;
    adk_acl_rule_t cp_rule;
    bool allSets = FALSE; //single PBR rule set is used
    bool noRules = TRUE;
    adk_ipv6_pbr_policy_set_handle_t pbrPolicySetHandle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
  
    
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_pbr_commit_return;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    /*! Acquire the Pbr PolicySet handle */
    pbrPolicySetHandle = ipv6ModuleInfo_p->adk_ipv6_dp_tables->adk_ipv6_pbr_policy_set_handle_t;

    if (!adk_ipv6_valid_policy_set_handle(pbrPolicySetHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid NULL input parameter \n");
        goto adk_ipv6_pbr_commit_return;
    }

    if (allSets == FALSE) {

        policy_info_p = (ipv6_pbr_rule_set_info_t *)
                         pbrPolicySetHandle;

        /*
         * Clear the Ruleset & add the rules from the ADK CP rule list.
         * Then commit the ruleset.
         */
        if(policy_info_p->pbr_tree_index)
        {
            old_index = 1;
            new_index = 0;
        }
        else
        {
            old_index = 0;
            new_index = 1;
        }

        /* Clear the rules from both active & shadow copies.
         * This will not update the runtime lookup DB */
        adk_ipv6_pbr_ruleset_clear(sysHandle, ipv6Handle, policy_info_p->pbr_tree[old_index]);
        adk_ipv6_pbr_ruleset_clear(sysHandle, ipv6Handle, policy_info_p->pbr_tree[new_index]);

        /*Get the rules from CP list & add them to S/W ACL*/
        adk_acl_get_reset(policy_info_p->cp_list);
        rule_index = 0;
        while(ADK_SUCCESS == adk_acl_next_rule_get(
                                    policy_info_p->cp_list, &cp_rule))
        {
            flow_info_p = (adk_ipv6_pbr_rule_handle_info_t *)
                          ADK_SYS_FLOW_ATTR_PTR_GET(cp_rule.user_data);
            pbrRules_p = &(flow_info_p->rule);
            pbrRuleStatsId = ADK_SYS_FLOWID_GET(cp_rule.user_data);
            ADK_IPV6_LOG_INFO(ipv6Handle, "Rule committed for SIP: %x:%x:%x:%x  prefix: %u  mask: %x ruleIndex: %x \n",
				pbrRules_p->srcAddrInfo.Ipv6Addr.addr[0], pbrRules_p->srcAddrInfo.Ipv6Addr.addr[1], 
				pbrRules_p->srcAddrInfo.Ipv6Addr.addr[2], pbrRules_p->srcAddrInfo.Ipv6Addr.addr[3],
			       	pbrRules_p->srcAddrInfo.prefixLength, pbrRules_p->PBR_RuleMask, rule_index);
            /*TODO: enhance this call to perform bulk add*/
            retValue = adk_ipv6_pbr_dpdk_sw_rule(ipv6Handle, policy_info_p,
                           (IPV6_MAX_PBR_RULE_PRIORITY - rule_index),
                           pbrRuleStatsId,
                           pbrRules_p);
            if (retValue != ADK_SUCCESS)
            {
                ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                        "Failed to add the rule in DPDK S/W\n");
                goto adk_ipv6_pbr_commit_return;
            }
            rule_index++;
	    noRules = FALSE;
        }

	/*The default rule is added only when all the user configured 
	 * policies are deleted.
	 */
	if(noRules)
	{
            /*
	     * Add a default to S/W PBR rule to mark as not found
	     */
	    memset(&pbrRule, 0, sizeof(pbrRule));
	    retValue = adk_ipv6_pbr_dpdk_sw_rule(ipv6Handle, policy_info_p,
				IPV6_MAX_PBR_RULE_PRIORITY, 0,
				&pbrRule);
	    if (retValue != ADK_SUCCESS)
	    {
                ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                    "Failed to add the PBR rule in DPDK S/W\n");
                goto adk_ipv6_pbr_commit_return;
            }
	}
        /*Update shadow copy*/
        ret = adk_ipv6_pbr_ruleset_commit(sysHandle, ipv6Handle, policy_info_p->pbr_tree[new_index]);
        if(!ret)
        {
            retValue = ADK_E_INTERNAL_ERROR;
            ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                    "Error: Could not commit PBR ACL rules...\n");
            goto adk_ipv6_pbr_commit_return;
        }

        /* Toggle the index */
        policy_info_p->pbr_tree_index = new_index;
        adk_rcu_synchronize();

        /* Update active copy */
        ret = adk_ipv6_pbr_ruleset_commit(sysHandle, ipv6Handle, policy_info_p->pbr_tree[old_index]);
        if(!ret)
        {
            retValue = ADK_E_INTERNAL_ERROR;
            ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                    "Error: Could not commit PBR ACL rules...\n");
            goto adk_ipv6_pbr_commit_return;
        }
        /* CP ACL List update to re-arrange the List */
        retValue = adk_acl_sw_only_commit(policy_info_p->cp_list);
        if(retValue != ADK_SUCCESS)
        {
            retValue = ADK_E_INTERNAL_ERROR;
            ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                               "Error: Could not re-arrange CP PBR ACL List...\n");
            goto adk_ipv6_pbr_commit_return;
        }
    }
    /* The else case is for multiple pbr  policy set based implementation.
     * Since we are using single policy set, the else case will never be hit.
     * Still keeping it for future use, if needed.
     */
    else
    {
        /* Re-build from scratch */
        for (cnt = 0; cnt < ADK_IPV6_PBR_RULE_SETS_MAX; cnt++)
        {
            /* Make sure that the ruleset handle is a valid one */
            if (ipv6ModuleInfo_p->policySetHandles[cnt] == 0)
                continue;
            policy_info_p = (ipv6_pbr_rule_set_info_t *)
                ipv6ModuleInfo_p->policySetHandles[cnt];

            /*
             * Clear the Ruleset & add the rules from the ADK CP rule list.
             * Then commit the ruleset.
             */
            if(policy_info_p->pbr_tree_index)
            {
                old_index = 1;
                new_index = 0;
            }
            else
            {
                old_index = 0;
                new_index = 1;
            }

            /* Clear the rules from both active & shadow copies.
             * This will not update the runtime lookup DB */
            adk_ipv6_pbr_ruleset_clear(sysHandle, ipv6Handle, policy_info_p->pbr_tree[old_index]);
            adk_ipv6_pbr_ruleset_clear(sysHandle, ipv6Handle, policy_info_p->pbr_tree[new_index]);

            /*Get the rules from CP list & add them to S/W ACL*/
            adk_acl_get_reset(policy_info_p->cp_list);
            rule_index = 0;
            while(ADK_SUCCESS == adk_acl_next_rule_get(
                                    policy_info_p->cp_list, &cp_rule))
            {
                flow_info_p = (adk_ipv6_pbr_rule_handle_info_t *)
                    ADK_SYS_FLOW_ATTR_PTR_GET(cp_rule.user_data);

                pbrRules_p = &(flow_info_p->rule);
                ADK_IPV6_LOG_INFO(ipv6Handle," Rule committed for SIP: %x:%x:%x:%x  prefix: %u  mask: %x ruleIndex: %x \n",
                        pbrRules_p->srcAddrInfo.Ipv6Addr.addr[0], pbrRules_p->srcAddrInfo.Ipv6Addr.addr[1], 
                        pbrRules_p->srcAddrInfo.Ipv6Addr.addr[2], pbrRules_p->srcAddrInfo.Ipv6Addr.addr[3],
                        pbrRules_p->srcAddrInfo.prefixLength, pbrRules_p->PBR_RuleMask, rule_index);
                pbrRuleStatsId = ADK_SYS_FLOWID_GET(cp_rule.user_data);
                /*TODO: enhance this call to perform bulk add*/
                retValue = adk_ipv6_pbr_dpdk_sw_rule(ipv6Handle, policy_info_p,
                               (IPV6_MAX_PBR_RULE_PRIORITY - rule_index),
                               pbrRuleStatsId, pbrRules_p);
                if (retValue != ADK_SUCCESS)
                {
                    ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                            "Failed to add the PBR rule in DPDK S/W\n");
                    goto adk_ipv6_pbr_commit_return;
                }
                rule_index++;
                noRules = FALSE;
            }
#if 1
            if(noRules)
            {
                /*
                 * Add a default to S/W ACL rule to mark as not found
                 */
                memset(&pbrRule, 0, sizeof(pbrRule));
                retValue = adk_ipv6_pbr_dpdk_sw_rule(ipv6Handle, policy_info_p,
                        IPV6_MAX_PBR_RULE_PRIORITY, 0,
                        &pbrRule);
                if (retValue != ADK_SUCCESS)
                {
                    ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                            "Failed to add the PBR rule in DPDP S/W\n");
                    goto adk_ipv6_pbr_commit_return;
                }
            }
#endif

            /*Update shadow copy*/
            ret = adk_ipv6_pbr_ruleset_commit(sysHandle, ipv6Handle, policy_info_p->pbr_tree[new_index]);
            if(!ret)
            {
                retValue = ADK_E_INTERNAL_ERROR;
                ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                        "Error: Could not commit PBR ACL rules...\n");
                goto adk_ipv6_pbr_commit_return;
            }
            /* Toggle the index */
            policy_info_p->pbr_tree_index = new_index;
        }

        adk_rcu_synchronize();

        /* Update the old copy */
        for (cnt = 0; cnt < ADK_IPV6_PBR_RULE_SETS_MAX; cnt++)
        {
            /* Make sure that the ruleset handle is a valid one */
            if (ipv6ModuleInfo_p->policySetHandles[cnt] == 0)
                continue;
            policy_info_p = (ipv6_pbr_rule_set_info_t *)
                ipv6ModuleInfo_p->policySetHandles[cnt];

            if(policy_info_p->pbr_tree_index)
                old_index = 0;
            else
                old_index = 1;

            ret = adk_ipv6_pbr_ruleset_commit(sysHandle, ipv6Handle, policy_info_p->pbr_tree[old_index]);
            if(!ret)
            {
                retValue = ADK_E_INTERNAL_ERROR;
                ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                        "Error: Could not commit PBR ACL rules...\n");
                goto adk_ipv6_pbr_commit_return;
            }

            /* CP ACL List update to re-arrange the List */
            retValue = adk_acl_sw_only_commit(policy_info_p->cp_list);
            if(retValue != ADK_SUCCESS)
            {
                retValue = ADK_E_INTERNAL_ERROR;
                ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                                "Error: Could not re-arrange CP ACL List...\n");
                goto adk_ipv6_pbr_commit_return;
            }
        }
	rte_acl_dump(policy_info_p->pbr_tree[0]);
	rte_acl_dump(policy_info_p->pbr_tree[1]);

    }
 
    adk_sys_flow_cache_update_version(sysHandle);

    /* Once the commit is complete, remove all the configuration for deleted
     * rules since last commit */
    for (cnt = 0; cnt < ADK_IPV6_PBR_RULE_SETS_MAX; cnt++)
    {
        /* If commit is for all sets, then get all the rule sets one by one and
         * clean the resources */
            /* Make sure that the ruleset handle is a valid one */
        if (ipv6ModuleInfo_p->policySetHandles[cnt] == 0)
            continue;
        policy_info_p = (ipv6_pbr_rule_set_info_t *)
                ipv6ModuleInfo_p->policySetHandles[cnt];

        for (idx = 0; idx < policy_info_p->num_pending_rule_ids;
             idx++)
        {
            pbrRuleStatsId = policy_info_p->pendingRuleIds[idx];
            /*pbrRuleStatsId is comming as 0, which is causing issue Jira:362
             *Since 0 is reserve fixed, it should not be freed 
             *TODO: Have to remove the whole pendingRuleIds[] , as its not really needed
              and resource is freed already in del_policy, before comming to pbr_commit*/
            if(pbrRuleStatsId)
                adk_rm_resource_free(ipv6ModuleInfo_p->ipv6PbrPolicyIDHandle,
                      pbrRuleStatsId);
        }
        policy_info_p->num_pending_rule_ids = 0;

    }

adk_ipv6_pbr_commit_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
#endif
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ipv6_pbr_config
 *
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    ipv6IfHandle - The IPv6 interface handle for which the PBR needs to
 * be enabled. 
 *
 * @param[in]    pbrPolicySetHandle - PBR policy set handle
 * 
 * @retval ADK_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER -  one of the direct arguments
 * to the API call is invalid.
 *
 * @brief
 * This API enables PBR on an IPv6 interface and configures IP PBR policy-set 
 * for an IPv6 interface. The packets received on the interface are then subject
 * to the policy based routing. Note that this API must be called for every
 * IPv6 interface on which PBR needs to be enabled. Also note that an IPv6
 * interface can only be associated with one PBR policy-set. So, calling this 
 * API will replace the existing association to any PBR policy-set, 
 * if any present.
 *
 * =============================================================
 */
adk_status_t 
adk_ipv6_pbr_config (
    adk_module_handle_t ipv6Handle,
    adk_interface_handle_t  *ipv6IfHandle,
    adk_ipv6_pbr_policy_set_handle_t pbrPolicySetHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_sys_handle_t sysHandle;
    uint32_t pbrPolicySetId = 0;
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;
    ipv6_interface_info_t *if_pvt_info = NULL;
   /* ipv6_tunnel_interface_info_t  *tun_if_pvt_info = NULL;*/
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    adk_ipv6_dp_interface_entry_t *dp_entry;

    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Module Info is NULL or cookie mismatches\n");
        goto _adk_ipv6_pbr_config_exit;
    }

    if ( NULL == *ipv6IfHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "NULL pointer parameter specified\n");
        goto _adk_ipv6_pbr_config_exit;
    }

    if (!adk_ipv6_interface_valid_handle(*ipv6IfHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Invalid Interface Handle type\n");
        goto _adk_ipv6_pbr_config_exit;
    }

    if_pvt_info = (ipv6_interface_info_t *)
                  ADK_SYS_INTERFACE_ATTR_PTR_GET(*ipv6IfHandle);
    if ( if_pvt_info->ifType == IPV6_INTERFACE_TYPE_VIRTUAL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                      "This attribute is not applicable for Virtual Interface\n");
        goto _adk_ipv6_pbr_config_exit;
    }
   
    if(!adk_ipv6_valid_policy_set_handle(pbrPolicySetHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Invalid IP PBR Policy Set Handle\n");
        goto _adk_ipv6_pbr_config_exit;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    policy_info_p = (ipv6_pbr_rule_set_info_t *) pbrPolicySetHandle;

    pbrPolicySetId = policy_info_p->ipv6_pbr_rule_set_id;
    if_pvt_info->pbr_policy_set_id = pbrPolicySetId;
    
    /*Update dp*/
    dp_entry = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_if_table.tbl[if_pvt_info->stats_id]);
    dp_entry->pbr_policy_set_id = pbrPolicySetId;
    if(pbrPolicySetId) {
        dp_entry->enable_pbr = TRUE;
    } else {
        dp_entry->enable_pbr = FALSE;
    }

    /* Update the dp tables version */
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_fc_ver[dp_entry->vrd_id]++;

_adk_ipv6_pbr_config_exit:

#endif    
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_ipv6_vrd_pbr_config
 *
 * @param[in] ipv6Handle            IPv6 module handle.
 *
 * @param[in] vrd_id                VRD ID for which the PBR rule set
 * handle needs to be configured.
 *
 * @param[in] pbrPolicySetHandle    The PBR rule set handle that needs
 * to be associated with VRD. If specified as 0, any existing association
 * is removed for the VRD.
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv6 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER  One of the direct arguments
 * to the API call is invalid.
 *
 * @brief
 *
 * This API associates or dissociates a PBR rule set for fast path 
 * originating traffic in specified Virtual Routing domain.
 * =============================================================
 */
adk_status_t adk_ipv6_vrd_pbr_config (
    adk_module_handle_t __attribute__((unused)) ipv6Handle,
    uint32_t __attribute__((unused)) vrd_id,
    adk_ipv6_pbr_policy_set_handle_t __attribute__((unused)) pbrPolicySetHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    uint32_t pbrPolicySetId = 0;
    ipv6_pbr_rule_set_info_t *policy_info_p =
        (ipv6_pbr_rule_set_info_t *) pbrPolicySetHandle;
    adk_ipv6_dp_vrd_pbr_ruleset_map_table_t *dp_vrd_pbr_policy_map;
    adk_sys_handle_t sysHandle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NULL or cookie mismatch \n");
        goto _adk_ipv6_vrd_pbr_config_exit;
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if(pbrPolicySetHandle)
    {
        if (!adk_ipv6_valid_policy_set_handle(pbrPolicySetHandle))
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid NULL input parameter \n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto _adk_ipv6_vrd_pbr_config_exit;
        }
        pbrPolicySetId = policy_info_p->ipv6_pbr_rule_set_id;
    }

    /*
     * Program the data plane tree for vrd id to policy set id mapping
    */
    dp_vrd_pbr_policy_map = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->
                             ipv6_dp_pbr_ruleset_map_table[vrd_id]);
    if(pbrPolicySetHandle) {
        dp_vrd_pbr_policy_map->pbr_policy_set_id = pbrPolicySetId;
        dp_vrd_pbr_policy_map->pbr_enabled = TRUE;
    } else {
        dp_vrd_pbr_policy_map->pbr_enabled = FALSE;
        dp_vrd_pbr_policy_map->pbr_policy_set_id = 0;
    }

_adk_ipv6_vrd_pbr_config_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
#endif
    return retValue;
}

/*!
 * =============================================================
 * adk_ipv6_pbr_rule_stats_get
 *
 * @param[in] ipv6Handle              IPv6 module handle
 *
 * @param[in] pbrRuleHandle         IPv6 PBR rule handle for which
 * the statistics needs to be retrieved
 *
 * @param[out] ipv6PbrRuleStats_p     Pointer to the structure having the
 * statistics related to the PBR rule.
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv6 module handle specified
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid parameter e.g. NULL pointer
 * parameter.
 *
 * @brief
 * This API gets the statistics for a PBR rule. 
 * =============================================================
 */
adk_status_t adk_ipv6_pbr_rule_stats_get (
    adk_module_handle_t __attribute__ ((unused))ipv6Handle,
    adk_flow_handle_t __attribute__ ((unused))pbrRuleHandle,
    adk_ipv6_pbr_rule_stats_t __attribute__ ((unused))*ipv6PbrRuleStats_p)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    uint32_t pbrRuleStatsId, i;
    adk_sys_handle_t sysHandle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, 
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_pbr_rule_stats_get_return;
    }

    if (!adk_ipv6_valid_pbr_rule_handle(pbrRuleHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Invalid IPv6 Pbr Rule Set Handle\n");
        goto adk_ipv6_pbr_rule_stats_get_return;
    }

    pbrRuleStatsId = ADK_SYS_FLOWID_GET(pbrRuleHandle);
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    ipv6PbrRuleStats_p->matchedPkts = 0;
    for(i=0;i<ADK_MAX_NUM_CORES;i++)
    {
        ipv6PbrRuleStats_p->matchedPkts +=
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_pbr_table.pbrRuleStats[i*IPV6_PBR_MAX_RULES_PER_SET+pbrRuleStatsId-1];
    }

adk_ipv6_pbr_rule_stats_get_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
#endif
    return retValue;
}


/*!
 * =============================================================
 * adk_ipv6_pbr_rule_stats_print
 *
 * @param[in] ipv6Handle              IPv6 module handle.
 *
 * @param[in] pbrRuleHandle      IPv6 pbr rule handle
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE     Invalid IPv6 module handle specified
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid parameter e.g. invalid PBR
 * rule handle.
 *
 * @brief
 *
 * This API dumps the statistics for a pbr rule on the console.
 * =============================================================
 */
adk_status_t adk_ipv6_pbr_rule_stats_print (
    adk_module_handle_t __attribute__ ((unused))ipv6Handle,
    adk_flow_handle_t __attribute__ ((unused))pbrRuleHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_ipv6_pbr_rule_stats_t ipv6PbrRuleStats;

    retValue = adk_ipv6_pbr_rule_stats_get(ipv6Handle, pbrRuleHandle,
                                            &ipv6PbrRuleStats);
    if (retValue != ADK_SUCCESS)
        return retValue;

    ADK_INFO_WRITE("\n IPv6 PBR Rule %"ADKPRIh" Statistics\n", pbrRuleHandle);
    ADK_INFO_WRITE("Total Pkts Matched: %u\n", ipv6PbrRuleStats.matchedPkts);
#endif
    return retValue;
}

/*!
 * =============================================================
 * @internal
 * adk_ipv6_pbr_dpdk_sw_rule
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief
 *
 * Internal support function for PBR ACL rule addition.
 *
 * =============================================================
 */
#if ADK_IPV6_PBR_ENABLE
static adk_status_t
adk_ipv6_pbr_dpdk_sw_rule (
    adk_module_handle_t ipv6Handle,
    ipv6_pbr_rule_set_info_t *policy_info_p,
    uint32_t precedence,
    uint32_t pbrRuleStatsId,
    const adk_ipv6_pbr_rule_t *pbrRules_p
)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_ipv6_pbr_acl_rule_t pbr_acl_rule;
    adk_sys_handle_t sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    int ret;

    /*****   Build the Individual Patterns ******/
    ADK_MEMSET(&pbr_acl_rule, 0, sizeof(pbr_acl_rule));

    if (pbrRules_p->PBR_RuleMask & ADK_IPV6_PBR_SRC_ADDR_MASK)
    {
        ADK_MEMCPY(&pbr_acl_rule.sip_addr[0], 
            &pbrRules_p->srcAddrInfo.Ipv6Addr.addr[0], sizeof(adk_ipv6_addr_t));
        pbr_acl_rule.sip_mask = pbrRules_p->srcAddrInfo.prefixLength;
    }
    
    /* PBR rule add will not touch the lookup DB.
     * Hence add rule to both active & shadow copy.
     */
    ret = adk_ipv6_pbr_rule_add(sysHandle, ipv6Handle, policy_info_p->pbr_tree[0],
       &pbr_acl_rule, (adk_ipv6_pbr_val_t)(uintptr_t)pbrRuleStatsId, precedence);
    if(!ret)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ipv6Handle, 
                           "Error inserting PBR rule in IPv6\n");
        goto _adk_ipv6_pbr_dpdk_sw_rule_exit;
    }
    ret = adk_ipv6_pbr_rule_add(sysHandle, ipv6Handle, policy_info_p->pbr_tree[1],
       &pbr_acl_rule, (adk_ipv6_pbr_val_t)(uintptr_t)pbrRuleStatsId, precedence);
    if(!ret)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ipv6Handle, 
                           "Error inserting PBR rule in IPv6\n");
        goto _adk_ipv6_pbr_dpdk_sw_rule_exit;
    }

_adk_ipv6_pbr_dpdk_sw_rule_exit:

    ADK_RETURN(retValue);

}    /* -----  end of function adk_ip_pbr_dpdk_sw_rule  ----- */
#endif

/*!
 * =============================================================
 * adk_ipv6_get_num_pbr_rules
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]   pbrRuleSetHandle - policyset handle for which want to 
 * get number of rules. This can be NULL
 *
 * @param[out]   numOfRules_p - Number of pbr rules present
 * on the particular policyset or total number of rules present
 *  
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 * @retval ADK_E_INVALID_PARAMETER - IPv6 inteface 
 * handle not present or invalid.  
 * 
 * @retval ADK_IPV6_E_INVALID_PARAMETER -  one of the direct 
 * arguments to the API call is invalid.
 *
 * @retval ADK_IPV6_E_FAILED_TO_GET_NUM_RULES - Failed to 
 * get number of pbr rules.
 *
 * @brief  
 *
 * This API is not supported
 *
 * This API validates the input parameter. If policyset handle 
 * is non NULL then it get number of pbr rules present 
 * on that policy set . If policyset handle is NULL then it gets 
 * rules present on all the policysets
 *
 * =============================================================
 */
adk_status_t  
adk_ipv6_get_num_pbr_rules ( 
        adk_module_handle_t __attribute__ ((unused))ipv6Handle,
        adk_ipv6_pbr_policy_set_handle_t __attribute__ ((unused))pbrRuleSetHandle,
        uint32_t __attribute__ ((unused))*numOfRules_p)
{
    adk_status_t retValue = ADK_IPV6_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    uint32_t numRules = 0, Ruleset = 0;
    adk_sys_handle_t sysHandle;
    adk_acl_rule_t cp_rule;
    ipv6_pbr_rule_set_info_t *policy_info_p;
    adk_ipv6_info_t       	*ipv6ModuleInfo_p=NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*Check if valid input interface input is given*/
    if ( pbrRuleSetHandle )
    {
        if (!adk_ipv6_valid_policy_set_handle(pbrRuleSetHandle)) 
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid IPv6 pbr RulesetHandle specified\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto _adk_ipv6_get_num_pbr_rules_return;
        }
        policy_info_p = (ipv6_pbr_rule_set_info_t *)(pbrRuleSetHandle);

        adk_acl_get_reset(policy_info_p->cp_list);
        while( ADK_SUCCESS == adk_acl_next_rule_get(
                                    policy_info_p->cp_list, &cp_rule))
        {
            numRules++;
        }
    }
    else
    {
        for ( Ruleset = 0; Ruleset < ADK_IPV6_PBR_RULE_SETS_MAX; Ruleset++)
        {
            pbrRuleSetHandle = ipv6ModuleInfo_p->policySetHandles[Ruleset];
            if ( pbrRuleSetHandle )
            {
                if (!adk_ipv6_valid_policy_set_handle(pbrRuleSetHandle)) 
                {
                    ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid IPv6 pbr RulesetHandle specified\n");
                    retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                    goto _adk_ipv6_get_num_pbr_rules_return;
                }
                policy_info_p = (ipv6_pbr_rule_set_info_t *)(pbrRuleSetHandle);

                adk_acl_get_reset(policy_info_p->cp_list);
                while( ADK_SUCCESS == adk_acl_next_rule_get(
                                            policy_info_p->cp_list, &cp_rule))
                {
                    numRules++;
                }
            }
        }
    }
    *numOfRules_p = numRules;
_adk_ipv6_get_num_pbr_rules_return:
#endif
    ADK_RETURN ( retValue );
}

/*!
 * =============================================================
 * adk_ipv6_get_pbr_rules
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @param[in]    pbrRuleSetHandle - pbr rule set handle for which want to 
 * get list of pbr rules. This can be NULL
 *
 *
 * @param[in]    numOfRules - Number of filter rules to be 
 * returned
 *  
 * @param[out]  pbr_entry_p - Pointer to 
 * adk_ipv6_pbr_info_t to return filter list
 * 
 * @retval ADK_IPV6_SUCCESS Successful completion
 *
 * @retval ADK_E_INVALID_PARAMETER - IPv6 pbr policyset
 * handle not present or invalid.  
 * 
 * @retval ADK_IPV6_E_INVALID_PARAMETER -  one of the direct 
 * arguments to the API call is invalid.
 *
 * @retval ADK_IPV6_E_FAILED_TO_GET_NUM_RULES - Failed to 
 * get number of pbr rules.
 *
 * @brief  
 *
 * This API is not supported, used only as part of utility to dump tables
 *
 * This API validates the input parameter. If policyset handle 
 * is non NULL then it gets pbr rules for that policy set. 
 * If polict set handle is NULL then it gets rules present on all the 
 * policy sets.
 * 
 * =============================================================
 */
adk_status_t  
adk_ipv6_get_pbr_rules ( 
        adk_module_handle_t __attribute__ ((unused))ipv6Handle,
        adk_ipv6_pbr_policy_set_handle_t __attribute__ ((unused))pbrRuleSetHandle,
        uint32_t __attribute__ ((unused))numOfRules,
        adk_ipv6_pbr_info_t __attribute__ ((unused))*pbr_entry_p)
{
    adk_status_t retValue = ADK_IPV6_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    uint32_t numRules = 0, Ruleset = 0;
    adk_sys_handle_t sysHandle;
    adk_acl_rule_t cp_rule;
    adk_ipv6_pbr_rule_handle_info_t *rule_handle;
    ipv6_pbr_rule_set_info_t *policy_info_p;
    adk_ipv6_info_t       	*ipv6ModuleInfo_p=NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*Check if valid input interface input is given*/
    if ( pbrRuleSetHandle)
    {
        if (!adk_ipv6_valid_policy_set_handle(pbrRuleSetHandle)) 
        {
            ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid IPv6 pbr RulesetHandle specified\n");
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto _adk_ipv6_get_pbr_rules_return;
        }
        policy_info_p = (ipv6_pbr_rule_set_info_t *)(pbrRuleSetHandle);
        adk_acl_get_reset(policy_info_p->cp_list);
        while( (ADK_SUCCESS == adk_acl_next_rule_get(
                                    policy_info_p->cp_list, &cp_rule)) &&
               (numRules < numOfRules))
        {
            rule_handle = (adk_ipv6_pbr_rule_handle_info_t *)
                          ADK_SYS_FLOW_ATTR_PTR_GET(cp_rule.user_data);
            ADK_MEMCPY(&(pbr_entry_p[numRules].pbrRule), &(rule_handle->rule), sizeof(adk_ipv6_pbr_rule_t));
            ADK_MEMCPY(&(pbr_entry_p[numRules].pbrAction), &(rule_handle->action), sizeof(adk_ipv6_pbr_action_t)); 
            numRules++;
        }
    }
    else
    {
        for ( Ruleset = 0; Ruleset < ADK_IPV6_PBR_RULE_SETS_MAX; Ruleset++)
        {
            pbrRuleSetHandle = ipv6ModuleInfo_p->policySetHandles[Ruleset];
            if ( pbrRuleSetHandle)
            {
                if (!adk_ipv6_valid_policy_set_handle(pbrRuleSetHandle)) 
                {
                    ADK_LOG_ERROR(sysHandle, ipv6Handle, "Invalid IPv6 filter RulesetHandle specified\n");
                    retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                    goto _adk_ipv6_get_pbr_rules_return;
                }
                policy_info_p = (ipv6_pbr_rule_set_info_t *)(pbrRuleSetHandle);
                adk_acl_get_reset(policy_info_p->cp_list);
                while( (ADK_SUCCESS == adk_acl_next_rule_get(
                                            policy_info_p->cp_list, &cp_rule)) &&
                       (numRules < numOfRules))
                {
                    rule_handle = (adk_ipv6_pbr_rule_handle_info_t *) cp_rule.user_data;
                    ADK_MEMCPY(&(pbr_entry_p[numRules].pbrRule), &(rule_handle->rule), sizeof(adk_ipv6_pbr_rule_t));
                    ADK_MEMCPY(&(pbr_entry_p[numRules].pbrAction), &(rule_handle->action), sizeof(adk_ipv6_pbr_action_t)); 
                    numRules++;
                }
            }
            
        }
    }
    
_adk_ipv6_get_pbr_rules_return:
#endif
    ADK_RETURN ( retValue );
}

/*!
* ====================================================================
*  adk_ipv6_get_pbr_policies_all
*
* @param[in]    ipv6Handle - IPV6 module handle.
*
* @param[out]   *numPbrPolicies - Number of PBR rules
*
* @param[out]   **ipv6PbrPol_p - List of PBR rules
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_E_INVALID_PARAMETER  Invalid input parameters
*
* @retval ADK_E_INVALID_HANDLE   Ipv6 Module is invalid.
*
* @brief
*
* This API gets the PBR policies
*
* @details
*
* This API gets all the PBR policies along with the count of them.
* The array index of a rule represents its index in the tree.
* Application is expected to free ipv6PbrPol_p pointer once it is used.
* ====================================================================
*/
adk_status_t
adk_ipv6_get_pbr_policies_all(
        adk_module_handle_t      ipv6Handle,
        uint32_t                 *numPbrPolicies,
        adk_ipv6_pbr_info_t      **ipv6PbrPol_p)
{
    adk_status_t retValue = ADK_SUCCESS;
#if ADK_IPV6_PBR_ENABLE
    adk_acl_rule_t cp_rule;
    adk_sys_handle_t sysHandle;
    adk_ipv6_info_t *ipv6ModuleInfo_p = NULL;
    ipv6_pbr_rule_set_info_t *policy_info_p = NULL;
    adk_ipv6_pbr_rule_handle_info_t *flow_info_p = NULL;
    adk_ipv6_pbr_info_t *ipv6PbrPol = NULL;
    adk_ipv6_pbr_policy_set_handle_t pbrPolicySetHandle;
    uint32_t num_nodes=0;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue,
                           "Module Info NULL or cookie mismatch \n");
        goto adk_ipv6_get_pbr_policies_return;
    }
    if ((numPbrPolicies == NULL) || (ipv6PbrPol_p == NULL))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid input parameter %"ADKPRIh" %"ADKPRIh"\n",
                            numPbrPolicies, ipv6PbrPol_p);
        goto adk_ipv6_get_pbr_policies_return;
    }
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    /*! Acquire the Pbr PolicySet handle */
    pbrPolicySetHandle = ipv6ModuleInfo_p->adk_ipv6_dp_tables->adk_ipv6_pbr_policy_set_handle_t;

    if (!adk_ipv6_valid_policy_set_handle(pbrPolicySetHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle,retValue, "Invalid NULL input parameter \n");
        goto adk_ipv6_get_pbr_policies_return;
    }

    policy_info_p = (ipv6_pbr_rule_set_info_t *)pbrPolicySetHandle;

    adk_acl_get_num_nodes(policy_info_p->cp_list, &num_nodes);

    ipv6PbrPol = (adk_ipv6_pbr_info_t*)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_PBR_RULESETS_OBJ, num_nodes, sizeof(adk_ipv6_pbr_info_t));

    adk_acl_get_reset(policy_info_p->cp_list);

    for (uint32_t i=0; i<num_nodes; i++)
    {
        retValue = adk_acl_next_rule_get(policy_info_p->cp_list, &cp_rule);
        if (retValue != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LIST_NOT_FOUND);
            ADK_IPV6_LOG_ERROR(ipv6Handle, retValue,
                               "PBR policy is not found in CP ACL list\n");
            goto adk_ipv6_get_pbr_policies_return;
        }
        flow_info_p = (adk_ipv6_pbr_rule_handle_info_t *)
                      ADK_SYS_FLOW_ATTR_PTR_GET(cp_rule.user_data);

        if ((ipv6PbrPol != NULL) && (flow_info_p != NULL))
        {
            ADK_MEMCPY(&((ipv6PbrPol+i)->pbrRule), &(flow_info_p->rule), sizeof(adk_ipv6_pbr_rule_t));
            ADK_MEMCPY(&((ipv6PbrPol+i)->pbrAction), &(flow_info_p->action), sizeof(adk_ipv6_pbr_action_t));
        }
    }
    *ipv6PbrPol_p = ipv6PbrPol;
    *numPbrPolicies = num_nodes;

adk_ipv6_get_pbr_policies_return:
ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);

#endif
    ADK_RETURN(retValue);
}
/*!
 * end addtogroup _ipv6Functions  IPv6 Functions
 * @}
 */
