/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <adk_include.h>
#include <adk_ipv6_pbr_acl.h>
#include "adk_ipv6_internal.h"

#include "adk_dpdk.h"

/*
 * Rule and trace formats definitions.
 */

struct pbr_tuple {
        uint8_t  first_byte;
        uint32_t ip_src1;
        uint32_t ip_src2;
        uint32_t ip_src3;
        uint32_t ip_src4;
}__attribute__((__packed__));

enum {
        FIRST_BYTE_FIELD_IPV6_PBR,
        SRC1_FIELD_IPV6_PBR,
        SRC2_FIELD_IPV6_PBR,
        SRC3_FIELD_IPV6_PBR,
        SRC4_FIELD_IPV6_PBR,
	NUM_FIELDS_IPV6_PBR
};

enum {
        FIRST_BYTE_INPUT_IPV6_PBR,
        SRC1_INPUT_IPV6_PBR,
        SRC2_INPUT_IPV6_PBR,
        SRC3_INPUT_IPV6_PBR,
        SRC4_INPUT_IPV6_PBR,
};

static struct rte_acl_field_def pbr_defs[NUM_FIELDS_IPV6_PBR] = {

    /* first input field - always one byte long. */
    {
        .type = RTE_ACL_FIELD_TYPE_BITMASK,
        .size = sizeof (uint8_t),
        .field_index = FIRST_BYTE_FIELD_IPV6_PBR,
        .input_index = FIRST_BYTE_INPUT_IPV6_PBR,
        .offset = offsetof (struct pbr_tuple, first_byte),
    },
    /* next input field (IPv6 source address) - 4 consecutive bytes. */
    {
	.type = RTE_ACL_FIELD_TYPE_MASK,
	.size = sizeof(uint32_t),
	.field_index = SRC1_FIELD_IPV6_PBR,
	.input_index = SRC1_INPUT_IPV6_PBR,
	.offset = offsetof(struct pbr_tuple, ip_src1),
    },
    {
	.type = RTE_ACL_FIELD_TYPE_MASK,
	.size = sizeof(uint32_t),
	.field_index = SRC2_FIELD_IPV6_PBR,
	.input_index = SRC2_INPUT_IPV6_PBR,
	.offset = offsetof(struct pbr_tuple, ip_src2),
    },
    {
	.type = RTE_ACL_FIELD_TYPE_MASK,
	.size = sizeof(uint32_t),
	.field_index = SRC3_FIELD_IPV6_PBR,
	.input_index = SRC3_INPUT_IPV6_PBR,
	.offset = offsetof(struct pbr_tuple, ip_src3),
    },
    {
	.type = RTE_ACL_FIELD_TYPE_MASK,
	.size = sizeof(uint32_t),
	.field_index = SRC4_FIELD_IPV6_PBR,
	.input_index = SRC4_INPUT_IPV6_PBR,
	.offset = offsetof(struct pbr_tuple, ip_src4),
    },
};

RTE_ACL_RULE_DEF(pbr_rule, NUM_FIELDS_IPV6_PBR);

adk_ipv6_pbr_tree_t *adk_ipv6_pbr_create(adk_sys_handle_t sys_handle, 
                                         adk_module_handle_t ipv6Handle, 
                                         uint16_t ruleSetId, uint8_t index)
{
    char name[30];
    struct rte_acl_ctx * ctx=NULL;
    struct rte_acl_param prm;
    
    snprintf(name, 30, "ADK_IPV6_PBR%3d_%1d", ruleSetId, index);
    prm.name = name;
    prm.socket_id = SOCKET_ID_ANY;
    prm.rule_size = RTE_ACL_RULE_SZ(NUM_FIELDS_IPV6_PBR);
    prm.max_rule_num = IPV6_PBR_MAX_RULES_PER_SET + 1;


    ADK_IPV6_LOG_EXT_API(ipv6Handle, "rte_acl_create", "Creating IPv6 PBR ACL context \n");
    ctx = rte_acl_create(&prm);

    return((adk_ipv6_pbr_tree_t *)ctx);
}

int adk_ipv6_pbr_ruleset_free(adk_sys_handle_t sys_handle,
                              adk_module_handle_t ipv6Handle,
                              adk_ipv6_pbr_tree_t *pbr_tree)
{
    ADK_IPV6_LOG_EXT_API(ipv6Handle, "rte_acl_free", "Deleting IPv6 PBR ACL context \n");
    rte_acl_free((struct rte_acl_ctx *)pbr_tree);
    return 1;
}

int adk_ipv6_pbr_rule_add(adk_sys_handle_t sys_handle,
                        adk_module_handle_t ipv6Handle,
                        adk_ipv6_pbr_tree_t *pbr_tree,
		 	adk_ipv6_pbr_acl_rule_t *pbr_rule,
		 	adk_ipv6_pbr_val_t  pbr_val,
		 	uint32_t precedence)
{
    int rc, i;
    struct pbr_rule v;

    ADK_MEMSET(&v, 0, sizeof(v));

    for(i=0; i<4; i++ )
    {
        v.field[SRC1_FIELD_IPV6_PBR+i].value.u32 = pbr_rule->sip_addr[i];
        if( ((i+1)* sizeof(uint32_t) * CHAR_BIT) <= pbr_rule->sip_mask)
        {    
            v.field[SRC1_FIELD_IPV6_PBR+i].mask_range.u32 = sizeof(uint32_t) * CHAR_BIT;
        }
        else
        {
            if( ((i)* sizeof(uint32_t) * CHAR_BIT) < pbr_rule->sip_mask)
            {
                v.field[SRC1_FIELD_IPV6_PBR+i].mask_range.u32 = 
                        pbr_rule->sip_mask-((i)*sizeof(uint32_t)*CHAR_BIT);
            }
            else
            {
                v.field[SRC1_FIELD_IPV6_PBR+i].mask_range.u32 = 0;
            }
        }
    }

    v.data.category_mask = 1 /*LEN2MASK(RTE_ACL_MAX_CATEGORIES)*/;
    v.data.priority = precedence;
    v.data.userdata = (uint32_t) (size_t)pbr_val;

    ADK_IPV6_LOG_EXT_API(ipv6Handle, "rte_acl_add_rules", "Add rules to IPv6 Pbr ACL context with priority %d \n", precedence);
    rc = rte_acl_add_rules((struct rte_acl_ctx *)pbr_tree, (struct rte_acl_rule *)&v, 1);
    if (rc != 0) {
	return 0;
    }

    return 1;
}

int adk_ipv6_pbr_ruleset_commit(adk_sys_handle_t sys_handle,
                                adk_module_handle_t ipv6Handle,
                                adk_ipv6_pbr_tree_t *pbr_tree)
{
    struct rte_acl_config cfg;
    int ret;

    ADK_MEMSET(&cfg, 0, sizeof(cfg));
    cfg.num_fields = NUM_FIELDS_IPV6_PBR;
    ADK_MEMCPY(&cfg.defs, pbr_defs, sizeof(pbr_defs));
    cfg.num_categories = 1;
    cfg.max_size = 0;

    ADK_IPV6_LOG_EXT_API(ipv6Handle, "rte_acl_build", "Building IPv6 PBR ACL context \n");
    ret = rte_acl_build((struct rte_acl_ctx *)pbr_tree, &cfg);
    if (ret != 0)
    {
        return 0;
    }
    return 1;

}

void adk_ipv6_pbr_ruleset_clear(adk_sys_handle_t sys_handle,
                                adk_module_handle_t ipv6Handle,
                                adk_ipv6_pbr_tree_t *pbr_tree)
{
    /* Remove all the rules. Does not clear the internal run time memory */
    ADK_IPV6_LOG_EXT_API(ipv6Handle, "rte_acl_reset_rules", " Removing all the rules of IPv6 PBR ACL context \n");
    rte_acl_reset_rules((struct rte_acl_ctx *)pbr_tree);
}



