/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_stats.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief  
 *
 * This file contains all the APIs provided by the IPV6 module.
 *
 ***********************************************************************/
/* Infrastructure header files */
#include "adk_include.h"
#include "adk_ph_dev.h"
#include "acp_gen_interface.h"
#include "adk_utils_tree_update.h"
#include <netinet/in.h>
#include <arpa/inet.h>

#include "adk_shm.h"

/* IPV6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_const.h"

/* IPV6 internal header files */
#include "adk_ipv6_internal.h"
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_dp_export.h"
#include "adk_ph_dev_dp.h"
#define ADK_CONVERT_TO_V6_VRD__HANDLE(__DATA) ((adk_ipv6_vrd_handle_t)__DATA)
#define ADK_CONVERT_TO_V6_FLOW__HANDLE(__DATA) ((adk_flow_handle_t)__DATA)


#ifdef ADK_DEBUG
/*!
 * =============================================================
 * adk_ipv6_global_statistics_reset
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully reset the global statistics.
 * @retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 *
 * @brief 
 * This API is not supported
 * 
 * Reset the global statistics.
 * This function reset all the global statistics
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_global_statistics_reset (
    adk_module_handle_t __attribute__((unused)) ipv6Handle)
{
    ADK_RETURN( ADK_SUCCESS);
}    /* -----  end of function adk_ipv6_global_statistics_reset  ----- */
#endif
/*!
 * =============================================================
 * adk_ipv6_global_statistics_get
 * @param[in] ipv6Handle IPV6 module handle. Returned at the
 * time of IPV6 module's registration with the module manager.
 * @param[out] ipv6GlobalStats_p The statistics returned to
 * the caller.
 *
 * @retval ADK_SUCCESS Successfully read the global statistics.
 * @retval ADK_E_INCONSISTENT_STATE The IPV6 module info pointer
 * in the module handle is NULL.
 * @retval ADK_E_INVALID_PARAMETER The input parameters are
 * NULL.
 *
 * @brief  Returns the global statistics.
 *
 * This function reads the statistics that do not belong to any
 * flow.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_global_statistics_get (
    adk_module_handle_t __attribute__((unused)) ipv6Handle,
    adk_ipv6_global_statistics_t __attribute__((unused)) *ipv6GlobalStats_p)
{
    adk_status_t                     retValue = ADK_IPV6_SUCCESS;
    uint32_t                         i = 0;
    adk_sys_handle_t                 sys_handle;
    adk_ipv6_info_t                  *ipv6ModuleInfo_p = NULL;
    adk_ipv6_dp_global_statistics_t  **global_stats = NULL;

    /* Handle validations */
    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Module Info is NULL or cookie mismatches\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    /*! It is assumed that the module handle is valid. */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    /*Check if IPv6 module info is not NULL*/
    if ( !ipv6GlobalStats_p)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Invalid input parameter\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_IPV6_E_INVALID_PARAMETER);
        goto adk_ipv6_get_global_stats_return;
    }
    
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Module info is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto adk_ipv6_get_global_stats_return;
    }

    ADK_MEMSET(ipv6GlobalStats_p, 0, sizeof(adk_ipv6_global_statistics_t));

    global_stats = ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_global_stats_table;

    for(i=0; i<ADK_MAX_NUM_CORES; i++)
    {

        /*! Convert the statistics to host format */
        ipv6GlobalStats_p->ipv6ReassemblyTimedoutPkts +=
             				(global_stats[i]->ipv6ReassTimedContextFailures);
        ipv6GlobalStats_p->ipv6ClassifierDefaultDroppedPkts +=
             				(global_stats[i]->noIngressAssociation);
        ipv6GlobalStats_p->ipv6PolicerDrop +=
             				(global_stats[i]->ipv6RateLimitingDrops);
        ipv6GlobalStats_p->ipv6ReassemblyError +=
             				(global_stats[i]->ipv6ReassErrContextFailures +
					 global_stats[i]->ipv6ReassTableFullFailures);
        ipv6GlobalStats_p->ipv6TooBigPackets +=
             				(global_stats[i]->ipv6OutFragLimitExceed +
					 global_stats[i]->ipv6FwdMtuExceed);
	ipv6GlobalStats_p->ipv6OutSrcIpInvalid +=
		                        (global_stats[i]->ipv6OutSrcIpInvalid);

        ipv6GlobalStats_p->ipv6RouteNotFoundPkts +=
                                        (global_stats[i]->routeNotFoundDrop);

        ipv6GlobalStats_p->ipv6NoNextHopEntryDrop +=
                                        (global_stats[i]->noNextHopEntryDrop);

        ipv6GlobalStats_p->ipv6MbufAllocFailures +=
                                        (global_stats[i]->ipv6MbufAllocFailures);

        ipv6GlobalStats_p->ipv6ReassTimerCreateFailures +=
                                        (global_stats[i]->ipv6ReassTimerCreateFailures);
        ipv6GlobalStats_p->ipv6InvalidIfFail +=
                                        (global_stats[i]->InvalidIfFail);
        ipv6GlobalStats_p->ipv6GenericDroppedPkts +=
                                        (global_stats[i]->ipv6GenericDroppedPkts);
    }
adk_ipv6_get_global_stats_return:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sys_handle, ipv6Handle);

    ADK_RETURN( ADK_SUCCESS);
}    /* -----  end of function adk_ipv6_global_statistics_get  ----- */
/*!
 * =============================================================
 * adk_ipv6_global_statistics_print
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 *
 * @brief  
 * This API prints the global statistics 
 *
 * =============================================================
 */
adk_status_t 
adk_ipv6_global_statistics_print( adk_module_handle_t ipv6Handle )
{
    adk_status_t                       retValue = ADK_SUCCESS;
    adk_ipv6_global_statistics_t       ipv6GlobalStats;

    ADK_MEMSET(&ipv6GlobalStats, 0, sizeof(adk_ipv6_global_statistics_t));

    retValue = adk_ipv6_global_statistics_get(ipv6Handle, &ipv6GlobalStats);
    if (retValue != ADK_SUCCESS)
        return retValue;

    ADK_INFO_WRITE("\nIPv6 Module Global Statistics\n" );

    ADK_IPV6_NON_ZERO_CNTR_WRITE("Reassembly Tmr exp Dropped PDUs                 : %d\n", ipv6GlobalStats.ipv6ReassemblyTimedoutPkts);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Defaut ingress classifier dropped PDUs          : %d\n", ipv6GlobalStats.ipv6ClassifierDefaultDroppedPkts);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Reassembly error context failures               : %d\n", ipv6GlobalStats.ipv6ReassemblyError);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("ipv6 Too bigger size packets dropped            : %d\n", ipv6GlobalStats.ipv6TooBigPackets);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Incoming Rate limit exceeded packet dropped     : %d\n", ipv6GlobalStats.ipv6PolicerDrop);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("IP not configured on interface packets dropped  : %d\n", ipv6GlobalStats.ipv6OutSrcIpInvalid);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("No Route found to transmit the packts           : %d\n", ipv6GlobalStats.ipv6RouteNotFoundPkts);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Next hop entry not present dropped packets      : %d\n", ipv6GlobalStats.ipv6NoNextHopEntryDrop);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Mbuf alloc failure                              : %d\n", ipv6GlobalStats.ipv6MbufAllocFailures);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Reassembly timer create failure                 : %d\n", ipv6GlobalStats.ipv6ReassTimerCreateFailures);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Invalid interface failure                       : %d\n", ipv6GlobalStats.ipv6InvalidIfFail);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Generic drop counters                           : %d\n", ipv6GlobalStats.ipv6GenericDroppedPkts);
    
    return ADK_SUCCESS;
}

/*!
* =============================================================
* adk_ipv6_address_stats_get
* @param[in]   ipHandle  IP module handle.
*
* @param[in]   ifHandle Interface on which this IP Address address is created,
*
* @param[in]   ipv6Address  IP Address to retrieve stats for.
*
* @param[out]  stats_p  pointer to adk_ipv6_address_stats_t
 * data structure.
*
* @retval ADK_SUCCESS Successful completion
*
* @retval ADK_E_INVALID_PARAMETER  invalid IP module handle or
* stats_p pointer specified.
*
* @retval ADK_E_STATS_READ_FAILURE  failed to read statistics
* from the hardware.
*
* @brief
*
* This function reads IP address stats and fill the adk_ipv6_address_stats_t
* point with appropriate values.
*
* =============================================================
*/

adk_status_t adk_ipv6_address_stats_get (
                adk_module_handle_t ipv6Handle,
                adk_ipv6_addr_prefix_t ipv6Address,
                adk_ipv6_address_stats_t *stats_p)
{
    adk_dp_status_t dp_st;
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t stats_id;
    adk_sys_handle_t sys_handle;
    adk_ipv6_info_t   *ipv6ModuleInfo_p = NULL;
    uint32_t ipv6_addr[4];
    adk_ipv6_dp_route_entry_t  *route_entry = NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid module handle %"ADKPRIh"\n", ipv6Handle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);

    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (ipv6ModuleInfo_p == NULL)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Module info is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        goto _adk_ipv6_addr_stats_get_exit;
    }

    /* Handle validations */
    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_IPV6_LOG_FORCE("Invalid IPv6 module handle %"ADKPRIh"\n", ipv6Handle);
        goto _adk_ipv6_addr_stats_get_exit;
    }

    if (!stats_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Invalid input NULL pointer parameter\n");
        goto _adk_ipv6_addr_stats_get_exit;
    }
    /* TODO: Here Route lookup has been done from Control Path, This will increase API latency, further this should be replaed with address stats table
     * here CP is doing route lookup, and it needs adk_ipv6_dp_tables pointer to do route lookup, as adk_ipv6_dp_tables pointer will get initialised 
     * only in DP process in adk_ipv6_dp_init function, so copying adk_ipv6_dp_tables pointer from DP to CP.
     */

    adk_ipv6_dp_tables = ipv6ModuleInfo_p->adk_ipv6_dp_tables;

    ipv6_addr[0] = adk_htonl(ipv6Address.Ipv6Addr.addr[0]);
    ipv6_addr[1] = adk_htonl(ipv6Address.Ipv6Addr.addr[1]);
    ipv6_addr[2] = adk_htonl(ipv6Address.Ipv6Addr.addr[2]);
    ipv6_addr[3] = adk_htonl(ipv6Address.Ipv6Addr.addr[3]);

    retValue = adk_ipv6_dp_route_entry_lookup(ADK_IPV6_LOCAL_ROUTING_TABLE,
        (uint8_t *) &ipv6_addr[0], &route_entry);

    if ( likely (retValue == ADK_SUCCESS) && likely(route_entry))
    {
        stats_id = route_entry->ipv6_addr_stats_id;
    }
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_CFG_ENTRY_NOT_FOUND);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "No existing route found for given IP %x.%x.%x.%x \n", ipv6Address.Ipv6Addr.addr[0],
                 ipv6Address.Ipv6Addr.addr[1], ipv6Address.Ipv6Addr.addr[2], ipv6Address.Ipv6Addr.addr[3]);
        goto _adk_ipv6_addr_stats_get_exit;
    }

    dp_st = adk_ipv6_dp_ip_addr_stats_get(ipv6Handle, stats_id, stats_p);

    if (dp_st != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_IPV6_LOG_ERROR(ipv6Handle, retValue, "Error in reading IP address stats from dataplane\n");
        goto _adk_ipv6_addr_stats_get_exit;
    }

_adk_ipv6_addr_stats_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:

    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);

    ADK_RETURN(retValue);
}


adk_status_t
adk_ipv6_dp_ip_addr_stats_get(adk_module_handle_t ipv6Handle,
                              uint16_t ipv6_addr_stats_id,
                              adk_ipv6_address_stats_t *stats_p)
{
    uint32_t i;
    adk_ipv6_info_t  *ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipv6Handle);
    adk_ipv6_dp_ip_addr_stats_t *dp_stats;

    ADK_MEMSET(stats_p, 0, sizeof(adk_ipv6_address_stats_t));


    for(i=0;i<ADK_MAX_NUM_CORES;i++) {
        dp_stats = &(ipv6ModuleInfo_p->adk_ipv6_dp_tables->
                ipv6AddrStatsTble[i][ipv6_addr_stats_id]);

        stats_p->ipv6InOctets += dp_stats->ipv6InOctets;
        stats_p->ipv6InPackets += dp_stats->ipv6InPackets;
        stats_p->ipv6OutOctets += dp_stats->ipv6OutOctets;
        stats_p->ipv6OutPackets += dp_stats->ipv6OutPackets;
        stats_p->ipv6InReassembledPackets += dp_stats->ipv6InReassembledPackets;
        stats_p->ipv6InErrors += dp_stats->ipv6InErrors;
    }

    return ADK_SUCCESS;
}

adk_status_t
adk_ipv6_l4counter_statistics_get(adk_module_handle_t ipv6Handle,
                                  adk_ipv6_l4counter_statistics_t *stats_p)
{
    uint32_t i;
    adk_ipv6_info_t  *ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET_INTERNAL(ipv6Handle);
    adk_ipv6_dp_l4counter_statistics_t  **dp_stats = NULL;

    ADK_MEMSET(stats_p, 0, sizeof(adk_ipv6_l4counter_statistics_t));

    dp_stats = ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_l4counter_stats_table;

    for(i=0;i<ADK_MAX_NUM_CORES;i++) {
        stats_p->ipv6TcpInPackets += (dp_stats[i]->ipv6TcpInPackets);
        stats_p->ipv6TcpInOctets += (dp_stats[i]->ipv6TcpInOctets);
        stats_p->ipv6TcpOutPackets += (dp_stats[i]->ipv6TcpOutPackets);
        stats_p->ipv6TcpOutOctets += (dp_stats[i]->ipv6TcpOutOctets);
        stats_p->ipv6SctpInPackets += (dp_stats[i]->ipv6SctpInPackets);
        stats_p->ipv6SctpInOctets += (dp_stats[i]->ipv6SctpInOctets);
        stats_p->ipv6SctpOutPackets += (dp_stats[i]->ipv6SctpOutPackets);
        stats_p->ipv6SctpOutOctets += (dp_stats[i]->ipv6SctpOutOctets);
    }

    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * adk_ipv6_flow_cahce_statistics_print
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 *
 * @brief  
 * This API prints the IPv6 flow cache statistics 
 *
 * =============================================================
 */
adk_status_t 
adk_ipv6_flow_cache_statistics_print( adk_module_handle_t ipv6Handle )
{
    adk_status_t ret = ADK_SUCCESS; 
    adk_ipv6_dp_flow_cache_stats_t lli_fc_stats;
    adk_ipv6_dp_flow_cache_stats_t uli_fc_stats;
    adk_sys_handle_t sys_handle;
    ADK_MEMSET(&lli_fc_stats, 0, sizeof(adk_ipv6_dp_flow_cache_stats_t));
    ADK_MEMSET(&uli_fc_stats, 0, sizeof(adk_ipv6_dp_flow_cache_stats_t));

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    ret = adk_ipv6_flow_cache_statistics_get(ipv6Handle, &lli_fc_stats, &uli_fc_stats);
    if(ret == ADK_SUCCESS)
    {
        ADK_LOG_INFO(sys_handle, ipv6Handle,"LLI Flow cache stats: \n");
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_route_hit     %"PRId64" \n",lli_fc_stats.fc_route_hit);     
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_term_hit      %"PRId64" \n",lli_fc_stats.fc_term_hit);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_frag_term_hit %"PRId64" \n",lli_fc_stats.fc_frag_term_hit);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_bind_term_hit %"PRId64" \n",lli_fc_stats.fc_bind_term_hit);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_missed        %d        \n",lli_fc_stats.fc_miss);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_flows_added   %d        \n",lli_fc_stats.fc_flows);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_flows_cleared %d        \n",lli_fc_stats.fc_cleared);
        ADK_LOG_INFO(sys_handle, ipv6Handle,"ULI Flow cache stats: \n");
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_route_hit     %"PRId64" \n",uli_fc_stats.fc_route_hit);     
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_term_hit      %"PRId64" \n",uli_fc_stats.fc_term_hit);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_frag_term_hit %"PRId64" \n",uli_fc_stats.fc_frag_term_hit);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_bind_term_hit %"PRId64" \n",uli_fc_stats.fc_bind_term_hit);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_miss_missed   %d        \n",uli_fc_stats.fc_miss);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_flows_added   %d        \n",uli_fc_stats.fc_flows);
        ADK_LOG_INFO(sys_handle, ipv6Handle," fc_flows_cleared %d        \n",uli_fc_stats.fc_cleared);
    }

    return ret;
}

/*!
 * =============================================================
 * adk_ipv6_flow_cahce_wcore_statistics_print
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 *
 * @retval ADK_IPV6_SUCCESS Successful completion
 * 
 *
 * @brief  
 * This API prints the IPv6 flow cache statistics for each worker core
 *
 * =============================================================
 */
adk_status_t 
adk_ipv6_flow_cache_wcore_statistics_print( adk_module_handle_t ipv6Handle )
{
    adk_status_t retValue = ADK_SUCCESS; 
    adk_ipv6_dp_flow_cache_stats_t   *lli_stats;
    adk_ipv6_dp_flow_cache_stats_t   *uli_stats;
    int i;
    adk_ipv6_info_t                  *ipv6ModuleInfo_p = NULL;
    adk_sys_handle_t sys_handle;

    /* Handle validations */
    if (!ipv6Handle)
    {
        ADK_FORCE_LOG_ERROR( "Module handle is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        return (retValue);
    }

     /*! Check if module handle is valid */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Invalid module handle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    if (!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        ADK_FORCE_LOG_ERROR( "Module Info is NULL or cookie mismatches\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sys_handle, ipv6Handle);

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        ADK_LOG_ERROR(sys_handle, ipv6Handle, "Module info is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        return (retValue);
    }

    ADK_LOG_INFO(sys_handle, ipv6Handle,"LLI Flow cache stats:\n");
    for(i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        lli_stats = &ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_lli_flow_cache_table[i].lli_fc_stats;
        if(lli_stats->fc_route_hit||
           lli_stats->fc_term_hit||        
           lli_stats->fc_frag_term_hit||
           lli_stats->fc_bind_term_hit||
           lli_stats->fc_miss||
           lli_stats->fc_flows||
           lli_stats->fc_cleared)

        {
            ADK_LOG_INFO(sys_handle, ipv6Handle," Core id %d \n", i); 
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_route_hit     %"PRId64" \n",lli_stats->fc_route_hit);     
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_term_hit      %"PRId64" \n",lli_stats->fc_term_hit);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_frag_term_hit %"PRId64" \n",lli_stats->fc_frag_term_hit);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_bind_term_hit %"PRId64" \n",lli_stats->fc_bind_term_hit);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_missed        %d        \n",lli_stats->fc_miss);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_flows_added   %d        \n",lli_stats->fc_flows);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_flows_cleared %d        \n",lli_stats->fc_cleared);
        }
    }

    ADK_LOG_INFO(sys_handle, ipv6Handle,"ULI Flow cache stats: \n");
    for(i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        uli_stats = &ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_uli_flow_cache_table[i].uli_fc_stats;
        if(uli_stats->fc_route_hit||
           uli_stats->fc_term_hit||        
           uli_stats->fc_frag_term_hit||
           uli_stats->fc_bind_term_hit||
           uli_stats->fc_miss||
           uli_stats->fc_flows||
           uli_stats->fc_cleared)

        {
            ADK_LOG_INFO(sys_handle, ipv6Handle," Core id %d \n", i); 
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_route_hit     %"PRId64" \n",uli_stats->fc_route_hit);     
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_term_hit      %"PRId64" \n",uli_stats->fc_term_hit);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_frag_term_hit %"PRId64" \n",uli_stats->fc_frag_term_hit);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_bind_term_hit %"PRId64" \n",uli_stats->fc_bind_term_hit);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_missed        %d        \n",uli_stats->fc_miss);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_flows_added   %d        \n",uli_stats->fc_flows);
            ADK_LOG_INFO(sys_handle, ipv6Handle,"  fc_flows_cleared %d        \n",uli_stats->fc_cleared);
        }
    }

    return retValue;
}
 /*!
 * end addtogroup _ipv6Functions  IPv6 Functions
 * @}
 */
