/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipv6_utility_counters.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief
 *
 * This file contains APIs to enable/disable ipv6 utility counters from application
 * and get APIs to copy DP counters to CP.
 *
 ***********************************************************************/

#include "adk_ipv6_utility_counters.h"
#include "adk_ipv6_dp_tables.h"

/*!
 * =============================================================
 *
 * adk_ipv6_set_utility_counters
 *
 * @param[in]  ipv6Handle  -  IPv6 module handle. Returned at the
 *                            time of IPv6 module's registration
 *                            with the module manager.
 * @param[in]   utility_cnt_state - Boolean variable to enable/disable
                                    utility counter.
 *
 * @brief   Function to configure utility counters.
 *
 * This function configures utility counters.On disabling the utility
 * counters, all the counter values will get reset to zero.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_set_utility_counters(adk_module_handle_t ipv6Handle, bool utility_cnt_state)
{
    adk_status_t                 ret = ADK_SUCCESS;
    adk_sys_handle_t             sys_handle;
    adk_ipv6_info_t              *ipv6ModuleInfo_p = NULL;

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv6 module handle %"ADKPRIh"\n",
                           ipv6Handle);
        ADK_IPV6_LOG_FORCE("Failed to set IPv6 utility counter flag 0x%x\n",
                           utility_cnt_state);
        return(ADK_E_INVALID_HANDLE);
    }

    sys_handle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, ret, "Module info is NULL\n");
        goto ipv6_set_utility_counters_return;
    }

    ADK_IPV6_LOG_FUNC_ENTRY(ipv6Handle);
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, ret);
    /* setting ipv6 utility counters flag */
    if (utility_cnt_state)
    {
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->isIpv6UtilityCntEnabled = utility_cnt_state;
    }
    else
    {
        ipv6ModuleInfo_p->adk_ipv6_dp_tables->isIpv6UtilityCntEnabled = utility_cnt_state;

        /* Resetting utility counters */
        for (uint32_t i = 0; i < ADK_MAX_NUM_CORES; i++)
        {
            memset(&(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_utility_counters[i]->ipv6IngressUtilityCounters), 0,
                   sizeof(ipv6_ing_utility_counters_t));

            memset(&(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_utility_counters[i]->ipv6EgressUtilityCounters), 0,
                   sizeof(ipv6_egress_utility_counters_t));
        }
    }

    ADK_IPV6_LOG_INFO(ipv6Handle,
                      "IPv6 utility counters flag set to %d\n",
                       ipv6ModuleInfo_p->adk_ipv6_dp_tables->isIpv6UtilityCntEnabled);

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPV6_LOG_FUNC_EXIT(ipv6Handle);

ipv6_set_utility_counters_return:
    return(ret);
}

/*!
 * =============================================================
 * @fn adk_ipv6_utility_counters_get
 *
 * @param[in]    ipsecHandle - IPv6 module handle.
 *
 * @param[out]   ipv6_utility_counters - IPv6 related utility counter data structure pointer
 *               gets set by the IPv6 ADK
 *
 * @retval ADK_SUCCESS - Successfully fetched utility counters
 *
 * @brief
 *
 * This API allows the application to retreive all IPv6 related utility counters
 * maintained by the IPv6 ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size ipv6_utility_counters_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_utility_counters_get(adk_module_handle_t ipv6Handle, ipv6_utility_counters_t *ipv6_utility_counters)
{
    adk_status_t                 ret = ADK_SUCCESS;
    adk_ipv6_info_t              *ipv6ModuleInfo_p = NULL;
    ipv6_utility_counters_t      **ipv6_dp_utility_counters = NULL;

    if (!adk_ipv6_valid_handle(ipv6Handle))
    {
        ADK_IPV6_LOG_FORCE("Invalid IPv6 module handle %"ADKPRIh"\n",
                           ipv6Handle);
        return(ADK_E_INVALID_HANDLE);
    }

    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);
    if (ipv6ModuleInfo_p == NULL)
    {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INCONSISTENT_STATE);
        ADK_IPV6_LOG_ERROR(ipv6Handle, ret, "Module info is NULL\n");
        goto ipv6_utility_counters_return;
    }

    ADK_MEMSET(&ipv6_utility_counters->ipv6IngressUtilityCounters, 0, sizeof(ipv6_ing_utility_counters_t));
    ADK_MEMSET(&ipv6_utility_counters->ipv6EgressUtilityCounters, 0, sizeof(ipv6_egress_utility_counters_t));

    ipv6_dp_utility_counters = ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_utility_counters;

    for (uint32_t i = 0; i < ADK_MAX_NUM_CORES; i++)
    {
        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxPktSendToIpsecLli +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxPktSendToIpsecLli);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxPktFwdToIpv6Uli +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxPktFwdToIpv6Uli);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxClassifierSendToEo +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxClassifierSendToEo);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxClassifierSendToHost +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxClassifierSendToHost);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxInvalidVersion +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxInvalidVersion);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxInvalidMcastMac +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxInvalidMcastMac);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxInvalidMcastGroup +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxInvalidMcastGroup);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxInvalidMcastScope +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxInvalidMcastScope);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxUcastDstIpMcastDstMacDrop +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxUcastDstIpMcastDstMacDrop);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxHopLimitZero +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxHopLimitZero);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxInvalidRouteAction +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxInvalidRouteAction);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxClassifierSendToEoFail +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxClassifierSendToEoFail);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxCLassifierSendToHostFail +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxCLassifierSendToHostFail);

        ipv6_utility_counters->ipv6IngressUtilityCounters.ipv6RxInvalidClassifierAction +=
                                   (ipv6_dp_utility_counters[i]->ipv6IngressUtilityCounters.ipv6RxInvalidClassifierAction);

        /* Copy Egress counters from DP to CP */
        ipv6_utility_counters->ipv6EgressUtilityCounters.ipv6TxSendToIpsecUli +=
                                   (ipv6_dp_utility_counters[i]->ipv6EgressUtilityCounters.ipv6TxSendToIpsecUli);

        ipv6_utility_counters->ipv6EgressUtilityCounters.ipv6TxSendToEthUli +=
                                   (ipv6_dp_utility_counters[i]->ipv6EgressUtilityCounters.ipv6TxSendToEthUli);

        ipv6_utility_counters->ipv6EgressUtilityCounters.ipv6PbrDisabled +=
                                   (ipv6_dp_utility_counters[i]->ipv6EgressUtilityCounters.ipv6PbrDisabled);

        ipv6_utility_counters->ipv6EgressUtilityCounters.ipv6TxRouteActionDrop +=
                                   (ipv6_dp_utility_counters[i]->ipv6EgressUtilityCounters.ipv6TxRouteActionDrop);

        ipv6_utility_counters->ipv6EgressUtilityCounters.ipv6TxInvalidRouteAction +=
                                   (ipv6_dp_utility_counters[i]->ipv6EgressUtilityCounters.ipv6TxInvalidRouteAction);

        ipv6_utility_counters->ipv6EgressUtilityCounters.ipv6TxhdrRoomNotAvail +=
                                   (ipv6_dp_utility_counters[i]->ipv6EgressUtilityCounters.ipv6TxhdrRoomNotAvail);
    }

ipv6_utility_counters_return:
    return ret;
}

/*!
 * =============================================================
 * @fn adk_ipv6_utility_counters_print
 *
 * @param[in]    ipv6Handle - IPv6 module handle.
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully printed utility counters
 *
 * @brief
 *
 * This API allows the application to print all IPv6 related utility counters
 * maintained by the IPv6 ADK module.
 *
 * =============================================================
 */
adk_status_t
adk_ipv6_utility_counters_print(adk_module_handle_t ipv6Handle)
{
    adk_status_t                 ret = ADK_SUCCESS;
    ipv6_utility_counters_t      ipv6_utility_counters;

    ret = adk_ipv6_utility_counters_get(ipv6Handle, &ipv6_utility_counters);

    if (ret != ADK_SUCCESS)
    {
        return ret;
    }

    ADK_INFO_WRITE("\nIPv6 Module Ingress utility counters Statistics\n" );

    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts forwarded to ipsec lli                         : %"PRId64"\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxPktSendToIpsecLli);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts forwarded to ipv6 uli                          : %"PRId64"\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxPktFwdToIpv6Uli);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts send to EO                                     : %"PRId64"\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxClassifierSendToEo);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts send to host                                   : %"PRId64"\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxClassifierSendToHost);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped due to invalid ip version              : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxInvalidVersion);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped bcz of invalid Mcast mac               : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxInvalidMcastMac);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped bcz of invalid Mcast group             : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxInvalidMcastGroup);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped bcz of invalid Mcast scope             : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxInvalidMcastScope);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped bcz of ucast DST IP has DST MAC        : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxUcastDstIpMcastDstMacDrop);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped due to hop limit value is zero         : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxHopLimitZero);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped due to invalid DIP table lookup action : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxInvalidRouteAction);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts send to EO failed                              : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxClassifierSendToEoFail);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts send to Host failed                            : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxCLassifierSendToHostFail);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Rx pkts dropped due to invalid Ing classifier action   : %d\n",
                                  ipv6_utility_counters.ipv6IngressUtilityCounters.ipv6RxInvalidClassifierAction);

    ADK_INFO_WRITE("\nIPv6 Module Egress utility counters Statistics\n" );
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Tx pkts forwarded to ipsec uli                                             : %"PRId64"\n",
                                  ipv6_utility_counters.ipv6EgressUtilityCounters.ipv6TxSendToIpsecUli);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Tx pkts forwarded to Eth lli                                               : %"PRId64"\n",
                                  ipv6_utility_counters.ipv6EgressUtilityCounters.ipv6TxSendToEthUli);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Tx pkts dropped due to pbr is disabled                                     : %d\n",
                                  ipv6_utility_counters.ipv6EgressUtilityCounters.ipv6PbrDisabled);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Tx pkts dropped due to route action is drop                                : %d\n",
                                  ipv6_utility_counters.ipv6EgressUtilityCounters.ipv6TxRouteActionDrop);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Tx pkts dropped due to invalid route action                                : %d\n",
                                  ipv6_utility_counters.ipv6EgressUtilityCounters.ipv6TxInvalidRouteAction);
    ADK_IPV6_NON_ZERO_CNTR_WRITE("Tx pkts dropped due no space created for fragment header                   : %d\n",
                                  ipv6_utility_counters.ipv6EgressUtilityCounters.ipv6TxhdrRoomNotAvail);

    return ret;
}
