/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_ipv6_vrf.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipv6Functions  IPv6 Functions
 *
 * @{
 *
 * @brief
 *
 * This file contains APIs to manage Virtual Routing Domains
 * creation and deletions.
 *
 ***********************************************************************/
/* infrastructure header files */
#include "adk_include.h"
#include "acp_gen_interface.h"


/* IPv6 external header files */
#include "adk_ipv6.h"
#include "adk_ipv6_dp.h"
#include "adk_ipv6_dp_tables.h"
#include "adk_ipv6_acl.h"
#include "adk_ipv6_const.h"

/* IPv6 internal header files */
#include "adk_ipv6_internal.h"



/*!
 * =============================================================
 * adk_ipv6_vrd_create
 *
 * @param[in] ipv6Handle      IPv6 module handle.
 * @param[in] vrdAttr       Pointer to adk_ipv6_vrd_t data structure.
 *                          which represents routing domian attributes.
 *
 * @param[out] ipv6VrdHandle_p Pointer to adk_ipv6_vrd_handle_t
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv6 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          One of the direct arguments
 * to the API call is invalid e.g. NULL pointer paramter
 *
 * @retval ADK_E_MEM_UNAVAILABLE            Failed to allocate memory for
 * virtual routing domain object
 *
 *
 * @brief
 *
 * This API creates a virtual routing domain based on the application
 * supplied domain attributes. The created routing domain handle is 
 * returned to the application. Application has to use this domain handle while
 * associating a IPv6 object (like IPoE Interface, Route Entry, Ingress
 * association etc..) with this routing domain.
 * =============================================================
 */

adk_status_t adk_ipv6_vrd_create(
    adk_module_handle_t ipv6Handle,
    const adk_ipv6_vrd_t  *vrdAttr_p,
    adk_ipv6_vrd_handle_t *ipv6VrdHandle_p)
{

    adk_status_t            	retValue = ADK_SUCCESS;
    adk_ipv6_vrd_info_t         *vrd_info_p = NULL;
    uint32_t            	vrdId = 0;
    adk_ipv6_info_t        *ipv6ModuleInfo_p=NULL;
    adk_sys_handle_t sysHandle;
    adk_ipv6_addr_prefix_t ipv6_addr_info;
    adk_ipv6_dp_route_entry_t route_entry;
    adk_ipv6_dp_ingress_flow_sip_entry_t ingress_flow_sip_entry;
     
    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
     }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);

       /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NUll or cookie mismatch \n");
        goto _adk_ipv6_vrd_create_release_lock_exit;
    }

   /*Check if input pointers are non NULL*/
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (!ipv6VrdHandle_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Invalid NULL pointer parameter\n");
        goto _adk_ipv6_vrd_create_exit;
    }

    if (!vrdAttr_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle,
                           "Invalid NULL pointer parameter\n");
        goto _adk_ipv6_vrd_create_exit;
    }

    vrdId = vrdAttr_p->vrd_id;

    if( vrdId >= ADK_IPV6_VRD_ID_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Invalid Vrd Id. Supported Vrd Id range is 0 - %x\n", ADK_IPV6_VRD_ID_MAX - 1);
        goto _adk_ipv6_vrd_create_exit;
    }

    if(ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_vrd_table.vrd_state[vrdId])
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                " Failed to allocate vrd namespace ID, Seems like it is already used \n");
        goto _adk_ipv6_vrd_create_exit;
    }

    ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_vrd_table.vrd_state[vrdId] = TRUE;

    ADK_LOG_INFO(sysHandle,ipv6Handle, "Allocated VRD ID 0x%x\n",
                      vrdId);

    vrd_info_p = (adk_ipv6_vrd_info_t *) 
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipv6Handle), ADK_IPV6_VRD_OBJ, 1, sizeof(adk_ipv6_vrd_info_t));

    if (vrd_info_p== NULL)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_FATAL_ERROR(sysHandle,ipv6Handle, 
                           "Memory allocation failed\n");
        goto _adk_ipv6_vrd_create_exit;
   } 

   /* We need to add static entries to route lookup tree and SIP look up tree with this 
    * particular VRD ID. For route look up tree, we will add the invalid destination IPv6
    * addresses and the output is a goto function.
    * For SIP lookup we will return the action as Drop.
    */

    /* ADD STATIC ENTRY FOR Route lookup & SIP LOOKUP 
     *   - unspecified IP address
     *   - Loopbak adress
     *   - Link local address
     *   - multicast address 
     */   

    ADK_MEMSET(&ipv6_addr_info, 0, sizeof(adk_ipv6_addr_prefix_t));
    ADK_MEMSET(&route_entry, 0, sizeof(adk_ipv6_dp_route_entry_t));

    /* unspecified IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0;
    ipv6_addr_info.Ipv6Addr.addr[1] = 0;
    ipv6_addr_info.Ipv6Addr.addr[2] = 0;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LENGTH);
    
    route_entry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_DROP_WITH_COUNT;

    retValue = adk_ipv6_dp_route_entry_add(ipv6Handle,
                                vrdId,
                                ipv6_addr_info,
                                &route_entry, FALSE);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }

    /* loopback IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[3] = 1;
    retValue = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                                vrdId,
                                ipv6_addr_info,
                                &route_entry, FALSE);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }

    /* loopback IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFE800000;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LINK_LOCAL_PREFIX_LENGTH);
    retValue = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                                vrdId,
                                ipv6_addr_info,
                                &route_entry, FALSE);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }

#if 0
    /* TODO - Multicast group support*/
    /* multicast IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFF000000;
    retValue = adk_ipv6_dp_route_entry_add(ipv6Handle, 
                                vrdId,
                                ipv6_addr_info,
                                &route_entry);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }
#endif

    /*ADD STATIC ENTRY FOR SIP LOOKUP */
    ADK_MEMSET(&ingress_flow_sip_entry, 0, sizeof(adk_ipv6_dp_ingress_flow_sip_entry_t));

    ipv6_addr_info.Ipv6Addr.addr[0] = 0;
    ipv6_addr_info.Ipv6Addr.addr[1] = 0;
    ipv6_addr_info.Ipv6Addr.addr[2] = 0;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;

    ipv6_addr_info.prefixLength = (ADK_IPV6_PDU_ADDR_LENGTH);

    ingress_flow_sip_entry.action = ADK_IPV6_ACTION_COUNT_AND_DROP;
    ingress_flow_sip_entry.apply_to_fwd_traffic = TRUE;


    /* unspecified IPv6 address */
    retValue = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                            vrdId,
                            ipv6_addr_info,
                            &ingress_flow_sip_entry);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }


    /* Loopback address */
    ipv6_addr_info.Ipv6Addr.addr[3] = 1;
    ingress_flow_sip_entry.apply_to_fwd_traffic = FALSE;
    retValue = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                            vrdId,
                            ipv6_addr_info,
                            &ingress_flow_sip_entry);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }

    /* Multicast address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFF000000;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength = (ADK_IPV6_PDU_ADDR_MULTICAST_PREFIX_LENGTH);
    retValue = adk_ipv6_dp_ingress_flow_sip_entry_add(ipv6Handle, 
                            vrdId,
                            ipv6_addr_info,
                            &ingress_flow_sip_entry);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_create_exit;
    }

    /* Store the VRD information in the internal data structure */
    vrd_info_p->cookie = ADK_IPV6_VRD_ID_COOKIE;
    vrd_info_p->vrdAttr.vrd_id = (uint32_t)vrdId;

    ipv6ModuleInfo_p->vrdHandleList[vrdId] = (adk_ipv6_vrd_handle_t) vrd_info_p;
    *ipv6VrdHandle_p = (adk_ipv6_vrd_handle_t) vrd_info_p;

_adk_ipv6_vrd_create_exit:
    if (retValue != ADK_SUCCESS){
        if (vrd_info_p)
        {
            ADK_NVM_FREE(vrd_info_p);
        }
     }
_adk_ipv6_vrd_create_release_lock_exit:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
    ADK_RETURN(retValue);

}    /* -----  end of function adk_ipv6_vrd_create ----- */



/*!
 * =============================================================
 * adk_ipv6_vrd_delete
 *
 * @param[in] ipv6Handle      IPv6 module handle.
 *
 * @param[out] ipv6VrdHandle  VRD handle
 *
 * @retval ADK_SUCCESS                      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE             Invalid IPv6 module handle
 *
 * @retval ADK_E_INVALID_PARAMETER          One of the direct arguments
 * to the API call is invalid e.g. NULL pointer paramter
 *
 * @retval ADK_E_MEM_UNAVAILABLE            Failed to allocate memory for
 * virtual routing domain object
 *
 *
 * @brief
 *
 * This API deletes a routing domain instance and frees up the resources.
 * =============================================================
 */
adk_status_t adk_ipv6_vrd_delete(
    adk_module_handle_t ipv6Handle,
    adk_ipv6_vrd_handle_t ipv6VrdHandle)

{
    adk_status_t        retValue = ADK_SUCCESS;
    adk_ipv6_vrd_info_t       *vrd_info_p;
    uint32_t        vrd_id;
    adk_ipv6_info_t    *ipv6ModuleInfo_p;
    adk_sys_handle_t sysHandle;
    adk_ipv6_addr_prefix_t ipv6_addr_info;
    adk_ipv6_dp_route_entry_t route_entry;
    adk_ipv6_dp_ingress_flow_sip_entry_t ingress_flow_sip_entry;

    /* Handle validations */
   
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
     }
       /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);
   
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);

    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NUll or cookie mismatch \n");
        goto _adk_ipv6_vrd_delete_exit;
    }


   /*Check if input pointers are non NULL*/
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (!adk_ipv6_vrd_valid_handle (ipv6VrdHandle, &vrd_id) || !ipv6VrdHandle)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle,
                           "Invalid VRD Handle type\n");
        goto _adk_ipv6_vrd_delete_exit;
    }
   
    vrd_info_p = (adk_ipv6_vrd_info_t *) ipv6VrdHandle;

   /* We need to remove static entries to route lookup tree and SIP look up tree with this 
    * particular VRD ID. For route look up tree, we will remove the invalid destination IPv6
    * addresses and the output is a goto function.
    * For SIP lookup we will return the action as Drop.
    */

    /* REMOVE STATIC ENTRY FOR Route lookup & SIP LOOKUP 
     *   - unspecified IP address
     *   - Loopbak adress
     *   - Link local address
     *   - multicast address 
     */   

    ADK_MEMSET(&ipv6_addr_info, 0, sizeof(adk_ipv6_addr_prefix_t));
    ADK_MEMSET(&route_entry, 0, sizeof(adk_ipv6_dp_route_entry_t));

    /* unspecified IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0;
    ipv6_addr_info.Ipv6Addr.addr[1] = 0;
    ipv6_addr_info.Ipv6Addr.addr[2] = 0;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LENGTH);
    
    route_entry.action = ADK_IPV6_ROUTE_LOOKUP_ACTION_DROP_WITH_COUNT;

    retValue = adk_ipv6_dp_route_entry_rem(ipv6Handle,
                                (uint8_t)vrd_id,
                                ipv6_addr_info, FALSE);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to rem the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }

    /* loopback IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[3] = 1;
    retValue = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                                (uint8_t)vrd_id,
                                ipv6_addr_info, FALSE);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to rem the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }

    /* loopback IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFE800000;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength     = (ADK_IPV6_PDU_ADDR_LINK_LOCAL_PREFIX_LENGTH);
    retValue = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                                (uint8_t)vrd_id,
                                ipv6_addr_info, FALSE);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to rem the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }

#if 0
    /* TODO Multicast group feature support */
    /* multicast IPv6 address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFF000000;
    retValue = adk_ipv6_dp_route_entry_rem(ipv6Handle, 
                                (uint8_t)vrd_id,
                                ipv6_addr_info);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to add the default route entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }
#endif
    /*ADD STATIC ENTRY FOR SIP LOOKUP */
    ADK_MEMSET(&ingress_flow_sip_entry, 0, sizeof(adk_ipv6_dp_ingress_flow_sip_entry_t));

    ipv6_addr_info.Ipv6Addr.addr[0] = 0;
    ipv6_addr_info.Ipv6Addr.addr[1] = 0;
    ipv6_addr_info.Ipv6Addr.addr[2] = 0;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;

    ipv6_addr_info.prefixLength = (ADK_IPV6_PDU_ADDR_LENGTH);

    ingress_flow_sip_entry.action = ADK_IPV6_ACTION_COUNT_AND_DROP;
    ingress_flow_sip_entry.apply_to_fwd_traffic = TRUE;


    /* unspecified IPv6 address */
    retValue = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                            (uint8_t)vrd_id,
                            ipv6_addr_info);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to rem the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }


    /* Loopback address */
    ipv6_addr_info.Ipv6Addr.addr[3] = 1;
    ingress_flow_sip_entry.apply_to_fwd_traffic = FALSE;
    retValue = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                            (uint8_t)vrd_id,
                            ipv6_addr_info);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to rem the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }

    /* Multicast address */
    ipv6_addr_info.Ipv6Addr.addr[0] = 0xFF000000;
    ipv6_addr_info.Ipv6Addr.addr[3] = 0;
    ipv6_addr_info.prefixLength = ADK_IPV6_PDU_ADDR_MULTICAST_PREFIX_LENGTH;
    retValue = adk_ipv6_dp_ingress_flow_sip_entry_rem(ipv6Handle, 
                            (uint8_t)vrd_id,
                            ipv6_addr_info);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle,
                ipv6Handle,
                "Failed to rem the default SIP entry \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INCONSISTENT_STATE); 
        goto _adk_ipv6_vrd_delete_exit;
    }

    ADK_LOG_INFO(sysHandle,ipv6Handle, "Freed VRD ID 0x%x\n", 
                      vrd_id);

    ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_vrd_table.vrd_state[vrd_id] = FALSE;
    ipv6ModuleInfo_p->vrdHandleList[vrd_id] = 0;
    vrd_info_p->cookie = 0;
    ADK_NVM_FREE(vrd_info_p);

_adk_ipv6_vrd_delete_exit:
    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);


}    /* -----  end of function adk_ipv6_vrd_delete ----- */

/*!
 *=============================================================
 * adk_ipv6_vrd_mode
 * 
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in]  ipv6VrdHandle - pointer to the IPv6 VRD handle
 *
 * @param[in]  host_mode - Enable/disable host mode per vrd
 *
 * @retval ADK_SUCCESS Successfully configured ADK IPv6 vrd host mode.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or VRD handle might be NULL.
 *
 * @brief
 *
 * This function enables/disables data forwarding (between TN ports) 
 * in dataplane per VRD without affecting outgoing terminating flows 
 * that use routing. 
 * =============================================================
 */
adk_status_t adk_ipv6_vrd_mode(adk_module_handle_t ipv6Handle,
                               adk_ipv6_vrd_handle_t ipv6VrdHandle,
                               bool host_mode)
{
    adk_status_t              retValue = ADK_SUCCESS;
    uint32_t              vrd_id;
    adk_ipv6_info_t           *ipv6ModuleInfo_p;
    adk_sys_handle_t          sysHandle;
    adk_ipv6_dp_vrd_host_mode_entry_t *vrd_host_mode_entry;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);   
    ADK_LOG_FUNC_ENTRY(sysHandle, ipv6Handle);
       /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NUll or cookie mismatch \n");
        goto _adk_ipv6_vrd_mode_exit;
    }

   /*Check if input pointers are non NULL*/
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (!adk_ipv6_vrd_valid_handle (ipv6VrdHandle, &vrd_id))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle,
                           "Invalid VRD Handle type\n");
        goto _adk_ipv6_vrd_mode_exit;
    }
  

    vrd_host_mode_entry = &ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_dp_vrd_table.tbl[vrd_id];

    if(host_mode == TRUE)
    {
        vrd_host_mode_entry->is_enabled     = TRUE;
        vrd_host_mode_entry->enable_routing = FALSE; /* routing disabled */
    }
    else
    {
        vrd_host_mode_entry->is_enabled     = FALSE;
        vrd_host_mode_entry->enable_routing = TRUE; /* routing enabled */
    }

    /* update the dp tables version */
    ipv6ModuleInfo_p->adk_ipv6_dp_tables->ipv6_fc_ver[vrd_id]++;

_adk_ipv6_vrd_mode_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ipv6Handle);
    ADK_RETURN(retValue);

}    /* -----  end of function adk_ipv6_vrd_mode ----- */


/*!
 *=============================================================
 * adk_ipv6_vrd_handle_get
 * 
 * @param[in]  ipv6Handle - IPv6 module handle.
 *
 * @param[in]  vrd_id - VRD ID for which handle is requested
 *
 * @param[out]  ipv6VrdHandle - pointer to the IPv6 VRD handle
 *
 * @retval ADK_SUCCESS Successfully configured ADK IPv6 vrd host mode.
 *
 * @retval ADK_E_INVALID_PARAMETER - Invalid input parameter. module
 * handle or VRD ID might be invalid
 *
 *
 * @brief
 *
 * This function returns VRD handle for the given VRD ID.
 * If the VRD ID is DEFAULT VRD ID, NULL Handle will be returned. 
 * =============================================================
 */
adk_status_t adk_ipv6_vrd_handle_get(
    adk_module_handle_t ipv6Handle,
    uint32_t vrd_id,
    adk_ipv6_vrd_handle_t *ipv6VrdHandle_p)
{
    adk_status_t              retValue = ADK_SUCCESS;
    uint32_t              vrdId;
    adk_ipv6_info_t           *ipv6ModuleInfo_p;
    adk_sys_handle_t          sysHandle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR( "Invalid IPv6 module handle \n");
        return (retValue);
    }

    if( vrd_id >= ADK_IPV6_VRD_ID_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR( "Invalid Vrd Id. Supported Vrd Id range is 0 - %x\n", ADK_IPV6_VRD_ID_MAX - 1);
        return (retValue);
    }
 
    /*! Aquire module lock */
    ADK_ACQUIRE_MODULE_LOCK(ipv6Handle, retValue);

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipv6Handle);   
    /*Validate Ipv6 private data and cookie */
    if(!adk_ipv6_valid_module_handle(ipv6Handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Module Info NUll or cookie mismatch \n");
        goto _adk_ipv6_vrd_handl_get_exit;
    }

    if (!ipv6VrdHandle_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle, 
                           "Invalid NULL pointer parameter\n");
        goto _adk_ipv6_vrd_handl_get_exit;
    }

    if( vrd_id == ADK_IPV6_DEFAULT_VRD_ID)
    {
        *ipv6VrdHandle_p = 0;
        goto _adk_ipv6_vrd_handl_get_exit;
    }

   /*Check if input pointers are non NULL*/
    ipv6ModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipv6Handle);

    if (!ipv6ModuleInfo_p->vrdHandleList[vrd_id] || 
        !adk_ipv6_vrd_valid_handle (ipv6ModuleInfo_p->vrdHandleList[vrd_id], &vrdId))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle,ipv6Handle,
                           "VRD Handle is not found for this VRD ID %x\n", vrd_id);
        goto _adk_ipv6_vrd_handl_get_exit;
    }

    *ipv6VrdHandle_p = ipv6ModuleInfo_p->vrdHandleList[vrdId];

_adk_ipv6_vrd_handl_get_exit:

    ADK_RELEASE_MODULE_LOCK(ipv6Handle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}

 /*!
 * end addtogroup _ipv6Functions  IPv6 Module
 * @}
 */
