/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_twamp.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _twamp  TWAMP Module
 * @{
 *
 * @brief 
 *
 * TWAMP module main header file
 *
 * This file contains all data types and API prototypes for TWAMP module.
 *
 ***********************************************************************/
#ifndef __ADK_TWAMP_H__
#define __ADK_TWAMP_H__

#include "adk_sys.h"

/*!
 * adk_twamp_responder_handle_t
 * TWAMP responder handle. 
 * This handle is returned at the time of TWAMP responder creation.
 */
typedef adk_flow_handle_t adk_twamp_responder_handle_t;


/*!
 * =============================================================
 * @struct adk_twamp_responder_flow_info_t adk_twamp.h "adk_twamp.h"
 *
 * @brief  TWAMP responder flow information
 *
 * =============================================================
 */
typedef struct adk_twamp_responder_info_s
{
    /*! The responder is created to handle either IPv4 or IPv6 test-sessions*/
    uint16_t        ip_version;
    /*! This is the session time-out value to be applied for 
     *  all test-sessions learned under this  responder context.
     *  the unit is in seconds. 
     */
    uint16_t        session_timeout;
    /*! Reserved for future use
     */
    uint32_t        reserved;
} adk_twamp_responder_info_t;


/*!
 * =============================================================
 * @struct adk_twamp_responder_stats_t adk_twamp.h "adk_twamp.h"
 *
 * @brief  TWAMP Responder stats 
 * 
 * =============================================================
 */
typedef struct adk_twamp_responder_stats_s
{
    /*! Total number of packet received from IP layer */
    uint32_t        rxPkts;
    /*! Total number of packet transmitted to IP layer*/
    uint32_t        txPkts;
    /*! Total number of received packet that are dropped due to invalid TWAMP packet size.
     *  that is TWAMP packet size (including header) less-than 41Bytes. 
     */
    uint32_t        pktLengthErrors;
    /*! Number of active test sessions under this responder.
     * Note:
     * a. This counter is not maintained in the fast path. It is managed
     * within adk_twamp_responder_stats_get() API. Due to linux context switch,
     * this counter might provide inaccurate values. For accurate values,
     * application must call the adk_twamp_responder_stats_get() API repeatedly,
     * till 2 successive reads provides the same numActiveSessions
     * counter value.
     */
    uint32_t        numActiveSessions;
    /*! Total number of packets dropped due to number of active test sessions (system wide) reached
     * maximum global limit.
     */
    uint32_t        pktsMaxSessions;
} adk_twamp_responder_stats_t;


/*!
 * =============================================================
 * @struct adk_twamp_global_stats_t adk_twamp.h "adk_twamp.h"
 *
 * @brief  TWAMP Global stats 
 * 
 * =============================================================
 */
typedef struct
{
    /*! Total number of packets received from IP layer.
     */
    uint32_t        rxPkts;
    /*! Total number of packets transmitted to IP layer.
     */
    uint32_t        txPkts;
    /*! number of packets dropped sue to invalid responder.
     *  either responder is not configured or deleted.
     */
    uint32_t        invalidRspndDrops;
    /*! Total number of packets dropped due to length check failed.
     *  twamp lenght < 41 bytes.
     */
    uint32_t        invalidLengthDrops;
    /*!
     *  This counter is not used. It is left as is for backward compatibility
     *  reason.
     */
    uint32_t        hashErrorDrops;
    /*! Total number of packets dropped due to TWAMP to IP bind not defined.
     */
    uint32_t        noBindDrops;

} adk_twamp_global_stats_t;

/*!
 * =============================================================
 * @enum adk_twamp_table_group_t
 *
 * @brief
 *
 * This enumeration defines different group of tables that can be
 * dumped in user readable csv format.
 * =============================================================
 */
typedef enum adk_twamp_table_group_s {
    /*! Non inclusive lower bound on table group type */
    ADK_TWAMP_TABLES_NONE = 0,
    /*! Responder lookup table */
    ADK_TWAMP_RESPONDER_TABLE = (1 << 0),
    /*! Custom bind table */
    ADK_TWAMP_CUSTOM_BIND_TABLE = (1 << 1),
    /* Inclusive upper bound on table group type for all tables */
    ADK_TWAMP_TABLES_ALL = ((1 << 2) - 1)
} adk_twamp_table_group_t;


/*!
 *  @addtogroup _twampFunctions  TWAMP Functions
 *  @{
 */
/* TWAMP API prototypes */

/*
 * TWAMP module init API.
 */
adk_status_t
adk_twamp_init(adk_module_handle_t twampHandle);

adk_status_t
adk_twamp_service_function_update (adk_module_handle_t twampHandle);

/*
 * API to create TWAMP responder.
 */
adk_status_t
adk_twamp_responder_create(adk_module_handle_t         twampHandle,
                     const adk_twamp_responder_info_t  *twampRespndInfo,
                     adk_twamp_responder_handle_t      *twampRespndHandle);

/* API to delete previously created TWAMP responder
 */
adk_status_t
adk_twamp_responder_delete(adk_module_handle_t        twampHandle,
                     adk_twamp_responder_handle_t     twampRespndHandle);

adk_status_t
/*
 * To get TWAMP responder level statistics
 */
adk_twamp_responder_stats_get(adk_module_handle_t     twampHandle,
                       adk_twamp_responder_handle_t   twampRespndHandle,
                       adk_twamp_responder_stats_t    *twampRespndStats);

/*
 * To get TWAMP global statistics
 */
adk_status_t
adk_twamp_global_stats_get(adk_module_handle_t        twampHandle,
                       adk_twamp_global_stats_t       *twampGlobalStats);

adk_status_t
adk_twamp_responder_stats_print(adk_module_handle_t   twampHandle,
                       adk_twamp_responder_handle_t   twampRespndHandle);

adk_status_t
adk_twamp_global_stats_print(adk_module_handle_t       twampHandle);

adk_status_t 
adk_twamp_tables_dump(adk_module_handle_t twampHandle, 
                    adk_twamp_table_group_t table_group);


adk_status_t adk_twamp_dp_init(void);

#endif

/*!
 *      @} end addtogroup _twampFunctions
 *
 * @} end addtogroup _twamp
 *
 */
