/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_twamp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _twamp  TWAMP Module
 * @{
 *      @addtogroup _twampFunctions  TWAMP Functions
 *      @{
 *
 * @brief TWAMP module initialization and shutdown APIs
 *
 *
 ***********************************************************************/

#include "adk_include.h"
#include "adk_twamp.h"
#include "adk_twamp_dp.h"
#include "adk_twamp_config.h"
#include "adk_twamp_internal.h"
#include <adk_rcu_distributed_util.h>
#include <sys/time.h>
#include "adk_shm.h"

/* TODO: Currently just stub code returning OK */
#define ADK_CONVERT_TO_TWAMP_RESP_HANDLE(__DATA) ((adk_twamp_responder_handle_t)(__DATA))

static void
adk_twamp_responder_stats_read(adk_module_handle_t twampHandle,
                               uint32_t respndId,
                               adk_twamp_responder_stats_t  *twampRespndStats_p);
/*!
 * =============================================================
 * adk_twamp_init
 *
 * @param[in]  twampHandle    TWAMP module handle whihc was created at
 * time of TWAMO module's registration with SIF module manager.
 *
 * @retval  ADK_SUCCESS     Successfully initialized TWAMP module.
 *
 * @retval ADK_E_INVALID_HANDLE    Invalid TWAMP module handle.
 *
 * @retval ADK_E_NS_RSRC_ALLOCATE_FAILED Resource allocation failed.
 *
 * @brief
 *
 * This function initializes the ADK TWAMP module. This function should
 * be passed as an argument to adk_module_register() SIF API at the
 * time of TWAMP module registration. This function must not be called
 * directly by the application. It gets called automatically by SIF
 * when module is started.
 * =============================================================
 */
adk_status_t
adk_twamp_init(adk_module_handle_t twampHandle)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_module_service_fn_t     svcFn;
    adk_twamp_module_info_t     *twampModuleInfo_p;
    adk_sys_handle_t sysHandle;
    adk_twamp_dp_pvt_data_t     *dp_pvt_p;
    int i;


    if ( !ADK_MODULE_HANDLE_VALIDATE(twampHandle) )
    {
        return (ADK_E_INVALID_HANDLE);
    }

    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    /* Check whether the module was already initialized */
    if (adk_twamp_valid_module_handle(twampHandle))
    {
        ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);
        return(ADK_SUCCESS);
    }

    /*
     * Allocate memory to store the module specific information.
     */
    twampModuleInfo_p = (adk_twamp_module_info_t *) ADK_NVM_CALLOC (ADK_MODULE_MODULEID_GET(twampHandle), 0,
						 1, sizeof(adk_twamp_module_info_t));
    if ( !twampModuleInfo_p )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _adk_twamp_init_exit;
    }

    ADK_MODULE_INFO_PTR_SET(twampHandle, twampModuleInfo_p);

    twampModuleInfo_p->cookie = ADK_TWAMP_MODULE_HANDLE_COOKIE;

    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleTerminate = adk_twamp_shutdown;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_TWAMP_LLI_FN_ID;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_RING] = ADK_TWAMP_LLI_RING_FN_ID;
    svcFn.srcModuleBind    = (module_bind_fn_t) adk_twamp_custom_bind;
    svcFn.srcModuleUnbind  = (module_bind_fn_t) adk_twamp_custom_unbind;
    svcFn.dstModuleBind    = (module_bind_fn_t) NULL;
    svcFn.dstModuleUnbind  = (module_bind_fn_t) NULL;
    svcFn.moduleLogSet     = adk_twamp_sys_log;

    twampModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_TWAMP_LLI_FN_ID;
    twampModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_RING] = ADK_TWAMP_LLI_RING_FN_ID;

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    twampModuleInfo_p->responder_rm_hdl =
                    adk_rm_resource_init(ADK_MODULE_MODULEID_GET(twampHandle), 0,
				TWAMP_MAX_PPF_RESPONDER);

    /* Initializing the toggle flag */
    for(i = 0; i < TWAMP_MAX_PPF_RESPONDER; i++)
    {
        twampModuleInfo_p->responderToggleFlag[i] = 0;
    }

    /* Create memzone to store the structure which will have pointers to all
     * the data plane tables */
    retValue = adk_shm_create(ADK_TWAMP_SHM_MEMZONE_NAME,
                        sizeof(adk_twamp_dp_pvt_data_t),
                        ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                        (adk_shm_handle_t **)&dp_pvt_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_FATAL_ERROR(sysHandle, twampHandle,
                          "Failed to create TWAMP memzone %s\n",\
                          ADK_TWAMP_SHM_MEMZONE_NAME);
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto _adk_twamp_init_exit;
    }

    /* initialize twamp dataplane module */
    retValue = adk_twamp_dp_table_init(twampHandle, dp_pvt_p);
    if (retValue != ADK_SUCCESS)
    {
      retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
      ADK_LOG_FATAL_ERROR(sysHandle, twampHandle,
                    "Failed to initialize the dataplane tables \n");
    }

    retValue = adk_module_service_function_register(sysHandle,
                                                    twampHandle, &svcFn);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                    "Failed to register the service functions\n");
        goto _adk_twamp_init_exit;
    }

    /* Hook up with TWAMP data plane tables to the module handle structure.
     * All the control plane access to the DP tables will be done via module
     * handle in the CP APIs. */
    twampModuleInfo_p->dp_pvt_p = dp_pvt_p;

#if 0
    retValue = adk_pal_stage_register("adk_twamp_dp_pkt_proc_lli",
                                      ADK_PAL_STAGE_LB_TYPE_ATOMIC,
                                      ADK_TWAMP_LLI_FN_ID,
                                      ADK_TWAMP_LLI_FN_ID);
    if (retValue != ADK_SUCCESS) {
        printf("adk_pal_stage_register Failed!\n");
        goto _adk_twamp_init_exit;
    }
#endif

_adk_twamp_init_exit:

    if (retValue != ADK_SUCCESS)
    {
        if (twampModuleInfo_p != NULL)
        {
            if (twampModuleInfo_p->responder_rm_hdl != NULL)
            {
                ADK_NVM_FREE(twampModuleInfo_p->responder_rm_hdl );
            }
            ADK_NVM_FREE(twampModuleInfo_p);
        }
    }
    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}

/*!
 * =============================================================
 *  adk_twamp_service_function_update
 *
 * @param[in]  twampHandle  - TWAMP module handle. Return at the
 * time of TWAMP module's registration with the module manager.
 *
 * @retval ADK_SUCCESS Successfully initialized TWAMP module.
 *
 * @brief
 *
 * This function registers the TWAMP module's service functions
 * with SIF and internally used function pointers. This function
 * should be passed as an argument to adk_module_handle_get() SIF API
 * when retrieving TWAMP module handle from a different process for
 * the first time. This function must not be called directly by
 * the application.
 * =============================================================
 */
adk_status_t
adk_twamp_service_function_update (adk_module_handle_t twampHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    adk_module_service_fn_t svcFn;
    adk_twamp_module_info_t *twampModuleInfo_p;

    /* Validate module handle */
    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle)) {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                             twampHandle);
        ADK_FORCE_LOG_ERROR("Service function update failed\n");
        return(ADK_E_INVALID_HANDLE);
    }

    /* Validate module handle */
    if (!adk_twamp_valid_module_handle(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                             twampHandle);
        ADK_FORCE_LOG_ERROR("Service function update failed\n");
        return(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    ADK_TWAMP_LOG_FUNC_ENTRY(twampHandle);

    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleTerminate = adk_twamp_shutdown;
    svcFn.moduleLogSet    = adk_twamp_sys_log;
    svcFn.srcModuleBind   = (module_bind_fn_t) adk_twamp_custom_bind;
    svcFn.srcModuleUnbind = (module_bind_fn_t) adk_twamp_custom_unbind;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] =
                  twampModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_DEFAULT];
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_RING] =
                  twampModuleInfo_p->cpuEntryFn_LLI[_DEST_FUNC_RING];

    retValue = adk_module_service_function_update(sysHandle, twampHandle,
                                                  &svcFn);

    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, twampHandle, "Failed to update service functions\n");
    }

    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);


_lock_release_failed:
_lock_acquire_failed:
    ADK_TWAMP_LOG_FUNC_EXIT(twampHandle);
    return retValue;
}

/*!
 * =============================================================
 * adk_twamp_responder_create
 *
 * @param[in] twampHandle       TWAMP module handle
 * @param[in] twampRspndInfo_p      Pointer to TWAMP responder setup info
 * @param[out] twampRspndHandle_p   Pointer to TWAMP responder flow handle
 *
 * @retval      ADK_SUCCESS
 * @retval      ADK_E_INVALID_PARAMETER
 * @retval      ADK_E_MEM_UNAVAILABLE
 * @retval      ADK_E_NS_RSRC_ALLOCATE_FAILED
 *
 * @brief
 *
 * Create a TWAMP responder based on the configuration info and
 * and return a pointer to the flow handle.
 * =============================================================
 */
adk_status_t
adk_twamp_responder_create(adk_module_handle_t         twampHandle,
                     const adk_twamp_responder_info_t  *twampRspndInfo_p,
                     adk_twamp_responder_handle_t      *twampRspndHandle_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    uint32_t                twampRspndID;
    adk_twamp_module_info_t *twampModuleInfo_p;
    struct timeval          timeVal;
    uint64_t                cpuTicksPerSec;
    adk_sys_handle_t sysHandle;
    void **pending_list = NULL;
    adk_twamp_pend_list_callback_data_t cb;

    /* Check validity of the input parameters */
    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle)) {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                             twampHandle);
        return(ADK_E_INVALID_HANDLE);
    }

    /* Validate passed  handle */
    if (!adk_twamp_valid_module_handle(twampHandle)) {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                             twampHandle);
        return(ADK_E_INVALID_HANDLE);
    }

    /* Serialize the function */
    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    if (( twampRspndInfo_p == NULL) ||
        (twampRspndHandle_p == NULL))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_twamp_responder_create_exit;
    }

    *twampRspndHandle_p = 0x00;
    /* Allocate a flow handle */
    ADK_SYS_FLOW_HANDLE_ALLOC (twampHandle, 0,
                               sizeof(adk_twamp_responder_info_t),
                               twampRspndHandle_p);
    if (*twampRspndHandle_p == 0) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _adk_twamp_responder_create_exit;
    }

    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    pending_list = (void **)twampModuleInfo_p->pend_list[ADK_TWAMP_RESPONDER_TBL].list;

    cb.tables = ADK_TWAMP_RESPONDER_TBL;
    cb.handle = twampHandle;
    ADK_RCU_DISTRIBUTED_PENDING_LIST_ARRAY_CHECK(pending_list,
            twampModuleInfo_p->pend_list[ADK_TWAMP_RESPONDER_TBL].num_elements,
                                           adk_twamp_rcu_pend_list_clear,
                                           &cb);

    if(twampModuleInfo_p->numActiveResponders == TWAMP_MAX_PPF_RESPONDER)
    {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                    "Exceeded the maximum responder limit \n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_ALLOCATE_FAILED);
        goto _adk_twamp_responder_create_exit;
    }

    twampRspndID =
        adk_rm_resource_reserve(twampModuleInfo_p->responder_rm_hdl);

    /* Initialize TWAMP flow handle. Store the responder Id */
    ADK_SYS_FLOWID_SET(*twampRspndHandle_p, twampRspndID);

    /*
     * Get the number of seconds and microseconds since the Epoch from linux
     */
     if(gettimeofday(&timeVal, NULL)) {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                    "Retrieval of time since the Epoch failed \n");
     }

    /*
     * Seconds since 1st Jan 1900 in NTP format. This will overflow in the
     * year 2036.
     */
    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].baseNtpTimeHi =
                                        timeVal.tv_sec + TWAMP_NTP_BASETIME;

    /* Seconds fraction since 1st Jan 1900 in NTP format. The granularity is at
     * the order 200 pico seconds. 1 sec = 200 picoseconds * 2^32 - 1 */
    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].baseNtpTimeLo =
              ((uint32_t)((timeVal.tv_usec * TWAMP_NTP_SCALE_FRAC)/1000000));


    /* Get the current clock ticks. This clock ticks does not map exactly with
     * the Epoch time. If a context switch happens before reading the clock
     * ticks, then the difference between linux time and CPU clock ticks will be
     * even more. The assumption is that the TWAMP session originator or the
     * sender will run without any preemption and correct the senders clock
     * periodically to take care of these timestamp errors.
     */
    ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_get_tsc_cycles", "Get "
                          "the number of TSC cycles since boot \n");
    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].epochCpuTicks =
                                                          rte_get_tsc_cycles();

    ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_get_tsc_hz", "Get "
                          "the measured frequency of RDTSC counter \n");
    cpuTicksPerSec = rte_get_tsc_hz();

    /* Required granularity is atleast 1 micro second */
    if(cpuTicksPerSec < 1000000) {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                      "CPU frequency is less than 1MHz \n");
        ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);
        return ADK_E_INTERNAL_ERROR;
    }

    /* Setting the time in seconds */
    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].session_timeout =
                                                twampRspndInfo_p->session_timeout;

    /* The timer event in ticks is set to 1/2 second, as the minimum session time
     * out interval is 1 second.
     */
    ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_get_tsc_hz", "Get "
                          "the measured frequency of RDTSC counter \n");
    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].timerEventInterval =
                                                             rte_get_tsc_hz()/2;

    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].ip_version =
                                            twampRspndInfo_p->ip_version;

    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].cpuTicksPerSec =
                                                             cpuTicksPerSec;

    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].isValid = 1;
    twampModuleInfo_p->numActiveResponders++;

#if ADK_DEBUG
    ADK_LOG_INFO(sysHandle, twampHandle,
                      "Created Responder ID = %x \n", twampRspndID);
#endif

_adk_twamp_responder_create_exit:
    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/*!
 * =============================================================
 * adk_twamp_responder_delete
 *
 * @param[in]   twampHandle         TWAMP module handle
 * @param[in]   twampFlowHandle     TWAMP flow handle
 *
 * @retval      ADK_SUCCESS
 * @retval      ADK_TWAMP_E_INVALID_PARAMETER
 * @retval      ADK_TWAMP_E_OUT_OF_RSRC
 * @retval      ADK_TWAMP_E_RSRC_FREE_FAIL
 *
 * @brief
 *
 * Free all resources associated with a TWAMP responder
 *
 * Note:
 *
 * 1. The IP->TWAMP and TWAMP->IP un-bind must be done before deleting the
 *    responder
 *
 * 2. The session age-out is implemented based on the responder level periodic
 * timer that is installed in the fast path during responder creation. The
 * periodic timer is set to minimum_session_age_out_time/2 => 1/2 => 0.5 seconds.
 * Deleting the responder does not un-install the responder level periodic timer
 * immediately in fast path. As ADK-TWAMP relies on auto-aging of the responder
 * level timer(i.e. The responder is marked invalid and, when the timer handler
 * is called in next timer expiry(0.5 seconds), the periodic timer is stopped.
 * So, the application must not recreate the same responder within 0.5 seconds.
 *
 * For example: If the application is creating a responder ID '1', then it must
 * not delete and re-create the same responder ID '1' within 0.5 seconds.
 * =============================================================
 */
adk_status_t
adk_twamp_responder_delete(adk_module_handle_t        twampHandle,
                     adk_twamp_responder_handle_t     twampRspndHandle_p)
{
    adk_status_t    retValue = ADK_SUCCESS;
    uint32_t    twampRspndID;
    adk_twamp_module_info_t *twampModuleInfo_p;
    adk_sys_handle_t sysHandle;
    void **pending_list = NULL;
    adk_twamp_pend_list_callback_data_t cb;

    /* Check validity of the input parameters */
    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle)) {
        ADK_FORCE_LOG_ERROR( "Invalid TWAMP module handle %"ADKPRIh"\n", twampHandle);
        return(ADK_E_INVALID_HANDLE);
    }

    /* Validate passed handle */
    if (!adk_twamp_valid_module_handle(twampHandle)) {
        ADK_FORCE_LOG_ERROR( "Invalid TWAMP module handle %"ADKPRIh"\n", twampHandle);
        return(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    if (!ADK_SYS_VALID_FLOW_HANDLE(twampRspndHandle_p)) {
        ADK_LOG_ERROR(sysHandle, twampHandle,
             "Invalid TWAMP Responder handle %"ADKPRIh"\n", twampRspndHandle_p);
        ADK_LOG_ERROR(sysHandle, twampHandle,
                      "Failed to delete Responder Handle\n");
        return(ADK_E_INVALID_HANDLE);
    }

    /* Serialize the function */
    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    pending_list = (void **)twampModuleInfo_p->pend_list[ADK_TWAMP_RESPONDER_TBL].list;

    cb.tables = ADK_TWAMP_RESPONDER_TBL;
    cb.handle = twampHandle;
    ADK_RCU_DISTRIBUTED_PENDING_LIST_ARRAY_CHECK(pending_list,
        twampModuleInfo_p->pend_list[ADK_TWAMP_RESPONDER_TBL].num_elements,
                                           adk_twamp_rcu_pend_list_clear,
                                           &cb);

    twampRspndID = ADK_SYS_FLOWID_GET(twampRspndHandle_p);

    /*
     * Marking the absolute responder as invalid. This will block the traffic in
     * all the sessions within this responder in data plane.
     */
    twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].isValid = 0;

    /* Check for inflight packets in data plane. The sessions have to be marked
     * invalid only after allowing all the inflight packets to next module.
     */

    ADK_RCU_RELEASE_DISTRIBUTED_PENDING_LIST_ARRAY(pending_list, twampRspndHandle_p,
                 twampModuleInfo_p->pend_list[ADK_TWAMP_RESPONDER_TBL].num_elements,
                                                 ADK_TWAMP_RESPONDER_TBL_THRESHOLD,
                                             adk_twamp_rcu_pend_list_clear, &cb);


    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/* Internal callback function to clear the pending list */
void
adk_twamp_rcu_pend_list_clear(void *rock, void *data)
{
    uint32_t i , relative_max_sessions;
    adk_twamp_responder_handle_t     twampRspndHandle_p;
    adk_twamp_pend_list_callback_data_t *cb_data;
    adk_twamp_module_info_t *twampModuleInfo_p;
    uint32_t    twampRspndID;
    adk_module_handle_t twampHandle;
    adk_sys_handle_t sysHandle;

    cb_data = (adk_twamp_pend_list_callback_data_t *)rock;
    twampHandle = cb_data->handle;

    /* Check validity of the input parameters */
    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle)) {
        ADK_FORCE_LOG_ERROR( "Invalid TWAMP module handle %"ADKPRIh"\n", twampHandle);
        return;
    }

    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    switch (cb_data->tables) {

    case ADK_TWAMP_RESPONDER_TBL:
        twampRspndHandle_p = (adk_twamp_responder_handle_t) data;

        ADK_LOG_INFO(sysHandle, twampHandle, "Clearing Twamp Responder "\
                     "pending list.\n");
        twampRspndID = ADK_SYS_FLOWID_GET(twampRspndHandle_p);

        /* Marking all the sessions as unused*/
        relative_max_sessions = (twampRspndID * TWAMP_MAX_REFLECTORS_PPF_RESPONDER) + TWAMP_MAX_REFLECTORS_PPF_RESPONDER;
        for(i = twampRspndID * TWAMP_MAX_REFLECTORS_PPF_RESPONDER; i < relative_max_sessions; i++) {
            twampModuleInfo_p->dp_pvt_p->sessionReflectorObj_p[i].seqNum = 0;
            /* Free all the Session IDs */
            if(twampModuleInfo_p->dp_pvt_p->sessionReflectorObj_p[i].isUsed) {
                twampModuleInfo_p->dp_pvt_p->sessionReflectorObj_p[i].isUsed = 0;
                if(adk_rm_resource_free(twampModuleInfo_p->dp_pvt_p->responderObj_p[twampRspndID].session_reflector_rm_hdl,
                                        twampModuleInfo_p->dp_pvt_p->sessionReflectorObj_p[i].sessionId ) == FALSE) {
                    ADK_LOG_ERROR(sysHandle, twampHandle,
                                  "Resource free failed for session id:%d Responder id:%d\n", i, twampRspndID );
                    return;
                }
            }
        }

        /* Reset the Responder stats */
        ADK_MEMSET(&twampModuleInfo_p->dp_pvt_p->dpTwampResponderStats_p[twampRspndID],
                   0, sizeof(adk_twamp_responder_stats_t));

        /* Reset fast path specific Responder Stats */
        twampModuleInfo_p->dp_pvt_p->dpTwampInternalResponderStats_p[twampRspndID].numActiveSessions = 0;

        adk_rm_resource_free(twampModuleInfo_p->responder_rm_hdl, twampRspndID);
        twampModuleInfo_p->numActiveResponders--;
        break;

    default:
        break;
    }

}

/*!
 * =============================================================
 * adk_twamp_responder_stats_get
 *
 * @param[in]  twampHandle      TWAMP module handle
 * @param[in]  twampFlowHandle  TWAMP responder flow handle
 * @param[out] twmpRspndInfo      Pointer to TWAMP responder info
 *
 * @retval      ADK_SUCCESS
 * @retval      ADK_E_INVALID_HANDLE
 * @retval      ADK_E_INVALID_PARAMETER
 * @retval      ADK_E_SE_STATE_GET_FAILED
 *
 * @brief
 *
 * Read TWAMP responder level statistics
 * =============================================================
 */

adk_status_t
adk_twamp_responder_stats_get(adk_module_handle_t   twampHandle,
                       adk_twamp_responder_handle_t twampRspndHandle_p,
                       adk_twamp_responder_stats_t  *twampRespndStats_p)
{
    adk_status_t    retValue = ADK_SUCCESS;
    adk_twamp_module_info_t *twampModuleInfo_p;
    uint32_t    respndId;
    adk_sys_handle_t sysHandle;

    /* Check validity of the input parameters */
    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle)) {
        ADK_FORCE_LOG_ERROR("Invalid module handle %"ADKPRIh"\n", twampHandle);
        return(ADK_E_INVALID_HANDLE);
    }

    /* Validate passed handle */
    if (!adk_twamp_valid_module_handle(twampHandle)) {
        ADK_FORCE_LOG_ERROR("Invalid module handle %"ADKPRIh"\n", twampHandle);
        return(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    if (!ADK_SYS_VALID_FLOW_HANDLE(twampRspndHandle_p)) {
        ADK_LOG_ERROR(sysHandle, twampHandle,\
            "Invalid TWAMP Responder handle %"ADKPRIh"\n", twampRspndHandle_p);
        ADK_LOG_ERROR(sysHandle, twampHandle,\
                     "Failed to get Responder Stats \n");
        return(ADK_E_INVALID_HANDLE);
    }

    /* Serialize the function */
    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);
    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    respndId = ADK_SYS_FLOWID_GET(twampRspndHandle_p);
    adk_twamp_responder_stats_read(twampHandle,
                                   respndId,
                                   twampRespndStats_p);

    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/* Dedicated function to avoid double lock when stats get is called from
 * responder delete API.
 */
static void
adk_twamp_responder_stats_read(adk_module_handle_t twampHandle,
                               uint32_t respndId,
                               adk_twamp_responder_stats_t  *twampRespndStats_p)
{
    adk_twamp_module_info_t *twampModuleInfo_p;
    uint32_t i, relative_max_sessions;
    uint64_t  prevTicks;
    uint64_t  currTxTicks;
    uint64_t  diffTicks;
    uint64_t  sessionTimeoutTicks;
    adk_twamp_session_reflector_object_t *sessionReflectorObj;
    adk_twamp_responder_object_t *responderObj;
    adk_sys_handle_t sysHandle;

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);
    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);
    ADK_MEMSET(twampRespndStats_p, 0, sizeof(adk_twamp_responder_stats_t));

    if (twampModuleInfo_p != NULL) {
        twampRespndStats_p->rxPkts =
            twampModuleInfo_p->dp_pvt_p->dpTwampResponderStats_p[respndId].rxPkts;
        twampRespndStats_p->txPkts =
            twampModuleInfo_p->dp_pvt_p->dpTwampResponderStats_p[respndId].txPkts;
        twampRespndStats_p->pktLengthErrors =
            twampModuleInfo_p->dp_pvt_p->dpTwampResponderStats_p[respndId].pktLengthErrors;
        twampRespndStats_p->pktsMaxSessions =
            twampModuleInfo_p->dp_pvt_p->dpTwampResponderStats_p[respndId].pktsMaxSessions;

        /*
         * Scan session DB for the specific responder ID to identify the active
         * sessions.
         */
        relative_max_sessions = (respndId * TWAMP_MAX_REFLECTORS_PPF_RESPONDER) +
            TWAMP_MAX_REFLECTORS_PPF_RESPONDER;
        sessionReflectorObj = twampModuleInfo_p->dp_pvt_p->sessionReflectorObj_p;
        responderObj = twampModuleInfo_p->dp_pvt_p->responderObj_p;

        for(i = respndId * TWAMP_MAX_REFLECTORS_PPF_RESPONDER; i < relative_max_sessions; i++) {
            if(sessionReflectorObj[i].isUsed) {
                ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_atomic64_read", "Atomically "
                                "read a 64-bit value\n");
                prevTicks = rte_atomic64_read(&sessionReflectorObj[i].prevTicks);
                ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_get_tsc_cycles", "Get "
                                "the number of TSC cycles since boot \n");
                currTxTicks = rte_get_tsc_cycles();
                /* To Do: The roll over case is not handled intentionally. The
                 * rte_get_tsc_cycles() API will return a 64 bit value in ticks
                 * since the system boot. The roll over will happen after several
                 * years and it depends on the CPU clock speed.
                 */
                diffTicks = currTxTicks - prevTicks;
                sessionTimeoutTicks = responderObj[respndId].cpuTicksPerSec *
                    responderObj[respndId].session_timeout;
                if(sessionTimeoutTicks >= diffTicks)
                {
                    twampRespndStats_p->numActiveSessions++;
                }
            } else {
                continue;
            }
        }
    }
}

/*!
 * =============================================================
 * adk_twamp_global_stats_get
 *
 * @param[in]  twampHandle      TWAMP module handle
 * @param[out] twmpStats_p      Pointer to TWAMP global stats
 *
 * @retval ADK_SUCCESS          Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE Invalid TWAMP module handle specified
 *
 * @brief
 *
 * This function reads the global level TWAMP statistics
 *
 * =============================================================
 */

adk_status_t
adk_twamp_global_stats_get(adk_module_handle_t  twampHandle,
                           adk_twamp_global_stats_t *twampGlobalStats_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t i;
    adk_twamp_module_info_t     *twampModuleInfo_p = NULL;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid module handle %"ADKPRIh"\n", twampHandle);
        return(ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE));
    }

    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    /* Validate passed  handles */
    if (!adk_twamp_valid_module_handle(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                          twampHandle);
        ADK_FORCE_LOG_ERROR("Failed to get TWAMP global statistics\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _adk_twamp_global_statistics_get_exit;
    }

    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    ADK_MEMSET(twampGlobalStats_p, 0, sizeof(adk_twamp_global_stats_t));

    for(i=0; i<ADK_MAX_NUM_CORES; i++) {
       twampGlobalStats_p->rxPkts += twampModuleInfo_p->dp_pvt_p->dpTwampGlobalStatistics_p[i]->rxPkts;
       twampGlobalStats_p->txPkts += twampModuleInfo_p->dp_pvt_p->dpTwampGlobalStatistics_p[i]->txPkts;
       twampGlobalStats_p->invalidRspndDrops += twampModuleInfo_p->dp_pvt_p->dpTwampGlobalStatistics_p[i]->invalidRspndDrops;
       twampGlobalStats_p->invalidLengthDrops += twampModuleInfo_p->dp_pvt_p->dpTwampGlobalStatistics_p[i]->invalidLengthDrops;
       twampGlobalStats_p->hashErrorDrops += twampModuleInfo_p->dp_pvt_p->dpTwampGlobalStatistics_p[i]->hashErrorDrops;
       twampGlobalStats_p->noBindDrops += twampModuleInfo_p->dp_pvt_p->dpTwampGlobalStatistics_p[i]->noBindDrops;
    }

_adk_twamp_global_statistics_get_exit:
    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return(retValue);
}

adk_status_t
adk_twamp_responder_stats_print(adk_module_handle_t twampHandle,
                       adk_twamp_responder_handle_t twampRespndHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_twamp_responder_stats_t twampRspndStats;

    if (adk_twamp_responder_stats_get(
                                      twampHandle,
                                      twampRespndHandle,
                                      &twampRspndStats) != ADK_SUCCESS)
    {
        return(retValue);
    }

    /* Print the statistics */
    ADK_INFO_WRITE("\nTWAMP responder statistics (Non-zero only)\n");

    ADK_TWAMP_STAT_PRINT("received packets: %u\n",
                         twampRspndStats.rxPkts);
    ADK_TWAMP_STAT_PRINT("transmitted packets: %u\n",
                         twampRspndStats.txPkts);
    ADK_TWAMP_STAT_PRINT("Packets dropped due to length check failed: %u\n",
                         twampRspndStats.pktLengthErrors);
    ADK_TWAMP_STAT_PRINT("Number of active sessions: %u\n",
                         twampRspndStats.numActiveSessions);
    ADK_TWAMP_STAT_PRINT("Packets dropped due to max number of sessions reached: %u\n",
                         twampRspndStats.pktsMaxSessions);

    return(retValue);
}

adk_status_t
adk_twamp_global_stats_print(adk_module_handle_t twampHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_twamp_global_stats_t twampGlobalStats;
    adk_sys_handle_t sysHandle;

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    retValue = adk_twamp_global_stats_get(
                                        twampHandle,
                                        &twampGlobalStats);
    if( retValue != ADK_SUCCESS )
    {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                         "Failed to get global statistics of TWAMP module\n");
        return(retValue);
    }

    /* Print the statistics */
    ADK_INFO_WRITE("\nTWAMP global statistics (Non-zero only)\n");

    ADK_TWAMP_STAT_PRINT("Total received packets: %d\n",
                         twampGlobalStats.rxPkts);
    ADK_TWAMP_STAT_PRINT("Total transmitted packets: %d\n",
                         twampGlobalStats.txPkts);
    ADK_TWAMP_STAT_PRINT("Packets dropped due to responder not defined : %d\n",
                         twampGlobalStats.invalidRspndDrops);
    ADK_TWAMP_STAT_PRINT("Packets dropped due to length check failed: %d\n",
                         twampGlobalStats.invalidLengthDrops);
    ADK_TWAMP_STAT_PRINT("Packets dropped due to unsupported packet length: %d\n",
                         twampGlobalStats.hashErrorDrops);
    ADK_TWAMP_STAT_PRINT("Packets dropped due to bind not defined: %d\n",
                         twampGlobalStats.noBindDrops);

    return retValue;
}

/*!
 * =============================================================
 * adk_twamp_tables_dump
 *
 * @param[in] twampHandle      TWAMP module handle.
 *
 * @param[in] table_group   Group of tables to dump
 *
 * @retval ADK_SUCCESS      Successful completion
 *
 * @retval ADK_E_INVALID_HANDLE  Invalid TWAMP module handle specified
 *
 * @retval ADK_E_INVALID_PARAMETER  Invalid table group type specified
 *
 * @brief
 *
 * This function dumps the specified group of tables.
 *
 =============================================================
 */

adk_status_t
adk_twamp_tables_dump(adk_module_handle_t twampHandle,
                    adk_twamp_table_group_t table_group)
{
    adk_twamp_module_info_t     *twampModuleInfo_p;
    int                          i;

    if ( !ADK_MODULE_HANDLE_VALIDATE(twampHandle) )
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP Module handle %"ADKPRIh" \n",
                            twampHandle);
        ADK_FORCE_LOG_ERROR("Failed to dump TWAMP tables\n");
        return(ADK_E_INVALID_HANDLE);
    }

    if (!adk_twamp_valid_module_handle(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP Module handle %"ADKPRIh" \n",
                            twampHandle);
        ADK_FORCE_LOG_ERROR("Failed to dump TWAMP tables\n");
        return(ADK_E_INVALID_HANDLE);
    }

    twampModuleInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    switch (table_group)
    {
     case ADK_TWAMP_RESPONDER_TABLE:
        for(i = 0; i < TWAMP_MAX_PPF_RESPONDER; i++) {
            ADK_TABLE_ENTRY_WRITE("0x%X", i);
            if(twampModuleInfo_p->dp_pvt_p->responderObj_p[i].isValid) 
            {
                ADK_TABLE_ENTRY_WRITE(", DEFINED");
            } else {
                ADK_TABLE_ENTRY_WRITE(", NOT_DEFINED");
            }
            ADK_TABLE_ENTRY_WRITE("\n");
        }
        break;
#if 0
     case ADK_TWAMP_CUSTOM_BIND_TABLE:
        break;
#endif
     default:
        break;
    }

    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * @fn adk_twamp_shutdown
 *
 * @param[in]  twampHandle    TWAMP module handle.
 *
 * @retval  ADK_SUCCESS   Successfully shut down TWAMP module.
 * @retval  ADK_E_INVALID_HANDLE   Invalid TWAMP module handle
 *
 * @brief  ADK TWAMP module shutdown
 * =============================================================
 */

adk_status_t
adk_twamp_shutdown(adk_module_handle_t twampHandle)
{
    adk_status_t              retValue = ADK_SUCCESS;
    adk_sys_handle_t          sysHandle;
    adk_twamp_module_info_t   *twampModuleInfo_p;
    adk_module_service_fn_t   svcFn;

    if ( !ADK_MODULE_HANDLE_VALIDATE(twampHandle) )
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP Module handle %"ADKPRIh" \n",
                            twampHandle);
        ADK_FORCE_LOG_ERROR("TWAMP module failed to shutdown\n");
        return(ADK_E_INVALID_HANDLE);
    }

    if (!adk_twamp_valid_module_handle(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP Module handle %"ADKPRIh" \n",
                            twampHandle);
        ADK_FORCE_LOG_ERROR("TWAMP module failed to shutdown\n");
        return(ADK_E_INVALID_HANDLE);
    }

    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    twampModuleInfo_p = (adk_twamp_module_info_t *) ADK_MODULE_INFO_PTR_GET(twampHandle);

    ADK_MEMSET(&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    retValue = adk_module_service_function_register(sysHandle, twampHandle,
                                                    &svcFn);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                            "Failed to register NULL functions\n");
    }

    /* Reset the cookie */
    twampModuleInfo_p->cookie = 0;

    /* Free the TWAMP module info structure */
    ADK_NVM_FREE (twampModuleInfo_p);
    ADK_MODULE_INFO_PTR_SET(twampHandle, NULL);

    if (retValue == ADK_SUCCESS) {
        ADK_LOG_INFO(sysHandle, twampHandle, "TWAMP module shutdown\n");
    } else {
        ADK_LOG_ERROR(sysHandle, twampHandle,
                            "TWAMP module failed to shutdown\n");
    }

    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    return retValue;
}

/*!
 * =============================================================
 * adk_twamp_sys_log
 *
 * @param[in] twampHandle           TWAMP Module Handle
 * @param[in] log_level             SIF defined log level
 *
 * @retval ADK_SUCCESS              Successful completion
 * @retval ADK_E_INVALID_HANDLE     Invalid TWAMP module handle
 * @retval ADK_E_UNSUPPORTED        Unsupported log level parameter.
 *
 * @brief
 *
 * This function is used to set the log-level for TWAMP module via SIF using
 * adk_module_attr_set() API. This function must not be called directly.
 * =============================================================
 */
adk_status_t adk_twamp_sys_log(adk_module_handle_t twampHandle,
                              adk_log_level_t sys_log_level)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_log_level_t twamp_log_level;
    adk_sys_handle_t sysHandle;

    if (!ADK_MODULE_HANDLE_VALIDATE(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                           twampHandle);
        ADK_FORCE_LOG_ERROR("Failed to set log-level %d for TWAMP\n",
                           sys_log_level);
        return(ADK_E_INVALID_HANDLE);
    }

    if (!adk_twamp_valid_module_handle(twampHandle))
    {
        ADK_FORCE_LOG_ERROR("Invalid TWAMP module handle %"ADKPRIh"\n",
                           twampHandle);
        ADK_FORCE_LOG_ERROR("Failed to set log-level 0x%x for TWAMP\n",
                           sys_log_level);
        return(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);
    ADK_TWAMP_LOG_FUNC_ENTRY(twampHandle);
    ADK_ACQUIRE_MODULE_LOCK(twampHandle, retValue);

    retValue = adk_twamp_map_sys_log(twampHandle, sys_log_level,
                                                 &twamp_log_level);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(retValue);
        ADK_LOG_ERROR(sysHandle, twampHandle,
                           "Failed to set log-level %d\n", sys_log_level);
        goto _adk_twamp_sys_log_exit;
    }

    ADK_MODULE_LOG_LEVEL_SET(twampHandle, twamp_log_level);

    ADK_LOG_INFO(sysHandle, twampHandle, "TWAMP module %"ADKPRIh" log-level set to "
                       "0x%x\n", twampHandle, twamp_log_level);

_adk_twamp_sys_log_exit:

    ADK_RELEASE_MODULE_LOCK(twampHandle, retValue);
    ADK_TWAMP_LOG_FUNC_EXIT(twampHandle);

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN(retValue);
}


/*!
 *      @} end addtogroup _twampFunctions
 * @} end addtogroup _twamp
 *
 */
