/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @internal
 * @file  adk_twamp_bind.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _twampFunctions  TWAMP Module
 *
 * @{
 *
 * @brief   This files has the functions which are used for binding TWAMP to
 *          other modules. These functions should not be called directly by the
 *          application but they are invoked indirectly via ADK SIF.
 *
 *  **********************************************************************/

#include <adk_twamp.h>
#include "adk_twamp_internal.h"
#include "adk_twamp_const_internal.h"


/*!
 * =============================================================
 * @internal
 * adk_twamp_custom_bind
 *
 * @param[in]   sysHandle       ADK System Handle
 * @param[in]   bindInfo_p      Binding information
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Function to perform custom bind from TWAMP to Transport layer
 *          modules
 *
 * This function is used to bind TWAMP to Transport modules without
 * going through the ADK SIF fast path code. This custom binding provides better
 * throughput compared to regular SIF binding.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called indirectly and automatically via adk_bind() API, if
 * isCustomBind parameter to adk_bind() API is set to TRUE.
 *
 * =============================================================
 */
adk_status_t adk_twamp_custom_bind(adk_sys_handle_t sysHandle,
                                   adk_bind_info_t *bindInfo_p)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_module_handle_t twampHandle;
    uint32_t            destModId = 0 ;
    uint32_t      srcFlowId;
    adk_twamp_module_info_t *twampModuleInfo_p = NULL;
    adk_twamp_responder_bind_attr_t *bind_attr = NULL;
    uint16_t entryFunc;

    /* NULL checking */
    if(!bindInfo_p)
    {
        ADK_TWAMP_LOG_FORCE("Invalid bind Info pointer !! \n");
        ADK_TWAMP_LOG_FORCE("Failed to create Custom bind entry \n");
        return(ADK_E_INVALID_PARAMETER);

    }

    twampHandle = bindInfo_p->srcModuleHandle;

    if (!adk_twamp_valid_module_handle(twampHandle) || !sysHandle)
    {
        ADK_TWAMP_LOG_FORCE("Invalid TWAMP module handle %"ADKPRIh" \n", 
                          twampHandle);
        ADK_TWAMP_LOG_FORCE("Failed to create Custom bind entry \n");
        return(ADK_E_INVALID_HANDLE);
    }

    ADK_TWAMP_LOG_FUNC_ENTRY(twampHandle);

    ADK_ACQUIRE_MODULE_LOCK(twampHandle, ret);

    twampModuleInfo_p = (adk_twamp_module_info_t *)
                            ADK_MODULE_INFO_PTR_GET(twampHandle);

    /* Validate the parameters in the bind_info */

    /* Custom bind must be a direct bind */
    if (bindInfo_p->path != ADK_BIND_PATH_DIRECT) {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                          "Only ADK_BIND_PATH_DIRECT is supported\n");
        goto _adk_twamp_custom_bind_exit;
    }

    /* Only F2M binding is supported via custom bind */
    if (bindInfo_p->type != ADK_BIND_FLOW_TO_MODULE) {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                          "Unsupported Binding Type %d\n", bindInfo_p->type);
        goto _adk_twamp_custom_bind_exit;
    }

    /* only custombind supported*/
    if (!bindInfo_p->isCustomBind) {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                          "Only custom bind supported\n");
        goto _adk_twamp_custom_bind_exit;
    }

    /* get destination module id */
    switch(ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle))
    {
    case ADK_IP_MODULE_ID:
        if (bindInfo_p->direction == ADK_BIND_EGRESS_PROCESSING) {
            destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV4_ULI;
        } else{
            ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                                "Only ADK_BIND_EGRESS_PROCESSING is supported\n");
        }
        break;
    case ADK_IPV6_MODULE_ID:
        if (bindInfo_p->direction == ADK_BIND_EGRESS_PROCESSING)
        {
            destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV6_ULI;
        }
        else
        {
            ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                                "Only ADK_BIND_EGRESS_PROCESSING is supported\n");
        }
        break;
    default:
         ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_MODULE);
         ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                "Following is only supported bind combination\n"
                " source module = TWAMP, destination module = IPV4\n"
                " source module = TWAMP, destination module = IPV6\n"
                );
         goto _adk_twamp_custom_bind_exit;
    }

    /* Get the responder ID */
    srcFlowId = ADK_SYS_FLOWID_GET(bindInfo_p->srcFlowHandle);
#if ADK_DEBUG
    ADK_LOG_INFO(sysHandle, twampHandle,
                 "Absolute Responder ID = %d \n", srcFlowId);
#endif

    bind_attr = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(twampHandle), 0,
		 1, sizeof(adk_twamp_responder_bind_attr_t));
    if (bind_attr == NULL) {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                          "Memory allocation failed !! \n");
        goto _adk_twamp_custom_bind_exit;
    }

    /* Update the Responder table with F2M bind information */
    bind_attr->destModuleId = destModId;
#if ADK_DEBUG
    ADK_LOG_INFO(sysHandle, twampHandle,
        "Destination module ID  = %d \n", bind_attr->destModuleId );
#endif

    entryFunc = ADK_MODULE_CPU_EGRESS_FN_GET(bindInfo_p->destModuleHandle,
                                                            _DEST_FUNC_DEFAULT);
    bind_attr->destFunc[_DEST_FUNC_DEFAULT] = entryFunc;

    /* Get the Stage Info through PAL */
    ret = adk_pal_dp_stage_info_get(
                         entryFunc,
                         &bind_attr->stage_info);
    if (ret != ADK_SUCCESS)
    {
        ret = ADK_E_INTERNAL_ERROR;
        ADK_LOG_ERROR(sysHandle, twampHandle,
                   "Failed to get stage info for dest module\n");
        goto _adk_twamp_custom_bind_exit;
    }

    /* Atomically update the bind attributes in Responder table */
    ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_atomic64_set", "Atomically "
                    "set a 64-bit value\n");
    rte_atomic64_set(&twampModuleInfo_p->dp_pvt_p->responderObj_p[srcFlowId].bind_attr, (int64_t) bind_attr);

_adk_twamp_custom_bind_exit:

    if (ret == ADK_SUCCESS) {
         ADK_LOG_INFO(sysHandle, twampHandle,
                       "Custom Bind Successfully done.\n");
    } else {
        ADK_TWAMP_LOG_FORCE("Failed to perform custom binding.\n");
    }

    ADK_RELEASE_MODULE_LOCK(twampHandle, ret);


_lock_release_failed:
_lock_acquire_failed:
    ADK_TWAMP_LOG_FUNC_EXIT(twampHandle);

    return(ret);
}

/*!
 * =============================================================
 * @internal
 * adk_twamp_custom_unbind
 *
 * @param[in]   sysHandle       ADK System Handle
 * @param[in]   bindInfo_p      Binding information
 *
 * @retval  ADK_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid module handle specified
 *
 * @brief   Function to perform custom unbind between TWAMP and transport 
 *          modules
 *
 * This function is used to unbind TWAMP from the transport module without
 * going through the ADK SIF fast path code.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called indirectly and automatically via adk_unbind() API, if
 * isCustomBind parameter to adk_unbind() API is set to TRUE.
 *
 * =============================================================
 */
adk_status_t adk_twamp_custom_unbind(adk_sys_handle_t sysHandle,
                                        adk_bind_info_t *bindInfo_p)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_module_handle_t twampHandle;
    adk_twamp_module_info_t *twampModuleInfo_p = NULL;
    uint32_t      srcFlowId;
    uint32_t            destModId = 0 ;
    adk_twamp_responder_bind_attr_t *bind_attr = NULL;

    /* NULL checking */
    if(!bindInfo_p)
    {
        ADK_TWAMP_LOG_FORCE("Invalid bind Info pointer !! \n");
        ADK_TWAMP_LOG_FORCE("Failed to create Custom bind entry \n");
        return(ADK_E_INVALID_PARAMETER);

    }

    twampHandle = bindInfo_p->srcModuleHandle;
    if (!adk_twamp_valid_module_handle(twampHandle) || !sysHandle)
    {
        ADK_TWAMP_LOG_FORCE("Invalid TWAMP module handle %"ADKPRIh" \n",
                          twampHandle);
        ADK_TWAMP_LOG_FORCE("Failed to remove Custom bind entry \n");
        return(ADK_E_INVALID_HANDLE);
    }

    ADK_TWAMP_LOG_FUNC_ENTRY(twampHandle);

    ADK_ACQUIRE_MODULE_LOCK(twampHandle, ret);

    twampModuleInfo_p = (adk_twamp_module_info_t *) ADK_MODULE_INFO_PTR_GET(twampHandle);

    /* Custom bind must be a direct bind */
    if (bindInfo_p->path != ADK_BIND_PATH_DIRECT) {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                          "Only ADK_BIND_PATH_DIRECT is supported\n");
        goto _adk_twamp_custom_unbind_exit;
    }

    if (bindInfo_p->type != ADK_BIND_FLOW_TO_MODULE) {
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                          "Unsupported Binding Type %d\n", bindInfo_p->type);
        goto _adk_twamp_custom_unbind_exit;
    }

    /* get destination module id */
    switch(ADK_MODULE_MODULEID_GET(bindInfo_p->destModuleHandle))
    {
    case ADK_IP_MODULE_ID:
        if (bindInfo_p->direction == ADK_BIND_EGRESS_PROCESSING) {
            destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV4_ULI;
        } else{
            ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                                "Only ADK_BIND_EGRESS_PROCESSING is supported\n");
        }
        break;
    case ADK_IPV6_MODULE_ID:
        if (bindInfo_p->direction == ADK_BIND_EGRESS_PROCESSING)
        {
            destModId = ADK_SIF_CFG_PIPELINE_MODULE_IPV6_ULI;
        }
        else
        {
            ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                                "Only ADK_BIND_EGRESS_PROCESSING is supported\n");
        }
        break;
    default:
         ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_MODULE);
         ADK_TWAMP_LOG_ERROR(twampHandle, ret,
                "Following is only supported bind combination\n"
                " source module = TWAMP, destination module = IPV4\n"
                " source module = TWAMP, destination module = IPV6\n"
                );
         goto _adk_twamp_custom_unbind_exit;
    }

    /* Get the responder ID */
    srcFlowId = ADK_SYS_FLOWID_GET(bindInfo_p->srcFlowHandle);

    /* Update the Responder table with F2M bind information */
    ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_atomic64_read", "Atomically "
                    "read a 64-bit value\n");
    bind_attr = (adk_twamp_responder_bind_attr_t *)
                     rte_atomic64_read(&twampModuleInfo_p->
                               dp_pvt_p->responderObj_p[srcFlowId].bind_attr);

    /* Setting NULL */
    ADK_LOG_EXT_API(sysHandle, twampHandle, "rte_atomic64_set", "Atomically "
                    "set a 64-bit value\n");
    rte_atomic64_set(&twampModuleInfo_p->dp_pvt_p->
                      responderObj_p[srcFlowId].bind_attr, 0);

    /* Free the pointer */
    ADK_NVM_FREE(bind_attr);

_adk_twamp_custom_unbind_exit:

    if (ret == ADK_SUCCESS) {
        ADK_LOG_INFO(sysHandle, twampHandle,
                      "Custom Un Bind Successfully done.\n");
    } else {
        ADK_FORCE_LOG_ERROR("Failed to perform custom un binding.\n");
    }

    ADK_RELEASE_MODULE_LOCK(twampHandle, ret);


_lock_release_failed:
_lock_acquire_failed:
    ADK_TWAMP_LOG_FUNC_EXIT(twampHandle);

    return(ret);
}

/*! end addtogroup _twamp
 * @}
 */
