/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_twamp_internal.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _twampFunctions  TWAMP Module
 *
 * @{
 *
 * @brief  
 *
 * This file contains all the internal functions provided by the TWAMP
 * module. None of the functions from this file should be called by
 * other modules.
 *
 ***********************************************************************/


/* Infrastructure header files */
#include "adk_include.h"

/* TWAMP external header files */
#include "adk_twamp.h"
#include "adk_twamp_config.h"
#include "adk_twamp_internal.h"


/*!
 * =============================================================
 * @internal
 * adk_twamp_valid_module_handle
 * @param[in] twampHandle TWAMP module handle. Returned at the
 * time of TWAMP module's registration with the module manager.
 *
 * @retval TRUE When TWAMP handle passes check
 * @retval FALSE When TWAMP handle fails check
 *
 * @brief TWAMP module handle validation
 *
 * This function checks whether the supplied TWAMP is valid.
 * In order to pass the checking the handle has to be non-NULL
 * and the its cookie value has to match the value programmed
 * during the TWAMP module initialization.
 *
 * =============================================================
 */
bool
adk_twamp_valid_module_handle(
    adk_module_handle_t twampHandle)
{
    adk_twamp_module_info_t *twampInfo_p = NULL;

    twampInfo_p = ADK_MODULE_INFO_PTR_GET(twampHandle);

    if (twampInfo_p == NULL)
    {
        return FALSE;
    }

    if (twampInfo_p->cookie != ADK_TWAMP_MODULE_HANDLE_COOKIE)
    {
        return FALSE;
    }

    return TRUE;
}   /* -----  end of function adk_twamp_valid_module_handle  ----- */

/*!
 * =============================================================
 * @internal
 * adk_twamp_map_sys_log
 * @param[in] twampHandle The handle of TWAMP module
 * @param[in] sysLogLevel Log level used by SIF
 * @param[out] twampLogLevel pass the mapped TWAMP Log
 * level to the caller
 *
 * @retval ADK_SUCCESS When SIF log level mapping to TWAMP
 * log level mapping is successful
 * @retval ADK_E_UNSUPPORTED When SIF log level mapping to
 * TWAMP log level mapping is unsuccessful
 *
 * @brief Maps SIF's log level to TWAMP's log level
 * =============================================================
 */
adk_status_t
adk_twamp_map_sys_log(
    adk_module_handle_t twampHandle,
    adk_log_level_t sysLogLevel,
    adk_log_level_t *twampLogLevel)
{
    adk_status_t  retValue = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(twampHandle);

    switch (sysLogLevel)
    {
    case ADK_LOG_LEVEL_INFO:
        *twampLogLevel = TWAMP_LOG_LEVEL_INFO; break;
    case ADK_LOG_LEVEL_WARNING:
        *twampLogLevel = TWAMP_LOG_LEVEL_WARNING; break;
    case ADK_LOG_LEVEL_ERROR:
        *twampLogLevel = TWAMP_LOG_LEVEL_ERROR; break;
    case ADK_LOG_LEVEL_FATAL:
        *twampLogLevel = TWAMP_LOG_FATAL; break;
    case ADK_LOG_LEVEL_DISABLE:
        *twampLogLevel = TWAMP_LOG_DISABLE; break;
    default:
        ADK_TWAMP_LOG_ERROR(twampHandle, ADK_E_UNSUPPORTED,
                            "Unsupported log-level in TWAMP!");
        retValue = ADK_E_UNSUPPORTED;
        break;
    }
    return(retValue);
}   /* -----  end of function adk_twamp_map_sys_log  ----- */


/*! end addtogroup _twamp
 * @}
 */
