/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <adk_include.h>
#include <adk_twamp_dp.h>
#include <adk_twamp.h>
#include <adk_twamp_internal.h>
#include <adk_twamp_const_internal.h>
#include <adk_pmu.h>
#include <adk_hash.h>
#include <adk_timer.h>
#include "adk_ip_dp_export.h"
#include "adk_ipv6_dp_export.h"
#include "adk_twamp_dp_export.h"
#include "adk_ip_config.h"
#include <rte_eventdev.h>
#include <rte_event_framework.h>
#include <rte_bridge_eth.h>
#include <rte_bridge_timer.h>

/* Module info pointer stored for DP process to access tables */
static adk_twamp_dp_pvt_data_t *adk_twamp_dp_pvt_g_p = NULL;

static void **reflectorFirstHashTbl;
static void **reflectorSecondHashTbl;
static adk_twamp_session_reflector_object_t *sessionReflectorObj;
static adk_twamp_responder_object_t *responderObj;

/* TWAMP global statistics */
static adk_twamp_global_stats_t **twampGlobalStatistics;

/* Responder level statistics */
static adk_twamp_responder_stats_t *twampResponderStats;
static adk_twamp_internal_responder_stats_t *twampInternalResponderStats;

static inline void
check_if_session_exists(uint32_t hash_val,
                        uint8_t responderID,
                        adk_twamp_transport_object_t transportObject,
                        adk_twamp_session_reflector_object_t **sessionObj,
                        bool *isPresent);

static inline void
create_session(uint32_t hash_val,
               uint8_t responderID,
               adk_twamp_transport_object_t transportObject,
               adk_twamp_session_reflector_object_t **sessionObj);

static inline void
check_if_session_object_is_valid(adk_twamp_session_reflector_object_t *sessionReflectorObj,
                                 adk_twamp_transport_object_t transportObject,
                                 uint8_t responderID,
                                 bool *isPresent);

/*!
 * =============================================================
 * adk_twamp_dp_table_init()
 *
 * Structure of task params :
 *
 * =============================================================
 */
adk_dp_status_t
adk_twamp_dp_table_init(__attribute__((unused)) adk_module_handle_t twampHandle,
                        adk_twamp_dp_pvt_data_t *twamp_dp_pvt_p)
{
    uint32_t i;
    const struct rte_memzone *mz = NULL;
    adk_status_t        ret = ADK_SUCCESS;

    /* Lookup the shared memory that gets created by control plane TWAMP
     * module initialization */
    mz = rte_memzone_lookup(ADK_TWAMP_SHM_MEMZONE_NAME);
    if (mz == NULL)
    {
        debug_print("Error: Failed to lookup TWAMP shared memory %s\n",
                    ADK_TWAMP_SHM_MEMZONE_NAME);
        ret = ADK_E_MEM_UNAVAILABLE;
        goto _adk_twamp_dp_table_init_exit;
    }

    /* allocate NVM memory for all the databases */
    reflectorFirstHashTbl = (void **) ADK_DP_SHM_CALLOC (
					ADK_MODULE_MODULEID_GET(twampHandle), 0, 0xFFFF,
                                       sizeof(adk_twamp_session_reflector_object_t *), 0);
    if(reflectorFirstHashTbl == NULL) {
        debug_print("Memory allocation failure \n");
        return ADK_DP_E_ERROR;
    }
    reflectorSecondHashTbl = (void **) ADK_DP_SHM_CALLOC (
					ADK_MODULE_MODULEID_GET(twampHandle), 0, 0xFFFF,
                                       sizeof(adk_twamp_session_reflector_object_t *), 0);
    if(reflectorSecondHashTbl == NULL) {
        debug_print("Memory allocation failure \n");
        return ADK_DP_E_ERROR;
    }

    sessionReflectorObj = (adk_twamp_session_reflector_object_t *)
                              ADK_DP_SHM_CALLOC (
				ADK_MODULE_MODULEID_GET(twampHandle), 0, TWAMP_TOTAL_REFLECTORS,
                              sizeof(adk_twamp_session_reflector_object_t), 0);
    if(sessionReflectorObj == NULL) {
        debug_print("Memory allocation failure \n");
        return ADK_DP_E_ERROR;
    }

    responderObj = (adk_twamp_responder_object_t *)
                              ADK_DP_SHM_CALLOC (
				ADK_MODULE_MODULEID_GET(twampHandle), 0, 2 * TWAMP_MAX_PPF_RESPONDER,
                              sizeof(adk_twamp_responder_object_t), 0);
    if(responderObj == NULL) {
        debug_print("Memory allocation failure \n");
        return ADK_DP_E_ERROR;
    }

    /* Allocating memory for responder level statistics */
    twampResponderStats = (adk_twamp_responder_stats_t *)
                                 ADK_DP_SHM_CALLOC (
				ADK_MODULE_MODULEID_GET(twampHandle), 0, TWAMP_MAX_PPF_RESPONDER,
                                 sizeof(adk_twamp_responder_stats_t), 0);
    if(twampResponderStats == NULL) {
        debug_print("Memory allocation failure \n");
        return ADK_DP_E_ERROR;
    }

    twampInternalResponderStats = (adk_twamp_internal_responder_stats_t *)
                                 ADK_DP_SHM_CALLOC (
				ADK_MODULE_MODULEID_GET(twampHandle), 0, TWAMP_MAX_PPF_RESPONDER,
                                 sizeof(adk_twamp_internal_responder_stats_t), 0);
    if(twampInternalResponderStats == NULL) {
        debug_print("Memory allocation failure \n");
        return ADK_DP_E_ERROR;
    }

    twamp_dp_pvt_p->reflectorFirstHashTbl_p = reflectorFirstHashTbl;
    twamp_dp_pvt_p->reflectorSecondHashTbl_p = reflectorSecondHashTbl;
    twamp_dp_pvt_p->sessionReflectorObj_p = sessionReflectorObj;
    twamp_dp_pvt_p->responderObj_p = responderObj;

    for(i=0; i < TWAMP_MAX_PPF_RESPONDER; i++) {
        responderObj[i].session_reflector_rm_hdl =
                      adk_rm_resource_init(ADK_MODULE_MODULEID_GET(twampHandle), 0,
							TWAMP_MAX_REFLECTORS_PPF_RESPONDER);
    }

    twampGlobalStatistics = (adk_twamp_global_stats_t **) ADK_DP_SHM_CALLOC (
						ADK_MODULE_MODULEID_GET(twampHandle),
						0, ADK_MAX_NUM_CORES,
                                                sizeof(adk_twamp_global_stats_t *), 0);
    if (twampGlobalStatistics == NULL)
    {
        debug_print("Error: Memory allocation failed \n");
        ret = ADK_E_MEM_UNAVAILABLE;
        goto _adk_twamp_dp_table_init_exit;
    }

    for(i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        twampGlobalStatistics[i] = (adk_twamp_global_stats_t *) ADK_DP_SHM_CALLOC (
								ADK_MODULE_MODULEID_GET(twampHandle), 0,
								1, sizeof(adk_twamp_global_stats_t), 0);
        if (twampGlobalStatistics[i] == NULL)
        {
            debug_print("Error: Memory allocation failed \n");
            ret = ADK_E_MEM_UNAVAILABLE;
            goto _adk_twamp_dp_table_init_exit;
        }
    }

    twamp_dp_pvt_p->dpTwampGlobalStatistics_p  = twampGlobalStatistics;
    twamp_dp_pvt_p->dpTwampResponderStats_p    = twampResponderStats;
    twamp_dp_pvt_p->dpTwampInternalResponderStats_p    = twampInternalResponderStats;

    /* Set the cookie to indicate data plane tables to be ready. Use memory
     * barrier to enforce this to be the last instruction generated by the
     * compiler. */
    barrier();
    twamp_dp_pvt_p->cookie = ADK_TWAMP_MODULE_COOKIE;

_adk_twamp_dp_table_init_exit:
    return ret;
}

adk_dp_status_t
adk_twamp_dp_pkt_proc_lli_single(void *task_p, int coreIndex,
                                 uint16_t *destFunc,
                                 adk_pal_dp_stage_info_t **stage_info,
                                 uint32_t *dst_mod_id)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t *task = (adk_pktio_task_t *)task_p;
    adk_twamp_pdu_t *twampPdu = NULL;
    uint32_t hash_val = 0, first_hash, second_hash;
    uint32_t responderID;
    uint32_t sourceIP = 0;
    uint32_t destIP = 0;
    uint32_t sourceIPv6[4] = {0};
    uint32_t destIPv6[4] = {0};
    uint32_t senderTTL = 0;
    uint16_t sourceUDP_port;
    uint16_t destUDP_port;
    uint16_t twampPktLength;
    adk_twamp_transport_object_t transportObject;
    bool isPresent = 0;
    bool isStale = 0;
    adk_twamp_session_reflector_object_t *sessionObj = NULL;
    adk_sif_ntp_time_t ntpTxTime;
    uint64_t  currTxTicks;
    adk_sif_ntp_time_t ntpRxTime;
    uint64_t  currRxTicks;
    /* Error Estimate fields */
    uint32_t s = 1, z = 0, scale = 32, multiplier = 1;
    adk_ip_dp_uli_imc_data_t *ip_uli_imc_param_p;
    adk_ipv6_dp_uli_imc_data_t *ipv6_uli_imc_param_p;
    uint32_t l3offset=0;
    uint32_t l4offset=0;
    uint32_t ipVrdId = 0;
    uint8_t  dscp = 0;
    uint16_t tmp_dscp = 0;
    uint8_t *pduData_p;
    adk_metadata_t *adk_md_p;
    adk_twamp_responder_bind_attr_t *bind_attr = NULL;

    adk_md_p = ADK_METADATA_PTR(task);

    adk_twamp_dp_lli_imc_data_t *twamp_imc_data_p =
                (adk_twamp_dp_lli_imc_data_t *)&(adk_md_p->imc_param_start);

    uint32_t twamp_imc_path;

    responderID = ADK_METADATA_PARAM(task, pdu_flow_id);

    /* Total number of packets received from IP layer. */
    twampGlobalStatistics[coreIndex]->rxPkts++;

    /* F2F unbind will be done if a responder is deleted. This is an additional
     * check. */
    if(unlikely(!responderObj[responderID].isValid)) {
#if ADK_DEBUG
        debug_print("Error: Invalid responder ID. Dropping packet! \n");
#endif
        /* Number of packets dropped due to invalid responder. */
        twampGlobalStatistics[coreIndex]->invalidRspndDrops++;
        ret = ADK_E_INTERNAL_ERROR;
        goto _twamp_processing_exit;
    }

#if ADK_DEBUG
    debug_print("Responder ID = %d \n", responderID);
#endif

    l3offset = ADK_METADATA_PARAM(task, ip_offset);
    l4offset = ADK_METADATA_PARAM(task, l4_offset);

    if(ADK_TWAMP_DP_IPV4_LLI_PATH == twamp_imc_data_p->path)
    {
        ipVrdId = twamp_imc_data_p->ip_info_u.ipv4_encap_info_p.ipVrdId;
        dscp = twamp_imc_data_p->ip_info_u.ipv4_encap_info_p.dscp;
    } else if(ADK_TWAMP_DP_IPV6_LLI_PATH == twamp_imc_data_p->path) {
        ipVrdId = twamp_imc_data_p->ip_info_u.ipv6_encap_info_p.ipVrdId;
        dscp = twamp_imc_data_p->ip_info_u.ipv6_encap_info_p.dscp;
    } else {
        ret = ADK_E_INTERNAL_ERROR;
        goto _twamp_processing_exit;
    }

    /*
     * Total number of packets dropped due to length check failed.
     * Twamp length must be atleast 41 bytes (Includes TWAMP header
     * and payload).
     * Task params 3 contains the L2 header length. pkt_len will contain the L2
     * payload length. Deducting IPv4 and UDP header length to arrive the TWAMP
     * packet length.
     */
    twampPktLength = task->pkt_len - (l4offset + ADK_UDP_HEADER_LEN);

    /* Total number of packets received from IP layer. */
    twampResponderStats[responderID].rxPkts++;

    if(unlikely(twampPktLength < ADK_TWAMP_MIN_PKT_LEN)) {
#if ADK_DEBUG
        debug_print("Error: Invalid Packet Length \n");
#endif
        twampGlobalStatistics[coreIndex]->invalidLengthDrops++;
        twampResponderStats[responderID].pktLengthErrors++;
        ret = ADK_E_INTERNAL_ERROR;
        goto _twamp_processing_exit;
    }

    /* Extracting the relevant fields from IP header */
    ADK_PDU_PTR_GET(task, pduData_p, void *);

    sourceUDP_port = *((uint16_t *) (pduData_p + l4offset));
    destUDP_port = *((uint16_t *) (pduData_p + l4offset + 2));

    if(ADK_TWAMP_DP_IPV4_LLI_PATH == twamp_imc_data_p->path)
    {
        /* Copying the TTL field from IPv4 header */
        senderTTL = *(pduData_p + l3offset + 8);

        sourceIP = *((uint32_t *) (pduData_p + l3offset + 12));
        destIP = *((uint32_t *) (pduData_p + l3offset + 16));

        /* Calculate the hash in network byte order */
        hash_val = SISD_CRC_HASH(0, responderID);
        hash_val = SISD_CRC_HASH(hash_val, sourceIP);
        hash_val = SISD_CRC_HASH(hash_val, sourceUDP_port);

#if ADK_DEBUG
        debug_print("SIP = 0x%x, DIP = 0x%x, SPORT = 0x%x, DPORT = 0x%x,\
               hash_val = 0x%x \n", sourceIP, destIP, sourceUDP_port,\
               destUDP_port, hash_val);
#endif

        transportObject.ipVersion = ADK_TWAMP_IPV4;
        transportObject.ipAddr.sourceIpv4addr = sourceIP;
        transportObject.sourceUDP_port = sourceUDP_port;
    }

    if(ADK_TWAMP_DP_IPV6_LLI_PATH == twamp_imc_data_p->path)
    {
        /* Copying the Hop Limit field from IPv6 header */
        senderTTL = *(pduData_p + l3offset + 7);

        sourceIPv6[0] = *((uint32_t *) (pduData_p + l3offset + 8));
        sourceIPv6[1] = *((uint32_t *) (pduData_p + l3offset + 12));
        sourceIPv6[2] = *((uint32_t *) (pduData_p + l3offset + 16));
        sourceIPv6[3] = *((uint32_t *) (pduData_p + l3offset + 20));
        destIPv6[0]   = *((uint32_t *) (pduData_p + l3offset + 24));
        destIPv6[1]   = *((uint32_t *) (pduData_p + l3offset + 28));
        destIPv6[2]   = *((uint32_t *) (pduData_p + l3offset + 32));
        destIPv6[3]   = *((uint32_t *) (pduData_p + l3offset + 36));

        /* Calculate the hash in network byte order */
        hash_val = SISD_CRC_HASH(0, responderID);
        hash_val = SISD_CRC_HASH(hash_val, sourceIPv6[0]);
        hash_val = SISD_CRC_HASH(hash_val, sourceIPv6[1]);
        hash_val = SISD_CRC_HASH(hash_val, sourceIPv6[2]);
        hash_val = SISD_CRC_HASH(hash_val, sourceIPv6[3]);
        hash_val = SISD_CRC_HASH(hash_val, sourceUDP_port);

        transportObject.ipVersion = ADK_TWAMP_IPV6;
        transportObject.ipAddr.sourceIpv6addr.addr[0] = sourceIPv6[0];
        transportObject.ipAddr.sourceIpv6addr.addr[1] = sourceIPv6[1];
        transportObject.ipAddr.sourceIpv6addr.addr[2] = sourceIPv6[2];
        transportObject.ipAddr.sourceIpv6addr.addr[3] = sourceIPv6[3];
        transportObject.sourceUDP_port = sourceUDP_port;
    }

    first_hash = hash_val & 0xFFFF;
    second_hash = hash_val >> 16;

    check_if_session_exists(hash_val, responderID, transportObject, &sessionObj,&isPresent);

    /* Stale session case */
    if(sessionObj) {
        if(isPresent) {
            adk_twamp_session_age_out(responderID, sessionObj->sessionId, isPresent, &isStale);
            if(isStale) { /* Treat it like a new session */
                isPresent = 0;
            }
        }
    }

    if(!isPresent) {
         /* Do TWAMP session age-out processing here */
         adk_twamp_session_age_out(responderID, 0, isPresent, &isStale);
         if(twampInternalResponderStats[responderID].numActiveSessions ==
                               TWAMP_MAX_REFLECTORS_PPF_RESPONDER) {
             debug_print("Error: Reached maximum number of reflector sessions \
                    for this responder. Dropping packet! \n");
             twampResponderStats[responderID].pktsMaxSessions++;

             ret = ADK_E_INTERNAL_ERROR;
             goto _twamp_processing_exit;
         }
#if ADK_DEBUG
         debug_print("\n Session doesn't exists. Creating session ... \n");
#endif
         create_session(hash_val, responderID, transportObject,
                        &sessionObj);
     }

     if(sessionObj) {
         if(isPresent) {
#if ADK_DEBUG
             debug_print("Session already exists... Processing the packets ...\n");
#endif
         } else {
             /* Creating the timer per reflector session */
             /* Reserving once all the session objects are set */
             sessionObj->isUsed = 1;
             /* Tracking the number of reflector sessions created per responder */
             twampInternalResponderStats[responderID].numActiveSessions++;
#if ADK_DEBUG
             debug_print(" Responder ID = %d, Number of Reflector sessions = %d \n",
                    responderID,
                    twampInternalResponderStats[responderID].numActiveSessions);
#endif
             sessionObj->responderId = responderID;

#if ADK_DEBUG
             if(ADK_TWAMP_DP_IPV4_LLI_PATH == twamp_imc_data_p->path) {
                 debug_print("SIP = 0x%x, SPORT = 0x%x, isUsed = %d \n",
                        sessionObj->transportObject.ipAddr.sourceIpv4addr,
                        sessionObj->transportObject.sourceUDP_port,
                        sessionObj->isUsed);
             }

             if(ADK_TWAMP_DP_IPV6_LLI_PATH == twamp_imc_data_p->path) {
                 debug_print("SIP[0] = 0x%x, SIP[1] = 0x%x, SIP[2] = 0x%x, SIP[3] = 0x%x,"
                        "SPORT = 0x%x, isUsed = %d \n",
                        sessionObj->transportObject.ipAddr.sourceIpv6addr.addr[0],
                        sessionObj->transportObject.ipAddr.sourceIpv6addr.addr[1],
                        sessionObj->transportObject.ipAddr.sourceIpv6addr.addr[2],
                        sessionObj->transportObject.ipAddr.sourceIpv6addr.addr[3],
                        sessionObj->transportObject.sourceUDP_port,
                        sessionObj->isUsed);
             }
#endif

         }

     } else {
         ret = ADK_E_INTERNAL_ERROR;
         goto _twamp_processing_exit;
     }

    /* TWAMP Header processing starts here */

    /* Moving pointer to start of TWAMP header */
    twampPdu = (adk_twamp_pdu_t *) (pduData_p + l4offset + ADK_UDP_HEADER_LEN);

    /*
     * The Reflector header's MBZ and senderTTL
     * are appended after sender's header of ADK_TWAMP_SENDER_HEADER_LEN size
     * in the outgoing packet i.e. 27 bytes of reflector header is split into 24
     * bytes in the beginning and 3 bytes at the end as shown below.
     *
     * Outgoing TWAMP header Format:
     * 24 bytes of new reflector header + 14 bytes of incoming sender header +
     *                                3 bytes of new reflector header
     *
     * The TWAMP sender header of 14 bytes is copied after 24 bytes of new
     * reflector header. The first 27 bytes of TWAMP payload is overwritten to
     * get the space for reflector header of 27 bytes.
     */
    ADK_MEMCPY(((uint8_t *)twampPdu) + 24, twampPdu,
                                     ADK_TWAMP_SENDER_HEADER_LEN);

    /* Updating sender sequence number */
    twampPdu->senderSeqNum = twampPdu->seqNum;
    twampPdu->seqNum = adk_htonl(sessionObj->seqNum);
    ++sessionObj->seqNum;

    currRxTicks = ADK_PKTMBUF_TIMESTAMP(task);
    /* Get the Rx NTP timestamp */
    adk_sif_get_ntp_time(currRxTicks, &ntpRxTime);
    twampPdu->integerRcvTS = adk_htonl(ntpRxTime.ntpSeconds);
    twampPdu->fractionalRcvTS = adk_htonl(ntpRxTime.ntpFractionSeconds);

    currTxTicks = rte_get_tsc_cycles();
    /* Store current time stamp in session DB for age-out calc purpose */
    rte_atomic64_set(&sessionObj->prevTicks, currTxTicks);

    /* Get the Tx NTP timestamp */
    adk_sif_get_ntp_time(currTxTicks, &ntpTxTime);
    /* Updating sender time stamp and error estimate fields */
    twampPdu->integerSendTS = twampPdu->integerTS;
    twampPdu->fractionalSendTS = twampPdu->fractionalTS;
    twampPdu->senderErrEstimate_And_MBZ =
                     adk_htonl((adk_ntohl(twampPdu->errEstimate_And_MBZ)) & 0xFFFF0000);
    twampPdu->integerTS = adk_htonl(ntpTxTime.ntpSeconds);
    twampPdu->fractionalTS = adk_htonl(ntpTxTime.ntpFractionSeconds);
    twampPdu->errEstimate_And_MBZ =
                  adk_htonl(s << 31 | z << 30 | scale << 24 | multiplier << 16);
    twampPdu->senderTTL = senderTTL;

#if ADK_DEBUG
    debug_print("Seqnum = %u \n", adk_ntohl(twampPdu->seqNum));
    debug_print("Rx integer TS = %u \n", adk_ntohl(twampPdu->integerRcvTS));
    debug_print("Rx fractional TS = %u \n", adk_ntohl(twampPdu->fractionalRcvTS));
    debug_print("Tx integer TS = %u \n", adk_ntohl(twampPdu->integerTS));
    debug_print("Tx fractional TS = %u \n", adk_ntohl(twampPdu->fractionalTS));
    debug_print("errEstimate_And_MBZ = 0x%x \n", adk_ntohl(twampPdu->errEstimate_And_MBZ));
#endif

    twamp_imc_path = twamp_imc_data_p->path;

    if(ADK_TWAMP_DP_IPV4_LLI_PATH == twamp_imc_path)
    {
        struct ipv4_hdr *ipHdr;
        struct udp_hdr *udpHdr;
        uint32_t tmp_32;
        uint16_t tmp_16;

        adk_md_p->exception_domain = ADK_FP_ORIGIN_DOMAIN;
        adk_md_p->func_idx = 0;
        
        /* Strip Ethernet*/
        ADK_PDU_HEAD_PULL(task, pduData_p, (void *), l3offset);
        adk_md_p->l4_offset -= l3offset;
        adk_md_p->ip_offset -= l3offset;
        
        /* Set flags to indicate ip/udp hdr present & update needed by IP */
        ip_uli_imc_param_p = (adk_ip_dp_uli_imc_data_t *)
            &(ADK_METADATA_PARAM(task, imc_param_start));

        ipHdr = (struct ipv4_hdr *)pduData_p;
        udpHdr = (struct udp_hdr *)(void *)(pduData_p + sizeof(struct ipv4_hdr));

        /* Update IP header */
        ipHdr->type_of_service = dscp << 2;
        ipHdr->time_to_live = 255;
        
        /* Swap ip addr */
        tmp_32 = ipHdr->src_addr;
        ipHdr->src_addr = ipHdr->dst_addr;
        ipHdr->dst_addr = tmp_32;

        /* Swap UDP ports */
        tmp_16 = udpHdr->src_port;
        udpHdr->src_port = udpHdr->dst_port;
        udpHdr->dst_port = tmp_16;
    }

    if(ADK_TWAMP_DP_IPV6_LLI_PATH == twamp_imc_path)
    {
        /* Set flags to indicate ip/udp hdr encapsulation */
        ipv6_uli_imc_param_p = (adk_ipv6_dp_uli_imc_data_t *)
            &(ADK_METADATA_PARAM(task, imc_param_start));

        /* Traffic Class */
        tmp_dscp = *((uint16_t *) (pduData_p + l3offset));
        tmp_dscp = (adk_htons(tmp_dscp)) & 0xF03F;
        dscp &= 0x3F;
        tmp_dscp |= (dscp << 6);
        *((uint16_t *) (pduData_p + l3offset)) = adk_ntohs(tmp_dscp);

        /* Hop Limit */
        *((uint8_t *) (pduData_p + l3offset + 7)) = 255;

        /* Swap IP addresses */
        /* Source IP */
        *((uint32_t *) (pduData_p + l3offset + 8)) = destIPv6[0];
        *((uint32_t *) (pduData_p + l3offset + 12)) = destIPv6[1];
        *((uint32_t *) (pduData_p + l3offset + 16)) = destIPv6[2];
        *((uint32_t *) (pduData_p + l3offset + 20)) = destIPv6[3];

        /* Destination IP */
        *((uint32_t *) (pduData_p + l3offset + 24)) = sourceIPv6[0];
        *((uint32_t *) (pduData_p + l3offset + 28)) = sourceIPv6[1];
        *((uint32_t *) (pduData_p + l3offset + 32)) = sourceIPv6[2];
        *((uint32_t *) (pduData_p + l3offset + 36)) = sourceIPv6[3];

        /* Swap UDP Ports */
        *((uint16_t *) (pduData_p + l4offset)) = destUDP_port;
        *((uint16_t *) (pduData_p + l4offset + 2)) = sourceUDP_port;

        /* Strip Ethernet header before F2M bind */
        ADK_PDU_HEAD_PULL(task, pduData_p, (void *), (l3offset));

        adk_md_p->exception_domain = ADK_FP_ORIGIN_DOMAIN;
        adk_md_p->func_idx = 0;
        adk_md_p->ip_offset = 0;
        adk_md_p->l4_offset -= l3offset;
    }

    bind_attr = (adk_twamp_responder_bind_attr_t *)
                        rte_atomic64_read(&responderObj[responderID].bind_attr);
    if(bind_attr != NULL) {
        *dst_mod_id = bind_attr->destModuleId ;
        *stage_info = bind_attr->stage_info;
        *destFunc =   bind_attr->destFunc[_DEST_FUNC_DEFAULT];
    } else {
        /*
         * Total number of packets dropped due to TWAMP to IP bind not defined.
         */
        twampGlobalStatistics[coreIndex]->noBindDrops++;
        debug_print("Error: IP Entry Function is not found \n");
        ret = ADK_E_ENTRY_FUNC_NOT_FOUND;
        goto _twamp_processing_exit;
    }

    /* Total number of packets transmitted to IP layer. */
    twampGlobalStatistics[coreIndex]->txPkts++;
    twampResponderStats[responderID].txPkts++;

_twamp_processing_exit:
    return (ret);
}

/* Create a reflector session */
static inline void
create_session(uint32_t hash_val,
               uint8_t responderID,
               adk_twamp_transport_object_t transportObject,
               adk_twamp_session_reflector_object_t **sessionObj)
{
    uint32_t first_hash, second_hash;
    first_hash = hash_val & 0xFFFF;
    second_hash = hash_val >> 16;
    adk_twamp_session_reflector_object_t *sessionRefObj = NULL;
    uint32_t sessionObjectId;
    uint32_t sessionId;

    sessionId =
        adk_rm_resource_reserve(responderObj[responderID].session_reflector_rm_hdl);

    sessionObjectId = (responderID * TWAMP_MAX_REFLECTORS_PPF_RESPONDER) + sessionId;

    /* sessionObjectId is the index in fallback array */
    *sessionObj = &sessionReflectorObj[sessionObjectId];

    (*sessionObj)->sessionId = sessionId;

    /* Copying the transport objects */
    ADK_MEMCPY (&((*sessionObj)->transportObject),
                &transportObject,
                sizeof(adk_twamp_transport_object_t));

    /* Checking first hash table */
    sessionRefObj = (adk_twamp_session_reflector_object_t *)
                                        reflectorFirstHashTbl[first_hash];
    /*
     * As the session reflector memory is pre-allocated during DP init, the
     * sessionRefObj will be NULL only for the first time when a session is
     * created. If the same session is recreated, then sessionRefObj will not be
     * NULL.
     */
    if((sessionRefObj && (!sessionRefObj->isUsed)) || (!sessionRefObj)) {
        reflectorFirstHashTbl[first_hash] = (void *) (*sessionObj);
#if ADK_DEBUG
        debug_print("Entry added in first hash table. Responder ID = %d,\
               session ID = %d \n", responderID, sessionId);
#endif
        return;
    }

    /* Checking second hash table */
    sessionRefObj = (adk_twamp_session_reflector_object_t *)
                                        reflectorSecondHashTbl[second_hash];
    if((sessionRefObj && (!sessionRefObj->isUsed)) || (!sessionRefObj)) {
        reflectorSecondHashTbl[second_hash] = (void *) (*sessionObj);
#if ADK_DEBUG
        debug_print("Entry added in second hash table. Responder ID = %d,\
               session ID = %d \n", responderID, sessionId);
#endif
        return;
    }

#if ADK_DEBUG
        debug_print("Entry added in fall back table. Responder ID = %d,\
               session ID = %d \n", responderID, sessionId);
#endif

}

/* Check if the session already exists within the responder */
static inline void
check_if_session_exists(uint32_t hash_val,
                        uint8_t responderID,
                        adk_twamp_transport_object_t transportObject,
                        adk_twamp_session_reflector_object_t **sessionObj,
                        bool *isPresent)
{
    uint32_t first_hash, second_hash;
    first_hash = hash_val & 0xFFFF;
    second_hash = hash_val >> 16;
    adk_twamp_session_reflector_object_t *sessionRefObj = NULL;
    uint32_t i;

    /* Checking first hash table */
    sessionRefObj = (adk_twamp_session_reflector_object_t *)
                                        reflectorFirstHashTbl[first_hash];
    if(sessionRefObj != NULL) {
         check_if_session_object_is_valid(sessionRefObj,
                                          transportObject,
                                          responderID,
                                          isPresent);
         if(*isPresent) {
             *sessionObj = sessionRefObj;
#if ADK_DEBUG
             debug_print("Hash table 1: Session already exists \n");
#endif
             return;
         }
    }

    /* Checking second hash table */
    sessionRefObj = (adk_twamp_session_reflector_object_t *)
                                        reflectorSecondHashTbl[second_hash];
    if(sessionRefObj != NULL) {
        check_if_session_object_is_valid(sessionRefObj,
                                         transportObject,
                                         responderID,
                                         isPresent);
        if(*isPresent) {
            *sessionObj = sessionRefObj;
#if ADK_DEBUG
            debug_print("Hash table 2: Session already exists \n");
#endif
            return;
        }
    }

    /* Linear search in fallback array */
    for(i = responderID * TWAMP_MAX_REFLECTORS_PPF_RESPONDER;
                        i < TWAMP_MAX_REFLECTORS_PPF_RESPONDER; i++) {
        check_if_session_object_is_valid(&sessionReflectorObj[i],
                                         transportObject,
                                         responderID,
                                         isPresent);
        if(*isPresent) {
            *sessionObj = &sessionReflectorObj[i];
#if ADK_DEBUG
            debug_print("Fallback: Session already exists \n");
#endif
            break;
        }
    }
    return;
}

static inline void
check_if_session_object_is_valid(adk_twamp_session_reflector_object_t *sessionReflectorObject,
                                 adk_twamp_transport_object_t transportObject,
                                 uint8_t responderId,
                                 bool *isPresent)
{
    if(sessionReflectorObject->isUsed) {
        if(sessionReflectorObject->transportObject.ipVersion == ADK_TWAMP_IPV4) {
            if((sessionReflectorObject->transportObject.ipAddr.sourceIpv4addr ==
                transportObject.ipAddr.sourceIpv4addr) &&
               (sessionReflectorObject->transportObject.sourceUDP_port ==
                transportObject.sourceUDP_port) &&
               (sessionReflectorObject->responderId == responderId)) {
                *isPresent = 1;
            }
        }

        if(sessionReflectorObject->transportObject.ipVersion == ADK_TWAMP_IPV6) {
            if((sessionReflectorObject->transportObject.ipAddr.sourceIpv6addr.addr[0] ==
                transportObject.ipAddr.sourceIpv6addr.addr[0]) &&
               (sessionReflectorObject->transportObject.ipAddr.sourceIpv6addr.addr[1] ==
                transportObject.ipAddr.sourceIpv6addr.addr[1]) &&
               (sessionReflectorObject->transportObject.ipAddr.sourceIpv6addr.addr[2] ==
                transportObject.ipAddr.sourceIpv6addr.addr[2]) &&
               (sessionReflectorObject->transportObject.ipAddr.sourceIpv6addr.addr[3] ==
                transportObject.ipAddr.sourceIpv6addr.addr[3]) &&
               (sessionReflectorObject->transportObject.sourceUDP_port ==
                transportObject.sourceUDP_port) &&
               (sessionReflectorObject->responderId == responderId)) {
                *isPresent = 1;
            }
        }
    }
}


/*
 * Session age out handler
 */
adk_dp_status_t adk_twamp_session_age_out (uint8_t responderId,
                                           uint32_t sessionId,
                                           bool isPresent,
                                           bool *isStale
                                           )
{
    uint32_t i, relative_max_sessions;
    uint64_t  currTxTicks;
    uint64_t  diffTicks;
    uint64_t  sessionTimeoutTicks;

#if ADK_DEBUG
    debug_print("Responder level session age-out processing.\
            Responder ID = %d \n", responderId);
#endif

    if(!isPresent) {
        /*
         * Scan session DB for the specific responder ID and age-out passive
         * sessions
         */
        relative_max_sessions = (responderId * TWAMP_MAX_REFLECTORS_PPF_RESPONDER) + TWAMP_MAX_REFLECTORS_PPF_RESPONDER;
        for(i = responderId * TWAMP_MAX_REFLECTORS_PPF_RESPONDER; i < relative_max_sessions; i++) {
            if(sessionReflectorObj[i].isUsed) {
                currTxTicks = rte_get_tsc_cycles();
                /* To Do: The roll over case is not handled intentionally. The
                 * rte_get_tsc_cycles() API will return a 64 bit value in ticks
                 * since the system boot. The roll over will happen after several
                 * years and it depends on the CPU clock speed.
                 */
                diffTicks = currTxTicks - (rte_atomic64_read(&sessionReflectorObj[i].prevTicks));
                sessionTimeoutTicks = responderObj[responderId].cpuTicksPerSec *
                    responderObj[responderId].session_timeout;
                if(diffTicks > sessionTimeoutTicks)
                {
                    adk_rm_resource_free(responderObj[responderId].session_reflector_rm_hdl,
                                         sessionReflectorObj[i].sessionId);
                    /* Tracking the number of reflector sessions per responder */
                    twampInternalResponderStats[responderId].numActiveSessions--;
                    rte_atomic64_set(&sessionReflectorObj[i].prevTicks, 0);
                    sessionReflectorObj[i].isUsed = 0;
                    sessionReflectorObj[i].seqNum = 0;
#if ADK_DEBUG
                    debug_print("\n Twamp Responder ID = %d Session ID %d aged out successfully.\
                           \n", sessionReflectorObj[i].responderId, sessionReflectorObj[i].sessionId);
#endif
                }
            } else {
                continue;
            }
        }
    } else {
        /* Stale session age-out handling */
        i = (responderId * TWAMP_MAX_REFLECTORS_PPF_RESPONDER) + sessionId;
        if(sessionReflectorObj[i].isUsed) {
            currTxTicks = rte_get_tsc_cycles();
            /* To Do: The roll over case is not handled intentionally. The
             * rte_get_tsc_cycles() API will return a 64 bit value in ticks
             * since the system boot. The roll over will happen after several
             * years and it depends on the CPU clock speed.
             */
            diffTicks = currTxTicks - (rte_atomic64_read(&sessionReflectorObj[i].prevTicks));
            sessionTimeoutTicks = responderObj[responderId].cpuTicksPerSec *
                responderObj[responderId].session_timeout;
            if(diffTicks > sessionTimeoutTicks)
            {
                adk_rm_resource_free(responderObj[responderId].session_reflector_rm_hdl,
                                     sessionReflectorObj[i].sessionId);
                /* Tracking the number of reflector sessions per responder */
                twampInternalResponderStats[responderId].numActiveSessions--;
                rte_atomic64_set(&sessionReflectorObj[i].prevTicks, 0);
                sessionReflectorObj[i].isUsed = 0;
                sessionReflectorObj[i].seqNum = 0;
                *isStale = 1;
#if ADK_DEBUG
                debug_print("\n Twamp Responder ID = %d Session ID %d aged out successfully.\
                       \n", sessionReflectorObj[i].responderId, sessionReflectorObj[i].sessionId);
#endif
            }
        }
    }

    return ADK_SUCCESS;
}

/*!
 * =============================================================
 * adk_twamp_dp_process_pkt
 *
 * Structure of task params:
 *
 * taskParams[0] :
 *
 * Brief:
 *
 * This function is the entry point for processing of TWAMP packets received
 * from NW.
 * Session create/update handler called at responder level atomicity
 *
 * =============================================================
 */
adk_dp_status_t adk_twamp_dp_pkt_proc_lli (void **task_p,
                        uint8_t task_count,
                        __attribute__((unused)) uint32_t task_mask,
                        __attribute__((unused)) void *ring,
                        __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry,
                        int coreIndex)
{
    adk_dp_status_t ret = ADK_SUCCESS;
    uint8_t drop_count = 0;
    uint32_t drop_mask = 0, cur_bit = 1;
    void *ring_hdl = NULL;
    adk_sif_cfg_pipeline_info_t *next_stage_entry_hdl = NULL;
    uint32_t dst_mod_id = 0;
    uint32_t dst_mod_task_count[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {0};
    uint32_t dst_mod_task_mask[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {0};
    uint16_t dst_mod_dst_func[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {0};
    uint16_t destFunc = 0;
    adk_pal_dp_stage_info_t *stage_info[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {NULL};
    adk_pal_dp_stage_info_t *dp_stage_info = NULL;
    void *dst_mod_ring_hdl[ADK_SIF_CFG_PIPELINE_MODULE_MAX] = {NULL};
    adk_sif_cfg_pipeline_info_t
        *dst_mod_next_stage_entry_hdl[ADK_SIF_CFG_PIPELINE_MODULE_MAX] =
        {NULL};
    int i;
    uint32_t num_pkts_sent = 0;
#if defined(ADK_PAL_DYNAMIC_PPL)
    uint32_t responderID;
#endif

    /* Session create/update handling */
#if ADK_DEBUG
    debug_print("Responder level processing to create/update sessions \n");
#endif

    for(i=0,cur_bit=1; i<ADK_DP_BULK_RX_SIZE; ++i,(cur_bit=cur_bit<<1))
    {
        if(!(task_mask & cur_bit))
            continue;

        ret = adk_twamp_dp_pkt_proc_lli_single(task_p[0], coreIndex, &destFunc, &dp_stage_info, &dst_mod_id);
        if (unlikely((ADK_E_INTERNAL_ERROR == ret) || (ADK_E_ENTRY_FUNC_NOT_FOUND == ret)))
        {
            /* Free the task */
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[0]));
            drop_mask |= cur_bit;
            drop_count++;
            return ADK_SUCCESS;
        }
        else if (unlikely(ADK_E_TASK_CONSUMED == ret))
        {
            /* Task consumed. Do not free the task */
            drop_mask |= cur_bit;
            drop_count++;
        }
        else
        {
            dst_mod_task_mask[dst_mod_id] |= cur_bit;
            dst_mod_task_count[dst_mod_id]++;
            dst_mod_dst_func[dst_mod_id] = destFunc;
            stage_info[dst_mod_id] = dp_stage_info;
            dst_mod_ring_hdl[dst_mod_id] = ring_hdl;
            dst_mod_next_stage_entry_hdl[dst_mod_id] = next_stage_entry_hdl;
        }
    }

    task_count -= drop_count;

    /* Bind with IP module. Currently there is no burst processing, so no need to
     * iterate all the SIF pipeline modules. */
#if defined(ADK_PAL_DYNAMIC_PPL)
    responderID = ADK_METADATA_PARAM(task_p[0], pdu_flow_id); /* To DO - Burst handling is not done */
#if ADK_DEBUG
    debug_print("Lock ID used in PAL staging = Source Flow ID / Responder ID = %d \n", responderID);
#endif
    num_pkts_sent = adk_pal_dp_next_stage_event_handle_with_stage_info(task_p, dst_mod_task_count[dst_mod_id],
                                      dst_mod_task_mask[dst_mod_id],
                                      /* TWAMP->IP is F2M bind.So, source flow ID is
                                       * passed. */
                                      stage_info[dst_mod_id], responderID, 0x00, ADK_TASK_TYPE_FWD);

    if(unlikely(num_pkts_sent != 1)) { /* To Do - Burst handling is not done */
        /* Freeing the task */
        ADK_TWAMP_DP_DROP_TASK(task_p[0]);
    }
#else
    if (adkDestEntryFunction_g[dst_mod_dst_func[dst_mod_id]])
        (adkDestEntryFunction_g[dst_mod_dst_func[dst_mod_id]])(task_p, dst_mod_task_count[dst_mod_id],
                                                          dst_mod_task_mask[dst_mod_id],
                                                          dst_mod_ring_hdl[dst_mod_id], dst_mod_next_stage_entry_hdl[dst_mod_id], coreIndex);
#endif
    return ADK_SUCCESS;
}


/* DP init function to be invoked by DP process */
adk_status_t adk_twamp_dp_init(void)
{
    adk_status_t        retVal = ADK_SUCCESS;
    const struct rte_memzone *mz = NULL;
    adk_twamp_dp_pvt_data_t *dp_pvt_p;

    /* Lookup the shared memory that gets created by control plane TWAMP
     * module initialization */
    mz = rte_memzone_lookup(ADK_TWAMP_SHM_MEMZONE_NAME);
    if (mz == NULL)
    {
        debug_print("Error: Failed to lookup TWAMP shared memory %s\n",
                    ADK_TWAMP_SHM_MEMZONE_NAME);
        retVal = ADK_E_MEM_UNAVAILABLE;
        goto _adk_twamp_dp_init_exit;
    }

    /* Check if the TWAMP module has been completely initialized */
    dp_pvt_p = (adk_twamp_dp_pvt_data_t *) mz->addr;
    if (dp_pvt_p->cookie != ADK_TWAMP_MODULE_COOKIE)
    {
        debug_print("TWAMP Module Initialization from control plane not yet"
                    " complete. Please wait and try again later.\n");
        retVal = ADK_E_MEM_UNAVAILABLE;
        goto _adk_twamp_dp_init_exit;
    }

    /* Attach to the TWAMP shared memory */
    adk_twamp_dp_pvt_g_p = dp_pvt_p;

    adkDestEntryFunction_g[ADK_TWAMP_LLI_FN_ID]= adk_twamp_dp_pkt_proc_lli;
    adkDestEntryFunction_g[ADK_TWAMP_LLI_RING_FN_ID]= NULL;

    reflectorFirstHashTbl = adk_twamp_dp_pvt_g_p->reflectorFirstHashTbl_p ;
    reflectorSecondHashTbl = adk_twamp_dp_pvt_g_p->reflectorSecondHashTbl_p ;
    sessionReflectorObj = adk_twamp_dp_pvt_g_p->sessionReflectorObj_p ;
    responderObj = adk_twamp_dp_pvt_g_p->responderObj_p ;
    twampGlobalStatistics = adk_twamp_dp_pvt_g_p->dpTwampGlobalStatistics_p ;
    twampResponderStats = adk_twamp_dp_pvt_g_p->dpTwampResponderStats_p;
    twampInternalResponderStats = adk_twamp_dp_pvt_g_p->dpTwampInternalResponderStats_p;

_adk_twamp_dp_init_exit:

    return retVal;
}
