/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#ifndef __ADK_TWAMP_DP_H__
#define __ADK_TWAMP_DP_H__

#include <adk_sys.h>
#include "adk_ip_dp_export.h"
#include "adk_ip_config.h"
#include "adk_rm.h"
#include "adk_timer.h"

#define ntohll(x) ( ( (uint64_t)(ntohl( (uint32_t)((x << 32) >> 32) )) << 32) |\
            ntohl( ((uint32_t)(x >> 32)) ) )

#define htonll(x) ntohll(x)

#define ADK_TWAMP_MIN_PKT_LEN 41

#define ADK_TWAMP_SENDER_HEADER_LEN 14
#define ADK_TWAMP_REFLECTOR_HEADER_LEN 27
#define ADK_UDP_HEADER_LEN 8

#define ADK_TWAMP_SHM_MEMZONE_NAME   "adk_twamp_shm_zone"

#define ADK_TWAMP_DP_DROP_TASK(mbuf) \
    do{ \
        if (mbuf != NULL) \
           rte_pktmbuf_free((struct rte_mbuf *)mbuf); \
    }while(0);

typedef enum adk_twamp_ip_version_e {
    ADK_TWAMP_IPV4 = 0,
    ADK_TWAMP_IPV6 = 1
}adk_twamp_ip_version_t;

typedef struct adk_twamp_pdu_s
{
    /*Reflector sequence number */
    uint32_t        seqNum;
    /* Reflector transmit time stamp */
    uint32_t        integerTS;
    uint32_t        fractionalTS;
    uint32_t        errEstimate_And_MBZ;
    /* Reflector receive time stamp */
    uint32_t        integerRcvTS;
    uint32_t        fractionalRcvTS;
    /* Sender sequence number */
    uint32_t        senderSeqNum;
    /* Sender time stamp */
    uint32_t        integerSendTS;
    uint32_t        fractionalSendTS;
    uint32_t        senderErrEstimate_And_MBZ;
    uint8_t         senderTTL;
} __attribute__((packed))adk_twamp_pdu_t;


typedef struct adk_twamp_ipv6_addr_s
{
    /*! The array of 32 bit words containing the IPv6 address.
     *  The higher bytes of the IPv6 address are stored starting
     *  in ipv6Addr[0].
     */
    uint32_t addr[4];
} adk_twamp_ipv6_addr_t;

typedef struct adk_twamp_transport_object_s {
    adk_twamp_ip_version_t ipVersion;
    /* A reflector session can be either IPv4 or IPv6 only */
    union {
       uint32_t sourceIpv4addr;
       adk_twamp_ipv6_addr_t sourceIpv6addr;
    }ipAddr;
    uint16_t sourceUDP_port;
}adk_twamp_transport_object_t;

/* Bind attributes to be atomically updated from control plane */
typedef struct adk_twamp_responder_bind_attr_s{
    /* Module ID for F2M with IP module */
    uint32_t destModuleId;
    uint16_t destFunc[_DEST_FUNC_MAX];
    adk_pal_dp_stage_info_t *stage_info;
}adk_twamp_responder_bind_attr_t;

/* Responder object to store the resource manager handle */
typedef struct adk_twamp_responder_object_s{
    uint16_t         ip_version;
    adk_rm_hdl_t     session_reflector_rm_hdl;
    bool             isValid;
    /*
     * Epoch time in NTP format.
     *
     * Note: The CPU clocks mapped to this time will be used
     * as the previous time ticks in data path. The time difference will be
     * added to this base NTP time and it will be embossed in TWAMP packet.
     *
     * The baseNtpTimeHi captures the time in seconds since 1st jan 1900.
     * This will roll over after 136 years since 1900 which is year 2036.
     * Calculation:
     *   (2^32 - 1) / ( 60(min) * 60(hrs) * 24(days) * 365(yrs)) => 136 years
     */
    uint32_t baseNtpTimeHi;
    uint32_t baseNtpTimeLo;

    uint64_t cpuTicksPerSec;

    /* Mapped CPU clock ticks when Epoch time was obtained from linux.
     *
     * Note: The clock ticks does not map exactly with
     * the Epoch time. If a context switch happens before reading the clock
     * ticks, then the difference between linux time and CPU clock ticks will be
     * even more. The assumption is that the TWAMP session originator or the
     * sender will run without any preemption and correct the senders clock
     * periodically to take care of these kind of timestamp errors introduced by
     * the TWAMP session reflectors.
     */
    uint64_t epochCpuTicks;

    uint32_t        session_timeout;
    /* Timer Event Interval in ticks */
    uint64_t        timerEventInterval;
    /* Timer Handle to be used during un-install */
    adk_handle_t    timer_handle;
    /* adk_twamp_responder_bind_attr_t */
    rte_atomic64_t  bind_attr;
}adk_twamp_responder_object_t;

/* Session reflector */
typedef struct adk_twamp_session_reflector_object_s{
    /* Is the object free to use */
    bool       isUsed;
    /* Responder to which this session object belongs to */
    uint8_t    responderId;
    uint32_t   seqNum;
    adk_twamp_transport_object_t transportObject;
    uint32_t sessionId;
    /* Previous time stamp for age-out purpose */
    rte_atomic64_t  prevTicks;
}adk_twamp_session_reflector_object_t;

adk_dp_status_t adk_twamp_session_age_out (uint8_t responderId,
                                           uint32_t sessionId,
                                           bool isPresent,
                                           bool *isStale);

adk_dp_status_t
adk_twamp_dp_pkt_proc_lli_single(void *task_p, int coreIndex, uint16_t *destFunc,
                                 adk_pal_dp_stage_info_t **stage_info,
                                 uint32_t *dst_mod_id);
#endif
