/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_twamp_internal.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _twamp  TWAMP Module
 *
 * @{
 *
 * @brief
 *
 * This file contains all the internal data structures and API prototypes
 * required for the TWAMP module.
 *
 ***********************************************************************/
#ifndef _ADK_TWAMP_INTERNAL_H_
#define _ADK_TWAMP_INTERNAL_H_

#include <unistd.h>

#include "adk_include.h"

#include "adk_twamp_const_internal.h"
#include "adk_twamp.h"
#include "adk_twamp_dp.h"
#include "adk_rm.h"

/* NTP related constant */
/*! Offset between NTP base and UTC base */
/* Number of seconds from 1st January 1900 to start of
 * Unix epoch which is 1st January 1970
 */
#define TWAMP_NTP_BASETIME                  2208988800ul
/*! NTP scaling fraction */
#define TWAMP_NTP_SCALE_FRAC                ((uint64_t)1 << 32)

#define ADK_TWAMP_MODULE_COOKIE \
    ((ADK_TWAMP_MODULE_ID << 24) | 0x793134)

#define ADK_TWAMP_MODULE_HANDLE_COOKIE \
    ((ADK_TWAMP_MODULE_ID << 24) | 0x496124)

#define ADK_TWAMP_RESPONDER_FLOW_HANDLE_COOKIE \
    ((ADK_TWAMP_MODULE_ID << 24) | 0x241249)

/*!
 * =============================================================
 *
 * @typedef adk_twamp_log_level_t
 *
 * @brief   The 32-bit TWAMP module's log-level
 *
 *  This data type is used for setting the log-level of TWAMP module. It is
 *  represented using unsigned long integer and is a logical OR or following
 *  log-types defined by TWAMP module:
 *
 *  TWAMP_LOG_FATAL       For logging any unrecoverable fatal error
 *
 *  TWAMP_LOG_ERROR       For logging all other errors
 *
 *  TWAMP_LOG_WARNING     For logging any any warnings
 *
 *  TWAMP_LOG_INFO        For logging all resource allocations, etc
 *
 *  TWAMP_LOG_FUNCS       For logging entry to all internal/external functions
 *
 *  TWAMP_LOG_HW          For logging all access to EIOA hardware
 *
 *  TWAMP_LOG_PTN         For logging all learning/unlearning of FPL Tree entries
 *
 *  TWAMP_LOG_SE_MEM      For logging all access to SE memory
 *
 *  TWAMP_LOG_ALL         For logging all above
 *
 *  The following are some of the pre-defined commonly used log-levels.
 *
 *  TWAMP_LOG_LEVEL_ERROR     enable TWAMP_LOG_ERROR & TWAMP_LOG_FATAL only
 *
 *  TWAMP_LOG_LEVEL_WARNING   enable TWAMP_LOG_WARNING, TWAMP_LOG_ERROR &
 *                           TWAMP_LOG_FATAL only
 *
 *  TWAMP_LOG_LEVEL_INFO      enable TWAMP_LOG_INFO, TWAMP_LOG_WARNING,
 *                           TWAMP_LOG_ERROR & TWAMP_LOG_FATAL only
 *
 *  TWAMP_LOG_LEVEL_VERBOSE   enable all types of logs. Same as TWAMP_LOG_ALL
 *
 *  TWAMP_LOG_LEVEL_DEFAULT   Compile time default log level setting
 *
 * =============================================================
 */
typedef uint32_t    adk_twamp_log_level_t;

#define TWAMP_LOG_DISABLE      (0 << 0)
#define TWAMP_LOG_FATAL        (1 << 0)
#define TWAMP_LOG_ERROR        (1 << 1)
#define TWAMP_LOG_WARNING      (1 << 2)
#define TWAMP_LOG_INFO         (1 << 3)
#define TWAMP_LOG_FUNCS        (1 << 4)
#define TWAMP_LOG_HW           (1 << 5)
#define TWAMP_LOG_PTN          (1 << 6)
#define TWAMP_LOG_SUBNET       (1 << 7)
#define TWAMP_LOG_EXT_API      (1 << 8)

#define TWAMP_LOG_ALL          0xFFFFFFFF

#define TWAMP_LOG_LEVEL_ERROR         (TWAMP_LOG_ERROR | TWAMP_LOG_FATAL)
#define TWAMP_LOG_LEVEL_WARNING       (TWAMP_LOG_LEVEL_ERROR | TWAMP_LOG_WARNING)
#define TWAMP_LOG_LEVEL_INFO          (TWAMP_LOG_LEVEL_WARNING | TWAMP_LOG_INFO)
#define TWAMP_LOG_LEVEL_VERBOSE       TWAMP_LOG_ALL

#define TWAMP_LOG_LEVEL_DEFAULT       TWAMP_LOG_LEVEL_ERROR

/*
 * TWAMP Responder stats
 */
typedef struct adk_twamp_internal_responder_stats_s
{
    /*
     * Number of active test sessions under this responder. This counter is
     * updated based on the packet events and not based on timer events.
     */
    uint32_t        numActiveSessions;
} adk_twamp_internal_responder_stats_t;

/***********************************************************************
******* MODULE LEVEL STRUCT HAVING POINTERS TO ALL DP TABLES **********
***********************************************************************/
typedef struct adk_twamp_dp_pvt_data_s
{
    /* Cookie for validation */
    uint32_t cookie;
    void **reflectorFirstHashTbl_p;
    void **reflectorSecondHashTbl_p;
    /* Max Sessions = Max responder * Max sessions per responder */
    adk_twamp_session_reflector_object_t *sessionReflectorObj_p;
    adk_twamp_responder_object_t *responderObj_p;
    /*
     * Global Stats
     *
     * Note: The Global stats is computed by aggregating all the per core level
     * stats, though the whole TWAMP module is running as an atomic stage in the
     * data plane. There is a scope for optimization here by maintaining a
     * common stats object instead of per core object. This holds good for
     * responder level stats and timers also.
     */
    adk_twamp_global_stats_t **dpTwampGlobalStatistics_p;
    /* Responder level stats. This is not implemented as per core stats. The
     * number of active sessions counter needs to be incremented and
     * decremented. If it is implemented as per core stats, to check the
     * maximum number of sessions, the core level stats needs to be summed
     * up which is expensive in data plane.
     */
    adk_twamp_responder_stats_t *dpTwampResponderStats_p;
    /* Internal TWAMP Responder Stats */
    adk_twamp_internal_responder_stats_t *dpTwampInternalResponderStats_p;
} adk_twamp_dp_pvt_data_t;

typedef enum adk_twamp_tables_e
{
    ADK_TWAMP_RESPONDER_TBL = 0,
    ADK_TWAMP_TABLES_MAX
} adk_twamp_tables_t;

typedef struct adk_twamp_pend_list_callback_data_s
{
    adk_twamp_tables_t tables;
    adk_module_handle_t handle;
} adk_twamp_pend_list_callback_data_t;

typedef struct adk_twamp_rcu_pend_list_s
{
    uint32_t num_elements;
    void     *list[ADK_TWAMP_RESPONDER_TBL_THRESHOLD];
} adk_twamp_rcu_pend_list_t;

/*!
 * =============================================================
 * @struc  adk_twamp_module_info_t
 *
 * @brief ADK TWAMP module information.
 *
 * =============================================================
 */
typedef struct adk_twamp_module_info_s
{
    /*! Cookie */
    uint32_t cookie;
    /*Entry function pointers in DP process*/
    uint16_t       cpuEntryFn_LLI[_DEST_FUNC_MAX];
#if 0
    void **reflectorFirstHashTbl_p;
    void **reflectorSecondHashTbl_p;
    /* Max Sessions = Max responder * Max sessions per responder */
    adk_twamp_session_reflector_object_t *sessionReflectorObj_p;
    adk_twamp_responder_object_t *responderObj_p;
#endif
    adk_rm_hdl_t     responder_rm_hdl;

    /*
     * There are 2 responder IDs used in the fast path.
     * One is the absolute responder Id and the other is relative responder Id.
     * The absolute and relative responder Id is a pair.
     *
     * Absolute responder Id can range from <0, 2 * TWAMP_MAX_PPF_RESPONDER - 1>.
     * It is used by Twamp Timer (first packet of absolute responder Id installs the
     * timer ), responder validity check, staging and bind in Fast path.
     *
     * Relative responder Id can range from <0, TWAMP_MAX_PPF_RESPONDER - 1>
     * It is used to look into the session Table, NTP time calculation and responder
     * level stats.
     *
     * This array maintains the toggle flag to toggle the most significant bit
     * of absolute responder Id.
     */
    uint8_t          responderToggleFlag[TWAMP_MAX_PPF_RESPONDER];
    uint8_t         numActiveResponders;
#if 0
    /*
     * Global Stats
     *
     * Note: The Global stats is computed by aggregating all the per core level
     * stats, though the whole TWAMP module is running as an atomic stage in the
     * data plane. There is a scope for optimization here by maintaining a
     * common stats object instead of per core object. This holds good for
     * responder level stats and timers also.
     */
    adk_twamp_global_stats_t **dpTwampGlobalStatistics_p;
    /* Responder level stats. This is not implemented as per core stats. The
     * number of active sessions counter needs to be incremented and
     * decremented. If it is implemented as per core stats, to check the
     * maximum number of sessions, the core level stats needs to be summed
     * up which is expensive in data plane.
     */
    adk_twamp_responder_stats_t *dpTwampResponderStats_p;
    /* Internal TWAMP Responder Stats */
    adk_twamp_internal_responder_stats_t *dpTwampInternalResponderStats_p;
#endif

    adk_twamp_dp_pvt_data_t     *dp_pvt_p;
    /* Table specific pending list */
    adk_twamp_rcu_pend_list_t   pend_list[ADK_TWAMP_TABLES_MAX];
} adk_twamp_module_info_t;

adk_dp_status_t
adk_twamp_dp_table_init(adk_module_handle_t twampHandle,
                                adk_twamp_dp_pvt_data_t *twamp_dp_pvt_p);

#define ADK_TWAMP_LOG_FATAL(mod_handle, err, fmt, args ...) \
        do { \
                    ADK_LOG_FATAL_ERROR(sysHandle, mod_handle, fmt, ## args); \
                    ADK_LOG_FATAL_ERROR(sysHandle, mod_handle, "%s\n", \
                                                                    adk_sys_status_get(sysHandle, err)); \
                } while (0)

#define ADK_TWAMP_LOG_ERROR(mod_handle, err, fmt, args ...) \
        do { \
                    ADK_LOG_ERROR(sysHandle, mod_handle, fmt, ## args); \
                    ADK_LOG_ERROR(sysHandle, mod_handle, "%s\n", \
                                                        adk_sys_status_get(sysHandle, err)); \
                } while (0)

#define ADK_TWAMP_LOG_WARNING(mod_handle, fmt, args ...) \
        ADK_LOG_WARNING(sysHandle, mod_handle, fmt, ## args)

#define ADK_TWAMP_LOG_INFO(mod_handle, fmt, args ...) \
        ADK_LOG_INFO(sysHandle, mod_handle, fmt, ## args)

#define ADK_TWAMP_LOG_FORCE(fmt, args ...) \
        ADK_FORCE_LOG_ERROR(fmt, ## args);

#define ADK_TWAMP_STAT_PRINT(name, counter) \
    if ((counter) != 0)                     \
{                                           \
    ADK_INFO_WRITE((name), (counter));      \
}

#define ADK_TWAMP_LOG_FUNC_ENTRY(mod_handle) \
        ADK_LOG_FUNC_ENTRY(sysHandle, mod_handle)

#define ADK_TWAMP_LOG_FUNC_EXIT(mod_handle) \
        ADK_LOG_FUNC_EXIT(sysHandle, mod_handle)

#define ADK_TWAMP_LOG_INTERNAL_FUNC_ENTRY(mod_handle) \
        ADK_LOG_INTERNAL_FUNC_ENTRY(sysHandle, mod_handle)

#define ADK_TWAMP_LOG_INTERNAL_FUNC_EXIT(mod_handle) \
        ADK_LOG_INTERNAL_FUNC_EXIT(sysHandle, mod_handle)


adk_status_t adk_twamp_shutdown(adk_module_handle_t twampHandle);

bool adk_twamp_valid_module_handle(adk_module_handle_t twampHandle);

adk_status_t adk_twamp_custom_bind(adk_sys_handle_t sysHandle,
                                      adk_bind_info_t *bindInfo_p);

adk_status_t adk_twamp_custom_unbind(adk_sys_handle_t sysHandle,
                                        adk_bind_info_t *bindInfo_p);

adk_status_t adk_twamp_sys_log(adk_module_handle_t twampHandle,
                              adk_log_level_t sys_log_level);

const char *adk_twamp_error_name(adk_status_t err);

void
adk_twamp_rcu_pend_list_clear(void *rock, void *data);

adk_status_t
adk_twamp_map_sys_log(
    adk_module_handle_t twampHandle,
    adk_log_level_t sysLogLevel,
    adk_log_level_t *twampLogLevel);

#endif /* End of _ADK_TWAMP_INTERNAL_H_*/

/*! end addtogroup _twamp
 * @}
 */

