/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ncp_internal.h"

/* bitmap handling */

#define BITS_PER_INT 32

typedef struct BitField
{
    uint32_t     size;
    uint32_t     words_needed;
    uint32_t    *bits;
    uint32_t     firstFree;
} BitField_t;

static void
bit_set(
    BitField_t      *bits,
    uint32_t     size,
    uint32_t     strt)
{
    uint32_t i;
    uint32_t offs;

    /*
     * If the requested size is longer than the
     * size of the bit field then truncate size.
     */
    if (size + strt > bits->size)
    {
        size = bits->size - strt;
    }

    /*
     * Calculate the starting NCP_U32 index in the bit array (i)
     * and the bit offset in that NCP_U32 (offs).
     */
    i = strt / BITS_PER_INT;
    offs = strt % BITS_PER_INT;

    /*
     * Handle the trivial case were all the bits to set
     * are contained in the first word.
     */
    if (size < (BITS_PER_INT - offs))
    {
        /*
         * Set 'size' bits 'offs' bits into word 'i' of the
         * bit array.
         */
        bits->bits[i] |= ((1 << size) - 1) << offs;
        return;
    }

    /*
     * Set bits from offs to the end of the first word
     * and update 'i' and size' to reflect this.
     */
    if (offs)
    {
        /*
         * Set all bits from 'offs' to the end of word 'i'.
         */
        bits->bits[i] |= ((1 << (BITS_PER_INT - offs)) - 1) << offs;

        /*
         * Update 'i' to point to the next word in the array
         * and 'size' to account for bits already set.
         */
        i++;
        size -= BITS_PER_INT - offs;
    }

    /*
     * While we have full words to set, fill them.
     */
    while (size >= BITS_PER_INT)
    {
        /* Set all bits in word i, increment i */
        bits->bits[i++] = 0xffffffff;

        /* Decrement size by count of bits in one int. */
        size -= BITS_PER_INT;
    }

    /* Set bits remaining in last word. */
    if (size)
    {
        bits->bits[i] |= (1 << size) - 1;
    }
}

static void
bit_clr(
    BitField_t      *bits,
    uint32_t     size,
    uint32_t     strt)
{
    uint32_t i;
    uint32_t offs;

    /*
     * If the requested size is longer than the
     * size of the bit field then truncate size.
     */
    if (size + strt > bits->size)
    {
        size = bits->size - strt;
    }

    /*
     * Calculate the starting NCP_U32 index in the bit array (i)
     * and the bit offset in that NCP_U32 (offs).
     */
    i = strt / BITS_PER_INT;
    offs = strt % BITS_PER_INT;

    /*
     * Handle the trivial case were all the bits to clear
     * are contained in the first word.
     */
    if (size < (BITS_PER_INT - offs))
    {
        bits->bits[i] &= ~(((1 << size) - 1) << offs);
        return;
    }

    /*
     * Clear bits from offs to the end of the first word
     * and update 'i' and size' to reflect this.
     */
    if (offs)
    {
        /*
         * Clear all bits from 'offs' to the end of word 'i'.
         */
        bits->bits[i] &= ~(((1 << (BITS_PER_INT - offs)) - 1) << offs);

        /*
         * Update 'i' to point to the next word in the array
         * and 'size' to account for bits already cleared.
         */
        i++;
        size -= BITS_PER_INT - offs;
    }

    /*
     * While we have full words to clear, fill them.
     */
    while (size >= BITS_PER_INT)
    {
        /* Set all bits in word i, increment i */
        bits->bits[i++] = 0;

        /* Decrement size by count of bits in one NCP_U32. */
        size -= BITS_PER_INT;
    }

    /* Set bits remaining in last word. */
    if (size)
    {
        bits->bits[i] &= ~((1 << size) - 1);
    }
}

static uint32_t
bit_num_clr(
    BitField_t      *bits,
    uint32_t     strt,
    uint32_t     limit)
{
    uint32_t i;
    uint32_t offs;
    uint32_t count = 0;

    /*
     * Calculate the starting NCP_U32 index in the bit array (i)
     * and the bit offset in that NCP_U32 (offs).
     */
    i = strt / BITS_PER_INT;
    offs = strt % BITS_PER_INT;

    /*
     * If we start in the middle of a word, count the
     * bits in that word 1 by 1.
     */
    if (offs)
    {
        /* While we still haven't reached the end of the word. */
        while (offs < BITS_PER_INT)
        {
            /* If we hit a set bit we are done, return current count. */
            if (bits->bits[i] & (1 << offs))
            {
                return count;
            }

            /* Increment count and offset to handle this set bit. */
            count++;
            offs++;
        }

        /* We are finished with this word, point to the next. */
        i++;
    }

    /*
     * While we are:
     *   1) not at the end of the bit field
     *   2) looking at a word that is all cleared.
     *   3) haven't seen 'limit' cleared bits yet.
     * count the entire word as clear and increment the word pointer (i).
     */
    while ((i < bits->words_needed) &&
           (bits->bits[i] == 0) &&
           (count < limit))
    {
        count += BITS_PER_INT;
        i++;
    }

    /*
     * If we exitted the previous loop with
     * count < limit and i != bits->words_needed
     * then we exitted with bits->bits[i] != 0.
     * We count the number of leading cleard bits
     * remaining in this word (up to limit).
     */
    if ((count < limit) && (i != bits->words_needed))
    {
        for (offs = 0;
             ((bits->bits[i] & (1 << offs)) == 0) && (count < limit);
             offs++)
        {
            count++;
        }
    }

    return count;
}

static uint32_t
bit_next_clr(
    BitField_t      *bits,
    uint32_t     strt)
{
    uint32_t i;
    uint32_t offs;
    uint32_t count = 0;

    /*
     * Calculate the starting NCP_U32 index in the bit array (i)
     * and the bit offset in that NCP_U32 (offs).
     */
     i = strt / BITS_PER_INT;
     offs = strt % BITS_PER_INT;

     /* --- scan for a cleared bit from [start, end] --- */

     /*
      * If the bit at 'strt' is cleared, advance start
      * by the count of cleared bits.
      */
     if ((bits->bits[i] & (1 << offs)) == 0)
     {
         /* start at strt and look for upto size - strt bits. */
         strt += bit_num_clr(bits, strt, bits->size - strt);
     }
     /* If we don't have any cleared bits in this range, try the other */
     if (strt < bits->size)
     {
         /*
          * Calculate the starting NCP_U32 index in the bit array (i)
          * and the bit offset in that NCP_U32 (offs).
          */
         i = strt / BITS_PER_INT;
         offs = strt % BITS_PER_INT;

         /*
          * If starting in the middle of the word, look for first cleared bit in
          * that word.
          */
         if (offs)
         {
             /* While still in the word. */
             while (offs < BITS_PER_INT)
             {
                 /* If current bit is cleared, return pointer to that bit. */
                 if ((bits->bits[i] & (1 << offs)) == 0)
                 {
                     return strt + count;
                 }

                 /* else increment count and offs. */
                 count++;
                 offs++;
             }

             /* We are done with the first word, advance the pointer. */
             i++;
         }

         /* Search for first word that is not all set. */
         while ((i < bits->words_needed) && (bits->bits[i] == 0xffffffff ))
         {
             count += BITS_PER_INT;
             i++;
         }

         /* If we ran out of bits in the bit_field, return size. */
         if (i >= bits->words_needed)
         {
             return bits->size;
         }

         /* Count leading one bits in this word. */
         for (offs = 0; bits->bits[i] & (1 << offs); offs++)
         {
             count++;
         }

         /*
          * Count is now the number of bits we skipped to get to
          * the bit we found. strt + count is index of that bit.
          */
         return strt + count;

     }
     else
     {
         /*
          * nothing in [strt,end] ,scan for a cleared bit from [0, strt)
          * we do not care if the last bits in the map are cleared..if the
          * start is cleared, we stop.
          */
         strt = 0;
         i = strt / BITS_PER_INT;
         offs = strt % BITS_PER_INT;
         /*
          * If the bit at 'strt' is cleared,
          * we are done.
          */
         if ((bits->bits[i] & (1 << offs)) == 0)
         {
             return strt;
         }
         /*
          * If starting in the middle of the word, look for first cleared bit in
          * that word.
          */
         if (offs)
         {
             /* While still in the word. */
             while (offs < BITS_PER_INT)
             {
                 /* If current bit is cleared, return pointer to that bit. */
                 if ((bits->bits[i] & (1 << offs)) == 0)
                 {
                     return strt + count;
                 }

                 /* else increment count and offs. */
                 count++;
                 offs++;
             }

             /* We are done with the first word, advance the pointer. */
             i++;
         }

         /* Search for first word that is not all set. */
         while ((i < bits->words_needed) && (bits->bits[i] == 0xffffffff ))
         {
             count += BITS_PER_INT;
             i++;
         }

         /* If we ran out of bits in the bit_field, return size. */
         if (i >= bits->words_needed)
         {
             return bits->size;
         }

         /* Count leading one bits in this word. */
         for (offs = 0; bits->bits[i] & (1 << offs); offs++)
         {
             count++;
         }

         /*
          * Count is now the number of bits we skipped to get to
          * the bit we found. strt + count is index of that bit.
          */
         return strt + count;
     }
}

ncp_bitmap_t *
ncp_bitmap_create(
    uint32_t size)
{
    uint32_t i;
    uint32_t words_needed;
    BitField_t *bits = NULL;
    uint32_t fill;
    uint32_t last_word;

    /*
     * Verify that the size is valid. The only check we
     * make now is that the size is greater than zero.
     */
    if (size <= 0)
    {
        return NULL;
    }

    /*
     * Calculate how many 'NCP_U32' values it will take to store
     * the requested bit field.
     */
    words_needed = (size + BITS_PER_INT - 1) / BITS_PER_INT;

    /*
     * Allocate the bit field descriptor.
     *   - if the allocation fails, return -1.
     */
    bits = (BitField_t *)ncp_nvm_malloc(sizeof(BitField_t));
    if (bits == (BitField_t *)0)
    {
        return NULL;
    }

    /*
     * Initialize the fields of the descriptor.
     */
    bits->size = size;
    bits->words_needed  = words_needed;
    bits->firstFree = 0;

    /*
     * Allocate the NCP_U32 array that will hold the bit field.
     *  - if the allocation fails, free the descriptor and
     *    return -1.
     */
    bits->bits = (uint32_t *)ncp_nvm_malloc(words_needed * sizeof(uint32_t));
    if (bits->bits == (uint32_t *)0)
    {
        ncp_nvm_free(bits);
        return NULL;
    }

    /*
     * Initialize the bit field to 0's.
     */
    for (i = 0; i < words_needed; i++)
    {
        bits->bits[i] = 0;
    }

    /*
     * calculate the number of unused bits in the last word
     * and initialize them to be set. This will avoid any issues
     * with them being misinterpreted as available entries.
     */
    fill = size % BITS_PER_INT;
    if (fill > 0)
    {
        last_word = words_needed - 1;
        while (fill < BITS_PER_INT)
        {
            bits->bits[last_word] |= (1 << fill);
            fill++;
        }
    }

    /*
     * Return
     */
    return (ncp_bitmap_t *)bits;
}

int32_t
ncp_bitmap_delete(
    ncp_bitmap_t *map)
{
    BitField_t *bits = (BitField_t *)map;

    if (bits)
    {
        if (bits->bits)
        {
            ncp_nvm_free(bits->bits);
        }
        ncp_nvm_free(bits);
    }

    return 0;
}

int32_t
ncp_bitmap_reserve(
    ncp_bitmap_t *map,
    uint32_t  size,
    uint32_t *base)
{
    BitField_t *bits = (BitField_t *)map;
    uint32_t strt,ostart,wdetect = 0;
    uint32_t offs,i;
    bool  wrapped = FALSE;

    /* Handle the case where no bitmap was created */
    if (!bits)
    {
        return -1;
    }

    /* bitmap must contain at least 'size' bits */
    if (size > bits->size)
    {
        return -2;
    }

    /*
     * Start at the first free bit
     */
    strt = bits->firstFree;

    /*
     * Calculate the starting NCP_U32 index in the bit array (i)
     * and the bit offset in that NCP_U32 (offs).
     */
    i = strt / BITS_PER_INT;
    offs = strt % BITS_PER_INT;

    /*
     * is the start really free ?
     */
    if ((bits->bits[i] & (1 << offs)) == 1)
    {
        /* nope,scan for a new start */
        strt = bit_next_clr(bits, strt);
        if (strt == bits->size)
        {
            /* nothing ! */
            return -1;
        }
    }

    /*
     * If the current bit is not the start of
     * a range of 'size' cleared bits, go to
     * the start of the next clear range.
     */
    ostart = strt;
    while ((strt < bits->size) &&
           (bit_num_clr(bits, strt, size) < size))
    {
        strt = bit_next_clr(bits, strt);
        if (strt == bits->size)
        {
            /* could not find a single bit! */
            return -1;
        }

        /* once we have wrapped, we need to detect when we leave the remaining
           valid region between the original start and the wrap detection slot */
        if (wrapped)
        {
            if (strt <= wdetect)
            {
                return -1;
            }
            if (strt >= ostart)
            {
                return -1;
            }
        }
        /* we to detect when we wrap around */
        if ((strt <= ostart) && !wrapped)
        {
            wrapped = TRUE;
            wdetect = strt;
        }

    }
    /* If no range of 'size' clear bits was found, return -1. */
    if (strt >= bits->size)
    {
        return -2;
    }

    /* Set the bits we just found. */
    bit_set(bits, size, strt);

    /* If start was at firstFree, move firstFree. */
    if (strt == bits->firstFree)
    {
        bits->firstFree = bit_next_clr(bits, strt);
        if (bits->firstFree == bits->size)
        {
            /* nothing found,go back to zero */
            bits->firstFree = 0;
        }
    }

    *base = strt;
    return 0;
}

int32_t
ncp_bitmap_unreserve(
    ncp_bitmap_t *map,
    uint32_t  size,
    uint32_t  base)
{
    BitField_t *bits = (BitField_t *)map;

    /* Clear the bits. */
    bit_clr(bits, size, base);

    /*
     * If the start of this range was before
     * firstFree, move firstFree.
     */
    if (base < bits->firstFree)
    {
        bits->firstFree = base;
    }
    return 0;
}

bool
ncp_bitmap_isset(
    ncp_bitmap_t *map,
    uint32_t  id)
{
    uint32_t i;
    uint32_t offs;

    BitField_t *bits = (BitField_t *)map;

    if(bits == NULL)
    {
        return FALSE;
    }

    /*
     * Calculate the starting NCP_U32 index in the bit array (i)
     * and the bit offset in that NCP_U32 (offs).
     */
    i = id / BITS_PER_INT;
    offs = id % BITS_PER_INT;

    if( id > (bits->size - 1) )
    {
        return FALSE;
    }

    if (bits == (BitField_t *)0)
    {
        return FALSE;
    }

    return (bits->bits[i] & (1 << offs)) != 0;
}


