/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <string.h>

#include "ncp_internal.h"

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

extern unsigned long ncp_nvm_vaddr;
extern unsigned long ncp_nvm_size;

/* 
 * definitions of the ncp_malloc/ncp_nvm_malloc family of functions.
 */

void *
ncp_malloc(size_t size)
{
    void *p;

    if (size <= 0)
    {
        /* NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_malloc(%d) passed a zero or less size.\n", size); */
        return 0;
    }

    /*
     * force cache line independence
     * just force min allocation size equal to cache line size
     */
#ifdef PROC_ARMA15
    if (size < 64)
    {
        size = 64;
    }
#else
    if (size < 128)
    {
        size = 128;
    }
#endif

    if ((p = __ncp_malloc(size)) == 0)
    {
        NCP_MSG(NCP_MSG_ERROR, "ncp_malloc(%zu) failed.\n", size);
    }

#ifdef NCP_MEMORY_CORUPTION_DEBUG
    if(p != NULL)
    {
        mem_block_alloc(p, size);
    }
#endif
    return p;
}

void *
ncp_calloc(uint32_t no, size_t size)
{
    void *p;

    if (size <= 0 || no <= 0)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_calloc(no=%lu, size=%lu) passed a zero or less size.\n", 
               (unsigned long)no, (unsigned long)size);
#endif
        return 0;
    }

    /*
     * force cache line independence
     * just force min allocation size equal to cache line size
     */
    size = (no * size); 
#ifdef PROC_ARMA15
    if (size < 64)
    {
        size = 64;
    }
#else
    if (size < 128)
    {
        size = 128;
    }
#endif

    if (NULL != (p = __ncp_malloc(size)))
    {
        memset(p, 0,  size);
    }  
    else
    {
        NCP_MSG(NCP_MSG_ERROR, "ncp_calloc(no=%u, size=%zu) failed.\n", no, size);
    }

    return p;
}

void *
ncp_realloc(void *ptr, size_t size)
{
    if (size <= 0)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_realloc(%lu) passed a zero or less size.\n", (unsigned long)size);
#endif
        return 0;
    }
    
    /*
     * force cache line independence
     * just force min allocation size equal to cache line size
     */
#ifdef PROC_ARMA15
    if (size < 64)
    {
        size = 64;
    }
#else
    if (size < 128)
    {
        size = 128;
    }
#endif    

    if ((ptr = __ncp_realloc(ptr, size)) == 0)
    {
        NCP_MSG(NCP_MSG_ERROR, "ncp_realloc(ptr=%p, size=%zu) failed.\n", ptr, size);
    }
    return ptr;
}

void 
ncp_free(void *ptr)
{
    if (ptr == NULL)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_free(%p) passed a NULL pointer.\n", ptr);
#endif
        return;
    }
#ifdef NCP_USE_NVM
    if ((ptr > (void *) ncp_nvm_vaddr) &&
        (ptr < (void *) (ncp_nvm_vaddr + ncp_nvm_size)))
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING: ncp_free(%p) passed a pointer in NVM region!\n", ptr);
#endif
        return;
    }
#endif

#ifdef NCP_MEMORY_CORUPTION_DEBUG
    if(mem_block_free(ptr) < 0)
        return;
#endif
    __ncp_free(ptr);
}

void *
ncp_nvm_malloc(size_t size)
{
    void *p;

    if (size <= 0)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_nvm_malloc(%lu) passed a zero or less size.\n", (unsigned long)size);
#endif
        return 0;
    }

    /*
     * force cache line independence
     * just force min allocation size equal to cache line size
     */
#ifdef PROC_ARMA15
    if (size < 64)
    {
        size = 64;
    }
#else
    if (size < 128)
    {
        size = 128;
    }
#endif

    if ((p = __ncp_nvm_malloc(size)) == 0)
    {
        NCP_MSG(NCP_MSG_ERROR, "ncp_nvm_malloc(%zu) failed.\n", size);
    }
    return p;
}

void *
ncp_nvm_calloc(uint32_t no, size_t size)
{
    void *p;

    if (size <= 0 || no <= 0)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_nvm_calloc(no=%lu, size=%lu) passed a zero or less size.\n", 
                (unsigned long)no, (unsigned long)size);
#endif
        return 0;
    }

    /*
     * force cache line independence
     * just force min allocation size equal to cache line size
     */

    size = no * size;
#ifdef PROC_ARMA15
    if (size < 64)
    {
        size = 64;
    }
#else
    if (size < 128)
    {
        size = 128;
    }
#endif

    if (NULL != (p = __ncp_nvm_malloc(size)))
    {
        memset(p, 0, size);
    }  
    else        
    {
        NCP_MSG(NCP_MSG_ERROR, "ncp_nvm_calloc(no=%u, size=%zu) failed.\n", no, size);
    }
    return p;
}

void *
ncp_nvm_realloc(void *ptr, size_t size)
{
    if (size <= 0)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_nvm_realloc(ptr=%p, size=%lu) passed a zero or less size.\n", 
                ptr, (unsigned long)size);
#endif
        return 0;
    }

    /*
     * force cache line independence
     * just force min allocation size equal to cache line size
     */
#ifdef PROC_ARMA15
    if (size < 64)
    {
        size = 64;
    }
#else
    if (size < 128)
    {
        size = 128;
    }
#endif

    if ((ptr = __ncp_nvm_realloc(ptr, size)) == 0)
    {
        NCP_MSG(NCP_MSG_ERROR, "ncp_nvm_realloc(ptr=%p, size=%zu) failed.\n", ptr, size);
    }
    return ptr;
}
           
void 
ncp_nvm_free(void *ptr)
{
    if (ptr == NULL)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING:  ncp_nvm_free(%p) passed a NULL pointer.\n", ptr);
#endif
        return;
    }

#ifdef NCP_USE_NVM
    if (((unsigned long) ptr < ncp_nvm_vaddr) ||
        ((unsigned long) ptr > (ncp_nvm_vaddr + ncp_nvm_size)))
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "WARNING: ncp_nvm_free(%p) passed a pointer not in NVM region!\n", ptr);
#endif
        return;
    }
#endif

    __ncp_nvm_free(ptr);

    return;
}

