/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <stdio.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "ncp_internal.h"

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */



#ifdef NCP_DEBUG
int ncp_trace_level = (NCP_MSG_INFO | NCP_MSG_ERROR);
#else
int ncp_trace_level = 0;
#endif

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * NVM Access Interface
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * The NVM Access Interface provides a means by which API state 
 * information can be stored in a non-volatile memory for the 
 * purpose of allowing a warm-restart. The interface consits of
 * a routine to memory map the NVM into the application address
 * space, a library of memory allocation functions (e.g. malloc)
 * which use the user provided memory pool, and routines to get
 * and set a pointer to the root API state information in NVM.
 *
 * The NVM is currently simulated by using file-backed memory
 * mapped I/O. On linux systems this could be modified to map
 * onto /dev/mem to allow access to a physical non-volatile memory.
 */

#include <stdlib.h>

#ifdef NCP_USE_NVM

/*
 * in order to support optional runtime specification of the
 * NVM address, we now use a global variable for the ncp_nvm_vaddr
 * TODO: NCP_NVM_VADDR and NCP_NVM_MAX_SIZE ought to be set via config file, not hard-coded at build time.
 */
static bool ncp_nvm_initialized = FALSE;
unsigned long ncp_nvm_vaddr = 0;
unsigned long ncp_nvm_size = 0;
ncp_sem_t *pNvmLock = NULL;
bool *pNvmActive = NULL;

char *
ncp_nvm_mmap(void)
{
    char *pnvm_base         = NULL;
    int         fd          = -1;
    off_t   offset;
    
    char *tmpDir            = NULL;
    char nvmFileName[256]   = {0};
    char null               = 0;

    /* on the external host we use a memory mapped file */
    /* 
     * get the path for the nvm.dat file. 
     */
    tmpDir = getenv("NCP_TMP_DIR");
    if (tmpDir != NULL)
    {
        strcpy(nvmFileName, tmpDir);
        strcat(nvmFileName, "/ncpNvmState.dat");
    }
    else
    {
        strcpy(nvmFileName, "/tmp/ncpNvmState.dat");
    }
    offset = (off_t) 0;

    /*
     * open the file for the memory-mapped I/O, creating it if it
     * does not exist. The open() does not seem to set write permissions
     * for group and others, so explicitly set the file permissions.
     * Grow the file size to be the max NVM size.
     */

    fd = open(nvmFileName, O_RDWR | O_CREAT , 0666);
    if (fd == -1)
    {
        printf("Could not open NVM file %s\n", nvmFileName);
        perror("open");
        return NULL; 
    }

    if (chmod(nvmFileName, 0666) < 0)
    {
        /*
         * Generally this only fails if the file already exists
         * and was created by root, and this process is running
         * as non-root. Since in this case the file-permissions
         * have already been set, ignore the error.
         */
        if (errno != EPERM)
        {
            perror("chmod");
            return NULL;
        }
    }

    /* set the file size */
    if (lseek(fd, ncp_nvm_size, SEEK_SET) < 0)
    {
        perror("lseek");
        return NULL;
    }

    if (write(fd, (void *) &null, sizeof(char)) == -1)
    {
        perror("write");
        return NULL;
    }

    /*
     * map the nvm.dat file to virtual address space at ncp_nvm_vaddr
     */
    pnvm_base = (char *)mmap((void *)ncp_nvm_vaddr, ncp_nvm_size, 
                                     (PROT_READ | PROT_WRITE), 
                                     MAP_SHARED, fd, offset);

    if (pnvm_base == (char *) MAP_FAILED)
    {
        perror("mmap");
        return NULL;
    }

    if (pnvm_base != (char *) ncp_nvm_vaddr)
    {
        printf("mmap returned wrong addr %p (expected 0x%lx)\n", pnvm_base, ncp_nvm_vaddr);
        return NULL;
    }

    close(fd);

    return pnvm_base;
}

int
ncp_nvm_munmap(void)
{
    int rc;

    rc = munmap((void *) ncp_nvm_vaddr, ncp_nvm_size);
    ncp_nvm_initialized = FALSE;

    return rc;
}

int
ncp_nvm_init(bool coldStart)
{
    ncp_nvm_addr_t   *pNvmAddr = NULL;
    char             *pnvm_base = NULL;

    /* 
     * check if the user has specified an alternative value for
     * the NVM address. If so reset the global ncp_nvm_vaddr.
     */
    char *vaddrEnv   = NULL;
    uint32_t val = 0;

    vaddrEnv = getenv("NCP_NVM_VADDR");
    if (vaddrEnv != NULL) {
        if (sscanf(vaddrEnv, "0x%x", &val) == 1) {
            ncp_nvm_vaddr = val;
        } else {
            printf("NCP_NVM_VADDR defined with invalid address %s\n", vaddrEnv);
        }
    }

    if (coldStart == TRUE) 
    {
        if (ncp_nvm_initialized == TRUE) {
            pnvm_base = (char *) ncp_nvm_vaddr;
            ncp_nvm_munmap();
        }

        ncp_nvm_initialized = FALSE;
    }

    /* return immediately if already initialized */
    if (ncp_nvm_initialized == TRUE)
    {
        return 0;
    }

    pnvm_base = ncp_nvm_mmap();

    if ( pnvm_base == NULL )
    {
        return -1;
    }

    /*
     * allocate space at the beginning of the NVM area for local address 
     * pointers and the mp_info structure required by the mpmlib.
     */
    pNvmAddr = (ncp_nvm_addr_t *)pnvm_base;    


    /* 
     * Initialize the mpmlib; either with the mp_initialize
     * routine for cold starts, or with the mp_restart routine
     * for a warm restart.
     * For now, we determine whether or not it's a cold start by
     * checking the rootPtr. If it's zero, its cold start.
     *
     */
    if ((pNvmAddr->rootPtr == NULL) || (coldStart == TRUE))
    {
        memset(pNvmAddr, 0, sizeof(ncp_nvm_addr_t));
        pNvmAddr->base = pnvm_base;
#if 0
        pNvmAddr->mp_info_ptr = (mp_info *) pnvm_base + sizeof(ncp_nvm_addr_t);
        pNvmAddr->nxt_avail = (char *) pNvmAddr->mp_info_ptr + sizeof(mp_info);
        if (!mp_initialize(pNvmAddr->mp_info_ptr, ncp_nvm_get_more_core))
        {
            NCP_MSG(NCP_MSG_ERROR, "mp_init error\n");
            return -1;
        }
#endif
        
        /*
         * Initialize NVM locking semaphore to make the NVM alloc/free
         * functions multi-thread and multi-process safe
         */
        pNvmLock = &(pNvmAddr->nvmLock);
        pNvmActive = &(pNvmAddr->nvmActive);
        memset(pNvmLock, 0, sizeof(ncp_sem_t));
        if (NCP_ST_SUCCESS != ncp_sem_init(pNvmLock, TRUE, 0))
        {
            printf("NVM sem_init failed\n");
            return -1;
        }
        pNvmAddr->cookie = NCP_NVM_COOKIE;
    }
    else
    {
        /*
         * For a restart we make sure that the addresses we've mapped matches.
         */
        if ( pNvmAddr->base != pnvm_base) {
            printf(
                "NVM base address mismatch!  mapped %p, but NVM has %p\n",
                pnvm_base, pNvmAddr->base);
            return -1;
        }
#if 0
        if (!mp_restart(pNvmAddr->mp_info_ptr, ncp_nvm_get_more_core))
        {
            NCP_MSG(NCP_MSG_ERROR, "mp_restart error\n");
            return -1;
        }
#endif
        pNvmLock = &(pNvmAddr->nvmLock);
        pNvmActive = &(pNvmAddr->nvmActive);
    }

    ncp_nvm_initialized = TRUE;

    return 0;
}

void
ncp_nvm_ptr_set(void *rootPtr)
{
    ncp_nvm_addr_t   *pNvmAddr = (ncp_nvm_addr_t *) ncp_nvm_vaddr;

    if (pNvmAddr->cookie != NCP_NVM_COOKIE)
    {
        pNvmAddr->rootPtr = 0;
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "ncp_nvm_ptr_set: bad cookie!\n");
#endif
        return;
    }
    pNvmAddr->rootPtr = rootPtr;
}

void *
ncp_nvm_ptr_get(void)
{
    ncp_nvm_addr_t   *pNvmAddr = (ncp_nvm_addr_t *) ncp_nvm_vaddr;

    if (pNvmAddr->cookie != NCP_NVM_COOKIE)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "ncp_nvm_ptr_get: bad cookie!\n");
#endif
        pNvmAddr->rootPtr = 0;
    }
    return (pNvmAddr->rootPtr);
}

void
ncp_nvm_user_ptr_set(int idx, void *userPtr)
{
    ncp_nvm_addr_t   *pNvmAddr = (ncp_nvm_addr_t *) ncp_nvm_vaddr;

    if (idx >= NCP_NVM_MAX_USER_POINTERS)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "ncp_nvm_user_ptr_set: invalid index (%d)\n", idx);
#endif
        return;
    }

    if (pNvmAddr->cookie != NCP_NVM_COOKIE)
    {
        pNvmAddr->userPtr[idx] = 0;
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "ncp_nvm_user_ptr_set: bad cookie!\n");
#endif
        return;
    }
    pNvmAddr->userPtr[idx] = userPtr;
}

void *
ncp_nvm_user_ptr_get(int idx)
{
    ncp_nvm_addr_t   *pNvmAddr = (ncp_nvm_addr_t *) ncp_nvm_vaddr;

    if (idx >= NCP_NVM_MAX_USER_POINTERS)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "ncp_nvm_user_ptr_get: invalid index (%d)\n", idx);
#endif
        return NULL;
    }

    if (pNvmAddr->cookie != NCP_NVM_COOKIE)
    {
#ifdef NCP_DEBUG
        NCP_MSG(NCP_MSG_DEBUG, "ncp_nvm_user_ptr_get: bad cookie!\n");
#endif
        pNvmAddr->userPtr[idx] = 0;
    }
    return (pNvmAddr->userPtr[idx]);
}

void 
ncp_nvm_stats_get(ncp_nvm_stats_t *nvmStats)
{
    memset(nvmStats, 0, sizeof(ncp_nvm_stats_t));
}

bool ncp_nvm_is_active(void)
{
    ncp_nvm_addr_t     *pNvmAddr = (ncp_nvm_addr_t *)ncp_nvm_vaddr;
    bool          pActive = FALSE;
    ncp_sem_wait(pNvmLock);
    pActive =  (pNvmAddr->nvmActive);
    ncp_sem_post(pNvmLock);
    return (pActive);
}

#else /* !#ifdef NCP_USE_NVM */

/* empty functions when NCP_USE_NVM not defined */

int ncp_nvm_init(bool coldStart)
{
    (void)coldStart;
    return 0;
}

int   ncp_nvm_munmap(void)
{
    return 0;
}


void ncp_nvm_stats_get(ncp_nvm_stats_t *nvmStats) 
{
    memset(nvmStats, 0, sizeof(ncp_nvm_stats_t));
}


void
ncp_nvm_ptr_set(void *rootPtr)
{
    (void)rootPtr;
}

void *
ncp_nvm_ptr_get(void)
{
    return NULL;
}

void
ncp_nvm_user_ptr_set(int idx, void *userPtr)
{
    (void)idx;
    (void)userPtr;
}

void *
ncp_nvm_user_ptr_get(int idx)
{
    (void)idx;
    return NULL;
}

bool ncp_nvm_is_enabled(void)
{
    return FALSE;
}

bool ncp_nvm_is_active(void)
{
    return FALSE;
}

#endif /* #ifdef NCP_USE_NVM */


ncp_st_t
ncp_sem_init(ncp_sem_t *sem, int32_t shared, uint32_t value)
{
    ncp_st_t ncpStatus = NCP_ST_SUCCESS;

    (void)value;

    if(pthread_mutexattr_init(&sem->mutexAttr))
    { 
        return NCP_ST_MUTEX_ATTR_INIT_ERROR;
    }

    if (shared)
    {
        if(pthread_mutexattr_setpshared(&sem->mutexAttr,
                                          PTHREAD_PROCESS_SHARED))
        {
            return NCP_ST_MUTEX_SHARED_INIT_ERROR;
        }
    }

    if(pthread_mutex_init(&sem->mutex, &sem->mutexAttr))
    { 
        return NCP_ST_MUTEX_INIT_ERROR;
    }

NCP_RETURN_LABEL
    return ncpStatus;
}

ncp_st_t
ncp_sem_post(ncp_sem_t *sem)
{
    ncp_st_t ncpStatus = NCP_ST_SUCCESS;

    if(pthread_mutex_unlock(&sem->mutex))
    {
        return NCP_ST_MUTEX_POST_ERROR;
    }

NCP_RETURN_LABEL
    return ncpStatus;
}

ncp_st_t
ncp_sem_wait(ncp_sem_t *sem)
{
    ncp_st_t ncpStatus = NCP_ST_SUCCESS;

    if(pthread_mutex_lock(&sem->mutex))
        return NCP_ST_MUTEX_WAIT_ERROR;

NCP_RETURN_LABEL
    return ncpStatus;
}

ncp_st_t
ncp_sem_destroy(ncp_sem_t *sem)
{
    ncp_st_t ncpStatus = NCP_ST_SUCCESS;

    if (pthread_mutex_destroy(&sem->mutex))
    {
        return NCP_ST_MUTEX_DESTROY_ERROR;
    }

NCP_RETURN_LABEL
    return ncpStatus;
}

