/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include "ncp_sorted_set.h"
#include "ncp_sal.h"

typedef struct ncp_sorted_set_node_s {
    void *key; /* user intreprets the key as they see fit */
    void *value;      
} ncp_sorted_set_node_t; 

typedef struct ncp_sorted_set_pvt_s {
    uint32_t maxSortedSetSize;
    uint32_t curSortedSetSize;
    uint32_t mallocSizeBytes;
    ncp_sorted_set_node_t *sortedSet; 
} ncp_sorted_set_pvt_t;


static ncp_sorted_set_pvt_t * _ncp_sorted_set_create(
    uint32_t msize,
    uint32_t numInitialNodes)
{
    ncp_sorted_set_pvt_t *_sorted_set;
    
    /* ensure malloc size is a multiple of the node size */
    if (msize % sizeof(ncp_sorted_set_node_t)){
        return NULL;
    }
    _sorted_set = (ncp_sorted_set_pvt_t*)ncp_nvm_malloc(sizeof(ncp_sorted_set_pvt_t));
    _sorted_set->curSortedSetSize = 0;
    _sorted_set->maxSortedSetSize = numInitialNodes;
    _sorted_set->mallocSizeBytes  = msize;
    _sorted_set->sortedSet = (ncp_sorted_set_node_t*)ncp_nvm_malloc(sizeof(ncp_sorted_set_node_t)*numInitialNodes);
    if ( _sorted_set->sortedSet != NULL){
        memset(_sorted_set->sortedSet,0,sizeof(ncp_sorted_set_node_t)*numInitialNodes);
    }
    return _sorted_set;
}

static ncp_sorted_set_node_t *_ncp_sorted_set_get_node_n(
    ncp_sorted_set_pvt_t *sorted_set,
    uint32_t n)
{
    return &sorted_set->sortedSet[n];
}

static ncp_sorted_set_node_t * _ncp_sorted_set_set_node_n(
    ncp_sorted_set_pvt_t *sorted_set,
    ncp_sorted_set_node_t *node,
    uint32_t n)
{
    ncp_sorted_set_node_t *rnode;

    rnode = &sorted_set->sortedSet[n];
    memcpy(&sorted_set->sortedSet[n],node,sizeof(ncp_sorted_set_node_t));
    return rnode;
}

static int _ncp_sorted_set_grow(ncp_sorted_set_pvt_t *sorted_set){
    ncp_sorted_set_node_t *rset;

    rset = (ncp_sorted_set_node_t*)ncp_nvm_realloc(sorted_set->sortedSet,sorted_set->mallocSizeBytes + (sizeof(ncp_sorted_set_node_t)*sorted_set->maxSortedSetSize));
    if (rset){
        sorted_set->sortedSet = rset; 
        /* clear out the new nodes */
        memset(sorted_set->sortedSet + sorted_set->maxSortedSetSize,0,sorted_set->mallocSizeBytes);
        sorted_set->maxSortedSetSize += sorted_set->mallocSizeBytes/sizeof(ncp_sorted_set_node_t);
            
    } else {
        return -1;
    }
    return 0;
}

static int _ncp_sorted_set_insert_node_n(
    ncp_sorted_set_pvt_t *sorted_set,
    ncp_sorted_set_node_t *node,
    uint32_t n)
{
    uint32_t numMoves;
   
    /* push all down from n on */
    numMoves = sorted_set->curSortedSetSize - n;
    if (numMoves){
        memmove(&sorted_set->sortedSet[n+1],&sorted_set->sortedSet[n],sizeof(ncp_sorted_set_node_t)*numMoves);
    }
    /* set node n */
    memcpy(&sorted_set->sortedSet[n],node,sizeof(ncp_sorted_set_node_t));
    /* update current size */
    sorted_set->curSortedSetSize++;

    return 0;
}

static int  
_ncp_sorted_set_remove_node_n(
    ncp_sorted_set_pvt_t *sorted_set,uint32_t n,ncp_sorted_set_node_t *outNode)
{
    uint32_t numMoves;
 
    /* is it in range ? */
    if (n >= sorted_set->curSortedSetSize){
        return -1;
    }
    /* copy out the node value */
    memcpy(&sorted_set->sortedSet[n],outNode,sizeof(ncp_sorted_set_node_t));
    
    /* compress around n */
    numMoves = sorted_set->curSortedSetSize - n;
    if (numMoves){
        memmove(&sorted_set->sortedSet[n],&sorted_set->sortedSet[n+1],sizeof(ncp_sorted_set_node_t)*numMoves);
    }
    /* update current size */
    sorted_set->curSortedSetSize--;
   
   return 0;
}

static int _ncp_sorted_set_iterate_through_set(
    ncp_sorted_set_pvt_t *sorted_set,
    ncp_sorted_set_element_function_t fn,
    void *args)
{
    uint32_t i;
    int rval = 0;

    for (i = 0; i < sorted_set->curSortedSetSize; i++){
        rval = fn(i,sorted_set->sortedSet[i].key,sorted_set->sortedSet[i].value,args);
        if (rval != 0){
            goto cleanup; 
        }
    }
 cleanup:
    return rval;
}

static void _ncp_sorted_set_delete_nodes_and_set(ncp_sorted_set_pvt_t *sorted_set,ncp_sorted_set_nodeDeleteFn_t delFn)
{
    uint32_t i;
    int rval = 0;
    
    if (delFn){
        for (i = 0; i < sorted_set->curSortedSetSize; i++){
            rval = delFn(sorted_set->sortedSet[i].key,sorted_set->sortedSet[i].value);
            if (rval != 0){
                goto cleanup; 
            }

        }
    }
    /* free the array */
    ncp_nvm_free(sorted_set->sortedSet);
    /* free the struct */
    ncp_nvm_free(sorted_set);
    
 cleanup:
    return;
}


static uint32_t _ncp_sorted_set_size(ncp_sorted_set_pvt_t *sorted_set){
    return  sorted_set->curSortedSetSize;
}

static uint32_t 
_ncp_sorted_set_find_insertion_location(
    ncp_sorted_set_pvt_t *sorted_set,
    void *key,
    ncp_sorted_set_element_compare_function_t compareFn,
    bool *match)
{
    uint32_t location = 0,left,right,center,size = 0;
    ncp_sorted_set_node_t *node;

    if (compareFn == NULL) {
        *match = FALSE;
        goto cleanup;
    }

    left   = 0;
    right  =  _ncp_sorted_set_size(sorted_set);
    if (right == 0){
        location = 0;
        *match   = FALSE;
        goto cleanup;
    }
    right--;
    do {
        center = (left + right) / 2; 
        node = _ncp_sorted_set_get_node_n(sorted_set,center);
        switch(compareFn(key,node->key)){
        case NCP_SORTED_SET_CMP_EQUAL:
            *match = TRUE;
            location = center;
            goto cleanup;
            break;
        case NCP_SORTED_SET_CMP_LEFT:
            if (center){
                size     =  center - left;
                right    =  center - 1;
                location =  center;
            } else {
                size     = 0;
                location = 0;
            }
            break;
        case NCP_SORTED_SET_CMP_RIGHT:
            size = right - center;
            left = center + 1;
            location = left;
            break;
        default:
            break;
        }
    } while (size >= 1);
    
    *match = FALSE; 
    

 cleanup:
    return location;

} 

ncp_sorted_set_t* ncp_sorted_set_create(
    uint32_t mallocSizeBytes,
    uint32_t numInitialNodes)
{
    return (ncp_sorted_set_t*)(_ncp_sorted_set_create(mallocSizeBytes,numInitialNodes));
}

void ncp_sorted_set_delete_nodes_and_set(ncp_sorted_set_t *sorted_set,ncp_sorted_set_nodeDeleteFn_t delFn)
{
    _ncp_sorted_set_delete_nodes_and_set((ncp_sorted_set_pvt_t*)sorted_set,delFn);
    return;
}

int ncp_sorted_set_iterate_through_set(
    ncp_sorted_set_t *sorted_set,
    ncp_sorted_set_element_function_t fn,
    void *args)
{
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set;
    return _ncp_sorted_set_iterate_through_set(_sorted_set,fn,args);
}

void* 
ncp_sorted_set_get(ncp_sorted_set_t *sorted_set,
                   void *key,
                   ncp_sorted_set_element_compare_function_t compareFn)
{
    void *returnvalue = NULL;
    ncp_sorted_set_node_t *node = NULL;
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set;
    uint32_t location,size;
    bool match;
    
    size  =  _ncp_sorted_set_size(_sorted_set);
    if (size == 0){
        goto cleanup;
    }
    
    location = _ncp_sorted_set_find_insertion_location(
                        _sorted_set,
                        key,
                        compareFn,
                        &match);
    if (match){
        node = _ncp_sorted_set_get_node_n(_sorted_set,location);
        returnvalue = node->value;
    } 
    
 cleanup:
    return returnvalue;
}

void* ncp_sorted_set_get_n(ncp_sorted_set_t *sorted_set, uint32_t n){
    uint32_t size;
    void *returnvalue = NULL;
    ncp_sorted_set_node_t *node = NULL;
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set;

    size  =  _ncp_sorted_set_size(_sorted_set);
    if (n >= size){
        goto cleanup;
    }
    node = _ncp_sorted_set_get_node_n(_sorted_set,n);
    returnvalue = node->value;

 cleanup:
    return returnvalue;
}

void* ncp_sorted_set_put_n(ncp_sorted_set_t *sorted_set,uint32_t n,void *value){
    uint32_t size;
    void *returnvalue = NULL;
    ncp_sorted_set_node_t *node = NULL;
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set; 
    
    size  =  _ncp_sorted_set_size(_sorted_set);
    if (n >= size){
        goto cleanup;
    }
    node = _ncp_sorted_set_get_node_n(_sorted_set,n);
    returnvalue = node->value;
    node->value = value;
    _ncp_sorted_set_set_node_n(_sorted_set,node,n);
    
 cleanup:
    return returnvalue;
}

uint32_t ncp_sorted_set_size(ncp_sorted_set_t *sorted_set){
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set;
    
    return _ncp_sorted_set_size(_sorted_set);
}

void* 
ncp_sorted_set_put(ncp_sorted_set_t *sorted_set,
                   void *key,
                   ncp_sorted_set_element_compare_function_t compareFn,
                   void* value)
{
    ncp_sorted_set_node_t *rnode;
    ncp_sorted_set_node_t snode;
    void *returnvalue = NULL;
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set;
    uint32_t location;
    bool match;
    
    snode.value = value;
    snode.key   = key;
    location = _ncp_sorted_set_find_insertion_location(
                        _sorted_set,
                        key,
                        compareFn,
                        &match);
    if (match){
        rnode = _ncp_sorted_set_get_node_n(_sorted_set,location);
        returnvalue = rnode->value;
        /* replace the value */
        _ncp_sorted_set_set_node_n(_sorted_set,&snode,location);
    } else {
        /* does it need to grow ? */
        if (_sorted_set->curSortedSetSize == _sorted_set->maxSortedSetSize){
            if (_ncp_sorted_set_grow(_sorted_set) != 0){
                goto cleanup;
            }
        }
        /* do the insert */
        if (_ncp_sorted_set_insert_node_n(_sorted_set,&snode,location) != 0){
            goto cleanup;
        }
    } 
    
 cleanup:
    return returnvalue;
}

void 
*ncp_sorted_set_remove(ncp_sorted_set_t *sorted_set,
                       void *key,
                       ncp_sorted_set_element_compare_function_t compareFn)
{
    ncp_sorted_set_node_t rnode;
    void *returnvalue = NULL;
    ncp_sorted_set_pvt_t *_sorted_set = (ncp_sorted_set_pvt_t*)sorted_set;
    uint32_t location,size;
    bool match;
    
    size  =  _ncp_sorted_set_size(_sorted_set);
    if (size == 0){
        goto cleanup;
    }
    location = _ncp_sorted_set_find_insertion_location(
                        _sorted_set,
                        key,
                        compareFn,
                        &match);
    if (match){
        if( _ncp_sorted_set_remove_node_n(_sorted_set,location,&rnode) == 0){
            returnvalue = rnode.value;
        }
    }
 cleanup:
    return returnvalue;
}

