/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_pal.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _pal  PAL Layer
 * @{
 *      @addtogroup _palFunctions  PAL Functions
 *      @{
 *
 * @brief
 *
 * This file contains all the APIs provided by the QM  Abstraction Layer (PAL)
 *
 ***********************************************************************/

/* PAL external header files */
#include <adk_include.h>
#include "adk_pal.h"
#include "adk_pal_internal.h"
#include "adk_pal_dp.h"
#include "adk_pal_dp_export.h"
#include "adk_pal_cp_pvt.h"
#include "adk_sys_services.h"

#include "adk_profile.h"


#include "adk_shm.h"

adk_pal_cp_info_t *adk_pal_cp_info_p = NULL;
adk_pal_dp_info_t *adk_pal_cp_dp_info_p = NULL;
uint64_t          *adk_pal_cp_min_cycles = NULL;

#ifdef ADK_PAL_CYCLE_MEASUREMENT
static adk_pal_proc_fn_stats_t proc_fn_stats[ADK_MAX_PROC_FUNC_ID];
#endif


static adk_status_t
adk_pal_update_ppl( adk_sys_rsrc_evdev_info_t *evdev_info_p);

static adk_status_t
adk_pal_update_cbd(adk_sys_rsrc_crypto_dev_info_t *cdev_info_p);

adk_status_t
adk_pal_update_eth_dev(adk_sys_rsrc_eth_dev_info_t *eth_dev_info_p);

/*!
 * =============================================================
 * adk_pal_init
 * @param[in] sysHandle The ADK system handle (adk_sys_handle_t)
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  Initialises the PAL.
 *
 * This function initializes the QM Abstraction Layer and
 * is internally called by adk_sys_init(). It performs
 * the following actions:
 *  > Creates PAL CP and DP shared memory
 *  > Query the event-dev, crypto-dev and eth-dev resources from
 *  adk_sys and does the PAL initialization.
 *
 * =============================================================
 */
adk_status_t
adk_pal_init (
    adk_sys_handle_t  sysHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_pal_shm_config_t        *shm_config_p = NULL;
    adk_sys_resource_info_t     *evdev_rsrc_info_list_p;
    adk_sys_rsrc_evdev_info_t   *evdev_info_p;
    uint8_t                     num_evdevs;
    adk_sys_resource_info_t     *cdev_rsrc_info_list_p;
    adk_sys_rsrc_crypto_dev_info_t  *crypto_dev_info_p;
    uint8_t                     num_crypto_devs;

    adk_sys_resource_info_t     *ethdev_rsrc_info_list_p;
    adk_sys_rsrc_eth_dev_info_t  *eth_dev_info_p;
    uint8_t                     num_eth_devs;

    uint32_t    index;

    if (ADK_INVALID_SYS_HANDLE (sysHandle))
    {
        /*! Invalid system handle */
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR ("Invalid system handle\n");
        return (retValue);
    }

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));

    retValue = adk_sys_config_evdev_info_get (&evdev_rsrc_info_list_p,
                                              &num_evdevs);
    if (retValue != ADK_SUCCESS)
    {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "[%s()] event dev resource get function failed.\n",
                     __func__);
        return (retValue);
    }

    retValue = adk_sys_config_crypto_dev_info_get(&cdev_rsrc_info_list_p,
                                              &num_crypto_devs);
    if (retValue != ADK_SUCCESS)
    {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "[%s()] crypto-dev resource get function failed.\n",
                     __func__);
        return (retValue);
    }

    retValue = adk_sys_config_eth_dev_info_get(&ethdev_rsrc_info_list_p,
                                              &num_eth_devs);
    if (retValue != ADK_SUCCESS)
    {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                      "[%s()] eth-dev resource get function failed.\n",
                      __func__);
        return (retValue);
    }

    /*
    * Allocated shared memory for IPC between primary(Control) and
    * secondary process(dataplne)*/
    retValue = adk_shm_create(ADK_PAL_SHM_MEMZONE_NAME,
                              sizeof(adk_pal_shm_config_t),
                              ADK_SHM_LOOKUP_LCORE,
                              (adk_shm_handle_t **)&shm_config_p);
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "[%s()] IPC shared memory allocation failed.\n",
                     __func__);
        return (retValue);
    }

    /*
    * Allocate Memory PAL CP info structure
    */
    adk_pal_cp_info_p = (adk_pal_cp_info_t *)
        ADK_NVM_CALLOC(ADK_PAL_MODULE_ID, 0, 1, sizeof(adk_pal_cp_info_t));
    if(adk_pal_cp_info_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "[%s()] memory allocation failed for pal-cp.\n",
                     __func__);
        return (retValue);
    }

    /*
     * Create a packet type table resource handle
     */
    adk_pal_cp_info_p->pkt_type_tbl_rm_handle = 
        adk_rm_resource_init(ADK_PAL_MODULE_ID, 0, ADK_PAL_PKT_TYPE_TBL_IDS);

    if(adk_pal_cp_info_p->pkt_type_tbl_rm_handle == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "[%s()] memory allocation failed for pal-pkt-type-tbl.\n",
                     __func__);
        return (retValue);
    }

    /*
     * reserve the ADK_PAL_PKT_TYPE_TBL_ID_RSVD id
     * */
    if(!adk_rm_resource_reserve_fixed(
                    adk_pal_cp_info_p->pkt_type_tbl_rm_handle,
                    ADK_PAL_PKT_TYPE_TBL_ID_RSVD))
    {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "Failed to reserve ADK_PAL_PKT_TYYPE_TBL_ID_RSVD ID.\n");
        return (retValue);
    }

    /* Allocate Shared memory to store pal dataplane information */
    adk_pal_cp_dp_info_p = (adk_pal_dp_info_t *) 
                    ADK_DP_SHM_CALLOC(ADK_PAL_MODULE_ID, 0, 1, sizeof(adk_pal_dp_info_t), 0);
    if(adk_pal_cp_dp_info_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                     "[%s()] memory allocation failed for pal-dataplane-info.\n",
                     __func__);
        return (retValue);
    }
    adk_pal_cp_dp_info_p->sys_handle   = NULL;

    /*Allocte dataplane objects in shared memory for all processing functions*/
    for (index = 0; index < ADK_MAX_PROC_FUNC_ID; index ++)
    {
        adk_pal_cp_dp_info_p->stage_info_p[index] = (adk_pal_stage_info_t *) 
                    ADK_DP_SHM_CALLOC(ADK_PAL_MODULE_ID, 0, 1, sizeof(adk_pal_stage_info_t), 0);

        /*Check for memory allocaiton failures */
        if(adk_pal_cp_dp_info_p->stage_info_p[index]) {

            adk_pal_cp_dp_info_p->stage_info_p[index]->fn_id =  index;
            adk_pal_cp_dp_info_p->stage_info_p[index]->entry_info.entry_type =
                ADK_PAL_STAGE_ENTRY_TYPE_FN_CALL;

            adk_pal_cp_dp_info_p->dp_stage_info[index].fn_id      =   index;
            adk_pal_cp_dp_info_p->dp_stage_info[index].event_q_id = -1;
        } else {
            /*TODO: free up memory allocated so far */
            retValue = ADK_E_MEM_UNAVAILABLE;
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                         "[%s()] memory allocation failed for pal-stage-info.\n",
                         __func__);
            return (retValue);
        }
    }

#ifdef ADK_PAL_CYCLE_MEASUREMENT
    /* Allocate Shared memory to store dataplne cycle measurement data*/
    adk_pal_cp_min_cycles = (uint64_t *)ADK_DP_SHM_CALLOC(ADK_PAL_MODULE_ID, 0, 1, sizeof(uint64_t) * \
                                    ADK_PAL_MAX_FN_IDS * ADK_MAX_NUM_CORES * \
                                    ADK_PAL_MAX_SAMPLES, sizeof(uint64_t));
    /* By default enable the stats collection */
    adk_pal_cp_dp_info_p->xstats_state = ADK_PAL_DP_XSTATS_DISABLE;
#endif

    ADK_MEMSET ((void *)adk_pal_cp_dp_info_p->eth_dev_info,
        0xFFFFFFFF,
        (sizeof (adk_pal_dp_eth_dev_t) * ADK_PAL_MAX_NUM_NIC_PORTS));

    ADK_MEMSET ((void *)adk_pal_cp_dp_info_p->eth_port_to_event_queue_id,
        0xFF,
        (sizeof (uint8_t) * ADK_PAL_MAX_NUM_NIC_PORTS));

    ADK_MEMSET ((void *)adk_pal_cp_dp_info_p->evdev_port_ids,
        0xFF,
        (sizeof (uint8_t) * ADK_MAX_NUM_CORES + ADK_PAL_MAX_FLOATING_THREADS));

#ifdef ADK_HOST_PATH_USE_NETD
    ADK_MEMSET ((void *)adk_pal_cp_dp_info_p->linux_intf_id_to_evnt_qid,
        -1,
        (sizeof (int16_t) * ADK_PAL_N_LINUX_INTF_IDS));
#endif

    /* Set the  PAL handle to shared memory segment. This will be read
     * by the secondary process.
     */
    shm_config_p->adk_pal_dp_handle = (adk_handle_t) adk_pal_cp_dp_info_p;
    shm_config_p->min_cycles     = adk_pal_cp_min_cycles;

    /* Clear state of  dp_proc_fn_ptr_table_state */
    adk_pal_cp_dp_info_p->dp_proc_fn_ptr_table_state =
        ADK_PAL_DP_PROC_FN_PTR_TBL_INIT_NOT_DONE;


    if((num_evdevs != 0) &&(evdev_rsrc_info_list_p != NULL)) {
        evdev_info_p    = (adk_sys_rsrc_evdev_info_t *)
                            evdev_rsrc_info_list_p[0].evdev_info_p;

        retValue = adk_pal_update_ppl(evdev_info_p);
        if (retValue != ADK_SUCCESS)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                          "[%s()] default pipeline init failed.\n",
                          __func__);
        }
    }

    if((num_crypto_devs!= 0) &&(cdev_rsrc_info_list_p != NULL)) {
        for (index=0; index< num_crypto_devs; index++) {
            crypto_dev_info_p    = (adk_sys_rsrc_crypto_dev_info_t *)
                                cdev_rsrc_info_list_p[index].crypto_dev_info_p;
            retValue = adk_pal_update_cbd(crypto_dev_info_p);
            if (retValue != ADK_SUCCESS) {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                            "[%s()] crypto-bd update failed.\n",
                            __func__);
            }
        }
    }

    if((num_eth_devs!= 0) &&(ethdev_rsrc_info_list_p != NULL)) {
        for (index=0; index< num_eth_devs; index++) {
            eth_dev_info_p    = (adk_sys_rsrc_eth_dev_info_t *)
                                ethdev_rsrc_info_list_p[index].eth_dev_info_p;
            retValue = adk_pal_update_eth_dev(eth_dev_info_p);
            if (retValue != ADK_SUCCESS) {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                            "[%s()] eth-dev update failed.\n",
                            __func__);
            }
        }
    }
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));
    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_init  ----- */


/*!
 * =============================================================
 * adk_pal_register
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  registres event-dev and worker-event-port information with 
 * ADK PAL.
 *
 * registers event-dev and worker-event-port information with ADK PAL.
 *
 * =============================================================
 */
adk_status_t
adk_pal_register (
    adk_sys_handle_t __attribute__((unused)) sysHandle,
    uint8_t             evdev_id,
    adk_pal_worker_info_t *w_qm_info,
    uint32_t            num_workers)
{
    int lcore_index = 0;
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t worker_index;
    adk_pal_shm_config_t *shm_config_p = NULL;

    if((w_qm_info == NULL) || (num_workers == 0))
        return (ADK_E_INVALID_PARAMETER);

    /* Allocate Memory PAL info structure
     * Allignment to cache-line
     * */

    /* Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_PAL_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS)
    {
        ADK_FORCE_LOG_ERROR ("adk_pal_register : IPC shared memory lookup failed.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    /* Read the  PAL handle from shared memory segment.
     * This has been set by the primary process.
     */
    adk_pal_cp_dp_info_p = (adk_pal_dp_info_t *)shm_config_p->adk_pal_dp_handle;

    adk_pal_cp_dp_info_p->event_dev_id = evdev_id;

    for(worker_index=0; worker_index < num_workers; worker_index++)
    {
        lcore_index = ADK_LCORE_INDEX_LOOKUP(w_qm_info[worker_index].lcore_id);
        if(lcore_index < 0)
        {
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_RETURN( retValue );
        }
        adk_pal_cp_dp_info_p->evdev_port_ids[lcore_index] = \
            w_qm_info[worker_index].evdev_port_id;
    }
    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_register ----- */

/*!
 * =============================================================
 * adk_pal_handle_get
 * @param[in] sysHandle The ADK system handle (adk_sys_handle_t)
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  returns previously created PAL handle.
 *
 * This function restores the previously created PAL shared memory handles
 * (CP and DP).
 *
 * =============================================================
 */
adk_status_t
adk_pal_handle_get(
    adk_sys_handle_t sysHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_pal_shm_config_t *shm_config_p = NULL;
    /*! Validate the input parameters */
    if (ADK_INVALID_SYS_HANDLE (sysHandle))
    {
        /*! Invalid system handle */
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR ("Invalid system handle\n");
        return (retValue);
    }

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));

    /* Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_PAL_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                      "adk_pal_handle_get : IPC shared memory lookup failed.\n");
        return (ADK_E_INVALID_PARAMETER);
    }

    /* Read the  PAL handle from shared memory segment.
     * This has been set by the primary process.
     */

    adk_pal_cp_dp_info_p = (adk_pal_dp_info_t *)shm_config_p->adk_pal_dp_handle;
    adk_pal_cp_info_p = (adk_pal_cp_info_t *)shm_config_p->adk_pal_cp_handle;
    adk_pal_cp_min_cycles = (uint64_t *)shm_config_p->min_cycles;

    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));
    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_handle_get----- */



adk_status_t adk_pal_rsrc_notify_fn_cbfn
            (adk_module_handle_t mod_handle, 
             adk_sys_rsrc_oper_t oper,
             adk_sys_resource_type_t type,
             adk_sys_resource_info_t rsrc)
{
    adk_status_t                retValue = ADK_SUCCESS;

    RTE_SET_USED(mod_handle);
    RTE_SET_USED(type);
    RTE_SET_USED(rsrc);

    ADK_LOG_INFO(adkSysShmConf_p->sysHandle,
                 ADK_SYS_MODULE_HANDLE_GET(adkSysShmConf_p->sysHandle),
                 "resource type %d\n", type);
    if(oper == ADK_SYS_RSRC_ADD) 
    {
        switch(type){
        case ADK_SYS_RSRC_TYPE_EVNT_DEV:
            {
                adk_sys_rsrc_evdev_info_t   *evdev_info_p;
                evdev_info_p    = (adk_sys_rsrc_evdev_info_t *)
                                    rsrc.evdev_info_p;

                retValue = adk_pal_update_ppl(evdev_info_p);
                if (retValue != ADK_SUCCESS)
                {
                    retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                    ADK_FORCE_LOG_ERROR ("[%s()] default pipeline init failed.\n",
                                        __func__);
                }
            }
            break;
        case ADK_SYS_RSRC_TYPE_CRYPTO_DEV:
            {
                adk_sys_rsrc_crypto_dev_info_t   *cdev_info_p;
                cdev_info_p    = (adk_sys_rsrc_crypto_dev_info_t *)
                                    rsrc.crypto_dev_info_p;

                retValue = adk_pal_update_cbd(cdev_info_p);
                if (retValue != ADK_SUCCESS)
                {
                    retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                    ADK_FORCE_LOG_ERROR ("[%s()] Crypto-dev update failed.\n",
                                        __func__);
                }
            }
            break;
        case ADK_SYS_RSRC_TYPE_ETH_DEV:
            {
                adk_sys_rsrc_eth_dev_info_t *eth_dev_info_p;
                eth_dev_info_p  = (adk_sys_rsrc_eth_dev_info_t *) rsrc.eth_dev_info_p;

                retValue = adk_pal_update_eth_dev(eth_dev_info_p);
                if (retValue != ADK_SUCCESS)
                {
                    retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                    ADK_FORCE_LOG_ERROR ("[%s()] eth-dev update failed.\n",
                                        __func__);
                }
            }
            break;
        default:
            break;
        }
    } else {
        ADK_FORCE_LOG_ERROR("Only resource ADD is supported, \
                            other operations are not supported yet\n"); 
    }

    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * adk_pal_veth_port_to_nic_port_mapping
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  mapping veth port to nic port.
 *
 * This function maps veth port ot nic port.
 *
 * Note: veth support will be deprecated.
 * This API will be removed once netdriver becomes default host path.
 *
 * =============================================================
 */
adk_status_t
adk_pal_veth_port_to_nic_port_mapping (
    adk_sys_handle_t __attribute__((unused)) sysHandle,
    uint32_t veth_port,
    uint32_t nic_port)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_pal_dp_eth_dev_t nic, veth;

    /* Check for array outof bound */
    if((veth_port >= ADK_PAL_MAX_NUM_NIC_PORTS)
      || (nic_port >= ADK_PAL_MAX_NUM_NIC_PORTS)) {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_FORCE_LOG_ERROR ("[%s()]invalid veth_port or nic_port \n", __func__);
        ADK_RETURN( retValue );
    }

    nic.aggr_info = 0x00;
    veth.aggr_info = 0x00;

    nic.entry_fn_id  = ADK_ETH_LLI_FN_ID;
    nic.dev_type     = ADK_PAL_ETH_PORT_TYPE_NIC;
    nic.port_id      = veth_port;
    nic.type_tbl_idx = ADK_PAL_PKT_TYPE_TBL_ID_RSVD;

    veth.entry_fn_id  = ADK_ETH_ULI_NETD_FN_ID;
    veth.port_id      = nic_port;
    veth.type_tbl_idx = ADK_PAL_PKT_TYPE_TBL_ID_RSVD;


#ifndef ADK_HOST_PATH_USE_NETD
    veth.dev_type     = ADK_PAL_ETH_PORT_TYPE_VETH;
#endif

    adk_pal_cp_dp_info_p->eth_dev_info[nic_port].aggr_info = nic.aggr_info;
    adk_pal_cp_dp_info_p->eth_dev_info[veth_port].aggr_info = veth.aggr_info;


    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_veth_port_to_nic_port_mapping ----- */


/*!
 * =============================================================
 * adk_pal_tun_dev_register
 * @param[in] sysHandle The ADK system handle (adk_sys_handle_t)
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  to register tunnel-dev/pmd and associated adk-tunnel-if-id
 *
 * =============================================================
 */
adk_status_t
adk_pal_tun_dev_register(
    adk_sys_handle_t __attribute__((unused)) sysHandle,
    uint32_t tun_dev_id,
    uint32_t adk_tunnel_if_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_pal_dp_eth_dev_t tun_dev;

    /* Validate the input params */
    if((tun_dev_id >= ADK_PAL_MAX_NUM_NIC_PORTS) ||
       ((adk_tunnel_if_id + ADK_PAL_TUNNEL_IF_BASE_ID) >= ADK_PAL_MAX_TUNNEL_IF_ID))
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_FORCE_LOG_ERROR ("[%s()]invalid tun-dev-id %d or tunnel-if-id %d \n", __func__, tun_dev_id, adk_tunnel_if_id);
        return (retValue);
    }


    tun_dev.aggr_info = 0x00;

    tun_dev.entry_fn_id  = ADK_SIF_TUND_IP_DEMUX_FN_ID;
    tun_dev.port_id      = adk_tunnel_if_id + ADK_PAL_TUNNEL_IF_BASE_ID;
    tun_dev.dev_type     = ADK_PAL_ETH_PORT_TYPE_TUND;
    tun_dev.type_tbl_idx = ADK_PAL_PKT_TYPE_TBL_ID_RSVD;

    adk_pal_cp_dp_info_p->adk_tun_if_id_to_tundev[adk_tunnel_if_id + ADK_PAL_TUNNEL_IF_BASE_ID]  = tun_dev_id;
    adk_pal_cp_dp_info_p->eth_dev_info[tun_dev_id].aggr_info = tun_dev.aggr_info;

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_tun_dev_register----- */


/*!
 * =============================================================
 * adk_pal_tun_dev_unregister
 * @param[in] sysHandle The ADK system handle (adk_sys_handle_t)
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  to unregister tunnel-dev/pmd and associated adk-tunnel-if-id
 *
 * =============================================================
 */
adk_status_t
adk_pal_tun_dev_unregister(
    adk_sys_handle_t __attribute__((unused)) sysHandle,
    uint32_t tun_dev_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t adk_tunnel_if_id = 0;

    adk_pal_dp_eth_dev_t tun_dev;

    /* Validate the input params */
    if(tun_dev_id >= ADK_PAL_MAX_NUM_NIC_PORTS)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_FORCE_LOG_ERROR ("[%s()]invalid tun-dev-id %d \n", __func__, tun_dev_id);
        return (retValue);
    }
    adk_tunnel_if_id = adk_pal_cp_dp_info_p->eth_dev_info[tun_dev_id].port_id;

    if(adk_tunnel_if_id >= ADK_PAL_MAX_TUNNEL_IF_ID) {
        retValue = ADK_E_INTERNAL_ERROR;
        ADK_FORCE_LOG_ERROR ("[%s()]invalid tunnel-if-id %d \n", __func__, adk_tunnel_if_id);
        return (retValue);
    }

    tun_dev.aggr_info = 0x00;

    tun_dev.entry_fn_id  = ADK_PAL_ERR_HANDLER_FUNC_ID;
    tun_dev.port_id      =  0xFFFF;
    tun_dev.dev_type     = ADK_PAL_ETH_PORT_TYPE_INVALID;
    tun_dev.type_tbl_idx = ADK_PAL_PKT_TYPE_TBL_ID_RSVD;

    adk_pal_cp_dp_info_p->adk_tun_if_id_to_tundev[adk_tunnel_if_id ]  = 0xFF;
    adk_pal_cp_dp_info_p->eth_dev_info[tun_dev_id].aggr_info = tun_dev.aggr_info;

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_tun_dev_unregister----- */



/*!
 * =============================================================
 * adk_pal_eth_bridge_add
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  adds eth-bridge to PAL.
 *
 * This function adds ethernet bridge driver info to PAL.
 *
 * =============================================================
 */

adk_status_t
adk_pal_bridge_eth_add (
    adk_sys_handle_t __attribute__((unused))  sysHandle,
    adk_pal_bridge_eth_info_t   *bd_eth_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t   eth_port_index;
    uint8_t   eth_port_id;
    uint8_t   *eth_port_list_p;

    eth_port_list_p = bd_eth_info_p->eth_ports_list;
    
    for(eth_port_index = 0; eth_port_index < bd_eth_info_p->num_ports; eth_port_index++)
    {
        eth_port_id  = eth_port_list_p[eth_port_index];
        adk_pal_cp_dp_info_p->eth_port_to_event_queue_id[eth_port_id] = bd_eth_info_p->event_queue_id;
    }

    adk_pal_cp_dp_info_p->nic_rx_md_offset = bd_eth_info_p->eth_dev_rx_metadata_offset;
    adk_pal_cp_dp_info_p->nic_tx_md_offset = bd_eth_info_p->eth_dev_tx_metadata_offset;


    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_bridge_eth_add ----- */

/*
 * API to register ethernet-port type
 */
adk_status_t
adk_pal_eth_port_type_register(
    uint8_t                     port_num, 
    adk_pal_eth_port_type_e     port_type)
{
    adk_status_t retValue = ADK_SUCCESS;
    retValue = ADK_E_UNSUPPORTED;
    RTE_SET_USED(port_num);
    RTE_SET_USED(port_type);
    ADK_RETURN( retValue );
}

#ifdef ADK_HOST_PATH_USE_NETD
/*!
 * =============================================================
 * adk_pal_bridge_netd_add
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  adds netdriver interface configs to PAL.
 *
 * This function adds netdriver configurations info to PAL.
 *
 * =============================================================
 */

adk_status_t
adk_pal_bridge_netd_add (
    adk_sys_handle_t __attribute__((unused))  sysHandle,
    adk_pal_bridge_netd_info_t   *bd_netd_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t   netd_port_index;
    uint8_t   netd_port_id;
    uint8_t   *netd_port_list_p;
    if (!bd_netd_info_p)
        goto __adk_pal_bridge_netd_add_invalid_param;

    netd_port_list_p = bd_netd_info_p->netd_ports_list;
    if (!netd_port_list_p)
        goto __adk_pal_bridge_netd_add_invalid_param;

    for(netd_port_index = 0; netd_port_index < bd_netd_info_p->num_ports;
            netd_port_index++)
    {
        netd_port_id  = netd_port_list_p[netd_port_index];
        adk_pal_cp_dp_info_p->linux_intf_id_to_evnt_qid[netd_port_id] =
            bd_netd_info_p->event_queue_id;
    }

    adk_pal_cp_dp_info_p->netd_md_offset =
        bd_netd_info_p->netd_metadata_offset;
    ADK_RETURN( retValue );

__adk_pal_bridge_netd_add_invalid_param:
    retValue = ADK_E_INVALID_PARAMETER;
    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_bridge_netd_add ----- */

/*!
 * =============================================================
 * adk_pal_dev_net_add
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  adds netdriver interface configs to PAL.
 *
 * This function adds netdriver configurations info to PAL.
 *
 * =============================================================
 */

adk_status_t
adk_pal_dev_net_add(
    adk_sys_handle_t __attribute__((unused))  sysHandle,
    uint8_t     event_queue_id,
    uint8_t     netd_metadata_offset)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint16_t    linux_intf_id;
    for(linux_intf_id = 0; linux_intf_id< ADK_PAL_N_LINUX_INTF_IDS;
            linux_intf_id++)
    {
        adk_pal_cp_dp_info_p->linux_intf_id_to_evnt_qid[linux_intf_id] =
            event_queue_id;
    }

    adk_pal_cp_dp_info_p->netd_md_offset = netd_metadata_offset;
    ADK_RETURN( retValue );

}    /* -----  end of function adk_pal_dev_net_add----- */

#endif /* ADK_HOST_PATH_USE_NETD */

/*!
 * =============================================================
 * adk_pal_worker_to_event_port_map_add
 *
 * @retval ADK_SUCCESS Successfully added woker to PAL
 * @retval ADK_E_INVALID_PARAMETER Invalid paramater
 *
 * @brief  Adds a worker to PAL.
 *
 * This function adds a worker to PAL.
 *
 * =============================================================
 */
adk_status_t
adk_pal_worker_to_event_port_map_add(
    adk_sys_handle_t sysHandle,
    adk_pal_worker_info_t *winfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    if (ADK_INVALID_SYS_HANDLE (sysHandle))
    {
        /*! Invalid system handle */
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR ("Invalid system handle\n");
        return (retValue);
    }

    ADK_LOG_FUNC_ENTRY (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));

    /* check whether lcore id is with-in the range or not */
    if( winfo_p->lcore_id >= (ADK_MAX_NUM_CORES + ADK_PAL_MAX_FLOATING_THREADS))
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                      "Invalid lcore id\n");
        ADK_RETURN( retValue );
    }

    adk_pal_cp_dp_info_p->evdev_port_ids[winfo_p->lcore_id] = winfo_p->evdev_port_id;

    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));
    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_worker_to_event_port_map_add----- */

/*!
 * =============================================================
 * adk_pal_worker_to_event_port_map_rem
 * @param[in] sysHandle The ADK system handle (adk_sys_handle_t)
 *
 * @retval ADK_SUCCESS Successfully removed a given woker
 * @retval ADK_E_INVALID_PARAMETER Invalid paramater
 *
 * @brief  removes a given worker from PAL.
 *
 * This function removes a given worker from PAL and cleans up
 * any corresponding internal resources.
 *
 * =============================================================
 */
adk_status_t
adk_pal_worker_to_event_port_map_rem(
    adk_sys_handle_t sysHandle,
    adk_pal_worker_info_t *winfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    int lcore_index;

    if (ADK_INVALID_SYS_HANDLE (sysHandle))
    {
        /*! Invalid system handle */
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR ("Invalid system handle\n");
        return (retValue);
    }

    ADK_LOG_FUNC_ENTRY (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));

    /* validate lcore id */
    if( winfo_p->lcore_id >= (ADK_MAX_NUM_CORES + ADK_PAL_MAX_FLOATING_THREADS))
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                      "Invalid lcore id\n");
        ADK_RETURN( retValue );
    }


    lcore_index = ADK_LCORE_INDEX_LOOKUP(winfo_p->lcore_id);
    if(lcore_index < 0)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                      "Invalid lcore index\n");
        ADK_RETURN( retValue );
    }

    adk_pal_cp_dp_info_p->evdev_port_ids[lcore_index] = ADK_PAL_INVALID_EV_PORT_ID;

    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle));
    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_worker_to_event_port_map_rem----- */

/*!
 * =============================================================
 * adk_pal_bridge_crypto_add
 * @param[in] sysHandle The ADK system handle (adk_sys_handle_t)
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  adds crypto-bridge to PAL.
 *
 * This function adds crypto bridge driver info to PAL.
 *
 * =============================================================
 */

adk_status_t
adk_pal_bridge_crypto_add (
    adk_sys_handle_t  sysHandle,
    adk_pal_bridge_crypto_info_t   *bd_crypto_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    RTE_SET_USED(sysHandle);

    if (bd_crypto_info_p->crypto_dev_id >= ADK_PAL_MAX_NUM_CRYPTO_DEV) {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid crypto-dev-id .\n", __func__);
        ADK_RETURN(ADK_E_INVALID_PARAMETER);
    }

    adk_pal_cp_dp_info_p->crypto_dev_to_event_queue_id[bd_crypto_info_p->crypto_dev_id] =
            bd_crypto_info_p->event_queue_id;

    adk_pal_cp_dp_info_p->crypto_bd_md_offset = ADK_RTE_CRYPTO_BD_TX_METADATA_OFFSET;

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_bridge_crypto_add ----- */


/*!
 * =============================================================
 * adk_pal_update_ppl
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  registres event-dev and worker-event-port information with 
 * ADK PAL.
 *
 * registres event-dev and worker-event-port information with ADK PAL.
 *
 * =============================================================
 */
adk_status_t
adk_pal_update_ppl( adk_sys_rsrc_evdev_info_t *evdev_info_p)
{
    adk_sys_rsrc_evp_info_t *evp_info_p;
    adk_sys_rsrc_evq_info_t *evq_info_p;
    uint8_t                 *wc_index_p;

    uint32_t            num_workers;
    uint8_t             num_ev_queues;
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t worker_index;
    uint32_t evq_index;
    int lcore_index;

    if(evdev_info_p == NULL)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_RETURN( retValue );
    }
    adk_pal_cp_dp_info_p->event_dev_id = evdev_info_p->event_dev_id ;
    evp_info_p                         = evdev_info_p->event_port_list_p;
    evq_info_p                         = evdev_info_p->event_q_list_p;
    num_workers                        = evdev_info_p->num_ev_ports;
    num_ev_queues                      = evdev_info_p->num_ev_queues;
    wc_index_p                         = (uint8_t *) evdev_info_p->reserved;

    for(worker_index=0; worker_index < num_workers; worker_index++)
    {
        lcore_index = wc_index_p[evp_info_p[worker_index].lcore_id];
        if(lcore_index < 0)
        {
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_RETURN( retValue );
        }
        adk_pal_cp_dp_info_p->evdev_port_ids[lcore_index] = \
            evp_info_p[worker_index].ev_port_id;
    }
    for(evq_index=0; evq_index < num_ev_queues; evq_index++)
    {
        adk_pal_cp_dp_info_p->event_q_list[evq_index] = evq_info_p[evq_index].event_queue_id;
    }
    /* if the number of ev-queues are less than the ADK_PAL_MAX_NUM_EVQS
     * then fill in the last event-queue in remaing indexes*/
    for(; evq_index < ADK_PAL_MAX_NUM_EVQS; evq_index++)
    {
        adk_pal_cp_dp_info_p->event_q_list[evq_index] = 
                adk_pal_cp_dp_info_p->event_q_list[evq_index-1];
    }

    adk_pal_cp_dp_info_p->num_ev_queues = num_ev_queues ;

    /* 
     * Step1: read list of staged proc-fns 
     * Step2: for each proc-fn identify the group-id, and event-queu-index 
     * Step3: create proc-stage-entry in the dataplane 
     * */
    {

        extern adk_pal_proc_func_cp_info_t adk_proc_fn_list[];
        adk_pal_stage_entry_info_t  entry_info;
        int index1=0, index2=0, index3=0;
        int num_staged_fns = sizeof(adk_stage_fn_list) / sizeof(adk_stage_fn_list[0]);
        int num_groups     = sizeof(adk_proc_group_list) / sizeof(adk_proc_group_list[0]);

        for(index1=0;index1< num_staged_fns; index1++)
        {
            /* run through the all fn-list array
             * and identify the matching proc_fn_info */
            for(index2=0;index2< ADK_MAX_PROC_FUNC_ID; index2++)
            {
                if(adk_stage_fn_list[index1] == adk_proc_fn_list[index2].fn_id)
                {
                    break;
                }
            }
            if(index2==ADK_MAX_PROC_FUNC_ID)
            {
                ADK_FORCE_LOG_ERROR("\n countn't find the staged-fn-id in the list, un-usal Error case!!!");
                continue;
            }
            /* identify the group-id and associated event-queue-index */
            for(index3=0;index3<num_groups; index3++)
            {
                if(adk_proc_fn_list[index2].group_id == adk_proc_group_list[index3].group_id)
                {
                    break;
                }

            }
            if(index3==num_groups)
            {
                ADK_FORCE_LOG_ERROR("\n countn't find the group-id in the list, un-usal Error case!!!");
                continue;
            }

            /*Create an entry  */
            /*ADK PAL update pipeline / stages to ev-queue mapping */
            entry_info.entry_type = ADK_PAL_STAGE_ENTRY_TYPE_EV_Q;
            entry_info.evnt_q_id  =  adk_pal_cp_dp_info_p->event_q_list[adk_proc_group_list[index3].event_queue_index];

            retValue = adk_pal_stage_entry_update(adk_proc_fn_list[index2].fn_name,
                                          &entry_info);
            if (retValue != ADK_SUCCESS)
            {
                ADK_FORCE_LOG_ERROR("\n couldn't update the state entry for stage %s!!!", adk_proc_fn_list[index2].fn_name);
                continue;
            }
        }

    }


    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_update_ppl----- */


/*!
 * =============================================================
 * adk_pal_eth_dev_subscribe
 *
 * param[in] eth_dev_id  dpdk assigned eth-dev id
 * param[in] pkt_type    packet type filed applicable for VSI eth-devs 
 *                       pkt_type ADK_PAL_PKT_TYPE_ALL means the pkt_type 
 *                       field is ignored. 
 * param[in] fn-id       ADK dataplane entry function id
 *
 * @retval ADK_SUCCESS Successfully updted the PAL tables
 *
 * @brief  ADK modules to cliam ownership of traffic on a given eth-dev
 * =============================================================
 */
adk_status_t 
adk_pal_eth_dev_subscribe(uint16_t eth_dev_id,
                          uint8_t pkt_type,
                          uint8_t fn_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    int table_id=ADK_PAL_PKT_TYPE_TBL_ID_RSVD;
    int id = 0;
    
    /* Validate the eth-dev-id */
    if(eth_dev_id >= ADK_PAL_MAX_NUM_NIC_PORTS) {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_RETURN( retValue );
    }

    adk_pal_dp_eth_dev_t eth_dev;
    
    eth_dev.aggr_info =  adk_pal_cp_dp_info_p->eth_dev_info[eth_dev_id].aggr_info;

    if(pkt_type != ADK_PAL_PKT_TYPE_ALL) {
        /* module requesting for per packet-type demux means assume that the
         * requested eth-dev is a VSI-ETH-DEV. 
         * 1. First check the validity of the eth-dev-type
         * 2. check if table id is already allocated, if not allocate a new
         * table id
         * 3.
         * */
        table_id  = eth_dev.type_tbl_idx;
        if(table_id == ADK_PAL_PKT_TYPE_TBL_ID_RSVD) {
            /*Allocate new table  */
            table_id = adk_rm_resource_reserve(
                            adk_pal_cp_info_p->pkt_type_tbl_rm_handle);
            if(table_id == ADK_RM_RSRC_INVALID) {
                retValue = ADK_E_RESOURCE_UNAVAILABLE;
                ADK_RETURN( retValue );
            }

            /* pre-fill the table with PAL error handler */
            for(id=0;id<ADK_PAL_PKT_TYPE_MAX; id++) {
                adk_pal_cp_dp_info_p->pkt_type_fn_id[table_id][id] = 
                    ADK_PAL_ERR_HANDLER_FUNC_ID;
            }

            /*Store table id in eth-dev information 
                **/
            eth_dev.type_tbl_idx = table_id;
        }
        adk_pal_cp_dp_info_p->pkt_type_fn_id[table_id][pkt_type] = fn_id;
    } else {
        /* Per eth-dev demux  */
        eth_dev.entry_fn_id  = fn_id;
        eth_dev.dev_type     = ADK_PAL_ETH_PORT_TYPE_NIC;
        eth_dev.type_tbl_idx = ADK_PAL_PKT_TYPE_TBL_ID_RSVD;
#ifdef ADK_HOST_PATH_USE_NETD
        eth_dev.port_id      = eth_dev_id;
#endif
    }

    adk_pal_cp_dp_info_p->eth_dev_info[eth_dev_id].aggr_info =  eth_dev.aggr_info;

    ADK_RETURN( retValue );
}

/*!
 * =============================================================
 * adk_pal_eth_dev_unsubscribe
 *
 * param[in] eth_dev_id  dpdk assigned eth-dev id
 * param[in] pkt_type    packet type filed applicable for VSI eth-devs 
 *                       pkt_type ADK_PAL_PKT_TYPE_ALL means the pkt_type 
 *                       field is ignored. 
 *
 * @retval ADK_SUCCESS Successfully updted the PAL tables
 *
 * @brief  ADK modules to cliam ownership of traffic on a given eth-dev
 * =============================================================
 */
adk_status_t 
adk_pal_eth_dev_unsubscribe(uint16_t eth_dev_id,
                          uint8_t pkt_type)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t     table_id=ADK_PAL_PKT_TYPE_TBL_ID_RSVD;
    int    id = 0;
    adk_pal_dp_eth_dev_t eth_dev;

    /* Validate the eth-dev-id */
    if(eth_dev_id >= ADK_PAL_MAX_NUM_NIC_PORTS) {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_RETURN( retValue );
    }
    
    eth_dev.aggr_info = adk_pal_cp_dp_info_p->eth_dev_info[eth_dev_id].aggr_info;
    table_id          = eth_dev.type_tbl_idx;

    if(pkt_type != ADK_PAL_PKT_TYPE_ALL) {
        /* module requesting for per packet-type demux
         * 1. First check the validity of the eth-dev-type
         * 2. check if table id is already allocated, if not allocate a new
         * table id
         * 3.
         * */
        if(table_id == ADK_PAL_PKT_TYPE_TBL_ID_RSVD) {
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_RETURN( retValue );
        }
        adk_pal_cp_dp_info_p->pkt_type_fn_id[table_id][pkt_type] = ADK_PAL_ERR_HANDLER_FUNC_ID;
    } else {
        /* Per eth-dev demux  */
        eth_dev.entry_fn_id  = ADK_PAL_ERR_HANDLER_FUNC_ID;
        eth_dev.dev_type     = ADK_PAL_ETH_PORT_TYPE_INVALID;
        eth_dev.type_tbl_idx = ADK_PAL_PKT_TYPE_TBL_ID_RSVD;
    }

    adk_pal_cp_dp_info_p->eth_dev_info[eth_dev_id].aggr_info = eth_dev.aggr_info;


    if(table_id != ADK_PAL_PKT_TYPE_TBL_ID_RSVD)
    {
        /* pre-fill the table with PAL error handler */
        for(id=0;id<ADK_PAL_PKT_TYPE_MAX; id++) {
            adk_pal_cp_dp_info_p->pkt_type_fn_id[table_id][id] = 
                ADK_PAL_ERR_HANDLER_FUNC_ID;
        }

        if(!adk_rm_resource_free(adk_pal_cp_info_p->pkt_type_tbl_rm_handle, 
                                table_id))
        {
            retValue = ADK_E_INTERNAL_ERROR;
            return (retValue);
        }
    }

    ADK_RETURN( retValue );
}

/*!
 * =============================================================
 * adk_pal_update_cbd
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  adds crypto-bridge to PAL.
 *
 * This function adds crypto bridge driver info to PAL.
 *
 * =============================================================
 */

adk_status_t
adk_pal_update_cbd(
    adk_sys_rsrc_crypto_dev_info_t *cdev_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    if (cdev_info_p->dev_id >= ADK_PAL_MAX_NUM_CRYPTO_DEV) {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid crypto-dev-id .\n", __func__);
        ADK_RETURN(ADK_E_INVALID_PARAMETER);
    }
    /* need Q-pair check only for lookaside device
     */
    if (cdev_info_p->dev_type == ADK_SYS_RSRC_CRYPTO_DEV_TYPE_LA)
    {
        if (cdev_info_p->num_q_pairs == 0) {
            ADK_FORCE_LOG_ERROR ("[%s()]: invalid number of q pairs .\n", __func__);
            ADK_RETURN(ADK_E_INVALID_PARAMETER);
        }

        adk_pal_cp_dp_info_p->crypto_dev_to_event_queue_id[cdev_info_p->dev_id] =
            cdev_info_p->qp_info_p[0].bd_ev_queue_id;
        adk_pal_cp_dp_info_p->crypto_bd_md_offset =  ADK_RTE_CRYPTO_BD_TX_METADATA_OFFSET;
    }


    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_update_cbd ----- */

/*!
 * =============================================================
 * adk_pal_update_eth_dev
 *
 * @retval ADK_SUCCESS Successfully initialized the internal structures.
 * @retval ADK_E_RESOURCE_UNAVAILABLE Unable to get the resource.
 *
 * @brief  adds eth-dev to PAL.
 *
 * This function adds eth dev info to PAL.
 *
 * =============================================================
 */

adk_status_t
adk_pal_update_eth_dev(adk_sys_rsrc_eth_dev_info_t *eth_dev_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint16_t                    *dev_list, eth_dev_id;
    uint8_t                      num_devs;
#if 0
    struct rte_eth_dev_info      dev_info;
#endif
    dev_list        = eth_dev_info_p->eth_dev_list;
    num_devs        = eth_dev_info_p->num_eth_devs;
    
    /* Update the eth-dev table*/
    while(num_devs!=0){
        num_devs--;
        eth_dev_id =  dev_list[num_devs];
        if(eth_dev_id >= ADK_PAL_MAX_NUM_NIC_PORTS) {
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_RETURN(retValue);
        }
        adk_pal_cp_dp_info_p->eth_port_to_event_queue_id[eth_dev_id] = 
            eth_dev_info_p->event_queue_id;
        /*adk_pal_cp_dp_info_p->eth_dev_info[eth_dev_id].entry_fn_id =  ADK_PAL_ERR_HANDLER_FUNC_ID;*/
        
#if 0
        /*Check if the dev is a VSI / NAC type */
         ADK_LOG_EXT_API(adkSysShmConf_p->sysHandle,
                         ADK_SYS_MODULE_HANDLE_GET(adkSysShmConf_p->sysHandle),
                         "rte_eth_dev_info_get",
                         "Retrieve the contextual information of an Ethernet device.\n");
         rte_eth_dev_info_get(eth_dev_id, &dev_info);
#ifdef ADK_DEVICE_SNR
        adk_pal_cp_info_p->vsi_dev_list[eth_dev_id] = false;
        if(dev_info.driver_name) {
            if(!strcmp(dev_info.driver_name, "net_ice_dsi")) {
                adk_pal_cp_info_p->vsi_dev_list[eth_dev_id] = true;
            }
        }
#endif
#endif

    }
    adk_pal_cp_dp_info_p->nic_rx_md_offset = ADK_RTE_ETH_BD_RX_METADATA_OFFSET;
    adk_pal_cp_dp_info_p->nic_tx_md_offset = ADK_RTE_ETH_BD_TX_METADATA_OFFSET;

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_update_eth_dev ----- */

/*
 * API to enable the PAL dataplane extended stats collection
 */
adk_status_t
adk_pal_dp_xstats_enable(void)
{

#ifdef ADK_PAL_CYCLE_MEASUREMENT
    rte_atomic16_set((rte_atomic16_t *)&(adk_pal_cp_dp_info_p->xstats_state), 
                     ADK_PAL_DP_XSTATS_ENABLE);
#endif
    return (ADK_SUCCESS);
}

/*
 * API to disable the PAL dataplane extended stats collection
 */
adk_status_t
adk_pal_dp_xstats_disable(void)
{

#ifdef ADK_PAL_CYCLE_MEASUREMENT
    rte_atomic16_set((rte_atomic16_t *)&(adk_pal_cp_dp_info_p->xstats_state), 
                     ADK_PAL_DP_XSTATS_DISABLE);
#endif
    return (ADK_SUCCESS);
}

#ifdef ADK_PAL_CYCLE_MEASUREMENT
/*
 * API to get PAL dataplane extended stats
 */
static adk_status_t
adk_pal_dp_xstats_get(void)
{
    int fn_index = 0, lcore_index = 0, sample_index=0, base, count =0, state=0;
    uint64_t  core_min;
    uint64_t  core_max;
    uint64_t  core_avg;
    uint64_t  global_avg = 0;
    uint64_t  global_min = 0;
    uint64_t  global_max = 0;

    for ( fn_index= 0 ; fn_index< ADK_PAL_MAX_FN_IDS; fn_index++)
    {
        count = 0;
        global_avg = 0;
        state = 0;
        for ( lcore_index= 0 ; lcore_index < ADK_MAX_NUM_CORES; lcore_index++)
        {
            base = (fn_index * ADK_MAX_NUM_CORES * ADK_PAL_MAX_SAMPLES) +
                   (lcore_index*ADK_PAL_MAX_SAMPLES);
            core_min = adk_pal_cp_min_cycles[base + 0];
            core_max = adk_pal_cp_min_cycles[base + 0];
            core_avg = adk_pal_cp_min_cycles[base + 0];
            for ( sample_index= 0 ; sample_index < ADK_PAL_MAX_SAMPLES; sample_index++)
            {
                if(core_min > adk_pal_cp_min_cycles[base+ sample_index]) {
                        core_min = adk_pal_cp_min_cycles[ base + sample_index];
                }
                if(core_max < adk_pal_cp_min_cycles[ base + sample_index]) {
                        core_max = adk_pal_cp_min_cycles[base + sample_index];
                }
                core_avg += adk_pal_cp_min_cycles[base + sample_index ];
            }
            core_avg = core_avg / ADK_PAL_MAX_SAMPLES;
            proc_fn_stats[fn_index].avg_cycles[lcore_index]  = core_avg;
            proc_fn_stats[fn_index].min_latency[lcore_index] = core_min;
            proc_fn_stats[fn_index].max_latency[lcore_index] = core_max;
            proc_fn_stats[fn_index].avg_latency[lcore_index] = core_avg;

            if(core_avg != 0) {
                global_avg +=core_avg;
                count++;
                if(state == 0) {
                    global_min = core_min;
                    state = 1;
                }
                if(global_min > core_min) {
                    global_min = core_min;
                }
                if(global_max < core_max) {
                    global_max = core_max;
                }
            }
        }
        if(global_avg!=0)
        {
            proc_fn_stats[fn_index].avg_cycles[ADK_MAX_NUM_CORES]  = global_avg / count;
            proc_fn_stats[fn_index].min_latency[ADK_MAX_NUM_CORES] = global_min;
            proc_fn_stats[fn_index].max_latency[ADK_MAX_NUM_CORES] = global_max;
            proc_fn_stats[fn_index].avg_latency[ADK_MAX_NUM_CORES] = global_avg / count;

        }
    }
    return (ADK_SUCCESS);
}

/*
 * API to get PAL dataplane extended stats
 */
static adk_status_t
adk_pal_dp_xstats_print(void)
{
    int fn_index = 0;

    for ( fn_index= 0 ; fn_index< ADK_PAL_MAX_FN_IDS; fn_index++)
    {
        if(proc_fn_stats[fn_index].avg_cycles[ADK_MAX_NUM_CORES] != 0) {
            if(fn_index==255) {
                ADK_INFO_WRITE("\n Measurements for processing function[%s]::\n",
                    "adk_fps_send");
            } else {
                ADK_INFO_WRITE("\n Measurements for processing function[%s]::\n",
                    adk_pal_proc_fn_id_to_name(fn_index));
            }
            ADK_INFO_WRITE("\t Average Cycles ::%d\n",
                           proc_fn_stats[fn_index].avg_cycles[ADK_MAX_NUM_CORES]);
            ADK_INFO_WRITE("\t Minumum Latency::%0.3f usec\n",
                           (((double)proc_fn_stats[fn_index].min_latency[ADK_MAX_NUM_CORES] / rte_get_tsc_hz()) * 1000000));
            ADK_INFO_WRITE("\t Maximum Latency::%0.3f usec\n",
                           (((double)proc_fn_stats[fn_index].max_latency[ADK_MAX_NUM_CORES] / rte_get_tsc_hz()) * 1000000));
            ADK_INFO_WRITE("\t Average Latency::%0.3f usec\n\n",
                           (((double)proc_fn_stats[fn_index].avg_latency[ADK_MAX_NUM_CORES] / rte_get_tsc_hz()) * 1000000));
        }
    }
    return (ADK_SUCCESS);
}
#endif


/*!
* =============================================================
*
* adk_pal_global_stats_get
*
* @param[in]   stats      of type adk_pal_global_stats_t
*
* @retval  ADK_SUCCESS                                            Successful completion
* @retval  ADK_E_INVALID_PARAMETER                 Invalid parameter (stats is NULL)
* @retval  ADK_E_INTERNAL_ERROR                        Internal error while getting stats
*
* @brief   API to get the stats of ADK PAL  infrastructure 
*
* =============================================================
*/

adk_status_t adk_pal_global_stats_get (adk_pal_global_stats_t *stats)
{
    adk_status_t ret = ADK_SUCCESS;
    uint8_t i =0;

    if(stats == NULL)
    {
        return ADK_E_INVALID_PARAMETER;
    }

    ADK_MEMSET((void *)stats, 0x00, sizeof(adk_pal_global_stats_t));
    for (i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        /*utility counters*/
        stats->arrival_evnts += adk_pal_cp_dp_info_p->stats[i].arrival_evnts;
        stats->nic_arrival_evnts += adk_pal_cp_dp_info_p->stats[i].nic_arrival_evnts;
#ifdef ADK_HOST_PATH_USE_NETD
        stats->netd_arrival_events += adk_pal_cp_dp_info_p->stats[i].netd_arrival_events;
#endif      
        stats->timer_arrival_events += adk_pal_cp_dp_info_p->stats[i].timer_arrival_events;
        stats->crypto_arrival_events += adk_pal_cp_dp_info_p->stats[i].crypto_arrival_events;
        stats->em_arrival_events += adk_pal_cp_dp_info_p->stats[i].em_arrival_events;
        stats->cpu_arrival_events += adk_pal_cp_dp_info_p->stats[i].cpu_arrival_events; 
        stats->pal_batched_em_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_em_events;
        stats->pal_batched_nic_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_nic_events;
        stats->pal_batched_ph_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_ph_events;
        stats->pal_batched_netd_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_netd_events;
        stats->pal_batched_stage_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_stage_events;
        stats->pal_batched_crypto_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_crypto_events;
        stats->pal_batched_release_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_release_events;
        stats->pal_batched_timer_events += adk_pal_cp_dp_info_p->stats[i].pal_batched_timer_events;
        stats->em_xmt_events += adk_pal_cp_dp_info_p->stats[i].em_xmt_events;
        stats->rte_xmt_events += adk_pal_cp_dp_info_p->stats[i].rte_xmt_events;

        /*Error counters*/    
        stats->pal_dispatcher_err_hdl += adk_pal_cp_dp_info_p->stats[i].pal_dispatcher_err_hdl;
        stats->nic_arrival_null_mbuf_evnts_drop += adk_pal_cp_dp_info_p->stats[i].nic_arrival_null_mbuf_evnts_drop;
        stats->em_xmt_fail_events += adk_pal_cp_dp_info_p->stats[i].em_xmt_fail_events;
        stats->rte_xmt_fail_events += adk_pal_cp_dp_info_p->stats[i].rte_xmt_fail_events;
        stats->batch_fail_safe_err += adk_pal_cp_dp_info_p->stats[i].pal_batch_fail_safe_err;
        stats->batch_fail_safe_drops += adk_pal_cp_dp_info_p->stats[i].pal_batch_fail_safe_drop;
    }
    return ret;
}
    
#ifdef ADK_DEBUG
adk_status_t
adk_pal_stats_clear(void)
{

    adk_status_t ret = ADK_SUCCESS;
    ADK_MEMSET((void *)adk_pal_cp_dp_info_p->stats, 0x00, (sizeof(adk_pal_dp_stats_t) * ADK_MAX_NUM_CORES));
    return (ret);
}

adk_status_t
adk_pal_stats_print(void)
{
    return adk_pal_dp_stats_print();

}
#endif


adk_status_t
adk_pal_set_utility_counters(bool utility_cnt_state)
{
    if(utility_cnt_state)
    {
        adk_pal_cp_dp_info_p->isUtilityCntEnabled = true;
    }
    else
    {
        adk_pal_cp_dp_info_p->isUtilityCntEnabled = false;
        for (int i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            adk_pal_cp_dp_info_p->stats[i].arrival_evnts = 0;
            adk_pal_cp_dp_info_p->stats[i].nic_arrival_evnts = 0;
#ifdef ADK_HOST_PATH_USE_NETD
            adk_pal_cp_dp_info_p->stats[i].netd_arrival_events = 0;
#endif
            adk_pal_cp_dp_info_p->stats[i].timer_arrival_events = 0;
            adk_pal_cp_dp_info_p->stats[i].crypto_arrival_events = 0;
            adk_pal_cp_dp_info_p->stats[i].em_arrival_events = 0;
            adk_pal_cp_dp_info_p->stats[i].cpu_arrival_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_em_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_nic_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_ph_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_netd_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_stage_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_crypto_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_release_events = 0;
            adk_pal_cp_dp_info_p->stats[i].pal_batched_timer_events = 0;
            adk_pal_cp_dp_info_p->stats[i].em_xmt_events = 0;
            adk_pal_cp_dp_info_p->stats[i].rte_xmt_events = 0;
        }
    }
    return ADK_SUCCESS;
}
/*
 * =============================================================
 *  ADK_DP_PAL_EXTERNAL
 *
 *
 * =============================================================
 */

adk_status_t
adk_pal_dp_stats_print(void)
{

    adk_status_t ret = ADK_SUCCESS;
    adk_pal_dp_stats_t global_stats;
    int i =0;

    ADK_MEMSET((void *)&global_stats, 0x00, sizeof(adk_pal_dp_stats_t));
#ifdef ADK_PAL_CYCLE_MEASUREMENT
    if(adk_pal_cp_dp_info_p->xstats_state) {
        adk_pal_dp_xstats_get();
        adk_pal_dp_xstats_print();
    }
#endif

    for (i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        global_stats.arrival_evnts += adk_pal_cp_dp_info_p->stats[i].arrival_evnts;
        global_stats.nic_arrival_evnts += adk_pal_cp_dp_info_p->stats[i].nic_arrival_evnts;
#ifdef ADK_HOST_PATH_USE_NETD
        global_stats.netd_arrival_events += adk_pal_cp_dp_info_p->stats[i].netd_arrival_events;
#endif
        global_stats.timer_arrival_events += adk_pal_cp_dp_info_p->stats[i].timer_arrival_events;
        global_stats.crypto_arrival_events += adk_pal_cp_dp_info_p->stats[i].crypto_arrival_events;
        global_stats.em_arrival_events += adk_pal_cp_dp_info_p->stats[i].em_arrival_events;
        global_stats.cpu_arrival_events += adk_pal_cp_dp_info_p->stats[i].cpu_arrival_events;
        global_stats.pal_dispatcher_err_hdl += adk_pal_cp_dp_info_p->stats[i].pal_dispatcher_err_hdl;
        global_stats.nic_arrival_null_mbuf_evnts_drop += adk_pal_cp_dp_info_p->stats[i].nic_arrival_null_mbuf_evnts_drop;

#ifdef ADK_PERFORMANCE_STATS
	if (adk_pal_cp_dp_info_p->stats[i].user_stats.loop_count != 0 ) {
		printf("************************************");
		printf("\nStart Loop count :%lu End loop count: %lu", adk_pal_cp_dp_info_p->stats[i].start_loop_count,
			adk_pal_cp_dp_info_p->stats[i].user_stats.loop_count);
		printf("\nActual Loop count :%lu events Dispatcher :%lu\n",
			(adk_pal_cp_dp_info_p->stats[i].user_stats.loop_count - adk_pal_cp_dp_info_p->stats[i].start_loop_count),
			adk_pal_cp_dp_info_p->stats[i].user_stats.events_dispatched);
		printf("Total numbers of dequeue:%lu", adk_pal_cp_dp_info_p->stats[i].total_num_of_dequeue);
		printf("\nCore %d: EMPTY DEQUEUES ::%ld \n",i,adk_pal_cp_dp_info_p->stats[i].batch_stats[0]);
		for (uint8_t j=1;j<=32;j+=4) {
			printf("\nBatch size:%2d numBatches::%3ld		\nBatch size:%2d NumBatches::%3ld		\nBatch size:%2d numBatches::%3ld		\nBatch size:%2d NumBatches::%3ld",
					j, adk_pal_cp_dp_info_p->stats[i].batch_stats[j],
					j+1, adk_pal_cp_dp_info_p->stats[i].batch_stats[j+1],
					j+2, adk_pal_cp_dp_info_p->stats[i].batch_stats[j+2],
					j+3, adk_pal_cp_dp_info_p->stats[i].batch_stats[j+3]);
		}
		printf("\n\n");
		for(uint8_t j = 0; j < ADK_MAX_NUM_OF_FLOW_SUPPORT; j ++) {
			if(adk_pal_cp_dp_info_p->stats[i].num_flows[j].flowid != 0)
				printf("Core:%d Flows num:%d-->Flowid:0x%x,flow_count:%ld \n",i, j, adk_pal_cp_dp_info_p->stats[i].num_flows[j].flowid,
					adk_pal_cp_dp_info_p->stats[i].num_flows[j].flow_count);
		}
		printf("====================================");
		printf("\n\n");
	}
#endif

        global_stats.pal_batched_em_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_em_events;
        global_stats.pal_batched_nic_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_nic_events;
        global_stats.pal_batched_ph_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_ph_events;
        global_stats.pal_batched_netd_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_netd_events;
        global_stats.pal_batched_stage_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_stage_events;
        global_stats.pal_batched_crypto_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_crypto_events;
        global_stats.pal_batched_release_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_release_events;
        global_stats.pal_batched_timer_events+= adk_pal_cp_dp_info_p->stats[i].pal_batched_timer_events;
        global_stats.em_xmt_events += adk_pal_cp_dp_info_p->stats[i].em_xmt_events;
        global_stats.em_xmt_fail_events += adk_pal_cp_dp_info_p->stats[i].em_xmt_fail_events;
        global_stats.rte_xmt_events += adk_pal_cp_dp_info_p->stats[i].rte_xmt_events;
        global_stats.rte_xmt_fail_events += adk_pal_cp_dp_info_p->stats[i].rte_xmt_fail_events;
        global_stats.pal_batch_fail_safe_err += adk_pal_cp_dp_info_p->stats[i].pal_batch_fail_safe_err;
        global_stats.pal_batch_fail_safe_drop += adk_pal_cp_dp_info_p->stats[i].pal_batch_fail_safe_drop;

    }
    ADK_INFO_WRITE("\n ADK PAL Global Stats ");
    ADK_INFO_WRITE("\n =========================");
    ADK_INFO_WRITE("\n pal_dispatcher_err_hdl ::%ld ",global_stats.pal_dispatcher_err_hdl);
    ADK_INFO_WRITE("\n nic_arrival_null_mbuf_evnts_drop ::%ld ",global_stats.nic_arrival_null_mbuf_evnts_drop);

    ADK_INFO_WRITE("\n\n ============PAL BATCH COUNTER============");
    ADK_INFO_WRITE("\n arrival_evnts          ::%ld ",global_stats.arrival_evnts);
    ADK_INFO_WRITE("\n nic_arrival_evnts      ::%ld ",global_stats.nic_arrival_evnts);
    ADK_INFO_WRITE("\n em_arrival_events  ::%ld ",global_stats.em_arrival_events);
    ADK_INFO_WRITE("\n cpu_arrival_events  ::%ld",global_stats.cpu_arrival_events);
#ifdef ADK_HOST_PATH_USE_NETD
    ADK_INFO_WRITE("\n netd_arrival_events    ::%ld",global_stats.netd_arrival_events);
#endif
    ADK_INFO_WRITE("\n crypto_arrival_events  ::%ld ",global_stats.crypto_arrival_events);
    ADK_INFO_WRITE("\n timer_arrival_events   ::%ld ",global_stats.timer_arrival_events);

    ADK_INFO_WRITE("\n\n pal_batched_em_events::%ld ",global_stats.pal_batched_em_events);
    ADK_INFO_WRITE("\n pal_batched_nic_events::%ld ",global_stats.pal_batched_nic_events);
    ADK_INFO_WRITE("\n pal_batched_ph_events::%ld ",global_stats.pal_batched_ph_events);
    ADK_INFO_WRITE("\n pal_batched_netd_events::%ld ",global_stats.pal_batched_netd_events);
    ADK_INFO_WRITE("\n pal_batched_stage_events::%ld ",global_stats.pal_batched_stage_events);
    ADK_INFO_WRITE("\n pal_batched_crypto_events::%ld ",global_stats.pal_batched_crypto_events);
    ADK_INFO_WRITE("\n pal_batched_release_events::%ld \n",global_stats.pal_batched_release_events);
    ADK_INFO_WRITE("\n pal_batched_timer_events::%ld \n",global_stats.pal_batched_timer_events);
    ADK_INFO_WRITE("\n em_xmt_events::%ld ",global_stats.em_xmt_events);
    ADK_INFO_WRITE("\n em_xmt_fail_events ::%ld ",global_stats.em_xmt_fail_events);
    ADK_INFO_WRITE("\n rte_xmt_events ::%ld ",global_stats.rte_xmt_events);
    ADK_INFO_WRITE("\n rte_xmt_fail_events ::%ld ",global_stats.rte_xmt_fail_events);
    ADK_INFO_WRITE("\n pal_batch_fail_safe_err ::%ld \n",global_stats.pal_batch_fail_safe_err);
    ADK_INFO_WRITE("\n pal_batch_fail_safe_drop ::%ld \n",global_stats.pal_batch_fail_safe_drop);
    return(ret);
}    /* -----  end of function adk_pal_dp_stats_print ----- */

/*!
 * =============================================================
 * adk_pal_dequeue_burst
 * @param[in] lcoreid logical core/ thread on where de-queue is happening
 * @param[in] *dpdk_mbuf_p[] arrary of DPDK mbuf pointer.
 * @param[in] *sub_event_id[] arrary of sub-event-ids 1-1 mapping with mbufs.
 * @param[in] burst_size  max burst size.  should be less than 32.
 *
 * @retval Number of packets de-queued.
 *
 * @brief  de-queues burst of packets from logical cores port.
 *
 * =============================================================
 */

uint32_t
adk_pal_dequeue_burst (
    uint8_t lcore_id,
    struct rte_mbuf *dpdk_mbuf_p[],
    uint8_t sub_event_id[],
    uint8_t burst_size,
    bool release_send)
{
    int32_t numOfPktsRecv = 0;
    int32_t numOfReleaseSend = 0;
    struct rte_event ev[32];
    uint8_t index = 0;

    numOfPktsRecv = rte_event_dequeue_burst(adk_pal_cp_dp_info_p->event_dev_id,
                                            adk_pal_cp_dp_info_p->evdev_port_ids[lcore_id],
                                            ev, burst_size, false);
    if (numOfPktsRecv < 0)
    {
        debug_print("rte_event_dequeue_burst failed: %d\n", numOfPktsRecv);
    } else
    {
        if ( numOfPktsRecv > 32 )
        {
            debug_print("\n [%s()] un-supported de-queue burst size : %d\n", __func__, numOfPktsRecv);
            numOfPktsRecv = 0;
            return numOfPktsRecv;
        }
        for (index =0; index < numOfPktsRecv; index++)
        {
            dpdk_mbuf_p[index] = ev[index].mbuf;
            sub_event_id[index] = ev[index].sub_event_type;
        }
        if(release_send & (numOfPktsRecv > 0))
        {
            for (index =0; index < numOfPktsRecv; index++)
            {
                ev[index].event_type     = RTE_EVENT_TYPE_CPU;
                ev[index].mbuf = NULL;
                ev[index].op             = RTE_EVENT_OP_RELEASE;
            }
            numOfReleaseSend = rte_event_enqueue_burst(adk_pal_cp_dp_info_p->event_dev_id,
                                        adk_pal_cp_dp_info_p->evdev_port_ids[lcore_id], ev, numOfPktsRecv);
            if (numOfReleaseSend != numOfPktsRecv) {
                     debug_print("[%s()] Failed send release for PH events.\n",
                     __func__);
            }
            else
            {
                debug_print("[%d] RELEASE event sent successfully for [%d]PH received events\n",numOfReleaseSend, numOfPktsRecv);
            }
        }
    }

    return numOfPktsRecv;
}    /* -----  end of function adk_pal_dequeue_burst ----- */




/*!     end addtogroup _palFunctions
 *     @}
 *  end addtogroup _pal
 * @}
 */
