/****************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_pal_ppl.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _palFunctions  PAL Functions
 *
 * @{
 *
 * @brief
 *
 * This file contains all the APIs related to dynamic pipeline 
 *
 ***********************************************************************/

#include <string.h>

/* PAL external header files */
#include <adk_include.h>
#include "adk_pal.h"
#include "adk_pal_cp_pvt.h"
#include "adk_pal_dp.h"
#include "adk_pal_dp_export.h"

#include "adk_shm.h"

extern adk_pal_dp_info_t *adk_pal_cp_dp_info_p;

/*List of processing functions added by each module
 * Each module addes all of it's dataplane stageable processing functions
 * 
 * this is a control plane table
 */
adk_pal_proc_func_cp_info_t adk_proc_fn_list[ADK_MAX_PROC_FUNC_ID] ={

/*
* FUNCTION-NAME         MODULE_ID     FUNCTION_ID       LB_TYPE         GROUP_ID    FUNCTION_PRIORITY
* 
*/
/*Don't move the first include*/
/*This is for EM packets*/
#include "adk_sif_em_proc_fn_list.h"

#include "adk_sif_dp_proc_fn_list.h"

#include "adk_ph_dev_dp_proc_fn_list.h"

#ifdef ADK_MODULE_ETHERNET_ENABLED
#include "adk_eth_dp_proc_fn_list.h"
#endif

#ifdef ADK_MODULE_ETH_OAM_ENABLED
#include "adk_eth_oam_dp_proc_fn_list.h"
#endif

#ifdef ADK_MODULE_IP_ENABLED
#include "adk_ip_dp_proc_fn_list.h"
#endif

#ifdef ADK_MODULE_IPV6_ENABLED
#include "adk_ipv6_dp_proc_fn_list.h"
#endif

#ifdef ADK_MODULE_IPSEC_ENABLED
#include "adk_ipsec_dp_proc_fn_list.h"
#endif

#ifdef ADK_MODULE_BFD_ENABLED
#include "adk_bfd_dp_proc_fn_list.h"
#endif

#ifdef ADK_MODULE_TWAMP_ENABLED
#include "adk_twamp_dp_proc_fn_list.h"
#endif


};

/* 
 * ADK_CP_EXTERNAL 
 * This API to be called by application / configuration layer. 
 * to change a stage's entry type
 */
adk_status_t 
adk_pal_stage_entry_update(const char *stage_name,
			adk_pal_stage_entry_info_t  *new_entry_info_p)
{

    adk_status_t retValue = ADK_SUCCESS;
    uint16_t fn_index;
    uint16_t stage_id;
    adk_pal_stage_entry_info_t *stage_entry_info_p;
    adk_pal_stage_lb_type_e  	stage_lb_type;

    /* function-id (fn_id) is reserved */
    for(fn_index=0; fn_index< 256; fn_index++)
    {
        if(adk_proc_fn_list[fn_index].fn_name[0] == 0)
            continue;
        if(ADK_STRNCMP(adk_proc_fn_list[fn_index].fn_name, stage_name,
                       sizeof(adk_proc_fn_list[fn_index].fn_name)) == 0)
        {
            stage_lb_type = adk_proc_fn_list[fn_index].lb_type;
            stage_id      = adk_proc_fn_list[fn_index].fn_id;
            break;
        }
    }
    if(fn_index== 256)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid stage name %s or \
                       stage has not registred yet\n", __func__, stage_name);
        return (ADK_E_INVALID_PARAMETER);
    }

    if(stage_id > 255)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid stage entry function id\n",__func__);
        return (ADK_E_INVALID_PARAMETER);
    }

    stage_entry_info_p = &(adk_pal_cp_dp_info_p->stage_info_p[stage_id]->entry_info);

    switch(new_entry_info_p->entry_type)
    {
    case ADK_PAL_STAGE_ENTRY_TYPE_EV_Q:
        stage_entry_info_p->entry_type =  ADK_PAL_STAGE_ENTRY_TYPE_EV_Q; 
        stage_entry_info_p->evnt_q_id  = new_entry_info_p->evnt_q_id;
        stage_entry_info_p->stage_priority= new_entry_info_p->stage_priority;
        adk_pal_cp_dp_info_p->stage_info_p[stage_id]->stage_lb_type = stage_lb_type;

        /*Update newly created dp-stage-information */
        adk_pal_cp_dp_info_p->dp_stage_info[stage_id].lb_type  = stage_lb_type;
        adk_pal_cp_dp_info_p->dp_stage_info[stage_id].priority = new_entry_info_p->stage_priority;
        adk_pal_cp_dp_info_p->dp_stage_info[stage_id].event_q_id = new_entry_info_p->evnt_q_id;

        break;

    case ADK_PAL_STAGE_ENTRY_TYPE_FN_CALL:
        stage_entry_info_p->entry_type =  ADK_PAL_STAGE_ENTRY_TYPE_FN_CALL; 
        /*Update newly created dp-stage-information */
        adk_pal_cp_dp_info_p->dp_stage_info[stage_id].event_q_id = -1;
        break;

    default:
        ADK_FORCE_LOG_ERROR ("adk_pal_stage_entry_update: un-supported stage entry type\n");
        return (ADK_E_UNSUPPORTED);
        break;
    }

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_stage_entry_update ----- */

/* 
 * ADK_CP_EXTERNAL 
 * This API to be called by application / configuration layer. 
 * to change a stage's LB type
 */
adk_status_t 
adk_pal_stage_lb_update(const char *stage_name,
			adk_pal_stage_lb_type_e  new_lb_type)
{

    adk_status_t retValue = ADK_SUCCESS;
    adk_pal_stage_info_t *stage_info_p;
    uint16_t stage_id;
    uint16_t fn_id;

    /* function-id (fn_id) is reserved */
    for(stage_id = 1; stage_id < 256; stage_id++)
    {
        if(adk_proc_fn_list[stage_id].fn_name[0] == 0)
            continue;
        if(ADK_STRNCMP(adk_proc_fn_list[stage_id].fn_name , stage_name,
                       sizeof(adk_proc_fn_list[stage_id].fn_name)) == 0)
        {
            fn_id = adk_proc_fn_list[stage_id].fn_id;
            break;
        }
    }
    if(stage_id == 256)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid stage name or \
                       stage has not registred yet\n",__func__);
        return (ADK_E_INVALID_PARAMETER);
    }

    if(fn_id > 255)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid stage entry function id\n",__func__);
        return (ADK_E_INVALID_PARAMETER);
    }

    stage_info_p = adk_pal_cp_dp_info_p->stage_info_p[fn_id];

    switch(stage_info_p->stage_lb_type)
    {
    case ADK_PAL_STAGE_LB_TYPE_ATOMIC:
        ADK_FORCE_LOG_ERROR ("LB type update can't be done for ATOMIC Stages\n");
        return (ADK_E_UNSUPPORTED);
        break;

    case ADK_PAL_STAGE_LB_TYPE_ORDERED:
        if (new_lb_type == ADK_PAL_STAGE_LB_TYPE_ATOMIC) 
        {
            stage_info_p->stage_lb_type =  ADK_PAL_STAGE_LB_TYPE_ATOMIC;
            adk_pal_cp_dp_info_p->dp_stage_info[fn_id].lb_type =  ADK_PAL_STAGE_LB_TYPE_ATOMIC;
        } else {
            ADK_FORCE_LOG_ERROR ("type can be update to ATOMIC only for ORDERED Stages\n");
            return (ADK_E_UNSUPPORTED);
        }
        break;
    case ADK_PAL_STAGE_LB_TYPE_PARALLE:
        if (new_lb_type == ADK_PAL_STAGE_LB_TYPE_DIRECT) 
        {
            ADK_FORCE_LOG_ERROR (" ADK_PAL_STAGE_LB_TYPE_DIRECT type is un-supported \n");
            return (ADK_E_UNSUPPORTED);
        } else {
            stage_info_p->stage_lb_type =  new_lb_type;
            adk_pal_cp_dp_info_p->dp_stage_info[fn_id].lb_type =  new_lb_type;
        }
        break;
    default:
        ADK_FORCE_LOG_ERROR ("adk_pal_stage_lb_update: un-supported lb type %d \n", new_lb_type);
        return (ADK_E_UNSUPPORTED);
        break;
    }

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_stage_lb_update ----- */

/*
 * ADK_CP_INTERNAL 
 * This is called by modules to receive traffic from a eth-dev
 * 
 */
adk_status_t 
adk_pal_stage_info_get(uint8_t fn_id, adk_pal_stage_info_t **info_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    if(info_p == NULL)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid address\n",__func__);
        return (ADK_E_INVALID_PARAMETER);
    }

    *info_p = adk_pal_cp_dp_info_p->stage_info_p[fn_id];

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_stage_info_get ----- */

/*
 * ADK_CP_INTERNAL 
 * This is called by modules to receive traffic from a eth-dev
 * 
 */
adk_status_t 
adk_pal_dp_stage_info_get(uint8_t fn_id, adk_pal_dp_stage_info_t **info_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    if(info_p == NULL)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid address\n",__func__);
        return (ADK_E_INVALID_PARAMETER);
    }

    *info_p = &(adk_pal_cp_dp_info_p->dp_stage_info[fn_id]);

    ADK_RETURN( retValue );
}    /* -----  end of function adk_pal_dp_stage_info_get ----- */


/*
 * PAL internal helper function
 * that returns function name from id
 */
char *adk_pal_proc_fn_id_to_name(adk_pal_proc_fn_id_e fn_id)
{
    uint16_t stage_id;

    /* function-id (fn_id) '0'  is reserved */
    for(stage_id = 0; stage_id < 256; stage_id++)
    {
        if(adk_proc_fn_list[stage_id].fn_id == fn_id)
            break;
    }
    if(stage_id == 256) {
        return ((char *)("Invalid fn-index"));
    } else {
        return ((char *)&adk_proc_fn_list[stage_id].fn_name);
    }
}    /* -----  end of function adk_pal_proc_fn_id_to_name----- */

/*
 * PAL internal helper function
 * that returns id from function name
 */
 adk_pal_proc_fn_id_e adk_pal_proc_fn_name_to_id(char *name)
{
    uint16_t stage_id;

    /* function-id (fn_id) '0'  is reserved */
    for(stage_id = 1; stage_id < 256; stage_id++)
    {
        if(adk_proc_fn_list[stage_id].fn_name[0] == 0 )
            continue;
        if(ADK_STRNCMP(adk_proc_fn_list[stage_id].fn_name , name,
                       sizeof(adk_proc_fn_list[stage_id].fn_name)) == 0)
            break;
    }
    if(stage_id == 256)
    {
        ADK_FORCE_LOG_ERROR ("[%s()]: invalid stage name or \
                       stage has not registred yet\n",__func__);
        stage_id = 0;
    }
    return (stage_id );
}    /* -----  end of function adk_pal_proc_fn_name_to_id ----- */


/*
 * =============================================================
 *  ADK_PAL_EXTERNAL
 *
 *
 * =============================================================
 */
adk_status_t
adk_pal_dp_ppl_dump(void)
{
    adk_status_t ret = ADK_SUCCESS;
    uint16_t stage_index;
    adk_pal_stage_info_t  *stage_info_p;
    adk_pal_stage_entry_info_t  *stage_entry_info_p;

    ADK_INFO_WRITE("\n ADK PAL Pipeline Dump");
    ADK_INFO_WRITE("\n ========================================================================\n");
    ADK_INFO_WRITE("\t STAGE-NAME\t\t\t\t EVENT-Q-ID  \t LB-TYPE");
    ADK_INFO_WRITE("\n ========================================================================\n");
    for(stage_index = 0; stage_index < ADK_MAX_PROC_FUNC_ID; stage_index++)
    {
        stage_info_p = adk_pal_cp_dp_info_p->stage_info_p[stage_index];
        stage_entry_info_p = &(stage_info_p->entry_info);

        if(stage_entry_info_p->entry_type == ADK_PAL_STAGE_ENTRY_TYPE_EV_Q)
        {
            ADK_INFO_WRITE("\t %5s",adk_pal_proc_fn_id_to_name(stage_index));
            ADK_INFO_WRITE("\t\t %d",stage_entry_info_p->evnt_q_id);
            ADK_INFO_WRITE("\t %d\n",stage_info_p->stage_lb_type);
        }
    }
    return(ret);
}    /* -----  end of function adk_pal_dp_ppl_dump----- */
/*!
 * end addtogroup _palFunctions  PAL
 *
 * @}
 */
