/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ph_dev.c
 *
 *      $HeadURL: ncp_adk/code/ph/include/adk_ph.h $
 *      $Revision: 366f377355b8b32e2740de4f058f578662bfab2c $
 *      $Date: Mon Oct 10 12:54:51 2016 -0500 $
 *
 * @addtogroup _ph  PH Module
 * @{
 *
 * @brief
 *
 * Contains the structure/union/macro definitions used by the PDU handler.
 * This file needs to be included by the user of the PDU handler.
 *
 * The PH Dev model is that a given device can have several instances. The
 * instance is where PDUs are transmitted and received.
 *
 * PDUs are received into queues, which are setup via a separate API call.
 * If there are callback functions configured they will be called after the
 * PDU(s) are received and before the adk_ph_dev_rx_burst() API returns. The
 * Rx path can use either ring buffers or Eventdev queues as the transport
 * from DP to CP.
 *
 * PH Dev will only use Eventdev for the transport from CP to DP. There are no
 * queues to be configured for the Tx. The Eventdev QID is passed to PH Dev in
 * the adk_ph_dev_tx_burst() API. There are no callback functions in the Tx
 * direction.
 *
 *  **********************************************************************/

#include "rte_config.h"
#include "rte_mbuf.h"
#include "rte_mempool.h"

#include "inttypes.h"

#include "adk_os.h"
#include "adk_sif.h"
#include "adk_sys.h"
#include "adk_ph_dev.h"
#include "adk_ph_dev_internal.h"
#include "adk_ph_dev_export.h"
#include "adk_log_lib.h"
#include "adk_error.h"
#include "rte_bridge_cppi.h"
#include "adk_pal_dp_export.h"
#include "adk_pal_internal.h"
#include "adk_qos_include_internal.h"
#include "adk_ph_dev_dp.h"
/*!
 * @addtogroup _phFunctions  PH Functions
 * @{
 *
 *
 * The API function declarations.   {{{1
 */

/*!
 *
 * Pointer to the CP shared memory segment that will hold the PH configuration
 *
 */
static adk_ph_dev_shm_info_t    *phShm_p;

/*!
 * =============================================================
 * adk_ph_dev_instance_create   {{{2
 *
 * @param[in] sysHandle: System handle
 * @param[in] config_p: Device configuration parameters
 *
 *
 * @retval ADK_SUCCESS
 *      ADK_SUCCESS
 *      ADK_E_INVALID_PARAMETER
 *      ADK_E_RESOURCE_UNAVAILABLE
 *      ADK_E_MEM_UNAVAILABLE
 *
 *
 * @brief  Initializes the PH device
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_instance_create(
    adk_sys_handle_t                sysHandle,
    adk_ph_dev_instance_config_t   *config_p)
{
    adk_status_t    retValue = ADK_SUCCESS;
    uint16_t        i;

    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        ADK_FORCE_LOG_ERROR("PH shared memory not allocated\n");
        goto _return;
    }

    if (NULL == config_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "config_p is NULL\n");
        goto _parameter_validation_fail;
    }

    retValue = adk_ph_dev_shm_create(sysHandle, config_p, &phShm_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH Dev Shared memory creation failed: 0x%x\n",
                      retValue);
        goto _runtime_failure_1;
    }

    /*
     * Reserve Device Net ID range for PH Dev use
     */
    for (i = ADK_PH_DEV_DEVICE_ID_BASE; i <= ADK_PH_DEV_DEVICE_ID_MAX; i++)
    {
        phShm_p->devices[i] = NULL;
    }


    for(i = 0; i < ADK_PH_CBFN_MAX; i++)
    {
        phShm_p->cbFnIdMap[i].cbFn = NULL;
        phShm_p->cbFnIdMap[i].cbParams = NULL;
    }

    for(i = 0; i <= ADK_PH_MODULE_ID_MAX; i++)
    {
        phShm_p->parseFnMap[i].parseFn = NULL;
        phShm_p->parseFnMap[i].moduleHandle = NULL;
    }

_runtime_failure_1:
_parameter_validation_fail:
_return:
    return retValue;
}


/*!
 * =============================================================
 * adk_ph_dev_instance_get   {{{2
 *
 * @param[in] sysHandle: System handle
 * @param[in] config_p: Device configuration parameters
 *
 *
 * @retval ADK_SUCCESS
 *      ADK_SUCCESS
 *      ADK_E_INVALID_PARAMETER
 *      ADK_E_RESOURCE_UNAVAILABLE
 *      ADK_E_MEM_UNAVAILABLE
 *
 *
 * @brief  Initialises the TLS PH SHM pointers
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_instance_get(
    adk_sys_handle_t                sysHandle)
{
    adk_status_t    retValue = ADK_SUCCESS;

    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        ADK_FORCE_LOG_ERROR("PH shared memory not allocated\n");
        goto _return;
    }

    retValue = adk_ph_dev_shm_get(sysHandle, &phShm_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH Dev Shared memory creation failed: 0x%x\n",
                      retValue);
        goto _runtime_failure_1;
    }

_runtime_failure_1:
_return:
    return retValue;
}


/*!
 * =============================================================
 * adk_ph_dev_device_configure   {{{2
 *
 * @param[in] config_p: Device configuration parameters
 * @param[out] deviceId_p: The ID for the device within the range of 0 to
 *                         (ADK_PH_DEV_MAX_NUM_DEVICES - 1)
 *
 *
 * @retval ADK_SUCCESS
 *      ADK_SUCCESS
 *      ADK_E_INVALID_PARAMETER
 *      ADK_E_RESOURCE_UNAVAILABLE
 *      ADK_E_MEM_UNAVAILABLE
 *
 *
 * @brief  Configures a PH Dev device
 *
 * The device either transmits and/or receives PDUs. The DP and CP data
 * structures needed are allocated and initialized. If the device is only
 * used for Tx, no more configuration is needed. For Rx, the queue and
 * callback function(s) need to be configured. CPPI structures are allocated
 * as needed. The API will return the device ID, so the application doesn't
 * have to manage them.
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_configure(
    const adk_ph_dev_device_config_t   *config_p,
    uint16_t                           *deviceId_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    int                     ret;
    adk_sys_handle_t        sysHandle;
    uint16_t                devId;
    uint16_t                devId_lessBaseId;
    uint16_t                freeRingSize = ADK_PH_DEV_CPPI_FREE_RING_SIZE;
    struct rte_ring        *ring_p;
    adk_pal_worker_info_t   palWorkerInfo = {0};
    bool                    lockActive = false;

#if RDK_REL_MAJOR >= 13
    struct rte_cppi_queue_config     cppiQueueConfig;
#else
    struct rte_bridge_cppi_queue_config     cppiQueueConfig;
#endif
    adk_ph_dev_device_internal_info_t      *cpDev_p;
    adk_ph_dev_dp_device_info_t            *dpDev_p;

    /*
     * Check to see if the shared memory segments have been configured
     */
    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    if ((NULL == config_p) || (NULL == deviceId_p))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "config_p or deviceId_p is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail;
    }

    if (phShm_p->numDevices >= ADK_PH_DEV_MAX_NUM_DEVICES)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "No devices available\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
        goto _parameter_validation_fail;
    }

    if (ADK_PH_DEV_MAX_NUM_QUEUES <= config_p->numRxQueues)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Number of Rx %d queues exceeds maximum %d\n",
                      config_p->numRxQueues,
                      ADK_PH_DEV_MAX_NUM_QUEUES);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail;
    }

    if ((ADK_PH_DEV_TYPE_EVENTDEV == config_p->type) &&
        (1 < config_p->numRxQueues))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Eventdev can only have 1 Rx queue; maxNumRxQueues: %d\n",
                      config_p->numRxQueues);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail;
    }

    cpDev_p = ADK_NVM_CALLOC(ADK_MISC_MODULE_ID, 0,
			 1, sizeof(adk_ph_dev_device_internal_info_t));
    if (NULL == cpDev_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Couldn't alloc memory for DP device\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _runtime_failure_1;
    }

    dpDev_p = ADK_DP_SHM_CALLOC(ADK_MISC_MODULE_ID, 3, 1,
                                sizeof(adk_ph_dev_dp_device_info_t),
                                ADK_PH_DP_PTR_ALIGN);
    if (NULL == dpDev_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Couldn't alloc memory for DP device\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _runtime_failure_2;
    }

    ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);

    lockActive = true;

    devId = adk_ph_dev_next_dev_get(phShm_p);
    if ((devId < ADK_PH_DEV_DEVICE_ID_BASE) ||
        (devId >= ADK_PH_DEV_DEVICE_ID_MAX))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "No devices available - device allocation\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
        goto _runtime_failure_3;
    }

    /*
     * Ensure the cookie is cleared. Set the devices pointer as this is used
     * in adk_ph_dev_next_dev_get() call to alloc a new device
     */
    cpDev_p->cookie = 0;
    phShm_p->devices[devId] = cpDev_p;

    devId_lessBaseId = devId - ADK_PH_DEV_DEVICE_ID_BASE;

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_device_get", "Get the CPPI device\n");
    retValue = adk_ph_dev_cppi_device_get(sysHandle,
                                          phShm_p->cppiInstanceId,
                                          devId_lessBaseId);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Unable to create CPPI device %d for devId %d\n",
                      devId_lessBaseId, devId);
        goto _runtime_failure_4;
    }

    /*
     * Alloc and init the PH DP lock
     */
    cpDev_p->lock_p = (void *)ADK_ALLOC_LOCK();
    if (cpDev_p->lock_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH lock alloc failed\n");
        goto _runtime_failure_4a;
    }

    retValue = ADK_INIT_LOCK(cpDev_p->lock_p,
                             ADK_PH_DEV_LOCK_SHARED,
                             ADK_PH_DEV_LOCK_VALUE);
    if (retValue != ADK_SUCCESS)
    {
        ADK_FORCE_LOG_ERROR("PH couldn't init lock\n");
        goto _runtime_failure_5;
    }

    /*
     * Create a ring for CPPI free queue
     */
    if (0 < config_p->numRxQueues)
    {
        /*
         * If the value is 0, we will use the default ADK value
         * (ADK_PH_DEV_CPPI_FREE_RING_SIZE), otherwise use the value passed
         * by the application.
         */
        if (0 != config_p->freeRingSize)
        {
            freeRingSize = config_p->freeRingSize;
        }

        if ( ! POWER_OF_2(freeRingSize))
        {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Invalid free ring size (%d) for devId (%d) must be "
                          "power of 2\n", freeRingSize, devId);
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto _runtime_failure_6;
        }

        retValue = adk_ph_dev_ring_create(sysHandle, devId,
                                          ADK_PH_CPPI_FREE_QUEUE_ID, 'F',
                                          freeRingSize, &ring_p);

        if (retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Unable to create CPPI free ring\n");
            goto _runtime_failure_6;
        }

        cpDev_p->freeRing_p = ring_p;

        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                     "[%s] Free Ring '%s' created: (%d:%d:%d)\n", __func__,
                     ring_p->name, phShm_p->cppiInstanceId, devId,
                     ADK_PH_CPPI_FREE_QUEUE_ID);
    }

    switch (config_p->type)
    {
        case ADK_PH_DEV_TYPE_EVENTDEV:
            dpDev_p->u.evd.portId = config_p->u.evd.portId;
            dpDev_p->palId = devId_lessBaseId + ADK_MAX_NUM_CORES;

            cpDev_p->u.evd.portId = config_p->u.evd.portId;
            cpDev_p->palId = devId_lessBaseId + ADK_MAX_NUM_CORES;

            palWorkerInfo.evdev_port_id = config_p->u.evd.portId;
            palWorkerInfo.lcore_id = dpDev_p->palId;
            ret = adk_pal_worker_to_event_port_map_add(sysHandle,
                                                               &palWorkerInfo);
            if (0 != ret)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                              "Failed to register PAL port\n");
                goto _runtime_failure_7;
            }
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "device type (%d) is invalid\n", config_p->type);
            goto _runtime_failure_7;
            break;
    }

    /*
     * Set internal state
     */
    dpDev_p->isRxEnabled    = false;
    dpDev_p->isActive       = false;

    cpDev_p->isRxEnabled    = false;
    cpDev_p->isActive       = false;

    cpDev_p->devId          = devId;
    cpDev_p->cppiDeviceId   = devId_lessBaseId;

    /*
     * Copy user config
     */
    dpDev_p->noZeroCopy             = config_p->noZeroCopy;
    dpDev_p->congestionControlStop  = config_p->congestionControlStop;
    dpDev_p->interruptEnabled       = config_p->interruptEnabled;
    dpDev_p->controlFd              = config_p->controlFd;
    dpDev_p->type                   = config_p->type;

    cpDev_p->noZeroCopy             = config_p->noZeroCopy;
    cpDev_p->congestionControlStop  = config_p->congestionControlStop;
    cpDev_p->interruptEnabled       = config_p->interruptEnabled;
    cpDev_p->controlFd              = config_p->controlFd;
    cpDev_p->type                   = config_p->type;
    cpDev_p->numRxQueues            = config_p->numRxQueues;
    cpDev_p->freeRingSize           = config_p->freeRingSize;

    cpDev_p->coalEnabled            = config_p->coalesceEnabled;
    cpDev_p->coalBufSize            = config_p->coalesceBufSize;
    if (true == cpDev_p->coalEnabled)
    {
        /*
         * If the buffer size is 0, we will use the ADK default size
         */
        if (0 == cpDev_p->coalBufSize)
        {
            cpDev_p->coalBufSize = ADK_PH_DEV_COALESCE_BUF_SIZE;
        }
        cpDev_p->coalBuf_p = (uint8_t *)ADK_RAM_CALLOC(1, cpDev_p->coalBufSize);
        if (NULL == cpDev_p->coalBuf_p)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Unable to alloc static coalesce buffer of size %d\n",
                          cpDev_p->coalBufSize);
            goto _runtime_failure_9;
        }
    }

    phShm_p->dp_p->devices[devId] = dpDev_p;
    phShm_p->devices[devId] = cpDev_p;

    phShm_p->numDevices++;

#ifdef ADK_QOS_ENABLE
    if (adk_qos_instance_init(sysHandle, devId) !=  ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                    "could not register PH dev with  ADK QoS \n");
        goto _runtime_failure_10;
    }
#endif

    dpDev_p->cookie = ADK_PH_DEV_DP_DEVICE_COOKIE;
    cpDev_p->cookie = ADK_PH_DEV_DEVICE_COOKIE;

    lockActive = false;
    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);

    /*
     * Create a ring for CPPI free queue
     */
    if (0 < config_p->numRxQueues)
    {
        cppiQueueConfig.queue_type = CPPI_RING_FQ;
        cppiQueueConfig.ring_p = cpDev_p->freeRing_p;
        cppiQueueConfig.mempool_p = NULL;
        cppiQueueConfig.mbuf_cache_sz = 0;
#if RDK_REL_MAJOR >= 13
        /* CPPI uses lock here */
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_cppi_free_queue_setup",
                        "setup CPPI Free Queue for devId_lessBaseId %d \n",
                        devId_lessBaseId);
        ret = rte_cppi_free_queue_setup(devId_lessBaseId,
                                        CPPI_FIFO_COUNT_MAX,
                                        &cppiQueueConfig);
        if (0 > ret)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Unable to setup CPPI queue %d:%d\n",
                          devId_lessBaseId, ADK_PH_CPPI_FREE_QUEUE_ID);
            goto _runtime_failure_10;
        }
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_bridge_cppi_free_queue_add",
                        "Add CPPI Free Queue for devId_lessBaseId %d \n",
                        devId_lessBaseId);
        ret = rte_bridge_cppi_free_queue_add(phShm_p->cppiInstanceId,
                                        devId_lessBaseId);
        if (0 > ret)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Unable to add CPPI queue %d:%d:%d\n",
                          phShm_p->cppiInstanceId, devId_lessBaseId,
                          ADK_PH_CPPI_FREE_QUEUE_ID);
            goto _runtime_failure_10;
        }
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                     "[%s] Free Queue %d:%d:%d added\n", __func__,
                     phShm_p->cppiInstanceId, devId_lessBaseId, ADK_PH_CPPI_FREE_QUEUE_ID);
#else
        ADK_LOG_EXT_API(sysHandle,
            ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "rte_bridge_cppi_queue_add",
            "setup CPPI Free Queue for devId_lessBaseId %d \n",
            devId_lessBaseId);
        ret = rte_bridge_cppi_queue_add(phShm_p->cppiInstanceId,
                                        devId_lessBaseId,
                                        ADK_PH_CPPI_FREE_QUEUE_ID,
                                        &cppiQueueConfig);
        if (0 > ret)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Unable to add CPPI queue %d:%d:%d\n",
                    phShm_p->cppiInstanceId, devId_lessBaseId,
                    ADK_PH_CPPI_FREE_QUEUE_ID);
            goto _runtime_failure_10;
        }
#endif
    }

    *deviceId_p = devId;

    goto _success_exit;

_runtime_failure_10:
    if (true == cpDev_p->coalEnabled)
    {
        ADK_RAM_FREE(cpDev_p->coalBuf_p);
    }

_runtime_failure_9:
    if (config_p->type == ADK_PH_DEV_TYPE_EVENTDEV)
    {
        ret = adk_pal_worker_to_event_port_map_rem(sysHandle, &palWorkerInfo);
        if (0 != ret)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Failed to remove event port map\n");
        }
    }


_runtime_failure_7:
     adk_ph_dev_ring_destroy(sysHandle, cpDev_p->freeRing_p);

_runtime_failure_6:
     ADK_DESTROY_LOCK(cpDev_p->lock_p);

_runtime_failure_5:
     ADK_FREE_LOCK(cpDev_p->lock_p);

_runtime_failure_4a:
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_device_free", "Free CPPI device");
    adk_ph_dev_cppi_device_free(sysHandle, phShm_p->cppiInstanceId,
                                devId_lessBaseId);

_runtime_failure_4:
    phShm_p->devices[devId] = NULL;

_runtime_failure_3:
_lock_acquire_failed:
    if (lockActive == true)
    {
        ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);
    }
    ADK_DP_SHM_FREE(dpDev_p);

_runtime_failure_2:
    ADK_NVM_FREE(cpDev_p);

_runtime_failure_1:
_success_exit:

_lock_release_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}
/*!
 * =============================================================
 * adk_ph_dev_device_start   {{{2
 *
 * @param[in] - devId  ID for the device being started
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Start the device
 *
 * Once started, no further configuration can happen.
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_start(uint16_t devId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    adk_ph_dev_dp_device_info_t        *dpDev_p;

    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DP_DEVICE_VALIDATE();

    dpDev_p->isRxEnabled = true;
    dpDev_p->isActive = true;

    cpDev_p->isRxEnabled = true;
    cpDev_p->isActive = true;

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}
/*!
 * =============================================================
 * adk_ph_dev_device_stop   {{{2
 *
 * @param[in] - devId  ID for the device being started
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Stop the device
 *
 * A stopped device can be restarted
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_stop(uint16_t devId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    adk_ph_dev_dp_device_info_t        *dpDev_p;

    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DP_DEVICE_VALIDATE();

    cpDev_p->isActive = false;
    cpDev_p->isRxEnabled = false;

    dpDev_p->isActive = false;
    dpDev_p->isRxEnabled = false;

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}
/*!
 * =============================================================
 * adk_ph_dev_device_close   {{{2
 *
 * @param[in] - devId  ID for the device being started
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 * @retval ADK_E_INTERNAL_ERROR
 *
 *
 * @brief  Close the device
 *
 * An error will be returned if the device is active.
 *
 * The API will free the CPPI and memory resources.
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_close(uint16_t devId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    int                                 ret;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    adk_ph_dev_dp_device_info_t        *dpDev_p;
    int                                 loopCount;
    uint16_t                            devId_lessBaseId;

    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DP_DEVICE_VALIDATE();

    /* If the device isn't stopped, we can't close it */
    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(true);

    /* Set the pointer to NULL, so that the DP won't grab it again */
    devId_lessBaseId = devId - ADK_PH_DEV_DEVICE_ID_BASE;
    phShm_p->dp_p->devices[devId] = NULL;
    phShm_p->devices[devId] = NULL;
    phShm_p->numDevices--;

#ifdef ADK_QOS_ENABLE
    if (adk_qos_instance_shutdown(sysHandle, devId) !=  ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                    "could not un-register PH dev with  ADK QoS \n");
        ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);
        goto _runtime_failure_1;
    }
#endif

    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);

    /*
     * Switch to the CP dev lock since the PH Dev has already been taken care
     * of.
     */
    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    /*
     * If the device isn't stopped, we can't close it.
     *
     * Check again in case it has been restarted
     */
    if (cpDev_p->isActive ==  true)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "devId (%d) has been reactivated\n", devId);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);

        goto _runtime_failure_1;
    }

    dpDev_p->cookie = 0;
    cpDev_p->cookie = 0;

    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);
    ADK_DESTROY_LOCK(cpDev_p->lock_p);
    ADK_FREE_LOCK(cpDev_p->lock_p);

    /*
     * Free the SHM here so that if the DP were using it, it should be done by
     * now.
     */
    loopCount = ADK_PH_DEV_SLEEP_LOOPS;
    while ((dpDev_p->inUse == true) && (loopCount-- > 0))
    {
        usleep(ADK_PH_DEV_SLEEP_TIME);
    }

    if (dpDev_p->inUse == true)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "DP dev still in use after 1 millisec: devId:%d\n",
                      devId);
        goto _runtime_failure_2;
    }

    ADK_DP_SHM_FREE(dpDev_p);
    dpDev_p = NULL;

    if (0 < cpDev_p->numRxQueues)
    {
        /* CPPI uses lock here */
#if RDK_REL_MAJOR >= 13
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_bridge_cppi_free_queue_rem",
                        "Remove CPPI Free Queue.\n");
        ret = rte_bridge_cppi_free_queue_rem(phShm_p->cppiInstanceId,
                                        devId_lessBaseId);
        if (0 != ret)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "CPPI free queue remove failed: devId:%d '%s' (%d)\n",
                        devId_lessBaseId, rte_strerror(-ret), ret);
            goto _runtime_failure_2;
        }
#else
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_bridge_cppi_queue_rem",
                        "Remove CPPI Free Queue.\n");
        ret = rte_bridge_cppi_queue_rem(phShm_p->cppiInstanceId,
                                        devId_lessBaseId,
                                        ADK_PH_CPPI_FREE_QUEUE_ID);
        if (0 != ret)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "CPPI free queue remove failed: devId:%d '%s' (%d)\n",
                        devId_lessBaseId, rte_strerror(-ret), ret);
            goto _runtime_failure_2;
        }
#endif
        adk_ph_dev_ring_destroy(sysHandle, cpDev_p->freeRing_p);
    }

    /* Free the CP dev here after we destroy the cppi rings */
    ADK_NVM_FREE(cpDev_p);
    cpDev_p = NULL;

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_device_free", "Free CPPI device");
    adk_ph_dev_cppi_device_free(sysHandle, phShm_p->cppiInstanceId,
                                devId_lessBaseId);

    goto _success_exit;

_runtime_failure_1:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);
    goto _success_exit;

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);
    goto _success_exit;

_runtime_failure_2:
_success_exit:
_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}
/*!
 * =============================================================
 * adk_ph_dev_device_info_get   {{{2
 *
 * @param[in]   devId  : Device ID
 * @param[out]  info_p : Pointer to the configuration structure filled in
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Get the device configuration information
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_info_get(
    uint16_t                    devId,
    adk_ph_dev_device_info_t   *info_p)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(false);

    info_p->isActive         = cpDev_p->isActive;
    info_p->isRxEnabled      = cpDev_p->isRxEnabled;

    info_p->noZeroCopy              = cpDev_p->noZeroCopy;
    info_p->congestionControlStop   = cpDev_p->congestionControlStop;
    info_p->interruptEnabled        = cpDev_p->interruptEnabled;
    info_p->controlFd               = cpDev_p->controlFd;
    info_p->numRxQueues             = cpDev_p->numRxQueues;

    info_p->type = cpDev_p->type;
    switch (cpDev_p->type)
    {
        case ADK_PH_DEV_TYPE_EVENTDEV:
            info_p->u.evd.portId = cpDev_p->u.evd.portId;
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "device type (%d) is invalid\n", cpDev_p->type);
            break;
    }

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_device_stats_get   {{{2
 *
 * @param[in]   devId  : Device ID
 * @param[otu]  stats  : Pointer to the stats structure returned by the API
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_INCONSISTENT_STATE
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Read the statistics for the device
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_stats_get(
    uint16_t                        devId,
    adk_ph_dev_device_statistics_t *stats)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    int                                 ret;
    adk_sys_handle_t                    sysHandle;
    struct rte_cppi_queue_stats         cppiStats;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    adk_ph_dev_dp_device_info_t        *dpDev_p;

    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DP_DEVICE_VALIDATE();

    ADK_PH_DEV_CP_DP_STATE_VALIDATE();

    if (NULL == stats)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "stats is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    ADK_MEMSET(stats, 0, sizeof(adk_ph_dev_device_statistics_t));

    if (cpDev_p->numRxQueues > 0)
    {
        ADK_MEMSET(&cppiStats, 0, sizeof(struct rte_cppi_queue_stats));
#if RDK_REL_MAJOR >= 13
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_cppi_queue_get_stats",
                        "Get statistics for free CPPI Queue.\n");
        ret = rte_cppi_queue_stats_get( cpDev_p->cppiDeviceId, 0,
                                        CPPI_RING_FQ,
                                        &cppiStats);
#else
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_bridge_cppi_get_stats",
                        "Get statistics for free CPPI Queue.\n");
        ret = rte_bridge_cppi_get_stats(phShm_p->cppiInstanceId,
                                        cpDev_p->cppiDeviceId,
                                        ADK_PH_CPPI_FREE_QUEUE_ID,
                                        &cppiStats);
#endif

        if (ret < 0)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "CPPI Get for Free Queue failed: devId:%d '%s' (%d)\n",
                          cpDev_p->cppiDeviceId, rte_strerror(-ret), ret);
            goto _runtime_failure_1;
        }
        stats->freeQ_jobsEnqueued = cppiStats.nb_requested;
        stats->freeQ_jobsProcessed = cppiStats.nb_processed;
        stats->rateLimitingDrops = dpDev_p->devStats.rateLimitingDrops;
    }

_parameter_validation_fail_need_unlock:
_runtime_failure_1:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_device_stats_print   {{{2
 *
 * @param[in] devId
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Prints the stats for the given device
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_device_stats_print(
    uint16_t        devId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_statistics_t      stats;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_MEMSET(&stats, 0, sizeof(adk_ph_dev_device_statistics_t));
    retValue = adk_ph_dev_device_stats_get(devId, &stats);
    if (ADK_SUCCESS != retValue)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "[%s] get failed\n", __func__);
        goto _runtime_failure_1;
    }

    ADK_INFO_WRITE("\nPH Device %d Stats"
                   " (Drop counters with zero count not shown):\n", devId);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tFree Queue Enqueued:     \t%10"PRIu64"\n",
                               stats.freeQ_jobsEnqueued);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tFree Queue Processed:    \t%10"PRIu64"\n",
                               stats.freeQ_jobsProcessed);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tRate limiting drops:    \t%10"PRIu64"\n",
                               stats.rateLimitingDrops);
    ADK_INFO_WRITE("\n");

_runtime_failure_1:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_callback_fn_add
 *
 * @param[in] cbFn : Pointer to the function that will be called
 * @param[in] userParam : A transparent parameter passed to the callback
 *                        function
 * @param[out] cbFnHandle_p : A pointer where the handle for this callback
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_INTERNAL_ERROR
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Add a callback function for the device's reception of a PDU
 *
 * The API can be called multiple times and it will return unique 6 bits
 * value for each call as cbFnHandle_p
 *
 * Maximum number of call back function adding supported is 64.
 *
 * =============================================================
 */

adk_status_t
adk_ph_dev_callback_fn_add(
    adk_ph_dev_callback_fn       cbFn,
    void                        *userParam,
    adk_ph_dev_cbfn_handle_t    *cbFnHandle_p)
{

    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    bool flag_cbfn_set;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;
    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);
    if ((NULL == cbFn) || (NULL == cbFnHandle_p))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Callback function pointer or handle pointer is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    for(int i = 0 ; i < ADK_PH_CBFN_MAX; i++)
    {
        if(NULL == phShm_p->cbFnIdMap[i].cbFn)
        {
            phShm_p->cbFnIdMap[i].cbFn = cbFn;
            phShm_p->cbFnIdMap[i].cbParams = userParam;
            flag_cbfn_set = true;
            *cbFnHandle_p = i;
            break;
        }
    }

    if(!flag_cbfn_set)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Could not store more than 64 Callback functions\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);

    }

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_callback_fn_rem
 *
 * @param[in] cbFnHandle_p returned by adk_ph_dev_callback_add()
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Remove a callback function

* =============================================================
 */
adk_status_t
adk_ph_dev_callback_fn_rem(
        adk_ph_dev_cbfn_handle_t    cbFnHandle_p)
{

    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;
    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);

    if (cbFnHandle_p > ADK_PH_CBFN_MAX)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Callback function handle deosn't valid\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    if(NULL == phShm_p->cbFnIdMap[cbFnHandle_p].cbFn)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Callback function is not registed for cbFnHandle\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }
    else
    {
        phShm_p->cbFnIdMap[cbFnHandle_p].cbFn = NULL;
        phShm_p->cbFnIdMap[cbFnHandle_p].cbParams = NULL;
    }

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

adk_status_t
adk_ph_dev_rx_parse_register(
    adk_ph_dev_rx_parse_fn     parseFn,
    uint8_t                    moduleId)
{

    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;
    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);
    if ((NULL == parseFn) || (moduleId > ADK_PH_MODULE_ID_MAX))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Unable to register the parse function\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    phShm_p->parseFnMap[moduleId].parseFn = parseFn;
    adk_module_handle_get(sysHandle, moduleId, NULL,
                            &phShm_p->parseFnMap[moduleId].moduleHandle);
    
_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

adk_status_t
adk_ph_dev_rx_parse_unregister(
    uint8_t                    moduleId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;
    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);
    if ((moduleId > ADK_PH_MODULE_ID_MAX) || (NULL == phShm_p->parseFnMap[moduleId].parseFn))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Unable to register the parse function\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    phShm_p->parseFnMap[moduleId].parseFn = NULL;
    phShm_p->parseFnMap[moduleId].moduleHandle = NULL;

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_rx_queue_setup   {{{2
 *
 * @param[in] devId: The PH device ID where the queue is being added
 * @param[in] queueId: The queue Id being added
 * @param[in] queueConf_p: Queue configuration data
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_INCONSISTENT_STATE
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Creates a queue for sending PDUs from the DP to CP
 *
 * For Eventdev, this will add the Eventdev queue. The Queue would be
 * used when enqueuing a PDU to the control plane.
 * For Rings, the ring name is used to lookup the ring buffer
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_rx_queue_setup(
    uint16_t                        devId,
    uint16_t                        queueId,
    adk_ph_dev_rx_queue_config_t   *queueConf_p)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    adk_ph_dev_dp_device_info_t        *dpDev_p;
    adk_ph_dev_dp_queue_info_t         *dpQueue_p;
    adk_ph_dev_queue_info_t            *cpQueue_p;

    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                 "[%s] devId: %d, queueId: %d, queueConf_p: %p\n",
                 __func__, devId, queueId, queueConf_p);

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DP_DEVICE_VALIDATE();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(true);

    ADK_PH_DEV_CP_DP_STATE_VALIDATE();

    if (queueId >= phShm_p->devices[devId]->numRxQueues)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid Rx queueId (%d) for devId (%d) max is %d\n",
                      queueId, devId, phShm_p->devices[devId]->numRxQueues);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    if (NULL != phShm_p->devices[devId]->rxQueues[queueId])
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Rx queueId %d exists for device\n", queueId);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    if (NULL == queueConf_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "queueConf_p is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _parameter_validation_fail_need_unlock;
    }

    dpQueue_p = (adk_ph_dev_dp_queue_info_t *)ADK_DP_SHM_CALLOC(ADK_MISC_MODULE_ID, 3, 1,
                                        sizeof(adk_ph_dev_dp_queue_info_t),
                                        ADK_PH_DP_PTR_ALIGN);
    if (NULL == dpQueue_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Couldn't alloc memory for DP queue\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _runtime_failure_1;
    }

    cpQueue_p = (adk_ph_dev_queue_info_t *)ADK_NVM_CALLOC(ADK_MISC_MODULE_ID, 0, 1,
                                        sizeof(adk_ph_dev_queue_info_t));
    if (NULL == cpQueue_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Couldn't alloc memory for CP queue\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _runtime_failure_2;
    }

    switch (cpDev_p->type)
    {
        case ADK_PH_DEV_TYPE_EVENTDEV:
            dpQueue_p->u.evd.queueId = queueConf_p->u.evd.queueId;
            dpQueue_p->u.evd.devId = devId;
            cpQueue_p->u.evd.queueId = queueConf_p->u.evd.queueId;
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "device type (%d) is invalid\n", cpDev_p->type);
            goto _runtime_failure_3;
            break;
    }

    cpDev_p->rxQueues[queueId] = cpQueue_p;
    dpDev_p->rxQueues[queueId] = dpQueue_p;

    goto _success_exit;


_runtime_failure_3:
    ADK_NVM_FREE(cpQueue_p);

_runtime_failure_2:
    ADK_DP_SHM_FREE(dpQueue_p);

_parameter_validation_fail_need_unlock:
_runtime_failure_1:
_success_exit:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}


/*!
 * =============================================================
 * adk_ph_dev_rx_queue_remove   {{{2
 *
 * @param[in] devId: The PH device ID where the queue is being added
 * @param[in] queueId: The queue Id being removed
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Remove a queue for sending PDUs from the DP to CP
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_rx_queue_remove(
    uint16_t                        devId,
    uint16_t                        queueId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    void                               *queue_p;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                 "[%s] devId: %d, queueId: %d\n", __func__, devId, queueId);

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(true);

    if (queueId >= phShm_p->devices[devId]->numRxQueues)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid Rx queueId (%d) for devId (%d) max is %d\n",
                      queueId, devId, phShm_p->devices[devId]->numRxQueues);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _parameter_validation_fail_need_unlock;
    }

    if (NULL == cpDev_p->rxQueues[queueId])
    {
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                     "Rx queueId %d already removed\n", queueId);
        goto _parameter_validation_fail_need_unlock;
    }

    queue_p = cpDev_p->rxQueues[queueId];
    cpDev_p->rxQueues[queueId] = NULL;
    ADK_NVM_FREE(queue_p);


    queue_p = phShm_p->devices[devId]->rxQueues[queueId];
    phShm_p->devices[devId]->rxQueues[queueId] = NULL;
    ADK_DP_SHM_FREE(queue_p);

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

    goto _success_exit;


_success_exit:
_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}


/*!
 * =============================================================
 * adk_ph_dev_rx_queue_stats_get   {{{2
 *
 * @param[in] devId : Device ID to read
 * @param[in] queueId : Queue ID to read
 * @param[in] stats_p : Pointer to a struct where the stats are stored
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_INCONSISTENT_STATE
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Read statistics for a given device's input queue
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_rx_queue_stats_get(
    uint16_t                            devId,
    uint16_t                            queueId,
    adk_ph_dev_rx_queue_statistics_t   *stats_p)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    adk_ph_dev_dp_device_info_t        *dpDev_p;
    adk_ph_dev_dp_queue_stats_t         dpStats;

    ADK_PH_DEV_SHM_VALIDATE();

    ADK_PH_DEV_DP_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DP_DEVICE_VALIDATE();

    ADK_PH_DEV_RX_QUEUE_VALIDATE();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(false);

    ADK_PH_DEV_DP_RX_QUEUE_VALIDATE();

    ADK_PH_DEV_CP_DP_STATE_VALIDATE();

    if (NULL == stats_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "stats_p is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _parameter_validation_fail_need_unlock;
    }

    retValue = adk_ph_dev_dp_rx_queue_stats_get(dpDev_p->rxQueues[queueId],
                                                &dpStats);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH Dev DP stats get failed\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _runtime_failure_1;
    }

    stats_p->pdusEnqueued = dpStats.rx.pdusEnqueued;
    stats_p->pdusDequeued = cpDev_p->rxQueues[queueId]->stats.rx.pdusDequeued;
/*    stats_p->pdusEnqDropped = dpStats.rx.pdusEnqDropped; */
    stats_p->pdusDeqDropped = cpDev_p->rxQueues[queueId]->stats.rx.pdusDeqDropped;
    stats_p->rxDisabled = dpStats.rx.rxDisabled;
    stats_p->bufferHighWatermark = dpStats.rx.bufferHighWatermark;
    stats_p->bufferFull = dpStats.rx.bufferFull;
    stats_p->congestStop = dpStats.rx.congestStop;
    stats_p->intrSent = dpStats.rx.intrSent;
    stats_p->intrSendFailed = dpStats.rx.intrSendFailed;
    /* stats_p->pdusDroppedQosRed = dpStats.rx.pdusDroppedQosRed; */
    stats_p->unknownType = dpStats.rx.unknownType;
    stats_p->palNextStageFailed = dpStats.rx.palNextStageFailed;


_parameter_validation_fail_need_unlock:
_runtime_failure_1:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_rx_queue_stats_print   {{{2
 *
 * @param[in] devId : Device ID to read
 * @param[in] queueId : Queue ID to read
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Prints the statistics for a given device and queue
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_rx_queue_stats_print(
    uint16_t                    devId,
    uint16_t                    queueId)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_rx_queue_statistics_t    stats;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_MEMSET(&stats, 0, sizeof(adk_ph_dev_rx_queue_statistics_t));
    retValue = adk_ph_dev_rx_queue_stats_get(devId, queueId, &stats);
    if (ADK_SUCCESS != retValue)
    {
        goto _runtime_failure_1;
    }

    ADK_INFO_WRITE("PH Device %d Rx Queue %d Stats"
                   " (Drop counters with zero count not shown):\n",
                   devId, queueId);

    ADK_INFO_WRITE("\tPDUs Received:                   \t%10"PRIu64"\n",
                                stats.pdusDequeued);
    ADK_INFO_WRITE("\tPDUs DP Enqueued:                \t%10"PRIu64"\n",
                                stats.pdusEnqueued);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tPDUs Dropped at enqueue:         \t%10"PRIu64"\n",
                                stats.pdusEnqDropped);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tPDUs Dropped at receive:         \t%10"PRIu64"\n",
                                stats.pdusDeqDropped);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tPDUs sent while device disabled: \t%10"PRIu64"\n",
                                stats.rxDisabled);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of high watermark hits:   \t%10"PRIu64"\n",
                                stats.bufferHighWatermark);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of buffer full hit:       \t%10"PRIu64"\n",
                                stats.bufferFull);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of congestion stop hit:   \t%10"PRIu64"\n",
                                stats.congestStop);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of interrupts sent:       \t%10"PRIu64"\n",
                                stats.intrSent);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of interrupts send fails: \t%10"PRIu64"\n",
                                stats.intrSendFailed);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of PDUs dropped PAL Next Stage Failed:  \t%10"PRIu64"\n",
                                stats.palNextStageFailed);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of PDUs dropped RED QoS:  \t%10"PRIu64"\n",
                                stats.pdusDroppedQosRed);
    ADK_SYS_NON_ZERO_CNTR_WRITE("\tNumber of PDUs dropped Unknown type:  \t%10"PRIu64"\n",
                                stats.unknownType);
    ADK_INFO_WRITE("\n");

_runtime_failure_1:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_rx_queue_info_get   {{{2
 *
 * @param[in] devId : Device ID to read
 * @param[in] queueId : Queue ID to read
 * @param[out] queue_p : Pointer to the struct for the queue information
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Read the queue configuration information
 *
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_rx_queue_info_get(
    uint16_t                    devId,
    uint16_t                    queueId,
    adk_ph_dev_rx_queue_info_t *queue_p)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_RX_QUEUE_VALIDATE();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(false);

    if (NULL == queue_p)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "queue_p is NULL\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto _parameter_validation_fail_need_unlock;
    }

    switch (cpDev_p->type)
    {
        case ADK_PH_DEV_TYPE_EVENTDEV:
            queue_p->u.evd.queueId = cpDev_p->rxQueues[queueId]->u.evd.queueId;
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "device type (%d) is invalid\n", cpDev_p->type);
            break;
    }

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}

/*!
 * =============================================================
 * adk_ph_dev_rx_burst   {{{2
 *
 * @param[in] devId : Device ID to read
 * @param[in] queueId : Queue ID to read; Current not use; should be set to 0
 * @param[in] rxPkts : A pointer to an array of mbuf pointers where the mbufs
 *                     will be stored
 * @param[in] maxNumPkts : The maximum burst size to read
 *
 *
 * @retval Number of PDUs read
 * @retval -ADK_E_SYS_NOT_INITIALIZED
 * @retval -ADK_E_INVALID_PARAMETER
 * @retval -ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Read a burst of PDUs
 *
 * The return value is overloaded. If it is < 0 then an error has happened.
 * The API does not block. For each PDU received, the callback function list
 * is walked. If configured, fragmented mbufs will be coalesced.
 *
 * =============================================================
 */
uint16_t
adk_ph_dev_rx_burst(
    uint16_t            devId,
    uint16_t            queueId,
    struct rte_mbuf   **rxPkts,
    const uint16_t      maxNumPkts)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    int32_t                             numPdus = 0;
    int32_t                             numBytes;
    uint8_t                             cbFnIds[maxNumPkts];
    int                                 i;
    uint8_t                            *buf_p;
    adk_ph_dev_cb_fn_id_map_t          *cbFnIdMap;
    struct rte_mbuf *mbuf_p;
    adk_ph_dev_dp_metadata_t *phMeta_p;
    adk_metadata_t *adk_md_ptr;
    void *dp_Message = NULL;
    adk_module_handle_t module_handle = NULL;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

#ifdef ADK_PH_DEV_BURST_DEBUG
    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));
#endif

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_RX_QUEUE_VALIDATE();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(false);

    if (ADK_PH_DEV_TYPE_EVENTDEV == cpDev_p->type)
    {
        numPdus = adk_pal_dequeue_burst(cpDev_p->palId,
                                        rxPkts,
                                        cbFnIds,
                                        maxNumPkts,
                                        true);
        if (numPdus > 0)
        {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                         "[%s] devId: %d, dequeued %d/%d Eventdev PDUs\n",
                         __func__, devId, numPdus, maxNumPkts);
        }

        if (numPdus < 0)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                           "adk_pal_dequeue_burst failed: '%s', %d, %d\n",
                           rte_strerror(-rte_errno), rte_errno, numPdus);
            goto _runtime_failure_1;
        }

    }
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                       "unknown device type: %d\n",
                       cpDev_p->type);
        goto _runtime_failure_1;
    }

    /*
     * If callbacks are enabled, then walk the PDUs received, calling the
     * callback functions
     */
    for (i = 0; i < numPdus; i++)
    {
        buf_p = rte_pktmbuf_mtod(rxPkts[i], void *);
        /*
         * If coalesce is enabled and we have a chained mbuf, flatten it
         * before calling the callback
         */
        if ((true == cpDev_p->coalEnabled) &&
            (false == rte_pktmbuf_is_contiguous(rxPkts[i])))
        {
            buf_p = cpDev_p->coalBuf_p;
            numBytes = adk_ph_dev_coalesce_pdu(rxPkts[i], buf_p,
                                               cpDev_p->coalBufSize);
            if (0 == numBytes)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle,
                              ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                              "Failed to coalesce PDU. buffer size: %d, "
                              "packet size: %d. Callback not called\n",
                              cpDev_p->coalBufSize,
                              rte_pktmbuf_pkt_len(rxPkts[i]));
                continue;
            }
        }
        mbuf_p = (struct rte_mbuf *)rxPkts[i];
        adk_md_ptr = ADK_METADATA_PTR(mbuf_p);
        phMeta_p = (adk_ph_dev_dp_metadata_t *)&(adk_md_ptr->imc_param_start);
#ifdef ADK_DEUG
        debug_printf("phMeta_p->moduleId %d\n",phMeta_p->moduleId);
#endif
        if(phMeta_p->moduleId <= ADK_PH_MODULE_ID_MAX &&
            NULL != phShm_p->parseFnMap[phMeta_p->moduleId].parseFn)
        {
            ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);
            //adk_module_handle_get(sysHandle, phMeta_p->moduleId, NULL, &module_handle);
            module_handle = phShm_p->parseFnMap[phMeta_p->moduleId].moduleHandle;
            phShm_p->parseFnMap[phMeta_p->moduleId].parseFn(rxPkts[i], &dp_Message, module_handle);
            ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);
        }

        if(cbFnIds[i] < ADK_PH_CBFN_MAX)
        { 
            cbFnIdMap = &phShm_p->cbFnIdMap[cbFnIds[i]];

            if(NULL != cbFnIdMap->cbFn)
            {
                ADK_ACQUIRE_LOCK(phShm_p->lock_p, retValue);
                ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "PH_rxCbFn_p",
                        "devId: %d, qId: %d, i: %d\n",
                        devId, queueId, i);

                ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "[%s] devId: %d, queue %d ,PH Event[%d] Priority %d \n",
                        __func__, devId, queueId, i, phMeta_p->priority);

                cbFnIdMap->cbFn(phMeta_p->moduleId,dp_Message,cbFnIdMap->cbParams);
                ADK_RELEASE_LOCK(phShm_p->lock_p, retValue);
            }
        }
        /* recheck to ensure device hasn't been removed */
        if ((NULL == cpDev_p) ||
            (ADK_PH_DEV_DEVICE_COOKIE != cpDev_p->cookie))
        {
            ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);
            ADK_LOG_ERROR(sysHandle,
                          ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Device devId (%d) has been removed\n",
                          devId);
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            goto _parameter_validation_fail;
        }
    }

    cpDev_p->rxQueues[queueId]->stats.rx.pdusDequeued += numPdus;

    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

    goto _success_exit;

_runtime_failure_1:
    phShm_p->devices[devId]->rxQueues[queueId]->stats.rx.pdusDeqDropped += (maxNumPkts - numPdus);

_parameter_validation_fail_need_unlock:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    numPdus = -retValue;

_success_exit:
#ifdef ADK_PH_DEV_BURST_DEBUG
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));
#endif

_return:
    return (uint16_t)numPdus;
}

/*!
 * =============================================================
 * adk_ph_dev_buffer_free   {{{2
 *
 * @param[in] devId : Device ID to read
 * @param[in] mbuf_p : A pointer to an mbuf to free
 *
 *
 * @retval ADK_SUCCESS
 * @retval ADK_E_SYS_NOT_INITIALIZED
 * @retval ADK_E_INVALID_PARAMETER
 * @retval ADK_E_LOCK_FAILURE
 *
 *
 * @brief  Pointer to an mbuf to free
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_buffer_free(
    uint16_t            devId,
    struct rte_mbuf    *mbuf_p)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sys_handle_t                    sysHandle;
    adk_ph_dev_device_internal_info_t  *cpDev_p;
    int                                 rteRetValue;

    ADK_PH_DEV_SHM_VALIDATE();

    sysHandle = phShm_p->sysHandle;

    ADK_LOG_FUNC_ENTRY(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

    ADK_PH_DEV_DEVICE_VALIDATE();

    ADK_ACQUIRE_LOCK(cpDev_p->lock_p, retValue);

    ADK_PH_DEV_RECHECK_CP_DEV_VALID();

    ADK_PH_DEV_DEVICE_ERROR_IF_ACTIVE_STATE_IS(false);

#if RDK_REL_MAJOR >= 13
    ADK_LOG_EXT_API(sysHandle,
        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
        "rte_cppi_mbuf_free", "free mbuf using CPPI device\n");
    rteRetValue = rte_cppi_mbuf_free(
                                        cpDev_p->cppiDeviceId,
                                        mbuf_p);
#else
    ADK_LOG_EXT_API(sysHandle,
        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
        "rte_bridge_cppi_mbuf_free", "free mbuf using CPPI device\n");
    rteRetValue = rte_bridge_cppi_mbuf_free(
                                        phShm_p->cppiInstanceId,
                                        cpDev_p->cppiDeviceId,
                                        ADK_PH_CPPI_FREE_QUEUE_ID,
                                        mbuf_p);
#endif
    if (rteRetValue < 0)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                       "CPPI mbuf free failed: '%s' %d\n",
                       rte_strerror(-rteRetValue), -rteRetValue);
        retValue = ADK_E_INTERNAL_ERROR;
        goto _runtime_failure_1;
    }

_parameter_validation_fail_need_unlock:
_runtime_failure_1:
    ADK_RELEASE_LOCK(cpDev_p->lock_p, retValue);

_lock_release_failed:
_lock_acquire_failed:
_parameter_validation_fail:
    ADK_LOG_FUNC_EXIT(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle));

_return:
    return retValue;
}


/*!
 *
 * Helper functions   {{{1
 */
/*!
 * =============================================================
 * adk_ph_dev_cb_fn_id_get   {{{2
 * @param[in]  mbuf_p    pointer to the mbuf
 *
 * @retval The callback function ID
 *
 * @brief  Helper function to retrieve the callback function ID of the PDU
 *
 * =============================================================
 */
uint16_t
adk_ph_dev_cb_fn_id_get(
    struct rte_mbuf *mbuf_p)
{
    if (NULL == mbuf_p)
    {
        return 0xffff;
    } else {
        return ADK_METADATA_PARAM((mbuf_p), func_idx);
    }
}

/*!
 * =============================================================
 * adk_ph_dev_pdu_offset   {{{2
 * @param[in]  mbuf_p    pointer to the mbuf
 *
 * @retval The offset to the PDU
 *
 * @brief  Helper function to retrieve the PDU offset
 *
 * =============================================================
 */
int
adk_ph_dev_pdu_offset(
    struct rte_mbuf *mbuf_p)
{
    RTE_SET_USED(mbuf_p);

    return 0;
}

/*!
 * =============================================================
 * adk_ph_dev_pdu_size   {{{2
 * @param[in]  mbuf_p    pointer to the mbuf
 *
 * @retval The size of the PDU
 *
 * @brief  Helper function to retrieve the PDU size
 *
 * =============================================================
 */
uint16_t
adk_ph_dev_pdu_size(
    struct rte_mbuf *mbuf_p)
{
    return mbuf_p->pkt_len;
}

/*!
 * =============================================================
 * adk_ph_dev_coalesce_pdu   {{{2
 * @param[in]  inMbuf_p   pointer to the input mbuf
 * @param[in]  outPdu_p   pointer to the memory area to copy to
 * @param[in]  outPduSize Size of the buffer PDU is being copied into
 *
 * @retval number of bytes copied
 *
 * @brief  Flattens the mbuf PDU
 *
 * Helper function to flatten the mbuf PDU. It is the application
 * responsibility to ensure that the buffer passed is large enough to hold the
 * PDU.
 *
 * =============================================================
 */
uint32_t
adk_ph_dev_coalesce_pdu(
    struct rte_mbuf    *inMbuf_p,
    void               *outPdu_p,
    uint32_t            outPduSize)
{
    uint32_t            segmentSize;
    uint32_t            lenCopied = 0;
    struct rte_mbuf    *mbuf_p;

    if (outPduSize < rte_pktmbuf_pkt_len(inMbuf_p))
    {
        return 0;
    }

    mbuf_p = inMbuf_p;
    do
    {
        segmentSize = rte_pktmbuf_data_len(mbuf_p);

        rte_memcpy((uint8_t *)outPdu_p + lenCopied,
                   rte_pktmbuf_mtod_offset(mbuf_p, uint8_t *, 0),
                   segmentSize);

        lenCopied += segmentSize;
        mbuf_p = mbuf_p->next;
    } while (NULL != mbuf_p);

    return lenCopied;
}


/*! end addtogroup _phFunctions {{{1
 * @}
 *
 * end addtogroup _ph
 * @}
 */
