/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ph_dev_internal.c
 *
 *      $HeadURL: ncp_adk/code/ph/include/adk_ph.h $
 *      $Revision: 366f377355b8b32e2740de4f058f578662bfab2c $
 *      $Date: Mon Oct 10 12:54:51 2016 -0500 $
 *
 * @addtogroup _ph  PH Module
 * @{
 *
 * @brief
 *
 * Contains the structure/union/macro definitions used by the PDU handler.
 * This file needs to be included by the user of the PDU handler.
 *
 *  **********************************************************************/

#include "rte_config.h"
#include "rte_mbuf.h"
#include "rte_mempool.h"

#include "inttypes.h"

#include "adk_os.h"
#include "adk_sif.h"
#include "adk_sys.h"
#include "adk_shm.h"
#include "adk_ph_dev.h"
#include "adk_ph_dev_internal.h"
#include "adk_ph_dev_dp_internal.h"
#include "adk_log_lib.h"
#include "adk_error.h"
#include "rte_bridge_cppi.h"

/*!
 * @addtogroup _phFunctions  PH Functions
 * @{
 *
 *
 * The API function declarations.   {{{1
 */

/*!
 * =============================================================
 * adk_ph_dev_shm_create   {{{2
 *
 * @param[in]
 *
 *
 *
 * @retval ADK_SUCCESS Successful completion
 *
 *
 * @brief  Creates the shared memory (CP & DP) for PH
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_shm_create(
    adk_sys_handle_t                sysHandle,
    adk_ph_dev_instance_config_t   *conf_p,
    adk_ph_dev_shm_info_t         **phShm_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_ph_dev_shm_info_t      *phShm_local_p=NULL;
    adk_ph_dev_dp_shm_info_t   *dpShm_p=NULL;

    /*
     * We will call adk_shm_create here because if the memzone exists, it will
     * return it. Otherwise it creates and returns it.
     */
    retValue = adk_shm_create(ADK_PH_DEV_SHM_MEMZONE_NAME,
                              sizeof(adk_ph_dev_shm_info_t),
                              ADK_SHM_LOOKUP_LCORE,
                              (adk_shm_handle_t **)&phShm_local_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH shared memory allocation failed\n");
        goto _runtime_failure_1;
    }

    /*
     * We will call adk_shm_create here because if the memzone exists, it will
     * return it. Otherwise it creates and returns it.
     */
    retValue = adk_shm_create(ADK_PH_DEV_DP_SHM_MEMZONE_NAME,
                              sizeof(adk_ph_dev_dp_shm_info_t),
                              ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                              (adk_shm_handle_t **)&dpShm_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH DP shared memory allocation failed\n");
        goto _runtime_failure_2;
    }

    /*
     * If the cookie is valid, then we jump out assuming the SHM has already
     * been created. We'll return an error, but not clear what is already
     * there so we might be able to diagnose it.
     */
    if ((ADK_PH_DEV_SHM_COOKIE == phShm_local_p->cookie) ||
        (ADK_PH_DEV_DP_SHM_COOKIE == dpShm_p->cookie))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH CP or DP SHM already configured\n");
        goto _return;
    }

    phShm_local_p->sysHandle = sysHandle;
    phShm_local_p->dp_p = dpShm_p;
/*    ((adk_sys_info_t *)sysHandle)->utils.phHandle = (adk_handle_t)phShm_local_p;*/

    /*
     * Alloc and init the PH DP lock
     */
    dpShm_p->dpLock_p = (void *)ADK_ALLOC_LOCK();
    if (dpShm_p->dpLock_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH lock alloc failed\n");
        goto _runtime_failure_3;
    }

    retValue = ADK_INIT_LOCK(dpShm_p->dpLock_p,
                             ADK_PH_DEV_LOCK_SHARED,
                             ADK_PH_DEV_LOCK_VALUE);
    if (retValue != ADK_SUCCESS)
    {
        ADK_FORCE_LOG_ERROR("PH couldn't init lock\n");
        goto _runtime_failure_4;
    }

    /*
     * Alloc and init the PH lock
     */
    phShm_local_p->lock_p = (void *)ADK_ALLOC_LOCK();
    if (phShm_local_p->lock_p == NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH lock alloc failed\n");
        goto _runtime_failure_5;
    }

    retValue = ADK_INIT_LOCK(phShm_local_p->lock_p,
                             ADK_PH_DEV_LOCK_SHARED,
                             ADK_PH_DEV_LOCK_VALUE);
    if (retValue != ADK_SUCCESS)
    {
        ADK_FORCE_LOG_ERROR("PH couldn't init lock\n");
        goto _runtime_failure_6;
    }

    phShm_local_p->cppiInstanceId = conf_p->cppiInstanceId;

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle), "PH\n");
    /*
     * If we get here, then all was well. Set the cookie and exit.
     *
     * Set DP first since the CP code will only check the CP cookie. This will
     * ensure that if we are pre-empted, the CP code will still see the system
     * as not initialiezed.
     */
    dpShm_p->cookie = ADK_PH_DEV_DP_SHM_COOKIE;
    phShm_local_p->cookie = ADK_PH_DEV_SHM_COOKIE;
    *phShm_p = phShm_local_p;

    goto _return;

_runtime_failure_6:
    ADK_FREE_LOCK(phShm_local_p->lock_p);

_runtime_failure_5:
    ADK_DESTROY_LOCK(dpShm_p->dpLock_p);

_runtime_failure_4:
    ADK_FREE_LOCK(dpShm_p->dpLock_p);

_runtime_failure_3:
    phShm_local_p->sysHandle = NULL;
    phShm_local_p->dp_p = NULL;
    ((adk_sys_info_t *)sysHandle)->utils.phHandle = NULL;

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_memzone_free",
                    "Free a memzone.\n");
    rte_memzone_free((struct rte_memzone *)dpShm_p);
    dpShm_p = NULL;

_runtime_failure_2:
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_memzone_free",
                    "Free a memzone.\n");
    rte_memzone_free((struct rte_memzone *)phShm_local_p);
    phShm_local_p = NULL;

_runtime_failure_1:
_return:
    return retValue;
}


/*!
 * =============================================================
 * @internal
 * adk_ph_dev_shm_get   {{{2
 *
 * @param[in]
 *
 *
 *
 * @retval ADK_SUCCESS Successful completion
 *
 *
 * @brief  Creates the shared memory (CP & DP) for PH
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_shm_get(
    adk_sys_handle_t                sysHandle,
    adk_ph_dev_shm_info_t         **phShm_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_ph_dev_shm_info_t      *phShm_local_p=NULL;
    adk_ph_dev_dp_shm_info_t   *dpShm_p=NULL;

    /*
     * We will call adk_shm_create here because if the memzone exists, it will
     * return it. Otherwise it creates and returns it.
     */
    retValue = adk_shm_create(ADK_PH_DEV_SHM_MEMZONE_NAME,
                              sizeof(adk_ph_dev_shm_info_t),
                              ADK_SHM_LOOKUP_LCORE,
                              (adk_shm_handle_t **)&phShm_local_p);
    if ((retValue != ADK_SUCCESS) &&
        (ADK_STATUS_ERROR_CODE_GET(retValue) != ADK_W_SHM_EXISTS))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH shared memory allocation failed\n");
        goto _runtime_failure_1;
    }

    /*
     * We will call adk_shm_create here because if the memzone exists, it will
     * return it. Otherwise it creates and returns it.
     */
    retValue = adk_shm_create(ADK_PH_DEV_DP_SHM_MEMZONE_NAME,
                              sizeof(adk_ph_dev_dp_shm_info_t),
                              ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                              (adk_shm_handle_t **)&dpShm_p);
    if ((retValue != ADK_SUCCESS) &&
        (ADK_STATUS_ERROR_CODE_GET(retValue) != ADK_W_SHM_EXISTS))
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH DP shared memory allocation failed\n");
        goto _runtime_failure_1;
    }

    /*
     * Since we know that the SHM should have been created, expect the return
     * of ADK_W_SHM_EXISTS and set retValue to ADK_SUCCESS
     */
    retValue = ADK_SUCCESS;

    /*
     * If the cookie is valid, then we jump out assuming the SHM has already
     * been created. We'll return an error, but not clear what is already
     * there so we might be able to diagnose it.
     */
    if ((ADK_PH_DEV_SHM_COOKIE != phShm_local_p->cookie) ||
        (ADK_PH_DEV_DP_SHM_COOKIE != dpShm_p->cookie))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "PH CP or DP SHM not configured\n");
        goto _runtime_failure_1;
    }

    *phShm_p = phShm_local_p;

    goto _return;

_runtime_failure_1:
_return:
    return retValue;
}


/*!
 * =============================================================
 * @internal
 * adk_ph_dev_next_dev_get          {{{2
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief Get the first available device
 *
 * Walk the list of devices and return the first one. If an available entry
 * isn't found, then ADK_PH_DEV_MAX_NUM_DEVICES is return. This error should
 * never happen as we check for numDevices earlier and error out if we
 * exceed the max number.
 *
 * =============================================================
 */
uint16_t
adk_ph_dev_next_dev_get(
    adk_ph_dev_shm_info_t    *phShm_p)
{
    uint16_t    i;

    for (i = ADK_PH_DEV_DEVICE_ID_BASE; i <= ADK_PH_DEV_DEVICE_ID_MAX; i++)
    {
        if (NULL == phShm_p->devices[i])
        {
            break;
        }
    }

    ADK_RETURN(i);
}    /* -----  end of function adk_ph_dev_next_dev_get  ----- */


/*!
 * =============================================================
 * @internal
 * adk_ph_dev_cppi_device_get           {{{2
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief  Creates a CPPI device for PH free/alloc queues
 *
 * This function will search the CPPI list until it finds an open slot to use.
 * Since PH is currently the only user, 0 will most often be open.
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_cppi_device_get(
    adk_sys_handle_t    sysHandle,
    uint8_t             instanceId,
    uint16_t            deviceId)
{
    adk_status_t                            retValue = ADK_SUCCESS;
    int                                     ret;
    uint16_t                                cppiDevId;
#if RDK_REL_MAJOR >= 13
    struct rte_cppi_dev_config       cppiConfig = {0};
#else
    struct rte_bridge_cppi_dev_config       cppiConfig = {0};
#endif

    cppiDevId = deviceId - ADK_PH_DEV_DEVICE_ID_BASE;
#if RDK_REL_MAJOR >= 13
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_cppi_dev_configure",
                    "Configure a CPPI Bridge device. instId: %d, devId: %d\n",
                    instanceId, deviceId);
    ret = rte_cppi_dev_configure (deviceId, 0, 0, &cppiConfig);
    if (ret != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Could not configure CPPI device -- devId: %d \n",
                      deviceId);
        goto _return;
    }

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_dev_add",
                    "Add a CPPI device to BD. instId: %d, devId: %d\n",
                    instanceId, deviceId);
    ret = rte_bridge_cppi_dev_add(instanceId, deviceId);
    if (ret != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Could not add CPPI device to -- instId: %d, devId: %d \n",
                      instanceId, deviceId);
        goto _return;
    }
#else
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_dev_add",
                    "Configure a CPPI Bridge device. instId: %d, devId: %d\n",
                    instanceId, deviceId);
    ret = rte_bridge_cppi_dev_add(instanceId, deviceId, &cppiConfig);
    if (ret != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Could not add CPPI device -- instId: %d, devId: %d \n",
                      instanceId, deviceId);
        goto _return;
    }
#endif

_return:
    ADK_RETURN( retValue );
}    /* -----  end of function adk_ph_dev_cppi_device_get  ----- */


/*!
 * =============================================================
 * @internal
 * adk_ph_dev_cppi_device_free          {{{2
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief  Creates a CPPI device for PH free/alloc queues
 *
 * This function will search the CPPI list until it finds an open slot to use.
 * Since PH is currently the only user, 0 will most often be open.
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_cppi_device_free(
    adk_sys_handle_t    sysHandle,
    uint8_t             instanceId,
    uint16_t            deviceId)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int     ret;

#if RDK_REL_MAJOR >= 13
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_dev_rem",
                    "Free a CPPI Bridge device. instId: %d, devId: %d\n",
                    instanceId, deviceId);
    ret = rte_bridge_cppi_dev_rem(instanceId, deviceId);
    if (ret != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "remove CPPI-DEV from -- instId: %d, devId: %d \n",
                      instanceId, deviceId);
        goto _return;
    }

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_cppi_dev_reset",
                    "Free a CPPI device.  devId: %d\n",
                    deviceId);
    ret = rte_cppi_dev_reset(deviceId);
    if (ret != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "CPPI-DEV free Invalid value -- devId: %d \n",
                      deviceId);
        goto _return;
    }
#else
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_bridge_cppi_dev_free",
                    "Free a CPPI Bridge device. instId: %d, devId: %d\n",
                    instanceId, deviceId);
    ret = rte_bridge_cppi_dev_free(instanceId, deviceId);
    if (ret != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "CPPI free Invalid value -- instId: %d, devId: %d \n",
                      instanceId, deviceId);
        goto _return;
    }

#endif
_return:
    ADK_RETURN( retValue );
}    /* -----  end of function adk_ph_dev_cppi_device_free ----- */


/*!
 * =============================================================
 * @internal
 * adk_ph_dev_ring_create           {{{2
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief  Create an RTE ring for use in CPPI free/alloc
 *
 * Detailed description starts here
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_ring_create(
    adk_sys_handle_t    sysHandle,
    uint16_t            deviceId,
    uint16_t            queueId,
    char                ringType,
    uint16_t            allocFreeRingSize,
    struct rte_ring   **ring_p)
{
    adk_status_t    retValue = ADK_SUCCESS;
    char            ring_name[ADK_PH_DEV_MAX_RING_NAME_LEN+1];

    ADK_MEMSET(ring_name, 0, sizeof(char) * ADK_PH_DEV_MAX_RING_NAME_LEN);
    snprintf(ring_name, ADK_PH_DEV_MAX_RING_NAME_LEN, "ADK_PH_RING_%c_%u_%u",
             ringType, deviceId, queueId);
    ring_name[ADK_PH_DEV_MAX_RING_NAME_LEN] = '\0';

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_ring_create",
                    "Create a new named ring in memory '%s'.\n", ring_name);
    *ring_p = rte_ring_create(ring_name, allocFreeRingSize, rte_socket_id(), 0);
    if(NULL == *ring_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "rte_ring_create failed device for ring %s: '%s' (%d)\n",
                      ring_name, rte_strerror(rte_errno), rte_errno);
        goto _return;
    }

_return:
    ADK_RETURN( retValue );
}    /* -----  end of function adk_ph_dev_ring_create  ----- */

/*!
 * =============================================================
 * @internal
 * adk_ph_dev_ring_destory           {{{2
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief  Create an RTE ring for use in CPPI free/alloc
 *
 * Detailed description starts here
 *
 * =============================================================
 */
void
adk_ph_dev_ring_destroy(
    adk_sys_handle_t    sysHandle,
    struct rte_ring    *ring_p)
{
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_ring_free",
                    "De-allocate all memory used by the ring.");
    rte_ring_free(ring_p);
}    /* -----  end of function adk_ph_dev_ring_destory  ----- */


/*
 * adk_ph_dev_dp_rx_queue_stats_get           {{{2
 * @param[in,out]
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief  Add the callback function to the end of the callback list
 *
 * Detailed description starts here
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_dp_rx_queue_stats_get(
    adk_ph_dev_dp_queue_info_t     *queue_p,
    adk_ph_dev_dp_queue_stats_t    *stats_p)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             coreId;

    ADK_MEMSET(stats_p, 0, sizeof(adk_ph_dev_dp_queue_stats_t));

    for (coreId = 0; coreId < ADK_MAX_NUM_CORES; coreId++)
    {
        stats_p->rx.pdusEnqueued += queue_p->stats[coreId].rx.pdusEnqueued;
        stats_p->rx.pdusDequeued += queue_p->stats[coreId].rx.pdusDequeued;
        /* stats_p->rx.pdusEnqDropped += queue_p->stats[coreId].rx.pdusEnqDropped; */
        stats_p->rx.pdusDeqDropped += queue_p->stats[coreId].rx.pdusDeqDropped;
        stats_p->rx.rxDisabled += queue_p->stats[coreId].rx.rxDisabled;
        stats_p->rx.bufferHighWatermark += queue_p->stats[coreId].rx.bufferHighWatermark;
        stats_p->rx.bufferFull += queue_p->stats[coreId].rx.bufferFull;
        stats_p->rx.congestStop += queue_p->stats[coreId].rx.congestStop;
        stats_p->rx.intrSent += queue_p->stats[coreId].rx.intrSent;
        stats_p->rx.intrSendFailed += queue_p->stats[coreId].rx.intrSendFailed;
        /* stats_p->rx.pdusDroppedQosRed += queue_p->stats[coreId].rx.pdusDroppedQosRed; */
    }

    ADK_RETURN( retValue );
}    /* -----  end of function adk_ph_dev_dp_rx_queue_stats_get  ----- */


/*! end addtogroup _phFunctions {{{1
 * @}
 *
 * end addtogroup _ph
 * @}
 *
 * vim: set foldmethod=marker foldlevel=1 foldcolumn=3 columns=83:
 *
 */
