/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ph_dev_dp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _phFunctions  PH Module
 *
 * @{
 *
 * @brief  Contains the ADK PDU handler (PH) APIs.
 *
 *  **********************************************************************/

#include <inttypes.h>
#include <stdbool.h>

#include "rte_errno.h"
#include "rte_mbuf.h"

#include "adk_sys.h"
#include "adk_module_mgr.h"
#include "adk_sys_lcore.h"
#include "adk_sif.h"
#include "adk_shm.h"
#include "adk_pktio.h"
#include "adk_log_lib.h"
#include "adk_sif_dp_log_lib.h"

#include "adk_pal_dp_export.h"
#include "adk_pal_dp.h"

#include "adk_ph_dev.h"
#include "adk_ph_dev_internal.h"
#include "adk_ph_dev_dp_internal.h"
#include "adk_ph_dev_dp.h"

#include "adk_qos_dp_export.h"

#if defined(ADK_MODULE_SIF_LTTNG_DP_ENABLED)
#define TRACEPOINT_DEFINE
#define TRACEPOINT_PROBE_DYNAMIC_LINKAGE
#include "adk_ph_dev_dp_lttng_debug.h"
#endif

static adk_ph_dev_dp_shm_info_t    *dpShm_p = NULL;

/*  {{{1 */
/*!
 * =============================================================
 * @internal
 * adk_ph_dev_dp_to_cp_send_stage_2      {{{2
 * @param[in]   pdu_p: Pointer to the mbuf
 *
 * @retval ADK_SUCCESS:  Successful completion
 * @retval ADK_E_INVALID_HANDLE:  One of the handles passed is invalid
 * @retval ADK_E_RX_BUFFER_FULL:  Unable to write to the buffer, it is full
 * @retval ADK_DP_W_QOS_RED_PDU_DROPPED:
 * @retval ADK_W_RX_HIGHWATER_MARK_HIT:
 * @retval ADK_E_INVALID_PARAMETER:
 *
 * =============================================================
 */
static adk_status_t
adk_ph_dev_dp_to_cp_send_stage_2(
    void   *pdu_p,
    uint8_t CbFnBatch,
    void* CbParamBatch)
{
    int                             retValue = ADK_SUCCESS;
    int                             ret = ADK_E_TASK_CONSUMED;
    adk_ph_dev_dp_device_info_t    *dev_p;
    int                             coreIndex = ADK_LCORE_INDEX;
    uint8_t                         policeColor;
    uint16_t                        devId;
    uint16_t                        queueId;
    adk_ph_dev_dp_metadata_t       *phMeta_p;
    adk_ph_dev_dp_queue_info_t     *rxQ_p;
#ifdef ADK_QOS_GEN3_INTERFACE
    adk_slowpath_dev_dp_metadata_t *spDevMeta_p;
#endif

#ifdef ADK_QOS_GEN3_INTERFACE
    spDevMeta_p = (adk_slowpath_dev_dp_metadata_t *)&(ADK_METADATA_PARAM(pdu_p,
                                                            imc_param_start));
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(spDevMeta_p->imc_param_start);
#else
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(ADK_METADATA_PARAM(pdu_p,
                                                            imc_param_start));
#endif
    devId   = ADK_MD_GET_ADK_DEV_ID(pdu_p);
    queueId = ADK_MD_GET_DEV_TX_QUEUE_ID(pdu_p);
    dev_p   = dpShm_p->devices[devId];

    if (dev_p == NULL)
    {
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                            "dev_p is NULL for device %d\n", devId);
        retValue = ADK_E_INVALID_HANDLE;
        goto _error;
    }

    dev_p->inUse = true;

    rxQ_p = dev_p->rxQueues[queueId];

    if (rxQ_p == NULL)
    {
        ADK_DP_LOG_ERROR_2D(ADK_SYS_MODULE_ID,
                            "rxQueues is NULL for device: %d, queue: %d\n",
                            devId, queueId);
        retValue = ADK_E_INVALID_HANDLE;
        goto _exit;
    }

    /*
     * Check policier if it has been defined
     */
#ifdef ADK_QOS_GEN3_INTERFACE
    if (spDevMeta_p->policerFlowId != UINT_MAX)
#else 
    if (phMeta_p->policerFlowId != UINT_MAX)
#endif
    {
        retValue = adk_qos_dp_policing_srtcm_color_blind_pps(
#ifdef ADK_QOS_GEN3_INTERFACE
                                    spDevMeta_p->policerFlowId,
#else
                                    phMeta_p->policerFlowId,
#endif
                                    &policeColor);

        if (retValue != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Policing call failed\n");
            goto _exit;
        }

#ifdef ADK_QOS_GEN3_INTERFACE
        adk_sys_dp_queue_stats_update(pdu_p , policeColor);
#endif
        if (policeColor == ADK_QOS_DP_RED)
        {
            dev_p->devStats.rateLimitingDrops++; 
            retValue = ADK_DP_WARNING_SEV_LEVEL_SET(ADK_DP_W_QOS_RED_PDU_DROPPED);
            ADK_DP_LOG_INFO(ADK_SYS_MODULE_ID, "Policing call returned RED\n");
            adk_pal_dp_release_send((adk_pktio_task_t *)pdu_p, CbFnBatch, NULL);
            adk_pktio_task_buffer_free((adk_pktio_task_t *)pdu_p);
            goto _exit;
        }
    }
    /*
     * Need to set th callback function ID into the metadata so we can extract
     * it on the other size.
     */
    ADK_PH_DEV_CB_FN_ID_SET(pdu_p, phMeta_p->cbFnId);

    if (dev_p->isRxEnabled == true)
    {
        if (dev_p->type == ADK_PH_DEV_TYPE_EVENTDEV)
        {
            ADK_DP_LOG_INFO_2X_DEBUG(ADK_SYS_MODULE_ID,
                                     "Evdev: qId: %d, cbFnId: %d\n",
                                     queueId, phMeta_p->cbFnId);
            if(CbFnBatch == ADK_PAL_DP_BATCH_CB_FN_PH)  
            {
                CbParamBatch = rxQ_p;
            }

            ret = adk_pal_dp_ph_send(rxQ_p->u.evd.queueId,
                                  phMeta_p->cbFnId, phMeta_p->priority,
                                  (struct rte_mbuf *)pdu_p, CbFnBatch, CbParamBatch);
        }
        else
        {
            rxQ_p->stats[coreIndex].rx.unknownType++;
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        }

        if(ret != ADK_E_TASK_CONSUMED)
        {
            ADK_DP_LOG_INFO_ERRNO(ADK_SYS_MODULE_ID,
                    "PH DP PDU send failed: '%s' (0x%x)\n",
                    adk_sys_error_string(ADK_STATUS_ERROR_CODE_GET(retValue)),
                    retValue);
            retValue = ADK_E_INTERNAL_ERROR;
        }
        else
        {
            retValue = ADK_SUCCESS;
        }
    } else
    {
        rxQ_p->stats[coreIndex].rx.rxDisabled++;
        retValue = ADK_WARNING_SEV_LEVEL_SET(ADK_DP_W_PH_RX_DISABLED);
    }

_exit:
    dev_p->inUse = false;

_error:
    return retValue;
}

/*!
 * =============================================================
 * @internal
 * adk_ph_dev_rate_limit_stage_entry      {{{2
 * @param[in]   task_p: Array of pointers to mbufs. Assumption is there is
 *                      only one mbuf.
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief  Pipeline entry stage when policing is needed
 *
 * When policing is enabled, this will be the entry stage. This function will
 * call the maine function that does work. Only 1 PDU is expected and
 * processed.
 *
 * =============================================================
 */
static adk_status_t
adk_ph_dev_rate_limit_stage_entry(
    void                          **task_p,
    uint8_t                         taskCount,
    uint32_t                        taskMask,
    void                           *ring,
    adk_sif_cfg_pipeline_info_t    *nextStageEntry,
    int                             coreIndex)
{
    adk_status_t    retValue = ADK_SUCCESS;

    RTE_SET_USED(taskCount);
    RTE_SET_USED(taskMask);
    RTE_SET_USED(ring);
    RTE_SET_USED(nextStageEntry);
    RTE_SET_USED(coreIndex);

    retValue = adk_ph_dev_dp_to_cp_send_stage_2(task_p[0], ADK_PAL_DP_BATCH_CB_FN_PH, NULL);
    if (ADK_SUCCESS != ADK_STATUS_ERROR_CODE_GET(retValue))
    {
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID,
                           "Stage 2 failed: 0x%x\n", retValue);
    }

    return retValue;
}


/*!
 * =============================================================
 * adk_ph_dev_dp_instance_init      {{{2
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval ADK_E_SYS_NOT_INITIALIZED
 *
 * @brief  Initialize the globals per process
 *
 * This will initialize all the globals needed for the PH DP thread. This must
 * be called after the group handle has been created and needs to be called on
 * all cores running a PH data path.
 *
 * =============================================================
 */
adk_status_t
adk_ph_dev_dp_instance_init (
    void)
{
    adk_status_t retValue = ADK_SUCCESS;

    retValue = adk_shm_lookup(ADK_PH_DEV_DP_SHM_MEMZONE_NAME,
                              (adk_shm_handle_t **)&dpShm_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "PH DP shared memory lookup failed\n");
        goto _runtime_failue_1;
    }

    if (ADK_PH_DEV_DP_SHM_COOKIE != dpShm_p->cookie)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "System Initialization from control plane not yet "
                         "complete. Please wait and try again later.\n");
        retValue = ADK_E_SYS_NOT_INITIALIZED;
    }

    adkDestEntryFunction_g[ADK_PH_DEV_RATE_LIMIT_FN_ID] = adk_ph_dev_rate_limit_stage_entry;

_runtime_failue_1:
    return retValue;
}    /* -----  end of function adk_ph_dev_dp_instance_init  ----- */

/*!
 * =============================================================
 * adk_ph_dev_dp_to_cp_send    {{{2
 *
 * @param[in,out]   pdu_p: Pointer to the mbuf
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * Brief: Send PDu to the CP
 *
 * This API will forward a PDU to the control plane. All parameters in
 * adk_ph_dev_dp_metadata_t must be defined at the start of the IMC data. The
 * application specific data can be appeneded after the PH metadata by using
 * the imc_param_start variable.
 *
 * =============================================================
 */

adk_status_t adk_ph_dev_dp_to_cp_send(void *pdu_p, uint8_t CbFnBatch, void* CbParamBatch)
{
    int                             retValue = ADK_SUCCESS;
    adk_status_t                    ret = 0;
    adk_ph_dev_dp_device_info_t    *dev_p;
    int                             coreIndex = ADK_LCORE_INDEX;
    uint16_t                        devId;
    uint16_t                        queueId;
    adk_ph_dev_dp_metadata_t       *phMeta_p;
#ifdef ADK_QOS_GEN3_INTERFACE
    adk_slowpath_dev_dp_metadata_t *spDevMeta_p;
#endif

    if (NULL == pdu_p)
    {
        ADK_DP_LOG_ERROR_1P(ADK_SYS_MODULE_ID, "pdu_p is NULL %p\n", pdu_p);
        retValue = ADK_E_INVALID_HANDLE;
        goto _exit;
    }

#ifdef ADK_QOS_GEN3_INTERFACE
    spDevMeta_p = (adk_slowpath_dev_dp_metadata_t *)&(ADK_METADATA_PARAM(pdu_p,
                                                            imc_param_start));
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(spDevMeta_p->imc_param_start);
#else 
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(ADK_METADATA_PARAM(pdu_p,
                                                            imc_param_start));
#endif
    devId = ADK_MD_GET_ADK_DEV_ID(pdu_p);
    queueId = ADK_MD_GET_DEV_TX_QUEUE_ID(pdu_p);
    dev_p = dpShm_p->devices[devId];

    if (NULL == dev_p)
    {
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                            "dev_p is NULL for device %d\n", devId);
        retValue = ADK_E_INVALID_HANDLE;
        goto _exit;
    }

    dev_p->inUse = true;

    if (NULL == dev_p->rxQueues[queueId])
    {
        ADK_DP_LOG_ERROR_2D(ADK_SYS_MODULE_ID,
                            "queue_p is NULL for device:queue %d:%d\n",
                            devId, queueId);
        retValue = ADK_E_INVALID_HANDLE;
        goto _error;
    }

    ADK_DP_LOG_INFO_1P_4D_DEBUG(ADK_SYS_MODULE_ID,
                                "pdu_p: %p, devId: %d, qId: %d, "
                                "prio: %d, cbId: %d\n",
                                pdu_p, devId, queueId,
                                phMeta_p->priority, phMeta_p->cbFnId);
    /*
     * Check policier if it has been defined
     */
#ifdef ADK_QOS_GEN3_INTERFACE
    if (UINT_MAX != spDevMeta_p->policerFlowId)
#else 
    if (UINT_MAX != phMeta_p->policerFlowId)
#endif
    {
        if (adkDestEntryFunction_g[ADK_PH_DEV_RATE_LIMIT_FN_ID])
        {
#if defined(ADK_PAL_DYNAMIC_PPL)
            ret = adk_pal_dp_stage_send((void **)&pdu_p, 1, 1, ADK_PH_DEV_RATE_LIMIT_FN_ID,
#ifdef ADK_QOS_GEN3_INTERFACE
                    spDevMeta_p->policerFlowId,
#else
                    phMeta_p->policerFlowId,
#endif
                    0x00, CbFnBatch, ADK_PAL_DP_BATCH_CB_EVENT_STAGE_PH_POLICER, CbParamBatch);
                                                                                             

            if (ret == ADK_E_INTERNAL_ERROR)
            {
                dev_p->rxQueues[queueId]->stats[coreIndex].rx.palNextStageFailed++;
                ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID,
                        "Failed to send to next stage; ret: 0x%x\n",
                        ret);
                retValue = ADK_E_INVALID_HANDLE;
                goto _error;
            }
#endif
        }
    }else
    {
        retValue = adk_ph_dev_dp_to_cp_send_stage_2(pdu_p, CbFnBatch, CbParamBatch);
    }

_error:
    dev_p->inUse = false;

_exit:

    return (adk_status_t)retValue;
}

void adk_ph_dev_pal_batch_handle_return(
    uint16_t devId,
    uint16_t queueId,
    uint8_t cbEventType,
    bool sendSuccess)
{   
    int                    ret = 0;
    adk_ph_dev_dp_device_info_t    *dev_p;
    int                             coreIndex = ADK_LCORE_INDEX;
    uint8_t                         fdMesg[ADK_EVENT_SIZE];

    if((cbEventType != ADK_PAL_DP_BATCH_CB_EVENT_PH) ||
            (cbEventType != ADK_PAL_DP_BATCH_CB_EVENT_STAGE_PH_POLICER))
    {
        return;
    }

    debug_print("adk_ph_dev_pal_batch_handle_return\n");
    dev_p = dpShm_p->devices[devId];

    if (NULL == dev_p)
    {
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                            "dev_p is NULL for device %d\n", devId);
        return;
    }

    dev_p->inUse = true;

    if (NULL == dev_p->rxQueues[queueId])
    {
        ADK_DP_LOG_ERROR_2D(ADK_SYS_MODULE_ID,
                            "queue_p is NULL for device:queue %d:%d\n",
                            devId, queueId);
        dev_p->inUse = false;
        return;
    }

    if(cbEventType == ADK_PAL_DP_BATCH_CB_EVENT_PH)
    {
        if(sendSuccess)
        {
            dev_p->rxQueues[queueId]->stats[coreIndex].rx.pdusEnqueued++;
            debug_print("incremented  pdusEnqueued\n");

            if (dev_p->interruptEnabled == true)
            {
                fdMesg[0] = ADK_PH_DEV_INTR_MSG_EVENT_SENT;
                ret = write(dev_p->controlFd, fdMesg, ADK_EVENT_SIZE);
                if (ret == -1)
                {
                    dev_p->rxQueues[queueId]->stats[coreIndex].rx.intrSendFailed++;
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID,
                            "write fd: %d\n",
                            dev_p->controlFd);
                    ADK_DP_LOG_INFO_ERRNO(ADK_SYS_MODULE_ID,
                            "write failed: '%s' (0x%x)\n",
                            strerror(errno), errno);    
                    dev_p->inUse = false;
                    return;
                }
                dev_p->rxQueues[queueId]->stats[coreIndex].rx.intrSent++;
            }
        }
        else
        {
            dev_p->rxQueues[queueId]->stats[coreIndex].rx.bufferFull++;
#if 0
            /*
             *  Check to see congestion control. If set to STOP, then disable
             *  transmit. The application must call adk_ph_dev_rx_queue_start() to
             *  resume transmitting of PDUs.
             */
            if ((ret < 0) && (dev_p->congestionControlStop == true))
            {
                rxQ_p->stats[coreIndex].rx.congestStop++;
            }
#endif
        }
        dev_p->inUse = false;
        return;
    }
    else if(cbEventType == ADK_PAL_DP_BATCH_CB_EVENT_STAGE_PH_POLICER)
    {
        if(!sendSuccess)
        {
            dev_p->rxQueues[queueId]->stats[coreIndex].rx.palNextStageFailed++;
        }
    }
    dev_p->inUse = false;
    return;
}

void adk_ph_dp_pal_batch_cb_fn(
        adk_pktio_task_t *dpdkMbuf,
        uint8_t       cbEventType,
        void          *cbParam,
        bool          sendSuccess)
{
    RTE_SET_USED(cbParam);
    if(cbEventType == ADK_PAL_DP_BATCH_CB_EVENT_PH)
    {
        debug_print("Received ADK_PAL_DP_BATCH_CB_EVENT_PH in adk_ph_dp_pal_batch_cb_fn\n");
        if(cbParam != NULL)
        {
            adk_ph_dev_dp_queue_info_t *rxQ = cbParam;
            
            adk_ph_dev_pal_batch_handle_return(rxQ->u.evd.queueId, rxQ->u.evd.devId, cbEventType, sendSuccess);
        }
        if(!sendSuccess)
        {
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(dpdkMbuf));
        }
    }
    return;
}
/*!
 * }}}2
 *
 * @} end addtogroup _phFunctions }}}1
 */
