/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_sif_cfg_include.h"

/*!
 * =============================================================
 * adk_sif_cfg_init
 * @param[out] cfg_info  Handle to configuration info table.
 *
 * @retval ADK_SUCCESS Successfully initialized cfg module.
 * @retval ADK_E_MEM_UNAVAILABLE Dynamic memory allocation failed
 *
 * @brief  Initializes CFG module.
 *
 * This function initializes configuration file parsing module.
 * It allocates memory for cfg info tables.
 *
 * =============================================================
 */
adk_status_t
adk_sif_cfg_init(adk_sif_cfg_info_t **cfg_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_params_t *cfg_params = NULL;

    /* Create cfg info table which stores parsed cfg file sections 
     * information 
     */
    *cfg_info = ADK_NVM_CALLOC(ADK_MISC_MODULE_ID, 0, 1, sizeof(adk_sif_cfg_info_t));
    if (*cfg_info == NULL)
    {
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
        ADK_PREINIT_WRITE("Memory allocation failure. Error code: %#x\n", ret);
        return ret;
    }
   
    //cfg_params = (adk_cfg_params_t *)(*cfg_info)->cfg_params_tbl;
    
    ret = adk_cfg_init(&cfg_params);
    if (ret != ADK_SUCCESS)
    {
        ADK_NVM_FREE(*cfg_info);
        ADK_PREINIT_WRITE("Configuration init failed. Error code: %#x\n", ret);
        return ret;
    }
    (*cfg_info)->cfg_params_tbl = (void *)cfg_params;

    adk_sif_cfg_eal_init(*cfg_info);
    adk_sif_cfg_mempool_init(*cfg_info);
    adk_sif_cfg_port_init(*cfg_info);
    adk_sif_cfg_ring_init(*cfg_info);
    adk_sif_cfg_pkt_buffs_init(*cfg_info);
    adk_sif_cfg_pipeline_init(*cfg_info);
    
    return ret;
}

/*!
 * =============================================================
 * adk_sif_cfg_shutdown
 * @param[in] cfg_info  Handle to configuration info table.
 *
 * @retval ADK_SUCCESS Successfully shutdown cfg module.
 *
 * @brief  Shuts down CFG module.
 *
 * This function releases memory allocated in configuration 
 * file parsing module.
 *
 * =============================================================
 */
adk_status_t
adk_sif_cfg_shutdown(adk_sif_cfg_info_t *cfg_info_hdl)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_params_t *cfg_params = NULL;

    if (cfg_info_hdl == NULL)
    {
        ADK_PREINIT_WRITE("Invalid cfg info handle pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    cfg_params = (adk_cfg_params_t *)cfg_info_hdl->cfg_params_tbl;

    adk_cfg_shutdown(cfg_params);

    /* Delete cfg info table */
    free(cfg_info_hdl);
    
    return ret;
}

/*!
 * =============================================================
 * adk_sif_cfg_load
 * @param[in] cfg_info  Handle to configuration info table.
 * @param[in] sysConf_p Handle to system configuration. Contains
 *                      CLI argc and argv.
 *
 * @retval ADK_SUCCESS Successfully load cfg module.
 * @retval ADK_E_MEM_UNAVAILABLE Dynamic memory allocation failed
 *
 * @brief  Loads CFG module.
 *
 * This function parses configuration file and setups the 
 * components specified in the configuration file.
 *
 * =============================================================
 */
adk_status_t
adk_sif_cfg_load(int argc, char **argv, adk_sif_cfg_info_t *cfg_info_hdl)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_params_t *cfg_params = NULL;

    if (cfg_info_hdl == NULL)
    {
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
        ADK_PREINIT_WRITE("Invalid cfg info handle pointer. Error code: %d\n", ret);
        return ret;
    }

    cfg_params = (adk_cfg_params_t *)cfg_info_hdl->cfg_params_tbl;

    ret = adk_cfg_parse(argc, argv, cfg_params);
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Configuration load failed. Error code: %#x\n", ret);
        return ret;
    }

    ret = adk_sif_cfg_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("SIF Configuration load failed. Error code: %#x\n", ret);
        return ret;
    }

    return ret;
}


adk_status_t
adk_sif_cfg_configure(adk_sif_cfg_info_t *cfg_info_hdl)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_eal_info_t *eal_info = NULL;

    ret = adk_sif_cfg_eal_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }

    ret = adk_sif_cfg_eal_info_get(cfg_info_hdl->eal_info_tbl, &eal_info);
    if ((ret != ADK_SUCCESS) || (eal_info == NULL))
    {
        ADK_PREINIT_WRITE("Error: EAL info get failed\n");
        return ret;
    }

    if (eal_info->process_type == ADK_SIF_CFG_EAL_PROCESS_TYPE_PRIMARY) 
    {
    ret = adk_sif_cfg_mempool_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
    
    ret = adk_sif_cfg_port_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
    
    ret = adk_sif_cfg_ring_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
    
    ret = adk_sif_cfg_pkt_buffs_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }

    ret = adk_sif_cfg_pipeline_configure(cfg_info_hdl);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
    }
    return ret;
}
