/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_modules_enabled.h"
#include "adk_sif_constants.h"
#include "adk_sif_cfg_include.h"
#include "adk_ethernet.h"
#include "adk_eth_dp.h"
#include "adk_ethernet_export.h"
#include "adk_pktio.h"

/* Table to store pipeline configuration information */
adk_sif_cfg_pipeline_info_tbl_t         *adkPipelineInfoTbl_g;

/* Entry functions */
const uint16_t adkEntryFunctions_g[] = {
    ADK_ETH_LLI_FN_ID,
#if 0
    adk_pktio_mod_rx,
#endif
};

adk_status_t
adk_sif_cfg_pipeline_init(adk_sif_cfg_info_t *cfg_info)
{
    adk_status_t ret = ADK_SUCCESS;

    adkPipelineInfoTbl_g = calloc(1, sizeof(adk_sif_cfg_pipeline_info_tbl_t)); 
    if (adkPipelineInfoTbl_g == NULL)
    {
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
        ADK_PREINIT_WRITE("Memory allocation failure. Error code: %d\n", ret);
        return ret;
    }

    cfg_info->pipeline_info_tbl = (void *)adkPipelineInfoTbl_g;

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_configure(adk_sif_cfg_info_t *cfg_info_hdl)
{    
    adk_status_t ret = ADK_SUCCESS, get_next_ret = ADK_SUCCESS;
    adk_cfg_params_t *cfg_params_tbl = NULL;
    adk_cfg_pipeline_params_tbl_t *pipeline_params_tbl = NULL;
    adk_cfg_pipeline_params_t *pipeline_params = NULL, *prev_pipeline_params = NULL;
    adk_sif_cfg_pipeline_info_t *pipeline_info = NULL, *prev_pipeline_info = NULL;
    adk_sif_cfg_pipeline_bind_info_t bind_info = {0};
    adk_sif_cfg_pipeline_port_rx_bind_info_t port_rx_bind_info = {0};
    adk_sif_cfg_pipeline_ring_rx_bind_info_t ring_rx_bind_info = {0};
    adk_sif_cfg_pipeline_buff_rx_bind_info_t buff_rx_bind_info = {0};
    adk_sif_cfg_pipeline_port_tx_bind_info_t port_tx_bind_info = {0};
    adk_sif_cfg_pipeline_ring_tx_bind_info_t ring_tx_bind_info = {0};
    adk_sif_cfg_pipeline_buff_tx_bind_info_t buff_tx_bind_info = {0};
    uint32_t i = 0;

    cfg_params_tbl = (adk_cfg_params_t *)cfg_info_hdl->cfg_params_tbl;
    pipeline_params_tbl = (adk_cfg_pipeline_params_tbl_t *)cfg_params_tbl->pipeline_params_tbl;

    /* Check for the consistency of all pipelines in pipeline parameters tables */
    ret = adk_cfg_pipeline_param_check(pipeline_params_tbl);
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline configuration failed\n");
        return ret;
    }

    /* Start of pipeline params table */
    adk_cfg_pipeline_param_rewind(pipeline_params_tbl);

    for (i = 0; i < pipeline_params_tbl->size; i++)
    {
        /* To sanity check */
        pipeline_info = NULL;
        
        get_next_ret = adk_cfg_pipeline_param_get_next(pipeline_params_tbl, &pipeline_params);        
        if (get_next_ret == ADK_E_LIST_RANGE)
        {
            break;
        }

        switch(pipeline_params->stage)
        {
            case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR:
            case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR: 
                /* Create pipeline */ 
                ret = adk_sif_cfg_pipeline_create(cfg_info_hdl, pipeline_params, &pipeline_info);
                if (ret != ADK_SUCCESS) 
                {
                    return ret;
                }
                if (pipeline_info == NULL)
                {
                    ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                }

                /* Bind Rx port/ring to pipeline */
                if (pipeline_params->modules[_RX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_PORT_RX)
                {
                    port_rx_bind_info.info = pipeline_info;
                
                    ret = adk_sif_cfg_pipeline_port_rx_bind(cfg_info_hdl, pipeline_params, &port_rx_bind_info);
                    if (ret != ADK_SUCCESS)
                    {
                        ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                    }
                }
                else if (pipeline_params->modules[_RX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_RING_RX)
                {
                    ring_rx_bind_info.info = pipeline_info;
                
                    ret = adk_sif_cfg_pipeline_ring_rx_bind(cfg_info_hdl, pipeline_params, &ring_rx_bind_info);
                    if (ret != ADK_SUCCESS)
                    {
                        ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                    }
                }
                else if (pipeline_params->modules[_RX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_BUFF_RX)
                {
                    buff_rx_bind_info.info = pipeline_info;
                
                    ret = adk_sif_cfg_pipeline_buff_rx_bind(cfg_info_hdl, pipeline_params, &buff_rx_bind_info);
                    if (ret != ADK_SUCCESS)
                    {
                        ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                    }
                }
                else if (pipeline_params->modules[_RX_IDX] ==
                            ADK_SIF_CFG_PIPELINE_MODULE_QAT_RX)
                {
                    ring_rx_bind_info.info = pipeline_info;
                    ret = adk_sif_cfg_pipeline_qat_rx_bind (
                                cfg_info_hdl,
                                pipeline_params,
                                &ring_rx_bind_info);
                    if (ret != ADK_SUCCESS)
                    {
                        ADK_PREINIT_WRITE ("Pipeline - %s configuration "
                                "failed\n", pipeline_params->name);
                        return ADK_STATUS_SEV_LEVEL_SET (
                                    ADK_E_INVALID_PARAMETER,
                                    ADK_SEV_LEVEL_ERROR);
                    }
                }


                /* Bind pipeline to Tx port/ring */
                if (pipeline_params->stage == ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR)
                {
                    if (pipeline_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX)
                    {
                        port_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_port_tx_bind(cfg_info_hdl, pipeline_params, &port_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                        }
                    }
                    else if (pipeline_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_RING_TX)
                    {
                        ring_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_ring_tx_bind(cfg_info_hdl, pipeline_params, &ring_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                        }
                    }
                    else if (pipeline_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_BUFF_TX)
                    {
                        buff_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_buff_tx_bind(cfg_info_hdl, pipeline_params, &buff_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                        }
                    }
                    else if (pipeline_params->modules[_TX_IDX] ==
                                    ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
                    {
                        ring_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_qat_tx_bind (
                                    cfg_info_hdl,
                                    pipeline_params,
                                    &ring_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE ("Pipeline - %s configuration "
                                    "failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET (
                                    ADK_E_INVALID_PARAMETER,
                                    ADK_SEV_LEVEL_ERROR);
                        }
                    }
                }

                prev_pipeline_params = pipeline_params;
                prev_pipeline_info = pipeline_info;
                break;

            case ADK_SIF_CFG_PIPELINE_STAGE_INTERMEDIATE:
            case ADK_SIF_CFG_PIPELINE_STAGE_TERMINATOR:
                /* Create pipeline */ 
                ret = adk_sif_cfg_pipeline_create(cfg_info_hdl, pipeline_params, &pipeline_info);
                if (ret != ADK_SUCCESS) 
                {
                    return ret;
                }
                if (pipeline_info == NULL)
                {
                    ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                }
               
                bind_info.src_info = prev_pipeline_info;
                bind_info.dst_info = pipeline_info; 

                /* Bind previous pipeline with current pipeline*/
                ret = adk_sif_cfg_pipeline_bind(cfg_info_hdl, prev_pipeline_params, pipeline_params, &bind_info);
                if (ret != ADK_SUCCESS)
                {
                    ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                }

                /* Bind pipeline to Tx port/ring */
                if (pipeline_params->stage == ADK_SIF_CFG_PIPELINE_STAGE_TERMINATOR)
                {
                    if (pipeline_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX)
                    {
                        port_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_port_tx_bind(cfg_info_hdl, pipeline_params, &port_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                        }
                    }
                    else if (pipeline_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_RING_TX)
                    {
                        ring_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_ring_tx_bind(cfg_info_hdl, pipeline_params, &ring_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                        }
                    }
                    else if (pipeline_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_BUFF_TX)
                    {
                        buff_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_buff_tx_bind(cfg_info_hdl, pipeline_params, &buff_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE("Pipeline - %s configuration failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                        }
                    }
                    else if (pipeline_params->modules[_TX_IDX] ==
                                    ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
                    {
                        ring_tx_bind_info.info = pipeline_info;

                        ret = adk_sif_cfg_pipeline_qat_tx_bind (
                                    cfg_info_hdl,
                                    pipeline_params,
                                    &ring_tx_bind_info);
                        if (ret != ADK_SUCCESS)
                        {
                            ADK_PREINIT_WRITE ("Pipeline - %s configuration "
                                    "failed\n", pipeline_params->name);
                            return ADK_STATUS_SEV_LEVEL_SET (
                                    ADK_E_INVALID_PARAMETER,
                                    ADK_SEV_LEVEL_ERROR);
                        }
                    }
                }

                prev_pipeline_params = pipeline_params;
                prev_pipeline_info = pipeline_info;
                break;

            default:
                ADK_PREINIT_WRITE("Pipeline - %s configuration failed. Unsupported pipeline stage type\n", 
                                  pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
    }

//#ifdef ADK_DEBUG
    /* Pipeline info debug dump */
    ret = adk_sif_cfg_pipeline_info_dump(cfg_info_hdl->pipeline_info_tbl);
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline configuration failed\n");
        return ret;
    }
//#endif

    ADK_PREINIT_WRITE("Pipelines configured\n");
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_create(adk_sif_cfg_info_t *cfg_info_hdl,
                            adk_cfg_pipeline_params_t *params,
                            adk_sif_cfg_pipeline_info_t **pipeline_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *info = NULL;

    /* Validate params and pipeline info pointers */
    if (!params || !pipeline_info)
    {
        ADK_PREINIT_WRITE("Pipeline [%s] create failed. \
                          Invalid params or pipeline info pointer\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    /* Validate pipeline lcore */
    ret = adk_sif_cfg_eal_lcore_check(cfg_info_hdl->eal_info_tbl, params->lcore); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline - %s create failed. Invalid lcore id %d\n", 
                          params->name, params->lcore);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    /* Allocate memory to store configuration info */
    ret = adk_sif_cfg_pipeline_info_alloc(cfg_info_hdl->pipeline_info_tbl, 
                                          params->lcore, 
                                          params->modules[_ENTRY_MOD_IDX], &info);
    if (ret != ADK_SUCCESS)
    {
        if (ret == ADK_E_SYS_ALREADY_INITIALIZED)
        {
            ret = ADK_SUCCESS;
        }
    }
    
    *pipeline_info = info;

    return ret;

 }

adk_status_t
adk_sif_cfg_pipeline_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                          adk_cfg_pipeline_params_t *src_params,
                          adk_cfg_pipeline_params_t *dest_params,
                          adk_sif_cfg_pipeline_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *src_info = NULL, *dest_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!src_params || !dest_params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    /* Validate module type and ring id */
    if (src_params->modules[_TX_IDX] != ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
    {
        if (src_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_RING_TX)
        {
            if (dest_params->modules[_RX_IDX] != ADK_SIF_CFG_PIPELINE_MODULE_RING_RX)
            {
                ADK_PREINIT_WRITE("Bind ring type mismatch\n");
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
        }
        else
        {
            ADK_PREINIT_WRITE("Unsupported bind ring type: %d\n", src_params->modules[_TX_IDX]);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }

        if (src_params->ring_q_out[0].ringid != dest_params->ring_q_in[0].ringid)
        {
            ADK_PREINIT_WRITE("Bind ring id mismatch: src ring id %d, dest ring id %d\n", 
                              src_params->ring_q_out[0].ringid, dest_params->ring_q_in[0].ringid);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
    }

    src_info = bind_info->src_info;
    dest_info = bind_info->dst_info;

    /* Update pipeline source stage */
    ret = adk_sif_cfg_pipeline_info_exit_module_set(cfg_info_hdl, src_params, 
                                                    dest_params, dest_info, &src_info);     
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline bind failed.\n");
        goto adk_sif_cfg_pipeline_bind_return;
    }

    /* Update pipeline destination stage */
    ret = adk_sif_cfg_pipeline_info_entry_module_set(cfg_info_hdl, dest_params, &dest_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline bind failed.\n");
        goto adk_sif_cfg_pipeline_bind_return;
    }

adk_sif_cfg_pipeline_bind_return:
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_port_rx_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                                  adk_cfg_pipeline_params_t *params,
                                  adk_sif_cfg_pipeline_port_rx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *port_rx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline port rx bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    port_rx_info = bind_info->info;

    ret = adk_sif_cfg_pipeline_info_entry_module_set(cfg_info_hdl, params, &port_rx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline port rx bind failed.\n");
        goto adk_sif_cfg_pipeline_port_rx_bind_return;
    }

adk_sif_cfg_pipeline_port_rx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_ring_rx_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                                  adk_cfg_pipeline_params_t *params,
                                  adk_sif_cfg_pipeline_ring_rx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *ring_rx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline ring rx bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }
    
    ring_rx_info = bind_info->info;

    ret = adk_sif_cfg_pipeline_info_entry_module_set(cfg_info_hdl, params, &ring_rx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline ring rx bind failed.\n");
        goto adk_sif_cfg_pipeline_ring_rx_bind_return;
    }

adk_sif_cfg_pipeline_ring_rx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_buff_rx_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                                  adk_cfg_pipeline_params_t *params,
                                  adk_sif_cfg_pipeline_buff_rx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *buff_rx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline Buff rx bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }
    
    buff_rx_info = bind_info->info;

    ret = adk_sif_cfg_pipeline_info_entry_module_set(cfg_info_hdl, params, &buff_rx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline buff rx bind failed.\n");
        goto adk_sif_cfg_pipeline_buff_rx_bind_return;
    }

adk_sif_cfg_pipeline_buff_rx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_qat_rx_bind (
    adk_sif_cfg_info_t *cfg_info_hdl,
    adk_cfg_pipeline_params_t *params,
    adk_sif_cfg_pipeline_ring_rx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *ring_rx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline QAT rx bind failed. Invalid bind "
                          "params or bind info pointer\n");

        return ADK_STATUS_SEV_LEVEL_SET (
                    ADK_E_INVALID_HANDLE,
                    ADK_SEV_LEVEL_ERROR);
    }
    
    ring_rx_info = bind_info->info;

    ret = adk_sif_cfg_pipeline_info_entry_module_set (
                cfg_info_hdl,
                params,
                &ring_rx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE ("QAT ring rx bind failed.\n");

        goto adk_sif_cfg_pipeline_qat_rx_bind_return;
    }

adk_sif_cfg_pipeline_qat_rx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_port_tx_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                                  adk_cfg_pipeline_params_t *params,
                                  adk_sif_cfg_pipeline_port_tx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *port_tx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline port tx bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    port_tx_info = bind_info->info;
    
    ret = adk_sif_cfg_pipeline_info_tx_exit_module_set(cfg_info_hdl, params, &port_tx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline port tx bind failed.\n");
        goto adk_sif_cfg_pipeline_port_tx_bind_return;
    }

adk_sif_cfg_pipeline_port_tx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_ring_tx_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                                  adk_cfg_pipeline_params_t *params,
                                  adk_sif_cfg_pipeline_ring_tx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *ring_tx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline ring tx bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    ring_tx_info = bind_info->info;
    
    ret = adk_sif_cfg_pipeline_info_tx_exit_module_set(cfg_info_hdl, params, &ring_tx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline ring tx bind failed.\n");
        goto adk_sif_cfg_pipeline_ring_tx_bind_return;
    }

adk_sif_cfg_pipeline_ring_tx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_buff_tx_bind(adk_sif_cfg_info_t *cfg_info_hdl,
                                  adk_cfg_pipeline_params_t *params,
                                  adk_sif_cfg_pipeline_buff_tx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *buff_tx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE("Pipeline Buff tx bind failed. Invalid bind params or bind info pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    buff_tx_info = bind_info->info;
    
    ret = adk_sif_cfg_pipeline_info_tx_exit_module_set(cfg_info_hdl, params, &buff_tx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline buff tx bind failed.\n");
        goto adk_sif_cfg_pipeline_buff_tx_bind_return;
    }

adk_sif_cfg_pipeline_buff_tx_bind_return:    
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_qat_tx_bind (
        adk_sif_cfg_info_t *cfg_info_hdl,
        adk_cfg_pipeline_params_t *params,
        adk_sif_cfg_pipeline_ring_tx_bind_info_t *bind_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *ring_tx_info = NULL;

    /* Validate bind params and bind info pointers */
    if (!params || !bind_info)
    {
        ADK_PREINIT_WRITE ("Pipeline QAT tx bind failed. Invalid bind "
                "params or bind info pointer\n");

        return ADK_STATUS_SEV_LEVEL_SET (
                ADK_E_INVALID_HANDLE,
                ADK_SEV_LEVEL_ERROR);
    }

    ring_tx_info = bind_info->info;
    
    ret = adk_sif_cfg_pipeline_info_tx_exit_module_set (
                cfg_info_hdl,
                params,
                &ring_tx_info); 
    if (ret != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Pipeline QAT tx bind failed.\n");

        goto adk_sif_cfg_pipeline_qat_tx_bind_return;
    }

adk_sif_cfg_pipeline_qat_tx_bind_return:
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_entry_module_set(adk_sif_cfg_info_t *cfg_info_hdl,
                                           adk_cfg_pipeline_params_t *params,
                                           adk_sif_cfg_pipeline_info_t **info)
{
    adk_status_t ret = ADK_SUCCESS;

    switch(params->modules[_RX_IDX])
    {
        case ADK_SIF_CFG_PIPELINE_MODULE_RING_RX:
            ret = adk_sif_cfg_pipeline_info_ring_q_set(cfg_info_hdl,
                                                       params->ring_q_in,
                                                       params->num_ring_q_in,
                                                       params->entry_func,
                                                       (void *)*info,
                                                       (*info)->entry_module.ring_q_in,
                                                       &(*info)->entry_module.num_ring_q_in);
            if (ret != ADK_SUCCESS)
            {
                return ret;
            }
            break;

        case ADK_SIF_CFG_PIPELINE_MODULE_PORT_RX:
            ret = adk_sif_cfg_pipeline_info_port_q_set(cfg_info_hdl,
                                                       params->port_q_in,
                                                       params->num_port_q_in,
                                                       params->entry_func,
                                                       (void *)*info,
                                                       (*info)->entry_module.port_q_in,
                                                       &(*info)->entry_module.num_port_q_in); 
            if (ret != ADK_SUCCESS)
            {
                return ret;
            }
            break;

        case ADK_SIF_CFG_PIPELINE_MODULE_BUFF_RX:
            ret = adk_sif_cfg_pipeline_info_buff_set(cfg_info_hdl,
                                                       params->buff_in,
                                                       params->num_buff_in,
                                                       params->entry_func,
                                                       (*info)->entry_module.buff_in,
                                                       &(*info)->entry_module.num_buff_in);
            if (ret != ADK_SUCCESS)
            {
                return ret;
            }
            break;

        case ADK_SIF_CFG_PIPELINE_MODULE_QAT_RX:
            (*info)->entry_module.qat_rx = params->qat_rx;
            break;

        default:
            ADK_PREINIT_WRITE("Invalid rx ring type %d\n", params->modules[_RX_IDX]);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_exit_module_set(adk_sif_cfg_info_t *cfg_info_hdl,
                                          adk_cfg_pipeline_params_t *src_params,
                                          adk_cfg_pipeline_params_t *dest_params,
                                          adk_sif_cfg_pipeline_info_t *dest_info,
                                          adk_sif_cfg_pipeline_info_t **src_info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_dest_module_t  *src_info_exit_module_dest = NULL;
    uint32_t src_params_exit_module = 0, dest_params_entry_module = 0;
    void *ring = NULL;
    
    /* If exit module is set to wildcard, then entry and exit modules are same*/
    if (src_params->modules[_EXIT_MOD_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_IGNORE)
    {
        src_params_exit_module = src_params->modules[_ENTRY_MOD_IDX];
    }
    else
    {
        src_params_exit_module = src_params->modules[_EXIT_MOD_IDX];
    }   
    dest_params_entry_module = dest_params->modules[_ENTRY_MOD_IDX];
    if (src_params_exit_module >= ADK_SIF_CFG_PIPELINE_MODULE_MAX)
    {
          return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
    src_info_exit_module_dest = &(*src_info)->exit_module[src_params_exit_module].dest_module[dest_params_entry_module];
   
    /* Update src info -> exit module -> destination module */
    if (src_params->modules[_TX_IDX] != ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
    {
        if (src_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_RING_TX)
        {
            src_info_exit_module_dest->pktio = ADK_SIF_CFG_PIPELINE_MODULE_RING_TX;
            
            ret = adk_sif_cfg_ring_info_get(cfg_info_hdl->ring_info_tbl, src_params->ring_q_out[0].ringid, &ring);
            if ((ret != ADK_SUCCESS) || (ring == NULL))
            {
                ADK_PREINIT_WRITE("Ring %d not configured\n", src_params->ring_q_out[0].ringid);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            src_info_exit_module_dest->ring = ring;
            src_info_exit_module_dest->ring_id = src_params->ring_q_out[0].ringid;
        }
        else
        {
            ADK_PREINIT_WRITE("Invalid tx ring type %d\n", src_params->modules[_TX_IDX]);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
    }
    src_info_exit_module_dest->valid = _CONFIG; 
    src_info_exit_module_dest->func_idx = _DEST_FUNC_RING;
    
    /* Point to next stage in the pipeline */
    src_info_exit_module_dest->next_entry = dest_info;
    //src_info_exit_module_dest->next_stage = &dest_info->exit_module[dest_params->modules[_EXIT_MOD_IDX]];

    /* Update pipeline stage which is unique per entry,exit module pair */
    if (src_params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
    {
        (*src_info)->exit_module[src_params_exit_module].qat_tx = TRUE;
    }
    (*src_info)->exit_module[src_params_exit_module].stage = src_params->stage;
    (*src_info)->exit_module[src_params_exit_module].valid = _CONFIG;

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_tx_exit_module_set(adk_sif_cfg_info_t *cfg_info_hdl __attribute__((unused)),  
                                             adk_cfg_pipeline_params_t *params, 
                                             adk_sif_cfg_pipeline_info_t **info)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_dest_module_t  *tx_info_exit_module_dest = NULL;
    uint32_t params_exit_module = 0, params_entry_module = 0;

    /* If exit module is set to wildcard, then entry and exit modules are same*/
    if (params->modules[_EXIT_MOD_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_IGNORE)
    {
        params_exit_module = params->modules[_ENTRY_MOD_IDX];
    }
    else
    {
        params_exit_module = params->modules[_EXIT_MOD_IDX];
    }
    if (params_exit_module >= ADK_SIF_CFG_PIPELINE_MODULE_MAX)
    {
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    } 
    params_entry_module = params->modules[_TX_IDX];
    tx_info_exit_module_dest = &(*info)->exit_module[params_exit_module].dest_module[params_entry_module];
   
    /* Update src info -> exit module -> destination module */ 
    if (params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX)
    {
        tx_info_exit_module_dest->pktio = ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX;
    }
    else if (params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_RING_TX)
    {
        tx_info_exit_module_dest->pktio = ADK_SIF_CFG_PIPELINE_MODULE_RING_TX;
    }
    else if (params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_BUFF_TX)
    {
        tx_info_exit_module_dest->pktio = ADK_SIF_CFG_PIPELINE_MODULE_BUFF_TX;
    }
    else if (params->modules[_TX_IDX] == ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
    {
        tx_info_exit_module_dest->pktio = ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid tx type %d\n", params->modules[_TX_IDX]);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
    tx_info_exit_module_dest->valid = _CONFIG; 
    
    /* Point to next stage in the pipeline */
    tx_info_exit_module_dest->next_entry = NULL;

    /* Update pipeline stage which is unique per entry,exit module pair */
    (*info)->exit_module[params_exit_module].stage = params->stage;
    (*info)->exit_module[params_exit_module].valid = _CONFIG;

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_alloc(adk_sif_cfg_pipeline_info_tbl_t *pipeline_info_hdl,
                                uint32_t lcore_id,
                                uint32_t entry_mod,
                                adk_sif_cfg_pipeline_info_t **info)
{
    adk_sif_cfg_pipeline_info_t *p = pipeline_info_hdl->pipeline_info[lcore_id];

    if (p[entry_mod].valid == _INIT)
    {
        p[entry_mod].valid = _CONFIG;
        *info = &p[entry_mod];
        pipeline_info_hdl->size++;
        return ADK_SUCCESS;
    }
    else
    {
        *info = &p[entry_mod];
        return ADK_E_SYS_ALREADY_INITIALIZED;
    }
}

adk_status_t
adk_sif_cfg_pipeline_info_free(adk_sif_cfg_pipeline_info_tbl_t *pipeline_info_hdl,
                               adk_sif_cfg_pipeline_info_t *info)
{
    adk_status_t ret = ADK_SUCCESS;

    if (info == NULL)
    {
        ADK_PREINIT_WRITE("Invalid memory access\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
    }

    ADK_MEMSET(info, 0, sizeof(adk_sif_cfg_pipeline_info_t));
    info->valid = _INIT;
    pipeline_info_hdl->size--;

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_rewind(adk_sif_cfg_pipeline_info_tbl_t *pipeline_info_hdl)
{
    adk_status_t ret = ADK_SUCCESS;

    pipeline_info_hdl->cur = 0;

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_port_q_set(adk_sif_cfg_info_t *cfg_info_hdl,
                                     adk_sif_cfg_pipeline_port_q_t *params_pq,
                                     uint32_t num_params_pq,
                                     uint16_t entry_func,
                                     void *entry_mod_hdl,
                                     adk_sif_cfg_pipeline_port_q_t *info_pq,
                                     uint32_t *num_info_pq)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t p = 0, q = 0, p_idx = 0, q_idx = 0;

    for (p = 0; p < num_params_pq; p++)
    {
        ret = adk_sif_cfg_port_check(cfg_info_hdl->port_info_tbl, params_pq[p].portid);
        if (ret != ADK_SUCCESS)
        {
            ADK_PREINIT_WRITE("Port %d not configured\n", params_pq[p].portid);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }

        ret = adk_sif_cfg_pipeline_info_port_idx_get(params_pq[p].portid,
                                                info_pq, num_info_pq, &p_idx);
        if (ret == ADK_E_INVALID_PARAMETER)
        {
            return ret;
        }
        if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
        {
            info_pq[p_idx].portid = params_pq[p].portid;
            ret = ADK_SUCCESS;
        }

        for (q = 0; q < params_pq[p].num_queues; q++)
        {
            ret = adk_sif_cfg_port_rx_queue_check(cfg_info_hdl->port_info_tbl, 
                                 params_pq[p].portid, params_pq[p].queueid[q]);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Port %d Queue %d not configured\n", 
                                  params_pq[p].portid, params_pq[p].queueid[q]);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            ret = adk_sif_cfg_pipeline_info_queue_idx_get(params_pq[p].queueid[q],
                            info_pq[p_idx].queueid, &info_pq[p_idx].num_queues, &q_idx);
            if (ret == ADK_E_INVALID_PARAMETER)
            {
                return ret;
            }
            if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
            {
                info_pq[p_idx].queueid[q_idx] = params_pq[p].queueid[q];
                info_pq[p_idx].weight[q_idx] = params_pq[p].weight[q];
                info_pq[p_idx].entry_func[q_idx] = entry_func;
                info_pq[p_idx].entry_mod_hdl[q_idx] = entry_mod_hdl;
                ret = ADK_SUCCESS;
            }
        }
    }

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_ring_q_set(adk_sif_cfg_info_t *cfg_info_hdl,
                                     adk_sif_cfg_pipeline_ring_q_t *params_rq,
                                     uint32_t num_params_rq,
                                     uint16_t entry_func,
                                     void *entry_mod_hdl,
                                     adk_sif_cfg_pipeline_ring_q_t *info_rq,
                                     uint32_t *num_info_rq)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t r = 0, r_idx = 0;
    void *ring = NULL;

    for (r = 0; r < num_params_rq; r++)
    {
        ret = adk_sif_cfg_ring_info_get(cfg_info_hdl->ring_info_tbl, params_rq[r].ringid, &ring);
        if ((ret != ADK_SUCCESS) || (ring == NULL))
        {
            ADK_PREINIT_WRITE("Ring %d not configured\n", params_rq[r].ringid);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }

        ret = adk_sif_cfg_pipeline_info_ring_idx_get(params_rq[r].ringid,
                                            info_rq, num_info_rq, &r_idx);
        if (ret == ADK_E_INVALID_PARAMETER)
        {
            return ret;
        }
        if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
        {
            info_rq[r_idx].ringid = params_rq[r].ringid;
            info_rq[r_idx].ringhdl = ring;
            info_rq[r_idx].entry_func = entry_func;
            info_rq[r_idx].entry_mod_hdl = entry_mod_hdl;
            ret = ADK_SUCCESS;
        }
    }

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_buff_set(
        adk_sif_cfg_info_t *cfg_info_hdl,
        adk_sif_cfg_pipeline_buff_t *params_rq,
        uint32_t num_params_rq,
        uint16_t entry_func,
        adk_sif_cfg_pipeline_buff_t *info_rq,
        uint32_t *num_info_rq)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t b = 0, b_idx = 0;
    adk_sif_cfg_pkt_buffs_info_t *buff_info = NULL;

    for (b = 0; b < num_params_rq; b++)
    {
        ret = adk_sif_cfg_pkt_buffs_info_get(
                                    cfg_info_hdl->pkt_buffs_info_tbl,
                                    params_rq[b].buff_id, &buff_info);

        if (ret != ADK_SUCCESS)
        {
            ADK_PREINIT_WRITE("Buff %d not configured\n", params_rq[b].buff_id);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }

        ret = adk_sif_cfg_pipeline_info_buff_idx_get(params_rq[b].buff_id,
                                            info_rq, num_info_rq, &b_idx);
        if (ret == ADK_E_INVALID_PARAMETER)
        {
            return ret;
        }
        if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
        {
            info_rq[b_idx].buff_id = buff_info->pkt_buffs_id;
            info_rq[b_idx].entry_func = entry_func;
            info_rq[b_idx].transmit = buff_info->pkt_buffs_transmit;
            info_rq[b_idx].weight = buff_info->pkt_buffs_weight;
            ret = ADK_SUCCESS;
        }
    }

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_port_idx_get(const uint8_t portid,
                                       adk_sif_cfg_pipeline_port_q_t *p,
                                       uint32_t *n,
                                       uint32_t *idx)
{
    uint32_t i = 0;

    for (i = 0; i < *n; i++)
    {
        if (p[i].portid == portid)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_PORT)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }

    ADK_PREINIT_WRITE("Number of port exceeds max supported %d\n", ADK_SIF_CFG_MAX_PORT);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_sif_cfg_pipeline_info_queue_idx_get(const uint16_t queueid,
                                        uint16_t *q,
                                        uint32_t *n,
                                        uint32_t *idx)
{
    uint32_t i = 0;

    for (i = 0; i < *n; i++)
    {
        if (q[i] == queueid)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_QUEUES_PER_PORT)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }

    ADK_PREINIT_WRITE("Number of port queues exceeds max supported %d\n", ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_sif_cfg_pipeline_info_ring_idx_get(const uint16_t ringid,
                                       adk_sif_cfg_pipeline_ring_q_t *r,
                                       uint32_t *n,
                                       uint32_t *idx)
{
    uint32_t i = 0;

    for (i = 0; i < *n; i++)
    {
        if (r[i].ringid == ringid)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_RING_QUEUES)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }

    ADK_PREINIT_WRITE("Number of ring queues exceeds max supported %d\n", ADK_SIF_CFG_MAX_RING_QUEUES);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_sif_cfg_pipeline_info_buff_idx_get(const uint16_t buff_id,
                                       adk_sif_cfg_pipeline_buff_t *r,
                                       uint32_t *n,
                                       uint32_t *idx)
{
    uint32_t i = 0;

    for (i = 0; i < *n; i++)
    {
        if (r[i].buff_id == buff_id)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_PKT_BUFFS)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }

    ADK_PREINIT_WRITE("Number of buffers exceeds max supported %d\n", ADK_SIF_CFG_MAX_PKT_BUFFS);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_sif_cfg_pipeline_info_get(adk_sif_cfg_pipeline_info_tbl_t *pipeline_info_hdl,
                              uint32_t lcore_id,
                              uint32_t entry_mod,
                              adk_sif_cfg_pipeline_info_t **info)
{
    adk_status_t ret = ADK_SUCCESS;

    /* Validation */
    if (lcore_id >= ADK_MAX_NUM_CORES)
    {
        ADK_PREINIT_WRITE("Lcore id %d exceeds max supported %d\n", lcore_id, ADK_MAX_NUM_CORES);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (entry_mod >= ADK_SIF_CFG_PIPELINE_MODULE_MAX)
    {
        ADK_PREINIT_WRITE("Entry module id %d exceeds max supported %d\n", entry_mod, ADK_SIF_CFG_PIPELINE_MODULE_MAX);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    /* Fetch config info */
    *info = &pipeline_info_hdl->pipeline_info[lcore_id][entry_mod];

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_info_dest_get(adk_sif_cfg_pipeline_dest_module_t  *dest_module)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;

    for (i = 0; i < ADK_SIF_CFG_PIPELINE_MODULE_MAX; i++)
    {
        if (dest_module[i].valid == _CONFIG)
        {
            dest_module = &dest_module[i];
        }
    }

    if (i == ADK_SIF_CFG_PIPELINE_MODULE_MAX)
    {
        return ADK_E_LIST_RANGE;
    }

    return ret;
}

/* adk_status_t */
/* adk_sif_cfg_pipeline_lcore_entry_in_get(int32_t coreIndex, */ 
/*                                               adk_sif_cfg_pipeline_port_q_t *portq_in, uint32_t *num_portq_in, */
/*                                               adk_sif_cfg_pipeline_ring_q_t *ringq_in, uint32_t *num_ringq_in, */
/*                                               adk_sif_cfg_pipeline_buff_t   *buff_in, uint32_t *num_buff_in, */
/*                                               bool *qat_rx_decrypt, bool *qat_rx_encrypt) */
adk_status_t
adk_sif_cfg_pipeline_lcore_entry_in_get(uint32_t coreIndex, adk_lcore_input_receive_t *lcore_receive)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_port_q_t *pq_in = lcore_receive->portq_in;
    adk_sif_cfg_pipeline_ring_q_t *rq_in = lcore_receive->ringq_in;
    adk_sif_cfg_pipeline_buff_t *bf_in = lcore_receive->buffs_in;
    adk_sif_cfg_pipeline_info_t *pipeline_info = NULL;
    adk_sif_cfg_pipeline_entry_module_t *entry_module = NULL;
    uint32_t pq_in_idx = 0, rq_in_idx = 0, bf_in_idx = 0;
    uint32_t entry_mod = 0, p = 0, q = 0, r = 0, b = 0, p_idx = 0, q_idx = 0, r_idx = 0, b_idx = 0;
    bool *qat_rx_decrypt = &lcore_receive->qat_rx_decrypt;
    bool *qat_rx_encrypt = &lcore_receive->qat_rx_encrypt;
    uint32_t *num_portq_in = &lcore_receive->num_portq_in; 
    uint32_t *num_ringq_in = &lcore_receive->num_ringq_in; 
    uint32_t *num_buff_in = &lcore_receive->num_buffs_in; 

    ADK_MEMSET(pq_in, 0, (sizeof(adk_sif_cfg_pipeline_port_q_t) * ADK_SIF_CFG_MAX_PORT));
    ADK_MEMSET(rq_in, 0, (sizeof(adk_sif_cfg_pipeline_ring_q_t) * ADK_SIF_CFG_MAX_RING_QUEUES));
    ADK_MEMSET(bf_in, 0, (sizeof(adk_sif_cfg_pipeline_buff_t) * ADK_SIF_CFG_MAX_PKT_BUFFS));

    for (entry_mod = 0; entry_mod < ADK_SIF_CFG_PIPELINE_MODULE_MAX; entry_mod++)
    {
        ret = adk_sif_cfg_pipeline_info_get(adkPipelineInfoTbl_g, coreIndex, entry_mod, &pipeline_info);
        if (ret != ADK_SUCCESS)
        {
            return ret;
        }

        if (pipeline_info->valid == _INIT)
        {
            continue;
        }
        
        entry_module = &(pipeline_info->entry_module);
        
        if ((lcore_receive->types.all) || (lcore_receive->types.port))
        {
            for (p = 0; p < entry_module->num_port_q_in; p++)
            {
                ret = adk_sif_cfg_pipeline_info_port_idx_get(entry_module->port_q_in[p].portid,
                        pq_in, &pq_in_idx, &p_idx);
                if (ret == ADK_E_INVALID_PARAMETER)
                {
                    return ret;
                }
                if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
                {
                    pq_in[p_idx].portid =  entry_module->port_q_in[p].portid;
                    ret = ADK_SUCCESS;
                }

                for (q = 0; q < entry_module->port_q_in[p].num_queues; q++)
                {
                    ret = adk_sif_cfg_pipeline_info_queue_idx_get(entry_module->port_q_in[p].queueid[q],
                            pq_in[p_idx].queueid, &pq_in[p_idx].num_queues, &q_idx);
                    if (ret == ADK_E_INVALID_PARAMETER)
                    {
                        return ret;
                    }
                    if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
                    {
                        pq_in[p_idx].queueid[q_idx] = entry_module->port_q_in[p].queueid[q];
                        pq_in[p_idx].entry_func[q_idx] = entry_module->port_q_in[p].entry_func[q];
                        pq_in[p_idx].entry_mod_hdl[q_idx] = entry_module->port_q_in[p].entry_mod_hdl[q];
                        pq_in[p_idx].weight[q_idx] = entry_module->port_q_in[p].weight[q];
                        ret = ADK_SUCCESS;
                    }
                }
            }
        } 
        
        if ((lcore_receive->types.all) || (lcore_receive->types.ring))
        {
            for (r = 0; r < entry_module->num_ring_q_in; r++)
            {
                ret = adk_sif_cfg_pipeline_info_ring_idx_get(entry_module->ring_q_in[r].ringid,
                        rq_in, &rq_in_idx, &r_idx);
                if (ret == ADK_E_INVALID_PARAMETER)
                {
                    return ret;
                }
                if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
                {
                    rq_in[r_idx].ringid = entry_module->ring_q_in[r].ringid;
                    rq_in[r_idx].ringhdl = entry_module->ring_q_in[r].ringhdl;
                    rq_in[r_idx].entry_func = entry_module->ring_q_in[r].entry_func;
                    rq_in[r_idx].entry_mod_hdl = entry_module->ring_q_in[r].entry_mod_hdl;
                    ret = ADK_SUCCESS;
                }
            }
        }

        if ((lcore_receive->types.all) || (lcore_receive->types.buff))
        {
            for (b = 0; b < entry_module->num_buff_in; b++)
            {
                ret = adk_sif_cfg_pipeline_info_buff_idx_get(entry_module->buff_in[b].buff_id,
                        bf_in, &bf_in_idx, &b_idx);
                if (ret == ADK_E_INVALID_PARAMETER)
                {
                    return ret;
                }
                if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
                {
                    bf_in[b_idx].buff_id = entry_module->buff_in[b].buff_id;
                    bf_in[b_idx].entry_func = entry_module->buff_in[b].entry_func;
                    bf_in[b_idx].transmit = entry_module->buff_in[b].transmit;
                    bf_in[b_idx].weight = entry_module->buff_in[b].weight ;
                    ret = ADK_SUCCESS;
                }
            }
        }

        if (ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI_FROM_CRYPTO == entry_mod)
        {
            if (entry_module->qat_rx == TRUE)
            {
                *qat_rx_decrypt = TRUE;
            }
        }

        if (ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI_FROM_CRYPTO == entry_mod)
        {
            if (entry_module->qat_rx == TRUE)
            {
                *qat_rx_encrypt = TRUE;
            }
        }
    }
    
    if (lcore_receive->types.all)
    {
        *num_portq_in = pq_in_idx;
        *num_ringq_in = rq_in_idx;
        *num_buff_in  = bf_in_idx;
    }
    else 
    {
        if(lcore_receive->types.port)
        {
            *num_portq_in = pq_in_idx;
        }

        if(lcore_receive->types.ring)
        {
            *num_ringq_in = rq_in_idx;
        }

        if(lcore_receive->types.buff)
        {
            *num_buff_in  = bf_in_idx;
        }
    }
    return ADK_SUCCESS;
}

adk_status_t
adk_sif_cfg_pipeline_lcore_entry_dump (uint32_t coreIndex, adk_lcore_input_receive_t *lcore_receive)
{
    uint32_t p = 0, q = 0, r = 0;
    char tmp_name[ADK_SIF_CFG_MAX_NAME_LEN];

    adk_sif_cfg_pipeline_port_q_t *portq_in = lcore_receive->portq_in;
    adk_sif_cfg_pipeline_ring_q_t *ringq_in = lcore_receive->ringq_in;
    adk_sif_cfg_pipeline_buff_t *buff_in = lcore_receive->buffs_in;
    uint32_t num_portq_in = lcore_receive->num_portq_in;
    uint32_t num_ringq_in = lcore_receive->num_ringq_in;
    uint32_t num_buff_in  = lcore_receive->num_buffs_in;
    bool qat_rx_decrypt = lcore_receive->qat_rx_decrypt;
    bool qat_rx_encrypt = lcore_receive->qat_rx_encrypt;

    ADK_PREINIT_WRITE("Lcore: %d\n", coreIndex);
    if ((lcore_receive->types.all) || (lcore_receive->types.port))
    {
        ADK_PREINIT_WRITE("\tPort, Queues, Entry function: ");
        for (p = 0; p < num_portq_in; p++)
        {
            for (q = 0; q < portq_in[p].num_queues; q++)
            {
                adk_sif_cfg_pipeline_entry_func_name_get(portq_in[p].entry_func[q], tmp_name);
                ADK_PREINIT_WRITE("(%d, %d, %s) ", portq_in[p].portid, portq_in[p].queueid[q], tmp_name); 
            }
        }
        ADK_PREINIT_WRITE("\n\n");
    }
    
    if ((lcore_receive->types.all) || (lcore_receive->types.ring))
    {
        ADK_PREINIT_WRITE("\tRing Id, Ring Hdl, Entry function: ");
        for (r = 0; r < num_ringq_in ; r++)
        {
            adk_sif_cfg_pipeline_entry_func_name_get(ringq_in[r].entry_func, tmp_name);
            ADK_PREINIT_WRITE("(%d, %p, %s) ", ringq_in[r].ringid, ringq_in[r].ringhdl, tmp_name);
        }
        ADK_PREINIT_WRITE("\n\n");
    }
    
    if ((lcore_receive->types.all) || (lcore_receive->types.buff))
    {
        ADK_PREINIT_WRITE("\tBuffer Id, Buff Hdl, Entry function: ");
        for (r = 0; r < num_buff_in ; r++)
        {
            adk_sif_cfg_pipeline_entry_func_name_get(buff_in[r].entry_func, tmp_name);
            ADK_PREINIT_WRITE("(%d, %s) ", buff_in[r].buff_id, tmp_name);
        }
        ADK_PREINIT_WRITE("\n\n");
    }
    ADK_PREINIT_WRITE("QAT_RX: %s", (((qat_rx_decrypt == TRUE) || (qat_rx_encrypt == TRUE))?"TRUE" : "FALSE"));

    ADK_PREINIT_WRITE("\n\n");

    return ADK_SUCCESS;
}

adk_status_t
adk_sif_cfg_pipeline_info_dump(adk_sif_cfg_pipeline_info_tbl_t *pipeline_info_hdl)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sif_cfg_pipeline_info_t *info = NULL;
    uint32_t lcore_idx = 0, entry_idx = 0, exit_idx = 0, dest_idx = 0;
    uint32_t p = 0, q = 0, r = 0;
    char entry_name[ADK_SIF_CFG_MAX_NAME_LEN], exit_name[ADK_SIF_CFG_MAX_NAME_LEN];
    char stage_name[ADK_SIF_CFG_MAX_NAME_LEN], dest_name[ADK_SIF_CFG_MAX_NAME_LEN];
    char tmp_name[ADK_SIF_CFG_MAX_NAME_LEN];

    /* Validate info handle pointer */
    if (!pipeline_info_hdl)
    {
        ADK_PREINIT_WRITE("Invalid pipeline info handle pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }
    
    adk_sif_cfg_pipeline_info_rewind(pipeline_info_hdl);

    for (lcore_idx = 0; lcore_idx < ADK_MAX_NUM_CORES; lcore_idx++)
    {
        for (entry_idx = 1; entry_idx < ADK_SIF_CFG_PIPELINE_MODULE_MAX; entry_idx++)
        {
            ret = adk_sif_cfg_pipeline_info_get(pipeline_info_hdl, lcore_idx, entry_idx, &info);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Pipeline consistency check failed\n");
                return ret;
            }

            if (info->valid == _INIT)
            {
                continue;
            }
            
            /* [Lcore][Entry module] info */
            adk_sif_cfg_pipeline_module_name_get((enum adk_sif_cfg_pipeline_modules_e)entry_idx, entry_name);
            ADK_PREINIT_WRITE("Lcore: %d\n\n", lcore_idx);
            ADK_PREINIT_WRITE("\tEntry module: %s\n\n", entry_name);
            if (info->entry_module.num_port_q_in > 0 )
            {
                ADK_PREINIT_WRITE("\t\tPort, Queues, Entry function: ");
                for (p = 0; p < info->entry_module.num_port_q_in; p++)
                {
                    for (q = 0; q < info->entry_module.port_q_in[p].num_queues; q++)
                    {
                        adk_sif_cfg_pipeline_entry_func_name_get(
                                info->entry_module.port_q_in[p].entry_func[q], tmp_name);
                        ADK_PREINIT_WRITE("(%d, %d, %s) ", info->entry_module.port_q_in[p].portid, 
                                info->entry_module.port_q_in[p].queueid[q], tmp_name);
                    }
                }
                ADK_PREINIT_WRITE("\n\n");
            }

            if (info->entry_module.num_ring_q_in > 0 )
            {
                ADK_PREINIT_WRITE("\t\tRing Id, Ring Hdl, Entry function: ");
                for (r = 0; r < info->entry_module.num_ring_q_in; r++)
                {
                    adk_sif_cfg_pipeline_entry_func_name_get(
                            info->entry_module.ring_q_in[r].entry_func, tmp_name);
                    ADK_PREINIT_WRITE("(%d, %p, %s) ", info->entry_module.ring_q_in[r].ringid, 
                            info->entry_module.ring_q_in[r].ringhdl, tmp_name);
                }
                ADK_PREINIT_WRITE("\n\n");
            } 

            if (info->entry_module.num_buff_in > 0 )
            {
                ADK_PREINIT_WRITE("\t\tBuff Id, Weight, Entry function: ");
                for (r = 0; r < info->entry_module.num_buff_in; r++)
                {
                    adk_sif_cfg_pipeline_entry_func_name_get(
                            info->entry_module.buff_in[r].entry_func, tmp_name);
                    ADK_PREINIT_WRITE("(%d, %d,%s) ",
                            info->entry_module.buff_in[r].buff_id,
                            info->entry_module.buff_in[r].weight,
                            tmp_name);
                }
                ADK_PREINIT_WRITE("\n\n");
            } 

            for (exit_idx = 1; exit_idx < ADK_SIF_CFG_PIPELINE_MODULE_MAX; exit_idx++)
            {
                if (info->exit_module[exit_idx].valid == _INIT)
                {
                    continue;
                }

                adk_sif_cfg_pipeline_module_name_get((adk_sif_cfg_pipeline_modules_t)exit_idx, exit_name);
                adk_sif_cfg_pipeline_stage_name_get(info->exit_module[exit_idx].stage, stage_name);
                ADK_PREINIT_WRITE("\t\tEntry module: %s - Exit module: %s - Stage: %s\n\n", entry_name, exit_name, stage_name);
               
                ADK_PREINIT_WRITE("\t\t\tDestination Module\t      Pktio       \tFunc Idx\tRing Id\t   Ring Hdl     \n");
                ADK_PREINIT_WRITE("\t\t\t------------------\t------------------\t--------\t-------\t----------------\n");
                for (dest_idx = 1; dest_idx < ADK_SIF_CFG_PIPELINE_MODULE_MAX; dest_idx++)
                {
                    if (info->exit_module[exit_idx].dest_module[dest_idx].valid == _INIT)
                    {
                        continue;
                    }
                    adk_sif_cfg_pipeline_module_name_get((adk_sif_cfg_pipeline_modules_t)dest_idx, dest_name);
                    adk_sif_cfg_pipeline_module_name_get(info->exit_module[exit_idx].dest_module[dest_idx].pktio, tmp_name);
                    ADK_PREINIT_WRITE("\t\t\t%18s\t%18s\t%8d\t%7d\t%16p\n", 
                                      dest_name, 
                                      tmp_name, 
                                      info->exit_module[exit_idx].dest_module[dest_idx].func_idx,
                                      info->exit_module[exit_idx].dest_module[dest_idx].ring_id,
                                      info->exit_module[exit_idx].dest_module[dest_idx].ring);
                }
            }

        }
    }

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_module_name_get(adk_sif_cfg_pipeline_modules_t mod, char *name)
{
    adk_status_t ret = ADK_SUCCESS;

    switch (mod)
    {
        case ADK_SIF_CFG_PIPELINE_MODULE_PORT_RX:
        ADK_STRNCPY(name, "PORT_RX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX:
        ADK_STRNCPY(name, "PORT_TX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_RING_RX:
        ADK_STRNCPY(name, "RING_RX", ADK_SIF_CFG_MAX_NAME_LEN);
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_RING_TX:
        ADK_STRNCPY(name, "RING_TX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_PKTIO_RX:
        ADK_STRNCPY(name, "PKTIO_RX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_PKTIO_TX:
        ADK_STRNCPY(name, "PKTIO_TX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_BUFF_RX:
        ADK_STRNCPY(name, "BUFF_RX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_BUFF_TX:
        ADK_STRNCPY(name, "BUFF_TX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_QAT_RX:
        ADK_STRNCPY(name, "QAT_RX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX:
        ADK_STRNCPY(name, "QAT_TX", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_ETHERNET_LLI:
        ADK_STRNCPY(name, "ETHERNET_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_ETHERNET_ULI:
        ADK_STRNCPY(name, "ETHERNET_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPV4_LLI:
        ADK_STRNCPY(name, "IPV4_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPV4_ULI:
        ADK_STRNCPY(name, "IPV4_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPV6_LLI:
        ADK_STRNCPY(name, "IPV6_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPV6_ULI:
        ADK_STRNCPY(name, "IPV6_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI:
        ADK_STRNCPY(name, "IPSEC_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI_TO_CRYPTO:
        ADK_STRNCPY(name, "IPSEC_LLI_TO_CRYPTO", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI_FROM_CRYPTO:
        ADK_STRNCPY(name, "IPSEC_LLI_FROM_CRYPTO", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_L4_LLI:
        ADK_STRNCPY(name, "L4_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_L4_ULI:
        ADK_STRNCPY(name, "L4_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI:
        ADK_STRNCPY(name, "IPSEC_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI_TO_CRYPTO:
        ADK_STRNCPY(name, "IPSEC_ULI_TO_CRYPTO", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI_FROM_CRYPTO:
        ADK_STRNCPY(name, "IPSEC_ULI_FROM_CRYPTO", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_GTPU_LLI:
        ADK_STRNCPY(name, "GTPU_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_GTPU_ULI:
        ADK_STRNCPY(name, "GTPU_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

 
        case ADK_SIF_CFG_PIPELINE_MODULE_BFD_LLI:
        ADK_STRNCPY(name, "BFD_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_BFD_ULI:
        ADK_STRNCPY(name, "BFD_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;
        
        case ADK_SIF_CFG_PIPELINE_MODULE_ETH_OAM_LLI:
        ADK_STRNCPY(name, "ETH_OAM_LLI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_ETH_OAM_ULI:
        ADK_STRNCPY(name, "ETH_OAM_ULI", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;

        case ADK_SIF_CFG_PIPELINE_MODULE_IGNORE:
        ADK_STRNCPY(name, "*", ADK_SIF_CFG_MAX_NAME_LEN); 
        break;
       
        default:
        ADK_STRNCPY(name, "INVALID", ADK_SIF_CFG_MAX_NAME_LEN); 
    }

    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_stage_name_get(adk_sif_cfg_pipeline_stage_t stage, char *name)
{
    adk_status_t ret = ADK_SUCCESS;

    switch(stage)
    {
        case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR:
            ADK_STRNCPY(name, "INITIATOR", ADK_SIF_CFG_MAX_NAME_LEN); 
            break;

        case ADK_SIF_CFG_PIPELINE_STAGE_INTERMEDIATE:
            ADK_STRNCPY(name, "INTERMEDIATE", ADK_SIF_CFG_MAX_NAME_LEN); 
            break;

        case ADK_SIF_CFG_PIPELINE_STAGE_TERMINATOR:
            ADK_STRNCPY(name, "TERMINATOR", ADK_SIF_CFG_MAX_NAME_LEN); 
            break;

        case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR:
            ADK_STRNCPY(name, "INITIATOR_TERMINATOR", ADK_SIF_CFG_MAX_NAME_LEN); 
            break;

        default:
            ADK_STRNCPY(name, "INVALID", ADK_SIF_CFG_MAX_NAME_LEN); 

    }
    return ret;
}

adk_status_t
adk_sif_cfg_pipeline_entry_func_name_get(uint16_t entry_func, char *name)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;

    for (i = 0; i < ADK_SIF_CFG_PIPELINE_ENTRY_FUNC_MAX; i++)
    {
        if (adkEntryFunctions_g[i] == entry_func)
        {
            break;
        }
    }

    switch(i)
    {
        case __FN_ETH_DP_PROCESS_PKT:
            ADK_STRNCPY(name, "ETHERNET_DP_PROCESS", ADK_SIF_CFG_MAX_NAME_LEN); 
            break;

         
#if 0
        case __FN_PKTIO_RX_DP_PROCESS_PKT:
            ADK_STRNCPY(name, "PKTIO_RX_DP_PROCESS", ADK_SIF_CFG_MAX_NAME_LEN); 
            break;
#endif
         
        default:
            ADK_STRNCPY(name, "INVALID", ADK_SIF_CFG_MAX_NAME_LEN); 
    }

    return ret;
}
