/* ***************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_hos.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _hos  SIF HOS Component
 * @{
 *
 * This file contains ADK HOS data structures, macros and API signatures
 *
 *  **********************************************************************/

#ifndef __ADK_HOS_INCLUDE_H__
#define __ADK_HOS_INCLUDE_H__

#include <adk_error.h>
#include <adk_sys.h>
#include <adk_hos_constants.h>

#define ADK_HOS_MAC_MAPPER_ID_INVALID INT_MAX

enum ies_acl_entry_state;

/*!
* =============================================================
*
* @typedef adk_hos_object_t
*
* @brief   HOS table identifier
*
* These enum values are used to identify the HOS table on which operations like
* add or delete rules can be done.
*
* =============================================================
*/

typedef enum adk_hos_object_s
{
    ADK_HOS_INVALID_OBJ = 0,
    /* L2 termination-interface table, assumed to be completely in HLP */
    ADK_HOS_L2_TERM_INTF_TBL,
    /* IPv4 clear text classifier table */
    ADK_HOS_IP4_CLASSIFIER_TBL,
    /* IPv4 clear text classifier table */
    ADK_HOS_IP6_CLASSIFIER_TBL,
    /* Supported ethertype table, assumed to be completely in HLP */
    ADK_HOS_SUPPORTED_ETHERTYPE_TBL,
    /* Supported ip protocol table, assumed to be completely in HLP */
    ADK_HOS_SUPPORTED_IPPROTO_TBL,
    /* Ethernet ingress flow table */
    ADK_HOS_L2_INGRESS_TBL,
     /* L3 termination IPv4/IPv6 outer/inner table, assumed to be completely in HLP */
    ADK_HOS_IP_TERM_TBL,
    /* SPD In table */
    ADK_HOS_SPD_IN_TBL,
    /* Post crypto SPD Out table */
    ADK_HOS_SPD_OUT_TBL,
    ADK_HOS_MAX_OBJ
} adk_hos_object_t;

typedef int adk_hos_rule_id_t;

typedef struct adk_hos_rule_stats_s
{
    uint64_t num_pkts;
    uint64_t num_octets;
} adk_hos_rule_stats_t;

/* HOS object specific adk_hos_*_rule_attr_t structure typecasted
 * to this common type */
typedef void adk_hos_rule_attr_t;

typedef enum adk_hos_packet_type_s
{
    ADK_HOS_L2_SWITCHING = 0,
    ADK_HOS_L2_TERMINATING,
    ADK_HOS_IP_ROUTING,
    ADK_HOS_IP_TERMINATING,
    ADK_HOS_PKT_TYPE_ANY
} adk_hos_packet_type_t;
/*!
* =============================================================
*
* @struct adk_hos_l2_term_intf_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in L2 interface table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_L2_TERM_INTF_TBL table. This helps in identifying 
* the L2 interface on which packet is received
*
* This structure consists of following attributes:
*
* l2_domain Traffic can be mapped to L2 domain using domain mapper in HLP
*
* vlan_id The top most 12-bit VLAN tag
*
* is_l2_m_b TRUE for broadcast and multicast MAC 
*
* dmac Destination mac address
*
* vsi_id DSI VSI Id
*
* vsi_q DSI VSI queue Id
*
* is_l2_term TRUE for L2 terminating, FALSE for L2 switching
*
* l2_interface L2 interface ID. This will be provided to application in DSI VSI meta data
*
* =============================================================
*/
typedef struct adk_hos_l2_term_intf_tbl_rule_attr_s
{
    int l2_domain;
    uint16_t vlan_id;
    bool is_l2_m_b;
    uint64_t dmac;
    uint32_t vsi_id;
    uint32_t vsi_q;
    bool is_l2_term;
    uint16_t l2_interface;
} adk_hos_l2_term_intf_tbl_rule_attr_t;

/*!
* =============================================================
*
* @struct adk_hos_ip_classifier_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in IPv4 and IPv6 
*           inner termination table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_IP4_CLASSIFIER_TBL, ADK_HOS_INNER_IP4_CLASSIFIER_TBL and 
* ADK_HOS_INNER_IP6_CLASSIFIER_TBL table. This helps in identifying
* classifier rule ID
*
* This structure consists of following attributes:
*
* precedence Rule priority
*
* useVlan TRUE for VLAN matching
*
* vlan_id VLAN ID indirectly represents the IP interface handle
*
* proto IP protocol type
* 
* proto_mask IP protocol mask
*
* is_ipv6 Set to TRUE for IPv6 else FALSE for IPv4
*
* dip_addr DIP IP address
*
* dip_addr_mask DIP IP address mask
*
* sip_addr SIP IP address
*
* sip_addr_mask SIP IP address mask
*
* dst_port L4 destination port
*
* dst_port_mask L4 destination port mask
*
* src_port L4 source port
*
* src_port_mask L4 source port mask
*
* rule_id Classifier rule ID
* =============================================================
*/
typedef struct adk_hos_ip_classifier_tbl_rule_attr_s
{
    uint32_t precedence;
    bool useVlan;
    uint16_t vlan_id;
    uint8_t proto;
    uint8_t proto_mask;
    bool is_ipv6;
    uint32_t dip_addr[4];   /* Use only index 0 (dip_addr[0]) for IPv4 */
    uint32_t dip_addr_mask[4];
    uint32_t sip_addr[4];   /* Use only index 0 (sip_addr[0]) for IPv4 */
    uint32_t sip_addr_mask[4];
    uint16_t l4_info1;  /* Can be L4 source port, IGMP type or ICMP type,
                         *  depending upon proto field */
    uint16_t l4_info1_mask;
    uint16_t l4_info2;  /* Can be L4 destination port or ICMP code,
                         *  depending upon proto field */
    uint16_t l4_info2_mask;
    uint32_t rule_id;
    bool is_disabled;    /* Set to TRUE if rule to be kept in disabled state */
} adk_hos_ip_classifier_tbl_rule_attr_t;

/*!
* =============================================================
*
* @struct adk_hos_ethertype_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in supported ethertype table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_SUPPORTED_ETHERTYPE_TBL table. This helps in identifying 
* if ethertype is allowed
*
* This structure consists of following attributes:
*
* ether_type ether type
*
* Ethertype allowed flag. This will be provided to application in DSI VSI meta data
* =============================================================
*/
typedef struct adk_hos_ethertype_tbl_rule_attr_s
{
    uint16_t ether_type;
} adk_hos_ethertype_tbl_rule_attr_t;

/*!
* =============================================================
*
* @struct adk_hos_ipproto_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in supported ip protocol table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_SUPPORTED_IPPROTO_TBL table. This helps in identifying 
* if ip protocol is allowed
*
* This structure consists of following attributes:
*
* proto IP protocol type
*
* IP protocol allowed flag. This will be provided to application in DSI VSI meta data
* =============================================================
*/
typedef struct adk_hos_ipproto_tbl_rule_attr_s
{
    uint8_t proto;
} adk_hos_ipproto_tbl_rule_attr_t;

/*!
* =============================================================
*
* @struct adk_hos_l2_ingress_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in L2 ingress table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_L2_INGRESS_TBL table. This helps in identifying 
* the ethernet ingress flow
*
* This structure consists of following attributes:
*
* l2_domain Traffic can be mapped to L2 domain using domain mapper in HLP
*
* ether_type  Ethernet protocol type 
*
* dmac_mask Mask for DMAC. This is passed to HLP ACL rule as well
*
* dmac Destination mac address
*
* l2_ingress Ethernet ingress ID. This will be provided to application in DSI VSI meta data
* =============================================================
*/
typedef struct adk_hos_l2_ingress_tbl_rule_attr_s
{
    int l2_domain;
    uint16_t ether_type;
    uint64_t dmac_mask;
    uint64_t dmac;
    uint32_t l2_ingress;
} adk_hos_l2_ingress_tbl_rule_attr_t;

/*!
* =============================================================
*
* @struct adk_hos_ip_term_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in IPv4 and IPv6 
*           termination table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_IP4_TERM_TBL and ADK_HOS_IP6_TERM_TBL table.
* This helps in identifying whether packet is L3 terminating
*
* This structure consists of following attributes:
*
* precedence rule priority, -1 priority means HOS will calculate priority
*            based on number of set bits in IP address mask
*            precedence 0 is reserved for default rule       
* 
* pkt_type Set ADK_HOS_L2_SWITCHING for L2 terminating packet
*            else ADK_HOS_L2_TERMINATING
*
* is_ipv6 Set to TRUE for IPv6 else FALSE for IPv4
*
* ip_addr IP address
*
* ip_addr_mask IP address mask
*
* is_ip_term TRUE for L3 termination and FALSE for L3 routing
*
* is_mcast TRUE if multicast IP address
*
* is_bcast TRUE if broadcast IP address
*
* ip_addr_stats_id Stats ID Define default ADK_HOS_DEFAULT_IP_STATS_ID
* =============================================================
*/
typedef struct adk_hos_ip_term_tbl_rule_attr_s
{
    int precedence;
    uint8_t pkt_type;
    bool is_ipv6;
    uint32_t ip_addr[4];
    uint32_t ip_addr_mask[4];
    bool is_ip_term;
    bool is_mcast;
    bool is_bcast;
    uint8_t ip_addr_stats_id;
} adk_hos_ip_term_tbl_rule_attr_t;


/*!
* =============================================================
*
* @struct adk_hos_spd_tbl_rule_attr_t "adk_hos.h"
*
* @brief   Base attributes for adding a rule in SPD IN and 
*           SPD OUT table
*
* This structure is used to specify the attributes required to add
* entry in ADK_HOS_SPD_IN_TBL and ADK_HOS_SPD_OUT_TBL. 
* This helps in identifying SPD rule ID
*
* This structure consists of following attributes:
*
* dscp DSCP value
*
* dscp_mask DSCP mask
*
* proto IP protocol type
* 
* proto_mask IP protocol mask
*
* is_ipv6 Set to TRUE for IPv6 else FALSE for IPv4
*
* dip_addr DIP IP address
*
* dip_addr_mask DIP IP address mask
*
* sip_addr SIP IP address
*
* sip_addr_mask SIP IP address mask
*
* dst_port L4 destination port
*
* dst_port_mask L4 destination port mask
*
* src_port L4 source port
*
* src_port_mask L4 source port mask
*
* spd_rule_id SPD rule ID. This will be provided to application in DSI VSI meta data
* =============================================================
*/
typedef struct adk_hos_spd_tbl_rule_attr_s
{
    uint32_t precedence;
    uint8_t proto;
    uint8_t proto_mask;
    bool is_ipv6;
    uint32_t dip_addr[4];   /* Use only index 0 (dip_addr[0]) for IPv4 */
    uint32_t dip_addr_mask[4];
    uint32_t sip_addr[4];   /* Use only index 0 (sip_addr[0]) for IPv4 */
    uint32_t sip_addr_mask[4];
    uint16_t l4_info1;  /* Can be L4 source port, IGMP type or ICMP type,
                         *  depending upon proto field */
    uint16_t l4_info1_mask;
    uint16_t l4_info2;  /* Can be L4 destination port or ICMP code,
                         *  depending upon proto field */
    uint16_t l4_info2_mask;
    uint16_t spd_rule_id;
    bool is_disabled;    /* Set to TRUE if rule to be kept in disabled state */
} adk_hos_spd_tbl_rule_attr_t;

/* HLP Port statistics [Keep in sync with Ethernet stats structure] */
typedef struct adk_hos_port_statistics_s
{
    uint64_t tx_octets;
    uint64_t tx_bad_octets;
    uint64_t tx_frm;
    uint64_t tx_bad_frm;
    uint64_t tx_mcast;
    uint64_t tx_bcast;
    uint64_t tx_pause;
    uint64_t tx_fcs;
    uint64_t tx_defer;
    uint64_t tx_xsdfr;
    uint64_t tx_1col;
    uint64_t tx_mcol;
    uint64_t tx_lcol;
    uint64_t tx_xscol;
    uint64_t tx_frag;
    uint64_t tx_usize;
    uint64_t tx_64;
    uint64_t tx_65;
    uint64_t tx_128;
    uint64_t tx_256;
    uint64_t tx_512;
    uint64_t tx_1024;
    uint64_t tx_1519;
    uint64_t tx_osize;
    uint64_t tx_jab;
    uint64_t tx_urun;
    uint64_t tx_stat_ovfl;

    uint64_t rx_octets;
    uint64_t rx_bad_octets;
    uint64_t rx_frm;
    uint64_t rx_bad_frm;
    uint64_t rx_mcast;
    uint64_t rx_bcast;
    uint64_t rx_ctl;
    uint64_t rx_pause;
    uint64_t rx_unk_op;
    uint64_t rx_fcs;
    uint64_t rx_align;
    uint64_t rx_length;
    uint64_t rx_code;
    uint64_t rx_frag;
    uint64_t rx_usize;
    uint64_t rx_64;
    uint64_t rx_65;
    uint64_t rx_128;
    uint64_t rx_256;
    uint64_t rx_512;
    uint64_t rx_1024;
    uint64_t rx_1519;
    uint64_t rx_osize;
    uint64_t rx_jab;
    uint64_t rx_drop;
    uint64_t rx_carrier;
    uint64_t rx_urun;
    uint64_t rx_stat_ovfl;

} adk_hos_port_statistics_t;

adk_status_t
adk_hos_attach (
    void);

adk_status_t
adk_hos_init (
    adk_sys_handle_t sys_handle);

adk_status_t
adk_hos_shutdown (
    void);

adk_status_t
adk_hos_detach (
    void);

adk_status_t
adk_hos_mac_stats_get (
    int port,
    adk_hos_port_statistics_t *stats);

adk_status_t
adk_hos_mac_stats_reset (
    int port);

adk_status_t
adk_hos_rule_add (
    adk_hos_object_t obj,
    adk_hos_rule_attr_t *rule_attr,
    adk_hos_rule_id_t *rule_id);

adk_status_t
adk_hos_rule_delete (
    adk_hos_object_t obj,
    adk_hos_rule_id_t rule_id);

adk_status_t
adk_hos_rule_update (
    adk_hos_object_t obj,
    adk_hos_rule_id_t rule_id,
    adk_hos_rule_attr_t *rule_attr);

adk_status_t
adk_hos_rule_state_set (
    adk_hos_object_t obj,
    adk_hos_rule_id_t rule_id,
    bool enable);

adk_status_t
adk_hos_rule_precedence_state_set (
    adk_hos_object_t obj,
    uint32_t num_precedences,
    uint32_t *precedence,
    enum ies_acl_entry_state *entryStates);

adk_status_t
adk_hos_next_rule_id_get (
    adk_hos_object_t obj,
    adk_hos_rule_id_t rule_id,
    adk_hos_rule_id_t *next_rule_id);

adk_status_t
adk_hos_rule_get (
    adk_hos_object_t obj,
    adk_hos_rule_id_t rule_id,
    adk_hos_rule_attr_t *rule_attr);

adk_status_t
adk_hos_rule_stats_get (
    adk_hos_object_t obj,
    adk_hos_rule_id_t rule_id,
    adk_hos_rule_stats_t *stats);

adk_status_t
adk_hos_rule_num_get (
    adk_hos_object_t obj,
    uint32_t* num_rules);

adk_status_t
adk_hos_ingress_pri_map_set (
    uint8_t pcp[8]);

adk_status_t
adk_hos_dump_stats (
    int port);

void
adk_hos_dump_hlp_version (
    void);

adk_status_t
adk_hos_mac_mapper_add (
    uint64_t mac_addr,
    uint8_t  mask_bits,
    uint8_t cmac,
    int *entry_id);

adk_status_t
adk_hos_mac_mapper_delete (
    int entry_id);

adk_status_t
adk_hos_mac_mapper_get (
    uint64_t mac_addr,
    uint8_t* cmac,
    int* entry_id);

adk_status_t
adk_hos_mac_mapper_get_by_cmac (
    uint8_t cmac,
    uint64_t *mac_addr,
    int *entry_id);

void adk_hos_disable (
    void);

void adk_hos_enable (
    void);

/* FollowPrec forces rules deletion in increasing
    order. The rule with lowest precedence will be
    deleted first */
adk_status_t
adk_hos_table_clean(adk_hos_object_t obj, 
                    bool followPrec);

/* Set the logical size of a table.
    The physical size is fixed and defined while
    creating the table in HLP. */
adk_status_t
adk_hos_table_size_set (
    adk_hos_object_t obj,
    uint32_t size);

/* Get the logical size of the table */
adk_status_t
adk_hos_table_size_get (
    adk_hos_object_t obj,
    uint32_t* size);

adk_status_t
adk_hos_table_size_reset (
    adk_hos_object_t obj);

/* Returns false if space is available in table
    from logical standpoint */
bool
adk_hos_table_full (
    adk_hos_object_t obj);


#endif /* __ADK_HOS_INCLUDE_H__ */

/*!
 * @} end addtogroup _hos
 */
