/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_hos_dp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _HOS SIF HOS Component
 * @{
 *     @addtogroup _hosFunctions  HOS Functions
 *     @{
 *
 * @brief   This file contains the ADK HOS DP functions for use in packet
 * processing by ADK modules.
 *
 *  **********************************************************************/


#include <adk_metadata.h>
#include <adk_hos_constants.h>
#include <adk_hos_conf.h>
#include <adk_hos_dp.h>

/* Return true if L2 terminating else false */

bool adk_hos_is_l2_term(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (dsi_md_p->packet_type & HOS_PKT_TYPE_L2_MASK);
}

/* Return true if L2 interface is invalid */
bool adk_hos_is_l2_interface_invalid(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (((dsi_md_p->inner_id & HOS_L2_INTF_MASK) == ADK_HOS_DEFAULT_L2_INTERFACE) ? true : false);
}

/* Return true if L2 interface is valid */
bool adk_hos_is_l2_interface_valid(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (((dsi_md_p->inner_id & HOS_L2_INTF_MASK) != ADK_HOS_DEFAULT_L2_INTERFACE) ? true : false);
}

/* Return L2 interface ID */
uint8_t adk_hos_get_l2_interface(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return dsi_md_p->inner_id & HOS_L2_INTF_MASK;
}

/* Return true if ingress flow rule does not match */
bool adk_hos_is_ingress_flow_invalid(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (((dsi_md_p->outer_id & HOS_INGRESS_FLOW_MASK) == ADK_HOS_DEFAULT_L2_INGRESS_FLOW) ? true : false);
}

/* Return ingress flow ID */
uint32_t adk_hos_get_ingress_flow(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return dsi_md_p->outer_id & HOS_INGRESS_FLOW_MASK;
}

/* Return true if L3 terminating else false */

bool adk_hos_is_l3_term(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (dsi_md_p->packet_type & HOS_PKT_TYPE_L3_MASK);
}

/* Return true if IP multicast else false */

bool adk_hos_is_ip_mcast(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (dsi_md_p->packet_type & HOS_PKT_TYPE_L3_MCAST);
}

/* Return true if IP broadcast else false */

bool adk_hos_is_ip_bcast(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (dsi_md_p->packet_type & HOS_PKT_TYPE_L3_BCAST);
}

/* Return true if IP stats ID is invalid else false */

bool adk_hos_is_ip_stats_invalid(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (((dsi_md_p->outer_id >> HOS_INGRESS_FLOW_WIDTH) & HOS_IP_STATS_BYTE_MASK) ==
                ADK_HOS_DEFAULT_IP_STATS_ID);
}

/* Return IP stats ID */

uint32_t adk_hos_get_ip_stats_id(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (dsi_md_p->outer_id >> HOS_INGRESS_FLOW_WIDTH) & HOS_IP_STATS_BYTE_MASK;
}

/* Return true if IP Ingress classifier is invalid else false */

bool adk_hos_is_ip_classifier_invalid(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return ((dsi_md_p->acl_id & HOS_IP_CLASSIFIER_MASK) ==
                ADK_HOS_DEFAULT_IP_CLASSIFIER);
}

/* Return classifier rule ID */

uint32_t adk_hos_get_ip_classifier_rule_id(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return dsi_md_p->acl_id & HOS_IP_CLASSIFIER_MASK;
}

/* Return if ethertype is allowed */
bool adk_hos_is_ethertype_invalid(adk_dsi_rx_metadata_t *dsi_md_p)
{
    return !(dsi_md_p->flags & HOS_FLAG_ETHERTYPE_MASK);
}

/* Return if IP protocol is allowed */
bool adk_hos_is_ipproto_invalid(adk_dsi_rx_metadata_t *dsi_md_p)
{
    return !(dsi_md_p->flags & HOS_FLAG_IPPROTO_MASK);
}

/* Return true if packet went through successful inline decryption */

bool adk_hos_is_inline_decrypted(adk_dsi_rx_metadata_t *dsi_md_p)
{
    if (dsi_md_p->ipsec_offloaded && !dsi_md_p->ipsec_status)
    {
        return true;
    }
    return false;
}

/* Return true if SPD ID is invalid else false */

bool adk_hos_is_spd_id_invalid(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return ((dsi_md_p->spd_id & HOS_SPD_ID_MASK) ==
                ADK_HOS_DEFAULT_SPD_RULE_ID);
}

/* Return SPD ID */

uint16_t adk_hos_get_spd_id(const adk_dsi_rx_metadata_t *dsi_md_p)
{
    return (dsi_md_p->spd_id & HOS_SPD_ID_MASK); 
}

bool adk_hos_is_ip_hdr_bad(adk_dsi_rx_metadata_t *dsi_md_p)
{
#ifdef ADK_DEVICE_SNR
    if ((dsi_md_p->flags & HOS_FLAG_TRAP_MASK) == HOS_IP_HDR_ERROR)
    {
        return (true);
    }
    else
    {
        return (false);
    }
#else
    RTE_SET_USED(dsi_md_p);
    return (false);
#endif
}

bool adk_hos_is_policer_red(adk_dsi_rx_metadata_t *dsi_md_p)
{
#ifdef ADK_DEVICE_SNR
    if (dsi_md_p->pol_a == HOS_POLICER_RED)
    {
        return (true);
    }
    else
    {
        return (false);
    }
#else
    RTE_SET_USED(dsi_md_p);
    return (false);
#endif
}


bool adk_hos_is_l4_chksum_err(adk_dsi_rx_metadata_t *dsi_md_p)
{
#ifdef ADK_DEVICE_SNR
    return (dsi_md_p->flags & HOS_FLAG_L4_CHKSUM_ERR);
#else
    RTE_SET_USED(dsi_md_p);
    return (false);
#endif
}

bool adk_hos_trap_code_get(adk_dsi_rx_metadata_t *dsi_md_p,
                           uint8_t *trap_code)
{
#ifdef ADK_DEVICE_SNR
    if (dsi_md_p->flags & HOS_FLAG_VALID_TRAP)
    {
        *trap_code = ((dsi_md_p->flags >> HOS_FLAG_TRAP_CODE_BIT) &
                      HOS_FLAG_TRAP_CODE_MASK);
        return (true);
    }

#else
    RTE_SET_USED(dsi_md_p);
    RTE_SET_USED(trap_code);
#endif
    
    return (false);
}

void adk_hos_enable_l4_chksum(adk_dsi_tx_metadata_t *dsi_md_p)
{
#ifdef ADK_DEVICE_SNR
    dsi_md_p->flags |= HOS_FLAG_L4_CHKSUM_SET;
#else
    RTE_SET_USED(dsi_md_p);
#endif
    return;
}

void adk_hos_disable_l4_chksum(adk_dsi_tx_metadata_t *dsi_md_p)
{
#ifdef ADK_DEVICE_SNR
    dsi_md_p->flags &= ~HOS_FLAG_L4_CHKSUM_SET;
#else
    RTE_SET_USED(dsi_md_p);
#endif
    return;
}

#ifdef ADK_HOS_DP_TEST
adk_dsi_rx_metadata_t*
adk_hos_test_get_dsi_vsi_ptr(void* task)
{
    if(NULL == task)
        return NULL;

    return (adk_dsi_rx_metadata_t*)((uintptr_t) task + sizeof(struct rte_mbuf) + EM_HDR_SIZE);
}
#endif

/*!    end addtogroup _hosFunctions
 *     @}
 * end addtogroup _hos
 * @}
 */
