/* ***************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_lttng_trace_defn.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifFunctions  SIF Functions
 *     @{
 *
 *
 *
 * This file contains definition of functions that are
 * part of implementation for LTTng support in ADK.
 *
 ***********************************************************************/
#include "adk_modules_enabled.h"

#define TRACEPOINT_DEFINE
#define TRACEPOINT_PROBE_DYNAMIC_LINKAGE

#include "adk_tp_sif.h"
#include "adk_tp_sif_debug.h"
#include "adk_tp_sif_qos.h"
#include "adk_tp_sif_qos_debug.h"
#include "adk_tp_sif_fps.h"
#include "adk_tp_sif_fps_debug.h"
#include "adk_tp_ph.h"
#include "adk_tp_ph_debug.h"

#if defined(ADK_MODULE_ETHERNET_ENABLED)
#include "adk_tp_ethernet.h"
#include "adk_tp_ethernet_debug.h"
#endif

#if defined(ADK_MODULE_ETH_OAM_ENABLED)
#include "adk_tp_eth_oam.h"
#include "adk_tp_eth_oam_debug.h"
#endif

#if defined(ADK_MODULE_GTPU_ENABLED)
#include "adk_tp_gtpu.h"
#include "adk_tp_gtpu_debug.h"
#endif

#if defined(ADK_MODULE_IP_ENABLED)
#include "adk_tp_ip.h"
#include "adk_tp_ip_debug.h"
#endif

#if defined(ADK_MODULE_IPV6_ENABLED)
#include "adk_tp_ipv6.h"
#include "adk_tp_ipv6_debug.h"
#endif

#if defined(ADK_MODULE_BFD_ENABLED)
#include "adk_tp_bfd.h"
#include "adk_tp_bfd_debug.h"
#endif



#if defined(ADK_MODULE_IPSEC_ENABLED)
#include "adk_tp_ipsec.h"
#include "adk_tp_ipsec_debug.h"
#endif


#if defined(ADK_MODULE_L2TPV3_ENABLED)
#include "adk_tp_l2tpv3.h"
#include "adk_tp_l2tpv3_debug.h"
#endif

#if defined(ADK_MODULE_EHSFP_IWF_ENABLED)
#include "adk_tp_ehsfp_iwf.h"
#include "adk_tp_ehsfp_iwf_debug.h"
#endif

#if defined(ADK_MODULE_TWAMP_ENABLED)
#include "adk_tp_twamp.h"
#include "adk_tp_twamp_debug.h"
#endif

#include "adk_include.h"
#include "adk_module_id.h"

#include <stdarg.h>
#include <adk_lttng_trace.h>

/*
 * =============================================================
 * 
 * Filename fragments (and filename) for determining the submodule
 *
 * =============================================================
 */
static const char * ADK_QOS_FILE_PREFIX="adk_qos_";
static const char * ADK_FPS_FILE_PREFIX="adk_fps_";
static const char * ADK_PH_FILE_PREFIX="adk_ph_";
static const char * ADK_PH_FILE_PREFIX_ALT="adk_ph.c";


/*!
 * =============================================================
 * @internal
 * @enum adk_lttng_sys_submodule_t
 *
 * Internal enum to define the SIF type of logging to do
 *
 * =============================================================
 */
typedef enum adk_lttng_sys_submodule_e {
    ADK_LTTNG_SYS_SUBMODULE_QOS = 1,  /*!< Submodule is QoS */
    ADK_LTTNG_SYS_SUBMODULE_FPS,  /*!< Submodule is FPS */
    ADK_LTTNG_SYS_SUBMODULE_PH,  /*!< Submodule is PH */
    ADK_LTTNG_SYS_SUBMODULE_SYS,  /*!< Submodule is SYS by default */
} adk_lttng_sys_submodule_t;   /* ------  end of enum adk_lttng_sys_submodule_t  ------ */

/*!
 * =============================================================
 * @internal
 * @adk_lttng_sys_submodule_get
 * @param[in] fileName Name of the file
 * @param[out] subModule - the submodule to log against
 *
 *
 * We only have one module ID for PH, QoS and SIF. Since we want to have
 * different LTTng tags for PH and QoS, this function will return the
 * submodule ID based on the file name.
 *
 * Functiont get the submodule to log against based on file name. QoS and PH
 * are part of the system module, but we would like to log against their own
 * tracepoints.
 *
 * =============================================================
 */
static adk_status_t
adk_lttng_sys_submodule_get(
                            const char *fileName,
                            adk_lttng_sys_submodule_t *subModule
                           )
{
    adk_status_t    retValue = ADK_SUCCESS;
    uint32_t    submoduleFound = 1;

    *subModule = ADK_LTTNG_SYS_SUBMODULE_SYS;

    submoduleFound = ADK_STRNCMP(ADK_QOS_FILE_PREFIX, fileName,
                             ADK_STRLEN(ADK_QOS_FILE_PREFIX));
    if (0 == submoduleFound)
    {
        *subModule = ADK_LTTNG_SYS_SUBMODULE_QOS;
    } else
    {
        submoduleFound = ADK_STRNCMP(ADK_FPS_FILE_PREFIX, fileName,
                                 ADK_STRLEN(ADK_FPS_FILE_PREFIX));
        if (0 == submoduleFound)
        {
            *subModule = ADK_LTTNG_SYS_SUBMODULE_FPS;
        } else
        {
            submoduleFound = ADK_STRNCMP(ADK_PH_FILE_PREFIX, fileName,
                                         ADK_STRLEN(ADK_PH_FILE_PREFIX));
            if (0 == submoduleFound)
            {
                *subModule = ADK_LTTNG_SYS_SUBMODULE_PH;
            } else
            {
                submoduleFound = ADK_STRNCMP(ADK_PH_FILE_PREFIX_ALT, fileName,
                                             ADK_STRLEN(ADK_PH_FILE_PREFIX_ALT));
                if (0 == submoduleFound)
                {
                    *subModule = ADK_LTTNG_SYS_SUBMODULE_PH;
                }
            }
        }
    }

    ADK_RETURN( retValue );
}    /* -----  end of function adk_lttng_sys_submodule_get  ----- */

/*
 * =============================================================
 * Generic macro definition for INFO/ERROR/WARNING events. 
 * This function calls the respective module's tracepoint 
 * provider. 
 * =============================================================
 */
#define DEFINE_INFO_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  char debug_info[MAX_LEN]; \
  if(__tracepoint_com_intel_adk_##module_name##___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_##module_name, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for FORCE_ versions of the 
 * INFO/ERROR/WARNING events. 
 * This function calls the respective module's tracepoint 
 * provider. 
 * =============================================================
 */
#define DEFINE_FORCE_INFO_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  char debug_info[MAX_LEN]; \
  if(__tracepoint_com_intel_adk_##module_name##___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_##module_name, intel_adk_##event, \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Macro definition for INFO/ERROR/WARNING events. This function 
 * calls the SIF module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_SIF_INFO_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  char debug_info[MAX_LEN]; \
  adk_lttng_sys_submodule_t  subModule; \
  adk_lttng_sys_submodule_get(fileName, &subModule); \
  if ( subModule == ADK_LTTNG_SYS_SUBMODULE_QOS) \
  { \
  if(__tracepoint_com_intel_adk_sif_qos___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_qos, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_FPS) \
  { \
  if(__tracepoint_com_intel_adk_sif_fps___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_fps, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_PH) \
  { \
  if(__tracepoint_com_intel_adk_ph___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_ph, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else \
  { \
  if(__tracepoint_com_intel_adk_sif___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for FORCE_INFO/FORCE_WARNING/
 * FORCE_ERROR events. This function calls the respective 
 * module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_FORCE_FNS(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle,const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  char debug_info[MAX_LEN]; \
  adk_lttng_sys_submodule_t  subModule; \
  adk_lttng_sys_submodule_get(fileName, &subModule); \
  if ( subModule == ADK_LTTNG_SYS_SUBMODULE_QOS) \
  { \
  if(__tracepoint_com_intel_adk_sif_qos___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_qos, intel_adk_##event, \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_FPS) \
  { \
  if(__tracepoint_com_intel_adk_sif_fps___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_fps, intel_adk_##event, \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_PH) \
  { \
  if(__tracepoint_com_intel_adk_ph___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_ph, intel_adk_##event, \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_SYS) \
  { \
  if(__tracepoint_com_intel_adk_sif___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif, intel_adk_##event, \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  return ADK_SUCCESS; \
}

/*
 * ==============================================================
 * Macro definition for FUNCTION LOCK/UNLOCK events. This function
 * calls the SIF module's tracepoint provider.
 * ==============================================================
 */
#define DEFINE_SIF_LOCK_UNLOCK_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo, const char* fileName, const char* fmt, ...) \
{ \
    if(__tracepoint_com_intel_adk_sif___intel_adk_##event.state) \
    { \
        va_list   args; \
        char debug_info[MAX_LEN]; \
        va_start (args, fmt); \
        ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
        ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
        tracepoint(com_intel_adk_sif, intel_adk_##event, \
                   ADK_MODULE_NAME_GET(handle), \
                   lineNo, fileName, \
                   debug_info); \
        va_end (args); \
    } \
    return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for FUNC_LOCK/UNLOCK events. This 
 * function calls the respective module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_LOCK_UNLOCK_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo, const char* fileName, const char* fmt, ...) \
{ \
    if(__tracepoint_com_intel_adk_##module_name##___intel_adk_##event.state) \
    { \
        va_list   args; \
        char debug_info[MAX_LEN]; \
        va_start (args, fmt); \
        ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
        ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
        tracepoint(com_intel_adk_##module_name, intel_adk_##event, \
                   ADK_MODULE_NAME_GET(handle), \
                   lineNo, fileName, \
                   debug_info); \
        va_end (args); \
    } \
    return ADK_SUCCESS; \
}

/*
 * ==============================================================
 * Macro definition for FUNCTION ENTRY/EXIT events. This function 
 * calls the SIF module's tracepoint provider. 
 * ==============================================================
 */
#define DEFINE_SIF_ENTRY_EXIT_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  adk_lttng_sys_submodule_t  subModule; \
  adk_lttng_sys_submodule_get(fileName, &subModule); \
  if ( subModule == ADK_LTTNG_SYS_SUBMODULE_QOS) \
  { \
  if(__tracepoint_com_intel_adk_sif_qos___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_sif_qos, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_FPS) \
  { \
  if(__tracepoint_com_intel_adk_sif_fps___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_sif_fps, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_PH) \
  { \
  if(__tracepoint_com_intel_adk_ph___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_ph, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  else \
  { \
  if(__tracepoint_com_intel_adk_sif___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_sif, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for FUNC_ENTRY/EXIT events. This 
 * function calls the respective module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_ENTRY_EXIT_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  if(__tracepoint_com_intel_adk_##module_name##___intel_adk_##event.state) \
  { \
    va_list   args; \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_##module_name, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for INTERNAL_FUNC_ENTRY/EXIT events. 
 * This function calls the respective module's tracepoint 
 * provider. 
 * =============================================================
 */
#define DEFINE_INTERNAL_ENTRY_EXIT_FN(module_name, event) \
adk_status_t adk_lttng_trace_internal_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  if(__tracepoint_com_intel_adk_##module_name##_debug___intel_adk_##event.state) \
  { \
    va_list   args; \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_##module_name, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Macro definition for INTERNAL FN. ENTRY/EXIT events. This 
 * function calls the SIF module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_SIF_INTERNAL_ENTRY_EXIT_FN(module_name, event) \
adk_status_t adk_lttng_trace_internal_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list	args; \
  adk_lttng_sys_submodule_t  subModule; \
  adk_lttng_sys_submodule_get(fileName, &subModule); \
  if ( subModule == ADK_LTTNG_SYS_SUBMODULE_QOS) \
  { \
  if(__tracepoint_com_intel_adk_sif_qos_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_sif_qos_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_FPS) \
  { \
  if(__tracepoint_com_intel_adk_sif_fps_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_sif_fps_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_PH) \
  { \
  if(__tracepoint_com_intel_adk_ph_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_ph_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  else \
  { \
  if(__tracepoint_com_intel_adk_sif_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    tracepoint(com_intel_adk_sif_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   va_arg (args, const char*)); \
    va_end (args); \
  } \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for FORCE_FUNC_ENTRY/EXIT events. This 
 * function calls the respective module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_FORCE_ENTRY_EXIT_FN(module_name, event) \
adk_status_t adk_lttng_trace_force_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  adk_lttng_sys_submodule_t subModule; \
  adk_lttng_sys_submodule_get(fileName, &subModule); \
  va_start (args, fmt); \
  if ( subModule == ADK_LTTNG_SYS_SUBMODULE_QOS) \
  { \
     tracepoint(com_intel_adk_sif_qos, intel_adk_##event, \
                  ADK_SYS_QOS_MODULE_NAME, lineNo, fileName, \
                  va_arg (args, const char*)); \
  } \
  if( subModule == ADK_LTTNG_SYS_SUBMODULE_FPS) \
  { \
     tracepoint(com_intel_adk_sif_fps, intel_adk_##event, \
                ADK_PH_MODULE_NAME, lineNo, fileName, \
                va_arg (args, const char*)); \
  } \
  if( subModule == ADK_LTTNG_SYS_SUBMODULE_PH) \
  { \
     tracepoint(com_intel_adk_ph, intel_adk_##event, \
                ADK_PH_MODULE_NAME, lineNo, fileName, \
                va_arg (args, const char*)); \
  } \
  if( subModule == ADK_LTTNG_SYS_SUBMODULE_SYS) \
  { \
     tracepoint(com_intel_adk_sif, intel_adk_##event, \
               ADK_SYS_MODULE_NAME, lineNo, fileName, \
               va_arg (args, const char*)); \
  } \
  va_end (args); \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Generic macro definition for HW events. This 
 * function calls the respective module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_HW_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  if(__tracepoint_com_intel_adk_##module_name##_debug___intel_adk_##event.state) \
  { \
    va_list   args; \
    char debug_info[MAX_LEN]; \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_##module_name##_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  return ADK_SUCCESS; \
}

/*
 * =============================================================
 * Macro definition for HW events. This function 
 * calls the SIF module's tracepoint provider. 
 * =============================================================
 */
#define DEFINE_SIF_HW_FN(module_name, event) \
adk_status_t adk_lttng_trace_##module_name##_##event(const adk_module_handle_t handle, const int32_t lineNo,const char* fileName, const char* fmt, ...) \
{ \
  va_list   args; \
  char debug_info[MAX_LEN]; \
  adk_lttng_sys_submodule_t  subModule; \
  adk_lttng_sys_submodule_get(fileName, &subModule); \
  if ( subModule == ADK_LTTNG_SYS_SUBMODULE_QOS) \
  { \
  if(__tracepoint_com_intel_adk_sif_qos_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_qos_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_FPS) \
  { \
  if(__tracepoint_com_intel_adk_sif_fps_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_fps_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else if ( subModule == ADK_LTTNG_SYS_SUBMODULE_PH) \
  { \
  if(__tracepoint_com_intel_adk_ph_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_ph_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  else \
  { \
  if(__tracepoint_com_intel_adk_sif_debug___intel_adk_##event.state) \
  { \
    va_start (args, fmt); \
    ADK_MEMSET(debug_info, 0, sizeof(debug_info)); \
    ADK_VSNPRINTF(debug_info, MAX_LEN, fmt, args); \
    tracepoint(com_intel_adk_sif_debug, intel_adk_##event, \
                                   ADK_MODULE_NAME_GET(handle), \
                                   lineNo, fileName, \
                                   debug_info); \
    va_end (args); \
  } \
  } \
  return ADK_SUCCESS; \
}

DEFINE_SIF_INFO_FN(sif, error)
DEFINE_SIF_INFO_FN(sif, fatal_error)
DEFINE_FORCE_FNS(sif, force_error)
DEFINE_FORCE_FNS(sif, force_fatal)
DEFINE_FORCE_FNS(sif, force_info)
DEFINE_FORCE_FNS(sif, force_warning)
DEFINE_SIF_INFO_FN(sif, info)
DEFINE_SIF_INFO_FN(sif, warning)
DEFINE_SIF_ENTRY_EXIT_FN(sif, func_entry)
DEFINE_SIF_ENTRY_EXIT_FN(sif, func_exit)
DEFINE_FORCE_ENTRY_EXIT_FN(sif, func_entry)
DEFINE_FORCE_ENTRY_EXIT_FN(sif, func_exit)
DEFINE_SIF_INTERNAL_ENTRY_EXIT_FN(sif, func_entry)
DEFINE_SIF_INTERNAL_ENTRY_EXIT_FN(sif, func_exit)
DEFINE_SIF_HW_FN(sif, log_hw)
DEFINE_SIF_LOCK_UNLOCK_FN(sif, lock_acquire)
DEFINE_SIF_LOCK_UNLOCK_FN(sif, lock_release)

#define ADK_MODULE_LTTNG_FUNCS_DEFINED(__module_adk__) \
    DEFINE_INFO_FN(__module_adk__, info) \
    DEFINE_FORCE_INFO_FN(__module_adk__, force_info) \
    DEFINE_INFO_FN(__module_adk__, error) \
    DEFINE_FORCE_INFO_FN(__module_adk__, force_error) \
    DEFINE_INFO_FN(__module_adk__, fatal_error) \
    DEFINE_FORCE_INFO_FN(__module_adk__, force_fatal) \
    DEFINE_INFO_FN(__module_adk__, warning) \
    DEFINE_FORCE_INFO_FN(__module_adk__, force_warning) \
    DEFINE_ENTRY_EXIT_FN(__module_adk__, func_entry) \
    DEFINE_ENTRY_EXIT_FN(__module_adk__, func_exit) \
    DEFINE_INTERNAL_ENTRY_EXIT_FN(__module_adk__, func_entry) \
    DEFINE_INTERNAL_ENTRY_EXIT_FN(__module_adk__, func_exit) \
    DEFINE_HW_FN(__module_adk__, log_hw) \
    DEFINE_LOCK_UNLOCK_FN(__module_adk__, lock_acquire) \
    DEFINE_LOCK_UNLOCK_FN(__module_adk__, lock_release)

#if defined(ADK_MODULE_IP_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(ip)
#endif

#if defined(ADK_MODULE_IPV6_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(ipv6)
#endif

#if defined(ADK_MODULE_IPSEC_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(ipsec)
#endif

#if defined(ADK_MODULE_ETHERNET_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(ethernet)
#endif

#if defined(ADK_MODULE_GTPU_ENABLED)
DEFINE_INFO_FN(gtpu, error)
DEFINE_ENTRY_EXIT_FN(gtpu, func_entry)
DEFINE_ENTRY_EXIT_FN(gtpu, func_exit)
DEFINE_INFO_FN(gtpu, info)
DEFINE_INTERNAL_ENTRY_EXIT_FN(gtpu, func_entry)
DEFINE_INTERNAL_ENTRY_EXIT_FN(gtpu, func_exit)
DEFINE_LOCK_UNLOCK_FN(gtpu, lock_acquire)
DEFINE_LOCK_UNLOCK_FN(gtpu, lock_release)
#endif

#if defined(ADK_MODULE_BFD_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(bfd)
#endif


#if defined(ADK_MODULE_L2TPV3_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(l2tpv3)
#endif

#if defined(ADK_MODULE_EHSFP_IWF_ENABLED)
DEFINE_INFO_FN(ehsfp_iwf, error)
DEFINE_INFO_FN(ehsfp_iwf, fatal_error)
DEFINE_ENTRY_EXIT_FN(ehsfp_iwf, func_entry)
DEFINE_ENTRY_EXIT_FN(ehsfp_iwf, func_exit)
DEFINE_INFO_FN(ehsfp_iwf, info)
DEFINE_INFO_FN(ehsfp_iwf, warning)
DEFINE_INTERNAL_ENTRY_EXIT_FN(ehsfp_iwf, func_entry)
DEFINE_INTERNAL_ENTRY_EXIT_FN(ehsfp_iwf, func_exit)
DEFINE_LOCK_UNLOCK_FN(ehsfp_iwf, lock_acquire)
DEFINE_LOCK_UNLOCK_FN(ehsfp_iwf, lock_release)
#endif

#if defined(ADK_MODULE_ETH_OAM_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(eth_oam)
#endif

#if defined(ADK_MODULE_TWAMP_ENABLED)
ADK_MODULE_LTTNG_FUNCS_DEFINED(twamp)
#endif

#define ADK_LTTNG_MODULE_INIT(__adk_module_id__, __adk_module_name__) \
    [__adk_module_id__][TRACE_INFO_INDEX] = FN_NAME(__adk_module_name__, info), \
    [__adk_module_id__][TRACE_FORCE_INFO_INDEX] = FN_NAME(__adk_module_name__, force_info), \
    [__adk_module_id__][TRACE_ERROR_INDEX] = FN_NAME(__adk_module_name__, error), \
    [__adk_module_id__][TRACE_FORCE_ERROR_INDEX] = FN_NAME(__adk_module_name__, force_error), \
    [__adk_module_id__][TRACE_FATAL_ERROR_INDEX] = FN_NAME(__adk_module_name__, fatal_error), \
    [__adk_module_id__][TRACE_FORCE_FATAL_INDEX] = FN_NAME(__adk_module_name__, force_fatal), \
    [__adk_module_id__][TRACE_WARNING_INDEX] = FN_NAME(__adk_module_name__, warning), \
    [__adk_module_id__][TRACE_FORCE_WARNING_INDEX] = FN_NAME(__adk_module_name__, force_warning), \
    [__adk_module_id__][TRACE_FUNC_ENTRY_INDEX] = FN_NAME(__adk_module_name__, func_entry), \
    [__adk_module_id__][TRACE_FUNC_EXIT_INDEX] = FN_NAME(__adk_module_name__, func_exit), \
    [__adk_module_id__][TRACE_INTERNAL_FUNC_ENTRY_INDEX] = FN_NAME_INTERNAL(__adk_module_name__, func_entry), \
    [__adk_module_id__][TRACE_INTERNAL_FUNC_EXIT_INDEX] = FN_NAME_INTERNAL(__adk_module_name__, func_exit), \
    [__adk_module_id__][TRACE_HW_INDEX] = FN_NAME(__adk_module_name__, log_hw), \
    [__adk_module_id__][TRACE_LOCK_ACQUIRE_INDEX] = FN_NAME(__adk_module_name__, lock_acquire), \
    [__adk_module_id__][TRACE_LOCK_RELEASE_INDEX] = FN_NAME(__adk_module_name__, lock_release),

static const adk_lttng_info_t adk_lttng_info_const_ptr = {
    .cookie = ADK_LTTNG_COOKIE,
    .trace_fn_ptr = {
        [ADK_SYS_MODULE_ID][TRACE_INFO_INDEX] = FN_NAME(sif, info),
        [ADK_SYS_MODULE_ID][TRACE_FORCE_INFO_INDEX] = FN_NAME(sif, force_info),
        [ADK_SYS_MODULE_ID][TRACE_ERROR_INDEX] = FN_NAME(sif, error),
        [ADK_SYS_MODULE_ID][TRACE_FORCE_ERROR_INDEX] = FN_NAME(sif, force_error),
        [ADK_SYS_MODULE_ID][TRACE_FATAL_ERROR_INDEX] = FN_NAME(sif, fatal_error),
        [ADK_SYS_MODULE_ID][TRACE_FORCE_FATAL_INDEX] = FN_NAME(sif, force_fatal),
        [ADK_SYS_MODULE_ID][TRACE_WARNING_INDEX] = FN_NAME(sif, warning),
        [ADK_SYS_MODULE_ID][TRACE_FORCE_WARNING_INDEX] = FN_NAME(sif, force_warning),
        [ADK_SYS_MODULE_ID][TRACE_FUNC_ENTRY_INDEX] = FN_NAME(sif, func_entry),
        [ADK_SYS_MODULE_ID][TRACE_FUNC_EXIT_INDEX] = FN_NAME(sif, func_exit),
        [ADK_SYS_MODULE_ID][TRACE_INTERNAL_FUNC_ENTRY_INDEX] = FN_NAME_INTERNAL(sif, func_entry),
        [ADK_SYS_MODULE_ID][TRACE_INTERNAL_FUNC_EXIT_INDEX] = FN_NAME_INTERNAL(sif, func_exit),
        [ADK_SYS_MODULE_ID][TRACE_FORCE_FUNC_ENTRY_INDEX] = adk_lttng_trace_force_sif_func_entry,
        [ADK_SYS_MODULE_ID][TRACE_FORCE_FUNC_EXIT_INDEX] = adk_lttng_trace_force_sif_func_exit,
        [ADK_SYS_MODULE_ID][TRACE_HW_INDEX] = FN_NAME(sif, log_hw),
        [ADK_SYS_MODULE_ID][TRACE_LOCK_ACQUIRE_INDEX] = FN_NAME(sif, lock_acquire),
        [ADK_SYS_MODULE_ID][TRACE_LOCK_RELEASE_INDEX] = FN_NAME(sif, lock_release),

#if defined(ADK_MODULE_IP_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_IP_MODULE_ID, ip)
#endif

#if defined(ADK_MODULE_IPV6_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_IPV6_MODULE_ID, ipv6)
#endif

#if defined(ADK_MODULE_IPSEC_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_IPSEC_MODULE_ID, ipsec)
#endif

#if defined(ADK_MODULE_ETHERNET_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_ETHERNET_MODULE_ID, ethernet)
#endif

#if defined(ADK_MODULE_GTPU_ENABLED)
        [ADK_GTPU_MODULE_ID][TRACE_ERROR_INDEX] = FN_NAME(gtpu, error),
        [ADK_GTPU_MODULE_ID][TRACE_FUNC_ENTRY_INDEX] = FN_NAME(gtpu, func_entry),
        [ADK_GTPU_MODULE_ID][TRACE_FUNC_EXIT_INDEX] = FN_NAME(gtpu, func_exit),
        [ADK_GTPU_MODULE_ID][TRACE_INFO_INDEX] = FN_NAME(gtpu, info),
        [ADK_GTPU_MODULE_ID][TRACE_INTERNAL_FUNC_ENTRY_INDEX] = FN_NAME_INTERNAL(gtpu, func_entry),
        [ADK_GTPU_MODULE_ID][TRACE_INTERNAL_FUNC_EXIT_INDEX] = FN_NAME_INTERNAL(gtpu, func_exit),
        [ADK_GTPU_MODULE_ID][TRACE_LOCK_ACQUIRE_INDEX] = FN_NAME(gtpu, lock_nlock),
        [ADK_GTPU_MODULE_ID][TRACE_LOCK_RELEASE_INDEX] = FN_NAME(gtpu, lock_release),
#endif

#if defined(ADK_MODULE_BFD_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_BFD_MODULE_ID, bfd)
#endif


#if defined(ADK_MODULE_L2TPV3_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_L2TPV3_MODULE_ID, l2tpv3)
#endif

#if defined(ADK_MODULE_EHSFP_IWF_ENABLED)
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_ERROR_INDEX] = FN_NAME(ehsfp_iwf, error),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_FATAL_ERROR_INDEX] = FN_NAME(ehsfp_iwf, fatal_error),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_FUNC_ENTRY_INDEX] = FN_NAME(ehsfp_iwf, func_entry),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_FUNC_EXIT_INDEX] = FN_NAME(ehsfp_iwf, func_exit),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_INFO_INDEX] = FN_NAME(ehsfp_iwf, info),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_WARNING_INDEX] = FN_NAME(ehsfp_iwf, warning),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_INTERNAL_FUNC_ENTRY_INDEX] = FN_NAME_INTERNAL(ehsfp_iwf, func_entry),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_INTERNAL_FUNC_EXIT_INDEX] = FN_NAME_INTERNAL(ehsfp_iwf, func_exit),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_LOCK_ACQUIRE_INDEX] = FN_NAME(gtpu, lock_nlock),
        [ADK_EHSFP_IWF_MODULE_ID][TRACE_LOCK_RELEASE_INDEX] = FN_NAME(gtpu, lock_release),
#endif

#if defined(ADK_MODULE_ETH_OAM_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_ETH_OAM_MODULE_ID, eth_oam)
#endif

#if defined(ADK_MODULE_TWAMP_ENABLED)
        ADK_LTTNG_MODULE_INIT(ADK_TWAMP_MODULE_ID, twamp)
#endif
    }
};

/*!
 * =============================================================
 * @internal
 * @adk_lttng_fn_ptr_get
 *
 * Retruns a pointer to the function for LTTng logging based on module ID and
 * event.
 *
 * =============================================================
 */
void
adk_lttng_fn_ptr_get(uint32_t moduleId,
                     uint32_t index,
                     adk_lttng_trace_fn_ptr * const lttngFnPtr_p)
{
    if ((adk_lttng_info_const_ptr.trace_fn_ptr[moduleId][index]) == NULL)
    {
        moduleId = ADK_SYS_MODULE_ID;
    }

    *lttngFnPtr_p = adk_lttng_info_const_ptr.trace_fn_ptr[moduleId][index];
}




/*!    end addtogroup _sifFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */


