/* ***************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_qos_dscp_pcp.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _qos  QoS Module
 * @{
 *
 * @brief  
 *
 * This file contains the C functions for updating the DSCP & PCP mapping tree
 *
 * The tree is setup to allow for mappings between different Code Points (CP),
 * i.e. DSCP->DSCP, DSCP->PCP, PCP->DSCP.
 *
 *
 *  **********************************************************************/

#ifndef  __ADK_QOS_DSCP_PCP_H__
#define  __ADK_QOS_DSCP_PCP_H__




/*!
 * =============================================================
 * @def ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT
 *
 * Default mapping profile to indicate that the Port/VLAN/IP inteface is not
 * associated with a mapping profile.
 *
 * =============================================================
 */
#define ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT 0


/*!
 * =============================================================
 * @def ADK_QOS_CP_MAPPING_PROFILE_ID_MAX
 *
 * Maximum number of profile IDs used for DSCP & PCP mappings
 *
 * =============================================================
 */
#define ADK_QOS_CP_MAPPING_PROFILE_ID_MAX 32


/*!
 * =============================================================
 * @enum adk_qos_cp_mapping_type_t
 *
 * @brief DSCP & PCP input mapping type
 *
 * The mapping tree will use DSCP or PCP as an input and this flag indicates
 * which input should be programmed for the tree. Each Code Point (CP) has a
 * default that can be programmed, so this flag will indicte that.
 *
 * This enum is used for both the input and output type. For an output, the
 * _DEFAULT entries are invalid and will cause an error to be returned.
 *
 * =============================================================
 */
typedef enum adk_qos_cp_mapping_type_e {
    ADK_QOS_CP_MAPPING_START = 0,
    /*!
     * The input or output value is of type DSCP
     */
    ADK_QOS_CP_MAPPING_DSCP = 1,
    /*!
     * The input is the default value for type DSCP
     */
    ADK_QOS_CP_MAPPING_DSCP_DEFAULT,
    /*!
     * The input or output value is of type PCP
     */
    ADK_QOS_CP_MAPPING_PCP,
    /*!
     * The input is the default value for type PCP
     */
    ADK_QOS_CP_MAPPING_PCP_DEFAULT,

    ADK_QOS_CP_MAPPING_MAX,
} adk_qos_cp_mapping_type_t;   /* ------  end of enum adk_qos_cp_mapping_type_t  ------ */


/*!
 * =============================================================
 * @internal
 * @enum adk_qos_cp_mapping_mask_t
 *
 * Output mask type to use for applying the remarking.
 *
 * =============================================================
 */
typedef enum adk_qos_cp_mapping_mask_e {
    /*!
     * This value indicates that remarking only is to be performaed
     */
    ADK_QOS_CP_MAPPING_MODE_REMARK = 1,
    /*!
     * This value indicates that the output of the tree should only be used
     * for queue selection.
     */
    ADK_QOS_CP_MAPPING_MODE_QUEUE_SELECT = 2,
    /*!
     * This value indicates that the output of the tree should be used
     * for remarking and queue selection. This is set to 3 which is the OR of
     * the two other bits.
     */
    ADK_QOS_CP_MAPPING_MODE_REMARK_QUEUE_SELECT = 3,

    ADK_QOS_CP_MAPPING_MODE_MAX, /*!<  */
} adk_qos_cp_mapping_mask_t;   /* ------  end of enum adk_qos_cp_mapping_mask_t  ------ */


/*!
 * =============================================================
 * @struct adk_qos_cp_mapping_t adk_qos_dscp_pcp.h "adk_qos_dscp_pcp.h"
 *
 * @brief DSCP & PCP input mapping data
 *
 * The input/output data used to update the mapping tree
 *
 * =============================================================
 */
typedef struct adk_qos_cp_mapping_s {
    uint32_t                profileId;
    adk_qos_cp_mapping_type_t   inputType;
    adk_qos_cp_mapping_type_t   outputType;
    adk_qos_cp_mapping_mask_t   outputMode;
    uint32_t                inputValue;
    uint32_t                outputValue;
} adk_qos_cp_mapping_t;   /* ------  end of struct adk_qos_cp_mapping_t  ------ */


/*!
 *     @addtogroup _qosFunctions  QoS Functions
 *     @{
 */

/*!
 * =============================================================
 * adk_qos_cp_mapping_add
 * @param[in]  sysHandle    System handle
 * @param[in]  mappingData  The structure containing the profile ID, type,
 *                          input and output maps
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @brief Fucntion to add a DSCP or PCP mapping entry
 *
 * This function is used to add or update a mapping entry. The two different
 * CP types, DSCP and PCP, can have a specific mapping associated with an
 * input value and a default for any unmapped values. The type field indicates
 * which is to be programmed. The DSCP and PCP maps can have different default
 * values.
 *
 * =============================================================
 */
adk_status_t
adk_qos_cp_mapping_add(const adk_sys_handle_t       sysHandle,
                       const adk_qos_cp_mapping_t  *mappingData);

/*!
 * =============================================================
 * adk_qos_cp_mapping_remove
 * @param[in]  sysHandle    System handle
 * @param[in]  mappingData  The structure containing the profile ID, type,
 *                          input and output maps
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @brief Fucntion to remove a DSCP or PCP mapping entry
 *
 * This function is used to remove a mapping entry. The specific as well as
 * the default values can be removed. Removing a specific value will cause
 * that to return the default value for the profile ID, if it is programmed,
 * or the value of 0 for both the DSCP and PCP mappings. If the default is
 * removed, the will return 0 for the DSCP and PCP input values that are not
 * mapped.
 *
 * =============================================================
 */
adk_status_t
adk_qos_cp_mapping_remove(const adk_sys_handle_t        sysHandle,
                          const adk_qos_cp_mapping_t   *mappingData);

#endif /* ifndef  __ADK_QOS_DSCP_PCP_H__ */


/*!    end addtogroup _qosFunctions
 *     @}
 * end addtogroup _qos
 * @}
 */
