/* ***************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_qos_policing_include.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _qos  QoS Module
 * @{
 *
 * This file contains ADK QoS data structures, macros and API signatures
 *
 *
 *  **********************************************************************/

#ifndef __ADK_QOS_POLICING_INCLUDE_H__
#define __ADK_QOS_POLICING_INCLUDE_H__

#include "adk_sys.h"
#include <rte_meter.h>


#define ADK_QOS_CORE_CLOCK_FREQUENCY        256000000
#define ADK_QOS_POLICING_PARAM_LENGTH_64    64
#define ADK_QOS_POLICING_PARAM_MOD_LENGTH   44
#define ADK_QOS_POLICING_PARAM_LENGTH_128   128

#define ADK_QOS_MAX_NUM_POLICER   1024

/*!
 * ADK policer handle 
 */
typedef adk_handle_t    adk_qos_policer_handle_t;

/*!
 * =============================================================
 * @enum adk_qos_policing_types_t
 *
 * The types of policers supported.
 * ADK supports only color blind algorithms mentioned below.
 *
 * =============================================================
 */
typedef enum adk_qos_policing_types_e {
    ADK_QOS_POLICE_TYPE_UNKNOWN_MIN,
    /* SRTCM packet rate*/
    ADK_QOS_POLICE_TYPE_SRTCM = 1,
    ADK_QOS_POLICE_TYPE_UNKNOWN_MAX
} adk_qos_policing_types_t;   

/*!
 * =============================================================
 * @struct adk_qos_policing_params_t adk_qos_policing_include.h "adk_qos_policing_include.h"
 *
 * Defines the generic policing parameters
 *
 * =============================================================
 */
#define  ADK_QOS_POLICING_FLOW_ID_GET  0xFFFFFFFF
typedef struct adk_qos_policing_params_s {
    adk_qos_policing_types_t        type;
    /*!< Committed Information Rate in packets per second */
    uint64_t                    cir; 
    /*!< Committed Burst Size in packets */
    uint32_t                    cbs; 
    /*!< Peak Burst Size in packets */
    uint32_t                    ebs; 
} adk_qos_policing_params_t;   /* ------  end of struct adk_qos_policing_params_t  ------ */

/*!
 * =============================================================
 * @struct adk_qos_policing_params_mod_t adk_qos_policing_include.h "adk_qos_policing_include.h"
 *
 * Defines the generic policing parameters
 ** For Single rate PPS(packet/sec policing algorithm 
 * pir = peak rate in packets
 * ebs = peak burst size in packets
 *
 * For Dual rate packet/sec policing algorithm 
 * CIR = peak rate in packets
 * cbs, ebs = peak burst size in packets
 *
 * =============================================================
 */
typedef struct adk_qos_policing_params_mod_s {
    uint64_t                    cir; /*!< Committed Information Rate in bytes per second */
    uint64_t                    pir; /*!< Peak Information Rate in bytes per second */
    uint32_t                    cbs; /*!< Committed Burst Size in bytes */
    uint32_t                    ebs; /*!< Peak Burst Size in bytes */
} adk_qos_policing_params_mod_t;   /* ------  end of struct adk_qos_policing_params_mod_t  ------ */


/*!
 * =============================================================
 * @struct adk_qos_policing_flow_stats_t adk_qos_policing_include.h "adk_qos_policing_include.h"
 *
 * @brief This data structure is used to get the policing flow statistics
 *
 * =============================================================
 */
typedef struct adk_qos_policing_flow_stats_s {
    uint64_t    green_pkts; /* Number of packets marked green */
    uint64_t    yellow_pkts; /* Number of packets marked Yellow/Amber */
    uint64_t    red_pkts; /* Number of packets marked Red */
} adk_qos_policing_flow_stats_t;   /* ------  end of struct adk_qos_policing_flow_stats_t  ------ */

/*!
 *   @addtogroup _qosFunctions  QoS Functions
 *   @{
 */
/*
 * =============================================================
 * ADK QoS API signatures
 * =============================================================
 */
adk_status_t 
adk_qos_policing_policy_get(
        adk_sys_handle_t            sysHandle,
        adk_qos_policer_handle_t    policeFlowHandle,
        adk_qos_policing_params_t  *policy);

adk_status_t
adk_qos_policing_flow_create ( 
        adk_sys_handle_t            sysHandle,
        const adk_qos_policing_params_t     *policingParams_p,
        adk_qos_policer_handle_t    *policingFlowHandle_p);

adk_status_t  
adk_qos_policing_flow_delete(
    adk_sys_handle_t                sysHandle,
    adk_qos_policer_handle_t        policeFlowHandle);

adk_status_t 
adk_qos_policing_flow_stats_get(
    adk_sys_handle_t                sysHandle,
    adk_qos_policer_handle_t        policeflowHandle,
    adk_qos_policing_flow_stats_t  *flowStats_p);

#endif /* __ADK_QOS_POLICING_INCLUDE_H__ */
 
/*!
 *   @} end addtogroup _qosFunctions
 *
 * @} end addtogroup _qos
 */
