/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_qos.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _qos  QoS Module
 * @{
 *      @addtogroup _qosFunctions  QoS Functions
 *      @{
 *
 * This file contains ADK QoS Dataplane processing procedures
 *
 *
 *  **********************************************************************/
 
#include "adk_dpdk.h"

#include "adk_qos_include.h"
#include "adk_qos_handles.h"
#include "adk_queue.h"
#include "adk_qos_private.h"
#include "adk_qos_policing_private.h"
#include "adk_list.h"
#include "adk_util_list.h"
#include "adk_qos_include_internal.h"
#include "adk_sys_services.h"
#include "adk_shm.h"

static adk_status_t
adk_qos_queue_stats_resource_init(adk_sys_handle_t sysHandle);

static adk_status_t
adk_qos_queue_stats_resource_shutdown(adk_sys_handle_t sysHandle);

/*!
 * =============================================================
 * @fn adk_qos_init       
 *
 * @param[in] sysHandle System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_MEM_UNAVAILABLE Invalid memory or memory not accessible
 * @retval ADK_E_MTM_TABLE_ERROR MTM table failure
 *
 *
 * This function initializes ADK QoS Module
 *
 * =============================================================
 */
adk_status_t
adk_qos_init(adk_sys_handle_t  sysHandle)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_sys_info_t          *sysInfo_p;
    uint32_t                i;
    uint16_t                *dev_list;
    uint16_t                 numDevices;

    adk_sys_resource_info_t     *ethdev_rsrc_info_list_p;
    adk_sys_rsrc_eth_dev_info_t *eth_dev_info_p;
    uint8_t                     num_eth_devs;
    adk_qos_shm_config_t        *shm_config_p=NULL;

    /*! Validate system handle */
    if (!ADK_VALID_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_qos_init_return;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    sysInfo_p = (adk_sys_info_t *)sysHandle;
    qos_p = (adk_qos_handle_t *) ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(sysHandle),
                                1, 1, sizeof(adk_qos_handle_t));

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "[%s()] QoS Handle created %p\n", __func__, qos_p);

    if (!ADK_VALID_HANDLE(qos_p))
    {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
            goto adk_qos_init_return;
    }

    if (sysInfo_p->multiThread) {
        qos_p->queue_mod_lock = (void *) ADK_ALLOC_LOCK();
        if (qos_p->queue_mod_lock) {
            if (ADK_INIT_LOCK(qos_p->queue_mod_lock,
                                    ADK_MODULE_LOCK_SHARED,
                                    ADK_MODULE_LOCK_VALUE)) {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] Failed to initialize qos queue module lock\n",__func__);
                goto adk_qos_init_return;
            }
        } else {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "[%s()] Failed to allocate memory for qos queue module lock\n",__func__);
            goto adk_qos_init_return;
        }
        qos_p->pol_mod_lock = (void *) ADK_ALLOC_LOCK();
        if (qos_p->pol_mod_lock) {
            if (ADK_INIT_LOCK(qos_p->pol_mod_lock,
                                    ADK_MODULE_LOCK_SHARED,
                                    ADK_MODULE_LOCK_VALUE)) {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] Failed to initialize qos police module lock\n",__func__);
                goto adk_qos_init_return;
            }
        } else {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "[%s()] Failed to allocate memory for qos police module lock\n",__func__);
            goto adk_qos_init_return;
        }
    } else {
        qos_p->queue_mod_lock = NULL;
        qos_p->pol_mod_lock = NULL;
    }

    /*Set the initial default queue to INVALID queue*/
    qos_p->defaultQueueId = ADK_QOS_INVALID_QUEUE_ID;
    qos_p->ptpQueueId = ADK_QOS_INVALID_QUEUE_ID;

    /*Reset the queueStatsTbl*/
    for (i = 0; i < ADK_QOS_MAX_PHB_QUEUES; i++) {
        qos_p->queueStatsIdTbl[i] = ADK_RM_RSRC_INVALID; 
    }

    /*Reset the SL queueStatsTbl*/
    for (i = 0; i < ADK_QOS_MAX_SL_QUEUES; i++) {
        qos_p->slQueueStatsIdTbl[i] = ADK_RM_RSRC_INVALID; 
    }
    /*
     * Store the QoS handle in SYS-INFO
     */
    sysInfo_p->qosHandle = (adk_handle_t)qos_p;

    /*QoS resource table init  */
    for(i=0;i<ADK_QOS_NUM_INSTANCES;i++) {
        qos_p->qosInstance[i].queueIdRmHdl = NULL;
        qos_p->qosInstance[i].cookie =  QOS_HANDLE_COOKIE;
    }

    /*Get the list of eth-dev's registred with ADK */
    retValue = adk_sys_config_eth_dev_info_get(&ethdev_rsrc_info_list_p,
                                              &num_eth_devs);
    if (retValue != ADK_SUCCESS)
    {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "[%s()] eth-dev resource get function failed.\n",
                             __func__);
        return (retValue);
    }

    if((num_eth_devs!= 0) &&(ethdev_rsrc_info_list_p != NULL)) {
        eth_dev_info_p  = (adk_sys_rsrc_eth_dev_info_t *)
                            ethdev_rsrc_info_list_p[0].eth_dev_info_p;
        dev_list        = eth_dev_info_p->eth_dev_list;
        numDevices      = eth_dev_info_p->num_eth_devs;
        while(numDevices!=0) {
            numDevices--;
            retValue = adk_qos_instance_init(sysHandle,dev_list[numDevices]);
            if (retValue != ADK_SUCCESS) {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "QoS initialization for device %d failed\n",
                                        dev_list[numDevices]);
            } else {
                qos_p->numDevices++;
            }
        }
    }
    /* policing flow table init */
    retValue = adk_qos_policing_init(sysHandle);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "QoS policing initialization failed \n");
    }

    /* 
    * Allocated shared memory for IPC between primary(Control) and
    * secondary process(dataplne)*/
    retValue = adk_shm_create(ADK_QOS_SHM_MEMZONE_NAME, 
                              sizeof(adk_qos_shm_config_t), 
                              ADK_SHM_LOOKUP_LCORE,
                              (adk_shm_handle_t **)&shm_config_p);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "IPC shared memory allocation failed.\n");
        return (retValue);
    }

    /*Initialize the Queues stats resource*/
    retValue = adk_qos_queue_stats_resource_init(sysHandle); 
    if (retValue != ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "[%s()] PHB resource initialization failed.\n",
                             __func__);
        return (retValue);
    }

    shm_config_p->dpPolTable_p = qos_p->polInstance[0].dpPolTable;
    shm_config_p->defaultQueueId = ADK_QOS_INVALID_QUEUE_ID;
    shm_config_p->ptpQueueId = ADK_QOS_INVALID_QUEUE_ID;
    /*Reset the queueStatsTbl*/
    for (i = 0; i < ADK_QOS_MAX_PHB_QUEUES; i++) {
        shm_config_p->queueStatsIdTbl[i] = ADK_RM_RSRC_INVALID; 
    }

    /*Reset the SL queueStatsTbl*/
    for (i = 0; i < ADK_QOS_MAX_SL_QUEUES; i++) {
        shm_config_p->slQueueStatsIdTbl[i] = ADK_RM_RSRC_INVALID; 
    }

adk_qos_init_return:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN (retValue);
}    /* -----  end of function adk_qos_init  ----- */


/*!
 * =============================================================
 * @fn adk_qos_instance_init 
 *
 * @param[in] sysHandle System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_NS_RSRC_CREATE_FAILED  Resource already allocated. 
 *
 * This function initializes ADK QoS Module
 *
 * =============================================================
 */
adk_status_t
adk_qos_instance_init(adk_sys_handle_t  sysHandle,
                uint32_t devNetId) 
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_instance_handle_t   *qos_instance_p = NULL;
    struct rte_eth_dev_info     dev_info;
    uint32_t                    numIds = 0;

    /* TODO: need to add other validation */
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)((adk_sys_info_t *) sysHandle)->qosHandle;

    if (ADK_INVALID_PTR(qos_p))
    {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Invalid QOS handle or QoS is not initialized.\n");
            return retValue;
    }
    qos_instance_p = &qos_p->qosInstance[devNetId];
    if (ADK_INVALID_QOS_HANDLE(qos_instance_p))
    {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "QoS is not initialized.\n");
            return retValue;
    }
    if(qos_p->qosInstance[devNetId].queueIdRmHdl) {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_NS_RSRC_CREATE_FAILED);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Resource has already been allocated. Port ID already initialized\n");
            return retValue;
    }
    /*set the number of id based on device type */
    if (devNetId < RTE_MAX_ETHPORTS) {
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_eth_dev_info_get",
                        "Retrieve the contextual information of an Ethernet device.\n");
        rte_eth_dev_info_get(devNetId, &dev_info);
        numIds = dev_info.nb_tx_queues;
    } else {
        numIds = ADK_QOS_DEV_NET_MAX_POLICERS;
    }

    qos_instance_p->numQueueIds = numIds;
    qos_instance_p->queueIdRmHdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(sysHandle), 1, numIds);

    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    return retValue;
}
 
/*!
 * =============================================================
 * @fn adk_qos_shutdown       {{{2
 *
 * @param[in] sysHandle System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_MTM_TABLE_NOT_FOUND Data not found in MTM Table
 * @retval ADK_E_MEM_UNAVAILABLE Invalid memory or memory not accessible
 * @retval ADK_E_MTM_TABLE_ERROR MTM table failure
 *
 *
 * This function deinitializes ADK QoS Module
 *
 * =============================================================
 */
adk_status_t
adk_qos_shutdown(adk_sys_handle_t  sysHandle)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_sys_info_t          *sysInfo_p;

    uint16_t                *dev_list;
    uint16_t                 numDevices;

    adk_sys_resource_info_t     *ethdev_rsrc_info_list_p;
    adk_sys_rsrc_eth_dev_info_t *eth_dev_info_p;
    uint8_t                     num_eth_devs;

    /*! Validate system handle */
    if (!ADK_VALID_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        goto adk_qos_shutdown_return;
    }

    
    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)((adk_sys_info_t *) sysHandle)->qosHandle;
    sysInfo_p = (adk_sys_info_t *)sysHandle;
    if (!ADK_VALID_HANDLE(qos_p))
    {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
            goto adk_qos_shutdown_return;
    }


    /*Get the list of eth-dev's registred with ADK */
    retValue = adk_sys_config_eth_dev_info_get(&ethdev_rsrc_info_list_p,
                                              &num_eth_devs);
    if (retValue != ADK_SUCCESS)
    {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "[%s()] eth-dev resource get function failed.\n",
                             __func__);
        return (retValue);
    }

    if((num_eth_devs!= 0) &&(ethdev_rsrc_info_list_p != NULL)) {
        eth_dev_info_p  = (adk_sys_rsrc_eth_dev_info_t *)
                            ethdev_rsrc_info_list_p[0].eth_dev_info_p;
        dev_list        = eth_dev_info_p->eth_dev_list;
        numDevices      = eth_dev_info_p->num_eth_devs;
        while(numDevices!=0) {
            numDevices--;
            retValue = adk_qos_instance_shutdown(sysHandle,dev_list[qos_p->numDevices]);
            if (retValue != ADK_SUCCESS) {
                    retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                    ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                                    "QoS instance shutdown for device %d failed\n",
                                    dev_list[qos_p->numDevices]);
            }
        }
    }
    
    if (sysInfo_p->multiThread) {
        ADK_DESTROY_LOCK(qos_p->queue_mod_lock);
        ADK_FREE_LOCK(qos_p->queue_mod_lock);
        ADK_DESTROY_LOCK(qos_p->pol_mod_lock);
        ADK_FREE_LOCK(qos_p->pol_mod_lock);
    }
    /* Add code to delete policing information*/
    /* policing flow table init */
    retValue = adk_qos_policing_shutdown(sysHandle);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "QoS policing shutdown failed \n");
    }

    retValue = adk_qos_queue_stats_resource_shutdown(sysHandle);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "QoS PHB resource delete failed \n");
    }

    retValue = adk_shm_destroy(ADK_QOS_SHM_MEMZONE_NAME);
    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "IPC shared memory free failed.\n");
    }

    ADK_NVM_FREE(qos_p);

adk_qos_shutdown_return:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN (retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_instance_shutdown 
 *
 * @param[in] sysHandle System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_NS_RSRC_CREATE_FAILED  Resource already allocated. 
 *
 * This function initializes ADK QoS Module
 *
 * =============================================================
 */
adk_status_t
adk_qos_instance_shutdown(adk_sys_handle_t  sysHandle,
                uint32_t devNetId) 
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_instance_handle_t   *qos_instance_p = NULL;

    /* TODO: need to add other validation */
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)((adk_sys_info_t *) sysHandle)->qosHandle;

    if (ADK_INVALID_PTR(qos_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                       "Invalid QOS handle or QoS is not initialized.\n");
        return retValue;
    }
    qos_instance_p = &qos_p->qosInstance[devNetId];
    if (ADK_INVALID_QOS_HANDLE(qos_instance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                       "QoS is not initialized.\n");
        return retValue;
    }

    if(ADK_VALID_PTR(qos_instance_p->queueIdRmHdl)) {
        adk_rm_resource_destroy(qos_instance_p->queueIdRmHdl);
    }

    qos_instance_p->queueIdRmHdl = NULL;

    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    return retValue;
}

/*!
 * =============================================================
 * @fn adk_qos_policing_init
 *
 * @param[in] sysHandle System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_NS_RSRC_CREATE_FAILED  Resource already allocated. 
 *
 * This function initializes ADK QoS Module
 *
 * =============================================================
 */
adk_status_t
adk_qos_policing_init(adk_sys_handle_t  sysHandle)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_qos_handle_t            *qos_p = NULL;
    uint32_t                    numIds = 0;
    adk_qos_pol_instance_handle_t   *polInstance_p;

    /* TODO: need to add other validation */
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)((adk_sys_info_t *) sysHandle)->qosHandle;

    if (ADK_INVALID_PTR(qos_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Invalid QOS handle or QoS is not initialized.\n");
        return retValue;
    }

    polInstance_p = &(qos_p->polInstance[0]);
    polInstance_p->cookie = 0x1234;
    polInstance_p->numPolIds = numIds = ADK_QOS_DEV_NET_MAX_POLICERS;
    polInstance_p->dpPolTable = (adk_qos_police_flow_handle_t *) 
                           ADK_DP_SHM_CALLOC(ADK_MODULE_MODULEID_GET(sysHandle),
                            1, numIds, sizeof(adk_qos_police_flow_handle_t), 0);

    if (ADK_INVALID_PTR( polInstance_p->dpPolTable)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "memory allocation failed for DP policer table.\n");
        return retValue;
    }
    
    polInstance_p->polRmHdl = adk_rm_resource_init(
                               ADK_MODULE_MODULEID_GET(sysHandle), 1, numIds);

    if (ADK_INVALID_PTR( polInstance_p->polRmHdl)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "resource handle init for policer table.\n");

        goto adk_qos_policing_init_return;
    }

adk_qos_policing_init_return:
    if(retValue != ADK_SUCCESS) {

        if (ADK_VALID_PTR( polInstance_p->dpPolTable))
            ADK_DP_SHM_FREE(polInstance_p->dpPolTable);

        if (ADK_VALID_PTR( polInstance_p->polRmHdl))
            adk_rm_resource_destroy(polInstance_p->polRmHdl);
    }
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    return retValue;
}    /* -----  end of function adk_qos_policing_init  ----- */


/*!
 * =============================================================
 * @fn adk_qos_policing_shutdown
 *
 * @param[in] sysHandle System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_NS_RSRC_CREATE_FAILED  Resource already allocated. 
 *
 * This function initializes ADK QoS Module
 *
 * =============================================================
 */
adk_status_t
adk_qos_policing_shutdown(adk_sys_handle_t  sysHandle)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_pol_instance_handle_t   *polInstance_p;

    /* TODO: need to add other validation */
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)((adk_sys_info_t *) sysHandle)->qosHandle;

    if (ADK_INVALID_PTR(qos_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Invalid QOS handle or QoS is not initialized.\n");
        return retValue;
    }

    polInstance_p = &(qos_p->polInstance[0]);

    if (ADK_INVALID_PTR(polInstance_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Invalid QOS policer instance handle.\n");
        return retValue;
    }

    if (ADK_VALID_PTR( polInstance_p->dpPolTable))
        ADK_DP_SHM_FREE(polInstance_p->dpPolTable);

    if (ADK_VALID_PTR( polInstance_p->polRmHdl))
        adk_rm_resource_destroy(polInstance_p->polRmHdl);

    polInstance_p->cookie = 0x0;
    polInstance_p->numPolIds = 0x0;
    polInstance_p->dpPolTable = NULL;
    polInstance_p->polRmHdl = NULL;

    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    return retValue;
}

static adk_status_t
adk_qos_queue_stats_resource_init(adk_sys_handle_t sysHandle)
{
    adk_status_t retVal;
    adk_qos_shm_config_t    *shm_config_p=NULL;

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retVal = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retVal != ADK_SUCCESS) {
        retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    /* Create FL PHB queue resources */
    shm_config_p->phbQueuesRsrcHdl =
            adk_rm_resource_init(ADK_SYS_MODULE_ID, 0, ADK_QOS_MAX_PHB_QUEUES);
    if(shm_config_p->phbQueuesRsrcHdl== NULL)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to Allocate RM for PHB stats\n");
        goto _adk_qos_queue_stats_resource_init_exit;
    }

    /* Create SL queues resources */
    shm_config_p->slQueuesRsrcHdl =
            adk_rm_resource_init(ADK_SYS_MODULE_ID, 0, ADK_QOS_MAX_SL_QUEUES);
    if(shm_config_p->slQueuesRsrcHdl== NULL)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to Allocate RM for SL queue stats\n");
        goto _adk_qos_queue_stats_resource_init_exit;
    }

    /* Create memzone for FL queue stats */
    retVal = adk_shm_create(ADK_QOS_PHB_MEMZONE_NAME,
             (sizeof(adk_qos_phb_dp_stats_t) * ADK_QOS_MAX_PHB_QUEUES),
             ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
             (adk_shm_handle_t **)&(shm_config_p->phbStats_p));
    if (retVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to create memzone for PHB stats\n");
        goto _adk_qos_queue_stats_resource_init_exit;
    }

    /* Create memzone for SL queue stats */
    retVal = adk_shm_create(ADK_QOS_SL_MEMZONE_NAME,
             (sizeof(adk_qos_sl_queue_dp_stats_t) * ADK_QOS_MAX_SL_QUEUES),
             ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
             (adk_shm_handle_t **)&(shm_config_p->slQueueStats_p));
    if (retVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to create memzone for PHB stats\n");
        goto _adk_qos_queue_stats_resource_init_exit;
    }
    return ADK_SUCCESS;
_adk_qos_queue_stats_resource_init_exit:

    if (shm_config_p->phbQueuesRsrcHdl) {
        adk_rm_resource_destroy(shm_config_p->phbQueuesRsrcHdl); 
    }

    if (shm_config_p->slQueuesRsrcHdl) {
        adk_rm_resource_destroy(shm_config_p->slQueuesRsrcHdl); 
    }

    if (shm_config_p->phbStats_p) {
        /* Delete memzone for the PHB Queue stats */
        adk_shm_destroy(ADK_QOS_PHB_MEMZONE_NAME);
    }
    if (shm_config_p->slQueueStats_p) {
        /* Delete memzone for the PHB Queue stats */
        adk_shm_destroy(ADK_QOS_SL_MEMZONE_NAME);
    }
    return retVal;
}

static adk_status_t
adk_qos_queue_stats_resource_shutdown(adk_sys_handle_t sysHandle)
{
    adk_status_t retVal;
    adk_qos_shm_config_t    *shm_config_p=NULL;

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retVal = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retVal != ADK_SUCCESS) {
        retVal = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    /* Destroy PHB queue resources */
    adk_rm_resource_destroy(shm_config_p->phbQueuesRsrcHdl);

    /* Destroy SL queue resources */
    adk_rm_resource_destroy(shm_config_p->slQueuesRsrcHdl);

    /* Delete memzone for the PHB Queue stats */
    retVal = adk_shm_destroy(ADK_QOS_PHB_MEMZONE_NAME);
    if (retVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to destroy memzone for PHB stats\n");
        return ADK_E_MEM_UNAVAILABLE;
    }

    /* Delete memzone for the SL Queue stats */
    retVal = adk_shm_destroy(ADK_QOS_SL_MEMZONE_NAME);
    if (retVal != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to destroy memzone for SL queue stats\n");
        return ADK_E_MEM_UNAVAILABLE;
    }
    return ADK_SUCCESS;
}
/*!
 *      @} end addtogroup _qosFunctions
 * @} end addtogroup _qos
 *
 * vim: set foldmethod=marker foldlevel=1 foldcolumn=3 columns=83:
 */

