/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2001 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_qos_queue.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _qos  QoS Module
 * @{
 *     @addtogroup _qosFunctions  ADK QoS Functions
 *     @{
 *
 * @brief
 *
 * This file contains the C functions for creating/deleting queues. 
 * All utilities used along with qos queues would be placed here.
 *
 *  **********************************************************************/

#include "adk_include.h"
#include "adk_qos_include.h"
#include "adk_qos_handles.h"
#include "adk_qos_private.h"
#include "adk_sys_services.h"
#include "adk_shm.h"

#ifdef ADK_QOS_GEN3_INTERFACE
/*!
 * =============================================================
 * @fn adk_qos_queue_create     {{{2
 *
 * @param[in] sysHandle         System handle
 * @param[in] *queue_p          Pointer to queue parameters desired
 * @param[in] private_data_size Caller can store callers private
 *                              information in Queue data structures
 *                              This parameter helps in allocating
 *                              that private data along with the queue
 *                              internal data(used by ADK QOS profiles).
 *                              Application is expected to provide a
 *                              value of zero
 * @param[in,out] pointer to queue Handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_MEM_UNAVAILABLE Invalid memory or memory not accessible
 * @retval ADK_E_INVALID_SHAPING_MODE Invalid shaping mode
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_INVALID_SCHEDULING_MODE Invalid scheduling type
 * @retval ADK_E_QOS_SCHEDULER_INVALID Invalid Parent Scheduler
 * @retval ADK_E_HW_PROGRAMMING_ERROR  Failed to configure hardware
 *
 * @details
 *
 * This function adds a queue to the scheduling hierarchy
 * and returns the queue id. 
 *
 * =============================================================
 */
adk_status_t
adk_qos_queue_create(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_params_t *queue_p,
    uint32_t                private_data_size,
    adk_handle_t           *queueHandle_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_queue_t        *queueInfo_p = NULL;
    adk_qos_handle_t       *qos_p = NULL;
    adk_qos_instance_handle_t *qosInstance_p = NULL;    
    uint16_t                    qos_instance_id=0;
    uint32_t                stats_id;
    RTE_SET_USED(private_data_size);

    /* TODO: need to add other validation */    
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    /* Validate queue handle pointer */
    if (ADK_INVALID_PTR(((adk_sys_info_t *) sysHandle)->qosHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "QoS not Initialized.\n");
        goto adk_qos_queue_create_return;
    }

    /* Validate queue handle pointer */
    if (ADK_INVALID_PTR(queueHandle_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle_p.\n");
        goto adk_qos_queue_create_return;
    }
    /* Validate queue pointer */
    if (ADK_INVALID_PTR(queue_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queue_p.\n");
        goto adk_qos_queue_create_return;
    }

    adk_qos_queue_param_print(sysHandle, queue_p);

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    if(adk_qos_get_dev_instance(queue_p->deviceId, &qos_instance_id) 
       != ADK_SUCCESS)    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid device id %d\n", queue_p->deviceId);
        goto adk_qos_queue_create_return;
    } 
    qosInstance_p = &qos_p->qosInstance[qos_instance_id];
    if (ADK_INVALID_QOS_HANDLE(qosInstance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "QOS Instance is not initialized.\n");
        goto adk_qos_queue_create_return;
    }

    /* Validate queue params*/
    retValue = adk_qos_queue_params_validate(sysHandle,queue_p);
    if(retValue!=ADK_SUCCESS) {
       retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Queue Params Validation failed\n");
    goto adk_qos_queue_create_return;
    }
    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    /* Allocate QoS object and id*/
    if (adk_qos_qid_alloc(sysHandle, queue_p, &queueInfo_p)!=ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Queue ID allocation failed\n");
        goto adk_qos_queue_create_error;
    }

    /* Allocate stats id for queue stats */
    retValue = adk_sys_egress_stats_id_alloc(sysHandle,
                    &stats_id);
    if (retValue != ADK_SUCCESS) {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Stats ID Resource unavailable\n");
        goto adk_qos_queue_create_error;
    }
    queueInfo_p->statsFlow.stats_id = stats_id;
    /* Call the hard/soft queue create based on dev type/id */
    if(queue_p->deviceId <= ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = adk_qos_hard_queue_create(sysHandle, queue_p, queueInfo_p);
    } else {
        retValue = adk_qos_soft_queue_create(sysHandle, queue_p, queueInfo_p);
    }
    if(retValue!= ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Error in queue creation\n");
        goto adk_qos_queue_create_error;
    }

    *queueHandle_p = (adk_handle_t) queueInfo_p;
    goto adk_qos_queue_create_lock_return;

adk_qos_queue_create_error:
    if(queueInfo_p->nsId[0]) {
        adk_rm_resource_free(qosInstance_p->queueIdRmHdl, queueInfo_p->nsId[0]);
    }

    if(queueInfo_p->statsFlow.stats_id) {
        adk_sys_egress_stats_id_free(sysHandle, queueInfo_p->statsFlow.stats_id);
    }
    if(queueInfo_p) {
        ADK_NVM_FREE(queueInfo_p);
    }

adk_qos_queue_create_lock_return:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_acquire_failed:
_lock_release_failed:
adk_qos_queue_create_return:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_queue_stats_counters_get     {{{2
 *
 * @param[in] sysHandle     System handle
 * @param[in] queueHandle   Pointer to queue Handle
 * @param[in] queueStats_p Pointer to an instance of
 *            adk_qos_queue_stats_counters_t structure that defines the
 *            queue statistics
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_HW_PROGRAMMING_ERROR  Failed to configure hardware
 *
 * @details
 *
 * This function gets the queue statistics of a given queue
 *
 * =============================================================
 */
adk_status_t
adk_qos_queue_stats_counters_get(
   adk_sys_handle_t                 sysHandle,
   adk_handle_t                     queueHandle,
   adk_qos_queue_stats_counters_t   *queueStats_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_qos_queue_t             *queueInfo_p = NULL;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_instance_handle_t   *qos_instance_p = NULL;
    /* TODO: need to add other validation */    
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)
                ((adk_sys_info_t *) sysHandle)->qosHandle;
    if (ADK_INVALID_PTR(qos_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "QoS not Initialized\n");
        goto adk_qos_queue_stats_counters_get_return;
    }

    /* TODO: Validate queue  */
    if (ADK_INVALID_PTR(queueHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle.\n");
        goto adk_qos_queue_stats_counters_get_return;
    }
    
    queueInfo_p = (adk_qos_queue_t *) queueHandle;
    if (queueInfo_p->cookie != ADK_QOS_QUEUE_COOKIE) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Queue is invalid.\n");
        goto adk_qos_queue_stats_counters_get_return;
    }
    
    qos_instance_p = &qos_p->qosInstance[queueInfo_p->parentEthDevId];
    if (ADK_INVALID_QOS_HANDLE(qos_instance_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid QOS handle or QoS is not initialized.\n");
        goto adk_qos_queue_stats_counters_get_return;
    }
    
    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);
    if(queueInfo_p->deviceId <= ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = adk_qos_hard_queue_stats_counters_get(sysHandle, 
                                    queueInfo_p, queueStats_p);
    } else {
        retValue = adk_qos_soft_queue_stats_counters_get(sysHandle, 
                                    queueInfo_p, queueStats_p);
    }
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_acquire_failed:
_lock_release_failed:
    
adk_qos_queue_stats_counters_get_return:    
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_queue_rate_modify     {{{2
 *
 * @param[in] sysHandle     System handle
 * @param[in] queueHandle   Pointer to queue Handle
 * @param[in] newQosrates_p Pointer to structure that defines the
 *                          queue shaping rates
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_MEM_UNAVAILABLE Invalid memory or memory not accessible
 * @retval ADK_E_INVALID_SHAPING_MODE Invalid shaping mode
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_INVALID_SCHEDULING_MODE Invalid scheduling type
 * @retval ADK_E_QOS_SCHEDULER_INVALID Invalid Parent Scheduler
 * @retval ADK_E_HW_PROGRAMMING_ERROR  Failed to configure hardware
 *
 * @details
 *
 * This function modifies the shaping rate of an existing queue
 *
 * =============================================================
 */
adk_status_t
adk_qos_queue_rate_modify(
    adk_sys_handle_t        sysHandle,
    adk_handle_t            queueHandle,
    const adk_qos_rates_t   *newQosrates_p)
{
    adk_status_t retValue=ADK_SUCCESS;
    adk_qos_queue_t             *queueInfo_p = NULL;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_instance_handle_t   *qos_instance_p = NULL;

    /* TODO: need to add other validation */    
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)
                ((adk_sys_info_t *) sysHandle)->qosHandle;
    if (ADK_INVALID_PTR(qos_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "QoS not initialized \n");
        goto adk_qos_queue_rate_modify_return;
    }

    /* TODO: Validate queue  */
    if (ADK_INVALID_PTR(queueHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle.\n");
        goto adk_qos_queue_rate_modify_return;
    }
    
    queueInfo_p = (adk_qos_queue_t *) queueHandle;
    if (queueInfo_p->cookie != ADK_QOS_QUEUE_COOKIE)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Queue is invalid.\n");
        goto adk_qos_queue_rate_modify_return;
    }
    
    qos_instance_p = &qos_p->qosInstance[queueInfo_p->parentEthDevId];
    if (ADK_INVALID_QOS_HANDLE(qos_instance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid QOS handle or QoS is not initialized.\n");
        return retValue;
    }
    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);
    /* Call the hard/soft queue create based on dev type/id */
    if(queueInfo_p->deviceId <= ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = adk_qos_hard_queue_rate_modify(sysHandle, 
                                    queueInfo_p, newQosrates_p );
    } else {
/*
        retValue = adk_qos_soft_queue_rate_modify(sysHandle, 
                                    queueInfo_p, newQosrates_p);
*/
    }
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_acquire_failed:
_lock_release_failed:
    
adk_qos_queue_rate_modify_return:
     ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_queue_service_quantum_modify     {{{2
 *
 * @param[in] sysHandle System handle
 * @param[in] queueHandle  Queue Handle
 * @param[in] service_quantum service Quantum value
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_MEM_UNAVAILABLE Invalid memory or memory not accessible
 * @retval ADK_E_INVALID_SHAPING_MODE Invalid shaping mode
 * @retval ADK_E_INVALID_PARAMETER Invalid queue parameter
 * @retval ADK_E_INVALID_SCHEDULING_MODE Invalid scheduling type
 * @retval ADK_E_HW_PROGRAMMING_ERROR  Failed to configure hardware
 * @retval ADK_E_INTERNAL_ERROR Internal error
 *
 * @details
 *
 * This function modifies the service quantum of a Queue
 *
 * =============================================================
 */
adk_status_t
adk_qos_queue_service_quantum_modify(
    adk_sys_handle_t        sysHandle,
    adk_handle_t            queueHandle,
    const adk_qos_service_quantum_t service_quantum)
{
    adk_status_t retValue;
    adk_qos_queue_t             *queueInfo_p = NULL;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_instance_handle_t   *qos_instance_p = NULL;
    struct rte_tm_error         tm_error;
    int                         ret; 

    /* TODO: need to add other validation */    
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)
                ((adk_sys_info_t *) sysHandle)->qosHandle;
    if (ADK_INVALID_PTR(qos_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle_p.\n");
        goto adk_qos_queue_service_quantum_modify_return;
    }

    /* TODO: Validate queue  */
    if (ADK_INVALID_PTR(queueHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle_p.\n");
        goto adk_qos_queue_service_quantum_modify_return;
    }
    
    queueInfo_p = (adk_qos_queue_t *) queueHandle;
    if (queueInfo_p->cookie != ADK_QOS_QUEUE_COOKIE)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "queue_p->cookie is invalid.\n");
        goto adk_qos_queue_service_quantum_modify_return;
    }

    if (queueInfo_p->parentEthDevId > ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Queue rate modify is not supported for this queue\n");
        goto adk_qos_queue_service_quantum_modify_return; 
    }
    
    qos_instance_p = &qos_p->qosInstance[queueInfo_p->parentEthDevId];
    if (ADK_INVALID_QOS_HANDLE(qos_instance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid QOS handle or QoS is not initialized.\n");
        goto adk_qos_queue_service_quantum_modify_return;
    }
    if (queueInfo_p->parentEthDevId > ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "BM policy set is not supported for this queue\n");
        goto adk_qos_queue_service_quantum_modify_return;
    }
    queueInfo_p->weight = service_quantum;
    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_node_parent_update",
                    "Traffic manager node parent update.\n");
    ret = rte_tm_node_parent_update(queueInfo_p->deviceId,
                    queueInfo_p->queueId, queueInfo_p->parentNodeId,
                    queueInfo_p->priority, queueInfo_p->weight,
                    &tm_error);
    if (ret) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tm_error.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] rte_tm_node_parent_update() failed"
                    " for port %u, service_quantum %u with error %d\n",
                    __func__, queueInfo_p->deviceId,
                     queueInfo_p->weight, ret);
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Service Quantum %u modified for node %u\n", 
                    queueInfo_p->weight,queueInfo_p->queueId);
    }
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_acquire_failed:
_lock_release_failed:
adk_qos_queue_service_quantum_modify_return:
     ADK_RETURN(retValue);
}

adk_status_t
adk_qos_queue_bm_policy_set(
    adk_sys_handle_t        sysHandle,
    adk_handle_t           queueHandle,
    adk_qos_bm_policy_t   *bm_policy_p)
{
    adk_status_t retValue;
    adk_qos_queue_t             *queueInfo = NULL;
    adk_qos_handle_t            *qos_p = NULL;
    adk_qos_instance_handle_t   *qos_instance_p = NULL;
    struct rte_tm_error         tm_error;
    int                         ret; 
    uint32_t id=0, prevWredId,currWredId;

    /* TODO: need to add other validation */    
    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)
                ((adk_sys_info_t *) sysHandle)->qosHandle;
    if (ADK_INVALID_PTR(qos_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "QoS not initialized \n");
        goto adk_qos_queue_bm_policy_set_return;
    }

    /* TODO: Validate queue  */
    if (ADK_INVALID_PTR(queueHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle_p.\n");
        goto adk_qos_queue_bm_policy_set_return;
    }
    queueInfo = (adk_qos_queue_t *) queueHandle;
    if (queueInfo->cookie != ADK_QOS_QUEUE_COOKIE)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Queue is invalid.\n");
        goto adk_qos_queue_bm_policy_set_return;
    }

    if (queueInfo->parentEthDevId > ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "BM policy set is not supported for this queue\n");
        goto adk_qos_queue_bm_policy_set_return;
    }

    id  = queueInfo->parentEthDevId;
    qos_instance_p = &qos_p->qosInstance[id];
    if (ADK_INVALID_QOS_HANDLE(qos_instance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid QOS handle or QoS is not initialized.\n");
        return retValue;
    }

    prevWredId = queueInfo->wredId;
    /* This will reset the bit(even for initialized value of 2)
     */ 
    currWredId = (queueInfo->wredId & 1) ^ 1;
    /*ToDO: Fill WRED information from queue params into profile*/
    retValue = adk_qos_bm_params_fill(sysHandle, bm_policy_p, queueInfo);
    if (retValue != ADK_SUCCESS) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid parameter\n");
        goto adk_qos_queue_bm_policy_set_return;
    }
    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);
    /* Create a WRED profile to be added to all queues */
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_wred_profile_add",
                    "Traffic manager WRED profile add.\n");
    ret = rte_tm_wred_profile_add(queueInfo->deviceId, 
                    queueInfo->nsId[currWredId], &queueInfo->wredProfile, 
                    &tm_error);
    if (ret) {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tm_error.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] rte_tm_wred_profile_add() failed"
                    " for port %u, wred profile id %u with error %d\n",
                    __func__, queueInfo->deviceId,
                    queueInfo->nsId[currWredId], ret);
            goto adk_qos_queue_bm_policy_set_lock_return;
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "WRED Profile created with id %u\n", 
                    queueInfo->nsId[currWredId]); 
    }
    
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_node_wred_context_update",
                    "Traffic manager node private WRED context update.\n");
    ret = rte_tm_node_wred_context_update(queueInfo->deviceId,
                    queueInfo->queueId, queueInfo->nsId[currWredId],
                    &tm_error);
    if (ret) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tm_error.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] rte_tm_node_wred_context_update() failed"
                    " for port %u, wred profile id %u with error %d\n",
                    __func__, queueInfo->deviceId,
                    queueInfo->nsId[currWredId], ret);
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_tm_wred_profile_delete",
                        "Traffic manager WRED profile delete.\n");
        ret = rte_tm_wred_profile_delete(queueInfo->deviceId, 
                        queueInfo->nsId[currWredId], &tm_error);
        goto adk_qos_queue_bm_policy_set_lock_return;
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "WRED Profile %u added to node %u\n", 
                    queueInfo->nsId[currWredId],queueInfo->queueId);
    }
    queueInfo->wredId = currWredId;
    if(prevWredId < 2) {
        /* Delete the previous  WRED profile*/
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_tm_wred_profile_delete",
                        "Traffic manager WRED profile delete.\n");
        ret = rte_tm_wred_profile_delete(queueInfo->deviceId, 
                        queueInfo->nsId[prevWredId], &tm_error);
        if (ret) {
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Error: %s\n", tm_error.message);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "[%s()] rte_tm_wred_profile_delete() failed"
                        " for port %u, wred profile id %u with error %d\n",
                        __func__, queueInfo->deviceId,
                        queueInfo->nsId[prevWredId], ret);
        goto adk_qos_queue_bm_policy_set_lock_return;
        } else {
                ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "WRED Profile deleted with id %u\n", 
                        queueInfo->nsId[prevWredId]);
        }
    }
adk_qos_queue_bm_policy_set_lock_return:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_acquire_failed:
_lock_release_failed:
adk_qos_queue_bm_policy_set_return:
     ADK_RETURN(retValue);
}

adk_status_t
adk_qos_queue_delete(
    adk_sys_handle_t        sysHandle,
    adk_handle_t           queueHandle,
    uint32_t        sleepTimeMicroSeconds,
    uint32_t        retryCount)
{
    adk_status_t retValue=ADK_SUCCESS, ret_val;
    adk_qos_queue_t        *queueInfo_p = NULL;
    adk_qos_handle_t    *qos_p = NULL;
    adk_qos_instance_handle_t   *qosInstance_p = NULL;
    uint32_t id;

    RTE_SET_USED(sleepTimeMicroSeconds);
    RTE_SET_USED(retryCount);
    /* TODO: need to add other validation */    

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *)
                ((adk_sys_info_t *) sysHandle)->qosHandle;
    if (ADK_INVALID_PTR(qos_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Qos Not Initialized\n");
        return retValue;
    }

    /* TODO: Validate queue  */
    if (ADK_INVALID_PTR(queueHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer: queueHandle_p.\n");
        return retValue;
    }
    queueInfo_p = (adk_qos_queue_t *) queueHandle;
    if (queueInfo_p->cookie != ADK_QOS_QUEUE_COOKIE)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "QueueHandle is invalid.\n");
        return retValue;
    }
    id  = queueInfo_p->parentEthDevId;

    qosInstance_p = &qos_p->qosInstance[id];
    if (ADK_INVALID_QOS_HANDLE(qosInstance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid QOS handle or QoS is not initialized.\n");
        return retValue;
    }

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);
    if(queueInfo_p->deviceId <= ADK_DEV_NET_ID_RSVD_MAX) {
        retValue = adk_qos_hard_queue_delete(sysHandle, queueInfo_p);
    } else {
        retValue = adk_qos_soft_queue_delete(sysHandle, queueInfo_p);
    }

    if(adk_rm_resource_free(qosInstance_p->queueIdRmHdl, queueInfo_p->nsId[0])) {
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Resource with ID %u deleted\n",
                            queueInfo_p->nsId[0]);
    } else {
        retValue += ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Resource with ID %u deletion failed \n",
                            queueInfo_p->nsId[0]);
    }

    retValue = adk_sys_egress_stats_id_free(sysHandle, queueInfo_p->statsFlow.stats_id);
    if (retValue != ADK_SUCCESS){
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Stats ID Resource free failed id %d \n", 
                        queueInfo_p->statsFlow.stats_id);
    }
    ADK_NVM_FREE(queueInfo_p);
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,ret_val);
_lock_acquire_failed:
_lock_release_failed:
     ADK_RETURN(retValue);
}

#endif

/*!
 * =============================================================
 * @fn adk_qos_default_queue_id_set 
 *
 * @param[in] sysHandle         System handle
 * @param[in] queueId           Queue Id 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available 
 *
 * @details
 *
 * This function set the default transmit queue id. 
 * This queue is used only in cases where QoS is disabled on an 
 * interface. 
 * =============================================================
 */
adk_status_t
adk_qos_default_queue_id_set(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t       *qos_p = NULL;
    adk_qos_shm_config_t *shm_config_p = NULL;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);    

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        goto _adk_qos_default_queue_add_exit;
    }

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    qos_p->defaultQueueId = queueId;
    shm_config_p->defaultQueueId = queueId;

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
_adk_qos_default_queue_add_exit:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_default_queue_id_reset 
 *
 * @param[in] sysHandle         System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available 
 *
 * @details
 *
 * This function reset the default transmit queue. 
 * This queue is used only in cases where QoS is disabled on an 
 * interface. 
 * =============================================================
 */
adk_status_t
adk_qos_default_queue_id_reset(
    adk_sys_handle_t        sysHandle)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t       *qos_p = NULL;
    adk_qos_shm_config_t *shm_config_p = NULL;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);    

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    qos_p->defaultQueueId = ADK_QOS_INVALID_QUEUE_ID;
    shm_config_p->defaultQueueId = ADK_QOS_INVALID_QUEUE_ID;

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_ptp_queue_id_set
 *
 * @param[in] sysHandle         System handle
 * @param[in] queueId           Queue Id
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function set the transmit queue for PTP traffic.
 * =============================================================
 */
adk_status_t
adk_qos_ptp_queue_id_set(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t       *qos_p = NULL;
    adk_qos_shm_config_t *shm_config_p = NULL;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        goto _adk_qos_ptp_queue_id_set_exit;
    }

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    qos_p->ptpQueueId = queueId;
    shm_config_p->ptpQueueId = queueId;

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
_adk_qos_ptp_queue_id_set_exit:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_ptp_queue_id_reset
 *
 * @param[in] sysHandle         System handle
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function reset the transmit queue for PTP traffic.
 * =============================================================
 */
adk_status_t
adk_qos_ptp_queue_id_reset(
    adk_sys_handle_t        sysHandle)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t       *qos_p = NULL;
    adk_qos_shm_config_t *shm_config_p = NULL;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    qos_p->ptpQueueId = ADK_QOS_INVALID_QUEUE_ID;
    shm_config_p->ptpQueueId = ADK_QOS_INVALID_QUEUE_ID;

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_phb_queue_register
 *
 * @param[in] sysHandle         System handle
 * @param[in] queueId           Queue Id
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function register the FL PHB queue in QoS module to maintain
 * statistics.
 * =============================================================
 */
adk_status_t
adk_qos_phb_queue_register(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_qos_shm_config_t    *shm_config_p = NULL;
    uint32_t                stats_id = 0;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    stats_id = adk_rm_resource_reserve(shm_config_p->phbQueuesRsrcHdl);
    if(stats_id == ADK_RM_RSRC_INVALID)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to Allocate Id for PHB stats \n");
        return ADK_E_RESOURCE_UNAVAILABLE;
    }

    /*Reset the counters*/
    memset(&shm_config_p->phbStats_p[stats_id], 0 , sizeof(adk_qos_phb_dp_stats_t)) ;
    /*Store the Statistics Id to the CP table*/
    qos_p->queueStatsIdTbl[queueId] = stats_id;
    shm_config_p->queueStatsIdTbl[queueId] = stats_id;
    adk_rcu_synchronize();

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_phb_queue_unregister
 *
 * @param[in] sysHandle         System handle
 * @param[in] queueId           Queue Id
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function unregister the PHB queue from QoS module.
 * Application should make sure that statistics for the given 
 * queue are fetched before calling this API.
 * =============================================================
 */
adk_status_t
adk_qos_phb_queue_unregister(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_qos_shm_config_t    *shm_config_p = NULL;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    if (ADK_RM_RSRC_INVALID != qos_p->queueStatsIdTbl[queueId]) {

        if (!adk_rm_resource_isset(shm_config_p->phbQueuesRsrcHdl, qos_p->queueStatsIdTbl[queueId])) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Invalid stats id, Failed to free id for PHB stats\n");
            return ADK_E_INVALID_PARAMETER;
        }

        adk_rm_resource_free(shm_config_p->phbQueuesRsrcHdl, qos_p->queueStatsIdTbl[queueId]);
    } else {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid queue id passed, Not registered queue in QoS\n");
        return ADK_E_INVALID_PARAMETER;
    }

    /*Reset the queue id with invalid statistics id */
    qos_p->queueStatsIdTbl[queueId] = ADK_RM_RSRC_INVALID;
    shm_config_p->queueStatsIdTbl[queueId] = ADK_RM_RSRC_INVALID;
    adk_rcu_synchronize();

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * @fn adk_qos_get_phb_queue_statistics
 *
 * @param[in]  sysHandle         System handle
 * @param[in]  queueId           Queue Id
 * @param[out] phbQueueStats_p   Queue statistics
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function fetches the PHB statistics for the given queue Id.
 * =============================================================
 */
adk_status_t
adk_qos_get_phb_queue_statistics(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId,
    adk_qos_phb_queue_statistics_t *phbQueueStats_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_qos_shm_config_t    *shm_config_p = NULL;
    uint32_t                stats_id = 0;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        return retValue;
    }
    /* Validate stats pointer*/
    if ( NULL == phbQueueStats_p) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    stats_id = qos_p->queueStatsIdTbl[queueId];
    
    if (ADK_RM_RSRC_INVALID == stats_id) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Queue not registered \n");
        goto _adk_qos_get_phb_queue_statistics_exit;
    }

    /*Fetch the counters*/
    phbQueueStats_p->ipv4TxPkts          = shm_config_p->phbStats_p[stats_id].ipv4TxPkts;
    phbQueueStats_p->ipv4TxBytes         = shm_config_p->phbStats_p[stats_id].ipv4TxBytes;
    phbQueueStats_p->ipv4TxPktsDropped   = shm_config_p->phbStats_p[stats_id].ipv4TxPktsDropped;
    phbQueueStats_p->ipv4TxBytesDropped  = shm_config_p->phbStats_p[stats_id].ipv4TxBytesDropped;
    phbQueueStats_p->ipv6TxPkts          = shm_config_p->phbStats_p[stats_id].ipv6TxPkts;
    phbQueueStats_p->ipv6TxBytes         = shm_config_p->phbStats_p[stats_id].ipv6TxBytes;
    phbQueueStats_p->ipv6TxPktsDropped   = shm_config_p->phbStats_p[stats_id].ipv6TxPktsDropped;
    phbQueueStats_p->ipv6TxBytesDropped  = shm_config_p->phbStats_p[stats_id].ipv6TxBytesDropped;

_adk_qos_get_phb_queue_statistics_exit:
_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_RETURN(retValue);
    
}

/*!
 * =============================================================
 * @fn adk_qos_sl_queue_register
 *
 * @param[in] sysHandle         System handle
 * @param[in] queueId           Queue Id
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function register the SL queue in QoS module to maintain
 * statistics.
 * =============================================================
 */
adk_status_t
adk_qos_sl_queue_register(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_qos_shm_config_t    *shm_config_p = NULL;
    uint32_t                stats_id = 0;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    stats_id = adk_rm_resource_reserve(shm_config_p->slQueuesRsrcHdl);
    if(stats_id == ADK_RM_RSRC_INVALID)
    {
        ADK_LOG_ERROR (sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                "Failed to Allocate Id for SL queue stats \n");
        return ADK_E_RESOURCE_UNAVAILABLE;
    }

    /*Reset the counters*/
    memset(&shm_config_p->slQueueStats_p[stats_id], 0 , sizeof(adk_qos_sl_queue_dp_stats_t)) ;
    /*Store the Statistics Id to the CP table*/
    qos_p->slQueueStatsIdTbl[queueId] = stats_id;
    shm_config_p->slQueueStatsIdTbl[queueId] = stats_id;
    adk_rcu_synchronize();

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_qos_sl_queue_unregister
 *
 * @param[in] sysHandle         System handle
 * @param[in] queueId           Queue Id
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function unregister the SL queue from QoS module.
 * Application should make sure that statistics for the given 
 * queue are fetched before calling this API.
 * =============================================================
 */
adk_status_t
adk_qos_sl_queue_unregister(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_qos_shm_config_t    *shm_config_p = NULL;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    ADK_QOS_ENTRY_LOG_DEBUG(sysHandle);
    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    if (ADK_RM_RSRC_INVALID != qos_p->slQueueStatsIdTbl[queueId]) {

        if (!adk_rm_resource_isset(shm_config_p->slQueuesRsrcHdl, qos_p->slQueueStatsIdTbl[queueId])) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Invalid stats id, Failed to free id for SL queue stats resource \n");
            return ADK_E_INVALID_PARAMETER;
        }

        adk_rm_resource_free(shm_config_p->slQueuesRsrcHdl, qos_p->slQueueStatsIdTbl[queueId]);
    } else {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid queue id passed, Not registered queue in QoS\n");
        return ADK_E_INVALID_PARAMETER;
    }

    /*Reset the queue id with invalid statistics id */
    qos_p->slQueueStatsIdTbl[queueId] = ADK_RM_RSRC_INVALID;
    shm_config_p->slQueueStatsIdTbl[queueId] = ADK_RM_RSRC_INVALID;
    adk_rcu_synchronize();

_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_QOS_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * @fn adk_qos_get_sl_queue_statistics
 *
 * @param[in]  sysHandle         System handle
 * @param[in]  queueId           Queue Id
 * @param[out] slQueueStats_p    Queue statistics
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_MEM_UNAVAILABLE Memory not available
 *
 * @details
 *
 * This function fetches the SL queue statistics for the given 
 * queue Id.
 * =============================================================
 */
adk_status_t
adk_qos_get_sl_queue_statistics(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_id_t      queueId,
    adk_qos_sl_queue_statistics_t *slQueueStats_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_qos_handle_t        *qos_p = NULL;
    adk_qos_shm_config_t    *shm_config_p = NULL;
    uint32_t                stats_id = 0;

    /* Validate system handle */
    if (ADK_INVALID_SYS_HANDLE(sysHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Invalid System Handle.\n");
        return retValue;
    }

    /* Validate queue Id */
    if ( ADK_QOS_INVALID_QUEUE_ID == queueId) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid queue id.\n");
        return retValue;
    }
    /* Validate stats pointer*/
    if ( NULL == slQueueStats_p) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid pointer.\n");
        return retValue;
    }

    /* Lookup Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);

    if (retValue != ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to fetch QoS DP memory.\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;

    ADK_QOS_ACQUIRE_LOCK(qos_p->queue_mod_lock,retValue);

    stats_id = qos_p->slQueueStatsIdTbl[queueId];
    
    if (ADK_RM_RSRC_INVALID == stats_id) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Queue not registered \n");
        goto _adk_qos_get_sl_queue_statistics_exit;
    }

    /*Fetch the counters*/
    slQueueStats_p->nonIPTxPkts         = shm_config_p->slQueueStats_p[stats_id].nonIPTxPkts;
    slQueueStats_p->nonIPTxBytes        = shm_config_p->slQueueStats_p[stats_id].nonIPTxBytes;
    slQueueStats_p->nonIPTxPktsDropped  = shm_config_p->slQueueStats_p[stats_id].nonIPTxPktsDropped;
    slQueueStats_p->nonIPTxBytesDropped = shm_config_p->slQueueStats_p[stats_id].nonIPTxBytesDropped;
    slQueueStats_p->ipv4TxPkts          = shm_config_p->slQueueStats_p[stats_id].ipv4TxPkts;
    slQueueStats_p->ipv4TxBytes         = shm_config_p->slQueueStats_p[stats_id].ipv4TxBytes;
    slQueueStats_p->ipv4TxPktsDropped   = shm_config_p->slQueueStats_p[stats_id].ipv4TxPktsDropped;
    slQueueStats_p->ipv4TxBytesDropped  = shm_config_p->slQueueStats_p[stats_id].ipv4TxBytesDropped;
    slQueueStats_p->ipv6TxPkts          = shm_config_p->slQueueStats_p[stats_id].ipv6TxPkts;
    slQueueStats_p->ipv6TxBytes         = shm_config_p->slQueueStats_p[stats_id].ipv6TxBytes;
    slQueueStats_p->ipv6TxPktsDropped   = shm_config_p->slQueueStats_p[stats_id].ipv6TxPktsDropped;
    slQueueStats_p->ipv6TxBytesDropped  = shm_config_p->slQueueStats_p[stats_id].ipv6TxBytesDropped;

_adk_qos_get_sl_queue_statistics_exit:
_lock_acquire_failed:
    ADK_QOS_RELEASE_LOCK(qos_p->queue_mod_lock,retValue);
_lock_release_failed:
    ADK_RETURN(retValue);
    
}
