/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2001 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_qos_queue_internal.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _qos  QoS Module
 * @{
 *     @addtogroup _qosFunctions  ADK QoS Functions
 *     @{
 *
 * @brief
 *
 * This file contains the C functions for creating/deleting queues. 
 * All utilities used along with qos queues would be placed here.
 *
 *  **********************************************************************/

#include "adk_include.h"
#include "adk_qos_include.h"
#include "adk_qos_handles.h"
#include "adk_qos_private.h"
#include "adk_qos_include_internal.h"
#include "adk_qos_policing_include.h"

#ifdef ADK_DEVICE_SNR
#include <rte_pmd_ice_dsi_tm.h>
#endif

#ifdef ADK_QOS_GEN3_INTERFACE

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_queue_params_validate        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 * @retval ADK_E_INVALID_SHAPING_MODE Invalid shaping mode
 *
 * @details
 *
 * Validates queue parameters
 *
 * =============================================================
 */
adk_status_t
adk_qos_queue_params_validate(
    adk_sys_handle_t           sysHandle,
    adk_qos_queue_params_t    *queueParams_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

    /*check shaping mode based on device type */
    if(queueParams_p->deviceId > ADK_DEV_NET_ID_RSVD_MAX) {
        
        if (queueParams_p->shapingMode != ADK_QOS_PPS) {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),  
                "Invalid shaping mode: supports ADK_QOS_PPS only \n");
            goto adk_qos_queue_params_validate_return;
        }
    }
    retValue = adk_qos_queue_shaper_params_validate(sysHandle,
                    queueParams_p->shapingMode, queueParams_p->qosRates);

adk_qos_queue_params_validate_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_queue_params_validate  ----- */

adk_status_t
adk_qos_queue_shaper_params_validate(
    adk_sys_handle_t        sysHandle,
    adk_qos_shaping_mode_t  shapingMode,
    const adk_qos_rates_t   qosRates)
{
    adk_status_t retValue = ADK_SUCCESS;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

    /* Validate QoS rates against the parent's QoS rates */
    switch (shapingMode)
    {
        case ADK_QOS_PR:
            break;

        case ADK_QOS_PR_SR:
            if ((qosRates.sr_mr_u.sustain.rate > 
                                    qosRates.peak.rate)) 
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),  
                              "Invalid Parameter: "
                              "queue sustained rate (%lu) > "
                              "queue peak rate (%lu)\n",
                              qosRates.sr_mr_u.sustain.rate,
                              qosRates.peak.rate);
                goto adk_qos_queue_params_validate_return;
            }
            break;
        case ADK_QOS_NO_SHAPING:
            /* Rates are ignored */
            break;
        case ADK_QOS_PPS:
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_SHAPING_MODE);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),  
                          "Invalid Shaping Mode.\n");
            goto adk_qos_queue_params_validate_return; 
            break;
    }

adk_qos_queue_params_validate_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_queue_params_validate  ----- */


/*!
 * =============================================================
 * @internal
 * @fn adk_qos_qid_alloc        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_qid_alloc(
    adk_sys_handle_t           sysHandle,
    adk_qos_queue_params_t    *queueParams_p,
    adk_qos_queue_t           **queue_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_qos_handle_t            *qos_p;
    adk_qos_instance_handle_t   *qosInstance_p;
    uint16_t                    deviceId;
    uint16_t                    qosInstanceId;
    uint32_t                    queueId;
    adk_qos_queue_t             *queueInfo_p;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

    deviceId = queueParams_p->deviceId;

    if(adk_qos_get_dev_instance(queueParams_p->deviceId, &qosInstanceId) 
       != ADK_SUCCESS)    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid device id %d\n", deviceId);
        goto adk_qos_qid_alloc_return;
    }

    qos_p = (adk_qos_handle_t *) ((adk_sys_info_t *) sysHandle)->qosHandle;
    qosInstance_p = &qos_p->qosInstance[qosInstanceId];

    if (ADK_INVALID_QOS_HANDLE(qosInstance_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "QOS Instance is not initialized.\n");
        goto adk_qos_qid_alloc_return;
    }

    *queue_p = (adk_qos_queue_t *)ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(sysHandle), 1, 1, (sizeof(adk_qos_queue_t)));

    if (!*queue_p)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Memory allocation failed for queue object.\n");
        goto adk_qos_qid_alloc_return;
    }
    queueInfo_p = *queue_p;

    /*Allocate Queue ID  */
    queueId = adk_rm_resource_reserve(qosInstance_p->queueIdRmHdl);

    if (queueId == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_RESOURCE_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Queue ID Namespace Resource unavailable\n");
        goto adk_qos_qid_alloc_return;
    }

    queueInfo_p->nsId[0]=queueId;
    /*This instance is used only by HW queue. Eventhough we assign a number here,
    * we are not allocating the same from Resource Manager. So we don't have to
    * free the same during and has no significance for SW Queue in ingress path    
    */

    queueInfo_p->nsId[1]= queueInfo_p->nsId[0] + qosInstance_p->numQueueIds;
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Reserved NS ID - %u and %u\n", queueInfo_p->nsId[0],
                        queueInfo_p->nsId[1]);

    queueInfo_p->parentEthDevId = qosInstanceId;

adk_qos_qid_alloc_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_qid_alloc  ----- */

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_hard_queue_create        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_hard_queue_create(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_params_t  *queue_p,
    adk_qos_queue_t         *queueInfo_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    int                         ret;
    struct rte_tm_error         tmError;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

    /* TODO: Fill shaper profile information*/
    retValue = adk_qos_queue_info_fill(sysHandle, queue_p, queueInfo_p);
    if(retValue!= ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Error in filling queue params\n");
        goto adk_qos_hard_queue_create_return;
    }
    if(queue_p->shapingMode != ADK_QOS_NO_SHAPING) {
        queueInfo_p->shaperId = 0;
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_tm_shaper_profile_add",
                        "Traffic manager shaper profile add.\n");
        ret = rte_tm_shaper_profile_add(queue_p->deviceId, 
                        queueInfo_p->nsId[queueInfo_p->shaperId], 
                        &queueInfo_p->shaperProfile, &tmError);
        if(ret) {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "\n  rte_tm_shaper_profile_add failed with error %s\n",
            tmError.message);
            /*TODO: Add an error*/
            goto adk_qos_hard_queue_create_return;
        } else {
           ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Shaper created with id %u\n",
                queueInfo_p->nsId[queueInfo_p->shaperId]);
            /*TODO: Add an error*/

        }
        queueInfo_p->nodeParams.shaper_profile_id = 
                queueInfo_p->nsId[queueInfo_p->shaperId];
    }

    /* Create a WRED profile to be added to all queues */
    queueInfo_p->wredId = 0;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].min_th =  0xFFFFFFFF - 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].max_th =  0xFFFFFFFF;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].maxp_inv = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].wq_log2 = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].min_th = 10;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].max_th = 11;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].maxp_inv = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].wq_log2 = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].min_th = 10;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].max_th = 11;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].maxp_inv = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].wq_log2 = 1;
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_wred_profile_add",
                    "Traffic manager WRED profile add.\n");
    ret = rte_tm_wred_profile_add(queueInfo_p->deviceId,
                    queueInfo_p->nsId[queueInfo_p->wredId], &queueInfo_p->wredProfile,
                    &tmError);
    if (ret) {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tmError.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] rte_tm_wred_profile_add() failed"
                    " for port %u, wred profile id %u with error %d\n",
                    __func__, queueInfo_p->deviceId,
                    queueInfo_p->nsId[queueInfo_p->wredId], ret);
            goto adk_qos_hard_queue_create_return;
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "WRED Profile created with id %u\n",
                    queueInfo_p->nsId[queueInfo_p->wredId]);
    }
    queueInfo_p->nodeParams.leaf.wred.wred_profile_id = 
            queueInfo_p->nsId[queueInfo_p->wredId];

    /* TODO: 
     * 1. Create node_params from queue_params
     * 2. Assign node_id from namespace
     * 3. 
     */
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_node_add",
                    "Traffic manager node add.\n");
    ret = rte_tm_node_add(queue_p->deviceId,
                queueInfo_p->queueId,
                queueInfo_p->parentNodeId,
                queueInfo_p->priority, /* fill from queue_params*/
                queueInfo_p->weight, /* fill from queue params*/
                RTE_TM_NODE_LEVEL_ID_ANY,
                &queueInfo_p->nodeParams,
                &tmError);
    if (ret) {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tmError.message);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "rte_tm_node_add() failed to add queue %d leaf node."
                        " errno = %d\n", queueInfo_p->queueId, ret);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
        goto adk_qos_hard_queue_create_return; /* TODO: Add ADK failure*/
    } else {
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Created queue with queue id %d\n", 
                        (queueInfo_p->queueId));
    }

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_eth_dev_tx_queue_start",
                    "Start TX for specified queue of a port. It is used when tx_deferred_start flag of the specified queue is true.\n");
    ret = rte_eth_dev_tx_queue_start(queueInfo_p->parentEthDevId, 
                            queueInfo_p->queueId);
    if (ret) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            " rte_eth_dev_tx_queue_start() failed to eth-dev-id:%d "
                     " queue-id:%d with errno = %d\n", queueInfo_p->parentEthDevId, 
                         queueInfo_p->queueId, ret);
            goto adk_qos_hard_queue_create_return;
    }

    goto adk_qos_hard_queue_create_return;

adk_qos_hard_queue_create_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_hard_queue_create ----- */

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_hard_queue_delete        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Delete HW Queue 
 *
 * =============================================================
 */
adk_status_t
adk_qos_hard_queue_delete(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_t         *queueInfo_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    int                         ret;
    struct rte_tm_error         tmError;
    struct rte_eth_dev_info     dev_info;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
    
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_eth_dev_info_get",
                    "Retrieve the contextual information of an Ethernet device.\n");
    rte_eth_dev_info_get(queueInfo_p->deviceId, &dev_info);
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_eth_dev_tx_queue_stop",
                    "Stop specified TX queue of a port.\n");
    ret = rte_eth_dev_tx_queue_stop(queueInfo_p->parentEthDevId, 
                            queueInfo_p->queueId);
    if (ret) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            " rte_eth_dev_tx_queue_stop() failed to eth-dev-id:%d "
                     " queue-id:%d with errno = %d\n", queueInfo_p->parentEthDevId, 
                         queueInfo_p->queueId, ret);
            goto adk_qos_hard_queue_delete_return;
    }

    ret = rte_tm_node_delete(queueInfo_p->deviceId,
                queueInfo_p->queueId,
                &tmError);
    if (ret) {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tmError.message);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_tm_node_delete() failed to delete queue %d leaf node."
                        " errno = %d\n", queueInfo_p->queueId, ret);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
        goto adk_qos_hard_queue_delete_return; /* TODO: Add ADK failure*/
    } else {
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "Deleted queue with queue id %d\t", 
                        (queueInfo_p->queueId));
    }

    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_wred_profile_delete",
                    "Traffic manager WRED profile delete.\n");
    ret = rte_tm_wred_profile_delete(queueInfo_p->deviceId, 
                    queueInfo_p->nsId[queueInfo_p->wredId], &tmError);
    if (ret) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Error: %s\n", tmError.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "[%s()] rte_tm_wred_profile_delete() failed"
                            " for port %u, wred profile id %u with error %d\n",
                            __func__, queueInfo_p->deviceId,
                            queueInfo_p->nsId[queueInfo_p->wredId], ret);
            retValue += ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "WRED Profile deleted with id %u\n", 
                            queueInfo_p->nsId[queueInfo_p->wredId]);
    }
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_shaper_profile_delete",
                    "Traffic manager shaper profile delete.\n");
    ret = rte_tm_shaper_profile_delete(queueInfo_p->deviceId, 
                    queueInfo_p->nsId[queueInfo_p->shaperId],
                    &tmError);
    if(ret) {
            retValue += ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "\n  rte_tm_shaper_profile_delete() failed with error %s\n",
                            tmError.message);
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "Shaper deleted with id %u\n",
                            queueInfo_p->nsId[queueInfo_p->shaperId]);
            /*TODO: Add an error*/

    }

    goto adk_qos_hard_queue_delete_return;

adk_qos_hard_queue_delete_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_hard_queue_delete ----- */

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_soft_queue_create        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_soft_queue_create(
    adk_sys_handle_t           sysHandle,
    adk_qos_queue_params_t    *queueParams_p,
    adk_qos_queue_t           *queueInfo_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_qos_policing_params_t   policeParams;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

    queueInfo_p->queueId = queueInfo_p->nsId[0];    
    queueInfo_p->deviceId = queueParams_p->deviceId;
    queueInfo_p->cookie = ADK_QOS_QUEUE_COOKIE;
    queueInfo_p->dropPolicy = queueParams_p->bmPolicy;
    queueInfo_p->shapingMode = queueParams_p->shapingMode;  

    /*Fill in policing params */
    policeParams.cir =  queueParams_p->qosRates.peak.rate;
    policeParams.cbs = queueParams_p->qosRates.peak.bucketSize;
    policeParams.type = ADK_QOS_POLICE_TYPE_SRTCM_COLOR_BLIND_PPS;

    /*Create policer */
    retValue= adk_qos_policing_flow_create(sysHandle, &policeParams, 
                                           &(queueInfo_p->policeFlowHandle));
    if (retValue!=ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            " adk_qos_policing_flow_create failed");
        goto adk_qos_soft_queue_create_return;
    }

adk_qos_soft_queue_create_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_soft_queue_create ----- */

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_hard_queue_stats_counters_get        {{{2
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_hard_queue_stats_counters_get(
    adk_sys_handle_t        		sysHandle,
    adk_qos_queue_t         		*queueInfo_p,
    adk_qos_queue_stats_counters_t	*queueStats_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    int                         ret;
    struct rte_tm_error         tmError;
    struct rte_tm_node_stats 	nodeStats;
    uint64_t stats_mask;
    int clear = 0;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
	memset(queueStats_p,0x0,sizeof(*queueStats_p));	
    /* Create a WRED profile to be added to all queues */
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_node_stats_read",
                    "Traffic manager node statistics counters read.\n");
    ret = rte_tm_node_stats_read(queueInfo_p->deviceId, 
					queueInfo_p->queueId, &nodeStats,
					&stats_mask, clear, &tmError);
    if (ret) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
						"Error: %s\n", tmError.message);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
						"[%s()] rte_tm_node_stats_read() failed"
                " for port %u, queueId id %u with error %d\n",
                __func__, queueInfo_p->deviceId,
                queueInfo_p->queueId, ret);
		goto adk_qos_hard_queue_stats_counters_get_return;
    } else {
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
				"Stats received for Queue with id %u\n", 
				queueInfo_p->queueId); 
    }
	queueStats_p->q_tx_octets_count	= nodeStats.n_bytes;
	queueStats_p->q_tx_pdu_count  = nodeStats.n_pkts	;
	queueStats_p->q_dropOctetCount = nodeStats.leaf.n_bytes_dropped[RTE_TM_GREEN] +
			nodeStats.leaf.n_bytes_dropped[RTE_TM_YELLOW] + 
			nodeStats.leaf.n_bytes_dropped[RTE_TM_RED];
	queueStats_p->q_dropCount = nodeStats.leaf.n_pkts_dropped[RTE_TM_GREEN] +
	        nodeStats.leaf.n_pkts_dropped[RTE_TM_YELLOW] +
	        nodeStats.leaf.n_pkts_dropped[RTE_TM_RED];
	queueStats_p->q_outStandingPacketCount = nodeStats.leaf.n_pkts_queued;
	queueStats_p->q_rx_octets_count = queueStats_p->q_dropOctetCount +
			queueStats_p->q_tx_octets_count + 
			nodeStats.leaf.n_bytes_queued;
	queueStats_p->q_rx_pdu_count = queueStats_p->q_dropCount +
			queueStats_p->q_tx_pdu_count +
			queueStats_p->q_outStandingPacketCount;

	ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
adk_qos_hard_queue_stats_counters_get_return:
	ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * @internal
 * @fn adk_qos_soft_queue_stats_counter_get        {{{2
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_soft_queue_stats_counters_get(
    adk_sys_handle_t        		sysHandle,
    adk_qos_queue_t         		*queueInfo_p,
	adk_qos_queue_stats_counters_t	*queueStats_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    adk_sys_egress_stats_t      stats;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

	memset(queueStats_p,0x0,sizeof(*queueStats_p));	
	memset(queueStats_p,0x0,sizeof(&stats));	

    /*  read the egress stats*/
    retValue = adk_sys_egress_stats_get(sysHandle, 
                                        queueInfo_p->statsFlow.stats_id, &stats);

    if (retValue) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
						" adk_sys_egress_stats_get() returned Error\n");
		goto adk_qos_soft_queue_stats_counters_get_return;
    } else {
        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
				"Stats read for Queue[%d] with stats flow id %u\n", 
				queueInfo_p->queueId, queueInfo_p->statsFlow.stats_id); 
    }

	queueStats_p->q_tx_octets_count	= stats.byte_xmits;
	queueStats_p->q_tx_pdu_count    = stats.pkt_xmits;
	queueStats_p->q_dropOctetCount  = stats.byte_drops;
	queueStats_p->q_dropCount       = stats.pkt_drops;
	queueStats_p->q_rx_octets_count = stats.byte_drops + stats.byte_xmits;
	queueStats_p->q_rx_pdu_count    = stats.pkt_drops  + stats.pkt_xmits; 

adk_qos_soft_queue_stats_counters_get_return:
	ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
	ADK_RETURN(retValue);
}


/*!
 * =============================================================
 * @internal
 * @fn adk_qos_hard_queue_rate_modify        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_hard_queue_rate_modify(
    adk_sys_handle_t        sysHandle,
    adk_qos_queue_t         *queueInfo_p,
	const adk_qos_rates_t   *newQosrates_p)
{
    adk_status_t                retValue = ADK_SUCCESS;
    int                         ret;
	struct rte_tm_error         tmError;
	uint32_t prevShaperId,currShaperId;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);
	
	prevShaperId = queueInfo_p->shaperId;
	/* This will reset the bit(even for initialized value of 2)
	 */ 
	currShaperId = (queueInfo_p->shaperId & 1) ^ 1;
	/*ToDO: Fill WRED information from queue params into profile*/
	retValue = adk_qos_fill_shaper_params(queueInfo_p->shapingMode, 
							newQosrates_p, &queueInfo_p->shaperProfile);
    if (retValue != ADK_SUCCESS) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid Shaper Mode \n");
        goto adk_qos_hard_queue_rate_modify_return;
    }
    /* Create a WRED profile to be added to all queues */
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_shaper_profile_add",
                    "Traffic manager shaper profile add.\n");
    ret = rte_tm_shaper_profile_add(queueInfo_p->deviceId, 
					queueInfo_p->nsId[currShaperId], &queueInfo_p->shaperProfile, 
					&tmError);
    if (ret) {
        	retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_HW_PROGRAMMING_ERROR);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
							"Error: %s\n", tmError.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
							"[%s()] rte_tm_wred_profile_add() failed"
                    " for port %u, wred profile id %u with error %d\n",
                    __func__, queueInfo_p->deviceId,
                    queueInfo_p->nsId[currShaperId], ret);
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
					"Shaper Profile created with id %u\n", 
					queueInfo_p->nsId[currShaperId]); 
    }
	
    ADK_LOG_EXT_API(sysHandle,
                    ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "rte_tm_node_shaper_update",
                    "Traffic manager node private shaper update.\n");
	ret = rte_tm_node_shaper_update(queueInfo_p->deviceId,
					queueInfo_p->queueId, queueInfo_p->nsId[currShaperId],
					&tmError);
    if (ret) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
							"Error: %s\n", tmError.message);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
							"[%s()] rte_tm_node_shaper_update() failed"
                    " for port %u, Shaper profile id %u with error %d\n",
                    __func__, queueInfo_p->deviceId,
                    queueInfo_p->nsId[currShaperId], ret);
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_tm_wred_profile_delete",
                        "Traffic manager WRED profile delete.\n");
    	ret = rte_tm_wred_profile_delete(queueInfo_p->deviceId, 
						queueInfo_p->nsId[currShaperId], &tmError);
        goto adk_qos_hard_queue_rate_modify_return;
    } else {
            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
					"WRED Profile %u added to node %u\n", 
					queueInfo_p->nsId[currShaperId],queueInfo_p->queueId);
    }
	queueInfo_p->shaperId = currShaperId;
	if(prevShaperId < 2) {
    	/* Delete the previous  WRED profile*/
        ADK_LOG_EXT_API(sysHandle,
                        ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                        "rte_tm_shaper_profile_delete",
                        "Traffic manager shaper profile delete.\n");
    	ret = rte_tm_shaper_profile_delete(queueInfo_p->deviceId, 
						queueInfo_p->nsId[prevShaperId], &tmError);
    	if (ret) {
    	        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
						"Error: %s\n", tmError.message);
    	        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
						"[%s()] rte_tm_wred_profile_delete() failed"
    	                " for port %u, wred profile id %u with error %d\n",
    	                __func__, queueInfo_p->deviceId,
    	                queueInfo_p->nsId[prevShaperId], ret);
        goto adk_qos_hard_queue_rate_modify_return;
    	} else {
    	        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
						"WRED Profile deleted with id %u\n", 
						queueInfo_p->nsId[prevShaperId]);
    	}
	}
adk_qos_hard_queue_rate_modify_return:
	ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_soft_queue_delete        {{{2
 *
 * @param[in] *queueParams_p Pointer to an instance of 
 *            adk_qos_queue_params_t structure that defines the 
 *            queue to be created 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid scheduler parameter
 *
 * @details
 *
 * Allocates Queue ID 
 *
 * =============================================================
 */
adk_status_t
adk_qos_soft_queue_delete(
    adk_sys_handle_t           sysHandle,
    adk_qos_queue_t           *queue_p)
{
    adk_status_t                retValue = ADK_SUCCESS;

    ADK_QOS_INTERNAL_ENTRY_LOG_DEBUG(sysHandle);

    /*delete policer */
    retValue= adk_qos_policing_flow_delete(sysHandle, queue_p->policeFlowHandle);
    if (retValue!=ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            " adk_qos_policing_flow_delete failed");
        goto adk_qos_soft_queue_delete_return;
    }

adk_qos_soft_queue_delete_return:
    ADK_QOS_INTERNAL_EXIT_LOG_DEBUG(sysHandle);
    ADK_RETURN(retValue);
}    /* -----  end of function adk_qos_soft_queue_delete ----- */


#if 0
/*!
 * =============================================================
 * @Internal
 * @fn adk_qos_bm_params_fill 
 * 
 * @param[in] *queue_p          Pointer to queue parameters desired
 * @param[in,out] *wred_profile     pointer to profile information 
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE Invalid system handle
 * @retval ADK_E_MEM_UNAVAILABLE Invalid memory or memory not accessible
 *
 * @details
 *
 * This function adds a queue to the scheduling hierarchy
 * and returns the queue id. 
 *
 * =============================================================
 */
adk_status_t
adk_qos_bm_params_fill(adk_qos_queue_params_t *queue_p, 
                            struct rte_tm_wred_params *profile)
{
    
}
#endif
adk_status_t
adk_qos_fill_shaper_params(
    adk_qos_shaping_mode_t  shapingMode,
    const adk_qos_rates_t   *qosRates_p,
    struct rte_tm_shaper_params *profile_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    switch(shapingMode) {
        case ADK_QOS_NO_SHAPING:
            break;
        case ADK_QOS_PR:
        case ADK_QOS_PPS:
            profile_p->committed.rate = 0;
            profile_p->committed.size = 0;
            profile_p->peak.rate = qosRates_p->peak.rate;
            profile_p->peak.size = qosRates_p->peak.bucketSize;
            break;
        case ADK_QOS_PR_MR:
        case ADK_QOS_PR_SR:
            profile_p->committed.rate = qosRates_p->sr_mr_u.sustain.rate;
            profile_p->committed.size = qosRates_p->sr_mr_u.sustain.bucketSize;
            profile_p->peak.rate = qosRates_p->peak.rate;
            profile_p->peak.size = qosRates_p->peak.bucketSize;
            break;
        default: 
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_SHAPING_MODE); 
                return retValue;
            }
            break;
    }
    profile_p->pkt_length_adjust = RTE_TM_ETH_FRAMING_OVERHEAD_FCS;
    return retValue;
}

adk_status_t
adk_qos_fill_scheduler_params(adk_qos_queue_params_t *queue_p, 
                            uint32_t *priority_p, uint32_t *weight_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    if(queue_p->schedulingMode == ADK_QOS_SP ||
            queue_p->schedulingMode == ADK_QOS_WFQ) 
    {
        *weight_p = queue_p->serviceQuantum;
        *priority_p = queue_p->queuePriority ;
    } else {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_SCHEDULING_MODE);  
    }
    ADK_RETURN(retValue);
}

adk_status_t
adk_qos_get_dev_instance( uint16_t deviceId, uint16_t *instanceId_p)
{

    if(deviceId >= ADK_DEV_NET_ID_MAX)
        return ADK_E_INVALID_PARAMETER;

    *instanceId_p = deviceId;

#ifdef ADK_DEVICE_SNR
    if(deviceId <= ADK_DEV_NET_ID_RSVD_MAX) {
        struct rte_eth_dev_info dev_info;

        /* DPDK Eth-Devs */
        rte_eth_dev_info_get(deviceId, &dev_info);
        if (strstr(dev_info.device->name, "net_ice_tm")) {
            if(rte_pmd_ice_dsi_tm_get_parent_port(deviceId, instanceId_p) !=0)
            {
                return ADK_E_INVALID_PARAMETER;
            }
        }
    }
#endif
    return ADK_SUCCESS;
}

void
adk_qos_queue_param_print(
        adk_sys_handle_t        sysHandle,
        adk_qos_queue_params_t  *queue_p)
{

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                 "Queue Params\n");
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\tdeviceId=%u\n", queue_p->deviceId);
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\tqueueId=%u,\t parentNodeHandle=%"ADKPRIh"\n",
                    queue_p->queueId,queue_p->parentNodeHandle);
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\t shapingMode=%u,\t priority=%u\n",
                    queue_p->shapingMode, queue_p->schedulingMode);
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\t serviceQuantum=%u,\t bmPolicy=%u\n",
                    queue_p->serviceQuantum, queue_p->bmPolicy);
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\t sharedBMContextId=%u,\t pktLenCorrection=%u\n",
                    queue_p->sharedBMContextId, queue_p->pktLenCorrection);
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\t peak.rate=%lu,\t peak.bucketSize=%u\n",
            queue_p->qosRates.peak.rate, queue_p->qosRates.peak.bucketSize);
    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "\t sr_mr_u.sustain.rate=%lu,\t sr_mr_u.sustain.bucketSize=%u\n",
                    queue_p->qosRates.sr_mr_u.sustain.rate, 
                    queue_p->qosRates.sr_mr_u.sustain.bucketSize);
}

adk_status_t
adk_qos_queue_info_fill(
        adk_sys_handle_t        sysHandle,
        adk_qos_queue_params_t  *queue_p,
        adk_qos_queue_t     *queueInfo)
{
    adk_status_t retValue=ADK_SUCCESS;
    uint64_t     nodeId = (uint64_t) queue_p->parentNodeHandle;

    queueInfo->queueId = queueInfo->nsId[0];    
    queueInfo->deviceId = queue_p->deviceId;
    queueInfo->cookie = ADK_QOS_QUEUE_COOKIE;
    queueInfo->dropPolicy = queue_p->bmPolicy;
    queueInfo->parentNodeId = (uint32_t) nodeId;
    queueInfo->shapingMode = queue_p->shapingMode;  
    queueInfo->shaperId = 2;
    queueInfo->wredId   = 2;
    queueInfo->nodeParams.shaper_profile_id = RTE_TM_SHAPER_PROFILE_ID_NONE;
    queueInfo->nodeParams.shared_shaper_id  = NULL;
    queueInfo->nodeParams.n_shared_shapers  = 0;

    switch(queue_p->bmPolicy){
        case ADK_BM_RED:
            queueInfo->nodeParams.leaf.cman = RTE_TM_CMAN_WRED;
            break;
        case ADK_BM_TAIL_DROP:
           queueInfo->nodeParams.leaf.cman = RTE_TM_CMAN_TAIL_DROP;
           break;
        default:
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "un supported bm policy type \n");
            goto adk_qos_queue_info_fill_error;
            break;
    }
    queueInfo->nodeParams.leaf.wred.wred_profile_id = RTE_TM_WRED_PROFILE_ID_NONE;
    queueInfo->nodeParams.leaf.wred.shared_wred_context_id = NULL;
    queueInfo->nodeParams.leaf.wred.n_shared_wred_contexts= 0;
    queueInfo->nodeParams.stats_mask = ( RTE_TM_STATS_N_PKTS |
                    RTE_TM_STATS_N_BYTES |
                    RTE_TM_STATS_N_PKTS_GREEN_DROPPED |
                    RTE_TM_STATS_N_PKTS_YELLOW_DROPPED |
                    RTE_TM_STATS_N_PKTS_RED_DROPPED |
                    RTE_TM_STATS_N_BYTES_GREEN_DROPPED |
                    RTE_TM_STATS_N_BYTES_YELLOW_DROPPED |
                    RTE_TM_STATS_N_BYTES_RED_DROPPED |
                    RTE_TM_STATS_N_PKTS_QUEUED |
                    RTE_TM_STATS_N_BYTES_QUEUED);

    retValue = adk_qos_fill_shaper_params(queue_p->shapingMode, 
                    &queue_p->qosRates, &queueInfo->shaperProfile);
    if(retValue!= ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Error in filling shaping parameters\n");
        goto adk_qos_queue_info_fill_error;
    }
    queueInfo->shaperProfile.pkt_length_adjust = queue_p->pktLenCorrection;
    retValue = adk_qos_fill_scheduler_params(queue_p, 
                    &queueInfo->priority, &queueInfo->weight);
    if(retValue!= ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Error in filling scheduler parameters\n");
        goto adk_qos_queue_info_fill_error;
    }
    goto adk_qos_queue_info_fill_return;

adk_qos_queue_info_fill_error:
adk_qos_queue_info_fill_return:
    ADK_RETURN(retValue);
}
    
adk_status_t
adk_qos_bm_params_fill(
        adk_sys_handle_t    sysHandle,
        adk_qos_bm_policy_t *bm_policy_p, 
        adk_qos_queue_t     *queueInfo_p)
{
    adk_status_t retValue=ADK_SUCCESS;
    uint32_t minThreshold, maxThreshold;
    uint8_t queueSizeAvgFactor;

    if(bm_policy_p->dropPolicy!= queueInfo_p->dropPolicy)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
            "Drop policy doesn't match the policy provided while creating queue\n");
        goto adk_qos_bm_params_fill_return;
    }
    switch(bm_policy_p->dropPolicy) {
        case ADK_BM_TAIL_DROP:
            if(bm_policy_p->params_u.tailConfigNewParams.dropThreshold == 0) 
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Drop Threshold cannot be zero\n");
                goto adk_qos_bm_params_fill_return;
            }
            minThreshold = bm_policy_p->params_u.tailConfigNewParams.dropThreshold;
            maxThreshold = bm_policy_p->params_u.tailConfigNewParams.dropThreshold;
            queueSizeAvgFactor = 1;
            break;
        case ADK_BM_RED:
            if(bm_policy_p->params_u.redConfigNewParams.max_queue_length == 0) 
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Max Queue Length cannot be zero\n");
                goto adk_qos_bm_params_fill_return;
            }
            if(((bm_policy_p->params_u.redConfigNewParams.minThreshold <1) ||
                   (bm_policy_p->params_u.redConfigNewParams.minThreshold > 100)) ||
                    ((bm_policy_p->params_u.redConfigNewParams.maxThreshold < 1) ||
                    (bm_policy_p->params_u.redConfigNewParams.maxThreshold > 100)) ||
                    (bm_policy_p->params_u.redConfigNewParams.minThreshold >
                        bm_policy_p->params_u.redConfigNewParams.maxThreshold))
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Invalid Min/Max Threshold range\n");
                goto adk_qos_bm_params_fill_return;
            }
            minThreshold = (bm_policy_p->params_u.redConfigNewParams.minThreshold
                        * bm_policy_p->params_u.redConfigNewParams.max_queue_length ) / 100;
            maxThreshold = (bm_policy_p->params_u.redConfigNewParams.maxThreshold
                        * bm_policy_p->params_u.redConfigNewParams.max_queue_length ) / 100;
            queueSizeAvgFactor = 
                bm_policy_p->params_u.redConfigNewParams.queueSizeAvgFactor;
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Invalid Drop Policy\n");
            goto adk_qos_bm_params_fill_return;
    }
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].min_th =  minThreshold;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].max_th =  maxThreshold;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].maxp_inv = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_GREEN].wq_log2 = 
            bm_policy_p->params_u.redConfigNewParams.queueSizeAvgFactor;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].min_th = minThreshold;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].max_th = maxThreshold;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].maxp_inv = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_YELLOW].wq_log2 = 
            bm_policy_p->params_u.redConfigNewParams.queueSizeAvgFactor;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].min_th = minThreshold;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].max_th = maxThreshold;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].maxp_inv = 1;
    queueInfo_p->wredProfile.red_params[RTE_TM_RED].wq_log2 = 
            bm_policy_p->params_u.redConfigNewParams.queueSizeAvgFactor;


adk_qos_bm_params_fill_return:
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @internal
 * @fn adk_qos_queue_info_get        {{{2
 *
 * @param[in]   sysHandle      ADk System handle 
 * @param[in]   queueHandle    queueHandle to fetch information from
 * @param[out]  *qId_p         pointer to Queue IDs structure.
 *
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_PARAMETER Invalid parameters
 * @retval ADK_E_QOS_QUEUE_INVALID Invalid queue
 * @details
 *
 * Returns queue IDs information from queueHandle
 *
 * =============================================================
 */
adk_status_t
adk_qos_queue_info_get(
    adk_sys_handle_t         sysHandle,
    adk_handle_t             queueHandle,
    adk_qos_queue_id_info_t *qId_p)
{
    adk_status_t retValue=ADK_SUCCESS;
    adk_qos_queue_t *queueInfo;

    /* Validate queue handle pointer */
    if (ADK_INVALID_PTR(queueHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_qos_queue_id_get_return;
    }
    if (ADK_INVALID_PTR(qId_p)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_qos_queue_id_get_return;
    }
    queueInfo = (adk_qos_queue_t *) queueHandle;
    if (queueInfo->cookie != ADK_QOS_QUEUE_COOKIE) {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_QOS_QUEUE_INVALID);
        goto adk_qos_queue_id_get_return;
    }

    qId_p->queueId  = queueInfo->queueId;
    qId_p->devId = queueInfo->parentEthDevId;
    qId_p->queueStatsId = queueInfo->statsFlow.stats_info;
    
    if (qId_p->devId > ADK_DEV_NET_ID_RSVD_MAX) {
        adk_qos_policing_flow_id_get(sysHandle, queueInfo->policeFlowHandle, 
                                     &(qId_p->policerId));
    } else {
        qId_p->policerId = UINT_MAX;
    }
adk_qos_queue_id_get_return:
    return retValue;
}
#else
adk_status_t
adk_qos_queue_info_get(
    adk_sys_handle_t         sysHandle,
    adk_handle_t             queueHandle,
    adk_qos_queue_id_info_t *qId_p)
{
    adk_status_t retValue=ADK_SUCCESS;
    RTE_SET_USED(sysHandle);
    RTE_SET_USED(queueHandle);

    qId_p->queueId  = 0;
    qId_p->devId = 0;
    qId_p->queueStatsId = 0;
    qId_p->policerId = UINT_MAX;
    return retValue;
}
#endif
