/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_qos_dp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _qos  QoS Module
 * @{
 *     @addtogroup _qosDpFunctions  QoS Dataplane Functions
 *     @{
 *
 * This file contains ADK QoS Dataplane processing procedures
 *
 *
 *  **********************************************************************/

#include "adk_dpdk.h"

#include <sched.h> 
#include "adk_qos_include.h"
#include "adk_qos_handles.h"
#include "adk_list.h"
#include "adk_util_list.h"
#include "adk_queue.h"
#include "adk_qos_private.h"
#include "adk_pktio.h"
#include "adk_qos_dp.h"
#include "adk_qos_dp_export.h"
#include "adk_util_list.h"
#include <adk_pmu.h>
#include <rte_cycles.h>
#include "adk_shm.h"

static adk_qos_police_flow_handle_t *dpPolTable_p=NULL;
static adk_qos_queue_id_t *defaultQueueId_p = NULL;
static adk_qos_queue_id_t *ptpQueueId_p = NULL;
static adk_qos_phb_dp_stats_t *phbStatsGlobal_p= NULL;
static adk_qos_sl_queue_dp_stats_t *slQueueStatsGlobal_p= NULL;
static uint32_t *queueStatsIdTbl_p = NULL;
static uint32_t *slQueueStatsIdTbl_p = NULL;

adk_status_t
adk_qos_dp_init(void)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_qos_shm_config_t *shm_config_p = NULL;

    /* Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_lookup(ADK_QOS_SHM_MEMZONE_NAME, (adk_shm_handle_t **)&shm_config_p);    

    if (retValue != ADK_SUCCESS) {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Error: Failed to do adk_qos_dp_init()!\n");
        return (ADK_E_MEM_UNAVAILABLE);
    }

    if (shm_config_p) {
        dpPolTable_p = shm_config_p->dpPolTable_p;
        defaultQueueId_p = &shm_config_p->defaultQueueId;
        ptpQueueId_p = &shm_config_p->ptpQueueId;
        phbStatsGlobal_p = shm_config_p->phbStats_p;
        slQueueStatsGlobal_p = shm_config_p->slQueueStats_p;
        queueStatsIdTbl_p = &shm_config_p->queueStatsIdTbl[0];
        slQueueStatsIdTbl_p = &shm_config_p->slQueueStatsIdTbl[0];
    }
    return retValue;
}


adk_status_t 
adk_qos_dp_trtcm_rate_shaper(
                adk_qos_ts_param_t *tsParams,
                uint32_t packetSize,
                bool *isReady)
                
{
#if 0
    struct timespec curTime;
#endif
    uint32_t deltaTime, currentTime;
    uint32_t color;


    /* The value of the incoming packet converted to the # of ticks for cir. */
    uint32_t cirBucketpacketSizeInTick;
    /* The value of the incoming packet converted to the # of ticks for pir. */
    uint32_t pirBucketpacketSizeInTick;

    *isReady = 0;
    if(!tsParams)
    {
        debug_print("Invalid Policing Param pointer. Dropping packet!\n");
	return ADK_E_INTERNAL_ERROR;
    }
    
    /*Get current time*/
    currentTime = adk_get_cycle_cntr();
    currentTime = currentTime <<10;
#if 0
    clock_gettime( CLOCK_REALTIME, &curTime);
    currentTime = (uint32_t) curTime.tv_nsec;
#endif

    /* Store the time elapse / delta time from the previous incoming
     * packet to the current packet */
    if (currentTime > tsParams->prev_time)
    {
        deltaTime = currentTime - tsParams->prev_time;
    }
    else
    {
        deltaTime = 0x3B9ACA00 - tsParams->prev_time + currentTime;
    }
    deltaTime = deltaTime * 1;

    tsParams->prev_time = currentTime;

    /*
     * Convert the incoming packet size in unit of tick.
     * Multiply packet size with pir and cir.
     */
    cirBucketpacketSizeInTick   = packetSize * tsParams->cir_ticks;
    pirBucketpacketSizeInTick   = packetSize * tsParams->pir_ticks;

    /*
     * Calculate the number of the token inside cir bucket
     * The value ranges from 0 to "committedBurstSize"
     * When the value is greater than committedBurstSize than set the value
     * to committedBurstSize
     */
    tsParams->cir_bucket_token += deltaTime;
    if (tsParams->cir_bucket_token > tsParams->cbs_ticks)
    {
        tsParams->cir_bucket_token = tsParams->cbs_ticks;
    }

    /*
     * Calculate the number of the token inside pir bucket
     * The value ranges from 0 to "peakBurstSize"
     * When the value is greater than peakBurstSize than set the value
     * to peakBurstSize
     */
    tsParams->pir_bucket_token += deltaTime;
    if (tsParams->pir_bucket_token > tsParams->pbs_ticks)
    {
        tsParams->pir_bucket_token = tsParams->pbs_ticks;
    }

    /*
     * Compare the incoming packet size with the number of pir bucket tokens
     * Which is, test if there is enough token for the incoming packet
     */

    /*
     * Then compare the incoming packet size with the number of cir bucket
     * tokens which is, test if there is enough token for the incoming
     * packet
     * Note: If the packet fails pir bucket it will be fail cir bucket
     * as well.
     */
    if (tsParams->pir_bucket_token < pirBucketpacketSizeInTick)
    {
        color =  ADK_QOS_DP_RED;
    }
    else if ( tsParams->cir_bucket_token <  cirBucketpacketSizeInTick)
    {
        color = ADK_QOS_DP_YELLOW;
        tsParams->pir_bucket_token -= pirBucketpacketSizeInTick;
    }
    else
    {
        color = ADK_QOS_DP_GREEN;
        tsParams->cir_bucket_token -= cirBucketpacketSizeInTick;
        tsParams->pir_bucket_token -= pirBucketpacketSizeInTick;
    }

#ifdef DEBUG
    if (*color == ADK_QOS_DP_RED)
    {
        /* Count RED packets. */
        tsParams->red_pkts++;
    } else if (*color == ADK_QOS_DP_YELLOW)
    {
        /* Count Yellow packets. */
        tsParams->yellow_pkts++;
    } else
    {
        /* Count Green packets. */
        tsParams->green_pkts++;
    }
#endif

    if(color != ADK_QOS_DP_RED) *isReady = 1;
    
    return ADK_SUCCESS;
}


adk_dp_status_t adk_qos_dp_policing_trtcm_color_blind(
                uint32_t policing_flow_id,
                uint32_t packetSize, 
                uint8_t * policeColor)
{
    adk_qos_police_flow_handle_t *police_flow_info_p;
    struct rte_meter_trtcm  *dpTrtcmParams;
    enum rte_meter_color pkt_color;
    uint64_t currentTime;

    if((!policeColor) || (!dpPolTable_p) || (policing_flow_id==UINT_MAX))
    {
        debug_print("NULL flow handle/color pointer. returnign error!\n");
        return ADK_DP_E_ERROR;
    }
    
    police_flow_info_p = &dpPolTable_p[policing_flow_id];
    dpTrtcmParams = &(police_flow_info_p->dpTrtcmParams);
    
    /*Get current time*/
    currentTime =rte_get_tsc_cycles();

    pkt_color = rte_meter_trtcm_color_blind_check(dpTrtcmParams, 
#if ADK_DPDK_VERSION >= 18050016
                            &(police_flow_info_p->dpTrTcmProfile),
#endif
                                                  currentTime,
                                                  packetSize);
    *policeColor = pkt_color;
    if (*policeColor == ADK_QOS_DP_RED)
    {
        /* Count RED packets. */
        police_flow_info_p->red_pkts++;
    } else if (*policeColor == ADK_QOS_DP_YELLOW)
    {
        /* Count Yellow packets. */
        police_flow_info_p->yellow_pkts++;
    } else
    {
        /* Count Green packets. */
        police_flow_info_p->green_pkts++;
    }
    
    return ADK_SUCCESS;
}

adk_dp_status_t adk_qos_dp_policing_srtcm_color_blind(
                uint32_t policing_flow_id,
                uint32_t packetSize, 
                uint8_t * policeColor)
{
    adk_qos_police_flow_handle_t *police_flow_info_p;
    struct rte_meter_srtcm  *dpSrtcmParams;
    enum rte_meter_color pkt_color;
    uint64_t currentTime;
    
    police_flow_info_p = &dpPolTable_p[policing_flow_id];
    dpSrtcmParams  = &(police_flow_info_p->dpSrtcmParams);
    
    /*Get current time*/
    currentTime = rte_get_tsc_cycles();

    pkt_color = rte_meter_srtcm_color_blind_check(dpSrtcmParams, 
#if ADK_DPDK_VERSION >= 18050016
                            &(police_flow_info_p->dpSrTcmProfile),
#endif
                                                  currentTime,
                                                  packetSize);
    *policeColor = pkt_color;
    if (*policeColor == ADK_QOS_DP_RED)
    {
        /* Count RED packets. */
        police_flow_info_p->red_pkts++;
    } else if (*policeColor == ADK_QOS_DP_YELLOW)
    {
        /* Count Yellow packets. */
        police_flow_info_p->yellow_pkts++;
    } else
    {
        /* Count Green packets. */
        police_flow_info_p->green_pkts++;
    }
    
    return ADK_SUCCESS;
}

adk_dp_status_t adk_qos_dp_policing_srtcm_color_blind_pps(
                uint32_t    policing_flow_id,
                uint8_t * policeColor)
{
    adk_qos_police_flow_handle_t *police_flow_info_p;
    struct rte_meter_srtcm  *dpSrtcmParams;
    enum rte_meter_color pkt_color;
    uint64_t currentTime;

    police_flow_info_p = &dpPolTable_p[policing_flow_id];
    dpSrtcmParams = &(police_flow_info_p->dpSrtcmParams);
    
    /*Get current time*/
    currentTime = rte_get_tsc_cycles();

    pkt_color = rte_meter_srtcm_color_blind_check(dpSrtcmParams, 
#if ADK_DPDK_VERSION >= 18050016
                            &(police_flow_info_p->dpSrTcmProfile),
#endif
                                                  currentTime,
                                                  1);
    *policeColor = pkt_color;
    if (*policeColor == ADK_QOS_DP_RED) {
        /* Count RED packets. */
        police_flow_info_p->red_pkts++;
    } else if (*policeColor == ADK_QOS_DP_YELLOW) {
        /* Count Yellow packets. */
        police_flow_info_p->yellow_pkts++;
    } else {
        /* Count Green packets. */
        police_flow_info_p->green_pkts++;
    }
    
    return ADK_SUCCESS;
}

adk_dp_status_t adk_qos_dp_policing_trtcm_color_blind_pps(
                uint32_t    policing_flow_id,
                uint8_t * policeColor)
{
    adk_qos_police_flow_handle_t *police_flow_info_p;
    struct rte_meter_trtcm  *dpTrtcmParams;
    enum rte_meter_color pkt_color;
    uint64_t currentTime;

    police_flow_info_p = &dpPolTable_p[policing_flow_id];
    dpTrtcmParams = &(police_flow_info_p->dpTrtcmParams);
    
    /*Get current time*/
    currentTime = rte_get_tsc_cycles();

    pkt_color = rte_meter_trtcm_color_blind_check(dpTrtcmParams, 
#if ADK_DPDK_VERSION >= 18050016
                            &(police_flow_info_p->dpTrTcmProfile),
#endif
                                                  currentTime,
                                                  1);
    *policeColor = pkt_color;
    if (*policeColor == ADK_QOS_DP_RED)
    {
        /* Count RED packets. */
        police_flow_info_p->red_pkts++;
    } else if (*policeColor == ADK_QOS_DP_YELLOW)
    {
        /* Count Yellow packets. */
        police_flow_info_p->yellow_pkts++;
    } else
    {
        /* Count Green packets. */
        police_flow_info_p->green_pkts++;
    }
    
    return ADK_SUCCESS;
}

adk_dp_status_t adk_qos_dp_policing_color_blind(
                uint32_t policing_flow_id,
                uint32_t __attribute__((unused)) packetSize, 
                uint8_t *policeColor)
{
    adk_qos_police_flow_handle_t *police_flow_info_p;
    if((!policeColor) || (!dpPolTable_p) || (policing_flow_id==UINT_MAX))
    {
        debug_print("NULL flow handle/color pointer. returnign error!\n");
        return ADK_DP_E_ERROR;
    }
    police_flow_info_p = &dpPolTable_p[policing_flow_id];
    switch(police_flow_info_p->type) {
        case ADK_QOS_POLICE_TYPE_SRTCM:
             return adk_qos_dp_policing_srtcm_color_blind_pps(policing_flow_id, policeColor);
        default:
             /*TODO: Handle case where policer configuration is invalid*/
             return ADK_DP_E_ERROR;
    }
}

inline adk_qos_queue_id_t adk_qos_dp_get_default_tx_queue(void)
{
    return *defaultQueueId_p;
}

inline adk_qos_queue_id_t adk_qos_dp_get_ptp_queue(void)
{
    return *ptpQueueId_p;
}

inline void adk_qos_dp_increment_phb_counters(uint8_t  ipVersion,
                                       uint16_t statsId,
                                       uint32_t ipLength,
                                       bool     pktDropped)
{
    adk_qos_phb_dp_stats_t *queuePhbStats_p = &(phbStatsGlobal_p[statsId]);
    /*Upper layers of this function has already sanitized v4/v6 version in the packet*/
    if (ADK_IPV4_VER == ipVersion) {
        if (!pktDropped) {
            queuePhbStats_p->ipv4TxPkts++;
            queuePhbStats_p->ipv4TxBytes += ipLength;
        }else{
            queuePhbStats_p->ipv4TxPktsDropped++;
            queuePhbStats_p->ipv4TxBytesDropped += ipLength;
        }
    } else {
        if (!pktDropped) {
            queuePhbStats_p->ipv6TxPkts++;
            queuePhbStats_p->ipv6TxBytes += ipLength;
        }else{
            queuePhbStats_p->ipv6TxPktsDropped++;
            queuePhbStats_p->ipv6TxBytesDropped += ipLength;
        }
    }
}

inline void adk_qos_dp_increment_sl_queue_counters(bool     isIPPacket,
                                            uint8_t  ipVersion,
                                            uint16_t statsId,
                                            uint32_t pktLength,
                                            bool     pktDropped)
{
    adk_qos_sl_queue_dp_stats_t *queueSlStats_p = &(slQueueStatsGlobal_p[statsId]);
    if (isIPPacket) {
        if (ADK_IPV4_VER == ipVersion) {
            if (!pktDropped) {
                queueSlStats_p->ipv4TxPkts++;
                queueSlStats_p->ipv4TxBytes += pktLength;
            }else{
                queueSlStats_p->ipv4TxPktsDropped++;
                queueSlStats_p->ipv4TxBytesDropped += pktLength;
            }
        } else {
            if (!pktDropped) {
                queueSlStats_p->ipv6TxPkts++;
                queueSlStats_p->ipv6TxBytes += pktLength;
            }else{
                queueSlStats_p->ipv6TxPktsDropped++;
                queueSlStats_p->ipv6TxBytesDropped += pktLength;
            }
        }
    } else {
        if (!pktDropped) {
            queueSlStats_p->nonIPTxPkts++;
            queueSlStats_p->nonIPTxBytes += pktLength;
        }else{
            queueSlStats_p->nonIPTxPktsDropped++;
            queueSlStats_p->nonIPTxBytesDropped += pktLength;
        }
    }
}

inline uint32_t adk_qos_dp_get_queue_stats_id(uint16_t txQueueId, bool *isPHBQueue)
{
    if( ADK_RM_RSRC_INVALID != slQueueStatsIdTbl_p[txQueueId]) {
        *isPHBQueue = false;
        return slQueueStatsIdTbl_p[txQueueId];
    }
    if( ADK_RM_RSRC_INVALID != queueStatsIdTbl_p[txQueueId]) {
        *isPHBQueue = true;
        return queueStatsIdTbl_p[txQueueId];
    }
    return ADK_RM_RSRC_INVALID;
}

/*!
 * }}}2
 *     @} end addtogroup _qosDpFunctions
 * @} end addtogroup _qos
 *
 * vim: set foldmethod=marker foldlevel=1 foldcolumn=3 columns=83:
 */
