/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

/*!
 * @file  adk_dev_net.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _devNet  ADK Device Net
 * @{
 *
 * @brief
 *
 * This file contains definitions and declarations of ADK Device Net
 * interface (APIs and data structure definitions).
 *
 ***********************************************************************/

#ifndef __ADK_DEV_NET_H__
#define __ADK_DEV_NET_H__

#ifdef __cplusplus
extern "C" {
#endif

/*!
 * =============================================================
 * 
 *  CPPI Bridge Driver metadata offset in mbuf private region
 *
 * =============================================================
 */
#define ADK_CPPI_BD_MD_OFFSET   0x00


/*!
 * =============================================================
 *
 * ADK Device ID
 *
 * =============================================================
 */
typedef uint16_t    adk_dev_id_t;

/*!
 * ========================================================================
 * @enum adk_dev_net_type_t
 *
 *  @brief  Data structure to to define different types of ADK Device Nets.
 *
 * =======================================================================
 */
typedef enum adk_dev_net_type_e {
    /*!
     * type to represent a ADK divice net that is to send and receive packets 
     * from user-space ADK-DATAPLANE to ADK NetDriver(Linux Kernel) via
     * CPPI-Bridge driver.
     */
    ADK_DEV_NET_TYPE_NETD,

    /*!
     * type to represent a ADK device net that is to send and receive packets 
     * from ADK-DATAPLANE to a userplane appliciaton and vise versa.
     * This type is not supported in the current release.
     */
    ADK_DEV_NET_TYPE_PH,

     /*!
     * type to represent a DPDK eth-dev(veth) that is to send and receive packets 
     * from ADK-DATAPLANE to DPDK veth via ethernet Bridger driver. 
     * This type is not supported in the current release.
     */
    ADK_DEV_NET_TYPE_DPDK_ETH_DEV_VETH,

     /*!
     * type to represent a DPDK eth-dev(tund) that is to send and receive packets 
     * from ADK-DATAPLANE to DPDK TUND via ethernet bridge driver.
     * This type is not supported in the current release.
     */   
    ADK_DEV_NET_TYPE_DPDK_ETH_DEV_TUND,

    /*!
     * type to represent a DPDK eth-dev that is to send and receive packets 
     * from ADK-DATAPLANE to regular DPDK Eth-DEV (External network) via
     * ethernet bridge driver.
     * This type is not supported in the current release.
     */
    ADK_DEV_NET_TYPE_DPDK_ETH_DEV,
    
} adk_dev_net_type_t;


/*!
 * ========================================================================
 * @enum adk_dev_net_dm_type_t
 *
 * @brief Data type to define different data mover types
 *
 * Data mover is required
 * to move packets from ADK-DATAPLANE to another applicaiton domain.
 * ========================================================================
 */
typedef enum adk_dev_net_dm_type_e {
    /*! The type to represent soft data mover.  */
    ADK_DEV_NET_DM_TYPE_SOFT,
    
} adk_dev_net_dm_type_t;


/*!
 * ========================================================================
 * @struct adk_dev_net_event_dev_bd_info_t
 *
 * @brief data structure to represent a event BD 
 *
 * ========================================================================
 */
typedef struct adk_dev_net_event_bd_info_s {
    /*! Bridge driver instance id.
     * The type of bridge driver depends on type of device net
     * (adk_dev_net_type_t)
     * */
    uint8_t     bd_instance_id;

    /*! event-queue on which bridge driver receive events from workers.  */
    uint8_t     bd_event_queue_id;
} adk_dev_net_event_bd_info_t;

/*!
 * ========================================================================
 * @enum adk_dev_net_mgmt_type_t
 *
 * @brief data structre to define the different management types.
 *
 * ========================================================================
 */
 typedef enum adk_dev_net_mgmt_type_e{
    /*! device is fully managed by ADK. 
     * 1. Device ID allocation. 
     * 2. Device configurations as part of adk_dev_net_configure() call.
     * 3. Device adding to corresponding bridge driver*/
    ADK_DEV_NET_MGMT_TYPE_ADK_FULL,
 } adk_dev_net_mgmt_type_t;

/*!
 * ========================================================================
 * @struct adk_dev_net_info_t
 * 
 * @brief data structure to define ADK device net parameters.  
 *
 * The ADK device net is created based on the appliciaton given parametrs. 
 *
 * ADK netdevs (object with attributes) / ADK netdev id   =  TN netdevs : 
 *     physical port netdevs, logical netdevs- e.g. bridges, tunnels,
 *     LAG ports etc.
 *
 *  ADK device_net / ADK device_net_id = device that is used to 
 *    transport network traffic associated with 1 or more netdevs -
 *    hardware devices used for network interfaces or shared memory 
 *
 * ========================================================================
 */
 
typedef struct adk_dev_net_info_s {    
    /*! ADK device net type.  */
    adk_dev_net_type_t     dev_type;
    
    /*!  data  mover details.  */
    struct {
        /*! The backend data mover type.  */
        adk_dev_net_dm_type_t dm_type;

        /*! Extensibility field - packet memory pool required to communicate 
        * with the peer. needed only if different from the data plane's memory pool */
        void               *pkt_mem_pool;
    };
    
    /* Information on the Eventdev bridge that converts to  ADK fast path
     * (Evdev workload) for this specific interface*/
    adk_dev_net_event_bd_info_t bd_info;

    /*! Number of netdev that are carried on this ADK device net (transport device).
     * 0 means ALL netdevs are carried. the current support is only for "0". 
     * extensible for veth/tund (1:1) etc.
     */
    uint16_t    n_member_netdevs;

    /*!  List of netdevs traffic that is trasported to ADK FastPath.
     *  Applicaiton need to give the list if the n_member_netdevs!=0*/
    uint16_t    *member_netdev_id_list;
    
    /*! type of indicate ADK managed resource vs ADK "used" resource that was
     *  pre-created by the application. current support is only for ADK_MANAGED.
     *  Extensibility field for future needs. 
     */
    adk_dev_net_mgmt_type_t  mgmt_type;
     
    /*!
     * if the device net is not full managed by ADK, then application need to
     * pass the respective DPDK-ETH-DEV-ID.
     * ID range [0 to RTE_MAX_ETHPORTS-1] is used by DPDK for DPDK-eth-dev-ids.
     * ID range [RTE_MAX_ETHPORTS to ADK_NW_DEVICE_ID_MAX] is managed by ADK. 
     * this filed is ignored if the mgmt_type= ADK_DEV_NET_MGMT_TYPE_ADK_FULL.
     */
    uint16_t  dev_id;

} adk_dev_net_info_t;

/*!
 * ========================================================================
 * @struct adk_dev_type_t
 *
 * @brief data structre to define the different ADK device types.
 *
 * ========================================================================
 */
 typedef enum adk_dev_type_e{
    /*!  Deive type Network   */
    ADK_DEVICE_TYPE_NET,
 } adk_dev_type_t;

/*!
 * ========================================================================
 * @struct adk_dev_t
 *
 * @brief data type to define ADK a device
 *
 * ========================================================================
 */
typedef struct adk_dev_s
{
    /*! type of the ADK device */ 
    adk_dev_type_t dev_type;

    /*! device id
     * ID range [0 to RTE_MAX_ETHPORTS-1] is used by DPDK for DPDK-eth-dev-ids.
     * ID range [RTE_MAX_ETHPORTS to ADK_NW_DEVICE_ID_MAX] is managed by ADK. 
     */
    uint16_t       dev_id;
} adk_dev_t;

/*!
 * ========================================================================
 * @struct adk_dev_net_sched_mode_t
 *
 * @brief Define the scheduling modes for rx/tx queues in a ADK Device net.
 *
 * ========================================================================
 */
 typedef enum adk_dev_net_sched_mode_e{
    /*!
     * Round robin scheduling mode; 
     */
    ADK_DEV_NET_SCHED_MODE_RR,
 } adk_dev_net_sched_mode_t;

/*!
 * ========================================================================
 * @struct adk_dev_net_config_params_t 
 * _
 * @brief Data structure to define configuration parameters of ADK Device Net.
 *
 * ========================================================================
 */
typedef struct adk_dev_net_config_params_s {
    /*! number of receive queues required in a ADK Device Net. 
     *  the number should be in range[1, 8]. A receive queue is to 
     *  receive packets from Linxu Kernel(ADK NetDriver) to ADK dataplane.
     *  for the device type NETD, the number of receive queues must be 8.
     */
    uint8_t                    num_rx_queues;
    
    /*! number of transmit queues required in a ADK Device Net. 
     *  the number should be in range[1, 8]. A transmit queue is to send 
     *  packets from ADK dataplane to Linxu Kernel (ADK NetDriver)
     */
    uint8_t                    num_tx_queues;
    
    /*! scheduling mode to define how the receive queues are serviced by ADK
     * dataplane 
     */
    adk_dev_net_sched_mode_t rx_sched_mode;
     
    /*! scheduling mode to define how the transmit queues are serviced by Linux
     * kernel (ADK NetDriver)  
     */
     adk_dev_net_sched_mode_t tx_sched_mode;
}adk_dev_net_config_params_t;

/*!
 *   @addtogroup _devNetFunctions  ADK Device Net Functions
 *   @{
 */

/*!
 * adk_dev_net_create()
 * @param[in] sysHandle    ADK System Handle
 * @param[in] info_p       pointer to adk_dev_net_info_t that represents information. 
 * @param[out] dev_id_P    pointer to store ADK allocated ADK-Device-ID. 
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to create a ADK Device Net.
 *
 * It returns created device id. 
 * Applicaiton is required to use this device to send/receive exception packets
 * to/from Linux Kernel(ADK NetDriver) from ADK dataplane.
 */
adk_status_t 
adk_dev_net_create(adk_sys_handle_t sysHandle, 
                     adk_dev_net_info_t *info_p,
                     adk_dev_id_t    *dev_id_p);
                     
/*!
 * adk_dev_net_configure()
 * @param[in]  sysHandle    ADK System Handle 
 * @param[in]  dev_id       ADK Device ID
 * @param[in]  config_p     pointer to adk_dev_net_config_params_t 
 *                          that define configuration parameters
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to configure previously created ADK Device Net. 
 *
 */
adk_status_t 
adk_dev_net_configure(adk_sys_handle_t sysHandle, 
                        adk_dev_id_t    dev_id,
                        adk_dev_net_config_params_t *config_p);

/*!
 * adk_dev_net_tx_queue_setup()
 * @param[in]  sysHandle    ADK System Handle 
 * @param[in]  dev_id       ADK Device ID
 * @param[in]  tx_queue_id  Transmit Queue ID (From dataplane to Linux Kernel
 *                          direction)
 * @param[in]  weight       weight of the queue used in case of device is 
 *                          configured for RR scheduling; priority in case of
 *                          SP scheduling mode
 * @param[in]  size         queue size in number of packets;
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to setup the queue size and scheduling weight of a transmit-queue in a Device Net.
 */
 adk_status_t 
adk_dev_net_tx_queue_setup(adk_sys_handle_t      sysHandle, 
                             adk_dev_id_t    dev_id,
                             uint16_t              tx_queue_id,
                             uint8_t               weight,
                             uint16_t              size);
/*!
 * adk_dev_net_rx_queue_setup()
 *
 * @param[in]  sysHandle    ADK System Handle 
 * @param[in]  dev_id       ADK Device ID
 * @param[in]  rx_queue_id  receive queue id (Receive from Linux Kernel to 
 *                          ADK dataplane direction)
 * @param[in]  weight       weight of the queue used in case of device is 
 *                          configured for RR scheduling; priority in case
 *                          of SP scheduling mode
 * @param[in]  size         queue size in number of packets;
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to setup the queue size and scheduling weight of a receive-queue in a Device net.
 */
 adk_status_t                           
adk_dev_net_rx_queue_setup(adk_sys_handle_t      sysHandle, 
                             adk_dev_id_t        dev_id,
                             uint16_t            rx_queue_id,
                             uint8_t             weight,
                             uint16_t            size);
/*!
 * adk_dev_net_start()
 * @param[in]  sysHandle    ADK System Handle 
 * @param[in]  dev_id       ADK Device ID
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 * ADK_E_NETD_FD_OPEN_FAILED    ADK NetDriver is not ready. 
 *
 * @brief API to start the given Device net. 
 */
 adk_status_t                                                        
adk_dev_net_start(adk_sys_handle_t      sysHandle, 
                    adk_dev_id_t        dev_id);
/*!
 * adk_dev_net_stop()
 * @param[in]  sysHandle    ADK System Handle 
 * @param[in]  dev_id       ADK Device ID
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to stop the given Device net 
 */
 adk_status_t                                               
adk_dev_net_stop(adk_sys_handle_t      sysHandle, 
                   adk_dev_id_t        dev_id);

/*!
 * adk_dev_net_delete()
 * @param[in] sys_handle    ADK System Handle
 * @param[in] dev_id        ADK Device-ID that is created previously. 
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to delete a ADK Device Net
 *
 * Applicaiton is required to
 * stop this device before deleting it.
 */
adk_status_t 
adk_dev_net_delete(adk_sys_handle_t sys_handle, 
                     adk_dev_id_t   dev_id);

/*!
 * adk_dev_net_stats_get()
 * @param[in] sys_handle    ADK System Handle
 * @param[in] dev_id        ADK Device-ID that is created previously. 
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to stats for a ADK Device Net
 *
 */
adk_status_t                           
adk_dev_net_stats_get(adk_sys_handle_t      sys_handle, 
                             adk_dev_id_t   dev_id);



#ifdef __cplusplus
}
#endif

#endif /* __ADK_DEV_NET_H__ */

/*!     end addtogroup _devNetFunctions
 *     @}
 * end addtogroup _devNet
 * @}
 */
