/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_error.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief
 *
 *  This file contains the error codes defined by ADK SIF. This file should not
 *  be directly in the source C file. Instead the adk_sys.h header file should
 *  be included which includes this header file as well.
 *
 *  **********************************************************************/
#ifndef __ADK_ERROR_H__
#define __ADK_ERROR_H__

#include <stdio.h>

#include <adk_module_id.h>

/*!
 * adk_status_t - a 32-bit unsigned integer.
 *
 * All ADK APIs return this value to the user application
 *
 * The structure of the adk_status_t is below:
 *
 * first 8-bits (MSB) - module ID
 * next 4-bits        - reserved (set to zero for now)
 * next 4-bits        - error severity
 * last 16-bits (LSB) - module specific error codes
 *
 * -----------------------------------------------------------------
 *      8-bit          | 4-bit   | 4-bit    |       16-bit         |
 *     module ID       | RSVD    | Severity |    error code        |
 * -----------------------------------------------------------------
 *
 *  If the caller is planning to compare the return value with the
 *  values defined in one of the error code enums (e.g. adk_sif_status_t,
 *  adk_ethernet_status_t etc), they must mask the severity bits by calling
 *  ADK_STATUS_SEV_LEVEL_MASK macro before performing the comparison.
 *
 **********************************************************************/
typedef uint32_t adk_status_t;

typedef uint32_t adk_dp_status_t;

/*!
 * =============================================================
 * @enum adk_error_sev_level_t
 *
 * @brief
 *
 * This enum defines the error severity levels.
 *
 * =============================================================
 */
typedef enum adk_error_sev_level_e
{
    /* Add ADK_FROM _DP here so it is in the same number space */
    ADK_FROM_DP = 0x80,

    ADK_SEV_LEVEL_DEBUG = 31,
    ADK_SEV_LEVEL_INFO = 15,
    ADK_SEV_LEVEL_WARNING = 7,
    ADK_SEV_LEVEL_ERROR = 3,
    ADK_SEV_LEVEL_FATAL = 1,
    ADK_SEV_LEVEL_DISABLE = 0
} adk_error_sev_level_t;

/*!
 * =============================================================
 * @enum adk_sif_status_t
 *
 * @brief
 *
 * This enum defines the error codes that are returned by the ADK
 * System infrastructure APIs (primarily bind and module manager).
 * Other broad SIF components (PDU handler, Ethernet, etc) may
 * define their own enum for specific error codes.
 *
 * =============================================================
 */
typedef enum adk_sif_status_e
{
    /* 0 << 24 is to reflect default module ID */
    ADK_SUCCESS = ((ADK_SYS_MODULE_ID << 24) + 0),

    /* System or chip level error codes */
    ADK_E_INVALID_HANDLE,
    ADK_E_SYS_NOT_INITIALIZED,
    ADK_E_SYS_ALREADY_INITIALIZED,
    ADK_E_RESYNC_FAILED,
    ADK_E_UNSUPPORTED_CHIP_INSTANCE,
    ADK_E_MODULE_NOT_REGISTERED,
    ADK_E_RESOURCE_UNAVAILABLE,
    ADK_E_INVALID_PARAMETER,
    ADK_E_INVALID_MODULE,
    ADK_E_UNREGISTERED_FUNCTION,
    ADK_E_INCONSISTENT_STATE,
    ADK_E_IMC_NOT_FOUND,
    ADK_E_ENTRY_FUNC_NOT_FOUND,
    ADK_E_UNKNOWN,
    ADK_E_UNSUPPORTED,
    ADK_E_HW_PROGRAMMING_ERROR,
    ADK_E_HW_CONFIG_LOAD_ERROR,
    ADK_E_MEM_UNAVAILABLE,
    ADK_E_INTERNAL_ERROR,
    ADK_E_VP_HANDLE_NOT_FOUND,
    ADK_E_VP_FLOW_DATA_WRITE_FAILED,
    ADK_E_VP_FLOW_DATA_READ_FAILED,
    ADK_E_VP_MC_FLOW_ADD_FAILED,
    ADK_E_VP_MC_FLOW_REM_FAILED,
    ADK_E_NS_RSRC_NOT_FOUND,
    ADK_E_NS_RSRC_CREATE_FAILED,
    ADK_E_NS_RSRC_DESTROY_FAILED,
    ADK_E_NS_RSRC_QUERY_FAILED,
    ADK_E_NS_RSRC_ALLOCATE_FAILED,
    ADK_E_NS_RSRC_FREE_FAILED,
    ADK_E_NS_RSRC_UNALIGNED,
    ADK_W_RX_HIGHWATER_MARK_HIT,
    ADK_E_RX_BUFFER_FULL,
    ADK_W_CBFN_EXISTS,
    ADK_W_CBFN_DOESNOT_EXIST,
    ADK_W_SHM_EXISTS,

    /* ADK HOS errors */
    ADK_E_HOS_TABLE_FULL,

    /* ADK NetDriver specific errors */
    ADK_E_NETD_FD_OPEN_FAILED,
    ADK_E_NETD_FP_INTF_MISMATCH,
    ADK_E_NETD_SOCK_OPEN_FAILED,
    ADK_E_NETD_SOCK_BIND_FAILED,
    ADK_E_NETD_SOCK_MSG_FAILED,

    /* MTM and SE related error codes */
    ADK_E_INVALID_SHAPING_MODE,
    ADK_E_INVALID_SCHEDULING_MODE,
    ADK_E_INVALID_ARBITRATION_MODE,
    ADK_E_QOS_SCHEDULER_LIST_NOT_EMPTY,
    ADK_E_QOS_SCHEDULER_INVALID,
    ADK_E_QOS_QUEUE_INVALID,
    ADK_E_RESOURCE_NOT_FOUND,
    ADK_E_MTM_TABLE_ERROR,
    ADK_E_MTM_TABLE_NOT_FOUND,
    ADK_E_MTM_DRAIN_BUSY,
    ADK_E_SE_ERROR,
    ADK_E_SE_STATE_GET_FAILED,
    ADK_E_SE_STATE_SET_FAILED,
    ADK_E_MTM_PARAMETER_LENGTH_INVALID,
    ADK_E_DS_POLICY_TABLE_ERROR,
    ADK_E_INVALID_DS_CLASS_TYPE,
    ADK_E_INVALID_DS_CLASS_ID,
    ADK_E_INVALID_BM_POLICY,
    ADK_E_INVALID_ATM_SERVICE_CATEGORY,
    ADK_E_UNSUPPORTED_POLICING_TYPE,
    ADK_E_INVALID_DS_POLICING_MODE,
    ADK_E_INVALID_POLICING_MODE,
    ADK_E_INVALID_POLICING_ID,
    ADK_E_DP_QUEUE_EMPTY,

    /* tree operations related error codes */
    ADK_E_INVALID_PATTERN,
    ADK_E_INVALID_ACTION,
    ADK_E_PATTERN_CREATE_FAILED,
    ADK_E_SUBNET_CREATE_FAILED,
    ADK_E_SUBNET_DESTROY_FAILED,
    ADK_E_SUBNET_ADD_FAILED,
    ADK_E_SUBNET_DELETE_FAILED,
    ADK_E_FIND_HIGHER_SUBNET_FAILED,
    ADK_E_FPL_SYMBOL_NOT_FOUND,
    ADK_E_LEARN_FAILED,
    ADK_E_UNLEARN_FAILED,
    ADK_E_UNLEARN_ANY_FAILED,
    ADK_E_UVTREE_LEARN_FAILED,
    ADK_E_UVTREE_UNLEARN_FAILED,
    ADK_E_UVTREE_UNLEARN_ANY_FAILED,
    ADK_E_OTREE_APPEND_FAILED,
    ADK_E_OTREE_INSERT_FAILED,
    ADK_E_OTREE_DELETE_FAILED,
    ADK_E_OTREE_CLEAR_FAILED,
    ADK_E_OTREE_BUILD_FAILED,
    ADK_E_OTREE_NOVRULES_GET_FAILED,
    ADK_E_OTREESET_APPEND_FAILED,
    ADK_E_OTREESET_INSERT_FAILED,
    ADK_E_OTREESET_DELETE_FAILED,
    ADK_E_OTREESET_CLEAR_FAILED,
    ADK_E_OTREESET_BUILD_FAILED,
    ADK_E_OTREESET_NOVRULES_GET_FAILED,
    ADK_E_OVTREESET_APPEND_FAILED,
    ADK_E_OVTREESET_INSERT_FAILED,
    ADK_E_OVTREESET_DELETE_FAILED,
    ADK_E_OVTREESET_CLEAR_FAILED,
    ADK_E_OVTREESET_BUILD_FAILED,
    ADK_E_OVTREESET_REBUILD_FAILED,
    ADK_E_OVTREESET_NOVRULES_GET_FAILED,

    /* EIOA specific error codes */
    ADK_E_EIOA_PORT_MAX_PKT_SIZE_SET_FAILED,
    ADK_E_EIOA_PORT_ENABLE_GET_FAILED,
    ADK_E_EIOA_PORT_ENABLE_SET_FAILED,
    ADK_E_EIOA_VP_ID_GET_FAILED,
    ADK_E_EIOA_VP_ID_SET_FAILED,
    ADK_E_EIOA_HASH_VALUE_GET_FAILED,
    ADK_E_EIOA_TAG_GET_FAILED,
    ADK_E_EIOA_TAG_SET_FAILED,
    ADK_E_EIOA_TAG_ENABLE_FAILED,
    ADK_E_EIOA_VLAN_CONFIG_FAILED,
    ADK_E_EIOA_SMAC_SET_FAILED,
    ADK_E_EIOA_QINQ_SET_FAILED,
    ADK_E_EIOA_CRC_SET_FAILED,
    ADK_E_EIOA_UC_FILTER_ENABLE_FAILED,
    ADK_E_EIOA_BC_FILTER_ENABLE_FAILED,
    ADK_E_EIOA_UC_FILTER_SET_FAILED,
    ADK_E_EIOA_SPEED_DUPLEX_GET_FAILED,
    ADK_E_EIOA_SPEED_DUPLEX_SET_FAILED,
    ADK_E_EIOA_MAC_STATS_GET_FAILED,
    ADK_E_EIOA_AUTONEG_SET_FAILED,
    ADK_E_EIOA_FLOW_CNTRL_SET_FAILED,
    ADK_E_EIOA_LOOPBACK_SET_FAILED,
    ADK_E_EIOA_TX_TIMESTAMP_GET_FAILED,
    ADK_E_EIOA_MAC_LINK_STATUS_GET_FAILED,

    /* Misc error codes */
    ADK_E_FG_UNKNOWN,
    ADK_E_FG_CFG_PARSE,
    ADK_E_FG_INVALID_CLASSIFIER_ID,
    ADK_E_FG_INVALID_FILTER_ID,
    ADK_E_FG_NOT_INITIALIZED,
    ADK_E_FG_INVALID_FILTER_TYPE,
    ADK_E_FG_NO_MEMORY,
    ADK_E_FG_GET_TOPLEVEL_MAP_FAILED,
    ADK_E_FG_GET__MAP_FAILED,
    ADK_E_FG_SYMBOL_NOT_FOUND,
    ADK_W_FG_NUM_CLASSIFIERS_PASSED_GREATER_THAN_ACTUAL_NUM,
    ADK_W_FG_NUM_FILTERS_PASSED_GREATER_THAN_ACTUAL_NUM,
    ADK_E_CFG_PARSE_ERROR,
    ADK_E_CFG_ENTRY_NOT_FOUND,
    ADK_W_CFG_SECTION_NOT_ENABLED,

    ADK_E_PTN_LIST_FAILED,
    ADK_E_SED_PARAM_SET_FAILED,

    /* Errors for Linked Lists */
    ADK_E_LIST_RANGE,
    ADK_E_LIST_INVALID_LIST,
    ADK_E_LIST_NOT_FOUND,

    /* Error for data path processing*/
    ADK_E_TASK_CONSUMED,
    ADK_E_TASK_PARKED,
    ADK_E_DROP_TASK,

    /* New EIOA Error Counters */
    ADK_E_EIOA_LPORT_SET_FAILED,
    ADK_E_EIOA_MAC_REG_MODE_GET_FAILED,
    ADK_E_EIOA_MAC_REG_MODE_SET_FAILED,
    ADK_E_EIOA_PORT_LEARN_ATTR_GET_FAILED,
    ADK_E_EIOA_PORT_LEARN_ATTR_SET_FAILED,
    ADK_E_EIOA_PORT_ATTR_GET_FAILED,
    ADK_E_EIOA_PORT_ATTR_SET_FAILED,
    ADK_E_EIOA_PORT_DEFAULT_VLAN_SET_FAILED,
    ADK_E_EIOA_PORT_VLAN_SET_FAILED,
    ADK_E_EIOA_PORT_VLAN_REM_FAILED,
    ADK_E_EIOA_PORT_CVLAN_SET_FAILED,
    ADK_E_EIOA_PORT_CVLAN_REM_FAILED,
    ADK_E_EIOA_PORT_INGRESS_OPTIONS_GET_FAILED,
    ADK_E_EIOA_PORT_INGRESS_OPTIONS_SET_FAILED,
    ADK_E_EIOA_PORT_INGRESS_PARSING_OPTIONS_GET_FAILED,
    ADK_E_EIOA_PORT_INGRESS_PARSING_OPTIONS_SET_FAILED,
    ADK_E_EIOA_UNKNOWN_MULTICAST_GET_FAILED,
    ADK_E_EIOA_UNKNOWN_MULTICAST_SET_FAILED,
    ADK_E_EIOA_PORT_ACL_CONFIG_GET_FAILED,
    ADK_E_EIOA_PORT_ACL_CONFIG_SET_FAILED,
    ADK_E_EIOA_ACL_BUILD_FAILED,
    ADK_E_EIOA_BRIDGE_DEST_VPS_SET_FAILED,
    ADK_E_EIOA_BRIDGE_ENTRY_LEARN_FAILED,
    ADK_E_EIOA_BRIDGE_ENTRY_UNLEARN_FAILED,
    ADK_E_EIOA_BRIDGE_ENTRY_LIST_FAILED,
    ADK_E_EIOA_PORT_MAX_STATS_SET_FAILED,
    ADK_E_RCU_SYNC_FAILED,

    ADK_DP_E_ERROR,
    ADK_DP_W_QOS_RED_PDU_DROPPED,

    ADK_DP_W_PH_RX_DISABLED,

    ADK_E_LOCK_FAILURE,
    ADK_I_LOCK_BUSY,
    /*Errors for EM attach*/
    ADK_E_EM_ATTACH_FAILURE,

    /*IES API ERRORS*/
    ADK_E_HLP_ERROR,

    /*QOS API ERRORS*/
    ADK_E_INVALID_POLICER_HANDLE,

    /*ADK CP LIST UTILS ERRORS */
    ADK_E_CP_MAX_OPERATIONS,
    ADK_E_CP_NO_RANGE_AVAILABLE,

    /*SYS FLOW CACHE API status*/
    ADK_E_FLOW_CACHE_ALLOC_FAILED,
    ADK_E_FLOW_CACHE_PKT_SLOW_PATH,
    ADK_E_FLOW_CACHE_PKT_FAST_PATH,

    /* Lookaside crypto session creation/initialization failure */
    ADK_E_LA_CRYPTO_CONFIG_FAILED,

    ADK_E_ENTRY_OVERWRITTEN,

    ADK_E_ERROR_MAX
} adk_sif_status_t;

/* Convenient macros to extract or mask certain fields in the returned
 * error code adk_status_t
 */

#define ADK_STATUS_SEV_LEVEL_MASK(errorValue) ((errorValue) & 0xfff0ffff)

#define ADK_STATUS_ERROR_CODE_GET(errorValue) ((errorValue) & 0x0000ffff)

#define ADK_STATUS_MODULE_ID_GET(errorValue) (((errorValue) & 0xff000000) >> 24)

#define ADK_STATUS_SEV_LEVEL_GET(errorValue) (((errorValue) & 0x000f0000) >> 16)

#endif /* #ifndef _ADK_ERROR_H_ */

/*! end addtogroup _sif
 * @}
 */
