/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_list.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief  
*
*    Description: This include file defines our list class.  For our
*                 internal purposes, we define a list as a linked list
*                 with a custom header and iterator controls.
**************************************************************************/


#ifndef _ADK_LIST_H_
#define _ADK_LIST_H_

#ifdef __cplusplus
extern "C" {
#endif

#define ADK_MUTEX_T  pthread_mutex_t
#define ADK_MUTEXATTR_T pthread_mutexattr_t
#define ADK_MUTEXATTR_RECURSIVE_NP  PTHREAD_MUTEX_RECURSIVE_NP
#define ADK_MUTEX_INIT_DEFAULT(m)      \
    pthread_mutex_init((ADK_MUTEX_T *) (m), NULL)
#define ADK_MUTEX_INIT_ATTRSET(m, attr)  \
    do { \
             ADK_MUTEXATTR_T pma; \
             pthread_mutexattr_init(&pma); \
             pthread_mutexattr_settype(&pma, attr); \
             pthread_mutex_init((ADK_MUTEX_T *) (m), &pma); \
             pthread_mutexattr_destroy(&pma); \
    } while (0)

#define ADK_MUTEX_LOCK(m)    pthread_mutex_lock(m)
#define ADK_MUTEX_UNLOCK(m)  pthread_mutex_unlock(m)
#define ADK_MUTEX_DESTROY(m) pthread_mutex_destroy(m)

/*!
 * Setups/macros
 */
#define ADK_LIST_INVALID(__list) \
    ((ADK_INVALID_PTR((__list))) || \
     ((__list)->cookie != ADK_SIF_LINKED_LIST_COOKIE) || \
     ((__list)->lh_size < 0) || \
     (((__list)->lh_size >0) && (ADK_INVALID_PTR((__list)->lh_head))) || \
     ((ADK_INVALID_PTR((__list)->lh_head)) && \
      (ADK_VALID_PTR((__list)->lh_tail))) || \
     ((ADK_VALID_PTR((__list)->lh_head)) && \
      (ADK_INVALID_PTR((__list)->lh_tail))))

/*!
 * Private types
 */

/*! The basic list node */
typedef struct _adk_list_item_s
{
    void   *li_data;                        /*! actual data storage */
    struct _adk_list_item_s *li_next;       /*! next item in list */
    struct _adk_list_item_s *li_prev;       /*! previous item in list */
} adk_list_item_t;

/*! 
 * Function Pointer Definitions
 */
typedef adk_status_t (adk_list_destroy_f)(  /*! blind destroy */
            void *data);

typedef adk_status_t  (adk_list_cmp_f)(     /*! compare */
            void *data1,
            const void *data2
            );

typedef adk_status_t  (adk_list_find_f)(    /*! matcher */
             void *data1,
             const void *data2,
             const void *extra
             );

typedef adk_status_t  (adk_list_update_f)(  /*! Update */
             const void **data1,
             const void *data2
             );

typedef adk_status_t  (adk_list_dump_f)(    /*! dump */
             void *data
             );

#define ADK_SIF_LINKED_LIST_COOKIE 0xd28ec59a
/*! The list head */
typedef struct _adk_list_head_s
{
    uint32_t        cookie;

    /*! Markers */
    adk_list_item_t    *lh_head; /*! head of item list */
    adk_list_item_t    *lh_tail; /*! tail of item list */
    adk_list_item_t    *lh_cur;  /*! pointer into list at current location */

    /*! Attributes */
    int32_t         lh_size;  /*! # of items in list */
    int32_t         lh_max;   /*! Maximum number of items possible.
                                      No limit, if set to 0 */
    bool          lh_atEnd; /*! are we at the end? */

    /*! Control functions */
    adk_list_cmp_f     *lhf_cmp;        /*! comparison for sorting */
    adk_list_find_f    *lhf_find;       /*! for matching from arg */
    adk_list_update_f  *lhf_update;     /*! for updating list data */
    adk_list_destroy_f *lhf_destroy;    /*! for detroying list items */
    adk_list_dump_f    *lhf_dump;       /*! for dumping list items */

    ADK_MUTEX_T         lh_mutex;      /*! for reentrancy protection */

} adk_list_head_t;

#define adk_linkedList_t adk_list_head_t

/*!
 * definitions for a simple linked list implementation. This utility
 * function can be extended on a need basis. These functions are defined in
 * adk_utils_tree_update.c
 */
typedef struct adk_list_entry_s
{
    uint32_t               listKey;
    void                  *listData;
    struct adk_list_entry_s    *nextEntry_p;
    struct adk_list_entry_s    *prevEntry_p;
} adk_list_entry_t;

typedef struct adk_list_s
{
    uint32_t        listSize;
    adk_list_entry_t   *listHead_p;
    adk_list_entry_t   *listTail_p;
} adk_list_t;

/*!
 * Prototypes for Linked Lists functions
 */

/*! Maintenance */
adk_status_t
adk_list_new(
        const adk_list_cmp_f       *compare_fp,
        const adk_list_find_f      *find_fp,
        const adk_list_update_f    *update_fp,
        const adk_list_destroy_f   *destroy_fp,
        const adk_list_dump_f      *dump_fp,
        const uint32_t          max_items,
        adk_linkedList_t          **newlist
        );

adk_status_t
adk_list_flush(
        adk_linkedList_t   *list
        );

adk_status_t
adk_list_append(
        adk_linkedList_t   *list,
        const void         *data
        );

adk_status_t
adk_list_prepend(
        adk_linkedList_t   *list,
        const void         *data
        );

adk_status_t
adk_list_count(
        const adk_linkedList_t *list,
        uint32_t           *count
        );

adk_status_t
adk_list_remove(
        adk_linkedList_t   *list,
        const void         *data
        );

adk_status_t
adk_list_destroy(
        adk_linkedList_t  **list
        );

/*! Positioning/iterator */
adk_status_t
adk_list_wind(
        adk_linkedList_t   *list
        );

adk_status_t
adk_list_rewind(
        adk_linkedList_t   *list
        );

adk_status_t
adk_list_sort(
        const adk_linkedList_t *list
        );

/*! Iterating */
adk_status_t
adk_list_get_next(
        adk_linkedList_t   *list,
        adk_list_item_t   **listItem,
        bool          wrapFlag
        );

adk_status_t
adk_list_get_prev(
        adk_linkedList_t   *list,
        adk_list_item_t   **listItem,
        bool          wrapFlag
        );

adk_status_t
adk_list_get_next_data(
        adk_linkedList_t   *list,
        void              **data,
        bool          wrapFlag
        );

adk_status_t
adk_list_get_prev_data(
        adk_linkedList_t   *list,
        void              **data,
        bool          wrapFlag
        );

adk_status_t
adk_list_find(
        const adk_linkedList_t *list,
        const void             *data,
        const void             *extra,
        void                  **listData
        );

adk_status_t
adk_list_update_data(
        const adk_linkedList_t *list,
        const void            **listData,
        const void             *newData
        );

adk_status_t
adk_list_dump(adk_linkedList_t *list_p);

#ifdef __cplusplus
}
#endif

#endif /* _ADK_LIST_H_ */

/*! end addtogroup _sif
 * @}
 */
