/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_log_lib.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief  
 *
 * This header file contains all macros and supporting declarations that
 * are required to support the logging mechanism.
 *
 ************************************************************************/
#ifndef __ADK_LOG_LIB_H__
#define __ADK_LOG_LIB_H__

#include <sys/time.h>
#include "adk_sys.h"
#include "adk_modules_enabled.h"

#ifdef ADK_MODULE_SIF_LTTNG_ENABLED
#include "adk_lttng_trace.h"
#define ADK_LTTNG_OUTPUT(__xx__) {__xx__}
#else
#define ADK_LTTNG_OUTPUT(__xx__)
#endif /* ifdef ADK_MODULE_SIF_LTTNG_ENABLED */


/*!
 *  ADK Debug Logging macros. By default we print all the debug information
 *  on the console. The user have to modify these macros to redirect the output
 *  to a different media (RAM, flash, network server etc). 
 *
 *  Each of the ADK modules has different messages that are written using
 *  the macros ADK_ERROR_WRITE and ADK_INFO_WRITE. By default, these macros
 *  are replaced with printf(). The applicaiton is free to change where/how
 *  the logging happens.
 *
 * At runtime, the application can change the logging level that is output
 * depending on need. The output is not a dump, but a stream of information
 * based on the level selected and the APIs being called. By default, we have
 * level ADK_LOG_LEVEL_WARNING enbled, which will print all warnings, errors,
 * and fatal errors to the console. ADK_LOG_LEVEL_INFO is typically used as a
 * debug tool.
 *
 * The ADK code is instrumented with different logging levels depending on
 * need, namely:
 *   - ADK_LOG_INFO()
 *   - ADK_LOG_WARNING()
 *   - ADK_LOG_ERROR()
 *   - ADK_LOG_FATAL_ERROR()
 *
 * The ADK_LOG macros above use the macros ADK_ERROR_WRITE() and
 * ADK_LOG_WRITE() to output the error/warning/info string. The macro
 * ADK_PREINIT_WRITE() is called for logging of errors that happen before
 * the logLevel is set in adk_sys_init(). These macros are defined in the
 * file adk_log_lib.h. The default definitions are
 *     #define ADK_ERROR_WRITE(__fmt, __args...) printf(__fmt, ##__args);
 *     #define ADK_INFO_WRITE(__fmt, __args...) printf(__fmt, ##__args);
 *     #define ADK_PREINIT_WRITE(__fmt, __args...) printf(__fmt, ##__args);
 *
 * The application can change the log level at runtime using the macro
 * ADK_MODULE_LOG_LEVEL_SET() defined in adk_module_mgr.h. The macro takes two
 * arguments, the module handle and the desired log level. The log levels are
 * defined in the data structure adk_log_level_t defined in the file adk_sys.h.
 * The following example would set ATM's log level to INFO.
 *     ADK_MODULE_LOG_LEVEL_SET(atmHandle, ADK_LOG_LEVEL_INFO);
 *
 * To set the system level log level, the variable logLevel in the data
 * structure adk_sys_conf_t (adk_sys.h) is set before calling adk_sys_init().
 * The value set here will be the default value for all modules that are
 * created, as well as the log level for the system level APIs, e.g. SIF and
 * QoS. The system log level cannot be changed at runtime.
 *
 */

#define ADK_INFO_WRITE(__fmt, __args...) printf((__fmt), ##__args);

/* Macros to dump Tree entries */
#define ADK_TABLE_ENTRY_WRITE(__fmt, __args...) printf((__fmt), ##__args);

#define ADK_INFO_BUFFER_WRITE_N(__buf, __len, __fmt, __args...) \
    snprintf((__buf), (__len), (__fmt), ##__args);
#define ADK_ERROR_BUFFER_WRITE_N(__buf, __len, __fmt, __args...) \
    snprintf((__buf), (__len), (__fmt), ##__args);
#define ADK_SNPRINTF(__buf, __len, __fmt, __args...) \
    snprintf((__buf), (__len), (__fmt), ##__args);
#define ADK_VSNPRINTF(__buf, __len, __fmt, __args) \
    vsnprintf((__buf), (__len), (__fmt), __args);

#define ADK_ERROR_WRITE(__fmt, __args...) printf((__fmt), ##__args);

#define ADK_PREINIT_WRITE(__fmt, __args...) \
    do { \
        printf((__fmt), ##__args); \
        ADK_LTTNG_OUTPUT( \
                          adk_lttng_trace_sif_force_error( \
                                                (adk_module_handle_t)NULL, \
                                                __LINE__, __FILE__, \
                                                (__fmt), ##__args); \
                        ) \
    } while(0)

/*!
 * The definitions below defines the default logging level for
 * release and debug builds. 
 */
#ifdef _ADK_DEBUG_BUILD_
#define ADK_DEFAULT_LOG_LEVEL  (ADK_LOG_LEVEL_INFO)
#else
#define ADK_DEFAULT_LOG_LEVEL  (ADK_LOG_LEVEL_WARNING)
#endif


/*!
 * This macro will return the default logging level for the system.
 * The value is a compile time option for now. If dynamic modification
 * of the default log level is desired, we may have
 * to replace the #define with a global variable.
 */
#define ADK_LOG_DEFAULT_LEVEL_GET  (ADK_DEFAULT_LOG_LEVEL) 


/*
 * Define macros for changing print colors and/or highlighting
 *
 * fields can be combined by placing a semicolon between them, like the foreground
 * bold colors.
 */

#define ADK_TEXT_BOLD                "\e[1m"
#define ADK_TEXT_UNDERLINE           "\e[4m"
#define ADK_TEXT_BLINK               "\e[5m"
#define ADK_TEXT_INVERSE             "\e[7m"
#define ADK_TEXT_RESET               "\e[m"
#define ADK_TEXT_BOLD_BLINK          "\e[1;5m"
#define ADK_TEXT_INVERSE_BLINK       "\e[7;5m"
#define ADK_TEXT_INVERSE_BLINK_RED   "\e[7;5;31m"

#define ADK_FORE_BLACK               "\e[0;30m"
#define ADK_FORE_RED                 "\e[0;31m"
#define ADK_FORE_GREEN               "\e[0;32m"
#define ADK_FORE_YELLOW_ORANGE       "\e[0;33m"
#define ADK_FORE_BLUE                "\e[0;34m"
#define ADK_FORE_MAGENTA             "\e[0;35m"
#define ADK_FORE_CYAN                "\e[0;36m"
#define ADK_FORE_LIGHT_GRAY_BLACK    "\e[0;37m"
#define ADK_FORE_DEFAULT             "\e[0;39m"


#define ADK_FORE_BOLD_GRAY           "\e[1;30m"
#define ADK_FORE_BOLD_RED            "\e[1;31m"
#define ADK_FORE_BOLD_GREEN          "\e[1;32m"
#define ADK_FORE_BOLD_YELLOW         "\e[1;33m"
#define ADK_FORE_BOLD_BLUE           "\e[1;34m"
#define ADK_FORE_BOLD_MAGENTA        "\e[1;35m"
#define ADK_FORE_BOLD_CYAN           "\e[1;36m"
#define ADK_FORE_BOLD_WHITE          "\e[1;37m"


#define ADK_BACK_BLACK               "\e[40m"
#define ADK_BACK_RED                 "\e[41m"
#define ADK_BACK_GREEN               "\e[42m"
#define ADK_BACK_YELLOW_ORANGE       "\e[43m"
#define ADK_BACK_BLUE                "\e[44m"
#define ADK_BACK_MAGENTA             "\e[45m"
#define ADK_BACK_CYAN                "\e[46m"
#define ADK_BACK_LIGHT_GRAY_BLACK    "\e[47m"
#define ADK_BACK_DEFAULT             "\e[49m"

#define IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE(__sysHandle, __handle, __mask) \
    (((__sysHandle) && ((ADK_SYS_LOG_LEVEL_GET & ADK_LOG_LEVEL_MASK) >= (__mask))) && \
     ((__handle) && ((ADK_MODULE_LOG_LEVEL_GET((__handle)) & ADK_LOG_LEVEL_MASK) >= (__mask))))

#define IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE(__sysHandle, __handle, __logType) \
    (((__sysHandle) && ((ADK_SYS_LOG_LEVEL_GET & (__logType)) == (__logType))) || \
     ((__handle) && \
      ((ADK_MODULE_LOG_LEVEL_GET((__handle)) & (__logType)) == (__logType))))

/*!
 * ADK_LOG_INFO macro will allow the caller to log messages to a 
 * terminal, file or a device. The caller is expected to set
 * up the macro ADK_LOG_WRITE to appropriate device write call.
 * input parameters:
 *
 * __sysHandle - system handle. This field is ignored for now. The user
 * can pass NULL.
 *
 * __handle - module handle. It is used to match the severity level
 * with the logging level configured for the module.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_LOG_INFO(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE \
            ((__sysHandle), (__handle), ADK_SEV_LEVEL_INFO)) \
        { \
            ADK_INFO_WRITE("[%s_log: info]", ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_INFO_WRITE((__fmt), ##__args); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_INFO_INDEX, &(__fn_ptr)); \
                        __fn_ptr((__handle), __LINE__, \
                                 __FILE__, (__fmt), ##__args); \
                          } \
                        ) \
    } while(0)


/*!
 * ADK_FORCE_LOG_INFO
 *
 * ADK_FORCE_LOG_INFO macro is similar to ADK_LOG_INFO but an unconditional
 * logging. This should be used only when both system handle and module handle
 * are not available.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_FORCE_LOG_INFO(__fmt, __args...) \
    do { \
        ADK_INFO_WRITE("[adk_unknown_log: info]"); \
        ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
        ADK_INFO_WRITE((__fmt), ##__args); \
        ADK_LTTNG_OUTPUT( \
        adk_lttng_trace_sif_force_info((adk_module_handle_t)NULL, __LINE__, \
                                       __FILE__, (__fmt), ##__args); \
                        ) \
    } while(0)

/*!
 * ADK_LOG_ERROR macro is almost same as ADK_LOG_INFO. The only difference
 * is, it is calling ADK_ERROR_WRITE instead of ADK_INFO_WRITE.
 *
 * __sysHandle - system handle. This field is ignored for now. The user
 * can pass NULL.
 *
 * __handle - module handle. It is used to match the severity level
 * with the logging level configured for the module.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_LOG_ERROR(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE \
            ((__sysHandle), (__handle), ADK_SEV_LEVEL_ERROR)) \
        { \
            ADK_ERROR_WRITE("[%s_log: ERROR]", \
                            ADK_MODULE_NAME_GET((__handle))); \
            ADK_ERROR_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_ERROR_WRITE((__fmt), ##__args); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_ERROR_INDEX, &(__fn_ptr)); \
                          __fn_ptr((__handle), __LINE__, \
                                   __FILE__, (__fmt), ##__args); \
                          } \
                        ) \
    } while(0)

/*!
 * ADK_FORCE_LOG_ERROR
 *
 * ADK_FORCE_LOG_ERROR macro is similar to ADK_LOG_ERROR but an unconditional
 * logging. This should be used only when both system handle and module handle
 * are not available.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_FORCE_LOG_ERROR(__fmt, __args...) \
    do { \
        ADK_ERROR_WRITE("[adk_unknown_log: %s]", "ERROR"); \
        ADK_ERROR_WRITE(" %s:%d ", __FILE__, __LINE__); \
        ADK_ERROR_WRITE((__fmt), ##__args); \
        ADK_LTTNG_OUTPUT( \
        adk_lttng_trace_sif_force_error((adk_module_handle_t)NULL, __LINE__, \
                                        __FILE__, (__fmt), ##__args); \
                        ) \
    } while(0)


/*!
 * ADK_LOG_FATAL_ERROR
 * This macro is almost same as ADK_LOG_ERROR. The only difference
 * is the severity level of the error that is logged is higher.
 *
 * __sysHandle - system handle. This field is ignored for now. The user
 * can pass NULL.
 *
 * __handle - module handle. It is used to match the severity level
 * with the logging level configured for the module.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_LOG_FATAL_ERROR(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE \
            ((__sysHandle), (__handle), ADK_SEV_LEVEL_FATAL)) \
        { \
            ADK_ERROR_WRITE("[%s_log: FATAL]", \
                            ADK_MODULE_NAME_GET((__handle))); \
            ADK_ERROR_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_ERROR_WRITE((__fmt), ##__args); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                          ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET(__handle), \
                                            TRACE_FATAL_ERROR_INDEX, \
                                            &(__fn_ptr)); \
                          __fn_ptr((__handle), __LINE__, \
                                   __FILE__, (__fmt), ##__args); \
                          } \
                        ) \
    } while(0)

/*!
 * ADK_FORCE_LOG_FATAL_ERROR
 *
 * ADK_FORCE_LOG_FATAL_ERROR macro is similar to ADK_LOG_FATAL_ERROR but an 
 * unconditional logging. This should be used only when both system handle 
 * and module handle are not available.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_FORCE_LOG_FATAL_ERROR(__fmt, __args...) \
    do { \
        ADK_ERROR_WRITE("[adk_unknown_log: FATAL]"); \
        ADK_ERROR_WRITE(" %s:%d ", __FILE__, __LINE__); \
        ADK_ERROR_WRITE((__fmt), ##__args); \
        ADK_LTTNG_OUTPUT( \
                    adk_lttng_trace_sif_force_fatal((adk_module_handle_t)NULL, \
                                                    __LINE__, __FILE__, \
                                                    (__fmt), ##__args); \
                        ) \
    } while(0)

/*!
 * ADK_LOG_WARNING
 * This macro is almost same as ADK_LOG_INFO. The only difference
 * is, it is calling ADK_WRITE_WARNING instead of ADK_INFO_WRITE.
 *
 * __sysHandle - system handle. This field is ignored for now. The user
 * can pass NULL.
 *
 * __handle - module handle. It is used to match the severity level
 * with the logging level configured for the module.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_LOG_WARNING(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE \
            ((__sysHandle), (__handle), ADK_SEV_LEVEL_WARNING)) \
        { \
            ADK_ERROR_WRITE("[%s_log: Warning]", \
                            ADK_MODULE_NAME_GET((__handle))); \
            ADK_ERROR_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_ERROR_WRITE((__fmt), ##__args); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                          ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET(__handle), \
                                            TRACE_WARNING_INDEX, &(__fn_ptr)); \
                          __fn_ptr((__handle), __LINE__, \
                                   __FILE__, (__fmt), ##__args); \
                          } \
                        ) \
    } while(0)

/*!
 * ADK_FORCE_LOG_WARNING
 *
 * ADK_FORCE_LOG_WARNING macro is similar to ADK_LOG_WARNING but an 
 * unconditional logging. This should be used only when both system handle 
 * and module handle are not available.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_FORCE_LOG_WARNING(__fmt, __args...) \
    do { \
        ADK_ERROR_WRITE("[adk_unknown_log: warning]"); \
        ADK_ERROR_WRITE(" %s:%d ", __FILE__, __LINE__); \
        ADK_ERROR_WRITE((__fmt), ##__args); \
        ADK_LTTNG_OUTPUT( \
        adk_lttng_trace_sif_force_warning((adk_module_handle_t)NULL, __LINE__, \
                                          __FILE__, (__fmt), ##__args); \
                        ) \
    } while(0)

/*!
 * Miscellaneous loggings */

#define ADK_LOG_FUNC_ENTRY(__sysHandle, __handle) \
    do { \
        struct timeval currTime; \
        if (IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE((__sysHandle), (__handle), \
                                                    SYS_LOG_FUNCS)) \
        { \
            gettimeofday(&currTime, NULL); \
            ADK_INFO_WRITE("[%s_log: function]", \
                           ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_INFO_WRITE("[%08lu:%08lu] Entering %s\n", \
                           currTime.tv_sec, currTime.tv_usec, __func__); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_FUNC_ENTRY_INDEX, \
                                          &(__fn_ptr)); \
                          __fn_ptr((adk_module_handle_t)(__handle), __LINE__, \
                                   __FILE__, (const char*)NULL, __func__); \
                          } \
                        ) \
    } while(0);

#define ADK_LOG_FUNC_EXIT(__sysHandle, __handle) \
    do { \
        struct timeval currTime; \
        if (IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE((__sysHandle), (__handle), \
                                                    SYS_LOG_FUNCS)) \
        { \
            gettimeofday(&currTime, NULL); \
            ADK_INFO_WRITE("[%s_log: function]", \
                           ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_INFO_WRITE("[%08lu:%08lu] Exiting %s\n", \
                           currTime.tv_sec, currTime.tv_usec, __func__); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_FUNC_EXIT_INDEX, \
                                          &(__fn_ptr)); \
                          __fn_ptr((__handle), __LINE__, \
                                   __FILE__, (const char*)NULL, __func__); \
                          } \
                        ) \
    } while(0);

#define ADK_LOG_LOCK_ACQUIRE(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE \
            ((__sysHandle), (__handle), ADK_SEV_LEVEL_INFO)) \
        { \
            ADK_INFO_WRITE("[%s_log: info]", ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_ERROR_WRITE((__fmt), ##__args); \
        } \
        ADK_LTTNG_OUTPUT( \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_LOCK_ACQUIRE_INDEX, \
                                          &(__fn_ptr)); \
                          __fn_ptr((adk_module_handle_t)(__handle), \
                               __LINE__, __FILE__, \
                               (__fmt), ##__args); \
                        ) \
    } while(0);

#define ADK_LOG_LOCK_RELEASE(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_LEVEL_ENABLED_FOR_THIS_MODULE \
            ((__sysHandle), (__handle), ADK_SEV_LEVEL_INFO)) \
        { \
            ADK_INFO_WRITE("[%s_log: info]", ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_ERROR_WRITE((__fmt), ##__args); \
        } \
        ADK_LTTNG_OUTPUT( \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_LOCK_RELEASE_INDEX, \
                                          &(__fn_ptr)); \
                          __fn_ptr((adk_module_handle_t)(__handle), \
                               __LINE__, __FILE__, \
                               (__fmt), ##__args); \
                        ) \
    } while(0);

#define ADK_LOG_INTERNAL_FUNC_ENTRY(__sysHandle, __handle) \
    do { \
        struct timeval currTime; \
        if (IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE((__sysHandle), (__handle), \
                                                    SYS_LOG_FUNCS)) \
        { \
            gettimeofday(&currTime, NULL); \
            ADK_INFO_WRITE("[%s_log: function]", \
                           ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_INFO_WRITE("[%08lu:%08lu] Entering %s\n", \
                           currTime.tv_sec, currTime.tv_usec, __func__); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_INTERNAL_FUNC_ENTRY_INDEX, \
                                          &(__fn_ptr)); \
                              __fn_ptr((__handle), __LINE__, \
                                       __FILE__, (const char*)NULL, __func__); \
                          } \
                        ) \
    } while(0);

#define ADK_LOG_INTERNAL_FUNC_EXIT(__sysHandle, __handle) \
    do { \
        struct timeval currTime; \
        if (IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE((__sysHandle), (__handle), \
                                                    SYS_LOG_FUNCS)) \
        { \
            gettimeofday(&currTime, NULL); \
            ADK_INFO_WRITE("[%s_log: function]", \
                           ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_INFO_WRITE("[%08lu:%08lu] Exiting %s\n", \
                           currTime.tv_sec, currTime.tv_usec, __func__); \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                        ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                          TRACE_INTERNAL_FUNC_EXIT_INDEX, \
                                          &(__fn_ptr)); \
                          __fn_ptr((__handle), __LINE__, \
                                   __FILE__, (const char*)NULL, __func__); \
                          } \
                        ) \
    } while(0);

#define ADK_LOG_FORCE_FUNC_ENTRY() \
    do { \
        ADK_LTTNG_OUTPUT( \
              adk_lttng_trace_force_sif_func_entry((adk_module_handle_t)NULL, \
                                                   __LINE__, __FILE__, \
                                                   (const char*)NULL, \
                                                   __func__); \
                         ) \
    } while(0);

#define ADK_LOG_FORCE_FUNC_EXIT() \
    do { \
        ADK_LTTNG_OUTPUT( \
              adk_lttng_trace_force_sif_func_exit((adk_module_handle_t)NULL, \
                                                  __LINE__, __FILE__, \
                                                  (const char*)NULL, \
                                                  __func__); \
                         ) \
    } while(0);


#define ADK_LOG_HW(__sysHandle, __handle, __fmt, __args...) \
    do { \
        if (IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE((__sysHandle), (__handle), \
                                                    SYS_LOG_HW)) \
        { \
            ADK_INFO_WRITE("[%s_log: hw]", ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d ", __FILE__, __LINE__); \
            ADK_INFO_WRITE((__fmt), ##__args);   \
        } \
        ADK_LTTNG_OUTPUT( \
                          { \
                              adk_lttng_trace_fn_ptr __fn_ptr; \
                      ADK_LTTNG_PTR_GET(ADK_MODULE_MODULEID_GET((__handle)), \
                                        TRACE_HW_INDEX, &(__fn_ptr)); \
                          __fn_ptr((__handle), __LINE__, \
                                   __FILE__, (__fmt), ##__args); \
                          } \
                        ) \
    } while(0);

#define ADK_LOG_EXT_API(__sysHandle, __handle, __api, __fmt, __args...) \
    do { \
        if (IS_LOGGING_TYPE_ENABLED_FOR_THIS_MODULE((__sysHandle), (__handle), \
                                                    SYS_LOG_EXT_API)) \
        { \
            ADK_INFO_WRITE("[%s_log: ext_API]", ADK_MODULE_NAME_GET((__handle))); \
            ADK_INFO_WRITE(" %s:%d %s() ", __FILE__, __LINE__, __api); \
            ADK_INFO_WRITE((__fmt), ##__args);   \
        } \
    } while(0);

/*!
 * ADK_FORCE_LOG_EXT_API
 *
 * ADK_FORCE_LOG_EXT_API macro is similar to ADK_LOG_EXT_API but an unconditional
 * logging. This should be used only when both system handle and module handle
 * are not available.
 *
 * __fmt, __args - charcter strings and appropriate parameters.
 * similar to printf inputs.
 */
#define ADK_FORCE_LOG_EXT_API(__api, __fmt, __args...) \
do { \
        ADK_INFO_WRITE("[adk_unknown_log: ext_API]"); \
        ADK_INFO_WRITE(" %s:%d %s() ", __FILE__, __LINE__, __api); \
        ADK_INFO_WRITE((__fmt), ##__args); \
} while(0);



#endif /* __ADK_LOG_LIB_H__ */

/*! end addtogroup _sif
 * @}
 */
