/* CHECK THIS FILE */
/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#ifndef __ADK_METADATA_H__
#define __ADK_METADATA_H__

#include <adk_os.h>
#include <adk_sif.h>
#include <adk_pktio.h>
#include <adk_domain.h>
#ifdef ADK_DEVICE_SNR
#include <rte_pmd_ice_dsi.h>
#endif
#include <rte_bridge_eth.h>

#define ADK_DSI_METADATA_SIZE_MAX   32
#define ADK_BD_METADATA_SIZE_MAX    32
#define EM_HDR_SIZE                 128 /*TODO: Need to check if we can get this from EM*/

#define ADK_IPV6_VER 6
#define ADK_IPV4_VER 4
#define ADK_SYS_INVALID_MDATA_ID 0xFFFFFFFF

#ifdef ADK_DEVICE_SNR

typedef struct ice_dsi_rx_metadata adk_dsi_rx_metadata_t;
typedef struct ice_dsi_tx_metadata adk_dsi_tx_metadata_t;

#else

typedef void adk_dsi_rx_metadata_t;
typedef void adk_dsi_tx_metadata_t;

#endif

typedef struct rte_bridge_eth_rx_metadata adk_bd_eth_rx_metadata_t;
typedef struct rte_bridge_eth_tx_metadata adk_bd_eth_tx_metadata_t;

#define ADK_METADATA_ALIGN __SIZEOF_POINTER__

/*!
 * =============================================================
 * @struct adk_metadata_t adk_sys.h "adk_sys.h"
 *
 * @brief 
 *
 * The data structure defines the common ADK metadata that is passed
 * along and in the packet from one module to another for the purpose
 * of packet processing. This metadata follow the RTE Mbuf and any
 * hardware (DSI) specific metadata at a fixed offset location. It is
 * assumed that this stucture is always aligned to uintptr_t but may
 * not be cache aligned. The IMC parameters should following this
 * common metadata structure.
 *
 * This structure does not include following elements which need
 * to be read from the RTE Mbuf data structure.
 *  - Input Port
 *  - Ingress Timestamp
 *  - PDU Start Offset
 *  - Packet Length
 *
 * =============================================================
 */
typedef struct adk_metadata_s {

    /* Chained RTE Mbuf pointer used in case of fragmentation.
     * TBD: Check if this is really needed */
    struct rte_mbuf *next_frag;        /* sizeof(uintptr_t) */

#ifdef ADK_PAL_CYCLE_MEASUREMENT
#ifndef ADK_PAL_CYCLE_MEASUREMENT_MBUF_UDATA
    /*
     * Timestamp marked at the start of ADK processing stage
     */
    uint64_t  stage_rx_ts;                /* 8B */
#endif
#endif

    /* Egress stats metadata set by protocol modules & used by Tx callback */
    adk_sys_egress_stats_mdata_t
    egress_stats[ADK_SYS_EGRESS_STATS_MDATA_ID_MAX];


    /* Input interface ID which includes physical and logical interfaces
     * like VLAN, Tunnels, BVIs, etc. */
    uint32_t input_intf_id;     /* 4B */

    /* Flow ID as per the PDU classification. This is a common IMC
     * parameter for passing destination flow ID */
    uint32_t pdu_flow_id;       /* 4B */

    /* Lock-ID, flow granularity of the curent packet that is received
     * from previous processing stage / bridge drivers. at the begining of each
     * stage processing ADK-PAL sets this filed. this is used by IP-reassmbly to
     * install timers*/
    uint32_t lock_id;       /* 4B */

    /* TimeStamp correlator. This field is set by ADK NetDriver before packet sent to ADK-FP.
     * This value is used by NetDriver in tx-time-stamp feature.
     */
    uint32_t ts_correlator;       /* 4B */

    /* Linux interface id
     * 0-255 : physical
     * 255-511: BVI
     * 512 onwards IP tunnlel*/
    uint16_t linux_intf_id;     /* 2B */

    /* IP header offset set by Ethernet at entry and valid for IPv4
     * and IPv6 packets */
    uint16_t ip_offset;         /* 2B */

    /* L4 header offset set by IP modules and used by L4 modules. */
    uint16_t l4_offset;         /* 2B */



    /* Entry function index used at the time of binding or at the time
     * of crossing processing stages */
    uint16_t func_idx;          /* 2B */

    /* Egress device-id. packet is sent on this device
     */
    uint16_t adk_dev_id;     /* 2B */

    /* Egress Queue-ID of the egress device on which the packet is
     * to be sent out.
     */
    uint16_t dev_tx_queue_id;     /* 2B */

    /* transmit completion notify function id. this function is called after a
     * packet transfer is complete.  This is set by ethernet module and used in
     * egress stats cb function.
     */
    uint8_t tx_notify_fn_id;    /* 1B */

    /* message type. defined for communication between NetDriver and ADK-FP
     * for time stamp index mgmt functionality.
     */
    uint8_t ts_msg_type;           /* 1B */

    /* Exception domain for the packet. It can be TN domain, FP Origin
     * domain or OS domain. */
    adk_sif_domain_t exception_domain;    /* 1B */

    /* event-queue-ID on what the current packet is received */
    uint8_t event_queue_id; /* 1B */



    /* Pipeline processing stage index used for multi-stage pipelined
     * processing. This needs to be revisited once static pipelining
     * model is changed to dynamic pipelining model. */
    uint8_t proc_stage_idx;     /* 1B */

    /* to indicate the current task/packet type whether the current packet/task
     * is newly (NEW) generated (IP fragments / multicast) within the ADK or an
     * incoming packet (FWD). This information is used by PAL when sending
     * packet to underlying framework.
     */
    adk_task_type_e task_type;          /* 1B */

    /* Timestamp index used for egress timestamp capturing. */
    uint8_t  timestamp_idx;     /* 1B */

    /* IP Identication used only for arp buffer. This will be removed
       post ADK R2 release. 0 mean unfragmented packet else it's a
       fragmented packet */
    uint32_t ip_identification;

    /* The neighbor status information being transmitted from inorder stage
       to buffer atomic stage */
    uint8_t neigh_status;

    /* DMAC information for packet encapsulation being transmitted from
       inorder stage to buffer atomic stage */
    uint8_t dest_mac[6];



    /* Nexthop buffer pointer being transmitted from inorder stage to
       buffer atomic stage */
    uintptr_t next_hop_buffer;

    adk_sys_common_egress_stats_mdata_t
    egress_mdata_stats[ADK_STATS_MDATA_ID_MAX];

    /* IPsec ULI would be filling innerIPVersion and innerIPHdrOffset with valid
     * values in case of Inline Encryption and Tunnel Mode.
     * In case of plain text packet/Lookaside Encryption/Transport Mode, both of these would
     * be set to 0 and IPv4/v6 should not use innerIPHdrOffset if innerIPVersion is 0.
     */
    uint8_t innerIPVersion;     /* 1B */

    /* This is application only in ULI processing */
    uint16_t innerIPHdrOffset;  /* 2B */

    uint8_t outerIPVersion;     /* 1B */

    /* This is intended to be used for while increment address stats in ingress path.
     * This should be used along with outerIPVersion to correctly handle address
     * stats update in all cases (plaintext, v4v4, v4v6, v6v6, v6v3) */
    uint32_t outerIpStatsId;    /* 4B */
    /* Scheduler synchronization type (RTE_SCHED_TYPE_*)
       of eventDev Queue on which event is received */
    uint8_t event_sched_type; /* 1B */
    /* This flag is used in ingress packet processing. This should be set by
     * first receiving IP module (v4/v6) and should be used after packet comes
     * back from ipsec to determine if forwarding is allowed in ingress interface
     */
    bool in_routing_disabled;
    /*QoS packet color - Green, Red, Yellow */
    uint8_t pktColor;
    /* Stats Id for QoS queues*/
    uint16_t qosStatsId;
    /* Flag to pass if this is PHB queue
       true -  PHB queue
       false - SL queue*/
    uint8_t  isPHBQueue:1;
    uint8_t  isIpPkt:1;
    uint8_t  isIpMcastPkt:1;
    uint8_t  isIpBcastPkt:1;
    uint8_t  isEthernetIpMcast:1;
    uint8_t  isEthernetIpBcast:1;
    uint8_t  isEthernetNonIpMcast:1;
    uint8_t  isEthernetNonIpBcast:1;
    uint16_t innerIpPktLen;
    uint16_t outerIpPktLen;
    adk_sys_dp_uli_flow_cache_entry_t *flow_cache_entry; 
    /* saID would be accessed in TX BD for all packets. This would be set with valid
     * data if the packet is going to be processed in Inline encryption mode. In all other
     * cases, this pointer would be set to ADK_IPSEC_INVALID_SA_ID.
     */
    uint16_t saID;
    /* Marker for word aligned beginning of IMC parameters structure.
     * The size of the marker is set to align the IMC parameters at
     * 8B boundary. The modules can typecast the pointer to this
     * location to the IMC parameters structure. */
    uintptr_t imc_param_start;
} __attribute__((aligned(ADK_METADATA_ALIGN))) adk_metadata_t;

/* Metadata Placement Assumptions:
 *
 *  - RX and Tx DB metadata are at offsets 0. Rx BD metadata is unused.
 *  - RX and TX DSI metadata are at default offset of 0 and 32.
 *
 * In Rx direction, ADK metadata is after the DSI metadata.
 * In Tx direction, ADK metadata falls after DB metadata and overlaps 
 * with DSI TX metadata. Therefore, before sending packet at egress,
 * the DSI Tx metadata should be set correctly, as it would occupy the
 * same space as ADK metadata and can cause random parameters to be passed
 * as DSI metadata. */

#ifdef ADK_DEVICE_SNR
#define ADK_METADATA_PTR(task) \
    ((adk_metadata_t *) ((uintptr_t) task + \
                         sizeof(struct rte_mbuf) + \
                         (ADK_DSI_METADATA_SIZE_MAX + \
                          ADK_BD_METADATA_SIZE_MAX) + \
			  EM_HDR_SIZE))
#endif

#define ADK_METADATA_PARAM(task, param) \
    (ADK_METADATA_PTR(task))->param

#define ADK_NETD_METADATA_PTR(task, ptr) \
    if(ADK_ETH_BD_TX_METADATA_OFFSET == 0) { \
        ptr = (struct adk_netd_pkt_md*) \
        ((uint8_t *)task + sizeof(struct rte_mbuf) + \
        sizeof(struct rte_bridge_cppi_tx_metadata)); \
    } else { \
        ptr = (struct adk_netd_pkt_md*) \
            ((uint8_t *)task + sizeof(struct rte_mbuf));}


#ifdef ADK_DEVICE_SNR
#define ADK_DSI_RX_METADATA_OFFSET  0
#define ADK_DSI_TX_METADATA_OFFSET  32

#define ADK_DSI_RX_METADATA_PTR(task) \
    ((adk_dsi_rx_metadata_t *) \
        rte_ice_dsi_mbuf_rx_meta((struct rte_mbuf *) task, \
                                 ADK_DSI_RX_METADATA_OFFSET))
#define ADK_DSI_TX_METADATA_PTR(task) \
    ((adk_dsi_tx_metadata_t *) \
        rte_ice_dsi_mbuf_tx_meta((struct rte_mbuf *) task, \
                                 ADK_DSI_TX_METADATA_OFFSET))
#define ADK_DSI_RX_METADATA_PARAM(task, param) \
    (ADK_DSI_RX_METADATA_PTR(task))->param
#define ADK_DSI_TX_METADATA_PARAM(task, param) \
    (ADK_DSI_TX_METADATA_PTR(task))->param

#else

#define ADK_DSI_RX_METADATA_PTR(task) \
    ((void *) ((uintptr_t) task + \
               sizeof(struct rte_mbuf)))
#define ADK_DSI_TX_METADATA_PTR(task) \
    ((void *) ((uintptr_t) task + \
               sizeof(struct rte_mbuf)))
#define ADK_DSI_RX_METADATA_PARAM(task, param) 0
#define ADK_DSI_TX_METADATA_PARAM(task, param) 0

#endif

#define ADK_BD_RX_METADATA_OFFSET  0
#define ADK_BD_TX_METADATA_OFFSET  0

#define ADK_BD_RX_METADATA_PTR(task) \
    ((adk_bd_eth_rx_metadata_t *) \
     ((uintptr_t) task + sizeof(struct rte_mbuf) + ADK_BD_RX_METADATA_OFFSET))
#define ADK_BD_TX_METADATA_PTR(task) \
    ((adk_bd_eth_tx_metadata_t *) \
     ((uintptr_t) task + sizeof(struct rte_mbuf) + ADK_BD_TX_METADATA_OFFSET))
#define ADK_BD_RX_METADATA_PARAM(task, param) \
    (ADK_BD_RX_METADATA_PTR(task))->param
#define ADK_BD_TX_METADATA_PARAM(task, param) \
    (ADK_BD_TX_METADATA_PTR(task))->param

#define ADK_MBUF_METADATA_PARAM(task, param) \
    ((struct rte_mbuf *) task)->param

#ifndef ADK_DEVICE_SNR
#define ADK_PACKET_INPUT_PORT(task) \
     ADK_MBUF_METADATA_PARAM(task, port)
#else
#define ADK_PACKET_INPUT_PORT(task) \
     ADK_DSI_RX_METADATA_PARAM(task, port)
#endif

#define ADK_MD_GET_LINUX_INTF_ID(task)  ADK_METADATA_PARAM(task, linux_intf_id)
#define ADK_MD_GET_TS_INDEX(task)       ADK_METADATA_PARAM(task, timestamp_idx)
#define ADK_MD_GET_TS_CORR(task)        ADK_METADATA_PARAM(task, ts_correlator)
#define ADK_MD_GET_ADK_DEV_ID(task)     ADK_METADATA_PARAM(task, adk_dev_id)
#define ADK_MD_GET_DEV_TX_QUEUE_ID(task)    ADK_METADATA_PARAM(task, dev_tx_queue_id)
#define ADK_MD_GET_TASK_TYPE(task)      ADK_METADATA_PARAM(task, task_type)
#define ADK_MD_GET_TS_MSG_TYPE(task)    ADK_METADATA_PARAM(task, ts_msg_type)
#define ADK_MD_GET_TX_NOTIFY_FN_ID(task) ADK_METADATA_PARAM(task, tx_notify_fn_id)

#define ADK_MD_GET_IP_OFF(task)             ADK_METADATA_PARAM(task, ip_offset)
#define ADK_MD_SET_IP_OFF(task, off)        ADK_METADATA_PARAM(task, ip_offset) = off

#define ADK_MD_GET_INNER_IP_OFF(task)       ADK_METADATA_PARAM(task, innerIPHdrOffset)
#define ADK_MD_SET_INNER_IP_OFF(task, off)       ADK_METADATA_PARAM(task, innerIPHdrOffset) = off

#define ADK_MD_GET_L4_OFF(task)             ADK_METADATA_PARAM(task, l4_offset)
#define ADK_MD_SET_L4_OFF(task, off)\
             ADK_METADATA_PARAM(task, l4_offset) = off

#define ADK_MD_GET_NEXT_FRAG(task)      ADK_METADATA_PARAM(task, next_frag)
#define ADK_MD_GET_EXCP_DOMAIN(task)         ADK_METADATA_PARAM(task, exception_domain)
#define ADK_MD_GET_IMC_PARAM_START(task)    ADK_METADATA_PARAM(task, imc_param_start)

#define ADK_MD_SET_LINUX_INTF_ID(task, val) \
     ADK_METADATA_PARAM(task, linux_intf_id) = val

#define ADK_MD_SET_TS_INDEX(task, val) \
     ADK_METADATA_PARAM(task, timestamp_idx) = val

#define ADK_MD_SET_TS_CORR(task, val) \
     ADK_METADATA_PARAM(task, ts_correlator) = val

#define ADK_MD_SET_ADK_DEV_ID(task, val) \
     ADK_METADATA_PARAM(task, adk_dev_id) = val

#define ADK_MD_SET_DEV_TX_QUEUE_ID(task, val) \
     ADK_METADATA_PARAM(task, dev_tx_queue_id) = val

#define ADK_MD_SET_TASK_TYPE(task, val) \
     ADK_METADATA_PARAM(task, task_type) = val

#define ADK_MD_SET_TS_MSG_TYPE(task, val) \
     ADK_METADATA_PARAM(task, ts_msg_type) = val

#define ADK_MD_SET_DEV_TX_NOTIFY_FN_ID(task, val) \
     ADK_METADATA_PARAM(task, tx_notify_fn_id) = val

#define ADK_MD_SET_EXCP_DOMAIN(task, val) \
     ADK_METADATA_PARAM(task, exception_domain) = val


/* Function to reset only DSI Tx metadata.
 * Note: If using this function, make sure to invoke other MD reset functions
 * like adk_pkt_adk_md_reset in the same path. Otherwise use common metadata
 * init function adk_pkt_metadata_init */
static inline void adk_pkt_dsi_tx_md_reset(adk_pktio_task_t *task_p)
{
#ifdef ADK_DEVICE_SNR
    adk_dsi_tx_metadata_t  *dsi_tx_md_p;

    dsi_tx_md_p = ADK_DSI_TX_METADATA_PTR(task_p);

    /* Reset only shared fields, remaining mandatory fields are set by
     * respective field owner*/
    dsi_tx_md_p->flags = 0;
#else
    RTE_SET_USED(task_p);
#endif
    return;
}

/* Function to reset only ADK metadata.
 * Note: If using this function, make sure to invoke other MD reset functions
 * like adk_pkt_dsi_tx_md_reset in the same path. Otherwise use common metadata
 * init function adk_pkt_metadata_init */
static inline void adk_pkt_adk_md_reset(
    adk_pktio_task_t *task_p,
    adk_sif_domain_t domain,
    uint16_t entry_func_idx,
    adk_task_type_e task_type)
{
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task_p);

    ADK_MEMSET((void *) adk_md_p, 0, sizeof(adk_metadata_t));

#ifdef ADK_DEVICE_SNR
    /* Reset mbuf tx offloads */
    task_p->tx_offload = 0;
    task_p->ol_flags &= ~(PKT_TX_OFFLOAD_MASK );

    adk_md_p->input_intf_id =
        ADK_DSI_RX_METADATA_PARAM(task_p, port);
#else
    adk_md_p->input_intf_id = task_p->port;
#endif

#ifdef ADK_DEVICE_SNR
    ADK_MD_SET_LINUX_INTF_ID(task_p, ADK_DSI_RX_METADATA_PARAM(task_p, port));
#else
    ADK_MD_SET_LINUX_INTF_ID(task_p, task_p->port);
#endif

    /* Setup the entry function index for the first stage for packets
     * received from the physical port */
    adk_md_p->func_idx = entry_func_idx;

    /* All packets received from the physical port start with TN Port
     * exception domain */
    adk_md_p->exception_domain = domain;

    /* Set the appropriate task type (NEW / FWD) */
    adk_md_p->task_type  = (uint8_t) task_type;

#if 0
    /* Rest all parameters are set to 0 */
    adk_md_p->next_frag = NULL;
    adk_md_p->pdu_flow_id = 0;
    adk_md_p->ip_offset = 0;
    adk_md_p->l4_offset = 0;
    adk_md_p->proc_stage_idx = 0;
    adk_md_p->timestamp_idx = 0;
#endif
    return;
}

/* Common packet metadata init function to be invoked by all entry points in
 * ADK.
 */
static inline void adk_pkt_metadata_init(
    adk_pktio_task_t *task_p,
    adk_sif_domain_t domain,
    uint16_t entry_func_idx,
    adk_task_type_e task_type)
{
    adk_pkt_dsi_tx_md_reset(task_p);

    adk_pkt_adk_md_reset(task_p, domain, entry_func_idx, task_type);

    return;
}

/* ADK defined macro to access timestamp field in the DPDK mbuf */
#define ADK_PKTMBUF_TIMESTAMP(m) (((struct rte_mbuf *)m)->timestamp)

/* ADK defined macro to access color field in the DPDK mbuf */
#define ADK_PKTMBUF_COLOR(m) (((struct rte_mbuf *)m)->hash.sched.lo)

#endif  /* #ifndef __ADK_METADATA_H__ */
