/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_module_mgr.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief  
 *
 * This file contains definitions and declarations of data structures 
 * and functions that are required by module manager.
 *
 ***********************************************************************/
#ifndef __ADK_MODULE_MGR_H__
#define __ADK_MODULE_MGR_H__

#include <adk_sys.h>
#include "adk_platform.h"
#include "adk_sif_cfg_pipeline.h"

/*!
 * prototype of module specific service function pointers. All the
 * modules must follow the prototypes described below.
 */
typedef adk_status_t (*module_terminate_fn_t)
    (adk_module_handle_t);
typedef adk_status_t (*module_create_fn_t)
    (adk_module_handle_t, const void *, void *);
typedef adk_status_t (*module_delete_fn_t)
    (adk_module_handle_t, const void *);
typedef adk_status_t (*module_modify_fn_t)
    (adk_module_handle_t, const void *);
typedef adk_status_t (*module_get_fn_t)
    (adk_module_handle_t, const void *);
typedef adk_status_t (*module_bind_fn_t)
    (adk_sys_handle_t, const void *);
typedef adk_status_t (*module_unbind_fn_t)
    (adk_sys_handle_t, const void *);
typedef adk_status_t (*module_extension_fn_t)
    (adk_module_handle_t, const void *);
typedef adk_status_t (*module_log_level_fn_t)
    (adk_module_handle_t, adk_log_level_t);
typedef const char *(*module_error_string_fn_t)
    (uint32_t);

/*! The void parameter should be adk_bind_info_t. Since module manager
 *  header files are higher in hierarchy (when compared to bind header
 *  files), it needs to be passed as void.
 */
typedef adk_status_t (*module_get_imc_t)
    (adk_module_handle_t, void *, uint32_t *);

/*! The void parameter should be adk_bind_info_t. Since module manager
 *  header files are higher in hierarchy (when compared to bind header
 *  files), it needs to be passed as void.
 */
typedef adk_status_t (*module_get_entry_fn_t)
    (adk_module_handle_t, void *, uint32_t *);

/*! The void pointer parameter should point to structure as defined by
 * the HOS, having the MAC event type and asociated information like 
 * MAC address, port number, etc. */
typedef void (*mac_addr_change_fn_t)
    (adk_module_handle_t, void *);

/*! The exact format of the CPU entry point for the module will
 *  be finalized after finalizing PDU handler APIs. This is just
 *  place holder for now. */
typedef adk_status_t (*module_cpu_fn_t) (void **task, uint8_t task_count,
                           uint32_t task_mask, void *ring, 
                           adk_sif_cfg_pipeline_info_t *next_stage_entry, int coreIndex);

/* List of service function IDs. The Function ID has to follow the following
 * template:
 * Module ID << ADK_SERVICE_FUNC_ID_SIZE | Service function ID
 */

#define ADK_MODULE_ID_SIZE       5
#define ADK_SERVICE_FUNC_ID_SIZE 3
#define ADK_MAX_DP_SERVICE_FN_ID 256

/* Global array of destination entry function pointers which are initialized to
 * NULL by default. This array will be populated by ADK module level DP
 * initialization functions. The DP init functions has to be called by the
 * application, after the corresponding ADK module level control plane
 * initializations.
 */

extern module_cpu_fn_t   adkDestEntryFunction_g[ADK_MAX_DP_SERVICE_FN_ID];

/*!
 * =============================================================
 * @internal
 *
 * @struct adk_module_config_t adk_module_config.h "adk_module_config.h"
 *
 * @brief 
 *
 * The configuration structure here defines the basic properties of 
 * a module. The module_name_p is a string that represent the module
 * name. module ID is a number that user assign for this module. This
 * field is mandatory and user must assign a unique number between 1
 * and 255 for each module that is added to the table below.
 * init_fn represent the intialization function for the module. The
 * user can set this field to NULL. In such cases the module 
 * initialization function must be provided with adk_module_register
 * API call.
 * Warning...This file is for exclusive use adk module manager. no
 * other file should be including this.
 * =============================================================
 */
typedef struct adk_module_config_s
{
    const char          *moduleName_p;
    uint32_t        moduleId;
    void                *initFn;
} adk_module_config_t;

/*!
 * =============================================================
 * @internal
 *
 * @struct adk_module_engine_service_fn_t adk_module_mgr.h "adk_module_mgr.h"
 *
 * @brief 
 *
 * This contains the MPP service function IDs. This data will be
 * common across all the instances of the SIF.
 * Internal to module manager 
 *
 * =============================================================
 */
typedef struct adk_module_engine_service_fn_s
{
    /*!
     * entry functions below are the indirect MPP entry point for
     * the module. MPP entry point for a module is the ID of the 
     * MPP function that process the PDU first. There are two indirect
     * entry points for a module (LLI and ULI) depending on the 
     * direction of the packet flow. 
     * The LLI function handles the packets received from lower
     * layer modules.
     */
    uint16_t          mppEntryFn_LLI;
    /*!
     * entry functions below are the indirect MPP entry point for
     * the module. MPP entry point for a module is the ID of the 
     * MPP function that process the PDU first. There are two indirect
     * entry points for a module (LLI and ULI) depending on the 
     * direction of the packet flow. 
     * The ULI function handles the packets received from 
     * upper layer modules.
     */
    uint16_t          mppEntryFn_ULI;
    /*!
     * Two entries below are the direct MPP entry point for a module.
     * The description for the indirect entry functions applies to 
     * these functions too. They are direct in the sense that the
     * source to destination module packet transfer occur via 
     * direct MPP function calls.
     */
    uint16_t          mppEntryFn_LLI_direct;
    /*!
     * Two entries below are the direct MPP entry point for a module.
     * The description for the indirect entry functions applies to 
     * these functions too. They are direct in the sense that the
     * source to destination module packet transfer occur via 
     * direct MPP function calls.
     */
    uint16_t          mppEntryFn_ULI_direct;
} adk_module_engine_service_fn_t;

/*!
 * =============================================================
 * @internal
 *
 * @struct adk_module_cpu_service_fn_t adk_module_mgr.h "adk_module_mgr.h"
 *
 * @brief 
 *
 * This contains the CPU service function IDs. This data will be
 * local to each instance of the SIF.
 * Internal to module manager 
 *
 * =============================================================
 */
typedef struct adk_module_cpu_service_fn_s
{
    /*!
     * The function pointers from _terminate to _extension are
     * the service functions supported by modules. The caller of the
     * adk_module_service_function_register can set unsupported
     * functions to NULL. The module manager will translate the ADK_xxx
     * calls to these function pointers internally.
     */
    module_terminate_fn_t moduleTerminate;
    module_create_fn_t    moduleCreate;
    module_delete_fn_t    moduleDelete;
    module_modify_fn_t    moduleModify;
    module_get_fn_t       moduleGet;
    module_bind_fn_t      moduleBind;
    module_bind_fn_t      moduleUnbind;
    module_bind_fn_t      srcModuleBind;
    /*! Custom unbind function when the module is a source module */
    module_bind_fn_t      srcModuleUnbind;
    /*! Custom bind function when the module is a destination module */
    module_bind_fn_t      dstModuleBind;
    /*! Custom unbind function when the module is a destination module */
    module_bind_fn_t      dstModuleUnbind;

    module_extension_fn_t moduleExtension;
    module_get_imc_t      moduleGetImc;
    module_get_entry_fn_t moduleGetEntryFunc;
    /*!
     * CPU entry point for a module is the pointer to the CPU function
     * that process the PDU first. 
     * The LLI function handles the packets received from lower
     * layer modules. It is the responsibility of the application create
     * PDU handler objects (receive group, receive queue, PH bind handle
     * etc) and spawn the processing thread as per PDU handler requirement.
     */
    uint16_t       cpuEntryFn_LLI[_DEST_FUNC_MAX];
    /*!
     * CPU entry point for a module is the pointer to the CPU function
     * that process the PDU first. 
     * The ULI function handles the packets received from lower
     * layer modules. It is the responsibility of the application create
     * PDU handler objects (receive group, receive queue, PH bind handle
     * etc) and spawn the processing thread as per PDU handler requirement.
     */
    uint16_t       cpuEntryFn_ULI[_DEST_FUNC_MAX];
    /*! Function to set the log-level of a module.
     */
    module_log_level_fn_t moduleLogSet;
    /*! error handler function that convert the module specific error 
     *  code to a meaningful string that user understands. 
     */
    module_error_string_fn_t errorString;
} adk_module_cpu_service_fn_t;

extern adk_module_cpu_service_fn_t adkModuleMgrCpuServiceFn_g[ADK_MAX_SUPPORTED_MODULES];

/*!
 * =============================================================
 * @internal
 *
 * @struct adk_module_info_t adk_module_mgr.h "adk_module_mgr.h"
 *
 * @brief 
 *
 * module info data structure. internal to module manager. 
 *
 * =============================================================
 */
typedef struct adk_module_info_s
{
    bool inUse;
    uint32_t cookie;
    /* Indicates that the module is registered */
    bool isRegistered;
    /* Indicates that the module is started successfully.
     * This is used in the adk_module_unregister API to
     * validate that the module is stopped.
     */
    bool isStarted;
    uint32_t reserve1;
    char *reserve2;
    adk_module_init_func_t moduleInit;
    uint32_t moduleId;
    void *module_lock; /* Can point to mutex, semaphore, spin-lock, etc */
    /*!
     * The system handle may not be needed by all modules. We are keeping
     * it here just for future reference.
     */
    adk_sys_handle_t sysHandle;
    adk_module_engine_service_fn_t  moduleEngineFunctions;
    /*!
     * moduleData_p field is used by the specific modules to store
     * their module specific information. The respective module can 
     * allocate memory for their global states and store the pointer 
     * here. This way the global information is readily available
     * with all the external API calls. 
     */
    void *moduleData_p;
} adk_module_info_t;

/*!
 * =============================================================
 * @struct adk_module_service_fn_t adk_module_mgr.h "adk_module_mgr.h"
 *
 * @brief
 *
 * Service function data structure. The modules will fill this
 * to register the service functions.
 *
 * =============================================================
 */
typedef struct adk_module_service_fn_s
{
    /*!
     * The function pointers from _terminate to _extension are
     * the service functions supported by modules. The caller of the
     * adk_module_service_function_register can set unsupported
     * functions to NULL. The module manager will translate the ADK_xxx
     * calls to these function pointers internally.
     */
    module_terminate_fn_t moduleTerminate;
    module_create_fn_t    moduleCreate;
    module_delete_fn_t    moduleDelete;
    module_modify_fn_t    moduleModify;
    module_get_fn_t       moduleGet;
    module_bind_fn_t      moduleBind;
    module_bind_fn_t      moduleUnbind;
    /*! Custom bind function when the module is a source module */
    module_bind_fn_t      srcModuleBind;
    /*! Custom unbind function when the module is a source module */
    module_bind_fn_t      srcModuleUnbind;
    /*! Custom bind function when the module is a destination module */
    module_bind_fn_t      dstModuleBind;
    /*! Custom unbind function when the module is a destination module */
    module_bind_fn_t      dstModuleUnbind;

    module_extension_fn_t moduleExtension;
    module_get_imc_t      moduleGetImc;
    module_get_entry_fn_t moduleGetEntryFunc;
    mac_addr_change_fn_t  macAddrChangeFunc;
    /*!
     * entry functions below are the indirect MPP entry point for
     * the module. MPP entry point for a module is the ID of the
     * MPP function that process the PDU first. There are two indirect
     * entry points for a module (LLI and ULI) depending on the
     * direction of the packet flow.
     * The LLI function handles the packets received from lower
     * layer modules.
     */
    uint16_t          mppEntryFn_LLI;
    /*!
     * entry functions below are the indirect MPP entry point for
     * the module. MPP entry point for a module is the ID of the
     * MPP function that process the PDU first. There are two indirect
     * entry points for a module (LLI and ULI) depending on the
     * direction of the packet flow.
     * The ULI function handles the packets received from
     * upper layer modules.
     */
    uint16_t          mppEntryFn_ULI;
    /*!
     * Two entries below are the direct MPP entry point for a module.
     * The description for the indirect entry functions applies to
     * these functions too. They are direct in the sense that the
     * source to destination module packet transfer occur via
     * direct MPP function calls.
     */
    uint16_t          mppEntryFn_LLI_direct;
    /*!
     * Two entries below are the direct MPP entry point for a module.
     * The description for the indirect entry functions applies to
     * these functions too. They are direct in the sense that the
     * source to destination module packet transfer occur via
     * direct MPP function calls.
     */
    uint16_t          mppEntryFn_ULI_direct;
    /*!
     * CPU entry point for a module is the pointer to the CPU function
     * that process the PDU first.
     * The LLI function handles the packets received from lower
     * layer modules. It is the responsibility of the application create
     * PDU handler objects (receive group, receive queue, PH bind handle
     * etc) and spawn the processing thread as per PDU handler requirement.
     */
    uint16_t       cpuEntryFn_LLI[_DEST_FUNC_MAX];
    /*!
     * CPU entry point for a module is the pointer to the CPU function
     * that process the PDU first.
     * The ULI function handles the packets received from lower
     * layer modules. It is the responsibility of the application create
     * PDU handler objects (receive group, receive queue, PH bind handle
     * etc) and spawn the processing thread as per PDU handler requirement.
     */
    uint16_t       cpuEntryFn_ULI[_DEST_FUNC_MAX];
    /*! Function to set the log-level of a module.
     */
    module_log_level_fn_t moduleLogSet;
    /*! error handler function that convert the module specific error
     *  code to a meaningful string that user understands.
     */
    module_error_string_fn_t errorString;
} adk_module_service_fn_t;

/* 
 * function proto types defined in adk_module_mgr_internal.c
 */
void print_module_dynamic_mem(void);
adk_status_t adk_module_mgr_init (adk_sys_handle_t);
adk_status_t adk_module_mgr_deinit (adk_sys_handle_t);
adk_status_t
adk_module_sif_module_register ( 
        adk_sys_handle_t sysHandle,
        const uint8_t *moduleName_p,
        adk_module_init_func_t moduleInit,
        adk_module_handle_t *moduleHandle_p);
adk_status_t
adk_module_sif_module_unregister (
        adk_module_handle_t moduleHandle);
adk_status_t adk_module_mgr_shutdown_ready (adk_sys_handle_t sysHandle);
adk_status_t adk_module_mgr_shutdown (adk_sys_handle_t);

adk_status_t adk_module_service_function_register ( 
        adk_sys_handle_t, adk_module_handle_t,
        const adk_module_service_fn_t *);

adk_status_t adk_module_service_function_update ( 
        adk_sys_handle_t, adk_module_handle_t,
        const adk_module_service_fn_t *);

/* 
 * function proto types defined in adk_module_apis.c
 */
const char * adk_module_error_string_get (
        adk_module_handle_t module_handle, uint32_t error_code);

adk_status_t adk_modify ( 
        adk_module_handle_t, const void *in_p);

adk_status_t adk_create ( 
        adk_module_handle_t, const void *in_p, void *out_p);

adk_status_t adk_get ( 
        adk_module_handle_t, const void *inout_p);

adk_status_t adk_delete ( 
        adk_module_handle_t, const void *in_p);

/*
 * all macros supported by module manager.
 */
#define ADK_MODULE_HDL_COOKIE  0x53248791

#define ADK_SIF_MODULE_PTR_GET(_mod_hdl)   \
    (ADK_SIF_DP_INFO_PTR_GET->module[(((adk_module_info_t *)(_mod_hdl))->moduleId)])

#define ADK_SIF_DP_MODULE_PTR_GET(_modId)   \
    (ADK_SIF_DP_INFO_PTR_GET->module[(_modId)])

#define ADK_MODULE_HDL_COOKIE_SET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->cookie = ADK_MODULE_HDL_COOKIE)

#define ADK_MODULE_HDL_COOKIE_CLEAR(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->cookie = 0)

#define ADK_MODULE_HANDLE_VALIDATE(_mod_hdl) \
    (((_mod_hdl)) && \
     (((adk_module_info_t *)(_mod_hdl))->cookie == ADK_MODULE_HDL_COOKIE))

#define ADK_DP_MODULE_NAME_GET(_mod_hdl)  \
    (((ADK_SIF_DP_MODULE_PTR_GET(_mod_hdl).name_p) != NULL) ? (ADK_SIF_DP_MODULE_PTR_GET(_mod_hdl).name_p) : "NULL")

#define ADK_MODULE_NAME_GET(_mod_hdl)  \
    (ADK_SIF_MODULE_PTR_GET(_mod_hdl).name_p)

#define ADK_MODULE_NAME_SET(_mod_hdl, _name_p)  \
    (ADK_SIF_MODULE_PTR_GET(_mod_hdl).name_p = (_name_p))

#define ADK_MODULE_ERROR_STRING_GET(_mod_hdl, code) \
        adk_module_error_string_get ((_mod_hdl), (code))

#define ADK_MODULE_LOG_LEVEL_SET(_mod_hdl, _logLevel) \
    (ADK_SIF_MODULE_PTR_GET(_mod_hdl).logLevel = (_logLevel))

#define ADK_MODULE_LOG_LEVEL_GET(_mod_hdl)  \
    (ADK_SIF_MODULE_PTR_GET(_mod_hdl).logLevel)

#define ADK_MODULE_MODULEID_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->moduleId)

#define ADK_MODULE_MODULEID_SET(_mod_hdl, moduleId) \
    (((adk_module_info_t *)(_mod_hdl))->moduleId = moduleId)

#define ADK_MODULE_LOCK_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->module_lock)

#define ADK_MODULE_LOCK_ENABLE(_mod_hdl) \
    (((adk_sys_info_t *)(((adk_module_info_t *)(_mod_hdl))->sysHandle))->multiThread)

#define ADK_MODULE_INFO_PTR_GET(_mod_hdl) \
    (((_mod_hdl)) ? (((adk_module_info_t *)(_mod_hdl))->moduleData_p) : NULL)

/*
 * Use this macro for when _mod_hdl has been checked to not be NULL in another
 * function. Using this macro will remove the Klocwork issue.
 */
#define ADK_MODULE_INFO_PTR_GET_INTERNAL(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->moduleData_p)

#define ADK_MODULE_INFO_PTR_SET(_mod_hdl, data_p) \
    (((adk_module_info_t *)(_mod_hdl))->moduleData_p = data_p)

#define ADK_MODULE_SYS_HANDLE_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->sysHandle)

#define ADK_MODULE_SYS_HANDLE_SET(_mod_hdl, data_p) \
    (((adk_module_info_t *)(_mod_hdl))->sysHandle = data_p)

#define ADK_MODULE_CPU_INGRESS_FN_GET(_mod_hdl, dest_func_idx) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    cpuEntryFn_LLI[dest_func_idx])

#define ADK_MODULE_CPU_EGRESS_FN_GET(_mod_hdl, dest_func_idx) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    cpuEntryFn_ULI[dest_func_idx])

#define ADK_MODULE_SRC_BIND_FN_GET(_mod_hdl) \
         (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
              srcModuleBind)
    #define ADK_MODULE_SRC_UNBIND_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    srcModuleUnbind)
                 
#define ADK_MODULE_DST_BIND_FN_GET(_mod_hdl) \
         (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
              dstModuleBind)
            
#define ADK_MODULE_DST_UNBIND_FN_GET(_mod_hdl) \
        (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
             dstModuleUnbind) 


#define ADK_MODULE_BIND_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleBind)

#define ADK_MODULE_UNBIND_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleUnbind)

#define ADK_MODULE_GET_IMC_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleGetImc)

#define ADK_MODULE_ENTRY_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleGetEntryFunc)

/*untested macros below */
#define ADK_MODULE_CREATE_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleCreate)

#define ADK_MODULE_SET_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleModify)

#define ADK_MODULE_TERMINATE_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleTerminate)

#define ADK_MODULE_DELETE_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleDelete)

#define ADK_MODULE_GET_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleGet)

#define ADK_MODULE_EXTENSION_FN_GET(_mod_hdl) \
    (adkModuleMgrCpuServiceFn_g[((adk_module_info_t *)(_mod_hdl))->moduleId].\
    moduleExtension)

#define ADK_MODULE_INGRESS_MPP_ENTRY_POINT_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->moduleEngineFunctions.mppEntryFn_LLI)

#define ADK_MODULE_EGRESS_MPP_ENTRY_POINT_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->moduleEngineFunctions.mppEntryFn_ULI)

#define ADK_MODULE_INGRESS_MPP_DIRECT_ENTRY_POINT_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->moduleEngineFunctions.mppEntryFn_LLI_direct)

#define ADK_MODULE_EGRESS_MPP_DIRECT_ENTRY_POINT_GET(_mod_hdl) \
    (((adk_module_info_t *)(_mod_hdl))->moduleEngineFunctions.mppEntryFn_ULI_direct)

#endif /* __ADK_MODULE_MGR_H__ */

/*! end addtogroup _sif
 * @}
 */
