/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_netd_fp_ts_index.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief
 *
 * This file contains definitions and declarations of objects that are
 * used for TN port timestamping.
 *
 *
 ***********************************************************************/

#ifndef __ADK_NETD_FP_TS_INDEX_H__
#define __ADK_NETD_FP_TS_INDEX_H__
#include <adk_include.h>
#include "adk_pal_dp.h"
#include "adk_sif_dp.h"
#include "adk_sif_dp_log_lib.h"

#include <adk_metadata.h>
#include "adk_nd_common.h"
#include <adk_netd_ts_index.h>

enum adk_fp_ts_error_code
{
    ADK_FP_TS_ALLOC_FAIL = 1,
    ADK_FP_TS_TX_FAIL,
    ADK_FP_TS_TX_QUEUE_FAIL
};

enum adk_fp_ts_state
{
    ADK_FP_TS_INDEX_FREE,
    ADK_FP_TS_INDEX_ALLOCATED,
    ADK_FP_TS_TRANSMITTED,
};

extern struct adk_netd_ts_index_t adk_netd_ts_index_table[ADK_NETD_MAX_PORT_GROUPS];

/* For storing the dev net id */
#ifdef ADK_DEVICE_SNR
extern int adk_fp_ts_dev_net_id;
#endif

/*
 * adk_netd_fp_ts_init
 * Called by ADK FP.
 * Initialize the Index manager for the particular port group
 * port_grp [in]: port group
 *
 * Returns 0 on success and negative value on error.
 */
int
adk_netd_fp_ts_init(uint8_t port_grp);

/*
 * adk_netd_fp_ts_index_alloc
 * Called by ADK FP.
 * On receiving the Egress packet, ADK FP call this API
 * to allocate an index.
 * This index is placed in DSI metadata before sending the packet out.
 *
 * port [in]: port number
 * corrl [in]: correlator
 * ts_index [out]: allocated timestamp index
 *
 * Returns 0 on success and negative value on error.
 */
int
adk_netd_fp_ts_index_alloc(uint8_t port,
    uint32_t corrl, uint8_t *ts_index);

/*
 * adk_netd_fp_ts_pkt_transmitted
 * Called by ADK FP.
 * The TX completion notification is received from CPK/PSM.
 * This information is passed to ADK netdriver uisng this API.
 *
 * port [in]: port number
 * ts_index [in]: timestamp index
 * corrl [in]: correlator
 *
 * Returns 0 on success and negative value on error.
 */
int
adk_netd_fp_ts_pkt_transmitted(uint8_t port, uint8_t ts_index,
        uint32_t corrl, struct rte_mbuf *dpdk_mbuf_p);

/*
 * adk_netd_fp_ts_index_free
 * Called by ADK FP.
 * Free the ts_index
 *
 * port [in]: port number
 * ts_index [in]: timestamp index
 * corrl [in]: correlator
 *
 * Returns 0 on success and negative value on error.
 */
int
adk_netd_fp_ts_index_free(uint8_t port, uint8_t ts_index,
    uint32_t corrl);

/*
 * adk_netd_fp_ts_index_provision
 * Called by ADK FP.
 * Provision the index space. This API will be useful for partitioning.
 *
 * port_grp [in]: port group
 * num_index [in]: number of indexes
 * start_offset [in]: start_offset in the index space
 * corrl [in]: correlator
 *
 * Returns 0 on success and negative value on error.
 */
int
adk_netd_fp_ts_index_provision(uint8_t port_grp, uint8_t num_idx,
        uint8_t start_offset, uint32_t corrl);

/*
 * adk_netd_fp_ts_fail
 * Called by ADK FP.
 * Report the failure to Index manager
 *
 * port [in]: port number. -1 if the packet dropped before identifying port.
 * ts_index [in]: timestamp index. -1 if the packet dropped before allocating
 * index to the right context.
 * corrl [in]: correlator. The TS_Reply message will carry the correlator,
 * which will be used by Netd to get to the right context.
 *
 * error_code [in]: error code
 *
 * Returns 0 on success and negative value on error.
 */
int
adk_netd_fp_ts_fail(int port, int ts_index,
    uint32_t corrl, enum adk_fp_ts_error_code error_code);

/*
 * adk_netd_ts_index_free_count
 * called by ADK FP
 * Report the free count in port group
 * port [in]: port number.
 *
 * Returns free count on success and negative value on error
 */
int
adk_netd_fp_ts_index_free_count(int port);

/*
 * adk_netd_ts_index_dump
 * called by ADK FP
 * dumps the index allocation status for port group
 *
 * port [in]: port number.
 */
void
adk_netd_fp_ts_index_dump(int port);

#endif /* __ADK_NETD_FP_TS_INDEX_H__ */

/*! end addtogroup _sif
 * @}
 */

