/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_os.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief
 *
 * This file contains all OS specific macros. The macros internally
 * calls RTE suppport OS wraper functions.
 *
 ***********************************************************************/

#ifndef __ADK_OS_H__
#define __ADK_OS_H__

#ifdef ADK_SAFESTRLIB
#include "safe_str_lib.h"
#include "safe_mem_lib.h"
#endif

#include "adk_sys.h"
#include <assert.h>
#include <pthread.h>

/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * A macro to map 'const char *' to 'char *' until the RTE fixes a couple
 * of their defects (Bugzilla 26746).
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */
#define ADK_MAP_CONST_CHAR_STAR char *

/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * Macros for memory allocation and free
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#if ! defined(__COVERITY__) && ! defined(KW_ARCH)
/*!
 * RAM memory management functions are used for allocating and freeing
 * scratch pad memory (used for temporary information storage). These
 * macros are always mapped to RTE RAM memory allocation functions.
 */
/*! malloc call for RAM */
#define ADK_RAM_MALLOC  malloc
/*! calloc call for RAM */
#define ADK_RAM_CALLOC  calloc
/*! free call for RAM */
#define ADK_RAM_FREE    free
/*! realloc call for RAM */
#define ADK_RAM_REALLOC realloc

/*!
 * NVM memory management functions are used for allocating and freeing
 * space in non-volatile memory to store permenant state information.
 * Macros here are mapped to RTE nvm alloc/free functions if ADK_USE_NVM
 * env is defined. The default behavior is to call RTE RAM management
 * functions. The user must use these macros to allocate storage for
 * all state information that will impact the recovery of the system
 * after warm restart.
 */

#ifndef ADK_USE_NVM
extern void* nvm_mod_alloc(unsigned int mod_id, unsigned int obj_id, size_t p, size_t s, size_t a, size_t b);

/*! Allocate single memory block using DPDK */
#define ADK_NVM_MALLOC(s)      rte_malloc(NULL, (s), 0)
/*! re-allocate single memory block using DPDK */
#define ADK_NVM_REALLOC(p, s)  rte_realloc((void*)(p), (s), 0)
/*! calloc call using DPDK */
#define ADK_NVM_CALLOC(a, b, p, s) 	nvm_mod_alloc((a), (b), (p), (s), 0, 0)
#define ADK_NVM_HANDLE_CALLOC(a, b, p, s) 	nvm_mod_alloc((a), (b), (p), (s), 0, 2)
/*! free call using DPDK */
#define ADK_NVM_FREE(ptr)      rte_free((void*)(ptr))

/*****************************************************************************
 * Aligned allocation for data plane structures and tables from shared memory.
 *****************************************************************************/

#define ADK_DP_SHM_MALLOC(s, a)      rte_malloc(NULL, (s), a)
/*! re-allocate single memory block using DPDK */
#define ADK_DP_SHM_REALLOC(p, s, a)  rte_realloc((void*)(p), (s), (a))
/*! calloc call using DPDK */
#define ADK_DP_SHM_CALLOC(b, c, p, s, a)   nvm_mod_alloc((b), (c), (p), (s), (a), 1)
/*! free call using DPDK */
#define ADK_DP_SHM_FREE(ptr)      rte_free((void*)(ptr))

#endif /* ADK_USE_NVM */

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
# else /* Running Coverity */
#define ADK_RAM_MALLOC(s)      malloc((s))
#define ADK_RAM_REALLOC(p, s)  realloc((p), (s))
#define ADK_RAM_CALLOC(c, s)   calloc((c), (s))
#define ADK_RAM_FREE(ptr)      free((void *)(ptr))

#define ADK_NVM_MALLOC(s)      malloc((s))
#define ADK_NVM_REALLOC(p, s)  realloc((s))
#define ADK_NVM_CALLOC(a, b, p, s)   calloc((p), (s))
#define ADK_NVM_FREE(ptr)      free((void *)(ptr))

#define ADK_DP_SHM_MALLOC(s)     malloc((s))
#define ADK_DP_SHM_REALLOC(p, s) realloc((s))
#define ADK_DP_SHM_CALLOC(b, c, p, s, a)  calloc((p), (s))
#define ADK_DP_SHM_FREE(ptr)        free((void *)(ptr))
# endif


/*
 * Change the ADK_*_FREE calls to call free when we are running splint. This will
 * ensure that it see the free and it won't give a memory leak warning.
 */
#ifdef S_SPLINT_S
#undef ADK_RAM_FREE
#define ADK_RAM_FREE free
#undef ADK_NVM_FREE
#define ADK_NVM_FREE free
#endif

/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * Macros for implementing multi-thread safe locking mechanism
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */

/*
 * Add an assert when we have certain errors while running static analysis.
 * This is needed as the tools aresn't tracking that we have already verified
 * the lock handle in the acquire call.
 */
#if defined(__COVERITY__)
#define ADK_STATIC_ANALYSIS_KILLPATH(__retValue) assert((__retValue) == 0)
#elif defined(ADK_KLOCWORK)
#define ADK_STATIC_ANALYSIS_KILLPATH(__retValue) assert((__retValue) == 0)
#else
#define ADK_STATIC_ANALYSIS_KILLPATH(__retValue)
#endif

#define ADK_LOCK_SHARED     TRUE
#define ADK_LOCK_VALUE      1    /* Not used internally by RTE as RTE uses mutex */

#define ADK_MODULE_LOCK_SHARED     TRUE
#define ADK_MODULE_LOCK_VALUE      1    /* Not used internally by RTE as RTE uses mutex */


/*!
 * =============================================================
 * @internal
 * @enum adk_sys_rwlock_rw_type_t
 *
 * @brief When calling the read-write lock API to lock, are we want a read or
 *        a write lock
 *
 * =============================================================
 */
typedef enum adk_sys_rwlock_rw_type_e {
    ADK_SYS_RWLOCK_WRITE = 1,
    ADK_SYS_RWLOCK_READ
} adk_sys_rwlock_rw_type_t;

typedef struct adk_sys_lock_s              *adk_sys_lock_t;
typedef struct adk_sys_lock_attr_s         *adk_sys_lock_attr_t;

/*
 * Define the types of locks we want
 */
#define ADK_USE_PTHREAD_SPIN_LOCKS
#define ADK_USE_PTHREAD_RW_LOCKS

/*
 * These includes need to come after the stuct adk_sys_lock_t
 */
#include "adk_os_pthread_locks.h"
#include "adk_os_dpdk_locks.h"


/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#if defined(ADK_USE_DPDK_SPIN_LOCKS)
/* Use DPDK spin lock if enabled */
#define ADK_ALLOC_LOCK()                    ADK_ALLOC_DPDK_SPIN_LOCK()
#define ADK_FREE_LOCK(lock)                 ADK_FREE_DPDK_SPIN_LOCK((lock))
#define ADK_INIT_LOCK(lock, shared, value)  ADK_INIT_DPDK_SPIN_LOCK((lock))
#define ADK_DESTROY_LOCK(lock)              ADK_DESTROY_DPDK_SPIN_LOCK((lock))
#define ADK_ACQUIRE_LOCK(lock, _retValue) \
        do { \
            (_retValue) = ADK_ACQUIRE_DPDK_SPIN_LOCK((lock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Lock acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)

#define ADK_RELEASE_LOCK(lock, _retValue) \
        do { \
            adk_status_t    __ret; \
            __ret = ADK_RELEASE_DPDK_SPIN_LOCK((lock)); \
            if (__ret == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
            } else \
            { \
                ADK_STATIC_ANALYSIS_KILLPATH(_retValue); \
                (_retValue) = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_FATAL_ERROR("Lock release failed\n"); \
                goto _lock_release_failed; \
            } \
        } while (0)

#define ADK_TRY_ACQUIRE_LOCK(lock, _retValue) \
        do { \
            (_retValue) = ADK_TRY_ACQUIRE_DPDK_SPIN_LOCK((lock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "Try lock acquire\n"); \
            } else if ((_retValue) == ADK_I_LOCK_BUSY) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_INFO(__sysHandle, \
                             ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                             "Lock try acquire busy\n"); \
                goto _lock_acquire_failed; \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Lock try acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#elif defined(ADK_USE_PTHREAD_SPIN_LOCKS)
/* Use pthread spin lock if enabled */
#define ADK_ALLOC_LOCK() \
    ADK_ALLOC_PTHREAD_SPIN_LOCK()
#define ADK_FREE_LOCK(lock) \
    ADK_FREE_PTHREAD_SPIN_LOCK((lock))
#define ADK_INIT_LOCK(lock, shared, value)  \
        ADK_INIT_PTHREAD_SPIN_LOCK((lock), (shared))
#define ADK_DESTROY_LOCK(lock) \
        ADK_DESTROY_PTHREAD_SPIN_LOCK((lock))

#define ADK_ACQUIRE_LOCK(lock, _retValue) \
        do { \
            (_retValue) = ADK_ACQUIRE_PTHREAD_SPIN_LOCK((lock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Lock acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)

#define ADK_RELEASE_LOCK(lock, _retValue) \
        do { \
            adk_status_t    __ret; \
            __ret = ADK_RELEASE_PTHREAD_SPIN_LOCK((lock)); \
            if (__ret == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
            } else \
            { \
                ADK_STATIC_ANALYSIS_KILLPATH(_retValue); \
                (_retValue) = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_FATAL_ERROR("Lock release failed\n"); \
                goto _lock_release_failed; \
            } \
        } while (0)

#define ADK_TRY_ACQUIRE_LOCK(lock, _retValue) \
        do { \
            (_retValue) = ADK_TRY_ACQUIRE_PTHREAD_SPIN_LOCK((lock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "Try lock acquire\n"); \
            } else if ((_retValue) == ADK_I_LOCK_BUSY) \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_INFO(__sysHandle, \
                             ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                             "Lock try acquire busy\n"); \
                goto _lock_acquire_failed; \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Lock try acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#else
#define ADK_ALLOC_LOCK()
#define ADK_FREE_LOCK(lock)
#define ADK_INIT_LOCK(lock, shared, value)
#define ADK_DESTROY_LOCK(lock)
#define ADK_ACQUIRE_LOCK(lock, _retValue)
#define ADK_RELEASE_LOCK(lock, _retValue)
#define ADK_TRY_ACQUIRE_LOCK(lock, _retValue)
#endif

/* Read-Write Locks related Macros */
/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#if defined(ADK_USE_DPDK_RW_LOCKS)
/* Use DPDK rwlock if enabled */
#define ADK_ALLOC_RW_LOCK()                 ADK_ALLOC_DPDK_RW_LOCK()
#define ADK_FREE_RW_LOCK(rwlock)            ADK_FREE_DPDK_RW_LOCK((rwlock))
#define ADK_INIT_RW_LOCK(rwlock, shared)    ADK_INIT_DPDK_RW_LOCK((rwlock))
#define ADK_DESTROY_RW_LOCK(rwlock)         ADK_DESTROY_DPDK_RW_LOCK((rwlock))
#define ADK_ACQUIRE_RW_R_LOCK(rwlock) \
        do { \
            (_retValue) = ADK_ACQUIRE_DPDK_RW_R_LOCK((rwlock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW R lock acquire\n"); \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("RW R lock acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)
#define ADK_ACQUIRE_RW_W_LOCK(rwlock) \
        do { \
            (_retValue) = ADK_ACQUIRE_DPDK_RW_W_LOCK((rwlock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW W lock acquire\n"); \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("RW W lock acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)
#define ADK_RELEASE_RW_R_UNLOCK(rwlock, _retValue) \
        do { \
            adk_status_t    __ret; \
            __ret = ADK_RELEASE_DPDK_RW_R_LOCK((rwlock)); \
            if (__ret == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW R lock release\n"); \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("RW R lock release failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)
#define ADK_RELEASE_RW_W_UNLOCK(rwlock, _retValue) \
        do { \
            adk_status_t    __ret; \
            __ret = ADK_RELEASE_DPDK_RW_W_LOCK((rwlock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW W lock release\n"); \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("RW W lock release failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#elif defined(ADK_USE_PTHREAD_RW_LOCKS)
/* Use pthread rwlock if enabled */
#define ADK_ALLOC_RW_LOCK() \
    ADK_ALLOC_PTHREAD_RW_LOCK()
#define ADK_FREE_RW_LOCK(rwlock) \
    ADK_FREE_PTHREAD_RW_LOCK((rwlock))
#define ADK_INIT_RW_LOCK(rwlock, shared) \
    ADK_INIT_PTHREAD_RW_LOCK((rwlock), (shared))
#define ADK_DESTROY_RW_LOCK(rwlock) \
    ADK_DESTROY_PTHREAD_RW_LOCK((rwlock))
#define ADK_ACQUIRE_RW_R_LOCK(rwlock, _retValue) \
        do { \
            (_retValue) = ADK_ACQUIRE_PTHREAD_RW_R_LOCK((rwlock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW R lock acquire\n"); \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("RW R lock acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)
#define ADK_ACQUIRE_RW_W_LOCK(rwlock, _retValue) \
        do { \
            (_retValue) = ADK_ACQUIRE_PTHREAD_RW_W_LOCK((rwlock)); \
            if ((_retValue) == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW W lock acquire\n"); \
            } else \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("RW W lock acquire failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)
#define ADK_RELEASE_RW_R_UNLOCK(rwlock, _retValue) \
        do { \
            adk_status_t    __ret; \
            __ret = ADK_RELEASE_PTHREAD_RW_LOCK((rwlock)); \
            if (__ret == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW R lock release\n"); \
            } else \
            { \
                ADK_STATIC_ANALYSIS_KILLPATH(_retValue); \
                (_retValue) = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_FATAL_ERROR("RW R lock release failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)
#define ADK_RELEASE_RW_W_UNLOCK(rwlock, _retValue) \
        do { \
            adk_status_t    __ret; \
            __ret = ADK_RELEASE_PTHREAD_RW_LOCK((rwlock)); \
            if (__ret == 0) \
            { \
                adk_sys_handle_t    __sysHandle; \
                __sysHandle = adkSysShmConf_p->sysHandle; \
                ADK_LOG_LOCK_ACQUIRE(__sysHandle, \
                                     ADK_SYS_MODULE_HANDLE_GET (__sysHandle), \
                                     "RW W lock release\n"); \
            } else \
            { \
                ADK_STATIC_ANALYSIS_KILLPATH(_retValue); \
                (_retValue) = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_FATAL_ERROR("RW W lock release failed\n"); \
                goto _lock_acquire_failed; \
            } \
        } while (0)

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#else
#define ADK_ALLOC_RW_LOCK()
#define ADK_FREE_RW_LOCK(rwlock)
#define ADK_INIT_RW_LOCK(rwlock, shared)
#define ADK_DESTROY_RW_LOCK(rwlock)
#define ADK_ACQUIRE_RW_R_LOCK(rwlock, _retValue)
#define ADK_ACQUIRE_RW_W_LOCK(rwlock, _retValue)
#define ADK_RELEASE_RW_R_UNLOCK(rwlock, _retValue)
#define ADK_RELEASE_RW_W_UNLOCK(rwlock, _retValue)
#endif

/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * Memory copy and string related macros
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#ifndef ADK_SAFESTRLIB
#define ADK_MEMCPY(to, from, size) \
     memcpy(to, from, (size_t) size)
#define ADK_MEMMOVE(to, from, size) \
     memmove(to, from, (size_t) size)
#define ADK_MEMCMP(m1, m2, size) \
     memcmp((void *) m1, (void *) m2, (size_t) size)
#define ADK_MEMSET(m1, val, sz)  \
    memset((void *) m1, (int32_t)val, (size_t)sz)


/*
 * C programming libraries for string manipulation.
 */
#define ADK_STRNCPY(s1, s2, n)        strncpy((char *)s1, (char *)s2, n)
#define ADK_STRNCMP(s1, s2, n)        strncmp((char *)s1, (char *)s2, n)
#define ADK_STRCPY(s1, s2)            echo "use ADK_STRNCPY()"
#define ADK_STRNCAT(s1, s2, n)        strncat((char *)s1, (char *)s2, n)
#define ADK_STRLEN(s1)                strlen((char *)s1)

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#else
#define ADK_MEMCPY(to, from, size) \
     memcpy_s(to, (size_t) size, from, (size_t) size)
#define ADK_MEMMOVE(to, from, size) \
     memmove_s(to, (size_t) size, from, (size_t) size)
#define ADK_MEMCMP(m1, m2, size) \
    { \
        int *diff = NULL; \
        memcmp_s((void *) m1, (size_t) size, \
                 (void *) m2, (size_t) size, diff); \
    }
#define ADK_MEMSET(m1, val, sz)  \
    memset_s((void *) m1, (size_t)sz, (uint8_t)val)


/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * C programming libraries for string manipulation.
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */
#define ADK_STRNCPY(s1, s2, n)        strncpy_s((char *)s1, n, (char *)s2, n)
#define ADK_STRNCMP(s1, s2, n)        strncmp((char *)s1, (char *)s2, n)
#define ADK_STRCPY(s1, s2)            echo "use ADK_STRNCPY()"
#define ADK_STRNCAT(s1, s2, n)        strncat_s((char *)s1, n, (char *)s2, n)
#define ADK_STRLEN(s1)                strlen((char *)s1)
#endif

#define ADK_BCOPY(src, dst, size)     bcopy(src, dst, size)

/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * File specific wrappers.
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */
#define ADK_FOPEN fopen
#define ADK_FCLOSE fclose
#define ADK_FREAD fread
#define ADK_FILE_DESCRIPTOR FILE *

#define ADK_FLOOR floor
#define ADK_CEIL ceil


/*
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 *
 * Adding macros to help with symbol versioning
 *
 * SYMBOL is the exported symbol name
 * FUNCTION is the actual function
 * VERSION is the version string to be added to the library
 *
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
 */
# define ADK_SYMBOL_VERSIONED(FUNCTION, SYMBOL, VERSION) \
    _ADK_SYMBOL_VERSION (FUNCTION, SYMBOL, VERSION)
# define ADK_SYMBOL_DEFAULT(FUNCTION, SYMBOL, VERSION) \
    _ADK_DEFAULT_SYMBOL_VERSION (FUNCTION, SYMBOL, VERSION)
# define ADK_SYMBOL_ALIAS(FUNCTION, ALIAS_NAME) \
    _ADK_STRONG_ALIAS(FUNCTION, ALIAS_NAME)
# define ADK_SYMBOL_UNDEFINED(FUNCTION, SYMBOL) \
    _ADK_UNDEF_SYMBOL_VERSION(FUNCTION, SYMBOL)

/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#ifdef ADK_USE_LIBTOOL
# define _ADK_SYMBOL_VERSION(real, name, version) \
    __asm__ (".symver " #real "," #name "@" #version)
# define _ADK_DEFAULT_SYMBOL_VERSION(real, name, version) \
    __asm__ (".symver " #real "," #name "@@" #version)
# define _ADK_STRONG_ALIAS(name, aliasname) \
    extern __typeof (name) aliasname __attribute__ ((alias (#name)));
# define _ADK_UNDEF_SYMBOL_VERSION(real, name) \
    __asm__ (".symver " #real "," #name "@" )
/* +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ */
#else
# define _ADK_SYMBOL_VERSION(real, name, version)
# define _ADK_DEFAULT_SYMBOL_VERSION(real, name, version) \
    extern __typeof (real) name __attribute__ ((alias (#real)));
# define _ADK_STRONG_ALIAS(name, aliasname)
# define _ADK_UNDEF_SYMBOL_VERSION(real, name)

#endif


#if 0
#define ADK_HIDDEN_PROTO(name, attrs...) \
    __ADK_HIDDEN_PROTO (name, __GI_##name, ##attrs)
#define ADK_HIDDEN_DEF(name) \
    __ADK_HIDDEN_VER(__GI_##name, name, name);
#define ADK_HIDDEN_DATA_DEF(name) \
    ADK_HIDDEN_DEF(name)

#define __ADK_HIDDEN_VER(local, internal, name) \
    extern __typeof (name) __EI_##name __asm__(__ADK_HIDDEN_ASMNAME (#internal)); \
    extern __typeof (name) __EI_##name __attribute__((alias (__ADK_HIDDEN_ASMNAME (#local))))

#define __ADK_HIDDEN_PROTO(name, internal, attrs...) \
    extern __typeof (name) name __asm__ (__ADK_HIDDEN_ASMNAME (#internal)) \
    __ADK_HIDDEN_PROTO_HIDDENATTR (attrs);
#define __ADK_HIDDEN_PROTO_HIDDENATTR(attrs...) \
    __attribute__ ((visibility ("hidden"), ##attrs))
#define __ADK_HIDDEN_ASMNAME(name) \
    __ADK_HIDDEN_ASMNAME1 (__USER_LABEL_PREFIX__, name)
#define __ADK_HIDDEN_ASMNAME1(prefix, name) \
    __ADK_HIDDEN_ASMNAME2(prefix, name)
#define __ADK_HIDDEN_ASMNAME2(prefix, name) \
    #prefix name
#endif


#endif /* __ADK_OS_H__ */

/*! end addtogroup _sif
 * @}
 */
