/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_os_dpdk_locks.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief  This file contains all OS specific macros for DPDK locks.
 *
 * This file should not be called directly, it should be inside adk_os.h
 *
 ***********************************************************************/

#ifndef __ADK_OS_DPDK_LOCKS_H__
#define __ADK_OS_DPDK_LOCKS_H__

#ifndef __ADK_OS_H__
#error "This file is to be called from adk_os.h"
#endif

#include "adk_error.h"
#include "adk_log_lib.h"
#include "adk_sif.h"

#define VALIDATE_DPDK_LOCK(lock) \
    do { \
        if ((lock) == NULL) \
        { \
            adk_sys_handle_t    sysHandle; \
            /* only error if locks are enabled */ \
            sysHandle = adkSysShmConf_p->sysHandle; \
            if (ADK_SYS_LOCK_ENABLE(sysHandle) == true) \
            { \
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("lock is NULL"); \
            } \
            goto _exit; \
        } \
    } while (0)

/*
 * Need to invert the logic of the return value for the RTE spinlocks so they
 * match pthread returns; 0 success, !0 failure
 */

/*!
 * =============================================================
 * DPDK rwlocks
 * =============================================================
 */

static inline int
adk_sys_rte_rwlock_free(
    adk_sys_lock_t     *rwlock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(rwlock);

    ADK_NVM_FREE(rwlock);

_exit:
    return retValue;
}

static inline int
adk_sys_rte_rwlock_init(
    adk_sys_lock_t     *rwlock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(rwlock);

    rte_rwlock_init((rte_rwlock_t *)rwlock);

_exit:
    return retValue;
}

static inline int
adk_sys_rte_rwlock_lock(
    adk_sys_lock_t             *rwlock,
    adk_sys_rwlock_rw_type_t    rwFlag)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(rwlock);

    if (rwFlag == ADK_SYS_RWLOCK_READ)
    {
        rte_rwlock_read_lock((rte_rwlock_t *)rwlock);
    } else if (rwFlag == ADK_SYS_RWLOCK_WRITE)
    {
        rte_rwlock_write_lock((rte_rwlock_t *)rwlock);
    } else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Invalid rwlock type: '0x%x'", rwFlag);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_rte_rwlock_unlock(
    adk_sys_lock_t             *rwlock,
    adk_sys_rwlock_rw_type_t    rwFlag)
{
    adk_status_t    retValue = 0;

    VALIDATE_DPDK_LOCK(rwlock);

    if (rwFlag == ADK_SYS_RWLOCK_READ)
    {
        rte_rwlock_read_unlock((rte_rwlock_t *)rwlock);
    } else if (rwFlag == ADK_SYS_RWLOCK_WRITE)
    {
        rte_rwlock_write_unlock((rte_rwlock_t *)rwlock);
    } else
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_FATAL_ERROR("Invalid rwlock type: '0x%x'", rwFlag);
        goto _exit;
    }

_exit:
    return retValue;
}

/* DPDK based Read-Write Locks related Macros */
#define ADK_ALLOC_DPDK_RW_LOCK() \
    ADK_NVM_MALLOC(sizeof(rte_rwlock_t))

#define ADK_FREE_DPDK_RW_LOCK(rwlock) \
    adk_sys_rte_rwlock_free((rwlock))

#define ADK_INIT_DPDK_RW_LOCK(rwlock) \
    adk_sys_rte_rwlock_init((rwlock))

#define ADK_ACQUIRE_DPDK_RW_R_LOCK(rwlock, shared) \
    adk_sys_rte_rwlock_lock((rwlock), ADK_SYS_RWLOCK_READ)

#define ADK_RELEASE_DPDK_RW_W_LOCK(rwlock) \
    adk_sys_rte_rwlock_lock((rwlock), ADK_SYS_RWLOCK_WRITE)

#define ADK_ACQUIRE_DPDK_RW_R_UNLOCK(rwlock, shared) \
    adk_sys_rte_rwlock_unlock((rwlock), ADK_SYS_RWLOCK_READ)

#define ADK_RELEASE_DPDK_RW_W_UNLOCK(rwlock) \
    adk_sys_rte_rwlock_unlock((rwlock), ADK_SYS_RWLOCK_WRITE)

#define ADK_DESTROY_DPDK_RW_LOCK(rwlock)


/*!
 * =============================================================
 * DPDK spin locks
 * =============================================================
 */

static inline int
adk_sys_rte_spinlock_free(
    adk_sys_lock_t *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(lock);

    ADK_NVM_FREE(lock);

_exit:
    return retValue;
}

static inline int
adk_sys_rte_spinlock_init(
    adk_sys_lock_t *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(lock);

    rte_spinlock_init((rte_spinlock_t *)lock);

_exit:
    return retValue;
}

static inline int
adk_sys_rte_spinlock_lock(
    adk_sys_lock_t *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(lock);

    rte_spinlock_lock((rte_spinlock_t *)lock);

_exit:
    return retValue;
}

static inline int
adk_sys_rte_spinlock_unlock(
    adk_sys_lock_t *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(lock);

    rte_spinlock_unlock((rte_spinlock_t *)lock);

_exit:
    return retValue;
}

static inline int
adk_sys_rte_spinlock_trylock(
    adk_sys_lock_t *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_DPDK_LOCK(lock);

    if (rte_spinlock_trylock((rte_spinlock_t *)lock) == 0)
    {
        retValue = ADK_I_LOCK_BUSY;
    }

_exit:
    return retValue;
}

/* DPDK based Spin Locks related Macros */
#define ADK_ALLOC_DPDK_SPIN_LOCK() \
    ((rte_spinlock_t *)(ADK_NVM_MALLOC(sizeof(rte_spinlock_t))))

#define ADK_FREE_DPDK_SPIN_LOCK(lock) \
    adk_sys_rte_spinlock_free((lock))

#define ADK_INIT_DPDK_SPIN_LOCK(lock) \
    adk_sys_rte_spinlock_init((lock))

#define ADK_ACQUIRE_DPDK_SPIN_LOCK(lock) \
    adk_sys_rte_spinlock_lock((lock))

#define ADK_TRY_ACQUIRE_DPDK_SPIN_LOCK(lock) \
    adk_sys_rte_spinlock_trylock((lock))

#define ADK_RELEASE_DPDK_SPIN_LOCK(lock) \
    adk_sys_rte_spinlock_unlock((lock))

#define ADK_DESTROY_DPDK_SPIN_LOCK(lock)


#endif /* __ADK_OS_DPDK_LOCKS_H__ */

/*! end addtogroup _sif
 * @}
 */
