/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_os_pthread_locks.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief  
 *
 * This file contains all OS specific macros. The macros internally 
 * calls RTE suppport OS wraper functions.
 *
 ***********************************************************************/

#ifndef __ADK_OS_PTHREAD_LOCKS_H__
#define __ADK_OS_PTHREAD_LOCKS_H__

#ifndef __ADK_OS_H__
#error "This file is to be called from adk_os.h"
#endif

#include "adk_error.h"
#include "adk_log_lib.h"
#include "adk_sif.h"

#define VALIDATE_PTHREAD_LOCK(lock) \
    do { \
        if ((lock) == NULL) \
        { \
            adk_sys_handle_t    sysHandle; \
            /* only error if locks are enabled */ \
            sysHandle = adkSysShmConf_p->sysHandle; \
            if (ADK_SYS_LOCK_ENABLE(sysHandle) == true) \
            { \
                retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("lock is NULL"); \
            } \
            ADK_STATIC_ANALYSIS_KILLPATH(retValue); \
            goto _exit; \
        } \
    } while (0)

/*!
 * =============================================================
 * pthread rwlocks
 * =============================================================
 */

static inline int
adk_sys_pthread_rwlock_free(
    adk_sys_lock_t     *rwlock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_PTHREAD_LOCK(rwlock);

    ADK_NVM_FREE(rwlock);

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_rwlock_init(
    adk_sys_lock_t         *rwlock,
    adk_sys_lock_attr_t    *attr,
    bool                    shared)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             rwlockRet;
    int             pref;

    VALIDATE_PTHREAD_LOCK(rwlock);

    rwlockRet = pthread_rwlockattr_init((pthread_rwlockattr_t *)attr);
    if (rwlockRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to init rwlock attr: '0x%x'", rwlockRet);
        goto _exit;
    }

    rwlockRet = pthread_rwlockattr_getkind_np((pthread_rwlockattr_t *)attr,
                                              &pref);
    if (rwlockRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to read rwlock pref: '0x%x'", rwlockRet);
        goto _exit;
    }

#if 0
    printf(" RW LOCK preference = %d \n", pref);
int             rwlockRet;

    pref = PTHREAD_RWLOCK_PREFER_WRITER_NONRECURSIVE_NP ;

    rwlockRet = pthread_rwlockattr_setkind_np((pthread_rwlockattr_t *)attr),
                                              pref);
    if (rwlockRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Faild to set rwlock pref: '0x%x'", rwlockRet);
        goto _exit;
    }
#endif

    if (shared == true)
    {
        rwlockRet = pthread_rwlockattr_setpshared((pthread_rwlockattr_t *)attr,
                                                  PTHREAD_PROCESS_SHARED);
        if (rwlockRet != 0)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
            ADK_FORCE_LOG_ERROR("Failed to set rwlock share: '0x%x'",
                                rwlockRet);
            goto _exit;
        }
    }

    rwlockRet = pthread_rwlock_init((pthread_rwlock_t *)rwlock,
                                    (pthread_rwlockattr_t *)attr);
    if (rwlockRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to init rwlock: '0x%x'", rwlockRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_rwlock_destroy(
    adk_sys_lock_t     *rwlock)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             rwlockRet;

    VALIDATE_PTHREAD_LOCK(rwlock);

    rwlockRet = pthread_rwlock_destroy((pthread_rwlock_t *)rwlock);
    if (rwlockRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to destroy rwlock: '0x%x'", rwlockRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_rwlock_lock(
    adk_sys_lock_t             *rwlock,
    adk_sys_rwlock_rw_type_t    rwFlag)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             rwlockRet = 0;

    VALIDATE_PTHREAD_LOCK(rwlock);

    if (rwFlag == ADK_SYS_RWLOCK_READ)
    {
        rwlockRet = pthread_rwlock_rdlock((pthread_rwlock_t *)rwlock);
    } else if (rwFlag == ADK_SYS_RWLOCK_WRITE)
    {
        rwlockRet = pthread_rwlock_wrlock((pthread_rwlock_t *)rwlock);
    } else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Invalid rwlock type: '0x%x'", rwFlag);
        goto _exit;
    }

    if (rwlockRet == EBUSY)
    {
        retValue = ADK_I_LOCK_BUSY;
        goto _exit;
    } else if (rwlockRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to rwlock rwlock: '0x%x'", rwlockRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_rwlock_unlock(
    adk_sys_lock_t     *rwlock)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             rwlockRet;

    VALIDATE_PTHREAD_LOCK(rwlock);

    rwlockRet = pthread_rwlock_unlock((pthread_rwlock_t *)rwlock);
    if (rwlockRet != 0)
    {
        ADK_STATIC_ANALYSIS_KILLPATH(retValue);
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_FATAL_ERROR("Failed to unlock rwlock: '0x%x'", rwlockRet);
        goto _exit;
    }

_exit:
    return retValue;
}

/* Pthread based Read-Write Locks related Macros */

#define ADK_ALLOC_PTHREAD_RW_LOCK() \
    ADK_NVM_MALLOC(sizeof(pthread_rwlock_t))

#define ADK_FREE_PTHREAD_RW_LOCK(rwlock) \
    adk_sys_pthread_rwlock_free((rwlock))

#define ADK_INIT_PTHREAD_RW_LOCK(rwlock, attr, shared) \
    adk_sys_pthread_rwlock_init((rwlock), (attr), (shared))

#define ADK_DESTROY_PTHREAD_RW_LOCK(rwlock) \
    adk_sys_pthread_rwlock_destroy((rwlock))

#define ADK_ACQUIRE_PTHREAD_RW_R_LOCK(rwlock) \
    adk_sys_pthread_rwlock_lock((rwlock), ADK_SYS_RWLOCK_READ)

#define ADK_ACQUIRE_PTHREAD_RW_W_LOCK(rwlock) \
    adk_sys_pthread_rwlock_lock((rwlock), ADK_SYS_RWLOCK_WRITE)

#define ADK_RELEASE_PTHREAD_RW_LOCK(rwlock) \
    adk_sys_pthread_rwlock_unlock((rwlock))


/*!
 * =============================================================
 * pthread spin locks
 * =============================================================
 */

static inline int
adk_sys_pthread_spin_free(
    adk_sys_lock_t     *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;

    VALIDATE_PTHREAD_LOCK(lock);

    ADK_NVM_FREE(lock);

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_spin_init(
    adk_sys_lock_t    *lock,
    bool               shared)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             spinRet;
    int             pshared = PTHREAD_PROCESS_PRIVATE;

    VALIDATE_PTHREAD_LOCK(lock);

    if (shared == true)
    {
        pshared = PTHREAD_PROCESS_SHARED;
    }

    spinRet = pthread_spin_init((pthread_spinlock_t *)lock, pshared);
    if (spinRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to init lock: '0x%x'", spinRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_spin_destroy(
    adk_sys_lock_t    *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             spinRet;

    VALIDATE_PTHREAD_LOCK(lock);

    spinRet = pthread_spin_destroy((pthread_spinlock_t *)lock);
    if (spinRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to destroy lock: '0x%x'", spinRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_spin_lock(
    adk_sys_lock_t     *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             spinRet = 0;

    VALIDATE_PTHREAD_LOCK(lock);

    spinRet = pthread_spin_lock((pthread_spinlock_t *)lock);
    if (spinRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to acquire lock: '0x%x'", spinRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_spin_trylock(
    adk_sys_lock_t    *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             spinRet = 0;

    VALIDATE_PTHREAD_LOCK(lock);

    spinRet = pthread_spin_trylock((pthread_spinlock_t *)lock);
    if (spinRet == EBUSY)
    {
        retValue = ADK_I_LOCK_BUSY;
        goto _exit;
    } else if (spinRet != 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_ERROR("Failed to acquire try lock: '0x%x'", spinRet);
        goto _exit;
    }

_exit:
    return retValue;
}

static inline int
adk_sys_pthread_spin_unlock(
    adk_sys_lock_t    *lock)
{
    adk_status_t    retValue = ADK_SUCCESS;
    int             spinRet = 0;

    VALIDATE_PTHREAD_LOCK(lock);

    spinRet = pthread_spin_unlock((pthread_spinlock_t *)lock);
    if (spinRet != 0)
    {
        ADK_STATIC_ANALYSIS_KILLPATH(retValue);
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE);
        ADK_FORCE_LOG_FATAL_ERROR("Failed to release lock: '0x%x'", spinRet);
        goto _exit;
    }

_exit:
    return retValue;
}



/* Pthread based Spin Locks related Macros */
#define ADK_ALLOC_PTHREAD_SPIN_LOCK() \
    ((pthread_spinlock_t *)(ADK_NVM_MALLOC(sizeof(pthread_spinlock_t))))

#define ADK_FREE_PTHREAD_SPIN_LOCK(lock) \
    adk_sys_pthread_spin_free((lock))

#define ADK_INIT_PTHREAD_SPIN_LOCK(lock, shared) \
    adk_sys_pthread_spin_init((lock), (shared))

#define ADK_DESTROY_PTHREAD_SPIN_LOCK(lock) \
    adk_sys_pthread_spin_destroy((lock))

#define ADK_ACQUIRE_PTHREAD_SPIN_LOCK(lock) \
    adk_sys_pthread_spin_lock((lock))

#define ADK_TRY_ACQUIRE_PTHREAD_SPIN_LOCK(lock) \
    adk_sys_pthread_spin_trylock((lock))

#define ADK_RELEASE_PTHREAD_SPIN_LOCK(lock) \
    adk_sys_pthread_spin_unlock((lock))


#endif /* __ADK_OS_PTHREAD_LOCKS_H__ */

/*! end addtogroup _sif
 * @}
 */
