/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_sif.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief
 *
 * This file contains definitions and declarations of objects that are
 * part of system infrastructure.
 *
 *
 ***********************************************************************/

#ifndef __ADK_SIF_H__
#define __ADK_SIF_H__

#include <adk_sys.h>
#include "adk_sys_lcore.h"
#include "adk_sif_constants.h"
#include "adk_sif_dp.h"
#include "adk_dev_net_internal.h"
/* #include <adk_include.h> */
#include <adk_rcu.h>
#include <adk_rm.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <adk_misc.h>

#define adk_htonl htonl
#define adk_htons htons
#define adk_ntohl ntohl
#define adk_ntohs ntohs


/*! defines the max size of the error string. This is the size
 *  of the array that the module manager is going to pass to the
 *  modules when calling the module specific error_handler.
 *  */
#define ADK_ERROR_STRING_MAX_SIZE 100
#define ADK_QOS_NUM_INSTANCES  ADK_DEV_NET_ID_MAX

/*!
 * Name of shared memory used in IPC between primary & secondary process
 */
#define ADK_SYS_SHM_MEMZONE_NAME "ADK_SYS_SHM_MEMZONE"
#define ADK_SIF_DP_SHM_MEMZONE_NAME "ADK_SIF_DP_SHM_MEMZONE"
#define ADK_SYS_EGRESS_STATS_SHM_MEMZONE_NAME "ADK_SYS_TX_STATS_SHM_MEMZONE"

/*!
 * This macro will set the severity level for the error and warning
 * error code. The user has to call the macro with the error code.
 * The macro will insert the appropriate severity level at the right
 * locaton.
 */
#define ADK_FATAL_SEV_LEVEL_SET(errorValue) \
    ((ADK_STATUS_SEV_LEVEL_MASK(errorValue)) | (ADK_SEV_LEVEL_FATAL << 16))
#define ADK_ERROR_SEV_LEVEL_SET(errorValue) \
    ((ADK_STATUS_SEV_LEVEL_MASK(errorValue)) | (ADK_SEV_LEVEL_ERROR << 16))
#define ADK_WARNING_SEV_LEVEL_SET(errorValue) \
    ((ADK_STATUS_SEV_LEVEL_MASK(errorValue)) | (ADK_SEV_LEVEL_WARNING << 16))

#define ADK_DP_FATAL_SEV_LEVEL_SET(errorValue) \
    ((ADK_STATUS_SEV_LEVEL_MASK(errorValue)) | \
    (ADK_SEV_LEVEL_FATAL << 16) | \
     (ADK_FROM_DP << 16))
#define ADK_DP_ERROR_SEV_LEVEL_SET(errorValue) \
    ((ADK_STATUS_SEV_LEVEL_MASK(errorValue)) | \
    (ADK_SEV_LEVEL_ERROR << 16) | \
     (ADK_FROM_DP << 16))
#define ADK_DP_WARNING_SEV_LEVEL_SET(errorValue) \
    ((ADK_STATUS_SEV_LEVEL_MASK(errorValue)) | \
    (ADK_SEV_LEVEL_WARNING << 16) | \
     (ADK_FROM_DP << 16))

/*!
 * This macro will allow the user to pass both error code and
 * severity level. It is same as previous macros except that the user
 * has the flexibilty to select severity level.
 */
#define ADK_STATUS_SEV_LEVEL_SET(errorValue, sev_level) \
    ((errorValue) | ((sev_level) << 16))

/*!
 * This macro will allow the user to generate the error code
 * by specifying module ID, sub component ID, severiyt level
 * and the actual error value.
 */
#define ADK_ERROR_CODE_GENERATE(mod_id, sub_cm, sev_level, code) \
    (((mod_id) << 24) | \
     ((sub_cm) << 20) | \
     ((sev_level) << 16) | \
     ((code) & 0x0000ffff))

const char *adk_sys_error_string(adk_sif_status_t errorCode);

/*!
 * =============================================================
 * @struct adk_flow_info_t adk_sys.h "adk_sys.h"
 *
 * @brief
 *
 * The data structure defines the key common parameters of
 * flow data structure.
 *
 * =============================================================
 */
typedef struct adk_flow_info_s {
    /* Cookie to correlate pointers */
    uint32_t cookie;
    /* flowId is a 24-bit field to uniquelt identify a flow
     * within ADK.
     */
    uint32_t flowId;
    /* moduleHandle identifies the owner module of the flow handle.
     */
    adk_module_handle_t moduleHandle;
    /* flowAttr indicate the first byte of the flow specific information.
     * The module is expected to allocate memory for sizeof(adk_flow_info_t)
     * + sizeof (module_specific_flow_attr) - sizeof(uintptr_t). The flowAttr
     * then gives the pointer to module_specific_flow_attr structure. Having
     * this model allows us to reduce the c(m)alloc calls from 2 to 1.
     */
    uintptr_t flowAttr;
} adk_flow_info_t;   /* ------  end of struct adk_flow_info_t  ------ */

#define ADK_FLOW_COOKIE 0x43215678

/*
 * Generic flow handle access macros
 */
/* ADK_FLOW_HANDLE_ALLOC macro takes module handle, module specific
 * flow information size and a pointer to store the allocated
 * flow handle.
 */

#define ADK_SYS_FLOW_HANDLE_ALLOC(modHandle, handleID, size, flowHandle_p) \
    do { \
        adk_flow_info_t *flowInfo_p; \
	unsigned int mod_id = ADK_MODULE_MODULEID_GET(modHandle); \
        flowInfo_p = (adk_flow_info_t *) \
            ADK_NVM_CALLOC(mod_id, handleID, 1, sizeof(adk_flow_info_t) + \
                                       (size) - sizeof(uintptr_t)); \
        if (flowInfo_p) { \
            flowInfo_p->moduleHandle = (modHandle); \
            flowInfo_p->cookie = ADK_FLOW_COOKIE; \
        } \
        *(flowHandle_p) = (adk_flow_handle_t) flowInfo_p; \
    } while (0)

#define ADK_SYS_FLOW_HANDLE_FREE(flowHandle) \
    do { \
        adk_flow_info_t *flowInfo_p = (adk_flow_info_t *) (flowHandle); \
        flowInfo_p->cookie = 0; \
        flowInfo_p->moduleHandle = 0; \
        ADK_NVM_FREE(flowInfo_p); \
    } while (0)

#define ADK_SYS_VALID_FLOW_HANDLE(flowHandle) \
    ((flowHandle) && \
     (((adk_flow_info_t *)(flowHandle))->cookie == ADK_FLOW_COOKIE))

#define ADK_SYS_FLOWID_GET(flowHandle) \
    (((adk_flow_info_t *)(flowHandle))->flowId)

#define ADK_SYS_FLOW_ATTR_PTR_GET(flowHandle) \
    (&((adk_flow_info_t *)(flowHandle))->flowAttr)

#define ADK_SYS_FLOW_MODULE_GET(flowHandle) \
    (((adk_flow_info_t *)(flowHandle))->moduleHandle)

#define ADK_SYS_FLOWID_SET(flowHandle, sysFlowId) \
    do { \
        adk_flow_info_t *flowInfo_p = (adk_flow_info_t *) (flowHandle); \
        if(flowInfo_p) { \
            flowInfo_p->flowId = (sysFlowId); \
        } \
    } while (0)

/*!
 * =============================================================
 * @struct adk_interface_info_t adk_sys.h "adk_sys.h"
 *
 * The data structure defines the key common parameters of
 * interface data structure.
 *
 * =============================================================
 */
typedef struct adk_interface_info_s {
    /*!
     * cookie to validate the interface handle */
    uint32_t cookie;

    /*!
     * interface ID is a 16-bit field to uniquely identify a flow
     * within ADK. keeping it as 32-bit in data structure for
     * future expansion.
     */
    uint32_t interfaceId;

    /*!
     * moduleHandle identifies the owner module of the interface handle.
     */
    adk_module_handle_t moduleHandle;

    /*!
     * qosNodeId field indicate the scheduler node ID in the MTM for
     * this interface. This value will be set to zero, if shaping
     * configuration for interface is not set.
     */
     uint32_t qosNodeId;

    /*!
     * The Virtual Routing Domain ID for the interface. This is set and read by
     * different modules and holding the value here made sense to keep modules
     * from accessing each other's data structures.
     */
    uint32_t    vrdId;

    /*!
     * Reserved data for later expansion.
     */
    uint32_t    reserved1;

    /*!
     * Reserved data for later expansion.
     */
    uint32_t    reserved2;

    /*!
     * ifAttr indicate the first byte of the interface specific information.
     * The module is expected to allocate memory for
     * sizeof(adk_interface_info_t) + sizeof(module_specific_interface_attr)
     *     - sizeof(uintptr_t).
     * The ifAttr then gives the pointer to module_specific_flow_attr
     * structure. Having this model allows us to reduce the (c)malloc calls
     * from 2 to 1.
     */
    uintptr_t ifAttr;
} adk_interface_info_t;   /* ------  end of struct adk_interface_info_t  ------ */

#define ADK_INTERFACE_COOKIE 0x43217856


/* ADK_INTERFACE_HANDLE_ALLOC macro takes module handle, module specific
 * interface information size and a pointer to store the allocated
 * interface handle.
 */
#define ADK_SYS_INTERFACE_HANDLE_ALLOC(modHandle, handleID, size, ifHandle_p) \
    do { \
        adk_interface_info_t *ifInfo_p; \
	unsigned int mod_id = ADK_MODULE_MODULEID_GET(modHandle); \
        ifInfo_p = (adk_interface_info_t *) \
                        ADK_NVM_CALLOC(mod_id, handleID, 1, (sizeof(adk_interface_info_t) + \
                                           (size) - sizeof(uintptr_t))); \
        if (ifInfo_p) { \
            ifInfo_p->moduleHandle = (modHandle); \
            ifInfo_p->cookie = ADK_INTERFACE_COOKIE; \
        } \
        *(ifHandle_p) = (adk_interface_handle_t) ifInfo_p; \
    } while (0)

#define ADK_SYS_INTERFACE_HANDLE_FREE(ifHandle) \
    do { \
        adk_interface_info_t *ifInfo_p; \
        ifInfo_p = (adk_interface_info_t *) (ifHandle); \
        ifInfo_p->moduleHandle = 0; \
        ifInfo_p->cookie = 0; \
        ADK_NVM_FREE((ifHandle)); \
    } while (0)

#define ADK_SYS_VALID_INTERFACE_HANDLE(ifHandle) \
    ((ifHandle) && ((adk_interface_info_t *)(ifHandle))->cookie == \
                                                    ADK_INTERFACE_COOKIE)

#define ADK_SYS_INTERFACE_ID_GET(ifHandle) \
    ((ifHandle) ? (((adk_interface_info_t *)(ifHandle))->interfaceId) : 0)

#define ADK_SYS_INTERFACE_ATTR_PTR_GET(ifHandle) \
    (&((adk_interface_info_t *)(ifHandle))->ifAttr)

#define ADK_SYS_INTERFACE_MODULE_GET(ifHandle) \
    (((adk_interface_info_t *)(ifHandle))->moduleHandle)

#define ADK_SYS_INTERFACE_ID_SET(ifHandle, ifId) \
    (((adk_interface_info_t *)(ifHandle))->interfaceId = ifId)

#define ADK_SYS_INTERFACE_QOS_NODEID_GET(ifHandle) \
    ((ifHandle) ? (((adk_interface_info_t *)(ifHandle))->qosNodeId) : 0)

#define ADK_SYS_INTERFACE_QOS_NODEID_SET(ifHandle, sched_id) \
    (((adk_interface_info_t *)(ifHandle))->qosNodeId = sched_id)

#define ADK_SYS_VRD_ID_SET(ifHandle, vrd_id) \
    (((adk_interface_info_t *)(ifHandle))->vrdId = vrd_id)

#define ADK_SYS_VRD_ID_GET(ifHandle) \
    ((ifHandle) ? (((adk_interface_info_t *)(ifHandle))->vrdId) : 0)

/*! all hardware engine handles */
#if 0
typedef struct adk_sys_hw_handle_s
{
    ncp_hdl_t ncpHandle;
} adk_sys_hw_handle_t;
#endif

/*! handle for all software sub components of SIF */
typedef struct adk_utils_handle_s
{
    adk_handle_t moduleMgrMasterHandle;
    adk_handle_t phHandle;
    adk_handle_t reserve1;
    adk_handle_t reserve2;
} adk_utils_handle_t;

/*!
 * =============================================================
 * @internal
 * @struct adk_sys_config_worker_lcore_info_t  ""
 *
 * @brief  Worker lcore info
 *
 * This is the structure returned by adk_sys_config_worker_lcore_info_get().
 * The worker core info is either the master mode EAL info or the slave mode
 * config data.
 *
 * =============================================================
 */
typedef struct adk_sys_config_worker_lcore_info_s {
    /*!
     * To indicate ADK configuration mode standalone or slave. IP and sys DP
     * need this info.
     */
    adk_sys_config_mode_t   adkMode;
    uint32_t                numLcores;
    struct adk_lcore_s {
        uint32_t    id;
        uint32_t    index;
        int32_t     socket;
    } lcore[ADK_MAX_NUM_CORES];
} adk_sys_config_worker_lcore_info_t;

/*!
 * =============================================================i
 * @internal
 * @struct adk_sys_info_t adk_sys.h "adk_sys.h"
 *
 * @brief
 *
 * aggregate handles and global states that are shared by all
 * components of SIF.
 *
 * =============================================================
 */
typedef struct adk_sys_info_s
{
    uint32_t            cookie;
    adk_utils_handle_t  utils;
    adk_handle_t        bindMasterHandle;
    adk_handle_t        reserve1;
    adk_handle_t        reserve2;
    bool                multiThread;
    void *sysLock; /* Can point to mutex, semaphore, spin-lock, etc */
    /*! debug_level for all sif sub components except PH and IF */
    uint32_t        reserve5;
    adk_module_handle_t sifPseudoHandle;
    adk_handle_t        qosHandle;

    adk_handle_t        reserve3;
    adk_handle_t        reserve4;
    /*!
     * Chip instance for which this ADK is initialized.
     */
    uint32_t        chipInstance;
    /*! Cfg handle */
    adk_handle_t        cfgHandle;
    /*! RCU state */
    adk_rcu_state_t     *rcuState_p;

    /*! Packet capture device ID */
    uint16_t pcap_ethDevId;
    /*! Packet Capture ID resource handle */
    adk_rm_hdl_t pcap_id_rm_hdl;

    /*! Egress stats*/
    adk_rm_hdl_t egress_stats_rm_hdl;
    adk_sys_dp_egress_stats_t *egress_stats_p;

    /*! slow path device ids*/
    adk_rm_hdl_t sp_dev_ids_rm_hdl;

} adk_sys_info_t;

#ifdef ADK_SYS_OVERLOAD_PROTECTION
/*!
 * =============================================================
 * @internal
 * @struct adk_sys_ovp_info_t adk_sif.h "adk_sif.h"
 *
 * @brief  ADk system overload protection information 
 *
 * =============================================================
 */
typedef struct adk_sys_ovp_info_s {
    /*!
     * Operating mode derived from below fileds
     * {system_wide_policy_flag, prio_type, dscp_mask} 
     */
    uint8_t   mode;
    /*!
     * flag to indicate system-wide policy vs per eth-dev policy
     * Allowed only fro XEON products as of now
     */
    uint8_t     system_wide_policy_flag;
    /*!
     * packet priority type DSCP Vs VLAN PCP
     */
    adk_sys_pkt_priority_type_t  prio_type;

    /*!
     * what is the default drop precedence that is applied 
     * for non-IP packets incase of type id IP_DSCP and un-tagged packets in
     * case of type is VLAN_PCP.
     */
    adk_sys_drop_precedence_t   default_drop_precedence;
    /*!
     * default drop action, TRUE: drop FALSE: enqueue 
     */
    bool    default_action_drop;

    /*!
     * this filed is valid only if the type is ADK_SYS_PRIO_TYPE_IP_DSCP.
     * dscp_mask indicates whether the priority->drop mapping is provided
     * for all the DSCP values or sub-set of DSCP values. 
     */
    uint8_t                     dscp_mask;

    /*!
     *  priority to drop-precedence mapping system global
     */
    adk_sys_pkt_prio_drop_precedence_map_t prio_to_drop_map_g[64];

    /*!
     *
     */
    uint64_t    marked_dscp_pkts;
    uint64_t    marked_non_dscp_pkts;
    uint64_t    non_marked_pkts;

    /*!
     *  priority to drop-precedence mapping list per eth-dev
     */
    adk_sys_pkt_prio_drop_precedence_map_t prio_to_drop_map_list[ADK_MAX_NUM_ETH_DEVICES][64];

} __attribute__ ((aligned(64))) adk_sys_ovp_info_t;

#endif

/*!
 * =============================================================
 * @struct adk_sys_config_info_t adk_sif.h "adk_sif.h"
 *
 * @brief Structure for ADK in slave mode.
 *
 * =============================================================
 */
typedef struct adk_sys_config_info_s {
    /*!
     * To indicate ADK configuration mode standalone or slave.
     */
    adk_sys_config_mode_t    adk_mode;
    /*!
     * Pointer to pktio memorypool to be used in packet generation
     * in ADK like BFD.
     */
    void          *pktio_mempool;
    /*!
     * Pointer to memorypool to be used in IP fragmentation.
     */
    void          *ip_frag_mempool;
    /*!
     * Worker lcore informat
     */
    adk_sys_config_worker_lcore_info_t lcoreInfo;

#ifdef ADK_SYS_OVERLOAD_PROTECTION
    /*!
     * System overload protection table
     */
    adk_sys_ovp_info_t  ovpInfo;
#endif

    /*!
     * resource database
     */
    adk_sys_resource_info_t  res_list[ADK_SYS_RSRC_TYPE_MAX][255];
    uint16_t                 num_res[ADK_SYS_RSRC_TYPE_MAX];

} adk_sys_config_info_t;

/*!
 * =============================================================
 * @struct adk_sys_shm_config_t adk_sys.h "adk_sys.h"
 *
 * @brief Structure to pass information from primary to secondary
 * process .
 *
 * =============================================================
 */
typedef struct adk_sys_shm_config_s {
    /*!
     * ADK system handle is stored in primary process and read in
     * secondary process.
     */
    adk_sys_handle_t    sysHandle;
    /*! This handle is required to choose between slave and master modes and
     * also to get the system configuration details from a different process.
     */
    adk_config_handle_t sysConfHandle;
    /*!
     * Pointer to DP's shared memory
     */
    adk_sys_handle_t    sifDpShmHandle;
} adk_sys_shm_config_t __rte_cache_aligned;

#define SYS_HANDLE_COOKIE  0x27443648

/* Is system handle invalid */
#define ADK_INVALID_SYS_HANDLE(sys) \
    (((sys) == 0) || (((adk_sys_info_t *)(sys))->cookie != SYS_HANDLE_COOKIE))

/* useful util macros that are tied to the sys_info data structure */
#define ADK_SIF_DP_INFO_PTR_GET \
    ((adk_sif_dp_shm_info_t *)(adkSysShmConf_p->sifDpShmHandle))

#define ADK_SYS_MODULE_HANDLE_GET(sys) \
    (((adk_sys_info_t *)(sys))->sifPseudoHandle)

#define ADK_SYS_LOCK_ENABLE(sys) \
    (((adk_sys_info_t *)(sys))->multiThread)

#define ADK_SYS_LOCK_GET(sys) \
    (((adk_sys_info_t *)(sys))->sysLock)

#define ADK_SYS_LOG_LEVEL_GET \
    (ADK_SIF_DP_INFO_PTR_GET->sysLogLevel)

#define ADK_SYS_LOG_LEVEL_SET(newlogLevel) \
    (ADK_SIF_DP_INFO_PTR_GET->sysLogLevel = newlogLevel)

#define ADK_SYS_UDP_VALIDATE_FLAG_SET(flag) \
    (ADK_SIF_DP_INFO_PTR_GET->udpValidateFlag = flag)

#define ADK_SYS_UDP_GEN_CONFIG_SET(flag) \
    (ADK_SIF_DP_INFO_PTR_GET->udpComputeFlag = flag)

#define ADK_SYS_IP_IPSEC_BYPASS_CONFIG_SET(bypass) \
    (ADK_SIF_DP_INFO_PTR_GET->ipsecBypass = bypass)

#define ADK_SYS_HDL_COOKIE_VALIDATE(sys) \
    (((adk_sys_info_t *)(sys))->cookie == SYS_HANDLE_COOKIE)

#define ADK_SYS_HDL_COOKIE_SET(sys) \
    (((adk_sys_info_t *)(sys))->cookie = SYS_HANDLE_COOKIE)

#define ADK_SYS_HDL_COOKIE_CLEAR(sys) \
    (((adk_sys_info_t *)(sys))->cookie = 0)

#define ADK_SYS_INTERFACE_RSRC_HDL_SET(sys, rsrcHdl) \
    (((adk_sys_info_t *) (sys))->interfaceRsrcHdl = rsrcHdl)

#define ADK_SYS_INTERFACE_RSRC_HDL_GET(sys) \
    (((adk_sys_info_t *)(sys))->interfaceRsrcHdl)

#define ADK_ACQUIRE_SYS_LOCK(sys, _retValue) \
    do { \
        if (ADK_SYS_LOCK_GET((sys)) != NULL) \
        { \
            ADK_ACQUIRE_LOCK(ADK_SYS_LOCK_GET((sys)), (_retValue)); \
        } else \
        { \
            if (ADK_SYS_LOCK_ENABLE((sys)) == true) \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Sys lock acquire failed: " \
                                    "invalid lock handle\n"); \
                goto _lock_acquire_failed; \
            } \
        } \
    } while (0)

/*
 * For RELEASE, we're going to make the assumption that the lock is defined.
 * This avoids the call to ADK_MODULE_LOCK_GET(), thus avoiding a complaint by
 * the static analysis tools about the lock not being unlocked.
 */
#define ADK_RELEASE_SYS_LOCK(sys, _retValue) \
    do { \
        ADK_RELEASE_LOCK(ADK_SYS_LOCK_GET((sys)), (_retValue)); \
    } while (0)

#define ADK_TRY_ACQUIRE_SYS_LOCK(sys, _retValue) \
    do { \
        if (ADK_SYS_LOCK_GET((sys)) != NULL) \
        { \
            ADK_TRY_ACQUIRE_LOCK(ADK_SYS_LOCK_GET((sys)), (_retValue)); \
        } else \
        { \
            if (ADK_SYS_LOCK_ENABLE((sys)) == true) \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Sys lock try acquire failed: " \
                                    "invalid lock handle\n"); \
                goto _lock_acquire_failed; \
            } \
        } \
    } while (0)

#define ADK_ACQUIRE_MODULE_LOCK(mod_hdl, _retValue) \
    do { \
        if (ADK_MODULE_LOCK_GET((mod_hdl)) != NULL) \
        { \
            ADK_ACQUIRE_LOCK(ADK_MODULE_LOCK_GET((mod_hdl)), (_retValue)); \
        } else \
        { \
            if (ADK_MODULE_LOCK_ENABLE((mod_hdl)) == true) \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Module acquire lock failed" \
                                    "invalid lock handle\n"); \
                goto _lock_acquire_failed; \
            } \
        } \
    } while (0)

/*
 * For RELEASE, we're going to make the assumption that the lock is defined.
 * This avoids the call to ADK_MODULE_LOCK_GET(), thus avoiding a complaint by
 * the static analysis tools about the lock not being unlocked.
 */
#define ADK_RELEASE_MODULE_LOCK(mod_hdl, _retValue) \
    do { \
        ADK_RELEASE_LOCK(ADK_MODULE_LOCK_GET((mod_hdl)), (_retValue)); \
    } while (0)

#define ADK_TRY_ACQUIRE_MODULE_LOCK(mod_hdl, _retValue) \
    do { \
        if (ADK_MODULE_LOCK_GET((mod_hdl)) != NULL) \
        { \
            ADK_TRY_ACQUIRE_LOCK(ADK_MODULE_LOCK_GET((mod_hdl)), (_retValue)); \
        } else \
        { \
            if (ADK_MODULE_LOCK_ENABLE((mod_hdl)) == true) \
            { \
                (_retValue) = ADK_ERROR_SEV_LEVEL_SET(ADK_E_LOCK_FAILURE); \
                ADK_FORCE_LOG_ERROR("Module try acquire lock failed" \
                                    "invalid lock handle\n"); \
                goto _lock_acquire_failed; \
            } \
        } \
    } while (0)

/*!
 * allocate interface Id from the common resource pool.
 * System handle and the pointer to interface ID location
 * are the input paramters. MACRO returns ADK_SUCCESS,
 * ADK_E_INVALID_PARAMETER or ADK_E_RESOURCE_UNAVAILABLE
 * as return value. Allocated interface Id is stored under
 * ifId_p.
 */

#define ADK_SYS_INTERFACE_ID_ALLOC(sys, ifId_p) \
    adk_sys_interface_id_alloc((sys), (ifId_p))

/*!
 * Free interface Id to the common resource pool.
 * System handle and interface Id are the input
 * parameter. MACRO return ADK_SUCCESS, ADK_E_INCONSISTENT_STATE
 * or ADK_E_INVALID_PARAMETER as return value.
 */

#define ADK_SYS_INTERFACE_ID_FREE(sys, ifId) \
    adk_sys_interface_id_free((sys), (ifId))

/*
 * commonly used macros.
 */
#define ADK_RETURN(x) \
    return(x)
#define ADK_VALID_HANDLE(adk_handle) \
    ((adk_handle) != 0)
#define ADK_VALID_PTR(adk_ptr) \
    (((adk_ptr)) != NULL)
#define ADK_INVALID_PTR(adk_ptr) \
    (((adk_ptr)) == NULL)
#define ADK_SYS_NON_ZERO_CNTR_WRITE(fmt, cntr) \
    do { \
        if ((cntr) != 0) { \
            ADK_INFO_WRITE((fmt), (cntr)) \
        } \
    } while (0)



/*!
 * =============================================================
 * @internal
 * @struct adk_sif_netDriver_demux_t adk_sif.h "adk_sif.h"
 *
 * @brief Parameters for updating sif_netDriver_demux_tree()
 *
 *
 * =============================================================
 */
typedef struct adk_sif_netDriver_demux_s {

    /*!
     * The output port either EIOA or IPv4/IPv6 logical port. Lower 10
     * bits are used.
     */
    uint16_t    outputPort;

    /*!
     * Used to set the PDU type, which is currently only used by IPv4
     * and IPv6. The lower 4 bytes are used.
     */
    uint8_t     pduType;

    /*!
     * If the PDU type has a valid value, this needs to be true
     */
    bool      pduTypeEnabled;
} adk_sif_netDriver_demux_t;   /* ------  end of struct adk_sif_netDriver_demux_t  ------ */

extern adk_sys_shm_config_t *adkSysShmConf_p;

adk_status_t adk_sys_interface_id_free (
        adk_sys_handle_t sysHandle,
        uint32_t     interfaceId);
adk_status_t adk_sys_interface_id_alloc (
        adk_sys_handle_t sysHandle,
        uint32_t     *interfaceId_p);
adk_status_t adk_sif_pcap_vsi_id_get (
        adk_sys_handle_t sysHandle,
        uint32_t *vsi_id_p);
adk_status_t adk_sif_pcap_id_alloc (
        adk_sys_handle_t sysHandle,
        uint32_t *pcapId);
adk_status_t adk_sif_pcap_id_free (
        adk_sys_handle_t sysHandle,
        uint32_t pcapId);
adk_status_t adk_sif_pcap_id_config (
        adk_sys_handle_t sysHandle,
        uint32_t pcap_id,
        uint32_t pcap_entry_fn_id);
adk_status_t adk_sif_netDriver_demux_add (
        adk_sys_handle_t            sysHandle,
        adk_sif_netDriver_demux_t  *netDriverDemuxData_p);
adk_status_t adk_sif_netDriver_demux_remove (
        adk_sys_handle_t            sysHandle,
        adk_sif_netDriver_demux_t  *netDriverDemuxData_p);
adk_status_t adk_sif_interface_to_outPort_mapping_add (
        adk_sys_handle_t    sysHandle,
        uint32_t        interfaceId,
        uint32_t        outputId,
        uint32_t        virtualHandleValue);
adk_status_t adk_sif_interface_to_outPort_mapping_remove (
        adk_sys_handle_t    sysHandle,
        uint32_t        interfaceId);
#if 0
adk_status_t adk_sif_interface_to_outPort_mapping_get (
        adk_sys_handle_t    sysHandle,
        uint32_t        interfaceId,
        uint32_t       *outputId_p);
#endif

adk_status_t
adk_sif_timestamp_offset_set (
        adk_sys_handle_t        sysHandle,
        const uint64_t      timestamp_offset );

adk_status_t
adk_sif_timestamp_offset_get (
        adk_sys_handle_t    sysHandle,
        uint64_t       *timestamp_offset_p );

extern adk_status_t adk_sys_config_mode_get (
        adk_sys_config_mode_t *adk_mode);

extern adk_status_t adk_sys_config_worker_lcore_info_get (
        adk_sys_config_worker_lcore_info_t **workerInfo_p);

adk_status_t adk_sys_global_stats_print (
        adk_sys_handle_t sysHandle);

adk_status_t adk_sys_overload_drop_stats_print (
        adk_sys_handle_t sysHandle);

adk_status_t adk_sys_global_debug_print (
        adk_sys_handle_t sysHandle);

#ifdef ADK_SYS_OVERLOAD_PROTECTION
adk_status_t adk_sys_overload_protection_stats_print (
        adk_sys_handle_t sysHandle);
#endif

adk_status_t adk_sif_module_egress_callback_register(
        adk_sys_handle_t sysHandle,
        uint8_t moduleId,
        adk_sif_module_egress_cb_fn_t moduleCbFn);

adk_status_t adk_sif_module_egress_callback_unregister(
        adk_sys_handle_t sysHandle,
        uint8_t moduleId);

adk_status_t adk_sif_module_slowpath_callback_register(
        adk_sys_handle_t sysHandle,
        uint8_t moduleId,
        adk_sif_module_slowpath_cb_fn_t moduleCbFn);

adk_status_t adk_sif_module_slowpath_callback_unregister(
        adk_sys_handle_t sysHandle,
        uint8_t moduleId);
#ifdef __cplusplus
}
#endif

#endif /* __ADK_SIF_H__ */

/*! end addtogroup _sif
 * @}
 */
