/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_sys_config.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief
 *
 * This file contains definitions and declarations of ADK system 
 * configuration interface definition (APIs and data structure definitions).
 *
 ***********************************************************************/

#ifndef __ADK_SYS_CONFIG_H__
#define __ADK_SYS_CONFIG_H__

#ifdef __cplusplus
extern "C" {
#endif

/*!
*/

/*! 
 *  @brief data structure to define resource types.
 */
typedef enum adk_sys_resource_type_e {
    ADK_SYS_RSRC_TYPE_EVNT_DEV,
    ADK_SYS_RSRC_TYPE_TIMER_DEV,
    ADK_SYS_RSRC_TYPE_ETH_DEV,
    ADK_SYS_RSRC_TYPE_PKT_MEMORY,
    ADK_SYS_RSRC_TYPE_WCORE,
    ADK_SYS_RSRC_TYPE_CRYPTO_DEV,
#if 0
    ADK_SYS_RSRC_TYPE_ADK_NETD,
    ADK_SYS_RSRC_TYPE_NAC,
    ADK_SYS_RSRC_TYPE_NIS_CPK,
    ADK_SYS_RSRC_TYPE_FPPS_HLP,
    ADK_SYS_RSRC_TYPE_CRYPTO_BD,
#endif
    ADK_SYS_RSRC_TYPE_MAX,
} adk_sys_resource_type_t;

/*!
 *  @brief data structure to define operation type
 *  
 */
typedef enum adk_sys_rsrc_oper_e {
    ADK_SYS_RSRC_ADD,
    ADK_SYS_RSRC_REM,
    ADK_SYS_RSRC_MODIFY,
} adk_sys_rsrc_oper_t;

/*! 
 *  @brief data structure to define sharing modes of any resource.
 */
typedef enum adk_sys_rsrc_mode_e{
    /*! mode to indicate a resource is shared between ADK software and
     * application software.
     */
    ADK_SYS_RSRC_MODE_SHARED,
    /*! mode to indicate a resource is dedicated to ADK software use only.
     */
    ADK_SYS_RSRC_MODE_DEDICATED,
} adk_sys_rsrc_mode_t;


/*!
* data structure to define crypto-dev type
*
*/
typedef enum adk_sys_rsrc_crypto_dev_type_e {
    ADK_SYS_RSRC_CRYPTO_DEV_TYPE_INLINE,
    ADK_SYS_RSRC_CRYPTO_DEV_TYPE_LA,
    ADK_SYS_RSRC_CRYPTO_DEV_TYPE_SOFT,
} adk_sys_rsrc_crypto_dev_type_t;

/*!
* @brief data structure to define Q-pair type
*        This is required only in case of non-inline
*        device types. For inline device, Q-pair
*        information is not applicable.
*
*/
typedef enum adk_sys_rsrc_crypto_dev_qp_use_e {
    /*! use it for IPSEC DL processing */
    ADK_SYS_RSRC_CRYPTO_DEV_QP_USE_IPSEC_DL,
    /*! use it for IPSEC UL processing */
    ADK_SYS_RSRC_CRYPTO_DEV_QP_USE_IPSEC_UL,
    /*! use it for IPSEC DL & UL processing */
    ADK_SYS_RSRC_CRYPTO_DEV_QP_USE_IPSEC_BOTH,
} adk_sys_rsrc_crypto_dev_qp_use_t;

/*!
* @brief data structure to define Q-pair information
*        This is required only in case of non-inline
*        device types. For inline device, Q-pair
*        information is not applicable.
*/
typedef struct adk_sys_rsrc_crypto_dev_qpair_info_s{
    /*! Queue pair id */
    uint8_t                         queue_pair_id;

    /*! mode to indicate whether this queue-pair is dedicated to ADK software 
     * use only or shared between application software and ADK software.
     */
    adk_sys_rsrc_mode_t  mode;

    /*! type indicates whether this queue-pair is to be used 
    * for IPSEC-DL only or IPSEC-UL only or IPSEC-DL&UL
    */
    adk_sys_rsrc_crypto_dev_qp_use_t usage;
    /*! list of traffic flow priorities to which this queue-pair to be used. 
    * maximum of up to 8 priorities supported.
    */
    int8_t                          pri_list[8];
    
    /*!
     * number of priorities that are valid in the pri_list. 
     * num_pri = 0 interpreted as this queue-pair is applicable for all 
     * priorities
     */
    uint8_t                         num_pri;

    /*!
     * Event-queue-id of the serving Crypto-BD.
     */
    uint8_t                         bd_ev_queue_id;

} adk_sys_rsrc_crypto_dev_qpair_info_t;

/*!
 * =============================================================
 *
 * @struct adk_sys_rsrc_crypto_dev_info_t adk_sys.h "adk_sys.h"
 *
 * @brief  
 * data structure to represent a Crypto device info. 
 * In case of "inline" device, only dev_type and dev_id
 * is required to be filled correctly, rest of the information
 * needs to be populated only for non-inline device types.
 *
 * =============================================================
 */
typedef struct adk_sys_rsrc_crypto_dev_info_s {
    /*! Type of Crypto device */
    adk_sys_rsrc_crypto_dev_type_t dev_type;
    /*! device id */
    uint8_t                dev_id;
    /*! Number of SA's that can be created on this QueuePair */
    uint16_t               num_sas; 
    /*! number of queue-pairs*/
    uint8_t                num_q_pairs;
    /*! List of queue pair info pointers. 
     */
    adk_sys_rsrc_crypto_dev_qpair_info_t   *qp_info_p;
    /*! reserved for ADK internal usage.
     */
    uint64_t               reserved;
} adk_sys_rsrc_crypto_dev_info_t;

/*!
 * =============================================================
 *
 * @struct adk_sys_rsrc_evp_info_t adk_sys.h "adk_sys.h"
 *
 * @brief  data structure to represent a worker to event-port-id mapping. 
 *
 *
 * =============================================================
 */
typedef struct adk_sys_rsrc_evdev_worker_info_s {
    /*! Worker lcore id */
    uint8_t    lcore_id;
    /*! event-dev-port associated with this lcore */
    uint8_t    ev_port_id;
} adk_sys_rsrc_evp_info_t;

/* Event queue configuration bitmap flags */
#define ADK_SYS_RSRC_EVQ_CFG_TYPE_MASK          (3ULL << 0)
/**< Mask for event queue schedule type configuration request */
#define ADK_SYS_RSRC_EVQ_CFG_ALL_TYPES          (0ULL << 0)
/**< Allow ATOMIC,ORDERED,PARALLEL schedule type enqueue
 */
#define ADK_SYS_RSRC_EVQ_CFG_ATOMIC_ONLY        (1ULL << 0)
/**< Allow only ATOMIC schedule type enqueue
 */
#define ADK_SYS_RSRC_EVQ_CFG_ORDERED_ONLY       (2ULL << 0)
/**< Allow only ORDERED schedule type enqueue
 */
#define ADK_SYS_RSRC_EVQ_CFG_PARALLEL_ONLY      (3ULL << 0)
/**< Allow only PARALLEL schedule type enqueue
 */
#define ADK_SYS_RSRC_EVQ_CFG_SINGLE_LINK        (1ULL << 2)
/**< This event queue links only to a single event port.
 */


/*!
 * @brief data structure to represent event-queue information
 */
typedef struct adk_sys_rsrc_evq_info_s {

    /*!  event queue id*/
    uint8_t     event_queue_id; 

    /*! mode to indicate whether this event-queue is dedicated to ADK software 
     * or shared between application software and ADK software.
     */
    adk_sys_rsrc_mode_t  mode;

    /*! Event queue configuration bitmap flags 
     * ADK_SYS_RSRC_EVQ_CFG_ATOMIC_ONLY, ADK_SYS_RSRC_EVQ_CFG_ORDERED_ONLY
     * ADK_SYS_RSRC_EVQ_CFG_PARALLEL_ONLY*/    
    uint32_t    evq_cfg_flags;  

    /*! list of event priorities that are assigned to ADK software 
     * on this event-queue.
     */
    uint8_t     event_pri_list[8]; 

    /*! number of event priorities that are assigned to ADK software 
     * on this event-queue.
     */
    uint8_t     num_priorities;

} adk_sys_rsrc_evq_info_t;

/*!
 * =============================================================
 *
 * @struct adk_sys_rsrc_evdev_info_t adk_sys.h "adk_sys.h"
 *
 * @brief eventdev resources information 
 *
 * This structure is used to specify details event-dev details to be 
 * registered with ADK software. 
 *
 *=============================================================
 */
typedef struct adk_sys_rsrc_evdev_info_s {
    /*! event device id */
    uint8_t                                 event_dev_id;
    /*! array of event ports. each element in the array
     *  represents a worker-core-id and corresponding event-port-id.
     */
    adk_sys_rsrc_evp_info_t                  *event_port_list_p;
    /*! number of event-ports 
     */
    uint8_t                                 num_ev_ports;
    /*! list of event-queues 
     */
    adk_sys_rsrc_evq_info_t                  *event_q_list_p;
    /*! number of event-queues
     */
    uint8_t                                 num_ev_queues;
    /*! reserved for ADK internal usage.
     */
    uint64_t               reserved;
} adk_sys_rsrc_evdev_info_t;


/*!
 * =============================================================
 * @struct adk_sys_rsrc_wcore_info_t adk_sys_config.h "adk_sys_config.h"
 *
 * @brief Structure for to represent dataplane worker cores.
 *
 * =============================================================
 */
typedef struct adk_sys_rsrc_wcore_info_s {
    /*!
     * Worker lcore list size.
     */
    uint32_t      num_worker_lcores;
    /*!
     * Pointer to list of DP logical core IDs on which ADK is executed.
     */
    uint8_t      *worker_lcores;
} adk_sys_rsrc_wcore_info_t;


/*!
 * =============================================================
 * @enum adk_sys_rsrc_timer_dev_type_e adk_sys_config.h "adk_sys_config.h"
 *
 * @brief Structure for passing timer device type information
 *
 * =============================================================
 */
typedef enum adk_sys_rsrc_timer_dev_type_s {

    ADK_SYS_RSRC_TIMER_DEV_TYPE_EVNT_TIMER_BD   = 1,
    ADK_SYS_RSCR_TIMER_DEV_TYPE_MAX
} adk_sys_rsrc_timer_dev_type_e;

/*!
 * =========================================================================
 * @struct adk_sys_rsrc_timer_dev_info_t adk_sys_config.h "adk_sys_config.h"
 *
 * @brief Structure for passing timer device information
 *
 * =========================================================================
 */
typedef struct adk_sys_rsrc_timer_dev_info_s {

    adk_sys_rsrc_timer_dev_type_e    type;
    union {
        uint8_t     timer_bd_instance_id;
    };
} adk_sys_rsrc_timer_dev_info_t;


/*!
 * =========================================================================
 * @struct adk_sys_rsrc_eth_dev_info_t adk_sys_config.h "adk_sys_config.h"
 *
 * @brief data structure to represent list of DPDK eth-devs and associated 
 * tx bridge driver event-queue.
 *
 * =========================================================================
 */
typedef struct adk_sys_rscr_eth_dev_info_s {
    /*! list of ethernet devs*/
    uint16_t     eth_dev_list[RTE_MAX_ETHPORTS];

    /*! number of ethernet devs*/
    uint8_t     num_eth_devs;  

    /*! event queue id of the serving bridge-eth tx driver */
    uint8_t     event_queue_id;
} adk_sys_rsrc_eth_dev_info_t;


/*! 
 *@brief General structure to represent any type of resource details.
 */
typedef union adk_sys_resource_info_u{
    /*! Crypto device information   */
    adk_sys_rsrc_crypto_dev_info_t  *crypto_dev_info_p;
    
    /*! event-dev information       */
    adk_sys_rsrc_evdev_info_t       *evdev_info_p;

    /*! worker core list information*/
    adk_sys_rsrc_wcore_info_t       *wcore_list_info_p;

    /*! timer device information    */
    adk_sys_rsrc_timer_dev_info_t   *timer_dev_info_p;

    /*! eth-dev information    */
    adk_sys_rsrc_eth_dev_info_t     *eth_dev_info_p;

    /*!
     * Pointer to pktio memorypool to be used in packet generation
     * in ADK dataplane like BFD etc..
     */
    void                            *pktio_mempool;

    /*!
     * Reserved for ADK internal usage 
     */
    void                            *reserve_p;
    
}adk_sys_resource_info_t;

/*!
 * =============================================================
 *
 * @struct adk_sys_resource_t adk_sys_config.h "adk_sys_config.h"
 *
 * @brief structure to represent a platform resources 
 *
 * =============================================================
 */
typedef struct adk_sys_resource_s{
    /*! Resource type */
    adk_sys_resource_type_t        type;
    /*! resource details*/
    adk_sys_resource_info_t        res_info_p;
} adk_sys_resource_t;

/*!
* adk_sys_resource_register
*
* @brief        API to register resources to ADK. 
*
* @param[in]    res_list   Pointer to hold resource info list
* @param[in]    num_res    number of resources 
*
* @return Returns ADK_SUCCESS if success, error code otherwise.
*/
adk_status_t adk_sys_resource_register(adk_sys_resource_t res_list[], uint8_t num_res);

/*!
* adk_sys_resource_unregister
*
* @brief        API to un-register resources that previously registered to ADK. 
*
* @param[in]    resource to be unregistered 
* @param[in]    num_res    number of resources 
*
* @return Returns ADK_SUCCESS if success, error code otherwise.
*/
adk_status_t adk_sys_resource_unregister(adk_sys_resource_t res_list[], uint8_t num_res);


#ifdef __cplusplus
}
#endif

#endif /* __ADK_SYS_CONFIG_H__ */

/*! end addtogroup _sif
 * @}
 */
