/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_bind_interface.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifFunctions  SIF Functions
 *     @{
 *
 * @brief 
 *
 * This file contains adk bind APIs and the internal functions that
 * are required to support bind. 
 *
 ***********************************************************************/

#include <adk_include.h>
#include <adk_bind_tree_cfg.h>
#include <adk_utils_tree_update.h>

/*
 * static functions used within this file.
 */
#if 0
static adk_status_t adk_bind_process_request (
        adk_sys_handle_t, const adk_bind_info_t *);

static adk_status_t adk_unbind_process_request (
        adk_sys_handle_t, const adk_bind_info_t *);
static adk_status_t adk_bind_process_dest_module_info ( 
        adk_sys_handle_t, const adk_bind_info_t *, uint32_t,  
        bool cancelProcess );

static adk_status_t adk_bind_process_src_module_info ( 
        adk_sys_handle_t, const adk_bind_info_t *, uint32_t, bool);

static adk_status_t adk_bind_forward2_destModule ( 
        adk_sys_handle_t, adk_bind_cpu_table_entry_t *, const void *);
#endif

static adk_status_t adk_bind_cpu_resource_init (adk_sys_handle_t);

/*!
 * =============================================================
 * @internal
 * @param[in] sysHandle   system handle returned by adk_sys_init.
 *
 * @retval ADK_SUCCESS the initialization was succesfull.
 * 
 * @retval ADK_E_MEM_UNAVAILABLE   failed to allocate memory
 * during initialization.
 *
 * @retval ADK_E_INCOSISTENT_STATE - Failed to get tree ID of one
 * of the bind trees.
 * 
 * @brief  
 *
 * This function initializes all the data structures that are 
 * required by the bind infrastructure. The only state information 
 * managed by bind infrastructure is the de-multiplexing index for 
 * MPP functions.
 *
 * =============================================================
 */
adk_status_t
adk_bind_infrastructure_init ( 
        adk_sys_handle_t sysHandle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_bind_master_handle_t *masterHandle_p;
    adk_module_handle_t	   handle;
    handle = ADK_SYS_MODULE_HANDLE_GET(sysHandle);
   
    masterHandle_p = (adk_bind_master_handle_t *)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(handle), 0,
		 1, sizeof(adk_bind_master_handle_t));

    ((adk_sys_info_t *)sysHandle)->bindMasterHandle = 
        (adk_handle_t) masterHandle_p;

    if (!ADK_VALID_HANDLE(masterHandle_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_FORCE_LOG_ERROR("Failed to init bind infrastructure. memory alloc failed.\n");
        goto adk_bind_infrastructure_init_done;
    }

    retValue = adk_bind_cpu_resource_init(sysHandle);

adk_bind_infrastructure_init_done:

    ADK_RETURN(retValue);            

}    /* -----  end of function adk_bind_infrastructure_init  ----- */


/*!
 * =============================================================
 * @internal
 * adk_bind_infrastructure_deinit
 * @param[in]    sysHandle   ADK system handle (of type adk_sys_handle_t).
 *
 * @retval ADK_SUCCESS    Successfully completed the API.
 * @retval ADK_E_INVALID_PARAMETER  Invalid bind infrastructure 
 * handle pointer.
 *
 * @brief  
 *
 * This function de-allocate the resources created/reserved by
 * the function adk_bind_infrastructure_init.
 *
 * =============================================================
 */
adk_status_t
adk_bind_infrastructure_deinit ( 
        adk_sys_handle_t sysHandle
        )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_info_t  *sysInfo_p = (adk_sys_info_t *) sysHandle;
    adk_bind_master_handle_t *bindHandle_p;

    bindHandle_p = (adk_bind_master_handle_t *) sysInfo_p->bindMasterHandle;

    if ( !bindHandle_p )
    {
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Failed to get Bind IMC tree ID\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        goto adk_bind_infrastructure_deinit;
    }

    /* Free the CPU bind resources */
    if ( bindHandle_p->cpuBindTable_p )
    {
        ADK_NVM_FREE(bindHandle_p->cpuBindTable_p);
    }
    /* This function is called only from adk_sys_shutdown.
     * So there is no need to validate the system handle.
     * It is already validated by the caller.
     */
    ADK_NVM_FREE (sysInfo_p->bindMasterHandle);

adk_bind_infrastructure_deinit:
    ADK_RETURN( retValue );
}    /* -----  end of function adk_bind_infrastructure_deinit  ----- */




/*!
 * =============================================================
 * @internal
 * @param[in]    sysHandle
 *
 * @retval ADK_SUCCESS Successfully completion
 *
 * @brief  
 *
 * Detailed description startes here
 *
 * =============================================================
 */
static adk_status_t
adk_bind_cpu_resource_init ( 
       adk_sys_handle_t sysHandle )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_bind_master_handle_t *bindHandle_p;
    adk_module_handle_t	   handle;
    handle = ADK_SYS_MODULE_HANDLE_GET(sysHandle);

    bindHandle_p = (adk_bind_master_handle_t *)
        ((adk_sys_info_t *)sysHandle)->bindMasterHandle;

    bindHandle_p->cpuBindTable_p = 
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(handle), 0,
		 ADK_MAX_SUPPORTED_MODULES, sizeof(adk_handle_t));
    
    if (!ADK_VALID_PTR(bindHandle_p->cpuBindTable_p)) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                "Failed to init bind infrastructure. memory alloc failed\n");
    }

    ADK_RETURN( retValue );
}    /* -----  end of function adk_bind_cpu_resource_init  ----- */

/*!
 * =============================================================
 * @internal 
 * @param[in]     system handle
 * @param[in]     pointer to bind information data structure. not 
 * all fields in adk_bind_info_t may be relevant here. The key 
 * fields of interest are the srcModuleHandle and srcFlowHandle.
 *
 * @retval        ADK_STATUS  successfully completed bind
 * @retval        ADK_E_INVALID_HANDLE  input handle/pointer is
 * not valid.
 * @retval        ADK_E_RESOURCE_UNAVAILABLE  failed to allocate
 * bind related resource.
 * @retval        ADK_E_INTERNAL  internal processing error
 * @brief  
 *
 * The same as that of adk_unbind except the high level validation.
 *
 * =============================================================
 */
#if 0
adk_status_t
adk_unbind_process_request ( 
        adk_sys_handle_t sysHandle,
        const adk_bind_info_t  *bindInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint16_t dmuxFnId = 0;

    /* handle validations are already performed by the caller */
    if ( bindInfo_p->destLocation == ADK_BIND_LOC_ENGINES )
    {
        if (bindInfo_p->direction == ADK_BIND_INGRESS_PROCESSING)
        {
            if (bindInfo_p->path == ADK_BIND_PATH_DIRECT) 
            {
                dmuxFnId = ADK_MODULE_INGRESS_MPP_DIRECT_ENTRY_POINT_GET(
                        bindInfo_p->destModuleHandle);
            } else 
            {
                dmuxFnId = ADK_MODULE_INGRESS_MPP_ENTRY_POINT_GET(
                        bindInfo_p->destModuleHandle);
            }

        } 
        else if (bindInfo_p->direction == ADK_BIND_EGRESS_PROCESSING)
        {
            if (bindInfo_p->path == ADK_BIND_PATH_DIRECT) 
            {
                dmuxFnId = ADK_MODULE_EGRESS_MPP_DIRECT_ENTRY_POINT_GET(
                        bindInfo_p->destModuleHandle);
            } 
            else 
            {
                dmuxFnId = ADK_MODULE_EGRESS_MPP_ENTRY_POINT_GET(
                        bindInfo_p->destModuleHandle);
            }
        } 
        else 
        {
            retValue = ADK_E_INVALID_PARAMETER;
            goto adk_unbind_process_request_return;
        }
    }

    /* update adk_bind_dmux_tree. The format is 
     * pipeline:8 dmuxIndex:4 :: goto dmux function.
     */
    /* This call will update the main bind tree and the IMC tree.
     */
    retValue = adk_bind_process_src_module_info(sysHandle, bindInfo_p, 
                                                dmuxFnId, TRUE);
    if (retValue != ADK_SUCCESS)
    {
        goto adk_unbind_process_request_failure1;
    } else 
    {
        goto adk_unbind_process_request_return;
    }

adk_unbind_process_request_failure1:

    adk_bind_process_dest_module_info(sysHandle, bindInfo_p, 
                                      dmuxFnId, TRUE);

adk_unbind_process_request_return:

    ADK_RETURN (retValue);

}    /* -----  end of function adk_unbind_process_request  ----- */
#endif

/*!
 * =============================================================
 * @param[in]   sysHandle    system handle
 * 
 * @param[in]   bindInfo_p    bind information data structure
 * pointer. This data structure specifies all properties of the
 * requested bind. individual fields in the data structure is
 * described in header file.
 *
 * @retval   ADK_STATUS   successfully completed bind
 * 
 * @retval   ADK_E_INVALID_HANDLE   input handle/pointer is
 * not valid.
 *
 * @retval   ADK_E_RESOURCE_UNAVAILABLE  failed to allocate
 * bind related resource.
 *
 * @retval   ADK_E_INVALID_PARAMS  invalid input parameter
 * passed.
 *
 * @brief  
 *
 * This function creates a linkage between source module and 
 * destination module. The type of linkage and the granularity
 * level is based on the bind type (M2M, F2M, F2F). The
 * information provided in the bindInfo_p data structure is used
 * to update the fast path trees and the bind tables in CPU.
 *
 * =============================================================
 */
adk_status_t adk_bind (
        adk_sys_handle_t sysHandle,
        const adk_bind_info_t *bindInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    module_bind_fn_t  bindFn;
    module_bind_fn_t srcModuleSrcBindFn;
    module_bind_fn_t dstModuleDstBindFn;

    /* To Do: Only custom bind is supported at present. Changes are required if
     * there is a need for non-custom bind. */
    bool                      isCustomBind = TRUE;

    /* Validate system handle and bind info pointer */
    if (ADK_INVALID_SYS_HANDLE(sysHandle) || !ADK_VALID_PTR (bindInfo_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Bind operation failed. invalid sys handle or bind"
                            "info pointer\n");
        goto _adk_bind_exit;
    }

    /* Validate the binding path type */
    if ((bindInfo_p->path != ADK_BIND_PATH_DIRECT) &&
        (bindInfo_p->path != ADK_BIND_PATH_INDIRECT)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind direction type\n");
        goto _adk_bind_exit;
    }

    /* Validate the binding direction */
    if ((bindInfo_p->direction != ADK_BIND_INGRESS_PROCESSING) &&
        (bindInfo_p->direction != ADK_BIND_EGRESS_PROCESSING)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind path type\n");
        goto _adk_bind_exit;
    }

    /* Validate the module handles */
    if (!ADK_MODULE_HANDLE_VALIDATE(bindInfo_p->srcModuleHandle) ||
        !ADK_MODULE_HANDLE_VALIDATE(bindInfo_p->destModuleHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid module handle\n");
        goto _adk_bind_exit;
    }

#if 0
### This 'if 0' make the if (isCustomBind != TRUE) below dead code
###    isCustomBind = bindInfo_p->isCustomBind ;
#endif

    srcModuleSrcBindFn = ADK_MODULE_SRC_BIND_FN_GET(bindInfo_p->srcModuleHandle);
    dstModuleDstBindFn = ADK_MODULE_DST_BIND_FN_GET(bindInfo_p->destModuleHandle);

    /* Validate the binding type and flow handles along with binding type */
    if (bindInfo_p->type == ADK_BIND_FLOW_TO_MODULE) 
    {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->srcFlowHandle)) 
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid source flow handle\n");
            goto _adk_bind_exit;
        }
    } 
    else if (bindInfo_p->type == ADK_BIND_FLOW_TO_FLOW) 
    {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->srcFlowHandle)) 
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid source flow handle\n");
            goto _adk_bind_exit;
        }
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->destFlowHandle)) 
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid destination flow handle\n");
            goto _adk_bind_exit;
        }
    } 
    else if (bindInfo_p->type == ADK_BIND_FLOW_TO_ID)
    {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->srcFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid source flow handle\n");
            goto _adk_bind_exit;
        }

#if 0
### Must be false as it never changes due to #if 0 above
###        if (isCustomBind != TRUE) {
###            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
###            ADK_FORCE_LOG_ERROR("Only custom bind is supported for "
###                    "the bind type ADK_BIND_FLOW_TO_ID\n");
###            goto _adk_bind_exit;
###        }
#endif

        if ((isCustomBind == TRUE) &&
                ((srcModuleSrcBindFn == NULL) /* || (dstModuleDstBindFn == NULL) */)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Source or Destination module does not "
                    "support custom bind\n");
            goto _adk_bind_exit;
        }
    }
    else if (bindInfo_p->type == ADK_BIND_ID_TO_FLOW)
    {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->destFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid destination flow handle\n");
            goto _adk_bind_exit;
        }   

#if 0
### Must be false as it never changes due to #if 0 above
###        if (isCustomBind != TRUE) {
###            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
###            ADK_FORCE_LOG_ERROR("Only custom bind is supported for "
###                    "the bind type ADK_BIND_ID_TO_FLOW\n");
###            goto _adk_bind_exit;
###        }   
#endif

        if ((isCustomBind == TRUE) && (srcModuleSrcBindFn == NULL)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Source module does not support custom bind\n");
            goto _adk_bind_exit;
        }
    }
    else if (bindInfo_p->type != ADK_BIND_MODULE_TO_MODULE) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind type\n");
        goto _adk_bind_exit;
    }

#if 0
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind type\n");
        goto _adk_bind_exit;
    }
#endif
    if ((isCustomBind == TRUE) && (srcModuleSrcBindFn)) {
        retValue = (*srcModuleSrcBindFn) (sysHandle, bindInfo_p);
        if (retValue != ADK_SUCCESS) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Source module custom bind function call failed\n");
            goto _adk_bind_exit;
        }
    }

    if (((bindInfo_p->type == ADK_BIND_ID_TO_MODULE) ||
                (bindInfo_p->type == ADK_BIND_FLOW_TO_ID)) &&
            (isCustomBind == TRUE) &&
            (dstModuleDstBindFn)) {
        retValue = (*dstModuleDstBindFn) (sysHandle, bindInfo_p);
        if (retValue != ADK_SUCCESS) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Destination module custom bind function call "
                    "failed\n");
            goto _adk_bind_exit;
        }
    }
    bindFn = ADK_MODULE_BIND_FN_GET(bindInfo_p->srcModuleHandle);

    if ((isCustomBind == TRUE) && (bindFn)) {
        retValue = (*bindFn) (sysHandle, bindInfo_p); 
        if (retValue != ADK_SUCCESS) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "custom bind function call failed\n");
            goto _adk_bind_exit;
        }
    } else {
         /*
          * TODO: shall be updated for CPU bindings 
          */
#if 0
        retValue = adk_bind_process_request (sysHandle, bindInfo_p);
#endif
    }

_adk_bind_exit:

    ADK_RETURN(retValue);
}    /* -----  end of function adk_bind  ----- */

/*!
 * =============================================================
 * @param[in]     sysHandle system handle
 *
 * @param[in]     bindInfo_p pointer to bind information data structure. not 
 * all fields in adk_bind_info_t may be relevant here. The key 
 * fields of interest are the srcModuleHandle and srcFlowHandle.
 *
 * @retval        ADK_STATUS  successfully completed bind
 * @retval        ADK_E_INVALID_HANDLE  input handle/pointer is
 * not valid.
 * @retval        ADK_E_INTERNAL  internal processing error
 *
 *
 * @brief  
 *
 * This function remove a previously setup bind. All the dataplane
 * objects and fastpath trees related to the specific bind function
 * is removed here.
 *
 * =============================================================
 */
adk_status_t adk_unbind (
        adk_sys_handle_t sysHandle,
        const adk_bind_info_t *bindInfo_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    module_unbind_fn_t unbindFn;
    module_unbind_fn_t srcModuleSrcUnbindFn;
    module_unbind_fn_t dstModuleDstUnbindFn;
    /* To Do: Only custom bind is supported at present. Changes are required if
     * there is a need for non-custom bind. */
    bool                      isCustomBind = TRUE;
    
    /* Validate system handle and bind info pointer */
    if (ADK_INVALID_SYS_HANDLE(sysHandle) || !ADK_VALID_PTR(bindInfo_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_FORCE_LOG_ERROR("Bind operation failed. invalid sys handle or bind"
                            "info pointer\n");
        goto _adk_unbind_exit;
    }

    /* Validate the binding path type */
    if ((bindInfo_p->path != ADK_BIND_PATH_DIRECT) &&
        (bindInfo_p->path != ADK_BIND_PATH_INDIRECT)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind path type\n");
        goto _adk_unbind_exit;
    }
#if 0
    /* Validate the location. The ADK_BIND_LOC_CPU is not supported */
    if ((bindInfo_p->srcLocation != ADK_BIND_LOC_ENGINES) ||
        (bindInfo_p->destLocation != ADK_BIND_LOC_ENGINES)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid or unsupported module location\n");
        goto _adk_unbind_exit;
    }
#endif
    /* Validate the binding direction */
    if ((bindInfo_p->direction != ADK_BIND_INGRESS_PROCESSING) &&
        (bindInfo_p->direction != ADK_BIND_EGRESS_PROCESSING)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind path type\n");
        goto _adk_unbind_exit;
    }

    /* Validate the module handles */
    if (!ADK_MODULE_HANDLE_VALIDATE(bindInfo_p->srcModuleHandle) ||
        !ADK_MODULE_HANDLE_VALIDATE(bindInfo_p->destModuleHandle)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid module handle\n");
        goto _adk_unbind_exit;
    }

#if 0
    isCustomBind = bindInfo_p->isCustomBind ;
#endif

    srcModuleSrcUnbindFn = ADK_MODULE_SRC_UNBIND_FN_GET(bindInfo_p->srcModuleHandle);
    dstModuleDstUnbindFn = ADK_MODULE_DST_UNBIND_FN_GET(bindInfo_p->destModuleHandle);

    /* Validate the binding type and flow handles along with binding type */
    if (bindInfo_p->type == ADK_BIND_FLOW_TO_MODULE) {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->srcFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid source flow handle\n");
            goto _adk_unbind_exit;
        }
    } else if (bindInfo_p->type == ADK_BIND_FLOW_TO_FLOW) {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->srcFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid source flow handle\n");
            goto _adk_unbind_exit;
        }
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->destFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid destination flow handle\n");
            goto _adk_unbind_exit;
        }
    }
    else if (bindInfo_p->type == ADK_BIND_FLOW_TO_ID)
    {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->srcFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid source flow handle\n");
            goto _adk_unbind_exit;
        }

#if 0
### Must be false as it never changes due to #if 0 above
###        if (isCustomBind != TRUE) {
###            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
###            ADK_FORCE_LOG_ERROR("Only custom bind is supported for "
###                    "the bind type ADK_BIND_FLOW_TO_ID\n");
###            goto _adk_unbind_exit;
###        }
#endif

        if ((isCustomBind == TRUE) &&
                (srcModuleSrcUnbindFn == NULL)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Source module does not "
                    "support custom bind\n");
            goto _adk_unbind_exit;
        }
    }

    else if (bindInfo_p->type == ADK_BIND_ID_TO_FLOW)
    {
        if (!ADK_SYS_VALID_FLOW_HANDLE(bindInfo_p->destFlowHandle)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Invalid destination flow handle\n");
            goto _adk_unbind_exit;
        }   

#if 0
### Must be false as it never changes due to #if 0 above
###        if (isCustomBind != TRUE) {
###            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
###            ADK_FORCE_LOG_ERROR("Only custom bind is supported for "
###                    "the bind type ADK_BIND_ID_TO_FLOW\n");
###            goto _adk_unbind_exit;
###        }   
#endif

        if ((isCustomBind == TRUE) && (srcModuleSrcUnbindFn == NULL)) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_FORCE_LOG_ERROR("Source module does not support custom bind\n");
            goto _adk_unbind_exit;
        }
    }
    else if (bindInfo_p->type != ADK_BIND_MODULE_TO_MODULE) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid bind type\n");
        goto _adk_unbind_exit;
    }

    if ((isCustomBind == TRUE) && (srcModuleSrcUnbindFn)) {
        retValue = (*srcModuleSrcUnbindFn) (sysHandle, bindInfo_p);
        if (retValue != ADK_SUCCESS) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Source module custom bind function call failed\n");
            goto _adk_unbind_exit;
        }
    }

    if (((bindInfo_p->type == ADK_BIND_ID_TO_MODULE) ||
                (bindInfo_p->type == ADK_BIND_FLOW_TO_ID)) &&
            (isCustomBind == TRUE) &&
            (dstModuleDstUnbindFn)) {
        retValue = (*dstModuleDstUnbindFn) (sysHandle, bindInfo_p);
        if (retValue != ADK_SUCCESS) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                    "Destination module custom bind function call "
                    "failed\n");
            goto _adk_unbind_exit;
        }
    }

    unbindFn = ADK_MODULE_UNBIND_FN_GET(bindInfo_p->srcModuleHandle);

    if ((isCustomBind == TRUE) && (unbindFn)) {
        retValue = (*unbindFn) (sysHandle, bindInfo_p); 
        if (retValue != ADK_SUCCESS) {
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                            "custom unbind function call failed\n");
            goto _adk_unbind_exit;
        }
    } else {
        /*
         * TODO: shall be updated for CPU bindings
         */
#if 0
         retValue = adk_unbind_process_request(sysHandle, bindInfo_p); 
#endif
         
    }

_adk_unbind_exit:

    ADK_RETURN ( retValue );
}    /* -----  end of function adk_unbind  ----- */

#if 0
###/*!
### * =============================================================
### * @param[in] sysHandle    ADK system handle.
### *
### * @param[in] moduleHandle    source module handle.
### *
### * @param[in] flowHandle    source flow handle.
### *
### * @param[in] pduInfo_p    pointer to the PDU buffer.
### *
### * @retval ADK_SUCCESS Successfully completion
### * 
### * @retval ADK_E_INVALID_HANDLE    invalid system or module handle.
### *
### * @brief  
### *
### * This function forwards the PDU from a CPU originated module 
### * to a hardware centric module or CPU centric module. The internal
### * implementation here is similar to the MPP based bind that 
### * we described under adk_bind. 
### *
### * =============================================================
### */
###adk_status_t
###adk_bind_forward_pdu ( 
###       adk_sys_handle_t sysHandle,
###       adk_module_handle_t moduleHandle,
###       adk_flow_handle_t flowHandle,
###       const void *pduInfo_p )
###{
###    adk_status_t retValue = ADK_SUCCESS;
###    adk_bind_master_handle_t *bindHandle_p;
###    adk_handle_t *cpuBindTable_p;
###    void *route_info;
###    uint32_t moduleId;
###    adk_bind_cpu_table_entry_t *route_info_p;
###
###    if (ADK_INVALID_SYS_HANDLE (sysHandle) ||
###            !ADK_VALID_HANDLE (moduleHandle))
###    {
###        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
###        ADK_FORCE_LOG_ERROR("Failed to route PDU. Invalid handle.\n");
###    } else
###    {
###        moduleId = ADK_MODULE_MODULEID_GET (moduleHandle);
###        bindHandle_p = (adk_bind_master_handle_t *) 
###            ((adk_sys_info_t *)sysHandle)->bindMasterHandle;
###    
###        cpuBindTable_p = bindHandle_p->cpuBindTable_p;
###        /* get the destination module info from the bind table
###         * and process it.
###         */
###        if ( cpuBindTable_p[moduleId] )
###        {
###            retValue = adk_utils_list_find (
###                   (adk_list_t *) cpuBindTable_p[moduleId],
###                   (intptr_t)flowHandle, &route_info);
###            route_info_p = (adk_bind_cpu_table_entry_t *) route_info;
###            if ( (retValue == ADK_SUCCESS) && (route_info_p != NULL) )
###            {
###                adk_bind_forward2_destModule (sysHandle, route_info_p, 
###                       pduInfo_p);
###            } else 
###            {
###                ADK_LOG_WARNING(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
###                                "Failed to find matching entry for "
###                                "module %"ADKPRIh" flow %"ADKPRIh"\n",
###                                moduleHandle, flowHandle);
###            }
###        }
###    }
###    
###    ADK_RETURN( retValue );
###}    /* -----  end of function adk_bind_forward_pdu  ----- */
###
###
###/*!
### * =============================================================
### * @internal
### * @param[in] sysHandle    ADK system handle.
### *
### * @param[in] route_info_p    pointer to adk_bind_cpu_table_entry
### * data structure. This data structure carries destination module
### * specific information.
### *
### *
### * @param[in] pduInfo_p    pointer to the PDU buffer.
### * 
### * @retval ADK_SUCCESS Successfully completion
### *
### * @brief  
### *
### * Internal helper function to support adk_bind_forward_pdu API.
### *
### * =============================================================
### */
###static adk_status_t
###adk_bind_forward2_destModule ( 
###        adk_sys_handle_t sysHandle,
###        adk_bind_cpu_table_entry_t *route_info_p,
###        const void *pduInfo_p )
###{
###    adk_status_t retValue = ADK_SUCCESS;
###
###    if ( route_info_p->cpuImcFn )
###    {
###        (*(route_info_p->cpuImcFn)) (sysHandle,
###                route_info_p->destModule, pduInfo_p);
###    }
###
###    if ( route_info_p->destLocation == ADK_BIND_LOC_CPU )
###    {
###        /* hand off the PDU to CPU entry function of the destination
###         * module.
###         */
###        if ( route_info_p->destEntryFn )
###        {
###            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
###                         "calling CPU entry function of destination module.\n");
###        }
###    } else 
###    {
###        /* call PDU handler send function here.
###         */
###        ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
###                     "calling PDU handler send to hand off PDU to MPP.\n");
###    }
###
###    ADK_RETURN( retValue );
###}    /* -----  end of function adk_bind_forward2_destModule  ----- */
#endif


/*!     end addtogroup _sifFunctions
 *      @}
 *
 * end addtogroup _sif
 * @}
 */
