/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_dev_net.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _devNet  ADK Device Net
 * @{
 *     @addtogroup _devNetFunctions  ADK Device Net Functions
 *     @{
 *
 * @brief
 *
 * This file contains the adk software initialization functions and
 * the ACP hardware initialization functions.
 *
 *
 *  **********************************************************************/

#include "adk_include.h"
#include "ncp.h"

#include "adk_sys.h"
#include "adk_sys_private.h"

#include "adk_release_id.h"

#include <adk_rm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "adk_config_overrides.h"

#include <adk_sys_dpdk.h>
#include "adk_shm.h"
#include <sys/time.h>

#include "adk_dev_net.h"
#include "adk_dev_net_internal.h"
#include <adk_nd_common.h>
#include "adk_qos_include_internal.h"
#include <adk_netd_fp_ts_index.h>


/*
 * file local global variables
 */
static adk_dev_net_pvt_info_t *dev_list_p[ADK_DEV_NET_ID_MAX];
static uint32_t                 num_dev_nets;

/*
 * file local static functions
 */

static adk_status_t
adk_dev_net_sdm_create(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p);
static adk_status_t
adk_dev_net_sdm_delete(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p);
static adk_status_t
adk_dev_net_sdm_config_params_validate(adk_sys_handle_t sys_handle,
                              adk_dev_net_config_params_t *config_p);
static adk_status_t
adk_dev_net_sdm_configure(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p,
                        adk_dev_net_config_params_t *config_p);
static adk_status_t
adk_dev_net_sdm_rx_queue_setup(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p,
                        uint16_t                 queue_id);
static adk_status_t
adk_dev_net_sdm_tx_queue_setup(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p,
                        uint16_t                 queue_id);

static adk_status_t
adk_dev_net_sdm_stats(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p);

/*
 * adk_dev_net_init()
 * @param[in] sys_handle    ADK System Handle
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to the ADK Slow Path Device init functionality
 */
adk_status_t
adk_dev_net_init(adk_sys_handle_t sys_handle)
{
    adk_status_t retValue=ADK_SUCCESS;
    adk_sys_info_t *sysInfo_p = (adk_sys_info_t *)sys_handle;
    int i =0;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return ADK_E_INVALID_PARAMETER;
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    /* Create User defined resources */
    sysInfo_p->sp_dev_ids_rm_hdl=
        adk_rm_resource_init(ADK_SYS_MODULE_ID, 0, ADK_DEV_NET_ID_MAX);
    if(sysInfo_p->sp_dev_ids_rm_hdl== NULL)
    {
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "Failed to Allocate RM for ADK Slow Path Device IDs DB\n");
        return ADK_E_INTERNAL_ERROR;
    }


#if 0
    /* Create memzone for the sys egress stats */
    retValue = adk_shm_create(ADK_SYS_EGRESS_STATS_SHM_MEMZONE_NAME,
             (sizeof(adk_sys_dp_egress_stats_t) * ADK_SYS_EGRESS_STATS_ID_MAX),
             ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
             (adk_shm_handle_t **)&(sysInfo_p->egress_stats_p));
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "Failed to create memzone for Sys egress stats DB\n");
        return ADK_E_MEM_UNAVAILABLE;
    }
#endif

    /*Clear the Global variables */
    for(i=0;i<ADK_DEV_NET_ID_MAX; i++)
    {
        dev_list_p[i] = NULL;
    }

    /*reserve ID range [0 - (RTE_MAX_ETHPORTS-1)]
     */
    for(i=ADK_DEV_NET_ID_RSVD_MIN;i<=ADK_DEV_NET_ID_RSVD_MAX_PH; i++)
    {
        adk_rm_resource_reserve_fixed(sysInfo_p->sp_dev_ids_rm_hdl, i);
        dev_list_p[i] = NULL;
    }

    num_dev_nets = 0;

    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    return retValue;
}

/*
 * adk_dev_net_shutdown()
 * @param[in] sys_handle    ADK System Handle
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to do the ADK Slow Path Device shutdfown functionality
 */

adk_status_t
adk_dev_net_shutdown(adk_sys_handle_t sys_handle)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_sys_info_t *sysInfo_p = (adk_sys_info_t *)sys_handle;
    int32_t i;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return ADK_E_INVALID_PARAMETER;
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);
    adk_rm_resource_destroy(sysInfo_p->sp_dev_ids_rm_hdl);

    /*Clear the Global variables */
    for(i=0;i<ADK_DEV_NET_ID_MAX; i++)
    {
        if(dev_list_p[i]!=NULL)
        {
            ADK_NVM_FREE(dev_list_p[i]);
            dev_list_p[i] = NULL;
        }
    }
    num_dev_nets = 0;

    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN (retValue);
}


/*
 * adk_dev_net_create()
 * @param[in] sys_handle    ADK System Handle
 * @param[in] info_p       pointer to adk_dev_net_info_t that represents information.
 * @param[out] dev_id_P    pointer to store ADK allocated ADK-Linux-Device-ID.
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to create a ADK Linux Device. It returns created device id.
 * Applicaiton is required to use this device to send/receive packets to/from
 * Linux Kernel(ADK NetDriver).
 */
adk_status_t
adk_dev_net_create(adk_sys_handle_t sys_handle,
                     adk_dev_net_info_t *info_p,
                     adk_dev_id_t    *dev_id_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;
    int32_t  id;
    adk_sys_info_t *sysInfo_p = (adk_sys_info_t *)sys_handle;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        goto adk_dev_net_create_return;
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    /* validate Application passed linus device params */
    if(info_p== NULL) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter \n");
        goto adk_dev_net_create_return;
    }

    if(info_p->dev_type != ADK_DEV_NET_TYPE_NETD ) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNSUPPORTED);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "un supported slowpath device type \n");
        goto adk_dev_net_create_return;
    }

    if(info_p->dm_type != ADK_DEV_NET_DM_TYPE_SOFT ) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNSUPPORTED);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "un supported data mover type \n");
        goto adk_dev_net_create_return;
    }

#if 0
    if(info_p->bd_info.bd_instance_id == 0 ) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "In valid CPPI BD instance id \n");
        goto adk_dev_net_create_return;
    }

    if(info_p->soft_dm.bd_event_queue_id == 0 ) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "In valid CPPI-BD event-queue-id \n");
        goto adk_dev_net_create_return;
    }
#endif

    /* Allocate memory for Linux Device*/
    dev_pvt_info_p = (adk_dev_net_pvt_info_t *) ADK_NVM_CALLOC (ADK_SYS_MODULE_ID, 0, 1, sizeof (adk_dev_net_pvt_info_t));
    if (!ADK_VALID_PTR (dev_pvt_info_p))
    {
        retValue = ADK_STATUS_SEV_LEVEL_SET (ADK_E_MEM_UNAVAILABLE,
                                             ADK_SEV_LEVEL_FATAL);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "Memory allocation failed \n");
        goto adk_dev_net_create_return;
    }
    /*Acquire ADK System lock */
    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    /*Allocate the ADK slowpath device id  */
    id = adk_rm_resource_reserve(sysInfo_p->sp_dev_ids_rm_hdl);
    if(id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_RESOURCE_UNAVAILABLE);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "Failed to Allocate ADK Linux Device Id \n");
        goto adk_dev_net_create_lock_return;
    }

    /* ID used for ADK QoS queue id namespace */
    dev_pvt_info_p->device_id = id;

    /* CPPI Device ID  */
    dev_pvt_info_p->device_id_internal = (id - ADK_DEV_NET_ID_BASE + ADK_PH_DEV_MAX_NUM_DEVICES);
    ADK_MEMCPY(&dev_pvt_info_p->info, info_p, sizeof(adk_dev_net_info_t));


    if(info_p->pkt_mem_pool == NULL) {
        ADK_LOG_INFO(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "using ADK DP mempool for dev_net \n");
        if(ADK_SUCCESS != adk_sys_config_pktio_mempool_get(&(dev_pvt_info_p->info.pkt_mem_pool)))
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "In valid packet memory pool \n");
            goto adk_dev_net_create_lock_return;

        }
    }

    if ((id < 0) || id>= ADK_DEV_NET_ID_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "In valid ADK Linux Device Id \n");
        goto adk_dev_net_create_lock_return;
    }

    retValue = adk_dev_net_sdm_create(sys_handle, dev_pvt_info_p);
    if(retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "Failed to create SDM based ADK Linux Device Id \n");
        goto adk_dev_net_create_lock_return;
    }

    snprintf(dev_pvt_info_p->cookie_name, ADK_DEV_NET_COOKIE_SIZE,
            "%s_%u", ADK_DEV_NET_COOKIE,id);

    dev_pvt_info_p->device_configured   = 0;
    dev_pvt_info_p->device_started      = 0;
    dev_pvt_info_p->adk_netd_fd         = -1;

    dev_list_p[id]  = dev_pvt_info_p;

    *dev_id_p       = id;
    /* Store the dev_net_id in TS index mgmt database */
#ifdef ADK_DEVICE_SNR
    adk_fp_ts_dev_net_id = id;
#endif

#ifdef ADK_QOS_ENABLE
    if (adk_qos_instance_init(sys_handle, id) !=  ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not register Linux dev with  ADK QoS \n");
        goto adk_dev_net_create_lock_return;
    }
#endif


adk_dev_net_create_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_create_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}


/*
 * adk_dev_net_configure()
 * @param[in]  sys_handle    ADK System Handle
 * @param[in]  dev_id       ADK Linux Device ID
 * @param[in]  config_p     pointer to adk_dev_net_config_params_t
 *                          that define configuration parameters
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to configure previously created ADK Linux Device.
 *
 */
adk_status_t
adk_dev_net_configure(adk_sys_handle_t sys_handle,
                        adk_dev_id_t    dev_id,
                        adk_dev_net_config_params_t *config_p)
{

    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        goto adk_dev_net_config_return;
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    /* validate Application passed linus device params */
    if(config_p== NULL) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter NULL config_p pointer\n");
        goto adk_dev_net_config_return;
    }

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)|| (dev_id >= ADK_DEV_NET_ID_MAX)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter slowpath device id %d \n", dev_id);
        goto adk_dev_net_config_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_config_lock_return;
    }
    if(dev_pvt_info_p->device_configured !=0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is already configured; \n", dev_id);
        goto adk_dev_net_config_lock_return;
    }

    dev_pvt_info_p->n_rx_queues     += config_p->num_rx_queues;
    dev_pvt_info_p->n_tx_queues     += config_p->num_tx_queues;
    dev_pvt_info_p->rx_sched_mode    = config_p->rx_sched_mode;
    dev_pvt_info_p->tx_sched_mode    = config_p->tx_sched_mode;

    switch(dev_pvt_info_p->info.dev_type)
    {
    case ADK_DEV_NET_TYPE_NETD:
        retValue = adk_dev_net_sdm_config_params_validate(sys_handle, config_p);
        if(retValue!= ADK_SUCCESS) {
            goto adk_dev_net_config_lock_return;
        }
        retValue = adk_dev_net_sdm_configure(sys_handle, dev_pvt_info_p, config_p);
        if(retValue!= ADK_SUCCESS) {
            goto adk_dev_net_config_lock_return;
        }
        break;
    default:
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid linux device DM Type \n");
        goto adk_dev_net_config_lock_return;
        break;
    }

    dev_pvt_info_p->device_configured = 1;

adk_dev_net_config_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_config_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}

/*
 * adk_dev_net_tx_queue_setup()
 * @param[in]  sys_handle    ADK System Handle
 * @param[in]  dev_id       ADK Linux Device ID
 * @param[in]  tx_queue_id  tx (From dataplane to Linux direction) queue id
 * @param[in]  weight       weight of the queue used in case of device is
 *                          configured for RR scheduling; priority in case of SP scheduling mode
 * @param[in]  size         size in number of packets; ignored in case of HWM-DM type of device
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to setup the size and weight of a tx-queue in a Linux Device
 */
 adk_status_t
adk_dev_net_tx_queue_setup(adk_sys_handle_t      sys_handle,
                             adk_dev_id_t    dev_id,
                             uint16_t               tx_queue_id,
                             uint8_t               weight,
                             uint16_t              size)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        goto adk_dev_net_tx_queue_setup_return;
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)|| (dev_id >= ADK_DEV_NET_ID_MAX)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter linux device id %d \n", dev_id);
        goto adk_dev_net_tx_queue_setup_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_tx_queue_setup_lock_return;
    }

    if(dev_pvt_info_p->device_configured !=1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is not yet configured; \n", dev_id);
        goto adk_dev_net_tx_queue_setup_lock_return;
    }

    if(dev_pvt_info_p->device_started !=0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is already started; \n", dev_id);
        goto adk_dev_net_tx_queue_setup_lock_return;
    }

    if(tx_queue_id+1 >= dev_pvt_info_p->n_tx_queues )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid queue id %d  \n", tx_queue_id);
        goto adk_dev_net_tx_queue_setup_lock_return;
    }
    if (size> ADK_DEV_NET_QUEUE_SIZE_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid queue size  %d  \n", size);
        goto adk_dev_net_tx_queue_setup_lock_return;
    }

    dev_pvt_info_p->tx_q_info[tx_queue_id+1].size = size;
    dev_pvt_info_p->tx_q_info[tx_queue_id+1].weight= weight;
#if RDK_REL_MAJOR >= 13
    dev_pvt_info_p->tx_q_info[tx_queue_id+1].queue_id= tx_queue_id + 1;

    retValue = adk_dev_net_sdm_tx_queue_setup(sys_handle, dev_pvt_info_p,
                                                tx_queue_id + 1);
#else
    dev_pvt_info_p->tx_q_info[tx_queue_id+1].queue_id= tx_queue_id;
    retValue = adk_dev_net_sdm_tx_queue_setup(sys_handle, dev_pvt_info_p,
                                                tx_queue_id);
#endif
    if(retValue!=ADK_SUCCESS)
    {
        goto adk_dev_net_tx_queue_setup_lock_return;
    }
    dev_pvt_info_p->tx_q_info[tx_queue_id+1].setup_done = 1;

adk_dev_net_tx_queue_setup_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_tx_queue_setup_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}
/*
 * adk_dev_net_rx_queue_setup()
 *
 * @param[in]  sys_handle    ADK System Handle
 * @param[in]  dev_id       ADK Linux Device ID
 * @param[in]  rx_queue_id  rx (From Linux todataplane direction) queue id
 * @param[in]  weight       weight of the queue used in case of device is
 *                          configured for RR scheduling; priority in case of SP scheduling mode
 * @param[in]  size         size in number of packets; ignored in case of HWM-DM type of device
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to setup the size and weight of a rx-queue in a Linux Device
 */
 adk_status_t
adk_dev_net_rx_queue_setup(adk_sys_handle_t      sys_handle,
                             adk_dev_id_t    dev_id,
                             uint16_t               rx_queue_id,
                             uint8_t               weight,
                             uint16_t              size)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return(retValue);
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)|| (dev_id >= ADK_DEV_NET_ID_MAX)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter linux device id %d \n", dev_id);
        goto adk_dev_net_rx_queue_setup_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_rx_queue_setup_lock_return;
    }

    if(dev_pvt_info_p->device_configured !=1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is not yet configured; \n", dev_id);
        goto adk_dev_net_rx_queue_setup_lock_return;
    }

    if(dev_pvt_info_p->device_started !=0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is already started; \n", dev_id);
        goto adk_dev_net_rx_queue_setup_lock_return;
    }

    if(rx_queue_id+1 >= dev_pvt_info_p->n_rx_queues )
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid queue id %d  \n", rx_queue_id);
        goto adk_dev_net_rx_queue_setup_lock_return;
    }
    if (size> ADK_DEV_NET_QUEUE_SIZE_MAX)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid queue size  %d  \n", size);
        goto adk_dev_net_rx_queue_setup_lock_return;
    }

    dev_pvt_info_p->rx_q_info[rx_queue_id+1].size = size;
    dev_pvt_info_p->rx_q_info[rx_queue_id+1].weight= weight;
#if RDK_REL_MAJOR >= 13
    dev_pvt_info_p->rx_q_info[rx_queue_id+1].queue_id= rx_queue_id + 1;
    retValue = adk_dev_net_sdm_rx_queue_setup(sys_handle, dev_pvt_info_p,
                                                rx_queue_id + 1);
#else
    dev_pvt_info_p->rx_q_info[rx_queue_id+1].queue_id= rx_queue_id;
    retValue = adk_dev_net_sdm_rx_queue_setup(sys_handle, dev_pvt_info_p,
                                                rx_queue_id);
#endif
    if(retValue!=ADK_SUCCESS)
    {
        goto adk_dev_net_rx_queue_setup_lock_return;
    }
    dev_pvt_info_p->rx_q_info[rx_queue_id+1].setup_done = 1;

adk_dev_net_rx_queue_setup_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_rx_queue_setup_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );

}
/*
 * adk_dev_net_start()
 * @param[in]  sys_handle    ADK System Handle
 * @param[in]  dev_id       ADK Linux Device ID
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to start the given Linux Device
 */
 adk_status_t
adk_dev_net_start(adk_sys_handle_t      sys_handle,
                    adk_dev_id_t    dev_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_status_t lretValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;
    struct adk_netd_device_info adk_netd_dev_info;
    int err, i;
    uint64_t  temp;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return(retValue);
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)||
       (dev_id >= ADK_DEV_NET_ID_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter linux device id %d \n", dev_id);
        goto adk_dev_net_start_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, lretValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_start_lock_return;
    }

    /*Check Whether device is confgiured or not */
    if(dev_pvt_info_p->device_configured !=1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is not configured \n", dev_id);
        goto adk_dev_net_start_lock_return;
    }

    if(dev_pvt_info_p->device_started !=0)
    {
        retValue = ADK_SUCCESS;
        ADK_LOG_WARNING(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is already started; \n", dev_id);
        goto adk_dev_net_start_lock_return;
    }

    /* Check FD and open */
    if (dev_pvt_info_p->adk_netd_fd < 0)
    {
        dev_pvt_info_p->adk_netd_fd = open("/dev/" ADK_NETD_FILE, O_RDWR);
        if (dev_pvt_info_p->adk_netd_fd < 0) {
            retValue = ADK_E_NETD_FD_OPEN_FAILED;
            ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                            "Can not open /dev/%s\n", ADK_NETD_FILE);
            goto adk_dev_net_start_lock_return;
        }
    }


    if (!dev_pvt_info_p->fq_info.cppi_info.mz || !dev_pvt_info_p->aq_info.cppi_info.mz ||
        !dev_pvt_info_p->rx_q_info[0].cppi_info.mz ||
            !dev_pvt_info_p->tx_q_info[0].cppi_info.mz)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_start_lock_return;

    }

    memset(&adk_netd_dev_info, 0, sizeof(struct adk_netd_device_info));
    snprintf(adk_netd_dev_info.name, ADK_NETD_NAMESIZE,
            "adk_nd_%u", 0);

    err = ioctl(dev_pvt_info_p->adk_netd_fd, ADK_NETD_IOCTL_GETINFO,
            &adk_netd_dev_info);
    if (err < 0) {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "%s() Failed to fetch info from ADK NetDriver[%d]  \n",
            __func__, err);
        goto adk_dev_net_start_lock_return;
    }

    /*Check for NETD MD version mis-match */
    if (adk_netd_dev_info.intf_ver_mj != ADK_NETD_FP_INTF_VERSION_MAJOR) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_NETD_FP_INTF_MISMATCH);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "%s() NetDriver Interface mismatch \n", __func__);
        goto adk_dev_net_start_lock_return;
    }

    if(adk_netd_dev_info.poll_status == 0x1) {
        memset(&adk_netd_dev_info, 0, sizeof(struct adk_netd_device_info));
        snprintf(adk_netd_dev_info.name, ADK_NETD_NAMESIZE,
                "adk_nd_%u", 0);

        err = ioctl(dev_pvt_info_p->adk_netd_fd, ADK_NETD_IOCTL_CLEANUP,
                &adk_netd_dev_info);
        if (err < 0)
        {

            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                "%s() Failed to clean linux device with ADK NetDriver[%d]  \n",
                __func__, err);
            goto adk_dev_net_start_lock_return;
        }
    }

    memset(&adk_netd_dev_info, 0, sizeof(struct adk_netd_device_info));
    snprintf(adk_netd_dev_info.name, ADK_NETD_NAMESIZE,
            "adk_nd_%u", 0);
    adk_netd_dev_info.dm_type    =  ADK_NETD_DM_SDM;
    adk_netd_dev_info.num_rx_queues =  dev_pvt_info_p->n_rx_queues;
    adk_netd_dev_info.num_tx_queues =  dev_pvt_info_p->n_tx_queues;
    adk_netd_dev_info.mbuf_size = RTE_MBUF_DEFAULT_DATAROOM;
    if (ADK_ETH_BD_TX_METADATA_OFFSET == 0)
    {
        adk_netd_dev_info.md_offset = sizeof(struct rte_mbuf) +
            sizeof(struct rte_bridge_cppi_tx_metadata);
    } else {
        adk_netd_dev_info.md_offset = sizeof(struct rte_mbuf);
    }

    adk_netd_dev_info.fq_fifo.fifo_type = ADK_NETD_FIFO_FQ;
    adk_netd_dev_info.fq_fifo.fifo_addr = dev_pvt_info_p->fq_info.cppi_info.mz->phys_addr;
    adk_netd_dev_info.fq_fifo.burst_size= ADK_DEV_NET_CPPI_BURST_SIZE;
    adk_netd_dev_info.fq_fifo.fifo_size =  ADK_DEV_NET_QUEUE_SIZE_MAX;
    adk_netd_dev_info.fq_fifo.priority  = 0;

    adk_netd_dev_info.aq_fifo.fifo_type = ADK_NETD_FIFO_AQ;
    adk_netd_dev_info.aq_fifo.fifo_addr = dev_pvt_info_p->aq_info.cppi_info.mz->phys_addr;
    adk_netd_dev_info.aq_fifo.burst_size= ADK_DEV_NET_CPPI_BURST_SIZE;
    adk_netd_dev_info.aq_fifo.fifo_size =  ADK_DEV_NET_QUEUE_SIZE_MAX;
    adk_netd_dev_info.aq_fifo.priority  = 0;

    for(i=0;i< ADK_NETD_NUM_FIFO ;i++)
    /*for(i=0;i< 1;i++)*/
    {
        adk_netd_dev_info.tx_fifo[i].fifo_type = ADK_NETD_FIFO_TX;
        if (i< dev_pvt_info_p->n_tx_queues) {
            adk_netd_dev_info.tx_fifo[i].fifo_addr =
                dev_pvt_info_p->tx_q_info[i].cppi_info.mz->phys_addr;
        } else {
            adk_netd_dev_info.tx_fifo[i].fifo_addr =
                dev_pvt_info_p->tx_q_info[dev_pvt_info_p->n_tx_queues -1].cppi_info.mz->phys_addr;
        }
        adk_netd_dev_info.tx_fifo[i].burst_size= ADK_DEV_NET_CPPI_BURST_SIZE;
        adk_netd_dev_info.tx_fifo[i].fifo_size =  ADK_DEV_NET_QUEUE_SIZE_MAX;
        adk_netd_dev_info.tx_fifo[i].priority  = 0;

        adk_netd_dev_info.rx_fifo[i].fifo_type = ADK_NETD_FIFO_RX;
        if (i< dev_pvt_info_p->n_rx_queues) {
            adk_netd_dev_info.rx_fifo[i].fifo_addr =
                dev_pvt_info_p->rx_q_info[i].cppi_info.mz->phys_addr;
        } else {
            adk_netd_dev_info.rx_fifo[i].fifo_addr =
                dev_pvt_info_p->rx_q_info[dev_pvt_info_p->n_rx_queues-1].cppi_info.mz->phys_addr;
        }
        adk_netd_dev_info.rx_fifo[i].burst_size= ADK_DEV_NET_CPPI_BURST_SIZE;
        adk_netd_dev_info.rx_fifo[i].fifo_size =  ADK_DEV_NET_QUEUE_SIZE_MAX;
        adk_netd_dev_info.rx_fifo[i].priority  = 0;
    }


    err = ioctl(dev_pvt_info_p->adk_netd_fd, ADK_NETD_IOCTL_CREATE,
            &adk_netd_dev_info);
    if (err < 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "Registering linux device with ADK NetDriver is failed[%d], %ld  \n",
            err, sizeof(adk_netd_dev_info));
        goto adk_dev_net_start_lock_return;
    }

    temp = 0x1234;
    err = ioctl(dev_pvt_info_p->adk_netd_fd, ADK_NETD_IOCTL_START,
            &temp);
    if (err < 0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "start ADK NetDriver is failed[%d]  \n",
            err);
        goto adk_dev_net_start_lock_return;
    }


#ifdef ADK_HOST_PATH_USE_NETD
    if(adk_pal_dev_net_add(sys_handle,
                             dev_pvt_info_p->info.bd_info.bd_event_queue_id,
                             ADK_CPPI_BD_MD_OFFSET)!=ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not register Linux dev with ADK PAL \n");
        goto adk_dev_net_start_lock_return;
    }
#endif

    dev_pvt_info_p->device_started = 1;

adk_dev_net_start_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, lretValue);

adk_dev_net_start_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    if(lretValue!=ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "ADK SYS LOCK macro Erro");

        ADK_RETURN( lretValue );
    }
    ADK_RETURN( retValue );
}

/*
 * adk_dev_net_stop()
 * @param[in]  sys_handle    ADK System Handle
 * @param[in]  dev_id       ADK Linux Device ID
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to stop the given Linux Device
 */
adk_status_t
adk_dev_net_stop(adk_sys_handle_t      sys_handle,
                   adk_dev_id_t    dev_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;
    struct adk_netd_device_info adk_netd_dev_info;
    int32_t err;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return(retValue);
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)||
       (dev_id >= ADK_DEV_NET_ID_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter linux device id %d \n", dev_id);
        goto adk_dev_net_stop_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_stop_lock_return;
    }

    if(dev_pvt_info_p->device_started !=1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is not yet started; \n", dev_id);
        goto adk_dev_net_stop_lock_return;
    }

    /* Check FD and open */
    if (dev_pvt_info_p->adk_netd_fd < 0) {
        dev_pvt_info_p->adk_netd_fd = open("/dev/" ADK_NETD_FILE, O_RDWR);
        if (dev_pvt_info_p->adk_netd_fd < 0) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                            "Can not open /dev/%s\n", ADK_NETD_FILE);
            goto adk_dev_net_stop_lock_return;
        }
    }

    memset(&adk_netd_dev_info, 0, sizeof(struct adk_netd_device_info));
    snprintf(adk_netd_dev_info.name, ADK_NETD_NAMESIZE,
            "adk_nd_%u", 0);

    err = ioctl(dev_pvt_info_p->adk_netd_fd, ADK_NETD_IOCTL_CLEANUP,
            &adk_netd_dev_info);
    if (err < 0)
    {

        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "Failed to clean linux device with ADK NetDriver[%d]  \n",
            err);
        goto adk_dev_net_stop_lock_return;
    }

    dev_pvt_info_p->device_started = 0;

    if(!(dev_pvt_info_p->adk_netd_fd < 0))
    {
        close(dev_pvt_info_p->adk_netd_fd);
    }

    dev_pvt_info_p->adk_netd_fd = -1;

adk_dev_net_stop_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_stop_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}

/*
 * adk_dev_net_delete()
 * @param[in] sys_handle    ADK System Handle
 * @param[in] dev_id        ADK slow path Device-ID that is created previously.
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to delete a ADK Slowpath Device. Applicaiton is required to
 * stop this device before deleting it.
 */
adk_status_t
adk_dev_net_delete(adk_sys_handle_t sys_handle,
                     adk_dev_id_t    dev_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;
    adk_sys_info_t *sysInfo_p = (adk_sys_info_t *)sys_handle;
    int32_t         id;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return(retValue);
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)||
       (dev_id >= ADK_DEV_NET_ID_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter linux device id %d \n", dev_id);
        goto adk_dev_net_delete_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_delete_lock_return;
    }

    if(dev_pvt_info_p->device_started !=0)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is not stopped; \n", dev_id);
        goto adk_dev_net_delete_lock_return;
    }

    switch(dev_pvt_info_p->info.dev_type)
    {
    case ADK_DEV_NET_TYPE_NETD:
        retValue = adk_dev_net_sdm_delete(sys_handle, dev_pvt_info_p);
        if(retValue!= ADK_SUCCESS) {
            goto adk_dev_net_delete_lock_return;
        }
        break;
    default:
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid linux device DM Type \n");
        goto adk_dev_net_delete_lock_return;
        break;
    }

#ifdef ADK_QOS_ENABLE
    if (adk_qos_instance_shutdown(sys_handle, dev_id) !=  ADK_SUCCESS) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not un-register Linux dev with ADK QoS \n");
        goto adk_dev_net_delete_lock_return;
    }
#endif

    dev_list_p[dev_id] = NULL;

    memset(dev_pvt_info_p , 0, sizeof(adk_dev_net_pvt_info_t));

    /* Allocate memory for Linux Device*/
    ADK_NVM_FREE (dev_pvt_info_p);

    /*Free the ADK slowpath device id  */
    id = adk_rm_resource_free(sysInfo_p->sp_dev_ids_rm_hdl, dev_id);
    if(id == ADK_RM_RSRC_INVALID)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_RESOURCE_UNAVAILABLE);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
          "Failed to Free ADK Slowpath Device Id %d with error ADK_RM_RSRC_INVALID \n", dev_id);
        goto adk_dev_net_delete_lock_return;
    }

adk_dev_net_delete_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_delete_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}


/*
 * adk_dev_net_stats_get()
 *
 * @param[in]  sys_handle    ADK System Handle
 * @param[in]  dev_id       ADK Linux Device ID
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief API to print device stats
 */
 adk_status_t
adk_dev_net_stats_get(adk_sys_handle_t      sys_handle,
                             adk_dev_id_t    dev_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_dev_net_pvt_info_t *dev_pvt_info_p;

    /* Validate adk system handle */
    if (ADK_INVALID_SYS_HANDLE(sys_handle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_PREINIT_WRITE("[%s]: invalid system handle\n", __func__);
        return(retValue);
    }

    ADK_LOG_FUNC_ENTRY(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

    if((dev_id <= ADK_DEV_NET_ID_RSVD_MAX_PH)|| (dev_id >= ADK_DEV_NET_ID_MAX)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "invalid parameter linux device id %d \n", dev_id);
        goto adk_dev_net_stats_get_return;
    }

    ADK_ACQUIRE_SYS_LOCK(sys_handle, retValue);

    dev_pvt_info_p =  dev_list_p[dev_id];

    /* Check whether the device is created or not */
    if(dev_pvt_info_p==NULL)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device id %d is not created  \n", dev_id);
        goto adk_dev_net_stats_get_lock_return;
    }

    if(dev_pvt_info_p->device_configured !=1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is not yet configured; \n", dev_id);
        goto adk_dev_net_stats_get_lock_return;
    }

    if(dev_pvt_info_p->device_started !=1)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                          "linux device %d is already started; \n", dev_id);
        goto adk_dev_net_stats_get_lock_return;
    }

    retValue = adk_dev_net_sdm_stats(sys_handle, dev_pvt_info_p);

    if(retValue!=ADK_SUCCESS)
    {
        goto adk_dev_net_stats_get_lock_return;
    }

adk_dev_net_stats_get_lock_return:
    ADK_RELEASE_SYS_LOCK(sys_handle, retValue);

adk_dev_net_stats_get_return:
    ADK_LOG_FUNC_EXIT(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle));

_lock_release_failed:
_lock_acquire_failed:
    ADK_RETURN( retValue );
}

/*
 * adk_dev_net_sdm_create()
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief function to create a soft DM based ADK Linux Dev
 */
static adk_status_t
adk_dev_net_sdm_create(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t cppi_device_id;
    uint8_t cppi_inst_id;
    uint16_t cppi_fq_id = ADK_DEV_NET_CPPI_FQ_ID_BASE;
    uint16_t cppi_aq_id = ADK_DEV_NET_CPPI_AQ_ID_BASE;
    uint16_t cppi_rxq_id = ADK_DEV_NET_CPPI_RXQ_ID_BASE;
    uint16_t cppi_txq_id = ADK_DEV_NET_CPPI_TXQ_ID_BASE;
    int32_t     err;
    adk_pal_dp_stage_info_t *stage_info_p;

    cppi_inst_id   = dev_pvt_info_p->info.bd_info.bd_instance_id,
    cppi_device_id = dev_pvt_info_p->device_id_internal;

#if RDK_REL_MAJOR >= 13
    struct rte_event ev;

    struct rte_cppi_dev_config dev_config =
                    {CPPI_DEV_ND, ADK_DEV_NET_CPPI_BURST_SIZE};
    struct rte_cppi_queue_config free_queue_config =
                    { .queue_type       = CPPI_FIFO_FQ,
                      .mbuf_cache_sz    = 0};
    struct rte_cppi_queue_config alloc_queue_config =
                    { .queue_type       = CPPI_FIFO_AQ,
                      .mbuf_cache_sz    = ADK_DEV_NET_CPPI_QUEUE_CACHE_SZ};
    struct rte_bridge_cppi_queue_config rx_queue_config =
                    { .queue_type       = CPPI_FIFO_RXQ,
                      .servicing_weight= 1};

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_dev_configure",
                    "Add a CPPI bridge device.\n");
    err = rte_cppi_dev_configure(cppi_device_id, 1, 1, &dev_config);
    if (err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "failed to configure cppi device %d  error: %d\n",
                        cppi_device_id, err);
        goto adk_dev_net_sdm_create_return;
    }
    ADK_LOG_EXT_API(sys_handle,
            ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "rte_bridge_cppi_dev_add",
            "Add a CPPI device to BD instance.\n");
    err = rte_bridge_cppi_dev_add(dev_pvt_info_p->info.bd_info.bd_instance_id,
            cppi_device_id);
    if (err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "could not add a cppi device %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }
#ifdef ADK_HOST_PATH_USE_NETD
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_tx_cb_register",
                    "Register a tx-cb with CPPI device.\n");
    err = rte_bridge_cppi_tx_cb_register(cppi_inst_id,
                        cppi_device_id, adk_sys_dp_slowpath_protection, NULL);
    if (err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "could not register a tx-cb with cppi device[%d] %d\n",
                        cppi_device_id,  err);
        goto adk_dev_net_sdm_create_return;
    }
#endif

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_free_queue_setup",
                    "Setup free queue of CPPI device.\n");
    err = rte_cppi_free_queue_setup(cppi_device_id,
                              CPPI_FIFO_COUNT_MAX, &free_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_queue_info_get",
                    "Get device queue info from CPPI-BD.\n");
    err = rte_cppi_queue_info_get(cppi_device_id,
            cppi_fq_id, CPPI_FIFO_FQ, &(dev_pvt_info_p->fq_info.cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_free_queue_add",
                    "Add CPPI device queue to BD.\n");
    err = rte_bridge_cppi_free_queue_add(cppi_inst_id, cppi_device_id);
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add cppi device queue to BD %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->fq_info.cppi_device_id= cppi_device_id;
    dev_pvt_info_p->fq_info.cppi_queue_id = cppi_fq_id;
    dev_pvt_info_p->fq_info.queue_id      =0;

    alloc_queue_config.mempool_p = dev_pvt_info_p->info.pkt_mem_pool;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_alloc_queue_add",
                    "Setup CPPI DEV alloc queue.\n");
    err = rte_cppi_alloc_queue_setup(cppi_device_id,
                          CPPI_FIFO_COUNT_MAX, &alloc_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not setup cppi device queue %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_cppi_queue_info_get(cppi_device_id,
            0, CPPI_FIFO_AQ, &(dev_pvt_info_p->aq_info.cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_alloc_queue_add",
                    "Add CPPI device queue to BD.\n");
    err = rte_bridge_cppi_alloc_queue_add(cppi_inst_id, cppi_device_id);
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add cppi device queue to BD %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }
    dev_pvt_info_p->aq_info.cppi_device_id= cppi_device_id;
    dev_pvt_info_p->aq_info.cppi_queue_id = cppi_aq_id;
    dev_pvt_info_p->aq_info.queue_id      =0;

    /* ADK First stage event queue id is to be assigned here
     *
     * TBD
     * */
    retValue = adk_pal_dp_stage_info_get(ADK_ETH_ULI_NETD_FN_ID, &stage_info_p);
    if(retValue!=ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get event queue associated with ETH-NETD processing\n");
        goto adk_dev_net_sdm_create_return;
    }

    ev.queue_id = stage_info_p->event_q_id;
    ev.priority = stage_info_p->priority;
    ev.sched_type = stage_info_p->lb_type;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_rx_queue_setup",
                    "Setup  CPPI DEV RX queue .\n");
    err = rte_cppi_rx_queue_setup(cppi_device_id,
            cppi_rxq_id, CPPI_FIFO_COUNT_MAX);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not create cppi device RX queue%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->rx_q_info[0].cppi_device_id= cppi_device_id;
    dev_pvt_info_p->rx_q_info[0].cppi_queue_id = cppi_rxq_id;
    dev_pvt_info_p->rx_q_info[0].queue_id      =0;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_cppi_queue_info_get(cppi_device_id,
            cppi_rxq_id, CPPI_FIFO_RXQ, &(dev_pvt_info_p->rx_q_info[0].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_rx_queue_add",
                    "Add cppi device RX queue to BD.\n");
    err = rte_bridge_cppi_rx_queue_add(cppi_inst_id, cppi_device_id, cppi_rxq_id,
                &ev, &rx_queue_config);
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add cppi device queue to BD %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_tx_queue_setup",
                    "Setup CPPI DEV TX queue.\n");
    err = rte_cppi_tx_queue_setup(cppi_device_id,
            cppi_txq_id, CPPI_FIFO_COUNT_MAX);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not setup cppi device queue%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->tx_q_info[0].cppi_device_id= cppi_device_id;
    dev_pvt_info_p->tx_q_info[0].cppi_queue_id = cppi_txq_id;
    dev_pvt_info_p->tx_q_info[0].queue_id      =0;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_cppi_queue_info_get(cppi_device_id,
            cppi_txq_id, CPPI_FIFO_TXQ, &(dev_pvt_info_p->tx_q_info[0].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_tx_queue_add",
                    "Add CPPI device queue to BD.\n");
    err = rte_bridge_cppi_tx_queue_add(cppi_inst_id, cppi_device_id, cppi_txq_id);
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add cppi device queue to BD %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }
#else
    struct rte_bridge_cppi_dev_config dev_config =
                    {CPPI_DEV_ND, ADK_DEV_NET_CPPI_BURST_SIZE};
    struct rte_bridge_cppi_queue_config free_queue_config =
                    { .queue_type       = CPPI_FIFO_FQ,
                      .mbuf_cache_sz    = 0};
    struct rte_bridge_cppi_queue_config alloc_queue_config =
                    { .queue_type       = CPPI_FIFO_AQ,
                      .mbuf_cache_sz    = ADK_DEV_NET_CPPI_QUEUE_CACHE_SZ};
    struct rte_bridge_cppi_queue_config rx_queue_config =
                    { .queue_type       = CPPI_FIFO_RXQ,
                      .mbuf_cache_sz    = 0};
    struct rte_bridge_cppi_queue_config tx_queue_config =
                    { .queue_type       = CPPI_FIFO_TXQ,
                      .mbuf_cache_sz    = ADK_DEV_NET_CPPI_QUEUE_CACHE_SZ};


    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_dev_add",
                    "Add a CPPI bridge device.\n");
    err = rte_bridge_cppi_dev_add(cppi_inst_id,
                        cppi_device_id, &dev_config);
    if (err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "failed to create a cppi device %d  error: %d\n",
                        cppi_device_id, err);
        goto adk_dev_net_sdm_create_return;
    }

#ifdef ADK_QOS_GEN3_INTERFACE
#ifdef ADK_HOST_PATH_USE_NETD
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_tx_cb_register",
                    "Register a tx-cb with CPPI device.\n");
    err = rte_bridge_cppi_tx_cb_register(cppi_inst_id,
                        cppi_device_id, adk_sys_dp_slowpath_protection, NULL);
    if (err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "could not register a tx-cb with cppi device[%d] %d\n",
                        cppi_device_id,  err);
        goto adk_dev_net_sdm_create_return;
    }
#endif
#endif

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_add",
                    "Add a CPPI device queue to CPPI-BD.\n");
    err = rte_bridge_cppi_queue_add(cppi_inst_id, cppi_device_id,
                                    cppi_fq_id, &free_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get device queue info from CPPI-BD.\n");
    err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
            cppi_fq_id, &(dev_pvt_info_p->fq_info.cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->fq_info.cppi_device_id= cppi_device_id;
    dev_pvt_info_p->fq_info.cppi_queue_id = cppi_fq_id;
    dev_pvt_info_p->fq_info.queue_id      =0;

    alloc_queue_config.mempool_p = dev_pvt_info_p->info.pkt_mem_pool;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_add",
                    "Add a CPPI device queue to CPPI-BD.\n");
    err = rte_bridge_cppi_queue_add(cppi_inst_id, cppi_device_id,
                                    cppi_aq_id, &alloc_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
            cppi_aq_id, &(dev_pvt_info_p->aq_info.cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->aq_info.cppi_device_id= cppi_device_id;
    dev_pvt_info_p->aq_info.cppi_queue_id = cppi_aq_id;
    dev_pvt_info_p->aq_info.queue_id      =0;

    /* ADK First stage event queue id is to be assigned here
     *
     * TBD
     * */
    retValue = adk_pal_dp_stage_info_get(ADK_ETH_ULI_NETD_FN_ID, &stage_info_p);
    if(retValue!=ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get event queue associated with ETH-NETD processing\n");
        goto adk_dev_net_sdm_create_return;
    }

    rx_queue_config.ev.queue_id = stage_info_p->event_q_id;
    rx_queue_config.ev.priority = stage_info_p->priority;
    rx_queue_config.ev.sched_type = stage_info_p->lb_type;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_add",
                    "Add a CPPI device queue to CPPI-BD.\n");
    err = rte_bridge_cppi_queue_add(cppi_inst_id, cppi_device_id,
            cppi_rxq_id, &rx_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->rx_q_info[0].cppi_device_id= cppi_device_id;
    dev_pvt_info_p->rx_q_info[0].cppi_queue_id = cppi_rxq_id;
    dev_pvt_info_p->rx_q_info[0].queue_id      =0;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
            cppi_rxq_id, &(dev_pvt_info_p->rx_q_info[0].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_add",
                    "Add a CPPI device queue to CPPI-BD.\n");
    err = rte_bridge_cppi_queue_add(cppi_inst_id, cppi_device_id,
            cppi_txq_id, &tx_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD%d\n", err);
        goto adk_dev_net_sdm_create_return;
    }

    dev_pvt_info_p->tx_q_info[0].cppi_device_id= cppi_device_id;
    dev_pvt_info_p->tx_q_info[0].cppi_queue_id = cppi_txq_id;
    dev_pvt_info_p->tx_q_info[0].queue_id      =0;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
            cppi_txq_id, &(dev_pvt_info_p->tx_q_info[0].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_create_return;
    }
#endif
    dev_pvt_info_p->n_rx_queues = 1;
    dev_pvt_info_p->n_tx_queues = 1;


adk_dev_net_sdm_create_return:
    return ( retValue );
}

/*
 * adk_dev_net_sdm_configure()
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief function to create a soft DM based ADK Linux Dev
 */
static adk_status_t
adk_dev_net_sdm_configure(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p,
                        adk_dev_net_config_params_t *config_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t num_dev_ids = 0;
    uint8_t cppi_device_base_id;
    int32_t err;

    cppi_device_base_id =  dev_pvt_info_p->device_id_internal;

    num_dev_ids = config_p->num_rx_queues > config_p->num_tx_queues ?
                    config_p->num_rx_queues : config_p->num_tx_queues;
#if RDK_REL_MAJOR >= 13
    struct rte_cppi_dev_config dev_config =
                    {CPPI_DEV_ND, ADK_DEV_NET_CPPI_BURST_SIZE};

    /* If requested number of queues are more than ADK_DEV_NET_N_QUEUES_MAX-1
     * the return an error*/
    if (num_dev_ids > (ADK_DEV_NET_N_QUEUES_MAX-1) ) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "invalid number of queue requested %d\n", num_dev_ids );
        goto adk_dev_net_sdm_configure_return;
    }
    ADK_LOG_EXT_API(sys_handle,
            ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "rte_cppi_dev_configure",
            "configure a CPPI device.\n");
    err = rte_cppi_dev_configure(cppi_device_base_id, dev_pvt_info_p->n_rx_queues,
            dev_pvt_info_p->n_tx_queues, &dev_config);
    if (err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "could not configure a cppi device %d\n", err);
        goto adk_dev_net_sdm_configure_return;
    }
#else
    uint8_t cppi_device_id;
    struct rte_bridge_cppi_dev_config dev_config =
                    {CPPI_DEV_ND, ADK_DEV_NET_CPPI_BURST_SIZE};

    for(cppi_device_id = cppi_device_base_id+1;
        cppi_device_id <= (cppi_device_base_id + num_dev_ids);
        cppi_device_id++)
    {
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_dev_add",
                        "Add a CPPI bridge device.\n");
        err = rte_bridge_cppi_dev_add(dev_pvt_info_p->info.bd_info.bd_instance_id,
                            cppi_device_id, &dev_config);
        if (err) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                            "could not create a cppi device %d\n", err);
            goto adk_dev_net_sdm_configure_return;
        }
#ifdef ADK_QOS_GEN3_INTERFACE
#ifdef ADK_HOST_PATH_USE_NETD
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_tx_cb_register",
                        "Register a tx-cb with CPPI device.\n");
        err = rte_bridge_cppi_tx_cb_register(dev_pvt_info_p->info.bd_info.bd_instance_id,
                    cppi_device_id, adk_sys_dp_slowpath_protection, NULL);
        if (err) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                            "could not register a tx-cb with cppi device[%d] %d\n",
                            cppi_device_id, err);
            goto adk_dev_net_sdm_configure_return;
        }
#endif
#endif
    }
#endif
adk_dev_net_sdm_configure_return:
    return ( retValue );
}


/*
 * adk_dev_net_sdm_delete()
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief function to create a soft DM based ADK Linux Dev
 */
static adk_status_t
adk_dev_net_sdm_delete(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint16_t cppi_device_id;
    uint8_t qid = 0;
    int32_t err;
    uint8_t cppi_inst_id = dev_pvt_info_p->info.bd_info.bd_instance_id;

#if RDK_REL_MAJOR >= 13
    for(qid= 0; qid < dev_pvt_info_p->n_rx_queues; qid++)
    {
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_rx_queue_rem",
                        "Remove a RX queue from a CPPI-BD.\n");
        err = rte_bridge_cppi_rx_queue_rem(cppi_inst_id,
                                        dev_pvt_info_p->rx_q_info[qid].cppi_device_id,
                                        qid);
        if(err) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "failed to remove RX cppi queue[%d] dev[%d] BD[%d] err%d\n",
                    qid, dev_pvt_info_p->rx_q_info[qid].cppi_device_id,
                    cppi_inst_id, err);
            goto adk_dev_net_sdm_delete_return;
        }
    }

    for(qid= 0; qid < dev_pvt_info_p->n_tx_queues; qid++)
    {
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_tx_queue_rem",
                        "Remove a TX queue from a CPPI-BD.\n");
        err = rte_bridge_cppi_tx_queue_rem(cppi_inst_id,
                                        dev_pvt_info_p->tx_q_info[qid].cppi_device_id,
                                        qid);
        if(err) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "failed to remove a cppi TX queue[%d] dev[%d] BD[%d] err%d\n",
                    qid, dev_pvt_info_p->tx_q_info[qid].cppi_device_id,
                    cppi_inst_id, err);
            goto adk_dev_net_sdm_delete_return;
        }
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_alloc_queue_rem",
                    "Remove a Alloc queue from a CPPI-BD.\n");
    err = rte_bridge_cppi_alloc_queue_rem(cppi_inst_id,
                                    dev_pvt_info_p->aq_info.cppi_device_id);
    if(err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "failed to remove a cppi alloc queue[%d] dev[%d] BD[%d] err%d\n",
                dev_pvt_info_p->aq_info.cppi_queue_id,
                dev_pvt_info_p->aq_info.cppi_device_id, cppi_inst_id, err);
        goto adk_dev_net_sdm_delete_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_free_queue_rem",
                    "Re,pve a queue from a CPPI-BD.\n");
    err = rte_bridge_cppi_free_queue_rem(cppi_inst_id,
                                    dev_pvt_info_p->fq_info.cppi_device_id);
    if(err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "failed to remove a cppi queue[%d] dev[%d] BD[%d] err%d\n",
                dev_pvt_info_p->fq_info.cppi_queue_id,
                dev_pvt_info_p->fq_info.cppi_device_id, cppi_inst_id, err);
        goto adk_dev_net_sdm_delete_return;
    }

    cppi_device_id = dev_pvt_info_p->device_id_internal;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_dev_rem",
                    "Remove a CPPI-DEV from a CPPI-BD.\n");
    err = rte_bridge_cppi_dev_rem(cppi_inst_id, cppi_device_id);
    if(err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "failed to remove a cppi device dev[%d] BD[%d] err%d\n",
                cppi_device_id, cppi_inst_id, err);
        goto adk_dev_net_sdm_delete_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_dev_reset",
                    "Reset a CPPI-DEV .\n");
    err = rte_cppi_dev_reset(cppi_device_id);
    if(err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "failed to reset a cppi device dev[%d] err%d\n",
                cppi_device_id, err);
        goto adk_dev_net_sdm_delete_return;
    }
#else
    uint8_t n_cppi_devices;
    uint8_t cppi_device_base_id;
    for(qid= 0; qid < dev_pvt_info_p->n_rx_queues; qid++)
    {
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_queue_rem",
                        "Remove a queue from a CPPI-BD.\n");
        err = rte_bridge_cppi_queue_rem(cppi_inst_id,
                                        dev_pvt_info_p->rx_q_info[qid].cppi_device_id,
                                        dev_pvt_info_p->rx_q_info[qid].cppi_queue_id);
        if(err) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "failed to remove a cppi queue[%d] dev[%d] BD[%d] err%d\n",
                    dev_pvt_info_p->rx_q_info[qid].cppi_queue_id,
                    dev_pvt_info_p->rx_q_info[qid].cppi_device_id, cppi_inst_id, err);
            goto adk_dev_net_sdm_delete_return;
        }
    }

    for(qid= 0; qid < dev_pvt_info_p->n_tx_queues; qid++)
    {
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_queue_rem",
                        "Remove a queue from a CPPI-BD.\n");
        err = rte_bridge_cppi_queue_rem(cppi_inst_id,
                                        dev_pvt_info_p->tx_q_info[qid].cppi_device_id,
                                        dev_pvt_info_p->tx_q_info[qid].cppi_queue_id);
        if(err) {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "failed to remove a cppi queue[%d] dev[%d] BD[%d] err%d\n",
                    dev_pvt_info_p->tx_q_info[qid].cppi_queue_id,
                    dev_pvt_info_p->tx_q_info[qid].cppi_device_id, cppi_inst_id, err);
            goto adk_dev_net_sdm_delete_return;
        }
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_rem",
                    "Remove a queue from a CPPI-BD.\n");
    err = rte_bridge_cppi_queue_rem(cppi_inst_id,
                                    dev_pvt_info_p->aq_info.cppi_device_id,
                                    dev_pvt_info_p->aq_info.cppi_queue_id);
    if(err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "failed to remove a cppi queue[%d] dev[%d] BD[%d] err%d\n",
                dev_pvt_info_p->aq_info.cppi_queue_id,
                dev_pvt_info_p->aq_info.cppi_device_id, cppi_inst_id, err);
        goto adk_dev_net_sdm_delete_return;
    }

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_rem",
                    "Re,pve a queue from a CPPI-BD.\n");
    err = rte_bridge_cppi_queue_rem(cppi_inst_id,
                                    dev_pvt_info_p->fq_info.cppi_device_id,
                                    dev_pvt_info_p->fq_info.cppi_queue_id);
    if(err) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                "failed to remove a cppi queue[%d] dev[%d] BD[%d] err%d\n",
                dev_pvt_info_p->fq_info.cppi_queue_id,
                dev_pvt_info_p->fq_info.cppi_device_id, cppi_inst_id, err);
        goto adk_dev_net_sdm_delete_return;
    }


    cppi_device_base_id = dev_pvt_info_p->device_id_internal;

    n_cppi_devices = dev_pvt_info_p->n_rx_queues > dev_pvt_info_p->n_tx_queues ?
                    dev_pvt_info_p->n_rx_queues : dev_pvt_info_p->n_tx_queues;

    for(cppi_device_id = cppi_device_base_id;
        cppi_device_id < (cppi_device_base_id + n_cppi_devices);
        cppi_device_id++)
    {
        ADK_LOG_EXT_API(sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                        "rte_bridge_cppi_dev_free",
                        "Free a CPPI device.\n");
        err = rte_bridge_cppi_dev_free(dev_pvt_info_p->info.bd_info.bd_instance_id,
                            cppi_device_id);
        if(err)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
            ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                            "could not free a cppi device %d\n", err);
            goto adk_dev_net_sdm_delete_return;
        }
    }
#endif
adk_dev_net_sdm_delete_return:
    return ( retValue );
}



/*
 * adk_dev_net_sdm_tx_queue_setup()
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief function to create a soft DM based ADK Linux Dev
 */
static adk_status_t
adk_dev_net_sdm_tx_queue_setup(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p,
                        uint16_t                 queue_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t cppi_device_id;
    uint8_t cppi_inst_id   = dev_pvt_info_p->info.bd_info.bd_instance_id;
    int32_t     err;

#if RDK_REL_MAJOR >= 13
    cppi_device_id = dev_pvt_info_p->device_id_internal;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_tx_queue_setup",
                    "Setup  cppi device TX queue \n");
    err = rte_cppi_tx_queue_setup(cppi_device_id,
            queue_id, dev_pvt_info_p->tx_q_info[queue_id].size);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not setup cppi device queue %d\n", err);
        goto adk_dev_net_sdm_tx_queue_setup_return;
    }

    dev_pvt_info_p->tx_q_info[queue_id].cppi_device_id = cppi_device_id;
    dev_pvt_info_p->tx_q_info[queue_id].cppi_queue_id = queue_id;
    dev_pvt_info_p->tx_q_info[queue_id].queue_id      = queue_id;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_cppi_queue_info_get(cppi_device_id,
            queue_id, CPPI_FIFO_TXQ, &(dev_pvt_info_p->tx_q_info[queue_id].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_tx_queue_setup_return;
    }
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_tx_queue_add",
                    "Add CPPI device queue to BD.\n");
    err = rte_bridge_cppi_tx_queue_add(cppi_inst_id, cppi_device_id, queue_id);
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add cppi device queue to BD %d\n", err);
        goto adk_dev_net_sdm_tx_queue_setup_return;
    }
#else
    uint16_t cppi_txq_id = ADK_DEV_NET_CPPI_TXQ_ID_BASE;
    struct rte_bridge_cppi_queue_config tx_queue_config =
                    { .queue_type       = CPPI_FIFO_TXQ,
                      .mbuf_cache_sz    = ADK_DEV_NET_CPPI_QUEUE_CACHE_SZ};

    cppi_device_id = dev_pvt_info_p->device_id_internal + queue_id + 1;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_add",
                    "Add a cppi device queue to CPPI-BD.\n");
    err = rte_bridge_cppi_queue_add(cppi_inst_id, cppi_device_id,
            cppi_txq_id, &tx_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD%d\n", err);
        goto adk_dev_net_sdm_tx_queue_setup_return;
    }

    dev_pvt_info_p->tx_q_info[queue_id+1].cppi_device_id = cppi_device_id;
    dev_pvt_info_p->tx_q_info[queue_id+1].cppi_queue_id = cppi_txq_id;
    dev_pvt_info_p->tx_q_info[queue_id+1].queue_id      = queue_id + 1;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
    err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
            cppi_txq_id, &(dev_pvt_info_p->tx_q_info[queue_id+1].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_tx_queue_setup_return;
    }
#endif
adk_dev_net_sdm_tx_queue_setup_return:
    return ( retValue );
}


/*
 * adk_dev_net_sdm_rx_queue_setup()
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief function to create a soft DM based ADK Linux Dev
 */
static adk_status_t
adk_dev_net_sdm_rx_queue_setup(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p,
                        uint16_t                 queue_id)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint8_t cppi_device_id;
    uint8_t cppi_inst_id   = dev_pvt_info_p->info.bd_info.bd_instance_id;
    adk_pal_dp_stage_info_t *stage_info_p;
    int32_t err;

#if RDK_REL_MAJOR >= 13
    struct rte_event ev;
    struct rte_bridge_cppi_queue_config rx_queue_config =
                    { .queue_type       = CPPI_FIFO_RXQ,
                      .servicing_weight    = dev_pvt_info_p->rx_q_info[queue_id].weight
                    };

    cppi_device_id =  dev_pvt_info_p->device_id_internal;

    /* ADK First stage event queue id is to be assigned here
     *
     * TBD
     * */
    retValue = adk_pal_dp_stage_info_get(ADK_ETH_ULI_NETD_FN_ID, &stage_info_p);
    if(retValue!=ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get event queue associated with ETH-NETD processing\n");
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }

    ev.queue_id = stage_info_p->event_q_id;
    ev.priority = stage_info_p->priority;
    ev.sched_type = stage_info_p->lb_type;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_rx_queue_setup",
                    "Setup CPPI device RX queue.\n");
    err = rte_cppi_rx_queue_setup(cppi_device_id,
            queue_id, dev_pvt_info_p->rx_q_info[queue_id].size);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not setup cppi device RX queue devid %d, qid %d err %d\n",
                    cppi_device_id, queue_id, err);
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }

    dev_pvt_info_p->rx_q_info[queue_id].cppi_device_id = cppi_device_id;
    dev_pvt_info_p->rx_q_info[queue_id].cppi_queue_id = queue_id;
    dev_pvt_info_p->rx_q_info[queue_id].queue_id      = queue_id;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_queue_info_get",
                    "Get cppi device queue info.\n");
    err = rte_cppi_queue_info_get(cppi_device_id,
            queue_id, CPPI_FIFO_RXQ, &(dev_pvt_info_p->rx_q_info[queue_id].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_rx_queue_add",
                    "Add cppi device RX queue to BD.\n");
    err = rte_bridge_cppi_rx_queue_add(cppi_inst_id, cppi_device_id, queue_id,
                &ev, &rx_queue_config);
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add cppi device queue to BD %d\n", err);
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }
#else
    uint16_t cppi_rxq_id = ADK_DEV_NET_CPPI_RXQ_ID_BASE;

    struct rte_bridge_cppi_queue_config rx_queue_config =
                    { .queue_type       = CPPI_FIFO_RXQ,
                      .mbuf_cache_sz    = 0};

    cppi_device_id =  dev_pvt_info_p->device_id_internal + queue_id + 1;

    /* ADK First stage event queue id is to be assigned here
     *
     * TBD
     * */
    retValue = adk_pal_dp_stage_info_get(ADK_ETH_ULI_NETD_FN_ID, &stage_info_p);
    if(retValue!=ADK_SUCCESS)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get event queue associated with ETH-NETD processing\n");
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }

    rx_queue_config.ev.queue_id = stage_info_p->event_q_id;
    rx_queue_config.ev.priority = stage_info_p->priority;
    rx_queue_config.ev.sched_type = stage_info_p->lb_type;

    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_add",
                    "Add a CPPI device queue to CPPI-BD.\n");
    err = rte_bridge_cppi_queue_add(cppi_inst_id, cppi_device_id,
            cppi_rxq_id, &rx_queue_config);
    if(err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not add a cppi device queue to CPPI-BD devid %d, qid %d err %d\n", cppi_device_id, cppi_rxq_id, err);
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }

    dev_pvt_info_p->rx_q_info[queue_id+1].cppi_device_id = cppi_device_id;
    dev_pvt_info_p->rx_q_info[queue_id+1].cppi_queue_id = cppi_rxq_id;
    dev_pvt_info_p->rx_q_info[queue_id+1].queue_id      = queue_id + 1;
    ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_bridge_cppi_queue_info_get",
                    "Get cppi device queue info.\n");
    err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
            cppi_rxq_id, &(dev_pvt_info_p->rx_q_info[queue_id+1].cppi_info));
    if (err)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
        ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                    "could not get cppi device queue info %d\n", err);
        goto adk_dev_net_sdm_rx_queue_setup_return;
    }
#endif
adk_dev_net_sdm_rx_queue_setup_return:
    return ( retValue );
}


/*
 * adk_dev_net_sdm_stats()
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief
 */
static adk_status_t
adk_dev_net_sdm_stats(adk_sys_handle_t sys_handle,
                        adk_dev_net_pvt_info_t *dev_pvt_info_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    int32_t err, i;

    uint8_t cppi_device_id;

#if RDK_REL_MAJOR >= 13
    uint16_t q_type[] = {CPPI_FIFO_FQ,
                       CPPI_FIFO_AQ,
                       CPPI_FIFO_RXQ,
                       CPPI_FIFO_TXQ};

    struct rte_cppi_queue_info q_info;
    struct rte_cppi_queue_stats q_stats;
    int32_t j, n_rx_qs;

    cppi_device_id = dev_pvt_info_p->device_id_internal;
    n_rx_qs = dev_pvt_info_p->n_rx_queues;
    for(j = 0; j < n_rx_qs; j++)
    {
        for(i=0;i<4;i++)
        {
            if (j && i <= 1)
                continue;
            ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_queue_stats_get",
                    "Get stats from CPPI-queue.\n");
            err = rte_cppi_queue_stats_get(cppi_device_id,
                    j, q_type[i], &q_stats);
            if(err)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "could not get stats for CPPI-QUEUE%d\n", err);
                goto adk_dev_net_sdm_stats_return;
            }

            ADK_LOG_EXT_API(sys_handle,
                    ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                    "rte_cppi_queue_info_get",
                    "Get CPPI device queue info.\n");
            err = rte_cppi_queue_info_get(cppi_device_id,
                    j, q_type[i], &q_info);
            if (err)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "could not get cppi device queue info %d\n", err);
                goto adk_dev_net_sdm_stats_return;
            }

            switch(q_info.queue_type){
                case CPPI_FIFO_AQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_AQ");
                    break;
                case CPPI_FIFO_FQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_FQ");
                    break;
                case CPPI_FIFO_RXQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_RXQ");
                    break;
                case CPPI_FIFO_TXQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_TXQ");
                    break;
                default:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "UNKNOWN");
                    break;
            }
            ADK_INFO_WRITE("nb_requested   ::%lu\n", q_stats.nb_requested);
            ADK_INFO_WRITE("nb_processed   ::%lu\n", q_stats.nb_processed);
            ADK_INFO_WRITE("nb_error       ::%lu\n", q_stats.nb_error);
            ADK_INFO_WRITE("free_count     ::%u\n", q_info.free_count);
            ADK_INFO_WRITE("used_count     ::%u\n", q_info.used_count);
            ADK_INFO_WRITE("write_offset   ::%u\n", q_info.write);
            ADK_INFO_WRITE("read_offset    ::%u\n", q_info.read);
        }
    }
#else
    uint16_t q_id[] = {ADK_DEV_NET_CPPI_RXQ_ID_BASE,
                        ADK_DEV_NET_CPPI_TXQ_ID_BASE,
                        ADK_DEV_NET_CPPI_FQ_ID_BASE,
                        ADK_DEV_NET_CPPI_AQ_ID_BASE};

    struct rte_bridge_cppi_queue_info q_info;
    uint8_t cppi_inst_id   = dev_pvt_info_p->info.bd_info.bd_instance_id;
    struct rte_cppi_queue_stats q_stats;

    cppi_device_id = dev_pvt_info_p->device_id_internal;
    for(;cppi_device_id < (dev_pvt_info_p->device_id_internal + ADK_DEV_NET_N_QUEUES_MAX);
           cppi_device_id++)
    {
        ADK_INFO_WRITE("cppi_device_id: %d\n",cppi_device_id);
        for(i=0;i<4;i++)
        {
            if(cppi_device_id > dev_pvt_info_p->device_id_internal && i > 1)
                continue;
            ADK_LOG_EXT_API(sys_handle,
                            ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                            "rte_bridge_cppi_get_stats",
                            "Get stats from CPPI-BD.\n");
            err = rte_bridge_cppi_get_stats(cppi_inst_id, cppi_device_id,
                    q_id[i], &q_stats);
            if(err)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "could not get stats from CPPI-BD%d\n", err);
                goto adk_dev_net_sdm_stats_return;
            }

            ADK_LOG_EXT_API(sys_handle,
                            ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                            "rte_bridge_cppi_queue_info_get",
                            "Get CPPI device queue info.\n");
            err = rte_bridge_cppi_queue_info_get(cppi_inst_id, cppi_device_id,
                    q_id[i], &q_info);
            if (err)
            {
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INTERNAL_ERROR);
                ADK_LOG_ERROR (sys_handle, ADK_SYS_MODULE_HANDLE_GET (sys_handle),
                        "could not get cppi device queue info %d\n", err);
                goto adk_dev_net_sdm_stats_return;
            }

            switch(q_info.queue_type){
                case CPPI_FIFO_AQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_AQ");
                    break;
                case CPPI_FIFO_FQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_FQ");
                    break;
                case CPPI_FIFO_RXQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_RXQ");
                    break;
                case CPPI_FIFO_TXQ:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "CPPI_FIFO_TXQ");
                    break;
                default:
                    ADK_INFO_WRITE("\n Queue TYPE   ::%s\n", "UNKNOWN");
                    break;
            }
            ADK_INFO_WRITE("nb_requested   ::%lu\n", q_stats.nb_requested);
            ADK_INFO_WRITE("nb_processed   ::%lu\n", q_stats.nb_processed);
            ADK_INFO_WRITE("nb_error       ::%lu\n", q_stats.nb_error);
            ADK_INFO_WRITE("free_count     ::%u\n", q_info.free_count);
            ADK_INFO_WRITE("used_count     ::%u\n", q_info.used_count);
            ADK_INFO_WRITE("write_offset   ::%u\n", q_info.write);
            ADK_INFO_WRITE("read_offset    ::%u\n", q_info.read);

        }
    }
#endif
adk_dev_net_sdm_stats_return:
    return ( retValue );
}

/*
 * adk_dev_net_sdm_config_params_validate()
 *
 * @return Returns ADK_SUCCESS if success, error code otherwise.
 *
 * @brief function to validate the soft DM configuration parameters
 */
adk_status_t
adk_dev_net_sdm_config_params_validate(adk_sys_handle_t sys_handle,
                              adk_dev_net_config_params_t *config_p)
{
    adk_status_t retValue = ADK_SUCCESS;

    if((config_p->num_rx_queues > ADK_DEV_NET_SDM_RX_QUEUES_MAX ) ||
      (config_p->num_tx_queues > ADK_DEV_NET_SDM_TX_QUEUES_MAX )) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNSUPPORTED);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "un supported number of rx queues[%d] / tx queues[%d] are requested\n",
            config_p->num_rx_queues, config_p->num_tx_queues);
        return ( retValue );
    }

    if((config_p->rx_sched_mode != ADK_DEV_NET_SCHED_MODE_RR ) ||
      (config_p->tx_sched_mode != ADK_DEV_NET_SCHED_MODE_RR)) {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNSUPPORTED);
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
            "un supported scheduling mode \n");
        return ( retValue );
    }

    return ( retValue );

}

/*!  @} end addtogroup _sifFunctions
 *
 * @} end addtogroup _sif
 */
