/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_list.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifFunctions  SIF Functions
 *     @{
 *
 * @brief  
 *
 * This file contains generic linked list APIs
 *
 ***********************************************************************/

#include <adk_include.h>
#include <adk_list.h>

/*!*****************************************************************************
 * adk_list_new() -- Create and initialize a new list head
 *
 * Args:
 *      cmp_f -- function to compare two items
 *      find_f -- function to match entry with generic input
 *      destroy_f -- function for destroying an item
 *      max_item -- Maximum number of items that can be added to the list
 *      newlist  -- Newly created/initialized struct
 *
 * Returns:
 *   ADK_SUCCESS  on success
 *   The new list is returned in the function parameter newlist.
 *   Following error codes on failure
 *      1. ADK_E_MEM_UNAVAILABLE     if not enough memory
 *
 * Strategy:
 *  (1) Allocate mem for struct
 *  (2) Initialize
 *
 *  Note:
 *
 *  If the max_item is passed as zero, it indicates no limit on number of
 *  items that can be added to the list.
 */

adk_status_t
adk_list_new(
    const adk_list_cmp_f       *cmp_fp,
    const adk_list_find_f      *find_fp,
    const adk_list_update_f    *update_fp,
    const adk_list_destroy_f   *destroy_fp,
    const adk_list_dump_f      *dump_fp,
    const uint32_t          max_items,
    adk_linkedList_t          **newlist
    )
{
    /*! Step 1: Allocate memory */
    *newlist = (adk_linkedList_t *)ADK_NVM_CALLOC(ADK_SYS_MODULE_ID, 0, 1, sizeof(adk_linkedList_t));
    if (ADK_INVALID_PTR(*newlist))
    {
        return ADK_E_MEM_UNAVAILABLE;
    }

    (*newlist)->lhf_cmp = cmp_fp;
    (*newlist)->lhf_find = find_fp;
    (*newlist)->lhf_update = update_fp;
    (*newlist)->lhf_destroy = destroy_fp;
    (*newlist)->lhf_dump = dump_fp;

    (*newlist)->lh_max = max_items;

    /* ADK_MUTEX_INIT_DEFAULT(&((*newlist)->lh_mutex)); */

    /*! Step 3: Set the cookie */
    (*newlist)->cookie = ADK_SIF_LINKED_LIST_COOKIE;

    /*! Step 4: Return success */
    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_flush() -- Clear all entries from a list
 *
 * Args:
 *    list -- the list to clear
 *
 * Returns:
 *    ADK_SUCCESS on success
 *    Following error codes on failure
 1. ADK_E_LIST_INVALID_LIST  if invalid list
 *
 * Do:
 *    Loop on each item, removing as we go
 */

adk_status_t
adk_list_flush(
    adk_linkedList_t   *list_p
    )
{
    adk_list_item_t* li;
    adk_list_item_t* li_next;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Walk the list */
    li = list_p->lh_head;
    while (ADK_VALID_PTR(li))
    {
        /*! Destroy item */
        if (ADK_VALID_PTR(list_p->lhf_destroy))
        {
            (*list_p->lhf_destroy)(li->li_data);
        } else
        {
            ADK_NVM_FREE(li->li_data);
        }

        /*! Move to next */
        li_next = li->li_next;

        /*! Free item container */
        ADK_NVM_FREE(li);
        li = li_next;
    }

    /*! Step 3: Update bits and pieces */
    list_p->lh_head = NULL;
    list_p->lh_tail = NULL;
    list_p->lh_cur = NULL;
    list_p->lh_atEnd = FALSE;
    list_p->lh_size = 0;

    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_append() -- Add an item at the end of a list
 *
 * Args:
 *    list_p -- the list to update
 *    data -- the data to add
 *
 * Returns:
 *    ADK_SUCCESS on success
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 *      2. ADK_E_LIST_RANGE         if maximum number of items limit reached for
 *                                 the list
 *      3. ADK_E_MEM_UNAVAILABLE           if not enough memory
 *
 * Do:
 *    Allocate new storage and tack on to end
 */

adk_status_t
adk_list_append(
    adk_linkedList_t   *list_p,
    const void         *data
    )
{
    adk_list_item_t* li;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! If maximum limit reached */
    if (list_p->lh_max > 0)
    {
        if (list_p->lh_size == list_p->lh_max)
        {
            return ADK_E_LIST_RANGE;
        }
    }

    /*! Step 2: Allocate space for item */
    li = (adk_list_item_t *)ADK_NVM_CALLOC(ADK_SYS_MODULE_ID, 0, 1, sizeof(adk_list_item_t));
    if (ADK_INVALID_PTR(li))
    {
        return(ADK_E_MEM_UNAVAILABLE);
    }

    /*! Step 3: Fill in item */
    li->li_prev = list_p->lh_tail;
    li->li_next = NULL;
    li->li_data = (void *)data;

    /*! Step 4: Update list */
    if (ADK_VALID_PTR(list_p->lh_tail))
    {
        list_p->lh_tail->li_next = li;
    }

    list_p->lh_tail = li;
    /* Set the head if it isn't set */
    if (ADK_INVALID_PTR(list_p->lh_head))
    {
        list_p->lh_head = li;
    }

    list_p->lh_size++;

    return(ADK_SUCCESS);
}


/*!*****************************************************************************
 * adk_list_prepend() -- Add an item at the head of a list
 *
 * Args:
 *    list_p -- the list to update
 *    data -- the data to insert
 *
 * Returns:
 *    ADK_SUCCESS on success
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 *      2. ADK_E_LIST_RANGE         if maximum number of items limit reached for
 *                                 the list
 *      3. ADK_E_MEM_UNAVAILABLE           if not enough memory
 *
 */

adk_status_t
adk_list_prepend(
    adk_linkedList_t* list_p,
    const void* data
    )
{
    adk_list_item_t* li;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! If maximum limit reached */
    if (list_p->lh_max > 0)
    {
        if (list_p->lh_size == list_p->lh_max)
        {
            return ADK_E_LIST_RANGE;
        }
    }

    /*! Step 2: Allocate space */
    li = (adk_list_item_t *)ADK_NVM_CALLOC(ADK_SYS_MODULE_ID, 0, 1, sizeof(adk_list_item_t));
    if (ADK_INVALID_PTR(li))
    {
        return(ADK_E_MEM_UNAVAILABLE);
    }

    /*! Step 3: Fill in item */
    li->li_prev = NULL;
    li->li_next = list_p->lh_head;
    li->li_data = (void *)data;

    /*! Step 4: Update list */
    if (ADK_VALID_PTR(list_p->lh_head))
    {
        list_p->lh_head->li_prev = li;
    }

    list_p->lh_head = li;
    /* Set the tail if it isn't set */
    if (ADK_INVALID_PTR(list_p->lh_tail))
    {
        list_p->lh_tail = li;
    }

    list_p->lh_size++;

    return(ADK_SUCCESS);
}


/*!*****************************************************************************
 * adk_list_count() -- Count items in a list
 *
 * Args:
 *    list_p -- list whose items we are to count
 *
 * Returns:
 *    ADK_SUCCESS on success and list item count is returned in the function
 *    parameter "count".
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 *      2. ADK_E_INVALID_PARAMETER      if invalid(null) parameter
 */

adk_status_t
adk_list_count(
    const adk_linkedList_t* list_p,
    uint32_t* count
    )
{
    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }
    if (ADK_INVALID_PTR(count))
    {
        return ADK_E_INVALID_PARAMETER;
    }

    /*! Step 2: Return */
    *count = list_p->lh_size;
    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_wind() -- Move a list to the end
 *
 * Args:
 *    list_p --list to be wound
 *
 * Returns:
 *    ADK_SUCCESS on success
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 */

adk_status_t
adk_list_wind(
    adk_linkedList_t* list_p
    )
{
    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Update */
    list_p->lh_cur = NULL;
    list_p->lh_atEnd = TRUE;

    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_rewind() -- Move a list to the beginning
 *
 * Args:
 *    list_p --list to be rewound
 *
 * Returns:
 *    ADK_SUCCESS on success
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 */

adk_status_t
adk_list_rewind(
    adk_linkedList_t* list_p
    )
{
    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Update */
    list_p->lh_cur = NULL;
    list_p->lh_atEnd = FALSE;

    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_sort() -- sort a list
 *
 * Args:
 *    list_p -- list to be sorted
 *
 * Returns:
 *    NPRM_ERROR  (Not implemented as present).
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 */

adk_status_t
adk_list_sort(
    const adk_linkedList_t* list_p
    )
{
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    return ADK_E_UNSUPPORTED; /*! Unimplemented yet */
}


/*!*****************************************************************************
 * adk_list_get_next() -- Return the next item in a list
 *
 * Args:
 *    list_p -- the list whose next item we want
 *
 * Returns:
 *    ADK_SUCCESS on successful retrieval and list item is returned in the
 *    function parameter "listItem".
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 2. ADK_E_LIST_RANGE         if at the end of list
 *
 * Notes:
 *    This is probably _not_ what you want.  See adk_list_get_next_data().
 */

adk_status_t
adk_list_get_next(
    adk_linkedList_t* list_p,
    adk_list_item_t** listItem,
    bool          wrapFlag
    )
{
    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Check for degenerate cases when lh_cur isn't set */
    if (ADK_INVALID_PTR(list_p->lh_cur))
    {
        if (list_p->lh_atEnd == TRUE)
        {
            if (wrapFlag == FALSE)
            {
                /*! At end of list */
                *listItem = NULL;

                return ADK_E_LIST_RANGE;
            } else
            {
                /*
                 * Since current isn't set and were not at the end, set current
                 * to head
                 */
                list_p->lh_cur = list_p->lh_head;
            }
        } else
        {
            if (ADK_INVALID_PTR(list_p->lh_head))
            {
                return ADK_E_LIST_NOT_FOUND;
            }

            /*
             * Since current isn't set and were not at the end, set current
             * to head
             */
            list_p->lh_cur = list_p->lh_head;
        }
    } else 
    {

        /*! Step 3: Are we at the last item? */
        if ((list_p->lh_cur == list_p->lh_tail) ||
            (list_p->lh_cur->li_next == NULL))
        {
            if (wrapFlag == FALSE)
            {
                /*! already at end */
                list_p->lh_cur = NULL;
                list_p->lh_atEnd = TRUE;

                return ADK_E_LIST_NOT_FOUND;
            } else
            {
                /* wrap back to the beginning */
                list_p->lh_cur = list_p->lh_head;
            }
        } else
        {
            /*! Step 4: Must be somewhere in the middle */
            list_p->lh_cur = list_p->lh_cur->li_next;
        }
    }

    *listItem = list_p->lh_cur;

    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_get_prev() -- return the previous item in a list
 *
 * Args:
 *    list_p -- the list whose previous item we want
 *
 * Returns:
 *    ADK_SUCCESS on successful retrieval and list item is returned in the
 *    function parameter "listItem".
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 2. ADK_E_LIST_RANGE         if at the beginning of list
 *    Error code on failure
 *
 * Notes:
 *    This is probably _not_ what you want.  See adk_list_prev_data().
 */

adk_status_t
adk_list_get_prev(
    adk_linkedList_t* list_p,
    adk_list_item_t** listItem,
    bool          wrapFlag
    )
{
    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Check for degenerate cases when lh_cur isn't set */
    if (ADK_INVALID_PTR(list_p->lh_cur))
    {
        if (list_p->lh_atEnd == TRUE)
        {
            /*! At end of list */
            list_p->lh_cur = list_p->lh_tail;
        } else
        {
            /*! At beginning of the list */
            *listItem = NULL;

            return ADK_E_LIST_RANGE;
        }
    } else
    {
        /*! Step 3: Are we at the first item? */
        if ((list_p->lh_cur == list_p->lh_head) ||
            (list_p->lh_cur->li_prev == NULL))
        {
            if (wrapFlag == FALSE)
            {
                /*! Already at beginning */
                list_p->lh_cur = NULL;

                return ADK_E_LIST_RANGE;
            } else
            {
                /*! Wrap from the beginning of the list to the end */
                list_p->lh_cur = list_p->lh_tail;
            }
        } else
        {
            /*! Step 4: Must be somewhere in the middle */
            list_p->lh_cur = list_p->lh_cur->li_prev;
        }
    }

    list_p->lh_atEnd = FALSE;
    *listItem = list_p->lh_cur;

    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_get_next_data() -- return the data of the next item
 *
 * Args:
 *    list_p -- the list to search
 *    data
 *    wrapFlag
 *
 * Returns:
 *    ADK_SUCCESS on successful retrieval and the data in the list item is
 *    returned in the function parameter "data".
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 2. ADK_E_LIST_RANGE         if at the end of list
 */

adk_status_t
adk_list_get_next_data(
    adk_linkedList_t*   list_p,
    void**              data,
    bool          wrapFlag
    )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_list_item_t *li;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Find next item */
    retValue = adk_list_get_next(list_p, &li, wrapFlag);
    if (retValue != ADK_SUCCESS)
    {
        *data = NULL;
    } else
    {
        /*! Step 3: Return data */
        *data = li->li_data;
    }

    return retValue;
}


/*!*****************************************************************************
 * adk_list_get_prev_data() -- return the data of the prev item
 *
 * Args:
 *    list_p -- the list to search
 *
 * Returns:
 *    ADK_SUCCESS on successful retrieval and the data in the list item is
 *    returned in the function parameter "data".
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list_p
 2. ADK_E_LIST_RANGE         if at the beginning of list
 */

adk_status_t
adk_list_get_prev_data(
    adk_linkedList_t*   list_p,
    void**              data,
    bool          wrapFlag
    )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_list_item_t *li;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Find prev item */
    retValue = adk_list_get_prev(list_p, &li, wrapFlag);
    if (retValue != ADK_SUCCESS)
    {
        *data = NULL;
    } else
    {
        /*! Step 3: return data */
        *data = li->li_data;
    }


    return retValue;
}


/*!*****************************************************************************
 * adk_list_find() -- find data in a list by user's find func
 *
 * Args:
 *    list_p --  the list to search
 *    data --  the data key
 *    extra -- extra user info (passed opaquely)
 *    listData -- the data in the list item if found(output parameter)
 *
 * Returns:
 *    ADK_SUCCESS on successful retrieval and the data in the list item is
 *    returned in the function parameter "listData".
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 2. ADK_E_LIST_NOT_FOUND      if item not found
 */

adk_status_t
adk_list_find(
    const adk_linkedList_t *list_p,
    const void             *data,
    const void             *extra,
    void                  **listData
    )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_list_item_t* li;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    if (ADK_INVALID_PTR(listData))
    {
        return ADK_E_INVALID_PARAMETER;
    }
    /*! Step 2: Loop on all nodes */
    li = list_p->lh_head;
    while (ADK_VALID_PTR(li))
    {
        if (list_p->lhf_find)
        {
            /*! User defined match function */
            retValue = (*list_p->lhf_find)(li->li_data, data, extra);
            if (retValue == ADK_SUCCESS)
            {
                *listData = li->li_data;
                return retValue;
            }
        } else
        {
            /*! No user defined match function, just simple match */
            if (li->li_data == data)
            {
                *listData = li->li_data;
                return retValue;
            }
        }
        li = li->li_next;
    }

    if (listData != NULL)
    {
        *listData = NULL;
    }

    return ADK_E_LIST_NOT_FOUND;
}


/*!*****************************************************************************
 * adk_list_update_data() -- Updates data of a particular item in the list
 *
 * Args:
 *    list_p --  the list to update
 *    list data --  data in list that needs to be updated
 *    new data -- updated data from user which is updated transparently in the list
 *
 * Returns:
 *    ADK_SUCCESS on successful updation.
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 2. ADK_E_LIST_NOT_FOUND      if item not found
 3. ADK_E_INVALID_PARAMETER
 */

adk_status_t
adk_list_update_data(
    const adk_linkedList_t* list_p,
    const void ** listData,
    const void* newData
    )
{
    adk_status_t retValue = ADK_SUCCESS;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Sanity check */
    if ((ADK_INVALID_PTR(listData)) || (ADK_INVALID_PTR(newData)))
    {
        return ADK_E_INVALID_PARAMETER;
    }

    /*! Step 4: Update data */
    if (list_p->lhf_update)
    {
        /*! User defined update function */      
        retValue = (*list_p->lhf_update)(listData, newData);
    }else
    {
        /*!
         * No user defined update function, simple pointer assignment
         *
         * Need to free the old listData block before the pointer assignment
         * so we don't have a memory leak.
         */
        ADK_NVM_FREE(*listData);

        *listData = newData; 
    }

    return retValue;
}


/*!*****************************************************************************
 * adk_list_remove() -- remove an item from a list
 *
 * Args:
 *    list_p -- the list to update
 *    data -- the data of the item to remove
 *
 * Returns:
 *    ADK_SUCCESS on successful retrieval
 *    Following error codes on failure
 *      1. ADK_E_LIST_INVALID_LIST  if invalid list
 2. ADK_E_LIST_NOT_FOUND      if item not found
 */

adk_status_t
adk_list_remove(
    adk_linkedList_t* list_p,
    const void* data
    )
{
    adk_list_item_t *li, *li_match = NULL;

    /*! Step 1: Sanity check */
    if (ADK_LIST_INVALID(list_p))
    {
        return ADK_E_LIST_INVALID_LIST;
    }

    /*! Step 2: Loop on all nodes */
    li = list_p->lh_head;
    while (ADK_VALID_PTR(li))
    {
        if (li->li_data == data)
        {
            li_match = li;
            break;
        }
        li = li->li_next;
    }

    /*! Step 3: If not found, bail */
    if (ADK_INVALID_PTR(li_match))
    {
        return ADK_E_LIST_NOT_FOUND;
    }

    /*! Step 4: Free up resources */
    if (ADK_VALID_PTR(list_p->lhf_destroy))
    {
        (*list_p->lhf_destroy)(li_match->li_data);
    } else
    {
        ADK_NVM_FREE(li_match->li_data);
    }

    /*! Step 5: Remove -- carefully */
    if (list_p->lh_cur == li_match)
    {
        list_p->lh_cur = li_match->li_next;
    }

    if (list_p->lh_head == li_match)
    {
        if (list_p->lh_tail == li_match)
        {
            /*! Single-entry list */
            list_p->lh_head = NULL;
            list_p->lh_tail = NULL;
            list_p->lh_cur  = NULL;
            list_p->lh_atEnd = FALSE;
        } else
        {
            /*! At head of multi-item list */
            list_p->lh_head->li_next->li_prev = NULL;
            list_p->lh_head = list_p->lh_head->li_next;
        }
    } else if (list_p->lh_tail == li_match)
    {
        /*! At tail of multi-item list */
        list_p->lh_tail->li_prev->li_next = NULL;
        list_p->lh_tail = list_p->lh_tail->li_prev;
    } else
    {
        /*! In middle somewhere */
        li_match->li_prev->li_next = li_match->li_next;
        li_match->li_next->li_prev = li_match->li_prev;
    }

    /*! Step 6: Free our container */
    ADK_NVM_FREE(li_match);
    list_p->lh_size--;

    return ADK_SUCCESS;
}


/*!*****************************************************************************
 * adk_list_destroy() -- Completely destroy a list
 *
 * Args:
 *   list_p -- the list to destroy
 *
 * Do:
 *   Flush, then free main struct
 *
 * Returns:
 *    ADK_SUCCESS on successful destroy
 *    Following error codes on failure
 *      1. NPRM_BAD_PARAMETER      if bad(null) parameter
 *      2. ADK_E_LIST_INVALID_LIST  if invalid list
 */

adk_status_t
adk_list_destroy(
    adk_linkedList_t **list_p
    )
{
    adk_status_t retValue;

    /*! Step 1: Sanity check */
    if (ADK_INVALID_PTR(list_p))
    {
        return ADK_E_INVALID_PARAMETER;
    }

    /*! Step 2: Flush list */
    retValue = adk_list_flush(*list_p);
    if (retValue != ADK_SUCCESS)
    {
        return retValue;
    }

    /*! Step 3: Free main struct */
    ADK_NVM_FREE(*list_p);

    return ADK_SUCCESS;
}

adk_status_t
adk_list_dump(adk_linkedList_t *list_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_list_item_t    *list_item;

    list_item = list_p->lh_head;

    while(list_item != NULL) {
        /*! User defined dump function */
        retValue = (list_p->lhf_dump)(list_item->li_data);
        list_item = list_item->li_next;
    }

    return retValue;
}


/*!    end addtogroup _sifFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */
