/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_module_apis.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifFunctions  SIF Functions
 *     @{
 *
 * @brief  
 *
 * This file contains global ADK APIs to communicate with ADK modules.
 *
 ***********************************************************************/

#include <adk_include.h>

/*
 * =============================================================
 *
 * @param[in]    moduleHandle    handle for the module for which 
 * the set function is called.
 *
 * @param[in]    *inParams_p    pointer to the data structure that
 * contains the input parameters for the set operation.
 *
 * @retval    The return values here are determined the individual 
 * modules.
 *
 * @brief  
 *
 * This function provides a generic set interface for the application.
 * The caller has to specify the module handle as one of the input 
 * parameter. The function internally calls the registered set function
 * of the appropriate module.
 *
 * =============================================================
 */
adk_status_t
adk_modify (
       adk_module_handle_t moduleHandle,
       const void *inParams_p )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_module_info_t *moduleInfo_p;

    if ( !ADK_VALID_HANDLE (moduleHandle) ||
            !ADK_VALID_PTR (inParams_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR (NULL, moduleHandle,
                "Failed to set attribute. invalid input handle\n");
    }
    else
    {
        moduleInfo_p = (adk_module_info_t *)moduleHandle;

        if (adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleModify)
        {
            retValue = 
                (*(adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleModify)) 
                    (moduleHandle, inParams_p);
        } else {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNREGISTERED_FUNCTION);
            ADK_LOG_WARNING (NULL, moduleHandle,
                    "There is no registered Set function for the module\n");
        }
    }

    ADK_RETURN (retValue);
}    /* -----  end of function adk_modify  ----- */

/*
 * =============================================================
 *
 * @param[in]    moduleHandle    handle for the module for which 
 * the create function is called.
 *
 * @param[in]    *inParams_p    pointer to the data structure that
 * contains the parameters for the create operation.
 *
 * @retval    The return values here are determined the individual 
 * modules.
 *
 * @brief  
 *
 * This function provides a generic create interface for the application.
 * The caller has to specify the module handle as one of the input 
 * parameter. The function internally calls the registered create function
 * of the appropriate module.
 *
 * =============================================================
 */
adk_status_t
adk_create (
       adk_module_handle_t moduleHandle,
       const void *inParams_p,
       void *outParams_p )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_module_info_t *moduleInfo_p;

    if ( !ADK_VALID_HANDLE (moduleHandle) ||
            !ADK_VALID_PTR (inParams_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR (NULL, moduleHandle,
                "Failed to create object. invalid input handle\n");
    } else
    {
        moduleInfo_p = (adk_module_info_t *) moduleHandle;

        if (adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleCreate)
        {
            retValue = 
                (*(adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleCreate)) 
                    (moduleHandle, inParams_p, outParams_p);
        } else {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNREGISTERED_FUNCTION);
            ADK_LOG_ERROR (NULL, moduleHandle,
                    "There is no registered create function for the module\n");
        }
    }

    ADK_RETURN (retValue);
}    /* -----  end of function adk_create  ----- */

/*
 * =============================================================
 *
 * @param[in]    moduleHandle    handle for the module for which 
 * the create function is called.
 *
 * @param[in]    *inParams_p    pointer to the data structure that
 * contains the parameters for the delete operation.
 *
 * @retval    The return values here are determined the individual 
 * modules.
 *
 * @brief  
 *
 * This function provides a generic delete interface for the application.
 * The caller has to specify the module handle as one of the input 
 * parameter. The function internally calls the registered delete function
 * of the appropriate module.
 *
 * =============================================================
 */
adk_status_t
adk_delete (
       adk_module_handle_t moduleHandle,
       const void *inParams_p )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_module_info_t *moduleInfo_p;

    if ( !ADK_VALID_HANDLE (moduleHandle) ||
            !ADK_VALID_PTR (inParams_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR (NULL, moduleHandle,
                "Failed to delete object. invalid input handle\n");
    } else
    {
        moduleInfo_p = (adk_module_info_t *) moduleHandle;

        if (adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleDelete)
        {
            retValue = 
                (*(adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleDelete)) 
                    (moduleHandle, inParams_p);
        } else {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNREGISTERED_FUNCTION);
            ADK_LOG_ERROR (NULL, moduleHandle,
                    "There is no registered delete function for the module\n");
        }
    }

    ADK_RETURN (retValue);
}    /* -----  end of function adk_delete  ----- */

/*
 * =============================================================
 * @param[in]    moduleHandle    handle for the module for which 
 * the create function is called.
 *
 * @param[in]    *inParams_p    pointer to the data structure that
 * contains the get request type and the location for filling the
 * requested info.
 *
 * @retval    The return values here are determined the individual 
 * modules.
 *
 * @brief  
 *
 * This function provides a generic get interface for the application.
 * The caller has to specify the module handle as one of the input 
 * parameter. The function internally calls the registered get function
 * of the appropriate module.
 *
 * =============================================================
 */
adk_status_t
adk_get (
       adk_module_handle_t moduleHandle,
       const void *inParams_p )
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_module_info_t *moduleInfo_p;

    if ( !ADK_VALID_HANDLE (moduleHandle) ||
            !ADK_VALID_PTR (inParams_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_LOG_ERROR (NULL, moduleHandle,
                "Failed to get the object. invalid input handle\n");
    } else
    {
        moduleInfo_p = (adk_module_info_t *) moduleHandle;

        if (adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleGet)
        {
            retValue = 
                (*(adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].moduleGet)) 
                    (moduleHandle, inParams_p);
        } else {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNREGISTERED_FUNCTION);
            ADK_LOG_ERROR (NULL, moduleHandle,
                    "There is no registered get function for the module\n");
        }
    }

    ADK_RETURN (retValue);
}    /* -----  end of function adk_get  ----- */

/*
 * =============================================================
 * @internal
 * adk_module_error_string_get
 * @param[in,out]  
 * @return 
 * @retval 

 * @brief  
 *
 * This function is a utility function that will convert the module
 * specific error value to error string. It is called from macro. 
 * The function indirectlly call registered module specific error
 * code conversion function. if there is no
 * registerd function for the module "failed to decode" string 
 * is returned. THIS FUNCTION SHOULD NOT BE CALLED FROM OTHER PLACES.
 * =============================================================
 */
const char *adk_module_error_string_get (
        adk_module_handle_t moduleHandle,
        uint32_t errorCode)
{
    adk_module_info_t *moduleInfo_p;

    moduleInfo_p = (adk_module_info_t *) moduleHandle;

    if (adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].errorString)
    {
        return((*(adkModuleMgrCpuServiceFn_g[moduleInfo_p->moduleId].
                  errorString))(errorCode));
    }

    return("Failed to decode");
}    /* -----  end of function adk_module_error_string_get  ----- */
 
/*!    end addtogroup _sifFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */
