/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2001 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @file  adk_sif_dscp_pcp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifFunctions  ADK SIF Functions
 *     @{
 *
 * @brief
 *
 * This file contains the C functions for updating the DSCP/PCP mapping tree
 *
 * These functions are called from the QoS module.
 *
 *
 *  **********************************************************************/

#include "adk_include.h"

#include "adk_os.h"
#include "adk_sif_dp.h"
#include "adk_qos_dscp_pcp.h"
#include "adk_sif_dscp_pcp.h"


/*!
 * =============================================================
 * @internal
 * adk_sif_cp_mapping_init
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval 
 *
 * @brief  Initialize the DSCP->PCP mapping structures
 *
 * =============================================================
 */
adk_status_t
adk_sif_cp_mapping_init(void)
{
    adk_status_t                        retValue = ADK_SUCCESS;

    /*
     * Initialize the tree to -1 for all return values. This indicates the
     * entry hasn't been defined.
     */
    ADK_MEMSET(ADK_SIF_DP_INFO_PTR_GET->dscpPcpMapTree, 0,
               (sizeof(ADK_SIF_DP_INFO_PTR_GET->dscpPcpMapTree)));

    return retValue;
}    /* -----  end of function adk_sif_cp_mapping_init  ----- */

/*!
 * =============================================================
 * adk_sif_cp_mapping_add
 * @param[in]  sysHandle    System handle
 * @param[in]  mappingData  The structure containing the profile ID, type,
 *                          input and output maps
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief Fucntion to add a DSCP or PCP mapping entry
 *
 * This function is used to add a mapping entry. The two different CP types,
 * DSCP and PCP, can have a specific mapping associated with an input value
 * and a default for any unmapped values. The type field indicates which is to
 * be programmed. The DSCP and PCP maps can have different default values.
 *
 * =============================================================
 */
adk_status_t
adk_sif_cp_mapping_add(const adk_sys_handle_t       sysHandle,
                       const adk_qos_cp_mapping_t  *mappingData)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sif_dp_dscp_pcp_new_value_t     outValue;
    adk_sif_dp_shm_info_t              *dpShm_p;
    adk_qos_cp_mapping_type_t           lookupType;

    uint32_t        maxWalkRules = 0;
    uint32_t        controlPointId = 0;

    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid system handle 0x%p\n", sysHandle);
        goto adk_sif_cp_mapping_add_return;
    }

    ADK_LOG_INTERNAL_FUNC_ENTRY(sysHandle,
                                ADK_SYS_MODULE_HANDLE_GET (sysHandle));

    if (ADK_INVALID_PTR(mappingData))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "mappingData handle is NULL\n");
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->profileId >= ADK_QOS_CP_MAPPING_PROFILE_ID_MAX) ||
        (mappingData->profileId == ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid profile ID %d. It must be less than %d "
                      "and and not equal to %d\n",
                      mappingData->profileId,
                      ADK_QOS_CP_MAPPING_PROFILE_ID_MAX,
                      ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->outputType <= ADK_QOS_CP_MAPPING_START) ||
        (mappingData->outputType >= ADK_QOS_CP_MAPPING_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid outputType %d not in range\n",
                      mappingData->outputType);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->outputType == ADK_QOS_CP_MAPPING_DSCP_DEFAULT) ||
        (mappingData->outputType == ADK_QOS_CP_MAPPING_PCP_DEFAULT))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid outputType. It must not be _DEFAULT\n");
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->outputType == ADK_QOS_CP_MAPPING_DSCP) &&
        (mappingData->outputValue > ADK_CP_DSCP_VALUE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "outputValue %d exceeds maximum %d\n",
                      mappingData->outputValue,
                      ADK_CP_DSCP_VALUE_MAX);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->outputType == ADK_QOS_CP_MAPPING_PCP) &&
        (mappingData->outputValue > ADK_CP_PCP_VALUE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "outputValue %d exceeds maximum %d\n",
                      mappingData->outputValue,
                      ADK_CP_PCP_VALUE_MAX);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->inputType == ADK_QOS_CP_MAPPING_DSCP) &&
        (mappingData->inputValue > ADK_CP_DSCP_VALUE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "inputValue %d exceeds maximum %d\n",
                      mappingData->inputValue,
                      ADK_CP_DSCP_VALUE_MAX);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->inputType == ADK_QOS_CP_MAPPING_PCP) &&
        (mappingData->inputValue > ADK_CP_PCP_VALUE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "inputValue %d exceeds maximum %d\n",
                      mappingData->inputValue,
                      ADK_CP_PCP_VALUE_MAX);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if ((mappingData->outputMode < ADK_QOS_CP_MAPPING_MODE_REMARK) ||
        (mappingData->outputMode >= ADK_QOS_CP_MAPPING_MODE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "outputMode %d not in range\n",
                      mappingData->outputMode);
        goto adk_sif_cp_mapping_add_return_error;
    }

    /* Flag unsupported modes here */
    if (((mappingData->outputType == ADK_QOS_CP_MAPPING_PCP) &&
         ((mappingData->inputType == ADK_QOS_CP_MAPPING_DSCP) ||
          (mappingData->inputType == ADK_QOS_CP_MAPPING_DSCP_DEFAULT))) ||
        ((mappingData->outputType == ADK_QOS_CP_MAPPING_DSCP) &&
         ((mappingData->inputType == ADK_QOS_CP_MAPPING_PCP) ||
          (mappingData->inputType == ADK_QOS_CP_MAPPING_PCP_DEFAULT))))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNSUPPORTED);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid combination of inputType %d and outputType %d;"
                      " both must be the same.\n",
                      mappingData->inputType, mappingData->outputType);
        goto adk_sif_cp_mapping_add_return_error;
    }

    if (mappingData->outputMode != ADK_QOS_CP_MAPPING_MODE_REMARK_QUEUE_SELECT)
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_UNSUPPORTED);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "outputMode %d is not supported.\n",
                      mappingData->outputMode);
        goto adk_sif_cp_mapping_add_return_error;
    }

    switch (mappingData->inputType)
    {
        case ADK_QOS_CP_MAPPING_DSCP:
            lookupType = DSCP_INDEX;
            break;
        case ADK_QOS_CP_MAPPING_PCP:
            lookupType = PCP_INDEX;
            break;
        case ADK_QOS_CP_MAPPING_DSCP_DEFAULT:
            lookupType = DSCP_INDEX;
            maxWalkRules = ADK_CP_DSCP_VALUE_MAX + 1;
            break;
        case ADK_QOS_CP_MAPPING_PCP_DEFAULT:
            lookupType = PCP_INDEX;
            maxWalkRules = ADK_CP_PCP_VALUE_MAX + 1;
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Invalid inputType\n");
            goto adk_sif_cp_mapping_add_return_error;
            break;
    }

    dpShm_p = ADK_SIF_DP_INFO_PTR_GET;
    outValue = dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                      [lookupType]
                                      [mappingData->inputValue];

    if (maxWalkRules == 0)
    {
        /*
         * Set the new output value
         */
        switch (mappingData->outputType)
        {
            case ADK_QOS_CP_MAPPING_DSCP:
                outValue.dscp = (mappingData->outputValue |
                                 ADK_DSCP_OUTPUT_VALID_MASK_SET);
                dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                       [lookupType]
                                       [mappingData->inputValue] = outValue;
                break;
            case ADK_QOS_CP_MAPPING_PCP:
                outValue.pcp = (mappingData->outputValue |
                                ADK_PCP_OUTPUT_VALID_MASK_SET);
                dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                       [lookupType]
                                       [mappingData->inputValue] = outValue;
                break;
            default:
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                              "Invalid outputType\n");
                goto adk_sif_cp_mapping_add_return_error;
                break;
        }
    } else
    {
        /*
         * If maxWalkRules > 0, then we are updating the default rule. We need
         * to check to see if any of the children of the default rule need to
         * be updated also. We will walk the subnet database looking for
         * children of the default rule to update.
         */
        for (controlPointId = 0;
             controlPointId < maxWalkRules;
             controlPointId++)
        {
            outValue = dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                              [lookupType]
                                              [controlPointId];

            if (mappingData->outputType == ADK_QOS_CP_MAPPING_DSCP)
            {
                outValue.dscpDefault = (mappingData->outputValue |
                                        ADK_DSCP_OUTPUT_VALID_MASK_SET);
                if ((outValue.dscp & ADK_DSCP_OUTPUT_INHERIT_MASK_SET) ||
                    ((outValue.dscp & ADK_DSCP_OUTPUT_VALID_MASK_SET) == 0))
                {
                    outValue.dscp = (mappingData->outputValue |
                                     ADK_DSCP_OUTPUT_VALID_MASK_SET |
                                     ADK_DSCP_OUTPUT_INHERIT_MASK_SET);
                }
            }
            else if (mappingData->outputType == ADK_QOS_CP_MAPPING_PCP)
            {
                outValue.pcpDefault = (mappingData->outputValue |
                                       ADK_PCP_OUTPUT_VALID_MASK_SET);
                if ((outValue.pcp & ADK_PCP_OUTPUT_INHERIT_MASK_SET) ||
                    ((outValue.pcp & ADK_PCP_OUTPUT_VALID_MASK_SET) == 0))
                {
                    outValue.pcp = (mappingData->outputValue |
                                    ADK_PCP_OUTPUT_VALID_MASK_SET |
                                    ADK_PCP_OUTPUT_INHERIT_MASK_SET);
                }
            } else
            {
                continue;
            }

            dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                   [lookupType]
                                   [controlPointId] = outValue;

            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                         "Successfully updated pattern for "
                         "profileId: %d, inputType: %d, inputValue: %d "
                         "outputType: %d, outputValue: %d "
                         "adk_sif_cp_mapping_tree\n",
                         mappingData->profileId,
                         mappingData->inputType,
                         controlPointId,
                         mappingData->outputType,
                         mappingData->outputValue);
        }
    }

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                 "Successfully added pattern for "
                 "profileId: %d, inputType: %d, inputValue: %d "
                 "outputType: %d, outputValue: %d "
                 "adk_sif_cp_mapping_tree\n",
                 mappingData->profileId,
                 mappingData->inputType,
                 mappingData->inputValue,
                 mappingData->outputType,
                 mappingData->outputValue);

adk_sif_cp_mapping_add_return_error:
    ADK_LOG_INTERNAL_FUNC_EXIT(sysHandle,
                               ADK_SYS_MODULE_HANDLE_GET (sysHandle));

adk_sif_cp_mapping_add_return:
    return retValue;
}    /* -----  end of function adk_sif_cp_mapping_add  ----- */

/*!
 * =============================================================
 * adk_sif_cp_mapping_remove
 * @param[in]  sysHandle    System handle
 * @param[in]  mappingData  The structure containing the profile ID, type,
 *                          input and output maps
 *
 * @retval ADK_SUCCESS Successful completion
 * @retval
 *
 * @brief Fucntion to remove a DSCP or PCP mapping entry
 *
 * This function is used to remove a mapping entry. The specific as well as
 * the default values can be removed. Removing a specific value will cause
 * that to return the default value for the profile ID, if it is programmed,
 * or the value of 0 for both the DSCP and PCP mappings. If the default is
 * removed, the will return 0 for the DSCP and PCP input values that are not
 * mapped.
 *
 * =============================================================
 */
adk_status_t
adk_sif_cp_mapping_remove(const adk_sys_handle_t      sysHandle,
                          const adk_qos_cp_mapping_t *mappingData)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sif_dp_dscp_pcp_new_value_t     outValue;
    uint32_t                            maxWalkRules = 0;
    uint32_t                            controlPointId = 0;
    uint32_t                            inputValueWidth = 0;
    adk_sif_dp_shm_info_t              *dpShm_p;
    adk_qos_cp_mapping_type_t           lookupType;
    uint8_t                             newOutValue;

    if (ADK_INVALID_SYS_HANDLE(sysHandle))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_FORCE_LOG_ERROR("Invalid system handle 0x%p\n", sysHandle);
        goto adk_sif_cp_mapping_remove_return;
    }

    ADK_LOG_INTERNAL_FUNC_ENTRY(sysHandle,
                                ADK_SYS_MODULE_HANDLE_GET (sysHandle));

    if (ADK_INVALID_PTR(mappingData))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "mappingData handle is NULL\n");
        goto adk_sif_cp_mapping_remove_error;
    }

    if ((mappingData->profileId >= ADK_QOS_CP_MAPPING_PROFILE_ID_MAX) ||
        (mappingData->profileId == ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Invalid profile ID %d. It must be less than %d "
                      "and and not equal to %d\n",
                      mappingData->profileId,
                      ADK_QOS_CP_MAPPING_PROFILE_ID_MAX,
                      ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT);
        goto adk_sif_cp_mapping_remove_error;
    }

    if ((mappingData->inputType == ADK_QOS_CP_MAPPING_DSCP) &&
        (mappingData->inputValue > ADK_CP_DSCP_VALUE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Input value %d exceeds maximum %d\n",
                      mappingData->inputValue,
                      ADK_CP_DSCP_VALUE_MAX);
        goto adk_sif_cp_mapping_remove_error;
    }

    if ((mappingData->inputType == ADK_QOS_CP_MAPPING_PCP) &&
        (mappingData->inputValue > ADK_CP_PCP_VALUE_MAX))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                      "Input value %d exceeds maximum %d\n",
                      mappingData->inputValue,
                      ADK_CP_PCP_VALUE_MAX);
        goto adk_sif_cp_mapping_remove_error;
    }

    switch (mappingData->inputType)
    {
        case ADK_QOS_CP_MAPPING_DSCP:
            lookupType = DSCP_INDEX;
            break;
        case ADK_QOS_CP_MAPPING_PCP:
            lookupType = PCP_INDEX;
            break;
        case ADK_QOS_CP_MAPPING_DSCP_DEFAULT:
            lookupType = DSCP_INDEX;
            maxWalkRules = ADK_CP_DSCP_VALUE_MAX + 1;
            inputValueWidth = ADK_DSCP_INPUT_VALUE_WIDTH;
            break;
            break;
        case ADK_QOS_CP_MAPPING_PCP_DEFAULT:
            lookupType = PCP_INDEX;
            maxWalkRules = ADK_CP_PCP_VALUE_MAX + 1;
            inputValueWidth = ADK_PCP_INPUT_VALUE_WIDTH;
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                          "Could not extend the pattern: inputValue\n");
            goto adk_sif_cp_mapping_remove_error;
            break;
    }

    dpShm_p = ADK_SIF_DP_INFO_PTR_GET;
    outValue = dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                      [lookupType]
                                      [mappingData->inputValue];

    if (maxWalkRules == 0)
    {
        /*
         * Set the new output value
         */
        newOutValue = 0;
        switch (mappingData->outputType)
        {
            case ADK_QOS_CP_MAPPING_DSCP:
                if (outValue.dscp != 0)
                {
                    newOutValue = (outValue.dscpDefault |
                                   ADK_DSCP_OUTPUT_INHERIT_MASK_SET);
                }
                outValue.dscp = newOutValue;
                dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                       [lookupType]
                                       [mappingData->inputValue] = outValue;
                break;
            case ADK_QOS_CP_MAPPING_PCP:
                if (outValue.pcp != 0)
                {
                    outValue.pcp = (outValue.pcpDefault |
                                    ADK_PCP_OUTPUT_INHERIT_MASK_SET);
                }
                outValue.pcp = newOutValue;
                dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                       [lookupType]
                                       [mappingData->inputValue] = outValue;
                break;
            default:
                retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
                ADK_LOG_ERROR(sysHandle, ADK_SYS_MODULE_HANDLE_GET(sysHandle),
                              "Invalid outputType\n");
                goto adk_sif_cp_mapping_remove_error;
                break;
        }
    } else
    {
        /*
         * If maxWalkRules > 0, then we are updating the default rule. We need
         * to check to see if any of the children of the default rule need to
         * be updated also. We will walk the subnet database looking for
         * children of the default rule to update.
         */
        for (controlPointId = 0;
             controlPointId < maxWalkRules;
             controlPointId++)
        {
            outValue = dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                              [lookupType]
                                              [controlPointId];

            if (mappingData->outputType == ADK_QOS_CP_MAPPING_DSCP)
            {
                outValue.dscpDefault = 0;
                if (outValue.dscp & ADK_DSCP_OUTPUT_INHERIT_MASK_SET)
                {
                    outValue.dscp = 0;
                }
            }
            else if (mappingData->outputType == ADK_QOS_CP_MAPPING_PCP)
            {
                outValue.pcpDefault = 0;
                if (outValue.pcp & ADK_PCP_OUTPUT_INHERIT_MASK_SET)
                {
                    outValue.pcp = 0;
                }
            } else
            {
                continue;
            }

            dpShm_p->dscpPcpMapTree[mappingData->profileId]
                                   [lookupType]
                                   [controlPointId] = outValue;

            ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                         "Successfully deleted pattern for "
                         "profileId: %d, inputType: %d, inputValue: %d "
                         "outputType: %d, outputValue: %d "
                         "adk_sif_cp_mapping_tree\n",
                         mappingData->profileId,
                         mappingData->inputType,
                         controlPointId,
                         mappingData->outputType,
                         mappingData->outputValue);

        }
    }

    ADK_LOG_INFO(sysHandle, ADK_SYS_MODULE_HANDLE_GET (sysHandle),
                 "Successfully removed pattern for "
                 "profileId: %d, type: %d, inputValue: %d "
                 "adk_sif_cp_mapping_tree\n",
                 mappingData->profileId,
                 mappingData->inputType,
                 mappingData->inputValue);

adk_sif_cp_mapping_remove_error:
    ADK_LOG_INTERNAL_FUNC_EXIT(sysHandle,
                               ADK_SYS_MODULE_HANDLE_GET (sysHandle));

adk_sif_cp_mapping_remove_return:
    return retValue;
}    /* -----  end of function adk_sif_cp_mapping_remove  ----- */

/*!    end addtogroup _sifFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */
