/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @internal
 * @file  adk_utils_tree_update.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifFunctions  SIF Functions
 *     @{
 *
 * @brief  
 *
 * This file contains helper functions to create and mange lists.
 *
 ***********************************************************************/

#include <adk_include.h>
#include <adk_utils_tree_update.h>

#include <adk_os.h>


/*!
 * =============================================================
 * @internal
 * @param[out] **list_p    pointer to list pointer.
 *
 * @retval ADK_SUCCESS 
 *
 *
 * @retval ADK_E_MEM_UNAVAILABLE    failed to allocate memory
 * for the list head.
 *
 * @brief  
 *
 * This function set up the basic data structure for holding the 
 * list specific information.
 *
 * =============================================================
 */
adk_status_t
adk_utils_list_create ( 
        adk_list_t **list_p )
{
    adk_status_t retValue = ADK_SUCCESS;

    *list_p = ADK_NVM_CALLOC (ADK_SYS_MODULE_ID,  0, 1, sizeof (adk_list_t));
    
    if ( !ADK_VALID_PTR (*list_p) ) 
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
    }
    ADK_RETURN ( retValue );
}    /* -----  end of function adk_list_create  ----- */

/*!
 * =============================================================
 * @internal
 * @param[in] list_p    pointer to the list head.
 *
 * @param[in] key    the value that is to be used for searching 
 * the list.
 *
 * @param[in] data   data that needs to be stored in the list.
 * 
 * @retval ADK_SUCCESS Successfully completion
 * @retval ADK_E_INVALID_HANDLE    invalid list pointer.
 *
 * @brief  
 *
 * This function will append a new entry to the tail of the list.
 *
 * =============================================================
 */
adk_status_t
adk_utils_list_append ( 
        adk_list_t *list_p,
        uint32_t key,
        void *data)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_list_entry_t *listEntry_p;
    
    if ( !ADK_VALID_PTR (list_p) ) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
    } else 
    {
        listEntry_p = (adk_list_entry_t *) ADK_NVM_MALLOC (
              sizeof (adk_list_entry_t));
        if ( !ADK_VALID_PTR (listEntry_p) )
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
            goto adk_list_append_return;
        }
        listEntry_p->listKey = key;
        listEntry_p->listData = data;
        listEntry_p->nextEntry_p = NULL;
        listEntry_p->prevEntry_p = list_p->listTail_p;
        if ( list_p->listTail_p )
        {
            list_p->listTail_p->nextEntry_p = listEntry_p;
        }
        list_p->listTail_p = listEntry_p;
        if ( list_p->listHead_p == NULL) 
        {
            list_p->listHead_p = listEntry_p;
        }
    }

adk_list_append_return:
    ADK_RETURN( retValue );
}    /* -----  end of function adk_list_append  ----- */

/*!
 * =============================================================
 * @internal
 * @param[in] list_p    pointer to the list head.
 *
 * @param[in] key    the value that is to be used for searching 
 * the list.
 *
 * @param[out] data_p    pointer to a location where the search 
 * result needs to be stored.
 * 
 * @retval ADK_SUCCESS Successfully completion
 *
 * @brief  
 *
 * This function search the list for the "key" and returns the
 * data that matches the key.
 *
 * =============================================================
 */
adk_status_t
adk_utils_list_find ( 
        adk_list_t *list_p,
        uint32_t key,
        void **data_p)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_list_entry_t *listEntry_p;

    if ( !ADK_VALID_PTR (list_p) || !ADK_VALID_PTR (*data_p) ) 
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
    } else 
    {
        listEntry_p =  list_p->listHead_p;
        while ( listEntry_p ) 
        {
            if ( listEntry_p->listKey == key )
            {
                *data_p = listEntry_p->listData;
                break;
            }
            listEntry_p = listEntry_p->nextEntry_p;
        }    /* -----  end while  ----- */

    }

    ADK_RETURN( retValue );
}    /* -----  end of function adk_list_find  ----- */

/*!
 * =============================================================
 * @internal
 * @param[in] list_p    pointer to the list head.
 *
 * @param[in] key    the key to the entry that needs to be 
 * removed from the list.
 * 
 * @retval ADK_SUCCESS Successfully completion
 *
 * @brief  
 *
 * This function removes the entry with appropriate "key" from 
 * the list.
 *
 * =============================================================
 */
adk_status_t
adk_utils_list_remove ( 
        adk_list_t __attribute__((unused)) *list_p, 
        uint32_t __attribute__((unused)) key )
{
    adk_status_t retValue = ADK_SUCCESS;

    ADK_RETURN( retValue );
}    /* -----  end of function adk_list_remove  ----- */

/*!    end addtogroup _sifFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */
