/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_sif_dp_dscp_pcp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifDpFunctions  SIF Data Plane Functions
 *     @{
 *
 * @brief
 *
 * This file contains the ADK Dataplane System Configuration function to be
 * called before System Initialization.
 *
 *  **********************************************************************/

#include "adk_sif_dp.h"
#include "adk_sif_dp_log_lib.h"
#include "adk_qos_dscp_pcp.h"

/*!
 * ====================================================================
 *
 * adk_sif_dp_code_point_mapping_get
 *
 * @param[in]   profileId : Profile ID to lookup
 * @param[in]   type : Enum indicating DSCP or PCP input
 * @param[in]   inputValue : DSCP or PCP input value
 * @param[in]   outputValues_p : Structure containing DSCP and PCP output values
 *
 * @retval       ADK_SUCCESS   system configuration was succesfull.
 *
 * @brief Map input DSPC/PCP value to output DSCP PCP tuple
 *
 * This DP API is used to map a DSCP or PCP value for a given profile into a
 * 2-tuple structure returning DSCP and PCP values. Both values are returned
 * and the caller needs to determine which to use. The caller checks to see if
 * the value us valid or not by checking one of the bits in the return value.
 * Two macros, ADK_DSCP_OUTPUT_VALID_MASK_SET and
 * ADK_PCP_OUTPUT_VALID_MASK_SET, are used for this. The caller should
 * complement the macro before ANDing with the return value.
 *
 * ====================================================================
 */
adk_status_t
adk_sif_dp_code_point_mapping_get(
    uint8_t                              profileId,
    adk_sif_dp_code_point_type_t         type,
    uint8_t                              inputValue,
    adk_sif_code_point_mapped_values_t  *outputValues_p)
{
    adk_status_t                        retValue = ADK_SUCCESS;
    adk_sif_dp_dscp_pcp_new_value_t     outValue;
    uint32_t                            lookupType;

    if ((profileId >= ADK_QOS_CP_MAPPING_PROFILE_ID_MAX) ||
        (profileId == ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                            "Invalid profile ID %d. It must be less than "
                            "ADK_QOS_CP_MAPPING_PROFILE_ID_MAX "
                            " and and not equal to "
                            "ADK_QOS_CP_MAPPING_PROFILE_ID_DEFAULT\n",
                            profileId);
        goto _error;
    }

    if ((type != ADK_SIF_DP_CODE_POINT_TYPE_DSCP) &&
        (type != ADK_SIF_DP_CODE_POINT_TYPE_PCP))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                            "Invalid input type %d \n", type);
        goto _error;
    }

    if (type == ADK_SIF_DP_CODE_POINT_TYPE_DSCP)
    {
        if (inputValue > ADK_CP_DSCP_VALUE_MAX)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                                "Invalid DSCP value: 0x%x\n", inputValue);
            goto _error;
        }

        /* Need to set an index based on the type since type is always > 0 */
        lookupType = DSCP_INDEX;
    } else if (type == ADK_SIF_DP_CODE_POINT_TYPE_PCP)
    {
        if (inputValue > ADK_CP_PCP_VALUE_MAX)
        {
            retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
            ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                                "Invalid PCP value: 0x%x\n", inputValue);
            goto _error;
        }

        /* Need to set an index based on the type since type is always > 0 */
        lookupType = PCP_INDEX;
    } else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_PARAMETER);
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                            "Invalid type: 0x%x\n", type);
        goto _error;
    }

    if ( ! ADK_VALID_HANDLE(outputValues_p))
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET (ADK_E_INVALID_HANDLE);
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "outputValues_p is NULL\n");
        goto _error;
    }

    outValue = ADK_SIF_DP_INFO_PTR_GET->dscpPcpMapTree[profileId]
                                                      [lookupType]
                                                      [inputValue];
    outputValues_p->dscpOut = outValue.dscp;
    outputValues_p->pcpOut = outValue.pcp;

_error:
    return retValue;
}
