/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_sys_dp.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifDpFunctions  SIF Data Plane Functions
 *     @{
 *
 * @brief
 *
 * This file contains the ADK Dataplane System Configuration function to be
 * called before System Initialization.
 *
 *  **********************************************************************/

#include <adk_include.h>
#include "adk_sif_dp.h"
#include "adk_sys_lcore.h"
#include <adk_sys_dpdk.h>
#include <adk_shm.h>
#include <adk_pmu.h>
#include <adk_hos.h>
#include <adk_timer.h>

#include <rte_bridge_timer.h>
#include <rte_bridge_eth.h>
#include "adk_pal_dp_export.h"
#include "adk_ipv6_const.h"
#include "adk_ipv6_dp_export.h"
#include "adk_ip_dp_export.h"
#include "adk_pal_dp.h"
#include "adk_sif_dp.h"
#include "adk_sif_dp_log_lib.h"
#include "adk_qos_dp_export.h"
#include "adk_nd_common.h"
#include "adk_em.h"
#include "adk_eth_dp.h"
#include "adk_eth_dp.h"
#include "adk_sys_dp_flow_cache_uli_func.h"

#if ADK_DPDK_VERSION >= 18050016
#include <rte_service.h>
#include <rte_service_component.h>
#endif

measure_cycle_t adkMeasCycleTbl_g[ADK_MAX_NUM_CORES][ADK_MEASURE_CYCLE_MAX_STAGE];
measure_cycle_t *adkMeasCycle_g[ADK_MAX_NUM_CORES][ADK_MEASURE_CYCLE_MAX_STAGE];

uint64_t *adk_pal_dp_min_cycles;
extern uint8_t* adk_dp_eth_auto_conn_status_g;
adk_sif_dp_shm_info_t   *adkSifDpShm_p = NULL;

#ifdef ADK_SYS_OVERLOAD_PROTECTION
adk_sys_ovp_info_t  *adk_sys_dp_ovp_table_p = NULL;
#endif

static adk_sys_dp_egress_stats_t *adk_sys_dp_egress_stats_p = NULL;
#if ADK_DPDK_VERSION >= 18050016
static unsigned int periodic_service_id;
#endif

#ifdef BLOCK_COUNT_ENABLED
    uint64_t adk_sys_egress_stats_rx_ts, adk_sys_egress_stats_tx_ts; 
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_EGRESS_STATS_RX_TS() adk_dp_measure_block_cycle_rx_ts(&adk_sys_egress_stats_rx_ts)
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_EGRESS_STATS_TX_TS() adk_dp_measure_block_cycle_tx_ts(&adk_sys_egress_stats_rx_ts, &adk_sys_egress_stats_tx_ts, "SYS EGRESS STATS CALLBACK", PMU_SYS_ULI_EGRESS_STATS)
#else
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_EGRESS_STATS_RX_TS()
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_EGRESS_STATS_TX_TS()
#endif
/*!
 * ====================================================================
 * @param[in]    sysConfig_p         pointer to adk_sys_config_t data
 * structure. The elements of this data structure are described
 * adk_sys.h
 *
 * @retval       ADK_SUCCESS   system configuration was succesfull.
 * @retval       ADK_E_MEM_UNAVAILABLE   failed to allocate memory
 * @retval       ADK_E_INVALID_PARAMETER invalid input parameter. This
 * could be invalid input pointers.
 *
 * @brief
 *
 * This function configures the ADK slave mode parameters. This API
 * must be called after DPDK EAL/platform initialization and before
 * ADK System and DP modules initializations.
 *
 * ====================================================================
 */
adk_status_t
adk_sys_dp_config(
        const adk_sys_config_t *sysConfig_p)
{
    adk_status_t            retValue = ADK_SUCCESS;
    adk_sys_config_info_t  *adk_slave_config_p = NULL;
    adk_sys_shm_config_t   *shm_config_p = NULL;
    adk_sif_dp_shm_info_t  *shm_dp_config_p = NULL;
    uint32_t                i, id;

    /* Validate input parameters */
    if (!sysConfig_p)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "adk_sys_config failed. Invalid pointers\n");
        retValue = ADK_E_INVALID_PARAMETER;
        goto _adk_sys_dp_config_exit;
    }

    if ((sysConfig_p->adk_mode != ADK_SYS_CONFIG_MODE_SLAVE) &&
        (sysConfig_p->adk_mode != ADK_SYS_CONFIG_MODE_SHARED_DATAPLANE))
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Invalid adk_mode configuration. "
                         "Must be shared-dataplane or slave mode.\n");
        retValue = ADK_E_INVALID_PARAMETER;
        goto _adk_sys_dp_config_exit;
    }

    if (sysConfig_p->num_worker_lcores != 0)
    {
        if (!sysConfig_p->worker_lcores)
        {
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                            "Invalid worker ID list pointer\n");
            retValue = ADK_E_INVALID_PARAMETER;
            goto _adk_sys_dp_config_exit;
        }
    }

    /* Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_create(ADK_SYS_SHM_MEMZONE_NAME,
                              sizeof(adk_sys_shm_config_t),
                              ADK_SHM_LOOKUP_LCORE,
                              (adk_shm_handle_t **)&shm_config_p);
    if ((retValue != ADK_SUCCESS) &&
        (ADK_STATUS_ERROR_CODE_GET(retValue) != ADK_W_SHM_EXISTS))
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, 
                         "Failed to create shared memory for IPC!\n");
        goto _adk_sys_dp_config_exit;
    }
    if (ADK_STATUS_ERROR_CODE_GET(retValue) != ADK_W_SHM_EXISTS) {
        adk_slave_config_p = ADK_NVM_CALLOC(ADK_SYS_MODULE_ID, 0, 1, sizeof(adk_sys_config_info_t));
        if (!ADK_VALID_PTR (adk_slave_config_p))
        {
            retValue = ADK_E_MEM_UNAVAILABLE;
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, 
                        "adk_sys_config failed. Memory allocation failed\n");
            goto _adk_sys_dp_config_cfg_mem_exit;
        }

        /* Set the system config handle to shared memory segment. This will be read
         * by the secondary process.
         */
        shm_config_p->sysConfHandle = (adk_config_handle_t) adk_slave_config_p;
    } else {
        adk_slave_config_p = (adk_sys_config_info_t *)shm_config_p->sysConfHandle;
    }

    /* store the ADK shared memory pointer in a global variable */
    adkSysShmConf_p = shm_config_p;

    /* Update ADK slave configuration */
    adk_slave_config_p->lcoreInfo.adkMode = sysConfig_p->adk_mode;
    adk_slave_config_p->pktio_mempool = sysConfig_p->pktio_mempool;
    adk_slave_config_p->ip_frag_mempool = sysConfig_p->ip_frag_mempool;
    adk_slave_config_p->lcoreInfo.numLcores = sysConfig_p->num_worker_lcores;
    for (i = 0; i < sysConfig_p->num_worker_lcores; i++)
    {
        id = sysConfig_p->worker_lcores[i];
        adk_slave_config_p->lcoreInfo.lcore[i].id = id;
        adk_slave_config_p->lcoreInfo.lcore[i].index = ADK_LCORE_INDEX_LOOKUP(id);
        adk_slave_config_p->lcoreInfo.lcore[i].socket = rte_lcore_to_socket_id(id);
    }

    /* Allocated shared memory for IPC between primary and secondary process */
    retValue = adk_shm_create(ADK_SIF_DP_SHM_MEMZONE_NAME,
                              sizeof(adk_sif_dp_shm_info_t),
                              ADK_SHM_LOOKUP_LCORE,
                              (adk_shm_handle_t **)&shm_dp_config_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Failed to create DP shared memory for IPC!\n");
        goto _adk_sys_dp_config_exit;
    }

    /* Set the system config handle to shared memory segment. This will be read
     * by the secondary process.
     */
    shm_config_p->sysConfHandle = (adk_config_handle_t) adk_slave_config_p;
    shm_config_p->sifDpShmHandle = (adk_sys_handle_t) shm_dp_config_p;
    adkSysShmConf_p = shm_config_p;
    adkSifDpShm_p = shm_dp_config_p;

    ADK_SIF_DP_HDL_COOKIE_SET;

_adk_sys_dp_config_exit:

    ADK_RETURN( retValue );

_adk_sys_dp_config_cfg_mem_exit:
    /* remove the shared memory */
    if(adk_shm_destroy(ADK_SYS_SHM_MEMZONE_NAME) !=ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "shared memory clean up failed !\n");
    }
    ADK_RETURN( retValue );
}

/*!
 * ====================================================================
 * adk_sys_dp_init
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * This function initializes the ADK maintenance manager
 * for each worker core configured in the system. This function
 * has to be called from DP process.
 * ====================================================================
 */
adk_status_t
adk_sys_dp_init(
    void)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_sys_shm_config_t   *shm_config_p = NULL;
    adk_sys_config_info_t  *adk_slave_config_p = NULL;
    adk_sif_dp_shm_info_t *shmDpInfo_p = NULL;
#if ADK_DPDK_VERSION >= 18050016
    struct rte_service_spec service = {0};
    uint32_t j;
#endif
    /* This API does nothing unless we are operating in slave mode with
     * eventDev framework
     */

    /* Restore the handles from shared memory that is created 
     * either as part of control APIs adk_sys_config / adk_sys_init
     */

    /* Allocated shared memory for IPC between primary and secondary process */
    ret = adk_shm_lookup(ADK_SYS_SHM_MEMZONE_NAME,
                              (adk_shm_handle_t **)&shm_config_p);
    if (ret != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Failed to lookup shared memory, "
                         "ADK control plane is not initialized !\n");
        return ret;
    }
    adk_slave_config_p = (adk_sys_config_info_t *)shm_config_p->sysConfHandle;

#ifdef ADK_SYS_OVERLOAD_PROTECTION
    adk_sys_dp_ovp_table_p  = &adk_slave_config_p->ovpInfo;
#endif

    ret = adk_shm_lookup(ADK_SYS_EGRESS_STATS_SHM_MEMZONE_NAME,
                        (adk_shm_handle_t **)&adk_sys_dp_egress_stats_p);
    if (ret != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Failed to lookup sys egress stats shared memory, "
                         "ADK control plane is not initialized !\n");
        return ret;
    }

#if !defined(ADK_MASTER_MODE)

    adk_sys_config_worker_lcore_info_t *workerInfo_p;
    uint32_t i;

    /*
     * Lookup the worker info so we can get the NUMA socket to allocate the
     * memzone from
     */
    ret = adk_sys_config_worker_lcore_info_get(&workerInfo_p);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }

    if (workerInfo_p->adkMode == ADK_SYS_CONFIG_MODE_STANDALONE)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                     "Error: System not built to operate in Slave Mode!\n");
        ret = ADK_E_UNSUPPORTED;
        goto _adk_sys_dp_init_exit;
    }

    /*
     * Need to run across all possible indices since there could be holes in
     * the list
     */
    for (i = 0; i < workerInfo_p->numLcores; i++)
    {
        ret = adk_dp_maintenance_init_v2(workerInfo_p->lcore[i].id);
        if (ret != ADK_SUCCESS)
        {
            ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                        "Failed to initialize maintenance timer on core %d\n",
                            workerInfo_p->lcore[i].id);
            goto _adk_sys_dp_init_exit;
        }
    }
#if ADK_DPDK_VERSION >= 18050016
    /* Register the periodic callback function */
    strncpy(service.name, "rcu_quiescent_state", RTE_SERVICE_NAME_MAX - 1);
    service.callback = (rte_service_func)adk_rcu_quiescent_state;
    service.callback_userdata = NULL;
    service.capabilities |= RTE_SERVICE_CAP_MT_SAFE;
    ret = rte_service_component_register(&service, &periodic_service_id);
    if (ret)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Error: Failed to register periodic callback fn as a service.\n");
        goto _adk_sys_dp_init_exit;
    }
    for (i = 0; i < workerInfo_p->numLcores; i++)
    {
        ret = rte_service_map_lcore_set(periodic_service_id,
                                        workerInfo_p->lcore[i].id, true);
        if (ret)
        {
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Error: Failed to set the periodic service's map.\n");
            for (j = 0; j < i; j++)
            {
                /* Unmap previously mapped Lcores */
                rte_service_map_lcore_set(periodic_service_id,
                                          workerInfo_p->lcore[j].id, false);
            }
            rte_service_component_unregister(periodic_service_id);
            goto _adk_sys_dp_init_exit;
        }
    }
    rte_service_component_runstate_set(periodic_service_id, 1);
#endif
    ret = adk_pal_dp_init();
    if (ret != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Error: Failed to do adk_pal_dp_init()!\n");
        goto _adk_sys_dp_init_exit;
    }
#ifdef ADK_QOS_ENABLE
    ret = adk_qos_dp_init();
    if (ret != ADK_SUCCESS)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Error: Failed to do adk_qos_dp_init()!\n");
        goto _adk_sys_dp_init_exit;
    }
#endif

#ifdef ADK_MEASURE_CYCLE
    for(i=0; i<ADK_MAX_NUM_CORES; i++)
    {
        adkMeasCycle_g[i][0] = &adkMeasCycleTbl_g[i][0];
        ADK_MEMSET(adkMeasCycle_g[i][0], 0, sizeof(measure_cycle_t));
        adkMeasCycle_g[i][1] = &adkMeasCycleTbl_g[i][1];
        ADK_MEMSET(adkMeasCycle_g[i][1], 0, sizeof(measure_cycle_t));
        adkMeasCycle_g[i][2] = &adkMeasCycleTbl_g[i][2];
        ADK_MEMSET(adkMeasCycle_g[i][2], 0, sizeof(measure_cycle_t));
        adkMeasCycle_g[i][3] = &adkMeasCycleTbl_g[i][3];
        ADK_MEMSET(adkMeasCycle_g[i][3], 0, sizeof(measure_cycle_t));
    }
#endif

    /*get DP configuration, it is not expected to be NULL*/
    shmDpInfo_p = (adk_sif_dp_shm_info_t *) adkSysShmConf_p->sifDpShmHandle;
    /*get DP configuration, it is not expected to be NULL*/
    adkSifDpShm_p = (adk_sif_dp_shm_info_t *) adkSysShmConf_p->sifDpShmHandle;

    if (shmDpInfo_p) {
        /* Flow cache tables */
        for(i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            /* ULI flow cache tables */
            /* allocate memory for Hash entries */
            adk_slave_config_p = ADK_NVM_CALLOC(ADK_SYS_MODULE_ID, 0, 1, sizeof(adk_sys_config_info_t));
            shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].hash_tbl = ADK_NVM_CALLOC(ADK_SYS_MODULE_ID,
                    0, 
                    (ADK_SYS_ULI_FLOW_CACHE_ENTRIES * ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES),
                    sizeof(adk_sys_dp_flow_cache_hash_entry_t));

            if (NULL == shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].hash_tbl) {
                ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                        "Error: Failed to allocate memory for flow cache hash table on coreIndex:%d !\n", i);
                goto _adk_sys_dp_init_exit;
            }

            /* allocate memory for flow cache table */
            shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].tbl = ADK_NVM_CALLOC(ADK_SYS_MODULE_ID,
                    0, 
                    ADK_SYS_ULI_FLOW_CACHE_ENTRIES,
                    sizeof(adk_sys_dp_uli_flow_cache_entry_t));

            if (NULL == shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].tbl) {
                ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                        "Error: Failed to allocate memory for flow cache entry table on coreIndex:%d !\n", i);
                goto _adk_sys_dp_init_exit;
            }

            /* allocate memory for flow cache entry stats*/
            shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].fc_entry_stats = ADK_NVM_CALLOC(ADK_SYS_MODULE_ID,
                    0, 
                    ADK_SYS_ULI_FLOW_CACHE_ENTRIES,
                    sizeof(adk_sys_dp_uli_flow_cache_entry_stats_t));

            if (NULL == shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].fc_entry_stats) {
                ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID,
                        "Error: Failed to allocate memory for flow cache entry stats table on coreIndex:%d !\n", i);
                goto _adk_sys_dp_init_exit;
            }

            ADK_SYS_DP_TABLE_IDX_INIT(shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i],
                    ADK_SYS_ULI_FLOW_CACHE_ENTRIES);

        }
    }

    /* Initialize function pointers used in the data plane processing */
    adkDestEntryFunction_g[ADK_SIF_TUND_IP_DEMUX_FN_ID] = adk_sys_dp_tund_ip_demux;
    adkDestEntryFunction_g[ADK_SIF_PCAP_DEMUX_FN_ID] = adk_sys_dp_pcap_demux;
    adkDestEntryFunction_g[ADK_SIF_BRIDGE_DEMUX_FN_ID] = adk_sys_dp_bridge_demux;
    /* Initialize function pointers used to receive EM event */
    adkDestEntryFunction_g[ADK_SIF_EM_DEMUX_FN_ID] = adk_sys_em_demux_fn_id;
    adkDestEntryFunction_g[ADK_SIF_EM_IPV6_DEMUX_FN_ID] = adk_sys_em_ipv6_demux_fn_id;
    adkDestEntryFunction_g[ADK_SIF_EM_LLI_FN_ID] = adk_sys_em_pkt_proc_lli;
    adkDestEntryFunction_g[ADK_SIF_EM_LLI_DIRECT_FN_ID] = adk_sys_em_pkt_proc_lli;
    return ret;

_adk_sys_dp_init_exit:
#endif /* #if !defined(ADK_MASTER_MODE) */

    if (shmDpInfo_p) {
        for(i=0; i<ADK_MAX_NUM_CORES; i++) {
            if (shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].hash_tbl) {
                ADK_NVM_FREE(shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].hash_tbl);
            }
            if (shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].tbl) {
                ADK_NVM_FREE(shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].tbl);
            }
            if (shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].fc_entry_stats) {
                ADK_NVM_FREE(shmDpInfo_p->sys_dp_uli_flow_cache_tbl[i].fc_entry_stats);
            }
        }
    }

    return ret;
}

/*!
 * ====================================================================
 * adk_sys_dp_pcap_demux
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * Entry function for packet capture from hardware. This function checks
 * the packet capture ID in the metadata and jumps to the function registerd
 * against it. This function, thus, serves as the entry demux function
 * for all packets captured in the hardware (HLP). The registered functions
 * are owned by different protocol modules.
 * ====================================================================
 */
adk_status_t
adk_sys_dp_pcap_demux(
    void **task_p,
    uint8_t task_count,
    uint32_t task_mask,
    void *ring,
    adk_sif_cfg_pipeline_info_t *next_stage_entry,
    int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;

#ifdef ADK_DEVICE_SNR
    uint32_t entry_fn_id;
    adk_pktio_task_t *task_in = (adk_pktio_task_t *)(task_p[0]);
    adk_dsi_rx_metadata_t *dsi_md_p = ADK_DSI_RX_METADATA_PTR(task_in);

    entry_fn_id = ADK_SIF_DP_INFO_PTR_GET->pcapIdEntryFnIdMap[dsi_md_p->acl_id];

    if (entry_fn_id && adkDestEntryFunction_g[entry_fn_id]) {
        ret = adkDestEntryFunction_g[entry_fn_id](task_p, task_count, task_mask,
                                                  NULL, NULL, coreIndex);
    }
    else
    {
        adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[0]));
        ADK_DP_LOG_ERROR_1D(ADK_SYS_MODULE_ID, "No registered function for"
                            " pcap ID %d!", dsi_md_p->acl_id);
        goto _adk_sys_dp_pcap_demux_exit;
    }

_adk_sys_dp_pcap_demux_exit:

#else
    RTE_SET_USED(task_p);
    RTE_SET_USED(task_count);
    RTE_SET_USED(task_mask);
    RTE_SET_USED(coreIndex);
#endif
    RTE_SET_USED(ring);
    RTE_SET_USED(next_stage_entry);

    return ret;
}

/*!
 * ====================================================================
 * adk_sys_dp_bridge_demux 
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * Packet received on bridge from netdriver will be further processed
 * based on ethertype 
 * ====================================================================
 */

adk_status_t 
adk_sys_dp_bridge_demux(void **task_p,
          uint8_t task_count,
          uint32_t task_mask,
          void *ring,
          adk_sif_cfg_pipeline_info_t *next_stage_entry,
          int coreIndex)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t *task_out;
    uint8_t         dest_fn_id=0;
    int             index;
    uint8_t         *pdu = NULL;
    uint16_t        ether_type;
    uint16_t        *tmp_16;
    adk_metadata_t *adk_md_p;
    uint8_t         cbEventType = ADK_PAL_DP_BATCH_CB_EVENT_STAGE_DIRECT;

    RTE_SET_USED(task_count);
    RTE_SET_USED(task_mask);
    RTE_SET_USED(ring);
    RTE_SET_USED(next_stage_entry);
    RTE_SET_USED(coreIndex);

    for(index=0;index<task_count; index++) {

        task_out = (adk_pktio_task_t *)(task_p[index]);

        adk_md_p = ADK_METADATA_PTR(task_out);

        adk_md_p->ip_offset = 14;

        ADK_PDU_PTR_GET(task_out, pdu, uint8_t *);

        if((adk_dp_eth_auto_conn_status_g != NULL) &&
                 (ADK_DP_ETH_AUTO_CONN_DISABLE == *adk_dp_eth_auto_conn_status_g))
        {
            tmp_16 = (uint16_t *)(void *)&pdu[12];
            ether_type = adk_ntohs(*tmp_16);

            /* tagged packet */
            if(ether_type == 0x8100)
            {
                /* ether type */
                tmp_16 = (uint16_t *)(void *)&pdu[16];
                ether_type = adk_ntohs(*tmp_16);
                adk_md_p->ip_offset = 18;
            }

            if(ether_type == 0x0800) {
                /* IPv4 packet */
                dest_fn_id =  ADK_IPV4_NETD_FN_ID;
                cbEventType = ADK_PAL_DP_BATCH_CB_EVENT_STAGE_DIRECT;
            } else if(ether_type == 0x86DD) {
                /* IPv6 packet */
                dest_fn_id =  ADK_IPV6_NETD_FN_ID;
                cbEventType = ADK_PAL_DP_BATCH_CB_EVENT_STAGE_DIRECT;
            } else {
                /* Non ip packet goes to Etherent */
                dest_fn_id =  ADK_ETH_ULI_NETD_BRIDGE_FN_ID;
                cbEventType = ADK_PAL_DP_BATCH_CB_EVENT_STAGE_ETH_NETD;
            }
        }
        else
        {
            /* All packet goes to Etherent during autoconnection */
            dest_fn_id =  ADK_ETH_ULI_NETD_BRIDGE_FN_ID;
            cbEventType = ADK_PAL_DP_BATCH_CB_EVENT_STAGE_ETH_NETD;
        }
       if (adkDestEntryFunction_g[dest_fn_id]) {
#if defined(ADK_PAL_DYNAMIC_PPL)
           ret = adk_pal_dp_stage_send((void **)&task_out, 1, 0x1, dest_fn_id, 0x00,
                   0x00, ADK_PAL_DP_BATCH_CB_FN_SIF, cbEventType, NULL);
           if(ret == ADK_E_INTERNAL_ERROR)
           {
               adk_pal_dp_release_send(task_out, ADK_PAL_DP_BATCH_CB_FN_SIF, NULL);
               adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_out));
               continue;
           }
#else
            ret = adkDestEntryFunction_g[destFunc](&task_out, 1, task_mask,
                                                   NULL, NULL, coreIndex);
#endif
       } else {
           /* NULL destination funciton, drop the packet  */
            adk_pal_dp_release_send(task_out,ADK_PAL_DP_BATCH_CB_FN_SIF, NULL);
            adk_pktio_task_buffer_free(task_out);
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Dest module is not initialized. Dropping packet!");
            continue;
       }

    }

    return ret;
}

/*!
 * ====================================================================
 * adk_sys_dp_tund_ip_demux
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * Overload protection call back funtion. called by Ethernet-BD at Rx for 
 * every packet to check drop eligibility of that packet.
 * ====================================================================
 */

adk_status_t 
adk_sys_dp_tund_ip_demux(void **task_p,
          uint8_t task_count,
          uint32_t task_mask,
          void *ring,
          adk_sif_cfg_pipeline_info_t *next_stage_entry,
          int coreIndex)                         
                         
{
    adk_status_t ret = ADK_SUCCESS;
    adk_pktio_task_t *task_out = (adk_pktio_task_t *)(*task_p);
    adk_metadata_t *adk_md_p = ADK_METADATA_PTR(task_out);
    uint8_t         *ip_type_p;
    uint8_t         dest_fn_id=0;
    int             index, cnt;

    RTE_SET_USED(task_count);
    RTE_SET_USED(task_mask);
    RTE_SET_USED(ring);
    RTE_SET_USED(next_stage_entry);
    RTE_SET_USED(coreIndex);
    RTE_SET_USED(adk_md_p);

    for(index=0;index<task_count; index++) {
        ip_type_p = rte_pktmbuf_mtod_offset(task_out, uint8_t *, 0 );


        if(((*ip_type_p) &(0xF0)) == 0x40) {
            /* IPv4 packet */
            dest_fn_id =  ADK_IPV4_NETD_FN_ID;
        } else if(((*ip_type_p) &(0xF0)) == 0x60) {
            /* IPv6 packet */
            dest_fn_id =  ADK_IPV6_NETD_FN_ID;
        } else {
            /* Un supported packet type */
            dest_fn_id =  0;
        }

       if (adkDestEntryFunction_g[dest_fn_id]) { 
#if defined(ADK_PAL_DYNAMIC_PPL)
            cnt = adk_pal_dp_next_stage_event_handle(task_p, task_count, task_mask, 
                                       dest_fn_id, 0x00, 0x00, ADK_TASK_TYPE_FWD);
            if(cnt != 1)
            {
                /* Free the task */
                adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[0]));
                ret = ADK_E_INTERNAL_ERROR;
            }
#else
            ret = adkDestEntryFunction_g[destFunc](task_p, task_count, task_mask,
                                                   NULL, NULL, coreIndex);
#endif
       } else {
           /* NULL destination funciton, drop the packet  */
            adk_pktio_task_buffer_free((adk_pktio_task_t *)(task_p[0]));
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "not an IPv4/6 packet. Dropping packet!");
            continue;
       }

    }

    return ret;
}


#ifdef ADK_SYS_OVERLOAD_PROTECTION
/*!
 * ====================================================================
 * adk_sys_dp_overlod_protection
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * Overload protection call back funtion. called by Ethernet-BD at Rx for 
 * every packet to check drop eligibility of that packet.
 * ====================================================================
 */

uint16_t
#if ((RDK_REL_MAJOR >= 12) || ((RDK_REL_MAJOR == 10) && (RDK_REL_MINOR == 2)))
adk_sys_dp_overlod_protection(uint16_t port_id,
#else
adk_sys_dp_overlod_protection(uint8_t port_id,
#endif
        uint16_t queue_id,
        uint32_t enqueue_buf_size,
        uint32_t enqueue_buf_count,
        struct rte_event *ev,
        uint16_t nb_events,
        void *cb_arg,
        uint16_t *nb_dropped)
{
    int index=0, index_store =0;
    RTE_SET_USED(port_id);
    RTE_SET_USED(queue_id);
    RTE_SET_USED(enqueue_buf_size);
    RTE_SET_USED(enqueue_buf_count);
    RTE_SET_USED(ev);
    RTE_SET_USED(nb_events);
    RTE_SET_USED(cb_arg);

    uint16_t *eth_type_p;
    uint16_t dscp;

    /* validate system ov memory ptr */
    if(!adk_sys_dp_ovp_table_p) {
        /* dont run the overload protection is ADK system is not initialized */
        *nb_dropped = 0;
        return (nb_events);
    }
    /* Default start with Green zone*/
    if(enqueue_buf_count ==0) {
        *nb_dropped = 0;
        adk_sys_dp_ovp_table_p->non_marked_pkts +=nb_events;
        return (nb_events);
    } else {
        switch(adk_sys_dp_ovp_table_p->mode){
            case 0x01: 
            {
                /* Yellow zone*/
                /* System wide global policy run for each packet*/
                /* if IP_DSCP based policy*/
                /* Treat All DSCP with same priority. Hence parsing is done to
                * identify IPvX Vs non-IP packet 
                */
                for(index=0;index<nb_events; index++) {
                    eth_type_p = rte_pktmbuf_mtod_offset(ev[index].mbuf, uint16_t *, 12 );

                    if(*eth_type_p == 0x0081) {
                        /* Single tagged */
                        eth_type_p = rte_pktmbuf_mtod_offset(ev[index].mbuf, uint16_t *, 16 );
                        if(*eth_type_p == 0x0081) {
                            /* double tagged */

                            eth_type_p = rte_pktmbuf_mtod_offset(ev[index].mbuf, uint16_t *, 20 );
                        }
                    }

                    /* Extract dscp  field same way from header whether it is IPv4/IPv6*/
                    dscp = *(eth_type_p+1);

                    if((*eth_type_p == 0x0008) || (*eth_type_p == 0xDD86)){
                        /* IPv4/6 packet*/
                        if(adk_sys_dp_ovp_table_p->prio_to_drop_map_g[0].action_drop)
                        {
                            rte_pktmbuf_free(ev[index].mbuf);
                            (*nb_dropped)++;
                        }
                        else
                        {
                            ev[index_store++] = ev[index];
                        }
                        adk_sys_dp_ovp_table_p->marked_dscp_pkts++;
                    } else {
                        /* non-ip packet */
                        if(adk_sys_dp_ovp_table_p->default_action_drop)
                        {
                            rte_pktmbuf_free(ev[index].mbuf);
                            (*nb_dropped)++;
                        }
                        else
                        {
                            ev[index_store++] = ev[index];
                        }
                        adk_sys_dp_ovp_table_p->marked_non_dscp_pkts++;
                    }
                }
            } break;
            case 0x02: 
            {
                /* Yellow zone*/
                /* System wide global policy run for each packet*/
                /* if IP_DSCP based policy*/
                /* Treat All DSCP with same priority. Hence parsing is done to
                * identify IPvX Vs non-IP packet 
                */
                for(index=0;index<nb_events; index++) {
                    eth_type_p = rte_pktmbuf_mtod_offset(ev[index].mbuf, uint16_t *, 12 );

                    if(*eth_type_p == 0x0081) {
                        /* Single tagged */
                        eth_type_p = rte_pktmbuf_mtod_offset(ev[index].mbuf, uint16_t *, 16 );
                        if(*eth_type_p == 0x0081) {
                            /* double tagged */
                            eth_type_p = rte_pktmbuf_mtod_offset(ev[index].mbuf, uint16_t *, 20 );
                        }
                    }

                    /* Extract dscp  field same way from header whether it is IPv4/IPv6*/
                    dscp = *(eth_type_p+1);

                    if(*eth_type_p == 0x0008){
                        /* IPv4 packet*/
                        /* get the dscp value*/
                        dscp = (dscp & 0xFC00) >> 10;
                        if(adk_sys_dp_ovp_table_p->prio_to_drop_map_g[dscp].action_drop)
                        {
                            rte_pktmbuf_free(ev[index].mbuf);
                            (*nb_dropped)++;
                        }
                        else
                        {
                            ev[index_store++] = ev[index];
                        }
                    } else if(*eth_type_p == 0xDD86){
                        /* IPv6 packet*/
                        /* get the dscp value*/
                        dscp = (((dscp & 0x000F) << 2) | ((dscp & 0xC000) >> 14));
                        if(adk_sys_dp_ovp_table_p->prio_to_drop_map_g[dscp].action_drop)
                        {
                            rte_pktmbuf_free(ev[index].mbuf);
                            (*nb_dropped)++;
                        }
                        else
                        {
                            ev[index_store++] = ev[index];
                        }
                    } else {
                        /* non-ip packet */
                        if(adk_sys_dp_ovp_table_p->default_action_drop)
                        {
                            rte_pktmbuf_free(ev[index].mbuf);
                            (*nb_dropped)++;
                        }
                        else
                        {
                            ev[index_store++] = ev[index];
                        }
                    }
                }
            } break;

            default:
                *nb_dropped = 0;
                return (nb_events);
                break;
        }
    }

    return index_store;
}
#endif


#ifdef ADK_QOS_GEN3_INTERFACE

/*!
 * ====================================================================
 * adk_sys_dp_queue_stats_update
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * Funxtion to update queue stats drop/xmt pkt & bytes
 * ====================================================================
 */
void adk_sys_dp_queue_stats_update(struct rte_mbuf *mbuf, uint8_t policeColor)
{
    adk_metadata_t * adk_md_p = ADK_METADATA_PTR(mbuf);
    uint16_t id = 
        adk_md_p->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_QUEUE_LOC].stats_id &
                ADK_SYS_EGRESS_STATS_ID_MASK;

    uint8_t version = (adk_md_p->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_QUEUE_LOC].stats_id >>
                        ADK_SYS_EGRESS_STATS_ID_LEN) &
                        ADK_SYS_EGRESS_STATS_VERSION_MASK;
    int8_t len_corr = adk_md_p->egress_stats[ADK_SYS_EGRESS_STATS_MDATA_QUEUE_LOC].len_correction;

    if(id >= ADK_SYS_EGRESS_STATS_ID_MAX) {
        /* Inavalid stats id, Use the global error stats id */
        id = 0;
    } else if((adk_sys_dp_egress_stats_p[id].version &
                ADK_SYS_EGRESS_STATS_VERSION_MASK) != version) {
        /* Inavalid stats id, Use the global error stats id */
        id = 0;
    }

    if (id) {
        if (policeColor==ADK_QOS_DP_RED) {
            adk_sys_dp_egress_stats_p[id].pkt_drops++;
            adk_sys_dp_egress_stats_p[id].byte_drops +=
                        (mbuf->pkt_len + len_corr);
        } else {
            adk_sys_dp_egress_stats_p[id].pkt_xmits++;
            adk_sys_dp_egress_stats_p[id].byte_xmits +=
                        (mbuf->pkt_len + len_corr);
        }
    }
    return;
}
#endif

/*!
 * ====================================================================
 * adk_sys_dp_slowpath_protection
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * Overload protection call back funtion. called by Ethernet-BD at Rx for 
 * every packet to check drop eligibility of that packet.
 * ====================================================================
 */

enum rte_bridge_cppi_cb_ret_type
adk_sys_dp_slowpath_protection(uint16_t dev_id,
        uint16_t queue_id,
        uint32_t enqueue_buf_size,
        uint32_t enqueue_buf_count,
        struct rte_mbuf **mbuf,
        uint16_t nb_mbuf,
        void *cb_arg,
        enum rte_bridge_cppi_cb_action action[])
{

    enum rte_bridge_cppi_cb_ret_type ret_type = RTE_BRIDGE_CPPI_CB_RET_ALL;
    int index=0;
    adk_sif_dp_shm_info_t *shmInfo_p;
    adk_slowpath_user_param_t params;
    void *user_param = &params;

    RTE_SET_USED(dev_id);
    RTE_SET_USED(queue_id);
    RTE_SET_USED(enqueue_buf_size);
    RTE_SET_USED(enqueue_buf_count);
    RTE_SET_USED(mbuf);
    RTE_SET_USED(nb_mbuf);
    RTE_SET_USED(cb_arg);

    /*get DP configuration, it is not expected to be NULL*/
    shmInfo_p = (adk_sif_dp_shm_info_t *) adkSysShmConf_p->sifDpShmHandle;
    if (NULL == shmInfo_p) {
        return RTE_BRIDGE_CPPI_CB_RET_ALL;
    }
    /*Loop through all mbuf/packets  */
    for(index = 0;index<nb_mbuf;index++) {
        struct rte_mbuf *mbp = mbuf[index];
        adk_slowpath_dev_dp_metadata_t *spDevMeta_p;
        uint8_t policeColor = ADK_QOS_DP_RED;

        spDevMeta_p = (adk_slowpath_dev_dp_metadata_t *)&(ADK_METADATA_PARAM(mbp,
                    imc_param_start));

        if(spDevMeta_p->policerFlowId != UINT_MAX) {
            adk_qos_dp_policing_srtcm_color_blind_pps(spDevMeta_p->policerFlowId,
                    &policeColor);
#ifdef ADK_QOS_GEN3_INTERFACE
            adk_sys_dp_queue_stats_update(mbp,policeColor);
#endif

            if(ADK_QOS_DP_GREEN == policeColor) {
                action[index] = RTE_BRIDGE_CPPI_CB_ENQUEUE;
            } else {
                action[index] = RTE_BRIDGE_CPPI_CB_DROP;
                ret_type = RTE_BRIDGE_CPPI_CB_RET_PARTIAL;
            }
        } else {
            action[index] = RTE_BRIDGE_CPPI_CB_ENQUEUE;
        }

        params.flow_id = spDevMeta_p->flowId;
        params.pkt_color = policeColor;
        params.policeFlowId = spDevMeta_p->policerFlowId;

        if (spDevMeta_p->moduleId == ADK_IPV6_MODULE_ID)
        {
            params.l4_len = spDevMeta_p->l4_len;
            params.l4_protocol = spDevMeta_p->l4_protocol;
            shmInfo_p->moduleSlowPathCbFn[ADK_IPV6_MODULE_ID](user_param);
        }
        else if (spDevMeta_p->moduleId == ADK_IP_MODULE_ID)
        {
            params.l4_len = spDevMeta_p->l4_len;
            params.l4_protocol = spDevMeta_p->l4_protocol;
            shmInfo_p->moduleSlowPathCbFn[ADK_IP_MODULE_ID](user_param);
        }
        else if (spDevMeta_p->moduleId == ADK_ETHERNET_MODULE_ID)
        {
            shmInfo_p->moduleSlowPathCbFn[ADK_ETHERNET_MODULE_ID](user_param);
        }
    }
    return ret_type;
}

#if defined(ADK_DP_MAINTENANCE)

/*!
 * ============================================================================
 * adk_dp_maintenance_fn
 * @param[in] event_dev_id  event device id.
 * @param[in] event_port_id event port id.
 * @param[in] arg           The call back data that was registered
 * when this function was registered with the RTE event framework.
 *
 * @retval None
 *
 * @brief  ADK dataplane periodic callback function to be registered with
 * RTE event framework by application. This function manages the ADK dataplane
 * maintenance functions like RCU.
 *
 * ============================================================================
 */

int
adk_dp_maintenance_fn(
	uint8_t event_dev_id,
	uint8_t event_port_id,
	void *arg)
{
    RTE_SET_USED(event_dev_id);
    RTE_SET_USED(event_port_id);
    RTE_SET_USED(arg);

    adk_rcu_quiescent_state();
    
    return 0;
}

static void adk_sys_dp_increment_qos_counters(struct rte_mbuf *pkt,
                                              void *user_param)
{
    RTE_SET_USED(user_param);
    adk_metadata_t * adk_md_p = ADK_METADATA_PTR(pkt);
    enum rte_mbuf_tx_enqueue_status enqueueStatus = rte_pktmbuf_tx_enqueue_status(pkt); 
    bool pktDropped = 0;
    bool isIpPacket = adk_md_p->isIpPkt;
    uint16_t pktLength = 0;

    if (isIpPacket) {
        pktLength = adk_md_p->outerIpPktLen;
    }else {
        pktLength = rte_pktmbuf_pkt_len(pkt);
    }
    pktDropped = (enqueueStatus == RTE_MBUF_TX_ENQUEUE_SUCCESS) ? false : true;

    if (adk_md_p->isPHBQueue) {
        /*Call QoS DP API to increment the PHB Counters
          If queue is PHB queue then counters will be incremented inside the called function*/
        adk_qos_dp_increment_phb_counters(adk_md_p->outerIPVersion, adk_md_p->qosStatsId, pktLength, pktDropped); 
    }else {
        adk_qos_dp_increment_sl_queue_counters(isIpPacket, adk_md_p->outerIPVersion, adk_md_p->qosStatsId, pktLength, pktDropped); 
    }
}

/*!
 * ============================================================================
 * adk_sys_dp_egress_stats_enqueue_cb
 *
 * @brief  ADK dataplane Tx callback function to be registered with Eth Dev.
 * This function manages the egress stats for enqueue.
 *
 * ============================================================================
 */
#ifdef ADK_DEVICE_SNR
void
adk_sys_dp_egress_stats_enqueue_cb(
    struct rte_mbuf *pkt,
    void *user_param)
{
    ADK_DP_MEASURE_BLOCK_CYCLE_SYS_EGRESS_STATS_RX_TS();
    adk_egress_user_param_t params_v6 = {0};
    adk_egress_user_param_t params = {0};
    adk_metadata_t * adk_md_p = ADK_METADATA_PTR(pkt);
    adk_sif_dp_shm_info_t *shmInfo_p;
    user_param = &params; // Not used for QoS counters

    /*get DP configuration, it is not expected to be NULL*/
    shmInfo_p = (adk_sif_dp_shm_info_t *) adkSysShmConf_p->sifDpShmHandle;
    if (NULL == shmInfo_p) {
        return;
    }
    for(int i=0; i<ADK_STATS_MDATA_ID_MAX; ++i)
    {
        if(adk_md_p->egress_mdata_stats[i].stats_id == ADK_SYS_INVALID_MDATA_ID)
            continue;

        switch(i) {
            case ADK_STATS_MDATA_ID_INTERFACE_STATS:
                if (ADK_IPV4_VER == adk_md_p->outerIPVersion) {
                    params.stats_mask |= ADK_STATS_IP_INTERFACE_MASK;
                }
                else {
                    params_v6.stats_mask |= ADK_STATS_IP_INTERFACE_MASK;
                }
                break;
            case ADK_STATS_MDATA_ID_INNER_IP_STATS:
                if (ADK_IPV4_VER == adk_md_p->innerIPVersion) {
                    params.stats_mask |= ADK_STATS_INNER_IP_ADDRESS_MASK;
                }
                else {
                    params_v6.stats_mask |= ADK_STATS_INNER_IP_ADDRESS_MASK;
                }
                break;
            case ADK_STATS_MDATA_ID_OUTER_IP_STATS:
                if (ADK_IPV4_VER == adk_md_p->outerIPVersion) {
                    params.stats_mask |= ADK_STATS_OUTER_IP_ADDRESS_MASK;
                }
                else {
                    params_v6.stats_mask |= ADK_STATS_OUTER_IP_ADDRESS_MASK;
                }
                break;
            default:
                /*Add more cases if needed*/
                break;
        }
    }

    if ((params.stats_mask) &&
        (shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID]))
    {
        user_param = &params; 
        shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID](pkt, user_param);
    }

    if ((params_v6.stats_mask) &&
        (shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID]))
    {
        user_param = &params_v6;
        shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID](pkt, user_param);
    }

    if ((ADK_SYS_INVALID_MDATA_ID != 
                adk_md_p->egress_mdata_stats[ADK_STATS_MDATA_ID_ETHERNET_INTERFACE_STATS].stats_id) &&
            (shmInfo_p->moduleEgressCbFn[ADK_ETHERNET_MODULE_ID]))
    {
        params.stats_mask = ADK_STATS_ETHERNET_INTERFACE_MASK;
        user_param = &params;
        shmInfo_p->moduleEgressCbFn[ADK_ETHERNET_MODULE_ID](pkt, user_param);
    }

    /*Increment the QoS PHB counters here*/
    adk_sys_dp_increment_qos_counters(pkt, user_param);     

#ifdef ADK_DEVICE_SNR
    /* Generate event to netd only incase of pkt drops*/
    if(rte_pktmbuf_tx_enqueue_status(pkt) != RTE_MBUF_TX_ENQUEUE_SUCCESS)
    {
        /* If timestamp capture is enabled,
         * send TX drop notification to Netdriver */
        if(ADK_MD_GET_TS_MSG_TYPE(pkt) == ADK_NETD_TS_CAPTURE_IDX_ALLOC ||
                ADK_MD_GET_TS_MSG_TYPE(pkt) == ADK_NETD_TS_CAPTURE)
        {
            if(ADK_MD_GET_TX_NOTIFY_FN_ID(pkt)) {
                /* If function id set by Ethernet is valid */
                (adkDestEntryFunction_g[ADK_MD_GET_TX_NOTIFY_FN_ID(pkt)])(
                        (void **)&pkt, 1, 0x1, NULL, NULL, ADK_LCORE_INDEX);
            }
        }
    }
#endif
#endif
    ADK_DP_MEASURE_BLOCK_CYCLE_SYS_EGRESS_STATS_TX_TS();
}
#else
uint16_t
adk_sys_dp_egress_stats_cb(
#if ADK_DPDK_VERSION >= 18050016
    uint16_t port_id,
#else
    uint8_t port_id,
#endif
    uint16_t queue,
    struct rte_mbuf *pkts[],
    uint16_t nb_pkts,
    void *user_param)
{
    RTE_SET_USED(port_id);
    RTE_SET_USED(queue);
    RTE_SET_USED(user_param);

    for(int p=0; p<nb_pkts && pkts[p]; ++p)
    {
        adk_metadata_t * adk_md_p = ADK_METADATA_PTR(pkts[p]);
        for(int i=0; i<ADK_SYS_EGRESS_STATS_MDATA_ID_MAX; ++i)
        {
            if(adk_md_p->egress_stats[i].flags)
            {
                uint16_t id = adk_md_p->egress_stats[i].stats_id &
                            ADK_SYS_EGRESS_STATS_ID_MASK;
                uint8_t version = (adk_md_p->egress_stats[i].stats_id >>
                                   ADK_SYS_EGRESS_STATS_ID_LEN) &
                                  ADK_SYS_EGRESS_STATS_VERSION_MASK;

                if(id >= ADK_SYS_EGRESS_STATS_ID_MAX)
                {
                    /* Inavalid stats id, Use the global error stats id */
                    id = 0;
                }
                else if((adk_sys_dp_egress_stats_p[id].version &
                         ADK_SYS_EGRESS_STATS_VERSION_MASK) != version)
                {
                    /* Inavalid stats id, Use the global error stats id */
                    id = 0;
                }

                switch(adk_md_p->egress_stats[i].flags)
                {
                case ADK_SYS_EGRESS_STATS_MDATA_FLAG_PKT_CNTR:
                    adk_sys_dp_egress_stats_p[id].pkt_xmits++;
                    break;

                case ADK_SYS_EGRESS_STATS_MDATA_FLAG_BYTES_CNTR:
                    adk_sys_dp_egress_stats_p[id].byte_xmits +=
                        (pkts[p]->pkt_len +
                         adk_md_p->egress_stats[i].len_correction);
                    break;

                case ADK_SYS_EGRESS_STATS_MDATA_FLAG_TS:
                    adk_sys_dp_egress_stats_p[id].timestamp = rte_rdtsc();
                    break;

                case (ADK_SYS_EGRESS_STATS_MDATA_FLAG_PKT_CNTR|
                      ADK_SYS_EGRESS_STATS_MDATA_FLAG_BYTES_CNTR):
                    adk_sys_dp_egress_stats_p[id].pkt_xmits++;
                    adk_sys_dp_egress_stats_p[id].byte_xmits +=
                        (pkts[p]->pkt_len +
                         adk_md_p->egress_stats[i].len_correction);
                    break;

                default:
                    /*Add more cases if needed*/
                    break;
                }
            }
        }
#ifdef ADK_DEVICE_SNR
        /* If timestamp capture is enabled,
         * send TX complete notification to Netdriver */
        if(ADK_MD_GET_TS_MSG_TYPE(pkts[p]) == ADK_NETD_TS_CAPTURE_IDX_ALLOC || ADK_MD_GET_TS_MSG_TYPE(pkts[p]) == ADK_NETD_TS_CAPTURE)  
        {
            if(ADK_MD_GET_TX_NOTIFY_FN_ID(pkts[p])) { /* If function id set by Ethernet is valid */
                (adkDestEntryFunction_g[ADK_MD_GET_TX_NOTIFY_FN_ID(pkts[p])])((void **)&pkts[p], 1, 0x1, NULL, NULL, ADK_LCORE_INDEX); 
            }    
        }
#endif
    }

    return nb_pkts;
}
#endif

#if defined(ADK_DEVICE_SNR) && defined(ADK_COMPLETION_CALLBACK_SUPPORT) 
/*!
 * ============================================================================
 * adk_sys_dp_egress_stats_xmit_cb
 *
 * @brief  ADK dataplane Tx callback function to be registered with Eth Dev.
 * This function manages the egress stats for xmit counters.
 *
 * ============================================================================
 */
void
adk_sys_dp_egress_stats_xmit_cb(
    struct rte_mbuf *pkt,
    void *user_param)
{
    adk_egress_user_param_t params;
    adk_metadata_t * adk_md_p = ADK_METADATA_PTR(pkt);
    adk_sif_dp_shm_info_t *shmInfo_p;
    params.xmit_cb = 1;
    user_param = &params;

    /*get DP configuration, it is not expected to be NULL*/
    shmInfo_p = (adk_sif_dp_shm_info_t *) adkSysShmConf_p->sifDpShmHandle;

    if (NULL == shmInfo_p) {
        return;
    }

    for(int i=0; i<ADK_STATS_MDATA_ID_MAX; ++i)
    {
        if(adk_md_p->egress_mdata_stats[i].stats_id == ADK_SYS_INVALID_MDATA_ID)
            continue;

        switch(i) {
            case ADK_STATS_MDATA_ID_INTERFACE_STATS:
                /*Interface stats is always based on Outer IP*/
                if(adk_md_p->outerIPVersion == ADK_IPV4_VER)
                {
                    if (shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID])
                    {
                        params.stats_id = ADK_STATS_MDATA_ID_INTERFACE_STATS;
                        shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID](pkt, user_param);
                    }
                }
                else if(adk_md_p->outerIPVersion == ADK_IPV6_VER)
                {
                    if (shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID])
                    {
                        params.stats_id =  ADK_STATS_MDATA_ID_INTERFACE_STATS;
                        shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID](pkt, user_param);
                    }
                }

                break;

            case ADK_STATS_MDATA_ID_OUTER_IP_STATS:
                if((adk_md_p->outerIPVersion == ADK_IPV4_VER) && (shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID]))
                {
                    params.stats_id =  ADK_STATS_MDATA_ID_OUTER_IP_STATS;
                    shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID](pkt, user_param);
                }
                else if((adk_md_p->outerIPVersion == ADK_IPV6_VER) && (shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID]))
                {
                    params.stats_id =  ADK_STATS_MDATA_ID_OUTER_IP_STATS;
                    shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID](pkt, user_param);
                }
                break;

            case ADK_STATS_MDATA_ID_INNER_IP_STATS:
                if((adk_md_p->innerIPVersion == ADK_IPV4_VER) && (shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID]))
                {
                    params.stats_id = ADK_STATS_MDATA_ID_INNER_IP_STATS;
                    shmInfo_p->moduleEgressCbFn[ADK_IP_MODULE_ID](pkt, user_param);
                }
                else if((adk_md_p->innerIPVersion == ADK_IPV6_VER) && (shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID]))
                {
                    params.stats_id = ADK_STATS_MDATA_ID_INNER_IP_STATS;
                    shmInfo_p->moduleEgressCbFn[ADK_IPV6_MODULE_ID](pkt, user_param);
                }
                break;
            case ADK_STATS_MDATA_ID_ETHERNET_INTERFACE_STATS:
                    if (shmInfo_p->moduleEgressCbFn[ADK_ETHERNET_MODULE_ID])
                    {
                        params.stats_id = ADK_STATS_MDATA_ID_ETHERNET_INTERFACE_STATS;
                        shmInfo_p->moduleEgressCbFn[ADK_ETHERNET_MODULE_ID](pkt, user_param);
                    }
                    break;
            default:
                /*Add more cases if needed*/
                break;
        }
    }
    /* Generate event to netd only incase of pkt drops*/
    if(rte_pktmbuf_tx_cmpltn_status(pkt) != RTE_MBUF_TX_CMPLTN_SUCCESS)
    {
        /* If timestamp capture is enabled,
         * send TX drop notification to Netdriver */
        if(ADK_MD_GET_TS_MSG_TYPE(pkt) == ADK_NETD_TS_CAPTURE_IDX_ALLOC ||
                ADK_MD_GET_TS_MSG_TYPE(pkt) == ADK_NETD_TS_CAPTURE)
        {
            if(ADK_MD_GET_TX_NOTIFY_FN_ID(pkt)) {
                /* If function id set by Ethernet is valid */
                (adkDestEntryFunction_g[ADK_MD_GET_TX_NOTIFY_FN_ID(pkt)])(
                        (void **)&pkt, 1, 0x1, NULL, NULL, ADK_LCORE_INDEX);
            }
        }
    }
}
#endif

/* Per Core DPDK Timer Array */
struct rte_timer adk_per_core_timer[ADK_MAX_NUM_CORES];

/*!
 * ====================================================================
 * adk_dp_maintenance_init_v2
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * This function initializes the ADK maintenance manager and
 * create 1ms DPDK timer for the given worker core.
 * ====================================================================
 */
adk_status_t
adk_dp_maintenance_init_v2(
    uint8_t lcore_id)
{
    uint64_t    ticks;
    int         ret;
    int32_t     coreIndex;

    coreIndex = ADK_LCORE_INDEX_LOOKUP(lcore_id);

    if (-1 == coreIndex)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Error: coreIndex cannot be -1 for DP cores\n");
        return(ADK_E_INTERNAL_ERROR);
    }

#ifdef ADK_TIMER
    adk_timer_init(coreIndex, adk_get_cycle_cntr());
#endif

    rte_timer_init(&adk_per_core_timer[coreIndex]);

    /* Calculate ticks for 1ms */
    ticks = rte_get_timer_hz()/1000;

    ret = rte_timer_reset(&adk_per_core_timer[coreIndex], ticks, PERIODICAL,
                          lcore_id, adk_dp_maintenance_fn_v2, NULL);
    if (ret != 0)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "Error: Failed to install RCU periodic timer\n");
        return(ADK_E_INTERNAL_ERROR);
    }

    return ADK_SUCCESS;
}

/*!
 * ====================================================================
 * adk_dp_maintenance_fn_v2
 *
 * @param[in] iteration_frequency   The frequency on which this function
 *                                  is invoked.
 *
 * @retval ADK_SUCCESS              Successful completion
 *
 * @brief
 *
 * This function manages the ADK dataplane maintenance functions like
 * timer, RCU, etc.. It is invoked on expiry of DPDK timer created on
 * adk_dp_maintenance_init_v2.
 * ====================================================================
 */
void
adk_dp_maintenance_fn_v2(
    struct rte_timer __attribute__((unused)) *timer,
    void __attribute__((unused)) *args)
{

#ifdef ADK_TIMER
    adk_timer_proc(ADK_LCORE_INDEX, adk_get_cycle_cntr());
#endif
#ifdef ADK_RCU
    adk_rcu_quiescent_state();
#endif
    adk_sys_dp_flow_cache_monitor();
    return;
}

/*!
 * ====================================================================
 * @brief
 *
 * This function installs a timer.
 *
 * ====================================================================
 */
adk_dp_status_t
adk_sys_dp_service_timer_install(
        adk_sys_timer_dev_info_t        *timer_dev_info,
 	adk_sys_dp_service_timer_type_e	timer_type,
	uint64_t			time_value,
 	uint8_t				timer_handler_entry_fn_id,
 	uint16_t			load_balance_id,
 	uintptr_t			user_op_data,
 	uint16_t			event_queue_id,
        adk_handle_t		        *timer_handle)
{

    adk_dp_status_t retValue = ADK_SUCCESS;
    struct rte_event_dev_timer *eventTimer = NULL;
    uint8_t adkTimerBdInstance_g;

    enum rte_timer_type type;
#if 0
    uint8_t  evq_id;
#endif

    /*Validate input params */
    switch (timer_type)
    {
        case ADK_SYS_DP_SERVICE_TIMER_TYPE_SINGE_SHOT:
            type = SINGLE;
            break;
        case ADK_SYS_DP_SERVICE_TIMER_TYPE_PEARODIC:
            type = PERIODICAL;
            break;
        default:
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Invalid timer type\n");
            return ADK_DP_E_ERROR;
            break;
    }

    if(timer_dev_info->type == ADK_SYS_TIMER_DEV_TYPE_EVNT_TIMER_BD) {
        adkTimerBdInstance_g = timer_dev_info->timer_bd_instance_id;
    } else {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Invalid timer device type\n");
	    return ADK_DP_E_ERROR;
    }

    eventTimer = rte_malloc(NULL, sizeof(struct rte_event_dev_timer), 0);

    if (eventTimer == NULL)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Memory allocation failed\n");
        return ADK_DP_E_ERROR;
    }

    rte_event_dev_timer_init(eventTimer);
    rte_event_dev_timer_util_set_instance_id(eventTimer, adkTimerBdInstance_g);

#if 0
    adk_pal_dp_stage_evq_id_get(timer_handler_entry_fn_id,
                                &evq_id);
    eventTimer->event.queue_id = evq_id;
#endif

#if 1
    /*
     * HACK: Getting the EventQueue from DSI metadata
     *
     * The Eventdev queue can be shared by multiple processing functions.
     * For Example: TWAMP LLI and TWAMP Session age out handler must share the
     * same queue. PAL will provide an init time API in which all the processing
     * functions attached to same queue will be grouped together. This grouping
     * concept will avoid the calling of adk_pal_stage_entry_update() API by the
     * application in the default run-to-completion/Single-stage model. The
     * above said PAL API must be called only for core-to-core/Multi-stage
     * model.
     */

    eventTimer->event.queue_id = event_queue_id;
#endif

    /* The timer installation can only be called from DP process context as
    * the call back function is registered here.
    *
    * eventTimer->drop_cb = event_timer_drop_cb;
    */
    eventTimer->drop_cb = NULL;

    eventTimer->event.event_ptr = (void *)user_op_data;
    eventTimer->event.sched_type = RTE_SCHED_TYPE_ATOMIC;
    eventTimer->event.flow_id = load_balance_id;
    eventTimer->event.sub_event_type = timer_handler_entry_fn_id;

    ADK_DP_LOG_INFO_DEBUG(ADK_SYS_MODULE_ID, "In Timer Event Generator\n");

    if( rte_event_dev_timer_reset(eventTimer, time_value, type))
    {
        retValue = ADK_DP_E_ERROR;
        if(eventTimer !=NULL)
        {
            rte_free(eventTimer);
            eventTimer = NULL;
        }
    }

    *timer_handle = (adk_handle_t) (void *)eventTimer;

    return retValue;
}

adk_dp_status_t
adk_sys_dp_service_timer_remain_time(adk_handle_t  timer_handle,
                                     uint64_t *remain_time)
{
    uint64_t    curr_time;
    struct rte_event_dev_timer *eventTimer = NULL;

    if(timer_handle == NULL)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Invalid timer handle\n");
        return ADK_DP_E_ERROR;
    }

    eventTimer = (struct rte_event_dev_timer *)timer_handle;

    adk_dp_measure_block_cycle_rx_ts(&curr_time);

    *remain_time = (eventTimer->timer_impl.expire - curr_time );

    ADK_RETURN(ADK_SUCCESS);

}


adk_dp_status_t
adk_sys_dp_service_timer_uninstall(adk_handle_t  timer_handle)
{

    int ret = 0;
    if(timer_handle == NULL)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Invalid timer handle\n");
        return ADK_DP_E_ERROR;
    }

    ret = rte_event_dev_timer_stop((struct rte_event_dev_timer *)timer_handle);
    if(ret!=0)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "rte_event_dev_timer_stop failed\n");
        ADK_RETURN(ADK_DP_E_ERROR);
    }

    rte_free(timer_handle);

    ADK_RETURN(ADK_SUCCESS);

}

adk_dp_status_t
adk_sys_dp_service_timer_reset(adk_handle_t  timer_handle,
                            adk_sys_dp_service_timer_type_e	timer_type,
                            uint64_t  new_time_value)
{

    int ret = 0;
    struct rte_event_dev_timer *eventTimer;
    enum rte_timer_type type;

    eventTimer = (struct rte_event_dev_timer *) timer_handle;

    if(timer_handle == NULL)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Invalid timer handle\n");
        return ADK_DP_E_ERROR;
    }

    /*Validate input params */
    switch(timer_type)
    {
        case ADK_SYS_DP_SERVICE_TIMER_TYPE_SINGE_SHOT:
            type = SINGLE;
            break;
        case ADK_SYS_DP_SERVICE_TIMER_TYPE_PEARODIC:
            type = PERIODICAL;
            break;
        default:
            ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID, "Invalid timer type\n");
            return ADK_DP_E_ERROR;
            break;
    }




    ret = rte_event_dev_timer_reset(eventTimer, new_time_value, type);
    if (ret != 0)
    {
        ADK_DP_LOG_ERROR(ADK_SYS_MODULE_ID,
                         "rte_event_dev_timer_reset failed\n");
        ADK_RETURN(ADK_DP_E_ERROR);
    }

    ADK_RETURN(ADK_SUCCESS);
}

#ifdef ADK_PAL_CYCLE_MEASUREMENT
adk_dp_status_t
adk_sys_dp_proc_cycle_measure(struct rte_mbuf *dpdk_mbuf_p)
{
    if(adk_pal_dp_info_p->xstats_state) {
        uint8_t  fn_id;
        uint8_t  coreIndex;
        uint32_t  diff;
        uint64_t  now = rte_rdtsc();
        uint32_t  base = 0;
        static uint32_t index[ADK_MAX_PROC_FUNC_ID][ADK_MAX_NUM_CORES];
        coreIndex = ADK_LCORE_INDEX;


        fn_id = ADK_METADATA_PARAM(dpdk_mbuf_p, proc_stage_idx);
#ifdef  ADK_PAL_CYCLE_MEASUREMENT_MBUF_UDATA
        diff = now - dpdk_mbuf_p->udata64;
#else
        diff = now - (ADK_METADATA_PARAM(dpdk_mbuf_p, stage_rx_ts));
#endif
        base = (fn_id * ADK_MAX_NUM_CORES * ADK_PAL_MAX_SAMPLES ) +
            (coreIndex * ADK_PAL_MAX_SAMPLES);
        base = base + (index[fn_id][coreIndex] % ADK_PAL_MAX_SAMPLES);
        adk_pal_dp_min_cycles[base] = diff;
        index[fn_id][coreIndex]++;
#if 0
        adk_pal_dp_min_cycles[base+1]++;
#endif
    } else {
        RTE_SET_USED(dpdk_mbuf_p);
    }
    ADK_RETURN(ADK_SUCCESS);
}
#endif


/*!
 * =============================================================
 * @internal
 * adk_sif_get_ntp_time
 *
 * @param[in]  cpuTicks         CPU time in ticks.
 *                              The current CPU time can be obtained using the
 *                              DPDK API rte_get_tsc_cycles().
 *
 * @param[out] ntpTime          NTP Time
 *
 *
 * @retval ADK_SUCCESS Successful completion
 *
 * @brief
 *
 * This function will convert the CPU time ticks to NTP time. The NTP time is
 * computed by applying the delta CPU time ticks in data plane, to the base NTP
 * time ticks precomputed in the control plane. The base NTP time is a relative
 * time i.e. CPU ticks equivalent to the linux time.
 *
 * =============================================================
 */
adk_dp_status_t
adk_sif_get_ntp_time( uint64_t cpuTicks,
                      adk_sif_ntp_time_t *ntpTime)
{
    uint64_t deltaTicks = 0;
    uint64_t delta_in_micro_sec = 0;
    uint64_t delta_in_sec = 0;
    uint64_t ticksPerMicroSec;
    uint64_t ntpFracSec;
    adk_sif_dp_shm_info_t *shmInfo;
    adk_sif_base_ntp_info_t *baseTime;

    shmInfo = (adk_sif_dp_shm_info_t *) adkSysShmConf_p->sifDpShmHandle;
    ticksPerMicroSec = shmInfo->cpuTicksPerSec / 1000000;

    /* Atomically read the baseTime pointer */
    baseTime = (adk_sif_base_ntp_info_t *)rte_atomic64_read(&shmInfo->baseTime);

    /*
     * Delta between CPU clock ticks and relative CPU clock ticks since
     * Epoch time that was retrieved from linux.
     *
     * Note: Due to burst processing the timestamp updated in 1st packet of
     * the batch will have more delta ticks than in subsequent packets in a
     * batch. The application has to measure the performance by
     * averaging several thousands of packets.
     */
    deltaTicks = cpuTicks - baseTime->epochCpuTicks;

    delta_in_micro_sec = deltaTicks / ticksPerMicroSec;
    delta_in_sec = delta_in_micro_sec / 1000000;

    delta_in_micro_sec = delta_in_micro_sec - (delta_in_sec * 1000000);

    /* Adding the delta seconds to NTP base time */
    ntpTime->ntpSeconds = baseTime->baseNtpTimeHi + delta_in_sec;

    /*
     * Ticks per "1 / (2^32 -1) ~= 200 pico seconds". NTP fractional
     * second is of 200 pico seconds granularity.
     * Converting fractional seconds from micro seconds to 200 pico seconds
     * granularity.
     */
    ntpFracSec = delta_in_micro_sec * ADK_SIF_NTP_SCALE_FRAC / 1000000;

    /* Adding the delta fractional seconds to NTP base time */
    ntpFracSec = baseTime->baseNtpTimeLo + ntpFracSec;

    /* Roll over check */
    if(ntpFracSec > 0xFFFFFFFF) {
        ++ntpTime->ntpSeconds;
        ntpFracSec = ntpFracSec - 0xFFFFFFFF;
    }
    ntpTime->ntpFractionSeconds = ntpFracSec;

    return ADK_SUCCESS;
}

void
adk_sys_dp_scratchpad_dump(adk_sif_dp_scratchpad_space_t *scratch_pad_p)
{
    if (scratch_pad_p) {
        ADK_DP_LOG_INFO_1P(ADK_SYS_MODULE_ID, "Flow cache pointer %p ", scratch_pad_p->flow_cache_entry); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "pktPath %u", scratch_pad_p->pktPath); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "ip_version %u", scratch_pad_p->ip_version); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "pktOrigin %u", scratch_pad_p->pktOrigin); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "l4_offset %u", scratch_pad_p->l4_offset); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "mtu %u", scratch_pad_p->mtu);
    }
    return;
}

void adk_sif_dp_pal_batch_cb_fn(
        adk_pktio_task_t *dpdkMbuf,
        uint8_t       cbEventType,
        void          *cbParam,
        bool          sendSuccess)
{
    RTE_SET_USED(cbParam);

    switch(cbEventType)
    {
        case ADK_PAL_DP_BATCH_CB_EVENT_STAGE_ETH_NETD:
            if(!sendSuccess)
            {
                adk_pktio_task_buffer_free(dpdkMbuf);
            }
            break;
        default:
            break;

    }
    return;
}


/*!    end addtogroup _sifDpFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */
