/***************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_sys_dp_flow_cache.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 * @addtogroup _sif  SIF Module
 * @{
 *     @addtogroup _sifDpFunctions  SIF Data Plane Functions
 *     @{
 *
 * @brief
 *
 * This file contains the ADK Dataplane flow cache function.
 ***********************************************************************/

#include <adk_include.h>
#include "adk_sif_dp.h"
#include "adk_sys_lcore.h"
#include <adk_sys_dpdk.h>
#include <adk_shm.h>
#include <adk_pmu.h>
#include <adk_timer.h>
#include <adk_hash.h>

#include <rte_bridge_timer.h>
#include <rte_bridge_eth.h>
#include "adk_pal_dp_export.h"
#include "adk_ipv6_const.h"
#include "adk_ipv6_dp_export.h"
#include "adk_ip_dp_export.h"
#include "adk_pal_dp.h"
#include "adk_sif_dp_log_lib.h"
#include "adk_sif_dp.h"
#include "adk_sys_dp_flow_cache.h"
#include "adk_sys_dp_flow_cache_uli_func.h"

#if ADK_DPDK_VERSION >= 18050016
#include <rte_service.h>
#include <rte_service_component.h>
#endif

/* Macro definition for module level cycle count */
#ifdef SYS_FLOW_CACHE_BLOCK_COUNT_ENABLED 
    uint64_t sys_fc_uli_rx_ts, sys_fc_uli_tx_ts;

    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_FC_ULI_RX_TS() adk_dp_measure_block_cycle_rx_ts(&sys_fc_uli_rx_ts)
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_FC_ULI_TX_TS() adk_dp_measure_block_cycle_tx_ts(&sys_fc_uli_rx_ts, &sys_fc_uli_tx_ts, "SYS FLOW CACHE ULI", PMU_SYS_FC_ULI_MODULE)
#else
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_FC_ULI_RX_TS()
    #define ADK_DP_MEASURE_BLOCK_CYCLE_SYS_FC_ULI_TX_TS()
#endif

adk_dp_status_t
adk_sys_dp_uli_flow_cache_entry_alloc(
        adk_sys_dp_flow_cache_hash_entry_t **uli_flow_cache_hash_entry,
        adk_sys_dp_uli_flow_cache_entry_t **flow_cache_entry,
        uint32_t hash_value)
{
    adk_sys_dp_uli_flow_cache_table_t *uli_flow_cache_table;
    uint16_t first_key, second_key, i, index;
    uint8_t coreIndex = ADK_LCORE_INDEX;

    /* hash lookup */
    uli_flow_cache_table = &adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex];
    first_key = hash_value & ADK_SYS_ULI_FLOW_CACHE_MASK;
    first_key = first_key*ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES;

    *uli_flow_cache_hash_entry = &uli_flow_cache_table->hash_tbl[first_key];
    for(i=0; i<ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES; i++, (*uli_flow_cache_hash_entry)++)
    {
        if(!(*uli_flow_cache_hash_entry)->isActive)
        {
            ADK_SYS_DP_FC_RSRV_IDX(uli_flow_cache_table->free_list,
                                   uli_flow_cache_table->tbl,
                                   index);
            *flow_cache_entry = &uli_flow_cache_table->tbl[index];
            (*uli_flow_cache_hash_entry)->index = index;
            (*uli_flow_cache_hash_entry)->hashValue = hash_value;
            (*uli_flow_cache_hash_entry)->isActive = true;

            /* update the Flow cache hash info */
            (*flow_cache_entry)->hashIdx = first_key+i;
            /* update the Flow cache state */
            (*flow_cache_entry)->state = ADK_SYS_DP_FLOW_CACHE_NEW;
            (*flow_cache_entry)->isActive = true;

            /* set active & aging flag */
            adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.aging_bitmap, index);
            adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.active_flows, index);
            if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.flows_active++;
                adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.alloc++;
                /*Initialize counters*/
                uli_flow_cache_table->fc_entry_stats[index].hit = 0;
                uli_flow_cache_table->fc_entry_stats[index].alloc = 1;
            }

            return ADK_SUCCESS;
        }
    }

    second_key = (hash_value >> ADK_SYS_ULI_FLOW_CACHE_WIDTH) &
        (ADK_SYS_ULI_FLOW_CACHE_MASK);
    second_key = second_key*ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES;

    *uli_flow_cache_hash_entry = &uli_flow_cache_table->hash_tbl[second_key];
    for(i=0; i<ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES; i++, (*uli_flow_cache_hash_entry)++)
    {
        if(!(*uli_flow_cache_hash_entry)->isActive)
        {
            ADK_SYS_DP_FC_RSRV_IDX(uli_flow_cache_table->free_list,
                    uli_flow_cache_table->tbl,
                    index);
            *flow_cache_entry = &uli_flow_cache_table->tbl[index];
            (*uli_flow_cache_hash_entry)->index = index;
            (*uli_flow_cache_hash_entry)->hashValue = hash_value;
            (*uli_flow_cache_hash_entry)->isActive = true;

            /* update the Flow cache hash info */
            (*flow_cache_entry)->hashIdx = second_key+i;
            /* update the Flow cache state */
            (*flow_cache_entry)->state = ADK_SYS_DP_FLOW_CACHE_NEW;
            (*flow_cache_entry)->isActive = true;

            /* set active & aging flag */
            adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.aging_bitmap, index);
            adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.active_flows, index);
            if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.flows_active++;
                adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.alloc++;
                /*Initialize counters*/
                uli_flow_cache_table->fc_entry_stats[index].hit = 0;
                uli_flow_cache_table->fc_entry_stats[index].alloc = 1;
            }

            return ADK_SUCCESS;
        }
    }

    return ADK_E_FLOW_CACHE_ALLOC_FAILED;
}

adk_status_t
adk_sys_dp_uli_flow_cache_entry_release(adk_sys_dp_uli_flow_cache_entry_t *uli_fc_entry)
{
    uint16_t hash_idx;
    uint8_t coreIndex = ADK_LCORE_INDEX;
    adk_sys_dp_flow_cache_hash_entry_t *fc_hash_entry;
    adk_sys_dp_uli_flow_cache_table_t  *uli_fc_table;


    /* release FC Hash entry */
    hash_idx = (uli_fc_entry->hashIdx);
    fc_hash_entry = &adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].hash_tbl[hash_idx];
    *((uint64_t *)fc_hash_entry) = 0;

    /* mark the FC entry disabled */
    uli_fc_entry->isActive = false;
    /* update the Flow cache state */
    uli_fc_entry->state = ADK_SYS_DP_FLOW_CACHE_RELEASED;

    /* release the FC entry index */
    uli_fc_table = &adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex];
    ADK_SYS_DP_FC_RELEASE_IDX(uli_fc_table->free_list,
            uli_fc_table->tbl,
            uli_fc_entry->idx);

    /* reset active flag */
    adk_sys_dp_fc_bitmap_reset(uli_fc_table->age_info.active_flows, uli_fc_entry->idx);
    adk_sys_dp_fc_bitmap_reset(uli_fc_table->age_info.aging_bitmap, uli_fc_entry->idx);
    if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.flows_active--;
        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.freed++;
        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.flows_stale--;
    }

    return(ADK_SUCCESS);
}

static inline void 
adk_sys_dp_flow_cache_key_generate(adk_sys_ip_addr_t *ipAddr,
                                   adk_ip_hdr_type_e hdrType,
                                   uint8_t nextHdr,
                                   uint8_t dscp,
                                   uint16_t srcPort,
                                   uint16_t dstPort,
                                   uint16_t *first_key,
                                   uint16_t *second_key,
                                   uint32_t *hash_key)
{
    uint64_t hash_value = 0;

    if (ADK_IPV4 == hdrType) {
        hash_value = SISD_CRC_HASH64(hash_value,
                *((uint64_t *) &ipAddr->v4Addr.srcAddr));
    } else {
        hash_value = SISD_CRC_HASH64(hash_value,
                *((uint64_t *) &ipAddr->v6Addr.srcAddr[0]));
        hash_value = SISD_CRC_HASH64(hash_value,
                *((uint64_t *) &ipAddr->v6Addr.srcAddr[2]));
        /*TODO: Why DIP is not considered for hash key genr 
         * for IPv4 , but only for IPv6??*/
        hash_value = SISD_CRC_HASH64(hash_value,
                *((uint64_t *) &ipAddr->v6Addr.dstAddr[0]));
        hash_value = SISD_CRC_HASH64(hash_value,
                *((uint64_t *) &ipAddr->v6Addr.dstAddr[2]));
    }

    hash_value = SISD_CRC_HASH64(hash_value,
            (((uint64_t)srcPort <<32)|
             ((uint64_t)dstPort <<16)|
             ((uint64_t)nextHdr<<8)|
             dscp));

    *hash_key = hash_value & 0xFFFFFFFF;
    *first_key = hash_value & ADK_SYS_ULI_FLOW_CACHE_MASK;
    *second_key = (hash_value >> ADK_SYS_ULI_FLOW_CACHE_WIDTH) &
                   (ADK_SYS_ULI_FLOW_CACHE_MASK);

    return;
}


inline adk_dp_status_t
adk_sys_dp_uli_flow_cache_lookup(adk_ip_hdr_t *ipHdr,
                                 adk_ip_hdr_type_e hdrType,
                                 uint8_t nextHdr,
                                 uint16_t srcPort, 
                                 uint16_t dstPort, 
                                 adk_sys_dp_uli_flow_cache_entry_t **flow_cache_entry,
                                 uint32_t *hash_key)
{
    uint8_t coreIndex = ADK_LCORE_INDEX;
    uint16_t i, first_key, second_key, index;
    adk_sys_dp_flow_cache_hash_entry_t     *uli_flow_cache_hash_entry;
    adk_sys_dp_uli_flow_cache_entry_t      *uli_flow_cache_entry;
    adk_sys_dp_uli_flow_cache_table_t      *uli_flow_cache_table;
    adk_sys_ip_addr_t                      *ipAddr_p = NULL;
    uint8_t                                dscp = 0;

    uli_flow_cache_table = &adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex];

    if (ADK_IPV4 == hdrType) {
        ipAddr_p = (adk_sys_ip_addr_t *)&ipHdr->v4Hdr.src_addr;
        dscp = (ipHdr->v4Hdr.type_of_service>>2) & 0x3F;
    } else {
        ipAddr_p = (adk_sys_ip_addr_t *)&ipHdr->v6Hdr.src_addr;
        dscp = (adk_ntohl(ipHdr->v6Hdr.vtc_flow) >> 22) & 0x3F;
    }
    /*Point to the IP header source IP address*/
    adk_sys_dp_flow_cache_key_generate(ipAddr_p, hdrType, nextHdr, dscp, srcPort, dstPort, &first_key, &second_key, hash_key);


    /* hash lookup */
    first_key = *hash_key & (ADK_SYS_ULI_FLOW_CACHE_MASK);
    first_key = first_key*ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES;

    uli_flow_cache_hash_entry = &uli_flow_cache_table->hash_tbl[(first_key)];
    for(i=0; i<ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES; i++, uli_flow_cache_hash_entry++)
    {
        if((uli_flow_cache_hash_entry->isActive) &&
           (uli_flow_cache_hash_entry->hashValue == *hash_key))
        {
            /*compare flow cache tuples */
            index = uli_flow_cache_hash_entry->index;
            uli_flow_cache_entry = &uli_flow_cache_table->tbl[index];
            if ( 
                ( /*IPv4 check*/
                (ADK_IPV4 == hdrType) &&
                ((ipHdr->v4Hdr.src_addr ==
                        (uli_flow_cache_entry)->ipAddr.v4Addr.srcAddr) &&
                 (ipHdr->v4Hdr.dst_addr ==
                     (uli_flow_cache_entry)->ipAddr.v4Addr.dstAddr) &&
                    (srcPort == (uli_flow_cache_entry)->srcPort) &&
                    (dstPort == (uli_flow_cache_entry)->dstPort) &&
                    (nextHdr == (uli_flow_cache_entry)->nextHdr))
                ) ||
                ( /*IPv6 check*/
                     ((ADK_IPV6 == hdrType) && !memcmp(&ipHdr->v6Hdr.src_addr, &((uli_flow_cache_entry)->ipAddr.v6Addr.srcAddr), 16) &&
                      !memcmp(&ipHdr->v6Hdr.dst_addr, &((uli_flow_cache_entry)->ipAddr.v6Addr.dstAddr), 16) &&
                      (srcPort == (uli_flow_cache_entry)->srcPort) &&
                      (dstPort == (uli_flow_cache_entry)->dstPort) &&
                      (nextHdr == (uli_flow_cache_entry)->nextHdr)))
                ) 
            {
                if(likely((uli_flow_cache_entry)->version ==
                        adkSifDpShm_p->fc_ver))
                {
                    if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                        ++adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.hit;
                        ++adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[index].hit;
                    }
                    /*Update the state of the flow cache*/
                    (uli_flow_cache_entry)->state = ADK_SYS_DP_FLOW_CACHE_ACTIVE;
                    /* set aging flag */
                    adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.aging_bitmap, index);

                    *flow_cache_entry = uli_flow_cache_entry;
                    return ADK_E_FLOW_CACHE_PKT_FAST_PATH;
                }
                else
                {
                    /*Update the state of the flow cache*/
                    (uli_flow_cache_entry)->state = ADK_SYS_DP_FLOW_CACHE_NEW;
                    /*Update the global flow cache version
                      TODO : check is rcu_sync is needed here ?*/
                    (uli_flow_cache_entry)->version = adkSifDpShm_p->fc_ver;
                    /* set ageing flag as this entry will be reused for slow path */
                    adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.aging_bitmap, index);
                    if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                        ++adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.miss;
                        /*Initialize the counters*/
                        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[index].hit = 0;
                        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[index].alloc = 1;
                    }

                    *flow_cache_entry = uli_flow_cache_entry;
                    return ADK_E_FLOW_CACHE_PKT_SLOW_PATH;
                }
            }
        }
    }

    second_key = (*hash_key >> ADK_SYS_ULI_FLOW_CACHE_WIDTH) &
                   (ADK_SYS_ULI_FLOW_CACHE_MASK);
    second_key = second_key*ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES;

    uli_flow_cache_hash_entry = &uli_flow_cache_table->hash_tbl[(second_key)];
    for(i=0; i<ADK_SYS_NUM_FLOW_CACHE_COL_ENTRIES; i++, uli_flow_cache_hash_entry++)
    {
        if((uli_flow_cache_hash_entry->isActive) &&
           (uli_flow_cache_hash_entry->hashValue == *hash_key))
        {
            /*compare flow cache tuples */
            index = uli_flow_cache_hash_entry->index;
            uli_flow_cache_entry = &uli_flow_cache_table->tbl[index];

            if ( 
                ( /*IPv4 check*/
                (ADK_IPV4 == hdrType) &&
                ((ipHdr->v4Hdr.src_addr ==
                        (uli_flow_cache_entry)->ipAddr.v4Addr.srcAddr) &&
                 (ipHdr->v4Hdr.dst_addr ==
                     (uli_flow_cache_entry)->ipAddr.v4Addr.dstAddr) &&
                    (srcPort == (uli_flow_cache_entry)->srcPort) &&
                    (dstPort == (uli_flow_cache_entry)->dstPort) &&
                    (nextHdr == (uli_flow_cache_entry)->nextHdr))
                ) ||
                ( /*IPv6 check*/
                  ((ADK_IPV6 == hdrType) && !memcmp(&ipHdr->v6Hdr.src_addr, &((uli_flow_cache_entry)->ipAddr.v6Addr.srcAddr), 16) &&
                   !memcmp(&ipHdr->v6Hdr.dst_addr, &((uli_flow_cache_entry)->ipAddr.v6Addr.dstAddr), 16) && 
                   (srcPort == (uli_flow_cache_entry)->srcPort) &&
                   (dstPort == (uli_flow_cache_entry)->dstPort) &&
                   (nextHdr == (uli_flow_cache_entry)->nextHdr)))
               )
            {
                if(likely((uli_flow_cache_entry)->version ==
                        adkSifDpShm_p->fc_ver))
                {
                    if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                        ++adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.hit;
                        ++adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[index].hit;
                    }
                    /*Update the state of the flow cache*/
                    (uli_flow_cache_entry)->state = ADK_SYS_DP_FLOW_CACHE_ACTIVE;
                    /* set aging flag */
                    adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.aging_bitmap, index);

                    *flow_cache_entry = uli_flow_cache_entry;
                    return ADK_E_FLOW_CACHE_PKT_FAST_PATH;
                }
                else
                {
                    /*Update the state of the flow cache*/
                    (uli_flow_cache_entry)->state = ADK_SYS_DP_FLOW_CACHE_NEW;
                    /*Update the global flow cache version
                      TODO : check is rcu_sync is needed here ?*/
                    (uli_flow_cache_entry)->version = adkSifDpShm_p->fc_ver;
                    /* set ageing flag as this entry will be reused for slow path */
                    adk_sys_dp_fc_bitmap_set(uli_flow_cache_table->age_info.aging_bitmap, index);
                    if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                        ++adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.miss;
                        /*Initialize the counters*/
                        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[index].hit = 0;
                        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[index].alloc = 1;
                    }

                    *flow_cache_entry = uli_flow_cache_entry;
                    return ADK_E_FLOW_CACHE_PKT_SLOW_PATH;
                }
            }
        }
    }

    *flow_cache_entry = NULL;
    return ADK_DP_E_ERROR;
}

inline adk_dp_status_t
adk_sys_dp_uli_flow_cache_get(adk_ip_hdr_t *ipHdr,
                                 adk_ip_hdr_type_e hdrType,
                                 uint8_t nextHdr,
                                 uint16_t srcPort, 
                                 uint16_t dstPort, 
                                 adk_sys_dp_uli_flow_cache_entry_t **flow_cache_entry,
                                 uint32_t *hash_key)
{
    ADK_DP_MEASURE_BLOCK_CYCLE_SYS_FC_ULI_RX_TS();
    adk_dp_status_t ret = ADK_SUCCESS;
    adk_sys_dp_flow_cache_hash_entry_t *fc_hash_entry = NULL;
    uint8_t dscp = 0;
    ret = adk_sys_dp_uli_flow_cache_lookup((adk_ip_hdr_t *)ipHdr,
            hdrType,
            nextHdr,
            srcPort,
            dstPort,
            flow_cache_entry,
            hash_key);

    if (unlikely(ADK_DP_E_ERROR == ret)) {
        /*If not created yet, allocate it. If outdated then update it*/
        if (NULL == *flow_cache_entry) {
            ret = adk_sys_dp_uli_flow_cache_entry_alloc(&fc_hash_entry, flow_cache_entry, *hash_key);
            if ( ADK_SUCCESS == ret) {
                if (ADK_IPV4 == hdrType) {
                    ADK_MEMCPY(&((*flow_cache_entry)->ipAddr), &ipHdr->v4Hdr.src_addr, (sizeof(ipHdr->v4Hdr.src_addr) + sizeof(ipHdr->v4Hdr.dst_addr))); 
                    dscp = (ipHdr->v4Hdr.type_of_service>>2) & 0x3F;
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "src address 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v4Addr.srcAddr)); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dst address 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v4Addr.dstAddr)); 
                    (*flow_cache_entry)->isInnerIPv6 = false;
                } else {
                    ADK_MEMCPY(&((*flow_cache_entry)->ipAddr), &ipHdr->v6Hdr.src_addr, 
                               (sizeof(ipHdr->v6Hdr.src_addr) + sizeof(ipHdr->v6Hdr.dst_addr))); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "src address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.srcAddr[0])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "src address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.srcAddr[1])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "src address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.srcAddr[2])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "src address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.srcAddr[3])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dst address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.dstAddr[0])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dst address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.dstAddr[1])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dst address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.dstAddr[2])); 
                    ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dst address [0] 0x%x", adk_ntohl((*flow_cache_entry)->ipAddr.v6Addr.dstAddr[3])); 
                    dscp = (adk_ntohl(ipHdr->v6Hdr.vtc_flow) >> 22) & 0x3F;
                    (*flow_cache_entry)->isInnerIPv6 = true;
                }
                (*flow_cache_entry)->srcPort = srcPort;
                (*flow_cache_entry)->dstPort = dstPort;
                (*flow_cache_entry)->nextHdr = nextHdr;
                (*flow_cache_entry)->dscp = dscp;
                ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dscp %u", (*flow_cache_entry)->dscp); 
                ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "srcPort %u", (*flow_cache_entry)->srcPort); 
                ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "dstPort %u", (*flow_cache_entry)->dstPort); 
                ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "nextHdr %u", (*flow_cache_entry)->nextHdr); 
                return ADK_E_FLOW_CACHE_PKT_SLOW_PATH;
            }
        }
    }
    ADK_DP_MEASURE_BLOCK_CYCLE_SYS_FC_ULI_TX_TS();
    /*Flow found */
    return ret;
}

void
adk_sys_dp_flow_cache_monitor(void)
{
    uint8_t coreIndex = ADK_LCORE_INDEX;    
    uint16_t i,j, fc_index, prev_fc_idx, max_iter;
    adk_sys_dp_uli_flow_cache_aging_info_t *fc_age_flags;
    adk_sys_dp_uli_flow_cache_entry_t  *fc_entry;

    fc_age_flags = &adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].age_info;
    fc_age_flags->time_in_msec++;
    if((fc_age_flags->time_in_msec >= ADK_SYS_FLOW_CACHE_AGING_TIMER) || 
           (!adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].free_list.count && 
                (!(fc_age_flags->time_in_msec & 0xFF))))
    {
        fc_age_flags->time_in_msec = 0;
        /* check threshhold for aging,
         * if occupancy is greater than 75% then check for clean up. */
        if(adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].free_list.count <=
                (ADK_SYS_ULI_FLOW_CACHE_AGING_THRESHOLD))
        {
            /* Flow cache monitor */
            prev_fc_idx = fc_age_flags->prev_fc_idx;
            max_iter = prev_fc_idx + (ADK_SYS_ULI_FLOW_CACHE_AGING_SEG_SIZE);
            for(i=prev_fc_idx; i<max_iter; i++)
            {
                /* scan active flows */
                if(fc_age_flags->active_flows[i])
                {
                    /* TODO : optimize */
                    for(j = 0; j<64; j++)
                    {
                        if(fc_age_flags->active_flows[i] & (1ll<<j))
                        {
                            /* check for aging_bitmap */
                            if((fc_age_flags->aging_bitmap[i] & (1ll<<j)))
                            {
                                fc_index = (i<<6)+j;
                                /* release the flow cache entry, it is aged out */
                                fc_entry = &adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].tbl[fc_index];
                                /*2 step process :
                                   Step1: Mark the state of FC entry as STALE for the first round and exit.
                                   Step2: if packets are not flowing through it then state remains stale
                                          in the second round and those flows are eligible for release*/
                                if (fc_entry->state != ADK_SYS_DP_FLOW_CACHE_STALE) {
                                    fc_entry->state = ADK_SYS_DP_FLOW_CACHE_STALE; 
                                    adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].stats.flows_stale++;
                                }else{
                                    if (ADK_SYS_FLOW_CACHE_COUNTERS_ENABLED) {
                                        adkSifDpShm_p->sys_dp_uli_flow_cache_tbl[coreIndex].fc_entry_stats[fc_index].hit = 0;
                                    }
                                    adk_sys_dp_uli_flow_cache_entry_release(fc_entry);
                                }
                            }
                        }
                    }
                }
            }

            fc_age_flags->prev_fc_idx = i;
            if(i >= (ADK_SYS_ULI_FLOW_CACHE_ENTRIES>>6))
                fc_age_flags->prev_fc_idx = 0;
        }
    }
    return;
}

void
adk_sys_dp_flow_cache_entry_dump(adk_sys_dp_uli_flow_cache_entry_t *flow_cache_entry)
{
    if (flow_cache_entry) {
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "idx %u", flow_cache_entry->idx); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "nextIdx %u", flow_cache_entry->nextIdx); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "isActive %u", flow_cache_entry->isActive); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "state %u", flow_cache_entry->state); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "polIndex %u", flow_cache_entry->polIndex); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "pathMtu %u", flow_cache_entry->pathMtu); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "SAId %u", flow_cache_entry->SAId); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "polAction %u", flow_cache_entry->polAction); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "eSAstate %u", flow_cache_entry->eSAstate); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "preFragEnabled %u", flow_cache_entry->preFragEnabled); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "tunnelMode %u", flow_cache_entry->tunnelMode); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "espProto %u", flow_cache_entry->espProto); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "useSingleSAInPrefrag %u", flow_cache_entry->useSingleSAInPrefrag); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "nonInitialESAstate %u", flow_cache_entry->nonInitialESAstate); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "configDF %u", flow_cache_entry->configDF); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "nonInitialpolAction %u", flow_cache_entry->nonInitialpolAction); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "nonIntialPolIndex %u", flow_cache_entry->nonIntialPolIndex); 
        ADK_DP_LOG_INFO_1D(ADK_SYS_MODULE_ID, "nonInitialSAId %u", flow_cache_entry->nonInitialSAId); 
    }
    return;
}

/*!    end addtogroup _sifDpFunctions
 *     @}
 * end addtogroup _sif
 * @}
 */
