/* **************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
/*!
 * @internal
 * @file  adk_sif_dp_log_lib.h
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _sif  SIF Module
 *
 * @{
 *
 * @brief  Logging primatives for DP
 *
 * This header file contains all macros and supporting declarations that
 * are required to support the logging mechanism in the DP.
 *
 ************************************************************************/
#ifndef __ADK_SIF_DP_LOG_LIB_H__
#define __ADK_SIF_DP_LOG_LIB_H__

#include <sys/time.h>
#include "adk_modules_enabled.h"
#include "adk_sif.h"
#include "adk_sif_dp.h"

#define STRINGIFY(x) #x
#define TO_STRING(y) STRINGIFY(y)
#define FILE_LINE __FILE__":"TO_STRING(__LINE__)

#if defined(ADK_DEBUG)
#define error_print(fmt, args...) printf(fmt, ##args);

#define INTERNAL_PRINT_MACRO(fmt, args...) \
    do \
    { \
        printf("%s (%s:%d) ", __func__, __FILE__, __LINE__); \
        printf(fmt, ##args); \
        printf("\n"); \
    } while(0)
#else
#define INTERNAL_PRINT_MACRO(fmt, args...)
#define error_print(fmt, args...)
#endif

#if defined(ADK_DEBUG)
#define debug_print(fmt, args...) printf(fmt, ##args);

#define DEBUG_PRINT_MACRO(fmt, args...) \
    do \
    { \
        printf("%s (%s:%d) ", __func__, __FILE__, __LINE__); \
        printf(fmt, ##args); \
        printf("\n"); \
    } while(0)

#else
#define DEBUG_PRINT_MACRO(fmt, args...)
#define debug_print(fmt, args...)
#endif

/*
 * Define a macro to put around all debug mode LTTng calls
 */
#if 0
####if defined(ADK_MODULE_SIF_LTTNG_DP_ENABLED) && defined(ADK_DEBUG)
####define ADK_LTTNG_DP_DEBUG_MODE(...) \
###    do { \
###    __VA_ARGS__; \
###    asdfasdf \
###    } while(0)
####else
####define ADK_LTTNG_DP_DEBUG_MODE(...)
####endif /* ifdef ADK_MODULE_SIF_LTTNG_DP_ENABLED */
#endif


/*
 * Local macro used to guard band the ADK_DP_ macros when LTTng is disabled
 * for DP
 */
#ifdef ADK_MODULE_SIF_LTTNG_DP_ENABLED
#include "adk_lttng_dp_trace.h"
#define ADK_DP_LTTNG_OUTPUT(...) {__VA_ARGS__}
#  ifdef ADK_DEBUG
#define ADK_DP_LTTNG_OUTPUT_DEBUG(...) {__VA_ARGS__}
#  else
#define ADK_DP_LTTNG_OUTPUT_DEBUG(...)
#  endif
#else
#define ADK_DP_LTTNG_OUTPUT(...)
#define ADK_DP_LTTNG_OUTPUT_DEBUG(...)
#endif


/*!
 * ADK_DP_LOG_INFO is a generic macro to log LTTng generic messages.
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 */
#define ADK_DP_LOG_INFO(__mod_id, __message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_INFO_1D is a generic macro to log LTTng message with an integer
 * parameter.
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1 - Integer value to be logged
 *
 */
#define ADK_DP_LOG_INFO_1D(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1D_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_INFO_1P is a generic macro to log LTTng message with a pointer
 * value parameter.
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1 - pointer address to be logged
 *
 */
#define ADK_DP_LOG_INFO_1P(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1P_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)

#define ADK_DP_LOG_INFO_2D(__mod_id, __message, __data1, __data2) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_2D_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2)); \
                           ) \
    } while(0)

#define ADK_DP_LOG_INFO_ERRNO(__mod_id, __message, __errStr, __my_errno) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__errStr), (__my_errno)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_ERRNO_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__errStr), (__my_errno)); \
                           ) \
    } while(0)

#define ADK_DP_LOG_INFO_1S(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1S_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)


/*!
 * ADK_DP_LOG_INFO_1X is a generic macro to log LTTng message with an integer
 * parameter, displayed as hex.
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1 - Integer value to be logged and displayed as hex
 *
 */
#define ADK_DP_LOG_INFO_1X(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1X_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)


/*!
 * ADK_DP_FORCE_LOG_INFO same as ADK_DP_LOG_INFO except that it doesn't
 * lookup the module info.
 *
 * __message - character strings and appropriate parameters.
 *
 */
#define ADK_DP_FORCE_LOG_INFO(__message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_sif_force_info(0,  \
                                                               FILE_LINE, \
                                                               (__message)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_ERROR is a generic LTTng tracepoint to log an error message
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 */
#define ADK_DP_LOG_ERROR(__mod_id, __message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_ERROR_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_ERROR_1D is a generic LTTng tracepoint to log an error message
 * and an integer value
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1 - Integer value to be logged
 *
 */
#define ADK_DP_LOG_ERROR_1D(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_ERROR_1D_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_ERROR_2D is a generic LTTng tracepoint to log an error message
 * and two integer values
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1, __data2 - Integer values to be logged
 *
 */
#define ADK_DP_LOG_ERROR_2D(__mod_id, __message, __data1, __data2) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_ERROR_2D_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_ERROR_1P is a generic LTTng tracepoint to log an error message
 * and a pointer value
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1 - pointer address to be logged
 *
 */
#define ADK_DP_LOG_ERROR_1P(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_ERROR_1P_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_ERROR_ERRNO will display a user message, a string of the error,
 * as well as the error number. The caller needs to call the appropriate
 * function to get the string representation of the error, i.e. calling
 * adk_sys_error_string() or rte_strerror().
 *
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - User message
 *
 * __errStr - the string representation of the error number
 *
 * __my_errno - the error number
 *
 */
#define ADK_DP_LOG_ERROR_ERRNO(__mod_id, __message, __errStr, __my_errno) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__errStr), (__my_errno)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_ERROR_ERRNO_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__errStr), (__my_errno)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_ERROR_1S is a generic LTTng tracepoint to log an error message
 * and a pointer value
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 * __data1 - string to be logged
 *
 */
#define ADK_DP_LOG_ERROR_1S(__mod_id, __message, __data1) \
    do { \
        INTERNAL_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_ERROR_1S_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_FORCE_LOG_ERROR same as ADK_DP_LOG_ERROR except that it doesn't
 * lookup the module info.
 *
 * __message - user message
 *
 */
#define ADK_DP_FORCE_LOG_ERROR(__message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_sif_force_error(0,  \
                                                                FILE_LINE, \
                                                                (__message)); \
                           ) \
    } while(0)


/*!
 * ADK_DP_LOG_FATAL_ERROR is a generic LTTng tracepoint to log a fatal error
 * message. This macro is almost same as ADK_DP_LOG_ERROR. The only difference
 * is the severity level of the error that is logged is higher.
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 */
#define ADK_DP_LOG_FATAL_ERROR(__mod_id, __message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_FATAL_ERROR_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, \
                                      (__message)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_FORCE_LOG_FATAL_ERROR same as ADK_DP_LOG_FATAL_ERROR except that
 * it doesn't lookup the module info.
 *
 * __message - user message
 *
 */
#define ADK_DP_FORCE_LOG_FATAL_ERROR(__message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_sif_force_fatal(0,  \
                                                                FILE_LINE, \
                                                                (__message)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_LOG_WARNING is a generic LTTng tracepoint to log a warning message
 *
 * __mod_id - module handle. It is used to match the severity level
 *            with the logging level configured for the module.
 *
 * __message - user message
 *
 */
#define ADK_DP_LOG_WARNING(__mod_id, __message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_WARNING_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, \
                                      (__message)); \
                           ) \
    } while(0)

/*!
 * ADK_DP_FORCE_LOG_WARNING same as ADK_DP_LOG_WARNING except that
 * it doesn't lookup the module info, thus it is unconditional logging. This
 * should be used only when the module ID is not available.
 *
 * __message - user message
 *
 */
#define ADK_DP_FORCE_LOG_WARNING(__message) \
    do { \
        INTERNAL_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_sif_force_warning(0, \
                                                                   \
                                                                  FILE_LINE, \
                                                                  (__message)); \
                           ) \
    } while(0)

/*!
 * Miscellaneous loggings
 */

#define ADK_DP_LOG_FUNC_ENTRY(__mod_id) \
    do { \
        INTERNAL_PRINT_MACRO("%s Entry", __func__); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_FUNC_ENTRY_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, __func__); \
                           ) \
    } while(0);

#define ADK_DP_LOG_FUNC_EXIT(__mod_id) \
    do { \
        INTERNAL_PRINT_MACRO("%s Exit", __func__); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_FUNC_EXIT_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, __func__); \
                           ) \
    } while(0);

#define ADK_DP_FORCE_LOG_FUNC_ENTRY() \
    do { \
        INTERNAL_PRINT_MACRO("%s Entry", __func__); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_force_sif_func_entry(0, \
                                                                FILE_LINE, \
                                                                __func__); \
                           ) \
    } while(0);

#define ADK_DP_FORCE_LOG_FUNC_EXIT() \
    do { \
        INTERNAL_PRINT_MACRO("%s Exit", __func__); \
        ADK_DP_LTTNG_OUTPUT( \
                             adk_lttng_dp_trace_force_sif_func_exit(0, \
                                                                FILE_LINE, \
                                                                __func__); \
                           ) \
    } while(0);


#define ADK_DP_LOG_FUNC_ENTRY_DEBUG(__mod_id) \
    do { \
        DEBUG_PRINT_MACRO("%s Entry", __func__); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                            TRACE_DP_FUNC_ENTRY_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, __func__); \
                           ) \
    } while(0);

#define ADK_DP_LOG_FUNC_EXIT_DEBUG(__mod_id) \
    do { \
        DEBUG_PRINT_MACRO("%s Exit", __func__); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                            TRACE_DP_FUNC_EXIT_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, __func__); \
                           ) \
    } while(0);

#define ADK_DP_LOG_HW_DEBUG(__mod_id, __message) \
    do { \
        DEBUG_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_HW_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message)); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_DEBUG(__mod_id, __message) \
    do { \
        DEBUG_PRINT_MACRO((__message)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message)); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_1D_DEBUG(__mod_id, __message, \
                                 __data1) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_2D_DEBUG(__mod_id, __message, \
                                 __data1, __data2) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_2D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_3D_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2), (__data3)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_3D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_4D_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3, __data4) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), (__data4)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_4D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_5D_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3, __data4, __data5) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), \
                          (__data4), (__data5)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_5D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4), (__data5) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_1X_DEBUG(__mod_id, __message, \
                                 __data1) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_2X_DEBUG(__mod_id, __message, \
                                 __data1, __data2) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_2X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_3X_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2), (__data3)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_3X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_4X_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3, __data4) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), (__data4)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_4X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_5X_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3, __data4, __data5) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), \
                          (__data4), (__data5)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_5X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4), (__data5) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_6X_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3, __data4, \
                                 __data5, __data6) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), \
                          (__data4), (__data5), (__data6)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_5X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4), (__data5), (__data6) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_1P_DEBUG(__mod_id, __message, \
                                 __data1) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1P_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_2P_DEBUG(__mod_id, __message, \
                                 __data1, __data2) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_2P_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_ERRNO_DEBUG(__mod_id, __message, __errStr, __my_errno) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__errStr), (__my_errno)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_ERRNO_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__errStr), (__my_errno)); \
                           ) \
    } while(0)

#define ADK_DP_LOG_INFO_1S_DEBUG(__mod_id, __message, __data1) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1S_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1)); \
                           ) \
    } while(0)

#define ADK_DP_LOG_INFO_1D_1X_DEBUG(__mod_id, __message, \
                                 __data1, __data2) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1D_1X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_1P_4D_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3, __data4, __data5) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), \
                          (__data4), (__data5)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1P_4D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4), (__data5) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_1P_1X_DEBUG(__mod_id, __message, \
                                 __data1, __data2) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1P_1X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_1P_2X_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2), (__data3)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_1P_2X_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_2P_1D_DEBUG(__mod_id, __message, \
                                 __data1, __data2, __data3) \
    do { \
        DEBUG_PRINT_MACRO((__message), (__data1), (__data2), (__data3)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_2P_1D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3) \
                                      ); \
                           ) \
    } while(0);

#define ADK_DP_LOG_INFO_2P_2D_DEBUG(__mod_id, __message, \
                                    __data1, __data2, __data3, __data4) \
    do { \
        DEBUG_PRINT_MACRO((__message), \
                          (__data1), (__data2), (__data3), (__data4)); \
        ADK_DP_LTTNG_OUTPUT_DEBUG( \
                             adk_lttng_dp_trace_fn_ptr __fn_ptr; \
                             ADK_LTTNG_DP_PTR_GET((__mod_id), \
                                                  TRACE_DP_INFO_2P_2D_DEBUG_INDEX, \
                                                  &(__fn_ptr)); \
                             __fn_ptr((__mod_id), FILE_LINE, (__message), \
                                      (__data1), (__data2), (__data3), \
                                      (__data4) \
                                      ); \
                           ) \
    } while(0);


#endif /* __ADK_SIF_DP_LOG_LIB_H__ */

/*! end addtogroup _sif
 * @}
 */
