/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#ifndef __ADK_PMU_H__
#define __ADK_PMU_H__

#define NUM_MAX_PKTS	30000000
#define NUM_MAX_ITRNS	(NUM_MAX_PKTS)
#include <stdint.h>
#include <stdbool.h>
#include <adk_sif_constants.h>
#include <adk_os.h>

#define ADK_MEASURE_CYCLE_FRAG_COUNT_PER_PACKET 2

#ifdef PIPELINE_STAGE_PKT_COUNT_ENABLED
/* Structure to hold the packet count in a pipeline stage.
 * By reading the counters across all the stages in a
 * pipeline, it is possible to know which stage is the
 * bottleneck.
 * It can also store the sent count as well. Each CPU
 * will have its own structure occupying a complete
 * cache line. As with any debugging tool, this may have
 * impact on performance.
 */
typedef struct pipeline_stage_pkt_count_s
{
    /* Packets received from the NIC ports */
    uint32_t portRecv;
    /* Packets received from the DPDK rings */
    uint32_t ringRecv;
    /* Responses received from QAT device */
    uint32_t qatRecv;
    /* Packets received from the PH ring */
    uint32_t phRecv;
} __attribute__((__aligned__(RTE_CACHE_LINE_SIZE))) pipeline_stage_pkt_count_t;

/* Maximum number of CPUs supported is 8. Since the
 * counters are stored in an array, it is the responsibility
 * of the user to provide the correct index into the array.
 */
#define PIPELINE_STAGE_PKT_COUNT_MAX_CPU 8

extern pipeline_stage_pkt_count_t
        pipelineStagePktCount[PIPELINE_STAGE_PKT_COUNT_MAX_CPU];

/* CPU ID is used as an index into the array. It is the responsibility
 * of the user to map the actual CPU ID to array index.
 */
#define PIPELINE_STAGE_PORT_PACKET_COUNT_ADD(__cpuId, __numTask) \
             pipelineStagePktCount[(__cpuId)].portRecv += (__numTask);

#define PIPELINE_STAGE_RING_PACKET_COUNT_ADD(__cpuId, __numTask) \
             pipelineStagePktCount[(__cpuId)].ringRecv += (__numTask);

#define PIPELINE_STAGE_QAT_PACKET_COUNT_ADD(__cpuId, __numTask) \
             pipelineStagePktCount[(__cpuId)].qatRecv += (__numTask);

#define PIPELINE_STAGE_PH_PACKET_COUNT_ADD(__cpuId, __numTask) \
             pipelineStagePktCount[(__cpuId)].phRecv += (__numTask);
#else
#define PIPELINE_STAGE_PORT_PACKET_COUNT_ADD(__cpuId, __numTask)
#define PIPELINE_STAGE_RING_PACKET_COUNT_ADD(__cpuId, __numTask)
#define PIPELINE_STAGE_QAT_PACKET_COUNT_ADD(__cpuId, __numTask)
#define PIPELINE_STAGE_PH_PACKET_COUNT_ADD(__cpuId, __numTask)
#endif /* PIPELINE_STAGE_PKT_COUNT_ENABLED */

/* enumeration for hw event ids available for cortex-a7 */
typedef enum {
    SW_INCR 			= 0x00,
    L1I_CACHE_REFILL		= 0x01,	
    L1I_TLB_REFILL		= 0x02,		
    L1D_CACHE_REFILL		= 0x03,
    L1D_CACHE			= 0x04,
    L1D_TLB_REFILL		= 0x05,
    LD_RETIRED			= 0x06,
    ST_RETIRED			= 0x07,
    INST_RETIRED		= 0x08,
    EXC_TAKEN			= 0x09,
    EXC_RETURN			= 0x0A,
    CID_WRITE_RETIRED		= 0x0B,
    PC_WRITE_RETIRED		= 0x0C,
    BR_IMMED_RETIRED		= 0x0D,
    BR_RETURN_RETIRED		= 0x0E,
    UNALIGNED_LDST_RETIRED	= 0x0F,
    BR_MIS_PRED			= 0x10,
    CPU_CYCLES			= 0x11,
    BR_PRED			= 0x12,
    MEM_ACCESS			= 0x13,
    L1I_CACHE			= 0x14,
    L1D_CACHE_WB		= 0x15,
    L2D_CACHE			= 0x16,
    L2D_CACHE_REFILL		= 0x17,
    L2D_CACHE_WB		= 0x18,
    BUS_ACCESS			= 0x19,
    BUS_CYCLES			= 0x1D,
    BUS_ACCESS_LD		= 0x60,
    BUS_ACCESS_ST		= 0x61,
    EXC_IRQ			= 0x86,
    EXC_FIQ			= 0x87,
    EXT_MEM_REQ			= 0xC0,
    EXT_MEM_REQ_NC		= 0xC1,
    PREFETCH_LINEFILL		= 0xC2,
    PREFETCH_LINEFILL_DROP	= 0xC3,
    READ_ALLOCATE_ENTER		= 0xC4,
    READ_ALLOC			= 0xC5,
    STALL_SB_FULL		= 0xC9,
    LOCAL_CLUSTER_SNOOP		= 0xCA
} adk_pmu_event_type_t;

/* enumeration for hw performance counter ids */
typedef enum {
    PMU_CNTR_ID_0 = 0x0,
    PMU_CNTR_ID_1 = 0x1,
    PMU_CNTR_ID_2 = 0x2,
    PMU_CNTR_ID_3 = 0x3
} adk_pmu_cntr_id_t;

/* enumeration for module ids - used to access module specific static variables */
typedef enum {
    PMU_ETHERNET_LLI_MODULE=0,
    PMU_ETHERNET_ULI_MODULE,
    PMU_IP_LLI_MODULE_PASS_1,
    PMU_IP_LLI_MODULE_PASS_2,
    PMU_IP_ULI_MODULE_PASS_1,
    PMU_IP_ULI_MODULE_PASS_2,
    PMU_IP_ULI_MODULE_EM_PATH_INIT,
    PMU_IPV6_LLI_PASS1_MODULE,
    PMU_IPV6_LLI_PASS2_MODULE,
    PMU_IPV6_ULI_PASS1_MODULE,
    PMU_IPV6_ULI_PASS2_MODULE,
    PMU_IPV6_ULI_EXTRA_MODULE,
    PMU_IPSEC_LLI_MODULE,
    PMU_IPSEC_ULI_MODULE,
    PMU_IPSEC_LLI_POST_CRYPTO_MODULE,
    PMU_IPSEC_ULI_POST_CRYPTO_MODULE,
    PMU_GTPU_LLI_MODULE,
    PMU_GTPU_ULI_MODULE,
    PMU_BFD_LLI_MODULE,
    PMU_BFD_ULI_MODULE,
    PMU_BFD_ATOMIC_LLI_MODULE,
    PMU_FPS_LLI_MODULE,
    PMU_FPS_ULI_MODULE,
    PMU_PAL_LLI_MODULE,
    PMU_PAL_ULI_MODULE,
    PMU_ADK_LLI_MODULE,
    PMU_ADK_ULI_MODULE,
    PMU_END_TO_END_LLI_MODULE,
    PMU_END_TO_END_ULI_MODULE,
    PMU_IPSEC_LLI_POST_CRYPTO_SINGLE_SA_MODULE,
    PMU_IPSEC_LLI_MODULE_UNEXPECTED,
    PMU_IPSEC_LLI_MODULE_POLICY_LKUP,
    PMU_IPSEC_LLI_MODULE_AFTER_POLICY_LKUP,
    PMU_IPSEC_LLI_MODULE_PROCESS_INBOUND,
    PMU_IPSEC_ULI_MODULE_VARIABLES_INIT,
    PMU_IPSEC_ULI_MODULE_PARSE_PKT,
    PMU_IPSEC_ULI_MODULE_SPD_LKUP,
    PMU_IPSEC_ULI_MODULE_ACTION_IPSEC,
    PMU_IPSEC_ULI_MODULE_PREPARE_PROTO,
    PMU_IPSEC_ULI_MODULE_PREPARE_CRYPTO,
    PMU_IPSEC_ULI_MODULE_AFTER_CRYPTO,
    PMU_SYS_FC_ULI_MODULE,
    PMU_SYS_ULI_EGRESS_STATS,
    PMU_MODULE_MAX
} adk_pmu_blk_cycle_cnt_module_id_t;

/* Structure to store module specific compute info */
typedef struct {
    uint64_t    diffTime;
    uint32_t    avgTime;
    uint32_t    totalCnt;
} adk_pmu_blk_cycle_cnt_info_t;


/* function to enable PMU counters */
void adk_enable_pmu(void);

/* function to reset cycle counter */
void adk_reset_cycle_cntr(void);

/* function to start cycle counter */
void adk_start_cycle_cntr(void);

/* function to stop cycle counter */
void adk_stop_cycle_cntr(void);

/* function to get cycle counter */
static inline uint32_t adk_get_cycle_cntr(void)
{
    /*return 0;*/
    uint32_t tmp, ts;
    __asm__ __volatile__("rdtsc":"=a" (ts) , "=d" (tmp));
    return ts;
}

/* function to reset event counters */
static inline void adk_reset_evt_cntr(void)
{
}

/* function to configure event counter */
void adk_config_evt_cntr(uint32_t cntr_id, uint32_t evt_id);

/* function to start event counter */
void adk_start_evt_cntr(uint32_t cntr_id);

/* function to stop event counter */
void adk_stop_evt_cntr(uint32_t cntr_id);

/* function to read event counter */
static inline 
uint32_t adk_get_evt_cntr(uint32_t __attribute__((unused)) cntr_id)
{
    return 0;
}

static inline void
ADK_INSTRUCTION_BARRIER(void)
{
    unsigned int regs[4];
    __asm__ __volatile__(
            "pushq %%rbx     \n\t" /* save %rbx */
            "cpuid            \n\t"
            "movl %%ebx ,%[ebx]  \n\t" /* write the result into output var */
            "popq %%rbx \n\t"
            : "=a"(regs[0]), [ebx] "=r"(regs[1]), "=c"(regs[2]), "=d"(regs[3])
            : "a"(index));
}

/* function to return event string from event id */
char * adk_get_evt_str(uint32_t eventid);

void adk_dp_measure_cycle_config(void);
void adk_dp_measure_cycle_rx_ts(void);
void adk_dp_measure_cycle_tx_ts(void);
uint32_t adk_dp_get_cycle_count(bool reset);

void adk_dp_measure_block_cycle_rx_ts(uint64_t* rx_ts);
void adk_dp_measure_block_cycle_tx_ts(uint64_t* rx_ts, uint64_t* tx_ts, const char *module,
        uint32_t module_id);

void adk_dp_measure_block_cycle_tx_inter_ts(uint64_t* rx_ts, uint64_t* tx_ts, uint32_t module_id);
void adk_pmu_pipeline_stage_pkt_count_print (void);

typedef struct _measure_cycle_params
{
   uint64_t mark;           /* mark/start processor cycle collection */
   uint64_t proc_cycles;    /* total cycle count */
   uint64_t iter;           /* total iteration/packets */
   uint32_t print_only_once;/* flag to enable or disable prints in 
                                   case of Measure Cycle.*/
   uint32_t resume_intel_itt;/* flag to enable vtune data collection. */
   uint32_t pause_intel_itt; /* flag to pause vtune data collection*/
   uint32_t stage_id;
} measure_cycle_t;


#define ADK_MEASURE_CYCLE_MAX_STAGE 4
#define ADK_TERMINATION_STAGE 0
#define ADK_ORIGINATION_STAGE 1
#define ADK_REASSEMBLY_STAGE  2
#define ADK_TIMER_STAGE       3
#define ADK_ROUTING_STAGE     1
extern measure_cycle_t adkMeasCycleTbl_g[ADK_MAX_NUM_CORES][ADK_MEASURE_CYCLE_MAX_STAGE];
extern measure_cycle_t *adkMeasCycle_g[ADK_MAX_NUM_CORES][ADK_MEASURE_CYCLE_MAX_STAGE];

static inline measure_cycle_t *ADK_MEASURE_CYCLE_INIT(void)
{
#if defined(ADK_MEASURE_CYCLE) || defined(INTEL_ITT)
    measure_cycle_t *mcd =
    (measure_cycle_t *) ADK_NVM_CALLOC(ADK_MISC_MODULE_ID, 0, 1, sizeof(measure_cycle_t));
    return mcd;
#else
    return NULL;
#endif
}

static inline
void ADK_MEASURE_CYCLE_START(measure_cycle_t __attribute__((unused))**mcd)
{
#ifdef ADK_MEASURE_CYCLE
    uint32_t cycles_lo;
    uint32_t cycles_hi;

    _mm_mfence();
    ADK_INSTRUCTION_BARRIER();
    __asm__ __volatile__("rdtsc" : "=a" (cycles_lo), "=d" (cycles_hi));
    (*mcd)->mark = (((uint64_t)cycles_hi << 32) | cycles_lo);
#endif
}

static inline
void ADK_MEASURE_CYCLE_PAL_START(measure_cycle_t __attribute__((unused))**mcd1,
                                 measure_cycle_t __attribute__((unused))**mcd2)
{
#ifdef ADK_MEASURE_CYCLE
    uint32_t cycles_lo;
    uint32_t cycles_hi;

    _mm_mfence();
    ADK_INSTRUCTION_BARRIER();
    __asm__ __volatile__("rdtsc" : "=a" (cycles_lo), "=d" (cycles_hi));
    (*mcd1)->mark = (((uint64_t)cycles_hi << 32) | cycles_lo);
    (*mcd2)->mark = (((uint64_t)cycles_hi << 32) | cycles_lo);
#endif
}

static inline
void ADK_MEASURE_CYCLE_DIFF(measure_cycle_t __attribute__((unused))**mcd,
                            uint32_t __attribute__((unused))tc)
{
#ifdef ADK_MEASURE_CYCLE
    uint32_t cycles_lo;
    uint32_t cycles_hi;

    __asm__ __volatile__("rdtscp" : "=a" (cycles_lo), "=d" (cycles_hi));
    ADK_INSTRUCTION_BARRIER();

    (*mcd)->proc_cycles = (*mcd)->proc_cycles + 
                      (((uint64_t)cycles_hi << 32) | cycles_lo) - (*mcd)->mark;
    (*mcd)->iter = (*mcd)->iter + tc;
#endif
}

static inline
void ADK_MEASURE_CYCLE_REPORT(measure_cycle_t __attribute__((unused))*mcd,
                              uint32_t __attribute__((unused))lcore_id,
                              uint32_t __attribute__((unused))stage_id)
{
#ifdef ADK_MEASURE_CYCLE
    if (unlikely(mcd->print_only_once == 0))
    {
        mcd->print_only_once = 1;
        printf("\nCPU id: %d", lcore_id);
        printf("\nSTAGE id: %d", stage_id);
        printf("\nCycles Count: %"PRIu64"\n", mcd->proc_cycles);
        printf("\nTotal Packets: %"PRIu64"\n", mcd->iter);
        printf("\nCycles Count per packet(Approx): %"PRIu64"\n\n",
                (mcd->proc_cycles/mcd->iter));
    }
#endif
}

static inline
void ADK_MEASURE_CYCLE_CHECK_EXIT(measure_cycle_t __attribute__((unused))*mcd,
                                  uint32_t __attribute__((unused))lcore_id,
                                  uint32_t __attribute__((unused))stage_id)
{
#ifdef ADK_MEASURE_CYCLE
    if (unlikely(mcd->iter >= NUM_MAX_PKTS))
            ADK_MEASURE_CYCLE_REPORT(mcd, lcore_id, stage_id);
#endif
}

static inline
void INTEL_ITT_RESUME(measure_cycle_t __attribute__((unused))**ittd)
{
#ifdef INTEL_ITT
    if (unlikely((*ittd)->resume_intel_itt == 0))
    {
        (*ittd)->resume_intel_itt = 1;
        _mm_mfence();
        ADK_INSTRUCTION_BARRIER();
        __itt_resume();
    }
#endif
}

static inline
void INTEL_ITT_PAUSE(measure_cycle_t __attribute__((unused))**ittd)
{
#ifdef INTEL_ITT
    if (unlikely(((*ittd)->iter >= NUM_MAX_PKTS) &&
                 ((*ittd)->pause_intel_itt == 0)))
    {
        (*ittd)->pause_intel_itt = 1;
        __itt_pause();
        ADK_INSTRUCTION_BARRIER();
    }
#endif
}

static inline
void INTEL_ITT_INCR_ITER(measure_cycle_t __attribute__((unused))**ittd,
                         uint32_t __attribute__((unused))tc )
{
#ifdef INTEL_ITT 
    (*ittd)->iter = (*ittd)->iter + tc;
#endif
}
#if 0
static inline
void ADK_MEASURE_CYCLE_INCR_ITER(measure_cycle_t __attribute__((unused))**mcd,
                                        uint32_t __attribute__((unused))tc );
#endif

#endif
