/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_sif_cfg_include.h"
#include "adk_cfg_pvt.h"

/* Section name, Valid digits, parse function */
static const adk_cfg_config_section_t cfg_file_scheme_g[] = {
        {"EAL",      0, adk_cfg_eal_parse},
        {"MEMPOOL",  1, adk_cfg_mempool_parse},
        {"PORT",     1, adk_cfg_port_parse},
        {"RING",     1, adk_cfg_ring_parse},
        {"PACKET_BUFFERS",     1, adk_cfg_pkt_buffs_parse},
        {"PIPELINE", 3, adk_cfg_pipeline_parse},
};


/*!
 * =============================================================
 * adk_cfg_init
 * @param[out] cfg_params  Handle to configuration info table.
 *
 * @retval ADK_SUCCESS Successfully initialized cfg module.
 * @retval ADK_E_MEM_UNAVAILABLE Dynamic memory allocation failed
 *
 * @brief  Initializes CFG module.
 *
 * This function initializes configuration file parsing module.
 * It allocates memory for cfg info tables.
 *
 * =============================================================
 */
adk_status_t
adk_cfg_init(adk_cfg_params_t **cfg_params)
{
    adk_status_t ret = ADK_SUCCESS;

    /* Create cfg info table which stores parsed cfg file sections 
     * information 
     */
    *cfg_params = calloc(1, sizeof(adk_cfg_params_t));
    if (*cfg_params == NULL)
    {
        ADK_PREINIT_WRITE("Memory allocation failure\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
    }

    adk_cfg_eal_init(*cfg_params);
    adk_cfg_mempool_init(*cfg_params);
    adk_cfg_port_init(*cfg_params);
    adk_cfg_ring_init(*cfg_params);
    adk_cfg_pipeline_init(*cfg_params);
    adk_cfg_pkt_buffs_init(*cfg_params);
    return ret;
}

/*!
 * =============================================================
 * adk_cfg_shutdown
 * @param[in] cfg_params  Handle to configuration info table.
 *
 * @retval ADK_SUCCESS Successfully shutdown cfg module.
 *
 * @brief  Shuts down CFG module.
 *
 * This function releases memory allocated in configuration 
 * file parsing module.
 *
 * =============================================================
 */
adk_status_t
adk_cfg_shutdown(adk_cfg_params_t *cfg_params)
{
    adk_status_t ret = ADK_SUCCESS;

    /* Delete cfg info table */
    free(cfg_params);
    
    return ret;
}

/*!
 * =============================================================
 * adk_cfg_parse 
 * @param[in] argc  handle. 
 *
 * @retval ADK_SUCCESS Successfully parsed the cfg file.
 * @retval ADK_E_CFG_PARSE_ERROR Config file parse failed.
 * @retval ADK_E_INVALID_PARAMETER Invalid configuration parameter.
 * @retval ADK_E_MEM_UNAVAILABLE Dynamic memory allocation failed
 *
 * @brief  Parses configuration file.
 *
 * This function parses ADK configuration file. Configuration files
 * contains multiple sections and each section will have multiple 
 * entries. Entries are parsed as (name,value) pair. Parsed 
 *
 * =============================================================
 */
adk_status_t
adk_cfg_parse(int argc, char **argv, adk_cfg_params_t *cfg_params)
{
    adk_status_t retValue = ADK_SUCCESS;
    char *config_filename = NULL;
    int32_t section_cnt = 0, i, j;
    char **section_names;
    struct rte_cfgfile *cfg_handle_p = NULL;
    const adk_cfg_config_section_t *section;
    int len, cfg_name_len;
    char resolved_filename[PATH_MAX] = "";
    FILE *fp1;
    char cmdstr[PATH_MAX+10];

    /* Validate cfg param pointer */
    if (cfg_params == NULL)
    {
        ADK_PREINIT_WRITE("Invalid cfg param pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER,
                                        ADK_SEV_LEVEL_ERROR);
    }

    /* Get the configuration file name */ 
    retValue = adk_cfg_get_filename(argc, argv, &config_filename);
    if (retValue != ADK_SUCCESS)
    {
        ADK_PREINIT_WRITE("Configuration file get failed\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR,
                                        ADK_SEV_LEVEL_ERROR);
    }

    /* Need to expand the filename as it might include environment variable(s)
     * which isn't supported by rte_cfgfile_load as DPDK assumes the shell
     * has expanded them
     */

    snprintf(cmdstr, PATH_MAX+10, "echo %s 2>/dev/null", config_filename);
    if ((fp1 = popen(cmdstr, "r")) == NULL || /* do echo cmd */
        fgets(resolved_filename, PATH_MAX, fp1) == NULL) /* Get echo results */
    {
        /* open/get pipe failed */
        if(fp1)
            pclose(fp1);
        perror("Unable to resolve filename ");
        ADK_PREINIT_WRITE("Unable to resolve filename %s\n", config_filename);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR,
                                        ADK_SEV_LEVEL_ERROR);
    }
    pclose(fp1);
    if ((ADK_STRLEN(resolved_filename)) &&
        (resolved_filename[ADK_STRLEN(resolved_filename) - 1] == 0xa))
    {
        resolved_filename[ADK_STRLEN(resolved_filename) - 1] = '\0'; /* remove newline */
    }

    /* Load the configuration file */ 
    cfg_handle_p = rte_cfgfile_load(resolved_filename, 0);
    if (cfg_handle_p == NULL)
    {
        ADK_PREINIT_WRITE("Unable to load config file %s\n", resolved_filename);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR,
                                        ADK_SEV_LEVEL_ERROR);
    }

    /* Start configuration file parsing */
    section_cnt = rte_cfgfile_num_sections(cfg_handle_p, NULL, 0);
    ADK_PREINIT_WRITE("\nNumber of sections = %d\n", section_cnt);

    if (section_cnt > ADK_SIF_CFG_MAX_SECTIONS)
    {
        ADK_PREINIT_WRITE("Number of sections configured (%d) exceeds "
                          "max supported sections (%d)\n", 
                          section_cnt, ADK_SIF_CFG_MAX_SECTIONS);
        retValue = ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER,
                                            ADK_SEV_LEVEL_ERROR);
        goto _runtime_error_1;
    }

    section_names = calloc(section_cnt, sizeof(char *));
    if (section_names == NULL)
    {
        ADK_PREINIT_WRITE("Memory allocation failure\n");
        retValue = ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE,
                                            ADK_SEV_LEVEL_ERROR);
        goto _runtime_error_1;
    }

    for (i = 0; i < section_cnt; i++)
    {
        section_names[i] = calloc(1, CFG_NAME_LEN);
        if (section_names[i] == NULL)
        {
            ADK_PREINIT_WRITE("Memory allocation failure\n");
            retValue = ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE,
                                                ADK_SEV_LEVEL_ERROR);
            goto adk_cfg_parse_return;
        }
    }

    if (rte_cfgfile_sections(cfg_handle_p, section_names, section_cnt) == 0)
    {
        ADK_PREINIT_WRITE("Unable to get sections from configuration file\n");
        retValue = ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR,
                                            ADK_SEV_LEVEL_ERROR);
        goto adk_cfg_parse_return;
    }

    cfg_params->cfg_handle = (void *)cfg_handle_p;

    for (i=0; i < section_cnt; i++)
    {
        debug_print("Section[%d]: %s\n", i, section_names[i]);
    }

    for (i=0; i < section_cnt; i++)
    {
        cfg_name_len = ADK_STRLEN(section_names[i]);

        /* Find section type */
        for (j = 0; j < (int)ADK_SIF_CFG_ARRAY_SIZE(cfg_file_scheme_g); j++) {
            section = &cfg_file_scheme_g[j];
            len = ADK_STRLEN(section->prefix);

            if (cfg_name_len < len)
                continue;

            if ((section_names[i][len] != '\0') &&
                    !isdigit(section_names[i][len]))
                continue;

            if (ADK_STRNCMP(section->prefix, section_names[i], len) == 0)
                break;
        }

        if(j >= (int)ADK_SIF_CFG_ARRAY_SIZE(cfg_file_scheme_g))
        {
            ADK_PREINIT_WRITE("Unsupported config file section - %s\n",
                              section_names[i]);
            retValue = ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER,
                                                ADK_SEV_LEVEL_ERROR);
            goto adk_cfg_parse_return;
        }

        retValue = adk_cfg_validate_name(section_names[i], section->prefix,
                                         section->numbers);
        if (retValue != ADK_SUCCESS)
        {
            ADK_PREINIT_WRITE("Invalid section name - %s\n", section_names[i]);
            goto adk_cfg_parse_return;
        }

        retValue = section->load(cfg_params, section_names[i]);
        if (retValue == ADK_W_CFG_SECTION_NOT_ENABLED)
        {
            /* Skipping disabled section */
            retValue = ADK_SUCCESS;
            continue;
        }
        else if (retValue != ADK_SUCCESS)
        {
            goto adk_cfg_parse_return;
        }
    }

adk_cfg_parse_return:
    for (i = 0; i < section_cnt; i++)
    {
        if (NULL != section_names[i])
        {
            free(section_names[i]);
        }
    }

    free(section_names);

_runtime_error_1:
    if (ADK_SUCCESS != retValue)
    {
        rte_cfgfile_close(cfg_handle_p);
    }

    return retValue;
}

adk_status_t
adk_cfg_get_filename(int argc, char **argv, char **config_filename)
{
    adk_status_t ret = ADK_SUCCESS;
    int opt = 0;

    opt = getopt(argc, argv, "f:");
    if (opt == -1)
    {
        ADK_PREINIT_WRITE("Usage: %s -f <config file>\n", argv[0]);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR, ADK_SEV_LEVEL_ERROR);
    }

    switch(opt)
    {
        case 'f':
            if (config_filename == NULL)
            {
                ADK_PREINIT_WRITE("Memory allocation failure\n");
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
            }
            *config_filename = strdup(optarg);

            break;

        default:
            ADK_PREINIT_WRITE("Usage Error: %s -f <config file>\n", argv[0]);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR, ADK_SEV_LEVEL_ERROR);
    }

    /* Reset getopt index */
    optind = 0;

    return ret;
}

static size_t
adk_cfg_skip_digits(const char *src)
{
        size_t i;

        for (i = 0; isdigit(src[i]); i++);

        return i;
}

adk_status_t
adk_cfg_validate_name(const char *name, const char *prefix, int num)
{
    adk_status_t ret = ADK_SUCCESS;
    size_t i, j;

    for (i = 0; (name[i] != '\0') && (prefix[i] != '\0'); i++) {
        if (name[i] != prefix[i])
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (prefix[i] != '\0')
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);

    if (!num) {
        if (name[i] != '\0')
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        else
            return ADK_SUCCESS;
    }

    if (num == 2) {
        j = adk_cfg_skip_digits(&name[i]);
        i += j;
        if ((j == 0) || (name[i] != '.'))
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        i++;
    }

    if (num == 1) {
        j = adk_cfg_skip_digits(&name[i]);
        i += j;
        if ((j == 0) || (name[i] != '\0'))
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

