/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_sif_cfg_include.h"
#include <errno.h>

/* Table to store pipeline section parameters */
static adk_cfg_pipeline_params_tbl_t   pipeline_params_tbl_g;

adk_status_t
adk_cfg_pipeline_init(adk_cfg_params_t *cfg_params)
{
    adk_status_t ret = ADK_SUCCESS;

    cfg_params->pipeline_params_tbl = (void *)&pipeline_params_tbl_g;

    return ret;
}

adk_status_t
adk_cfg_pipeline_parse(adk_cfg_params_t *cfg_params_hdl,
                       const char *section_name)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_pipeline_params_t *pipeline_params = NULL;
    int32_t section_entries_cnt = 0, i = 0, idx = 0;
    uint32_t r_idx = 0;
    struct rte_cfgfile_entry *section_entries = NULL, *entry = NULL;
    char tmp_buf[CFG_VALUE_LEN * 2], *p = NULL;
    adk_cfg_pipeline_flags_t flags;

    ADK_MEMSET(&flags, 0, sizeof(adk_cfg_pipeline_flags_t));

    /* Allocate memory to store parsed parameters */
    ret = adk_cfg_pipeline_param_alloc(cfg_params_hdl->pipeline_params_tbl, &pipeline_params);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
 
    /* Get number of entries present in current section */
    section_entries_cnt = rte_cfgfile_section_num_entries(cfg_params_hdl->cfg_handle, section_name);
    ADK_PREINIT_WRITE("\nSection:%s entries = %d\n", section_name, section_entries_cnt);

    /* Get all entries of current sections */
    section_entries = calloc(section_entries_cnt, sizeof(struct rte_cfgfile_entry));
    if (section_entries == NULL)
    {
        ADK_PREINIT_WRITE("Memory allocation failure\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
    }

    if (rte_cfgfile_section_entries(cfg_params_hdl->cfg_handle, section_name, 
                                    section_entries, section_entries_cnt) == 0)
    {
        ADK_PREINIT_WRITE("Get section entries failed\n");
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR, ADK_SEV_LEVEL_ERROR);
        goto adk_cfg_pipeline_parse_return;
    }



#ifdef ADK_DEBUG
    ADK_PREINIT_WRITE("[%s]\n", section_name);
    for (i = 0; i < section_entries_cnt; i++)
    {
        ADK_PREINIT_WRITE("\t%s = %s\n", section_entries[i].name, section_entries[i].value);
    }
#endif

    ADK_MEMCPY(pipeline_params->name, section_name, CFG_NAME_LEN);
    if(pipeline_params->name[CFG_NAME_LEN - 1] != '\0')
        pipeline_params->name[CFG_NAME_LEN - 1] = '\0';

    /* Parse each entry of current section */
    for (idx = 0; idx < section_entries_cnt; idx++)
    {
        entry = &section_entries[idx];

        /* Skip commneted lines */ 
        if (ADK_STRNCMP(entry->name, "#", 1) == 0)
        {
            continue;
        }

        if (ADK_STRNCMP(entry->name, "enable", sizeof("enable")) == 0)
        {
            if (strcasecmp(entry->value, "false") == 0)
            {
                /* Skip disabled section */
                ADK_MEMSET(pipeline_params, 0, sizeof(adk_cfg_pipeline_params_t));
                ret = ADK_W_CFG_SECTION_NOT_ENABLED;
                goto adk_cfg_pipeline_parse_return;
            }
            else
            {
                continue;
            }
        }
        else if (ADK_STRNCMP(entry->name, "stage", sizeof("stage")) == 0)
        {
            ret = adk_cfg_pipeline_stage_get(entry->value, &pipeline_params->stage);
            if (ret != ADK_SUCCESS)
            {
                goto adk_cfg_pipeline_parse_return;
            }
            flags.stage++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "lcore", sizeof("lcore")) == 0)
        {
            pipeline_params->lcore = (uint32_t)atoi(entry->value);
            flags.lcore++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "modules", sizeof("modules")) == 0)
        {
            ADK_MEMSET(tmp_buf, 0, sizeof(tmp_buf));
            ADK_STRNCPY(tmp_buf, entry->value, CFG_VALUE_LEN);
            i = 0;
            p = strtok (tmp_buf, " ");
            while (p != NULL)
            {
                ret = adk_cfg_pipeline_module_get(p, &pipeline_params->modules[i++]);
                if (ret != ADK_SUCCESS)
                {
                    goto adk_cfg_pipeline_parse_return;
                }

                if (pipeline_params->modules[i - 1] ==
                        ADK_SIF_CFG_PIPELINE_MODULE_QAT_RX)
                {
                    pipeline_params->qat_rx = TRUE;
                }
                else if (pipeline_params->modules[i - 1] ==
                        ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX)
                {
                    pipeline_params->qat_tx = TRUE;
                }

                if (i >= ADK_SIF_CFG_PIPELINE_MODULE_MAX)
                {
                    ADK_PREINIT_WRITE("Number of modules exceeds max supported modules %d\n", 
                                       ADK_SIF_CFG_PIPELINE_MODULE_MAX);
                    ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                    goto adk_cfg_pipeline_parse_return;
                }
                p = strtok(NULL, " ");
            }
            pipeline_params->num_modules = i;
            flags.modules++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "port_q_in", sizeof("port_q_in")) == 0)
        {
            ADK_MEMSET(tmp_buf, 0, sizeof(tmp_buf));
            ADK_STRNCPY(tmp_buf, entry->value, CFG_VALUE_LEN);
            p = strtok (tmp_buf, " ");
            while (p != NULL)
            {
                ret = adk_cfg_pipeline_port_q_get(p, cfg_params_hdl->port_params_tbl,
                        pipeline_params->port_q_in, &pipeline_params->num_port_q_in);
                if (ret != ADK_SUCCESS)
                {
                    goto adk_cfg_pipeline_parse_return;
                }
                p = strtok(NULL, " ");
            }
            flags.portqin++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "port_q_out", sizeof("port_q_out")) == 0)
        {
            ADK_MEMSET(tmp_buf, 0, sizeof(tmp_buf));
            ADK_STRNCPY(tmp_buf, entry->value, CFG_VALUE_LEN);
            p = strtok (tmp_buf, " ");
            while (p != NULL)
            {
                ret = adk_cfg_pipeline_port_q_get(p, cfg_params_hdl->port_params_tbl,
                        pipeline_params->port_q_out, &pipeline_params->num_port_q_out);
                if (ret != ADK_SUCCESS)
                {
                    goto adk_cfg_pipeline_parse_return;
                }
                p = strtok(NULL, " ");
            }
            flags.portqout++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "ring_q_in", sizeof("ring_q_in")) == 0)
        {
            ret = adk_cfg_pipeline_ring_idx_get((uint32_t)atoi(entry->value), pipeline_params->ring_q_in, 
                                       &pipeline_params->num_ring_q_in, &r_idx);
            if (ret == ADK_E_INVALID_PARAMETER)
            {
                goto adk_cfg_pipeline_parse_return;
            }
            if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
            {
                pipeline_params->ring_q_in[r_idx].ringid = (uint32_t)atoi(entry->value);
                flags.ringqin++;
                ret = ADK_SUCCESS;
                continue;
            }
        }
        else if (ADK_STRNCMP(entry->name, "ring_q_out", sizeof("ring_q_out")) == 0)
        {
            ret = adk_cfg_pipeline_ring_idx_get((uint32_t)atoi(entry->value), pipeline_params->ring_q_out, 
                                       &pipeline_params->num_ring_q_out, &r_idx);
            if (ret == ADK_E_INVALID_PARAMETER)
            {
                goto adk_cfg_pipeline_parse_return;
            }
            if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
            {
                pipeline_params->ring_q_out[r_idx].ringid = (uint32_t)atoi(entry->value);
                flags.ringqout++;
                ret = ADK_SUCCESS;
                continue;
            }
        }
        else if (ADK_STRNCMP(entry->name, "buff_in", sizeof("buff_in")) == 0)
        {
            ADK_MEMSET(tmp_buf, 0, sizeof(tmp_buf));
            ADK_STRNCPY(tmp_buf, entry->value, CFG_VALUE_LEN);
            p = strtok (tmp_buf, " ");
            while (p != NULL)
            {
                ret = adk_cfg_pipeline_buff_get(p, pipeline_params->buff_in,
                        &pipeline_params->num_buff_in);
                if (ret != ADK_SUCCESS)
                {
                    goto adk_cfg_pipeline_parse_return;
                }
                p = strtok(NULL, " ");
            }
            flags.buffin++;
            continue;
        }
        else
        {
            flags.unknown++;
        }
    }
    
    /* Validation */
    ret = adk_cfg_pipeline_stage_validate(pipeline_params, &flags);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_pipeline_parse_return;
    }

    /* Set defaults */
    ret = adk_cfg_pipeline_stage_defaults_set(pipeline_params);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_pipeline_parse_return;
    }

    free(section_entries);
    return ret;

adk_cfg_pipeline_parse_return:
    free(section_entries);
    adk_cfg_pipeline_param_free(cfg_params_hdl->pipeline_params_tbl, pipeline_params);
    return ret;
}

adk_status_t
adk_cfg_pipeline_param_alloc(adk_cfg_pipeline_params_tbl_t *pipeline_params_hdl,
                             adk_cfg_pipeline_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;
    adk_cfg_pipeline_params_t *p = pipeline_params_hdl->pipeline_params;

    for (i = 0; i < ADK_SIF_CFG_MAX_PIPELINES; i++)
    {
        if (p[i].valid == _INIT)
        {
            p[i].valid = _CONFIG;
            *params = &p[i];
            pipeline_params_hdl->size++;
            return ADK_SUCCESS;
        }
    }

    if (i == ADK_SIF_CFG_MAX_PIPELINES)
    {
        ADK_PREINIT_WRITE("Number of pipelines exceeds max supported %d\n", ADK_SIF_CFG_MAX_PIPELINES);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);        
    }
    
    return ret;
}

adk_status_t
adk_cfg_pipeline_param_free(adk_cfg_pipeline_params_tbl_t *pipeline_params_hdl,
                            adk_cfg_pipeline_params_t *params)
{
    adk_status_t ret = ADK_SUCCESS;

    if (params == NULL)
    {
        ADK_PREINIT_WRITE("Invalid memory access\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);        
    }

    params->valid = _INIT;
    pipeline_params_hdl->size--;
    
    return ret;
}

adk_status_t
adk_cfg_pipeline_param_rewind(adk_cfg_pipeline_params_tbl_t *pipeline_params_hdl)
{
    adk_status_t ret = ADK_SUCCESS;

    pipeline_params_hdl->cur = 0;
    
    return ret;
}

adk_status_t
adk_cfg_pipeline_param_get_next(adk_cfg_pipeline_params_tbl_t *pipeline_params_hdl,
                                adk_cfg_pipeline_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    
    /* Boundary check */
    if (pipeline_params_hdl->cur == pipeline_params_hdl->size)
    {
        pipeline_params_hdl->cur = 0;
        return ADK_E_LIST_RANGE;
    }

    /* Fetch config params */
    *params = &pipeline_params_hdl->pipeline_params[pipeline_params_hdl->cur];

    /* Point to next entry */
    pipeline_params_hdl->cur++;

    return ret;
}

adk_status_t
adk_cfg_pipeline_param_check(adk_cfg_pipeline_params_tbl_t *pipeline_params_hdl)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_pipeline_params_t *params = NULL;
    uint32_t start_new_stage = 0;
    uint32_t curr_stage = ADK_SIF_CFG_PIPELINE_STAGE_MIN;
    uint32_t prev_stage = ADK_SIF_CFG_PIPELINE_STAGE_MIN;
    uint32_t i = 0;

    /* Validate params handle pointer */
    if (!pipeline_params_hdl)
    {
        ADK_PREINIT_WRITE("Invalid pipeline param handle pointer\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE, ADK_SEV_LEVEL_ERROR);
    }

    adk_cfg_pipeline_param_rewind(pipeline_params_hdl);

    start_new_stage = 1;

    for (i = 0; i < ADK_SIF_CFG_MAX_PIPELINES; i++)
    {
        ret = adk_cfg_pipeline_param_get_next(pipeline_params_hdl, &params);
 
        if (ret == ADK_E_LIST_RANGE)
        {
            ret = ADK_SUCCESS;
            break;
        }

        curr_stage = params->stage;
        
        if (start_new_stage == 1)
        {
            if (curr_stage == ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR)
            {
                start_new_stage = 0;
                prev_stage = curr_stage;
            }
            else if (curr_stage == ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR)
            {
                start_new_stage = 1;
            }
            else
            {
                ADK_PREINIT_WRITE("Pipeline - %s consistency check failed\n", params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
        }
        else
        {
            if ((prev_stage == ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR) || 
                (prev_stage == ADK_SIF_CFG_PIPELINE_STAGE_INTERMEDIATE))
            {
                if (curr_stage == ADK_SIF_CFG_PIPELINE_STAGE_INTERMEDIATE)
                {
                    prev_stage = curr_stage;
                }
                else if (curr_stage == ADK_SIF_CFG_PIPELINE_STAGE_TERMINATOR)
                {
                    start_new_stage = 1;
                }
                else
                {
                    ADK_PREINIT_WRITE("Pipeline - %s consistency check failed\n", params->name);
                    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
                }
            }
            else
            {
                ADK_PREINIT_WRITE("Pipeline - %s consistency check failed\n", params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
        }
    }

    /* Check for partial stages in the last pipeline */
    if (start_new_stage != 1)
    {
        ADK_PREINIT_WRITE("Pipeline - %s consistency check failed. Partial pipeline\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
    
    return ret;
}

adk_status_t
adk_cfg_pipeline_stage_get(const char *value, adk_sif_cfg_pipeline_stage_t *stage)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "INITIATOR", sizeof("INITIATOR")) == 0)
    {
        *stage = ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR;
    }
    else if (ADK_STRNCMP(value, "INTERMEDIATE", sizeof("INTERMEDIATE")) == 0)
    {
        *stage = ADK_SIF_CFG_PIPELINE_STAGE_INTERMEDIATE;
    }
    else if (ADK_STRNCMP(value, "TERMINATOR", sizeof("TERMINATOR")) == 0)
    {
        *stage = ADK_SIF_CFG_PIPELINE_STAGE_TERMINATOR;
    }
    else if (ADK_STRNCMP(value, "INITIATOR_TERMINATOR", sizeof("INITIATOR_TERMINATOR")) == 0)
    {
        *stage = ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid pipeline stage - %s\n", value);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_cfg_pipeline_module_get(const char *value, adk_sif_cfg_pipeline_modules_t *modules)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "PORT_RX", sizeof("PORT_RX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_PORT_RX;
    }
    else if (ADK_STRNCMP(value, "PORT_TX", sizeof("PORT_TX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_PORT_TX;
    }
    else if (ADK_STRNCMP(value, "RING_RX", sizeof("RING_RX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_RING_RX;
    }
    else if (ADK_STRNCMP(value, "RING_TX", sizeof("RING_TX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_RING_TX;
    }
    else if (ADK_STRNCMP(value, "PKTIO_RX", sizeof("PKTIO_RX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_PKTIO_RX;
    }
    else if (ADK_STRNCMP(value, "PKTIO_TX", sizeof("PKTIO_TX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_PKTIO_TX;
    }
    else if (ADK_STRNCMP(value, "BUFF_RX", sizeof("BUFF_RX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_BUFF_RX;
    }
    else if (ADK_STRNCMP(value, "BUFF_TX", sizeof("BUFF_TX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_BUFF_TX;
    }
    else if (ADK_STRNCMP(value, "QAT_RX", sizeof("QAT_RX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_QAT_RX;
    }
    else if (ADK_STRNCMP(value, "QAT_TX", sizeof("QAT_TX")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX;
    }
    else if (ADK_STRNCMP(value, "ETHERNET_LLI", sizeof("ETHERNET_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_ETHERNET_LLI;
    }
    else if (ADK_STRNCMP(value, "ETHERNET_ULI", sizeof("ETHERNET_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_ETHERNET_ULI;
    }
    else if (ADK_STRNCMP(value, "IPV4_LLI", sizeof("IPV4_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPV4_LLI;
    }
    else if (ADK_STRNCMP(value, "IPV4_ULI", sizeof("IPV4_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPV4_ULI;
    }
    else if (ADK_STRNCMP(value, "IPSEC_LLI", sizeof("IPSEC_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI;
    }
    else if (ADK_STRNCMP(value, "IPSEC_LLI_TO_CRYPTO", sizeof("IPSEC_LLI_TO_CRYPTO")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI_TO_CRYPTO;
    }
    else if (ADK_STRNCMP(value, "IPSEC_LLI_FROM_CRYPTO", sizeof("IPSEC_LLI_FROM_CRYPTO")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_LLI_FROM_CRYPTO;
    }
    else if (ADK_STRNCMP(value, "IPSEC_ULI", sizeof("IPSEC_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI;
    }
    else if (ADK_STRNCMP(value, "L4_LLI", sizeof("L4_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_L4_LLI;
    }
    else if (ADK_STRNCMP(value, "L4_ULI", sizeof("L4_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_L4_ULI;
    }
    else if (ADK_STRNCMP(value, "IPSEC_ULI_TO_CRYPTO", sizeof("IPSEC_ULI_TO_CRYPTO")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI_TO_CRYPTO;
    }
    else if (ADK_STRNCMP(value, "IPSEC_ULI_FROM_CRYPTO", sizeof("IPSEC_ULI_FROM_CRYPTO")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IPSEC_ULI_FROM_CRYPTO;
    }
    else if (ADK_STRNCMP(value, "GTPU_LLI", sizeof("GTPU_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_GTPU_LLI;
    }
    else if (ADK_STRNCMP(value, "GTPU_ULI", sizeof("GTPU_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_GTPU_ULI;
    }
    else if (ADK_STRNCMP(value, "BFD_LLI", sizeof("BFD_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_BFD_LLI;
    }
    else if (ADK_STRNCMP(value, "BFD_ULI", sizeof("BFD_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_BFD_ULI;
    }
    else if (ADK_STRNCMP(value, "ETH_OAM_LLI", sizeof("ETH_OAM_LLI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_ETH_OAM_LLI;
    }
    else if (ADK_STRNCMP(value, "ETH_OAM_ULI", sizeof("ETH_OAM_ULI")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_ETH_OAM_ULI;
    }
    else if (ADK_STRNCMP(value, "*", sizeof("*")) == 0)
    {
        *modules = ADK_SIF_CFG_PIPELINE_MODULE_IGNORE;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid pipeline module - %s\n", value);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_cfg_pipeline_port_q_get(const char *value,
                            adk_cfg_port_params_tbl_t *port_params_hdl,
                            adk_sif_cfg_pipeline_port_q_t *port_q,
                            uint32_t *num)
{
    adk_status_t ret = ADK_SUCCESS;
    char s[256];
    const char *p, *p0 = value;
    char *end;
    enum fieldnames {
        FLD_PORT = 0,
        FLD_QUEUE,
        _NUM_FLD
    };
    unsigned long int_fld[_NUM_FLD];
    char *str_fld[_NUM_FLD];
    int i;
    unsigned size;
    uint32_t p_idx = 0, q_idx = 0;
    uint16_t weight = 0;

    while ((p = strchr(p0,'(')) != NULL)
    {
        ++p;
        if((p0 = strchr(p,')')) == NULL)
        {
            ADK_PREINIT_WRITE("Invalid port queue configuration\n");
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }

        size = p0 - p;
        if(size >= sizeof(s))
        {
            ADK_PREINIT_WRITE("Invalid port queue configuration\n");
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }

        snprintf(s, sizeof(s), "%.*s", size, p);
        if (rte_strsplit(s, sizeof(s), str_fld, _NUM_FLD, ',') != _NUM_FLD)
        {
            ADK_PREINIT_WRITE("Invalid port queue configuration\n");
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
        
        for (i = 0; i < _NUM_FLD; i++)
        {
            errno = 0;
            int_fld[i] = strtoul(str_fld[i], &end, 0);
            if (errno != 0 || end == str_fld[i] || int_fld[i] > 255)
            {
                ADK_PREINIT_WRITE("Invalid port queue configuration\n");
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
        }

        /* Update port queue info */
        ret = adk_cfg_pipeline_port_idx_get((uint8_t)int_fld[FLD_PORT], port_q, num, &p_idx);
        if (ret == ADK_E_INVALID_PARAMETER)
        {
            return ret;
        }
        if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
        {
            port_q[p_idx].portid = (uint8_t)int_fld[FLD_PORT];
            ret = ADK_SUCCESS;
        }
        ret = adk_cfg_pipeline_queue_idx_get((uint8_t)int_fld[FLD_QUEUE], port_q[p_idx].queueid, &port_q[p_idx].num_queues, &q_idx);
        if (ret == ADK_E_INVALID_PARAMETER)
        {
            return ret;
        }
        if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
        {
            port_q[p_idx].queueid[q_idx] = (uint8_t)int_fld[FLD_QUEUE];
            ret = adk_cfg_port_param_queue_weight_get(port_params_hdl, 
                                                      port_q[p_idx].portid, 
                                                      port_q[p_idx].queueid[q_idx], 
                                                      &weight);
            if (ret == ADK_E_INVALID_PARAMETER)
            {
                return ret;
            }
            port_q[p_idx].weight[q_idx] = weight;
            ret = ADK_SUCCESS;
        }
    }

    return ret;
}

adk_status_t
adk_cfg_pipeline_port_idx_get(const uint8_t portid,
                              adk_sif_cfg_pipeline_port_q_t *p,
                              uint32_t *n,
                              uint32_t *idx)
{
    uint32_t i = 0;

    for (i = 0; i < *n; i++)
    {
        if (p[i].portid == portid)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_PORT)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }
    
    ADK_PREINIT_WRITE("Number of port exceeds max supported %d\n", ADK_SIF_CFG_MAX_PORT);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_cfg_pipeline_queue_idx_get(const uint16_t queueid,
                               uint16_t *q,
                               uint32_t *n,
                               uint32_t *idx)
{
    uint32_t i = 0;

    for (i = 0; i < *n; i++)
    {
        if (q[i] == queueid)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_QUEUES_PER_PORT)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }
    
    ADK_PREINIT_WRITE("Number of port queues exceeds max supported %d\n", ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_cfg_pipeline_ring_idx_get(const uint16_t ringid,
                              adk_sif_cfg_pipeline_ring_q_t *r,
                              uint32_t *n,
                              uint32_t *idx)
{
    uint32_t i = 0;
    
    for (i = 0; i < *n; i++)
    {
        if (r[i].ringid == ringid)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_RING_QUEUES)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }

    ADK_PREINIT_WRITE("Number of ring queues exceeds max supported %d\n", ADK_SIF_CFG_MAX_RING_QUEUES);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_cfg_pipeline_buff_get(const char *value,
                            adk_sif_cfg_pipeline_buff_t __attribute__((unused)) *buff,
                            uint32_t __attribute__((unused))*num)
{
    adk_status_t ret = ADK_SUCCESS;
    char s[256];
    char *str_bid[256];
    char *endPtr;
    unsigned int int_bid[ADK_SIF_CFG_MAX_PKT_BUFFS];
    int i, tot;
    uint32_t r_idx;

    if(ADK_STRLEN(value) >= sizeof(s))
    {
        ADK_PREINIT_WRITE("Too many buffers configured for a pipeline.\
                Max Allowed Buffers: %d \n", ADK_SIF_CFG_MAX_PKT_BUFFS);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    ADK_MEMCPY((char *) s, value, ADK_STRLEN(value));

    tot = rte_strsplit(s, ADK_STRLEN(s), str_bid, ADK_SIF_CFG_MAX_PKT_BUFFS + 1, ',');
    if (tot > ADK_SIF_CFG_MAX_PKT_BUFFS)
    {
        ADK_PREINIT_WRITE("Too many buffers configured for a pipeline.\
                Max Allowed Buffers: %d \n", ADK_SIF_CFG_MAX_PKT_BUFFS);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    for (i=0;i < tot;i++)
    {
        endPtr = str_bid[i];
        errno = 0;
        int_bid[i] = (unsigned int) strtol(str_bid[i], &endPtr, 10);

        if (errno == 0) 
        {
            ret = adk_cfg_pipeline_buff_idx_get(int_bid[i], buff,
                    num, &r_idx);
            if (ret == ADK_E_INVALID_PARAMETER)
            {
                continue;
            }
            if (ret == ADK_E_CFG_ENTRY_NOT_FOUND)
            {
                buff[r_idx].buff_id = int_bid[i];
                ret = ADK_SUCCESS;
            }
        }
        else
        {
            ADK_PREINIT_WRITE("Invalid input\n");
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
    }
    return ret;
}

adk_status_t
adk_cfg_pipeline_buff_idx_get(const uint16_t buff_id,
                              adk_sif_cfg_pipeline_buff_t *r,
                              uint32_t *n,
                              uint32_t *idx)
{
    uint32_t i = 0;
    
    for (i = 0; i < *n; i++)
    {
        if (r[i].buff_id == buff_id)
        {
            *idx = i;
            return ADK_SUCCESS;
        }
    }
    if (*n < ADK_SIF_CFG_MAX_PKT_BUFFS)
    {
        *idx = *n;
        (*n)++;
        return ADK_E_CFG_ENTRY_NOT_FOUND;
    }

    ADK_PREINIT_WRITE("Number of Packet Buffers exceeds max supported %d\n", ADK_SIF_CFG_MAX_PKT_BUFFS);
    return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
}

adk_status_t
adk_cfg_pipeline_entry_function_get(const adk_sif_cfg_pipeline_modules_t entry_mod,
                                    const adk_sif_cfg_pipeline_modules_t exit_mod,
                                    uint16_t *entry_func)
{
    adk_status_t ret = ADK_SUCCESS;

    if (entry_mod == ADK_SIF_CFG_PIPELINE_MODULE_ETHERNET_LLI)
    {
        *entry_func = adkEntryFunctions_g[__FN_ETH_DP_PROCESS_PKT];
    }
    else if (entry_mod == ADK_SIF_CFG_PIPELINE_MODULE_BFD_ULI)
    { 
        *entry_func = 0;
    }
    else if (entry_mod == ADK_SIF_CFG_PIPELINE_MODULE_PKTIO_RX)
    {
        switch(exit_mod)
        {
#if 0
            case ADK_SIF_CFG_PIPELINE_MODULE_IGNORE:
                *entry_func = adkEntryFunctions_g[__FN_PKTIO_RX_DP_PROCESS_PKT];
                break;
#endif
            default:
                *entry_func = adkEntryFunctions_g[__FN_ETH_DP_PROCESS_PKT];
        }
    }

    else
    {
        ADK_PREINIT_WRITE("Invalid entry function\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_cfg_pipeline_stage_validate(adk_cfg_pipeline_params_t *pipeline_params,
                                adk_cfg_pipeline_flags_t *flags)
{
    adk_status_t ret = ADK_SUCCESS;

    if (pipeline_params->num_modules != (_TX_IDX + 1))
    {
        ADK_PREINIT_WRITE("Section - %s: number of modules is invalid\n", pipeline_params->name);
        ADK_PREINIT_WRITE("Valid Format: modules = <Rx I/O type> <Entry module> <Exit module> <Tx I/O type>\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    switch(pipeline_params->stage)
    {
        case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR:
            if (!flags->stage)
            {
                ADK_PREINIT_WRITE("Section - %s: pipleine stage entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->lcore)
            {
                ADK_PREINIT_WRITE("Section - %s: logical core entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->modules)
            {
                ADK_PREINIT_WRITE("Section - %s: modules entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            
            if ((!flags->portqin) && (!flags->ringqin) && (!flags->buffin))
            {
                ADK_PREINIT_WRITE("Section - %s: input port queue or ring or packet buffer entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            break;

        case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR:
            if (!flags->stage)
            {
                ADK_PREINIT_WRITE("Section - %s: pipleine stage entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->lcore)
            {
                ADK_PREINIT_WRITE("Section - %s: logical core entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->modules)
            {
                ADK_PREINIT_WRITE("Section - %s: modules entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            
            if ((!flags->portqin) && (!flags->ringqin) && (!flags->buffin))
            {
                ADK_PREINIT_WRITE("Section - %s: input port or ring queue entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if ((!flags->ringqout) && (!flags->buffout))
            {
                ADK_PREINIT_WRITE("Section - %s: output ring queue entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            break;

        case ADK_SIF_CFG_PIPELINE_STAGE_INTERMEDIATE:
            if (!flags->stage)
            {
                ADK_PREINIT_WRITE("Section - %s: pipleine stage entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->lcore)
            {
                ADK_PREINIT_WRITE("Section - %s: logical core entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->modules)
            {
                ADK_PREINIT_WRITE("Section - %s: modules entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if ((pipeline_params->qat_rx != TRUE) && (!flags->ringqin) && (!flags->buffin))
            {
                ADK_PREINIT_WRITE("Section - %s: input ring queue entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if ((pipeline_params->qat_tx != TRUE) && (!flags->ringqout) && (!flags->buffout))
            {
                ADK_PREINIT_WRITE("Section - %s: output ring queue entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            break;

        case ADK_SIF_CFG_PIPELINE_STAGE_TERMINATOR:
            if (!flags->stage)
            {
                ADK_PREINIT_WRITE("Section - %s: pipleine stage entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->lcore)
            {
                ADK_PREINIT_WRITE("Section - %s: logical core entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if (!flags->modules)
            {
                ADK_PREINIT_WRITE("Section - %s: modules entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }

            if ((pipeline_params->qat_rx != TRUE) && (!flags->ringqin) && (!flags->buffin))
            {
                ADK_PREINIT_WRITE("Section - %s: input ring queue entry not configured\n", pipeline_params->name);
                return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
            }
            break;

        default:
            ADK_PREINIT_WRITE("Section - %s: Unsupported pipeline stage type\n", pipeline_params->name);
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_cfg_pipeline_module_validate(adk_sif_cfg_pipeline_modules_t mod)
{
    adk_status_t ret = ADK_SUCCESS;

    if ((mod <= ADK_SIF_CFG_PIPELINE_MODULE_MIN) || (mod >= ADK_SIF_CFG_PIPELINE_MODULE_MAX))
    {
        ADK_PREINIT_WRITE("Unsupported module type\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_cfg_pipeline_stage_defaults_set(adk_cfg_pipeline_params_t *pipeline_params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t p = 0, r = 0;

    switch(pipeline_params->stage)
    {
        case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR_TERMINATOR:
        case ADK_SIF_CFG_PIPELINE_STAGE_INITIATOR:
            ret = adk_cfg_pipeline_entry_function_get(pipeline_params->modules[_ENTRY_MOD_IDX], 
                                                      pipeline_params->modules[_EXIT_MOD_IDX],
                                                      &pipeline_params->entry_func);
            if (ret != ADK_SUCCESS)
            {
                return ret;
            }
        break;

        default:
        ;
    }

    switch (pipeline_params->modules[_RX_IDX])
    {
        case ADK_SIF_CFG_PIPELINE_MODULE_PORT_RX:
            for (p = 0; p < pipeline_params->num_port_q_in; p++)
            {
                pipeline_params->port_q_in[p].type = ADK_SIF_CFG_PIPELINE_PORT_TYPE_DPDK_PORT;
            }
            break;

        case ADK_SIF_CFG_PIPELINE_MODULE_RING_RX:
            for (r = 0; r < pipeline_params->num_ring_q_in; r++)
            {
                pipeline_params->ring_q_in[r].type = ADK_SIF_CFG_PIPELINE_PORT_TYPE_DPDK_RING;
            }

            break;

        case ADK_SIF_CFG_PIPELINE_MODULE_QAT_RX:
            pipeline_params->qat_rx = TRUE;
            break;

        default:
            ;
    }

    switch (pipeline_params->modules[_TX_IDX])
    {
        case ADK_SIF_CFG_PIPELINE_MODULE_RING_TX:
            for (r = 0; r < pipeline_params->num_ring_q_out; r++)
            {
                pipeline_params->ring_q_out[r].type = ADK_SIF_CFG_PIPELINE_PORT_TYPE_DPDK_RING;
            }

            break;

        case ADK_SIF_CFG_PIPELINE_MODULE_QAT_TX:
            pipeline_params->qat_tx = TRUE;
            break;

        default:
            ;
    }

    return ret;
}

