/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_sif_cfg_include.h"

/* Table to store pkt_buffs section parameters */
static adk_cfg_pkt_buffs_params_tbl_t pkt_buffs_params_tbl_g;

adk_status_t
adk_cfg_pkt_buffs_init(
        adk_cfg_params_t *cfg_params)
{
    adk_status_t ret = ADK_SUCCESS;

    cfg_params->pkt_buffs_params_tbl = (void *)&pkt_buffs_params_tbl_g;

    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_parse(
        adk_cfg_params_t *cfg_params_hdl,
        const char *section_name)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_pkt_buffs_params_t *pkt_buffs_params = NULL;
    int32_t section_entries_cnt = 0, idx = 0;
    struct rte_cfgfile_entry *section_entries = NULL, *entry = NULL;
    adk_cfg_pkt_buffs_flags_t flags;
#ifdef ADK_DEBUG
    int32_t i = 0;
#endif

    ADK_MEMSET(&flags, 0, sizeof(adk_cfg_pkt_buffs_flags_t));

    /* Allocate memory to store parsed parameters */
    ret = adk_cfg_pkt_buffs_param_alloc(cfg_params_hdl->pkt_buffs_params_tbl, &pkt_buffs_params);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
 
    section_entries_cnt = rte_cfgfile_section_num_entries(cfg_params_hdl->cfg_handle, section_name);
    ADK_PREINIT_WRITE("\nSection:%s entries = %d\n", section_name, section_entries_cnt);

    /* Get all entries of current sections */
    section_entries = ADK_RAM_CALLOC(section_entries_cnt, sizeof(struct rte_cfgfile_entry));
    if (section_entries == NULL)
    {
        ADK_PREINIT_WRITE("Memory allocation failure\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
    }

    if (rte_cfgfile_section_entries(cfg_params_hdl->cfg_handle, section_name, 
                                    section_entries, section_entries_cnt) == 0)
    {
        ADK_PREINIT_WRITE("Get section entries failed\n");
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR, ADK_SEV_LEVEL_ERROR);
        goto adk_cfg_pkt_buffs_parse_return;
    }

#ifdef ADK_DEBUG
    ADK_PREINIT_WRITE("[%s]\n", section_name);
    for (i = 0; i < section_entries_cnt; i++)
    {
        ADK_PREINIT_WRITE("\t%s = %s\n", section_entries[i].name, section_entries[i].value);
    }
#endif

    ADK_MEMCPY(pkt_buffs_params->name, section_name, CFG_NAME_LEN);
    if(pkt_buffs_params->name[CFG_NAME_LEN - 1] != '\0')
        pkt_buffs_params->name[CFG_NAME_LEN - 1] = '\0';

    /* Parse each entry of current section */
    for (idx = 0; idx < section_entries_cnt; idx++)
    {
        entry = &section_entries[idx];

        /* Skip commneted lines */ 
        if (ADK_STRNCMP(entry->name, "#", 1) == 0)
        {
            continue;
        }

        if (ADK_STRNCMP(entry->name, "enable", sizeof("enable")) == 0)
        {
            if (strcasecmp(entry->value, "false") == 0)
            {
                /* Skip disabled section */
                ADK_MEMSET(pkt_buffs_params, 0, sizeof(adk_cfg_pkt_buffs_params_t));
                ret = ADK_W_CFG_SECTION_NOT_ENABLED;
                goto adk_cfg_pkt_buffs_parse_return;
            }
            else
            {
                continue;
            }
        }
        else if (ADK_STRNCMP(entry->name, "pkt_buffs", sizeof("pkt_buffs")) == 0)
        {
            pkt_buffs_params->pkt_buffsid = atoi(entry->value); 
            flags.pkt_buffs++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_mempool", sizeof("rx_mempool")) == 0)
        {
            pkt_buffs_params->rx_mempool = atoi(entry->value); 
            flags.rx_mempool++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "transmit", sizeof("transmit")) == 0)
        {
            adk_cfg_pkt_buffs_transmit_type_get(entry->value, &pkt_buffs_params->transmit);
            flags.transmit++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "weight", sizeof("weight")) == 0)
        {
            pkt_buffs_params->weight = atoi(entry->value); 
            flags.weight++;
            continue;
        }
        else
        {
            flags.unknown++;
        }
    }
    
    /* Validation */
    ret = adk_cfg_pkt_buffs_validate(pkt_buffs_params, &flags);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_pkt_buffs_parse_return;
    }

    /* Set defaults */
    ret = adk_cfg_pkt_buffs_defaults_set(pkt_buffs_params);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_pkt_buffs_parse_return;
    }

    free(section_entries);
    return ret;

adk_cfg_pkt_buffs_parse_return:
    free(section_entries);
    adk_cfg_pkt_buffs_param_free(cfg_params_hdl->pkt_buffs_params_tbl, pkt_buffs_params);
    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_param_alloc(
        adk_cfg_pkt_buffs_params_tbl_t *pkt_buffs_params_hdl,
        adk_cfg_pkt_buffs_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;
    adk_cfg_pkt_buffs_params_t *p = pkt_buffs_params_hdl->pkt_buffs_params;

    for (i = 0; i < ADK_SIF_CFG_MAX_PKT_BUFFS; i++)
    {
        if (p[i].valid == _INIT)
        {
            p[i].valid = _CONFIG;
            *params = &p[i];
            pkt_buffs_params_hdl->size++;
            return ADK_SUCCESS;
        }
    }

    if (i == ADK_SIF_CFG_MAX_PKT_BUFFS)
    {
        ADK_PREINIT_WRITE("Number of pkt_buffss exceeds max suppkt_buffsed %d\n", ADK_SIF_CFG_MAX_PKT_BUFFS);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);        
    }
    
    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_param_free(
        adk_cfg_pkt_buffs_params_tbl_t *pkt_buffs_params_hdl,
        adk_cfg_pkt_buffs_params_t *params)
{
    adk_status_t ret = ADK_SUCCESS;

    if (params == NULL)
    {
        ADK_PREINIT_WRITE("Invalid memory access\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);        
    }

    params->valid = _INIT;
    pkt_buffs_params_hdl->size--;
    
    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_param_rewind(
        adk_cfg_pkt_buffs_params_tbl_t *pkt_buffs_params_hdl)
{
    adk_status_t ret = ADK_SUCCESS;

    pkt_buffs_params_hdl->cur = 0;
    
    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_param_get_next(
        adk_cfg_pkt_buffs_params_tbl_t *pkt_buffs_params_hdl,
        adk_cfg_pkt_buffs_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;
    
    /* Boundary check */
    if (pkt_buffs_params_hdl->cur == pkt_buffs_params_hdl->size)
    {
        pkt_buffs_params_hdl->cur = 0;
        return ADK_E_LIST_RANGE;
    }

    /* Skip deleted entries */ 
    for (i = pkt_buffs_params_hdl->cur; i < pkt_buffs_params_hdl->size; i++)
    {
        if (pkt_buffs_params_hdl->pkt_buffs_params[pkt_buffs_params_hdl->cur].valid == _CONFIG)
        {
            break;
        }
    }
    pkt_buffs_params_hdl->cur = i;

    /* Fetch config params */
    *params = &pkt_buffs_params_hdl->pkt_buffs_params[pkt_buffs_params_hdl->cur];

    /* Point to next entry */
    pkt_buffs_params_hdl->cur++;

    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_transmit_type_get(const char *value, uint16_t *transmit)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "LOOP", sizeof("LOOP")) == 0)
    {
        *transmit = ADK_CFG_PKT_BUFFS_TYPE_LOOP;
    }
    else if (ADK_STRNCMP(value, "FINISH", sizeof("FINISH")) == 0)
    {
        *transmit = ADK_CFG_PKT_BUFFS_TYPE_FINISH;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid PKT Buffer transmit %s\n", value);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
 
    return ret;
}

/* Dummy Code: Future reference for extension */
#if 0
adk_status_t
adk_cfg_pkt_buffs_promisc_get(const char *value, uint8_t *promisc)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "YES", sizeof("YES")) == 0)
    {
        *promisc = 1;
    }
    else if (ADK_STRNCMP(value, "", sizeof("")) == 0)
    {
        *promisc = 0;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid pkt_buffs promisc setting\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
 
    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_rx_mq_mode_get(const char *value, uint32_t *rx_mq_mode)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "NONE", sizeof("NONE")) == 0)
    {
        *rx_mq_mode = ADK_CFG_PORT_RX_MQ_MODE_NONE;
    }
    else if (ADK_STRNCMP(value, "RSS", sizeof("RSS")) == 0)
    {
        *rx_mq_mode = ADK_CFG_PORT_RX_MQ_MODE_RSS;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid pkt_buffs rx multi-queue mode setting\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
 
    return ret;
}
#endif

adk_status_t
adk_cfg_pkt_buffs_validate(
        adk_cfg_pkt_buffs_params_t *params,
        adk_cfg_pkt_buffs_flags_t *flags)
{
    adk_status_t ret = ADK_SUCCESS;
    
    if (!flags->pkt_buffs)
    {
        ADK_PREINIT_WRITE("Section - %s:  pkt_buffs entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->rx_mempool)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx mempool entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->transmit)
    {
        ADK_PREINIT_WRITE("Section - %s:  transmit entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->weight)
    {
        ADK_PREINIT_WRITE("Section - %s:  weight entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
    return ret;
}

adk_status_t
adk_cfg_pkt_buffs_defaults_set(
        adk_cfg_pkt_buffs_params_t *params __attribute__((unused)))
{
    adk_status_t ret = ADK_SUCCESS;
    return ret;
}

