/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_sif_cfg_include.h"

/* Table to store port section parameters */
static adk_cfg_port_params_tbl_t port_params_tbl_g;


adk_status_t
adk_cfg_port_init(adk_cfg_params_t *cfg_params)
{
    adk_status_t ret = ADK_SUCCESS;

    cfg_params->port_params_tbl = (void *)&port_params_tbl_g;

    return ret;
}

adk_status_t
adk_cfg_port_parse(adk_cfg_params_t *cfg_params_hdl,
                   const char *section_name)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_port_params_t *port_params = NULL;
    int32_t section_entries_cnt = 0, idx = 0;
    struct rte_cfgfile_entry *section_entries = NULL, *entry = NULL;
    adk_cfg_port_flags_t flags;
#ifdef ADK_DEBUG
    int32_t i = 0;
#endif

    ADK_MEMSET(&flags, 0, sizeof(adk_cfg_port_flags_t));

    /* Allocate memory to store parsed parameters */
    ret = adk_cfg_port_param_alloc(cfg_params_hdl->port_params_tbl, &port_params);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
 
    section_entries_cnt = rte_cfgfile_section_num_entries(cfg_params_hdl->cfg_handle, section_name);
    ADK_PREINIT_WRITE("\nSection:%s entries = %d\n", section_name, section_entries_cnt);

    /* Get all entries of current sections */
    section_entries = calloc(section_entries_cnt, sizeof(struct rte_cfgfile_entry));
    if (section_entries == NULL)
    {
        ADK_PREINIT_WRITE("Memory allocation failure\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
    }

    if (rte_cfgfile_section_entries(cfg_params_hdl->cfg_handle, section_name, 
                                    section_entries, section_entries_cnt) == 0)
    {
        ADK_PREINIT_WRITE("Get section entries failed\n");
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR, ADK_SEV_LEVEL_ERROR);
        goto adk_cfg_port_parse_return;
    }

#ifdef ADK_DEBUG
    ADK_PREINIT_WRITE("[%s]\n", section_name);
    for (i = 0; i < section_entries_cnt; i++)
    {
        ADK_PREINIT_WRITE("\t%s = %s\n", section_entries[i].name, section_entries[i].value);
    }
#endif

    ADK_MEMCPY(port_params->name, section_name, CFG_NAME_LEN);
    if(port_params->name[CFG_NAME_LEN-1] != '\0')
        port_params->name[CFG_NAME_LEN-1] = '\0';

    /* Parse each entry of current section */
    for (idx = 0; idx < section_entries_cnt; idx++)
    {
        entry = &section_entries[idx];

        /* Skip commneted lines */ 
        if (ADK_STRNCMP(entry->name, "#", 1) == 0)
        {
            continue;
        }

        if (ADK_STRNCMP(entry->name, "enable", sizeof("enable")) == 0)
        {
            if (strcasecmp(entry->value, "false") == 0)
            {
                /* Skip disabled section */
                ADK_MEMSET(port_params, 0, sizeof(adk_cfg_port_params_t));
                ret = ADK_W_CFG_SECTION_NOT_ENABLED;
                goto adk_cfg_port_parse_return;
            }
            else
            {
                continue;
            }
        }
        else if (ADK_STRNCMP(entry->name, "port", sizeof("port")) == 0)
        {
            port_params->portid = atoi(entry->value); 
            flags.port++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "speed", sizeof("speed")) == 0)
        {
            port_params->speed = atoi(entry->value); 
            flags.speed++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "duplex", sizeof("duplex")) == 0)
        {
            ret = adk_cfg_port_duplex_get(entry->value, &port_params->duplex);
            if (ret != ADK_SUCCESS)
            {
                goto adk_cfg_port_parse_return;
            }
            flags.duplex++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "promisc", sizeof("promisc")) == 0)
        {
            ret = adk_cfg_port_promisc_get(entry->value, &port_params->promisc);
            if (ret != ADK_SUCCESS)
            {
                goto adk_cfg_port_parse_return;
            }
            flags.promisc++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_mq_mode", sizeof("rx_mq_mode")) == 0)
        {
            ret = adk_cfg_port_rx_mq_mode_get(entry->value, &port_params->rx_mq_mode);
            if (ret != ADK_SUCCESS)
            {
                goto adk_cfg_port_parse_return;
            }
            flags.rx_mq_mode++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_queues", sizeof("rx_queues")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->rx_queues, &port_params->num_rx_queues);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of rx queues exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            flags.rx_queues = port_params->num_rx_queues;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_queue_mempool", sizeof("rx_queue_mempool")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->rx_queue_mempool, &(flags.rx_queue_mempool));
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of rx queue mempool exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_queue_size", sizeof("rx_queue_size")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->rx_queue_size, &flags.rx_queue_size);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of rx queue size exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_queue_weight", sizeof("rx_queue_weight")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->rx_queue_weight, &flags.rx_queue_weight);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of rx queue weight exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "rx_queue_burst", sizeof("rx_queue_burst")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->rx_queue_burst, &flags.rx_queue_burst);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of rx queue burst exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "tx_queues", sizeof("tx_queues")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->tx_queues, &port_params->num_tx_queues);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of tx queues exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            flags.tx_queues = port_params->num_tx_queues;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "tx_queue_socket", sizeof("tx_queue_socket")) == 0)
        {
            ret = adk_cfg_port_util_socket_list_get(entry->value, 
                                                    ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                                    port_params->tx_queue_socket, &flags.tx_queue_socket);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of tx queue size exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "tx_queue_size", sizeof("tx_queue_size")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->tx_queue_size, &flags.tx_queue_size);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of tx queue size exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "tx_queue_burst", sizeof("tx_queue_burst")) == 0)
        {
            ret = adk_cfg_port_util_list_get(entry->value, 
                                             ADK_SIF_CFG_MAX_QUEUES_PER_PORT, 
                                             port_params->tx_queue_burst, &flags.tx_queue_burst);
            if (ret != ADK_SUCCESS)
            {
                ADK_PREINIT_WRITE("Number of tx queue burst exceeds max supported %d\n", 
                        ADK_SIF_CFG_MAX_QUEUES_PER_PORT);
                goto adk_cfg_port_parse_return;
            }
            continue;
        }
        else
        {
            flags.unknown++;
        }
    }
    
    /* Validation */
    ret = adk_cfg_port_validate(port_params, &flags);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_port_parse_return;
    }

    /* Set defaults */
    ret = adk_cfg_port_defaults_set(port_params);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_port_parse_return;
    }

    free(section_entries);
    return ret;

adk_cfg_port_parse_return:
    free(section_entries);
    adk_cfg_port_param_free(cfg_params_hdl->port_params_tbl, port_params);
    return ret;
}

adk_status_t
adk_cfg_port_param_alloc(adk_cfg_port_params_tbl_t *port_params_hdl,
                         adk_cfg_port_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;
    adk_cfg_port_params_t *p = port_params_hdl->port_params;

    for (i = 0; i < ADK_SIF_CFG_MAX_PORT; i++)
    {
        if (p[i].valid == _INIT)
        {
            p[i].valid = _CONFIG;
            *params = &p[i];
            port_params_hdl->size++;
            return ADK_SUCCESS;
        }
    }

    if (i == ADK_SIF_CFG_MAX_PORT)
    {
        ADK_PREINIT_WRITE("Number of ports exceeds max supported %d\n", ADK_SIF_CFG_MAX_PORT);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);        
    }
    
    return ret;
}

adk_status_t
adk_cfg_port_param_free(adk_cfg_port_params_tbl_t *port_params_hdl,
                        adk_cfg_port_params_t *params)
{
    adk_status_t ret = ADK_SUCCESS;

    if (params == NULL)
    {
        ADK_PREINIT_WRITE("Invalid memory access\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);        
    }

    params->valid = _INIT;
    port_params_hdl->size--;
    
    return ret;
}

adk_status_t
adk_cfg_port_param_rewind(adk_cfg_port_params_tbl_t *port_params_hdl)
{
    adk_status_t ret = ADK_SUCCESS;

    port_params_hdl->cur = 0;
    
    return ret;
}

adk_status_t
adk_cfg_port_param_get_next(adk_cfg_port_params_tbl_t *port_params_hdl,
                            adk_cfg_port_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;
    
    /* Boundary check */
    if (port_params_hdl->cur == port_params_hdl->size)
    {
        port_params_hdl->cur = 0;
        return ADK_E_LIST_RANGE;
    }

    /* Skip deleted entries */ 
    for (i = port_params_hdl->cur; i < port_params_hdl->size; i++)
    {
        if (port_params_hdl->port_params[port_params_hdl->cur].valid == _CONFIG)
        {
            break;
        }
    }
    port_params_hdl->cur = i;

    /* Fetch config params */
    *params = &port_params_hdl->port_params[port_params_hdl->cur];

    /* Point to next entry */
    port_params_hdl->cur++;

    return ret;
}

adk_status_t
adk_cfg_port_param_queue_weight_get(adk_cfg_port_params_tbl_t *port_params_hdl,
                                    uint8_t port_id,
                                    uint16_t queue_id,
                                    uint16_t *weight)
{
    adk_cfg_port_params_t *port_params = port_params_hdl->port_params;
    uint32_t p = 0, q = 0;

    for (p = 0; p < port_params_hdl->size; p++)
    {
        if ((port_params[p].valid == _CONFIG) && (port_params[p].portid == port_id))
        {
            for (q = 0; q < port_params[p].num_rx_queues; q++)
            {
                if ( port_params[p].rx_queues[q] == queue_id)
                {
                    *weight = port_params[p].rx_queue_weight[q];
                    return ADK_SUCCESS;
                }
            }
        }
    }

    return ADK_E_LIST_NOT_FOUND;
}

adk_status_t
adk_cfg_port_duplex_get(const char *value, uint32_t *duplex)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "FULL_DUPLEX", sizeof("FULL_DUPLEX")) == 0)
    {
        *duplex = ADK_CFG_PORT_DUPLEX_FULL;
    }
    else if (ADK_STRNCMP(value, "HALF_DUPLEX", sizeof("HALF_DUPLEX")) == 0)
    {
        *duplex = ADK_CFG_PORT_DUPLEX_HALF;
    }
    else if (ADK_STRNCMP(value, "AUTO", sizeof("AUTO")) == 0)
    {
        *duplex = ADK_CFG_PORT_DUPLEX_AUTO;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid port duplex mode\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
 
    return ret;
}

adk_status_t
adk_cfg_port_promisc_get(const char *value, uint8_t *promisc)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "YES", sizeof("YES")) == 0)
    {
        *promisc = 1;
    }
    else if (ADK_STRNCMP(value, "", sizeof("")) == 0)
    {
        *promisc = 0;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid port promisc setting\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
 
    return ret;
}

adk_status_t
adk_cfg_port_rx_mq_mode_get(const char *value, uint32_t *rx_mq_mode)
{
    adk_status_t ret = ADK_SUCCESS;

    if (ADK_STRNCMP(value, "NONE", sizeof("NONE")) == 0)
    {
        *rx_mq_mode = ADK_CFG_PORT_RX_MQ_MODE_NONE;
    }
    else if (ADK_STRNCMP(value, "RSS", sizeof("RSS")) == 0)
    {
        *rx_mq_mode = ADK_CFG_PORT_RX_MQ_MODE_RSS;
    }
    else
    {
        ADK_PREINIT_WRITE("Invalid port rx multi-queue mode setting\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
 
    return ret;
}

adk_status_t
adk_cfg_port_util_list_get(const char *value,
                           uint16_t max_list, 
                           uint16_t *list, 
                           uint16_t *num_list)
{
    adk_status_t ret = ADK_SUCCESS;
    char tmp_buf[CFG_VALUE_LEN * 2], *p = NULL;
    uint32_t i = 0;

    ADK_MEMSET(tmp_buf, 0, sizeof(tmp_buf));
    ADK_STRNCPY(tmp_buf, value, CFG_VALUE_LEN);
    i = 0;
    p = strtok (tmp_buf, " ");
    while (p != NULL)
    {
        list[i++] = atoi(p);
        if (i > max_list)
        {
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
        p = strtok(NULL, " ");
    }
    *num_list = i;

    return ret;
}
 
adk_status_t
adk_cfg_port_util_socket_list_get(const char *value, 
                                  uint16_t max_list, 
                                  int32_t *list, 
                                  uint16_t *num_list)
{
    adk_status_t ret = ADK_SUCCESS;
    char tmp_buf[CFG_VALUE_LEN * 2], *p = NULL;
    uint32_t i = 0;

    ADK_MEMSET(tmp_buf, 0, sizeof(tmp_buf));
    ADK_STRNCPY(tmp_buf, value, CFG_VALUE_LEN);
    i = 0;
    p = strtok (tmp_buf, " ");
    while (p != NULL)
    {
        list[i++] = atoi(p);
        if (i > max_list)
        {
            return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
        }
        p = strtok(NULL, " ");
    }
    *num_list = i;

    return ret;
}

adk_status_t
adk_cfg_port_validate(adk_cfg_port_params_t *params,
                      adk_cfg_port_flags_t *flags)
{
    adk_status_t ret = ADK_SUCCESS;
    
    if (!flags->port)
    {
        ADK_PREINIT_WRITE("Section - %s:  port entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->speed)
    {
        ADK_PREINIT_WRITE("Section - %s:  speed entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->duplex)
    {
        ADK_PREINIT_WRITE("Section - %s:  duplex entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->promisc)
    {
        ADK_PREINIT_WRITE("Section - %s:  promisc entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->rx_mq_mode)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx multi queue mode entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->rx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx queues entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->rx_queue_mempool)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx queue mempool entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->rx_queue_size)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx queue size entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->rx_queue_weight)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx queue weight entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

#if 0
    /* Not supported. For performance consideration using system wide burst size */
    if (!flags->rx_queue_burst)
    {
        ADK_PREINIT_WRITE("Section - %s:  rx queue burst entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
#endif

    if (!flags->tx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  tx queues entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->tx_queue_socket)
    {
        ADK_PREINIT_WRITE("Section - %s:  tx queue socket entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->tx_queue_size)
    {
        ADK_PREINIT_WRITE("Section - %s:  tx queue size entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

#if 0
    /* Not supported. For performance consideration using system wide burst size */
    if (!flags->tx_queue_burst)
    {
        ADK_PREINIT_WRITE("Section - %s:  tx queue burst entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
#endif    

    if (flags->rx_queue_mempool != flags->rx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  num rx_queue_mempool (%d) & num of rx_queues (%d) mismatch\n", 
        params->name, flags->rx_queue_mempool, flags->rx_queues);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (flags->rx_queue_size != flags->rx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  num rx_queue_size (%d) & num of rx_queues (%d) mismatch\n", 
        params->name, flags->rx_queue_size, flags->rx_queues);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

#if 0
    /* Not supported. For performance consideration using system wide burst size */
    if (flags->rx_queue_burst != flags->rx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  num rx_queue_burst (%d) & num of rx_queues (%d) mismatch\n", 
        params->name, flags->rx_queue_burst, flags->rx_queues);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
#endif    

    if (flags->tx_queue_socket != flags->tx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  num tx_queue_socket (%d) & num of tx_queues (%d) mismatch\n", 
        params->name, flags->tx_queue_socket, flags->tx_queues);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (flags->tx_queue_size != flags->tx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  num tx_queue_size (%d) & num of tx_queues (%d) mismatch\n", 
        params->name, flags->tx_queue_size, flags->tx_queues);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

#if 0
    /* Not supported. For performance consideration using system wide burst size */
    if (flags->tx_queue_burst != flags->tx_queues)
    {
        ADK_PREINIT_WRITE("Section - %s:  num tx_queue_burst (%d) & num of tx_queues (%d) mismatch\n", 
        params->name, flags->tx_queue_burst, flags->tx_queues);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
#endif

    return ret;
}

adk_status_t
adk_cfg_port_defaults_set(adk_cfg_port_params_t *params __attribute__((unused)))
{
    adk_status_t ret = ADK_SUCCESS;
    return ret;
}


