/* ***************************************************************************
 **                                                                          *
 ** INTEL CONFIDENTIAL                                                       *
 ** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
 **                                                                          *
 ** The source code contained or described herein and all documents related  *
 ** to the source code ("Material") are owned by Intel Corporation or its    *
 ** suppliers or licensors. Title to the Material remains with Intel         *
 ** Corporation or its suppliers and licensors. The Material contains trade  *
 ** secrets and proprietary and confidential information of Intel or its     *
 ** suppliers and licensors. The Material is protected by worldwide          *
 ** copyright and trade secret laws and treaty provisions. No part of the    *
 ** Material may be used, copied, reproduced, modified, published, uploaded, *
 ** posted, transmitted, distributed, or disclosed in any way without        *
 ** Intel's prior express written permission.                                *
 **                                                                          *
 ** No license under any patent, copyright, trade secret or other            *
 ** intellectual property right is granted to or conferred upon you by       *
 ** disclosure or delivery of the Materials, either expressly, by            *
 ** implication, inducement, estoppel or otherwise. Any license under such   *
 ** intellectual property rights must be express and approved by Intel       *
 ** in writing.                                                              *
 **                                                                          *
 ** **************************************************************************
 **/
#include "adk_include.h"
#include "adk_sif_cfg_include.h"

/* Table to store ring section parameters */
static adk_cfg_ring_params_tbl_t ring_params_tbl_g;


adk_status_t
adk_cfg_ring_init(adk_cfg_params_t *cfg_params)
{
    adk_status_t ret = ADK_SUCCESS;

    cfg_params->ring_params_tbl = (void *)&ring_params_tbl_g;

    return ret;
}

adk_status_t
adk_cfg_ring_parse(adk_cfg_params_t *cfg_params_hdl,
                   const char *section_name)
{
    adk_status_t ret = ADK_SUCCESS;
    adk_cfg_ring_params_t *ring_params = NULL;
    int32_t section_entries_cnt = 0, idx = 0;
    struct rte_cfgfile_entry *section_entries = NULL, *entry = NULL;
    adk_cfg_ring_flags_t flags;
#ifdef ADK_DEBUG
    int32_t i = 0;
#endif

    ADK_MEMSET(&flags, 0, sizeof(adk_cfg_ring_flags_t));

    /* Allocate memory to store parsed parameters */
    ret = adk_cfg_ring_param_alloc(cfg_params_hdl->ring_params_tbl, &ring_params);
    if (ret != ADK_SUCCESS)
    {
        return ret;
    }
 
    section_entries_cnt = rte_cfgfile_section_num_entries(cfg_params_hdl->cfg_handle, section_name);
    ADK_PREINIT_WRITE("\nSection:%s entries = %d\n", section_name, section_entries_cnt);

    /* Get all entries of current sections */
    section_entries = calloc(section_entries_cnt, sizeof(struct rte_cfgfile_entry));
    if (section_entries == NULL)
    {
        ADK_PREINIT_WRITE("Memory allocation failure\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);
    }

    if (rte_cfgfile_section_entries(cfg_params_hdl->cfg_handle, section_name, 
                                    section_entries, section_entries_cnt) == 0)
    {
        ADK_PREINIT_WRITE("Get section entries failed\n");
        ret = ADK_STATUS_SEV_LEVEL_SET(ADK_E_CFG_PARSE_ERROR, ADK_SEV_LEVEL_ERROR);
        goto adk_cfg_ring_parse_return;
    }

#ifdef ADK_DEBUG
    ADK_PREINIT_WRITE("[%s]\n", section_name);
    for (i = 0; i < section_entries_cnt; i++)
    {
        ADK_PREINIT_WRITE("\t%s = %s\n", section_entries[i].name, section_entries[i].value);
    }
#endif

    ADK_MEMCPY(ring_params->name, section_name, CFG_NAME_LEN);
    if(ring_params->name[CFG_NAME_LEN-1] != '\0')
        ring_params->name[CFG_NAME_LEN-1] = '\0';

    /* Parse each entry of current section */
    for (idx = 0; idx < section_entries_cnt; idx++)
    {
        entry = &section_entries[idx];

        /* Skip commneted lines */ 
        if (ADK_STRNCMP(entry->name, "#", 1) == 0)
        {
            continue;
        }

        if (ADK_STRNCMP(entry->name, "enable", sizeof("enable")) == 0)
        {
            if (strcasecmp(entry->value, "false") == 0)
            {
                /* Skip disabled section */
                ADK_MEMSET(ring_params, 0, sizeof(adk_cfg_ring_params_t));
                ret = ADK_W_CFG_SECTION_NOT_ENABLED;
                goto adk_cfg_ring_parse_return;
            }
            else
            {
                continue;
            }
        }
        else if (ADK_STRNCMP(entry->name, "ring", sizeof("ring")) == 0)
        {
            ring_params->ring_id = atoi(entry->value); 
            flags.ring++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "ring_size", sizeof("ring_size")) == 0)
        {
            ring_params->ring_size = atoi(entry->value); 
            flags.ring_size++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "ring_enqueue_burst", sizeof("ring_enqueue_burst")) == 0)
        {
            ring_params->ring_enqueue_burst = atoi(entry->value); 
            flags.ring_enqueue_burst++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "ring_dequeue_burst", sizeof("ring_dequeue_burst")) == 0)
        {
            ring_params->ring_dequeue_burst = atoi(entry->value); 
            flags.ring_dequeue_burst++;
            continue;
        }
        else if (ADK_STRNCMP(entry->name, "socket", sizeof("socket")) == 0)
        {
            ring_params->socket = atoi(entry->value); 
            flags.socket++;
            continue;
        }
        else
        {
            flags.unknown++;
        }
    }
    
    /* Validation */
    ret = adk_cfg_ring_validate(ring_params, &flags);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_ring_parse_return;
    }

    /* Set defaults */
    ret = adk_cfg_ring_defaults_set(ring_params);
    if (ret != ADK_SUCCESS)
    {
        goto adk_cfg_ring_parse_return;
    }

    free(section_entries);
    return ret;

adk_cfg_ring_parse_return:
    free(section_entries);
    adk_cfg_ring_param_free(cfg_params_hdl->ring_params_tbl, ring_params);
    return ret;
}

adk_status_t
adk_cfg_ring_param_alloc(adk_cfg_ring_params_tbl_t *ring_params_hdl,
                         adk_cfg_ring_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;
    adk_cfg_ring_params_t *p = ring_params_hdl->ring_params;

    for (i = 0; i < ADK_SIF_CFG_MAX_RING_QUEUES; i++)
    {
        if (p[i].valid == _INIT)
        {
            p[i].valid = _CONFIG;
            *params = &p[i];
            ring_params_hdl->size++;
            return ADK_SUCCESS;
        }
    }

    if (i == ADK_SIF_CFG_MAX_RING_QUEUES)
    {
        ADK_PREINIT_WRITE("Number of rings exceeds max supported %d\n", ADK_SIF_CFG_MAX_RING_QUEUES);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);        
    }
    
    return ret;
}

adk_status_t
adk_cfg_ring_param_free(adk_cfg_ring_params_tbl_t *ring_params_hdl,
                        adk_cfg_ring_params_t *params)
{
    adk_status_t ret = ADK_SUCCESS;

    if (params == NULL)
    {
        ADK_PREINIT_WRITE("Invalid memory access\n");
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE, ADK_SEV_LEVEL_ERROR);        
    }

    params->valid = _INIT;
    ring_params_hdl->size--;
    
    return ret;
}

adk_status_t
adk_cfg_ring_param_rewind(adk_cfg_ring_params_tbl_t *ring_params_hdl)
{
    adk_status_t ret = ADK_SUCCESS;

    ring_params_hdl->cur = 0;
    
    return ret;
}

adk_status_t
adk_cfg_ring_param_get_next(adk_cfg_ring_params_tbl_t *ring_params_hdl,
                            adk_cfg_ring_params_t **params)
{
    adk_status_t ret = ADK_SUCCESS;
    uint32_t i = 0;

    /* Boundary check */
    if (ring_params_hdl->cur == ring_params_hdl->size)
    {
        ring_params_hdl->cur = 0;
        return ADK_E_LIST_RANGE;
    }  

    /* Skip deleted entries */ 
    for (i = ring_params_hdl->cur; i < ring_params_hdl->size; i++)
    {
        if (ring_params_hdl->ring_params[ring_params_hdl->cur].valid == _CONFIG)
        {
            break;
        }
    }
    ring_params_hdl->cur = i;

    /* Fetch config params */
    *params = &ring_params_hdl->ring_params[ring_params_hdl->cur];

    /* Point to next entry */
    ring_params_hdl->cur++;

    return ret;
}


adk_status_t
adk_cfg_ring_validate(adk_cfg_ring_params_t *params,
                      adk_cfg_ring_flags_t *flags)
{
    adk_status_t ret = ADK_SUCCESS;
    
    if (!flags->ring)
    {
        ADK_PREINIT_WRITE("Section - %s:  ring entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->ring_size)
    {
        ADK_PREINIT_WRITE("Section - %s:  ring size entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

#if 0
    /* Not supported. For performance consideration using system wide burst size */
    if (!flags->ring_enqueue_burst)
    {
        ADK_PREINIT_WRITE("Section - %s:  ring enqueue burst entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    if (!flags->ring_dequeue_burst)
    {
        ADK_PREINIT_WRITE("Section - %s:  ring dequeue burst entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }
#endif    

    if (!flags->socket)
    {
        ADK_PREINIT_WRITE("Section - %s:  socket entry not configured\n", params->name);
        return ADK_STATUS_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER, ADK_SEV_LEVEL_ERROR);
    }

    return ret;
}

adk_status_t
adk_cfg_ring_defaults_set(adk_cfg_ring_params_t *params __attribute__((unused)))
{
    adk_status_t ret = ADK_SUCCESS;
    return ret;
}


