/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

#include <ies_sdk.h>
#include <adk_include.h>
#include <adk_shm.h>
#include <adk_hos.h>
#include <adk_cp_acl_list.h>

#define ADK_CP_LIST_SHM_MEMZONE_NAME   "adk_acl_shm_zone"

#define ADK_ACL_EFFEC_PREC(node) (node->hos_precedence_mod ? node->hos_precedence_mod : node->hos_precedence)

/* default configuration */
#define ADK_ACL_MAX_NUM_OPS 512
#define ADK_ACL_DEFAULT_NODE_SPACING 50
#define ADK_ACL_MAX_NODE_PRECEDENCE 0xFFFF
#define ADK_ACL_LIST_MIN_NODE_SPACING 2

        
typedef struct adk_acl_pvt_data_s
{
    adk_sys_handle_t sys_handle;

    /* MAX Node precedence */
    uint32_t max_precedence;

    /* Ideal spacing between nodes */
    uint32_t node_spacing;

    /* ACL Handles */
    adk_acl_handle_t acl_hdls[ADK_ACL_MAX];
} adk_acl_pvt_data_t;

adk_acl_pvt_data_t *adk_acl_pvt_g_p;

typedef enum adk_acl_op_type_e {
    ADK_ACL_ADD = 0,
    ADK_ACL_DEL,
    ADK_ACL_MODIFY,
    ADK_ACL_NONE,
} adk_acl_op_type_t;

typedef struct adk_acl_entry_s adk_acl_entry_t;

typedef struct adk_acl_entry_s {
    adk_acl_entry_t *next;
    adk_acl_entry_t *prev;
    adk_acl_op_type_t op_type;
    bool hos_committed;
    uint16_t index;
    adk_acl_rule_t rule;
    adk_acl_rule_t modify_rule;
    uint32_t hos_precedence;
    uint32_t hos_precedence_mod;
    adk_hos_rule_id_t hos_rule_id[ADK_ACL_MAX_HOS_RULES_PER_RULE];
    /*Bkp to be used for modify operation*/
    uint32_t bkp_stats;
    adk_hos_rule_id_t bkp_hos_rule_id[ADK_ACL_MAX_HOS_RULES_PER_RULE];
    uint16_t bkp_num_hos_rules;
} adk_acl_entry_t;

typedef struct adk_acl_s {
    adk_acl_entry_t *head;
    adk_acl_entry_t *get;
    adk_acl_info_t info;
    uint16_t num_ops;
    uint32_t num_nodes;

    /* This is just to track if second iteration is required during commit.
        If any rule deletion or modification is done, only
        then second iteration is required */
    bool secondIterReq;
} adk_acl_t;

/* This data structure keeps track of node range
    availability while traversing node */
typedef struct adk_acl_space_data_s {
    adk_acl_entry_t* node;
    uint32_t num_nodes;
} adk_acl_space_data_t;



/* Should be called only once for the initial CP LIST configuration. This can be
 * done from any one process preferably, the one performing the ADK system
 * initialization at the beginning. */

adk_status_t
adk_acl_init (
    adk_sys_handle_t sys_handle)
{
    adk_status_t retValue = ADK_SUCCESS;

    /* Create memzone to store all the HOS private data */
    retValue = adk_shm_create(ADK_CP_LIST_SHM_MEMZONE_NAME,
                              sizeof(adk_acl_pvt_data_t),
                              ADK_SHM_FIRST_WORKER_IN_USE_LCORE,
                              (adk_shm_handle_t **) &adk_acl_pvt_g_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_LOG_ERROR(sys_handle, ADK_SYS_MODULE_HANDLE_GET(sys_handle),
                      "Failed to create CP LIST memzone %s\n",
                      ADK_CP_LIST_SHM_MEMZONE_NAME);
        return retValue;
    }

    adk_acl_pvt_g_p->sys_handle = sys_handle;
    adk_acl_pvt_g_p->max_precedence = ADK_ACL_MAX_NODE_PRECEDENCE;
    adk_acl_pvt_g_p->node_spacing = ADK_ACL_DEFAULT_NODE_SPACING;
    return retValue;
}

/* This API should be called from one process to remove all the default
 * configurations done at the CP LIST initialization time. Preferably, from
 * the process performing ADK sys-shutdown. */

adk_status_t
adk_acl_shutdown (
    void)
{
    /* Destroy the private memzone created during init time */
    adk_shm_destroy(ADK_CP_LIST_SHM_MEMZONE_NAME);
    return ADK_SUCCESS;
}

/* Should be called from each secondary application process which requires
 * ADK ACL API services. This should not be called from the process which performs
 * the ADK sys-init. */

adk_status_t
adk_acl_attach (
    void)
{
    adk_status_t retValue = ADK_SUCCESS;
    const struct rte_memzone *mz = NULL;

    /* Get access to the HOS private data */
    ADK_FORCE_LOG_EXT_API("rte_memzone_lookup", "Lookup for a memzone.\n");
    mz = rte_memzone_lookup(ADK_CP_LIST_SHM_MEMZONE_NAME);
    if (mz == NULL)
    {
        ADK_FORCE_LOG_ERROR("Failed to retrieve ADK ACL private data\n");
        retValue = ADK_E_MEM_UNAVAILABLE;
        goto _adk_acl_attach_exit;
    }

    adk_acl_pvt_g_p = (adk_acl_pvt_data_t*) mz->addr;

_adk_acl_attach_exit:

    return retValue;
}

void
adk_acl_set_default_spacing (
    uint32_t spacing)
{
    adk_acl_pvt_g_p->node_spacing = spacing; 
}

uint32_t
adk_acl_get_default_spacing (
    void)
{
    return adk_acl_pvt_g_p->node_spacing; 
}

void
adk_acl_reset_default_spacing (
    void)
{
    adk_acl_pvt_g_p->node_spacing = ADK_ACL_DEFAULT_NODE_SPACING; 
}

void
adk_acl_set_max_precedence (
    uint32_t max_precedence)
{
    adk_acl_pvt_g_p->max_precedence = max_precedence; 
}

uint32_t
adk_acl_get_max_precedence (
    void)
{
    return adk_acl_pvt_g_p->max_precedence; 
}

void
adk_acl_reset_max_precedence (
    void)
{
    adk_acl_pvt_g_p->max_precedence = ADK_ACL_MAX_NODE_PRECEDENCE; 
}

adk_status_t
adk_acl_rule_get_hos_committed(adk_acl_handle_t acl_hdl,
                        uint32_t index,
                        bool* hos_committed)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL;
    uint32_t i;

    if(!acl_hdl)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ADK ACL Error: Failed to get hos committed flag from ACL List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    /* Update the main CP list */
    cur = list->head;
    i=0;
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            /* Already deleted entry, skip to next */
            cur = cur->next;
            continue;
        }

        if(i == index)
        {
            *hos_committed = cur->hos_committed;
            return retValue;
        }

        ++i;
        cur = cur->next;
    }

    /*Not found*/
    retValue = ADK_E_INVALID_PARAMETER;
    return retValue;
}

adk_status_t
adk_acl_set_num_ops_allowed(adk_acl_handle_t acl_hdl,
                        uint32_t num_ops_allowed)
{
    adk_acl_t *list = (adk_acl_t *)acl_hdl;

    if(!list)
        return ADK_E_INVALID_PARAMETER;

    list->info.num_ops_allowed = num_ops_allowed;
    return ADK_SUCCESS;
}

adk_status_t
adk_acl_get_num_ops_allowed(adk_acl_handle_t acl_hdl,
                        uint32_t* num_ops_allowed)
{
    adk_acl_t *list = (adk_acl_t *)acl_hdl;

    if(!list)
        return ADK_E_INVALID_PARAMETER;

    *num_ops_allowed = list->info.num_ops_allowed;
    return ADK_SUCCESS;
}

adk_status_t
adk_acl_reset_num_ops_allowed(adk_acl_handle_t acl_hdl)
{
    adk_acl_t *list = (adk_acl_t *)acl_hdl;

    if(!list)
        return ADK_E_INVALID_PARAMETER;

    list->info.num_ops_allowed = ADK_ACL_MAX_NUM_OPS;
    return ADK_SUCCESS;
}

adk_status_t
adk_acl_create(adk_acl_handle_t *acl_hdl,
                const adk_acl_info_t *list_info)
{
    adk_acl_t *list;

    if(!acl_hdl || !list_info)
    {
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to create CP List\n");
        return ADK_E_INVALID_PARAMETER;
    }

    if(list_info->user > ADK_ACL_MAX)
    {
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to create CP List, "
                      "Invalid user %d\n", list_info->user);
        return ADK_E_INVALID_PARAMETER;
    }

    list = (adk_acl_t *)ADK_NVM_CALLOC(ADK_MISC_MODULE_ID, 0, 1, sizeof(adk_acl_t));
    if(!list)
    {
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to allocate memory\n");
        return ADK_E_MEM_UNAVAILABLE;
    }

    ADK_MEMCPY(&(list->info), list_info, sizeof(adk_acl_info_t));
    if(!list->info.num_ops_allowed)
        list->info.num_ops_allowed = ADK_ACL_MAX_NUM_OPS; 
    if(!list->info.num_hw_rules_per_cpu_rule)
        list->info.num_hw_rules_per_cpu_rule = ADK_ACL_MAX_HOS_RULES_PER_RULE; 

    *acl_hdl = (adk_acl_handle_t)list;
    adk_acl_pvt_g_p->acl_hdls[list_info->user] = *acl_hdl;
    return ADK_SUCCESS;
}

adk_status_t
adk_acl_free(adk_acl_handle_t acl_hdl)
{
    adk_acl_t *list = (adk_acl_t *)acl_hdl;

    if(!list)
        return ADK_E_INVALID_PARAMETER;

    adk_acl_pvt_g_p->acl_hdls[list->info.user] = NULL;
    ADK_NVM_FREE(list);
    return ADK_SUCCESS;
}

adk_acl_handle_t
adk_acl_hdl_get(adk_acl_user_t list_user)
{
   if(list_user >= ADK_ACL_MAX)
        return NULL;

    return adk_acl_pvt_g_p->acl_hdls[list_user]; 
}

adk_status_t
adk_acl_clear(adk_acl_handle_t acl_hdl)
{
    adk_acl_t *list = (adk_acl_t *)acl_hdl;

    if(!list)
        return ADK_E_INVALID_PARAMETER;

    adk_acl_entry_t* cur = list->head;
    while(cur)
    {
        adk_acl_entry_t* temp = cur;
        cur = cur->next;
        ADK_NVM_FREE(temp);        
    }

    list->head = NULL;
    list->num_ops = 0;
    list->num_nodes = 0;
    list->secondIterReq = false;
    return ADK_SUCCESS;
}

adk_status_t
adk_acl_get_num_nodes(adk_acl_handle_t acl_hdl,
                        uint32_t* num_nodes)
{
    adk_acl_t *list = (adk_acl_t *)acl_hdl;

    if(!list)
        return ADK_E_INVALID_PARAMETER;

    *num_nodes = list->num_nodes;
    return ADK_SUCCESS;
}

static adk_status_t
adk_acl_create_node(adk_acl_entry_t** node,
                    const adk_acl_rule_t *rule)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_entry_t* newNode = (adk_acl_entry_t *)ADK_NVM_CALLOC(
                            ADK_MISC_MODULE_ID, 0, 1, sizeof(adk_acl_entry_t));
    if(!newNode)
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to create node, "
                      "memory allocation failed\n");
        return retValue;
    }

    ADK_MEMCPY(&(newNode->rule), rule, sizeof(adk_acl_rule_t));
    newNode->op_type = ADK_ACL_ADD;
    *node = newNode;
    return ADK_SUCCESS;
}

/* This is an alternative algorithm for finding range
    while traversing forward to find ith node */
#ifdef ADK_HOS_FINE_TUNE_ALGO
/* The range is available between node_space_data->node and 
    node_space_data->node->prev */
static void
adk_acl_fill_node_space_data(adk_acl_space_data_t* node_space_data,
                                adk_acl_entry_t* cur, uint32_t index)
{
    printf("*********adk_acl_fill_node_space_data called for index %d*******\n", index);
    /* This is the head node */
    if(!cur->prev)
    {
        printf("adk_acl_fill_node_space_data head node\n");
        int prevRange = adk_acl_pvt_g_p->max_precedence - ADK_ACL_EFFEC_PREC(cur);
        /* Left shift is required for two min node range on both
            sides of new inserted node */
        printf("adk_acl_fill_node_space_data head node %d\n", prevRange);
        if(prevRange > (int)(adk_acl_pvt_g_p->node_spacing*2))
        {
            node_space_data->node = cur;
            node_space_data->num_nodes = 1;
            printf("adk_acl_fill_node_space_data head node node space data filled node index %d\n", index);
        }
        return;
    }

    int32_t prevRangeAvailable = 0;
    if(node_space_data->node)
    {
        printf("adk_acl_fill_node_space_data prevRangeAvailable logic enters with num_nodes= %d\n", node_space_data->num_nodes);
        uint32_t prevPrevPrecedence = node_space_data->node->prev ?
                                ADK_ACL_EFFEC_PREC(node_space_data->node->prev) : adk_acl_pvt_g_p->max_precedence;
        uint32_t prevRange = prevPrevPrecedence - ADK_ACL_EFFEC_PREC(cur);

        printf("adk_acl_fill_node_space_data prevPrevPrecedence= %d, prevRange= %d\n", prevPrevPrecedence, prevRange);
        node_space_data->num_nodes++;

        /* Addition of 2 is required for two min node range on both
            sides of new inserted node.
            NOTE: ADK_ACL_DEFAULT_NODE_SPACING is used for range availability instead of 
            adk_acl_pvt_g_p->node_spacing to enable  possibility of accumulation of ranges
            while traversing forward */
        prevRangeAvailable = prevRange - ((node_space_data->num_nodes+2)*ADK_ACL_DEFAULT_NODE_SPACING);
        printf("adk_acl_fill_node_space_data prevRangeAvailable= %d\n", prevRangeAvailable);
        /* Is the last range still available? */
        if(prevRangeAvailable < 0)
        {
            printf("adk_acl_fill_node_space_data prevRangeAvailable negative reset \n");
            node_space_data->node = NULL;
            node_space_data->num_nodes = prevRangeAvailable = 0;
        }
    }

    int curRange = ADK_ACL_EFFEC_PREC(cur->prev)
                            - ADK_ACL_EFFEC_PREC(cur);

    printf("adk_acl_fill_node_space_data CurRange= %d\n", curRange);
    if(curRange <= (int)(adk_acl_pvt_g_p->node_spacing*3))
        return;

    /* Already a node with range was found,
        compare to find who wins */
    if(curRange < prevRangeAvailable)
        return;

    printf("adk_acl_fill_node_space_data curRange wins\n");
    node_space_data->node = cur;
    node_space_data->num_nodes = 1;
}
#endif

/* Node Index returned here is actual node index counting deleted nodes as well */
static adk_status_t
adk_acl_get_node_at(adk_acl_entry_t** cur,
                    adk_acl_entry_t** prev,
                    const adk_acl_t* list,
                    uint32_t index,
                    uint32_t* nodeIndex)
{
    adk_status_t retValue = ADK_SUCCESS;
    bool append = (index == UINT16_MAX);
    *cur = list->head;
    for(uint32_t i=0; i < index; )
    {
        if(*cur == NULL)
        {
            if(append)
                break;
            
            retValue = ADK_E_INVALID_PARAMETER;
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "CP List Error: Failed to find node in CP List, "
                          "invalid index %d\n", index);
            return retValue;
        }

        if((*cur)->op_type != ADK_ACL_DEL)
            ++i;

        *prev = *cur;
        *cur = (*cur)->next;
        ++(*nodeIndex);
    }

    return retValue;
}

/* The range is available between the node returned by this function and 
    the prev node */
static void
adk_acl_get_left_node_space_data(adk_acl_space_data_t* left_space_data,
                                    adk_acl_entry_t* node, uint32_t node_space)
{
    adk_acl_entry_t *cur = node;
   
    /*Looks like a prepend at the beginning */
    if(!node)
        return;

    cur = node;
    while(cur)
    {
        uint32_t prevPrec = cur->prev ? ADK_ACL_EFFEC_PREC(cur->prev) : adk_acl_pvt_g_p->max_precedence;
        int32_t curPrecRange = prevPrec - ADK_ACL_EFFEC_PREC(node);
        left_space_data->num_nodes++;
        int32_t min_space_req = (left_space_data->num_nodes+2)*node_space;
        if(curPrecRange > min_space_req)
        {
            left_space_data->node = cur;
            return;
        }

        cur = cur->prev;
    }

    return;
}

/* The range is available between the node returned by this function and 
    the next node */
static void
adk_acl_get_right_node_space_data(adk_acl_space_data_t* right_space_data,
                                    adk_acl_entry_t* node, uint32_t node_space)
{
    adk_acl_entry_t *cur = node;
   
    /*Looks like a append at the end */
    if(!node)
        return;

    cur = node;
    while(cur)
    {
        uint32_t nextPrec = cur->next ? ADK_ACL_EFFEC_PREC(cur->next) : 0;
        int32_t curPrecRange = ADK_ACL_EFFEC_PREC(node) - nextPrec;
        right_space_data->num_nodes++;
        int32_t min_space_req = (right_space_data->num_nodes+2)*node_space;
        if(curPrecRange > min_space_req)
        {
            right_space_data->node = cur;
            return;
        }

        cur = cur->next;
    }

    return; 
}

static void 
adk_acl_distribute_range_left(adk_acl_entry_t* node,
                            adk_acl_space_data_t* node_space_data,
                            int32_t range)
{
    uint32_t basePrec = node_space_data->node->prev ?
                            ADK_ACL_EFFEC_PREC(node_space_data->node->prev) : adk_acl_pvt_g_p->max_precedence;
    adk_acl_entry_t* cur = node_space_data->node;

    while(cur != node)
    {
        cur->hos_precedence_mod = (basePrec - range) | 0x1;
        basePrec = (basePrec - range);
        cur = cur->next;
    }
    cur->hos_precedence_mod = (basePrec - range) | 0x1;
}

static void
adk_acl_distribute_range_right(adk_acl_entry_t* node,
                            adk_acl_space_data_t* node_space_data,
                            int32_t range)
{
    uint32_t basePrec = node_space_data->node->next ?
                            ADK_ACL_EFFEC_PREC(node_space_data->node->next) : 0;
    adk_acl_entry_t* cur = node_space_data->node;

    while(cur != node)
    {
        cur->hos_precedence_mod = (basePrec + range) | 0x1;
        basePrec = basePrec + range;
        cur = cur->prev;
    }
    cur->hos_precedence_mod = (basePrec + range) | 0x1;
}

static void
adk_acl_find_preferrd_node_space(adk_acl_t* list,
                                    adk_acl_entry_t* node,
                                    adk_acl_space_data_t* left_space_data,
                                    adk_acl_space_data_t* right_space_data,
                                    uint32_t nodeIndex)
{
    uint32_t avgLeftDist = 0, avgRightDist = 0;

    if(node->prev)
    {
        avgLeftDist = (adk_acl_pvt_g_p->max_precedence - ADK_ACL_EFFEC_PREC(node->prev)) / nodeIndex;
        avgRightDist = ADK_ACL_EFFEC_PREC(node->prev) / (list->num_nodes - nodeIndex);

    }
    else
    {
        /* it's insertion at head */
        avgLeftDist = (adk_acl_pvt_g_p->max_precedence - ADK_ACL_EFFEC_PREC(node->next));
        avgRightDist = ADK_ACL_EFFEC_PREC(node->next) / (list->num_nodes - nodeIndex);
    }

    /* Range likely available on left side, try that first */
    if(avgLeftDist > avgRightDist)
    {
        /* Find left space data */
        adk_acl_get_left_node_space_data(left_space_data, node->prev, adk_acl_pvt_g_p->node_spacing);
        if(NULL != left_space_data->node)
            return;

        right_space_data->num_nodes = 0;
        /* Find right space data */
        adk_acl_get_right_node_space_data(right_space_data, node->next, adk_acl_pvt_g_p->node_spacing);
        if(NULL != right_space_data->node)
            return;

        /* If no range available on either side with calculated node spacing, try with default node spacing */
        left_space_data->num_nodes = 0;
        adk_acl_get_left_node_space_data(left_space_data, node->prev, ADK_ACL_DEFAULT_NODE_SPACING);
        if(NULL != left_space_data->node)
            return;

        right_space_data->num_nodes = 0;
        adk_acl_get_right_node_space_data(right_space_data, node->next, ADK_ACL_DEFAULT_NODE_SPACING);
    }
    else
    {
        /* Find right space data */
        adk_acl_get_right_node_space_data(right_space_data, node->next, adk_acl_pvt_g_p->node_spacing);
        if(NULL != right_space_data->node)
            return;

        left_space_data->num_nodes = 0;
        /* Find left space data */
        adk_acl_get_left_node_space_data(left_space_data, node->prev, adk_acl_pvt_g_p->node_spacing);
        if(NULL != left_space_data->node)
            return;

        /* If no range available on either side with calculated node spacing, try with default node spacing */
        right_space_data->num_nodes = 0;
        adk_acl_get_right_node_space_data(right_space_data, node->next, ADK_ACL_DEFAULT_NODE_SPACING);
        if(NULL != right_space_data->node)
            return;

        left_space_data->num_nodes = 0;
        adk_acl_get_left_node_space_data(left_space_data, node->prev, ADK_ACL_DEFAULT_NODE_SPACING);
    }
}

/* Node index here includes deleted nodes as well */
static adk_status_t
adk_acl_assign_precedence(adk_acl_t* list,
                            adk_acl_entry_t* node,
                            uint32_t nodeIndex)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_space_data_t right_space_data, left_space_data;
    int32_t leftRange = 0, rightRange = 0;

    /* First Node, assign first default precedence */
    if(!node->prev && !node->next)
    {
        node->hos_precedence_mod = (adk_acl_pvt_g_p->max_precedence >> 1) | 0x1;
        return retValue;
    }
    /* has room for precedence? simply assign and return */
    else
    {
        uint32_t prev_precedence = node->prev ? ADK_ACL_EFFEC_PREC(node->prev) : adk_acl_pvt_g_p->max_precedence;
        uint32_t next_precedence = node->next ? ADK_ACL_EFFEC_PREC(node->next) : 0;
        int32_t curRange = prev_precedence - next_precedence;
        if(curRange >= (ADK_ACL_LIST_MIN_NODE_SPACING << 1))
        {
            node->hos_precedence_mod = ((prev_precedence + next_precedence) >> 1) | 0x1;
            return retValue;
        }
    }

    memset(&left_space_data, 0, sizeof(left_space_data));
    memset(&right_space_data, 0, sizeof(right_space_data));

    adk_acl_find_preferrd_node_space(list, node, &left_space_data, &right_space_data, nodeIndex);

    if(NULL == left_space_data.node && NULL == right_space_data.node)
    {
        retValue = ADK_E_CP_NO_RANGE_AVAILABLE;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to add rule to CP List, "
                      "No HLP precedence range is available\n");
        return retValue;
    }

    if(left_space_data.node)
    {
        uint32_t prevLeftPrec = left_space_data.node->prev ?
                                ADK_ACL_EFFEC_PREC(left_space_data.node->prev) : adk_acl_pvt_g_p->max_precedence;
        /* Node range should include all IDs between node space data node to new node inserted */
        leftRange = (prevLeftPrec - ADK_ACL_EFFEC_PREC(node->prev))/
                            (left_space_data.num_nodes + 2);
        adk_acl_distribute_range_left(node, &left_space_data, leftRange);
        return retValue;
    }

    if(right_space_data.node)
    {
        uint32_t nextRightPrec = right_space_data.node->next ?
                                    ADK_ACL_EFFEC_PREC(right_space_data.node->next) : 0;
        /* Node range should include all IDs between node space data node to new node inserted */
        rightRange = (ADK_ACL_EFFEC_PREC(node->next) - nextRightPrec)/
                            (right_space_data.num_nodes + 2);
        adk_acl_distribute_range_right(node, &right_space_data, rightRange);
        return retValue;
    }

    return retValue;
}

adk_status_t
adk_acl_rule_add(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    const adk_acl_rule_t *rule)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL, *newNode=NULL, *prev=NULL;
    uint32_t nodeIndex = 0;

    if(!acl_hdl || !rule)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to add rule to CP List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    if(list->num_ops > list->info.num_ops_allowed)
    {
        retValue = ADK_E_CP_MAX_OPERATIONS;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to add rule to CP List, "
                      "exceeds max operations allowed on CP LIST\n");
        return retValue;
    }

    /* Adjust node spacing depending on availability */
    if(list->num_nodes)
    {
        adk_acl_pvt_g_p->node_spacing = (ADK_ACL_MAX_NODE_PRECEDENCE / list->num_nodes) >> 1;
        if(adk_acl_pvt_g_p->node_spacing < ADK_ACL_DEFAULT_NODE_SPACING)
            adk_acl_pvt_g_p->node_spacing = ADK_ACL_DEFAULT_NODE_SPACING;
    }

    if(ADK_SUCCESS != (retValue = adk_acl_get_node_at(&cur, &prev,
                        list, index, &nodeIndex)))
        return retValue;

    if(ADK_SUCCESS != (retValue = adk_acl_create_node(&newNode, rule)))
        return retValue;

    list->num_nodes++;
    /*Add the entry at this pos*/
    newNode->next = cur;
    newNode->prev = prev;
    if(prev)
        prev->next = newNode;
    else
        list->head = newNode;
    if(cur)
        cur->prev = newNode;

    if(ADK_SUCCESS != (retValue = adk_acl_assign_precedence(list, newNode, nodeIndex)))
        return retValue;

    list->num_ops++;
    return retValue;
}

adk_status_t
adk_acl_rule_delete(adk_acl_handle_t acl_hdl,
                    uint16_t index)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL;
    int i;

    if(!acl_hdl)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to delete rule from CP List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    if(list->num_ops > list->info.num_ops_allowed)
    {
        retValue = ADK_E_CP_MAX_OPERATIONS;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to delete rule from CP List, "
                      "exceeds max operations allowed on CP LIST\n");
        return retValue;
    }

    /* Update the main CP list */
    cur = list->head;
    i=0;
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            /* Already deleted entry, skip to next */
            cur = cur->next;
            continue;
        }

        if(i != index)
        {
            ++i;
            cur = cur->next;
            continue;
        }

        /* Mark as to be deleted,
         * will be removed from the list during the commit */
        cur->op_type = ADK_ACL_DEL;

        list->num_ops++;
        list->secondIterReq = true;
        return retValue;
    }
    
    /*Not found*/
    retValue = ADK_E_INVALID_PARAMETER;
    return retValue;
}

adk_status_t
adk_acl_rule_modify(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    const adk_acl_rule_t *rule)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL;
    int i;

    if(!acl_hdl || !rule)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to modify rule in CP List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    if(list->num_ops > list->info.num_ops_allowed)
    {
        retValue = ADK_E_CP_MAX_OPERATIONS;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to modify rule in CP List, "
                      "exceeds max operations allowed on CP LIST\n");
        return retValue;
    }

    /* Update the main CP list */
    cur = list->head;
    i=0;
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            /* Already deleted entry, skip to next */
            cur = cur->next;
            continue;
        }

        if(i == index)
        {
            /*Mark as to be modified, will be updated during the commit*/
            cur->op_type = ADK_ACL_MODIFY;

            ADK_MEMCPY(&(cur->modify_rule), rule,
                       sizeof(adk_acl_rule_t));
            
            list->num_ops++;
            list->secondIterReq = true;
            return retValue;
        }

        ++i;
        cur = cur->next;
    }

    /*Not found*/
    retValue = ADK_E_INVALID_PARAMETER;
    return retValue;
}

adk_status_t
adk_acl_rule_hos_precedences_get(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    uint32_t* hos_precedence,
                    uint32_t* hos_precedence_mod)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL;
    int i;

    if(!acl_hdl)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ADK ACL List Error: Failed to get precedences of a rule from ACL List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    /* Update the main CP list */
    cur = list->head;
    i=0;
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            /* Already deleted entry, skip to next */
            cur = cur->next;
            continue;
        }

        if(i == index)
        {
            *hos_precedence = cur->hos_precedence;
            *hos_precedence_mod = cur->hos_precedence_mod;         
            return retValue;
        }

        ++i;
        cur = cur->next;
    }

    /*Not found*/
    retValue = ADK_E_INVALID_PARAMETER;
    return retValue;
}

adk_status_t
adk_acl_rule_hos_precedences_set(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    uint32_t hos_precedence_mod)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL;
    int i;

    if(!acl_hdl)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ADK ACL List Error: Failed to set precedence of a rule in ACL List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    /* Update the main CP list */
    cur = list->head;
    i=0;
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            /* Already deleted entry, skip to next */
            cur = cur->next;
            continue;
        }

        if(i == index)
        {
            /*Mark as to be modified, will be updated during the commit*/
            cur->hos_precedence_mod = hos_precedence_mod;
            return retValue;
        }

        ++i;
        cur = cur->next;
    }

    /*Not found*/
    retValue = ADK_E_INVALID_PARAMETER;
    return retValue;
}

adk_status_t
adk_acl_get_reset(adk_acl_handle_t acl_hdl)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;

    if(!acl_hdl)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to reset CP List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    list->get = list->head;

    return retValue;
}

adk_status_t
adk_acl_next_rule_get(adk_acl_handle_t acl_hdl,
                              adk_acl_rule_t *rule)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;

    if(!acl_hdl || !rule)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to get next rule in CP List, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;
    while(list->get)
    {
        if(list->get->op_type == ADK_ACL_MODIFY)
        {
            *rule = list->get->modify_rule;
            list->get = list->get->next;
            return retValue;
        }
        if(list->get->op_type != ADK_ACL_DEL)
        {
            *rule = list->get->rule;
            list->get = list->get->next;
            return retValue;
        }
        
        list->get = list->get->next;
    }

    retValue = ADK_E_LIST_NOT_FOUND;
    return retValue;
}

static adk_status_t
adk_acl_rules_state_alloc(uint32_t** precedencesP,
                            enum ies_acl_entry_state** entryStatesP,
                            const adk_acl_t* list)
{
    adk_status_t retValue = ADK_SUCCESS;
    *entryStatesP = (enum ies_acl_entry_state *)ADK_NVM_CALLOC(ADK_MISC_MODULE_ID,
                        0, list->info.num_ops_allowed, sizeof(enum ies_acl_entry_state));
    if(!*entryStatesP)
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to entry states array, "
                      "memory allocation failed\n");
        return retValue;
    }

    *precedencesP = (uint32_t *)ADK_NVM_CALLOC(ADK_MISC_MODULE_ID,
                        0, list->info.num_ops_allowed, sizeof(uint32_t));
    if(!*precedencesP)
    {
        retValue = ADK_E_MEM_UNAVAILABLE;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "CP List Error: Failed to precedence array, "
                      "memory allocation failed\n");
        return retValue;
    }
    
    return retValue;
}

static void
adk_acl_rules_state_free(uint32_t* precedences,
                            enum ies_acl_entry_state* entryStates)
{
    if(NULL != precedences)
        ADK_NVM_FREE(precedences);

    if(NULL != entryStates)
        ADK_NVM_FREE(entryStates);
}

static adk_status_t
adk_acl_get_hos_table(adk_hos_object_t* hos_obj,
                        adk_acl_user_t user)
{
    if(user == ADK_ACL_IPV4_ING_CLASSIFIER)
        *hos_obj = ADK_HOS_IP4_CLASSIFIER_TBL;
    else if(user == ADK_ACL_IPV6_ING_CLASSIFIER)
        *hos_obj = ADK_HOS_IP6_CLASSIFIER_TBL;
    else if(user == ADK_ACL_SPD_IN)
        *hos_obj = ADK_HOS_SPD_IN_TBL;
    else if(user == ADK_ACL_SPD_OUT)
        *hos_obj = ADK_HOS_SPD_OUT_TBL;
    else
        return ADK_E_INVALID_PARAMETER;

    return ADK_SUCCESS;
}

static void
adk_acl_find_potential_candidates_offload(adk_acl_entry_t** offload_node,
                                            adk_hos_object_t hos_obj,
                                            const adk_acl_t* list)
{
    adk_acl_entry_t *cur = NULL, *prev = NULL;
    uint32_t num_hos_rules = 0;
    uint32_t tbl_size;

    cur = list->head;
    adk_hos_table_size_get(hos_obj, &tbl_size);

    while(cur)
    {
        /* For atomic enable/disable operation, the rules scheduled to be
            deleted needs to be in HLP along with new rules added or modified
            or unmodified rules. Only half of slice is used.
            So it can be assuemd that slice will have enough space to keep the
            scheduled deleted rules */            
        if(cur->op_type != ADK_ACL_DEL)
            num_hos_rules +=  cur->rule.num_hos_rules;

        /* For modify rule, the hos rules for modified rules needs to be
            accounted since for atomic operation, the origianl and modified
            HOS rules need to be in HLP */
        if(cur->op_type == ADK_ACL_MODIFY)
            num_hos_rules += cur->modify_rule.num_hos_rules;

        if(num_hos_rules > tbl_size)
        {
            *offload_node = prev;
            return;
        }
        prev = cur;
        cur = cur->next; 
    }

    *offload_node = prev;    
    return;
}

static adk_status_t
adk_acl_delete_overflow_hos_rules(adk_acl_t* list,
                                    adk_acl_entry_t* offload_node,
                                    adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_entry_t* cur = offload_node ? offload_node->next : list->head;
    adk_acl_entry_t* prev = NULL;

    /* First find the right most node committed to HOS and
        delete HOS rules from right to left */
    while(cur)
    {
        prev = cur;
        cur = cur->next;
    }

    cur = prev;
    if(!cur)
        return retValue;

    while(cur != offload_node)
    {
        if(!cur->hos_committed)
        {
            cur = cur->prev;
            continue;
        }

        /*Delete the rules*/
        for(uint32_t iRule = 0; iRule < cur->rule.num_hos_rules; ++iRule)
        {
            if(cur->op_type == ADK_ACL_DEL && NULL != cur->rule.hos_rule_attr[iRule])
            {
                ADK_NVM_FREE(cur->rule.hos_rule_attr[iRule]);
                cur->rule.hos_rule_attr[iRule] = NULL;
            }
            retValue = adk_hos_rule_delete(hos_obj,
                             cur->hos_rule_id[iRule]);
            if(retValue != ADK_SUCCESS)
            {
                ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                              ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                              "ACL List Error: Failed commit ACL List to HOS, "
                              "cant't delete HOS rule\n");
                return retValue;
            }
        }
        if(cur->op_type != ADK_ACL_DEL)
            cur->op_type = ADK_ACL_ADD;
        cur->hos_committed = false;
        if(!cur->hos_precedence_mod)
            cur->hos_precedence_mod = cur->hos_precedence;
        cur->hos_precedence = 0;
        cur = cur->prev;
    }

    return retValue;
}

static adk_status_t
acl_rule_hos_add_internal(adk_acl_entry_t* cur,
                            adk_hos_object_t hos_obj,
                            bool is_disabled)
{
    adk_status_t retValue = ADK_SUCCESS;

    for(uint32_t iRule = 0; iRule < cur->rule.num_hos_rules ; ++iRule)
    {
        if((hos_obj == ADK_HOS_IP4_CLASSIFIER_TBL) ||
           (hos_obj == ADK_HOS_IP6_CLASSIFIER_TBL))
        {
            adk_hos_ip_classifier_tbl_rule_attr_t *classifier_rule =
                (adk_hos_ip_classifier_tbl_rule_attr_t *)
                (cur->rule.hos_rule_attr[iRule]);
            /*Keep the rule in disabled state*/
            classifier_rule->is_disabled = is_disabled; 
            classifier_rule->precedence = ADK_ACL_EFFEC_PREC(cur);
        }
        else
        {
            /*ADK_ACL_SPD*/
            adk_hos_spd_tbl_rule_attr_t *spd_rule =
                (adk_hos_spd_tbl_rule_attr_t *)
                (cur->rule.hos_rule_attr[iRule]);
            /*Keep the rule in disabled state*/
            spd_rule->is_disabled = is_disabled;
            spd_rule->precedence = ADK_ACL_EFFEC_PREC(cur);
        }
        retValue = adk_hos_rule_add(hos_obj,
                         cur->rule.hos_rule_attr[iRule],
                         &(cur->hos_rule_id[iRule]));
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: Failed to add rule to HOS\n");
            return retValue;
        }
    }

    return retValue;
}

static adk_status_t
adk_acl_rule_hos_add(adk_acl_entry_t* cur,
                        uint32_t* num_prec,
                        uint32_t* precedences,
                        enum ies_acl_entry_state* entryStates,
                        adk_hos_object_t hos_obj,
                        const adk_acl_t* list)
{
    adk_status_t retValue = ADK_SUCCESS;

    if(!cur->rule.num_hos_rules)
    {
        cur->op_type = ADK_ACL_NONE;
        cur->hos_committed = true;
        return retValue;
    }

    if(ADK_SUCCESS != (retValue = acl_rule_hos_add_internal(cur, hos_obj, true)))
        return retValue;

    entryStates[*num_prec] = IES_ACL_RULE_ENTRY_STATE_VALID;
    precedences[*num_prec] = cur->hos_precedence_mod;
    ++(*num_prec);
    if(*num_prec > list->info.num_ops_allowed)
    {
        printf("num_prec %d and list->info.num_ops_allowed %d\n", *num_prec, list->info.num_ops_allowed);
        retValue = ADK_E_UNSUPPORTED; 
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ACL List Error: User operation is resulting into too many HLP operations\n");
        return retValue;
    }

    /* This node is committed to HOS*/
    cur->op_type = ADK_ACL_NONE;
    cur->hos_committed = true;
    cur->hos_precedence = cur->hos_precedence_mod;
    cur->hos_precedence_mod = 0;
 
    return retValue;
}

static adk_status_t
adk_acl_rule_hos_precedence_modify(adk_acl_entry_t* cur,
                        adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;

    /* If this not was not hos committed earlier, just
        change precedence and return */
    if(!cur->hos_committed)
    {
        if(!cur->hos_precedence_mod)
            cur->hos_precedence_mod = cur->hos_precedence;
        cur->hos_precedence = 0;
        return retValue;
    }
        
    for(uint32_t iRule = 0; iRule < cur->rule.num_hos_rules ; ++iRule)
    {
        if((hos_obj == ADK_HOS_IP4_CLASSIFIER_TBL) ||
           (hos_obj == ADK_HOS_IP6_CLASSIFIER_TBL))
        {
            adk_hos_ip_classifier_tbl_rule_attr_t *classifier_rule =
                (adk_hos_ip_classifier_tbl_rule_attr_t *)
                (cur->rule.hos_rule_attr[iRule]);
            classifier_rule->precedence = cur->hos_precedence_mod;
        }
        else
        {
            /*ADK_ACL_SPD*/
            adk_hos_spd_tbl_rule_attr_t *spd_rule =
                (adk_hos_spd_tbl_rule_attr_t *)
                (cur->rule.hos_rule_attr[iRule]);
            spd_rule->precedence = cur->hos_precedence_mod;
        }
        retValue = adk_hos_rule_update(hos_obj,
                        cur->hos_rule_id[iRule],
                        cur->rule.hos_rule_attr[iRule]);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: Failed to modify ACL rule\n");
            return retValue;
        }
    }

    cur->hos_precedence = cur->hos_precedence_mod;
    cur->hos_precedence_mod = 0;
    return retValue;
}

static adk_status_t
adk_acl_rule_hos_modify(adk_acl_entry_t* cur,
                        uint32_t* num_prec,
                        uint32_t* precedences,
                        enum ies_acl_entry_state* entryStates,
                        adk_hos_object_t hos_obj,
                        const adk_acl_t* list)
{
    adk_status_t retValue = ADK_SUCCESS;

    /*Free the old hos rules */
    for(uint32_t iRule=0; iRule < cur->rule.num_hos_rules; ++iRule)
    {
        if(cur->rule.hos_rule_attr[iRule])
            ADK_NVM_FREE(cur->rule.hos_rule_attr[iRule]);
        cur->rule.hos_rule_attr[iRule] = NULL;
        cur->bkp_hos_rule_id[iRule] = cur->hos_rule_id[iRule];
    }
    cur->bkp_num_hos_rules = cur->rule.num_hos_rules;
    cur->rule = cur->modify_rule;

    /*Add the modified rules in disabled state*/
    /*toggle last bit*/
    cur->hos_precedence = cur->hos_precedence ^ 0x1;
    for(uint32_t iRule = 0; iRule < cur->rule.num_hos_rules ; ++iRule)
    {
        if((hos_obj == ADK_HOS_IP4_CLASSIFIER_TBL) ||
            (hos_obj == ADK_HOS_IP6_CLASSIFIER_TBL))
        {
            adk_hos_ip_classifier_tbl_rule_attr_t *classifier_rule =
                (adk_hos_ip_classifier_tbl_rule_attr_t*)
                (cur->rule.hos_rule_attr[iRule]);
            /*Add the new rule in disabled state*/
            classifier_rule->is_disabled = true;
            classifier_rule->precedence = cur->hos_precedence;
        }
        else
        {
            /*ADK_ACL_SPD*/
            adk_hos_spd_tbl_rule_attr_t *spd_rule =
                (adk_hos_spd_tbl_rule_attr_t *)
                (cur->rule.hos_rule_attr[iRule]);
            /*Add the new rule in disabled state*/
            spd_rule->is_disabled = true;
            spd_rule->precedence = cur->hos_precedence;
        }
        retValue = adk_hos_rule_add(hos_obj,
                        cur->rule.hos_rule_attr[iRule],
                        &(cur->hos_rule_id[iRule]));
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: Failed to add ACL rule\n");
            return retValue;
        }
    }

    entryStates[*num_prec] = IES_ACL_RULE_ENTRY_STATE_VALID;
    precedences[*num_prec] = cur->hos_precedence;
    ++(*num_prec);
    if(*num_prec > list->info.num_ops_allowed)
    {
        retValue = ADK_E_UNSUPPORTED; 
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ACL List Error: User operation is resulting into too many HLP operations\n");
        return retValue;
    }

    if(cur->rule.num_hos_rules & cur->hos_committed)
    {
        entryStates[*num_prec] = IES_ACL_RULE_ENTRY_STATE_INVALID;
        precedences[*num_prec] = cur->hos_precedence ^ 0x1;
        ++(*num_prec);

        if(*num_prec > list->info.num_ops_allowed)
        {
            retValue = ADK_E_UNSUPPORTED; 
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: User operation is resulting into too many HLP operations\n");
            return retValue;
        }
    }
    
    memset(&cur->modify_rule, 0, sizeof(cur->modify_rule));
    cur->hos_committed = true;
    return retValue;
 }

static adk_status_t
adk_acl_surrender_range_left(adk_acl_entry_t* prevNode,
                                adk_acl_entry_t* nextNode,
                                adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t numNodes = 1;
    uint32_t basePrec = nextNode ? ADK_ACL_EFFEC_PREC(nextNode) : 0;
    adk_acl_entry_t *cur, *leftMostNode;

    /* prevNode is the only node in the list*/
    if(!nextNode && !(prevNode->prev))
    {
        prevNode->hos_precedence_mod = (ADK_ACL_MAX_NODE_PRECEDENCE >> 1) | 0x1;
        if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(prevNode, hos_obj)))
            return retValue;

        return retValue;
    }

    cur = prevNode;
    while(cur)
    {
        uint32_t prevHosPrec = cur->prev ? ADK_ACL_EFFEC_PREC(cur->prev) : adk_acl_pvt_g_p->max_precedence;
        uint32_t curRange = prevHosPrec - ADK_ACL_EFFEC_PREC(cur);
        if(curRange > ADK_ACL_LIST_MIN_NODE_SPACING)
            break;

        int32_t availableRange = prevHosPrec - basePrec;
        if(availableRange < (int)(adk_acl_pvt_g_p->node_spacing*numNodes + 1))
            break;
        numNodes += 1;
        cur = cur->prev;
    }

    if(numNodes == 1)
        return retValue;    
   
    leftMostNode = cur;
    uint32_t spacePerNode = ((leftMostNode ? ADK_ACL_EFFEC_PREC(leftMostNode) : adk_acl_pvt_g_p->max_precedence) -  basePrec) / numNodes;

    cur = prevNode;
    while(cur != leftMostNode)
    {
        basePrec += spacePerNode;
        cur->hos_precedence_mod = basePrec | 0x1;
        if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(cur, hos_obj)))
            return retValue;
        cur = cur->prev;
    }

    return retValue;
}

static adk_status_t
adk_acl_surrender_range_right(adk_acl_entry_t* prevNode,
                                adk_acl_entry_t* nextNode,
                                adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t numNodes = 1;
    uint32_t basePrec = prevNode ? ADK_ACL_EFFEC_PREC(prevNode) : adk_acl_pvt_g_p->max_precedence;
    adk_acl_entry_t *cur, *rightMostNode;

    /* NextNode is the only node in the list*/
    if(!prevNode && !(nextNode->next))
    {
        nextNode->hos_precedence_mod = (ADK_ACL_MAX_NODE_PRECEDENCE >> 1) | 0x1;    
        if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(nextNode, hos_obj)))
            return retValue;

        return retValue;
    }

    cur = nextNode;
    while(cur)
    {
        uint32_t nextHosPrec = cur->next ? ADK_ACL_EFFEC_PREC(cur->next) : 0;
        uint32_t curRange = ADK_ACL_EFFEC_PREC(cur) - nextHosPrec;
        if(curRange > ADK_ACL_LIST_MIN_NODE_SPACING)
            break;

        int32_t availableRange = basePrec - nextHosPrec;
        if(availableRange < (int)(adk_acl_pvt_g_p->node_spacing*numNodes + 1))
            break;
        numNodes += 1;
        cur = cur->next;
    }

    if(numNodes == 1)
        return retValue;    
   
    rightMostNode = cur; 
    uint32_t spacePerNode = (basePrec - (rightMostNode ? ADK_ACL_EFFEC_PREC(rightMostNode) : 0)) / numNodes;

    cur = nextNode;
    while(cur != rightMostNode)
    {
        basePrec -= spacePerNode;
        cur->hos_precedence_mod = basePrec | 0x1;
        if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(cur, hos_obj)))
            return retValue;
        cur = cur->next;
    }

    return retValue;
}

static adk_status_t
adk_acl_surrender_range(adk_acl_entry_t* prevNode,
                        adk_acl_entry_t* nextNode,
                        uint32_t index,
                        adk_hos_object_t hos_obj,
                        adk_acl_t* list)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t avgLeftDist = 0, avgRightDist = 0;
    bool surrenderRangeRight = false, surrenderRangeLeft = false;

    if(!index)
    {
        if(nextNode)
            surrenderRangeRight = true;
    }
    else if(prevNode)
    {
        surrenderRangeLeft = true;
        if(nextNode)
        {
            avgLeftDist = (adk_acl_pvt_g_p->max_precedence - ADK_ACL_EFFEC_PREC(prevNode)) / index;
            avgRightDist = ADK_ACL_EFFEC_PREC(prevNode) / (list->num_nodes - index + 1);
            if(avgLeftDist > avgRightDist)
            {
                surrenderRangeRight = true;
                surrenderRangeLeft = false;
            }
        }
    }
    else if(nextNode)
    {
        surrenderRangeRight = true;
        if(prevNode)
        {
            avgLeftDist = (adk_acl_pvt_g_p->max_precedence - ADK_ACL_EFFEC_PREC(nextNode)) / (index + 1);
            avgRightDist = ADK_ACL_EFFEC_PREC(nextNode) / (list->num_nodes - index);
            if(avgLeftDist < avgRightDist)
            {
                surrenderRangeLeft = true;
                surrenderRangeRight = false;
            }
        }
    }

    if(surrenderRangeRight)
    { 
        retValue = adk_acl_surrender_range_right(prevNode, nextNode, hos_obj);
        return retValue;
    }

    if(surrenderRangeLeft)
    {
        retValue = adk_acl_surrender_range_left(prevNode, nextNode, hos_obj);
        return retValue;
    }
   
    return retValue;
}

static adk_status_t
adk_acl_rule_hos_delete(adk_acl_entry_t* cur,
                        uint32_t index,
                        adk_hos_object_t hos_obj,
                        adk_acl_t* list)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_entry_t *prevNode, *nextNode;
    uint32_t range, prevHosPrec, nextHosPrec;

    /* Delete the rules from HLP. These rules were atomically disabled earlier*/
    if(cur->hos_committed)
    {
        for(uint32_t iRule = 0; iRule < cur->rule.num_hos_rules; ++iRule)
        {
            if(cur->rule.hos_rule_attr[iRule])
                ADK_NVM_FREE(cur->rule.hos_rule_attr[iRule]);
            cur->rule.hos_rule_attr[iRule] = NULL;
            retValue = adk_hos_rule_delete(hos_obj,
                             cur->hos_rule_id[iRule]);
            if(retValue != ADK_SUCCESS)
            {
                ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                              ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                              "ACL List Error: H/W Error deleting old ACL rule in HOS\n");
                return retValue;
            }
        }
    }

    prevNode = cur->prev;
    nextNode = cur->next;
    prevHosPrec = prevNode ? ADK_ACL_EFFEC_PREC(prevNode) : adk_acl_pvt_g_p->max_precedence;
    nextHosPrec = nextNode ? ADK_ACL_EFFEC_PREC(nextNode) : 0;
    range = prevHosPrec - nextHosPrec;

    if(cur->prev)
        cur->prev->next = cur->next;
    else
        list->head = cur->next;
    if(cur->next)
        cur->next->prev = cur->prev;
    ADK_NVM_FREE(cur);

    list->num_nodes--;

    if(NULL == list->head)
        return retValue;

    if(range > adk_acl_pvt_g_p->node_spacing)
    {
        if(ADK_SUCCESS != (retValue = adk_acl_surrender_range(prevNode, nextNode, index, hos_obj, list)))
        {
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: Error surrender range fails\n");
            return retValue;
        }
    }        

    return retValue;
}

static adk_status_t
adk_acl_rule_hos_modified_delete(adk_acl_entry_t* cur,
                                    adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;

    if(!cur->hos_committed)
    {
        cur->bkp_num_hos_rules = 0;
        return retValue;
    }

    /*Delete the old rules*/
    for(uint32_t iRule = 0; iRule < cur->bkp_num_hos_rules; ++iRule)
    {
        adk_hos_rule_stats_t hos_stats;
        /* Save the bkp rule stats & delete it*/
        if(ADK_SUCCESS == adk_hos_rule_stats_get(hos_obj,
                                        cur->bkp_hos_rule_id[iRule],
                                        &hos_stats))
        {
            cur->bkp_stats += hos_stats.num_pkts;
        }
        retValue = adk_hos_rule_delete(hos_obj,
                        cur->bkp_hos_rule_id[iRule]);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: H/W Error deleting pre-modification ACL rule in HOS\n");
            return retValue;
        }
    }

    cur->bkp_num_hos_rules = 0;
    return retValue;
}

#ifdef ADK_CP_LIST_DEBUG
static void dumpHosTable(void)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_hos_rule_id_t rule_id = -1;  /* Set to -1 for first rule */
    adk_hos_ip_classifier_tbl_rule_attr_t classifier_attr;

    while (true)
    {
        return;
        retValue = adk_hos_next_rule_id_get(ADK_HOS_IP4_CLASSIFIER_TBL, rule_id, &rule_id);
        if(retValue != ADK_SUCCESS || -1 ==  rule_id)
            break;

        retValue = adk_hos_rule_get(ADK_HOS_IP4_CLASSIFIER_TBL, rule_id, &classifier_attr);
        if(retValue != ADK_SUCCESS)
            break;

        printf("Classifier rule ID: %d ", rule_id);
        printf("precedence: %d\n", classifier_attr.precedence);
#if 0
        printf("Classifier IPv6: %d\n", classifier_attr.is_ipv6);
        printf("Classifier proto: %d\n", classifier_attr.proto);
        printf("Classifier proto mask: 0x%x\n", classifier_attr.proto_mask);
        printf("Classifier SIP: 0x%x\n", classifier_attr.sip_addr[0]);
        printf("Classifier SIP mask: 0x%x\n", classifier_attr.sip_addr_mask[0]);
        printf("Classifier DIP: 0x%x\n", classifier_attr.dip_addr[0]);
        printf("Classifier DIP mask: 0x%x\n", classifier_attr.dip_addr_mask[0]);
        printf("Classifier SPORT: %d\n", classifier_attr.l4_info1);
        printf("Classifier SPORT mask: 0x%x\n", classifier_attr.l4_info1_mask);
        printf("Classifier DPORT: %d\n", classifier_attr.l4_info2);
        printf("Classifier DPORT mask: 0x%x\n", classifier_attr.l4_info2_mask);
        printf("Classifier ACL ID: 0x%x\n", classifier_attr.rule_id);
#endif
    }
}
#endif

static adk_status_t
acl_add_rules_enabled(adk_acl_entry_t* cur,
                        adk_hos_object_t hos_obj,
                        uint32_t* num_hos_rules)
{
    adk_status_t retValue = ADK_SUCCESS;
    uint32_t tbl_size;

    adk_hos_table_size_get(hos_obj, &tbl_size);

    *num_hos_rules += cur->rule.num_hos_rules;
    if(*num_hos_rules <= tbl_size)
    {
        if(ADK_SUCCESS != (retValue = acl_rule_hos_add_internal(cur, hos_obj, false)))
            return retValue;

        cur->op_type = ADK_ACL_NONE;
        if(cur->hos_precedence_mod)
            cur->hos_precedence = cur->hos_precedence_mod;
        cur->hos_precedence_mod = 0;
        cur->hos_committed = true;
    }
    else
    {
        cur->op_type = ADK_ACL_ADD;
        if(!cur->hos_precedence_mod)
            cur->hos_precedence_mod = cur->hos_precedence;
        cur->hos_precedence = 0;
        cur->hos_committed = false;
    }

    return retValue;
}


static adk_status_t
adk_acl_trigger_clean_slate(adk_acl_t* list,
                            adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_entry_t* cur = list->head;
    uint32_t num_hos_rules = 0;

    /* First clean the HLP table */
    retValue = adk_hos_table_clean(hos_obj, true);
    if(retValue != ADK_SUCCESS)
    {
        retValue = ADK_E_HW_PROGRAMMING_ERROR;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ACL List Error: Failed to run clean slate mechanism, "
                      "HLP table clean up fails\n");
        return retValue;
    }

    while(cur)
    {
        if(cur->op_type == ADK_ACL_ADD || cur->op_type == ADK_ACL_NONE)
        {
            if(ADK_SUCCESS != (retValue = acl_add_rules_enabled(cur, hos_obj, &num_hos_rules)))
                return retValue;

        }
        else if(cur->op_type == ADK_ACL_DEL)
        {
            adk_acl_entry_t* temp = cur->next;
            for(uint32_t iRule = 0; iRule < cur->rule.num_hos_rules; ++iRule)
            {
                if(cur->rule.hos_rule_attr[iRule])
                    ADK_NVM_FREE(cur->rule.hos_rule_attr[iRule]);
                cur->rule.hos_rule_attr[iRule] = NULL;
            }

            if(cur->prev)
                cur->prev->next = cur->next;
            else
                list->head = cur->next;
            if(cur->next)
                cur->next->prev = cur->prev;
            ADK_NVM_FREE(cur);

            cur = temp;
            continue;
        }
        else if(cur->op_type == ADK_ACL_MODIFY)
        {
            /*Free the old hos rules */
            for(uint32_t iRule=0; iRule < cur->rule.num_hos_rules; ++iRule)
            {
                if(cur->rule.hos_rule_attr[iRule])
                    ADK_NVM_FREE(cur->rule.hos_rule_attr[iRule]);
                cur->rule.hos_rule_attr[iRule] = NULL;
            }
            cur->bkp_num_hos_rules = 0;
            cur->bkp_stats = 0;
            cur->rule = cur->modify_rule;
            memset(&(cur->modify_rule), 0, sizeof(cur->modify_rule));
   
            cur->hos_precedence_mod = ADK_ACL_EFFEC_PREC(cur) | 0x1; 
            if(ADK_SUCCESS != (retValue = acl_add_rules_enabled(cur, hos_obj, &num_hos_rules)))
                return retValue;
        }

        cur = cur->next;
    }

    return retValue;
}

static adk_status_t
adk_acl_post_commit(adk_acl_t* list,
                            adk_hos_object_t hos_obj)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_entry_t* cur = list->head;
    uint32_t index = 0;

    /* Reset node spacing to default for surrender ID algorithm */
    adk_acl_pvt_g_p->node_spacing = ADK_ACL_DEFAULT_NODE_SPACING;
 
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            adk_acl_entry_t* prev = cur;
            cur = cur->next;
            retValue = adk_acl_rule_hos_delete(prev, index, hos_obj, list);
            if(ADK_SUCCESS != retValue)
                return retValue;

            continue;
        }
        else if(cur->op_type == ADK_ACL_MODIFY)
        {
    
            retValue = adk_acl_rule_hos_modified_delete(cur, hos_obj);
            if(ADK_SUCCESS != retValue)
                return retValue;

            /* Reassign the odd precedence for the rule, This is just for sanity and
                debugging. Algorithms can run even without this */
            cur->hos_precedence_mod = ADK_ACL_EFFEC_PREC(cur) ^ 0x1; 
            if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(cur, hos_obj)))
                return retValue;

            if(cur->hos_committed)
                cur->op_type = ADK_ACL_NONE;
            else
                cur->op_type = ADK_ACL_ADD;
        }

        cur = cur->next;
        ++index;
    }

    return retValue;
 }

adk_status_t
adk_acl_hos_commit(adk_acl_handle_t acl_hdl)
{
    adk_status_t retValue = ADK_SUCCESS;
    adk_acl_t *list;
    adk_acl_entry_t *cur = NULL, *offload_node = NULL;
    adk_acl_entry_t *revNodeStart = NULL, *revNodeEnd = NULL;
    uint32_t num_prec = 0;
    adk_hos_object_t hos_obj;
    enum ies_acl_entry_state* entryStates = NULL;
    uint32_t* precedences = NULL;
    bool reverse_iter = false;
    bool fast_forward = false;

    if(!acl_hdl)
    {
        retValue = ADK_E_INVALID_PARAMETER;
        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                      "ACL List Error: Failed commit CP List to HOS, "
                      "invalid parameter\n");
        return retValue;
    }

    list = (adk_acl_t *)acl_hdl;

    cur = list->head;
    /* There is nothing to commit */
    if(!cur)
        return ADK_SUCCESS;

    if(ADK_SUCCESS != (retValue = adk_acl_get_hos_table(&hos_obj, list->info.user)))
        return retValue;

    adk_acl_find_potential_candidates_offload(&offload_node, hos_obj, list);
    if(NULL == offload_node)
    {
        ADK_LOG_INFO(adk_acl_pvt_g_p->sys_handle,
                        ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                        "No candidates to offload\n");
   
        if(ADK_SUCCESS != (retValue = adk_acl_delete_overflow_hos_rules(list, offload_node, hos_obj)))
            goto _adk_acl_rules_state_free_exit;

        if(ADK_SUCCESS != (retValue = adk_acl_sw_only_commit(acl_hdl)))
            goto _adk_acl_rules_state_free_exit;

        return retValue; 
    }

    if(ADK_SUCCESS != (retValue = adk_acl_delete_overflow_hos_rules(list, offload_node, hos_obj)))
        goto _adk_acl_rules_state_free_exit;
     
    if(ADK_SUCCESS  != (retValue = adk_acl_rules_state_alloc(&precedences, &entryStates, list)))
        goto _adk_acl_rules_state_free_exit;

    /*iterate forward*/
    while(cur != offload_node->next || fast_forward)
    {
        if(fast_forward)
        {
            if(cur->op_type == ADK_ACL_ADD && cur != offload_node->next)
            {
                cur = cur->next;
                continue;
            }

            if(!(cur->hos_precedence_mod) ||
                (cur->hos_precedence_mod >= cur->hos_precedence)||
                (NULL == cur->next) || (cur == offload_node->next))
            {
                fast_forward = false;
                reverse_iter = true;
                if(NULL != cur->next)
                    cur = cur->prev;
                /* Store the end node of fast forward, the last node for which
                    the modified precedence is less than original precedence */
                revNodeEnd = cur;
            }
            else
            {
                cur = cur->next;
            }

            continue;
        }

        if(cur->op_type == ADK_ACL_ADD)
        {
            if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_add(cur, &num_prec, precedences,
                        entryStates, hos_obj, list)))
                goto _adk_acl_rules_state_free_exit;
        }
        else
        {
            /* An existing rule: modify or del or none */
            /* First shift to new index if different*/
            if(cur->hos_precedence_mod &&
                (cur->hos_precedence != cur->hos_precedence_mod))
            {
                /* Change precedence immediately, it's not going to mess up the existing order */
                if(cur->hos_precedence_mod > cur->hos_precedence)
                {
                    if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(cur, hos_obj)))
                        goto _adk_acl_rules_state_free_exit;
                }
                else
                {
                    if(reverse_iter)
                    {
                        if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_precedence_modify(cur, hos_obj)))
                            goto _adk_acl_rules_state_free_exit;
                    }
                    else if(!fast_forward)
                    {                        
                        /* Changing precedence immediately can mess up the order
                        Fast forward until a node is found with modified 
                        precedence greater or equal to old precedence */ 
                        fast_forward = true;
                        revNodeStart = cur;
                        continue;
                    }
               }
            }
            
            if(cur->op_type == ADK_ACL_MODIFY)
            {
                if(ADK_SUCCESS != (retValue = adk_acl_rule_hos_modify(cur, &num_prec, precedences,
                                    entryStates, hos_obj, list)))
                    goto _adk_acl_rules_state_free_exit;
            }
            else if(cur->op_type == ADK_ACL_DEL)
            {
                if(cur->rule.num_hos_rules && cur->hos_committed)
                {
                    entryStates[num_prec] = IES_ACL_RULE_ENTRY_STATE_INVALID;
                    precedences[num_prec] = cur->hos_precedence;
                    num_prec++;
                    if(num_prec > list->info.num_ops_allowed)
                    {
                        retValue = ADK_E_UNSUPPORTED; 
                        ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                                      ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                                      "ACL List Error: User operation is resulting "
                                        "into too many HLP operations\n");
                        goto _adk_acl_rules_state_free_exit;
                    }
                }
            }
        }
       
        if(reverse_iter)
        {
            if(cur == revNodeStart)
            {
                reverse_iter = false;
                /* All nodes upto revNodeEnd already processed */
                while(cur != revNodeEnd)
                    cur = cur->next;
                cur = cur->next;
            }
            else
            {
                cur = cur->prev;
            }                
        }
        else
        {
            cur = cur->next;
        }
    }

    if(num_prec)
    {
        /*Bulk update*/
        retValue = adk_hos_rule_precedence_state_set(hos_obj,
                                 num_prec,
                                 precedences,
                                 entryStates);
        if(retValue != ADK_SUCCESS)
        {
            retValue = ADK_E_UNSUPPORTED; 
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                          ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                          "ACL List Error: H/W Error in bulk updating the ACL rule status\n");
            goto _adk_acl_rules_state_free_exit;
        }
    }

    /* If no rule deletion or modification, second iteration
        is not required */
    if(!list->secondIterReq)
    {
        list->num_ops = 0;
        goto _adk_acl_rules_state_free_exit;
    }

    /* Iterate again to remove the del entry and old entries of modified rules */
    if(ADK_SUCCESS != (retValue = adk_acl_post_commit(list, hos_obj)))
        goto _adk_acl_rules_state_free_exit;
    
    list->num_ops = 0;
    list->secondIterReq = false;

_adk_acl_rules_state_free_exit:
    adk_acl_reset_default_spacing();
    adk_acl_rules_state_free(precedences, entryStates);
    if(ADK_SUCCESS != retValue)
    {
        retValue = adk_acl_trigger_clean_slate(list, hos_obj);
        if(retValue != ADK_SUCCESS)
        {
            ADK_LOG_ERROR(adk_acl_pvt_g_p->sys_handle,
                            ADK_SYS_MODULE_HANDLE_GET(adk_acl_pvt_g_p->sys_handle),
                            "Normal commit to HLP failed. Unfortunately fail safe "
                            "also failed. Now system is in undefiend state. Application "
                            "may want to undo/redo all ACL rules and rebuild\n");
        }
    }

    return retValue;
}

adk_status_t
adk_acl_sw_only_commit(adk_acl_handle_t list_hdl)
{
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL, *tmp=NULL;

    if(!list_hdl)
        return ADK_E_INVALID_PARAMETER;

    list = (adk_acl_t *)list_hdl;

    /* Reset the main CP list */
    cur = list->head;
    while(cur)
    {
        if(cur->op_type == ADK_ACL_DEL)
        {
            /*Del the entry at this pos*/
            if(cur->prev)
                cur->prev->next = cur->next;
            else
                list->head = cur->next;
            if(cur->next)
                cur->next->prev = cur->prev;

            tmp = cur;
            cur = cur->next;
            ADK_NVM_FREE(tmp);
            list->num_nodes--;
        }
        else
        {
            if(cur->op_type == ADK_ACL_MODIFY)
            {
                cur->rule = cur->modify_rule;
            }
            if(cur->hos_committed)
                cur->op_type = ADK_ACL_NONE;
            else
                cur->op_type = ADK_ACL_ADD;
            cur = cur->next;
        }
    }
    list->num_ops = 0;
    list->secondIterReq = false;
    return ADK_SUCCESS;
}

adk_status_t
adk_acl_hos_stats_get(adk_acl_handle_t acl_hdl,
                        void *user_data,
                        uint32_t *hit_counter)
{
    adk_acl_t *list;
    adk_acl_entry_t *cur=NULL;
    adk_hos_object_t hos_obj;
    adk_hos_rule_stats_t hos_stats;

    if(!acl_hdl || !user_data || !hit_counter)
        return ADK_E_INVALID_PARAMETER;

    list = (adk_acl_t *)acl_hdl;
    if(list->info.user == ADK_ACL_IPV4_ING_CLASSIFIER)
        hos_obj = ADK_HOS_IP4_CLASSIFIER_TBL;
    else if(list->info.user == ADK_ACL_IPV6_ING_CLASSIFIER)
        hos_obj = ADK_HOS_IP6_CLASSIFIER_TBL;
    else if(list->info.user == ADK_ACL_SPD_IN)
        hos_obj = ADK_HOS_SPD_IN_TBL;
    else if(list->info.user == ADK_ACL_SPD_OUT)
        hos_obj = ADK_HOS_SPD_OUT_TBL;
    else
        return ADK_E_INVALID_PARAMETER;

    cur = list->head;
    while(cur)
    {
        if(cur->rule.user_data == user_data)
        {
            *hit_counter = 0;

            if(cur->op_type == ADK_ACL_ADD)
            {
                /*Newly added rule & not commited yet*/
                return ADK_SUCCESS;
            }

            for(int i=0; i<cur->rule.num_hos_rules; ++i)
            {
                if(ADK_SUCCESS == adk_hos_rule_stats_get(hos_obj,
                                                         cur->hos_rule_id[i],
                                                         &hos_stats))
                {
                    *hit_counter += hos_stats.num_pkts;
                }
            }

            /* Include the bkp stats due to rule modify */
            *hit_counter += cur->bkp_stats;

            return ADK_SUCCESS;
        }

        cur = cur->next;
    }

    return ADK_E_LIST_NOT_FOUND;
}

static const char*
adk_acl_action_string(adk_acl_op_type_t op_type)
{
    switch(op_type)
    {
        case ADK_ACL_ADD:
            return "ADK_ACL_ADD";
        case ADK_ACL_DEL:
            return "ADK_ACL_DEL";
        case ADK_ACL_MODIFY:
            return "ADK_ACL_MODIFY";
        case ADK_ACL_NONE:
            return "ADK_ACL_NONE";
        default:
            return "ADK_ACL_INVALID";
    }

    return "ADK_ACL_INVALID";
}

adk_status_t
adk_acl_dump(adk_acl_handle_t acl_hdl)
{
    adk_acl_t *list;
    adk_acl_entry_t *cur = NULL;
    uint32_t index = 0;

    if(!acl_hdl)
        return ADK_E_INVALID_PARAMETER;

    list = (adk_acl_t *)acl_hdl;

    cur = list->head;
    while(cur)
    {
        printf("Index %.3d\t", index);
        printf("Action %s\t", adk_acl_action_string(cur->op_type));
        printf("hos_precedence %d\t", cur->hos_precedence);
        printf("hos_precedence_mod %d\t", cur->hos_precedence_mod);
        printf("hos_committed %s\n", cur->hos_committed ? "true" : "false");

        cur = cur->next;
        ++index; 
    }
    
    return ADK_SUCCESS; 
}

/*
 * This function compares the 2 128 bit values & return values as follows
    0  if start is equal to end
    1  if start is greater than end
    -1 if start is less than end
*/
int adk_cp_acl_compare128(uint64_t start[], uint64_t end[]);

/* This function returns the first bit set position from LSB 
 * accepts 128 bit value */
int adk_cp_acl_ffs128(uint64_t num[]);

/* This function supports 128 bit arithmatic add */
void adk_cp_acl_add128(uint64_t num1[], uint64_t num2[], uint64_t result[]);

/* This function supports 128 bit arithmatic subtract */
void adk_cp_acl_sub128(uint64_t num1[], uint64_t num2[], uint64_t result[]);

/* returns mask for input size bits */
void adk_cp_acl_get_mask(uint8_t size_bits, uint64_t mask[]);

int
adk_acl_convert_range_bitmask(uint16_t start,
        uint16_t end, uint8_t size_bits, adk_acl_range_bitmask_t *arr)
{
    uint16_t first_set_position, rule_count =0;
    uint32_t new_start, range;
    uint16_t mask = ((1 << size_bits) -1);
    adk_acl_range_bitmask_t *local = arr;

    if(end > mask)
    {
        return -1;
    }

    new_start = start;
    while(new_start <= end)
    {
        first_set_position = __builtin_ffs(new_start);
        /*Handle Zero start value*/
        if(first_set_position == 0)
        {
            /*Check for complete Range usage, to generate single rule*/
            if(end == (1<<size_bits)-1)
            {
                if(local)
                    local->mask = ~mask;
                rule_count++;
                return rule_count;
            }
            first_set_position = size_bits;
        }

        if(local)
            local->val = new_start;
        range = 1 << (first_set_position -1);

        while((new_start + range-1) >end)
        {
            first_set_position -=1;
            range = 1 << (first_set_position - 1);
        }

        if(local)
            local->mask = (mask <<(first_set_position-1)) & mask;

        rule_count++;
        if(rule_count > ADK_ACL_MAX_HOS_RULES_PER_RULE)
        {
            return -1;
        }
        new_start += range;
        if(local)
            local +=1;
    }
    return rule_count;
}

int adk_cp_acl_compare128(uint64_t start[], uint64_t end[])
{
        return ((((start[1] > end[1]) || ((start[1] == end[1]) && (start[0] > end[0]))) ? 1 : 0) 
                     - (((start[1] < end[1]) || ((start[1] == end[1]) && (start[0] < end[0]))) ? 1 : 0));
}

int adk_cp_acl_ffs128(uint64_t num[])
{
    int first_set_position;
    if(num[0])
        first_set_position = __builtin_ffsll(num[0]);
    else if(num[1])
        first_set_position = 64 + __builtin_ffsll(num[1]);
    else
        first_set_position = 0;

    return first_set_position;
}

void adk_cp_acl_add128(uint64_t num1[], uint64_t num2[], uint64_t result[])
{

    if((num1[0] + num2[0])<num1[0])
    {
        result[1] = num1[1]+num2[1]+1;
    }
    else
    {
        result[1] = num1[1]+num2[1];
    }

    result[0] = num1[0]+num2[0];
    return;
}

void adk_cp_acl_sub128(uint64_t num1[], uint64_t num2[], uint64_t result[])
{

    if((num1[0] - num2[0]) > num1[0])
    {
        result[1] = num1[1]-num2[1]-1;
    }
    else
    {
        result[1] = num1[1]-num2[1];
    }

    result[0] = num1[0]-num2[0];
    return;
}

void adk_cp_acl_get_mask(uint8_t size_bits, uint64_t mask[])
{
    if(size_bits == 128)
    {
        mask[1] = 0xFFFFFFFFFFFFFFFF;
        mask[0] = 0xFFFFFFFFFFFFFFFF;
   
    }
    else if(size_bits >= 64)
    {
        mask[1] = ((1ll<<(size_bits-64))-1);
        mask[0] = 0xFFFFFFFFFFFFFFFF;
    }
    else
    {
        mask[1] = 0;
        mask[0] = ((1ll<<size_bits)-1);
    }
}

int
adk_acl_convert_addr_range_bitmask(uint64_t start[],
        uint64_t end[], uint8_t size_bits, adk_acl_addr_range_bitmask_t *arr)
{
    uint16_t first_set_position, rule_count =0;
    uint64_t new_start[2], tmp_new_start[2], range[2], tmp[2];
    uint64_t mask[2];
    bool     is128bitwrapped;

    adk_acl_addr_range_bitmask_t *local = arr;
    adk_cp_acl_get_mask(size_bits, mask);

    new_start[0] = start[0];
    new_start[1] = start[1];
    
    while(adk_cp_acl_compare128(new_start, end) <= 0)
    {
        first_set_position = adk_cp_acl_ffs128(new_start);
        /*Handle Zero start value*/
        if(first_set_position == 0)
        {
            /*Check for complete Range usage, to generate single rule*/
            if(end[0] == mask[0] && end[1] == mask[1])
            {
                if(local)
                {
                    local->val[1] = new_start[1];
                    local->val[0] = new_start[0];
                    local->mask[1] = ~mask[1];
                    local->mask[0] = ~mask[0];
                }
                rule_count++;
                return rule_count;
            }
            first_set_position = size_bits;
        }

        if(local)
        {
            local->val[1] = new_start[1];
            local->val[0] = new_start[0];
        }


        if(first_set_position > 64)
        {
            range[1] = 1ll << ((first_set_position-64)-1);
            range[0] = 0;
        }
        else
        {
            range[1] = 0;
            range[0] = 1ll << (first_set_position -1);
        }

        tmp[0]=1;tmp[1]=0;
        adk_cp_acl_sub128(range, tmp, tmp_new_start);
        adk_cp_acl_add128(new_start, tmp_new_start, tmp_new_start);
        is128bitwrapped = (new_start[1]>tmp_new_start[1])?1:0;
        while(((adk_cp_acl_compare128(tmp_new_start, end)) == 1)||(is128bitwrapped))
        {
            first_set_position -=1;
            if(first_set_position > 64)
            {
                range[1] = 1ll << ((first_set_position-64)-1);
                range[0] = 0;
            }
            else
            {
                range[1] = 0;
                range[0] = 1ll << (first_set_position -1);
            }

            adk_cp_acl_add128(new_start, range, tmp_new_start);
            is128bitwrapped = (new_start[1]>tmp_new_start[1])?1:0;
            tmp[0]=1;tmp[1]=0;
            adk_cp_acl_sub128(tmp_new_start, tmp, tmp_new_start);
        }

        if(local)
        {
            if(first_set_position > 64)
            {
                local->mask[1] = (mask[1] << ((first_set_position-64)-1)) & mask[1];
                local->mask[0] = 0;
            }
            else
            {
                local->mask[1] = mask[1];
                local->mask[0] = (mask[0] <<(first_set_position-1)) & mask[0];
            }
        }

        rule_count++;
        if(rule_count > ADK_ACL_MAX_HOS_RULES_PER_RULE)
        {
            return -1;
        }

        adk_cp_acl_add128(new_start, range, tmp_new_start);
        if(local)
            local +=1;
        is128bitwrapped = (new_start[1]>tmp_new_start[1])?1:0;
        if(is128bitwrapped)
        {
            return rule_count;
        }
        new_start[0] = tmp_new_start[0];
        new_start[1] = tmp_new_start[1];
    }
    return rule_count;
}


