/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#ifndef __ADK_ACL_H__
#define __ADK_ACL_H__

/******************************************************************************
 * CP ACL List APIs facilitates the control plane list management of ACL rules.
 * The intended users are IPv4 ACL & IPv6 ACL. IPsec SPD can also leverage
 * the utility. The CP list will be useful in case of S/W (DPDK) based ACL,
 * as well as H/W (like SNR/HLP) based ACL implementations.
 *
 * In case of S/W (DPDK) based ACL implementations, the updates need complete
 * rebuild of ACL table. There are no rule delete or rule dump DPDK APIs,
 * hence the complete list of ACL rules need to be maintained  by the user.
 * This utility can be used for maintaining such a list.
 *
 * In case of H/W based ACL implementations, incremental updates are
 * possible. Hence the list of operations since last user commit will need
 * to be maintained along with complete list of updated rule list for handling
 * partial offload. This utility handles such different list management & also
 * provides h/w commit APIs which takes care of h/w bulk updates, unique
 * precedence management, partial offloads etc..
 *
 * This utility is under EXPERIMENTAL, APIs will change/evolve
 ******************************************************************************/

#define ADK_ACL_MAX_HOS_RULES_PER_RULE 128

typedef adk_handle_t adk_acl_handle_t;

typedef enum adk_acl_user_e {
    ADK_ACL_IPv4 = 0,
    ADK_ACL_IPv6,
    ADK_ACL_SPD_IN,
    ADK_ACL_SPD_OUT,
    ADK_ACL_IPV4_ING_CLASSIFIER,
    ADK_ACL_IPV6_ING_CLASSIFIER,
    ADK_ACL_IPV4_EGRESS_QOS_CLASSIFIER,
    ADK_ACL_IPV6_EGRESS_QOS_CLASSIFIER,
    ADK_ACL_IPV4_ROUTE_TABLE,
    ADK_ACL_IPV6_ROUTE_TABLE,
    ADK_ACL_MAX
} adk_acl_user_t;

typedef struct adk_acl_info_s {
    /* The user of the list*/
    adk_acl_user_t user;
    /* This is user level configuration to allow
        num operations in bulk commit for this list.
        Each add/delete/modify results in an operaton
        Specify 0 for default */
    uint32_t num_ops_allowed;
    /* This is a user level configuration to allow num HLP
        rules per CPU rule for this list. Since HLP does
        not support range e.g. port range, a CPU rule containing
        range can translate into multiple HLP rules using masks
        Specify 0 for default */
    uint32_t num_hw_rules_per_cpu_rule;
    adk_sys_handle_t sys_handle;
} adk_acl_info_t;

adk_status_t
adk_acl_init (
    adk_sys_handle_t sys_handle);

adk_status_t
adk_acl_shutdown (
    void);

adk_status_t
adk_acl_attach (
    void);

void
adk_acl_set_default_spacing (
    uint32_t spacing);

uint32_t
adk_acl_get_default_spacing (
    void);

void
adk_acl_reset_default_spacing (
    void);

void
adk_acl_set_max_precedence (
    uint32_t max_precedence);

uint32_t
adk_acl_get_max_precedence (
    void);

void
adk_acl_reset_max_precedence (
    void);

/* Get number of nodes in ADK ACL list */
adk_status_t
adk_acl_get_num_nodes(adk_acl_handle_t acl_hdl,
                        uint32_t* num_nodes);

/* Get whether a rule at given index is offloaded to HLP */
adk_status_t
adk_acl_rule_get_hos_committed(adk_acl_handle_t acl_hdl,
                        uint32_t index,
                        bool* hos_committed);

/* Set number of operations allowed on an ACL list.
    This API is supposed to be used for unit testing */
adk_status_t
adk_acl_set_num_ops_allowed(adk_acl_handle_t acl_hdl,
                        uint32_t num_ops_allowed);

adk_status_t
adk_acl_get_num_ops_allowed(adk_acl_handle_t acl_hdl,
                        uint32_t* num_ops_allowed);

adk_status_t
adk_acl_reset_num_ops_allowed(adk_acl_handle_t acl_hdl);

/* Create a new CP ACL list. */
adk_status_t
adk_acl_create(adk_acl_handle_t *acl_hdl,
                const adk_acl_info_t* info);

/* Free an existing CP ACL list. */
adk_status_t
adk_acl_free(adk_acl_handle_t acl_hdl);

/* Get handle of CP ACL list */
adk_acl_handle_t
adk_acl_hdl_get(adk_acl_user_t list_user);

/* Delete all the nodes for all rules in CP ACL list
   This does not remove HOS entries */
adk_status_t
adk_acl_clear(adk_acl_handle_t acl_hdl);

typedef void adk_hos_rule_attr_t;

typedef struct adk_acl_rule_s {
    /* User data like rule handle */
    void *user_data;
    /* Number of h/w rules for the user rule. H/W does not support ranges, &
     * each user rule need to be represented as multiple h/w rules in case of
     * ranges in the user rule. */
    uint16_t num_hos_rules;
    /* User allocated h/w rule attribute. */
    adk_hos_rule_attr_t *hos_rule_attr[ADK_ACL_MAX_HOS_RULES_PER_RULE];
} adk_acl_rule_t;

/* Add a rule to a CP ACL list at an index */
adk_status_t
adk_acl_rule_add(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    const adk_acl_rule_t *rule);

/* Delete a rule from a CP ACL list at the index*/
adk_status_t
adk_acl_rule_delete(adk_acl_handle_t acl_hdl,
                    uint16_t index);

/* Modify a rule from a CP ACL list at the index*/
adk_status_t
adk_acl_rule_modify(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    const adk_acl_rule_t *rule);

/* Get HOS precedences of a rule from a CP ACL list at an index*/
adk_status_t
adk_acl_rule_hos_precedences_get(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    uint32_t* hos_precedence,
                    uint32_t* hos_precedence_mod);

/* Set HOS precedences mod of a rule from a CP ACL list at an index
    This API is only for unit testing to create corner scenarios */
adk_status_t
adk_acl_rule_hos_precedences_set(adk_acl_handle_t acl_hdl,
                    uint16_t index,
                    uint32_t hos_precedence_mod);

/* Get the next rule in the List.
 * Use adk_acl_get_reset() & followed by adk_acl_next_rule_get()
 * to get the first rule in the list.
 * To be used by S/W ACL to get the updated list of rule for S/W rebuild or
 * for dumping the entries */
adk_status_t
adk_acl_next_rule_get(adk_acl_handle_t acl_hdl,
                        adk_acl_rule_t *rule);

/* Reset the get API to start from first rule.
 * Use adk_acl_get_reset() & followed by adk_acl_next_rule_get()
 * to get the first rule in the list */
adk_status_t
adk_acl_get_reset(adk_acl_handle_t acl_hdl);

/* H/W commit of all the operations on the List since the last commit. All the
 * operations will be updated atomically */
adk_status_t
adk_acl_hos_commit(adk_acl_handle_t acl_hdl);

/* When ACL is entirely in SW, this API needs to be called for SW commit */
adk_status_t
adk_acl_sw_only_commit(adk_acl_handle_t list_hdl);

/* Get H/W stats for the rule based on user data */
adk_status_t
adk_acl_hos_stats_get(adk_acl_handle_t acl_hdl,
                        void *user_data,
                        uint32_t *hit_counter);

/* Dump ACL precedences, indices and oper action */
adk_status_t
adk_acl_dump(adk_acl_handle_t acl_hdl);

typedef struct adk_acl_range_bitmask
{
    uint16_t val;
    uint16_t mask;
}adk_acl_range_bitmask_t;


typedef struct adk_acl_addr_range_bitmask
{
    uint64_t val[2];
    uint64_t mask[2];
}adk_acl_addr_range_bitmask_t;


/* This function is used for converting given range into "value/bitmask"
 * combination. HLP does not support range lookup in the FFU tables, only
 * bitmask is supported. 
 * @start      Start value of the range
 * @end        End value of range.
 * @size_bits  Size of the field(sport/dport/protocol/dscp) in bits.
 *             Supports upto 16 bit size values as input.
 * @arr        Pointer to store splitted value/bitmask combination. It is the
 *             responsibility of the caller to allocate memory for max number
 *             of entries and pass. The maximum number of splits supported is
 *             ADK_ACL_MAX_HOS_RULES_PER_RULE.
 * 
 * The functoin returns the number of splits done for given range.
 * returns  "0" incase start is greater than end value.
 * returns "-1" incase the splits is crossing the limit of
 *              ADK_ACL_MAX_HOS_RULES_PER_RULE.
 * returns "-1" incase end value is greater than the range supported by bit
 *              size passed.
 */
int
adk_acl_convert_range_bitmask(uint16_t start,
     uint16_t end, uint8_t size_bits, adk_acl_range_bitmask_t *arr);

/* This function is used for converting given IPv4/IPv6 address range into "value/bitmask"
 * combination. HLP does not support range lookup in the FFU tables, only
 * bitmask is supported. 
 * @start      Start value of the IPv4/IPv6 address range
 * @end        End value of IPv4/IPv6 address range.
 * @size_bits  Size of the field(IPv4/IPv6 address size) in bits.
 *             Supports upto 128 bit size values as input.
 * @arr        Pointer to store splitted value/bitmask combination. It is the
 *             responsibility of the caller to allocate memory for max number
 *             of entries and pass. The maximum number of splits supported is
 *             ADK_ACL_MAX_HOS_RULES_PER_RULE.
 * 
 * The functoin returns the number of splits done for given range.
 * returns  "0" incase start is greater than end value.
 * returns "-1" incase the splits is crossing the limit of
 *              ADK_ACL_MAX_HOS_RULES_PER_RULE.
 * returns "-1" incase end value is greater than the range supported by bit
 *              size passed.
 */
int
adk_acl_convert_addr_range_bitmask(uint64_t start[],
        uint64_t end[], uint8_t size_bits, adk_acl_addr_range_bitmask_t *arr);

#endif

