/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

#include <adk_include.h>
#include <adk_lpm.h>
#include <adk_rcu.h>

typedef struct adk_lpm_node_s adk_lpm_node_t;

struct adk_lpm_node_s
{
    adk_handle_t data;
    adk_lpm_node_t *bit_child[2];
    adk_lpm_node_t *byte_child[256];
    adk_lpm_node_t *pending_next;
    uint16_t use_count;
}__attribute__((aligned(64)));

typedef struct
{
    adk_lpm_node_t *root_node;
    adk_lpm_node_t *pending_list;
    uint16_t width;
    uint16_t pending_count;
    uint16_t pending_max;
}adk_lpm_t;

adk_lpm_handle_t adk_lpm_create(size_t mod_id, size_t obj_id, uint16_t width, uint16_t pending_size)
{
    adk_lpm_t *new_lpm = (adk_lpm_t *)
                         ADK_DP_SHM_CALLOC(mod_id, obj_id, 1, sizeof(adk_lpm_t), 0);

    //debug_print("LPM Create\n");

    if (new_lpm == NULL)
    {
        return (adk_lpm_handle_t)NULL;
    }

    new_lpm->root_node = NULL;
    new_lpm->pending_list = NULL;
    new_lpm->pending_count = 0;
    new_lpm->pending_max = pending_size;
    new_lpm->width = width;

    return (adk_lpm_handle_t)new_lpm;
}

static void lpm_node_free(adk_lpm_node_t *node)
{
    uint32_t i;

    /* TO DO: Uncomment: debug_print("LPM node free\n"); */

    /*Free all the byte children*/
    for(i = 0; i < 256; ++i)
    {
        if(NULL != node->byte_child[i])
        {
            lpm_node_free(node->byte_child[i]);
        }
    }

    /*Free the bit children*/
    if(NULL != node->bit_child[0])
    {
        lpm_node_free(node->bit_child[0]);
    }
    if(NULL != node->bit_child[1])
    {
        lpm_node_free(node->bit_child[1]);
    }

    /*Now all the children are freed.. So free the node*/
    ADK_NVM_FREE(node);
}

static void lpm_pending_list_node_free(adk_lpm_node_t *node)
{
    uint32_t i;

    //debug_print("LPM pending list free\n");

    /*Atmost 1 child will be present for a node in pending list*/

    /*Look for the byte child*/
    for(i = 0; i < 256; ++i)
    {
        if(node->byte_child[i])
        {
            lpm_pending_list_node_free(node->byte_child[i]);
            goto _child_found;
        }
    }

    /*Look for the bit child*/
    if(node->bit_child[0])
    {
        lpm_pending_list_node_free(node->bit_child[0]);
        goto _child_found;
    }

    if(node->bit_child[1])
    {
        lpm_pending_list_node_free(node->bit_child[1]);
    }

_child_found:
    if(node->pending_next)
    {
        lpm_pending_list_node_free(node->pending_next);
    }

    /*Free the node*/
    ADK_NVM_FREE(node);
}

void adk_lpm_free(adk_lpm_handle_t lpm_handle)
{
    adk_lpm_t * lpm = (adk_lpm_t *)lpm_handle;
    //debug_print("LPM Free\n");
    if(lpm)
    {
        if(lpm->root_node)
        {
            lpm_node_free(lpm->root_node);
        }
        if(lpm->pending_list)
        {
            lpm_pending_list_node_free(lpm->pending_list);
        }
        ADK_NVM_FREE(lpm);
    }
}

/* Searches if key is already present or not. Returns the node or NULL if not
 * present*/
static adk_lpm_node_t *lpm_node_find(adk_lpm_t *lpm,
                                     uint8_t *key, uint16_t depth)
{
    adk_lpm_node_t *node;
    uint8_t	*p;
    uint8_t	c, count, bit_depth, byte_depth;

    /* Uncommentdebug_print("LPM find.\n"); */

    node = lpm->root_node;
    byte_depth = depth/8;
    bit_depth = depth%8;

    /* Traverse byte wise*/
    for (count=0, p=key; count < byte_depth; ++count, ++p)
    {
        if (node == NULL)
        {
            /* Not found in the tree. Return. */
            return NULL;
        }

	/* Jump to the next node */
	c = *p;
	node = node->byte_child[c];
    }

    /*Traverese bit wise*/
    for (count=0; count < bit_depth; ++count)
    {
        if (node == NULL)
        {
            /* Not found in the tree. Return. */
            return NULL;
	}

        /* Jump to the next node */
	c = (*p >> (7-count)) & 0x1;
	node = node->bit_child[c];
    }

    /* This key is present if the value at the last node is not INVALID */
    return node;
}

int adk_lpm_is_rule_present(adk_lpm_handle_t lpm_handle, uint8_t *key,
                            uint16_t depth, adk_handle_t *value)
{
    adk_lpm_t * lpm = (adk_lpm_t *)lpm_handle;
    adk_lpm_node_t *node;

    //debug_print("LPM is rule present\n");

    if (!lpm_handle)
    {
        //debug_print("LPM is rule present: NULL LPM handle\n");
        return -1;
    }

    /* Find & replace the existing value if the key is already added*/
    node = lpm_node_find(lpm, key, depth);
    if (node != NULL && node->data != (adk_handle_t)(ADK_LPM_VALUE_INVALID))
    {
        //debug_print("LPM is rule present: Entry is present.\n");
        *value = node->data;
	return 1;
    }
    else
    {
        //debug_print("LPM is rule present: Entry not present.\n");
	return 0;
    }
}

int adk_lpm_add(size_t mod_id, size_t obj_id, adk_lpm_handle_t lpm_handle, uint8_t *key,
                uint16_t depth, adk_handle_t value)
{
    adk_lpm_t * lpm = (adk_lpm_t *)lpm_handle;
    adk_lpm_node_t **parent,  *node;
    uint8_t *byte_p, bit, count, bit_depth = depth%8, byte_depth = depth/8;

    /* Uncomment debug_print("LPM Add\n"); */

    if (value == (adk_handle_t)(ADK_LPM_VALUE_INVALID))
    {
        /* Uncomment debug_print("LPM Add: Invalid value.\n");*/
        return 1;
    }

    /* Check if nodes can be freed from pending list */
    if(lpm->pending_list)
    {
        if(lpm->pending_count == lpm->pending_max)
        {
            adk_rcu_distributed_wait();
            lpm_pending_list_node_free(lpm->pending_list);
            lpm->pending_list = NULL;
            lpm->pending_count = 0;
        }
        else if(adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
        {
            lpm_pending_list_node_free(lpm->pending_list);
            lpm->pending_list = NULL;
            lpm->pending_count = 0;
        }
    }

    /* Find & replace the existing value if the key is already added*/
    node = lpm_node_find(lpm, key, depth);
    if (node != NULL && node->data != (adk_handle_t)(ADK_LPM_VALUE_INVALID))
    {
    	/* Uncomment debug_print("LPM Add: entry already present, replacing the value.\n"); */
        node->data = value;
	return 0;
    }

    /* Traverese the tree byte wise, create nodes if not present*/
    parent = &lpm->root_node;
    for (count = 0, byte_p = key; ; ++count,++byte_p)
    {
        node = *parent;
        if (node == NULL)
        {
            /* Create the node */
            node = (adk_lpm_node_t *)
                   ADK_DP_SHM_CALLOC(mod_id, obj_id, 1, sizeof(adk_lpm_node_t), 0);
            if (node == NULL)
            {
                /* Allocation failed.*/
                return 1;
            }
            /* Uncomment debug_print("LPM Add: Byte node created.\n"); */
            node->data = (adk_handle_t)(ADK_LPM_VALUE_INVALID);

            /* Link the created node to parent*/
            *parent = node;
        }

        ++node->use_count;

        /* End of byte lookup*/
        if (count == byte_depth)
        {
            break;
        }

        /* next node */
        parent = &node->byte_child[*byte_p];
    }

    /* Traverese the tree bit wise, create the node if not present*/
    for (count = 0; count < bit_depth; ++count)
    {
        bit = (*byte_p >> (7-count)) & 0x1;
        node = (*parent)->bit_child[bit];
        if (node == NULL)
        {
            /* Create node */
            node = (adk_lpm_node_t *)
                   ADK_DP_SHM_CALLOC(mod_id, obj_id, 1, sizeof(adk_lpm_node_t), 0);
            if (node == NULL)
            {
                /* Allocation failed*/
                return 1;
            }
            /* Uncomment debug_print("LPM Add: Bit node created.\n"); */
            node->data = (adk_handle_t)(ADK_LPM_VALUE_INVALID);

            /* Link in to the parent */
            (*parent)->bit_child[bit] = node;
        }
        ++node->use_count;
        parent = &(*parent)->bit_child[bit];
    }

    /* Reached the end. Set the data at the node we have reached */
    node->data = value;
    return 0;
}

int adk_lpm_delete(adk_lpm_handle_t lpm_handle, uint8_t *key, uint16_t depth)
{
    adk_lpm_t * lpm = (adk_lpm_t *)lpm_handle;
    adk_lpm_node_t *node, *next, **prev_ptr, **next_prev_ptr;
    uint8_t *byte_p, count, byte_depth, bit_depth, bit_value;

    //debug_print("LPM Delete\n");

    if (!lpm_handle)
    {
        //debug_print("NULL LPM handle\n");
        return -1;
    }

    /* Check if nodes can be freed from pending list */
    if(lpm->pending_list)
    {
        if(lpm->pending_count == lpm->pending_max)
        {
            adk_rcu_distributed_wait();
            lpm_pending_list_node_free(lpm->pending_list);
            lpm->pending_list = NULL;
            lpm->pending_count = 0;
        }
        else if(adk_rcu_distributed_ver_verify() == ADK_SUCCESS)
        {
            lpm_pending_list_node_free(lpm->pending_list);
            lpm->pending_list = NULL;
            lpm->pending_count = 0;
        }
    }

    /* Find the end node and remove the value */
    node = lpm_node_find(lpm, key, depth);

    if (node != NULL && node->data != (adk_handle_t)(ADK_LPM_VALUE_INVALID))
    {
        node->data = (adk_handle_t)(ADK_LPM_VALUE_INVALID);
    }
    else
    {
        //debug_print("LPM delete: node not found.\n");
        return 1;
    }

    /* Now traverse the tree again as before, decrementing the use
     * count of each node.  Free back nodes as necessary. */
    node = lpm->root_node;
    prev_ptr = &lpm->root_node;
    byte_p = key;

    /*Traverse byte wise*/
    byte_depth = depth / 8;
    for(count = 0; count < byte_depth; ++count, ++byte_p)
    {
        /* Find the next node now. We might free this node. */
	next_prev_ptr = &node->byte_child[(int) *byte_p];
	next = *next_prev_ptr;

        /* Decrease the use count and free the node if it
	 * reaches zero. */
        --node->use_count;
        //debug_print("LPM delete: Decresing use count for node :%"ADKPRIh"\n", node->data);

        if (node->use_count == 0)
        {
#if 0
            debug_print("LPM delete: freeing node :%"ADKPRIh"\n",node->data);
	    ADK_NVM_FREE(node);
            if (prev_ptr != NULL)
            {
	        *prev_ptr = NULL;
            }
            next_prev_ptr = NULL;
#else
            /*break the link from parent*/
            *prev_ptr = NULL;
            //debug_print("LPM delete: moving node :%"ADKPRIh" to pending list\n", node->data);
            node->pending_next = lpm->pending_list;
            lpm->pending_list = node;
            lpm->pending_count++;
            adk_rcu_distributed_ver_update();
            return 0;
#endif
        }

        /* Update pointers */
        node = next;
	prev_ptr = next_prev_ptr;
    }

    /*Traverse bit wise*/
    bit_depth = depth % 8;
    for(count = 0; count < bit_depth; count++)
    {
        /* Find the next node through bit lookup. */
        bit_value = (*byte_p >> (7-count)) & 0x1;
	next_prev_ptr = &node->bit_child[bit_value];
	next = *next_prev_ptr;

        /* Decrease the use count and free the node if it
	 * reaches zero. */
        --node->use_count;
        //debug_print("LPM delete: Decresing use count for node :%"ADKPRIh"\n", node->data);

        if (node->use_count == 0)
        {
#if 0
            //debug_print("LPM delete: freeing node :%"ADKPRIh"\n",node->data);
	    ADK_NVM_FREE(node);
            if (prev_ptr != NULL)
            {
		*prev_ptr = NULL;
	    }
            next_prev_ptr = NULL;
#else
            /*break the link from parent*/
            *prev_ptr = NULL;
            //debug_print("LPM delete: moving node :%"ADKPRIh" to pending list\n", node->data);
            node->pending_next = lpm->pending_list;
            lpm->pending_list = node;
            lpm->pending_count++;
            adk_rcu_distributed_ver_update();
            return 0;
#endif
	}

	/* Update pointers */
	node = next;
	prev_ptr = next_prev_ptr;
    }

    /* Reached the end node. Decrease the use count & free if it reaches 0*/
    --node->use_count;
    //debug_print("LPM delete: Decresing use count for end node\n");
    if (node->use_count == 0)
    {
#if 0
        debug_print("LPM delete: freeing end node\n");
        ADK_NVM_FREE(node);
        if (prev_ptr != NULL)
        {
	    *prev_ptr = NULL;
	}
        next_prev_ptr = NULL;
#else
        /*break the like from parent*/
        *prev_ptr = NULL;
        //debug_print("LPM delete: moving node :%"ADKPRIh" to pending list\n", node->data);
        node->pending_next = lpm->pending_list;
        lpm->pending_list = node;
        lpm->pending_count++;
        adk_rcu_distributed_ver_update();
        return 0;
#endif
    }

    /* Removed successfully */
    return 0;
}

adk_handle_t adk_lpm_lookup(adk_lpm_handle_t lpm_handle, uint8_t *key)
{
    adk_lpm_t * lpm = (adk_lpm_t *)lpm_handle;
    adk_lpm_node_t *node=NULL, *prev=NULL, *tmp=NULL;
    adk_handle_t value = (adk_handle_t)(ADK_LPM_VALUE_INVALID);
    uint8_t *byte_p, bit, count, byte, bit_count;

    /* TO DO: Uncomment debug_print("LPM Lookup\n"); */

    if((lpm == NULL) || (lpm->root_node == NULL))
    {
        //debug_print("LPM Lookup: NULL LPM or NULL root node\n");
        return (adk_handle_t)(ADK_LPM_VALUE_INVALID);
    }

    /* Traverse the tree byte wise*/
    node = lpm->root_node;
    for(count = 0, byte_p = key; node != NULL; count ++, byte_p++)
    {
    	/* TO DO: Uncomment  debug_print("LPM Lookup: cur node data:%"ADKPRIh"\n", node->data); */
        if(node->data != (adk_handle_t)(ADK_LPM_VALUE_INVALID))
        {
            /*Node with a valid data*/
	    value = node->data;
        }
        else if(prev)
        {
            /* Intermediate node, traverse bit wise as well &
             * update lpm match value if any*/
            tmp = prev;
            for(bit_count = 0; bit_count < 7; bit_count++)
            {
                /* Jump to the next node */
                bit = (byte >> (7-bit_count) ) & 0x1;
                tmp = tmp->bit_child[bit];
                /* TO DO: Uncomment  debug_print("LPM Lookup: bit lookup:%x\n", bit); */
                if(tmp == NULL)
                {
                    break;
                }
                /* TO DO: Uncomment  debug_print("LPM Lookup: cur node data:%"ADKPRIh"\n", tmp->data); */
                if(tmp->data != (adk_handle_t)(ADK_LPM_VALUE_INVALID))
                {
                    value = tmp->data;
                }
            }
        }

        if(count*8 == lpm->width)
        {
            /*Reached the end of tree*/
            return value;
        }
        prev = node;
        byte = *byte_p;
        node = node->byte_child[byte];
        /* TO DO: Uncomment  debug_print("LPM Lookup: byte lookup:%x\n", byte); */
    }

    /* Traverse the tree bit wise*/
    node = prev;
    for(count = 0; count < 7; count ++)
    {
        /* Jump to the next node */
	bit = (byte >> (7-count) ) & 0x1;
	node = node->bit_child[bit];
	/* TO DO: Uncomment debug_print("LPM Lookup: bit lookup:%x\n", bit); */
        if(node == NULL)
        {
            break;
        }
        /* TO DO: Uncomment  debug_print("LPM Lookup: cur node data:%"ADKPRIh"\n", node->data); */
        if(node->data != (adk_handle_t)(ADK_LPM_VALUE_INVALID))
        {
	    value = node->data;
        }
    }

    return value;
}

static void
adk_lpm_dump_bit_node(adk_lpm_node_t *node, uint8_t *key, uint16_t depth,
                      adk_lpm_handle_t lpm_handle, adk_lpm_dump_cb_fn_t cb,
                      void *rock)
{
    uint16_t i, bit_depth = depth%8, byte_depth = depth/8;

    if(node->data != (adk_handle_t)ADK_LPM_VALUE_INVALID)
    {
        cb(lpm_handle, rock, key, depth, node->data);
    }

    for(i=0; i<2; ++i)
    {
        if(node->bit_child[i])
        {
            key[byte_depth] |= (i << (7 - bit_depth));
            adk_lpm_dump_bit_node(node->bit_child[i], key, depth + 1,
                                  lpm_handle, cb, rock);
        }
    }

    return;
}

static void
adk_lpm_dump_byte_node(adk_lpm_node_t *node, uint8_t *key, uint16_t depth,
                       adk_lpm_handle_t lpm_handle, adk_lpm_dump_cb_fn_t cb,
                       void *rock)
{
    uint16_t i;

    if(node->data != (adk_handle_t)ADK_LPM_VALUE_INVALID)
    {
        cb(lpm_handle, rock, key, depth, node->data);
    }

    /*Dump the bit children*/
    for(i=0; i<2; ++i)
    {
        if(node->bit_child[i])
        {
            key[depth/8] = (i << 7);
            adk_lpm_dump_bit_node(node->bit_child[i], key, depth + 1,
                                  lpm_handle, cb, rock);
        }
    }

    /*Dump the byte children*/
    for(i=0; i<256; ++i)
    {
        if(node->byte_child[i])
        {
            key[depth/8] = i;
            adk_lpm_dump_byte_node(node->byte_child[i], key, depth + 8,
                                   lpm_handle, cb, rock);
        }
    }

    return;
}

void adk_lpm_dump(adk_lpm_handle_t lpm_handle, adk_lpm_dump_cb_fn_t cb,
                  void *rock)
{
    adk_lpm_t * lpm = (adk_lpm_t *)lpm_handle;
    uint8_t *key;

    //debug_print("LPM Dump\n");

    if(lpm && (lpm->root_node))
    {
        key = (uint8_t *)ADK_DP_SHM_CALLOC(ADK_MISC_MODULE_ID, 2, 1, 1+(lpm->width)/8, 0);
        adk_lpm_dump_byte_node(lpm->root_node, key, 0, lpm_handle, cb, rock);
        ADK_NVM_FREE(key);
    }

    return;
}


