/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#ifndef __ADK_PKTIO_H__
#define __ADK_PKTIO_H__

#include "adk_sif_cfg_include.h"
#include "adk_sif_cfg_pipeline.h"

#include <adk_ph_dev.h>
#include <arpa/inet.h>

/*
 * This file contains exported data structures & APIs for packet input/output
 * implementation.
 * 
 * All types of packet io are implemented using non-blocking modes.
 */

/* Handle for packet io instance */
typedef adk_handle_t adk_pktio_hdl_t;

/* structure for packet io task is same as RTE's task.
 * This helps in using it across nca-based packet io
 * instances seamlessly.
 */
typedef struct rte_mbuf adk_pktio_task_t;


/* Enumeration for packet io APIs return codes. 
 */
typedef enum {
    ADK_PKTIO_ERROR 		= 0,
    ADK_PKTIO_SUCCESS 		= 1,
    ADK_PKTIO_QUEUE_EMPTY 	= 2,
    ADK_PKTIO_NO_MEMORY
} adk_pktio_ret_t;

/* Enumeration for different types of packet io's
 * supported.
 *
 * NCAV2_LITE is special implementation of NCAv2 APIs
 * optimized for scaling.
 *
 * BUFFER type is special implementation with pre-defined
 * packet buffers.
 *
 */
typedef enum {
    ADK_PKTIO_TYPE_NCAV2        = 0,
    ADK_PKTIO_TYPE_NCAV2_LITE   = 1,
    ADK_PKTIO_TYPE_DPDK_BUFF    = 2,
    ADK_PKTIO_TYPE_SOCKET       = 3,
    ADK_PKTIO_TYPE_QUEUE        = 4,
    ADK_PKTIO_TYPE_TUNTAP       = 5,
    ADK_PKTIO_TYPE_PCAP         = 6,
    ADK_PKTIO_TYPE_DPDK_PORT    = 7,
    ADK_PKTIO_TYPE_DPDK_RING    = 8,
    ADK_PKTIO_TYPE_PH           = 9
} adk_pktio_type_t;

/* Enumeration for packet flow types used in
 * BUFFER based packet io instances.
 */
typedef enum {
    ADK_PKTIO_BUFF_TYPE_TERM_W_REASSEMBLY       = 1,
    ADK_PKTIO_BUFF_TYPE_TERM_WO_REASSEMBLY      = 2,
    ADK_PKTIO_BUFF_TYPE_TN_TO_WIFI_FWD          = 3,
    ADK_PKTIO_BUFF_TYPE_WIFI_TO_TN_FWD          = 4,
    ADK_PKTIO_BUFF_TYPE_ORIG                    = 5
} adk_pktio_buffer_flow_type_t;


/* 
 * Structure for configuring socket based packet
 * io instances.
 */
typedef struct {
    struct in_addr remote_ip;
    struct in_addr local_ip;
    int udp_port;
    uint32_t inPort;
} adk_pktio_socket_config_t;

/*
 * Structure for configuring pcap based packet
 * io instances.
 */
typedef struct {
    char dev_name[10];
} adk_pktio_pcap_config_t;

/*
 * Structure for configuring queue based packet
 * io instances.
 */
typedef struct {
    int size;
} adk_pktio_queue_config_t;

/* 
 * Structure for configuring tun/tap based packet
 * io instances.
 */
typedef struct {
    bool isTun;
    char dev_name[20];
    char dev_addr[6];
} adk_pktio_tuntap_config_t;

/* 
 * Structure for configuring dpdk based packet
 * io instances.
 */
typedef struct {
    struct rte_mempool *pktmbuf_pool;
} adk_pktio_dpdk_config_t;

/* 
 * Structure for configuring a packet io instance.
 */
typedef struct {
    adk_pktio_type_t type;
    union {
        adk_pktio_socket_config_t       sock_config;
        adk_pktio_pcap_config_t         pcap_config;
        adk_pktio_queue_config_t        queue_config;
        adk_pktio_tuntap_config_t       tuntap_config;
        adk_pktio_dpdk_config_t         dpdk_config;
    }u;
} adk_pktio_cfg_t;

/* Structure for configuring dpdk buffers for pktio buffers */
typedef struct adk_dpdk_pkt_buffers_config_s {
    uint32_t *adkPacketsInBuffer;
    uint32_t *adkPacketsSizeInBuffer;
    char (*adkPktBuffs)[RTE_MAX_MBUFS_IN_BUFFER][RTE_MBUF_DEFAULT_BUF_SIZE];
} adk_dpdk_pkt_buffers_config_t;

/* 
 * Function to configure a packet io instance of given type.
 * Packet IO handle is returned to caller which must be used in
 * send/receive APIs.
 */
adk_pktio_ret_t adk_pktio_config(adk_pktio_cfg_t cfg, adk_pktio_hdl_t *hdl);

/*
 * Function to receive packet on a given packet io instance.
 * Task is returned back, if packet is available.
 * In case when no packet is avaiable, QUEUE_EMPTY code is returned back.
 * 
 * All Packet IOs are implemented using non-blocking mode.
 */
adk_pktio_ret_t adk_pktio_recv(adk_pktio_hdl_t hdl, void *userData,
                               adk_pktio_task_t **task_p, uint8_t *task_count,
                               uint32_t *task_mask);

/*
 * Function to send packet on a given packet io instance.
 */
adk_pktio_ret_t adk_pktio_send(adk_pktio_hdl_t hdl, adk_pktio_task_t **task,
                               uint8_t task_count, uint32_t task_mask);

adk_pktio_ret_t adk_pktio_dpdk_port_flush(int coreIndex);

adk_pktio_ret_t adk_pktio_dpdk_buff_flush(adk_pktio_hdl_t hdl, int coreIndex);

void adk_pktio_task_buffer_alloc(adk_pktio_task_t **task_p);

void adk_pktio_task_buffer_free(adk_pktio_task_t *task_p);

/*
 * Function to stage pktio rx and tx on a separate lcore
 */
adk_status_t
adk_pktio_mod_rx(void **task_p, 
                 uint8_t task_count,
                 uint32_t task_mask,
                 int coreIndex);

adk_status_t
adk_pktio_mod_tx(void **task_p, 
                 uint8_t task_count,
                 uint32_t task_mask,
                 __attribute__((unused)) void *ring,
                 __attribute__((unused)) adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                 int coreIndex);

adk_status_t
adk_pktio_mod_ring_tx(void **task_p, 
                      __attribute__((unused)) uint8_t task_count,
                      uint32_t task_mask,
                      void *ring,
                      adk_sif_cfg_pipeline_info_t *next_stage_entry, 
                      int __attribute__((unused)) coreIndex);

void adk_pktio_buffs_stats_print(void);
#endif
