/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <string.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#include <errno.h>
#include <net/if_arp.h>

#include "adk_include.h"
#include "adk_pktio.h"
#include "adk_pktio_pvt.h"

#define BUFSIZE 2048
#define BUF_HEADROOM 64

static int tuntap_config(char *dev, int flags, char *hw_addr) 
{
    struct ifreq ifr;
    int fd, err;

    if( (fd = open("/dev/net/tun", O_RDWR)) < 0 )
    {
        debug_print("\n[pktio] Could not open the device!\n");
        return fd;
    }

    ADK_MEMSET(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = flags;
    if(*dev)
    {
        ADK_MEMCPY(ifr.ifr_name, dev, IFNAMSIZ-1);
        if(ifr.ifr_name[IFNAMSIZ - 1] != '\0')
            ifr.ifr_name[IFNAMSIZ - 1] = '\0';
    }
    if( (err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 )
    {
        debug_print("\n[pktio] Failed to attach to device!\n");
        close(fd);
        return err;
    }

    /*set mac address for tap interface*/
    if(IFF_TAP & flags)
    {
        ADK_MEMCPY(ifr.ifr_hwaddr.sa_data, hw_addr, sizeof(char)*6);
        ifr.ifr_hwaddr.sa_family =  ARPHRD_ETHER;
        if( (err = ioctl(fd, SIOCSIFHWADDR, (void *) &ifr)) < 0 )
        {
            debug_print("\n[pktio] Failed to set hw addr to device!\n");
            close(fd);
            return err;
        }
    }

    return fd;
}

adk_pktio_ret_t adk_pktio_tuntap_config(adk_pktio_cfg_t cfg,
			    adk_pktio_info_t *hdl)
{
    int fd, flags = IFF_NO_PI;

    flags += (cfg.u.tuntap_config.isTun == TRUE) ? IFF_TUN : IFF_TAP;

    fd = tuntap_config(cfg.u.tuntap_config.dev_name, flags,
		   cfg.u.tuntap_config.dev_addr);
    if(fd == -1)
    {
	perror("Failed to create tun/tap interface\n");
	return ADK_PKTIO_ERROR;
    }

    /* make tun/tap interface as non-blocking */
    flags = fcntl(fd, F_GETFL, 0);
    if(fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
    {
	perror("Failed to make tun/tap interface as non-blocking\n");
        close(fd);
	return ADK_PKTIO_ERROR;
    }
  
    /* store the information */
    hdl->u.tuntap_info.fd = fd;
    hdl->u.tuntap_info.flags = flags;

    return ADK_PKTIO_SUCCESS;
}
    
static unsigned char tap_rx_buf[BUFSIZE];

adk_pktio_ret_t adk_pktio_tuntap_recv(adk_pktio_hdl_t hdl,
                          __attribute__((unused)) void *userData,
                          adk_pktio_task_t **task,
                          uint8_t *task_count,
                          uint32_t *task_mask)
{
    adk_pktio_info_t *info = &((adk_pktio_internal_t *)hdl)->info;
    int fd = info->u.tuntap_info.fd;
    unsigned char *buf = NULL;
    int recvlen = 0;
    adk_pktio_task_t *tap_task_p = NULL;

    recvlen = read(fd, tap_rx_buf, BUFSIZE);
    if(recvlen == -1) {
	if(errno == EAGAIN || errno == EWOULDBLOCK) {
	    return ADK_PKTIO_QUEUE_EMPTY; /* try again - socket empty */
	} else {
	    perror("tuntap : receive failed!\n");
	    return ADK_PKTIO_ERROR;
	}
    }
    
    tap_task_p = (adk_pktio_task_t *)rte_pktmbuf_alloc(adkPktmbufPool_g);
    if(!tap_task_p)
    {
        debug_print("Error allocating task for tap interface receive. Exiting...\n");
        return ADK_PKTIO_ERROR;
    }

    debug_print("\ntuntap : received %d bytes\n", recvlen);

    adk_pkt_metadata_init(tap_task_p, ADK_OS_ORIGIN_DOMAIN, 0, ADK_TASK_TYPE_FWD);
    ADK_TASK_PDU_SIZE((tap_task_p)) = 0;
    ADK_PDU_TAIL_PUSH((tap_task_p), recvlen);
    buf = rte_pktmbuf_mtod((struct rte_mbuf *) tap_task_p, void*);
    ADK_MEMCPY(buf, tap_rx_buf, recvlen);

    *task = tap_task_p;
    *task_count = 1;
    *task_mask = 0x1;

    return ADK_PKTIO_SUCCESS;
}

adk_pktio_ret_t adk_pktio_tuntap_send(adk_pktio_hdl_t hdl,
		          adk_pktio_task_t **task,
                          __attribute__((unused)) uint8_t task_count,
                          __attribute__((unused)) uint32_t task_mask)
{
    adk_pktio_info_t *info = &((adk_pktio_internal_t *)hdl)->info;
    adk_pktio_task_t *chained_task = task[0];
    int fd = info->u.tuntap_info.fd, i;
    int sendlen;
    unsigned char *buf_tmp;
    struct iovec buff_ptr[10]; /* number of chained buffers*/
    int buf_size = 0;

    if(task[0]->nb_segs>1)
    {
        for(i = 0; i<task[0]->nb_segs; i++)
        {
            buff_ptr[i].iov_base = 
                (void *) ((uint8_t *)(chained_task->buf_addr) + chained_task->data_off);
            buff_ptr[i].iov_len = chained_task->data_len;
            chained_task = (adk_pktio_task_t *)chained_task->next;
        }

        sendlen = writev(fd, buff_ptr, task[0]->nb_segs); 
    }
    else
    {
        buf_tmp = rte_pktmbuf_mtod((struct rte_mbuf *)(task[0]), unsigned char *);
        buf_size = ADK_TASK_PDU_SIZE((task[0]));

        sendlen = write(fd, buf_tmp, buf_size);
    }
    if(sendlen == -1) {
	perror("tuntap : Send failed!");
	return ADK_PKTIO_ERROR;
    }
    debug_print("tuntap : sent %d bytes\n", sendlen);
    
    rte_pktmbuf_free((struct rte_mbuf *)task[0]);
    /* success */
    return ADK_PKTIO_SUCCESS;
}
