/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#include <adk_queue.h>
#include <adk_mutex.h>

typedef struct _adk_sp_mc_queue_node_s
{
    void *data;
} adk_sp_mc_queue_node_t;

typedef struct _adk_sp_mc_queue_s
{
    uint32_t head; /* producer index */
    uint32_t pad[15];
    uint32_t tail; /* consumer index */
    uint32_t size;
    adk_sp_mc_queue_node_t *root;
} adk_sp_mc_queue_t;

adk_queue_handle_t 
adk_sp_mc_queue_create(uint32_t size)
{
    adk_sp_mc_queue_t *newQueue = NULL;
    adk_sp_mc_queue_node_t *newList = NULL;
    adk_queue_handle_t qHdl;

    newQueue = (adk_sp_mc_queue_t *) 
	ADK_NVM_CALLOC(ADK_MISC_MODULE_ID, 0, 1, sizeof(adk_sp_mc_queue_t));
    if(newQueue == NULL)
    {
	debug_print("Error: Failed to create Queue!\n");
	return 0;
    }

    /* this will make all data entries as NULL */
    newList = (adk_sp_mc_queue_node_t *)
    		ADK_NVM_CALLOC(
		ADK_MISC_MODULE_ID, 0, size, sizeof(adk_sp_mc_queue_node_t));
    if(newList == 0)
    {
	debug_print("Error: Failed to allocate memory for queue!\n");
	return 0;
    }

    newQueue->root = newList;
    newQueue->head = 0; /* initial value */
    newQueue->tail = 0; /* initial value */
    newQueue->size = size;
    qHdl = (adk_queue_handle_t)newQueue;
    debug_print("Successfully created queue of size - %d\n", size);

    return qHdl;
}

int adk_sp_mc_queue_delete(adk_queue_handle_t qHdl)
{
    adk_sp_mc_queue_t *queue_p = (adk_sp_mc_queue_t *)qHdl;
    ADK_NVM_FREE(queue_p->root);
    ADK_NVM_FREE(queue_p);
    return 1;
}

int adk_sp_mc_queue_dequeue(adk_queue_handle_t qHdl,
		      void **data)
{
    adk_sp_mc_queue_t *queue_p = (adk_sp_mc_queue_t *)qHdl;
    adk_sp_mc_queue_node_t *node = NULL;

    node = &(queue_p->root[queue_p->tail]);

    /* check if packet is available for processing */
    if(NULL == node->data) {
	/* queue is emtpy */
	*data = NULL;
	return 0;
    }
    *data = node->data;
    if(adk_compare_and_set(&node->data, (int)(*data), (int)NULL))
    {
	/* other core got the packet for processing
	 * so retry for next packet.
	 */
	*data = NULL;
	return 0;
    }
    queue_p->tail++;
    if(queue_p->tail == queue_p->size){
	/* circular buffer - roll over to start */
	queue_p->tail = 0;
    }
    /* complete memory operation */
    __asm__ __volatile__("dmb":::"memory");

    debug_print("Successfully dequeued from queue!\n");
    /* dequeue successful */
    return 1;
}

int adk_sp_mc_queue_enqueue(adk_queue_handle_t qHdl,
		      void *data)
{
    adk_sp_mc_queue_t *queue_p = (adk_sp_mc_queue_t *)qHdl;
    adk_sp_mc_queue_node_t *node = NULL;

    /* read from head */
    node = &(queue_p->root[queue_p->head]);
    if(NULL != node->data) {
	/* queue is full */
	return 0;
    }
    node->data = data;
    queue_p->head++;
    if(queue_p->head == queue_p->size) {
	/* circular buffer - roll over to start */
	queue_p->head = 0;
    }

    debug_print("Successfully inserted in queue\n");
    /* enqueue successful */
    return 1;
}

uint32_t adk_sp_mc_queue_get_size(adk_queue_handle_t qHdl)
{
    adk_sp_mc_queue_t *queue_p = (adk_sp_mc_queue_t *)qHdl;

    return (queue_p->size);
}
