/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
#ifndef __ADK_SHM__
#define __ADK_SHM__

#include <sys/syscall.h>
#include "adk_include.h"
#include "adk_shm.h"

adk_status_t
adk_shm_create(
        const char *name,
        const size_t mem_size,
        uint32_t lookupMethod,
        adk_shm_handle_t **shm_hdl)
{
    adk_status_t retValue = ADK_SUCCESS;
    const struct rte_memzone *mz = NULL;
    adk_sys_config_info_t *sysConf_p;
    uint32_t socketId = 0;
    uint32_t lcoreId;
    uint32_t i;

    /* Validation */
    if (!name)
    {
        ADK_FORCE_LOG_ERROR("Invalid pointer\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_shm_create_return;
    }

    mz = rte_memzone_lookup(name);
    if (mz)
    {
        /* Update user pointer */
        *shm_hdl = mz->addr;
        ADK_FORCE_LOG_INFO("Shared memory is already created: '%s'\n", name);
        retValue = ADK_WARNING_SEV_LEVEL_SET(ADK_W_SHM_EXISTS);
        goto adk_shm_create_return;
    }

    /* Get socket info */
    if (ADK_SHM_LOOKUP_LCORE == lookupMethod)
    {
        socketId = rte_socket_id();
    } else
    {
        sysConf_p = (adk_sys_config_info_t *)adkSysShmConf_p->sysConfHandle;
        /*
         * Look for the first core that is in use and us that
         */
        for (i = 0; i < ADK_MAX_NUM_CORES; i++)
        {
            lcoreId = sysConf_p->lcoreInfo.lcore[i].id;
            if (0xff != lcoreId)
            {
                socketId = rte_lcore_to_socket_id(lcoreId);
                break;
            }
        }
    }

    /* Allocate a memory zone */
    mz = rte_memzone_reserve(name, mem_size, socketId, 0);
    if (mz == NULL)
    {
        ADK_FORCE_LOG_ERROR("Unable create shared memory - '%s', "
                            "errno: '%s' (%d)\n", name,
                            rte_strerror(rte_errno), rte_errno);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto adk_shm_create_return;
    }

    /* Initialize the reverved shared memory */
    ADK_MEMSET(mz->addr, 0, mem_size);

    /* Update user pointer */
    *shm_hdl = mz->addr;

adk_shm_create_return:
    return retValue;
}

adk_status_t
adk_shm_lookup(
        const char *name,
        adk_shm_handle_t **shm_hdl)
{
    adk_status_t retValue = ADK_SUCCESS;
    const struct rte_memzone *mz = NULL;

    /* Validation */
    if (!name)
    {
        ADK_FORCE_LOG_ERROR("Invalid pointer\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto adk_shm_lookup_return;
    }

    /* Get memzone handle */
    mz = rte_memzone_lookup(name);
    if (mz == NULL)
    {
        ADK_FORCE_LOG_ERROR("Unable attach to shared memory - %s\n", name);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto adk_shm_lookup_return;
    }

    /* Update user pointer */
    *shm_hdl = mz->addr;

adk_shm_lookup_return:
    return retValue;
}

adk_status_t
adk_shm_destroy(
        const char *name)
{
    adk_status_t retValue = ADK_SUCCESS;
    const struct rte_memzone *mz = NULL;
    int             rteErr = 0;

    /* Validation */
    if (!name)
    {
        ADK_FORCE_LOG_ERROR("Invalid pointer\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _return;
    }

    /* Get memzone handle */
    mz = rte_memzone_lookup(name);
    if (mz == NULL)
    {
        ADK_FORCE_LOG_ERROR("Unable find shared memory - %s\n", name);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _return;
    }

    rteErr = rte_memzone_free(mz);
    if (0 != rteErr)
    {
        ADK_FORCE_LOG_ERROR("Unable remove shared memory - %s, "
                            "error code: %s (%d)\n", name,
                            rte_strerror(-rteErr), rteErr);
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        goto _return;
    }

_return:
    return retValue;
}

#endif
