/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2018 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/

#ifdef ADK_TIMER

#include "adk_include.h"
#include "adk_timer.h"

static adk_timer_wheel_instance_t *cpu_timer_wheel[ADK_MAX_NUM_CORES];
static adk_sys_config_worker_lcore_info_t  *workerInfo_p;

static void adk_timer_per_tick_bookkeeping (uint16_t coreIndex);

void adk_timer_init(uint16_t coreIndex, uint32_t cpu_tick)
{
    adk_status_t    retValue = ADK_SUCCESS;
    uint16_t i;

    cpu_timer_wheel[coreIndex] = ADK_DP_SHM_CALLOC(ADK_MISC_MODULE_ID, 1,
				 1, sizeof(adk_timer_wheel_instance_t), 0);
    cpu_timer_wheel[coreIndex]->cpu_tick_prev = cpu_tick;
    /*Timer tick is 1 milli sec*/
    cpu_timer_wheel[coreIndex]->cpu_ticks_per_timer_tick = rte_get_tsc_hz() / 1000;
    /* Initialize the node free list */
    cpu_timer_wheel[coreIndex]->timer_node_free_list = 0;
    for(i=0; i<ADK_TIMER_MAX_NUMBER; ++i)
    {
        cpu_timer_wheel[coreIndex]->timer_node_pool[i].cur = i;
        cpu_timer_wheel[coreIndex]->timer_node_pool[i].next_free = i+1;
    }
    /* Initialize the list head of each timer wheel index*/
    for(i=0; i<ADK_TIMER_WHEEL_SIZE; ++i)
    {
        cpu_timer_wheel[coreIndex]->timer_wheel[i].list_head = ADK_TIMER_MAX_NUMBER;
    }

    /*
     * Lookup the worker info so we can get the NUMA socket to allocate the
     * memzone from
     */
    retValue = adk_sys_config_worker_lcore_info_get(&workerInfo_p);
    if (ADK_SUCCESS != retValue)
    {
           rte_panic("  {%s} Failed to get worker core info %s \n",
                     __func__, adk_sys_error_string(retValue));
    }

    /*
     * Initialize the early expiry rings
     */
    for (i = 0; i < workerInfo_p->numLcores; ++i)
    {
        struct rte_ring  *ring_p;
        char             ring_name[RTE_RING_NAMESIZE];

        /* No early list needed for same lcore timer wheel */
        if (workerInfo_p->lcore[i].index == coreIndex)
        {
            continue;
        }

        ADK_MEMSET(ring_name, 0, sizeof(char) * RTE_RING_NAMESIZE);
        snprintf(ring_name, RTE_RING_NAMESIZE,
                 "TW_EARLY_EXPIRY_%u_%u",coreIndex, i);
        ring_p = rte_ring_create(ring_name, 1024, rte_socket_id(), 0);
        if (ring_p == NULL)
        {
           rte_panic("  {%s} Failed to create ring FAILED %s \n",
                     __func__, rte_strerror(rte_errno));
        }

        cpu_timer_wheel[coreIndex]->early_expiry_list[workerInfo_p->lcore[i].index]
        = ring_p;
    }
}

int adk_timer_create(uint16_t coreIndex,
                     adk_timer_info_t timer_info,
		     adk_timer_cb_fn_t cbFn,
		     void *rock,
		     adk_timer_hdl_t *timer_hdl)
{
    uint8_t insert_wheel_index;
    adk_timer_list_node_info_t * new_node;
    adk_timer_wheel_instance_t * wheel_instance = cpu_timer_wheel[coreIndex];

    if(unlikely(NULL == cpu_timer_wheel[coreIndex]))
    {
        debug_print("\nTimer wheel not initialized. Failed to create timer");
        return 0;
    }

    if(unlikely(ADK_TIMER_MAX_NUMBER == wheel_instance->timer_node_free_list))
    {
        debug_print("\nReached max num of timers");
        return 0;
    }

    /*calculate the index to be inserted in timer wheel*/
    insert_wheel_index = wheel_instance->cur_timer_wheel_index +
                         (timer_info.interval_milli_sec & 0xFF);

    /*Get free node from the pool*/
    new_node = &(cpu_timer_wheel[coreIndex]->
                 timer_node_pool[wheel_instance->timer_node_free_list]);
    wheel_instance->timer_node_free_list = new_node->next_free;

    /*prepare the node*/
    new_node->timer_wheel_index = insert_wheel_index;
    new_node->timer_hash_offset = (timer_info.interval_milli_sec >> 8);
    new_node->cbFn = cbFn;
    new_node->rock = rock;
    new_node->version_id++;

    debug_print("CPU Id:%d, Current wheel index:%d, Inserting at wheel index:%d, \
           with hash_offset:%d\n",coreIndex, wheel_instance->cur_timer_wheel_index,
           insert_wheel_index, new_node->timer_hash_offset);

    /*Insert at the head of the list*/
    if(wheel_instance->timer_wheel[insert_wheel_index].list_head
       < ADK_TIMER_MAX_NUMBER)
    {
        /*Non-empty list*/
        new_node->prev = ADK_TIMER_MAX_NUMBER;
        new_node->next = wheel_instance->timer_wheel
                         [insert_wheel_index].list_head;
        wheel_instance->timer_wheel[insert_wheel_index].list_head
                       = new_node->cur;
        cpu_timer_wheel[coreIndex]->timer_node_pool[new_node->next].prev
                       = new_node->cur;
    }
    else
    {
        /*Empty list*/
        new_node->next = ADK_TIMER_MAX_NUMBER;
        new_node->prev = ADK_TIMER_MAX_NUMBER;
        wheel_instance->timer_wheel[insert_wheel_index].list_head
                       = new_node->cur;
    }

    *timer_hdl = (adk_timer_hdl_t)
                 ((new_node->cur << 16) |
                  (coreIndex << 8) |
                  new_node->version_id);

    return 1;
}

int adk_timer_delete(uint16_t coreIndex, adk_timer_hdl_t timer_hdl)
{
    adk_timer_wheel_instance_t * wheel_instance;
    adk_timer_list_node_info_t * del_node;
    uint8_t del_wheel_index, version_id, dst_cpu_id;
    uint16_t del_node_index;

    del_node_index = timer_hdl >> 16;
    if(del_node_index >= ADK_TIMER_MAX_NUMBER)
    {
        debug_print("\nDeleting invalid timer\n");
        return 1;
    }
    
    dst_cpu_id = (timer_hdl >> 8) & 0xFF;
    wheel_instance = cpu_timer_wheel[dst_cpu_id];
    del_node = &(wheel_instance->timer_node_pool[del_node_index]);

    version_id = timer_hdl & 0xFF;
    if(del_node->version_id != version_id)
    {
        debug_print("\nDeleting a timer with invalid version\n");
        return 1;
    }
    
    if(dst_cpu_id == coreIndex)
    {
        /*Directly work on the timer wheel to delete*/
        del_wheel_index = del_node->timer_wheel_index;

        debug_print("\nCPU Id:%d, Current wheel index:%d, Deleting at wheel index:%d\n",
            coreIndex, wheel_instance->cur_timer_wheel_index, del_wheel_index);

        if(del_node->prev < ADK_TIMER_MAX_NUMBER)
        {
            /*Cur node is not the head*/
            cpu_timer_wheel[coreIndex]->timer_node_pool[del_node->prev].next
            = del_node->next; 
        }
        else
        {
            /*Cur node is the head*/
            wheel_instance->timer_wheel[del_wheel_index].list_head = del_node->next;
        }

        if(del_node->next < ADK_TIMER_MAX_NUMBER)
        {
            cpu_timer_wheel[coreIndex]->timer_node_pool[del_node->next].prev
            = del_node->prev;
        }
                
        del_node->cbFn(del_node->rock, ADK_TIMER_EVENT_TYPE_DELETE);
        
        /*Add the node to free node list*/
        /*ADK_MEMSET(del_node, 0x0, sizeof(adk_timer_list_node_info_t));*/
        del_node->version_id++;
        del_node->next_free = wheel_instance->timer_node_free_list;
        wheel_instance->timer_node_free_list = del_node->cur;
    }
    else
    {
        /*Add to the delete request list of destination timer wheel*/
        if(wheel_instance->early_expiry_list[coreIndex])
            rte_ring_sp_enqueue(wheel_instance->early_expiry_list[coreIndex],
                            (void *)(size_t)timer_hdl);
    }

    return 1;
}

int adk_timer_update(uint16_t coreIndex,
                     adk_timer_hdl_t timer_hdl,
                     adk_timer_info_t timer_info)
{
    adk_timer_wheel_instance_t * wheel_instance = cpu_timer_wheel[coreIndex];
    adk_timer_list_node_info_t * cur_node;
    uint8_t wheel_index, version_id;
    uint16_t cur_node_index;

    cur_node_index = timer_hdl >> 8;
    if(cur_node_index >= ADK_TIMER_MAX_NUMBER)
    {
        debug_print("\nUpdating invalid timer\n");
        return 1;
    }

    version_id = timer_hdl & 0xFF;
    cur_node = &(wheel_instance->timer_node_pool[cur_node_index]);

    if(cur_node->version_id != version_id)
    {
        debug_print("\nUpdating a timer with invalid version\n");
        return 1;
    }

    wheel_index = cur_node->timer_wheel_index;

    debug_print("\nCPU Id:%d, Current wheel index:%d, Deleting at wheel index:%d\n",
           coreIndex, wheel_instance->cur_timer_wheel_index, wheel_index);

    if(cur_node->prev < ADK_TIMER_MAX_NUMBER)
    {
        cpu_timer_wheel[coreIndex]->timer_node_pool[cur_node->prev].next
        = cur_node->next;
    }
    else
    {
        /*Cur node is the head*/
        wheel_instance->timer_wheel[wheel_index].list_head = cur_node->next;
    }
    if(cur_node->next < ADK_TIMER_MAX_NUMBER)
    {
        cpu_timer_wheel[coreIndex]->timer_node_pool[cur_node->next].prev
        = cur_node->prev;
    }

    /*calculate the index to be inserted in timer wheel*/
    wheel_index = wheel_instance->cur_timer_wheel_index +
                         (timer_info.interval_milli_sec & 0xFF);

    /*prepare the node*/
    cur_node->timer_wheel_index = wheel_index;
    cur_node->timer_hash_offset = (timer_info.interval_milli_sec >> 8);

    debug_print("\nCPU Id:%d, Current wheel index:%d, Inserting at wheel index:%d, \
           with hash_offset:%d",coreIndex, wheel_instance->cur_timer_wheel_index,
           wheel_index, cur_node->timer_hash_offset);

    /*Insert at the head of the list*/
    if(wheel_instance->timer_wheel[wheel_index].list_head
       != ADK_TIMER_MAX_NUMBER)
    {
        cur_node->prev = ADK_TIMER_MAX_NUMBER;
        cur_node->next = wheel_instance->timer_wheel[wheel_index].list_head;
        wheel_instance->timer_wheel[wheel_index].list_head = cur_node->cur;
        cpu_timer_wheel[coreIndex]->timer_node_pool[cur_node->next].prev
        = cur_node->cur;
    }
    else
    {
        /*Empty list*/
        cur_node->next = ADK_TIMER_MAX_NUMBER;
        cur_node->prev = ADK_TIMER_MAX_NUMBER;
        wheel_instance->timer_wheel[wheel_index].list_head = cur_node->cur;
    }

    return 1;
}


int adk_timer_remaining_time(uint16_t coreIndex,
                             adk_timer_hdl_t timer_hdl,
                             adk_timer_info_t *timer_info)
{
    adk_timer_wheel_instance_t * wheel_instance = cpu_timer_wheel[coreIndex];
    adk_timer_list_node_info_t * cur_node;
    uint8_t wheel_index, version_id;
    uint16_t cur_node_index;

    cur_node_index = timer_hdl >> 8;
    if(cur_node_index >= ADK_TIMER_MAX_NUMBER)
    {
        debug_print("\nTime remaining called for invalid timer\n");
        return 1;
    }

    version_id = timer_hdl & 0xFF;
    cur_node = &(wheel_instance->timer_node_pool[cur_node_index]);

    if(cur_node->version_id != version_id)
    {
        debug_print("\nTime remaining caller for a timer with invalid version\n");
        return 1;
    }

    wheel_index = cur_node->timer_wheel_index;

    timer_info->interval_milli_sec = cur_node->timer_hash_offset << 8;
    if(wheel_index < wheel_instance->cur_timer_wheel_index) {
        timer_info->interval_milli_sec += 0xFF - wheel_instance->cur_timer_wheel_index + wheel_index;
    } else {
        timer_info->interval_milli_sec += wheel_index - wheel_instance->cur_timer_wheel_index;
    }

    debug_print("\nCPU Id:%d, Current wheel index:%d, Inserting at wheel index:%d, \
           with hash_offset:%d remaining time:%d",coreIndex, wheel_instance->cur_timer_wheel_index,
           wheel_index, cur_node->timer_hash_offset, timer_info->interval_milli_sec);

    return 1;
}


void adk_timer_proc(uint16_t coreIndex, uint32_t cpu_tick_cur)
{
    adk_timer_wheel_instance_t * wheel_instance = cpu_timer_wheel[coreIndex];
    uint32_t cpu_tick_diff = cpu_tick_cur - wheel_instance->cpu_tick_prev;
    uint16_t i;

    /*
     * Look into the early expiry list
     */
    for (i = 0; i < workerInfo_p->numLcores; ++i)
    {
        struct rte_ring *r;
        void * timer_list[8];
        adk_timer_hdl_t timer;
        uint8_t nb, j;

        /*No early list needed for same lcore timer wheel*/
        if(workerInfo_p->lcore[i].index == coreIndex)
            continue;

        r = wheel_instance->early_expiry_list[workerInfo_p->lcore[i].index];
        if(!r)
            continue;

        do
        {
            nb = rte_ring_sc_dequeue_burst(r, timer_list, 8, NULL);
            for(j=0; j<nb; ++j)
            {
                uint8_t version_id, dst_cpu_id, wheel_index;
                uint16_t node_index;
                adk_timer_list_node_info_t * node;

                timer = (adk_timer_hdl_t)(size_t)(timer_list[j]);
                node_index =  timer >> 16;
                dst_cpu_id = (timer >> 8) & 0xFF;
                if(dst_cpu_id != coreIndex)
                    continue; /*Ignore this request*/

                node = &(wheel_instance->timer_node_pool[node_index]);
                version_id = timer & 0xFF;
                if(version_id != node->version_id)
                    continue; /*Ignore this request*/

                /*Early expire*/
                wheel_index = node->timer_wheel_index;
                if(node->prev < ADK_TIMER_MAX_NUMBER)
                {
                    /*Cur node is not the head*/
                    cpu_timer_wheel[coreIndex]->timer_node_pool[node->prev].next
                    = node->next;
                }
                else
                {
                    /*Cur node is the head*/
                    wheel_instance->timer_wheel[wheel_index].list_head = node->next;
                }
                if(node->next < ADK_TIMER_MAX_NUMBER)
                {
                    cpu_timer_wheel[coreIndex]->timer_node_pool[node->next].prev
                    = node->prev;
                }

                node->cbFn(node->rock, ADK_TIMER_EVENT_TYPE_DELETE);
                
                /*Add the node to free node list*/
                node->version_id++;
                node->next_free = wheel_instance->timer_node_free_list;
                wheel_instance->timer_node_free_list = node->cur;
            }
        }while(nb);
    }

    if(cpu_tick_diff >= wheel_instance->cpu_ticks_per_timer_tick)
    {
        /*1 timer tick exceeded.. Do the per timer tick bookkeeping*/
        adk_timer_per_tick_bookkeeping(coreIndex);
        /*Store the current cpu tick with offset corrected as previous cpu tick*/
        wheel_instance->cpu_tick_prev = cpu_tick_cur -
                        (cpu_tick_diff - wheel_instance->cpu_ticks_per_timer_tick);
    }
}

void adk_timer_per_tick_bookkeeping(uint16_t coreIndex)
{
    adk_timer_list_node_info_t * cur_node, *del_node;
    adk_timer_wheel_instance_t * wheel_instance = cpu_timer_wheel[coreIndex];
    uint16_t cur_node_index;

    wheel_instance->cur_timer_wheel_index++;

    /*Traverse the list*/
    cur_node_index = wheel_instance->timer_wheel
                     [wheel_instance->cur_timer_wheel_index].list_head;
    while(cur_node_index < ADK_TIMER_MAX_NUMBER)
    {
#if 0
        debug_print("\nCPU Id:%d, adk_timer_per_tick_bookkeeping, cur_timer_wheel_index:%d",
               coreIndex, wheel_instance->cur_timer_wheel_index);
#endif
        cur_node = &(cpu_timer_wheel[coreIndex]->timer_node_pool[cur_node_index]);
        if(cur_node->timer_hash_offset)
        {
            cur_node->timer_hash_offset--;
            cur_node_index = cur_node->next;
        }
        else
        {
            /* call application registered callback function */
            cur_node->cbFn(cur_node->rock, ADK_TIMER_EVENT_TYPE_TIMEOUT);
            del_node = cur_node;
            cur_node_index = cur_node->next;

            /* delete the node */
            if(del_node->prev < ADK_TIMER_MAX_NUMBER)
            {
                cpu_timer_wheel[coreIndex]->timer_node_pool[del_node->prev].next
                = del_node->next;
            }
            else
            {
                /*Cur node is the head*/
                wheel_instance->timer_wheel[wheel_instance->
                                            cur_timer_wheel_index].list_head
                = del_node->next;
            }
            if(del_node->next < ADK_TIMER_MAX_NUMBER)
            {
                cpu_timer_wheel[coreIndex]->timer_node_pool[del_node->next].prev
                = del_node->prev;
            }
            /*ADK_MEMSET(del_node, 0x0, sizeof(adk_timer_list_node_info_t));*/
            cur_node->version_id++;
            del_node->next_free = wheel_instance->timer_node_free_list;
            wheel_instance->timer_node_free_list = del_node->cur;
        }
    }
}

#endif /*ADK_TIMER*/
