/* **************************************************************************
**                                                                          *
** INTEL CONFIDENTIAL                                                       *
** Copyright 2011 - 2019 Intel Corporation All Rights Reserved.             *
**                                                                          *
** The source code contained or described herein and all documents related  *
** to the source code ("Material") are owned by Intel Corporation or its    *
** suppliers or licensors. Title to the Material remains with Intel         *
** Corporation or its suppliers and licensors. The Material contains trade  *
** secrets and proprietary and confidential information of Intel or its     *
** suppliers and licensors. The Material is protected by worldwide          *
** copyright and trade secret laws and treaty provisions. No part of the    *
** Material may be used, copied, reproduced, modified, published, uploaded, *
** posted, transmitted, distributed, or disclosed in any way without        *
** Intel's prior express written permission.                                *
**                                                                          *
** No license under any patent, copyright, trade secret or other            *
** intellectual property right is granted to or conferred upon you by       *
** disclosure or delivery of the Materials, either expressly, by            *
** implication, inducement, estoppel or otherwise. Any license under such   *
** intellectual property rights must be express and approved by Intel       *
** in writing.                                                              *
**                                                                          *
** **************************************************************************
**/
/*!
 * @file  adk_ipsec.c
 *
 *      $HeadURL$
 *      $Revision$
 *      $Date$
 *
 * @addtogroup _ipsec IPsec Module
 * @{
 *      @addtogroup _ipsecFunctions IPsec Functions
 *      @{
 *
 * @brief
 *
 * The file contains all APIs that are used by the application
 * to configure and manage IPsec Security Policies and Security
 * Associations into ACP system memory. These are used to facilitate
 * IPsec processing of packets by the ACP data plane software
 *
 ***********************************************************************/
#ifdef ADK_DEVICE_SNR
#include <rte_security.h>
#endif
#include "adk_include.h"
#include <adk_cksum.h>
#include <adk_rm.h>
#include <adk_sys_services.h>
#include <adk_cp_acl_list.h>
#include <adk_hos.h>

/* ipsec header file */
#include "adk_ipsec.h"
#include "adk_ipsec_common.h"
#include "adk_ipsec_cp.h"

/* PH dev header file */
#include "adk_ph_dev_dp.h"
/*QOS internal header file*/
#include "adk_qos_include_internal.h"

/*!
 * =============================================================
 *
 * adk_ipsec_init
 *
 * @param[in]  ipsecHandle  - IPsec module handle. Returned at the
 *                            time of IPsec module's registration
 *                            with the module manager.
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully initialized IPsec module.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM - Invalid i/p parameter.
 * @retval ADK_E_IPSEC_NVM_CALLOC_FAIL - RTE call to allocate memory
 *                                     NV RAM failed.
 * @retval ADK_E_IPSEC_INIT_FAIL - RTE call to get one of the tree
 *                               Id's, or namespace Id's has failed,
 *                               Leading to init fail.
 *
 * @brief   Function to initialize IPsec ADK module
 *
 * This function initializes the ADK IPsec module. It stores
 * frequestly used namespace Id's/Tree Ids used by IPsec module
 * into local NV RAM mmeory.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_init (adk_module_handle_t ipsecHandle)
{
    adk_status_t                    retValue = ADK_IPSEC_SUCCESS;
    ipsec_module_info_t             *ipsecModuleInfo_p;
    adk_sys_handle_t                sysHandle;
    adk_module_service_fn_t         svcFn;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    ipsecModuleInfo_p =
        (ipsec_module_info_t *)ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
			 ADK_IPSEC_MOD_INFO_OBJ, 1, sizeof(ipsec_module_info_t));
    if ( ipsecModuleInfo_p  == NULL )
    {
        ADK_IPSEC_LOG_FORCE("IPSEC: Failed to allocate memory\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    /* set module info data structure pointer */
    ADK_MODULE_INFO_PTR_SET(ipsecHandle, ipsecModuleInfo_p);

    /* Set IPsec module info cookie */
    ipsecModuleInfo_p->cookie = ADK_IPSEC_MODULE_COOKIE;

    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleLogSet = adk_ipsec_log_level_set;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_IPSEC_LLI_FN_ID;
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_IPSEC_ULI_FN_ID;
    svcFn.moduleTerminate = adk_ipsec_shutdown;
    retValue = adk_module_service_function_register(sysHandle, ipsecHandle, &svcFn);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_FORCE("IPSEC: Failed to register service function\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    /* Allocate and initialize SPD IN/OUT CP ACL list */
    retValue = adk_ipsec_spd_cp_create(ipsecHandle, ipsecModuleInfo_p);
    if (ADK_SUCCESS != retValue)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, retValue,
                     "IPSEC: Failed to create CP ACL list\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(retValue);
    }

    retValue = adk_ipsec_dp_table_init(ipsecHandle, sysHandle, ipsecModuleInfo_p);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_FORCE("IPSEC: DP Table init failed!\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ipsecModuleInfo_p->policyHdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipsecHandle), ADK_IPSEC_RULES_PER_SET_OBJ,
                                                        ADK_IPSEC_MAX_NUM_SPD_RULES);
    if (ipsecModuleInfo_p->policyHdl == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                            "IPSEC: Failed to allocate memory\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    ipsecModuleInfo_p->policyInHdl = adk_rm_resource_init(
                                           ADK_MODULE_MODULEID_GET(ipsecHandle), ADK_IPSEC_IN_RULES_PER_SET_OBJ,
                                           ADK_IPSEC_MAX_NUM_SPD_IN_RULES);
    if (ipsecModuleInfo_p->policyInHdl == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                            "IPSEC: Failed to allocate memory for SPD IN policies\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    ipsecModuleInfo_p->eSASetHdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipsecHandle), ADK_IPSEC_ING_NUM_SA_OBJ,
							ADK_IPSEC_NUM_OF_SA);
    if (ipsecModuleInfo_p->eSASetHdl == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                     "IPSEC: Failed to allocate memory\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    ipsecModuleInfo_p->iSASetHdl = adk_rm_resource_init(ADK_MODULE_MODULEID_GET(ipsecHandle), ADK_IPSEC_EG_NUM_SA_OBJ,
							ADK_IPSEC_NUM_OF_SA);
    if (ipsecModuleInfo_p->iSASetHdl == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                     "IPSEC: Failed to allocate memory\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    if(!adk_rm_resource_reserve_fixed(ipsecModuleInfo_p->policyHdl, 0))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                     "Failed to reserve 0th index in SPD-OUT.\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        return ADK_E_INTERNAL_ERROR;
    }
    if(!adk_rm_resource_reserve_fixed(ipsecModuleInfo_p->policyInHdl, 0))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                     "Failed to reserve 0th index in SPD-IN.\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        return ADK_E_INTERNAL_ERROR;
    }
    /* Set default log level to ADK_LOG_LEVEL_INFO */
    ipsecModuleInfo_p->logLevel = ADK_LOG_LEVEL_INFO;

    ipsecModuleInfo_p->ipsec_fc_ver_status = ADK_IPSEC_FC_VER_UPDATED;
    
    ipsecModuleInfo_p->numIngSAEntries = 0;
    ipsecModuleInfo_p->numEgSAEntries = 0;

    LOGIPsecModuleInfo(ipsecModuleInfo_p,"IPSEC_INIT - ");
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
} /* -----  end of function adk_ipsec_init  ----- */

/*!
 * =============================================================
 *
 * adk_ipsec_service_function_update
 *
 * @param[in]   ipsecHandle  IPsec module handle
 *
 * @retval  ADK_IPSEC_SUCCESS             Successful completion
 * @retval  ADK_E_INVALID_HANDLE    Invalid IPsec module handle specified
 *
 * @brief       Function to register the CPU based service fns
 *
 * This function registers the CPU based service functions with
 * the SIF.
 *
 * Note: This function must not be called directly from the application. This
 * function gets called from adk_module_handle_get() API.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_service_function_update (adk_module_handle_t ipsecHandle)
{
    adk_status_t            retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t        sysHandle;
    ipsec_module_info_t     *ipsecModuleInfo_p;
    adk_module_service_fn_t svcFn;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    /* Handle validations */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p, ipsecHandle, retValue);

    ADK_MEMSET (&svcFn, 0x00, sizeof (adk_module_service_fn_t));
    svcFn.moduleLogSet = adk_ipsec_log_level_set;
    svcFn.cpuEntryFn_LLI[_DEST_FUNC_DEFAULT] = ADK_IPSEC_LLI_FN_ID;
    svcFn.cpuEntryFn_ULI[_DEST_FUNC_DEFAULT] = ADK_IPSEC_ULI_FN_ID;
    svcFn.moduleTerminate = adk_ipsec_shutdown;

    /* Register ipsec parser function in PH to Decode exception params*/
    retValue = adk_ph_dev_rx_parse_register((void*)adk_ipsec_ph_rx_parse,
            ADK_IPSEC_MODULE_ID);
    if (retValue != ADK_SUCCESS)
    {
        printf("Failed to register IPsec parser function in PH\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        return 1;
    }

    retValue = adk_module_service_function_update (sysHandle,
                                                   ipsecHandle,
                                                   &svcFn);

    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle,
                            retValue, "Failed to register the functions\n");
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
}

/*!
* =============================================================
* @fn adk_ipsec_add_sp
*
* @param[in]    ipsecHandle  -  IPsec module handle.
* @param[in]    *secPol_p    -  Pointer to security policy data structure.
* @param[in]    polIndex     -  Index in the SPD OUT tree where the policy will be added.
* @param[in]    polByIndex   -  Flag which indicates whether to insert policy into an
*                               index or append the policy into SPD OUT tree.
* @param[in]    polCommit    -  This flag when set, indicates that the policy
*                               will be added to DP immediately resulting
*                               into rebuilding of SPD OUT tree. If not set, the
*                               policy will be added only to CP and
*                               will take effect in DP only when a commit is
*                               performed later. It will be handled only in case of
*                               bulk policy addition.
* @param[in]     eSANsHandle -  Namespace handle for egress SA. If egress SA is already
*                               configured and need to be attached with policy then namespace
*                               of this egress SA will be input to this API else NULL shall be passed
*                               as eSANsHandle.
* @param[in]     iSANsHandle -  Namespace handle for ingress SA. If ingress SA is already
*                               configured and need to be attached with policy then namespace
*                               of this ingress SA will be input to this API else NULL shall be passed
*                               as iSANsHandle.
*
* @retval ADK_IPSEC_SUCCESS                     - Successfully added new Security Policy
* @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE     - i/p IPsec module handle
*                                                 is invalid.
* @retval ADK_E_IPSEC_INVALID_INPUT_PARAM       - Invalid i/p parameter.
* @retval ADK_E_IPSEC_INDEX_OUT_OF_BOUND        - Given polIndex is out of boundary
* @retval ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL        - Failed to add SPD OUT Entry
* @retval ADK_E_IPSEC_SPD_TREE_BUILD_FAIL       - Failed to build SPD OUT TREE
*
* @brief
* The API does following things internally
* STEP 1: validate all the i/p arguments
* STEP 2: program policy tree
* STEP 3: set ipsec_sp_context at policy index based on eSA namespace handle
* STEP 4: program SPI tree based on iSA namespace handle
* STEP 5: Build Policy tree.
*
* Note:
* For IPv4 fragments to match IPsec policy the L4 protocol ports configuration
* should be for ‘any’.
* For Ipv6 fragments to match IPsec policy, the protocol value in the policy
* configuration should be set to 0 apart from the protocol ports configuration
* being wildcarded, else the fragmented packets would be discarded
*
* Ipsec tunnel mode encryption will lead to bigger packet size due to
* IPv4/IPv6 header creation and other processing. Application is expected
* consider this increase in packet size while configuring path mtu parameter.
* For sec policy selector protocol value other that TCP/UDP/SCTP, the policy
* selectors considered for policy lookup are currently restricted only to src IP
* addr, dst IP addr and protocol value. src port and dst port configuration
* will be ignored in such cases.
* The source and destination address information in a policy specification
* adk_ipsec_sec_pol_t implies directionality. The implied direction is "into"
* the tunnel (going for encryption). The word tunnel is loosely used here to
* cover both tunnel and transport IPsec PDU creation. The policy is stateful
* in so far as it expects to have two SAs, one for PDUs with matching source
* and destination addresses (egress side SA/PDU) and the other with swapped
* addresses (ingress side SA/PDU).
* When a PDU "exits" a tunnel (after decryption) its source and destination
* IP addresses are swapped to match against the policy eliminating the need
* to add a mirror policy in the reverse direction for the PDUs coming out
* of the tunnel. If one adds a policy with exactly reversed source and
* destination addresses to the policy database it will be looked up just
* like any other unrelated policy and not as a "reverse" policy of some policy.eg.
* policy config: src addr = 10.10.10.1, dst addr = 10.20.20.1
* PDU with these IP addresses is going into tunnel, policy lookup will
* happen successfully for such PDU's and these will get encrypted as per egress
* SA configuration.
* Now if a PDU with src addr = 10.20.20.1 and dst addr = 10.10.10.1
* and protocol as ESP/AH is received (SPI value as per ingress SA
* configuration), it will be sent for decryption and then a policy
* lookup on swapped IP addresses will be done.
* If the configured policy falls under double tunnel case, where this
* policy is the second policy, 'configDF' bit configuration is ignored
* for such a case.
* 
* Note: For multi-selector policy rules, calls to adk_ipsec_add_sp API are assumed
* to be having identical input parameters except the selectors and polIndex.
* Multi-selector policy rules should have same spIDEx.
*
* =========================================================================================
*/
adk_status_t
adk_ipsec_add_sp(adk_module_handle_t ipsecHandle,
                 adk_ipsec_sec_pol_t *secPol_p,
                 uint16_t            polIndex,
                 bool                polByIndex,
                 bool                polCommit,
                 adk_handle_t        eSANsHandle,
                 adk_handle_t        iSANsHandle)
{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    uint32_t  idx = 0, i = 0, ingress_sa_idx = 0;
    adk_acl_rule_t *cpRule_p;
    adk_ipsec_spd_rule_t spdRulePtn;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                             ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer*/
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);
    
    retValue = adk_ipsec_crypto_dev_init(sysHandle, 
                                         ipsecHandle,
                                         ipsecModuleInfo_p);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                            "Crypto Devices initialization failed\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
    }

    retValue = adk_ipsec_verify_add_sp_input_args(ipsecHandle,
                                                  ipsecModuleInfo_p,
                                                  secPol_p);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    
    if ((polByIndex) && (polIndex > (ADK_IPSEC_MAX_NUM_SPD_RULES - 1)))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "Given Policy index#%d is out of boundary\n", polIndex);
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
    if ((polByIndex) && (polIndex > ipsecModuleInfo_p->numSpdRules))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "Application should not leave any gaps between polIndexes\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ADK_LOG_INFO (sysHandle, ipsecHandle,
        "ADD_SP: ipsecHandle: %"ADKPRIh", polIndex: 0x%x, polByIndex: 0x%x, "
        "polCommit: 0x%x, eSANsHandle: %"ADKPRIh", iSANsHandle: %"ADKPRIh"\n",
        ipsecHandle, polIndex, polByIndex, polCommit, eSANsHandle, iSANsHandle);

    idx = adk_rm_resource_reserve(ipsecModuleInfo_p->policyHdl);
    if (idx == ADK_RM_RSRC_INVALID)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                            "Reached max number of IPsec policies\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ipsecSecPolLocal = (ipsec_sec_pol_info_t*)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
		 ADK_IPSEC_SP_OBJ, 1, sizeof(ipsec_sec_pol_info_t));
    if (ipsecSecPolLocal == NULL)
    {
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                        "IPSEC: Failed to allocate memory\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SP_MEMORY_UNAVAILABLE);
    }

    ipsecSecPolLocal->policyResourceID = idx;

   /*
     * Add rule to ADK CP rule list.
     * The rules will be added during the next commit.
     */
    cpRule_p = &(ipsecSecPolLocal->cpRule);
    cpRule_p->user_data = (void *)(ipsecSecPolLocal);   /* This user data points to security policy info structure */


    /* first normalize whatver application gave us */
    ADK_MEMSET(&spdRulePtn, 0, sizeof(adk_ipsec_spd_rule_t));
    ipsec_policy_ptn_create(&spdRulePtn, &secPol_p->secPolSel);

     /* Convert this rule to a expanded set of hos rules */
    adk_hos_spd_tbl_rule_attr_t *hos_rule[ADK_ACL_MAX_HOS_RULES_PER_RULE];

    retValue = ipsec_spd_convert_to_hos_acls(ipsecHandle, secPol_p, &spdRulePtn,
                                             idx, &(cpRule_p->num_hos_rules), hos_rule);
    if (retValue != ADK_SUCCESS)
    {
        ADK_NVM_FREE(ipsecSecPolLocal);
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
        ADK_IPSEC_LOG_ERROR(ipsecHandle, retValue,
                "Failed to convert SPD OUT"
                " rule to HOS rules\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL);
    }

    for(i = 0; i< cpRule_p->num_hos_rules; i++)
    {
        cpRule_p->hos_rule_attr[i] = (adk_hos_rule_attr_t *)hos_rule[i];
    }

    /* Convert this rule to dpdk rule */
    ipsec_spd_convert_to_dpdk_acl (ipsecHandle, &spdRulePtn, idx,
                                        &(ipsecSecPolLocal->dpdkAclRule));

    LOGSecPolAttr(secPol_p->secPolAttr,"ADD_SP - ");

   /* Add this rule to CP list */
    retValue = adk_acl_rule_add(
            ipsecModuleInfo_p->cpSpdAclList,
            polByIndex ? polIndex : UINT16_MAX,
            cpRule_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_NVM_FREE(ipsecSecPolLocal);
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
        ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_INCONSISTENT_STATE,
                             "Failed to add SPD Rule\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL);
    }

    (ipsecModuleInfo_p->numSpdRules)++;

    /* we do all this irrespective of commit flag, because
     * application might do "bulk" operations for multiple
     * adds/deletes in any crazy combinations. In that case
     * the packets anyway will follow what is in the final policy
     * DB, till then no harm in populting other DS as they won't
     * get used till the policy commit anyway.
     */

    ADK_MEMSET(&ipsecModuleInfo_p->dpTbl->polSAInfo[idx], 0x00,
               sizeof(adk_ipsec_sec_pol_sainfo_t));
    adk_ipsec_dp_policy_attributes(&ipsecModuleInfo_p->dpTbl->polSAInfo[idx],
                                   &secPol_p->secPolAttr, TRUE);
    if (eSANsHandle)
    {
        ipsecModuleInfo_p->dpTbl->polSAInfo[idx].eSAIdx =
                                        ((ipsec_sa_idx_t *)eSANsHandle)->index;
        ipsecModuleInfo_p->dpTbl->polSAInfo[idx].eSAstate = VALID_SA;
    }
    else
        ipsecModuleInfo_p->dpTbl->polSAInfo[idx].eSAstate = INVALID_SA;

    if (iSANsHandle)
    {
        ingress_sa_idx = ((ipsec_sa_idx_t *)iSANsHandle)->index;
        if (secPol_p->secPolAttr.preFragEnable &&
                secPol_p->secPolAttr.useSingleSAInPrefrag)
        {
            ipsecModuleInfo_p->dpTbl->iSATbl[ingress_sa_idx].useSingleSaInPreFrag = TRUE;
            ipsecModuleInfo_p->dpTbl->iSATbl[ingress_sa_idx].policy_version++;
        }
    }
    ipsecModuleInfo_p->ipsec_fc_ver_status = ADK_IPSEC_FC_VER_TO_BE_UPDATED;
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
    if (polCommit == TRUE)
    {
        retValue = adk_ipsec_table_build(ipsecHandle, ADK_IPSEC_SPD_OUT_TBL);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            adk_acl_rule_delete(ipsecModuleInfo_p->cpSpdAclList,
                polByIndex ? polIndex : (ipsecModuleInfo_p->numSpdRules -1));
            ADK_NVM_FREE(ipsecSecPolLocal);
            adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
            ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_INCONSISTENT_STATE,
                                 "Failed to build SPD OUT Rules\n");
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_TREE_BUILD_FAIL);
        }
    }

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
} /* -----  end of function adk_ipsec_add_sp  ----- */

/*!
 * * =============================================================
 * * @fn adk_ipsec_add_sp_in
 * *
 * * @param[in]    ipsecHandle -  IPsec module handle.
 * * @param[in]    *secPol_p   -  Pointer to SPD IN security policy data structure.
 * * @param[in]    polIndex    -  Index in the SPD IN tree where the policy will be added.
 * * @param[in]    polByIndex  -  Flag which indicates whether to insert policy into an
 * *                              index or append the policy into SPD IN tree.
 * * @param[in]    polCommit   -  This flag when set, indicates that the policy
 * *                              will be added to DP immediately resulting
 * *                              into rebuilding of SPD IN tree. If not set, the
 * *                              policy will be added only to CP side tree and
 * *                              will take effect in hardware only when a commit is
 * *                              performed later.
 * *
 * * @retval ADK_IPSEC_SUCCESS                    - Successfully added new Security Policy.
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE    - Invalid IPsec module handle
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM      - Invalid input parameter.
 * * @retval ADK_E_IPSEC_INDEX_OUT_OF_BOUND       - Input param index is out of bounds, no policy
 * *                                                exists at this index.
 * * @retval ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL       - Failed to add SPD IN Entry.
 * *
 * * @brief
 * *
 * * This API allows the application to add a new security policy into
 * * SPD IN tree. Entry can be added at a specific index OR can be appended to the tree
 * * 1. When polByIndex is True and polIndex is 0, the given policy rule will be prepended and all the
 * *    existing acl rules will be pushed down by 1 level.
 * * 2. When polByIndex is false, the given policy rule will be appended.
 * * 3. When polByIndex is True and polIndex is n, policy rule can be inserted at nth index provided
 * *    that 0 < n <= numOfExistingRules.
 * * 4. When application wants to add/insert rules at subsequent indices, application has to set polByIndex as
 * *    true and indices should be given in incrementing order. Application should pass polIndices sequentially and
 * *    not randomly.
 * * 5. Top ACL rule has the highest priority.
 * * =========================================================================================
 * */
adk_status_t
adk_ipsec_add_sp_in(adk_module_handle_t        ipsecHandle,
                    adk_ipsec_sec_pol_in_t     *secPol_p,
                    uint16_t                   polIndex,
                    bool                       polByIndex,
                    bool                       polCommit)

{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    ipsec_sec_pol_in_info_t *ipsecSecPolInLocal = NULL;
    adk_acl_rule_t *cpRule_p;
    adk_ipsec_spd_in_rule_t spdRulePtn;
    uint32_t  idx = 0, i;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                             ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer*/
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    retValue = adk_ipsec_verify_add_sp_in_input_args(ipsecHandle,
                                                     secPol_p);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    
    if ((polByIndex) && (polIndex > (ADK_IPSEC_MAX_NUM_SPD_IN_RULES - 1)))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "Given Policy index is out of boundary\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
    if ((polByIndex) && (polIndex > ipsecModuleInfo_p->numSpdInRules))
    {
         ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                             "Application should not leave any gaps between polIndexes\n");
         ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
         ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ADK_LOG_INFO (sysHandle, ipsecHandle,
                  "ADD_SP_IN: ipsecHandle: %"ADKPRIh", polIndex: 0x%x, polByIndex: 0x%x, polCommit: 0x%x\n",
        ipsecHandle, polIndex, polByIndex, polCommit);

    idx = adk_rm_resource_reserve(ipsecModuleInfo_p->policyInHdl);
    if (idx == ADK_RM_RSRC_INVALID)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                            "Reached max number of IPsec SPD IN policies\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ipsecSecPolInLocal = (ipsec_sec_pol_in_info_t*)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
         ADK_IPSEC_SP_IN_OBJ, 1, sizeof(ipsec_sec_pol_in_info_t));
    if (ipsecSecPolInLocal == NULL)
    {
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                        "IPSEC: Failed to allocate memory\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SP_MEMORY_UNAVAILABLE);
    }

    ipsecSecPolInLocal->policyResourceID = idx;

   /*
     * Add rule to ADK CP rule list.
     * The rules will be added during the next commit.
     */
    cpRule_p = &(ipsecSecPolInLocal->cpRule);
    cpRule_p->user_data = (void *)(ipsecSecPolInLocal);   /* This user data points to security policy info structure */

    /* first normalize whatver application gave us */
    ADK_MEMSET(&spdRulePtn, 0, sizeof(adk_ipsec_spd_in_rule_t));
    ipsec_spd_in_policy_ptn_create(&spdRulePtn, &secPol_p->secPolSel);

     /* Convert this rule to a expanded set of hos rules */
    adk_hos_spd_tbl_rule_attr_t *hos_rule[ADK_ACL_MAX_HOS_RULES_PER_RULE];

    retValue = ipsec_spd_in_convert_to_hos_acls(ipsecHandle, secPol_p,
                                                &spdRulePtn, idx, &(cpRule_p->num_hos_rules), 
                                                hos_rule);
    if (retValue != ADK_SUCCESS)
    {
        ADK_NVM_FREE(ipsecSecPolInLocal);
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
        ADK_IPSEC_LOG_ERROR(ipsecHandle, retValue,
                "Failed to convert SPD OUT"
                " rule to HOS rules\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL);
    }

    for(i = 0; i< cpRule_p->num_hos_rules; i++)
    {
        cpRule_p->hos_rule_attr[i] = (adk_hos_rule_attr_t *)hos_rule[i];
    }

    /* Convert this rule to dpdk acl rule */
    ipsec_spd_in_convert_to_dpdk_acl (ipsecHandle, &spdRulePtn, idx,
                                        &(ipsecSecPolInLocal->dpdkAclInRule));

    ADK_IPSEC_LOG_INFO(ipsecHandle, "ADD_SP_IN - ACTION: 0x%x\n", secPol_p->secPolAttr.action);

    /*Add this rule to CP list */
    retValue = adk_acl_rule_add(
            ipsecModuleInfo_p->cpSpdInAclList,
            polByIndex ? polIndex : UINT16_MAX,
            cpRule_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_NVM_FREE(ipsecSecPolInLocal);
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, idx);
        ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_INCONSISTENT_STATE,
                             "Failed to add SPD In Rule\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_ENTRY_ADD_FAIL);
    }

    (ipsecModuleInfo_p->numSpdInRules)++;

    ipsecModuleInfo_p->dpTbl->spdInAction[idx] = secPol_p->secPolAttr.action;
    
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
    
    if (polCommit == TRUE) 
    {
        retValue = adk_ipsec_table_build(ipsecHandle, ADK_IPSEC_SPD_IN_TBL);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            adk_acl_rule_delete(ipsecModuleInfo_p->cpSpdInAclList,
                polByIndex ? polIndex : (ipsecModuleInfo_p->numSpdInRules -1));
            ADK_NVM_FREE(ipsecSecPolInLocal);
            adk_rm_resource_free(ipsecModuleInfo_p->policyInHdl, idx);
            ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_INCONSISTENT_STATE,
                                 "Failed to build SPD IN Rules\n");
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_TREE_BUILD_FAIL);
        }
    }

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);

} /* -----  end of function adk_ipsec_add_sp_in  ----- */

/*!
 * =============================================================
 * @fn adk_ipsec_update_sp
 *
 * @param[in]    ipsecHandle      - IPsec module handle.
 * @param[in]    polIndex         - Index of the entry to be modified
 * @param[in]    *newSecPolAttr_p - pointer to new security policy attributes
 *
 * @retval ADK_IPSEC_SUCCESS                  - Successfully updated Security Policy
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE  - i/p ipsec module handle
 *                                              is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM    - Invalid i/p parameter.
 * @retval ADK_E_IPSEC_INDEX_OUT_OF_BOUND     - i/p param index is out of bound, no policy
 *                                              exists at this index.
 * @brief
 *
 * This API allows the application to update an existing security policy
 * Note: Update SP should not be called to update critical fields of existing
 * policy. Critical fields are all selectors, tunnel/transport, ESP/AH.
 *
 * If prefragEnable and useSingleSA fields are to be changed then update_sp
 * should be followed with update_sa (ingress and egress).
 *
 * Note: For multi-selector policy rules, calls to adk_ipsec_update_sp API are assumed
 * to be having identical input parameters except the selectors and polIndex.
 * Multi-selector policy rules should have same spIDEx.
 *  
 * =============================================================
 */
adk_status_t
adk_ipsec_update_sp (adk_module_handle_t      ipsecHandle,
                     uint16_t                 polIndex,
                     adk_ipsec_sec_pol_attr_t *newSecPolAttr_p)
{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    uint16_t index=0;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;
    uint16_t polHdlIdx;
    uint8_t idx;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p, ipsecHandle, retValue);

    retValue = adk_ipsec_verify_upd_sp_input_args(ipsecHandle,
                                                  newSecPolAttr_p);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                           "SPD OUT Attributes verification failed \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    
    if(FALSE == adk_rm_resource_isset(ipsecModuleInfo_p->policyHdl,
                                      polIndex))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid SPD OUT policy handle \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
 
    if (polIndex > (ADK_IPSEC_MAX_NUM_SPD_RULES - 1))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "Given Policy index#%d is out of boundary\n", polIndex);
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
    
    if (polIndex >= ipsecModuleInfo_p->numSpdRules)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "Application should update only existing policy entries\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipsecModuleInfo_p->cpSpdAclList, &cpRule))
    {
        if(index==polIndex)
        {
            ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
            break;
        }
        index++;
    }

    if (NULL == ipsecSecPolLocal)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                            "The policy is not found in CP ACL List\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
    }

    polHdlIdx = ipsecSecPolLocal->policyResourceID; 

    LOGSecPolUpdateAttr(newSecPolAttr_p,"IPSEC: UPDATE_SP - ");
    adk_ipsec_dp_policy_attributes(&ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx],
                                   newSecPolAttr_p, TRUE);
    if (ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx].preFragEnable &&
        ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx].useSingleSAInPrefrag)
    {
        for (idx = 0; idx < ipsecModuleInfo_p->numIngSAEntries; idx++)
        {
            if (ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx].spIDEx
                == ipsecModuleInfo_p->dpTbl->iSATbl[idx].spIDEx)
            {
                ipsecModuleInfo_p->dpTbl->iSATbl[idx].useSingleSaInPreFrag = TRUE;
                ipsecModuleInfo_p->dpTbl->iSATbl[idx].policy_version++;
            }
        }
    }
    retValue = adk_sys_flow_cache_update_version(sysHandle);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                "Could not update Flow cache version\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
    }
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * * =============================================================
 * * @fn adk_ipsec_update_sp_in
 * *
 * * @param[in]    ipsecHandle      - IPsec module handle.
 * * @param[in]    polIndex         - Index of the entry to be modified
 * * @param[in]    newAction        - New security policy action
 * *
 * * @retval ADK_IPSEC_SUCCESS                  - Successfully updated Security Policy
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE  - i/p ipsec module handle
 * *                                              is invalid.
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM    - Invalid i/p parameter.
 * * @retval ADK_E_IPSEC_INDEX_OUT_OF_BOUND     - i/p param index is out of bound, no policy
 * *                                              exists at this index.
 * * @brief
 * *
 * * This API allows the application to update the action of an existing security policy from SPD IN tree.
 * * =============================================================
 * */
adk_status_t
adk_ipsec_update_sp_in (adk_module_handle_t    ipsecHandle,
                        uint16_t               polIndex,
                        adk_ipsec_pol_action_t newAction)
{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    ipsec_sec_pol_in_info_t *ipsecSecPolInLocal = NULL;
    adk_acl_rule_t cpRule;
    uint16_t index=0;
    uint16_t polHdlIdx;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p, ipsecHandle, retValue);

    if ((newAction != ADK_IPSEC_SP_ACTION_BYPASS) &&
        (newAction != ADK_IPSEC_SP_ACTION_DROP))
    {
	 
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    if(FALSE == adk_rm_resource_isset(ipsecModuleInfo_p->policyInHdl,
                                      polIndex))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                           "Invalid SPD IN policy handle \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
 
    if (polIndex > (ADK_IPSEC_MAX_NUM_SPD_IN_RULES - 1))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "Given Policy index is out of boundary\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
    if (polIndex >= ipsecModuleInfo_p->numSpdInRules)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "Application should update only existing policy entries\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdInAclList);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipsecModuleInfo_p->cpSpdInAclList, &cpRule))
    {
        if(index==polIndex)
        {
            ipsecSecPolInLocal = (ipsec_sec_pol_in_info_t*)(cpRule.user_data);
            break;
        }
        index++;
    }

    if (NULL == ipsecSecPolInLocal)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                            "The policy is not found in CP ACL List\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
    }

    polHdlIdx = ipsecSecPolInLocal->policyResourceID;

    ADK_IPSEC_LOG_INFO(ipsecHandle, "UPDATE_SP_IN - ACTION: 0x%x\n", newAction);

    ipsecModuleInfo_p->dpTbl->spdInAction[polHdlIdx] = newAction;

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;

}

/*!
* =============================================================
* @fn adk_ipsec_del_sp
*
* @param[in]    ipsecHandle       -   IPsec module handle.
* @param[in]    polIndex          -   Index in the SPD OUT tree from where the policy will be deleted.
* @param[in]    polCommit         -   This flag when set, indicates that the policy
*                                     will be added to DP immediately resulting
*                                     into rebuilding of SPD tree. If not set, the
*                                     policy will be added only to CP side tree and
*                                     will take effect in DP only when a commit is
*                                     performed later. It will be handled only in case of
*                                     bulk policy addition.
*
*
* @retval ADK_IPSEC_SUCCESS                       -    Successfully added new Security Policy
* @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE       -    i/p ipsec module handle
*                                                      is invalid.
* @retval ADK_E_IPSEC_INVALID_INPUT_PARAM         -    Invalid i/p parameter.
* @retval ADK_E_IPSEC_INDEX_OUT_OF_BOUND          -    i/p param index is out of bound, no policy
*                                                      exists at this index.
* @retval ADK_E_IPSEC_SPD_TREE_ENTRY_DEL_FAIL     -    Failed to delete SPD OUT tree entry.
* @retval ADK_E_IPSEC_SPD_TREE_BUILD_FAIL         -    Failed to build SPD OUT TREE
*
* @brief
*
* This API allows the application to delete an existing security policy
* from the SPD tree maintained by ACP infrastructure.
* STEP 1: Validate i/p arguments
* STEP 2: Check if polIndex is out of bound, if yes error
* STEP 3: Delete SPD entry from SPD OUT tree from polIndex
* STEP 4: Build SPD OUT tree
* NOTE: polIndex and policy selectors uniquely identify a policy.
*       The application should not use a different polIndex
*       and policy selectors other than the one that was given to it during
*       the time of policy addition
*
* =============================================================
 */

adk_status_t
adk_ipsec_del_sp (adk_module_handle_t    ipsecHandle,
                  uint16_t               polIndex,
                  bool                   polCommit)
{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    uint16_t index=0;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (polIndex > (ADK_IPSEC_MAX_NUM_SPD_RULES - 1))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "Given Policy index#%d is out of boundary\n", polIndex);
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
    
    if (polIndex >= ipsecModuleInfo_p->numSpdRules)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "No policy entry exist at the given policy index#%d\n", polIndex);
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipsecModuleInfo_p->cpSpdAclList, &cpRule))
    {
        if(index==polIndex)
        {
            ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
            break;
        }
        index++;
    }

    if (NULL == ipsecSecPolLocal)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                            "The policy is not found in CP ACL List\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
    }

    if (ADK_RM_RSRC_INVALID != ipsecSecPolLocal->policyResourceID)
    {
        adk_rm_resource_free(ipsecModuleInfo_p->policyHdl, ipsecSecPolLocal->policyResourceID);
    }

    ADK_NVM_FREE(ipsecSecPolLocal);

    retValue = adk_acl_rule_delete(ipsecModuleInfo_p->cpSpdAclList,
            polIndex);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_SPD_ENTRY_DEL_FAIL,
                            "DEL_SP: Failed \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_ENTRY_DEL_FAIL);
    }

    (ipsecModuleInfo_p->numSpdRules)--;
    
    ipsecModuleInfo_p->ipsec_fc_ver_status = ADK_IPSEC_FC_VER_TO_BE_UPDATED;

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
    if (polCommit == TRUE)
    {
        retValue = adk_ipsec_table_build(ipsecHandle, ADK_IPSEC_SPD_OUT_TBL);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_INCONSISTENT_STATE,
                                 "Failed to build SPD OUT Rules\n");
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_TREE_BUILD_FAIL);
        }
    }

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
}

/*!
 * * =============================================================
 * * @fn adk_ipsec_del_sp_in
 * *
 * * @param[in]    ipsecHandle     -  IPsec module handle.
 * * @param[in]    polIndex        -  Index in the SPD IN tree where the policy will be removed.
 * * @param[in]    polCommit       -  This flag when set, indicates that the policy
 * *                                  will be removed from hardware tree immediately resulting
 * *                                  into rebuilding of SPD tree. If not set, the
 * *                                  policy will be removed only from CP side tree and
 * *                                  will take effect in hardware only when a commit is
 * *                                  performed later.
 * *
 * *
 * * @retval ADK_IPSEC_SUCCESS                    -     Successfully deleted Security Policy
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE    -     i/p ipsec module handle
 * *                                                    is invalid.
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM      -     Invalid i/p parameter.
 * * @retval ADK_E_IPSEC_INDEX_OUT_OF_BOUND       -     i/p param index is out of bound, no policy
 * *                                                    exists at this index.
 * * @retval ADK_E_IPSEC_SPD_ENTRY_DEL_FAIL       -     Failed to delete SPD IN tree entry.
 * *
 * * @brief
 * *
 * * This API allows the application to delete an existing security policy from SPD IN Tree
 * * tree.
 * * 1. A Policy rule can be deleted at nth index provided that 0 <= n < numOfExistingRules.
 * * 2. When application wants to delete rules at subsequent indices, then indices should be given in decrementing order.
 * * =============================================================
 * */
adk_status_t
adk_ipsec_del_sp_in (adk_module_handle_t    ipsecHandle,
                     uint16_t               polIndex,
                     bool                   polCommit)
{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    uint16_t index=0;
    ipsec_sec_pol_in_info_t *ipsecSecPolInLocal = NULL;
    adk_acl_rule_t cpRule;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (polIndex > (ADK_IPSEC_MAX_NUM_SPD_IN_RULES - 1))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "Given Policy index is out of boundary\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
    if (polIndex >= ipsecModuleInfo_p->numSpdInRules)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                            "No policy entry exist at the given policy index#%d\n", polIndex);
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
    }
	
    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdInAclList);
    index=0;
    while(ADK_SUCCESS == adk_acl_next_rule_get(
            ipsecModuleInfo_p->cpSpdInAclList, &cpRule))
    {
        if(index==polIndex)
        {
            ipsecSecPolInLocal = (ipsec_sec_pol_in_info_t*)(cpRule.user_data);
            break;
        }
        index++;
    }

    if (NULL == ipsecSecPolInLocal)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                            "The policy is not found in CP ACL List\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
    }

    if (ADK_RM_RSRC_INVALID != ipsecSecPolInLocal->policyResourceID)
    {
        adk_rm_resource_free(ipsecModuleInfo_p->policyInHdl, ipsecSecPolInLocal->policyResourceID);
    }

    ADK_NVM_FREE(ipsecSecPolInLocal);

    retValue = adk_acl_rule_delete(ipsecModuleInfo_p->cpSpdInAclList,
            polIndex);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_SPD_ENTRY_DEL_FAIL,
                            "DEL_SP: Failed \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_ENTRY_DEL_FAIL);
    }

    (ipsecModuleInfo_p->numSpdInRules)--;

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
    
    if (polCommit == TRUE) 
    {
        retValue = adk_ipsec_table_build(ipsecHandle, ADK_IPSEC_SPD_IN_TBL);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR (ipsecHandle, ADK_E_INCONSISTENT_STATE,
                                 "Failed to build SPD IN Rules\n");
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_TREE_BUILD_FAIL);
        }
    }

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);

}


/*!
* =============================================================
* @fn adk_ipsec_get_num_sp
*
* @param[in]    ipsecHandle - IPsec module handle.
* @param[out]   numSP_p     - pointer set to number of policies existing
*
* @retval ADK_IPSEC_SUCCESS                 - Successfully added new Security Policy
* @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
*                                             is invalid.
* @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
*
* @brief
*
* This API allows the application to get total number of SPD OUT policies
* configured
*
* =============================================================
*/
adk_status_t
adk_ipsec_get_num_sp (adk_module_handle_t ipsecHandle,
                      uint32_t            *numSP_p)
{
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t   *ipsecModuleInfo_p=NULL;
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture
     * pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (numSP_p == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                "IPSEC: Invalid SPD pointer\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    *numSP_p = ipsecModuleInfo_p->numSpdRules;

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * * =============================================================
 * * @fn adk_ipsec_get_num_sp_in
 * *
 * * @param[in]    ipsecHandle     - IPsec module handle.
 * * @param[out]   *numSP_p        - pointer set to number of policies existing
 * *
 * * @retval ADK_IPSEC_SUCCESS                 - Successfully added new Security Policy
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 * *                                             is invalid.
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 * *
 * * @brief
 * *
 * * This API allows the application to get total number of SP policies
 * * configured
 * *
 * * =============================================================
 * */

adk_status_t
adk_ipsec_get_num_sp_in (adk_module_handle_t ipsecHandle,
                         uint32_t            *numSP_p)
{
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t   *ipsecModuleInfo_p=NULL;
    adk_status_t          retValue = ADK_IPSEC_SUCCESS;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture
     * pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (numSP_p == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "IPSEC: Invalid SPD pointer\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    *numSP_p = ipsecModuleInfo_p->numSpdInRules;

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;

}

/*!
* =============================================================
* @fn adk_ipsec_get_sp
*
* @param[in]    ipsecHandle         - IPsec module handle.
* @param[in]    *secPolSelLkup_p    - pointer to security policy selector data lookup
*                                     structure.
* @param[out]   secPol_p            - pointer set by the API to point to required policy
* @param[in]    polIndex            - policy index of the policy that has to be fetched
* @param[in]    byIndex             - if set, indicates that the policy to be fetched
*                                     must be searched by index, else lookup with the 
*                                     help of policy selector values
*
* @retval ADK_IPSEC_SUCCESS                 - Successfully added new Security Policy
* @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
*                                             is invalid.
* @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
* @retval ADK_E_IPSEC_POLICY_NOT_FOUND      - Given policy selectors did not
*                                             match any configured policy
*
* @brief
*
* This API allows the application to get a specific security policy. It is
* expected that the application would have allocated a memory chunk equal to
* the size of data structure adk_ipsec_sec_pol_t where the API can dump the
* policy related values.
* 
* NOTE: The retrieved source IP and destination IP selectors will always in range format
* in output policy structure. So rangeSelFlag will be set for source IP and destination IP and
* application should expect these values in range format only as result of this API.
*
* =======================================================================================
*/
adk_status_t
adk_ipsec_get_sp (adk_module_handle_t             ipsecHandle,
                  adk_ipsec_sec_pol_sel_lkup_t    *secPolSelLkup_p,
                  adk_ipsec_sec_pol_t             *secPol_p,
                  uint16_t                        polIndex,
                  bool                            byIndex)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p = NULL;
    uint16_t  polHdlIdx = 0, polIdx = 0;
    uint16_t index=0;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;
    
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture
     * pointer */

    if (secPol_p == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                "IPSEC: Invalied SPD pointer\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    if (0 == ipsecModuleInfo_p->numSpdRules)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "No policy exist");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    if (byIndex)
    {
        if (polIndex > (ADK_IPSEC_MAX_NUM_SPD_RULES - 1))
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                 "Given Policy index is out of boundary\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
        }
        if (polIndex >= ipsecModuleInfo_p->numSpdRules)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                          "IPSEC: Given Policy index is greater than number of policies configured\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        
        adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
        index=0;
        while(ADK_SUCCESS == adk_acl_next_rule_get(
                ipsecModuleInfo_p->cpSpdAclList, &cpRule))
        {
            if(index==polIndex)
            {
                ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
                break;
            }
            index++;
        }
        
        if (NULL == ipsecSecPolLocal)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                "The policy is not found in CP ACL List\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }

        polHdlIdx = ipsecSecPolLocal->policyResourceID; 
        adk_ipsec_dp_policy_attributes(&ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx],
                                       &secPol_p->secPolAttr,
                                       FALSE);
        adk_ipsec_get_spd_rule(&secPol_p->secPolSel,
                                &ipsecSecPolLocal->dpdkAclRule);
    }
    else
    {
        if (secPolSelLkup_p == NULL)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "IPSEC: Invalid pointer\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        retValue = adk_ipsec_get_cp_spd_rule_idx(ipsecModuleInfo_p, 
                       secPolSelLkup_p, ipsecModuleInfo_p->numSpdRules, &polIdx,
                       &ipsecSecPolLocal);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                "No such policy exist");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }
        polHdlIdx = ipsecSecPolLocal->policyResourceID;
        adk_ipsec_dp_policy_attributes(&ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx],
                                       &secPol_p->secPolAttr,
                                       FALSE);
        adk_ipsec_get_spd_rule(&secPol_p->secPolSel, &ipsecSecPolLocal->dpdkAclRule);
    
    }
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * * =============================================================
 * * @fn adk_ipsec_get_sp_in
 * *
 * * @param[in]    ipsecHandle         - IPsec module handle.
 * * @param[in]    *secPolSelLkup_p    - pointer to security policy selector data lookup
 * *                                     structure.
 * * @param[in]    polIndex            - policy index of the policy that has to be fetched
 * * @param[in]    byIndex             - if set, indicates that the policy to be fetched
 * *                                     must be searched by index, else with the help of policy
 *                                       selector values
 * * @param[out]   *secPol_p           - pointer set by the API to point to required SDP IN policy
 * *
 * * @retval ADK_IPSEC_SUCCESS                 - Successfully fetched a Security IN Policy
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 * *                                             is invalid.
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 * * @retval ADK_E_IPSEC_POLICY_NOT_FOUND      - Given policy selectors did not
 * *                                             match any configured policy
 * *
 * * @brief
 * *
 * * This API allows the application to get a specific security policy. It is
 * * expected that the application would have allocated a memory chunk equal to
 * * the size of data structure adk_ipsec_sec_pol_in_t where the API can dump the
 * * policy related values.
 * *
 * * NOTE: The retrieved source IP and destination IP selectors will always in range format
 * * in output policy structure. So rangeSelFlag will be set for source IP and destination IP and
 * * application should expect these values in range format only as result of this API.
 * * =======================================================================================
 * */
adk_status_t
adk_ipsec_get_sp_in (adk_module_handle_t             ipsecHandle,
                     adk_ipsec_sec_pol_sel_lkup_t    *secPolSelLkup_p,
                     adk_ipsec_sec_pol_in_t          *secPol_p,
                     uint16_t                        polIndex,
                     bool                            byIndex)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p = NULL;
    uint16_t polHdlIdx = 0, polIdx = 0;
    uint16_t index=0;
    ipsec_sec_pol_in_info_t *ipsecSecPolInLocal = NULL;
    adk_acl_rule_t cpRule;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (secPol_p == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "IPSEC: Invalid pointer\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    if (0 == ipsecModuleInfo_p->numSpdInRules)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "No policy exist\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    if (byIndex)
    {
        if (polIndex > (ADK_IPSEC_MAX_NUM_SPD_IN_RULES - 1))
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INDEX_OUT_OF_BOUND,
                                "Given Policy index is out of boundary\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INDEX_OUT_OF_BOUND);
        }
        if (polIndex >= ipsecModuleInfo_p->numSpdInRules)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                          "IPSEC: Given Policy index is greater than number of policies configured\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }

        adk_acl_get_reset(ipsecModuleInfo_p->cpSpdInAclList);
        index=0;
        while(ADK_SUCCESS == adk_acl_next_rule_get(
                ipsecModuleInfo_p->cpSpdInAclList, &cpRule))
        {
            if(index==polIndex)
            {
                ipsecSecPolInLocal = (ipsec_sec_pol_in_info_t*)(cpRule.user_data);
                break;
            }
            index++;
        }

        if (NULL == ipsecSecPolInLocal)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                "The policy is not found in CP ACL List\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }

        polHdlIdx = ipsecSecPolInLocal->policyResourceID; 
        ADK_MEMCPY(&secPol_p->secPolAttr,
                   &ipsecModuleInfo_p->dpTbl->spdInAction[polHdlIdx],
                   (sizeof(adk_ipsec_pol_action_t)));
        adk_ipsec_get_spd_in_rule(&secPol_p->secPolSel,
                    &ipsecSecPolInLocal->dpdkAclInRule);
    }
    else
    {
        if (secPolSelLkup_p == NULL)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "IPSEC: Invalid pointer\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        
        retValue = adk_ipsec_get_cp_spd_in_rule_idx(ipsecModuleInfo_p,
                       secPolSelLkup_p, ipsecModuleInfo_p->numSpdInRules, &polIdx,
                       &ipsecSecPolInLocal);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                "No such policy exist\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }
        polHdlIdx = ipsecSecPolInLocal->policyResourceID;
        ADK_MEMCPY(&secPol_p->secPolAttr,
                   &ipsecModuleInfo_p->dpTbl->spdInAction[polHdlIdx],
                   sizeof(adk_ipsec_pol_action_t));
        adk_ipsec_get_spd_in_rule(&secPol_p->secPolSel,
                   &ipsecSecPolInLocal->dpdkAclInRule);
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;

}
/*!
 * =============================================================
 * @fn adk_ipsec_get_sp_all
 *
 * @param[in]    ipsecHandle     - IPsec module handle.
 * @param[in]    numSp           - Desired number of SPD OUT policies to be retrieved.
 * @param[out]   secPolList_p    - list of security policies configured within
 *                                 the IPsec ADK system
 *
 * @retval ADK_IPSEC_SUCCESS                 - Successfully retrieved new Security Policy
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 *
 * @brief
 *
 * This API allows the application to retrieve the desired number of security policies
 * configured into IPsec ADK module. secPolList_p would return numSp policy rules laid out
 * in the array per the index of the respective rules from 0 to numSp-1. It is expected that
 * the calling application would have allocated a chunk of memory equivalent to hold the
 * desired number of policies configured. The total number of policies configured
 * can be found by using API: adk_ipsec_get_num_sp.
 *
 * NOTE: The retrieved source IP and destination IP selectors will always in range format
 * in output policy structure. So rangeSelFlag will be set for source IP and destination IP and
 * application should expect these values in range format only as result of this API.
 *
 * ======================================================================================
 */
adk_status_t
adk_ipsec_get_sp_all(adk_module_handle_t    ipsecHandle,
                     uint32_t               numSp,
                     adk_ipsec_sec_pol_t    *secPolList_p)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p;
    uint16_t numPol, i, polHdlIdx;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;
    
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (secPolList_p == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle,  ADK_E_IPSEC_INVALID_INPUT_PARAM,
                "IPSEC: Security policy list pointer is invalid\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    numPol = ipsecModuleInfo_p->numSpdRules;

    if (!numPol)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "No policies configured");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
  
    if (numSp > numPol)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "Desired number of policies is more than policies configured\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
    for (i = 0; i < numSp; i++)
    {
        retValue = adk_acl_next_rule_get(ipsecModuleInfo_p->cpSpdAclList, &cpRule);
 
        if (ADK_SUCCESS != retValue)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                "The policy is not found in CP ACL List\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }

        ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
        polHdlIdx = ipsecSecPolLocal->policyResourceID;
        adk_ipsec_dp_policy_attributes(&ipsecModuleInfo_p->dpTbl->polSAInfo[polHdlIdx],
                                       &secPolList_p[i].secPolAttr,
                                       FALSE);
        adk_ipsec_get_spd_rule(&secPolList_p[i].secPolSel,
                   &ipsecSecPolLocal->dpdkAclRule);
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * * =============================================================
 * * @fn adk_ipsec_get_sp_in_all
 * *
 * * @param[in]    ipsecHandle     - IPsec module handle.
 * * @param[in]    numSp           - Desired number of SPD IN policies to be retrieved.
 * * @param[out]   *secPolList_p   - list of security policies configured within
 * *                                 the IPsec ADK system
 * *
 * * @retval ADK_IPSEC_SUCCESS                 - Successfully fetched out Policies
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 * *                                             is invalid.
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 * *
 * * @brief
 * *
 * * This API allows the application to retrieve the desired number of security policies
 * * configured into IPsec ADK module. It is expected that the calling
 * * application would have allocated a chunk of memory equivalent to hold the
 * * desired number of configured policies. The total number of SPD IN policies configured
 * * can be found by using API: adk_ipsec_get_num_sp_in. 
 * *
 * * NOTE: The retrieved source IP and destination IP selectors will always in range format
 * * in output policy structure. So rangeSelFlag will be set for source IP and destination IP and
 * * application should expect these values in range format only as result of this API.
 * * ======================================================================================
 * */
adk_status_t
adk_ipsec_get_sp_in_all(adk_module_handle_t      ipsecHandle,
                        uint32_t                 numSp,
                        adk_ipsec_sec_pol_in_t   *secPolList_p)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p;
    uint8_t numPol, i, polHdlIdx;
    ipsec_sec_pol_in_info_t *ipsecSecPolInLocal = NULL;
    adk_acl_rule_t cpRule;
 
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (secPolList_p == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle,  ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "IPSEC: Security policy list pointer is invalid\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    numPol = ipsecModuleInfo_p->numSpdInRules;
    
    if (!numPol)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "No policies configured\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }  
    
    if (numSp > numPol)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                            "Desired number of policies is more than policies configured\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdInAclList);
    for (i = 0; i < numSp; i++)
    {
        retValue = adk_acl_next_rule_get(ipsecModuleInfo_p->cpSpdInAclList, &cpRule);
        
        if (ADK_SUCCESS != retValue)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                "The policy is not found in CP ACL List\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }

        ipsecSecPolInLocal = (ipsec_sec_pol_in_info_t*)(cpRule.user_data);
        polHdlIdx = ipsecSecPolInLocal->policyResourceID;
        ADK_MEMCPY(&secPolList_p[i].secPolAttr,
                   &ipsecModuleInfo_p->dpTbl->spdInAction[polHdlIdx],
                   (sizeof(adk_ipsec_pol_action_t)));
        adk_ipsec_get_spd_in_rule(&secPolList_p[i].secPolSel,
                   &ipsecSecPolInLocal->dpdkAclInRule);
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
  
}
/*!
 * * =============================================================
 * * @fn adk_ipsec_table_build
 * *
 * * @param[in]    ipsecHandle -  IPsec module handle.
 * *
 * * @param[in]    ipsecTblId  -  IPsec table id
 * *
 * * @retval ADK_IPSEC_SUCCESS                    - Successfully added new Security Policy.
 * *
 * * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE    - Invalid IPsec module handle
 * *
 * * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM      - Invalid input parameter.
 * *
 * * @retval ADK_E_IPSEC_SPD_TREE_BUILD_FAIL      - RTE call to build SPD tree failed
 * *
 * * @brief
 * *
 * * This API builds a given ordered tree. Application can add rules to tree by calling tree
 * * specific add/delete API's with commit flag false. After the last rule addition/deletion
 * * application can do a single build that builds all the rules.
 * *
 * * =========================================================================================
 * */

adk_status_t
adk_ipsec_table_build(adk_module_handle_t        ipsecHandle,
                      adk_ipsec_table_t          ipsecTblId)
{
    adk_status_t         retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t     sysHandle;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);
    
    if ((ipsecTblId != ADK_IPSEC_SPD_OUT_TBL) &&
        (ipsecTblId != ADK_IPSEC_SPD_IN_TBL))
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    if (ipsecTblId == ADK_IPSEC_SPD_OUT_TBL)
    {
        retValue = adk_ipsec_spd_commit(sysHandle, ipsecHandle, ipsecModuleInfo_p);
        if(retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_TREE_BUILD_FAIL);
        }
    }
    else 
    {
        retValue = adk_ipsec_spd_in_commit(sysHandle, ipsecHandle, ipsecModuleInfo_p);
        if(retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_SPD_TREE_BUILD_FAIL);
        }
    }
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);

}

/*!
 * =============================================================
 * @fn adk_ipsec_add_sa
 *
 * @param[in]    ipsecHandle     - IPsec module handle.
 * @param[in]    *secAsc_p       - Pointer to security association data structure.
 * @param[in]    dscpVal         - DSCP value for ingress sa, unused for egress SA.
 *                                 It should be same as configured in the corresponding
 *                                 policy configuration.
 * @param[in]    numPol          - Number of policies attached to this SA. Indexes for 
 *                                 these policies should be given from startPolIndex.
 * @param[in]    startPolIndex   - Starting policy index in the beginning of Policy indexes list.
 *                                 Note that number of policies attached to this SA should be in 
 *                                 contiguous order.
 * @param[out]   *saNsHandle     - This value is set by the ADK code and is used
 *                                 as SA identifier for each SA
 *
 * @retval ADK_IPSEC_SUCCESS                  - Successfully added new Security Policy
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE  - i/p ipsec module handle
 *                                              is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM    - Invalid i/p parameter.
 * @retval ADK_E_IPSEC_INVALID_SA             - input security association pointer is NULL
 * @retval ADK_E_LA_CRYPTO_CONFIG_FAILED      - Inline session is configured successfully but failed
 *                                              to configure lookaside crypto session when SA mode
 *                                              is inline and direction is inbound.
 *
 * @brief
 *
 * This API allows the application to add a new security association
 * Egress SA Addition:
 *
 * STEP 1:If Direction in SA structure is OUTBOUND, allocate eSANsHandle from
 *        egress namespace..
 *
 * Ingress SA Addition:
 *
 * STEP 1:If Direction in SA structure is INBOUND, allocate iSANsHandle from
 *        ingress namespace.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_add_sa (adk_module_handle_t ipsecHandle,
                  adk_ipsec_sa_t      *secAsc_p,
                  uint8_t             dscpVal,
                  uint8_t             numPol,
                  uint16_t            startPolIndex,
                  adk_handle_t        *saNsHandle)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_status_t        laConfigRetValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t    sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p = NULL;
    ipsec_sa_idx_t      *ipsecSaIndex = NULL;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;
    uint32_t  polIndex;
    uint32_t  saIndex = ADK_IPSEC_INVALID_SA_ID, i;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE("Invalid IPSec module handle %"ADKPRIh"\n", ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = (ipsec_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipsecHandle);

    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);
    
    retValue = adk_ipsec_crypto_dev_init(sysHandle,
                                         ipsecHandle,
                                         ipsecModuleInfo_p);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                            "Crypto Devices initialization failed\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
    }

    retValue = adk_ipsec_verify_add_sa_input_args(sysHandle, ipsecHandle,
                                                  ipsecModuleInfo_p,
                                                  secAsc_p, saNsHandle,
                                                  numPol,
                                                  startPolIndex);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                            "ADD_SA: Failed \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ADK_LOG_INFO(sysHandle, ipsecHandle,
            "ADD_SA: ipsecHandle: %"ADKPRIh"\n",
            ipsecHandle);

    ipsecSaIndex = (ipsec_sa_idx_t *)
        ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
		 ADK_IPSEC_SA_ID_OBJ, 1, sizeof(ipsec_sa_idx_t));
    if (ipsecSaIndex == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                        "IPSEC: Failed to allocate memory\n");

        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    /*  Prior to any SA additions, the stale array is purged of stale entries that
     *  do not have any inflight packets.
     */

    retValue = adk_ipsec_delete_stale_entries(sysHandle,
                                   ipsecHandle,
                                   ipsecModuleInfo_p);
    if (retValue != ADK_SUCCESS)
    {
        /*  Above function will fail if the egress_stats_id_get function
         *  returns an error. This will be applicable only for inline.
         *  In such an error condition, the update functionality shouldn't
         *  get interrupted, since only the deletion of internally managed
         *  stale entries has failed here. API functionality will continue to
         *  proceed after logging the error.
         */
        ADK_IPSEC_LOG_INFO(ipsecHandle, "Failed to retrieve Tx notification counter");
    }

    LOGSecAsc(secAsc_p,"ADD_SA - ");
    if (secAsc_p->SASPPContext.direction == ADK_IPSEC_OUTBOUND)
    {
        saIndex = adk_rm_resource_reserve(ipsecModuleInfo_p->eSASetHdl);
        if (saIndex == ADK_RM_RSRC_INVALID)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                                "Reached max number of IPsec SAs\n");
            ADK_NVM_FREE (ipsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        retValue = adk_ipsec_add_eSA(sysHandle, ipsecHandle,
                                     ipsecModuleInfo_p, &saIndex,
                                     secAsc_p, dscpVal);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
            ADK_NVM_FREE (ipsecSaIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                                "ADD_eSA: Failed\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        ipsecSaIndex->index = saIndex;
        ipsecModuleInfo_p->numEgSAEntries++;
    }
    else
    {
        saIndex = adk_rm_resource_reserve(ipsecModuleInfo_p->iSASetHdl);
        if (saIndex == ADK_RM_RSRC_INVALID)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                                "Reached max number of IPsec SAs\n");
            ADK_NVM_FREE (ipsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        /* need to update the SPI table */
        retValue = adk_ipsec_add_iSA(sysHandle, ipsecHandle, ipsecModuleInfo_p,
                                     &saIndex, secAsc_p,
                                     dscpVal);
        if((ADK_IPSEC_SA_MODE_INLINE == secAsc_p->saMode) && (ADK_E_LA_CRYPTO_CONFIG_FAILED == retValue))
        {
            ADK_IPSEC_LOG_WARNING(ipsecHandle, "ADD_iSA: Failed to configure Lookaside crypto,"
                                 "only inline decryption operation is supported\n");
            laConfigRetValue = ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }
        else if ((retValue != ADK_IPSEC_SUCCESS) ||
                 ((ADK_IPSEC_SA_MODE_LOOKASIDE == secAsc_p->saMode) &&
                 (ADK_E_LA_CRYPTO_CONFIG_FAILED == retValue)))
        {
            /* TODO how to handle this?
             */
            adk_rm_resource_free(ipsecModuleInfo_p->iSASetHdl, saIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                                "ADD_iSA: Failed\n");
            ADK_NVM_FREE (ipsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        ipsecSaIndex->index = saIndex;
        ipsecModuleInfo_p->numIngSAEntries++;
        ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].iSAHandle =
                (adk_handle_t)ipsecSaIndex;
    }

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
    if(numPol)
    {
        uint16_t index=0;
        while(ADK_SUCCESS == adk_acl_next_rule_get(
                    ipsecModuleInfo_p->cpSpdAclList, &cpRule))
        {
            if(index==startPolIndex)
            {
                ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
                break;
            }
            index++;
        }

        if (NULL == ipsecSecPolLocal)
        {
            if (secAsc_p->SASPPContext.direction == ADK_IPSEC_OUTBOUND)
            {
                adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
            }
            else
            {
                adk_rm_resource_free(ipsecModuleInfo_p->iSASetHdl, saIndex);
            }
            ADK_NVM_FREE (ipsecSaIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                    "The policy is not found in CP ACL List\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
        }

        for (i=0; i< numPol; i++)
        {
            /* key assumption here is that, same SA (ingress or egress) should be
             * applied to multiple policies in a single rule set ID and not across
             * rule set IDs. If that really becomes a requirement, then this will
             * break and need to redesign this
             */
            polIndex = ipsecSecPolLocal->policyResourceID;

            if (secAsc_p->SASPPContext.direction == ADK_IPSEC_OUTBOUND)
            {
                ipsecModuleInfo_p->dpTbl->polSAInfo[polIndex].eSAIdx = saIndex;
                ipsecModuleInfo_p->dpTbl->polSAInfo[polIndex].eSAstate = VALID_SA;
            }
            else
            {
                if (ipsecModuleInfo_p->dpTbl->polSAInfo[polIndex].preFragEnable &&
                    ipsecModuleInfo_p->dpTbl->polSAInfo[polIndex].useSingleSAInPrefrag)
                {
                    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].useSingleSaInPreFrag = TRUE;
                    ipsecModuleInfo_p->dpTbl->iSATbl[saIndex].policy_version++;
                }
            }
            retValue = adk_acl_next_rule_get(ipsecModuleInfo_p->cpSpdAclList, &cpRule);
            /* Here, we are not resetting to HEAD in the CP ACL list. So we should return
             * error only when we didn't iterate the loop till numPol but returned value is 
             * ADK_E_LIST_NOT_FOUND.
             */
            if ((ADK_SUCCESS != retValue) && (i+1 < numPol))
            {
                if (secAsc_p->SASPPContext.direction == ADK_IPSEC_OUTBOUND)
                {
                    adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
                }
                else
                {
                    adk_rm_resource_free(ipsecModuleInfo_p->iSASetHdl, saIndex);
                }
                ADK_NVM_FREE (ipsecSaIndex);
                ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                        "The policy is not found in CP ACL List\n");
                ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
                ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
            }
            else
            {
                /* Resetting retValue to ADK_IPSEC_SUCCESS here or else
                 * API would be returning error code of adk_cp_acl_list_next_rule_get 
                 * to user application.
                 */
                retValue = ADK_IPSEC_SUCCESS;
            }
            ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
        }
    }
    if (secAsc_p->SASPPContext.direction == ADK_IPSEC_OUTBOUND)
    {
        retValue = adk_sys_flow_cache_update_version(sysHandle);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
            ADK_NVM_FREE (ipsecSaIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                    "Could not update Flow cache version\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
        }
    }
    *saNsHandle = (adk_handle_t ) ipsecSaIndex;
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    if((ADK_E_LA_CRYPTO_CONFIG_FAILED != laConfigRetValue)
       || (ADK_IPSEC_SUCCESS != retValue))
    {
        ADK_RETURN(retValue);
    }
    else
    {
        ADK_RETURN(laConfigRetValue);
    }
} /* -----  end of function adk_ipsec_add_sa  ----- */

/*!
 * =============================================================
 * @fn adk_ipsec_del_sa
 *
 * @param[in]    ipsecHandle       - IPsec module handle.
 * @param[in]    saDir             - Indicates the direction of SA need to be deleted.
 * @param[in]    saNsHandle        - SA namespace index.
 * @param[in]    numPol            - Number of policies attached to this SA. Indexes for 
 *                                   these policies should be given from startPolIndex.
 * @param[in]    startPolIndex     - Starting policy index in the beginning of Policy indexes list.
 *                                   Note that number of policies attached to this SA should be in 
 *                                   contiguous order.
 * @retval    ADK_IPSEC_SUCCESS                 - Successfully deleted new Security Policy
 * @retval    ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                                is invalid.
 * @retval    ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 * @retval    ADK_E_IPSEC_INVALID_SA            - input security association pointer is NULL
 *
 * @brief
 *
 * This API allows the application to delete a security association from
 * ACP infrastructure
 * STEP 1: If saDir== OUTBOUND and if polIndex is not NULL,Set eSAIndex
 *         for the corresponding policy with zero and 
 *        set eSAState invalid. Free the egress SA namespace indexed by eSANshandle
 * STEP 2: If saDir==INBOUND,
 *         Free the ingress SA namespace indexed by iSAHandle
 * =============================================================
 */
adk_status_t
adk_ipsec_del_sa(adk_module_handle_t      ipsecHandle,
                 adk_ipsec_sa_direction_t saDir,
                 adk_handle_t             saNsHandle,
                 uint8_t                  numPol,
                 uint16_t                 startPolIndex)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t    sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p = NULL;
    ipsec_sa_idx_t      *ipsecSaIndex = NULL;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    retValue = adk_ipsec_verify_del_sa_input_args(sysHandle, ipsecHandle,
                                                  ipsecModuleInfo_p,
                                                  saDir, numPol,
                                                  startPolIndex,
                                                  saNsHandle);

    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                            "DEL_SA:Failed\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ipsecSaIndex = (ipsec_sa_idx_t *)saNsHandle;

    if (saDir == ADK_IPSEC_OUTBOUND)
    {
        /* first of all, let us stop accepting packets on this SA
         * by making the SA invalid. This will not affect inflight packets.
         */
        retValue = adk_ipsec_stop_eSA(ipsecHandle, ipsecModuleInfo_p,
                           ipsecSaIndex->index, numPol, startPolIndex);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                                "Failed while stopping traffic on this SA\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
        }
        ipsecModuleInfo_p->dpTbl->eSATbl[ipsecSaIndex->index].bytesUsed = 0;

        /* The SA should now be added to the array of stale entries
         * Operations on the stale SA that remain :
         * 1. Destroying the CPM session (in case of inline)
         * 2. Deleting the session (in case of lookaside)
         * 3. Reducing the number of SAs using the device (in case of lookaside)
         */

        if (ipsecModuleInfo_p->numStaleEntries < ADK_IPSEC_NUM_OF_SA)
        {
            ipsecModuleInfo_p->staleSaEntries[ipsecModuleInfo_p->numStaleEntries++] =
                (ADK_IPSEC_OUTBOUND_SA_MASK | (ipsecSaIndex->index));
        }
        ipsecModuleInfo_p->numEgSAEntries--;
        retValue = adk_sys_flow_cache_update_version(sysHandle);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                    "Could not update Flow cache version\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
        }
    }
    else
    {
        /* first of all, let us stop accepting packets on this SA
         * by making the SA invalid in the SPI hash table. Note that 'polSAInfo'
         * shouldn't be tampered with here, as there may be inflight packets
         * that are pending post-decryption SA validation.
         */
        adk_ipsec_invalidate_iSA (sysHandle, ipsecHandle, ipsecModuleInfo_p,
                                  ipsecSaIndex->index);
        ipsecModuleInfo_p->dpTbl->iSATbl[ipsecSaIndex->index].bytesUsed = 0;

        /* The SA should now be added to the array of stale entries.
         * Operations on the stale SA that remain to be done :
         * 1. Invalidating the SA ID in polSAInfo
         * 2. Deleting the session (in case of lookaside)
         * 3. Reducing the number of SAs using the device (in case of lookaside)
         */

        if (ipsecModuleInfo_p->numStaleEntries < ADK_IPSEC_NUM_OF_SA)
        {
            ipsecModuleInfo_p->staleSaEntries[ipsecModuleInfo_p->numStaleEntries++] =
                (ADK_IPSEC_INBOUND_SA_MASK | (ipsecSaIndex->index));
        }
    }

    ADK_NVM_FREE(ipsecSaIndex);

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_ipsec_update_sa
 *
 * @param[in]    ipsecHandle                  - IPsec module handle.
 * @param[in]    *newSecAscUpdate_p           - Pointer to security association data structure.
 * @param[in]    dscpVal                      - dscp value is required for ingress sa.
 *                                              It should be same as configured in the corresponding
 *                                              policy configuration.
 * @param[in]    numPol                       - Number of policies attached to this SA. Indexes for 
 *                                              these policies should be given from startPolIndex.
 * @param[in]    startPolIndex                - Starting policy index in the beginning of Policy indexes list.
 *                                              Note that number of policies attached to this SA should be in 
 *                                              contiguous order.
 * @param[in]    oldSaNsHandle                - Handle of the SA which needs to be updated.
 * @param[in]    *newSaNsHandle               - Handle of the new SA created.
 * @retval ADK_IPSEC_SUCCESS                  - Successfully updated new Security Policy
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE  - i/p ipsec module handle
 *                                              is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM    - Invalid i/p parameter.
 * @retval ADK_E_IPSEC_INVALID_SA             - input security association pointer is NULL
 * @retval ADK_E_IPSEC_POLICY_NS_MEM_GET_FAIL - RTE call to get policy namespace memory failed.
 * @retval ADK_E_LA_CRYPTO_CONFIG_FAILED      - Inline session is configured successfully but failed
 *                                              to configure lookaside crypto session when SA mode
 *                                              is inline and direction is inbound.
 *
 * @brief
 *
 * This API allows the application to update a new security association into
 * ACP infrastructure
 * NOTE: Currently IPsec ADK allows only 64 bits and 128 bits anti replay
 *       window size configuration.
 *       Enabling sequence rollover is not allowed when ESN "extended sequence
 *       number" is enabled in outbound SA
 *
 * In case of egress SA update, older SA index would be marked as stale and added to a repository
 * of stale SAs. Here, older SA Handle would be deleted.
 * In case of ingress SA update, the older SA index would not be considered as stale since the new SA
 * being added is considered as one of the many inbound SAs for the given policy. 
 * Here, new SA handle is returned but the older SA Handle would be preserved. 
 * =============================================================
 */
adk_status_t
adk_ipsec_update_sa (adk_module_handle_t ipsecHandle,
                     adk_ipsec_upd_sa_t  *newSecAscUpdate_p,
                     uint8_t             dscpVal,
                     uint8_t             numPol,
                     uint16_t            startPolIndex,
                     adk_handle_t        oldSaNsHandle,
                     adk_handle_t        *newSaNsHandle)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_status_t        laConfigRetValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t    sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p = NULL;
    adk_ipsec_sa_t      saConfig;
    ipsec_sa_idx_t *oldIpsecSaIndex = (ipsec_sa_idx_t *)oldSaNsHandle;
    ipsec_sa_idx_t *newIpsecSaIndex = NULL;
    uint32_t  polIndex, i, saIndex = 0;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    ipsecModuleInfo_p = (ipsec_module_info_t *) ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    retValue = adk_ipsec_verify_upd_sa_input_args(sysHandle, ipsecHandle,
                                                  ipsecModuleInfo_p,
                                                  newSecAscUpdate_p, oldSaNsHandle,
                                                  numPol, startPolIndex);
    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    /*  Due to limitation of CPM, a particular SA entry cannot be updated
     *  in-place. An updation would have to be treated as creation of a new
     *  SA. This behavior is made applicable for lookaside as well, in order
     *  to keep functionality uniform.The SA index would need to be changed
     *  in 'polSAInfo' as well, so that the correct SA index is picked up
     *  during data plane processing. Updation of the SA index also serves as
     *  an atomic way of switching SA indices on updation, since future packets
     *  can begin using the updated SA index and associated information. This
     *  updation of index however, should be done after the new SA is created and
     *  the eSATbl/iSATbl is written to with the latest information, in order to
     *  make sure there is consistency.
     *
     *  This change of scheme also brings the concept of making an SA
     *  'stale'. In the case of outbound SAs, when an SA is updated, since there
     *  would exist two SAs referring to the same SA entity, the old entry needs
     *  to be deleted. In order to differentiate between the two representatives
     *  of the same SA, the older one is marked as stale and added to a
     *  repository of stale SAs.
     *
     *  Prior to any operations, the stale array is purged of stale entries that
     *  do not have any inflight packets.
     */

    retValue = adk_ipsec_delete_stale_entries(sysHandle,
                                   ipsecHandle,
                                   ipsecModuleInfo_p);

    if (retValue != ADK_SUCCESS)
    {
        /*  Above function will fail if the egress_stats_id_get function
         *  returns an error. This will be applicable only for inline.
         *  In such an error condition, the update functionality shouldn't
         *  get interrupted, since only the deletion of internally managed
         *  stale entries has failed here. API functionality will continue to
         *  proceed after logging the error. TODO - TBD on error message to be
         *  logged, since we don't want to expose existence of stale arrays to
         *  application
         */
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                            "Error in deleting stale entries\n");
    }
    newIpsecSaIndex = (ipsec_sa_idx_t *)
           ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
           ADK_IPSEC_SA_ID_OBJ, 1, sizeof(ipsec_sa_idx_t));

    if (newIpsecSaIndex == NULL)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                            "IPSEC: Failed to allocate memory\n");
        ADK_NVM_FREE (oldIpsecSaIndex);
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_MEM_UNAVAILABLE);
    }

    LOGSecAscUpdate(newSecAscUpdate_p,"UPDATE_SA - ");

    if (newSecAscUpdate_p->SASPPContext.direction == ADK_IPSEC_OUTBOUND)
    {
        saIndex = adk_rm_resource_reserve(ipsecModuleInfo_p->eSASetHdl);
        if (saIndex == ADK_RM_RSRC_INVALID)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                                "Reached max number of IPsec SAs\n");
            ADK_NVM_FREE (oldIpsecSaIndex);
            ADK_NVM_FREE (newIpsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }

        ADK_MEMCPY(&saConfig,
                   &ipsecModuleInfo_p->eSAConfig[oldIpsecSaIndex->index],
                   sizeof(adk_ipsec_sa_t));
        ADK_MEMCPY(&saConfig.SASPPContext, &newSecAscUpdate_p->SASPPContext,
                   sizeof(adk_crypto_ipsec_context_t));
        retValue = adk_ipsec_add_eSA(sysHandle, ipsecHandle,
                                     ipsecModuleInfo_p, &saIndex,
                                     &saConfig, dscpVal);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            /* TODO how to handle this?
             */
            adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                                "update_eSA: Failed\n");
            ADK_NVM_FREE (oldIpsecSaIndex);
            ADK_NVM_FREE (newIpsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
        if(numPol)
        {
            uint16_t index=0;
            while(ADK_SUCCESS == adk_acl_next_rule_get(
                        ipsecModuleInfo_p->cpSpdAclList, &cpRule))
            {
                if(index==startPolIndex)
                {
                    ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
                    break;
                }
                index++;
            }

            if (NULL == ipsecSecPolLocal)
            {
                adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
                ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                        "The policy is not found in CP ACL List\n");
                ADK_NVM_FREE (oldIpsecSaIndex);
                ADK_NVM_FREE (newIpsecSaIndex);
                ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
                ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
            }

            for (i = 0; i < numPol; i++)
            {
                polIndex = ipsecSecPolLocal->policyResourceID;
                ipsecModuleInfo_p->dpTbl->polSAInfo[polIndex].eSAIdx = saIndex;
                ipsecModuleInfo_p->dpTbl->polSAInfo[polIndex].eSAstate = VALID_SA;
                retValue = adk_acl_next_rule_get(ipsecModuleInfo_p->cpSpdAclList, &cpRule);
                /* Here, we are not resetting to HEAD in the CP ACL list. So we should return
                 * error only when we didn't iterate the loop till numPol but returned value is 
                 * ADK_E_LIST_NOT_FOUND.
                 */
                if ((ADK_SUCCESS != retValue) && (i+1 < numPol))
                {
                    adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
                    ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                            "The policy is not found in CP ACL List\n");
                    ADK_NVM_FREE (oldIpsecSaIndex);
                    ADK_NVM_FREE (newIpsecSaIndex);
                    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
                    ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
                }
                else
                {
                    /* Resetting retValue to ADK_IPSEC_SUCCESS here or else
                     * API would be returning error code of adk_cp_acl_list_next_rule_get 
                     * to user application.
                     */
                    retValue = ADK_IPSEC_SUCCESS;
                }
                ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
            }

            if (ipsecModuleInfo_p->numStaleEntries < ADK_IPSEC_NUM_OF_SA)
            {
                ipsecModuleInfo_p->staleSaEntries[ipsecModuleInfo_p->numStaleEntries++] =
                    (ADK_IPSEC_OUTBOUND_SA_MASK | (oldIpsecSaIndex->index));
            }
        }
        retValue = adk_sys_flow_cache_update_version(sysHandle);
        if (retValue != ADK_IPSEC_SUCCESS)
        {
            adk_rm_resource_free(ipsecModuleInfo_p->eSASetHdl, saIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INTERNAL_ERROR,
                    "Could not update Flow cache version\n");
            ADK_NVM_FREE (oldIpsecSaIndex);
            ADK_NVM_FREE (newIpsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_INTERNAL_ERROR);
        }
        newIpsecSaIndex->index = saIndex;
        ADK_NVM_FREE(oldIpsecSaIndex);
    }
    else
    {
        /* the new scheme works for inbound direction as well, since in
         * case of ingress SA update, because the SPI is new,
         * this is equivalent of adding a new SA. However, the previous SA ID
         * will not be considered stale, since the new SA being added is
         * considered as one of many inbound SAs for a given policy.
         */
        saIndex = adk_rm_resource_reserve(ipsecModuleInfo_p->iSASetHdl);
        if (saIndex == ADK_RM_RSRC_INVALID)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_MEM_UNAVAILABLE,
                                "Reached max number of IPsec SAs\n");
            ADK_NVM_FREE (oldIpsecSaIndex);
            ADK_NVM_FREE (newIpsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }

        ADK_MEMCPY(&saConfig,
                   &ipsecModuleInfo_p->iSAConfig[oldIpsecSaIndex->index],
                   sizeof(adk_ipsec_sa_t));
        ADK_MEMCPY(&saConfig.SASPPContext, &newSecAscUpdate_p->SASPPContext,
                   sizeof(adk_crypto_ipsec_context_t));
        /* need to update the SPI table */
        retValue = adk_ipsec_add_iSA(sysHandle, ipsecHandle, ipsecModuleInfo_p,
                                     &saIndex, &saConfig,
                                     dscpVal);
        if((ADK_IPSEC_SA_MODE_INLINE == saConfig.saMode) && (ADK_E_LA_CRYPTO_CONFIG_FAILED == retValue))
        {
            ADK_IPSEC_LOG_WARNING(ipsecHandle, "ADD_iSA: Failed to configure Lookaside crypto,"
                                 "only inline decryption operation is supported\n");
            laConfigRetValue = ADK_E_LA_CRYPTO_CONFIG_FAILED;
        }
        else if ((retValue != ADK_IPSEC_SUCCESS) ||
                 ((ADK_IPSEC_SA_MODE_LOOKASIDE == saConfig.saMode) &&
                 (ADK_E_LA_CRYPTO_CONFIG_FAILED == retValue)))
        {
            adk_rm_resource_free(ipsecModuleInfo_p->iSASetHdl, saIndex);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                                "ADD_iSA: Failed\n");
            ADK_NVM_FREE (oldIpsecSaIndex);
            ADK_NVM_FREE (newIpsecSaIndex);
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
        adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);

        ipsecModuleInfo_p->numIngSAEntries++;

        newIpsecSaIndex->index = saIndex;
    }
    *newSaNsHandle = (adk_handle_t ) newIpsecSaIndex;
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    if((ADK_E_LA_CRYPTO_CONFIG_FAILED != laConfigRetValue)
       || (ADK_IPSEC_SUCCESS != retValue))
    {
        return retValue;
    }
    else
    {
        return laConfigRetValue;
    }
}

/*!
 * =============================================================
 * @fn adk_ipsec_get_sa
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 * @param[in]    saNsHandle  - Identify the SA for which get SA is required
 * @param[in]    saDirection - security association direction (EGRESS or INGRESS)
 * @param[out]   *secAsc_p   - Memory pointed at by the caller of the API for copying the required SA.
 *
 * @retval ADK_IPSEC_SUCCESS                 - Successfully retrieved new Security Policy
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 *
 * @brief
 *
 * This API allows the application to get a specific security association.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_sa (adk_module_handle_t        ipsecHandle,
                  adk_handle_t               saNsHandle,
                  adk_ipsec_sa_direction_t   saDirection,
                  adk_ipsec_sa_t             *secAsc_p)
{
    adk_status_t          retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    uint32_t saIdx;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    retValue = adk_ipsec_verify_get_sa_input_args(ipsecHandle,
                                                  ipsecModuleInfo_p,
                                                  saNsHandle,
                                                  saDirection);

    if (retValue != ADK_IPSEC_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                            "GET_SA: Failed \n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    saIdx = ((ipsec_sa_idx_t *)saNsHandle)->index;

    if (saDirection == ADK_IPSEC_OUTBOUND)
    {
        ADK_MEMCPY(secAsc_p,
            &ipsecModuleInfo_p->eSAConfig[saIdx],
            sizeof(adk_ipsec_sa_t));
    }
    else
    {
        ADK_MEMCPY(secAsc_p,
            &ipsecModuleInfo_p->iSAConfig[saIdx],
            sizeof(adk_ipsec_sa_t));
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return (ADK_IPSEC_SUCCESS);
}

/*!
 * =============================================================
 * @fn adk_ipsec_get_stat
 *
 * @param[in]    ipsecHandle - IPsec module handle.
 *
 * @param[out]   ipsecStat_p - IPsec related statistics data structure pointer
 *               gets set by the IPsec ADK
 *
 * @retval ADK_IPSEC_SUCCESS - Successfully added new Security Policy
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 *
 * @brief
 *
 * This API allows the application to retreive all IPsec related staistics
 * maintained by the IPsec ADK module. It is expected that the application
 * would have allocated a chunk of memory equivalent to hold data structure of
 * size adk_ipsec_stat_t.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_stat (adk_module_handle_t  ipsecHandle,
                    adk_ipsec_ip_version_t  ipVersion,
                    adk_ipsec_stat_t     *ipsecStat_p)
{
    adk_sys_handle_t      sysHandle;
    adk_status_t retValue = ADK_IPSEC_SUCCESS;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    adk_ipsec_stat_t      *stats;

    uint16_t  i;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (ipsecStat_p == NULL)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ADK_MEMSET (ipsecStat_p, 0x00, sizeof(adk_ipsec_stat_t));

    if (ipVersion == ADK_IPSEC_IPV4)
    {
        for (i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            stats = &ipsecModuleInfo_p->dpTbl->ipsecGblStatsV4[i];
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_pol_discard_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_in_pol_discard_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_pol_bypass_count +=
                    stats->ingressStats.ipsec_ing_global_spd_in_pol_bypass_count;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_default_discard_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_in_default_discard_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_keepalive_pdu_drop +=
                    stats->ingressStats.ipsec_ing_global_keepalive_pdu_drop ;
            ipsecStat_p->ingressStats.ipsec_ing_global_unknown_natt_pdu_drop +=
                    stats->ingressStats.ipsec_ing_global_unknown_natt_pdu_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_ip_fragment_drop +=
                    stats->ingressStats.ipsec_ing_global_ip_fragment_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_sa_not_present_drop +=
                    stats->ingressStats.ipsec_ing_global_sa_not_present_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_auth_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_auth_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_pad_verification_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pad_verification_fail_drop ;
            ipsecStat_p->ingressStats.ipsec_ing_global_seqno_check_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_seqno_check_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_sa_hard_life_expire_drop +=
                    stats->ingressStats.ipsec_ing_global_sa_hard_life_expire_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_sa_validation_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_sa_validation_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_out_default_discard_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_out_default_discard_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_out_not_ipsec_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_out_not_ipsec_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_dummy_pkt_drop +=
                    stats->ingressStats.ipsec_ing_global_dummy_pkt_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_exception_not_registered +=
                    stats->ingressStats.ipsec_ing_global_exception_not_registered;
            ipsecStat_p->ingressStats.ipsec_ing_global_tfc_pad_drop +=
                    stats->ingressStats.ipsec_ing_global_tfc_pad_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_inline_decryption_failed +=
                    stats->ingressStats.ipsec_ing_global_inline_decryption_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_la_send_to_crypto_failed +=
                    stats->ingressStats.ipsec_ing_global_la_send_to_crypto_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_pad_removal_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pad_removal_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_arw_update_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_arw_update_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_send_to_next_stage_failed +=
                    stats->ingressStats.ipsec_ing_global_send_to_next_stage_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_pkt_process_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pkt_process_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_buffered_single_sa_pkt_process_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_buffered_single_sa_pkt_process_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_mbuf_clone_failed +=
                    stats->ingressStats.ipsec_ing_global_mbuf_clone_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_too_small_pkt_drop +=
                    stats->ingressStats.ipsec_ing_global_too_small_pkt_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_la_no_buffer_drop +=
                    stats->ingressStats.ipsec_ing_global_la_no_buffer_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_pad_alloc_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pad_alloc_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_reassembly_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_reassembly_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_invalid_dst_fn_id_drop +=
                    stats->ingressStats.ipsec_ing_global_invalid_dst_fn_id_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_lookup_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_in_lookup_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_unsupported_l4_proto_drop +=
                    stats->ingressStats.ipsec_ing_global_unsupported_l4_proto_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_exception_send_failure_drop +=
                    stats->ingressStats.ipsec_ing_global_exception_send_failure_drop;

            ipsecStat_p->egressStats.ipsec_eg_global_spd_pol_discard_drop +=
                    stats->egressStats.ipsec_eg_global_spd_pol_discard_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_exception_not_registered +=
                    stats->egressStats.ipsec_eg_global_exception_not_registered;
            ipsecStat_p->egressStats.ipsec_eg_global_spd_pol_default_discard_drop +=
                    stats->egressStats.ipsec_eg_global_spd_pol_default_discard_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_seqno_overflow_drop +=
                    stats->egressStats.ipsec_eg_global_seqno_overflow_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_sa_hardlife_expire_drop +=
                    stats->egressStats.ipsec_eg_global_sa_hardlife_expire_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_sa_not_found_drop +=
                    stats->egressStats.ipsec_eg_global_sa_not_found_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_spd_pol_bypassCnt +=
                    stats->egressStats.ipsec_eg_global_spd_pol_bypassCnt;
            ipsecStat_p->egressStats.ipsec_eg_global_prefrag_reqd_df_bit_set_drop_count +=
                    stats->egressStats.ipsec_eg_global_prefrag_reqd_df_bit_set_drop_count;
            ipsecStat_p->egressStats.ipsec_eg_global_pre_frag_error_drop +=
                    stats->egressStats.ipsec_eg_global_pre_frag_error_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_pad_alloc_fail_drop +=
                    stats->egressStats.ipsec_eg_global_pad_alloc_fail_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_no_security_session_drop +=
                    stats->egressStats.ipsec_eg_global_no_security_session_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_la_send_to_crypto_failed +=
                    stats->egressStats.ipsec_eg_global_la_send_to_crypto_failed;
            ipsecStat_p->egressStats.ipsec_eg_global_la_no_buffer_drop +=
                    stats->egressStats.ipsec_eg_global_la_no_buffer_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_too_small_pkt_drop +=
                    stats->egressStats.ipsec_eg_global_too_small_pkt_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_la_crypto_op_failed +=
                    stats->egressStats.ipsec_eg_global_la_crypto_op_failed;
            ipsecStat_p->egressStats.ipsec_eg_global_mbuf_clone_failed +=
                    stats->egressStats.ipsec_eg_global_mbuf_clone_failed;
            ipsecStat_p->egressStats.ipsec_eg_global_outer_hdr_buf_alloc_fail_drop +=
                    stats->egressStats.ipsec_eg_global_outer_hdr_buf_alloc_fail_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_exception_send_failure_drop +=
                    stats->egressStats.ipsec_eg_global_exception_send_failure_drop;
        }
    }
    else
    {
        for (i=0; i<ADK_MAX_NUM_CORES; i++)
        {
            stats = &ipsecModuleInfo_p->dpTbl->ipsecGblStatsV6[i];
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_pol_discard_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_in_pol_discard_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_pol_bypass_count +=
                    stats->ingressStats.ipsec_ing_global_spd_in_pol_bypass_count;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_default_discard_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_in_default_discard_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_keepalive_pdu_drop +=
                    stats->ingressStats.ipsec_ing_global_keepalive_pdu_drop ;
            ipsecStat_p->ingressStats.ipsec_ing_global_unknown_natt_pdu_drop +=
                    stats->ingressStats.ipsec_ing_global_unknown_natt_pdu_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_ip_fragment_drop +=
                    stats->ingressStats.ipsec_ing_global_ip_fragment_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_sa_not_present_drop +=
                    stats->ingressStats.ipsec_ing_global_sa_not_present_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_auth_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_auth_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_pad_verification_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pad_verification_fail_drop ;
            ipsecStat_p->ingressStats.ipsec_ing_global_seqno_check_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_seqno_check_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_sa_hard_life_expire_drop +=
                    stats->ingressStats.ipsec_ing_global_sa_hard_life_expire_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_sa_validation_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_sa_validation_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_out_default_discard_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_out_default_discard_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_out_not_ipsec_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_out_not_ipsec_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_dummy_pkt_drop +=
                    stats->ingressStats.ipsec_ing_global_dummy_pkt_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_exception_not_registered +=
                    stats->ingressStats.ipsec_ing_global_exception_not_registered;
            ipsecStat_p->ingressStats.ipsec_ing_global_tfc_pad_drop +=
                    stats->ingressStats.ipsec_ing_global_tfc_pad_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_inline_decryption_failed +=
                    stats->ingressStats.ipsec_ing_global_inline_decryption_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_la_send_to_crypto_failed +=
                    stats->ingressStats.ipsec_ing_global_la_send_to_crypto_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_pad_removal_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pad_removal_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_arw_update_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_arw_update_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_send_to_next_stage_failed +=
                    stats->ingressStats.ipsec_ing_global_send_to_next_stage_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_pkt_process_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pkt_process_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_buffered_single_sa_pkt_process_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_buffered_single_sa_pkt_process_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_mbuf_clone_failed +=
                    stats->ingressStats.ipsec_ing_global_mbuf_clone_failed;
            ipsecStat_p->ingressStats.ipsec_ing_global_too_small_pkt_drop +=
                    stats->ingressStats.ipsec_ing_global_too_small_pkt_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_la_no_buffer_drop +=
                    stats->ingressStats.ipsec_ing_global_la_no_buffer_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_pad_alloc_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_pad_alloc_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_reassembly_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_reassembly_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_invalid_dst_fn_id_drop +=
                    stats->ingressStats.ipsec_ing_global_invalid_dst_fn_id_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_spd_in_lookup_fail_drop +=
                    stats->ingressStats.ipsec_ing_global_spd_in_lookup_fail_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_unsupported_l4_proto_drop +=
                    stats->ingressStats.ipsec_ing_global_unsupported_l4_proto_drop;
            ipsecStat_p->ingressStats.ipsec_ing_global_exception_send_failure_drop +=
                    stats->ingressStats.ipsec_ing_global_exception_send_failure_drop;

            ipsecStat_p->egressStats.ipsec_eg_global_spd_pol_discard_drop +=
                    stats->egressStats.ipsec_eg_global_spd_pol_discard_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_exception_not_registered +=
                    stats->egressStats.ipsec_eg_global_exception_not_registered;
            ipsecStat_p->egressStats.ipsec_eg_global_spd_pol_default_discard_drop +=
                    stats->egressStats.ipsec_eg_global_spd_pol_default_discard_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_seqno_overflow_drop +=
                    stats->egressStats.ipsec_eg_global_seqno_overflow_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_sa_hardlife_expire_drop +=
                    stats->egressStats.ipsec_eg_global_sa_hardlife_expire_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_sa_not_found_drop +=
                    stats->egressStats.ipsec_eg_global_sa_not_found_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_spd_pol_bypassCnt +=
                    stats->egressStats.ipsec_eg_global_spd_pol_bypassCnt;
            ipsecStat_p->egressStats.ipsec_eg_global_pre_frag_error_drop +=
                    stats->egressStats.ipsec_eg_global_pre_frag_error_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_pad_alloc_fail_drop +=
                    stats->egressStats.ipsec_eg_global_pad_alloc_fail_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_no_security_session_drop +=
                    stats->egressStats.ipsec_eg_global_no_security_session_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_la_send_to_crypto_failed +=
                    stats->egressStats.ipsec_eg_global_la_send_to_crypto_failed;
            ipsecStat_p->egressStats.ipsec_eg_global_la_no_buffer_drop +=
                    stats->egressStats.ipsec_eg_global_la_no_buffer_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_too_small_pkt_drop +=
                    stats->egressStats.ipsec_eg_global_too_small_pkt_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_la_crypto_op_failed +=
                    stats->egressStats.ipsec_eg_global_la_crypto_op_failed;
            ipsecStat_p->egressStats.ipsec_eg_global_mbuf_clone_failed +=
                    stats->egressStats.ipsec_eg_global_mbuf_clone_failed;
            ipsecStat_p->egressStats.ipsec_eg_global_outer_hdr_buf_alloc_fail_drop +=
                    stats->egressStats.ipsec_eg_global_outer_hdr_buf_alloc_fail_drop;
            ipsecStat_p->egressStats.ipsec_eg_global_exception_send_failure_drop +=
                    stats->egressStats.ipsec_eg_global_exception_send_failure_drop;
        }
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * =============================================================
 * @fn adk_ipsec_get_sa_stat
 *
 * @param[in]    ipsecHandle      - IPsec module handle.
 * @param[in]    saNsHandle       - identifies the SA for which the SA statistics
 *                                  are required.
 * @param[in]    saDirection      - security association direction (EGRESS or INGRESS)
 *
 * @param[out]   *saStat_p        - pointer of SA stat data structure type carrying all
 *                                  the required statistics
 *
 * @retval ADK_IPSEC_SUCCESS                 - Successfully retrieved SA stats.
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 * @retval ADK_E_IPSEC_SA_NOT_FOUND          - SA not found
 *
 * @brief
 *
 * This API allows the application to retrieve SA specific statistics
 * maintained by the IPsec ADK module. It is expected that the application has
 * allocated a chunk of memory equivalent to size of structure
 * adk_ipsec_sa_stat_t where the ADK will dump the statistics.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_get_sa_stat(adk_module_handle_t         ipsecHandle,
                      adk_handle_t                saNsHandle,
                      adk_ipsec_sa_direction_t    saDirection,
                      adk_ipsec_sa_stat_t         *saStat_p)
{
    adk_status_t            retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t        sysHandle;
    ipsec_module_info_t     *ipsecModuleInfo_p=NULL;
    adk_ipsec_ing_sa_stat_t *ingStats;
    adk_ipsec_eg_sa_stat_t  *egStats;

    uint32_t saIndex;
    uint16_t  i;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    if (saStat_p == NULL || saNsHandle == NULL)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    ADK_MEMSET (saStat_p, 0x00, sizeof(adk_ipsec_sa_stat_t));

    saIndex = ((ipsec_sa_idx_t *)saNsHandle)->index;

    if ( saDirection  == ADK_IPSEC_OUTBOUND )
    {
        if (adk_rm_resource_isset(ipsecModuleInfo_p->eSASetHdl, saIndex))
        {
            for (i=0; i<ADK_MAX_NUM_CORES; i++)
            {
                egStats = &ipsecModuleInfo_p->dpTbl->egSAStats[i][saIndex];
                saStat_p->sa_stat.eg_sa_stat.ipsec_eg_sa_tot_out_pkts +=
                      egStats->ipsec_eg_sa_tot_out_pkts;
            }
        }
        else
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                    "Invalid eSA handle \n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
    }
    else
    {
        if (adk_rm_resource_isset(ipsecModuleInfo_p->iSASetHdl, saIndex))
        {
            for (i=0; i<ADK_MAX_NUM_CORES; i++)
            {
                ingStats = &ipsecModuleInfo_p->dpTbl->ingSAStats[i][saIndex];
                saStat_p->sa_stat.ing_sa_stat.ipsec_ing_sa_tot_in_pkts +=
                      ingStats->ipsec_ing_sa_tot_in_pkts;
                saStat_p->sa_stat.ing_sa_stat.ipsec_ing_sa_seq_num_mismatch_drop +=
                      ingStats->ipsec_ing_sa_seq_num_mismatch_drop;
                saStat_p->sa_stat.ing_sa_stat.ipsec_ing_sa_esp_cryptographic_error +=
                      ingStats->ipsec_ing_sa_esp_cryptographic_error;
                saStat_p->sa_stat.ing_sa_stat.ipsec_ing_sa_traffic_sel_mismatch_drop +=
                      ingStats->ipsec_ing_sa_traffic_sel_mismatch_drop;
            }
        }
        else
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                    "Invalid iSA handle \n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
        }
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * =============================================================
 * adk_ipsec_register_exception
 * @param[in]   ipsecHandle  IPSec module handle.
 *
 * @param[in]   exceptionHandler_p  Pointer to adk_ipsec_exception_handler_t.
 *
 * @retval   ADK_IPSEC_SUCCESS - successfully registerd exception handler
 *
 * @retval  ADK_E_INVALID_HANDLE - Invalid  ipsec handle specified
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function allows the user to register exception handlers
 * for various IPsec SA Trigger/Audit event exceptions. enum adk_ipsec_exception_type_t defines
 * the type of exceptions that are supported by ADK.
 * Application can modify the parameter of already registered exception 
 * using the same register API *
 * =============================================================
 */
adk_status_t
adk_ipsec_register_exception (adk_module_handle_t ipsecHandle,
                  const adk_ipsec_exception_handler_t *exceptionHandler_p)
{
    adk_status_t                      retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t                  sysHandle;
    ipsec_module_info_t               *ipsecModuleInfo_p = NULL;
    adk_ipsec_dp_exception_handler_t *dpExceptionHdl_p = NULL,*temp = NULL;
    uint32_t                          policerFlowId = ADK_QOS_INVALID_POLICER_FLOW_ID;

    uint8_t dpExcepIndex;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    if (!exceptionHandler_p)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                "Invalid input parameter\n");
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipsec_register_exception_exit;
    }


    switch(exceptionHandler_p->exceptionType)
    {
        case ADK_IPSEC_EXCEPTION_SA_ACQUIRE:
        case ADK_IPSEC_EXCEPTION_SA_EXPIRE:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_IP_FRAGMENT_DROP:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_NOT_PRESENT:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SEQNO_CHECK_FAILURE:    
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_AUTH_CHECK_FAILURE:    
        case ADK_IPSEC_EXCEPTION_AUDIT_EG_SEQNO_OVERFLOW:    
        case ADK_IPSEC_EXCEPTION_AUDIT_EG_SPD_OUT_LKUP_DISCARD:    
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPD_IN_LKUP_DISCARD:
        case ADK_IPSEC_EXCEPTION_NO_POLICY_2ND_FRAGMENT_2SA_PREFRAG:
        case ADK_IPSEC_EXCEPTION_EG_PREFRAG_REQD_BUT_DF_BIT_SET:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_MISMATCH:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_DISCARD:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_NO_IPSEC_DROP:
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                "Invalid input parameter\n");
            goto _adk_ipsec_register_exception_exit;
    }

    dpExcepIndex = exceptionHandler_p->exceptionType -1;

    /*change device and queue id, if host policer enabled*/
    if (exceptionHandler_p->policerHandle)
    {
        retValue = adk_qos_policing_flow_id_get(sysHandle, exceptionHandler_p->policerHandle, &policerFlowId);
        if (ADK_SUCCESS != retValue) {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, retValue,
                    "Invalid policer handle\n");
            goto _adk_ipsec_register_exception_exit;
        }
    }

    dpExceptionHdl_p = ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
            ADK_IPSEC_EXC_OBJ, 1, sizeof(adk_ipsec_dp_exception_handler_t));
    if (!dpExceptionHdl_p)
    {
        retValue = ADK_FATAL_SEV_LEVEL_SET(ADK_E_MEM_UNAVAILABLE);
        ADK_IPSEC_LOG_ERROR(ipsecHandle, retValue,
                "Failed to allocate memory\n");
        goto _adk_ipsec_register_exception_exit;
    }

    dpExceptionHdl_p->devId = exceptionHandler_p->devId;
    dpExceptionHdl_p->queueId = exceptionHandler_p->queueId;
    dpExceptionHdl_p->priority = exceptionHandler_p->priority;
    dpExceptionHdl_p->cbFnId = exceptionHandler_p->cbFnHandle;
    dpExceptionHdl_p->policerId = policerFlowId;

    if(ipsecModuleInfo_p->dpTbl->ipsecExceptionHandlerTble[dpExcepIndex])
    {
        temp = ipsecModuleInfo_p->dpTbl->ipsecExceptionHandlerTble[dpExcepIndex];
        ipsecModuleInfo_p->dpTbl->ipsecExceptionHandlerTble[dpExcepIndex] = dpExceptionHdl_p;
        adk_rcu_synchronize();
        ADK_DP_SHM_FREE(temp);
    }
    else
    {
        ipsecModuleInfo_p->dpTbl->ipsecExceptionHandlerTble[dpExcepIndex] = dpExceptionHdl_p;
    }

_adk_ipsec_register_exception_exit:

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);

}    /* -----  end of function adk_ipsec_register_exception  ----- */

/*!
 * =============================================================
 * adk_ipsec_ph_rx_parse adk_ipsec.h "adk_ipsec.h"
 *
 * @param[in]   task Pointer to rte_mbuf.
 *
 * @param[out]  dp_message   Pointer to structure having the decoded parameters
 *                            sent by ipsec data path code to application.
 *
 * @retval   ADK_SUCCESS - successfully decoded the parameters
 * handler
 *
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 *                                    passed. The pointer might be NULL.
 *
 * @brief
 *
 * This function helps in interpreting the parameters present in the task
 * Parameters that IPsec module send along with the packet to the Application
 * via PH Dev.
 *
 * =============================================================
 */

adk_status_t
adk_ipsec_ph_rx_parse (
        void *task,
        void **dp_message,
        adk_module_handle_t ipsecHandle)
{
    adk_status_t ret = ADK_SUCCESS;
    struct rte_mbuf *mbuf_p;
    adk_metadata_t *adk_md_ptr;
    adk_ph_dev_dp_metadata_t *phMeta_p;
    adk_ipsec_dp_exception_data_t *ipsec_excep_data;
    struct timeval tv;

    adk_sys_handle_t                  sysHandle;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_E_INTERNAL_ERROR;
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    if (!task || !dp_message)
    {
        ADK_IPSEC_LOG_FORCE("Invalid input paramenter in [%s]",__func__);
        ret = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        goto _adk_ipsec_ph_rx_parse_decode_exit;
    }

    mbuf_p = (struct rte_mbuf *)task;
    adk_md_ptr = ADK_METADATA_PTR(mbuf_p);
    phMeta_p = (adk_ph_dev_dp_metadata_t *)&(adk_md_ptr->imc_param_start);
    ipsec_excep_data = (adk_ipsec_dp_exception_data_t *)&(phMeta_p->imc_param_start);
    switch (ipsec_excep_data->exceptionType)
    {
        case ADK_IPSEC_EXCEPTION_SA_ACQUIRE:
        {
            adk_ipsec_extract_headers_sa_message(task,&ipsec_excep_data->exceptionData.saAquireData.saMsg);
        }
        break;
        case ADK_IPSEC_EXCEPTION_SA_EXPIRE:
        {
            adk_ipsec_extract_expiry_exception_data(task,&ipsec_excep_data->exceptionData.saExpiryData);
        }
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_IP_FRAGMENT_DROP:
            adk_ipsec_fill_encrypted_pkt_audit_info(task, 
                                                    &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngFragmentedIpsecPktEvent);
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_NOT_PRESENT:
            adk_ipsec_fill_encrypted_pkt_audit_info(task, 
                                                    &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngSaNotPresentEvent);
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SEQNO_CHECK_FAILURE:
            adk_ipsec_fill_encrypted_pkt_audit_info(task, 
                                                    &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngSeqCheckFailureEvent);
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_AUTH_CHECK_FAILURE:
            adk_ipsec_fill_encrypted_pkt_audit_info(task, 
                                                    &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngAuthCheckFailureEvent);
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_EG_SEQNO_OVERFLOW:
            /* Audit event was filled from SA Info in DP */
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_EG_SPD_OUT_LKUP_DISCARD:
            adk_ipsec_fill_before_encr_pkt_audit_info(task,
                                                      &ipsec_excep_data->exceptionData.auditEventData.event.ipsecEgSpdOutLkupDiscardEvent);
        break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPD_IN_LKUP_DISCARD:
            adk_ipsec_fill_before_encr_pkt_audit_info(task,
                                                      &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngSpdInLkupDiscardEvent);
            break;
        case ADK_IPSEC_EXCEPTION_NO_POLICY_2ND_FRAGMENT_2SA_PREFRAG:
            adk_ipsec_fill_before_encr_pkt_audit_info(task,
                                                      &ipsec_excep_data->exceptionData.auditEventData.event.ipsecEgNoPolicy2ndFrag2SaPrefragEvent);
            break;
        case ADK_IPSEC_EXCEPTION_EG_PREFRAG_REQD_BUT_DF_BIT_SET:
            adk_ipsec_fill_before_encr_pkt_audit_info(task,
                                                      &ipsec_excep_data->exceptionData.auditEventData.event.ipsecEgPrefragReqdButDfBitSetEvent);
            break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_DISCARD:
            adk_ipsec_fill_ing_sa_validation_audit_info(task,
                                                        &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngSpdOutLkupDiscardEvent);
            break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_NO_IPSEC_DROP:
            adk_ipsec_fill_ing_sa_validation_audit_info(task,
                                                   &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngSpdOutLkupNoIpsecPolicyEvent);
            break;
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_MISMATCH:
            adk_ipsec_fill_ing_sa_validation_audit_info(task,
                                                        &ipsec_excep_data->exceptionData.auditEventData.event.ipsecIngSaMismatchEvent);
            break;
        default:
            ADK_IPSEC_LOG_FORCE("Invalid exception type [%d] received in [%s]",
                        ipsec_excep_data->exceptionType,__func__);
            ADK_RETURN(ret);
    }
    gettimeofday(&tv, NULL);
    if ((ipsec_excep_data->exceptionType != ADK_IPSEC_EXCEPTION_SA_ACQUIRE) &&
        (ipsec_excep_data->exceptionType != ADK_IPSEC_EXCEPTION_SA_EXPIRE))
    {
        ipsec_excep_data->exceptionData.auditEventData.timeInSec = tv.tv_sec;
        ipsec_excep_data->exceptionData.auditEventData.timeInUsec = tv.tv_usec;
    }

    *dp_message = (adk_ipsec_dp_exception_data_t*)ipsec_excep_data;

_adk_ipsec_ph_rx_parse_decode_exit:

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_RETURN(ret);
}


/*!
 * =============================================================
 * adk_ipsec_unregister_exception
 * @param[in]   ipsecHandle  IPSec module handle.
 *
 * @param[in]   exceptionType of  adk_ipsec_exception_type_t 
 *
 * @retval   ADK_IPSEC_SUCCESS - successfully unregisterd exception
 * handler
 *
 * @retval  ADK_E_INVALID_HANDLE - Invalid  ipsec handle specified
 *
 * @retval  ADK_E_INVALID_PARAMETER - Invalid input parameter
 * passed.
 *
 * @brief
 *
 * This function allows the user to unregister exception handlers
 * for various exceptions. enum adk_ipsec_exception_type_t defines
 * the type of exceptions that are supported by ADK.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_unregister_exception (adk_module_handle_t ipsecHandle,
        const adk_ipsec_exception_type_t exceptionType)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t    sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p;
    adk_ipsec_dp_exception_handler_t *dpExceptionHdl_p = NULL,*temp = NULL;
    uint8_t dpExcepIndex;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    if (!exceptionType)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                "Invalid input parameter\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }


    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    switch(exceptionType)
    {
        case ADK_IPSEC_EXCEPTION_SA_ACQUIRE:
        case ADK_IPSEC_EXCEPTION_SA_EXPIRE:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_IP_FRAGMENT_DROP:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_NOT_PRESENT:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SEQNO_CHECK_FAILURE:    
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_AUTH_CHECK_FAILURE:   
        case ADK_IPSEC_EXCEPTION_AUDIT_EG_SEQNO_OVERFLOW:    
        case ADK_IPSEC_EXCEPTION_AUDIT_EG_SPD_OUT_LKUP_DISCARD:    
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPD_IN_LKUP_DISCARD:
        case ADK_IPSEC_EXCEPTION_NO_POLICY_2ND_FRAGMENT_2SA_PREFRAG:
        case ADK_IPSEC_EXCEPTION_EG_PREFRAG_REQD_BUT_DF_BIT_SET:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SA_MISMATCH:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_DISCARD:
        case ADK_IPSEC_EXCEPTION_AUDIT_ING_SPP_SPD_OUT_LKUP_NO_IPSEC_DROP:
            break;
        default:
            retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                    "Invalid Exception Type\n");
            goto _adk_ipsec_unregister_exception_exit;
    }

    dpExcepIndex =  exceptionType-1;
    if(ipsecModuleInfo_p->dpTbl->ipsecExceptionHandlerTble[dpExcepIndex])
    {
        ipsecModuleInfo_p->dpTbl->ipsecExceptionHandlerTble[dpExcepIndex] = dpExceptionHdl_p;
        adk_rcu_synchronize();
        ADK_DP_SHM_FREE(temp);
    }
    else
    {
        retValue = ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_INVALID_PARAMETER,
                "Couldn't unregister exception \n");
        goto _adk_ipsec_unregister_exception_exit;
    }


_adk_ipsec_unregister_exception_exit:

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    ADK_RETURN(retValue);
}

/*!
 * =============================================================
 * @fn adk_ipsec_prefrag_buffering_timeout_config
 *
 * @param[in]    ipsecHandle -  IPsec module handle.
 *
 * @param[in]    timeOutVal  -  PreFrag buffering timeout in milliseconds used if the inner fragments 
 *                              are forwarded fragments.
 *
 * @retval ADK_SUCCESS              - Successfully configured the timeout value.
 * 
 * @retval ADK_E_INVALID_PARAMETER  - Invalid IPsec module handle.
 *
 * @retval ADK_E_UNSUPPORTED        - Timeout value greater than 2000 seconds or less than 100 milliseconds is not supported.
 *
 * @brief
 *
 * This API configures the global timeout value that can be used only in inbound processing when,
 * 1. preFragEnable and useSingleSAInPrefrag are set.
 * 2. Inner fragments should be forwarded fragments.
 * Timeout value greater than 2000 seconds is not supported.
 * The default preFrag buffering timeout value is 100 milliseconds. Minimum preFrag buffering
 * timeout value that can be configured is 100 milliseconds.
 *
 * If the inner fragments are locally terminated, then the preFrag buffering timeout value configured through 
 * adk_ip_reassembly_timeout_config/adk_ipv6_reassembly_timeout_config would be used.
 * =============================================================
 */
adk_status_t 
adk_ipsec_prefrag_buffering_timeout_config (adk_module_handle_t ipsecHandle,
                                            uint32_t timeOutVal)
{
    adk_status_t         retValue = ADK_E_INTERNAL_ERROR;
    ipsec_module_info_t  *ipsecModuleInfo_p = NULL;
    adk_sys_handle_t    sysHandle; 
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                             ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }
    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    /* Get the ipsec module info data strcuture pointer*/
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);
    if (retValue != ADK_SUCCESS)
    {
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_INVALID_PARAMETER);
    }
    if ((timeOutVal > 2000000) || (timeOutVal < 100))
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_UNSUPPORTED,
                "The given timeout value is not supported\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_UNSUPPORTED);
    }
    ipsecModuleInfo_p->dpTbl->preFragBufferTimeOutVal = timeOutVal;

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

static inline const char *cipherAlgoNames(adk_ipsec_encr_t encr)
{
    switch(encr)
    {
    case ADK_IPSEC_ENCR_DES_CBC: return "DES CBC";
    case ADK_IPSEC_ENCR_3DES_CBC: return "3DES CBC";
    case ADK_IPSEC_ENCR_NULL: return "NULL";
    case ADK_IPSEC_ENCR_AES_CBC: return "AES CBC";
    case ADK_IPSEC_ENCR_AES_CTR: return "AES CTR";
    case ADK_IPSEC_ENCR_AES_CCM_ICV8: return "AES CCM ICV8";
    case ADK_IPSEC_ENCR_AES_CCM_ICV12: return "AES CCM ICV12";
    case ADK_IPSEC_ENCR_AES_CCM_ICV16: return "AES CCM ICV16";
    case ADK_IPSEC_ENCR_AES_GCM_ICV8: return "AES GCM ICV8";
    case ADK_IPSEC_ENCR_AES_GCM_ICV12: return "AES GCM ICV12";
    case ADK_IPSEC_ENCR_AES_GCM_ICV16: return "AES GCM ICV16";
    case ADK_IPSEC_ENCR_NULL_GMAC: return "NULL GMAC";
    default: return "Invalid";
    }
}

static inline const char *authAlgoNames(adk_ipsec_auth_t auth)
{
    switch(auth)
    {
    case ADK_IPSEC_AUTH_NULL: return "NULL";
    case ADK_IPSEC_AUTH_HMAC_MD5: return "HMAC MD5";
    case ADK_IPSEC_AUTH_HMAC_SHA1: return "HMAC SHA1";
    case ADK_IPSEC_AUTH_AES_CMAC: return "AES CMAC";
    case ADK_IPSEC_AUTH_AES_XCBC_MAC: return "AES XCBC MAC";
    case ADK_IPSEC_AUTH_HMAC_SHA_256: return "HMAC SHA-256";
    case ADK_IPSEC_AUTH_HMAC_SHA_384: return "HMAC SHA-384";
    case ADK_IPSEC_AUTH_HMAC_SHA_512: return "HMAC SHA-512";
    case ADK_IPSEC_AUTH_AES_128_GMAC: return "AES GMAC-128";
    case ADK_IPSEC_AUTH_AES_192_GMAC: return "AES GMAC-192";
    case ADK_IPSEC_AUTH_AES_256_GMAC: return "AES GMAC-256";
    default: return "Invalid";
    }
}

#ifdef ADK_DEBUG
/*!
 * =============================================================
 * adk_ipsec_dump_sa
 * @param[in]  ipsecHandle  - IPsec module handle
 * @param[in]  direction    - IPsec SA direction
 *
 * @retval ADK_IPSEC_SUCCESS Successfully dumped all the information
 *
 * @brief
 *
 * Dumps all the configured information for all the SAs
 * in the given direction.
 *
 * =============================================================
 */
adk_status_t adk_ipsec_dump_sa(adk_module_handle_t ipsecHandle,
                               adk_ipsec_sa_direction_t direction)
{
    adk_status_t          retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    adk_ipsec_sa_t        *saConfig = NULL;

    uint16_t  saCount = 0, i = 0;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);
    if (direction == ADK_IPSEC_OUTBOUND)
    {
        ADK_TABLE_ENTRY_WRITE(" \t IPsec outbound SA Table \n");
        for (i = 0; i < ADK_IPSEC_NUM_OF_SA; i++)
        {
            saConfig = &ipsecModuleInfo_p->eSAConfig[i];
            /* key assumption, for any valid SA, SPI cannot
             * be 0!
             */
            if(saConfig->SASPPContext.spi == 0)
                continue;
            ADK_TABLE_ENTRY_WRITE(" \t SA # %d \n", saCount++);
            ADK_TABLE_ENTRY_WRITE(" \t Packets on this SA would undergo %s processing ",
                                  ipsecModuleInfo_p->dpTbl->eSATbl[i].saMode == ADK_IPSEC_SA_BOTH ? "Inline and Lookaside":"only Lookaside");
            ADK_TABLE_ENTRY_WRITE(" \t SA version = %d \n",
                                  saConfig->ipVersion);
            ADK_TABLE_ENTRY_WRITE(" \t Tunnel Information =  \n");
            if (ADK_IPSEC_IPV4 == saConfig->ipVersion)
            {
                ADK_TABLE_ENTRY_WRITE(" \t SIP = %x \n",
                                      adk_ntohl(saConfig->tunAddr.ipv4TunAddr.tunnelSrcIP));
                ADK_TABLE_ENTRY_WRITE(" \t DIP = %x \n",
                                      adk_ntohl(saConfig->tunAddr.ipv4TunAddr.tunnelDstIP));
            }
            else
            {
                ADK_TABLE_ENTRY_WRITE(" \t SIP = %x::%x::%x::%x \n",
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[0],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[1],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[2],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[3]);
                ADK_TABLE_ENTRY_WRITE(" \t DIP = %x::%x::%x::%x \n",
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[0],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[1],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[2],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[3]);
            }
            ADK_TABLE_ENTRY_WRITE(" \t softByteLifetime = %"PRIu64 "\n",
                                  saConfig->SASPPContext.softByteLifetime);
            ADK_TABLE_ENTRY_WRITE(" \t hardByteLifetime = %"PRIu64 "\n",
                                  saConfig->SASPPContext.hardByteLifetime);
            ADK_TABLE_ENTRY_WRITE(" \t Cipher algo = %s \n",
                                  cipherAlgoNames(saConfig->SASPPContext.encr));
            if (ADK_IPSEC_ENCR_NULL != saConfig->SASPPContext.encr)
            {
                ADK_TABLE_ENTRY_WRITE(" \t Cipher Key = %s \n",
                                      saConfig->SASPPContext.encrKey.buf);
            }
            ADK_TABLE_ENTRY_WRITE(" \t Auth algo = %s \n",
                                  authAlgoNames(saConfig->SASPPContext.auth));
            if (ADK_IPSEC_AUTH_NULL != saConfig->SASPPContext.auth)
            {
                ADK_TABLE_ENTRY_WRITE(" \t Auth Key = %s \n",
                                      saConfig->SASPPContext.authKey.buf);
            }
            ADK_TABLE_ENTRY_WRITE(" \t SPI = %x \n",
                                  saConfig->SASPPContext.spi);
            ADK_TABLE_ENTRY_WRITE(" \t Extended seq number enabled = %s \n",
                                  saConfig->SASPPContext.extSeqNum == true ? "Yes":"No");
            ADK_TABLE_ENTRY_WRITE(" \t Sequence number roll-over enabled = %s \n",
                                  saConfig->SASPPContext.sn.seqRollover == true ? "Yes":"No");
            ADK_TABLE_ENTRY_WRITE(" \t Lifetime enabled = %s \n",
                                  saConfig->SASPPContext.lifetimesEnabled == true ? "Yes":"No");
            ADK_TABLE_ENTRY_WRITE(" \t Antireplay window size = %d \n",
                                  saConfig->SASPPContext.antiReplayWindowSize);
            ADK_TABLE_ENTRY_WRITE(" \t ##################################### \n");
        }
    }
    else
    {
        ADK_TABLE_ENTRY_WRITE(" \t IPsec Inbound SA Table \n");
        for (i = 0; i < ADK_IPSEC_NUM_OF_SA; i++)
        {
            saConfig = &ipsecModuleInfo_p->iSAConfig[i];
            /* key assumption, for any valid SA, SPI cannot
             * be 0!
             */
            if(saConfig->SASPPContext.spi == 0)
                continue;
            ADK_TABLE_ENTRY_WRITE(" \t SA # %d \n", saCount++);
            ADK_TABLE_ENTRY_WRITE(" \t Packets on this SA would undergo %s processing ",
                                  ipsecModuleInfo_p->dpTbl->iSATbl[i].saMode == ADK_IPSEC_SA_BOTH ? "Inline and Lookaside":"only Lookaside");
            ADK_TABLE_ENTRY_WRITE(" \t SA version = %d \n",
                                  saConfig->ipVersion);
            ADK_TABLE_ENTRY_WRITE(" \t Tunnel Information =  \n");
            if (ADK_IPSEC_IPV4 == saConfig->ipVersion)
            {
                ADK_TABLE_ENTRY_WRITE(" \t SIP = %x \n",
                                      adk_ntohl(saConfig->tunAddr.ipv4TunAddr.tunnelSrcIP));
                ADK_TABLE_ENTRY_WRITE(" \t DIP = %x \n",
                                      adk_ntohl(saConfig->tunAddr.ipv4TunAddr.tunnelDstIP));
            }
            else
            {
                ADK_TABLE_ENTRY_WRITE(" \t SIP = %x::%x::%x::%x \n",
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[0],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[1],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[2],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelSrcIP[3]);
                ADK_TABLE_ENTRY_WRITE(" \t DIP = %x::%x::%x::%x \n",
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[0],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[1],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[2],
                                      saConfig->tunAddr.ipv6TunAddr.tunnelDstIP[3]);
            }
            ADK_TABLE_ENTRY_WRITE(" \t softByteLifetime = %"PRIu64 "\n",
                                  saConfig->SASPPContext.softByteLifetime);
            ADK_TABLE_ENTRY_WRITE(" \t hardByteLifetime = %"PRIu64 "\n",
                                  saConfig->SASPPContext.hardByteLifetime);
            ADK_TABLE_ENTRY_WRITE(" \t Cipher algo = %s \n",
                                  cipherAlgoNames(saConfig->SASPPContext.encr));
            if (ADK_IPSEC_ENCR_NULL != saConfig->SASPPContext.encr)
            {
                ADK_TABLE_ENTRY_WRITE(" \t Cipher Key = %s \n",
                                      saConfig->SASPPContext.encrKey.buf);
            }
            ADK_TABLE_ENTRY_WRITE(" \t Auth algo = %s \n",
                                  authAlgoNames(saConfig->SASPPContext.auth));
            if (ADK_IPSEC_AUTH_NULL != saConfig->SASPPContext.auth)
            {
                ADK_TABLE_ENTRY_WRITE(" \t Auth Key = %s \n",
                                      saConfig->SASPPContext.authKey.buf);
            }
            ADK_TABLE_ENTRY_WRITE(" \t SPI = %x \n",
                                  saConfig->SASPPContext.spi);
            ADK_TABLE_ENTRY_WRITE(" \t Extended seq number enabled = %s \n",
                                  saConfig->SASPPContext.extSeqNum == true ? "Yes":"No");
            ADK_TABLE_ENTRY_WRITE(" \t Sequence number roll-over enabled = %s \n",
                                  saConfig->SASPPContext.sn.seqRollover == true ? "Yes":"No");
            ADK_TABLE_ENTRY_WRITE(" \t Lifetime enabled = %s \n",
                                  saConfig->SASPPContext.lifetimesEnabled == true ? "Yes":"No");
            ADK_TABLE_ENTRY_WRITE(" \t Antireplay window size = %d \n",
                                  saConfig->SASPPContext.antiReplayWindowSize);
            ADK_TABLE_ENTRY_WRITE(" \t ##################################### \n");
        }
    }
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * =============================================================
 * adk_ipsec_dump_policies
 * @param[in]  ipsecHandle  - IPsec module handle.
 *
 * @retval ADK_IPSEC_SUCCESS Successfully dumped all the information
 *
 * @brief
 *
 * =============================================================
 */
adk_status_t adk_ipsec_dump_policies(adk_module_handle_t ipsecHandle)
{
    adk_status_t          retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    adk_ipsec_sec_pol_sel_t   polSel = { 0 };
    uint16_t  i;
    ipsec_sec_pol_info_t *ipsecSecPolLocal = NULL;
    adk_acl_rule_t cpRule;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);
    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);

    adk_acl_get_reset(ipsecModuleInfo_p->cpSpdAclList);
    for (i = 0; i < ipsecModuleInfo_p->dpTbl->numPolicies[ipsecModuleInfo_p->dpTbl->spdTblIdx]; i++)
    {
         retValue = adk_acl_next_rule_get(ipsecModuleInfo_p->cpSpdAclList, &cpRule);
         if (ADK_SUCCESS != retValue)
         {
             ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_POLICY_NOT_FOUND,
                                 "The policy is not found in CP ACL List\n");
             ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
             ADK_RETURN_ERROR_SET(ADK_E_IPSEC_POLICY_NOT_FOUND);
         }
         ipsecSecPolLocal = (ipsec_sec_pol_info_t*)(cpRule.user_data);
         adk_ipsec_get_spd_rule(&polSel, &ipsecSecPolLocal->dpdkAclRule);
         ADK_TABLE_ENTRY_WRITE(" \t IPsec Policy # %d \n", i);
         if (ADK_IPSEC_IPV4 == polSel.ipVersion)
         {
             ADK_TABLE_ENTRY_WRITE(" \t SIP begin = %x \n",
                                   polSel.ipAddr.ipv4Addr.srcAddr.addr.addr_range_t.addrBegin);
             ADK_TABLE_ENTRY_WRITE(" \t SIP end = %x \n",
                                   polSel.ipAddr.ipv4Addr.srcAddr.addr.addr_range_t.addrEnd);
             ADK_TABLE_ENTRY_WRITE(" \t DIP begin = %x \n",
                                   polSel.ipAddr.ipv4Addr.dstAddr.addr.addr_range_t.addrBegin);
             ADK_TABLE_ENTRY_WRITE(" \t DIP end = %x \n",
                                   polSel.ipAddr.ipv4Addr.dstAddr.addr.addr_range_t.addrEnd);
         }
         else
         {
             ADK_TABLE_ENTRY_WRITE(" \t SIP = %x::%x::%x::%x \n",
                                   polSel.ipAddr.ipv6Addr.srcAddr.addr.addr_range_t.addrBegin[0],
                                   polSel.ipAddr.ipv6Addr.srcAddr.addr.addr_range_t.addrBegin[1],
                                   polSel.ipAddr.ipv6Addr.srcAddr.addr.addr_range_t.addrBegin[2],
                                   polSel.ipAddr.ipv6Addr.srcAddr.addr.addr_range_t.addrBegin[3]);
             ADK_TABLE_ENTRY_WRITE(" \t DIP = %x::%x::%x::%x \n",
                                   polSel.ipAddr.ipv6Addr.dstAddr.addr.addr_range_t.addrBegin[0],
                                   polSel.ipAddr.ipv6Addr.dstAddr.addr.addr_range_t.addrBegin[1],
                                   polSel.ipAddr.ipv6Addr.dstAddr.addr.addr_range_t.addrBegin[2],
                                   polSel.ipAddr.ipv6Addr.dstAddr.addr.addr_range_t.addrBegin[3]);
         }
         ADK_TABLE_ENTRY_WRITE(" \t DSCP value = %d \n", polSel.dscpVal);
         ADK_TABLE_ENTRY_WRITE(" \t Protocol = %d \n", polSel.protocol);
         ADK_TABLE_ENTRY_WRITE(" \t Source port begin = %d \n", polSel.l4Info.portInfo.srcPort.portBegin);
         ADK_TABLE_ENTRY_WRITE(" \t Source port end = %d \n", polSel.l4Info.portInfo.srcPort.portEnd);
         ADK_TABLE_ENTRY_WRITE(" \t Destination port begin = %d \n", polSel.l4Info.portInfo.dstPort.portBegin);
         ADK_TABLE_ENTRY_WRITE(" \t Destination port end = %d \n", polSel.l4Info.portInfo.dstPort.portEnd);
         ADK_TABLE_ENTRY_WRITE(" \t ##################################### \n");
        }
        ADK_TABLE_ENTRY_WRITE(" \t ##################################### \n");

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}
#endif

/*!
 * =============================================================
 * adk_ipsec_shutdown
 * @param[in]  ipsecHandle  - IPsec module handle. Return at the
 * time of IPsec module's registration with the module manager.
 *
 * @retval ADK_IPSEC_SUCCESS Successfully shut down IPsec module.
 *
 * @brief
 *
 * This function de-initializeds the ADK IPsec module. The primary
 * responsibility of the function are
 *  - release module level resource. The resources that are
 *    allocated by module's objects won't be cleared.
 *  - Free up memory allocated for module data structure.
 *  - It is expected that before calling this API, application
 *    deletes all the SAs, policies, rulesets etc configuration
 *    related to IPsec.
 *
 * =============================================================
 */
adk_status_t
adk_ipsec_shutdown (adk_module_handle_t ipsecHandle)
{
    adk_status_t        retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t    sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p;

    /* Handle validations */
    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);

    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);


    /* Free the DP tables */

    retValue = adk_ipsec_dp_table_shutdown(ipsecHandle, sysHandle, ipsecModuleInfo_p);

    if (retValue != ADK_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR (ipsecHandle, retValue,
                             "Failed to shutdown DP tables\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
        return retValue;
    }

    /* Delete SPD IN/OUT CP ACL list */
    adk_ipsec_spd_cp_delete(ipsecHandle, ipsecModuleInfo_p);

    ADK_NVM_FREE(ipsecModuleInfo_p);

    /*Unregister ethernet parser function in PH to Decode exception params*/
    retValue = adk_ph_dev_rx_parse_unregister(ADK_IPSEC_MODULE_ID);
    if (retValue != ADK_SUCCESS)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle,ADK_E_INTERNAL_ERROR,
                "Failed to unregister the ph_dev_rx_parse function for IPsec Module\n");
    }

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!
 * =============================================================
 *
 * adk_ipsec_log_level_set
 *
 * @param[in]  ipsecHandle   IPsec Module Handle
 * @param[in]  log_level     Different log types
 *
 * @retval ADK_SUCCESS          Successful completion
 * @retval ADK_E_INVALID_HANDLE Invalid IPsec module handle parameter
 *
 * @brief  Function to set the log-level for IPsec module
 *
 * This function is used to set the IPsec module log-level for IPsec module. This
 * completely overwrites the existing log-level setting.
 *
 * =============================================================
 */
adk_status_t adk_ipsec_log_level_set(adk_module_handle_t ipsecHandle,
                                     adk_log_level_t log_level)
{
    adk_status_t                    ret = ADK_SUCCESS;
    adk_sys_handle_t sysHandle;
    ipsec_module_info_t *ipsecModuleInfo_p;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE("Invalid IPsec module handle %"ADKPRIh"\n",
                           ipsecHandle);
        ADK_IPSEC_LOG_FORCE("Failed to set IPsec log-level 0x%x\n",
                           log_level);
        return(ADK_E_INVALID_HANDLE);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);

    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, ret);

    /* Set log level */
    ipsecModuleInfo_p->logLevel = log_level;

    ADK_IPSEC_LOG_INFO(ipsecHandle,
                      "IPsec module %"ADKPRIh" log-level set to 0x%x\n",
                      ipsecHandle, ADK_IPSEC_LOG_LEVEL_GET(ipsecHandle));

    ADK_RELEASE_MODULE_LOCK(ipsecHandle, ret);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return(ret);
}

/*!
 * =============================================================
 * @fn adk_ipsec_get_sa_all
 *
 * @param[in]    ipsecHandle     - IPsec module handle.
 * @param[in]    saDirection     - security association direction (EGRESS or INGRESS)
 * @param[out]   *numSa          - Number of SAs configured in the given direction. 
 * @param[out]   **secAscList_p  - Pointer to list of all the SA informations.
 *
 * @retval ADK_IPSEC_SUCCESS                 - Successfully retrieved Security Associations
 * @retval ADK_E_IPSEC_INVALID_MODULE_HANDLE - i/p ipsec module handle
 *                                             is invalid.
 * @retval ADK_E_IPSEC_INVALID_INPUT_PARAM   - Invalid i/p parameter.
 *
 * @brief
 *
 * This API allows the application to retrieve all security associations configured into ADK
 * IPsec module for the given direction. secAscList_p would return list of all the SA informations
 * laid out in the array.
 * It is expected that the calling application will free secAscList_p pointer after its use.
 * 
 * =============================================================
 */
adk_status_t
adk_ipsec_get_sa_all (adk_module_handle_t        ipsecHandle,
                      adk_ipsec_sa_direction_t   saDirection,
                      uint32_t                   *numSa_p,
                      adk_ipsec_sa_t             **secAscList_p)
{
    adk_status_t          retValue = ADK_IPSEC_SUCCESS;
    adk_sys_handle_t      sysHandle;
    ipsec_module_info_t   *ipsecModuleInfo_p = NULL;
    adk_ipsec_sa_t        *localSecAscList_p = NULL;
    uint32_t saIdx;

    if (!ADK_MODULE_HANDLE_VALIDATE(ipsecHandle))
    {
        ADK_IPSEC_LOG_FORCE ("Invalid IPSec module handle %"ADKPRIh"\n",
                              ipsecHandle);
        return ADK_ERROR_SEV_LEVEL_SET(ADK_E_INVALID_PARAMETER);
    }

    sysHandle = ADK_MODULE_SYS_HANDLE_GET(ipsecHandle);
    ADK_IPSEC_LOG_FUNC_ENTRY(ipsecHandle);
    ADK_ACQUIRE_MODULE_LOCK(ipsecHandle, retValue);

    /* Get the ipsec module info data strcuture pointer */
    ipsecModuleInfo_p = ADK_MODULE_INFO_PTR_GET(ipsecHandle);
    mVerifyArguments(ipsecModuleInfo_p,ipsecHandle, retValue);
    if (!numSa_p)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                "IPSEC: invalid address passed for numSa_p pointer\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }
    if (!secAscList_p)
    {
        ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_INVALID_INPUT_PARAM,
                "IPSEC: invalid address passed for secAscList_p pointer\n");
        ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
        ADK_RETURN_ERROR_SET(ADK_E_IPSEC_INVALID_INPUT_PARAM);
    }

    if (saDirection == ADK_IPSEC_OUTBOUND)
    {
        localSecAscList_p =
            (adk_ipsec_sa_t *)ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
                    ADK_IPSEC_SA_ID_OBJ, ipsecModuleInfo_p->numEgSAEntries, sizeof(adk_ipsec_sa_t));
        if(NULL == localSecAscList_p)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle,  ADK_E_IPSEC_ERROR,
                    "IPSEC: Failed to allocate memory for Security Association list pointer\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_ERROR);
        }
        for(saIdx = 0; saIdx < ipsecModuleInfo_p->numEgSAEntries; saIdx++)
        {
            ADK_MEMCPY(&localSecAscList_p[saIdx],
                    &ipsecModuleInfo_p->eSAConfig[saIdx],
                    sizeof(adk_ipsec_sa_t));
        }
        *numSa_p = ipsecModuleInfo_p->numEgSAEntries;
    }
    else
    {
        localSecAscList_p =
            (adk_ipsec_sa_t *)ADK_NVM_CALLOC(ADK_MODULE_MODULEID_GET(ipsecHandle),
                    ADK_IPSEC_SA_ID_OBJ, ipsecModuleInfo_p->numIngSAEntries, sizeof(adk_ipsec_sa_t));
        if(NULL == localSecAscList_p)
        {
            ADK_IPSEC_LOG_ERROR(ipsecHandle, ADK_E_IPSEC_ERROR,
                    "IPSEC: Failed to allocate memory for Security Association list pointer\n");
            ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);
            ADK_RETURN_ERROR_SET(ADK_E_IPSEC_ERROR);
        }
        for(saIdx = 0; saIdx < ipsecModuleInfo_p->numIngSAEntries; saIdx++)
        {
            ADK_MEMCPY(&localSecAscList_p[saIdx],
                    &ipsecModuleInfo_p->iSAConfig[saIdx],
                    sizeof(adk_ipsec_sa_t));
        }
        *numSa_p = ipsecModuleInfo_p->numIngSAEntries;
    }
    *secAscList_p = localSecAscList_p;
    ADK_RELEASE_MODULE_LOCK(ipsecHandle, retValue);

_lock_release_failed:
_lock_acquire_failed:
    ADK_IPSEC_LOG_FUNC_EXIT(ipsecHandle);
    return retValue;
}

/*!     end addtogroup _ipsecFunctions
 *      @}
 *  end addtogroup _ipsec
 * @}
 */
